/////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TAttribute.h
 \b Created : Fri Oct  1 13:35:21 2004 \n
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>\n
 \brief Interface file for the class URANIE::DataServer::TAttribute

 \par Current version
 $Date$
 $Revision$
 
 
 */
#ifndef TAttribute_H
#define TAttribute_H
#include "TString.h"
#include <string>

//ROOT
#include "TEventList.h"
#include "TList.h"
#include "TROOT.h"
#include "TSystem.h"

#include "DataServer.h"

//URANIE
//#include "UExceptions.h"

//STL
#include <vector>
#include <map>
#include <list>
#include <iostream>
using std::vector;
using std::string;
using std::map;
using std::cout;
using std::endl;
using std::list;

//URANIE

//#include "UExceptions.h"

#define DLE_MAX 1.42387e+64

/*!
 @class URANIE::DataServer::TAttribute

 @brief Description of a variable (field).

 Avalaible informations are:
 \li name : the name of the attribute
 \li title : the title of the attribute on graphs. It can be be different than the name.
 \li _sunity : unit on graphs
 \li bounds : Min and Max Bounds if necessary
 \li default : the default value of this variable (can be used in optimization module)
 \li step : the step of variable (can be used in optimization module)
 \li _snote : additional note on variable
 - ordering
 - taxonomy
 - normalization

 For creation of variables to use in all incertainties and sensitivities studies.

 To define a pointer \em px on a non-bounded variable \f$ x\f$.
 \code
 TAttribute *px = new TAttribute("x");
 \endcode

 To define a pointer \em x on a bounded attribute \f$ x28\f$ comprised between \f$ \left[-1., 1.\right]\f$  with title \f$ \Delta P_{e}^{F_{iso}}\f$ on graphs.
 \code
 TAttribute *x = new TAttribute("x28", -1.0, 1.0);
 x->SetTitle("#Delta P_{e}^{F_{iso}}");
 \endcode
 To draw this attribute in a \ref TDataServer, apply the draw method
 \code
 tds->draw("x28");
 \endcode
 */

namespace URANIE
{
namespace DataServer
{

typedef struct STRUCT
{
    //! the enum of input file typ
    /*!
      \deprecated The \b kNewLine enum is deprecated from v2.1.10 to v2.3.0 
    */
    enum EFileType
    {
        kUnknown,
        kKey,
        kFlag,
        kNewKey,
        kNewLine,
        kNewRow,
        kNewColumn,	
        kXMLAttribute,
        kXMLField,
        kFMU,
	kNewTDS
    };
    TString _sFile; ///< File related to a key
    TString _sKey;  ///< Name of the key
    TString _sFormat; ///< Format to perform substitutions
    EFileType _sFileType; ///< Type of substitution
    TEventList * _sEventList; ///< EventList related to a specific couple file/key
} TAttributeFileKey;

class TAttribute: public TNamed
{
public:
    enum EOrigin
    {
        kInternal, kIterator, kConstant, kAttribute, kDeleted
    };
    enum EAttribute
    {
        kInput, kOutput
    };
    enum EType
    {
        kDefault, kReal, kVector, kString, kCategory, kInconnu
    };
    // Attributes
protected:
    //---------------------------------------------
    /** @name String informations of an attribute
     **/
    //@{
    TString _sunity;        ///< Unity.
    TString _snote;         ///< Note.
    Bool_t _blog;         ///< Log printing.
    Int_t _nshare; ///< The number of time this attribute is shared in TDataServer.
    TAttribute *_attSon;      ///<  Son attribute 
    Int_t _nlevel;                     ///< level to its "father"
    Bool_t _haveSon; ///< true if the attribute have a son
    //@}

    //---------------------------------------------
    /** @name Informations about inputs of an attribute
     **/
    //@{
    TString _sFormatSubstitute;   ///< Format to substitute the value.

    list<TEventList *> _nfields; ///< List of all TEventLists created for the attribute.

    list<TAttributeFileKey *> _KeyList; ///< List of Keys Attributes structures.
    //@}

    //---------------------------------------------
    /** @name Numerical information of an attribute
     **/
    //@{
    Double_t upperBound;   ///< Upper bound
    Bool_t _bHaveUpperBound;  ///< If have an upper bound

    Double_t lowerBound;   ///< Lower bound
    Bool_t _bHaveLowerBound;  ///< If have a lower bound

    Double_t _defaultValue;       ///< Default value
    Bool_t _bHaveDefaultValue;    ///< If have a default value

    vector<double> _defaultVector;       ///< Default value for vector
    Bool_t _bHaveDefaultVector;    ///< If have a default value for vector

    string _defaultString;       ///< Default value for string
    Bool_t _bHaveDefaultString;    ///< If have a default value for string
    
    Double_t _stepValue;          ///< Step value when using in Optimization
    Bool_t _bHaveStepValue;       ///< If have a step value
    //@}

    //---------------------------------------------
    /** @name Statistical information of an attribute
     **/
    //@{
    vector<Double_t> _vminimum;         ///< All minimun calculated
    vector<Double_t> _vmaximum;         ///< All maximun calculated
    vector<Double_t> _vmean;            ///< All mean calculated
    vector<Double_t> _vstd;             ///< All std calculated

    vector<Double_t> *_vbuffer;        ///> buffer pointer to check addresses
    vector<map<double,double>* > _vquantile;         ///< All Computed quantiles
  
    //@}
    EOrigin _norigin; ///< The origin of the attribute (Internal of uranie, attribute or Input, Output...)
    EAttribute _nAttribute; ///< The nature of attributes : kInput (default) or kOutput
    EType _nType; ///< The type of attribute: real (double), vector (of double), string...

    /** @name Allows to generate remote cluster scripts
     **/
    TString _slurmValue;

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     
        For creation of variables to use in all incertainties and sensitivities studies.
     
        To define a pointer \em px on a non-bounded variable \f$x\f$.
        \code
        TAttribute *px = new TAttribute("x");
        \endcode
     
        To define a pointer \em x on a bounded attribute \f$ x28\f$ comprised between \f$ \left[-1., 1.\right]\f$  with title \f$ \Delta P_{e}^{F_{iso}}\f$ on graphs.
        \code
        TAttribute *x = new TAttribute("x28", -1.0, 1.0);
        x->SetTitle("#Delta P_{e}^{F_{iso}}");
        \endcode
     
    **/
    //@{
    //! Default constructor
    TAttribute();
    TAttribute(const char* name, const char* title);

    //! Constructor by name only
    /*! 
     
      \param name Name of the attribute
    */
    TAttribute(TString name);
    //! Constructor by name and type only
    /*! 
     
      \param name Name of the attribute    
      \param type Type of the attribute
    */
    TAttribute(TString name, EType type);
    //! Constructor by name and range
    /*! 
     
      \param name Name of the attribute
      \param value Value of the attribute
      \param type Type of the attribute
    */
    TAttribute(TString name,TString value, EType type);
    //! Constructor by name and range
    /*!

      \param name (TString) Name of the attribut.
      \param value (TString) slurm field value.
    */
    TAttribute(TString name, Double_t lower, Double_t upper);
    //! Copy constructor
    TAttribute(const TAttribute &attr);
    //! Clone an attribute with a new name
    /*!
      Make a clone of an attribute using the Streamer facility with a new name.
      All the information of the origin attribute are stored in the new attribute (title, unity, bounds, ...).
      \param newname (const char *) The new name of the attribute
    */
    TAttribute *Clone(const char *newname) const;
    //! Default destructor
    virtual ~TAttribute();
    //@}


    //---------------------------------------------
    /** @name Setting and Getting Statistical and Quantiles
        The statistical informations (minimum, maximum, mean and std)  are computed from the data base inside the TDataServer.
    **/
    //@{
    //!Internal method used to check the validity of all setters when dealing with vectors
    /*!
      \param type: which statistical object considered amongst min, max, mean and std
      \param iel: number of the element in the vector
      \param val: value to be stored
      \exception if iel is negative, if iel greater than the considered vector size (works with push_back)
    */
    void checkSizeBeforePush(const char* type, int iel, Double_t val);
  
    //!Internal method used to check the validity of all getters when dealing with vectors and return the request
    /*!
      \param type: which statistical object considered amongst min, max, mean and std
      \param iel: number of the element in the vector
      \exception if iel is negative, if iel greater than the considered vector size (works with push_back)
    */
    Double_t checkSizeBeforeGet(const char* type, int iel=0);
  
    //! Set the quantile value
    /*!
      Store the quantile value computed for prob (used as key), for element iel (in case of a vector
      The method called for the first time for element iel creeate a new map and add the couple (key,value) as (proba,quantile).
      If this is not the first call for this element, only the second part is done.
      \exception if iel is negative, if iel greater than the considered vector size (works with push_back)
    */
    void setQuantile(double prob, double quantile, int iel=0);
    //! Returns the number of quantile computed and stored for element iel
    void getQuantilesSize(int &nb, int iel=0);
    //! Returns all the probabilities and corresponding quantiles computed for element iel
    /*!
      \warning the size of the prob and quantile array has to be carefully set. This method should be called like
      \code
      int nbQ=0, iel=0;
      tds->getAttribute("x")->getQuantilesSize(nbQ,iel);
      double prob[nbQ], quan[nbQ];
      tds->getAttribute("x")->getQuantiles(prob,quan,iel);
      \endcode    
      \warning Quantile map is cleared (as statistical vectors) as soon as the selection is changed in TDataServer.
      The value stored in this map is the latest one computed. No history is kept to know the method and option used
    */
    void getQuantiles(double* prob, double* quantile, int iel=0);
    //! Returns the value of the quantile for element iel, given the probability
    /*!
      \warning Quantile map is cleared (as statistical vectors) as soon as the selection is changed in TDataServer.
      The value stored in this map is the latest one computed. No history is kept to know the method and option used
    */
    void getQuantile(double &prob, double &quantile, int iel=0);
    //! Returns the vector of quantile given the probability
    /*!
      \warning Quantile map is cleared (as statistical vectors) as soon as the selection is changed in TDataServer.
      The value stored in this map is the latest one computed. No history is kept to know the method and option used
    */
    void getQuantileVector(double &prob, vector<double> &quantile);
     
    // Calculates with a sampling
    Double_t getMinimum(int iel=0){
        return checkSizeBeforeGet("Minimum", iel);
    }
    //! Returns the size of the vector of minimum
    Double_t getMinimumSize(){

        return _vminimum.size();
    }
    //! Details whether this attribute has  a default value
    Bool_t hasDefault() {

        switch(_nType)
        {
        case kVector :
            return hasDefaultVector();
        case kString :
            return hasDefaultString();
        default :
            return hasDefaultValue();
        }
    }
    //! Set the minimum value oef element iel
    /*!
      The vector of minimum is of size N+1 if the vector considered is of size N. The last element is the minimum of the global vector
    */
    void setMinimum(Double_t val,int iel=0){
        checkSizeBeforePush("Minimum", iel, val);
    }

    Double_t getMaximum(int iel=0){
        return checkSizeBeforeGet("Maximum", iel);
    }
    //! Returns the size of the vector of maximum
    Double_t getMaximumSize(){
        return _vmaximum.size();
    }
    //! Set the maximum value oef element iel
    /*!
      The vector of maximum is of size N+1 if the vector considered is of size N. The last element is the maximum of the global vector
    */
    void setMaximum(Double_t val,int iel=0){
        checkSizeBeforePush("Maximum", iel, val);
    }

    //! Returns the vector of considered statistical value
    /*!
      \param type: "min", "max", "mean" or "std"
      \param vec: vector in which results will be stored
    */
    void getStatisticalVector(const char *type, vector<double> &vec);
    
    Double_t getMean(int iel=0)
        {
            return checkSizeBeforeGet("Mean", iel);
        }
    //! Returns the size of the vector of mean
    Double_t getMeanSize()
        {
            return _vmean.size();
        }
    //! Set the mean value oef element iel
    /*!
      The vector of mean is of size N+1 if the vector considered is of size N. The last element is the mean of the global vector
    */
    void setMean(Double_t val,int iel=0);

    Double_t getStd(int iel=0)
        {
            return checkSizeBeforeGet("Std", iel);
        }
    //! Returns the size of the vector of std
    Double_t getStdSize()
        {
            return _vstd.size();
        }
    //! Set the std value oef element iel
    /*!
      The vector of std is of size N+1 if the vector considered is of size N. The last element is the std of the global vector
    */
    void setStd(Double_t val,int iel=0);
    void setSlurmOption(TString value)
        {
            _slurmValue=value;
        }

    //@}

    
    //---------------------------------------------
    /** @name Setting and Getting attributes general information
     **/
    //@{

    //! Returns the datatype of the considered attribute
    EType getDataType() const
        {
            return _nType;
        }
    //! Returns the datatype as TString
    /*!
      Returns
      "V" <=> vector
      "S" <=> string
      "D" <=> double
    */
    TString getDataTypeStr();

    //! Defines the title of the attribute
    /*!
      We must define this method to uniform all the method name
    */
    void setTitle(const char* title)
        {
            SetTitle(title);
        }
    void setTitle(TString str)
        {
            SetTitle(str.Data());
        }

    TString getTitle()
        {
            return TString(GetTitle());
        }

    //! Change the type of data
    void setDataType(EType thetype)
        {
            _nType = thetype;
        }
    //! Change the type of data
    /*!
      "real" or "double" or "d" <=> kReal
      "vector" or "v" <=> kVector
      "string" or "s" <=> kString
    */
    void setDataType(TString name);

    //! Returns the legend of the attribut
    /*!
      The legend is the Title plus, if exists, "[" unity "]"
    */
    TString getLegend();


    TString getUnit()
        {
            return _sunity;
        }
    void setUnit(TString str)
        {
            _sunity = str;
        }

    TString getUnity()
        {
            return _sunity;
        }
    void setUnity(TString str)
        {
            _sunity = str;
        }

    TString getNote()
        {
            return _snote;
        }
    void setNote(TString str)
        {
            _snote = str;
        }

    Bool_t setDefault(TString value);
    TString getDefault(TString format="%e");
    
    Bool_t setDefaultValue(Double_t val);
    Bool_t getDefaultValue(Double_t &val);
    Bool_t hasDefaultValue();
    
    Bool_t setDefaultVector(vector<double> &vec);
    Bool_t getDefaultVector(vector<double> *val);
    Bool_t hasDefaultVector();
    
    Bool_t setDefaultString(TString str);
    Bool_t getDefaultString(string &str);
    Bool_t hasDefaultString();
    
    void setStepValue(Double_t val);

    Bool_t getStepValue(Double_t &val);
    TString getSlurmOption()
        {
            return _slurmValue;
        }
    EOrigin getOrigin()
        {
            return _norigin;
        }
    void setOrigin(EOrigin ind)
        {
            _norigin = ind;
        }

    Int_t getShare()
        {
            return _nshare;
        }

    //! Sets the attribute as an output attribute
    void setOutput()
        {
            _nAttribute = kOutput;
        }
    //! Sets the attribute as Input
    void setInput()
        {
            _nAttribute = kInput;
        }

    //! Returns the input/Output information of the attribute
    EAttribute getInputOutput()
        {
            return _nAttribute;
        }


    //@}

    //---------------------------------------------
    /** @name Treament of inputs of an attribute
     
     **/
    //@{

    //!  Sets the file name for a key 
    /*!
      New implementation for old setKey method.
    */
    void setFileNameOfKey(TString str);

    //!  Gets KeyList for the attribute
    list<URANIE::DataServer::TAttributeFileKey *> getKeyList()
        {
            return _KeyList;
        }
    ;

    //!  Sets index ind in the unique field related to a unique key 
    /*!
      \deprecated Multiple fields can be defined for a same attribute. Use the method setFieldOfKey(Int_t sIndex, Int_t ind).
    */
    //void 	setFieldOfKey(Int_t ind) { _nfield = ind;}
    void setFieldOfKey(Int_t ind);
    void setFieldOfKey(Int_t sIndex, Int_t ind);

    //! Defines the input files and eventually others informations like key, format to substitute
    /*!
      In this method, you can associate the input file, the key and the format to store the value for all patterns in the design of experiments.
      The input must be existed before (\b kKey and \b kFlag), or does not exist before (\b kNewRow, \b kNewColumn, \b kNewKey).
      You can defines also the type of input file:
      \li TAttributeFileKey::kKey : the input file is key format (i.e. "key = value")
      \code
      Rw = 0.0500 ;
      R = 33366.67 ;
      Tu = 63070.0 ;
      \endcode
      \li TAttributeFileKey::kFlag  : the input file is "flag format (i.e. "information value1 value2 \@value3\@")
      \code
      kW                  @Kw@
      information_Tu      Champ_Uniforme 1        @Tu@
      information_Tl      Champ_Uniforme 1        @Tl@
      information_L {
      precision  @L@
      }
      convergence {
      criterion relative_max_du_dt
      precision  1.e-6
      }
      \endcode
      \li TAttributeFileKey::kNewRow : the input file is creted on the fly with values on \em line.
      \code
      1.123982e-01 1.507645e+04 1.101851e+05 1.000706e+02 1.086654e+03 7.556420e+02 1.600970e+03 1.194875e+04
      \endcode
      \li TAttributeFileKey::kNewColumn : the input file is creted on the fly with values on \em column.
      \code
      1.123982e-01
      1.507645e+04
      1.101851e+05
      ...
      \endcode
      \li TAttributeFileKey::kNewKey : the input file is creted on the fly with Key and values on \em line.
      \code
      Rw = 1.123982e-01
      Tu = 1.507645e+04
      Tl = 1.101851e+05
      ...
      \endcode

      \param sfile (TString) the input file where to change the values of the attribute
      \param skey (TString) ("") the key in the input file when it was different than the name of the attribute.
      The default value (empty string) means that is the name of the attribute.
      \param  sformatToSubstitute (TString)["%e"] the format to substitue. The default value ["%e"] means that it is with the scientifist format \em x.yzE-uv.
      \param FileType (TAttributeFileKey::EFileType) [TAttributeFileKey::kKey] the type of input files. The default value is \b kKey
    */
    void setFileKey(TString sfile, TString skey = "",
                    TString sformatToSubstitute = "%e",
                    TAttributeFileKey::EFileType FileType = TAttributeFileKey::kKey);

    //! Defines the input files as in the \em "flag" format and eventually others informations like key, format to substitute
    /*!
      This method is an alias to the \ref setFileKey method with TAttributeFileKey::kFlag as the fourth argument.

      \see setFileKey

      \param sfile (TString) the input file where to change the values of the attribute
      \param skey (TString) ("") the key in the input file when it was different than the name of the attribute.
      The default value (empty string) means that is the name of the attribute.
      \param  sformatToSubstitute (TString)["%e"] the format to substitue. The default value ["%e"] means that it is with the scientifist format \em x.yzE-uv.
    */
    void setFileFlag(TString sfile, TString skey = "", TString sformatToSubstitute = "%e")
        {
            setFileKey(sfile, skey, sformatToSubstitute, TAttributeFileKey::kFlag);
        }

    //! Defines the input files as an \em "XML" input file with XPath \b attribute format and eventually the format to substitute
    /*!
      This method is an alias to the \ref setFileKey method with TAttributeFileKey::kXMLAttribute as the fourth argument.

      \see setFileKey

      \param sfile (TString) the XML input filename where to change the values of the attribute
      \param sXPathAttribute (TString) the XPath for an attribute in the XML input file.
      \param  sformatToSubstitute (TString)["%e"] the format to substitue. The default value ["%e"] means that it is with the scientifist format \em x.yzE-uv.
    */
    void setFileXMLAttribute(TString sfile, TString sXPathAttribute, TString sformatToSubstitute = "%e")
        {
            setFileKey(sfile, sXPathAttribute, sformatToSubstitute, TAttributeFileKey::kXMLAttribute);
        }

    //! Defines the input files as an \em "XML" input file with XPath \b field and eventually the format to substitute
    /*!
      This method is an alias to the \ref setFileKey method with TAttributeFileKey::kXMLField as the fourth argument.

      \see setFileKey

      \param sfile (TString) the XML input filename where to change the values of the attribute
      \param sXPathField (TString) the XPath for a field in the XML input file.
      \param  sformatToSubstitute (TString)["%e"] the format to substitue. The default value ["%e"] means that it is with the scientifist format \em x.yzE-uv.
    */
    void setFileXMLField(TString sfile, TString sXPathField, TString sformatToSubstitute = "%e")
        {
            setFileKey(sfile, sXPathField, sformatToSubstitute, TAttributeFileKey::kXMLField);
        }

    void setFileFMU(TString sfile, TString sXPathField, TString sformatToSubstitute = "%e")
    {
        setFileKey(sfile, sXPathField, sformatToSubstitute, TAttributeFileKey::kFMU);
    }

    //@}

    //---------------------------------------------
    /** @name Substitution of value
     
     **/
    //@{
    //! Returns the format of substitution
    const char* getFormatToSubstitute()
        {
            return _sFormatSubstitute.Data();
        }

    //! Sets the format of substitution
    void setFormatToSubstitute(TString str)
        {
            _sFormatSubstitute = str;
        }

    //@}

    //---------------------------------------------
    /** @name Treatment of the properties of an attribute
     **/
    //@{
    //! Tests if the attribute is an input attribute
    Bool_t isInput()
        {
            return (((_norigin == kAttribute) || (_norigin == kConstant)) && (_nAttribute == kInput));
        }

    //! Tests if the attribute is an output attribute
    Bool_t isOutput()
        {
            return ((_norigin == kAttribute)  && (_nAttribute == kOutput));
        }

    void addShare()
        {
            _nshare++;
        }
    void delShare()
        {
            _nshare--;
        }
    void initShare()
        {
            _nshare = 1;
        }

    void clearVectors();
    //@}

    //---------------------------------------------
    /** @name Treatment the boundaries of an attribute
     **/
    //@{
    //! Sets the lower bound value of the attribute
    /*
      \param val (Double_t) The lower bound value of the attribute
    */
    virtual void setLowerBound(Double_t val, bool internalcall=false);
    //! Gets the lower bound value of the attribute
    /*s
      \return The lower bound value of the attribute
    */
    Double_t getLowerBound()
        {
            return lowerBound;
        }
    //! Returns a boolean if the attribute have a lower bound
    /*
      \return If the attribute have a lower bound value return kTRUE else return kFALSE
    */
    Bool_t isLowerBounded()
        {
            return _bHaveLowerBound;
        }

    //! Sets the upper bound value of the attribute
    /*
      \param val (Double_t) The upper bound value of the attribute
    */
    virtual void setUpperBound(Double_t val, bool internalcall=false);
    //! Gets the upper bound value of the attribute
    /*
      \return The upper bound value of the attribute
    */
    Double_t getUpperBound()
        {
            return upperBound;
        }
    //! Returns a boolean if the attribute have an upper bound
    /*
      \return If the attribute have an upper bound value return kTRUE else return kFALSE
    */
    Bool_t isUpperBounded()
        {
            return _bHaveUpperBound;
        }

    //! Sets the two bounds values (lower and  upper) of the attribute
    /*
      \exception if the lower bound is superior of the upper boun
      \exception if the upper bound is inferior of the lower boun
     
      \param lower (Double_t) The lower bound value of the attribute
      \param upper (Double_t) The upper bound value of the attribute
    */
    virtual void setBounds(Double_t lower, Double_t upper);
    //! Returns a boolean if the attribute have a lower AND an upper bounds
    /*
      \return If the attribute have a lower AND an upper bounds return kTRUE else return kFALSE
    */
    Bool_t isBounded()
        {
            return _bHaveLowerBound && _bHaveUpperBound;
        }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
        {
            _blog = kTRUE;
        }
    void unsetLog()
        {
            _blog = kFALSE;
        }
    void changeLog()
        {
            _blog = _blog ? kFALSE : kTRUE;
        }
    Bool_t getLog()
        {
            return _blog;
        }
    virtual void printLog();
    //@}

public:

    //---------------------------------------------
    /** @name Internal methode used to estimate the quantile by importance sampling
     **/
    //@{
    TAttribute *getSonAttribute();
    void setSonAttribute(TAttribute * tatt);
    void removeSonAttribute();
    Int_t getLevel();
    void setLevel(Int_t nlevel);
    Bool_t haveSon();
    //@}
    ClassDef(URANIE::DataServer::TAttribute, ID_DATASERVER)
    //Definition of an variable

};

} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
