//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TBetaDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-11-20 10:13:36

 \brief Implementation of the class URANIE::DataServer::TBetaDistribution
 \par Modifications
 \li Cedric Forgues (C.S.) 2010/11/17
 Tolerance in getInvCDF was lowered to get a precision on values more accurate according to values given by R.
 \li Cedric Forgues (C.S.) 2010/06/11
 Added Exceptions, computeMean added in constructor with an attribute.
 */
#include "TMath.h"
#include "TBetaDistribution.h"
#include <iostream>
using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TBetaDistribution)
namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TBetaDistribution::TBetaDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kBeta), _dparameterAlpha(
                0.5), _dparameterBeta(0.5)
{
    computeTheoreticalInformation();
    //  cout << "build TBetaDistribution" << endl; 
}

TBetaDistribution::TBetaDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kBeta), _dparameterAlpha(
                0.5), _dparameterBeta(0.5)
{
    //  cout << "Constructeur URANIE::DataServer::TBetaDistribution[" << this << "]str[" << str<<"]"<<endl;
    computeTheoreticalInformation();
}

TBetaDistribution::TBetaDistribution(TString str, Double_t alpha, Double_t beta,
        Double_t min, Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kBeta, min, max)
{
    setParameters(alpha, beta, min, max);
    //  cout << "Constructeur URANIE::DataServer::TBetaDistribution[" << this << "]str[" << str<<"] Min[" << min <<"] Max[" << max << "]" << endl;
}

TBetaDistribution::TBetaDistribution(URANIE::DataServer::TAttribute *att,
        Double_t alpha, Double_t beta) :
        TStochasticAttribute(att, TStochasticAttribute::kBeta)
{
    setParameterAlpha(alpha, kFALSE);
    setParameterBeta(beta, kFALSE);
    computeTheoreticalInformation();
    //  cout << "Constructeur URANIE::DataServer::TBetaDistribution att[" << this << "]att->getName()[" << att->getName()<<"]" << endl;
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------       

TBetaDistribution::~TBetaDistribution()
{
    //  cout << "detruit TBetaDistribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------
void TBetaDistribution::setParameters(Double_t alpha, Double_t beta,
        Double_t lower, Double_t upper, Bool_t recompute)
{
    TAttribute::setLowerBound(lower);
    TAttribute::setUpperBound(upper);
    setParameterAlpha(alpha, kFALSE);
    setParameterBeta(beta, kFALSE);
    if (recompute)
        computeTheoreticalInformation();
}

void TBetaDistribution::setParameterAlpha(Double_t alpha, Bool_t recompute)
{
    try
    {
        if (alpha < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TBetaDistribution::setParameterAlpha: parameter alpha [%f] is too close to zero: lower limit is 1e-4",
                            alpha));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterAlpha = alpha;

    if (recompute)
        computeTheoreticalInformation();
}

void TBetaDistribution::setParameterBeta(Double_t beta, Bool_t recompute)
{
    try
    {
        if (beta < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TBetaDistribution::setParameterBeta: parameter beta [%f] is too close to zero: lower limit is 1e-4",beta));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
    }

    _dparameterBeta = beta;

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TBetaDistribution::computeTheoreticalInformation()
{
    Double_t scalingCoef = (upperBound - lowerBound);

    // Mean
    _dtheoreticalMean = lowerBound
            + scalingCoef * _dparameterAlpha
                    / (_dparameterAlpha + _dparameterBeta);

    // StdDev
    _dtheoreticalStdDev = scalingCoef
            * TMath::Sqrt(
                    (_dparameterAlpha * _dparameterBeta)
                            / (_dparameterAlpha + _dparameterBeta + 1.0))
            / (_dparameterAlpha + _dparameterBeta);

    // Mode
    Double_t tmpMode;
    if (_dparameterAlpha > 1.0 && _dparameterBeta > 1.0)
        tmpMode = lowerBound
                + scalingCoef * (_dparameterAlpha - 1)
                        / (_dparameterAlpha + _dparameterBeta - 2);

    else
        tmpMode = THEO_DEF;

    // else if (_dparameterAlpha < 1.0 && _dparameterBeta < 1.0)
    // {
    //     if (_dparameterAlpha < _dparameterBeta)
    //         tmpMode = lowerBound;

    //     else if (_dparameterAlpha > _dparameterBeta)
    //         tmpMode = upperBound;

    //     else
    //         tmpMode = (upperBound - lowerBound) / 2.0;
    // }

    // else if ((_dparameterAlpha < 1.0 && _dparameterBeta >= 1.0)
    //         || (_dparameterAlpha == 1.0 && _dparameterBeta > 1.0))
    //     tmpMode = lowerBound;

    // else if ((_dparameterAlpha >= 1.0 && _dparameterBeta < 1.0)
    //         || (_dparameterAlpha > 1.0 && _dparameterBeta == 1.0))
    //     tmpMode = upperBound;

    // else
    //     // At this stage, _dparameterAlpha == _dparameterBeta == 1.0. 
    //     // In this case, the distribution is uniform, thus any value 
    //     // inside the range is a valid mode. 
    //     // Like for the TUniformDistribution, I'm taking the middle of 
    //     // the range.
    //     tmpMode = (upperBound - lowerBound) / 2.0;

    _dtheoreticalMode = tmpMode;
}


// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TBetaDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
        pdf = TMath::BetaDist((x - lowerBound) / (upperBound - lowerBound),
                _dparameterAlpha, _dparameterBeta) / (upperBound - lowerBound);

    return pdf;
}

Double_t TBetaDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
        cdf = TMath::BetaDistI((x - lowerBound) / (upperBound - lowerBound),
                _dparameterAlpha, _dparameterBeta);
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TBetaDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TBetaDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
                << p << "]" << endl;

    Double_t x0 = lowerBound, x1 = upperBound, px;
    Double_t xcrt = (upperBound + lowerBound) / 2.0;
    //Double_t xtol = x0 * 0.000000000000001, ptol = 0.000000000000001;
    Double_t xtol = ((x0!=0 && x0!= -DLE_MAX)?TMath::Abs(x0):0.1)*1e-14, ptol = 1e-15;
    Int_t ncpt = 0;

    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                 << "], current solution[" << xcrt
                 << "], corresponding probability[" << px
                 << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]"
                 << "]  diff versus tol ["<<setprecision(20)<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]";
        if (TMath::Abs(px - p) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                        << "], corresponding probability[" << px
                        << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < p)
            x0 = xcrt;
        else if (px > p)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                    << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                    << endl;
        ncpt++;

        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
    }

    if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TGamma");
        _localmess <<  " TGammaDistribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }

    return xcrt;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TBetaDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TStochasticAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TBetaDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TBetaDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Shape parameters: alpha[" << _dparameterAlpha << "], beta["
            << _dparameterBeta << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

