//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCauchyDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-11-20 10:09:30

 \brief Interface of the class URANIE::DataServer::TCauchyDistribution

 */
/**

 * @class URANIE::DataServer::TCauchyDistribution
 *
 * @brief Description of the class TCauchyDistribution.
 *
 This class defines a stochastic attribute following a Cauchy-Lorentz distribution.

 This distribution has two parameters: the location parameter \f$ x_{0} \f$ 
 (a.k.a. the median or mode) and the scale parameter \f$ \gamma \f$. The scale 
 parameter is strictly positive.

 It has to be noted that although this distribution has a defined mode and median, 
 it has no defined mean, variance or any other higher moment. 

 \par Probability Density Function (PDF):
 \li \f$ f(x) = \gamma/\left(\pi*\left(\gamma^{2}+(x-x_{0})^{2}\right)\right) \f$

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = \pi^{-1} * \arctan\left((x-x_{0})/\gamma\right) + 0.5 \f$

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{0} + \gamma * \tan\left(\pi*(p-0.5)\right) \f$ for \f$ p \in [0,1] \f$

 */
#ifndef TCAUCHYDISTRIBUTION_H
#define TCAUCHYDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TCauchyDistribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterGamma;       ///< The scale parameter \f$ \gamma\f$
    Double_t _dparameterMedian; ///< The location parameter \f$ x_{0} \f$, a.k.a. median parameter

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default location is 0.0, default scale is 1.0 
     */
    TCauchyDistribution();

    //! Constructor setting the name
    /*!
     Default location is 0.0, default scale is 1.0 

     \param name (TString) Name of the attribute
     */
    TCauchyDistribution(TString name);

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param gamma (Double_t) the scale parameter
     \param median (Double_t) the location parameter (Default = 0.0)
     */
    TCauchyDistribution(TString name, Double_t gamma, Double_t median = 0.0);

    //! Construction of a TCauchyDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)
     */
    TCauchyDistribution(URANIE::DataServer::TAttribute *att);

    //! Default destructor
    virtual ~TCauchyDistribution();
    //@}

    /** @name Law parameters
     This law has two parameters: the location \f$ x_{0} \f$, also known 
     as median parameter, and the scale \f$ \gamma \f$. Both are defined 
     here.            
     **/
    //@{
    //! Set the location parameter
    /*!
     \param median (Double_t) the location parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE) 
     */
    virtual void setParameterMedian(Double_t median, Bool_t recompute = kTRUE);

    //! Return the location parameter
    Double_t getParameterMedian()
    {
        return _dparameterMedian;
    }

    //! Set the scale parameter
    /*!
     \param gamma (Double_t) the scale parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if gamma < 1e-4.
     */
    virtual void setParameterGamma(Double_t gamma, Bool_t recompute = kTRUE);

    //! Return the scale parameter
    Double_t getParameterGamma()
    {
        return _dparameterGamma;
    }

    //! Set all the parameters
    /*!
     \param gamma (Double_t) the scale parameter                
     \param median (Double_t) the location parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE) 
     */
    virtual void setParameters(Double_t gamma, Double_t median,
            Bool_t recompute = kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    /*!
     \warning Theoretical mean and standard deviation are undefined for 
     the Cauchy distribution. As a consequence, their values 
     are not set by this function.
     */
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TCauchyDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
