//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TComposedDistribution.cxx
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \date   Wed Aug 24 12:00:00 CET 2022

 \brief Implementation of the class URANIE::DataServer::TComposedDistribution

 totot
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TRandom.h"

// URANIE
#include "UExceptions.h"
#include "TComposedDistribution.h"

ClassImp (URANIE::DataServer::TComposedDistribution)

namespace URANIE
{
namespace DataServer
{

//===============================//
// Constructors and Destructors //
//===============================//

TComposedDistribution::TComposedDistribution()
{
    _iNbLaws = 0;
    _dSumWeights = 0.0;
    _inputLaws.clear();
    _inputWeights.clear();
}

TComposedDistribution::TComposedDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kCompose)
{
    _iNbLaws = 0;
    _dSumWeights = 0.0;
    _inputLaws.clear();
    _inputWeights.clear();
}


TComposedDistribution::~TComposedDistribution()
{
    // //  cout << "detruit TComposedDistribution" << endl;
    for (vector<URANIE::DataServer::TStochasticAttribute *>::reverse_iterator i = _inputLaws.rbegin(); i != _inputLaws.rend(); ++i )
        (*i)->delShare();
}

//=====================//
// Getters and Setters //
//=====================//      

int TComposedDistribution::addDistribution(URANIE::DataServer::TStochasticAttribute *statt, double weight)
{

    int ret_value = 0;
        
    try{
        if( find(_inputLaws.begin(), _inputLaws.end(), statt) != _inputLaws.end() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TComposedDistribution::addDistribution ERROR : the distribution [%s] has already been added to TComposedDistribution [%s].",
                                                            statt->GetName(), GetName()));

        if(weight==0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TComposedDistribution::addDistribution ERROR : the distribution [%s] has beed added with a weight value of 0.",
                                                            GetName()));
        if(weight<0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TComposedDistribution::addDistribution ERROR : the distribution [%s] has beed added with a negative weight.",
                                                            GetName()));

        if(statt->getPMin() != 0  || statt->getPMax() != 1)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                       Form("TComposedDistribution::addDistribution WARNING : the distribution [%s] has beed added with a truncation \
(pmin = %d and pmax = %d).\n This might be messing with the way you might picture its distribution (in the way its weighted w.r.t other components.",
                                                            GetName(), statt->getPMin(), statt->getPMax()));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        ret_value = 1;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    //  If all is well, let's fill everything
    _inputLaws.push_back(statt);
    _inputWeights.push_back(weight);
    _dSumWeights += weight;    
    _iNbLaws += 1;

    // Increment the attribute share counter
    statt->addShare();

    return ret_value;
}


//=======================//
// Statistical Functions //
//=======================//

void TComposedDistribution::computeTheoreticalInformation(int info)
{
    
    double value = THEO_DEF;
    vector<double> values, max_probas, mean_val, std_val, mode_val;
    int i=0, j=0; 

    switch(info)
    {
    case 0:
        mean_val.clear();
        try{
            for(i=0; i<_iNbLaws; i++)
                mean_val.push_back(_inputLaws.at(i)->getTheoreticalMean());
        }catch(URANIE::Exceptions::UWarningExceptions& ue)
        {
            cout << "\033[1;34m  <URANIE::INFO> The theoretical mean of " << _inputLaws.at(i)->GetName() << " is undefined " << endl
                 << "  <URANIE::INFO> The theoretical mean of the composed law ("<<GetName()<<") is set to default value ("<<THEO_DEF<<")\033[0m" << endl;
            _dtheoreticalMean = THEO_DEF;
            break;
        }            
        value = 0;
        for(i=0; i<_iNbLaws; i++)
            value += _inputWeights.at(i) * mean_val.at(i) / _dSumWeights;
        _dtheoreticalMean = ((_iNbLaws==0) ? THEO_DEF : value);
        break;

    case 1:
        mean_val.clear();
        try{
            for(i=0; i<_iNbLaws; i++)
                mean_val.push_back(_inputLaws.at(i)->getTheoreticalMean());
        }catch(URANIE::Exceptions::UWarningExceptions& ue)
        {
            cout << "\033[1;34m  <URANIE::INFO> The theoretical mean of " << _inputLaws.at(i)->GetName() << " is undefined " << endl
                 << "  <URANIE::INFO> The theoretical standard deviation of the composed law ("<<GetName()<<") is set to default value ("<<THEO_DEF<<")\033[0m" << endl;
            _dtheoreticalStdDev = THEO_DEF;
            break;
        }
        std_val.clear();
        try{
            for(i=0; i<_iNbLaws; i++)
                std_val.push_back(_inputLaws.at(i)->getTheoreticalStdDev());
        }catch(URANIE::Exceptions::UWarningExceptions& ue)
        {
            cout << "\033[1;34m  <URANIE::INFO> The theoretical standard deviation of " << _inputLaws.at(i)->GetName() << " is undefined " << endl
                 << "  <URANIE::INFO> The theoretical standard deviation of the composed law ("<<GetName()<<") is set to default value ("<<THEO_DEF<<")\033[0m" << endl;
            _dtheoreticalStdDev = THEO_DEF;
            break;
        }   
        value = 0;
        for(i=0; i<_iNbLaws; i++)
        {
            double deltai = mean_val.at(i) * _inputWeights.at(i);
            value += _inputWeights.at(i) / _dSumWeights * std_val.at(i) * std_val.at(i);
            value += (_dSumWeights - _inputWeights.at(i)) / (_inputWeights.at(i) * _dSumWeights * _dSumWeights) * deltai * deltai ;
            for(j=i+1; j<_iNbLaws; j++)
            {
                double deltaj = mean_val.at(j) * _inputWeights.at(j);
                value += - 2 / (_dSumWeights*_dSumWeights) * deltai * deltaj;
            }
        }
        _dtheoreticalStdDev = ((_iNbLaws==0) ? THEO_DEF : sqrt(value));
        break;
        
    case 2:
        mode_val.clear();
        try{
            for(i=0; i<_iNbLaws; i++)
                mode_val.push_back(_inputLaws.at(i)->getTheoreticalMode());
        }catch(URANIE::Exceptions::UWarningExceptions& ue)
        {
            cout << "\033[1;34m  <URANIE::INFO> The theoretical mode of " << _inputLaws.at(i)->GetName() << " is undefined " << endl
                 << "  <URANIE::INFO> The theoretical mode of the composed law ("<<GetName()<<") is set to default value ("<<THEO_DEF<<")\033[0m" << endl;
            _dtheoreticalMode = THEO_DEF;
            break;
        }
        double tested_value = 0, tested_proba = 0, max_proba = -9999999;
        for(i=0; i<_iNbLaws; i++)
        {            
            tested_value = mode_val.at(i);
            tested_proba = _inputLaws.at(i)->getPDF(tested_value);

            if( (_inputWeights.at(i) * tested_proba) >= max_proba )
            {
                max_proba = (_inputWeights.at(i) * tested_proba);
                _dtheoreticalMode = tested_value;
                
                if( find(max_probas.begin(), max_probas.end(), max_proba) == max_probas.end())
                {
                    max_probas.clear();
                    values.clear();
                    max_probas.push_back(max_proba);
                    values.push_back(tested_value);
                }
                else
                    values.push_back(tested_value);
            }
        }
        break;
    }
    try{
        if(values.size()>1)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("TComposedDistribution::computeTheoreticalInformation WARNING: the composition [%s] has several modes with the same probability. \n\
Here is a list of values found [%s].",
                                                              GetName(), URANIE::DataServer::UEntry::dumpVectorStr(&values).c_str()));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    }
        
}

Double_t TComposedDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0;

    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
    {
        for(int iatt=0; iatt<_iNbLaws; iatt++)
            cdf += _inputLaws.at(iatt)->getCDF(x) * _inputWeights.at(iatt) / _dSumWeights;
        cdf = (cdf - _pmin) / (_pmax - _pmin) ;
    }

    return cdf ;
}

Double_t TComposedDistribution::getPDF(Double_t x)
{
    // If x is not in the distribution, its probability is 0.0
    Double_t p = 0.0;
    
    if ((x >= lowerBound) && (x <= upperBound))
    {
        for(int iatt=0; iatt<_iNbLaws; iatt++)
        {
            p += _inputLaws.at(iatt)->getPDF(x) * _inputWeights.at(iatt) / _dSumWeights;
        }
    }

    return p ;
}


Double_t TComposedDistribution::getInvCDF(Double_t p)
{

    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TComposedDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                                                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
             << p << "]" << endl;

    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p; 
    
    double x0 = lowerBound, x1 = upperBound;
    Double_t px;
    double xcrt = (upperBound + lowerBound) / 2.0;
    double xtol = ((x0!=0 && x0!= -DLE_MAX)?TMath::Abs(x0):0.1) * 1e-14, ptol = 1e-20;
    //double xtol = ((x0!=0)?x0:1) * nextafter(0., HUGE_VAL), ptol = 1e-20;

    Int_t ncpt = 0;

    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                 << "], current solution[" << xcrt
                 << "], corresponding probability[" << setprecision(20)<<px
                 << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]  diff versus tol ["<<setprecision(20)<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]"
                 << "] diffproba  ["<<setprecision(30)<<px<<"  "<<realP<<"/"<<TMath::Abs(px - realP)<<"]";

        if (TMath::Abs(px - realP) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                     << "], corresponding probability[" << setprecision(20)<<px
                     << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < realP)
            x0 = xcrt;
        else if (px > realP)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                 << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                 << endl;
        ncpt++;

        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
    }

    if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TCompose");
        _localmess <<  " TComposedDistribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }
    
    return xcrt;
}

//=====//
// Log //
//=====//

void TComposedDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    //     TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TComposedDistribution::printLog  ******" << endl;
    cout << "**   Name    [" << GetName() << "]" << endl;
    cout << "**   label   [" << GetTitle() << "]" << endl;
    cout << "**   unite   [" << getUnity() << "]" << endl;
    cout << "**   law     [" << _lawType << "]" << endl;
    cout << "**   type    [" << _nType << "]" << endl;
    cout << "**   nValues [" << _iNbLaws << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

