//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef TDSNtupleD_H
#define TDSNtupleD_H

// ROOT
#include "TROOT.h"
#include "TTree.h"
#include "TMatrixD.h"

// URANIE
#include "DataServer.h"
#include "UEntry.h"

#define SIZENAME string("size")

using namespace std;

namespace URANIE
{
namespace DataServer
{
class TDSNtupleD: public TTree
{
public:

    /** @name Constructors & destructor 
	The possible constructors
    **/
    //@{
    //! Standard empty constructor
    TDSNtupleD();
    
    //! Standard desctructor
    ~TDSNtupleD();

    //! Construct a new TDSNtupleD from an existing TTree/TDSNtupleD
    /*!
      It is used for compatibility reasons. The contents of the TTree
      are copied (i.e. duplicated) into the TDSNtupleD.
      \param ntd (TTree*): a pointer to the TTree to be copied.
      \param varlist: list of variable to be read separated by ":".
      \param cut: TTreeeFormula used to REJECT events 
      \param NewIsMainOne: MAke the new tree have the regular name, the original one will then be GetName()+"_copied". Otherwise the new one will be GetNAme()+"_copy"
    */
    TDSNtupleD(TTree* ntd, const char* varlist="*", const char* cut="", bool NewIsMainOne=false);

    //! Construct a new TDSNtupleD from a list of variable
    /*!
      This is the main constructor of the TDSNtupleD.
      \param name (const char*): name of the object.
      \param title (const char*): title of the object.
      \param varlist (const char*): a list of branch names separated by ":".
      \param vartype (const char*): a list of branch types separated by ":". 
      Types are "D" for double, "V" for vectpr<double>, "S" for string.
      \param bufsize (Int_t): buffer size (default = 32000).
    */
    TDSNtupleD(const char* name, const char* title, const char* varlist,
	       const char *vartype, Int_t bufsize = 32000);

    //! Construct a new TDSNtupleD from a list of variable
    /*!
      Other version kept for back-compatibility assume that all variable are double 
      \warning: vartype is defined as "D" by construction !!
      \param name (const char*): name of the object.
      \param title (const char*): title of the object.
      \param varlist (const char*): a list of branch names separated by ":".
      Types are "D" for double, "V" for vectpr<double>, "S" for string.
      \param bufsize (Int_t): buffer size (default = 32000).
    */
    TDSNtupleD(const char* name, const char* title, const char* varlist,
	       Int_t bufsize = 32000);

    //! Construct a new TDSNtupleD from a list of variable and a TMatrixD
    /*!
      This constructor allows to create a TDSNtupleD and fill it with data contained in a TMatrixD.
      The number of columns of the matrix must match the number of variables of the TDSNtupleD.
      \warning: Assume that all variable are double (==> vartype is made with "D" only)
      \param name (const char*): name of the object.
      \param title (const char*): title of the object.
      \param varlist (const char*): a list of branch names separated by ":".
      \param data (TMatrixD): a matrix containing the data.
      \param bufsize (Int_t): buffer size (default = 32000).
      \exception UranieError if the number of variables in varlist is different from the number of columns of data.
    */
    TDSNtupleD(const char* name, const char* title, const char* varlist,
	       TMatrixD data, Int_t bufsize = 32000);

    //@}
    

    /** @name Fillers
	Function used to fill data
    **/
    //@{
    //! General Fill form: return TTree::Fill()
    virtual Int_t  Fill();

    //! Fill tuple with values in AllValues
    /*!
      \param AllValues: values to be stored in the tuple
      \exception UranieError if the number of variables in varlist is different from the number of columns of data, 
      or if string or vector<double> are available in the tuple 
    */
    Int_t     Fill(Double_t* AllValues);
    
    //! Fill tuple with values, up to 15 arguments
    /*!
      Back compatibility with the older version of TDSNtupleD that inherited this method from TNtupleD
      \param x0, x1, .., x14: values to be stored in the tuple
      \exception UranieError if the number of variables in varlist is different from the number of columns of data, 
      or if string or vector<double> are available in the tuple 
    */
    Int_t     Fill(Double_t x0, Double_t x1, Double_t x2=0, Double_t x3=0, Double_t x4=0,
		   Double_t x5=0, Double_t x6=0, Double_t x7=0, Double_t x8=0, Double_t x9=0,
		   Double_t x10=0, Double_t x11=0, Double_t x12=0, Double_t x13=0, Double_t x14=0);

    
    //! Fill tuple with a line from an input file
    /*!
      \param line: contain an entry for all attribute of the tree 
      \param IterIndex: give the index in the line of the iterator (if there is one)
    */
    Int_t     Fill(string line, int IterIndex=-1);
    
    //! Fill tuple with a vector of UEntry
    /*!
      \param evt: Vector of UEntry that will be used to 
    */
    Int_t     Fill(vector<URANIE::DataServer::UEntry*> *event);

    //! Fill tuple with two vectors of UEntry
    /*!
      \param vinput: Vector of UEntry that will be used to fill inputs
      \param voutput: Vector of UEntry that will be used to fill outputs
    */
    Int_t     Fill(vector<URANIE::DataServer::UEntry*> *vinput, vector<URANIE::DataServer::UEntry*> *voutput);

    //@}

    
    /** @name Operation on branches
	Creation of new branches and reassignement of their addresses
    **/
    //@{

    //! Create a new Branch
    /*!
      This function properly add a new Branch in the TDSNtupleD. The 
      parameter _fNvar is incremented, the memory repository (_fDArgs or _vfVArgs) 
      is resized accordingly depending on the nature of the new branch, and 
      the addresses of the existing branches are reset.
      This function is written for the addAttribute method and does not concern 
      the string branch (no TTreeFormula can be computed).
      \warning do not forget to call TNtupleD::ResetBranchAddress(TBranch*) 
      or TNtupleD::ResetBranchAddresses() after you are done with the data 
      filling. This will insure that the new branch is properly accessible 
      by filler function.
      \param name (const char*): name of the new variable
      \param address (void*): pointer to a Double_t variable which will be 
      used to fill the new branch.
      \param leaflist (const char*): this parameter is left for 
      compatibility reasons. It is advised to leave this string empty 
      (default = "").
      \param bufsize (Int_t): buffer size (default = 32000)
    */

    TBranch* Branch(const char* name, void* address, const char* leaflist,
		    Int_t bufsize = 32000);
    
    //! Reset the branch addresses to the internal memory arrays. 
    /*!
      Use this method when the addresses were changed via calls to SetBranchAddress().
      * \param branch (TBranch*): the pointer to the branch we want to 
      * reset the addresses
      */
    void ResetBranchAddress(TBranch *branch);
    //! Reset all branches addresses to the internal memory arrays.
    /*! 
      Use this method when the addresses were changed via calls to SetBranchAddress().
    */
    void ResetBranchAddresses();
    //@}
    
 
    /** @name Get data / extract data / readfiles 
	Methods to read and extract data for file, tuples...
    **/
    //@{

    //! Return only the double argument of the tuple
    /*!
      Check whether this should be kept !!
    */
    Double_t *GetArgs();
    
    //! Return a pointer to the selected entries of a given branch.
    /*!
      \warning the returned pointer redirects to the results of a "TDSNtupleD::Draw()"
      call, which means that any new call will modify the contents. If you
      need to protect the values of the retrieved data, you either need to
      copy the contents of the returned array or use the function "copyBranchData".
      \param branchName (const char*): the name of the branch from which we
      want to retrieve the data.
      \param selection (const char*): a selection (in ROOT style) of the data to
      retrieve (default = "").
      \param nbDataPtr (Int_t *): address of an integer variable which will 
      receive the number of selected data (default = NULL).
      \exception UErrorException if the branch does not exist.
      \exception UErrorException if the branch is not a branch filled with regular double.
    */
    Double_t* getBranchData(const char* branchName, const char* selection = "",
			    Int_t* nbDataPtr = NULL);
    
    //! Copy the values of the selected entries of a given branch into the provided array
    /*!
      Contrary to "getBranchData" which does not copy the data, this function
      allows to safely retrieve the selected data of a given branch. The user
      must provide an array with enough allocated memory to contains all the
      data to copy.
      \param copyArray (Double_t*): an array of double allocated by the user.
      Unless the size of the selection is known, it is advised to allocate 
      the maximum number of data in the branch.
      \param arraySize (Int_t): the size of the "copyArray".
      \param branchName (const char*): the name of the branch from which we
      want to retrieve the data.
      \param selection (const char*): a selection (in ROOT style) of the data to
      retrieve (default = "").
      \param nbDataPtr (Int_t *): address of an integer variable which will 
      receive the number of selected data (default = NULL).            
      \exception UErrorException if the branch does not exist. 
      \exception UErrorException if the branch is not a branch filled with regular double.
      \exception UErrorException if the number of selected data is greater 
      than the size of the array. 
    */
    void copyBranchData(Double_t* copyArray, int arraySize,
			const char* branchName, const char* selection = "",
			Int_t* nbDataPtr = NULL);
    
    //! Extract selected data from the tuple to a flat array.
    /*!
      The extracted data are stored in a Double_t array passed as parameter 
      and allocated by the user.
      This is a "flat" array, i.e. if more than one attribute is selected,
      the user need to be aware of the order in which the values are stored.
      This order is given as option:
      - option == "line":   dataOut[i * nbCol + j]  = tds[i,j]
      - option == "column": dataOut[i + j * nbLine] = tds[i,j]
      where i is the event index, j is the attribute index, nbCol is the
      number of selected attributes and nbLine the number of selected
      events.
      PLEASE NOTE that the expression "tds[i,j]" does not exist in URANIE.
      It is used only for simplicity.
      \param dataOut (double_t*): an array of double which will store the extracted data.
      \param arraySize (Int_t): the size of the array dataOut. It must be at least 
      nCol * nLines, where nCol is the number of selected variables, and nLines is
      the number of selected elements. 
      \param varexpinput (const char*): a list of attribute names separated by semicolons (":").
      If "" or "*", all the variables of the tuple are used (default = "*").
      \param selection (const char*): a selection (in ROOT style) of the data to retrieve (default = "").
      \param option (const Option_t*): a character string defining how multi-dimensionnal data
      are stored in dataOut (cf. description above) (default: "line").
      \param nbDataPtr (Int_t *): address of an integer variable which will 
      receive the number of selected data (default = NULL).
      \exception UErrorException if one of the selected branches does not exist.
      \exception UErrorException if one of the branches is not a branch filled with regular double.
      \exception UErrorException if the number of selected data is greater 
      than the size of the array. 
    */
    void extractData(Double_t* dataOut, Int_t arraySize,
		     const char* varexpinput = "*", const char* selection = "",
		     const Option_t* option = "line", Int_t *nbDataPtr = NULL);

    //! Extract selected data from the tuple to a TMatrixD.
    /*!
      The extracted data are stored in a TMatrixD. The matrix will be resized as necessary and all
      its previous contents will be lost.
      Each column of the matrix corresponds to an input variable, and each line to an event.
      \param matrix (TMatrixD*): pointer to a matrix of double that will store the extracted values.
      \param varexpinput (const char*): a list of attribute names separated by semicolons (":").
      If "" or "*", all the variables of the tuple are used (default = "*").
      \param selection (const char*): a selection (in ROOT style) of the data to retrieve (default = "").
      \exception UErrorException if one of the selected branches does not exist.
      \exception UErrorException if one of the branches is not a branch filled with regular double.
    */
    void extractData(TMatrixD* matrix, const char* varexpinput = "*",
		     const char* selection = "");

    //! Fill the TDSNtupleD using data from a text file.
    /*!
      This function overrides the TNtupleD's one. There is
      no limit in the number of characters that can be read in a row.
      \param filename (const char*): the name of the file containign the data.
    */
    Long64_t ReadFile(const char* filename);
    //@}
    
    /** @name  Utility 
	Utility function used mostly by other methods.
    **/
    //@{

    //! To be written, might be very usefull :D
    void Merger(TDSNtupleD *file, const string newbrname, const string oldbrname, bool tdsmerge=false, const char* iter1="", const char* iter2="");
    //! Parse a string using delim as a separation
    /*!
      This function parse a string using delim as a seperation and stores the results in the given vector
      \param input: the string to be parsed
      \param delim: the separation to be used
      \param res: the vector to be used to store all substring
    */

    void Browse(TBrowser *b);  
    //@}
    


    /** @name  Getters
	Getters simple function
    **/
    //@{    
    virtual Int_t     GetNvar() const { return _fNvar; }
    UInt_t getNVar(){return _fNvar;};
    string getVarName(){return _varname;};
    string getVarType(){return _vartype;};
    int getVectSize(){return _vfArgs.size();};
    //@}
    
    ClassDef(URANIE::DataServer::TDSNtupleD, ID_DATASERVER);    
protected:

    UInt_t       _i;                     ///< iterator vastly used
    UInt_t       _fNvar;                  ///<  Number of columns
    vector<UEntry*> _vfArgs; //! Vector of entries (memory buffer and their ccorresponding types)
    vector<Double_t>    _vfDArgs;                ///< Array of variables defined as double    
    vector<string*> _vfSArgs;     ///< Vector of variable defined as char*
    vector<vector<Double_t>* > _vfVArgs;  ///< Vector of Vector for variable defined that should be vector[double] (i.e. levele temporal)
    string   _varname; ///< string describing the name of the variable
    string   _vartype; ///< string describing the type of the variable
    
    map<string,unsigned int> _mTypeNb;      ///< Map to store total number of variable per type
    map<unsigned int,unsigned int> _mVarTypeNb;  ///< Map to the specific type index for the index i in total variable reference    
    
    /** @name Initialisation 
	Internal method used to initialise the ntuples
    **/
    //@{
    
    //! Initialise TTree features
    /*!
      \param name (const char*): name of the object.
      \param title (const char*): title of the object.
    */
    void _initTTree(const char* name, const char* title);
    
    //! Create the branches of the TDSNtupleD
    /*!
      \param bufsize (Int_t): buffer size.
    */
    void _branchConstruction(Int_t bufsize);

    //! Clear the vectors and maps
    /*!
      Clear the internal vectors and maps
    */
    void ClearVectorsAndMaps();
    //@}


    //Flag special because windows complains a lot
#ifdef WIN32 
    Int_t Branch(TCollection* list, Int_t bufsize = 32000, Int_t splitlevel = 99, const char* name = "")
	{
	    return TTree::Branch(list, bufsize, splitlevel, name );
	}
    Int_t Branch(TList* list, Int_t bufsize = 32000, Int_t splitlevel = 99)
	{
	    return TTree::Branch(list, bufsize , splitlevel );    
	}
    Int_t Branch(const char* folder, Int_t bufsize = 32000, Int_t splitlevel = 99)
	{
	    return TTree::Branch(folder, bufsize , splitlevel );
	}
    TBranch* Branch(const char* name, const char* classname, void* addobj, Int_t bufsize = 32000, Int_t splitlevel = 99)
	{
	    return TTree::Branch(name, classname, addobj, bufsize, splitlevel );
	}
#endif
    
};

}
}

#endif
