//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGeneralizedNormalDistribution.cxx
 \author Rudy Chocat <rudy.chocat@cea.fr>
 \date   28 oct. 2021 11:13:50

 \brief Implementation of the class URANIE::DataServer::TGeneralizedNormalDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TGeneralizedNormalDistribution.h"
 #include "Math/QuantFuncMathCore.h"

ClassImp (URANIE::DataServer::TGeneralizedNormalDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TGeneralizedNormalDistribution::TGeneralizedNormalDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kGeneralizedNormal), _dparameterMu(
                0.0), _dparameterAlpha(1.0), _dparameterBeta(2.0)
{
    //  cout << "construit TGeneralizedNormalDistribution" << endl;
    computeTheoreticalInformation();
}

TGeneralizedNormalDistribution::TGeneralizedNormalDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kGeneralizedNormal), _dparameterMu(
                0.0), _dparameterAlpha(1.0), _dparameterBeta(2.0)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalDistribution[" << this << "]str[" << str<<"]"<<endl;
    computeTheoreticalInformation();
}

TGeneralizedNormalDistribution::TGeneralizedNormalDistribution(TString str, Double_t mu,
        Double_t alpha, Double_t beta) :
        TInfiniteDistribution(str, TStochasticAttribute::kGeneralizedNormal)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalDistribution[" << this << "]str[" << str<<"] Mean[" << mean <<"] Alpha[" << alpha << "]" Beta[" << beta << "]" << endl;
    setParameters(mu, alpha, beta);
}

TGeneralizedNormalDistribution::TGeneralizedNormalDistribution(URANIE::DataServer::TAttribute *att,
        Double_t mu, Double_t alpha, Double_t beta) :
        TInfiniteDistribution(att, TStochasticAttribute::kGeneralizedNormal)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalDistribution att[" << this << "]att->getName()[" << att->getName()<<"] Mean[" << mean <<"] Alpha[" << alpha << "]" Beta[" << beta << "]" << endl;
    setParameters(mu, alpha, beta);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TGeneralizedNormalDistribution::~TGeneralizedNormalDistribution()
{
    //  cout << "detruit TGeneralizedNormalDistribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TGeneralizedNormalDistribution::setParameterMu(Double_t mu, Bool_t recompute)
{
    _dparameterMu = mu;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalDistribution::setParameterAlpha(Double_t alpha, Bool_t recompute)
{
    try
    {
        if (alpha <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalDistribution::setParameterAlpha: parameter alpha [%f] is negative: it must be strictly positive",
                            alpha));
        if (TMath::Abs(alpha) < 1e-6 * TMath::Abs(_dparameterMu))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalDistribution::setParameterAlpha: parameter alpha [%f] is too close to zero: lower limit is 1e-6 * mu [%f]",
                            alpha, 1e-6 * TMath::Abs(_dparameterMu)));
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterAlpha = alpha;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalDistribution::setParameterBeta(Double_t beta, Bool_t recompute)
{
    try
    {
        if (beta <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalDistribution::setParameterBeta: parameter beta [%f] is negative: it must be strictly positive",
                            beta));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterBeta = beta;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalDistribution::setParameters(Double_t mu, Double_t alpha, Double_t beta,
        Bool_t recompute)
{
    setParameterMu(mu, kFALSE);
    setParameterAlpha(alpha, kFALSE);
    setParameterBeta(beta, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TGeneralizedNormalDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterMu;

    _dtheoreticalStdDev = TMath::Sqrt( (_dparameterAlpha * _dparameterAlpha) * TMath::Gamma(3.0 / _dparameterBeta) / TMath::Gamma(1.0 / _dparameterBeta) );

    _dtheoreticalMode = _dparameterMu;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TGeneralizedNormalDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        pdf = (_dparameterBeta / ( 2.0 * _dparameterAlpha * TMath::Gamma(1.0 / _dparameterBeta) ) )
                * TMath::Exp(
                        (-1.0) * TMath::Power( TMath::Abs(x - _dparameterMu) / _dparameterAlpha, _dparameterBeta ) );
        // Take into account the case where at least one of the bounds is finite
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TGeneralizedNormalDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = 0.5 + 0.5 * TMath::Sign(1.0, x - _dparameterMu) 
                * TMath::Gamma( 1.0 / _dparameterBeta, TMath::Power( TMath::Abs(x - _dparameterMu) / _dparameterAlpha, _dparameterBeta ) );
//                 / ( TMath::Gamma(1.0 / _dparameterBeta) ) ;

        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TGeneralizedNormalDistribution::getInvCDF(Double_t p)
{
    
        // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TGeneralizedNormalDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                                                           p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
             << p << "]" << endl;

    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p;// * (_pmax - _pmin) + _pmin;
    
    Double_t x0 = lowerBound, x1 = upperBound, px;
    Double_t xcrt = (upperBound + lowerBound) / 2.0;
    //Double_t xtol = x0 * 0.000000000000001, ptol = 0.000000000000001;
    Double_t xtol = ((x0!=0 && x0!= -DLE_MAX)?TMath::Abs(x0):0.1) * 1e-14, ptol = 1e-20;
    //Double_t xtol = ((x0!=0)?x0:1) * nextafter(0., HUGE_VAL), ptol = 1e-15;

    Int_t ncpt = 0;
    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                 << "], current solution[" << xcrt
                 << "], corresponding probability[" << px
                 << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]  diff versus tol ["<<setprecision(20)<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]";
        if (TMath::Abs(px - realP) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                     << "], corresponding probability[" << px
                     << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < realP)
            x0 = xcrt;
        else if (px > realP)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                 << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                 << endl;
        ncpt++;

        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
    }

     if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TGeneralizedNormalDistribution");
        _localmess <<  " TGeneralizedNormalDistribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }
    
    return xcrt;
    
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TGeneralizedNormalDistribution::printLog(Option_t *option)
{
    //  TInfiniteDistribution::printLog();
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TGeneralizedNormalDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Normal" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mean parameter mu[" << _dparameterMu << "]" << endl;
    cout << "**   Scale parameter alpha [" << _dparameterAlpha << "]" << endl;
    cout << "**   Shape parameter beta [" << _dparameterBeta << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

