//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGeneralizedNormalV2Distribution.h
 \author Rudy Chocat <rudy.chocat@cea.fr>
 \date   28 oct. 2021 11:13:50
 
 \brief Interface of the class URANIE::DataServer::TGeneralizedNormalV2Distribution

 */
/**

 * @class URANIE::DataServer::TGeneralizedNormalV2Distribution
 *
 * @brief Description of the class TGeneralizedNormalV2Distribution.

 This class defines a stochastic attribute following an Normal (or Gaussian) distribution.

 This distribution has two parameters: the mean parameter \f$ \mu \f$ and the 
 standard deviation \f$ \sigma > 0 \f$.

 \par Probability Density Function (PDF):
 \li \f$ f(x) =  \exp\left( -( x - \mu)^2 / 2\sigma^2 \right) / \sqrt{2\pi\sigma^2}\f$

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0.5 * \left( 1 + erf\left( (x-\mu)/\sqrt{2\sigma^2}\right) \right)\f$

 \par Inverse CDF:
 \li \f$ F^{-1}(p) =  \mu + erf^{-1}\left( 2p-1 \right) * \sqrt{2\sigma^2}\f$ for \f$ p \in [0,1] \f$


 \note In order to reduce numerical errors, the algorithm used to compute \f$ F^{-1}(p) \f$ 
 is actually either Peter John Acklam's algorithm (http://home.online.no/~pjacklam/notes/invnorm) 
 or the ROOT implementation of AS241 algorithm (Michael J. Wichura, Applied Statistics, vol. 37, 1988, pp. 477-484). 
 The latter is used by default.


 */
#ifndef TGENERALIZEDNORMALV2DISTRIBUTION_H
#define TGENERALIZEDNORMALV2DISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TGeneralizedNormalV2Distribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterXi;    ///< Position parameter \f$ \xi \f$
    Double_t _dparameterAlpha; ///< Scale parameter \f$ \alpha \f$
    Double_t _dparameterKappa; ///< Shape parameter \f$ \kappa \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default mean is 0.0, default standard deviation is 1.0 
     */
    TGeneralizedNormalV2Distribution();

    //! Constructor setting the name
    /*!
     Default mean is 0.0, default standard deviation is 1.0 

     \param name (TString) Name of the attribute
     */
    TGeneralizedNormalV2Distribution(TString name);

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param xi (Double_t) Position parameter \f$ \xi \f$
     \param alpha (Double_t) Scale parameter \f$ \alpha \f$
     \param kappa (Double_t) Shape parameter \f$ \kappa \f$
     */
    TGeneralizedNormalV2Distribution(TString name, Double_t xi, Double_t alpha, Double_t kappa);

    //! Construction of a TGeneralizedNormalV2Distribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param xi (Double_t) Position parameter \f$ \xi \f$
     \param alpha (Double_t) Scale parameter \f$ \alpha \f$
     \param kappa (Double_t) Shape parameter \f$ \kappa \f$
     */
    TGeneralizedNormalV2Distribution(URANIE::DataServer::TAttribute *att, Double_t xi,
            Double_t alpha, Double_t kappa);

    //! Default destructor
    virtual ~TGeneralizedNormalV2Distribution();
    //@}

    /** @name Law parameters
     This law has three parameters: the position \f$ \xi \f$, the scale \f$ \alpha \f$ and the shape \f$ \beta \f$. Both are defined here. 
     **/
    //@{            
    //! Set the mean parameter
    /*!
     \param xi (Double_t) Position parameter \f$ \xi \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)                            
     */
    virtual void setParameterXi(Double_t xi, Bool_t recompute = kTRUE);

    //! Return the rate parameter
    Double_t getParameterXi()
    {
        return _dparameterXi;
    }

    //! Set the scale parameter
    /*!
     \param alpha (Double_t) Scale parameter \f$ \alpha \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)

     \exception UErrorExceptions if scale <  1e-6 * mu                   
     */
    virtual void setParameterAlpha(Double_t alpha, Bool_t recompute = kTRUE);

    //! Return the scale parameter
    Double_t getParameterAlpha()
    {
        return _dparameterAlpha;
    }
    
    //! Set the shape parameter
    /*!
     \param kappa (Double_t) Shape parameter \f$ \kappa \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)

     \exception UErrorExceptions if scale <  1e-6 * mu                
     */
    virtual void setParameterKappa(Double_t kappa, Bool_t recompute = kTRUE);

    //! Return the shape parameter
    Double_t getParameterKappa()
    {
        return _dparameterKappa;
    }

    //! Set all the parameters
    /*!
     \param xi (Double_t) Position parameter \f$ \xi \f$
     \param alpha (Double_t) Scale parameter \f$ \alpha \f$
     \param kappa (Double_t) Shape parameter \f$ \kappa \f$            
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t xi, Double_t alpha, Double_t kappa, Bool_t recompute =
            kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     The ROOT implementation of AS241 algorithm is used by default.
     
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TGeneralizedNormalV2Distribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
