//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGumbelDistribution.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   Mon Jan 19 12:00:00 CET 2009
 
 \brief Implementation of the class URANIE::DataServer::TGumbelDistribution

 totot
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TGumbelDistribution.h"

ClassImp (URANIE::DataServer::TGumbelDistribution)

namespace URANIE
{
namespace DataServer
{

//===============================//
// Constructors and Desctructors //
//===============================//

TGumbelDistribution::TGumbelDistribution()
{
    //     cout << "build TGumbelDistribution" << endl; 
  clearVectors();
}

TGumbelDistribution::TGumbelDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kGumbel)
{
    //     cout << "Constructor of URANIE::DataServer::TGumbelDistribution[" << this << "]str[" << str<<"]"<<endl;
    _dmod = 0.;
    dscale = 1.;
    computeMean();
}

TGumbelDistribution::TGumbelDistribution(TString str, Double_t mu,
        Double_t beta) :
        TInfiniteDistribution(str, TStochasticAttribute::kGumbel)
{
    //     cout << "Constructeur URANIE::DataServer::TGumbelDistribution[" << this << "]str[" << str<<"] Location [" << mu <<"] Scale [" << beta << "]" << endl;
    setLocation(mu);
    setScale(beta);
}

TGumbelDistribution::TGumbelDistribution(URANIE::DataServer::TAttribute *att,
        Double_t mu, Double_t beta) :
        TInfiniteDistribution(att, TStochasticAttribute::kGumbel)
{
    setLocation(mu);
    setScale(beta);

    //     cout << "Constructor of URANIE::DataServer::TGumbelDistribution att[" << this << "]att->getName()[" << att->getName()<<"] Location [" << mu <<"] Scale [" << beta << "]" << endl;
}

TGumbelDistribution::~TGumbelDistribution()
{
    //     cout << "detruit TGumbelDistribution" << endl; 

}

//=====================//
// Getters and Setters //
//=====================//      

void TGumbelDistribution::setLocation(Double_t mu)
{
    _dmod = mu;
    if (_blog)
    {
        cout << "** Changed Location [" << _dmod << "]" << endl;
    }
    computeMean();
}

void TGumbelDistribution::setScale(Double_t beta)
{
    if (beta <= 0.)
    {
        // An exception is necessary here:
        cout
                << "*** WARNING: Scale parameter must be strictly positive. Current value: ["
                << beta << "]." << endl;
        cout << "***          To avoid problems, value has been set to [1.0]."
                << endl;
        dscale = 1.0;
    }
    else
    {
        dscale = beta;
    }

    if (_blog)
    {
        cout << "** Changed Scale [" << dscale << "]" << endl;
    }
    computeMean();
}

Double_t TGumbelDistribution::getLocation()
{
    return _dmod;
}

Double_t TGumbelDistribution::getScale()
{
    return dscale;
}

//======================//
// Satistical Functions //
//======================//

Double_t TGumbelDistribution::getPDF(Double_t x)
{
    Double_t pdf;
    pdf = TMath::Exp(-TMath::Exp((_dmod - x) / dscale));
    return pdf;
}
/*
 \todo A faire si la loi est bornee
 */
void TGumbelDistribution::computeMean()
{
    dmean = _dmod + dscale * TMath::EulerGamma();
    dstddev = dscale * TMath::Pi() / TMath::Sqrt(6.);
}

Double_t TGumbelDistribution::getInvCDF(Double_t p)
{
    Double_t xhat = _dmod - dscale * TMath::Log(-TMath::Log(p));
    // fga::todo verifier les bornes pour une loi exponentielle
    if (_blog)
    {
        cout << "** Location [" << _dmod << "]" << endl;
        cout << "** Scale [" << dscale << "]" << endl;
        cout << "** p  [" << p << "]" << endl;
        cout << "** TMath::Log(p) [" << TMath::Log(p) << "]" << endl;
        cout << "TGumbelDistribution::getInvCDF p[" << p << "]  x[" << xhat
                << "]" << endl;
    }
    return xhat;
}

//=====//
// Log //
//=====//

void TGumbelDistribution::printLog(Option_t *option)
{
    //     TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TGumbelDistribution::printLog  ******" << endl;
    cout << "**   Name    [" << GetName() << "]" << endl;
    cout << "**   label   [" << GetTitle() << "]" << endl;
    cout << "**   unite   [" << getUnity() << "]" << endl;
    cout << "**   law     [" << lawType << "]" << endl;
    cout << "**   Location[" << _dmod << "]; Scale [" << dscale << "]" << endl;
    cout << "**   Mean    [" << dmean << "]; StdDev[" << dstddev << "]" << endl;
    cout << "**   Bound   [" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   PMin    [" << _pmin << "]; PMax  [" << _pmax << "]" << endl;
    cout << "**   type    [" << dataType << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

