//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogNormalDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TLogNormalDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TLogNormalDistribution.h"

ClassImp (URANIE::DataServer::TLogNormalDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TLogNormalDistribution::TLogNormalDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kLogNormal), _unormal(
                NULL), _dparameterXMin(0.0), _dparameterMean(0.01), _dparameterErrorFactor(2.0)
{
    if (_blog)
        cout << "Constructing TLogNormalDistribution" << endl;
    TAttribute::setLowerBound(_dparameterXMin,true);
    createUnderlyingNormal(_dparameterMean - _dparameterXMin,
            _dparameterErrorFactor);
    computeTheoreticalInformation();
}

TLogNormalDistribution::TLogNormalDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kLogNormal), _unormal(
                NULL), _dparameterXMin(0.0), _dparameterMean(0.01), _dparameterErrorFactor(2.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogNormalDistribution with: Name["
                << str << "]" << endl;
    TAttribute::setLowerBound(_dparameterXMin,true);
    createUnderlyingNormal(_dparameterMean - _dparameterXMin,
            _dparameterErrorFactor);
    computeTheoreticalInformation();
}

TLogNormalDistribution::TLogNormalDistribution(TString str, Double_t M,
        Double_t errorfactor, Double_t xmin) :
        TInfiniteDistribution(str, TStochasticAttribute::kLogNormal)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogNormalDistribution with: Name["
                << str << "] Mean[" << M << "] Error Factor[" << errorfactor
                << "] xmin[" << xmin << "]" << endl;
    TAttribute::setLowerBound(xmin,true);
    setParameters(M, errorfactor, xmin);
}

TLogNormalDistribution::TLogNormalDistribution(
        URANIE::DataServer::TAttribute *att, Double_t M, Double_t errorfactor,
        Double_t xmin) :
        TInfiniteDistribution(att, TStochasticAttribute::kLogNormal)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogNormalDistribution from URANIE::DataServer::TAttribute object ["
                << att->GetName() << "] " << endl;

    TAttribute::setLowerBound(xmin,true);
    setParameters(M, errorfactor, xmin);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TLogNormalDistribution::~TLogNormalDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TLogNormalDistribution object "
                << GetName() << endl;
    if (_unormal)
        delete _unormal;
    _unormal = NULL;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TLogNormalDistribution::createUnderlyingNormal(Double_t M,
        Double_t errorfactor)
{

    Double_t sigmacrt = TMath::Log(errorfactor) / 1.645;
    Double_t meancrt = TMath::Log(M) - 0.5 * sigmacrt * sigmacrt;

    _unormal = new TNormalDistribution(
            Form("guranie_unormal_underly_%s", GetName()), meancrt, sigmacrt);
    if (_blog)
        _unormal->printLog();
}

void TLogNormalDistribution::setParameterMean(Double_t M, Bool_t recompute)
{
    try
    {
        if (M < _dparameterXMin * 1.000001)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setParameterMean: parameter M [%f] is too smalle: (M - xmin) [%f] must be greater than %f",
                            M, M - _dparameterXMin,
                            _dparameterXMin * 1.000001));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterMean = M;
    if (recompute)
    {
        createUnderlyingNormal(M - _dparameterXMin, _dparameterErrorFactor);
        computeTheoreticalInformation();
    }

}

void TLogNormalDistribution::setParameterErrorFactor(Double_t errorfactor,
        Bool_t recompute)
{
    try
    {
        if ((errorfactor - 1.0) < 1e-6)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setParameterErrorFactor: parameter errorfactor [%f] is too small: (errorfactor - 1.0) must be greater than 1e-6",
                            errorfactor));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterErrorFactor = errorfactor;
    if (recompute)
    {
        createUnderlyingNormal(_dparameterMean - _dparameterXMin, errorfactor);
        computeTheoreticalInformation();
    }

}

void TLogNormalDistribution::setParameterXMin(Double_t xmin, Bool_t recompute)
{
    try
    {
        if (_dparameterMean < xmin * 1.000001)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setParameterXMin: parameter xmin [%f] is too high: (M - xmin) [%f] must be greater than %f",
                            xmin, _dparameterMean - xmin, xmin * 1.000001));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterXMin = xmin;
    if (recompute)
    {
        createUnderlyingNormal(_dparameterMean - xmin, _dparameterErrorFactor);
        computeTheoreticalInformation();
    }

}

void TLogNormalDistribution::setParameters(Double_t M, Double_t errorfactor,
        Double_t xmin, Bool_t recompute)
{
    try
    {
        if (M < xmin * 1.000001)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setParameters: parameters xmin [%f] and M [%f] are incompatible: (M - xmin) [%f] must be greater than %f",
                            xmin, M, M - xmin, xmin * 1.000001));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterXMin = xmin;
    setParameterMean(M, kFALSE);
    setParameterErrorFactor(errorfactor, kFALSE);
    if (recompute)
    {
        createUnderlyingNormal(M - xmin, errorfactor);
        computeTheoreticalInformation();
    }
}

void TLogNormalDistribution::setUnderlyingNormalParameters(Double_t mu,
        Double_t sigma, Bool_t recompute)
{
    _unormal = new TNormalDistribution(
            Form("guranie_unormal_underly_%s", GetName()), mu, sigma);

    _dparameterErrorFactor = TMath::Exp(sigma * 1.645);
    _dparameterMean = TMath::Exp(mu + 0.5 * sigma * sigma) + _dparameterXMin;

    if (recompute)
        computeTheoreticalInformation();

    if (_blog)
        _unormal->printLog();
}

void TLogNormalDistribution::setLowerBound(Double_t val)
{
    //------------------------------------------------//
    // Verify that the new lower bound is a valid one //
    //------------------------------------------------//
    try
    {
        if (val >= upperBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setLowerBound: the new lower bound [%f] is greater than or equal to the current upper bound [%f]",
                            val, upperBound));
        else if (val < _dparameterXMin * 1.000001)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setLowerBound: the new lower bound [%f] is too small. lower limit is %f",
                            val, _dparameterXMin * 1.000001));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    //----------------------------------//
    // Determine the new value of _pmin // 
    //----------------------------------//

    // To achieve this, we need to reset the bounds in order to restore 
    // the full CDF.
    //Double_t old_lowerBound = lowerBound; // not really necessary
    Double_t old_upperBound = upperBound;
    double old_pmax = _pmax;
    _pmax=1.;
    //TAttribute::setLowerBound(_dparameterXMin * 1.000001);
    TAttribute::setLowerBound(-DLE_MAX); // Values smaller than _dparameterXMin are not taken into account
    _pmin=0.;
    TAttribute::setUpperBound(DLE_MAX,true);

    // Now, calculate _pmin, the probability of being smaller than the 
    // new lower bound 
    _pmin = getCDF(val);

    // Finally, restore the attribute limits with the new lower bound
    TAttribute::setLowerBound(val);
    TAttribute::setUpperBound(old_upperBound,true);
    _pmax=old_pmax;

    //----------------------------------//
    // Update the underlying normal law //
    //----------------------------------//
    _unormal->setLowerBound(TMath::Log(val - _dparameterXMin));

    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
        cout << "** _pmin  [" << _pmin << "]" << endl;

}

void TLogNormalDistribution::setUpperBound(Double_t val)
{
    //------------------------------------------------//
    // Verify that the new upper bound is a valid one //
    //------------------------------------------------//
    try
    {
        if (val <= lowerBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setUpperBound: the new upper bound [%f] is less than or equal to the current lower bound [%f]",
                            val, lowerBound));
        else if (val > DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setUpperBound: the new upper bound [%f] is too large. Infinity is %f",
                            val, DLE_MAX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    //----------------------------------//
    // Determine the new value of _pmin // 
    //----------------------------------//

    // To achieve this, we need to reset the bounds in order to restore 
    // the full CDF.
    Double_t old_lowerBound = lowerBound;
    double old_pmin = _pmin;
    _pmin=0.;
    //Double_t old_upperBound = upperBound; // not really necessary
    //TAttribute::setLowerBound(_dparameterXMin * 1.000001);
    TAttribute::setLowerBound(-DLE_MAX,true); // Values smaller than _dparameterXMin are not taken into account
    _pmax=1.;
    TAttribute::setUpperBound(DLE_MAX);

    // Now, calculate _pmax, the probability of being smaller than the 
    // new upper bound 
    _pmax = getCDF(val);

    // Finally, restore the attribute limits with the new upper bound
    TAttribute::setLowerBound(old_lowerBound,true);
    _pmin=old_pmin;
    TAttribute::setUpperBound(val);

    //----------------------------------//
    // Update the underlying normal law //
    //----------------------------------//
    _unormal->setUpperBound(TMath::Log(val - _dparameterXMin));

    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
        cout << "** _pmax  [" << _pmin << "]" << endl;

}

void TLogNormalDistribution::setBounds(Double_t lower, Double_t upper)
{
    // Verify that the new bounds are valid //
    try
    {
        if (upper <= lower)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setBounds: the new upper bound [%f] is less than or equal to the new lower bound [%f]",
                            upper, lower));
        else if (upper > DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setBounds: the new upper bound [%f] is too large. Infinity is %f",
                            upper, DLE_MAX));
        else if (lower < _dparameterXMin * 1.000001)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::setBounds: the new lower bound [%f] is too small. Lower limit is %f",
                            lower, 1e-6));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

 
    setLowerBound(lower);
    setUpperBound(upper);
    
    //----------------------------------//
    // Update the underlying normal law //
    //----------------------------------//
    _unormal->setBounds(TMath::Log(lower - _dparameterXMin),
            TMath::Log(upper - _dparameterXMin));

    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
    {
        cout << "** _pmin  [" << _pmin << "]" << endl;
        cout << "** _pmax  [" << _pmax << "]" << endl;
    }
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TLogNormalDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterMean;

    Double_t tmpVar = TMath::Exp(
            2.0 * _unormal->getParameterMu()
                    + _unormal->getParameterSigma()
                            * _unormal->getParameterSigma())
            * (TMath::Exp(
                    _unormal->getParameterSigma()
                            * _unormal->getParameterSigma()) - 1);
    
    _dtheoreticalStdDev = TMath::Sqrt(tmpVar);

    _dtheoreticalMode = _dparameterXMin
            + TMath::Exp(
                    _unormal->getParameterMu()
                            - _unormal->getParameterSigma()
                                    * _unormal->getParameterSigma());
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TLogNormalDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if (x > _dparameterXMin)
    {
        if ((x >= lowerBound) && (x <= upperBound))
        {
            Double_t dmeancrt = _unormal->getParameterMu();
            Double_t dsigmacrt = _unormal->getParameterSigma();
            Double_t X = x - _dparameterXMin;

            pdf = (1 / (TMath::Sqrt(2 * TMath::Pi()) * dsigmacrt * X))
                    * TMath::Exp(
                            (-0.5) * (TMath::Log(X) - dmeancrt)
                                    * (TMath::Log(X) - dmeancrt)
                                    / (dsigmacrt * dsigmacrt));
            // Take into account the case where at least one of the bounds is finite
            pdf /= _pmax - _pmin;
        }
    }
    return pdf;
}

Double_t TLogNormalDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if (x > _dparameterXMin)
    {
        if ((x >= lowerBound) && (x <= upperBound))
        {
            Double_t dmeancrt = _unormal->getParameterMu();
            Double_t dsigmacrt = _unormal->getParameterSigma();
            Double_t X = x - _dparameterXMin;

            cdf =
                    0.5
                            * (1
                                    + TMath::Erf(
                                            (TMath::Log(X) - dmeancrt)
                                                    / TMath::Sqrt(
                                                            2 * dsigmacrt
                                                                    * dsigmacrt)));
            // Take into account the case where at least one of the bounds is finite
            //cdf /= _pmax - _pmin;
            cdf = (cdf - _pmin) / (_pmax - _pmin);
        }
        else if (x > upperBound)
            cdf = 1.0;
    }
    return cdf;
}

Double_t TLogNormalDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogNormalDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // The case where at least one of the bounds is finite is taken into 
    // account by the underlying normal distribution. 
    Double_t x = _dparameterXMin + TMath::Exp(_unormal->getInvCDF(p));

    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TLogNormalDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TLogNormalDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] LogNormal" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mean parameter M[" << _dparameterMean << "]" << endl;
    cout << "**   Error Factor parameter [" << _dparameterErrorFactor << "]"
            << endl;
    cout << "**   Minimum value parameter xmin [" << _dparameterXMin << "]"
            << endl;
    cout << "**   Underlying Normal parameters: mu["
            << _unormal->getParameterMu() << "] sigma["
            << _unormal->getParameterSigma() << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

