//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogTriangularDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TLogTriangularDistribution

 */
/**

 * @class URANIE::DataServer::TLogTriangularDistribution
 *
 * @brief Description of the class TLogTriangularDistribution.

 This class defines a stochastic attribute following a LogTriangular distribution.

 The definition of this distribution is that if a random variable \f$ x \f$ 
 follows a LogTriangular distribution, the random variable \f$ \ln(x) \f$  
 follows a Triangular distribution.

 User parameters are the bounds of the range \f$ [x_{min},x_{max}] \f$, and the 
 mode \f$ x_{mode} \f$. 

 The constraint on these parameters is: \f$ 0 < x_{min} < x_{mode} < x_{max} \f$

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 2 * \ln(x/x_{min}) / ( x * \ln(x_{max}/x_{min}) * \ln(x_{mode}/x_{min}) ) \f$ for \f$ x \in [x_{min},x_{mode}] \f$
 \li \f$ f(x) = 2 * \ln(x_{max}/x) / ( x * \ln(x_{max}/x_{min}) * \ln(x_{max}/x_{mode}) ) \f$ for \f$ x \in [x_{mode},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$
 \li \f$ F(x) = (\ln(x/x_{min}))^2 / ( \ln(x_{max}/x_{min}) * \ln(x_{mode}/x_{min}) ) \f$ for \f$ x \in [x_{min},x_{mode}] \f$
 \li \f$ F(x) = 1 - (\ln(x_{max}/x))^2 / ( \ln(x_{max}/x_{min}) * \ln(x_{max}/x_{mode}) ) \f$ for \f$ x \in [x_{mode},x_{max}] \f$
 \li \f$ F(x) = 1 \f$ for \f$ x > x_{max} \f$ 


 \par Inverse CDF:
 \li \f$ F^{-1}(p) = \exp \left( \ln(x_{min}) + \sqrt{p * \ln(x_{max}/x_{min}) * \ln(x_{mode}/x_{min})} \right)\f$ for \f$ p \in [0,F(x_{mode})] \f$
 \li \f$ F^{-1}(p) = \exp \left( \ln(x_{max}) - \sqrt{(1-p) * \ln(x_{max}/x_{min}) * \ln(x_{max}/x_{mode})} \right)\f$ for \f$ p \in [F(x_{mode}),1] \f$

 */
#ifndef TLOGTRIANGULARDISTRIBUTION_H
#define TLOGTRIANGULARDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"
#include "TTriangularDistribution.h"

namespace URANIE
{
namespace DataServer
{
class TLogTriangularDistribution: public TStochasticAttribute
{
private:
    Double_t _dleftArea; ///< The probability that \f$ x \leq x_{mod} \f$, equivalent to \f$ F(x_{mod}) \f$ or getCDF(\f$ x_{mod} \f$)

protected:
    Double_t _dparameterMode;  ///< The mode parameter

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     The default range is [0.01,10] with mode = 5
     */
    TLogTriangularDistribution();

    //! Constructor setting the name
    /*! 
     The default range is [0.01,10] with mode = 5
     
     \param name (TString) Name of the attribute
     */
    TLogTriangularDistribution(TString name);

    //! Constructor setting name and range
    /*!                 
     The default value for the mode is (max-min)/2
     
     \param name (TString) Name of the attribute
     \param min (Double_t) Lower bound
     \param max (Double_t) Upper bound
     */
    TLogTriangularDistribution(TString name, Double_t min, Double_t max);

    //! Constructor setting name, range and mode
    /*!                 
     \param name (TString)  Name of the attribut
     \param min  (Double_t) Lower bound
     \param max  (Double_t) Upper bound
     \param mode (Double_t) Mode value
     */
    TLogTriangularDistribution(TString name, Double_t min, Double_t max,
            Double_t mode);

    //! Construction of a TUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.09)
     
     \param att  (TAttribute *) Pointer to an existing TAttribute object
     \param mode (Double_t)     Mode value
     */
    TLogTriangularDistribution(URANIE::DataServer::TAttribute *att,
            Double_t mode);

    //! Default destructor
    virtual ~TLogTriangularDistribution();
    //@}
    /** @name Law parameters
     This law has three parameters: the lower bound, the upper bound and 
     the mode. The mode parameter is defined here. The bounds parameters 
     are defined in the class URANIE::DataServer::TAttribute.
     However, as there are specific constraints that must be checked, the setting 
     functions are redefined here.
     **/
    //@{
    //! Set the lower bound
    /*!
     \param lower (Double_t) The value of the lower bound
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if lower is less than or equal to zero, 
     or greater than or equal to upperBound.
     */
    virtual void setLowerBound(Double_t lower, Bool_t recompute = kTRUE);

    //! Set the upper bound
    /*!
     \param upper (Double_t) The value of the upper bound
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if upper is less than or equal to zero, 
     or less than or equal to lowerBound.
     */
    virtual void setUpperBound(Double_t upper, Bool_t recompute = kTRUE);

    //! Set the bounds
    /*!
     \param lower (Double_t) The value of the lower bound
     \param upper (Double_t) The value of the upper bound
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if lower or upper are less than or equal 
     to zero, or lower is greater than or equal to upper.
     */
    virtual void setBounds(Double_t lower, Double_t upper, Bool_t recompute =
            kTRUE);

    //! Set the value of the mode parameter
    /*!
     \param mode (Double_t) Value of the mode parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if mode is outside of the interval ]lowerBound,upperBound[
     */
    virtual void setParameterMode(Double_t mode, Bool_t recompute = kTRUE);

    //! Returns the value of the mode parameter
    virtual Double_t getParameterMode()
    {
        return _dparameterMode;
    }

    //! Set all parameters at once
    /*!
     \param lower  (Double_t) value of the lower bound
     \param upper  (Double_t) value of the upper bound
     \param mode   (Double_t) value of the mode
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if mode is outside of the interval [lower,upper]
     */
    virtual void setParameters(Double_t lower, Double_t upper, Double_t mode,
            Bool_t recompute = kTRUE);

    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{                
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    /*!
     \warning The computation of the Standard Deviation of the 
     log-triangular distribution is currently an approximation. 
     (Nicolas Gilardi, 2010.06.10)                            
     */
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TLogTriangularDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
