//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogUniformDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TLogUniformDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TLogUniformDistribution.h"

ClassImp (URANIE::DataServer::TLogUniformDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TLogUniformDistribution::TLogUniformDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kLogUniform)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TLogUniformDistribution"
                << endl;
    computeTheoreticalInformation();
}

TLogUniformDistribution::TLogUniformDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kLogUniform)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogUniformDistribution with: Name["
                << str << "]" << endl;
    computeTheoreticalInformation();
}

TLogUniformDistribution::TLogUniformDistribution(TString str, Double_t min,
        Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kLogUniform)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogUniformDistribution with: Name["
                << str << "] Min[" << min << "] Max[" << max << "]" << endl;
    setBounds(min, max);
}

TLogUniformDistribution::TLogUniformDistribution(
        URANIE::DataServer::TAttribute *att) :
        TStochasticAttribute(att, TStochasticAttribute::kLogUniform)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TLogUniformDistribution from an existing TAttribute ["
                << att->GetName() << "]" << endl;
    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------       

TLogUniformDistribution::~TLogUniformDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TLogUniformDistribution"
                << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TLogUniformDistribution::setLowerBound(Double_t lower)
{
    try
    {
        if (lower <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setLowerBound: parameter lower [%f] is negative or null",
                            lower));
        else if (lower == upperBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setLowerBound: parameter lower [%f] is equal to upperBound [%f]",
                            lower, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    TAttribute::setLowerBound(lower);
    computeTheoreticalInformation();
}

void TLogUniformDistribution::setUpperBound(Double_t upper)
{
    try
    {
        if (upper <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setUpperBound: parameter upper [%f] is negative or null",
                            upper));
        else if (upper == lowerBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setUpperBound: parameter upper [%f] is equal to lowerBound [%f]",
                            upper, lowerBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    TAttribute::setUpperBound(upper);
    computeTheoreticalInformation();
}

void TLogUniformDistribution::setBounds(Double_t lower, Double_t upper)
{
    try
    {
        if (lower <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setBounds: lower bound of %s is negative or null [%f]",
                            fName.Data(), lower));
        else if (upper <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setBounds: upper bound of %s is negative or null [%f]",
                            fName.Data(), upper));
        else if (upper <= lower)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::setBounds: upper bound [%f] of %s is smaller or equal to lower bound [%f]",
                            upper, fName.Data(), lower));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
//             TAttribute::setLowerBound(lower);
//             TAttribute::setUpperBound(upper);   
    TAttribute::setBounds(lower, upper);
    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TLogUniformDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = (upperBound - lowerBound)
            / (TMath::Log(upperBound) - TMath::Log(lowerBound));

    Double_t valNumerator = (upperBound - lowerBound)
            * ((TMath::Log(upperBound) - TMath::Log(lowerBound))
                    * (upperBound + lowerBound)
                    - 2.0 * (upperBound - lowerBound));
    Double_t valDenominator = 2.0
            * (TMath::Log(upperBound) - TMath::Log(lowerBound))
            * (TMath::Log(upperBound) - TMath::Log(lowerBound));

    _dtheoreticalStdDev = TMath::Sqrt(valNumerator / valDenominator);

    _dtheoreticalMode = lowerBound;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TLogUniformDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x >= lowerBound) && (x <= upperBound))
        pdf = 1.0 / (x * (TMath::Log(upperBound) - TMath::Log(lowerBound)));

    return pdf;
}

Double_t TLogUniformDistribution::getCDF(Double_t x)
{
    Double_t cdf;
    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
        cdf = (TMath::Log(x) - TMath::Log(lowerBound))
                / (TMath::Log(upperBound) - TMath::Log(lowerBound));

    return cdf;
}

Double_t TLogUniformDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLogUniformDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    // In order to be coherent in the formulas, replace the base 10 
    // logarithm by a base e logarithm.
    //return  TMath::Power(10., TMath::Log10(lowerBound) + ( TMath::Log10(upperBound) - TMath::Log10(lowerBound) ) * p);
    return TMath::Exp(
            TMath::Log(lowerBound)
                    + (TMath::Log(upperBound) - TMath::Log(lowerBound)) * p);
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------        

void TLogUniformDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TLogUniformDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TLogUniformDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

