//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLogUniformDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TLogUniformDistribution
 */

/**

 * @class URANIE::DataServer::TLogUniformDistribution
 *
 * @brief Description of the class TLogUniformDistribution.

 This class defines a stochastic attribute following a LogUniform distribution.

 The definition of this distribution is that if a random variable \f$ x \f$ 
 follows a LogUniform distribution, the random variable \f$ \ln(x) \f$  
 follows a Uniform distribution.

 User parameters are the minimum value \f$ x_{min} \f$ and the maximum value \f$ x_{max} \f$ .

 The constraint on these parameters is: \f$ 0 < x_{min} < x_{max} \f$

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 1/(x * \ln(x_{max}/x_{min})) \f$ for \f$ x \in [x_{min},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0.0 \f$ for \f$ x < x_{min} \f$
 \li \f$ F(x) = \ln(x/x_{min})/\ln(x_{max}/x_{min}) \f$ for \f$ x \in [x_{min},x_{max}] \f$
 \li \f$ F(x) = 1.0 \f$ for \f$ x > x_{max} \f$

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = 10^{Log_{10}(x_{min}) + p * Log_{10}(x_{max}/x_{min})  }\f$ for \f$ p \in [0,1] \f$ (logarithm in base 10)
 \li \f$ F^{-1}(p) = \exp \left( \ln(x_{min}) + p * \ln(x_{max}/x_{min}) \right)\f$ for \f$ p \in [0,1] \f$ (logarithm in base e)

 */
#ifndef TLOGUNIFORMDISTRIBUTION_H
#define TLOGUNIFORMDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TLogUniformDistribution: public TStochasticAttribute
{
private:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default range is [0.01, 10.0]
     */
    TLogUniformDistribution();

    //! Constructor setting the name
    /*! 
     Default range is [0.01, 10.0]
     
     \param name (TString) Name of the attribute
     */
    TLogUniformDistribution(TString name);

    //! Constructor setting name and range
    /*!                 
     \param name (TString) Name of the attribute
     \param min (Double_t) Lower bound
     \param max (Double_t) Upper bound                
     */
    TLogUniformDistribution(TString name, Double_t min, Double_t max);

    //! Construction of a TLogUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)
     
     \param att (TAttribute *) Pointer to an existing TAttribute object                                
     */
    TLogUniformDistribution(URANIE::DataServer::TAttribute *att);

    //! Default destructor
    virtual ~TLogUniformDistribution();

    //@}

    /** @name Law parameters
     This law has two parameters: the lower bound, the upper bound. They 
     are both defined in the class URANIE::DataServer::TAttribute. However, 
     as there are specific constraints that must be checked, the setting 
     functions are redefined here.
     **/
    //@{
    //! Set the lower bound
    /*!
     \param lower (Double_t) The value of the lower bound
     
     \exception UErrorExceptions if lower is less than or equal to zero, 
     or greater than or equal to upperBound.
     */
    virtual void setLowerBound(Double_t lower);

    //! Set the upper bound
    /*!
     \param upper (Double_t) The value of the upper bound
     
     \exception UErrorExceptions if upper is less than or equal to zero, 
     or less than or equal to lowerBound.
     */
    virtual void setUpperBound(Double_t upper);

    //! Set the bounds
    /*!
     \param lower (Double_t) The value of the lower bound
     \param upper (Double_t) The value of the upper bound
     
     \exception UErrorExceptions if lower or upper are less than or equal 
     to zero, or lower is greater than or equal to upper.
     */
    virtual void setBounds(Double_t lower, Double_t upper);
    virtual void setParameters(Double_t lower, Double_t upper){ setBounds(lower,upper); };
    //@}            

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TLogUniformDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
