//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMultinomialDistribution.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   Mon Jan 19 12:00:00 CET 2009

 \brief Implementation of the class URANIE::DataServer::TMultinomialDistribution

 totot
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"

// URANIE
#include "UExceptions.h"
#include "TMultinomialDistribution.h"

ClassImp (URANIE::DataServer::TMultinomialDistribution)

namespace URANIE
{
namespace DataServer
{

//===============================//
// Constructors and Desctructors //
//===============================//

TMultinomialDistribution::TMultinomialDistribution()
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
    clearVectors();
}

TMultinomialDistribution::TMultinomialDistribution(TString str) :
        TDiscreteAttribute(str, TStochasticAttribute::kMultinomial)
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
}

TMultinomialDistribution::TMultinomialDistribution(TString name, Int_t nbValues,
        Double_t* values, Double_t* frequencies) :
        TDiscreteAttribute(name, TStochasticAttribute::kMultinomial)
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
    setDistribution(nbValues, values, frequencies);
}

TMultinomialDistribution::~TMultinomialDistribution()
{
    //     cout << "detruit TMultinomialDistribution" << endl;
    if (pValues)
    {
        free(pValues);
    }
    if (pFrequencies)
    {
        free(pFrequencies);
    }

}

//=====================//
// Getters and Setters //
//=====================//      

Int_t TMultinomialDistribution::getNbValues()
{
    return iNbValues;
}

Double_t* TMultinomialDistribution::getAllValues()
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::getAllValues WARNING: the list of input values is empty.");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pValues;
}

Double_t TMultinomialDistribution::getValue(Int_t i)
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::getValue ERROR: the list of input values is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultinomialDistribution::getValue ERROR: index [%d] is greater than the number of values [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pValues[i];
}

Double_t* TMultinomialDistribution::getAllFrequencies()
{
    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::getAllFrequencies WARNING: the list of frequencies is empty.");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pFrequencies;
}

Double_t TMultinomialDistribution::getFrequency(Int_t i)
{
    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::getFrequency ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultinomialDistribution::getFrequency ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pFrequencies[i];
}

Double_t TMultinomialDistribution::getProba(Int_t i)
{
    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::getProba ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultinomialDistribution::getProba ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pFrequencies[i] / _dSumFreq;
}

void TMultinomialDistribution::setDistribution(Int_t nbValues, Double_t* values,
        Double_t* frequencies)
{
    if (pValues)
    {
        free(pValues);
        free(pFrequencies);
    }

    iNbValues = nbValues;
    pValues = (Double_t*) malloc(iNbValues * sizeof(Double_t));
    pFrequencies = (Double_t*) malloc(iNbValues * sizeof(Double_t));

    for (Int_t i = 0; i < iNbValues; i++)
    {
        pValues[i] = values[i];
        if (frequencies)
        {
            try
            {
                if (frequencies[i] <= 0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "TMultinomialDistribution::setDistribution ERROR: invalide frequency: frequency %d is negative or null [%e].",
                                    i, frequencies[i]));
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
            pFrequencies[i] = frequencies[i];
        }
        else
        {
            pFrequencies[i] = 1.0 / iNbValues;
        }
        _dSumFreq += pFrequencies[i];
    }
}

void TMultinomialDistribution::setAllValues(Double_t* values)
{
    for (Int_t i = 0; i < iNbValues; i++)
    {
        pValues[i] = values[i];
    }
}

void TMultinomialDistribution::setValue(Int_t i, Double_t value)
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::setValue ERROR: the list of input values is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultinomialDistribution::setValue ERROR: index [%d] is greater than the number of values [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    pValues[i] = value;
}

void TMultinomialDistribution::setAllFrequencies(Double_t* frequencies)
{
    _dSumFreq = 0.0;
    for (Int_t i = 0; i < iNbValues; i++)
    {
        pFrequencies[i] = frequencies[i];
        _dSumFreq += pFrequencies[i];
    }
}

void TMultinomialDistribution::setFrequency(Int_t i, Double_t frequency)
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultinomialDistribution::setFrequency ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultinomialDistribution::setFrequency ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _dSumFreq -= pFrequencies[i];
    pFrequencies[i] = frequency;
    _dSumFreq += pFrequencies[i];
}

//======================//
// Satistical Functions //
//======================//

void TMultinomialDistribution::computeMean()
{
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                "TMultinomialDistribution::computeMean WARNING : Mean and Variance are meaningless for discrete distributions");
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

Double_t TMultinomialDistribution::getCDF(Double_t x)
{
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                "TMultinomialDistribution::getCDF WARNING : Cumulative Distribution Function is meaningless for discrete distributions");
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return 0.0;
}

Double_t TMultinomialDistribution::getPDF(Double_t x, Double_t precision)
{
    // If x is not in the distribution, its probability is 0.0
    Double_t p = 0.0;

    for (Int_t i = 0; i < iNbValues; i++)
    {
        if (TMath::Abs(x - pValues[i]) < precision)
        {
            p = pFrequencies[i];
            break;
        }
    }

    if (_blog)
    {
        cout << "TMultinomialDistribution::getPDF x[" << x << "]  p[" << p
                << "]" << endl;
    }

    return p / _dSumFreq;
}

Double_t TMultinomialDistribution::getPDF(Double_t x)
{
    return getPDF(x, 1e-9);
}

Double_t TMultinomialDistribution::getInvCDF(Double_t p)
{
    Double_t x;
    Double_t cumul = 0.0;

    for (Int_t i = 0; i < iNbValues; i++)
    {
        cumul += pFrequencies[i] / _dSumFreq;
        if (p <= cumul)
        {
            x = pValues[i];
            break;
        }
    }

    if (_blog)
    {
        cout << "TMultinomialDistribution::getInvCDF p[" << p << "]  x[" << x
                << "]" << endl;
    }
    return x;
}

//=====//
// Log //
//=====//

void TMultinomialDistribution::printLog(Option_t *option)
{
    //     TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TMultinomialDistribution::printLog  ******" << endl;
    cout << "**   Name    [" << GetName() << "]" << endl;
    cout << "**   label   [" << GetTitle() << "]" << endl;
    cout << "**   unite   [" << getUnity() << "]" << endl;
    cout << "**   law     [" << _lawType << "]" << endl;
    cout << "**   type    [" << dataType << "]" << endl;
    cout << "**   nValues [" << iNbValues << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

