//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMultinomialDistribution.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface of the class URANIE::DataServer::TMultinomialDistribution

 */
/**

 * @class URANIE::DataServer::TMultinomialDistribution
 *
 * @brief Description of the class TMultinomialDistribution.
 *
 * To be written by the developper.
 *
 */
#ifndef TMULTINOMIALDISTRIBUTION_H
#define TMULTINOMIALDISTRIBUTION_H

//Uranie
#include "TDiscreteAttribute.h"

namespace URANIE
{
namespace DataServer
{
class TMultinomialDistribution: public TDiscreteAttribute
{
private:
protected:
    Int_t iNbValues;
    Double_t _dSumFreq;
    Double_t* pValues;
    Double_t* pFrequencies;
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TMultinomialDistribution();

    //! Constructor by name only
    /*! 
     
     \param name (TString) Name of the attribute
     */
    TMultinomialDistribution(TString name);

    //! Constructor by name and statistics
    /*! 
     This constructor allows to define the distribution. The values 
     are considered as labels, and not numerical values. This means that 
     there is no need to sort them. The associated "frequencies" 
     are given in a separated array. They are not probabilities as 
     their sum does not need to be equal to 1. 
     However, it is assumed that the distribution is complete, i.e. 
     the discrete random variable can only take values from the one 
     given.
     \param name (TString) Name of the attribute
     \param nbValues (Int_t) Number of values taken by the discrete variable 
     \param values (Double_t*) List of values taken by the discrete variable
     \param frequencies (Double_t*) List of the frequencies of each value. 
     If NULL, the values are set to be equiprobable. Default = NULL.
     */
    TMultinomialDistribution(TString name, Int_t nbValues, Double_t* values,
            Double_t* frequencies = NULL);

    //! Default destructor
    virtual ~TMultinomialDistribution();
    //@}
    //---------------------------------------------
    /** @name Getters and Setters methods
     Getting and setting attribute values.
     **/
    //@{
    //! Get the number of values
    virtual Int_t getNbValues();

    //! Get the values as a Double_t array
    virtual Double_t* getAllValues();

    //! Get a value by index
    virtual Double_t getValue(Int_t i);

    //! Get the frequencies as a Double_t array
    virtual Double_t* getAllFrequencies();

    //! Get a frequency by index
    virtual Double_t getFrequency(Int_t i);

    //! Return the probability corresponding to the given index
    virtual Double_t getProba(Int_t i);

    //! Set the value and frequency arrays. This function erase any 
    //! previous arrays and reset the "iNbValues" parameter.
    /*!
     The values are considered as labels, and not numbers. Thus, they 
     don't need to be sorted. The associated "frequencies"      
     are given in a separated array. They are not probabilities as 
     their sum does not need to be equal to 1. 
     However, it is assumed that the distribution is complete, i.e. 
     the discrete random variable can only take values from the one 
     given.
     \param nbValues (Int_t) Number of values taken by the discrete variable 
     \param values (Double_t*) List of values taken by the discrete variable
     \param frequencies (Double_t*) List of the frequencies of each value. 
     If NULL, the values are set to be equiprobable. Default = NULL.
     */
    virtual void setDistribution(Int_t nbValues, Double_t* values,
            Double_t* frequencies = NULL);

    //! Set the values as a Double_t array. The number of the 
    //! array must match the one of the existing probability array
    virtual void setAllValues(Double_t* values);

    //! Set a value by index
    virtual void setValue(Int_t i, Double_t value);

    //! Set the frequencies as a Double_t array. The size of the 
    //! array must match iNbValues.
    virtual void setAllFrequencies(Double_t* frequencies);

    //! Set a single frequency by index. 
    virtual void setFrequency(Int_t i, Double_t frequencies);
    //@}
    /** @name Statistical methods
     The statistical methods ( pdf, cdf, quantile, ...).
     **/
    //@{
    //! Mean and Variance are meaningless for non ordered discrete 
    //! values: this function throws a UWarningException.
    virtual void computeMean();
    //! CDF is meaningless for non ordered discrete values: this 
    //! function throws a UWarningException.
    virtual Double_t getCDF(Double_t x);

    //! Returns the probability of a value.
    /*!
     If x belongs to the distribution, its probability will be equal 
     to the corresponding value in the pFrequencies array divided by the 
     sum of the pFrequencies values.
     NOTE: only values belonging to the distribution have a non-zero
     probability. 
     \param x (Double_t): the value for which the probability is computed 
     \param precision (Double_t): the precision to determine if x is 
     equal to a value of the distribution.
     */
    virtual Double_t getPDF(Double_t x, Double_t precision);
    //! Same as before, but with precision fixed to 1e-9
    virtual Double_t getPDF(Double_t x);

    //! Returns a value from the distribution.
    /*! 
     The choice of the value is based on a pseudo-cumulative 
     function where the order of the discrete values is the order 
     of appearance in the pValues array.
     \param p (Double_t): a probability value between 0 and 1
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TMultinomialDistribution, ID_DATASERVER)
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
