//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TPermutation.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2020/09/25

 \brief Implement the URANIE::DataServer::TPermutation class
 */

// ROOT
#include "TROOT.h"
#include "TMath.h"
#include "TEntryList.h"

// Uranie
#include "TPermutation.h"
#include "UExceptions.h"

#include <iostream>
using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TPermutation)
namespace URANIE
{
    namespace DataServer
    {

    //______________________________________________________________________________
    TPermutation::TPermutation() :
            TEventList("_permut_", "Default TPermutation"),
            _nSeed(0), _nMaxSize(-1), _bExact(kFALSE),
            _ntdOfExact(NULL), _pelExactMixed(NULL), _nExactIndex(0),
            _ntdOfSampled(NULL),
            _blog(kFALSE)
    {
        _sArchiveName = Form("%s.root", GetName());
        cout << " *** _sArchiveName[" << _sArchiveName << "]"<<endl;
    }

    TPermutation::TPermutation(const char *name, const char *title, Int_t initsize, Bool_t bExact) :
        TEventList(name, title, initsize),
        _nSeed(0), _nMaxSize(-1), _bExact(bExact),
        _ntdOfExact(NULL), _pelExactMixed(NULL), _nExactIndex(0),
        _ntdOfSampled(NULL),
        _blog(kFALSE)
    {
        _sArchiveName = Form("%s.root", GetName());
        if ( _blog ) cout << " *** _sArchiveName[" << _sArchiveName << "]"<<endl;

        for(Int_t i=0; i< initsize; i++)
            Enter(i);

        // Set the current values
        _vecCurrent = new Int_t[initsize];
        for(Int_t i=0; i<initsize; i++)
            _vecCurrent[i] = i+1;

        if ( _bExact ) {
            // case of Exact
            fillAllPermutations();
        } else {
            // case Of not-exact
            if ( _blog ) cout << " *** Not Exact constructor" << endl;
            if( gDirectory != gROOT )
                gDirectory=gROOT;
            _ntdOfSampled = new TTree("__permut_sampled_Event_List__", "Sampled permutations");
            _ntdOfSampled->SetDirectory(gDirectory);

            _ntdOfSampled->Branch("loc", &_nExactIndex, "loc/L");
            _ntdOfSampled->Branch("cor", _vecCurrent, Form("cor[%d]/I", initsize));
            _ntdOfSampled->SetScanField(0);
        }

}

    //______________________________________________________________________________
    TPermutation::~TPermutation()
    {
        if ( _blog )
            cout << " ** Destructor TPermutation Name[" << GetName()
                    << "] Title[" << GetTitle() << "] fN[" << fN << "]" << endl;

        if (_blog) cout << " *** case of _ntdOfExact" << endl;
        if ( _ntdOfExact != NULL ) {
            delete _ntdOfExact;
        }
        if (_blog) cout << " *** case of _pelExactMixed" << endl;
        if ( _pelExactMixed != NULL ) {
            delete _pelExactMixed;
        }

        if (_blog) cout << " *** case of _ntdOfSampled" << endl;
        if ( _ntdOfSampled != NULL ) {
            delete _ntdOfSampled;
        }

        if (_blog) cout << " *** case of _vecCurrent" << endl;
        if ( _vecCurrent != NULL ) {
            delete[] _vecCurrent;
        }

        if ( _blog )
            cout << " ** End Of Destructor TPermutation Name[" << GetName()
                    << "] Title[" << GetTitle() << "] fN[" << fN << "]" << endl;
    }

    //______________________________________________________________________________
    void TPermutation::Enter(Long64_t entry)
    {
        //          Enter element entry into the list
        if (!fList)
        {
            fList = new Long64_t[fSize];
            fList[0] = entry;
            fN = 1;
            return;
        }
        //if (entry==fList[fN-1]) return;
        if (fN >= fSize)
        {
            Int_t newsize = TMath::Max(2 * fSize, fN + fDelta);
            Resize(newsize - fSize);
        }

        fList[fN] = entry;
        ++fN;
    }


    //______________________________________________________________________________
    void TPermutation::generatePermutation(Int_t *vecindexes, Option_t *option) {
        if ( _blog ) cout << " **** function[TPermutation::" << __func__ << "] option[" << option << "] _nExactIndex[" << _nExactIndex << "/" << _ntdOfSampled->GetEntries() << "] _nMaxSize[" << _nMaxSize << "]" << endl;
        Int_t nP = GetN();
        if ( _blog ) cout << "    **** nP[" << nP << "]" << endl;
        try {
            if (_bExact) {
                if ( _nExactIndex > _nMaxSize )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TPermutation::generatePermutation ERROR :\n in the Exact method,"
                                                                    " the maximum number [%lld] of sample is reached [%lld]", _nMaxSize, _nExactIndex));

                _ntdOfExact->GetEntry(_pelExactMixed->GetEntry(_nExactIndex++));
                if ( _blog ) {
                    cout << " *** case of EXACT" << endl;
                    _ntdOfExact->Show();
                }
            }   else {
                if ( _blog ) cout << " *** case of NOT EXACT" << endl;
                _nExactIndex++;
                generateBySwapping(option);
                if ( _blog ) {
                    cout << " *** Candidate permutation [";
                    for(Int_t i=0; i<nP; i++)
                        cout << _vecCurrent[i] <<", ";
                    cout << "]"<<endl;
                }

                if ( !isPermutationAlreadyExist() ) {
                    if ( _blog )
                        cout << " *** FILLING the data for _nExactIndex[" << _nExactIndex << "]" << endl;

                    _ntdOfSampled->Fill();

                    if ( _blog )
                        _ntdOfSampled->Scan("*");
                }
            }

            for(Int_t i=0; i<nP; i++) {
                vecindexes[i] = _vecCurrent[i];
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
    }

    //______________________________________________________________________________
    void TPermutation::generateBySwapping(Option_t *option) {

        if ( _blog ) cout << " **** function[" << __func__ << "] option[" << option << "]" << endl;
        Int_t nP = GetN();
        if ( _blog ) {
            cout << "    **** nP[" << nP << "] _nExactIndex[" << _nExactIndex << "] Current[";
            for(Int_t i=0; i<nP; i++)
                cout << _vecCurrent[i];
            cout << "]" << endl;
        }

        Int_t indtmp, valtmp;
        for(Int_t i=nP-1; i>0; i--) {
            indtmp = (Int_t) gRandom->Integer(i+1);
            valtmp = _vecCurrent[indtmp];
            // cout<< "** i["<<i << "] val [" << _vecCurrent[indtmp] << "] indtmp["<<indtmp<<"] val["<<valtmp << "]";
            _vecCurrent[indtmp] = _vecCurrent[i];
            _vecCurrent[i] = valtmp;
            // cout<< "************* i["<<i << "] val [" << _vecCurrent[i] << "] indtmp["<<indtmp<<"] val["<<_vecCurrent[indtmp] << "]"<<endl;
        }
    }

    Bool_t TPermutation::isPermutationAlreadyExist() {
        if ( _blog ) cout << " **** function[TPermutation::" << __func__ << "]" << endl;
        Int_t nP = GetN();
        if ( _blog ) cout << "    **** nP[" << nP << "]" << endl;

        Int_t *vecCrt = new Int_t[nP];

        Long64_t nTheCrtExactIndex = _nExactIndex;
        if ( _blog ) cout << " Distance to (";
        for(Int_t i=0; i<nP; i++) {
            vecCrt[i] = _vecCurrent[i];
            if ( _blog ) cout << vecCrt[i];
        }
        if ( _blog ) cout << ")    nTheCrtExactIndex[" << nTheCrtExactIndex << "]" << endl;

        Long64_t nCrtEntries = _ntdOfSampled->GetEntries();
        if ( _blog ) cout << "    **** nCrtEntries[" << nCrtEntries << "/" << _nExactIndex << "]" << endl;
        if ( !nCrtEntries )
        {
            delete[] vecCrt; vecCrt = NULL;
            return kFALSE;
        }
        
        Int_t ndistance = 0;
        for(Long64_t i=0; i<nCrtEntries; i++) {
            _ntdOfSampled->GetEntry(i);
            ndistance = 0;
            for(Int_t i=0; i<nP; i++) {
                ndistance += TMath::Abs(_vecCurrent[i]-vecCrt[i]);
            }

            if ( ndistance == 0 ) {
                if ( _blog ) cout << " distance[" << ndistance << "] Before to  fill(";
                _nExactIndex = nTheCrtExactIndex;
                for(Int_t i=0; i<nP; i++) {
                    if ( _blog ) cout << "(" << _vecCurrent[i] << ")[" << vecCrt[i] << "] ";
                    _vecCurrent[i] = vecCrt[i];
                }
                if ( _blog ) cout << ")    _nExactIndex[" << _nExactIndex << "]" << endl;



                delete[] vecCrt; vecCrt = NULL;
                return kTRUE;
            }
        }

        if ( _blog ) cout << " Before to  fill(";
        _nExactIndex = nTheCrtExactIndex;
        for(Int_t i=0; i<nP; i++) {
            _vecCurrent[i] = vecCrt[i];
            if ( _blog ) cout << _vecCurrent[i];
        }
        if ( _blog ) cout << ")    _nExactIndex[" << _nExactIndex << "]" << endl;

        // Clean
        delete[] vecCrt; vecCrt = NULL;
        if ( _blog ) cout << " **** End function[" << __func__ << "]" << endl;
        return kFALSE;
    }

    //______________________________________________________________________________
    void TPermutation::fillAllPermutations() {

        cout << " **** function[TPermutation::" << __func__ << "]" << endl;
        try
        {
            const Int_t nP = GetN();
            _nMaxSize = TMath::Factorial(nP);

            if( gDirectory != gROOT )
                gDirectory=gROOT;
            _ntdOfExact = new TTree("__permut_exact_Event_List__", "List of all permutations");
            _ntdOfExact->SetDirectory(gDirectory);

            Long64_t ij = 1;

            std::vector<int> v;
            for(Int_t i=0; i< nP; i++) {
                v.push_back(i+1);
            }

            Int_t icrt=0;
            cout << " *** original vector [";
            for (int e : v) {
                cout << " (" << e << ")";
                _vecCurrent[icrt] =e;
                cout << "[" << _vecCurrent[icrt] << "]";
                icrt++;
            }
            cout << "]" << endl;

            _ntdOfExact->Branch("loc", &ij, "loc/L");
            _ntdOfExact->Branch("cor", _vecCurrent, Form("cor[%d]/I", nP));

            do {
                cout << Form("Case of ij[%lld] :: ", ij);

                icrt=0;
                for (int e : v) {
                    cout << " " << e;
                    _vecCurrent[icrt] =e;
                    cout << "[" << _vecCurrent[icrt] << "]";
                    icrt++;
                }
                cout << "]" << endl;

                // Fill the TTree
                _ntdOfExact->Fill();

                // increment the index
                ij++;
            } while (std::next_permutation(v.begin(), v.end()));

            Long64_t nsizeStored = _ntdOfExact->GetEntries();
            cout << " *** Exact nsizeStored[" << nsizeStored << "]" << endl;
            if (nsizeStored != _nMaxSize )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TPermutation::fillAllPermutations The number stored [%lld] is not equal to the computed value [%lld]", nsizeStored, _nMaxSize));

            _pelExactMixed = new TPatternsEventList("spel", "all patterns", _nMaxSize);
            for(Long64_t i=0; i<_nMaxSize; i++)
                _pelExactMixed->Enter(i) ;

            if ( nP < 7 ) _pelExactMixed->Print("all");
            _pelExactMixed->Mix();
            if ( nP < 7 ) _pelExactMixed->Print("all");


            _ntdOfExact->SetScanField(0);

            cout << " **** End Of function[TPermutation::" << __func__ << "] nP[" << nP << "] _nMaxSize[" << _nMaxSize << "]" << endl;
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    }

    //______________________________________________________________________________
    void TPermutation::setSeed(Int_t ind)
    {
        if (ind == 0)
        {
            time_t secondes;
            time(&secondes);
            _nSeed = (Int_t)(secondes);
        }
        else
        {
            _nSeed = ind;
        }
        gRandom->SetSeed(_nSeed);
    }


    //______________________________________________________________________________
    void TPermutation::Print(Option_t *option) const
    {
        //          Print contents of this list

        printf("TPermutation:%s/%s, number of entries =%d, size=%d\n",
                GetName(), GetTitle(), fN, fSize);
        cout << " *** Exact[" << ( _bExact ? "Yes": "No") << "]" << endl;
        cout << " **** _nMaxSize[" << _nMaxSize << "]" << endl;
        cout << " **** _nExactIndex[" << _nExactIndex << "]" << endl;
        if (!strstr(option, "all"))
            return;
        Int_t i;
        Int_t nbuf = 0;
        char element[10];
        char *line = new char[100];
        sprintf(line, "%5d : ", 0);
        for (i = 0; i < fN; i++)
        {
            nbuf++;
            if (nbuf > 10)
            {
                printf("%s\n", line);
                sprintf(line, "%5d : ", i);
                nbuf = 1;
            }
            sprintf(element, "%7lld ", fList[i]);
            strcat(line, element);
        }
        if (nbuf)
            printf("%s\n", line);
        delete[] line;
        line = NULL;
    }

    } // End Of namespace DataServer
} // End Of namespace URANIE


// End Of File
