//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPossibilityAttribute.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2010-07-07 15:52:43

 \brief  Class of attributes following a probability density function


 */
#ifndef TPOSSIBITYATTRIBUTE_H
#define TPOSSIBITYATTRIBUTE_H

#include "TString.h"
#include <iostream>

// Uranie
#include "DataServer.h"
#include "TAttribute.h"
#include "UExceptions.h"

namespace URANIE
{
namespace DataServer
{

/*! \class URANIE::DataServer::TPossibilityAttribute
 This is the parent class for all attributes which values can be generated by a 
 possibilty (cf RaFu IRSN). Child objects are possibility attributes following a specific possibility.
 
 Each must be redefined in the daughter classes.
 
 */
class TPossibilityAttribute: public TAttribute
{
public:

    //! The list of possibility that a TAttribute can follow
    enum EPossibilityType
    {
        kUniform, kTriangular, kTrapezium, kPolygonal, kUnknown
    };
    //! The list of monotonous relationship  
    enum EMonotonousRelationship
    {
        kIncreasing, kDecreasing, kNotDefined
    };

    EPossibilityType _PossibilityType; ///< The possibility type EPossibilityType
    EMonotonousRelationship _MonotonousRelationship; ///< The relationship of monotonous 
public:
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    //! Constructor setting default possibility by the name 
    /*!
     The default is possibility \b "kUnknown" and the relationshio of monotonous is \b "kNotDefined"
     \param str (TString)  Name of the attribute
     */
    TPossibilityAttribute(TString str);

    //! Constructor setting the name and the law type 
    /*!
     \param str (TString)    Name of the attribute
     \param value (EPossibilityType) The type of the probability law of the attribute.
     \param monotonous (EMonotonousRelationship) The relationship of monotonous
     The available types are defined in the EPossibilityType 
     enumerator
     */
    TPossibilityAttribute(TString str, EPossibilityType value,
            EMonotonousRelationship monotonous);

    //! Constructor setting the name, the law type, and the range (lower and upper bounds)
    /*!
     \param str (TString)    Name of the attribute
     \param value (EPossibilityType) The type of the probability law of the attribute.
     \param monotonous (EMonotonousRelationship) The relationship of monotonous
     \param min (Double_t)   Lower bound of the attribute's range
     \param max (Double_t)   Upper bound of the attribute's range
     */
    TPossibilityAttribute(TString str, EPossibilityType value, Double_t min,
            Double_t max, EMonotonousRelationship monotonous);

    //! Default Destructor
    virtual ~TPossibilityAttribute();

    //---------------------------------------------
    /** @name Treat the relationship of monotonous
     **/
    //@{
    //! Get the relationship of monotonous
    int getMonotonousRelationship()
    {
        return _PossibilityType;
    }
    //! is the relationship of monotonous of the attribute is \em Increasing
    Bool_t isIncreasing()
    {
        return (_MonotonousRelationship == kIncreasing);
    }
    //! is the relationship of monotonous of the attribute is \em Decreasing
    Bool_t isDecreasing()
    {
        return (_MonotonousRelationship == kDecreasing);
    }
    //@}

    //---------------------------------------------
    /** @name Possibility Representation
     **/
    //@{
    //! Returns the law type as an integer
    int getLawType()
    {
        return _PossibilityType;
    }

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    virtual void printLog();
    //@}

    ClassDef(TPossibilityAttribute, ID_DATASERVER)
    //Definition of a stochastic variable
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
