//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TStochAttribut.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $date$

 \brief  Implementation of the class URANIE::DataServer::TStochasticAttribute

 Created: Fri Oct  1 13:37:23 2004

 */

#include <iostream>

#include "TStochAttribut.h"

ClassImp (URANIE::DataServer::TStochasticAttribute)
ClassImp (URANIE::DataServer::TInfiniteDistribution)

namespace URANIE
{
namespace DataServer
{
// ========================================================================== //     
//                             TStochasticAttribute                           //
// ========================================================================== //

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TStochasticAttribute::TStochasticAttribute() :
    _lawType(kUnknown), _dtheoreticalMean(THEO_DEF), _dtheoreticalStdDev(THEO_DEF), _dtheoreticalMode(THEO_DEF),
    _pmin(0.0), _pmax(1.0), _nStochasticBasis(kUnknownBasis)
{
  clearVectors();
}

TStochasticAttribute::TStochasticAttribute(TString str) :
    TAttribute(str), _lawType(kUnknown), _dtheoreticalMean(THEO_DEF), _dtheoreticalStdDev(THEO_DEF), _dtheoreticalMode(THEO_DEF),
    _pmin(0.0), _pmax(1.0), _nStochasticBasis(kUnknownBasis)
{
}

TStochasticAttribute::TStochasticAttribute(TString str, ELawType value) :
    TAttribute(str), _lawType(value), _dtheoreticalMean(THEO_DEF), _dtheoreticalStdDev(THEO_DEF), _dtheoreticalMode(THEO_DEF),
    _pmin(0.0), _pmax(1.0)
{
    setDefaultStochasticBasis();

    switch (_lawType)
    {
    case TStochasticAttribute::kUniform:
    {
        setLowerBound(0.);
        setUpperBound(1.0);
        break;
    }
    case TStochasticAttribute::kLogUniform:
    {
        setLowerBound(0.01);
        setUpperBound(10.0);
        break;
    }
    case TStochasticAttribute::kLogNormal:
    {
        setLowerBound(0.00);
        break;
    }
    case TStochasticAttribute::kBeta:
    {
        setLowerBound(0.);
        setUpperBound(1.0);
        break;
    }
    default: break;
    }
}

TStochasticAttribute::TStochasticAttribute(TString str, ELawType value,
        Double_t min, Double_t max) :
    TAttribute(str, min, max), _lawType(value), _dtheoreticalMean(THEO_DEF), _dtheoreticalStdDev(THEO_DEF), _dtheoreticalMode(THEO_DEF),
    _pmin(0.0), _pmax(1.0)
{
    setDefaultStochasticBasis();
}

TStochasticAttribute::TStochasticAttribute(TAttribute *object, ELawType value) :
        _dtheoreticalMean(THEO_DEF), _dtheoreticalStdDev(THEO_DEF), _dtheoreticalMode(THEO_DEF), _pmin(0.0), _pmax(1.0)
{
    // ROOT inherited information
    SetName(object->GetName());
    SetTitle(object->GetTitle());

    // Information about the attribute
    _sunity = object->getUnity();
    _snote = object->getNote();
    _nType = object->getDataType();
    _lawType = value;
    setDefaultStochasticBasis();

    // User defined characteristics
    lowerBound = object->getLowerBound();
    _bHaveLowerBound = object->isLowerBounded();
    upperBound = object->getUpperBound();
    _bHaveUpperBound = object->isUpperBounded();
    try
    {
        switch (_lawType)
        {
        case TStochasticAttribute::kLogUniform:
        {
            if (lowerBound <= 0.0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TStochasticAttribute::TStochasticAttribute: LogUniform attribute [%s]  parameter lower bound [%f] is negative or null",
                                object->GetName(), lowerBound));
        }
	break;
        case TStochasticAttribute::kLogNormal:
        {
            if (_bHaveLowerBound && (lowerBound < 0.0))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TStochasticAttribute::TStochasticAttribute: LogNormal attribute [%s]  parameter lower bound [%f] is negative or null",
                                object->GetName(), lowerBound));
        }
	break;
	default: break;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    //_defaultValue      = object->getDefaultValue();
    //_bHaveDefaultValue = object->getHaveDefaultValue();
    //_stepValue         = object->getStepValue();
    //_bHaveStepValue    = object->getHaveStepValue();

    // Statistical information information calculated on data
    clearVectors();
    // for(int iel=0; iel<object->getMinimumSize(); iel++) _vminimum.push_back( object->getMinimum(iel) );
    // for(int iel=0; iel<object->getMaximumSize(); iel++) _vmaximum.push_back( object->getMaximum(iel) );
    // for(int iel=0; iel<object->getMeanSize(); iel++) _vmean.push_back( object->getMean(iel) );
    // for(int iel=0; iel<object->getStdSize(); iel++) _vstd.push_back( object->getStd(iel) );

}

TStochasticAttribute::~TStochasticAttribute()
{
    if (_blog)
    {
        cout << "Destructeur::TStochasticAttribute[" << this << "] sname["
                << GetName() << "]" << endl;
    }

}
// -----------------------------------------------------------------------
//  Getter and setter methods
// -----------------------------------------------------------------------

void TStochasticAttribute::setLawType(TString str)
{

    str.ToLower();
    if (str == "uniform")
        _lawType = kUniform;
    else if (str == "loguniform")
        _lawType = kLogUniform;
    else if (str == "normal")
        _lawType = kNormal;
    else if (str == "lognormal")
        _lawType = kLogNormal;
    else if (str == "trapezium")
        _lawType = kTrapezium;
    else if (str == "triangular")
        _lawType = kTriangular;
    else if (str == "logtriangular")
        _lawType = kLogTriangular;
    else if (str == "exponential")
        _lawType = kExponential;
    else if (str == "beta")
        _lawType = kBeta;
    else if (str == "cauchy")
        _lawType = kCauchy;
    else if (str == "uniformbyparts")
        _lawType = kUniformByParts;
    else if (str == "weibull")
        _lawType = kWeibull;
    else if (str == "gumbel")
        _lawType = kGumbel;
    else if (str == "multinomial")
        _lawType = kMultinomial;
    else if (str == "generalizednormal")
        _lawType = kGeneralizedNormal;
    else if (str == "generalizednormalv2")
        _lawType = kGeneralizedNormalV2;
    else
        _lawType = kUnknown;
}

void TStochasticAttribute::setDefaultStochasticBasis()
{
    switch (_lawType)
    {
    case TStochasticAttribute::kUniform:
    {
        setStochasticBasis(TStochasticAttribute::kLegendre);
        break;
    }
    case TStochasticAttribute::kLogUniform:
    {
        setStochasticBasis(TStochasticAttribute::kLegendre);
        break;
    }
    case TStochasticAttribute::kNormal:
    {
        setStochasticBasis(TStochasticAttribute::kHermite);
        break;
    }
    case TStochasticAttribute::kLogNormal:
    {
        setStochasticBasis(TStochasticAttribute::kHermite);
        break;
    }
    case TStochasticAttribute::kExponential:
    {
        setStochasticBasis(TStochasticAttribute::kLaguerre);
        break;
    }
    case TStochasticAttribute::kBeta:
    {
        setStochasticBasis(TStochasticAttribute::kJacobi);
        break;
    }
    default:
    {
        setStochasticBasis(TStochasticAttribute::kUnknownBasis);
        break;
    }
    }
}

Double_t TStochasticAttribute::getCDF(Double_t x)
{
    cout << "\n ** ERROR\n ** TStochasticAttribute::getCDF " << GetName()
            << " x[" << x << "]\n ** Function not yet implemented\n" << endl;
    return 0.0;
}

//! Returns the theoretical value of the mean
Double_t TStochasticAttribute::getTheoreticalMean()
{
    computeTheoreticalInformation();
    try{
        if(_dtheoreticalMean == THEO_DEF)
        {
            TString classname = (TString)this->IsA()->GetName();
            classname.ReplaceAll("URANIE::DataServer::","");
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("%s::getTheoreticalMean WARNING: the mean is not defined for this law [%s].",
                                                              classname.Data(), GetName()));
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _dtheoreticalMean;
}


//! Returns the theoretical value of the standard deviation
Double_t TStochasticAttribute::getTheoreticalStdDev()
{
    computeTheoreticalInformation();
    try{
        if(_dtheoreticalStdDev == THEO_DEF)
        {
            TString classname = (TString)this->IsA()->GetName();
            classname.ReplaceAll("URANIE::DataServer::","");
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("%s::getTheoreticalMean WARNING: the standard deviation is not defined for this law [%s].",
                                                              classname.Data(), GetName()));
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _dtheoreticalStdDev;
}

//! Returns the theoretical value of the mode
Double_t TStochasticAttribute::getTheoreticalMode()
{
    computeTheoreticalInformation();
    try{
        if(_dtheoreticalMode == THEO_DEF)
        {
            TString classname = (TString)this->IsA()->GetName();
            classname.ReplaceAll("URANIE::DataServer::","");
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("%s::getTheoreticalMean WARNING: the mode is not defined for this law [%s].",
                                                              classname.Data(), GetName()));
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _dtheoreticalMode;
}


void TStochasticAttribute::printLog()
{
    TAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TStochasticAttribute::printLog  ******" << endl;
    cout << "**   law [" << _lawType << "]" << endl;

    switch (_nStochasticBasis)
    {
    case kUnknownBasis:
        cout << "**   StochasticBasis[kUnknownBasis]" << endl;
        break;
    case kHermite:
        cout << "**   StochasticBasis[kHermite]" << endl;
        break;
    case kLegendre:
        cout << "**   StochasticBasis[kLegendre]" << endl;
        break;
    case kLaguerre:
        cout << "**   StochasticBasis[kLaguerre]" << endl;
        break;
    case kJacobi:
        cout << "**   StochasticBasis[kJacobi]" << endl;
        break;
    }
    cout << "*******************************" << endl;

}

// ========================================================================== //     
//                            TInfiniteDistribution                           //
// ========================================================================== //

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TInfiniteDistribution::TInfiniteDistribution()
{
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
    clearVectors();  
}

TInfiniteDistribution::TInfiniteDistribution(TString str) :
        TStochasticAttribute(str)
{
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
}

TInfiniteDistribution::TInfiniteDistribution(TString str, ELawType value) :
        TStochasticAttribute(str, value)
{
    lowerBound = -1.0 * DLE_MAX;
    upperBound = DLE_MAX;
}

TInfiniteDistribution::TInfiniteDistribution(TAttribute *object, ELawType value) :
        TStochasticAttribute(object, value)
{
}

TInfiniteDistribution::~TInfiniteDistribution()
{

}

void TInfiniteDistribution::setLowerBound(Double_t val)
{
    //------------------------------------------------//
    // Verify that the new lower bound is a valid one //
    //------------------------------------------------//
    try
    {
        if (val >= upperBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setLowerBound: the new lower bound [%f] is greater than or equal to the current upper bound [%f]",
                            val, upperBound));
        else if (val < -1.0 * DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setLowerBound: the new lower bound [%f] is too small. Lower infinity is %f",
                            val, -1.0 * DLE_MAX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    //----------------------------------//
    // Determine the new value of _pmin // 
    //----------------------------------//

    // To achieve this, we need to reset the bounds in order to restore 
    // the full CDF.
    //Double_t old_lowerBound = lowerBound; // not really necessary
    Double_t old_upperBound = upperBound;
    double old_pmax = _pmax;
    _pmax=1.;
    TAttribute::setLowerBound(-1.0 * DLE_MAX);
    _pmin=0.;
    TAttribute::setUpperBound(DLE_MAX,true);

    // Now, calculate _pmin, the probability of being smaller than the 
    // new lower bound 
    _pmin = getCDF(val);

    // Finally, restore the attribute limits with the new lower bound
    TAttribute::setLowerBound(val);
    TAttribute::setUpperBound(old_upperBound,true);
    _pmax=old_pmax;

    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
        cout << "** _pmin  [" << _pmin << "]" << endl;
}

void TInfiniteDistribution::setUpperBound(Double_t val)
{
    //------------------------------------------------//
    // Verify that the new upper bound is a valid one //
    //------------------------------------------------//
    try
    {
        if (val <= lowerBound)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setUpperBound: the new upper bound [%f] is less than or equal to the current lower bound [%f]",
                            val, lowerBound));
        else if (val > DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setUpperBound: the new upper bound [%f] is too large. Infinity is %f",
                            val, DLE_MAX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    //----------------------------------//
    // Determine the new value of _pmax // 
    //----------------------------------//

    // To achieve this, we need to reset the bounds in order to restore 
    // the full CDF.
    Double_t old_lowerBound = lowerBound;
    double old_pmin = _pmin;
    _pmin=0.;
    //Double_t old_upperBound = upperBound; // not really necessary
    TAttribute::setLowerBound(-1.0 * DLE_MAX,true);
    _pmax=1.;
    TAttribute::setUpperBound(DLE_MAX);

    // Now, calculate _pmax, the probability of being smaller than the 
    // new upper bound 
    _pmax = getCDF(val);

    // Finally, restore the attribute limits with the new upper bound
    TAttribute::setLowerBound(old_lowerBound,true);
    _pmin=old_pmin;
    TAttribute::setUpperBound(val);
    
    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
        cout << "** _pmax  [" << _pmax << "]" << endl;

}

void TInfiniteDistribution::setBounds(Double_t lower, Double_t upper)
{
    // Verify that the new bounds are valid //
    try
    {
        if (upper <= lower)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setBounds: the new upper bound [%f] is less than or equal to the new lower bound [%f]",
                            upper, lower));
        else if (upper > DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setBounds: the new upper bound [%f] is too large. Infinity is %f",
                            upper, DLE_MAX));
        else if (lower < -1.0 * DLE_MAX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TInfiniteDistribution::setBounds: the new lower bound [%f] is too small. Lower infinity is %f",
                            lower, -1.0 * DLE_MAX));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
 
    setLowerBound(lower);
    setUpperBound(upper);
    
    //-----------------------------------------//
    // Recalculate the theoretical information //
    //-----------------------------------------//
    computeTheoreticalInformation();

    //----------------//
    // Update the log //
    //----------------//
    if (_blog)
    {
        cout << "** _pmin  [" << _pmin << "]" << endl;
        cout << "** _pmax  [" << _pmax << "]" << endl;
    }
}

void TInfiniteDistribution::printLog()
{
    TStochasticAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TInfiniteDistribution::printLog  ******" << endl;
    cout << "**   _pmin[" << getPMin() << "]" << endl;
    cout << "**   _pmax[" << getPMax() << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE
