//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TStudentDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TStudentDistribution

 */
/**

 * @class URANIE::DataServer::TStudentDistribution
 *
 * @brief Description of the class TStudentDistribution.

 This class defines a stochastic attribute following an Student distribution.

 This distribution has one parameters: the dof parameter \f$ k \f$ 

 \par Probability Density Function (PDF):
 \li \f$ f(x) =  \exp\left( -( x - k)^2 / 2\sigma^2 \right) / \sqrt{2\pi\sigma^2}\f$


 \note In order to reduce numerical errors, the algorithm used to compute \f$ F^{-1}(p) \f$ 
 is actually either Peter John Acklam's algorithm (http://home.online.no/~pjacklam/notes/invnorm) 
 or the ROOT implementation of AS241 algorithm (Michael J. Wichura, Applied Statistics, vol. 37, 1988, pp. 477-484). 
 The latter is used by default.


 */
#ifndef TSTUDENTDISTRIBUTION_H
#define TSTUDENTDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TStudentDistribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterK;    ///< Dof parameter \f$ k \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default dof is 0.0, default standard deviation is 1.0 
     */
    TStudentDistribution();

    //! Constructor setting the name
    /*!
     Default dof is 0.0, default standard deviation is 1.0 

     \param name (TString) Name of the attribute
     */
    TStudentDistribution(TString name);

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param k (Double_t) Dof parameter \f$ k \f$
     */
    TStudentDistribution(TString name, Double_t k);

    //! Construction of a TStudentDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param k (Double_t) Dof parameter \f$ k \f$
     */
    TStudentDistribution(URANIE::DataServer::TAttribute *att, Double_t k);

    //! Default destructor
    virtual ~TStudentDistribution();
    //@}

    /** @name Law parameters
     This law has two parameters: the dof \f$ k \f$ and the standard deviation 
     \f$ \sigma \f$. Both are defined here. 
     **/
    //@{            
    //! Set the dof parameter
    /*!
     \param k (Double_t) Dof parameter \f$ k \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)                            
     */
    virtual void setParameterK(Double_t k, Bool_t recompute = kTRUE);

    //! Return the rate parameter
    Double_t getParameterK()
    {
        return _dparameterK;
    }

    //! Set all the parameters
    /*!
     \param k (Double_t) Dof parameter \f$ k \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t k, Bool_t recompute =kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     The ROOT implementation of AS241 algorithm is used by default.
     
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    /** @name Deprecated methods
     **/
    //@{
    //!  
    /*!
     \deprecated This method has been replaced by computeTheoreticalInformation
     */
    virtual void computeMean()
    {
    }
    //@}

    ClassDef(URANIE::DataServer::TStudentDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
