//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTemporalTree.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2014/11/27

 \brief Interface of the class URANIE::DataServer::TTemporalTree

*/
#ifndef TTEMPORALTREE_H
#define TTEMPORALTREE_H

// ROOT
#include "TROOT.h"
#include "TTree.h"
#include "TNamed.h"

// Uranie
#include "DataServer.h"
#include "TDataServer.h"

namespace URANIE
{
namespace DataServer
{
class TTemporalTree : public TNamed {

private:
    TString        _sIter;              // The iterator attribute
    TString        _sTime;              // The time attribute    (nT \eq 1)
    TString        _sY;                 // The output attribute  (nY \eq 1)
    TString         _sXs;               // The vector of X attributes (nX \leq 1)
    TList * _lstOfYTAttributes;         // The list of Output (Y) and Time(T) attributes
    TList * _lstOfXAttributes;          // The list of X attributes
    TString         _sAdditional;       // The additional attribute to store
    Bool_t          _bHaveAdditionnal;  // Boolean if have additional attribute
    Bool_t          _bHaveXs;           // Boolean if have nX attributes
    Bool_t _bdrawProgressBar;           // Boolean to know if the progress bar has to be drawn

    TTree *        _tTree;              //-> TTree contain vector of time values and of sY values
    TTree *        _tStatTree;          //-> TTree contain vector of time values and of sY values
    TDataServer *  _tds;                // Pointer to a TDS
    Int_t          _nNMaxTime;          // The maximun number of time
    Int_t          _nNMinTime;          // The minimum number of time
    Bool_t         _blog;               // Log printing.
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*!
     * \brief Default empty constructor
     */
    TTemporalTree();
    /*!
     * \brief Build a TTree object with one only temporal Y attribute
     * \param tds The TDataServer
     * \param stime (TString) The name of the time attribute (only one attribute)
     * \param sy (TString) The name of the Y Attribute (only one attribute)
     * \param sXs (TString) The list of inputs attributes to store. The attributes are separated by the ':' character
     */
    TTemporalTree(TDataServer *tds, TString stime, TString sy, TString sXs = "");
    /*!
    * \brief virtaul destructor
    */
   virtual ~TTemporalTree();
    //@}

    //---------------------------------------------
    /** @name User methods
     **/
    //@{
    /*!
    * \brief Build the data and statistics TTrees.
    * \param option Option to use (empty)
    * \return (Bool_t) return kTRUE if OK
    */
   Bool_t buildTree(Option_t *option = "");
   /*!
    * \brief Make two graphics with the data and statistics
    * \param option
    */
   void draw(Option_t *option = "");
   //@}

   //---------------------------------------------
   /** @name Setting and Getting attributes
    **/
   //@{
   /*!
    * \brief Set the iterator attributes if different than tds->getIteratorName
    * \param TString siter the name of the attribute to use as the iterator
    */
   void setIterator(TString siter);
   /*!
    * \brief Return the name of the attribute used as the iterator
    * \return the name of the iterator attribute
    */
   TString getIterator() { return _sIter;}
   /*!
    * \brief Get the number of minimum of times.
    * \return The minimum size of number of times
    */
   Int_t getNMinTime() { return _nNMinTime;}
   /*!
    * \brief Get the number of mximum of times.
    * \return The maximum size of number of times
    */
   Int_t getNMaxTime() { return _nNMaxTime;}
   /*!
    * \brief Get the pointer to the TTree object
    * \return the pointer to the TTree object
    */
   TTree * getTTree();
   /*!
    * \brief Build a TTree with statistics of the Y attribute versus times.
    * The statistic are :
    *\li Mean
    *\li Standard-Deviation
    *\li Median
    * \return The TTree of statistics
    */
   TTree * getStatTree();
   /*!
    * \brief setAdditinalAttribute
    * \param sname TString the name of the additional attribute to store in the TTemporalAttribute
    */
   void setAdditinalAttribute(TString sname);
   TString getAdditinalAttribute() { return _sAdditional; }
   //@}

   //---------------------------------------------
   /** @name Progress Bar
    **/
   //@{
   //! Set the "draw progress bar" flag
   /*!
    The progress bar will be drawn if and only if the "draw progress bar"
    flag is true.
    */
   void setDrawProgressBar(Bool_t bbool = kTRUE)
   {
       _bdrawProgressBar = bbool;
   }

   //! Get the "draw progress bar" flag
   /*!
    \return (Bool_t) the "draw progress bar" flag (boolean)
    */
   Bool_t getDrawProgressBar()
   {
       return _bdrawProgressBar;
   }
   //@}

   //---------------------------------------------
   /** @name Printing Log
    **/
   //@{
   void setLog()
   {
       _blog = kTRUE;
   }
   void unsetLog()
   {
       _blog = kFALSE;
   }
   void changeLog()
   {
       _blog = _blog ? kFALSE : kTRUE;
   }
   Bool_t getLog()
   {
       return _blog;
   }
   virtual void printLog();
   //@}

   ClassDef(URANIE::DataServer::TTemporalTree, ID_DATASERVER)  //TTemporalAttribute structure
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif  // TTEMPORALTREE_H
