//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTrapeziumDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TTrapeziumDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TTrapeziumDistribution.h"

ClassImp (URANIE::DataServer::TTrapeziumDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TTrapeziumDistribution::TTrapeziumDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kTrapezium, 0.0, 1.0)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TTrapeziumDistribution"
                << endl;
    _dparameterXlow = 0.25;
    _dparameterXup = 0.75;
    computeTheoreticalInformation();
}

TTrapeziumDistribution::TTrapeziumDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kTrapezium, 0.0, 1.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTrapeziumDistribution with: Name["
                << str << "]" << endl;
    _dparameterXlow = 0.25;
    _dparameterXup = 0.75;
    computeTheoreticalInformation();
}

TTrapeziumDistribution::TTrapeziumDistribution(TString str, Double_t xmin,
        Double_t xmax) :
        TStochasticAttribute(str, TStochasticAttribute::kTrapezium, xmin, xmax)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTrapeziumDistribution with: Name["
                << str << "] Min[" << xmin << "] Max[" << xmax << "]" << endl;
    _dparameterXlow = 0.25 * (xmax - xmin);
    _dparameterXup = 0.75 * (xmax - xmin);
    computeTheoreticalInformation();
}

TTrapeziumDistribution::TTrapeziumDistribution(TString str, Double_t xmin,
        Double_t xmax, Double_t xlow, Double_t xup) :
        TStochasticAttribute(str, TStochasticAttribute::kTrapezium, xmin, xmax)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTrapeziumDistribution with: Name["
                << str << "] Min[" << xmin << "] Max[" << xmax << "] Xlower["
                << xlow << "] Xupper[" << xup << "]" << endl;

    try
    {
        if ((xup < xlow) || (xup > xmax))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution: parameter xup [%f] is out of bounds [%f,%f]",
                            xup, xlow, xmax));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    try
    {
        if ((xlow < xmin) || (xlow > xup))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution: parameter xlow [%f] is out of bounds [%f,%f]",
                            xlow, xmin, xup));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setParameterXlower(xlow);
    setParameterXupper(xup);
}

TTrapeziumDistribution::TTrapeziumDistribution(
        URANIE::DataServer::TAttribute *att, Double_t xlow, Double_t xup) :
        TStochasticAttribute(att, TStochasticAttribute::kTrapezium)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTrapeziumDistribution from existing TAttribute ["
                << att->GetName() << "] and Xlower[" << xlow << "] Xupper["
                << xup << "]" << endl;

    try
    {
        if (xup < xlow)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution: parameter xup [%f] is smaller than parameter xlow [%f]",
                            xup, xlow));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setParameterXlower(xlow);
    setParameterXupper(xup);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TTrapeziumDistribution::~TTrapeziumDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TTrapeziumDistribution "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TTrapeziumDistribution::setParameterXlower(Double_t xlow)
{
    // Verify that xlow is between lowerBound and upperBound
    try
    {
        if ((xlow < lowerBound) || (xlow > upperBound))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution::setParameterXlower: parameter Xlower [%f] is out of bounds [%f,%f]",
                            xlow, lowerBound, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterXlow = xlow;
    computeTheoreticalInformation();
}

void TTrapeziumDistribution::setParameterXupper(Double_t xup)
{
    // Verify that xup is between lowerBound and upperBound
    try
    {
        if ((xup < lowerBound) || (xup > upperBound))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution::setParameterXupper: parameter Xupper [%f] is out of bounds [%f,%f]",
                            xup, lowerBound, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterXup = xup;
    computeTheoreticalInformation();
}

void TTrapeziumDistribution::setParameters(Double_t xmin, Double_t xmax,
        Double_t xlow, Double_t xup)
{
    // Verify that mode is between lower and upper
    try
    {
        if ((xup < xlow) || (xup > xmax))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution::setParameters: parameter xup [%f] is out of bounds [%f,%f]",
                            xup, xlow, xmax));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    try
    {
        if ((xlow < xmin) || (xlow > xup))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution::setParameters: parameter xlow [%f] is out of bounds [%f,%f]",
                            xlow, xmin, xup));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setBounds(xmin, xmax);
    _dparameterXlow = xlow;
    _dparameterXup = xup;

    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TTrapeziumDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMode = (_dparameterXup - _dparameterXlow) / 2.0;

    _dtrapeziumHeight = 2.0
            / (_dparameterXup - _dparameterXlow + upperBound - lowerBound);


    _dtheoreticalMean = _dtrapeziumHeight / 6;
    _dtheoreticalMean *= ( (upperBound*upperBound*upperBound - _dparameterXup*_dparameterXup*_dparameterXup) / (upperBound - _dparameterXup)
                           - (_dparameterXlow*_dparameterXlow*_dparameterXlow - lowerBound*lowerBound*lowerBound) / (_dparameterXlow - lowerBound) );
    
    double tmp = _dtrapeziumHeight / 12;
    tmp *= ( (upperBound*upperBound*upperBound*upperBound - _dparameterXup*_dparameterXup*_dparameterXup*_dparameterXup) / (upperBound - _dparameterXup)
             - (_dparameterXlow*_dparameterXlow*_dparameterXlow*_dparameterXlow - lowerBound*lowerBound*lowerBound*lowerBound) / (_dparameterXlow - lowerBound) );
    
    _dtheoreticalStdDev = sqrt(tmp);
    
    _dlowerTriangleArea = (_dparameterXlow - lowerBound) * _dtrapeziumHeight
            / 2.0;

    _dupperTriangleArea = (upperBound - _dparameterXup) * _dtrapeziumHeight
            / 2.0;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TTrapeziumDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x > lowerBound) && (x < upperBound))
    {
        if (x < _dparameterXlow)
            pdf = _dtrapeziumHeight * (x - lowerBound)
                    / (_dparameterXlow - lowerBound);
        else if ((x >= _dparameterXlow) && (x <= _dparameterXup))
            pdf = _dtrapeziumHeight;
        else
            pdf = _dtrapeziumHeight * (upperBound - x)
                    / (upperBound - _dparameterXup);
    }
    return pdf;
}

Double_t TTrapeziumDistribution::getCDF(Double_t x)
{
    Double_t cdf;

    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
    {
        if (x < _dparameterXlow)
            cdf = _dtrapeziumHeight * (x - lowerBound) * (x - lowerBound)
                    / (2 * (_dparameterXlow - lowerBound));
        else if ((x >= _dparameterXlow) && (x <= _dparameterXup))
            cdf = _dtrapeziumHeight * (x - _dparameterXlow)
                    + _dlowerTriangleArea;
        else
            cdf = 1.0
                    - _dtrapeziumHeight * (upperBound - x) * (upperBound - x)
                            / (2 * (upperBound - _dparameterXup));
    }
    return cdf;
}

Double_t TTrapeziumDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTrapeziumDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    Double_t x;
    if (p < _dlowerTriangleArea)
        x = lowerBound
                + TMath::Sqrt(
                        p * 2 * (_dparameterXlow - lowerBound)
                                / _dtrapeziumHeight);
    else if ((p >= _dlowerTriangleArea) && (p <= 1 - _dupperTriangleArea))
        x = p / _dtrapeziumHeight + (_dparameterXlow + lowerBound) / 2.0;
    else
        x = upperBound
                - TMath::Sqrt(
                        (1.0 - p) * 2 * (upperBound - _dparameterXup)
                                / _dtrapeziumHeight);
    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------      

void TTrapeziumDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TTrapeziumDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TTrapeziumDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Small base limits[" << _dparameterXlow << ", "
            << _dparameterXup << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

