//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTrapeziumDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TTrapeziumDistribution

 */
/**

 * @class URANIE::DataServer::TTrapeziumDistribution
 *
 * @brief Description of the class TTrapeziumDistribution.

 This class defines a stochastic attribute following a Trapezium distribution.

 The form of the density function is a trapezium of large base \f$ [x_{min}, x_{max}] \f$, 
 small base \f$ [x_{low}, x_{up} \f$. By definition, the area of the 
 trapezium is equal to 1.

 In the following expressions, \f$ H = 2 / ( (x_{up}-x_{low}) + (x_{max}-x_{min}) ) \f$ 
 \par Probability Density Function (PDF):
 \li \f$ f(x) = H * (x - x_{min}) / (x_{low} - x_{min})\f$ for \f$ x \in [x_{min},x_{low}] \f$
 \li \f$ f(x) = H \f$ for \f$ x \in [x_{low},x_{up}] \f$
 \li \f$ f(x) = H * (x_{max} - x) / (x_{max} - x_{up}) \f$ for \f$ x \in [x_{up},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$ 
 \li \f$ F(x) = H * (x-x_{min})^2 / ( 2 * (x_{low} - x_{min}) ) \f$ for \f$ x \in [x_{min},x_{low}] \f$
 \li \f$ F(x) = H * ( (x-x_{low}) + (x_{low}-x_{min})/2 )\f$ for \f$ x \in [x_{low},x_{up}] \f$
 \li \f$ F(x) = 1 - H * (x_{max}-x)^2 / ( 2 * (x_{max}-x_{up})) \f$ for \f$ x \in [x_{up},x_{max}] \f$
 \li \f$ F(x) = 1 \f$ for \f$ x > x_{max} \f$ 

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} + \sqrt{p * 2 * (x_{low}-x_{min})/H} \f$ for \f$ p \in [0,F(x_{low})] \f$
 \li \f$ F^{-1}(p) = p/H + (x_{low}+x_{min})/2 \f$ for \f$ p \in [F(x_{low},F(x_{up})] \f$
 \li \f$ F^{-1}(p) = x_{max} - \sqrt{(1-p) * 2 * (x_{max}-x_{up})/H} \f$ for \f$ p \in [F(x_{up}),1] \f$

 */
#ifndef TTRAPEZIUMDISTRIBUTION_H
#define TTRAPEZIUMDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TTrapeziumDistribution: public TStochasticAttribute
{
private:
    Double_t _dtrapeziumHeight;   ///< The height of the trapezium 
    Double_t _dlowerTriangleArea; ///< Area of the lower triangle
    Double_t _dupperTriangleArea; ///< Area of the upper triangle

protected:
    Double_t _dparameterXlow;  ///< The lower limit of the small base
    Double_t _dparameterXup;   ///< The upper limit of the small base

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     The default range is [0,1] with small base [0.25,0.75]
     */
    TTrapeziumDistribution();

    //! Constructor setting the name
    /*! 
     The default range is [0,1] with small base [0.25,0.75]
     
     \param name (TString) Name of the attribute
     */
    TTrapeziumDistribution(TString name);

    //! Constructor setting name and range
    /*!                 
     The default range of the small base is [0.25*(xmax-xmin), 0.75*(xmax-xmin)]
     
     \param name (TString) Name of the attribute
     \param xmin (Double_t) Lower bound
     \param xmax (Double_t) Upper bound
     */
    TTrapeziumDistribution(TString name, Double_t xmin, Double_t xmax);

    //! Constructor setting name, range and small base limits
    /*!                 
     \param name (TString)  Name of the attribut
     \param xmin  (Double_t) Lower bound
     \param xmax  (Double_t) Upper bound
     \param xlow  (Double_t) lower limit of the small base
     \param xup   (Double_t) upper limit of the small base
     
     \exception UErrorExceptions if xlow is outside of the interval [xmin,xup]
     \exception UErrorExceptions if xup is outside of the interval [xlow,xmax]
     */
    TTrapeziumDistribution(TString name, Double_t xmin, Double_t xmax,
            Double_t xlow, Double_t xup);

    //! Construction of a TUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.10.29)
     
     \param att  (TAttribute *) Pointer to an existing TAttribute object
     \param xlow  (Double_t) lower limit of the small base
     \param xup   (Double_t) upper limit of the small base
     
     \exception UErrorExceptions if xlow > xup
     */
    TTrapeziumDistribution(URANIE::DataServer::TAttribute *att, Double_t xlow,
            Double_t xup);

    //! Default destructor
    virtual ~TTrapeziumDistribution();
    //@}

    /** @name Law parameters
     This law has four parameters: the lower bound, the upper bound and lower and 
     upper limits of the small base. The bounds parameters are defined in the class 
     URANIE::DataServer::TAttribute. The small base limits are defined here.
     **/
    //@{
    //! Set the value of the lower limit of the small base
    /*!
     \param xlow (Double_t) Lower limit of the small base
     
     \exception UErrorExceptions if xlow is outside of the interval [lowerBound,upperBound]
     */
    virtual void setParameterXlower(Double_t xlow);

    //! Set the value of the lower limit of the small base
    /*!
     \param xup (Double_t) Upper limit of the small base
     
     \exception UErrorExceptions if xup is outside of the interval [lowerBound,upperBound]
     */
    virtual void setParameterXupper(Double_t xup);

    //! Returns the value of the lower limit of the small base
    virtual Double_t getParameterXlower()
    {
        return _dparameterXlow;
    }

    //! Returns the value of the upper limit of the small base
    virtual Double_t getParameterXupper()
    {
        return _dparameterXup;
    }

    //! Set all parameters at once
    /*!
     \param xmin (Double_t) value of the lower bound
     \param xmax (Double_t) value of the upper bound
     \param xlow (Double_t) Lower limit of the small base
     \param xup  (Double_t) Upper limit of the small base
     
     \exception UErrorExceptions if xlow is outside of the interval [xmin,xup]
     \exception UErrorExceptions if xup is outside of the interval [xlow,xmax]
     */
    virtual void setParameters(Double_t xmin, Double_t xmax, Double_t xlow,
            Double_t xup);

    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{              
    // \todo theoretical mean and standard deviation formulas must be implemented.
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.                 
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    /*!
     \param option (Option_t *) character chain containing a list of options (Default = "")
     
     \note definir une liste d'options et un exemple d'utilisation (Nicolas Gilardi 2010.05.31)
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TTrapeziumDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
