//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TStochAttribut.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $date$

 \brief  Implementation of the class URANIE::DataServer::TTrapeziumPossibility

 Created: 2010-07-07 15:52:43

 */
#include <iostream>

#include "TTrapeziumPossibility.h"
#include "UExceptions.h"

ClassImp (URANIE::DataServer::TTrapeziumPossibility)
namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TTrapeziumPossibility::TTrapeziumPossibility(TString str) :
        TPossibilityAttribute(str, TPossibilityAttribute::kTrapezium, 0.0, 1.0,
                TPossibilityAttribute::kNotDefined), _dlowerMode(0.25), _dupperMode(
                0.75)
{

}

TTrapeziumPossibility::TTrapeziumPossibility(TString str, Double_t min,
        Double_t max, Double_t lowMode, Double_t upperMode,
        TPossibilityAttribute::EMonotonousRelationship monotonous) :
        TPossibilityAttribute(str, TPossibilityAttribute::kTrapezium, min, max,
                monotonous), _dlowerMode(lowMode), _dupperMode(upperMode)
{
}

TTrapeziumPossibility::~TTrapeziumPossibility()
{
    if (_blog)
    {
        cout << "Destructeur::TTrapeziumPossibility[" << this << "] sname["
                << GetName() << "]" << endl;
    }
}
// -----------------------------------------------------------------------
//  lower and upper modes
// -----------------------------------------------------------------------
void TTrapeziumPossibility::setLowerMode(Double_t dval)
{
    try
    {
        if (dval < getLowerBound())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularPossibility::setMaximumPossibility\n the x value [%f] is lesser than the lower bound [%f]",
                            dval, getLowerBound()));
        if (dval > getUpperBound())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularPossibility::setMaximumPossibility\n the x value [%f] is greater than the upper bound [%f]",
                            dval, getUpperBound()));

        _dlowerMode = dval;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
}
void TTrapeziumPossibility::setUpperMode(Double_t dval)
{
    try
    {
        if (dval < getUpperBound())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularPossibility::setMaximumPossibility\n the x value [%f] is lesser than the upper bound [%f]",
                            dval, getUpperBound()));
        if (dval > getUpperBound())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularPossibility::setMaximumPossibility\n the x value [%f] is greater than the upper bound [%f]",
                            dval, getUpperBound()));

        _dupperMode = dval;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
}
void TTrapeziumPossibility::setModes(Double_t lowVal, Double_t upVal)
{
    setLowerMode(lowVal);
    setUpperMode(upVal);
}

// -----------------------------------------------------------------------
//  printLog
// -----------------------------------------------------------------------
void TTrapeziumPossibility::printLog()
{
    TAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TTrapeziumPossibility::printLog  ******" << endl;
    cout << "**   Possibility [" << TPossibilityAttribute::kTrapezium << "]"
            << endl;
    cout << "**     Inf[" << getLowerBound() << "]  Sup[" << getUpperBound()
            << "]" << endl;
    cout << "**   Modes lower[" << getLowerMode() << "] Upper["
            << getUpperMode() << "]" << endl;
    TPossibilityAttribute::printLog();
    cout << "** End Of TTrapeziumPossibility::printLog  ******" << endl;
    cout << "*******************************" << endl;

}

} // Fin du namespace DataServer
} // Fin du namespace URANIE
