//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TUniformByPartsDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-12-07 11:13:29

 \brief Implementation of the class URANIE::DataServer::TUniformByPartsDistribution

 */
#include "TMath.h"
#include "TUniformByPartsDistribution.h"
#include <iostream>
using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TUniformByPartsDistribution)
namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors                           
// -----------------------------------------------------------------------

TUniformByPartsDistribution::TUniformByPartsDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kUniformByParts, 0.0,
                1.0)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TUniformByPartsDistribution"
                << endl;

    _dparameterMedian = 0.5;
    computeTheoreticalInformation();
}

TUniformByPartsDistribution::TUniformByPartsDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kUniformByParts, 0.0,
                1.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformByPartsDistribution with: Name["
                << str << "]" << endl;

    _dparameterMedian = 0.5;
    computeTheoreticalInformation();
}

TUniformByPartsDistribution::TUniformByPartsDistribution(TString str,
        Double_t min, Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kUniformByParts, min,
                max)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformByPartsDistribution with: Name["
                << str << "] Min[" << min << "] Max[" << max << "]" << endl;

    _dparameterMedian = (min + max) / 2.0;
    computeTheoreticalInformation();
}

TUniformByPartsDistribution::TUniformByPartsDistribution(TString str,
        Double_t min, Double_t max, Double_t median) :
        TStochasticAttribute(str, TStochasticAttribute::kUniformByParts, min,
                max)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformByPartsDistribution with: Name["
                << str << "] Min[" << min << "] Max[" << max << "]Median["
                << median << "]" << endl;

    setParameterMedian(median);
}

TUniformByPartsDistribution::TUniformByPartsDistribution(
        URANIE::DataServer::TAttribute *att) :
        TStochasticAttribute(att, TStochasticAttribute::kUniformByParts)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformByPartsDistribution from existing TAttribute ["
                << att->GetName() << "]" << endl;

    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TUniformByPartsDistribution::~TUniformByPartsDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TUniformByPartsDistribution "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TUniformByPartsDistribution::setParameterMedian(Double_t median)
{
    // Verify that median is between lowerBound and upperBound
    try
    {
        if ((median < lowerBound) || (median > upperBound))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TUniformByPartsDistribution::setParameterMedian: new parameter median [%f] is out of bounds [%f,%f]",
                            median, lowerBound, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterMedian = median;
    computeTheoreticalInformation();
}

void TUniformByPartsDistribution::setParameters(Double_t lower, Double_t upper,
        Double_t median)
{
    // Verify that median is between lower and upper
    try
    {
        if ((median < lower) || (median > upper))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TUniformByPartsDistribution::setParameters: new parameter median [%f] is out of the new bounds [%f,%f]",
                            median, lower, upper));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setBounds(lower, upper);
    _dparameterMedian = median;

    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------
void TUniformByPartsDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = 0.25
            * (upperBound + lowerBound + 2.0 * _dparameterMedian);

    Double_t tmpvar = upperBound * upperBound + lowerBound * lowerBound
            + 2.0 * _dparameterMedian * _dparameterMedian;
    tmpvar += _dparameterMedian * (upperBound + lowerBound);
    tmpvar /= 6.0;

    _dtheoreticalStdDev = TMath::Sqrt(
            tmpvar - _dtheoreticalMean * _dtheoreticalMean);

    // Warning: as I chose to assign the median to the lower interval, 
    //          I can't say for sure that it is an acceptable value for 
    //          the mode (it depends if the lower interval is smaller 
    //          than the upper one). 
    //          Actually, any value inside the smallest interval is a 
    //          valid Mode. Thus I chose the mean of this interval... 
    if ((upperBound - _dparameterMedian) < (_dparameterMedian - lowerBound))
        _dtheoreticalMode = (upperBound - _dparameterMedian) / 2.0;
    else
        _dtheoreticalMode = (_dparameterMedian - lowerBound) / 2.0;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TUniformByPartsDistribution::getPDF(Double_t x)
{
    Double_t dpdf = 0.;
    if ((x > lowerBound) && (x < upperBound))
    {
        if (x < _dparameterMedian)
            dpdf = 0.5 / (_dparameterMedian - lowerBound);
        else
            dpdf = 0.5 / (upperBound - _dparameterMedian);
    }

    return dpdf;
}

Double_t TUniformByPartsDistribution::getCDF(Double_t x)
{
    Double_t dcdf = x;

    if (x < lowerBound)
        dcdf = 0.;
    else if (x > upperBound)
        dcdf = 1.0;
    else
    {
        if (x <= _dparameterMedian)
            dcdf = 0.5 * (x - lowerBound) / (_dparameterMedian - lowerBound);
        else
            dcdf = 0.5
                    + 0.5 * (x - _dparameterMedian)
                            / (upperBound - _dparameterMedian);
    }
    return dcdf;
}

Double_t TUniformByPartsDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TUniformDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    Double_t dinvcdf = 0.;

    if (p <= 0.5)
        dinvcdf = lowerBound + (_dparameterMedian - lowerBound) * 2.0 * p;
    else
        dinvcdf = _dparameterMedian
                + (upperBound - _dparameterMedian) * 2.0 * (p - 0.5);

    return dinvcdf;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------      

void TUniformByPartsDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TStochasticAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TUniformByPartsDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TUniformByPartsDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Median parameter[" << _dparameterMedian << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

