//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TUniformDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TUniformDistribution

 */
/**

 * @class URANIE::DataServer::TUniformDistribution
 *
 * @brief Description of the class TUniformDistribution.

 This class defines a stochastic attribute following a Uniform distribution over 
 the interval \f$ [x_{min}, x_{max}] \f$ .

 This means that the probability of sampling a value outside of this interval is 
 null, while any value inside the interval has an equal probability to be sampled.

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 1/(x_{max}-x_{min}) \f$ for \f$ x \in [x_{min},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$ 
 \li \f$ F(x) = (x-x_{min})/(x_{max}-x_{min})\f$ for \f$ x \in [x_{min},x_{max}] \f$
 \li \f$ F(x) = 1 \f$ for \f$ x > x_{max} \f$ 

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} + ( x_{max} - x_{min} ) * p \f$ for \f$ p \in [0,1] \f$

 */

#ifndef TUNIFORMDISTRIBUTION_H
#define TUNIFORMDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TUniformDistribution: public TStochasticAttribute
{
private:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     The default range is [0,1]
     */
    TUniformDistribution();

    //! Constructor setting the name
    /*! 
     The default range is [0,1]
     
     \param name (TString) Name of the attribute
     */
    TUniformDistribution(TString name);

    //! Constructor setting name and range
    /*!                 
     \param name (TString) Name of the attribute
     \param min (Double_t) Lower bound
     \param max (Double_t) Upper bound
     */
    TUniformDistribution(TString name, Double_t min, Double_t max);

    //! Construction of a TUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)
     
     \param att (TAttribute *) Pointer to an existing TAttribute object                                
     */
    TUniformDistribution(URANIE::DataServer::TAttribute *att);

    //! Default destructor
    virtual ~TUniformDistribution();
    //@}

    /** @name Law parameters
     This law has two parameters: the lower bound, the upper bound. They 
     are both defined in the class URANIE::DataServer::TAttribute.
     **/
    //@{
    //! Set all parameters at once
    /*!
     \param lower  (Double_t) value of the lower bound
     \param upper  (Double_t) value of the upper bound
     
     \exception UErrorExceptions if lower equals to or greater than upper
    */
    virtual void setParameters(Double_t lower, Double_t upper);
    
    //@}
    
    //@}            
    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{                
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    /*!
     \param option (Option_t *) character chain containing a list of options (Default = "")
     
     \note definir une liste d'options et un exemple d'utilisation (Nicolas Gilardi 2010.05.31)
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TUniformDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
