//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TWeibullDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TWeibullDistribution

 */

#include "TMath.h"
#include "TWeibullDistribution.h"

using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TWeibullDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TWeibullDistribution::TWeibullDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kWeibull), _dparameterLambda(
                1.0), _dparameterK(1.0), _dparameterXmin(0.0)
{
    if (_blog)
        cout << "constructing URANIE::DataServer::TWeibullDistribution" << endl;

    if (lowerBound < _dparameterXmin)
        TAttribute::setLowerBound(_dparameterXmin);
    computeTheoreticalInformation();
}

TWeibullDistribution::TWeibullDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kWeibull), _dparameterLambda(
                1.0), _dparameterK(1.0), _dparameterXmin(0.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TWeibullDistribution with: Name["
                << str << "]" << endl;

    if (lowerBound < _dparameterXmin)
        TAttribute::setLowerBound(_dparameterXmin);
    computeTheoreticalInformation();
}

TWeibullDistribution::TWeibullDistribution(TString str, Double_t lambda,
        Double_t k, Double_t xmin) :
        TInfiniteDistribution(str, TStochasticAttribute::kWeibull)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TWeibullDistribution with: Name["
                << str << "] lambda[" << _dparameterLambda << "] k["
                << _dparameterK << "] xmin[" << xmin << "]" << endl;
    setParameters(lambda, k, xmin);
}

TWeibullDistribution::TWeibullDistribution(URANIE::DataServer::TAttribute *att,
        Double_t lambda, Double_t k, Double_t xmin) :
        TInfiniteDistribution(att, TStochasticAttribute::kWeibull)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TWeibullDistribution from URANIE::DataServer::TAttribute object ["
                << att->GetName() << "] " << endl;
    setParameters(lambda, k, xmin);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TWeibullDistribution::~TWeibullDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TWeibullDistribution object ["
                << GetName() << "]" << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TWeibullDistribution::setParameterLambda(Double_t lambda, Bool_t recompute)
{
    try
    {
        if (lambda < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TWeibullDistribution::setParameterLambda: parameter lambda [%f] is too close to zero: lower limit is 1e-4",
                            lambda));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterLambda = lambda;

    if (recompute)
        computeTheoreticalInformation();
}

void TWeibullDistribution::setParameterK(Double_t k, Bool_t recompute)
{
    try
    {
        if (k < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TWeibullDistribution::setParameterK: parameter k [%f] is too close to zero: lower limit is 1e-4",
                            k));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterK = k;

    if (recompute)
        computeTheoreticalInformation();
}

void TWeibullDistribution::setParameterXmin(Double_t xmin, Bool_t recompute)
{
    _dparameterXmin = xmin;

    if (lowerBound < _dparameterXmin)
        TAttribute::setLowerBound(_dparameterXmin);

    if (recompute)
        computeTheoreticalInformation();
}

void TWeibullDistribution::setParameters(Double_t lambda, Double_t k,
        Double_t xmin, Bool_t recompute)
{
    setParameterLambda(lambda, kFALSE);
    setParameterK(k, kFALSE);
    setParameterXmin(xmin, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TWeibullDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = _dparameterXmin
            + _dparameterLambda * TMath::Gamma(1.0 + 1.0 / _dparameterK);

    Double_t tmpVar = _dparameterLambda * _dparameterLambda
            * TMath::Gamma(1.0 + 2.0 / _dparameterK)
        - (_dtheoreticalMean-_dparameterXmin) * (_dtheoreticalMean-_dparameterXmin);

    _dtheoreticalStdDev = TMath::Sqrt(tmpVar);

    Double_t tmpMode = _dparameterXmin;
    if (_dparameterK > 1.0)
        tmpMode += _dparameterLambda
                * TMath::Power((_dparameterK - 1) / _dparameterK,
                        1.0 / _dparameterK);
    else
        tmpMode = THEO_DEF;

    _dtheoreticalMode = tmpMode;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TWeibullDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.0;

    if (x >= _dparameterXmin)
    {
        if ((x >= lowerBound) && (x <= upperBound))
        {
            Double_t scaled_x = (x - _dparameterXmin) / _dparameterLambda;

            pdf = _dparameterK / _dparameterLambda
                    * TMath::Power(scaled_x, _dparameterK - 1)
                    * TMath::Exp(-1.0 * TMath::Power(scaled_x, _dparameterK));
            // Take into account the case where the law is truncated
            pdf /= _pmax - _pmin;
        }
    }
    return pdf;
}

Double_t TWeibullDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;

    if (x >= _dparameterXmin)
    {
        if ((x >= lowerBound) && (x <= upperBound))
        {
            Double_t scaled_x = (x - _dparameterXmin) / _dparameterLambda;

            cdf = 1.0 - TMath::Exp(-1.0 * TMath::Power(scaled_x, _dparameterK));
            // Take into account the case where the law is truncated
            //cdf /= _pmax - _pmin;
            cdf = (cdf - _pmin) / (_pmax - _pmin);
        }
        else if (x > upperBound)
            cdf = 1.0;
    }
    else if (x > DLE_MAX)
        cdf = 1.0;

    return cdf;
}

Double_t TWeibullDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TWeibullDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // Take into account the case where the law is truncated
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    Double_t x = _dparameterXmin
            + _dparameterLambda
                    * TMath::Power(-1.0 * TMath::Log(1.0 - realP),
                            1.0 / _dparameterK);

    return x;
}

void TWeibullDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TWeibullDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Weibull" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Scale parameter lambda[" << _dparameterLambda << "]" << endl;
    cout << "**   Shape parameter k[" << _dparameterK << "]" << endl;
    cout << "**   Location parameter xmin[" << _dparameterXmin << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE
