//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TXMLProblem.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu may 31 09:31:50 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Implementation of the class URANIE::DataServer::TXMLProblem
 
 */
#include <sstream>

// ROOT
#include <TDOMParser.h>
#include <TXMLAttr.h>
#include "TEventList.h"
#include "TList.h"
#include "TDirectory.h"
#include "TSystem.h"
#include "TROOT.h"

// Uranie
#include "TXMLProblem.h"
#include "TStochAttribut.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TSampling.h"
#include "TLauncherFunction.h"
#include "TAttributeFormula.h"
#include "TANNModeler.h"

using std::cout;
using std::endl;

/////////// class URANIE::Sampler::TSampling;

ClassImp (URANIE::DataServer::TXMLProblem)

namespace URANIE
{
namespace DataServer
{
TXMLProblem::TXMLProblem() :
        TNamed("_nom_", "_title_"), _tds(NULL), _tdsdatafield(NULL), _blog(
                kFALSE), _ndebuglevel(0), _ntask(0), _vectorOfModels(0)
{
    if (_blog)
        cout << "construit TXMLProblem" << endl;
    _listOfToDo = new TList();
}
TXMLProblem::TXMLProblem(TString filename, Bool_t blog) :
        TNamed("_nom_", "_title_"), _tds(NULL), _tdsdatafield(NULL), _blog(
                blog), _ndebuglevel(0), _ntask(0), _vectorOfModels(0)
{
    if (_blog)
        cout << "construit TXMLProblem File[" << filename << "]" << endl;
    _listOfToDo = new TList();
    parseFile(filename);
}
TXMLProblem::~TXMLProblem()
{
    cout << "detruit TXMLProblem" << endl;
}
///////////////////////////////////////////////////////////
//  Lecture des donn�es
///////////////////////////////////////////////////////////
Int_t TXMLProblem::parseFile(TString filename)
{
    if (_blog)
        cout << "TXMLProblem::ParseFile[" << filename << "]" << endl;

    TDOMParser *domParser = new TDOMParser();
    Int_t parsecode = domParser->ParseFile(filename);

    UranieError error;

    if (parsecode < 0)
    {
        cerr << domParser->GetParseCodeMessage(parsecode) << endl;
        error.value = -3;
        error.msg =
                "Problem not correctly specified in the input XML file (problem in ParseFile)";
        cerr << error.msg << endl;
        throw error;
    }

    TXMLNode * node = domParser->GetXMLDocument()->GetRootNode();
    if (strcmp(node->GetNodeName(), "Problem") == 0)
    {
        parseNodeHeader(node->GetChildren());
        parseNodeDataDictionary(node->GetChildren());

        // We treat the step as they come
        TXMLNode * node_child = node->GetChildren();
        while (node_child)
        {
            if (node_child->GetNodeType() == TXMLNode::kXMLElementNode)
            { // Element Node
//           cout << "**** TXMLNode::kXMLElementNode node[" << node_child->GetNodeName() << "]" << endl;

                if (strcmp(node_child->GetNodeName(), "Sampler") == 0)
                {
                    parseAndRunNodeSampler(node_child);
                }

                if (strcmp(node_child->GetNodeName(), "Launcher") == 0)
                {
                    parseAndRunNodeLauncher(node_child);
                }
                if (strcmp(node_child->GetNodeName(), "Modeler") == 0)
                {
                    parseAndRunNodeModeler(node_child);
                }
            }
            node_child = node_child->GetNextNode();
        }

    }
    else
    {

        error.value = -3;
        error.msg =
                "Problem not correctly specified in the input XML file (problem node)";
        cerr << error.msg << endl;
        throw error;
    }

    //ParsePersonList(node);

    return 0;

}

///////////////////////////////////////////////////////////
//  Parse the nodes
///////////////////////////////////////////////////////////
void TXMLProblem::parseNodeHeader(TXMLNode *node)
{

    Bool_t bfound = kFALSE;

    for (; !bfound; node = node->GetNextNode())
    {
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeHeader[" << node->GetNodeName()
                        << "]" << endl;
            if (strcmp(node->GetNodeName(), "Header") == 0)
            {
                bfound = kTRUE;
                Int_t id = 0;
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TIter next(attrList);
                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Description attr Name["
                                    << attr->GetName() << "] Value["
                                    << attr->GetValue() << "]" << endl;
                        if (strcmp(attr->GetName(), "name") == 0)
                            SetName(attr->GetValue());
                        if (strcmp(attr->GetName(), "title") == 0)
                            SetTitle(attr->GetValue());
                        if (strcmp(attr->GetName(), "debug") == 0)
                            _ndebuglevel = atoi(attr->GetValue());
                    }
                }
            }
        }
    }
}

///////////////////////////////////////////////////////////
//  Parse the Data node 
///////////////////////////////////////////////////////////
void TXMLProblem::parseNodeDataDictionary(TXMLNode *node)
{

    Bool_t bfound = kFALSE;

    for (; !bfound; node = node->GetNextNode())
    {
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeDataDictionary["
                        << node->GetNodeName() << "]" << endl;
            if (strcmp(node->GetNodeName(), "DataDictionary") == 0)
            {
                _tds = new TDataServer(GetName(), GetTitle());
                _tdsdatafield = new TDataServer(
                        Form("gURANIE__DataField__XML__%s__", GetName()),
                        GetTitle());
                //          _tds->printLog();
                bfound = kTRUE;
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TIter next(attrList);
                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Data attr Name[" << attr->GetName()
                                    << "] Value[" << attr->GetValue() << "]"
                                    << endl;
                        if (strcmp(attr->GetName(), "url") == 0)
                        {
                            _tds->fileDataRead(attr->GetValue());
                        }
                    }
                }
                else
                {
                    // Create the default iterator attributes
                    _tds->addAttribute(NULL);
                }
                if (node->HasChildren())
                {
                    parseNodeDataField(node->GetChildren());
                }

                //          cout << endl << endl << "********************************************************************" << endl ;
                cout << endl << " ** DataServer Name[" << _tds->GetName()
                        << "] Title[" << _tds->GetTitle() << "]" << endl; //////////- Name[" <<  _tds->GetName() << "]    title[" << _tds->GetTitle << "]" << endl;

            }
        }
    }

}

void TXMLProblem::parseNodeDataVariables(TXMLNode *node)
{

    Bool_t bfound = kFALSE;

    for (; !bfound; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeDataVariables["
                        << node->GetNodeName() << "]" << endl;
            if (strcmp(node->GetNodeName(), "variables") == 0)
            {
                bfound = kTRUE;
                if (node->HasChildren())
                {
                    parseNodeDataField(node->GetChildren());
                }
            }
        }
    }
}

void TXMLProblem::parseNodeDataField(TXMLNode *node)
{

    for (; node; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeDataField["
                        << node->GetNodeName() << "]" << endl;
            if (strcmp(node->GetNodeName(), "DataField") == 0)
            {

                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TString sname = "", sformula = "", slabel = "", sunit = "",
                            slaw = "";
                    Double_t dParamA = 0.0, dParamB = 1.0;
                    TIter next(attrList);
                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Data attr Name[" << attr->GetName()
                                    << "] Value[" << attr->GetValue() << "]"
                                    << endl;
                        if (strcmp(attr->GetName(), "name") == 0)
                        {
                            sname = attr->GetValue();
                            if (_tds->isAttribute(sname) == kFALSE)
                            {
			      _tds->addAttribute(new URANIE::DataServer::TAttribute(sname));
			      _tds->getAttribute(sname)->delShare();
                            }
                        }
                        if (strcmp(attr->GetName(), "forumula") == 0)
                            sformula = attr->GetValue();
                        if (strcmp(attr->GetName(), "label") == 0)
                            slabel = attr->GetValue();
                        if (strcmp(attr->GetName(), "unit") == 0)
                            sunit = attr->GetValue();
                        if (strcmp(attr->GetName(), "law") == 0)
                            slaw = attr->GetValue();
                        if (strcmp(attr->GetName(), "min") == 0)
                        {
                            dParamA = atof(attr->GetValue());
                            URANIE::DataServer::TAttribute* attr =
                                    _tds->getAttribute(sname);
                            if (attr)
                                attr->setMinimum(dParamA);
                            attr->setLowerBound(dParamA);
                        }
                        if (strcmp(attr->GetName(), "max") == 0)
                        {
                            dParamB = atof(attr->GetValue());
                            URANIE::DataServer::TAttribute* attr =
                                    _tds->getAttribute(sname);
                            if (attr)
                                attr->setMaximum(dParamB);
                            attr->setUpperBound(dParamB);
                        }
                        if (strcmp(attr->GetName(), "mean") == 0)
                            dParamA = atof(attr->GetValue());
                        if (strcmp(attr->GetName(), "std") == 0)
                            dParamB = atof(attr->GetValue());
                    }

                    if ((_blog) || (_ndebuglevel > 2))
                        cout << "*****************[" << sname << "]" << endl;
                    if (sname.Length() > 0)
                    {
                        URANIE::DataServer::TAttribute* attr =
                                _tds->getAttribute(sname);
                        //        if ( attr == NULL ) {
                        // Case of a formula attribute
                        if (sformula.Length() > 0)
                        {
                            delete attr;
                            attr = NULL;
                            _tds->addAttribute(sname, sformula);
                        }
                        // Case of random attribute
                        if (slaw.Length() > 0)
                        {
                            TString law = slaw;

                            URANIE::DataServer::TStochasticAttribute *varsto =
                                    NULL;

                            //            URANIE::DataServer::TAttribute* x = _tds->getDataSpecification()->getAttribute(sname);

                            if (law.CompareTo("normal", TString::kExact) == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TNormalDistribution(
                                                attr, dParamA, dParamB);
                            }
                            else if (law.CompareTo("lognormal", TString::kExact)
                                    == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TLogNormalDistribution(
                                                attr, dParamA, dParamB);
                            }
                            else if (law.CompareTo("uniform", TString::kExact)
                                    == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TUniformDistribution(
                                                attr);
                            }
                            else if (law.CompareTo("loguniform",
                                    TString::kExact) == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TLogUniformDistribution(
                                                attr);
                            }
                            else if (law.CompareTo("triangular",
                                    TString::kExact) == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TTriangularDistribution(
                                                attr, dParamA);
                            }
                            else if (law.CompareTo("logtriangular",
                                    TString::kExact) == 0)
                            {
                                varsto =
                                        new URANIE::DataServer::TLogTriangularDistribution(
                                                attr, dParamA);
                            }
                            else
                            {
                                cout << "*** ERROR TSampling::addVariable Loi ["
                                        << law << "] non connue" << endl;
                                throw Form(
                                        "*** ERROR TSampling::addVariable(TAttribute *attr, TString sLoi) Loi [%s] non connue\n",
                                        law.Data());
                            }
                            attr->setOrigin(
                                    URANIE::DataServer::TAttribute::kConstant);
                            attr->SetName(
                                    TString(attr->GetName()) + TString("_old"));
                            varsto->setOrigin(
                                    URANIE::DataServer::TAttribute::kAttribute);
                            _tds->addAttribute(varsto);
                            _tdsdatafield->addAttribute(varsto);
                        }
                        //        }
                    }
                }

            }

        }

    }
}

///////////////////////////////////////////////////////////
//  Parse and Run the Sampler node 
///////////////////////////////////////////////////////////
void TXMLProblem::parseAndRunNodeSampler(TXMLNode *node)
{

    if ((_blog) || (_ndebuglevel > 2))
        cout << " **** TXMLProblem::parseNodeSampling[" << node->GetNodeName()
                << "]" << endl;
    if (node->HasAttributes())
    {
        TList *attrList = node->GetAttributes();
        TXMLAttr *attr = 0;
        TIter next(attrList);
        TString smethod = "", ssize = "100", sexport = "";
        while ((attr = (TXMLAttr*) next()))
        {
            if ((_blog) || (_ndebuglevel > 2))
                cout << " ** Data attr Name[" << attr->GetName() << "] Value["
                        << attr->GetValue() << "]" << endl;
            if (strcmp(attr->GetName(), "method") == 0)
            {
                smethod = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "N") == 0)
            {
                ssize = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "export") == 0)
            {
                sexport = attr->GetValue();
            }
        }

        if (_tds)
        {
            delete _tds;
            _tds = new TDataServer(GetName(), GetTitle());
        }

        Int_t nattdf = _tdsdatafield->getNAttributes();
        for (Int_t iatt = 0; iatt < nattdf; iatt++)
            _tds->addAttribute(_tdsdatafield->getAttribute(iatt));

        if ((_ndebuglevel > 2) && (sexport.Length() == 0))
            sexport = Form("sampler_%s_%s.dat", smethod.Data(), ssize.Data());

        // Run the sampling
        if (smethod.Length() != 0)
        {

            Int_t nsize = atoi(ssize.Data());

            addInfoTask("Sampler", smethod + "," + ssize + "," + sexport);
            URANIE::Sampler::TSampling *fsamptds =
                    new URANIE::Sampler::TSampling(_tds, smethod, nsize);
            fsamptds->generateSample();

            if (sexport.Length() > 0)
                _tds->exportData(sexport.Data());

            //delete fsamptds;

            // Add the action to the list: Depreciated code
            _listOfToDo->Add(
                    new TNamed("Sampler_" + smethod + "_" + ssize,
                            "sampler[" + smethod + "," + ssize + "," + sexport
                                    + "]"));
        }

    }
}

///////////////////////////////////////////////////////////
//  Parse and Run the Launcher node 
///////////////////////////////////////////////////////////
void TXMLProblem::parseAndRunNodeLauncher(TXMLNode *node)
{

    if ((_blog) || (_ndebuglevel > 2))
        cout << "TXMLProblem::parseAndRunNodeLauncher[" << node->GetNodeName()
                << "]" << endl;
    if (node->HasAttributes())
    {
        TList *attrList = node->GetAttributes();
        TXMLAttr *attr = 0;
        TIter next(attrList);
        TString smacro = "", sfunction = "", sinput = "", soutput = "",
                snoutput = "", sexport = "";
        while ((attr = (TXMLAttr*) next()))
        {
            if ((_blog) || (_ndebuglevel > 2))
                cout << " ** Data attr Name[" << attr->GetName() << "] Value["
                        << attr->GetValue() << "]" << endl;
            if (strcmp(attr->GetName(), "macro") == 0)
            {
                smacro = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "function") == 0)
            {
                sfunction = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "input") == 0)
            {
                sinput = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "output") == 0)
            {
                soutput = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "noutput") == 0)
            {
                snoutput = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "export") == 0)
            {
                sexport = attr->GetValue();
            }
        }
        if ((_ndebuglevel > 2) && (sexport.Length() == 0))
            sexport = Form("launcher_function_%s_%s.dat", smacro.Data(),
                    sfunction.Data());

        addInfoTask("LauncherFunction",
                smacro + "," + sfunction + "," + sexport);
        // Run the launching
        if (sfunction.Length() != 0)
        {
            if (smacro.Length() > 0)
            {
                cout << " ** Loading macro \"" << smacro << "\"...";
                gROOT->LoadMacro(smacro);
                cout << "done" << endl;
            }

            try
            {
                URANIE::Launcher::TLauncherFunction *gUranie__tlf__ =
                        new URANIE::Launcher::TLauncherFunction(_tds,
                                TString(sfunction), sinput, soutput);

                if (snoutput.Length() > 0)
                    gUranie__tlf__->setNOutput(atoi(snoutput.Data()));

                gUranie__tlf__->run();

                if (sexport.Length() > 0)
                    _tds->exportData(sexport);

                //          gUranie__tlf__->delOutputAttributes();

                delete gUranie__tlf__;
                gUranie__tlf__ = NULL;
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                exit(-1);
            }

            // Add the action to the list: Depreciated code
            _listOfToDo->Add(
                    new TNamed(
                            "LauncherFunction_" + smacro + "_function_"
                                    + sfunction,
                            "LauncherFunction[" + smacro + "," + sfunction + ","
                                    + sexport + "]"));
        }
    }
}
///////////////////////////////////////////////////////////
//  Parse and Run the Modeler node 
///////////////////////////////////////////////////////////
void TXMLProblem::parseAndRunNodeModeler(TXMLNode *node)
{

    if ((_blog) || (_ndebuglevel > 2))
        cout << "TXMLProblem::parseAndRunNodeModeler[" << node->GetNodeName()
                << "]" << endl;
    if (node->HasAttributes())
    {
        TList *attrList = node->GetAttributes();
        TXMLAttr *attr = 0;
        TIter next(attrList);
        TString sinput = "", soutput = "", sname = "", sexport = "";
        Int_t ninit = 0, niter = 0, nhidden = -1;
        Double_t dlearn = -1.0;
        while ((attr = (TXMLAttr*) next()))
        {
            if ((_blog) || (_ndebuglevel > 2))
                cout << " ** Data attr Name[" << attr->GetName() << "] Value["
                        << attr->GetValue() << "]" << endl;
            if (strcmp(attr->GetName(), "input") == 0)
            {
                sinput = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "output") == 0)
            {
                soutput = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "name") == 0)
            {
                sname = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "export") == 0)
            {
                sexport = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "hidden") == 0)
            {
                nhidden = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "iter") == 0)
            {
                niter = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "init") == 0)
            {
                ninit = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "learn") == 0)
            {
                dlearn = atof(attr->GetValue());
            }
        }

        URANIE::DataServer::TBaseModel* gUranie__tabsmod__ =
                new URANIE::DataServer::TBaseModel(sinput, soutput);
        if (sname.Length() > 0)
        {
            gUranie__tabsmod__->setName(sname);
        }
        else
        {
            gUranie__tabsmod__->setName(soutput);
        }
        if (ninit > 0)
            gUranie__tabsmod__->setNInitialization(ninit);
        if (niter > 0)
            gUranie__tabsmod__->setNIteration(niter);
        if (dlearn > 0)
            gUranie__tabsmod__->setLearn(dlearn);

        if (sexport.Length() > 0)
        {
            sexport.ToLower();
            if (sexport == "c")
                gUranie__tabsmod__->setLanguage(TBaseModel::kC);
            if (sexport == "cxx")
                gUranie__tabsmod__->setLanguage(TBaseModel::kCXX);
            if (sexport == "fortran")
                gUranie__tabsmod__->setLanguage(TBaseModel::kFORTRAN);

        }

        // Parse the model structure

        if (node->HasChildren())
        {
            // We treat the step as they come
            TXMLNode * node_child = node->GetChildren();
            while (node_child)
            {
                if (node_child->GetNodeType() == TXMLNode::kXMLElementNode)
                { // Element Node
                    if (strcmp(node_child->GetNodeName(), "NeuralNetwork") == 0)
                    {
                        parseAndRunNodeModelerANN(node_child,
                                gUranie__tabsmod__);
                    }

                }
                node_child = node_child->GetNextNode();
            }

        }

        delete gUranie__tabsmod__;
        gUranie__tabsmod__ = NULL;

//       addInfoTask("Modeler", sinput + "," + Form("%d,", nhidden) + soutput+ "," + sname + "," + sexport);
        // Run the modeler
    }
}

void TXMLProblem::parseAndRunNodeModelerANN(TXMLNode *node, TBaseModel *model)
{

    if ((_blog) || (_ndebuglevel > 2))
        cout << "TXMLProblem::parseAndRunNodeModelerANN[" << node->GetNodeName()
                << "] Model[" << model->GetName() << "] title["
                << model->GetTitle() << "]" << endl;
    if (node->HasAttributes())
    {
        TList *attrList = node->GetAttributes();
        TXMLAttr *attr = 0;
        TIter next(attrList);
        TString sname = "", sexport = "";
        Int_t ninit = 0, niter = 0, nhidden = -1;
        Double_t dlearn = -1.0;
        while ((attr = (TXMLAttr*) next()))
        {
            if ((_blog) || (_ndebuglevel > 2))
                cout << " ** Data attr Name[" << attr->GetName() << "] Value["
                        << attr->GetValue() << "]" << endl;
            if (strcmp(attr->GetName(), "name") == 0)
            {
                sname = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "export") == 0)
            {
                sexport = attr->GetValue();
            }
            if (strcmp(attr->GetName(), "hidden") == 0)
            {
                nhidden = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "iter") == 0)
            {
                niter = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "init") == 0)
            {
                ninit = atoi(attr->GetValue());
            }
            if (strcmp(attr->GetName(), "learn") == 0)
            {
                dlearn = atof(attr->GetValue());
            }
        }

        TString sinput = model->getInput();
        TString soutput = model->getOutput();

        if (sname.Length() == 0)
            sname = Form(Form("%s_nh%03d", model->GetName(), nhidden));

        if (sexport.Length() == 0)
            sexport = sname;
        if (ninit == 0)
            ninit = model->getNInitialization();
        if (niter == 0)
            niter = model->getNIteration();


        addInfoTask("ModelerNeuralNetwork",
                sinput + "," + Form("%d,", nhidden) + soutput + ","
                        + Form("iter=%d,init=%d,", niter, ninit) + sname + ","
                        + sexport);
        URANIE::Modeler::TANNModeler* gUranie__tann__ =
                new URANIE::Modeler::TANNModeler(_tds,
                        Form("%s,%d,%s", sinput.Data(), nhidden,
                                soutput.Data()), model->getLearn());
        gUranie__tann__->SetName(sname);

        gUranie__tann__->train(niter, ninit, "text");

        switch (model->getLanguage())
        {
        case TBaseModel::kCXX:
        {
            gUranie__tann__->exportFunction("c++", Form("%s.C", sexport.Data()),
                    gUranie__tann__->GetName());
        }
            break;
        case TBaseModel::kFORTRAN:
        {
            cout << " ** export fortran" << endl;
        }
            break;
        }
        delete gUranie__tann__;
        gUranie__tann__ = NULL;

        // Run the modeler
    }

}

///////////////////////////////////////////////////////////
//  Parse the MODELS node 
///////////////////////////////////////////////////////////

void TXMLProblem::parseNodeModels(TXMLNode *node)
{

    Bool_t bfound = kFALSE;

    for (; node; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeModels[" << node->GetNodeName()
                        << "]" << endl;
            if (strcmp(node->GetNodeName(), "modeler") == 0)
            {
                bfound = kTRUE;

                TBaseModel * defaultModel = new TBaseModel();
                defaultModel->SetNameTitle("_uranie_default_model",
                        "Model by default");
                // Read the optional attributes of the Models
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TIter next(attrList);
                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Models attr Name[" << attr->GetName()
                                    << "] Value[" << attr->GetValue() << "]"
                                    << endl;
                        if (strcmp(attr->GetName(), "input") == 0)
                            defaultModel->setInput(attr->GetValue());
                        if (strcmp(attr->GetName(), "name") == 0)
                            defaultModel->SetName(attr->GetValue());
                        if (strcmp(attr->GetName(), "output") == 0)
                            defaultModel->setOutput(attr->GetValue());
                        if (strcmp(attr->GetName(), "export") == 0)
                        {
                            TString sexport = attr->GetValue();
                            sexport.ToLower();
                            if (sexport == "c")
                                defaultModel->setLanguage(TBaseModel::kC);
                            if (sexport == "cxx")
                                defaultModel->setLanguage(TBaseModel::kCXX);
                            if (sexport == "fortran")
                                defaultModel->setLanguage(TBaseModel::kFORTRAN);
                        }
                    }
                }

                // Read the child of this node : list of models with the input output string by default
                if (node->HasChildren())
                {
                    parseNodeModel(node->GetChildren(), defaultModel);
                }
                if ((Int_t)(_vectorOfModels.size()) > 0)
                {
                    TString stitle = TString(
                            Form("Construct Models --%d--",
                                    _vectorOfModels.size()));
                    _listOfToDo->Add(
                            new TNamed(defaultModel->GetName(),
                                    Form("modeler[%d]",
                                            _vectorOfModels.size())));

                }
                // Clean
                delete defaultModel;
                defaultModel = NULL;
            }
        }
    }
}

void TXMLProblem::parseNodeModel(TXMLNode *node, TBaseModel *model)
{

    for (; node; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodeModel[" << node->GetNodeName()
                        << "]" << endl;

            // Default values for the current model ( in the node models)
            TString sinput = model->getInput();
            TString soutput = model->getOutput();

            if (strcmp(node->GetNodeName(), "patterns") == 0)
            {
                parseNodePatterns(node, model);
            }
            if (strcmp(node->GetNodeName(), "algorithm") == 0)
            {
                parseNodeModelerAlgorithm(node, model);
            }

            if (strcmp(node->GetNodeName(), "NeuralNetwork") == 0)
            {
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TString sname = "", sclass = "", shidden = "", sfilename =
                            "", sfunctionname = "";
                    TIter next(attrList);
                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Model attr Name[" << attr->GetName()
                                    << "] Value[" << attr->GetValue() << "]"
                                    << endl;
                        if (strcmp(attr->GetName(), "name") == 0)
                            sname = attr->GetValue();
                        if (strcmp(attr->GetName(), "input") == 0)
                            sinput = attr->GetValue();
                        if (strcmp(attr->GetName(), "output") == 0)
                            soutput = attr->GetValue();
                        if (strcmp(attr->GetName(), "hidden") == 0)
                            shidden = attr->GetValue();
                        if (strcmp(attr->GetName(), "file") == 0)
                            sfilename = attr->GetValue();
                        if (strcmp(attr->GetName(), "function") == 0)
                            sfunctionname = attr->GetValue();
                    }
                    if ((sinput.Length() > 0) && (soutput.Length() > 0))
                    {
                        TBaseModel * crtModel = new TBaseModel(sname, sinput,
                                shidden, soutput);
                        crtModel->setLanguage(model->getLanguage());
                        crtModel->setNIteration(model->getNIteration());
                        crtModel->setNInitialization(
                                model->getNInitialization());
                        crtModel->setSeed(model->getSeed());
                        if (sfilename.Length() != 0)
                            crtModel->setFileNameExport(sfilename);
                        if (sfunctionname.Length() != 0)
                            crtModel->setFunctionNameExport(sfunctionname);
                        _vectorOfModels.push_back(crtModel);
                    }
                }
            }

        }
    }
}

void TXMLProblem::parseNodePatterns(TXMLNode *node, TBaseModel *model)
{

    Bool_t bfound = kFALSE;

    for (; !bfound; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodePatterns[" << node->GetNodeName()
                        << "]" << endl;
            if (strcmp(node->GetNodeName(), "patterns") == 0)
            {
                bfound = kTRUE;

                // Read the optional attributes of the Models
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TIter next(attrList);

                    Int_t npat = 100;
                    if (_tds->getTuple() != NULL)
                        npat = _tds->getTuple()->GetEntries();
                    if ((_blog) || (_ndebuglevel > 2))
                        cout << " ** init npat[" << npat << "]" << endl;

                    // Read the selection of patterns if exist
                    if (node->HasChildren())
                    {
                        TString sselect = parseNodePatternSelected(
                                node->GetChildren());

                        if (sselect != "")
                        {
                            _tds->getTuple()->Draw(">>elist1", sselect);
                            TEventList *telapp = (TEventList*) gDirectory->Get(
                                    "elist1");
                            npat = telapp->GetN();
                            if ((_blog) || (_ndebuglevel > 2))
                                cout << " ** APP Select[" << sselect << "] ["
                                        << npat << "]" << endl;
                        }
                    }

                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Patterns attr Name[" << attr->GetName()
                                    << "] Value[" << attr->GetValue() << "]"
                                    << endl;

                        if (strcmp(attr->GetName(), "learn") == 0)
                        {
                            Float_t valcrt = atof(attr->GetValue());
                            if (valcrt < 1)
                                npat = (Int_t)(npat * valcrt);
                            else
                            {
                                if ((Int_t)(valcrt) < npat)
                                    npat = (Int_t)(valcrt);
                            }

                            if ((_blog) || (_ndebuglevel > 2))
                                cout << " ** Finally npat[" << npat << "]"
                                        << endl;
                        }

                    }
                }

            }
        }
    }
}

void TXMLProblem::parseNodeModelerAlgorithm(TXMLNode *node, TBaseModel *model)
{

    Bool_t bfound = kFALSE;

    for (; !bfound; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodePatterns[" << node->GetNodeName()
                        << "]" << endl;
            if (strcmp(node->GetNodeName(), "algorithm") == 0)
            {
                bfound = kTRUE;

                // Read the optional attributes of the Models
                if (node->HasAttributes())
                {
                    TList *attrList = node->GetAttributes();
                    TXMLAttr *attr = 0;
                    TIter next(attrList);

                    // Read the selection of patterns if exist
                    if (node->HasChildren())
                    {
                        TString sselect = "ReadLibraryWith Constraint";
                    }

                    while ((attr = (TXMLAttr*) next()))
                    {
                        if ((_blog) || (_ndebuglevel > 2))
                            cout << " ** Algorithm attr Name["
                                    << attr->GetName() << "] Value["
                                    << attr->GetValue() << "]" << endl;

                        if (strcmp(attr->GetName(), "iter") == 0)
                            model->setNIteration(atoi(attr->GetValue()));
                        if (strcmp(attr->GetName(), "init") == 0)
                            model->setNInitialization(atoi(attr->GetValue()));
                        if (strcmp(attr->GetName(), "seed") == 0)
                            model->setSeed(atoi(attr->GetValue()));
                    }
                }

            }
        }
    }
}

TString TXMLProblem::parseNodePatternSelected(TXMLNode *node)
{

    TString sselect = "";
    for (; node; node = node->GetNextNode())
    { // node / !bfound
        if (node->GetNodeType() == TXMLNode::kXMLElementNode)
        { // Element Node
            if ((_blog) || (_ndebuglevel > 2))
                cout << "TXMLProblem::parseNodePatternSelected["
                        << node->GetNodeName() << "]" << endl;
            if (strcmp(node->GetNodeName(), "select") == 0)
            {
                sselect = node->GetChildren()->GetContent();
            }
        }
    }

    return sselect;
}

///////////////////////////////////////////////////////////
//   Execute the actions
///////////////////////////////////////////////////////////
void TXMLProblem::init(Option_t *option)
{
    if (_blog)
        cout << "TXMLProblem::init" << endl;
}

void TXMLProblem::run(Option_t *option)
{
    if (_listOfToDo)
    {
        cout << endl << "Liste des choses � faire Name[" << GetName() << "]"
                << endl;
        _listOfToDo->Print();

        init();

        // The treatment
        TNamed *obj = 0;
        Int_t ncrt = 1;
        TIter next(_listOfToDo);
        while (obj = (TNamed *) next())
        {
            TString sname = obj->GetName();
            TString stitle = obj->GetTitle();
            cout << " ******* Task[" << sname << "]" << endl;
            cout << endl << endl
                    << "********************************************************************"
                    << endl << " ** Task " << ncrt << " - Name[" << sname
                    << "]    title[" << obj->GetTitle() << "]" << endl;

            if (sname.BeginsWith("Sampler", TString::kExact))
            {
                cout << "Sampling" << endl;
                stitle.ReplaceAll("[", " ");
                stitle.ReplaceAll("]", " ");
                stitle.ReplaceAll(",", " ");
                istringstream inStream((char *) stitle.Data());
                string inName, inMethod, inSize, inExport;
                inStream >> inName >> inMethod >> inSize >> inExport;
                cout << "inName[" << inName << "]" << endl;
                cout << "inMethod[" << inMethod << "]" << endl;
                cout << "inSize[" << inSize << "]" << endl;
                cout << "inExport[" << inExport << "]" << endl;
                Int_t nsize = atoi(inSize.c_str());
                cout << "nsize[" << nsize << "]" << endl;

                URANIE::Sampler::TSampling *fsamptds =
                        new URANIE::Sampler::TSampling(_tds, inMethod.c_str(),
                                nsize);
                fsamptds->generateSample();

                if (inExport.length() > 0)
                    _tds->exportData(inExport.c_str());

                delete fsamptds;
                fsamptds = NULL;

            }

            ////////////////////////////////
            // Node Launcher
            ////////////////////////////////

            if (sname.BeginsWith("LauncherFunction", TString::kExact))
            {
                cout << "Launching" << endl;

                stitle.ReplaceAll("[", " ");
                stitle.ReplaceAll("]", " ");
                stitle.ReplaceAll(",", " ");
                istringstream inStream((char *) stitle.Data());
                string inName, inMacro, inFunction, inExport;
                inStream >> inName >> inMacro >> inFunction >> inExport;
                cout << "inName[" << inName << "]" << endl;
                cout << "inMacro[" << inMacro << "]" << endl;
                cout << "inFunction[" << inFunction << "]" << endl;
                cout << "inExport[" << inExport << "]" << endl;

                TString smacro = TString(inMacro.c_str());
                if (smacro.Length() > 0)
                {
                    cout << "Loading macro \"" << inMacro << "\"...";
                    gROOT->LoadMacro(inMacro.c_str());
                    cout << "done" << endl;
                }
                else
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "TXMLProblem::run ERROR : macro filename is empty."));

                }

                URANIE::Launcher::TLauncherFunction *glauncher =
                        new URANIE::Launcher::TLauncherFunction(_tds,
                                inFunction.c_str());
                glauncher->run();

                if (inExport.length() > 0)
                    _tds->exportData(inExport.c_str());

		delete glauncher;
                glauncher = NULL;
            }

            ncrt++;
        }

        terminate();
        clean();
    }
}

void TXMLProblem::terminate(Option_t *option)
{
    //    gSystem->cd(_sdirectory);
    if (_blog)
        cout << "TXMLProblem::terminate" << endl;

}
void TXMLProblem::clean(Option_t *option)
{
    if (_blog)
        cout << "TXMLProblem::clean" << endl;
}

void TXMLProblem::addInfoTask(TString sname, TString stitle)
{
    _ntask++;
//     cout << endl << "********************************************************************" << endl ;
    cout << endl << " ** Task " << _ntask << " - Name[" << sname
            << "]    title[" << stitle << "]" << endl;
}

///////////////////////////////////////////////////////////
//  Printing Log
///////////////////////////////////////////////////////////

void TXMLProblem::printLog(Option_t *option)
{
    cout << "TXMLProblem::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << "Debug[" << _ndebuglevel << "]" << endl;
    if (_tds)
        _tds->printLog();
    // Les models
    Int_t n = (Int_t)(_vectorOfModels.size());
    cout << "Model [" << n << "]" << endl;
    for (Int_t i = 0; i < n; i++)
    {
        cout << "Model index[" << i + 1 << "/" << n << "]" << endl;
        _vectorOfModels[i]->printLog();
    }
    _listOfToDo->Print();
    cout << "fin de TXMLProblem::printLog[" << option << "]" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

