//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UEntry.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu apr 26 14:14:11 CEST 2007

 $Revision$
 $Date$

 \brief Interface of the class URANIE::DataServer::UEntry

 add doc 
 */
/*!
 @class URANIE::DataServer::UEntry

 @brief Description of the class UEntry.

 This is the interface used to consider our input and outputs as vectors, disregarding the nature of their content
 (double, string or vectors).
 It contains a great deal of helping function (to deal with the UEntry) and also some useful functions (many for 
 parsing and splitting strings), defined here as it is included in most of the files (convenient location)

 */
#ifndef UENTRY_H
#define UENTRY_H

//std
#include <iostream>
#include <iomanip>
#include <fstream>
#include <map>
#include <string>
#include <vector>
using namespace std;

//ROOT
#include "TString.h"
#include "TObject.h"

// URANIE
#include "DataServer.h"

// Mess used to help debugging
#define XYZZY std::cout<<__FILE__<<"::"<<__LINE__<<std::endl
#define XYZ std::cout<<__FILE__<<"::"<<__LINE__<<"  "
#define SSTR_PREC 12

//#define UNUSED(...) (void)(__VA_ARGS__)
#define UNUSED1(z) (void)(z)
#define UNUSED2(y,z) UNUSED1(y),UNUSED1(z)
#define UNUSED3(x,y,z) UNUSED1(x),UNUSED2(y,z)
#define UNUSED4(b,x,y,z) UNUSED2(b,x),UNUSED2(y,z)
#define UNUSED5(a,b,x,y,z) UNUSED2(a,b),UNUSED3(x,y,z)
#define UNUSED6(a,b,c,x,y,z) UNUSED3(a,b,c),UNUSED3(x,y,z)

#define VA_NUM_ARGS_IMPL(_1,_2,_3,_4,_5, _6, N,...) N
#define VA_NUM_ARGS(...) VA_NUM_ARGS_IMPL(__VA_ARGS__, 6, 5, 4, 3, 2, 1)

#define ALL_UNUSED_IMPL_(nargs) UNUSED ## nargs
#define ALL_UNUSED_IMPL(nargs) ALL_UNUSED_IMPL_(nargs)
#define UNUSED(...) ALL_UNUSED_IMPL( VA_NUM_ARGS(__VA_ARGS__))(__VA_ARGS__ )

#ifndef WIN32
#include <time.h>
#define TSTART clock_t begin_time = begin_time = clock()
#define TBEG begin_time = clock()
#define TEND std::cout << __FILE__<<"::"<<__LINE__<<" time diff is " << float( clock () - begin_time ) /  CLOCKS_PER_SEC << std::endl
#endif

namespace URANIE
{
namespace DataServer
{

class UEntry : public TObject
{
    
public:

    enum TAG {Double, String, Vector, Empty};
    TAG tag; ///< Tag the nature of the considered entry
    string name; ///< name of the entry (not used normally, but helpful to debug)
    bool bisnew; ///< State wether this entry is created to hold an existing pointer (bisnew==false) or if it should contain a new memory allocation (bisnew==true)
    union pouet {
	double *dou;
	string *str;
	vector<double> *vec;
    } val; ///< the pointer to which the data are stored (UEntry::Double<->double*; UEntry::String<->string*; UEntry::Vector<->vector<double>*)
    
    
    /** @name  Constructor/Destructor
	Constructors and destructors of the entry
    **/
    //@{

    //! Default constructor
    /*!
      Pointer is set to NULL and tag to Empty (only created so that ROOT does not complains about it
    */
    UEntry();
    
    //! Tag/Name constructor
    /*!
      This constructor is used when the entry has to hold a new memory allocation to store data
      \param atag: chosen tag for the entry
      \param aname: chosen name for the entry
    */
    UEntry(UEntry::TAG atag, string aname="[default]");

    //! Tag/Name/Pointer constructor
    /*!
      This constructor is used when the entry should not hold a new memory allocation, but used an already existing one instead
      \param atag: chosen tag for the entry
      \param aname: chosen name for the entry
      \param ptr: pointer to the allocated memory. This should be passed as a void*, knowing that it will be converted accordingly to the chosen tag:
      UEntry::Double : void* -> double*; UEntry::String : void* -> string*; UEntry::Vector : void* -> vector<double>*; 
    */
    UEntry(UEntry::TAG atag, string aname, void *ptr);
    //! Default destructor
    /*!
      Delete the pointer when bisnew==true
     */
    ~UEntry();
    //@}

    
    /** @name  Playing with content
	Change the entry values, from different format.
    **/
    //@{

    //!Change the address to where the entry is pointing
     /*!
       Change the address to where the entry is pointing
       \param ptr: pointer to the allocated memory. This should be passed as a void*, knowing that it will be converted accordingly to the chosen tag:
       UEntry::Double : void* -> double*; UEntry::String : void* -> string*; UEntry::Vector : void* -> vector<double>*; 
    */
    void setAddress( void *ptr );

    //!Get the address to where the entry is pointing
     /*!
       Get the address to where the entry is pointing
    */
    string getAddress(  );

    //!Dump the content
    /*!
      Return the content of the entry as a string
      \param delim: chosen delimiter for vector
      \param beg: beginning boundary to delimite the field (vector or string)
      \param end: ending boundary to delimite  the field (vector or string)
    */
    string dumpStr( TString delim=",", TString beg="", TString end="", TString format="%e" );

    //!Dump the content
    /*!
      Return the content of the entry as a string
      \param nentry : the number of the entry (for vectors)
      \param moretowrite: a boolean to specify whether this entry is the last one (moretowrite==false) or not (moretowrite==true)
    */
    string dumpStr( unsigned int nentry, bool &moretowrite, TString beg="", TString end="", TString format="%e" );

    //! Return the dimension of the content
    /*!
      The returned value is a code for the MPI treatment: -1 is for a double, -2 for a string and the size of the vector for a vector
     */
    int getDim();
    
    //! Copy the content of an entry
    /*!
      Copy the content of an entry into this one
      \param src: pointer to another entry
    */
    void copyEntry(UEntry *src);

    //! Push Back the content of the pointer
    /*!
      Fill the entry memory with the value pointed by the pointer, without clearing the memory.
      It is equivalent to a simple assignement in the case of doubles and strings and to a push_back for vectors.
      \param ptr: pointer to the allocated memory. This should be passed as a void*, knowing that it will be converted accordingly to the chosen tag:
      UEntry::Double : void* -> double*; UEntry::String : void* -> string*; UEntry::Vector : void* -> vector<double>*; 
      \param isdouble: state wether the double case is a double pointer (isdouble==true) or a float one (isdouble==false)
    */
    void pushEntry( void *ptr, bool isdouble=true );

    //! Push back values
    /*!
      Same as previous one, but should be applied only for double and vector<double> cases.
      Exception if not ?
     */
    void pushEntry( double value );

    //! Push back values
    /*!
      Same as previous one, but should be applied only for double and vector<double> cases.
      Exception if not ?
    */
    void pushEntry( TString value, TString del="", TString beg="", TString end="" );
    void pushEntry( string value, string del="", string beg="", string end="" )
	{
	    pushEntry(TString(value), TString(del), TString(beg), TString(end) );
	}
    
    //! Assign the content of an entry
    /*!
      Fill the entry memory with the value pointed by the pointer
      \param ptr: pointer to the allocated memory. This should be passed as a void*, knowing that it will be converted accordingly to the chosen tag:
      UEntry::Double : void* -> double*; UEntry::String : void* -> string*; UEntry::Vector : void* -> vector<double>*; 
      \param isdouble: state wether the double case is a double pointer (isdouble==true) or a float one (isdouble==false)
    */
    void assignEntry( void *ptr, bool isdouble=true );

     //! Fill entry from a TString
    /*!
      Fill the entry memory with the content of the string provided
      \param line:  the provided line for the corresponding entry
      \param delim: chosen delimiter for vector
    */
    void fillEntry( TString line, TString delim=",", TString beg="", TString end="");

    //! Clear the content of the entry
    void clearEntry();
    //@}
    
    /** @name  Utility 
	Utility function used mostly by other methods.
    **/
    //@{

    //! Static method to split a string into a vector of substring
    /*!
      The input string is split following a delimiter, and the result is stored in the provided vector
      \param input: the string to split
      \param delim: the delimiter to be considered
      \param res: the vector of string provided that will be cleared and filled.
    */
    static void StringParser(string input, string delim, vector<string> &res, bool LeaveEmpty=false);

    //! Static method to dump the content of a double vector
    /*!
      It dumps the content following our internal convention of comma separators
      \param buffer: a pointer to the considered vector
      \param precision: if precision is precised, the display is done with scientific style.
      \param delim: chosen delimiter for vector
    */
    static string dumpVectorStr(vector<double> *buffer, int precision=-1, string delim=",", TString format="%e");
    
    //! Static method to dump the content of a double vector
    /*!
      It dumps the content following our internal convention of comma separators
      \param buffer: a pointer to the considered vector
      \param precision: if precision is precised, the display is done with scientific style.
    */
    static string dumpVectorStr(vector<double> *buffer, int precision, TString delim, TString format)
	{
	    return  dumpVectorStr( buffer, precision, string(delim), format );
	}

    static string dumpVectorStr(vector<double> *buffer, int precision, TString delim)
	{
	    return  dumpVectorStr( buffer, precision, string(delim) );
	}

    static void fillVectorFromStr(TString input, TString delim, vector<double> &buffer);
    

    //! Static method be sure to read properly file independently of the OS that produced it (Windows/linux)
    static std::istream& safeGetline(std::istream& is, std::string& t);
    
    //!Check whether a line is empty
    /*!
      Check whether a line is empty
      \param line: the line to be checked
    */
    static bool isWhiteLine(string line);
    
    //@}


    
    ClassDef(URANIE::DataServer::UEntry, ID_DATASERVER);
};

}
}
#endif
