//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UExceptions.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu ao� 9 11:58:09 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::Exceptions::UExceptions
 
 */

/*!
 @class URANIE::Exceptions::UExceptions
 
 @brief Description of the class UExceptions.
 
 To write by the developper.
 
 */
#ifndef UEXCEPTIONS_H
#define UEXCEPTIONS_H

#include <stdlib.h>

#include <iostream>
using std::cout;
using std::endl;
using std::string;
using std::istringstream;

// Uranie
#include "UVersion.h"
#include "TDataServer.h"
#include "UMessageLogger.h"

namespace URANIE
{
namespace Exceptions
{
class UExceptions
{
public:
    enum UMsgType
    {
        kVERBOSE = 1,
        kDEBUG,
        kINFO,
        kDEPRECATED,
        kWARNING,
        kERROR,
        kFATAL,
        kALWAYS
    };
protected:
    // the mutable declaration is needed to use the logger in const methods
    //GA static URANIE::DataServer::UMessageLogger * _fLogger; // message logger
    URANIE::DataServer::UMessageLogger _fLogger; // message logger
private:
    //Bool_t _blog;          ///< Affichage de la log
    TString _msg;           ///< The error message
    UMsgType _msgtype;       ///< The type of message
    TString _file;
    Int_t _line;
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    UExceptions();
    //! Default constructor
    UExceptions(TString msg);
    //! Default constructor
    UExceptions(TString file, Int_t line, TString msg);

    //! Default destructor
    virtual ~UExceptions();
    //@}

    virtual void printMessage();

    //---------------------------------------------
    /** @name Setter and Getter 
     **/
    //@{
    UMsgType getMessageType()
    {
        return _msgtype;
    }
    void setMessageType(UMsgType activeType, UMsgType minType = kINFO);

    TString getFile()
    {
        return _file;
    }
    TString getMessage()
    {
        return _msg;
    }
    void setMessage(TString str)
    {
        _msg = str;
    }
    void addMessage(TString str)
    {
        _msg += Form("\n%s", str.Data());
    }
    Int_t getLine()
    {
        return _line;
    }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
/*
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
*/
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Exceptions::UExceptions, ID_DATASERVER)
    //Abstract class of exceptions
};

////////////////////////////////////////////////////////////
//// UWarningExceptions
////////////////////////////////////////////////////////////
class UWarningExceptions: public UExceptions
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    UWarningExceptions() :
            UExceptions()
    {
        setMessageType(UExceptions::kWARNING);
// 	_fLogger->SetMinType(URANIE::DataServer::UMessageLogger::kWARNING);
    }

    //! Default constructor
    UWarningExceptions(TString msg) :
            UExceptions(msg)
    {
        setMessageType(UExceptions::kWARNING);
//  	_fLogger->SetMinType(URANIE::DataServer::UMessageLogger::kWARNING);
    }
    //! Default constructor
    UWarningExceptions(TString file, Int_t line, TString msg) :
            UExceptions(file, line, msg)
    {
        setMessageType(UExceptions::kWARNING);
// 	_fLogger->SetMinType(URANIE::DataServer::UMessageLogger::kWARNING);
    }

    //! Default destructor
    virtual ~UWarningExceptions();
    //@}

    virtual void printMessage();

    ClassDef(URANIE::Exceptions::UWarningExceptions, ID_DATASERVER)
    //Classe warning exceptionsx
};

////////////////////////////////////////////////////////////
//// UDepreciatedExceptions
////////////////////////////////////////////////////////////
class UDeprecatedExceptions: public UExceptions
{
private:
    Int_t _nfromVersion, _nlastVersion;
    TString _sfromVersion, _slastVersion;
    TString _sobject, _smethod;
    void evaluateType()
    {

        TString slastversion = _slastVersion;
        slastversion.ReplaceAll(".", " ");
        slastversion.ReplaceAll("/", " ");
        slastversion += " ";
        istringstream inStream((char *) slastversion.Data());
        string smajor, sminor;
        inStream >> smajor >> sminor;

        Int_t nmajor = atoi(smajor.c_str());
        Int_t nminor = atoi(sminor.c_str());

        if ( URANIE_VERSION_CODE < URANIE_VERSION(nmajor, nminor, 00))
        {
            setMessageType(kDEPRECATED);
        }
        else
        {
            setMessageType(kERROR);
            addMessage(
                    Form(
                            "The current version of uranie is %s \nand the last version authorized is %d.%d",
                            URANIE_RELEASE, nmajor, nminor));
        }

    }
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    UDeprecatedExceptions() :
            UExceptions()
    {
        evaluateType();
    }
    //! Constructor with all informations
    UDeprecatedExceptions(TString file, Int_t line, TString fromVersion,
            TString lastVersion, TString sobject, TString smethod, TString msg) :
            UExceptions(file, line, msg), _sfromVersion(fromVersion), _slastVersion(
                    lastVersion), _sobject(sobject), _smethod(smethod)
    {
        evaluateType();
    }
    //! Default destructor
    virtual ~UDeprecatedExceptions();

    virtual void printMessage();

    ClassDef(URANIE::Exceptions::UDeprecatedExceptions, ID_DATASERVER)
    //Classe warning exceptionsx
};

////////////////////////////////////////////////////////////
//// UErrorExceptions
////////////////////////////////////////////////////////////
class UErrorExceptions: public UExceptions
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    UErrorExceptions() :
            UExceptions()
    {
        setMessageType(UExceptions::kERROR);
    }

    //! Default constructor
    UErrorExceptions(TString msg) :
            UExceptions(msg)
    {
        setMessageType(UExceptions::kERROR);
    }
    //! Default constructor
    UErrorExceptions(TString file, Int_t line, TString msg) :
            UExceptions(file, line, msg)
    {
        setMessageType(UExceptions::kERROR);
    }

    //! Default destructor
    virtual ~UErrorExceptions();
    //@}

    virtual void printMessage();

    ClassDef(URANIE::Exceptions::UErrorExceptions, ID_DATASERVER)
    //Classe warning exceptionsx
};

////////////////////////////////////////////////////////////
//// UInfoExceptions
////////////////////////////////////////////////////////////
class UInfoExceptions: public UExceptions
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    UInfoExceptions() :
            UExceptions()
    {
        setMessageType(UExceptions::kINFO);
    }

    //! Default constructor
    UInfoExceptions(TString msg) :
            UExceptions(msg)
    {
        setMessageType(UExceptions::kINFO);
    }
    //! Default constructor
    UInfoExceptions(TString file, Int_t line, TString msg) :
            UExceptions(file, line, msg)
    {
        setMessageType(UExceptions::kINFO);
    }

    //! Default destructor
    virtual ~UInfoExceptions();
    //@}

    virtual void printMessage();

    ClassDef(URANIE::Exceptions::UInfoExceptions, ID_DATASERVER)
    //Classe warning exceptionsx
};

} // Fin du namespace Exceptions
} // Fin du namespace URANIE
#endif
