//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UMessageLogger.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu avr 26 14:14:11 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::DataServer::UMessageLogger
 
 */
/*!
 @class URANIE::DataServer::UMessageLogger

 @brief Description of the class UMessageLogger.

 ostreamstream derivative to redirect and format output
 */

#ifndef UMESSAGELOGGER_H
#define UMESSAGELOGGER_H

// STL include(s):
#include <string>
#include <sstream>
#include <map>

// ROOT include(s)
#include "TObject.h"
#include "TString.h"

// Uranie
#include "DataServer.h"

namespace URANIE
{
namespace DataServer
{

class UMessageLogger: public std::ostringstream, public TObject
{

public:
    // define outside of class to facilite access
    enum EMsgType
    {
        kVERBOSE = 1,
        kDEBUG = 2,
        kINFO = 3,
        kDEPRECATED = 4,
        kWARNING = 5,
        kERROR = 6,
        kFATAL = 7,
        kALWAYS = 8
    };

    UMessageLogger(const TObject* source, EMsgType minType = kINFO);
    UMessageLogger(const std::string& source, EMsgType minType = kINFO);
    UMessageLogger(EMsgType minType = kINFO);
    UMessageLogger(const UMessageLogger& parent);
    ~UMessageLogger();

    // Accessors
    void SetSource(const std::string& source)
    {
        fStrSource = source;
    }
    EMsgType GetMinType() const
    {
        return fMinType;
    }
    void SetMinType(EMsgType minType)
    {
        fMinType = minType;
    }
    UInt_t GetMaxSourceSize() const
    {
        return (UInt_t) fMaxSourceSize;
    }
    std::string GetPrintedSource() const;
    std::string GetFormattedSource() const;

    // Needed for copying
    UMessageLogger& operator=(const UMessageLogger& parent);

    // Stream modifier(s)
    static UMessageLogger& Endmsg(UMessageLogger& logger);

    // Accept stream modifiers
    UMessageLogger& operator<<(UMessageLogger& (*_f)(UMessageLogger&));
    UMessageLogger& operator<<(std::ostream& (*_f)(std::ostream&));
    UMessageLogger& operator<<(std::ios& (*_f)(std::ios&));

    // Accept message type specification
    UMessageLogger& operator<<(EMsgType type);

    // For all the "conventional" inputs
    template<class T> UMessageLogger& operator<<(T arg)
    {
        *(std::ostringstream*) this << arg;
        return *this;
    }

private:

    // private utility routines
    void Send();
    void InitMaps();
    void WriteMsg(EMsgType type, const std::string& line) const;

    const TObject* fObjSource;     // the source TObject (used for name)
    std::string fStrSource;     // alternative string source
    const std::string fPrefix;        // the prefix of the source name
    const std::string fSuffix;        // suffix following source name
    EMsgType fActiveType;    // active type
    const std::string::size_type fMaxSourceSize; // maximum length of source name
    static Int_t             fgInstanceCounter; // counts open MsgLogger instances

    static std::map<EMsgType, std::string> *fTypeMap; // matches output types with strings
    static std::map<EMsgType, std::string> *fColorMap; // matches output types with terminal colors
    EMsgType fMinType;       // minimum type for output

    ClassDef(UMessageLogger, 0) // ostringstream derivative to redirect and format logging output  
    ;
};
// class UMessageLogger

inline UMessageLogger& UMessageLogger::operator<<(
        UMessageLogger& (*_f)(UMessageLogger&))
{
    return (_f)(*this);
}

inline UMessageLogger& UMessageLogger::operator<<(
        std::ostream& (*_f)(std::ostream&))
{
    (_f)(*this);
    return *this;
}

inline UMessageLogger& UMessageLogger::operator<<(std::ios& (*_f)(std::ios&))
{
    (_f)(*this);
    return *this;
}

inline UMessageLogger& UMessageLogger::operator<<(EMsgType type)
{
    fActiveType = type;
    return *this;
}

// Although the proper definition of "Endl" as a function pointer
// would be nicer C++-wise, it introduces some "unused variable"
// warnings so let's use the #define definition after all...
//   static UMessageLogger& ( *Endl )( UMessageLogger& ) = &UMessageLogger::Endmsg;
#define Endl URANIE::DataServer::UMessageLogger::Endmsg

} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
