/**
 * @file   UserFunctions.C
 * @brief  User Functions
 */

/****************** User Function **********************/
//! The true flow rate model
/*!
  It describes the flow rate of water through a borehole from an upper aquifer to a lower aquifer separated by an impermeable rock layer.
  The response variable \f$ y\f$, the \b flow \b water through the borehole in \f$ m^{3}/yr\f$, is given by :
  \f[
  y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
  \f]

  \param x (double *) the input vector in the order :
  \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
  \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
  \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
  \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
  \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
  \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
  \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
  \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
  \param y (double *) the output vector

  \see waterboreholeHoXu
*/
void flowrateModel(double *x, double *y)
{
  Double_t drw = x[0], dr  = x[1];
  Double_t dtu = x[2], dtl = x[3];
  Double_t dhu = x[4], dhl = x[5];
  Double_t dl  = x[6], dkw = x[7];

  Double_t dnum = 2.0 * TMath::Pi() * dtu * ( dhu -dhl);
  Double_t dlnronrw = TMath::Log( dr / drw);
  Double_t dden = dlnronrw * ( 1.0 +  ( 2.0 * dl * dtu ) / ( dlnronrw * drw * drw * dkw) + dtu / dtl );
  
  y[0] = dnum / dden;
}

void flowrateCalib1D(double *x, double *y)
{
  Double_t drw = x[1], dr  = 25050;
  Double_t dtu = 89335, dtl = 89.55;
  Double_t dhu = 1050, dhl = x[0];
  Double_t dl  = x[2], dkw = 10950;

  Double_t dnum = 2.0 * TMath::Pi() * dtu * ( dhu -dhl);
  Double_t dlnronrw = TMath::Log( dr / drw);
  Double_t dden = dlnronrw * ( 1.0 +  ( 2.0 * dl * dtu ) / ( dlnronrw * drw * drw * dkw) + dtu / dtl );

  y[0] = dnum / dden;
}

void flowrateCalib2D(double *x, double *y)
{
    Double_t drw = x[2], dr  = 25050;
    Double_t dtu = 89335, dtl = 89.55;
    Double_t dhu = x[0], dhl = x[1];
    Double_t dl  = x[3], dkw = 10950;

    Double_t dnum = 2.0 * TMath::Pi() * dtu * ( dhu - dhl );
    Double_t dlnronrw = TMath::Log( dr / drw);
    Double_t dden = dlnronrw * ( 1.0 +  ( 2.0 * dl * dtu ) / ( dlnronrw * drw * drw * dkw) + dtu / dtl );

    y[0] = dnum / dden;
}

void flowrateModelnoH(double *x, double *y)
{
  Double_t drw = x[0], dr  = 25050;
  Double_t dtu = 89335, dtl = 89.55;
  Double_t dl  = x[1], dkw = 10950;

  Double_t dnum = 2.0 * TMath::Pi() * dtu;
  Double_t dlnronrw = TMath::Log( dr / drw);
  Double_t dden = dlnronrw * ( 1.0 +  ( 2.0 * dl * dtu ) / ( dlnronrw * drw * drw * dkw) + dtu / dtl );

  y[0] = dnum / dden;
}

void transf(double *x, double *res)
{
    res[0] = 1050 - x[0];  // simply H_l = \theta - H_u
}

void Toy(double *x, double *y)
{
    double in=x[0], t0=x[1], t1=x[2];    
    y[0] = t0  + t1 *in;
}

//! The surrogate \em flow \em rate model (polynomial regression) proposed Ho \em and Xu (2000) is
/*!
  The polynomial regression models is given by the formula :
  \f[
  \hat{log(y)}  = 4.156 + 1.9903(\log{r_{\omega}}+2.3544) - 0.0007292(L-1400) - .003554(H_l - 760) \\
  \f]
  \f[
  \hspace*{5cm} + .0035068(H_u-1050) + .000090868(K_{\omega}-10950) - .00087286 (\log{(r)} -8.8914) \\
  \f]
  \f[
  + .000015325 (H_u-1050) (H_l - 760) + .00000026487 (L-1400)^2 \\
  \f]
  \f[
  \hspace*{5cm} - .0000071759 (H_l - 760)^2 - .0000068021 (H_u-1050)^2 \\
  \f]
  
  \param x (double *) the input vector in the order :
  \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
  \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
  \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
  \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
  \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
  \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
  \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
  \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
  \param y (double *) the output vector

  \see flowrateModel
  
 */

void HoXuSurrogateModel(double *x, Double_t *y)
{
  Double_t drw = x[0], dr  = x[1];
//   Double_t dtu = x[2], dtl = x[3];
  Double_t dhu = x[4] - 1050.0, dhl = x[5] -   760.0;
  Double_t dl  = x[6] - 1400.0, dkw = x[7] - 10950.0;

  Double_t yHoXu = 4.1560 + 1.9903 * (TMath::Log(drw) + 2.3544);
  yHoXu -= .0007292 * dl;
  yHoXu -= .003554 * dhl;
  yHoXu += .0035068 * dhu;
  yHoXu += .000090868 * dkw;
  yHoXu += .000015325 * dhu * dhl;
  yHoXu += .00000026487 * dl * dl;
  yHoXu -= .0000071759 * dhl * dhl;
  yHoXu -= .0000068021 * dhu * dhu;
  yHoXu -= .00087286 * (TMath::Log(dr) - 8.8914);

  y[0] = TMath::Exp(yHoXu);
  
}

//! The rosenbrock Function ( a=10.0 and b=1.0)
/*!
  The rosenbrock function is defined by the equation 
  \$[ f(x_1, x_2) = a * ( x_2 - x_1^2 )^2 + b * ( 1.0 - x_1 )^2 \$]
  with a=10.0 and b=1.0
 */
void rosenbrock(double *x, Double_t *y)
{
  Double_t dx1 = x[0], dx2 = x[1];
  Double_t da=10.0, db=1.0;
  
  y[0] = da * (dx2-dx1*dx1)*(dx2-dx1*dx1)+db*(1.0-dx1)*(1.0-dx1);
  
}


//! Ishigami function
void FctIshigami(Double_t *x, Double_t *y)
{
    Double_t u1, u2, u3, s1, s2, s3, a, b;
    a = 7.;
    b = 0.1;
    u1 = x[0];
    u2 = x[1];
    u3 = x[2];
    s1 = TMath::Sin(u1);
    s2 = TMath::Sin(u2);
    s3 = TMath::Sin(u3);
    y[0] = s1 + a * s2 * s2 + b * u3 * u3 * u3 * u3 * s1;
}

void ModelPasture(Double_t *x, Double_t *y)
{
  Double_t theta1=69.95, theta2=61.68, theta3=-9.209, theta4=2.378;
  
  y[0] = theta1;
  y[0] -= theta2* TMath::Exp( -1.0 * TMath::Exp( theta3 + theta4 * TMath::Log(x[0])));
}

//! The hollow bar optimisation problem
/*!
We have a hollow metalic bar of square section. The inner length is y and the 
outter length is x. The problem is to choose x and y in [0.0, 1.0] such that:
- the weight (i.e. the area "x^2 - y^2") is minimal;
- the deformation (i.e. "1/(x^4-y^4)") is minimal.
In addition, the thickness (i.e. "x-y") of the bar can't be smaller than 0.4.
*/
void barAllCost(double *param, double *res)
{
  double x = param[0];
  double y = param[1];

  /* thickness constraint */
  res[0] = x-y; 
  /* section criterion */
  //res[1] = 10 * (x*x - y*y);
  res[1] = x*x - y*y;
  /* deformation criterion */
  res[2] = 1. / (  x*x*x*x - y*y*y*y + 1.e-66);
}

//! Dummy operation function
void operation(double *x, Double_t *y)
{
  Double_t x1 = x[0], x2 = x[1];
  // product
  y[0] = x1 * x2;
  // divide
  y[1] = ((x2!=0) ? x1 / x2 : -999);
  // sum
  y[2] = x1 + x2;
  // diff
  y[3] = x1 - x2;  
}

//! Generating a correlation matrix
/*!
  
 */
void GenCorr(TMatrixD *inCorr, bool correlated=true, bool linearOutput=false )
{

    int _nX = inCorr->GetNrows();
    gRandom->SetSeed(__rdtsc());
    //Define a randomly filled matrix 
    TMatrixD A(_nX,_nX);
    for (int i=0; i<_nX; i++)
        for (int j=0; j<_nX; j++)
            A(i,j)=gRandom->Gaus(0,1);

    // Compute AA^T and normalise it to get "covariance matrix"
    TMatrixD Gamma(A,TMatrixD::kMultTranspose,A);
    Gamma*=1./_nX;

    // Get the inverse of the diagonal matrix to do as if this was 1/sqrt(variance)
    TMatrixD Sig(_nX,_nX);
    for(int i=0;i<_nX;i++) Sig(i,i)=1./sqrt( Gamma(i,i) );

    // Compute the input correlation matrix 
    (*inCorr) = Sig * Gamma *Sig;
    if( !correlated )
        inCorr->UnitMatrix();
    double varY=inCorr->Sum();

    if( !linearOutput )
        return ;
    else
    {
        // Proper correlation, output included
        TMatrixD Corr(_nX+1,_nX+1);
        Corr.UnitMatrix();
        // putting already defined input
        Corr.SetSub(0,0,*inCorr);
        // Adjust the covariance of the output wrt to all inputs
        for(unsigned int i=0; i<_nX; i++)
        {
            double value=0;
            for(unsigned int j=0; j<_nX; j++)
                value+=(*inCorr)(i,j);
            Corr(_nX,i) = Corr(i,_nX) = value / sqrt(varY);
        }
        (*inCorr).ResizeTo(_nX+1,_nX+1);
        (*inCorr) = Corr;
        return ;
    }
}


/****************** End of User Function **********************/
