//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TCodeFMU.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/09/10 at 12:10:53

 \brief Interface file for the class URANIE::Launcher::TCodeFMU
 */
/**

 * @class URANIE::Launcher::TCodeFMU
 *
 * @brief A virtual class for the FMI/FMU encapsulation
 *
 * \li step \b init  ( \see \ref test_simulate_fmi2_cs ) :
 *       - Create the callback
 *       - Create the FMU pointer
 *       - Search the ReferenceValue in the FMU object for each input
 *       - Case of Log :
 *           - Print all the input start values
 * \li step \b preTraitment :
 *       - Set the input current values into the FMU object via setFMUReal() ..
 *
 * \li step \b run :
 *       - Simulate the FMU Object via the loop (time_begin, time_step, _time_end) AND fmi2_import_do_step function
 *
 * \li step \b postTraitment :
 *       - Catch the output values via the fmi2_import_get_real() function
 *
 * \li step \b terminate :
 *       - Delete all the pointer (callbacks, FMU, ...)
 *
 * @see https://fmi-standard.org/
 *
 */
#ifndef TCODEFMU_H
#define TCODEFMU_H

// ROOT

//URANIE
#include "TCode.h"
#include "TInputFileFMU.h"
#include "TOutputFileFMU.h"

// FMI
#ifdef WIN32
#include <Windows4Root.h>
#endif
#include "fmilib.h"
#include <JM/jm_portability.h>

#include <string>

namespace URANIE {
namespace Launcher {

class TCodeFMU : public TCode
{
public:

    //! The list of log level
    enum ELogLevel_t {
        log_level_nothing = 0,        /** \brief Must be first in this enum. May be usefull in application relying solely on jm_get_last_error() */
        log_level_fatal,              /** \brief Unrecoverable errors */
        log_level_error,              /** \brief Errors that may be not critical for some FMUs. */
        log_level_warning,            /** \brief Non-critical issues */
        log_level_info,               /** \brief Informative messages */
        log_level_verbose,            /** \brief Verbose messages */
        log_level_debug,              /** \brief Debug messages. Only enabled if library is configured with FMILIB_ENABLE_LOG_LEVEL_DEBUG */
        log_level_all                 /** \brief  Must be last in this enum. */
    };

    // Constructors/Destructors
    //  


    /*!
     * \brief Default constructor
     */
    TCodeFMU ();
    /*!
     * \brief constructor with a  name and a title
     * \param name
     * \param title
     */
    TCodeFMU(const TString &name, const TString &title);
    /*!
     * \brief Constructor from TDataServer and Command as string
     * \param tds
     * \param sFMUfile
     */
    TCodeFMU(URANIE::DataServer::TDataServer *tds, TString sFMUfile, Option_t *option ="");

    /*!
     * \brief Copy constructor
     * \param mycode
     */
    TCodeFMU(URANIE::Launcher::TCode &mycode);

    /**
     * Empty Destructor
     */
    virtual ~TCodeFMU ();

    // Static Public attributes
    //  

    // Public attributes
    //  
    //---------------------------------------------
    /** @name Encapsulates a code
     *
     **/
    //@{
    /*!
     * \brief Init phase of the code
     * \param option
     */
    virtual void init(Option_t *option = "");

    /*!
     * \brief Run the code
     * \param option
     */
    virtual void run(Option_t *option = "");

    /*!
     * \brief Terminate phase of the code
     * \param option
     */
    virtual void terminate(Option_t *option = "");

    void initIFirst(Option_t *option = "");

    //@}


    // Public attribute accessor methods
    //  
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    /*!
     * \brief Set the value of _sFMUFile
     *
     * The FMU filename which contains the XML File and the library.
     *
     * \param new_var the new value of _sFMUFile
     */
    void setFMUFile (TString new_var, Option_t *option = "");
    /*!
     * \brief Get the value of the _sFMUFile attribute
     *
     * The FMU filename which contains the XML File and the library.
     *
     * \return the value of _sFMUFile
     */
    TString getFMUFile () {
        FUNCT_TRACE();
        return _sFMUFile;
    }

    Double_t getStartSimulation() {
        FUNCT_TRACE();
        return _dStart;
    }
    void setStartSimulation(Double_t dval) {
        FUNCT_TRACE();
        _dStart = dval;
        _bStart = kTRUE;
    }
    Double_t getStoptSimulation() {
        FUNCT_TRACE();
        return _dStop;
    }
    void setStopSimulation(Double_t dval) {
        FUNCT_TRACE();
        _dStop = dval;
        _bStop = kTRUE;
    }
    Double_t getStepSimulation() {
        FUNCT_TRACE();
        return _dStep;
    }
    void setStepSimulation(Double_t dval) {
        FUNCT_TRACE();
        _dStep = dval;
        _bStep = kTRUE;
    }

    Double_t getToleranceSimulation() {
        return _dTolerance;
    }
    void setToleranceSimulation(Double_t dval) {
        FUNCT_TRACE();
        _dTolerance = dval;
        _bTolerance = kTRUE;
    }
    void setStartStopSimulation(Double_t dstart, Double_t dstop) {
        FUNCT_TRACE();
        try {
            if ( dstart < dstop ) {
                setStartSimulation(dstart);
                setStopSimulation(dstop);
            } else {
                throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__, __LINE__,
                    Form("TcodeFMU::%s Method\n The start time[%f] is greater than stop  [%e] time", __func__, dstart, dstop));
            }
        } catch (URANIE::Exceptions::UExceptions& ue) {
            ue.printMessage();
            if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
                exit(-1);

        }
    }


    //@}

    // Public FMI functions
    //
    //---------------------------------------------
    /** @name methods associated to the FMI Library
     **/
    //@{


    /*!
     * \brief do_exit
     * \param code
     */

    void do_exit(int code)
    {
            printf("Press 'Enter' to exit\n");
            /* getchar(); */
            exit(code);
    }

    //@}

    // Public attribute accessor methods
    //  

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    /*!
     * \brief Force the log
     */
    void setLog() {
        _blog = kTRUE;
    }
    /*!
     * \brief Force to desable the log
     */
    void unsetLog() {
        _blog = kFALSE;
    }
    /*!
     * \brief Swith the status of the log
     */
    void changeLog() {
        _blog = _blog ? kFALSE : kTRUE;
    }
    /*!
     * \brief Get the status of the log
     * \return
     */
    Bool_t getLog() {
        return _blog;
    }

    /*!
     * \brief Set the Log Level for the FMI Library
     * \see ELogLevel_t
     *
     * \param theLogLevel
     */
    void setLogLevel(ELogLevel_t theLogLevel = log_level_info) {
        _logLevel = theLogLevel;
    }

    /*!
     * \brief Get the Log Level in the FMI Library
     * \return
     */
    ELogLevel_t getLogLevel() {
        return _logLevel;
    }

    /*!
     * \brief Print the log of the object/class
     * \param option
     */
    virtual void printLog(Option_t *option = "");
    //@}

protected:

    // Static Protected attributes
    //  

    // Protected attributes
    //  

public:


    // Protected attribute accessor methods
    //  

    //---------------------------------------------
    /** @name protected attributes
     *
     **/
    //@{
protected:
    TString _sFMUFile;                  ///< The FMU filename
    TString _sFMUWorkingDirectory;      ///< the FMU import directory
    ELogLevel_t _logLevel;              ///< The log Level for FMI Library
    Double_t _dStart;                   ///< The start time for the simulation
    Double_t _dStop;                    ///< The stop time for the simulation
    Double_t _dStep;                    ///< The step time for the simulation
    Double_t _dTolerance;               ///< The tolerance for the simulation
    Bool_t   _bStart;                   ///< boolean to known if _dstart is defined by user
    Bool_t   _bStop;                    ///< boolean to known if _dstop is defined by user
    Bool_t   _bStep;                    ///< boolean to known if _dstep is defined by user
    Bool_t   _bTolerance;               ///< boolean to known if _dTolerance is defined by user
    Bool_t   _binitIFirst;              ///< boolean to known if passed into the initIFirst step

    // FMI structures
    fmi_version_enu_t _sVersion;        ///< The FMU version
    fmi_import_context_t* _context_FMI; ///< The FMU context
    fmi1_import_t* _fmi1Import;         ///< The FMU 1.0 Object
    fmi2_import_t* _fmi2Import;         ///< The FMU 2.0 Object

    // FMI General informations
    TString _sGenerationTool;            ///< The FMU generation tool
    TString _sGenerationDateTime;        ///< The FMU generation date and time
    Double_t _dFMUStart;                 ///< The start time for default experiment as specified in the XML file
    Double_t _dFMUStop;                  ///< The stop time for default experiment as specified in the XML file
    Double_t _dFMUStep;                  ///< The step time for default experiment as specified in the XML file
    Double_t _dFMUTolerance;             ///< The tolerance for default experiment as specified in the XML file

    //@}

public:


    // Protected attribute accessor methods
    //  

protected:


private:

    // Static Private attributes
    //  

    // Private attributes
    //  

public:


    // Private attribute accessor methods
    //  

private:
    /*!
     * \brief treat the soption for the class
     * \param soption
     */
    void treatOption(TString soption);

public:


    // Private attribute accessor methods
    //  

private:

    ClassDef(URANIE::Launcher::TCodeFMU, ID_LAUNCHER)


};

} // End of namespace Launcher
} // End of namespace URANIE

#endif // TCODEFMU_H
// End of File

