//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TEcCluster lun aoû 4 14:19:19 CEST 2008
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TEcCluster.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon august 4 14:19:19 CEST 2008

 \brief Implementation of the class URANIE::Launcher::TEcCluster

 */

#include "TEcCluster.h"

#include "TSystem.h"
#include <vector>
#include <TRegexp.h>
#include "TDataServer.h"

namespace URANIE
{
namespace Launcher
{


// Constructor
TEcCluster::TEcCluster() :
    TCluster("Ec CLUSTER", "none ", 1)
{
    if (_blog)
        cout << " ** Constructor TEcCluster::TEcCluster [" << GetName()
             << "] Title[" << GetTitle() << "]" << endl;
    init();
}
//Constructor Remote

#ifdef WITH_LIBSSH        
TEcCluster::TEcCluster(URANIE::DataServer::TDataServer * tds, TCode * mycode) :
    TCluster(tds,mycode) {

    _tds = tds;
    _code= mycode;
}

TEcCluster::TEcCluster(TCode * mycode) :
    TCluster(mycode) {
    _code= mycode;
}
#endif

TEcCluster::~TEcCluster()
{
    if (_blog)
        cout << " ** Destroy TEcCluster [" << GetName() << "] Title["
             << GetTitle() << "]" << endl;
}

///////////////////////////////////////////////////////////
//  Construct and return the list of node
///////////////////////////////////////////////////////////
void TEcCluster::init()
{
}

#ifdef WITH_LIBSSH
// Ec-access commands
Int_t TEcCluster::sendHeader()
{
    TString filename = gSystem->ConcatFileName(_sRemoteWorkingDirectory,_sSplitName);
    //"ecaccess-file-mkdir" Will create the folder if it doesn't exists'
    GetFromPipe("ecaccess-file-mkdir " + _sGateWay+":"+_sRemoteWorkingDirectory );
    GetFromPipe(TString("ecaccess-file-put -umask 755 ")  + gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName) + " "  + _sGateWay + ":" + filename );
    return 0;
}
Int_t TEcCluster::sendCompiledBinary(list<TString> filenames)
{
    TString remote = gSystem->ConcatFileName(_sRemoteWorkingDirectory,"bin");
    //"ecaccess-file-mkdir" Will create the folder if it doesn't exists'
    GetFromPipe("ecaccess-file-mkdir " +  _sGateWay + ":" +_sRemoteWorkingDirectory );
    GetFromPipe("ecaccess-file-mkdir " +  _sGateWay + ":" + remote  );
    auto it = filenames.begin();
    while ( it != filenames.end() )
    {

        TString folder_path = gSystem->ConcatFileName(_sWorkingDirectory,"bin/"+(*it));
        _sSshOutput = GetFromPipe("ecaccess-file-put -umask 755 " + folder_path + " " + _sGateWay + ":" +remote+"/"+(*it) );
        it++;
    }
    _sUserBinary = kTRUE;
    return 0;
}

Int_t TEcCluster::sendInputFiles(Int_t start, Int_t end)
{
    TString remoteFolder;
    TString localFolder;
    TString remoteFile;
    TString localFile;
    for (Int_t indx = start; indx < end; indx++)
    {
        localFolder = gSystem->ConcatFileName(_sWorkingDirectory,
                                              TString("UranieLauncher_") + std::to_string(indx + 1).c_str());
        remoteFolder = gSystem->ConcatFileName(_sRemoteWorkingDirectory,
                                               TString("UranieLauncher_") + std::to_string(indx + 1).c_str());
        GetFromPipe("ecaccess-file-mkdir " + _sGateWay + ":" +remoteFolder );
        for (int ifile=0; ifile<_code->getNInputFiles(); ifile++)
        {
            remoteFile= gSystem->ConcatFileName(remoteFolder,basename(_code->getInputFile(ifile)->getFileName()));
            localFile = gSystem->ConcatFileName(localFolder,basename(_code->getInputFile(ifile)->getFileName()));
            GetFromPipe("ecaccess-file-put -umask 755 " + localFile + " " + _sGateWay + ":" +remoteFile);
        }
    }
    return 0;

}
Int_t TEcCluster::retrieveInputFiles(const char * launcherFolder)
{
    TString remoteFolder = Form("%s/%s", _sRemoteWorkingDirectory.Data(), launcherFolder);
    TString localFolder = Form("%s/%s", _sWorkingDirectory.Data(), launcherFolder);

    GetFromPipe("ecaccess-file-mget -force " + _sGateWay + ":"+remoteFolder + "/* " + localFolder);

    return 0;
}

TString TEcCluster::submitJob(TString command)
{

    TString jobid = GetFromPipe(command,true);
    _sSshOutput = jobid;
    return jobid;
}



void TEcCluster::selectHeader(TString SkeletonHeaderName,TString localHeaderFolderPath)
{

    if (!localHeaderFolderPath.EndsWith("/"))
        localHeaderFolderPath.Append("/");
    _slocalHeaderFolderPath=localHeaderFolderPath;

    if (getClusterName().Length() > 0)
    {
        if (SkeletonHeaderName.EqualTo("") || SkeletonHeaderName.Length() <= 0 || SkeletonHeaderName.EqualTo("./") )
        {
            if (getClusterName().EqualTo("cca"))
            {
                cout << "A predefined cluster name is found: " ;
                SkeletonHeaderName = TString("cca_pbs_skeleton.in");
            }
            else if (getClusterName().EqualTo("ccb")) {
                cout << "A predefined cluster name is found: ";
                SkeletonHeaderName = TString("ccb_pbs_skeleton.in");
            }
            else
            {
                cout << getClusterName() <<endl;
                cout << "invalid clustername-skeleton combination" << endl;
                exit(-1);
            }
        }
        else
        {
            cout << "Custom skeleton selection: " << endl;
            cout << SkeletonHeaderName << endl ;
        }
    }
    else{
        cout << "Error, first you must set a cluster name" << endl ;
        exit(-1);
    }
    cout << SkeletonHeaderName << endl;
    setSkeleton(SkeletonHeaderName);
}

void TEcCluster::generateHeaderSplit(Option_t *option)
{
    TString soption =option;
    soption.ToLower();
    std::vector<TRegexp> regular;
    TRegexp regularStart("start=[0-9]+");
    TRegexp regularEnd("end=[0-9]+");
    TString outStart = soption(regularStart);
    TString outEnd = soption(regularEnd);
    TString start = outStart.Remove(0, 6);
    TString end = outEnd.Remove(0, 4);

    _sSplitName=start+"_to_"+end+"_"+getOutputHeaderName();
    //std::ifstream infile();
    substituteDirectives(gSystem->ConcatFileName(_slocalHeaderFolderPath,_sSkeletonHeader),gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName),  _sListClusterDirectives);
    std::ofstream os(gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName), ios::app );

    TString workingDirectory;

    if  (_sRemoteWorkingDirectory.BeginsWith('.'))
        os << "workingDirectory=\"\"$HOME\"/" <<  _sRemoteWorkingDirectory.Data()  << "\"" << endl;
    else
        os << "workingDirectory=\"" <<  _sRemoteWorkingDirectory.Data() << "\"" <<  endl;
    if (_sUserBinary)
    {
        os << "#Configuration of UserBinary" << endl;
        os << "export PATH=$PATH:$workingDirectory/bin" <<endl;
        os << "export LIBRARY_PATH=${LIBRARY_PATH:+LIBRARY_PATH:}$workingDirectory/lib" <<endl;
        os << "export CPATH=${CPATH:+CPATH:}$workingDirectory/include" <<endl;
        os << "export LD_LIBRARY_PATH=${LD_LIBRARY_PATH:+LD_LIBRARY_PATH:}$workingDirectory/lib" <<endl;
        os << "#END Configuration" <<endl;
    }
    os << "set -x" << endl;
    os << "sdir=\"" << "nothing" << "\"" <<  endl;
    os << "fixedRetries=" << _sRetries << endl;
    os << "ipat=" << start << endl;
    os << "ninput=" << end << endl;
    os << "command=\""<<_code->getCommand() << "\"" << endl;
    os << "ncores=" << _nAllocatedCores  << endl;
    os << "SECONDS=0" << endl;
    os << "for (( ipat=$ipat; ipat<$ninput; ipat++ ))" << endl;
    os << "do" << endl;
    os << "     sdir=\"\"$workingDirectory\""<< "/UranieLauncher_" << "$(($ipat+1))" << "\"" <<  endl;
    os << "     cd "  << "$sdir" << endl;
    os << "     echo " << "$PWD" << endl;
    os << "     fileExist=$(ls -l | grep '\\.in\\|\\.dat\\|\\.4' | wc -l)" << endl;
    os << "     retries=$fixedRetries" << endl;
    os << "     while [ $fileExist -eq 0 ]" << endl;
    os << "     do" << endl;
    os << "         echo \"Waiting 1m for recv an input file, retries left:\" ";
    os << "     \"$retries\"" << endl;
    os << "     ((retries--))" << endl;
    os << "     sleep 1m"  << endl;
    os << "     cd "  << "$sdir" << endl;
    os << "     echo " << "$PWD" << endl;
    os << "     fileExist=$(ls -l | grep '*.in \\| *.dat \\| *.4' | wc -l)" << endl;
    os << "     if [ $retries -eq 0 ]; then" << endl;
    os << "         break; fi" << endl;
    os << "     done " << endl;
    //        os << "     eval " <<  " $command" << " &" << endl;
    os << " $command" << endl;
    os << "     sleep 0.15s"  << endl;
    os << "done" << endl;
    os << "wait" << endl;
    os << "echo \"Completed in $SECONDS seconds\" " << endl;
    os.close();
}
void TEcCluster::generateHeader(Option_t *option)
{

    _sConnected = kTRUE;
    generateHeaderSplit(option);
    if (sendHeader()  != 0)
    {
        cout << "Error" << endl;
        exit(-1);
    }
}
TString TEcCluster::getJobStatusCommand(list<pair<TString, pair<int, int> > > JobList)
{
    TString JobCommand ="'";
    auto aJob = JobList.begin();
    while (aJob != JobList.end())
    {

        JobCommand+=(*aJob).first+"\\|";
        aJob++;
    }
    JobCommand.Chop();
    JobCommand.Chop();
    JobCommand = JobCommand + "'";

    return "ecaccess-job-list | grep " + JobCommand;
}
TString TEcCluster::checkJobStatus(TString JobId, TString Command)
{
    _sSshOutput = GetFromPipe(Command,true);
    TPRegexp COMPLETED("\\b"+JobId+"\\b.*\\bDONE\\b.*");
    TPRegexp QUEUING("\\b"+JobId+"\\b.*\\b(INIT|RETR|STDBY|WAIT)\\b.*");
    TPRegexp FAILED("\\b"+JobId+"\\b.*\\bSTOP\\b.*");
    TPRegexp RUNNING("\\b"+JobId+"\\b.*\\bEXEC\\b.*");
    if (_sSshOutput.Contains(COMPLETED))
    {
        return "COMPLETED";
    }
    else if (_sSshOutput.Contains(RUNNING))
    {
        return "RUNNING";
    }
    else if (_sSshOutput.Contains(QUEUING))
    {
        return "QUEUING";
    } else if (_sSshOutput.Contains(FAILED))
    {
        return "FAILED";
    }
    else
    {
        return "FAILED";
    }
}
#endif

void TEcCluster::getCommandPrefix(char *mpiexe, char* procs)
{
    UNUSED(mpiexe,procs);
    cout << "Not Implemented";
}

TString TEcCluster::GetFromPipe(const char *command,bool output)
{
    TString out;
    FILE *pipe = gSystem->OpenPipe(command , "r");
    //cout << command << endl;
    if (!pipe) {
        SysError("GetFromPipe", "cannot run command \"%s\"", command);
        return out;
    }
    if (output)
    {
        TString line;
        while (line.Gets(pipe)) {
            if (out != "")
                out += "\n";
            out += line;
        }
    }
    Int_t r = gSystem->ClosePipe(pipe);
    return out;

}

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TEcCluster::printLog(Option_t *option)
{
    cout << "TEcCluster::printLog[" << option << "]" << endl;
    TCluster::printLog(option);
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

