//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileXML.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Sep  6 10:46:46 2011

 \brief Implementation of the class URANIE::Launcher::TInputFileXML

 */
#include <iostream>
#include<fstream>
using std::cout;
using std::endl;

// ROOT
#include "TMath.h"

// URANIE
#include "UExceptions.h"
#include "TInputFileXML.h"

ClassImp (URANIE::Launcher::TInputFileXML)

namespace URANIE
{
namespace Launcher
{

TInputFileXML::TInputFileXML(TString str) :
        TInputFile(str)
{
    if (_blog)
        cout << "constructing TInputFileXML file[" << str << "]" << endl;
    SetTitle(Form("TInputFileXML with name[%s]", str.Data()));
    setSeparatorCharacter(" \t\r\n");
    _sstrbeg = _sstrend = "";
    _svecbeg = _svecend = ""; _svecdel = " ";
}

TInputFileXML::~TInputFileXML()
{
    if (_blog)
        cout << "destroying TInputFileXML" << endl;
}

void TInputFileXML::searchIndexeAttribute(URANIE::DataServer::TAttribute* att)
{
    if (_blog)
    {
        cout << endl
                << " ** TInputFileXML::searchIndexeAttribute(URANIE::DataServer::TAttribute*) method\t"
                << endl;
        cout << " ** File[" << getFileName() << "] Attribute[" << att->GetName()
                << "]" << endl;
        cout << " ** Nothing to do" << endl << endl;
    }
}

void TInputFileXML::substitue(Int_t &n, Double_t *val, Int_t *index)
{
    if (_blog)
        cout << endl
                << " ** TInputFileXML::substitue(Int_t &n, Double_t *val, Int_t *index) method\t"
                << endl;

    // Open the XSLT file
    TString sXSLTFile = Form("uranie_substitute_xmlfile_%s.xslt",
            gSystem->BaseName(_sFileName.Data()));
    if (_blog)
        cout << "** sXSLTFile[" << sXSLTFile << "] ****" << endl;

    std::ofstream fileXSLT(sXSLTFile.Data(), std::ios::trunc);
    fileXSLT << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>" << endl;
    fileXSLT
            << "<xsl:stylesheet xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" version=\"1.0\">"
            << endl;
    fileXSLT << endl;

    // Current value
    Double_t dValue = 0.0;

    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << " ** " << i << "/" << getAttributesNumber() << " -- name["
                    << _attributes[i]->GetName() << "]" << endl;
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
                _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
                KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileXML::substitue\n ** PWD[" << gSystem->pwd()
                        << "]\n ** sfilename[" << sfilename << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            TString sformattosubstitute = TString((*it)->_sFormat);

            switch ((*it)->_sFileType)
            {
            case URANIE::DataServer::TAttributeFileKey::kXMLAttribute:
            {

                if (index != NULL)
                {
                    dValue = val[index[n - 1]];
                    if (_blog)
                        cout << " ** dvalue [" << dValue
                                << "] with index and type kXMLAttribute"
                                << endl;
                }
                else
                {
                    dValue = val[n];
                    if (_blog)
                        cout << " ** dvalue [" << dValue
                                << "] with type kXMLAttribute" << endl;
                }

                fileXSLT << "  <xsl:template match=\"" << (*it)->_sKey << "\">"
                        << endl;
                fileXSLT << "    <xsl:attribute name=\""
                        << getAttributeInKey((*it)->_sKey) << "\">" << endl;

                if (sformattosubstitute.CompareTo("%d", TString::kExact) == 0)
                {
                    Int_t nval = (Int_t)(TMath::Ceil(dValue));
                    fileXSLT << "      <xsl:value-of select=\"" << nval
                            << "\"/>" << endl;
                }
                else
                    fileXSLT << "      <xsl:value-of select=\"" << dValue
                            << "\"/>" << endl;

                fileXSLT << "    </xsl:attribute>" << endl;
                fileXSLT << "  </xsl:template>" << endl;
                fileXSLT << endl;
            }
                break;
            case URANIE::DataServer::TAttributeFileKey::kXMLField:
            {
                if (index != NULL)
                {
                    dValue = val[index[n - 1]];
                    if (_blog)
                        cout << " ** dvalue [" << dValue
                                << "] with index and type kXMLField" << endl;
                }
                else
                {
                    dValue = val[n];
                    if (_blog)
                        cout << " ** dvalue [" << dValue
                                << "] with type kXMLField" << endl;
                }
                fileXSLT << "  <xsl:template match=\"" << (*it)->_sKey << "\">"
                        << endl;
                fileXSLT << "    <xsl:copy>" << endl;
                fileXSLT << "      <xsl:apply-templates select=\"@*\"/>"
                        << endl;

                if (sformattosubstitute.CompareTo("%d", TString::kExact) == 0)
                {
                    Int_t nval = (Int_t)(TMath::Ceil(dValue));
                    fileXSLT << "      <xsl:value-of select=\"" << nval
                            << "\"/>" << endl;
                }
                else
                    fileXSLT << "      <xsl:value-of select=\"" << dValue
                            << "\"/>" << endl;

                fileXSLT << "    </xsl:copy>" << endl;
                fileXSLT << "  </xsl:template>" << endl;
                fileXSLT << endl;
            }
                break;
	    default:
	    {
		try{		    
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,"TCodeTInputFileXML::The AttributeFileKey Format not compatible with XML file");
		    
		} catch (URANIE::Exceptions::UErrorExceptions& ue)
		{
		    ue.printMessage();
		    throw ue;
		}        		
		break;
            }
	    }
        }
        n++;
    }
    fileXSLT << "  <!-- Copy all the rest of the file -->" << endl;
    fileXSLT << "  <xsl:template match=\"node()|@*\">" << endl;
    fileXSLT << "    <xsl:copy>" << endl;
    fileXSLT << "      <xsl:apply-templates select=\"node()|@*\"/>" << endl;
    fileXSLT << "    </xsl:copy>" << endl;
    fileXSLT << "  </xsl:template>" << endl;

    fileXSLT << "" << endl;

    fileXSLT << "</xsl:stylesheet>" << endl;
    fileXSLT << endl;
    fileXSLT.close();

    // Transform the XML file by xsltproc
    TString xsltProc;
#ifdef WIN32
    xsltProc = TString("xsltproc.exe");
#else
    xsltProc = TString("xsltproc");
#endif
    TString sCmd = Form("%s -o %s %s %s", xsltProc.Data(),
            gSystem->BaseName(_sFileName.Data()), sXSLTFile.Data(),
            _sFileName.Data());
    if (_blog)
        cout << "** sCmd[" << sCmd << "]" << endl;

    gSystem->Exec(sCmd.Data());
}

void TInputFileXML::substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index)
{
    if (_blog)
        cout << endl
                << " ** TInputFileXML::substitue(Int_t &n, Double_t *val, Int_t *index) method\t"
                << endl;

    // Open the XSLT file
    TString sXSLTFile = Form("uranie_substitute_xmlfile_%s.xslt",
            gSystem->BaseName(_sFileName.Data()));
    if (_blog)
        cout << "** sXSLTFile[" << sXSLTFile << "] ****" << endl;

    ofstream fileXSLT(sXSLTFile.Data(), std::ios::trunc);
    fileXSLT << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>" << endl;
    fileXSLT
            << "<xsl:stylesheet xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" version=\"1.0\">"
            << endl;
    fileXSLT << endl;

    // Current value
///    Double_t dValue = 0.0;
    TString beg, end;
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << " ** " << i << "/" << getAttributesNumber() << " -- name["
                    << _attributes[i]->GetName() << "]" << endl;
	getBegEndProperties( _attributes[i], beg, end);
	
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
                _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
                KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileXML::substitue\n ** PWD[" << gSystem->pwd()
                        << "]\n ** sfilename[" << sfilename << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            TString sformattosubstitute = TString((*it)->_sFormat);

            switch ((*it)->_sFileType)
            {
            case URANIE::DataServer::TAttributeFileKey::kXMLAttribute:
            {

                fileXSLT << "  <xsl:template match=\"" << (*it)->_sKey << "\">"
                        << endl;
                fileXSLT << "    <xsl:attribute name=\""
                        << getAttributeInKey((*it)->_sKey) << "\">" << ent->at(((index==NULL) ? n : index[n-1]))->dumpStr( _svecdel, beg, end, (*it)->_sFormat)
			 << "</xsl:attribute>" << endl;

		//fileXSLT << "      <xsl:value-of select=\"" << ent->at(((index==NULL) ? n : index[n-1]))->dumpStr( _svecdel, beg, end, (*it)->_sFormat)<< "\"/>" << endl;

                //fileXSLT << "    </xsl:attribute>" << endl;
                fileXSLT << "  </xsl:template>" << endl;
                fileXSLT << endl;
            }
                break;
            case URANIE::DataServer::TAttributeFileKey::kXMLField:
            {

                fileXSLT << "  <xsl:template match=\"" << (*it)->_sKey << "\">"
                        << endl;
                fileXSLT << "    <xsl:copy>" << endl;
                // fileXSLT << "      <xsl:apply-templates select=\"@*\"/>"
                //         << endl;

                // fileXSLT << "      <xsl:value-of select=\"" << ent->at(((index==NULL) ? n : index[n-1]))->dumpStr( _svecdel, beg, end, (*it)->_sFormat)<< "\"/>" << endl;

		fileXSLT << "     " << ent->at(((index==NULL) ? n : index[n-1]))->dumpStr( _svecdel, beg, end, (*it)->_sFormat) << endl;

                fileXSLT << "    </xsl:copy>" << endl;
                fileXSLT << "  </xsl:template>" << endl;
                fileXSLT << endl;
            }
                break;
	    default:
	    {
		try{		    
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,"TCodeTInputFileXML::The AttributeFileKey Format not compatible with XML file");
		    
		} catch (URANIE::Exceptions::UErrorExceptions& ue)
		{
		    ue.printMessage();
		    throw ue;
		}        		
		break;
            }
            }
        }
        n++;
    }
    fileXSLT << "  <!-- Copy all the rest of the file -->" << endl;
    fileXSLT << "  <xsl:template match=\"node()|@*\">" << endl;
    fileXSLT << "    <xsl:copy>" << endl;
    fileXSLT << "      <xsl:apply-templates select=\"node()|@*\"/>" << endl;
    fileXSLT << "    </xsl:copy>" << endl;
    fileXSLT << "  </xsl:template>" << endl;

    fileXSLT << "" << endl;

    fileXSLT << "</xsl:stylesheet>" << endl;
    fileXSLT << endl;
    fileXSLT.close();

    // Transform the XML file by xsltproc
    TString xsltProc;
#ifdef WIN32
    xsltProc = TString("xsltproc.exe");
#else
    xsltProc = TString("xsltproc");
#endif
    TString sCmd = Form("%s -o %s %s %s", xsltProc.Data(),
            gSystem->BaseName(_sFileName.Data()), sXSLTFile.Data(),
            _sFileName.Data());
    if (_blog)
        cout << "** sCmd[" << sCmd << "]" << endl;

    gSystem->Exec(sCmd.Data());
}

TString TInputFileXML::getAttributeInKey(TString skey)
{
    TString sattname = skey;

    Ssiz_t nlast = sattname.Last('@');
    sattname.Remove(0, nlast + 1);

    if (_blog)
        cout << " *** TInputFileXML::getAttributeInKey[" << skey << "] last["
                << nlast << "] sattname[" << sattname << "]" << endl;

    return sattname;
}

TString TInputFileXML::getFieldInKey(TString skey)
{
    TString sfieldname = skey;

    Ssiz_t nlast = sfieldname.Last('/');
    sfieldname.Remove(0, nlast + 1);

    if (_blog)
        cout << " *** TInputFileXML::getFieldInKey[" << skey << "] last["
                << nlast << "] sfieldname[" << sfieldname << "]" << endl;

    return sfieldname;
}

void TInputFileXML::printLog(Option_t *option)
{
    cout << "TInputFileXML::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    cout << "_cComments[" << _cComments << "]" << endl;
    cout << "_cSeparators[" << _cSeparators << "]" << endl;
    Int_t nAttribute = (Int_t)(_attributes.size());

    TString soption = TString(option);
    soption.ToLower();

    if (soption.Contains("att"))
    {
        for (Int_t indx = 0; indx < nAttribute; indx++)
        {
            cout << "index[" << indx + 1 << "/" << nAttribute << "]" << endl;
            _attributes[indx]->printLog();
        }
    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

