//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TIntMPI.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 10:38:34 CEST 2006
 
 \brief Impl�mentation de la classe URANIE::Launcher::TIntMPI
 
 */
#include "TIntMPI.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Launcher::TIntMPI)

namespace URANIE
{
namespace Launcher
{
TIntMPI::TIntMPI()
{
    cout << "constructing TIntMPI" << endl;
}
TIntMPI::~TIntMPI()
{
    cout << "destroy TIntMPI" << endl;
}

// ----------------------------------------------------------------------
// Initialize & Exit
// ----------------------------------------------------------------------

int TIntMPI::init(int& argc, char**& argv)
{
    // This routine puts all processes in a communicator called
    // APPS_COMM.  Users who wish to divide the processes into more than
    // one communicator, in order to do multi-level parallelism for
    // instance, should add code to pare down worldgroup to just the set
    // of processes APPS should use. This function returns the process'
    // rank.

    if (argc == -1)
    {
        cerr
                << "URANIE::Launcher::TIntMPI::init - must specify argc and argv when using MPI"
                << endl;
        throw "URANIE::Launcher error";
    }

    MPI_Group worldGroup, appsGroup;
    int e;
    int groupSize;
    int* ranksInOldGroup;

    e = MPI_Init(&argc, &argv);

    if (e != MPI_SUCCESS)
    {
        cerr << "URANIE::Launcher::TIntMPI::init - error initializing MPI."
                << endl;
        throw "URANIE::Launcher error";
    }

    // Get the set of processes (group, in MPI lingo) associated with 
    // MPI_COMM_WORLD.
    e = MPI_Comm_group(MPI_COMM_WORLD, &worldGroup);
    if (e != MPI_SUCCESS)
    {
        cerr << "URANIE::Launcher::TIntMPI::init - error calling MPI_Comm_group"
                << endl;
        throw "URANIE::Launcher error";
    }

    // Find out how many process there are in the group.
    e = MPI_Group_size(worldGroup, &groupSize);
    if (e != MPI_SUCCESS)
    {
        cerr << "URANIE::Launcher::TIntMPI::init - error calling MPI_Group_size"
                << endl;
        throw "URANIE::Launcher error";
    }

    // Build an array of ranks to include in the new communicator.  
    // The process with rank ranksInOldGroup[0] will have rank 0 in the 
    // new groupd, ranksInOldGroup[1] will have rank 1, etc.
    // In this case, since we want the new group to be the same as the
    // old, we include all ranks in the original order.
    ranksInOldGroup = new int[groupSize];
    for (int i = 0; i < groupSize; i++)
        ranksInOldGroup[i] = i;

    // Create the new group.
    e = MPI_Group_incl(worldGroup, groupSize, ranksInOldGroup, &appsGroup);
    if (e != MPI_SUCCESS)
    {
        cerr << "URANIE::Launcher::TIntMPI::init - error calling MPI_Group_incl"
                << endl;
        throw "URANIE::Launcher error";
    }

    // Create a new communicator which has the processes in group
    // appsGroup, drawn from MPI_COMM_WORLD.
    e = MPI_Comm_create(MPI_COMM_WORLD, appsGroup, &APPS_COMM);
    if (e != MPI_SUCCESS)
    {
        cerr
                << "URANIE::Launcher::TIntMPI::init - error calling MPI_Comm_create"
                << endl;
        throw "URANIE::Launcher error";
    }

    delete[] ranksInOldGroup;
    ranksInOldGroup = NULL;

    return TIntMPI::getMyTid();

}

// ----------------------------------------------------------------------
// Status & Notify
// ----------------------------------------------------------------------

int TIntMPI::getMyTid()
{
    // Return my taskId (or rank in MPI lingo).

    int rank;

    int e = MPI_Comm_rank(APPS_COMM, &rank);
    if (e != MPI_SUCCESS)
    {
        cerr << "APPSPACK::GCI::mytid - error getting MPI rank." << endl;
        throw "APPSPACK error";
    }

    return rank;

}

int TIntMPI::getNumProcs()
{
    // Returns number of processes

    int nProcs;

    int e = MPI_Comm_size(APPS_COMM, &nProcs);
    if (e != MPI_SUCCESS)
    {
        cerr
                << "APPSPACK::GCI::getNumProcs - error getting number of processes."
                << endl;
        throw "APPSPACK error";
    }

    return nProcs;

}

void TIntMPI::printLog(Option_t *option)
{
    cout << "TIntMPI::printLog[" << option << "]" << endl;
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

