//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncher.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 11:37:19 CEST 2006
 
 \brief Implementation of the class URANIE::Launcher::TLauncher
 \par Modifications 
 \li Jean-Baptiste Blanchard (CEA) 2016/04/05
Change the way the files are read (to be able to read strings and vectors)
 \li Jean-Baptiste Blanchard (CEA) 2016/02/10
 Add a boolean to write the output file only at the end (speed up dummy runs and tests) + setter/getter
 modify if it is true the opening and writting of the ROOT outputfile.
 \li S. Pomarede (InckA) 2012/05/04
 Add of a boolean attribute to manage progress bar display or not (with setter and getter)
 \li S. Pomarede (InckA) 2012/02/16
 Throw an exception when trying to declare a TCode without any TOutputFile declared (DI 148).
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - replace Form("%s/%s",....) by gSystem->ConcatFileName(...) in createResultsDirectory() and init()
 - replace the sum of string by gSystem->ConcatFileName(...) in run()
 - replace the comman unix "cp" by the generic method gSystem->CopyFile in run()
 \li F. Gaudier (2010-12-14 11:38:09)
 - resolve the DI111 and DI113
 \li Fabrice Gaudier (CEA) 2010/07/29
 When the working directory ends with a slash (/)  or a backslash(\), we remove it when stored in the  _sWorkingDirectory attribut.
 \li Cedric Forgues (C.S.) 2010/04/08
 Use of the cleanOutputFiles() method of TCode if saves are not performed through setSave() method.
 \li Cedric Forgues (C.S.) 2010/04/06
 The clean() method deletes now all the working directory. A new private method createResultsDirectory() was created to ensure that the results directory is well built. The default value for the setSave() method parameter is set to -1 (all calculations for each pattern are saved).
 \li Cedric Forgues (C.S.) 2010/03/17
 The method setSave() has arguments now (see TLauncher.h file). _nMaxSave is the fixed number of mandatory saves. nMaxSave is reset to MAX_PATTERN with the unsetSave() method.
 \li Cedric Forgues (C.S.) 2009/12/08
 The method _code->initFileType() replaces the _code->init() method in the TLauncher constructor.
 *********************************************************************
 
 */

// URANIE
#include "TLauncher.h"
#include "TAttribute.h"
#include "UExceptions.h"
#include "UEntry.h"
#include "UTimer.h"
#ifndef WIN32
#include "TLauncherMulti.h"
#endif
#ifndef OLDCLUB
#include "TamponTexte.h"
#endif

// ROOT
#include "TBranch.h"
#include "TList.h"
#include "TLeaf.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TPad.h"
#include "TFile.h"
#include "TMatrixT.h"
#include "TMatrixTUtils.h"

#include <iostream>
#include <fstream>

//# define OPEN_MAX 2024
//# define NR_OPEN 2024
using std::cout;
using std::endl;

ClassImp (URANIE::Launcher::TLauncher)

namespace URANIE
{
namespace Launcher
{

TLauncher::TLauncher() :
    _code(NULL), _bemptyMasterNode(kFALSE), _delay(0), _tds(NULL)
{
    if (_blog)
	cout << "construit TLauncher" << endl;
    SetNameTitle("Uranie_Unknown_Launcher", "Unknown launcher with uranie");
    _fLogger = new URANIE::DataServer::UMessageLogger("TLauncher");
}

TLauncher::TLauncher(URANIE::DataServer::TDataServer *tds, TCode *code, Option_t *option) :
    _code(code), _sdraw(0), _sopt(0), _scut(0), _sdirectory(gSystem->pwd()), _sworkingDirectory(""), _sPreProCommand(""), _sPostProCommand(""), _bsave(kFALSE), _bclean(kFALSE),
    _blog(kFALSE), _DoLogger(true), _DoMultiDump(true), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE), _bemptyMasterNode(kFALSE),
    _nprocs(1), _nMaxSave(MAX_PATTERN), _delay(0), _listOfSaveFile(NULL), _tds(tds)
{
    if (_blog)
	cout << "Builds TLauncher by TDataServer[" << tds->GetName() << "] code"
	     << endl;

    try
    {
	if (code == NULL)
	    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						  "TLauncher::constructor ERROR : the pointer of \"code\" is empty");
    }catch(URANIE::Exceptions::UExceptions &ue)
    {
	ue.printMessage();
	throw ue;
    }
	
    
    SetNameTitle(Form("LauncherTDS%sCODE%s", _tds->GetName(), _code->GetName()),
		 Form("Launcher of %s with %s", _code->GetName(), _tds->GetName()));

    TString soption=option;
    soption.ToLower();

    if(soption.Contains("quiet"))
        _DoLogger=false;

    if(soption.Contains("multiquiet"))
        _DoMultiDump=false;

    if(_DoLogger)
    {
        _fLogger = new URANIE::DataServer::UMessageLogger("TLauncher");
        (*_fLogger) << Endl<< "TLauncher \"" << this << "\" TCode \"" << code->GetName() << "\" with TDataServer \"" << tds->GetName() << "\"" << Endl;
    }
    else
        _fLogger=NULL;
    //(*_fLogger) << " ** Patterns : " << _tds->getNPatterns() << Endl;

    //Cedric Forgues (C.S.) 2009/12/08
    //_code->init();
    try
    {
	_code->initFileType();
    }catch(URANIE::Exceptions::UExceptions &ue)
    {
	if(_fLogger!=NULL)
	    delete _fLogger;
	_fLogger=NULL;   
	throw ue;
    }
#ifndef OLDCLUB
    catch (TamponTexteError& tte)
    {
	if(_fLogger!=NULL)
	    delete _fLogger;
	_fLogger=NULL;   
	throw tte;
    }
#endif
    if (_blog)
    {
        _code->printLog();
        cout << " *** end of init in TLauncher constructor" << endl;
    }
}

TLauncher::~TLauncher()
{
    if (_blog)
	cout << "destroy TLauncher" << endl;

    if(_fLogger!=NULL)
	delete _fLogger;
    _fLogger=NULL;

    //delete the input and output vectors
    for(vector<URANIE::DataServer::UEntry*>::iterator it = _InputVector.begin(); it != _InputVector.end(); it++)
	delete (*it);
    _InputVector.clear();
    
    for(vector<URANIE::DataServer::UEntry*>::iterator it = _OutputVector.begin(); it != _OutputVector.end(); it++)
	delete (*it);
    _OutputVector.clear();    
}

// ----------------------------------------------------------------------
// Encapsulate a code
// ----------------------------------------------------------------------

///////////////////////////////////////////////////////////
//  Init Method
///////////////////////////////////////////////////////////
void TLauncher::init(Option_t *option)
{

    FUNCT_TRACE();

    if (_blog)
	cout << " ** TLauncher::init" << endl;

    try
    {
        // The working directory is not specified in the TLauncher object
        if (_sworkingDirectory.Length() == 0)
	{
            // Get the working directory in the TCode object
            _sworkingDirectory = _code->getWorkingDirectory();
            // Else the current directory is current path/URANIE
            if (_sworkingDirectory.Length() == 0)
	    {
                _sworkingDirectory = gSystem->pwd();
		char *foo=gSystem->ConcatFileName( _sworkingDirectory.Data(), "URANIE"); //Needed prevent memleak
                _sworkingDirectory = foo;
		delete[] foo;
	    }
            if (_sworkingDirectory == gSystem->pwd())
	    {
		char *foo=gSystem->ConcatFileName(_sworkingDirectory.Data(), "URANIE");
                _sworkingDirectory = foo;
		delete[] foo;
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form(
								 "Beware! You have set the current directory as your working directory. Performing a setClean() method will delete the current directory. The directory %s/URANIE is created as working directory.",
								 _sworkingDirectory.Data()));
	    }
	}
        if (_sworkingDirectory == gSystem->pwd())
	{
	    char *foo=gSystem->ConcatFileName(_sworkingDirectory.Data(), "URANIE");
            _sworkingDirectory = foo;
	    delete[] foo;
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "Beware! You have set the current directory as your working directory. Performing a setClean() method will delete the current directory. The directory %s/URANIE is created as working directory.",
							     _sworkingDirectory.Data()));
	}
        if (_sworkingDirectory[0] == '.')
	{
            TString tmpDir = _sworkingDirectory;
            tmpDir.Remove(0, 1);
            tmpDir.Prepend(_sdirectory);
            _sworkingDirectory = tmpDir;
	}
        if (_code->getNOutputFiles() == 0)
	{
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "Beware! The list of output files is empty. You have to set the output files by using the addOutputFile method on the TCode object.");
	}
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if(_DoLogger)
        (*_fLogger)  << " ** WorkingDirectory[" << _sworkingDirectory << "]" << Endl;

    _code->setWorkingDirectory(_sworkingDirectory);

    TString soption(option);
    soption.ToLower();
    if(soption.Contains("nointermed"))
	setNoIntermediateSaved();

    //Old lines removed by JBB (2016-03-16)    
}

void TLauncher::initiateTheEntryVector()
{
    int ifile, iatt;    
    URANIE::DataServer::TAttribute *att;

    if( _InputVector.size() != 0 &&  _OutputVector.size() != 0 )
	return ;
    
    //Iterators
    _InputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, "iterator") );
    _OutputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, "iterator_tot") );

    stringstream sstrN, sstrT;    
    sstrN << _tds->getIteratorName() << ":";
    sstrT << "D:" ;
    //Inputs
    for(ifile=0; ifile<_code->getNInputFiles(); ifile++)
    {
	for(iatt=0; iatt<_code->getNInputAttributes(ifile); iatt++)
	{
	    att=_code->getInputAttribute(ifile, iatt);
            sstrN << att->GetName() << ":";
            sstrT << att->getDataTypeStr() << ":"; 
	    switch( att->getDataType() )
	    {
	    case URANIE::DataServer::TAttribute::kVector:
		_InputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName()) ) );
		break;
	    case URANIE::DataServer::TAttribute::kString:
		_InputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, string(att->GetName()) ) );
		break;
	    default:
		_InputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(att->GetName()) ) );
		break;
	    }
	}
    }

    //Output
    _OutputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, "n__calculs__") );    
    for(ifile=0; ifile<_code->getNOutputFiles(); ifile++)
    {
	for(iatt=0; iatt<_code->getNOutputAttributes(ifile); iatt++)
	{
	    att=_code->getAttribute(ifile, iatt);
	    switch( att->getDataType() )
	    {
	    case URANIE::DataServer::TAttribute::kVector:
		_OutputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string(att->GetName()) ) );
		break;
	    case URANIE::DataServer::TAttribute::kString:
		_OutputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, string(att->GetName()) ) );
		break;
	    default:
		_OutputVector.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(att->GetName()) ) );
		break;
	    }
	}
    }	

    _sInputNames = sstrN.str();
    _sInputTypes = sstrT.str();
}

///////////////////////////////////////////////////////////
// Create the ntuple file
///////////////////////////////////////////////////////////
URANIE::DataServer::TDSNtupleD* TLauncher::createNtupleDFile(TString &sntlname, TString &slabelvarname, TString &slabelvartype)
{

    /* JBB, 04/2016:
       Code largely taken from the previous one in ::run of this class
       but separated to be used also by TLauncherByStep
    */
    //Run over output files stored in code to look for output attributes
    Int_t noutfile = _code->getNOutputFiles();
    if (_blog)
	cout << " ** N output Files [" << noutfile << "]" << endl;
    for (Int_t ioutfile = 0; ioutfile < noutfile; ioutfile++)
    {
        Int_t nattfile = _code->getNOutputAttributes(ioutfile);
        if (_blog)
	    cout << " ** N attributes output Files [" << ioutfile << "/" << noutfile << "] - natt[" << nattfile << "]" << endl;
        for (Int_t iattfile = 0; iattfile < nattfile; iattfile++)
	{
            URANIE::DataServer::TAttribute * att = _code->getAttribute(ioutfile, iattfile);
            TString sattname = att->GetName();

	    /* Assume that is there is vector or strings, it has been precised when adding
	       the attribute to the corresponding outputfile*/
	    if (!_tds->isAttribute(sattname))
		_tds->addAttribute(att);	   
            slabelvarname += TString(":") + sattname;
	    slabelvartype += TString(":") + att->getDataTypeStr();

	    //Add the size attribute
	    if( att->getDataTypeStr().EqualTo("V") )
	    {
		if (!_tds->isAttribute(SIZENAME+"__"+string(sattname)))
		    _tds->addAttribute( (SIZENAME+"__"+string(sattname)).c_str() );	   
		_tds->getAttribute( (SIZENAME+"__"+string(sattname)).c_str() )->setOrigin(URANIE::DataServer::TAttribute::kInternal);
		
		slabelvarname += TString(":") + TString(SIZENAME+"__")+sattname;
		slabelvartype += TString(":") + "D";
	    }
            if (_blog)
		cout << " **  Name[" << sattname << "] Share[" << att->getShare() << "]" << endl;
	}
    }
    if(_blog)
	cout<<"varname "<<slabelvarname<<" vartype  "<<slabelvartype<<endl;

    //Creation of the tuple
    sntlname = Form("Launcher_%s", _tds->getTuple()->GetName());
    URANIE::DataServer::TDSNtupleD *ntl = new URANIE::DataServer::TDSNtupleD(sntlname.Data(), Form("Launcher_%s", _tds->getTuple()->GetTitle()),
									     slabelvarname, slabelvartype);
    ntl->SetMarkerColor(2);
    ntl->SetMarkerSize(4);
    ntl->SetMarkerStyle(7);

    return ntl;
}
  
///////////////////////////////////////////////////////////
//  createResultsDirectory Method
///////////////////////////////////////////////////////////  
void TLauncher::createResultsDirectory(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    // Make the Results directory for distribution
    try
    {
	char *foo=gSystem->ConcatFileName(_sworkingDirectory.Data(),"UranieResults");
	TString sresultsDirectory = foo;
	delete[] foo;
        void * openDirResult = gSystem->OpenDirectory(sresultsDirectory);
        if (openDirResult == 0)
	    if (gSystem->mkdir(sresultsDirectory, kTRUE) == -1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
							       "TLauncher::createResultsDirectory ERROR\n the subdirectory \"UranieResults\" can not be created in the workindirectory \"%s\" with name\"%s\"",
							       _sworkingDirectory.Data(),
							       sresultsDirectory.Data()));
        gSystem->FreeDirectory(openDirResult);
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	    throw ue;
    }
}

///////////////////////////////////////////////////////////
//  Terminate Method
///////////////////////////////////////////////////////////
void TLauncher::terminate(Option_t *option)
{
    if (_blog)
	cout << " ** TLauncher::terminate" << endl;
    _code->terminate(option);
    if (_bclean && (_bsave == kFALSE))
	clean(option);

    gSystem->cd(_sdirectory);
    if (_blog)
    {
        cout << " ** TLauncher::terminate Current directory[" << gSystem->pwd()
	     << "]" << endl;
    }
}

///////////////////////////////////////////////////////////
//  Clean Method
///////////////////////////////////////////////////////////
void TLauncher::clean(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    if(_DoLogger)
        (*_fLogger) << " ** Clean the Working Directory ... ";

    // Place before in the current directory (TLauncher)
    gSystem->cd(_sdirectory);

    // Go to the working directoryPlace in the good directory
    if (_sworkingDirectory.Length() != 0)
    {
#ifdef WIN32
        gSystem->Exec(Form("if exist %s rmdir /s /q %s", _sworkingDirectory.Data(), _sworkingDirectory.Data()));
#else
        gSystem->Exec(Form("rm -fr %s", _sworkingDirectory.Data()));
#endif
        if(_DoLogger)            
            (*_fLogger) << "Done" << Endl;
    }
    else
    {
        (*_fLogger) << "Not realised (empty)" << Endl;
    }

}

// ----------------------------------------------------------------------
// Method RUN
// ----------------------------------------------------------------------
void TLauncher::run(Option_t *option)
{

    TString soption=option;
    soption.ToLower();

    if(soption.Contains("quiet"))
       _DoLogger=false;

    bool _bincludeInputs=false;
#ifndef WIN32    
    if(soption.Contains("withinputs"))
       _bincludeInputs=true;
#endif
        
    if (_blog)
	cout << " ** TLauncher::run" << endl;

    if (_bsave && _DoLogger)
	(*_fLogger) << " ** Save : Yes" << Endl;
    else if (_DoLogger)
	(*_fLogger) << " ** Save : No" << Endl;


    // Intialisation and Define the attributs _sworkingDirectory by default
    init(option);

    // Clean if necessary
    if (_bclean)
	clean(option);

    //       //Ensures that Results directory is well built
    //       createResultsDirectory();      

    // Init the code ( make sure to create the sub-directories UranieLauncher_1 and UranieResults)
    _code->init(option);

    TString srootfile = _code->getWorkingDirectory(); // getReferenceDirectory();
    // When there is no reference directory for the TCode
    if (srootfile.Length() == 0)
	srootfile = _sdirectory;

    /*
      JBB (2016-03-16) dont need to recreate a tuple with inputs. Only one with iterator + output should do the trick
      ==> Aim is not to touch original ntuple and concatenate at the end the various inputs
    */

    stringstream sstr; sstr.precision(SSTR_PREC);

    //Ntuple properties
    TString sntlname;
    TString slabelvarname = "Launcher__n__iter__:n__calculs__";
    TString slabelvartype = "D:D";
    URANIE::DataServer::TDSNtupleD *ntl = createNtupleDFile(sntlname, slabelvarname, slabelvartype);

    TCanvas *localcan=NULL;
    if(_sdraw.Length() > 0)
    {
	localcan = new TCanvas("__launcherDisplay__","__launcherDisplay__",1);	
	ntl->AddFriend( _tds->getTuple() );
    }
    
    TString svarcut;
    Int_t npat = (Int_t)(_tds->getNPatterns());

    // --------------
    // ---- NEWS ----
    Int_t nprocstot = 1;
#ifndef WIN32
    pid_t pidpere = getpid();
    TLauncherMulti multiLanceur(ntl->GetName(), ntl->GetTitle(), option,
				_bemptyMasterNode, _delay);
    if (_blog)
	cout << "_bemptyMasterNode = " << _bemptyMasterNode << endl;
    multiLanceur.setWorkingDir(_sworkingDirectory);

    nprocstot = multiLanceur.getAvailableProcs();
#endif
    TString originalCommand = _code->getCommand();
    Int_t ipatMulti = 0;
    // ---- NEWS ----
    // --------------
#ifndef WIN32
       if(_DoLogger)
           (*_fLogger) << " ** Proc : " << nprocstot << Endl;
#endif

    //Memory database initialisation
    initiateTheEntryVector();

#ifndef WIN32
    if(nprocstot>1)
    {
        multiLanceur.defineJob(_code->getCommand(), npat, _nprocs);
        setSave(-1);

        if(_bincludeInputs)
        {
            delete ntl;
            slabelvarname = _sInputNames + "Launcher__n__iter__:n__calculs__";
            slabelvartype = _sInputTypes + "D:D";
            ntl = createNtupleDFile(sntlname, slabelvarname, slabelvartype);
        }
    }
#endif          

    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
    Bool_t bCrtSave = _bsave;
    Int_t nFailure = 0;

    TFile *rootfile = NULL;
    TString srootfileipat;
    if (nprocstot == 1)
    {
	char *foo;
#ifdef WIN32
	foo = gSystem->ConcatFileName(srootfile.Data() ,"UranieResults\\uranielauncher.root");
#else
	foo = gSystem->ConcatFileName(srootfile.Data(),
				      "UranieResults//uranielauncher.root");
#endif

	srootfileipat = foo;
	delete[] foo;
    }
    
    // ----------------------------------------------------------------------
    //  LOOP
    // ----------------------------------------------------------------------
    _code->setCommand(originalCommand);
    for (Int_t ipat = 0; ipat < npat; ipat++)
    {

        if ((ipat == _nMaxSave) && (bCrtSave))
	{
            (*_fLogger) << Endl<< URANIE::DataServer::UMessageLogger::kWARNING << "Last save pattern [" << ipat << "] now in 9999" << Endl;
            (*_fLogger) << URANIE::DataServer::UMessageLogger::kINFO;
            bCrtSave = kFALSE;
            _code->resetCurrentSimulation();
	}

        if (bCrtSave)
	    _code->init(option);
        else
	    _code->cleanOutputFiles();
	
        _tds->getTuple()->GetEntry(ipat);

        if (_blog)
	    cout << endl << " ** TLauncher::run iter[" << 1.0 + ipat << "]"
		 << endl;

        _code->preTraitment(ipat, &_InputVector, option);
        // --------------
        // ---- NEWS ----
        if (nprocstot == 1)
	{
            _code->run(option);
            ipatMulti = ipat;
	    //           cout << " *** RUN launcher  [" << ipat << "]"<< endl;
	    //           cerr << " *** RUN launcher  [" << ipatMulti << "]"<< endl;
	}
#ifndef WIN32
        else
	{
	    //           cout << " *** RUN launcherMulti PID[" << getpid()<<"] -- [" << ipat << "]"<< endl;
	    //           cerr << " *** RUN launcherMulti PID[" << getpid()<<"] -- [" << ipatMulti << "]"<< endl;
            multiLanceur.createJob(ipat);
            if (getpid() != pidpere)
	    {
                // PreProCommand step if necessary
                if ( _sPreProCommand.Length() > 0 )
                       gSystem->Exec(_sPreProCommand.Data());

                multiLanceur.run(_code, option);

                // PostProCommand step if necessary
                if ( _sPostProCommand.Length() > 0 )
                       gSystem->Exec(_sPostProCommand.Data());
                ipatMulti = ipat;
                ipat = 0; // Child : nouveau donc doit etre initialise
	    }
            else
		continue;
	}
#endif
        // ---- NEWS ----
        // --------------

	/*Now the output are stored in this map having the key:value principle: no conversion is needed 
	  and the result is concatenate and transmit to the string fill method in the chosen order */
	// map<string,string> sout;
        // if (_code->postTraitment(&sout, _sopt) == kFALSE)
	//     nFailure++;

        if (_code->postTraitment(&_OutputVector, _sopt) == kFALSE)
	    nFailure++;

        // Save the Output Files (Untouched by JBB)
        if (_listOfSaveFile)
	{
            cout << endl << " *** Save the file" << endl;
            _listOfSaveFile->Print();
            cout << " ** la liste" << endl;
            TNamed *obj = NULL;
            TIter next(_listOfSaveFile);

            while ( (obj = (TNamed *) next()) )
	    {
                TString sfilename = obj->GetName();
                TString soutname = Form("%s.%d", obj->GetTitle(),
					ipatMulti + 1);
                cout << " isfilename  [" << sfilename << "] soutname ["
		     << soutname << "]" << endl;
                Int_t nret = gSystem->Exec(
		    Form("cp %s %s/%s", sfilename.Data(),
			 _sdirectory.Data(), soutname.Data()));
                if (nret == -1)
		    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
							  Form(
							      "TLauncher::run ERROR : for copy output file \"%s\" to \"%s\"",
							      sfilename.Data(), soutname.Data()));
	    }
	}

        if (_blog)
	    cout << " *** Analysis of several outputs" << endl;

	_OutputVector.at(0)->copyEntry( _InputVector.at(0) );
        if(_bincludeInputs)
            ntl->Fill(&_InputVector, &_OutputVector);
        else            
            ntl->Fill(&_OutputVector);

        if (_blog)
	    cout << " *** End of Analyse de plusieurs sorties" << endl;

	if( nprocstot !=1 )
	{
	    char *foo;
#ifdef WIN32
            foo = gSystem->ConcatFileName(_sworkingDirectory.Data(), "UranieResults\\uranielauncher_");
#else
            foo = gSystem->ConcatFileName(_sworkingDirectory.Data(),
					  "UranieResults/uranielauncher_");
#endif
	    srootfileipat = foo;
            srootfileipat += ipatMulti;
            srootfileipat += ".root";
	    delete[] foo;
	}
        if (_blog)
	{
            cout << "srootfileipat[" << srootfileipat << "]" << endl;
	}

	if (nprocstot != 1 || !_bnoIntermediateSaved )
	{
	    rootfile = TFile::Open(srootfileipat.Data(), "RECREATE");
	    ntl->SetDirectory(0);
	    ntl->Write();
	    rootfile->Close();
	}
        // ntall->SetDirectory(gDirectory);

        if (nprocstot == 1 && ipat % 5 == 0 && !_bnoIntermediateSaved)
	{
            TString ssavefile = srootfileipat + TString(".save");
            gSystem->CopyFile(srootfileipat.Data(), ssavefile.Data(), kTRUE);
	}

        // ---- NEWS ----
        // --------------

        // Draw the graph with the selection
        if (_sdraw.Length() > 0)
	{
            if (_scut.Length() > 0)
		svarcut = TString(
		    Form("( ( %s ) && (  %s < %d ) )", _scut.Data(),
			 _tds->getIteratorName(), ipatMulti));
            else
		svarcut = TString(
		    Form("%s<%d", _tds->getIteratorName(), ipatMulti + 1));

            ntl->Draw(Form("%s",_sdraw.Data()), svarcut.Data(), _sopt);
            localcan->Modified();
            localcan->Update();
	}

        // --------------
        // ---- NEWS ----
#ifndef WIN32  
        if (nprocstot == 1)
	{
            if (_bdrawProgressBar)
	    {
                timer.DrawProgressBar(ipat, nFailure);
	    }
	}
        else
	{
            if (getpid() != pidpere)
	    {
                ipat = npat; // Child : sortie de boucle
	    }
	}
#endif        
        // ---- NEWS ----
        // --------------

    }  // End of LOOP

    if(_sdraw.Length() > 0)
    {
	ntl->RemoveFriend( _tds->getTuple() );
	delete localcan;
    }

    if (nprocstot == 1 && _bnoIntermediateSaved )
    {
	rootfile = TFile::Open(srootfileipat.Data(), "RECREATE");
	rootfile->cd();
	ntl->SetDirectory(0);	
	ntl->Write();
	rootfile->Close();
    }

    if (nprocstot == 1 && _DoLogger)
    {
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
        if ( nFailure > 0 )
	    (*_fLogger) << "     Failure : \033[0;31m" << nFailure << "\033[0m/\033[0;31m" << npat << "\033[0m" << " --- \033[0;31m" << setw(1) << 100. * Double_t(nFailure)/Double_t(npat) << "%\033[0m" << Endl;
    }

    // Terminate
    //      terminate(option);

    TFile *fin=NULL;
    // --------------
    // ---- NEWS ----
    if (nprocstot == 1)
    {
        // Terminate
        terminate(option);
    }
#ifndef WIN32
    else
    {
        multiLanceur.waitAll();
        if (getpid() != pidpere)
	{
            _tds->keepFinalTuple(false);
            if( _DoMultiDump )
            {
                cerr << " ** TLauncher::run method:: Job Finished , pid = "
                     << getpid() << " Iteration[" << ipatMulti + 1 << "]"
                     << endl;
                cout << " ** TLauncher::run method:: Job Finished , pid = "
                     << getpid() << " Iteration[" << ipatMulti + 1 << "]"
                     << endl;
            }
		    delete _tds;
            exit(1);
	}
        else
	{
            if( _DoMultiDump )
            {
                cerr << " ********** pidpere[" << pidpere << "]" << endl;
                cerr << " ** End of Lancher : Begin the merge with all the data ["
                     << npat << "]" << endl << endl;
            }
            multiLanceur.mergeAll(npat);
            if( _DoMultiDump )
            {
                cout << "JOB FINISHED (master), pid = " << getpid() << endl;
                cerr << "JOB FINISHED (master), pid = " << getpid() << endl;
            }
	    if(ntl != NULL)
	    {
		delete ntl;
		ntl = NULL;
	    }

#ifdef WIN32
	    fin = TFile::Open(_sworkingDirectory + "\\UranieResults\\uranielauncher.root" );
#else
	    fin = TFile::Open(_sworkingDirectory + "/UranieResults/uranielauncher.root" );
#endif
	    ntl = (URANIE::DataServer::TDSNtupleD*)fin->Get( sntlname.Data() );//new URANIE::DataServer::TDSNtupleD( tt );

	    terminate(option);
	}
    }
#endif
    // ---- NEWS ----
    // --------------

    /*
      Final step: merge both input and output.
    */
    vector<string> buffer; 
    URANIE::DataServer::UEntry::StringParser( string(slabelvarname), ":", buffer);
    string thevars="";
    for (unsigned int ivar=1; ivar<buffer.size(); ivar++)
	thevars+=buffer.at(ivar)+((ivar!=buffer.size()-1)?":":"");

    char *foo=0;
    if ( ! (_bclean && (_bsave == kFALSE)) )
    {
#ifdef WIN32
        foo = gSystem->ConcatFileName(srootfile.Data() ,"UranieResults\\uranielauncher.root");
#else
        foo = gSystem->ConcatFileName(srootfile.Data(), "UranieResults//uranielauncher.root");
#endif

        rootfile = TFile::Open(foo, "RECREATE");
        rootfile->cd();
    }

    if(!_bincludeInputs)
    {
        _tds->getTuple()->Merger(ntl, thevars, thevars,false,_tds->getIteratorName(),"Launcher__n__iter__");
        _tds->getTuple()->SetName( (string("Launcher_")+string(_tds->getTuple()->GetName())).c_str() );
    }
    else
        _tds->setTuple(ntl);

    if ( ! (_bclean && (_bsave == kFALSE)) )
    {
        _tds->getTuple()->Write();
        _tds->getTuple()->LoadBaskets();//SetScanField(-1);
        rootfile->Close();
        // _tds->getTuple()->SetDirectory(0);
    }
    
    delete[] foo;
        
    if (_blog)
    {
        cout << " ** End of Lancher tds ****************" << endl;
        _tds->getTuple()->Print();
        _tds->getTuple()->Scan("*","","",1);
    }
    
    if(ntl != NULL && fin==NULL)
    {
        delete ntl;
        ntl = NULL;
    }
    
#ifndef WIN32
    if( nprocstot != 1 && fin!=NULL)
        fin->Close();

#endif

}
    
void TLauncher::addSaveFile(TString sfile)
{

    cout << endl << " **** sfile[" << sfile << "]" << endl;
    if (_listOfSaveFile == NULL)
    {
        _listOfSaveFile = new TList();
    }

    TString sracinefile = gSystem->BaseName(sfile.Data());
    _listOfSaveFile->Add(new TNamed(sfile, sracinefile));

    _listOfSaveFile->Print();

    cout << endl;

}

// ----------------------------------------------------------------------
// setWorkingDirectory() method
// ----------------------------------------------------------------------
//-----------------------------------------------------------------------
void TLauncher::setWorkingDirectory(TString str)
{
    TString strDirectory = str;
    try
    {
        if ((strDirectory.EndsWith("\\")) || (strDirectory.EndsWith("/")))
	{
            strDirectory.Remove(strDirectory.Length() - 1);
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TLauncher::setWorkingDirectory method\n the working directory \"%s\" is redefined without last backslash/slash \"%s\"",
							     str.Data(), strDirectory.Data()));

	}
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
    }
    _sworkingDirectory = strDirectory;
    if (_blog)
	cout << " *** _sworkingDirectory [" << _sworkingDirectory << "]"
	     << endl;
}

// ----------------------------------------------------------------------
// setSave() method
// ----------------------------------------------------------------------
//-----------------------------------------------------------------------
void TLauncher::setSave(Int_t nSave)
{
    try
    {
        if (nSave > 0)
	{
            _nMaxSave = nSave;
            if (nSave > _tds->getNPatterns())
	    {
                _nMaxSave = _tds->getNPatterns();
                if (nSave != MAX_PATTERN)
		    throw URANIE::Exceptions::UWarningExceptions(__FILE__,
								 __LINE__,
								 "You cannot perform more saves than the existing number of samples!");
	    }
	}
        else if (nSave == -1)
	    _nMaxSave = _tds->getNPatterns();

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        cout << "** Performing " << _nMaxSave << " (number of samples) saves."
	     << endl;
    }
    _bsave = kTRUE;
}

// ----------------------------------------------------------------------
// Printing Log
// ----------------------------------------------------------------------
//---------------------------------------------
void TLauncher::printLog(Option_t *option)
{
    cout << "TLauncher::printLog[" << option << "]" << endl;
    //_code->printLog();
    cout << " _sdraw[" << _sdraw << "]" << endl;
    cout << " _sopt[" << _sopt << "]" << endl;
    cout << " _scut[" << _scut << "]" << endl;
}

} // Fin du namespace LAUNCHER
} // Fin du namespace URANIE

