//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TLauncherByStep.cxx
 \author Fabrice Gaudier  <fabrice.gaudier@cea.fr>
 \date 2011-10-20 19:40:04

 \brief

 */

#include "TLauncherByStep.h"

#include <iostream>
#include <fstream>

using std::cout;
using std::endl;

// ROOT
#include "TFile.h"
#include "TLeaf.h"
#include "TPad.h"

// URANIE
#include "UExceptions.h"
#include "UTimer.h"

ClassImp (URANIE::Launcher::TLauncherByStep);

namespace URANIE
{
namespace Launcher
{

////////////////////////////////////////////////////////////
//  Default constructor
////////////////////////////////////////////////////////////
TLauncherByStep::TLauncherByStep(URANIE::DataServer::TDataServer *tds, TCode *code, Option_t *option) :
    TLauncher(tds, code, option), _nAllocatedSeconds(1800), _nAllocatedCores(0), _sProjectName(""), _sClusterName("")
{
    setSave();
    setClean();
}

////////////////////////////////////////////////////////////
//  Default destructor
////////////////////////////////////////////////////////////
TLauncherByStep::~TLauncherByStep()
{
}

////////////////////////////////////////////////////////////
//  preTreatment
////////////////////////////////////////////////////////////
void TLauncherByStep::preTreatment(Option_t *option)
{
    if (_blog)
        cout << " ** TLauncherByStep::preTreatment" << endl;
    if(_DoLogger)
        (*_fLogger) << Endl<< " ** TLauncherByStep::preTreatment" << Endl;

    if (_bsave && _DoLogger)
        (*_fLogger) << " ** Save : Yes" << Endl;
    else if (_DoLogger)
        (*_fLogger) << " ** Save : No" << Endl;

    // Intialisation and Define the attributs _sworkingDirectory by default
    initiateTheEntryVector();
    init(option);
    
    // Clean if necessary
    clean(option);

    // Init the code
    _code->init(option);

    Int_t npat = (Int_t)(_tds->getNPatterns());

#ifndef WIN32  
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
#endif        

    // ----------------------------------------------------------
    //  LOOP
    // ----------------------------------------------------------
    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        _code->init(option);
        _tds->getTuple()->GetEntry(ipat);

        _code->preTraitment(ipat, &_InputVector, option);

        if (_blog)
            cout << endl << " ** TLauncherByStep::preTreatment iter[" << 1.0 + ipat << "]" << endl;

#ifndef WIN32
        if (_bdrawProgressBar)            
            timer.DrawProgressBar(ipat, 0);
#endif        
    }  // End of LOOP

#ifndef WIN32
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
#endif        

    // Terminate
    gSystem->cd(_sdirectory);

    if (_blog)
    {
        cout << " **************************************" << endl;
        cout << " ** End Of TLauncherByStep::preTreatment" << endl;
        _tds->getTuple()->Print();
        cout << " **************************************" << endl;
    }

}

////////////////////////////////////////////////////////////
//  run
////////////////////////////////////////////////////////////
void TLauncherByStep::run(Option_t *option)
{
    if (_blog)
    {
        cout << " **************************************" << endl;
        cout << " ** TLauncherByStep::run" << endl;
    }
    if(_DoLogger)
        (*_fLogger) << Endl<< " ** TLauncherByStep::run" << Endl;

    Int_t npat = (Int_t)(_tds->getNPatterns());
#ifndef WIN32  
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");
#endif        
    if (strcmp(_sClusterName.Data(), "irene") == 0)
    {
    	TString sTmpFileName("__Uranie_Tmp__");
        // if _nb_allocated_cores is not set, use number of patterns
        Int_t nbcores = (_nAllocatedCores > 0) ? _nAllocatedCores : npat;
        //creating the SLURM batch file
        TString sBsubFile(Form("%s.bsub", sTmpFileName.Data()));
        std::ofstream os(sBsubFile.Data(), std::ofstream::out);
        os << "#!/bin/bash" << endl;
        os << "#MSUB -r URANIE_JOB " << endl;
        os << "#MSUB -n " << nbcores << endl;
        os << "#MSUB -t " << _nAllocatedSeconds << endl;
        os << "#MSUB -o URANIE_JOB.o" << endl;
        os << "#MSUB -e URANIE_JOB.e" << endl;
        os << "#MSUB -q skylake" << endl;
        for (UInt_t i = 0; i < _sListExtraDirectives.size(); i++)
            os << _sListExtraDirectives[i].Data() << endl;
        os << "set -x \n" << endl;
        os << "cd ${BRIDGE_MSUB_PWD}" << endl;

        for (Int_t ipat = 0; ipat < npat; ipat++)
        {
            TString sdir = Form("%s/UranieLauncher_%d",
                    _sworkingDirectory.Data(), ipat + 1);

            os << "cd " << sdir.Data() << endl;
            os << "ccc_mprun -E '--exclusive' -n " << _nprocs << " "
                    << _code->getCommand() << " &" << endl;
        }
        os << "wait" << endl;
        os.close();
        TString scommand(Form("ccc_msub "));

        // check if the project name is set and add it to the command
        if (_sProjectName.Length() > 0 )    
            scommand += Form(" -A %s", _sProjectName.Data());

        // Add to the command the bsub and log files 
        scommand += Form(" %s.bsub > %s.log", sTmpFileName.Data(), sTmpFileName.Data());
    
		
		try {
			Int_t result = gSystem->Exec(scommand.Data());
			
			if(result!=0){
			throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                   Form("TLauncherByStep::run Error\n The command [%s] is not valid. It works only on irene!", scommand.Data()));
												   }

		} catch (URANIE::Exceptions::UErrorExceptions& ue)
		{	
			ue.printMessage();
			throw ue;
		}
		
    }
    else
    {
        for (Int_t ipat = 0; ipat < npat; ipat++)
        {
            // locate in the evaluate directory
            TString sdir = Form("%s/UranieLauncher_%d",
                    _sworkingDirectory.Data(), ipat + 1);
            gSystem->cd(sdir);
            _code->run(option);
#ifndef WIN32
            if (_bdrawProgressBar)
                timer.DrawProgressBar(ipat, 0);
#endif 
        }
    }

#ifndef WIN32
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
#endif        
    // Terminate
    gSystem->cd(_sdirectory);

    if (_blog)
    {
        cout << endl;
        cout << " ** End Of TLauncherByStep::run" << endl;
        cout << " **************************************" << endl;
    }
}

////////////////////////////////////////////////////////////
//  postTreatment
////////////////////////////////////////////////////////////
void TLauncherByStep::postTreatment(Option_t *option)
{
    if (_blog)
        cout << " ** TLauncherByStep::postTreatment" << endl;

    if(_DoLogger)
        (*_fLogger) << Endl<< " ** TLauncherByStep::postTreatment" << Endl;

    TString srootfile = getWorkingDirectory();
    if (_blog)
        cout << " ** srootfile    getWorkingDirectory[" << srootfile << "]"
                << endl;
    srootfile += TString("/UranieResults/uranielauncher.root");
    if (_blog)
        cout << " ** srootfile[" << srootfile << "]" << endl;

    //Deal with the intermediate file
    TString soption(option);
    soption.ToLower();
    if(soption.Contains("nointermed"))
      setNoIntermediateSaved();
    
    Int_t npat = (Int_t)(_tds->getNPatterns());

    //Create the ntuple and go through the final results output file
    stringstream sstr; sstr.precision(SSTR_PREC);
    TString sntlname;
    TString slabelvarname = "Launcher__n__iter__:n__calculs__";
    TString slabelvartype = "D:D";
    URANIE::DataServer::TDSNtupleD *ntl = createNtupleDFile(sntlname, slabelvarname, slabelvartype);
    
    
    URANIE::DataServer::UTimer timer(npat, "Uranie Launcheur");

    Int_t nFailure = 0;
    // fga::Modification 2011-11-10 14:59:45
    // Clear the listOfOutput list
    //   At this step, only the number of output (nOutput) attributes and the vector of default values
    //   are used in the next

    TFile *rootfile;

    // -------------------------------------------------------------
    //  LOOP
    // -------------------------------------------------------------
    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        // locate in the evaluate directory
        TString sdir = Form("%s/UranieLauncher_%d", _sworkingDirectory.Data(),
                ipat + 1);
        gSystem->cd(sdir);

        if (_blog)
            cout << endl << " ** TLauncher::run iter[" << 1.0 + ipat << "]"
                    << endl;

  	// map<string,string> sout;
        // if (_code->postTraitment(&sout, _sopt) == kFALSE)
	
        if (_code->postTraitment(&_OutputVector, _sopt) == kFALSE)
	    nFailure++;

        // Store the Output Files if request by addSaveFile(TString) method
        if (_listOfSaveFile)
        {
            cout << endl << " ** Save the files" << endl;
            _listOfSaveFile->Print();
            TNamed *obj = NULL;
            TIter next(_listOfSaveFile);

            while ( (obj = (TNamed *) next()) )
            {
                TString sfilename = obj->GetName();
                TString soutname = Form("%s.%d", obj->GetTitle(), ipat + 1);
                cout << " ** isfilename  [" << sfilename << "] soutname ["
                        << soutname << "]" << endl;
                //            Int_t nret = gSystem->Exec(Form("cp %s %s/%s", sfilename.Data(), );
                Int_t nret = gSystem->CopyFile(sfilename.Data(),
                        Form("%s/%s", _sdirectory.Data(), soutname.Data()),
                        kTRUE);
                if (nret == -1)
                    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                            Form(
                                    "TLauncher::run ERROR : for copy output file \"%s\" to \"%s\"",
                                    sfilename.Data(), soutname.Data()));
            }
        }

	//Fill the ntuple
	// string theline;
	// sstr.str("");
	// sstr << ipat + 1 << " ";
	// theline=sstr.str()+" "+sout["ncalc"]+" ";
	// vector<string> buffer; 
	// URANIE::DataServer::UEntry::StringParser( string(slabelvarname), ":", buffer);

	// for (unsigned int ivar=2; ivar<buffer.size(); ivar++)
	//   theline+=sout[buffer.at(ivar)]+((ivar!=buffer.size()-1)?" ":"");

	// if(_blog)
	//   cout<<" *** Filling the tuple with  "<<theline<<endl<<endl;

	// ntl->Fill(theline);

	_OutputVector.at(0)->copyEntry( _InputVector.at(0) );
	ntl->Fill(&_OutputVector);
	
        if (_blog)
        {
            cout << " ** End Of Analysis of several outputs" << endl;
            cout << " *************************************" << endl << endl;

	}

        // --------------
        // ---- NEWS ---- 
        // Save the TDSNtupleD in a rootfile
	if(!_bnoIntermediateSaved )
	  {
	    rootfile = TFile::Open(srootfile.Data(), "RECREATE");
	    if (_blog)
	      {
		cout << " ** TLauncherByStep Current directory[" << gSystem->pwd()
		     << "]" << endl;
		cout << " **** srootfile[" << srootfile << "]" << endl;
	      }
	    ntl->SetDirectory(0);
	    ntl->Write();
	    rootfile->Close();
	    ntl->SetDirectory(gROOT);
	    if (ipat % 5 == 0)
	      gSystem->CopyFile(srootfile.Data(),
				Form("%s.save", srootfile.Data()), kTRUE);
	  }
        // ---- NEWS ----
        // --------------

#ifndef WIN32
        if (_bdrawProgressBar)
            timer.DrawProgressBar(ipat, nFailure);
#endif        
    }  // End of LOOP


    if(_bnoIntermediateSaved )
      {
	rootfile = TFile::Open(srootfile.Data(), "RECREATE");
	if (_blog)
	  {
	    cout << " ** TLauncherByStep Current directory[" << gSystem->pwd()
		 << "]" << endl;
	    cout << " **** srootfile[" << srootfile << "]" << endl;
	  }
	ntl->SetDirectory(0);
	ntl->Write();
	rootfile->Close();
      }
    
    // Merge the input and output ntupleD
    vector<string> buffer; 
    URANIE::DataServer::UEntry::StringParser( string(slabelvarname), ":", buffer);
    string thevars="";
    for (unsigned int ivar=1; ivar<buffer.size(); ivar++)
      thevars+=buffer.at(ivar)+((ivar!=buffer.size()-1)?":":"");
    
    _tds->getTuple()->Merger(ntl, thevars, thevars,false,_tds->getIteratorName(),"Launcher__n__iter__");
    
    if(_DoLogger)
        (*_fLogger) << Endl<< "time elapsed " << timer.GetElapsedTime() << Endl;
    if (nFailure > 0 && _DoLogger)
        (*_fLogger) << "     Failure : \033[0;31m" << nFailure
                << "\033[0m/\033[0;31m" << npat << "\033[0m"
                << " --- \033[0;31m"
                << setw(1) << 100. * Double_t(nFailure)/Double_t(npat) << "%\033[0m" << Endl;

    // Terminate
    terminate(option);

    // Affect the current TTree to the TDataServer
    if (_blog)
        cout << " ** End of Lancher ntall" << endl;

    delete ntl;
    
    if (_blog)
    {
        cout << " ** End of Lancher tds ****************" << endl;
        _tds->getTuple()->Print();
    }

}

////////////////////////////////////////////////////////////
//  print the Log
////////////////////////////////////////////////////////////
void TLauncherByStep::printLog(Option_t *option)
{
    cout << "TLauncherByStep::printLog[" << option << "]" << endl;
}

} // End Of namespace Launcher
} // End Of namespace URANIE

//  End Of File TLauncherByStep.cxx
