//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TLauncherByStep.h
 \author Fabrice Gaudier  <fabrice.gaudier@cea.fr>
 \date 2011-10-20 19:40:04

 \brief

 */

#ifndef TLAUNCHERBYSTEP_H
#define TLAUNCHERBYSTEP_H

#include "TNamed.h"

#include <iostream>
#include <fstream>

using std::cout;
using std::endl;

// ROOT
#include "Rtypes.h"

// Uranie
#include "TLauncher.h"
#include "Launcher.h"

namespace URANIE
{
namespace Launcher
{
class TLauncherByStep: public TLauncher
{

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor 
    TLauncherByStep(URANIE::DataServer::TDataServer *tds, TCode *code, Option_t *option="");
    //! Default destructor
    virtual ~TLauncherByStep();
    //@}
    //---------------------------------------------
    /** @name Manipulate the code
     **/
    //@{
    //! PreTreatment
    /*!
     Creates all the working subdirectory which contains the input files
     */
    virtual void preTreatment(Option_t *option = "");
    //! Evaluate the code in each working subdirectory
    /*!
     This step can be done by an another code
     */
    virtual void run(Option_t *option = "");
    //! PostTreatment
    /*!
     Extracts the values of the output attributes in the output files for each working subdirectory
     */
    virtual void postTreatment(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Print the log of the class
    virtual void printLog(Option_t *option = "");
    //@}

    //-----------------------------------------------
    /** @name manipulating variables for batch systems
     **/
    //@{
    //! Sets the number of cores requested
    /*!
     * \param ncores (Int_t) number of cores requested
     */
    void setNumberOfCores(Int_t ncores)
    {
        _nAllocatedCores = ncores;
    }
    /*!
     * \brief Get the Number of cores allocated  
     * \return Number of cores allocated  
     */
    Int_t getNumberOfCores() const
    {
        return _nAllocatedCores;
    }
    //! Sets the requested time in seconds
    /*!
     * \param nseconds (Int_t) number of second allocated
     */
    void setAllocatedTime(Int_t nseconds)
    {
        _nAllocatedSeconds = nseconds;
    }
    /*!
     * \brief Get the Number of second allocated 
     * \return Number of second allocated  
     */
    Int_t getAllocatedTime() const
    {
        return _nAllocatedSeconds;
    }
    //! Sets the project name
    /*!
     * \param project_name (TString) name of the project
     */
    void setProjectName(TString project_name)
    {
        _sProjectName = project_name;
    }
    /*!
     * \brief Get the Project Name
     * \return project Name  
     */
    TString getProjectName() const
    {
        return _sProjectName;
    }
    //! Sets the Cluster name
    /*!
     * \param cluster_name (TString) name of the cluster
     */
    void setClusterName(TString cluster_name)
    {
        _sClusterName = cluster_name;
    }
    /*!
     * \brief Get the Cluster Name
     * \return cluster Name  
     */
    TString getClusterName() const
    {
        return _sClusterName;
    }
    //! Add extra directive
    /*!
     * \param directive (TString) extra directive
     */
    void addExtraDirective(const TString& directive)
    {
        _sListExtraDirectives.push_back(directive);
    }
    //@}

private:
    /*! A private variable
     * Contain the Number of second allocated  
     */
    Int_t _nAllocatedSeconds;
    /*! A private variable
     * Contain the Number of cores allocated  
     */
    Int_t _nAllocatedCores;
    /*! A private variable
     * Contain the Name of the Project  
     */
    TString _sProjectName;
    /*! A private variable
     * Contain the Name of the Cluster  
     */
    TString _sClusterName;
    /*! A private variable
     * List of Extra Directives 
     */
    vector<TString> _sListExtraDirectives;

    ClassDef(URANIE::Launcher::TLauncherByStep, ID_LAUNCHER)
    //The description
};

} // End Of namespace Launcher
} // End Of namespace URANIE

#endif
//  End Of File TLauncherByStep.h
