//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TLauncherByStepRemote.h
 \author Daniel Beltran<daniel.beltran@bsc.es>
 \date 2019-02-12

 \brief

 */

#ifndef TLAUNCHERBYSTEPREMOTE_H
#define TLAUNCHERBYSTEPREMOTE_H

#include "TNamed.h"

#include <iostream>
#include <fstream>
//#define LIBSSH_STATIC 1
//#include <libssh/libssh.h>
using std::cout;
using std::endl;

// ROOT
#include "Rtypes.h"

// Uranie
#include "TLauncher.h"
#include "Launcher.h"
#include "TCluster.h"
#include "TEcCluster.h"

#include "TInputFileFlag.h"
#include "TAttribute.h"
#define LIBSSH_STATIC 1
#include <libssh/libssh.h>
#include <libssh/sftp.h>
#include <semaphore.h>

#ifdef WITH_LIBSSH
namespace URANIE
{
namespace Launcher
{
class TLauncherByStepRemote: public TLauncher
{


public:
    TCluster *_cluster;

    enum EDistrib { kUnknown,kLSB, kSLURM,kMPI, kPBS, kSGE , kEc};

    EDistrib  _edistrib;         ///< The type of distribution
    TString eType;

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor 
    TLauncherByStepRemote(URANIE::DataServer::TDataServer *tds, TCode *code,EDistrib _edistrib=kUnknown, Option_t *option="" );
    TLauncherByStepRemote(URANIE::DataServer::TDataServer *tds, TCode *code,TCluster *cluster, Option_t *option="" );

    //! Default destructor
    virtual ~TLauncherByStepRemote();
    //@}
    //---------------------------------------------
    /** @name Manipulate the code
     **/
    //@{
    //! PreTreatment
    /*!
     Creates all the working subdirectory which contains the input files
     */
    virtual void preTreatment(Option_t *option = "");


    //! Evaluate the code in each working subdirectory
    /*!
     This step can be done by an another code
     */
    virtual void run(Option_t *option = "");



    void clusterRun(Option_t *option = "");


    //! PostTreatment
    /*!
     Extracts the values of the output attributes in the output files for each working subdirectory
     */
    virtual void postTreatment(Option_t *option = "");
    //@}
    //! retrieveOutputs
    /*!
     Retrieves the outputfiles from the ssh server.
     */
    void *retrieveOutputs(void);
    static void  *staticRetrieveOutputs(void *dummyPtr)
    {
        return ((TLauncherByStepRemote *)dummyPtr)->retrieveOutputs();

    }

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Print the log of the class
    virtual void printLog(Option_t *option = "");
    //@}

    //-----------------------------------------------
    /** @name manipulating variables for batch systems
     **/
    //@{
    //! Sets the number of cores requested
    /*!
     * \param ncores (Int_t) number of cores requested
     */
    void setNumberOfCores(Int_t ncores)
    {
        _nAllocatedCores = ncores;
    }

    //! Set singlejob flag to true
    /*!
     * \param ncores (Int_t) number of cores requested
     */
    void setSingleJob()
    {
        _sSingleJob = !_sSingleJob;
    }
    //! get singlejob flag value
    /*!
     * \param bool_t
     */
    Bool_t getSingleJob()
    {
        return _sSingleJob;
    }
    /*!
     * \brief Get the Number of cores allocated  
     * \return Number of cores allocated  
     */
    Int_t getNumberOfCores() const
    {
        return _nAllocatedCores;
    }
    //! Sets the requested time in seconds
    /*!
     * \param nseconds (Int_t) number of second allocated
     */
    void setAllocatedTime(Int_t nseconds)
    {
        _nAllocatedSeconds = nseconds;
    }
    /*!
     * \brief Get the Number of second allocated 
     * \return Number of second allocated  
     */
    Int_t getAllocatedTime() const
    {
        return _nAllocatedSeconds;
    }
    //! Sets the project name
    /*!
     * \param project_name (TString) name of the project
     */
    void setProjectName(TString project_name)
    {
        _sProjectName = project_name;
    }

    /*!
     * \brief Get the Project Name
     * \return project Name  
     */
    TString getProjectName() const
    {
        return _sProjectName;
    }



    void setNumberOfChunks(Int_t numberofChunks)
    {
       _sNumberOfChunks = numberofChunks;

    }
    void setJobName(TString path)
    {
        _sJobName = path;
        _cluster->setJobName(path);
    }

    void addCluster(TCluster *cluster)
    {
        _cluster->_tds = _tds;
        _cluster->_code= _code;
        _cluster->setCluster(cluster);

    }

    //! getNumberOfChunks
    /*!
     * \brief Get the path to remoteWorkingDir
     * \return Output Path Header Name
     */
    Int_t getNumberOfChunks() const
    {
        return _sNumberOfChunks;
    }
                    //! getNRetries
    /*!
     * \brief Get the number of retries
     * \return Get the number of retries
     */
    Int_t getNRetries() const
    {
        return _sRetries;
    }

    void UpdateJobListStatus();
    void RetrieveJobOutput(std::pair<Int_t,Int_t> split);
    void compileBinary();

    TCluster* getCluster();
    TCluster* setCluster();

    //CODE-Compiler

    //! addCompilerDirectives
    /*!
     * \param directive  (TString) name
     * \param directive  (TString) value
     */

    void addCompilerDirective(TString directive, TString value);

    void addCodeDependency(TString destinationLib, TString sourceList);
    void addCodeLibrary(TString libName, TString sourcesList);
    void addCodeExternalLibrary(TString libName, TString libPath , TString includePath);
    void addCodeExternalLibrary(TString libName);


    void substituteCompilerCode();







    //@}

private:


    /*! A private variable
     * Contain the Number of second allocated  
     */

    Int_t _nAllocatedSeconds;
    /*! A private variable
     * Contain the Number of cores allocated  
     */
    Int_t _nAllocatedCores;
    /*! A private variable
     * Contain the Name of the Project  
     */
    TString _sProjectName;



    /*! A private variable
     * List of active Jobs
     */
    list<std::pair<TString,std::pair<Int_t,Int_t>>> _sActiveJobs;
    /*! A private variable
     * List of completed Jobs
     */
    list<std::pair<TString,std::pair<Int_t,Int_t>>> _sCompletedJobs;

     sem_t _sshSemaphore;

     /*! A private variable
     ! _sNumberOfChunks
     * Splits the Npatterns into  X Chunks in order to speedup the posttreatment function (default 4)
     */
     Int_t _sNumberOfChunks=1;
    /*! A private variable
    ! _sChunkSize
    * Splits the Npatterns into  X Chunks in order to speedup the posttreatment function (default 4)
    */
    Int_t _sChunkSize;


    /*! A private variable
    ! _sSingleJob
    * a flag to determine if _sJobName is not used, default false
    */
    Bool_t _sSingleJob=false;
    /*! A private variable
     *
! _sJobName
* jobname where the input/output files will be stored.
*/
     TString _sJobName;
    /*! A private variable
    ! _sRetries
    * Number Of retries if an inputfile is not found
    */
    Int_t _sRetries =5;
    /*! A private variable
    * Store the jobid submitted
    */
    list<std::pair<TString,std::pair<Int_t,Int_t>>> _sJobList;
    /*! A private variable
* Show job status submitted
*/
    list<TString> _sJobStatus;
    /*! A private variable
    * Store the jobid in command formant
    */
    TString _sJobIDs="";

    /*! A private variable
    * List of Cluster Directives
    */
    std::pair<Int_t,Int_t>  _sLastSplit;

    /*! A private variable
    * List of Compiler Directives
    */
    std::map<TString,TString> _sListCompilerDirectives;
    /*! A private variable
    * List of Code Libraries Directives
    */
    std::map<TString,TString> _sCodeLibs;
    /*! A private variable
* List of Code Libraries Directives
*/
    list<pair<TString,pair<TString,TString>>  >  _sExternalLib;

    /*! A private variable
    * List of Code Dependencies Directives
    */
    std::map<TString,TString> _sCodeDependencies;
    TString _sCompilerSkeleton = "CMakeLists.txt.in";
    ClassDef(URANIE::Launcher::TLauncherByStepRemote, ID_LAUNCHER)
    //The description
};

} // End Of namespace Launcher
} // End Of namespace URANIE
#endif
#endif
//  End Of File TLauncherByStepRemote.h
