//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileDataServer.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   tuesday june 3rd 16:06:57 CEST 2007

 \brief Implementation of the class URANIE::Launcher::TOutputFileDataServer

 \par Modifications
\li Jean-Baptiste Blanchard 2016/04/08
 - Add the searchOutput with a map to read strings/vectors
 \li F. Gaudier (2010-08-06 12:01:56)
 - change the deprecated setKey method by setFileKey method

 */

//Uranie
#include "TOutputFileDataServer.h"
#include "UExceptions.h"
#include "TTokenizer.h"


// ROOT
#include "TMatrixT.h"
#include "TMatrixTUtils.h"
#include "TArrayD.h"
#include "TLeaf.h"

using std::cout;
using std::endl;
using std::ostringstream;

ClassImp (URANIE::Launcher::TOutputFileDataServer)

namespace URANIE
{
namespace Launcher
{

TOutputFileDataServer::TOutputFileDataServer(TString str) :
    TOutputFile(str), _nLines(0), _nColumns(0), _nLength(0)
{
    if (_blog)
	cout << "constructor TOutputFileDataServer file[" << str << "]" << endl;
    setCommentCharacter("#");
    _sstrbeg = _sstrend = "";
    _svecbeg = _svecend = ""; _svecdel = ",";
}

TOutputFileDataServer::~TOutputFileDataServer()
{
    if (_blog)
	cout << "destroy TOutputFileDataServer" << endl;
}

/////////////////////
//// Add attribut to the Output File
/////////////////////
void TOutputFileDataServer::addAttribute(TString str, Int_t nline)
{
    URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
    att->delShare();
    addAttribute(att, nline);
}

void TOutputFileDataServer::addAttribute(URANIE::DataServer::TAttribute* att,
					 Int_t nline)
{
    Int_t nlinecrt = nline;

    if (nlinecrt < 1)
    {
        nlinecrt = 1;
        Int_t natt = getAttributesNumber();
        if (natt > 0)
	{
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
		_attributes[natt - 1]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		KeyList.begin();
            TEventList * eventfield = (*it)->_sEventList;
            nlinecrt = (Int_t) eventfield->GetEntry(0) + 1;
	}
    }

    att->setFileKey(_sFileName, att->GetName());
    att->setFieldOfKey(0, nlinecrt);
    TCodeFile::addAttribute(att);
}

/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileDataServer::searchOutput(Int_t &indcrt, Double_t *val,
					   TList *lstOut)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileDataServer::searchOutput indcrt["
	     << indcrt << "]" << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = getAttributesNumber();
    if (_blog)
    {
        cout << " *** natt[" << natt << "]" << endl;
    }

    TMatrixT<double> *matOut = new TMatrixT<double>(1, natt);

    // Default Value
    // In case of no output
    Double_t valdef = 1.23456789;

    Double_t * tempVal = matOut->GetMatrixArray();
    for (Int_t i = 0; i < natt; i++)
    {
        valdef = 1.23456789;
        _attributes[i]->getDefaultValue(valdef);
        tempVal[i] = valdef;
    }

    if (_blog)
    {
        cout << endl << " *** Default Value";
        matOut->Print();
        cout << " *** End Of Default Value" << endl;
    }

    std::ifstream file(_sFileName.Data(), ios::in);
    try
    {
        if (!file)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TOutputFileDataServer::searchOutput ERROR : can't open file [%s]",
							   _sFileName.Data()));

        // Reading of  headers: all line beginning with a "#" character
        Bool_t bentete = kTRUE;
        map < TString, TString > header_file;

        // Read the whole line each character by character until End Of Line is reached
        do
	{
            ostringstream sLine;
            char ch;
            while (file.get(ch) && ch != '\n')
	    {
                sLine.put(ch);
	    }

            //           if ( _blog )
            //             cout << " ** sLine [" << sLine.str() << ']' << endl;
            // Reads the whole line
            //  file.getline( fLine, fLen-1 , '\n');
            //if (fLine[strlen(fLine)-1] == '\n') fLine[strlen(fLine)-1] = 0;

            string stringSLine = sLine.str();
            TString logLine = TString(stringSLine.c_str());
            //           if( _blog )
            //             cout << " ** logLine[" << logLine << "]" << endl;

            // If the line is a header's one, continue, else breaks the loop
            if (logLine.BeginsWith("#"))
		bentete = kTRUE;
            else
	    {
                bentete = kFALSE;
                break;
	    }

            // Search of notion [name, unity, type, label]
            Ssiz_t npoint = logLine.Index(":");
            //           if( _blog )
            //             cout << "npoint[" << npoint << "]" << endl;
            TString sLabel;
            if (npoint > 0)
		sLabel = logLine(1, npoint - 1);
            //    else {
            //      npoint = 0;
            //      sLabel = TString("COLUMN_NAMES");
            //    }
            //           if( _blog )
            //             cout << "sLabel[" << sLabel << "]" << endl;
            header_file[sLabel] = logLine.Remove(0, npoint + 1);

	} while (bentete);

        // creation of the vector of attributes
        TString headerTag = TString("headertag");
        if (header_file["COLUMN_NAMES"].IsNull())
	{
            if (header_file["COLUMN_TITLES"].IsNull())
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileDataServer::searchOutput ERROR : neither header line COLUMN_NAMES nor header line COLUMN_TITLES specified in the file \"%s\".",
							       _sFileName.Data()));
            else
		headerTag = TString("COLUMN_TITLES");
	}
        else
	    headerTag = TString("COLUMN_NAMES");

        header_file[headerTag].ReplaceAll(" ", "");
        TTokenizer *tt = new TTokenizer();
        tt->setDelimiters("|");
        vector < string > tokens = tt->getTokens(header_file[headerTag].Data());
        _nColumns = (Int_t)(tokens.size());
        if (_blog)
	{
            cout << " ** headerTag[" << headerTag << "] --- ["
		 << header_file[headerTag].Data() << "] _nColumns["
		 << _nColumns << "]" << endl;
	}
        delete tt;
        tt = NULL;
        // Read all the data and fill then in the TDSNtupleD
        TString sVariableNames = header_file[headerTag];
        sVariableNames.ReplaceAll("|", ":");
        URANIE::DataServer::TDSNtupleD *datatree =
	    new URANIE::DataServer::TDSNtupleD(
		Form("%s__output__tree__", GetName()),
		"title of TOutputFileDataServer file",
		sVariableNames.Data());
        Double_t *inValues = new Double_t[_nColumns];
        _nLines = 0;
        string line;
        while (getline(file, line))
	{
            istringstream streamline(line);
            for (Int_t indx = 0; indx < _nColumns; indx++)
		streamline >> inValues[indx];

            if (!(line.length() > 0))
		break;

            datatree->Fill(inValues);
            _nLines++;

        }
        if ( _blog && datatree->GetEntries() < 15 )
            datatree->Scan("*");

        // Now fill the matrix
        // Case of several lines
        if (_blog)
	{
            cout << " *** _nLines[" << _nLines << "]" << endl;
            cout << " *** _nColumns[" << _nColumns << "]" << endl;
	}
	delete matOut;
        matOut = new TMatrixT<double>(_nLines, natt);

        Double_t * theVal = matOut->GetMatrixArray();

        Int_t ncrt = 0;
        Int_t nindcrt = indcrt;

        // Default Value for all the matrix
        for (Int_t ichamp = 0; ichamp < natt; ichamp++)
	{
            ncrt = ichamp;
            valdef = 1.23456789;
            _attributes[ichamp]->getDefaultValue(valdef);
            for (Int_t ilength = 0; ilength < _nLines; ilength++)
	    {
                if (ilength == 0)
		    val[nindcrt++] = valdef;
                theVal[ncrt] = valdef;
                ncrt += natt;
	    }
	}

        if (_blog)
	{
            cout << endl << " *** Default Value For the Matrix";
            matOut->Print();
            cout << " *** End Of Default Value For the Matrix" << endl;
	}

        // Case of all the matrix
        nindcrt = indcrt;
        for (Int_t iout = 0; iout < natt; iout++)
	{
            ncrt = iout;
            TLeaf *tleaf = datatree->GetLeaf(_attributes[iout]->GetName());
            if (tleaf == NULL)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileDataServer::searchOutput method\nthe output file [%s] does not contains the output variable [%s]",
							       _sFileName.Data(),
							       _attributes[iout]->GetName()));

            for (Int_t ilength = 0; ilength < _nLines; ilength++)
	    {
                datatree->GetEntry(ilength);
                if (_blog)
		    cout << " *** iout(" << iout << "/" << natt << ") -- "
			 << _attributes[iout]->GetName() << "  ** Field("
			 << ilength << "/" << _nLines << ") value["
			 << tleaf->GetValue(0) << "]" << endl;
                if (ilength == 0)
		    val[indcrt++] = tleaf->GetValue(0);
                theVal[ncrt] = tleaf->GetValue(0);
                ncrt += natt;
	    }
	}

        // Pour le retour
        if (lstOut != NULL)
	{
            if (_blog)
	    {
                cout << endl << " *** Returned Value For the Matrix";
                matOut->Print();
                cout << " *** End Of Returned Value For the Matrix" << endl;
	    }
            lstOut->Add(matOut);
	}
	else
	{
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	}
        // Clean
        delete[] inValues;
        inValues = NULL;
        delete datatree;
        datatree = NULL;

    } catch (URANIE::Exceptions::UErrorExceptions&)
    {
        if (lstOut != NULL)
	{
            if (_blog)
	    {
                cout << " *** catch UErrorExceptions" << endl;
                matOut->Print();
                cout << " *** End Of catch UErrorExceptions" << endl;
	    }
            lstOut->Add(matOut);
	}
	else
	{
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	    return kFALSE;
	}
    }

    return kTRUE;
}

/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileDataServer::searchOutput(vector<URANIE::DataServer::UEntry*> *out)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileDataServer::searchOutput " << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = getAttributesNumber();
    if (_blog)
	cout << " *** natt[" << natt << "]" << endl;

     // Create a default map containing just the default of Attribute of 1.23456789
    // // Default Value

    for (Int_t i = 0; i < natt; i++)
    {
        TString def=_attributes[i]->getDefault();
        out->at(i)->fillEntry( ((def=="" && out->at(i)->tag==URANIE::DataServer::UEntry::Double ) ? "1.23456789" : def) );
    }

    if (_blog)
    	cout << endl << " *** Default Value  "<< "" << " *** End Of Default Value" << endl;


    ifstream file(_sFileName.Data(), ios::in);
    try
    {
        if (!file)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TOutputFileDataServer::searchOutput ERROR : can't open file [%s]",
							   _sFileName.Data()));

        // Reading of  headers: all line beginning with a "#" character
        Bool_t bentete = kTRUE;
        map < TString, TString > header_file;

        // Read the whole line each character by character until End Of Line is reached
        do
	{
            ostringstream sLine;
            char ch;
            while (file.get(ch) && ch != '\n')
	    {
                sLine.put(ch);
	    }

            string stringSLine = sLine.str();
            TString logLine = TString(stringSLine.c_str());

            // If the line is a header's one, continue, else breaks the loop
            if (logLine.BeginsWith("#"))
		bentete = kTRUE;
            else
	    {
                bentete = kFALSE;
                break;
	    }

            // Search of notion [name, unity, type, label]
            Ssiz_t npoint = logLine.Index(":");
            //           if( _blog )
            //             cout << "npoint[" << npoint << "]" << endl;
            TString sLabel;
            if (npoint > 0)
		sLabel = logLine(1, npoint - 1);

            header_file[sLabel] = logLine.Remove(0, npoint + 1);

	} while (bentete);

        // creation of the vector of attributes
        TString headerTag = TString("headertag");
        if (header_file["COLUMN_NAMES"].IsNull())
	{
            if (header_file["COLUMN_TITLES"].IsNull())
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileDataServer::searchOutput ERROR : neither header line COLUMN_NAMES nor header line COLUMN_TITLES specified in the file \"%s\".",
							       _sFileName.Data()));
            else
		headerTag = TString("COLUMN_TITLES");
	}
        else
	    headerTag = TString("COLUMN_NAMES");

        header_file[headerTag].ReplaceAll(" ", "");
        TTokenizer *tt = new TTokenizer();
        tt->setDelimiters("|");
        vector < string > tokens = tt->getTokens(header_file[headerTag].Data());
        _nColumns = (Int_t)(tokens.size());
        if (_blog)
	{
            cout << " ** headerTag[" << headerTag << "] --- ["
		 << header_file[headerTag].Data() << "] _nColumns["
		 << _nColumns << "]" << endl;
	}
        delete tt;
        tt = NULL;


	string line;
	vector<string> subline;
	// Changed to used the fill (string) method of TDSNtupleD, JBB (2016-02) to cope with strings/vector<double>
	while ( !URANIE::DataServer::UEntry::safeGetline(file, line).eof() )
	{
	    if (!(line.length() > 0) || URANIE::DataServer::UEntry::isWhiteLine(line))
		break;
	    
	    URANIE::DataServer::UEntry::StringParser(line, " ", subline);
	    for(UInt_t indx = 0; indx < tokens.size(); indx++)
	    {

		for(Int_t indatt = 0; indatt < natt; indatt++)
		{
		    list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[indatt]->getKeyList();
		    list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin();
		    TString skey = (*it)->_sKey;
		
		    if( skey != tokens.at(indx) )
			continue;

		    out->at(indatt)->fillEntry(subline.at(indx), _svecdel);
		}
	    }
	}
    } catch (URANIE::Exceptions::UErrorExceptions&)
    {
	if (_blog)
	    cout << " *** catch UErrorExceptions" << endl;
        return kFALSE;
    }

    return kTRUE;
}

void TOutputFileDataServer::printLog(Option_t *option)
{
    cout << "*** TOutputFileDataServer::printLog[" << option << "]" << endl;
    cout << " *** _sFileName[" << _sFileName << "]" << endl;
    cout << " *** _nColumns[" << _nColumns << "]" << endl;
    cout << " *** _nLines[" << _nLines << "]" << endl;
    cout << " *** _nLength[" << _nLength << "]" << endl;
    TOutputFile::printLog(option);
    cout << "*** End of TOutputFileDataServer::printLog[" << option << "]"
	 << endl;
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

