//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileFMU.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/10/14 at 11:52:12

 \brief Implementation of the class URANIE::Launcher::TOutputFileFMU

 */
#include <stdlib.h>
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TMatrixT.h"

// Uranie
#include "TOutputFileFMU.h"
#include "UExceptions.h"

// LibXML2
#include <libxml/xpathInternals.h>

ClassImp (URANIE::Launcher::TOutputFileFMU)

namespace URANIE
{
namespace Launcher
{

TOutputFileFMU::TOutputFileFMU(TString str) :
    TOutputFile(str),
    _sVersion(fmi_version_unknown_enu),
    _fmi1Import(nullptr), _fmi2Import(nullptr),
    _nOutputVector(0), _nCrtOutputVector(0), _lstOfVector(nullptr),
    _vecOutputs(0)
{

    if (_blog)
        cout << " *** Method TOutputFileFMU::" << __func__ << " with str[" << str << "]" << endl;

    _sstrbeg = _sstrend = "";
    _svecbeg = _svecend = ""; _svecdel = " ";

    if ( !gSystem->IsAbsoluteFileName(_sFileName.Data()) ) {
        if (_blog)
            cout << " *** _sFileName (" << _sFileName << ") not absolute -- > TODO" << endl;
        TString sPWD = gSystem->pwd();
        // buffer used to prevent memleak, as gSystem::ConcatFileName result must be deleted
        char *foo=NULL;
        foo = gSystem->ConcatFileName(gSystem->pwd(), _sFileName.Data());
        _sFileName = foo;
        if (_blog)
            cout << " *** _sFileName (" << _sFileName << ") done" << endl;
        delete[] foo; foo=NULL;
    }

    _spwd = Form("%s/", gSystem->pwd());

    if (_blog)
        cout << " *** End of Method TOutputFileFMU::" << __func__ << " with str[" << str << "]" << endl;
}

TOutputFileFMU::~TOutputFileFMU()
{
    FUNCT_TRACE();
    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method [TOutputFileFMU::" << __func__  << "] Name[" << GetName() << "]" << endl;

    // Clean
    if ( _lstOfVector != nullptr )
        delete _lstOfVector;
    _lstOfVector = nullptr;

    if (_blog)
        cout << " *** End of Method [TOutputFileFMU::" << __func__  << "] Name[" << GetName() << "]" << endl <<
                " ********************************" << endl;
}

// ----------------------------------------------------------------------
//  addAttribute Method
// ----------------------------------------------------------------------
void TOutputFileFMU::addAttribute(TString str, TString sKey)
{
    FUNCT_TRACE();
    URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
//    att->delShare();
    addAttribute(att, sKey);
} // End of method addAttribute

// ----------------------------------------------------------------------
//  addAttribute Method
// ----------------------------------------------------------------------
void TOutputFileFMU::addAttribute(URANIE::DataServer::TAttribute* att, TString sKey)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method [TOutputFileFMU::" << __func__  << "]" <<
                "      attribute[" << att->GetName() << "]" <<
                "      sKey[" << sKey << "]" <<
                "      TOutputFileFMU::FileName[" << _sFileName << "]" << endl;

    try {
        // DataType kCategory not yet implemented
        if ( att->getDataType() == URANIE::DataServer::TAttribute::kCategory )
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TOutputFileKey::%s ERROR\n The Data-Type \"kCategory\" for output attribute [%s] is not yet implemented.",
                             __func__, att->GetName()));
        // DataType kVector
        if ( att->getDataType() == URANIE::DataServer::TAttribute::kVector )
            _nOutputVector++;
        // An Output attribute is associated with at most one file
        if ( !att->getKeyList().empty() ) {
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TOutputFileKey::%s ERROR\n The Output attribute [%s] have an Output File yet: only one is autorised",
                             __func__, att->GetName()));
        }
        // The FMU output file can be only kXMLFMU
        att->setFileFMU(_sFileName, sKey);
        // Add the attribute into the list of attributes
        TCodeFile::addAttribute(att);
        if (_blog)
            att->printLog();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << endl <<
                " *** End of Method [TOutputFileFMU::" << __func__  << "]" <<
                "      attribute[" << att->GetName() << "]" <<
                "      sKey[" << sKey << "]" <<
                "      TOutputFileFMU::FileName[" << _sFileName << "]" <<
                " ********************************" << endl;
} // End of method addAttribute

// ----------------------------------------------------------------------
//  searchIndexeOutputAttribute Method
// ----------------------------------------------------------------------
void TOutputFileFMU::searchIndexeOutputAttribute(URANIE::DataServer::TAttribute* att)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method [TOutputFileFMU::" << __func__  << "]" <<
                "      _sFileName[" << _sFileName << "] " <<
                "      Attribute [" << att->GetName() << "]" << endl;


    if ( _blog )
        cout << " *** _sVersion[" << _sVersion << "]" << endl;

    try {
        switch (_sVersion) {
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TOutputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not defined for attribute[%s].",
                             __func__, _sVersion, _sFileName.Data(), att->GetName()));
            break;
        case fmi_version_unsupported_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TOutputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not supported.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_1_enu:
        case fmi_version_2_0_enu:
            Int_t nfield = -1;
            TString skey = "", sfilecrt ="", srelativefilecrt ="";
            list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();

            for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it) {
                srelativefilecrt = sfilecrt = (*it)->_sFile;

                if (_blog)
                    cout << " *** _spwd[" << _spwd << "] sfilecrt[" << sfilecrt << "] nfield[" << nfield << "]" << endl;
                srelativefilecrt.ReplaceAll(_spwd.Data(), "");
                if (_blog)
                    cout << " *** srelativefilecrt[" << srelativefilecrt << "] _sFileName[" << _sFileName << "]" << endl;

                if ( ( sfilecrt.CompareTo(_sFileName, TString::kExact) == 0) ||
                     ( srelativefilecrt.CompareTo(_sFileName, TString::kExact) == 0) ) {
                    skey = (*it)->_sKey;
                    nfield = -2;

                    if (_blog)
                        cout << " *** Attribute[" << att->GetName() << "]" <<
                            " sfilecrt[" << sfilecrt.Data() << "]" <<
                            " :: skey[" << skey << "]" <<
                            " nfield[" << nfield << "]" << endl;

                    // Get the FMI Attribute ( 2.0 )
                    if ( _sVersion == fmi_version_2_0_enu ) {
                        fmi2_import_variable_t* fmiAtt = fmi2_import_get_variable_by_name(_fmi2Import, skey.Data());

                        if ( !fmiAtt) {
                            throw URANIE::Exceptions::UErrorExceptions(
                                        __FILE__,
                                        __LINE__,
                                        Form("TOutputFileFMU::%s ERROR\n: the key \"%s\" of attribute \"%s\" not found in the file \"%s\".",
                                             __func__, skey.Data(), att->GetName(), _sFileName.Data()));
                        }

                        if (_blog)
                            printf(" *** Att[%s]::fmi2_import_get_variable_name[%s]::Type[%d]\n",
                                   att->GetName(),
                                   fmi2_import_get_variable_name(fmiAtt),
                                   att->getDataType());

                        // Verify the type of ATT is the same in the FMU model
                        switch (att->getDataType()) {
                        case URANIE::DataServer::TAttribute::kReal:
                        case URANIE::DataServer::TAttribute::kDefault:
                        case URANIE::DataServer::TAttribute::kVector:
                            if ( !fmi2_import_get_variable_as_real(fmiAtt) &&
                                 !fmi2_import_get_variable_as_integer(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"Real\" is not an \"REAL|INTEGER\" in the FMU model.",
                                                 __func__,  att->GetName()));
                            break;
                        case URANIE::DataServer::TAttribute::kString:
                            if ( !fmi2_import_get_variable_as_string(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"String\" is not an \"STRING\" in the FMU model.",
                                                 __func__,  att->GetName()));
                            break;
                        case URANIE::DataServer::TAttribute::kCategory:
                            if ( !fmi2_import_get_variable_as_enum(fmiAtt)  &&
                                 !fmi2_import_get_variable_as_boolean(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"Category\" is not an \"ENUM|BOOLEAN\" in the FMU model.",
                                                 __func__,  att->GetName()));
                            break;
                        default:
                            break;
                        }

                        // Get the ValueReference index
                        nfield = fmi2_import_get_variable_vr(fmiAtt);
                        if (_blog)
                            cout << " *** from fmi2_import_get_variable_vr nfield[" << nfield << "]" << endl;

                        // Case of kVector
                        if ( att->getDataType() == URANIE::DataServer::TAttribute::kVector ) {
                            if (_blog)
                                cout << " *** att name[" << att->GetName() << "] is kVector :: " <<
                                        " _nCrtOutputVector[" <<_nCrtOutputVector << "] Vs" <<
                                        " _nOutputVector[" << _nOutputVector << "]";
                            if ( !_lstOfVector )
                                _lstOfVector = new TArrayI(_nOutputVector);
                            if ( _lstOfVector )
                                _lstOfVector->AddAt(nfield, _nCrtOutputVector++);
                            if (_blog)
                                cout << "   _lstOfVector[" << _nCrtOutputVector << "]    " << endl;
                        }
                    } // End of FMI Attribute ( 2.0 )

                    // Get the FMI Attribute ( 1.0 )
                    if ( _sVersion == fmi_version_1_enu ) {
                        fmi1_import_variable_t* fmiAtt = fmi1_import_get_variable_by_name(_fmi1Import, skey.Data());

                        if ( !fmiAtt) {
                            throw URANIE::Exceptions::UErrorExceptions(
                                        __FILE__,
                                        __LINE__,
                                        Form("TOutputFileFMU::%s ERROR\n: the key \"%s\" of attribute \"%s\" not found in the file \"%s\" (FMI 1).",
                                             __func__, skey.Data(), att->GetName(), _sFileName.Data()));
                        }

                        if (_blog)
                            printf(" *** Att[%s]::fmi1_import_get_variable_name[%s]::Type[%d]\n",
                                   att->GetName(),
                                   fmi1_import_get_variable_name(fmiAtt),
                                   att->getDataType());

                        switch (att->getDataType()) {
                        case URANIE::DataServer::TAttribute::kReal:
                        case URANIE::DataServer::TAttribute::kDefault:
                        case URANIE::DataServer::TAttribute::kVector:
                            if ( !fmi1_import_get_variable_as_real(fmiAtt) &&
                                 !fmi1_import_get_variable_as_integer(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"Real\" is not an \"REAL|INTEGER\" in the FMU model (FMI 1).",
                                                 __func__,  att->GetName()));
                            break;
                        case URANIE::DataServer::TAttribute::kString:
                            if ( !fmi1_import_get_variable_as_string(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"String\" is not an \"STRING\" in the FMU model (FMI 1).",
                                                 __func__,  att->GetName()));
                            break;
                        case URANIE::DataServer::TAttribute::kCategory:
                            if ( !fmi1_import_get_variable_as_enum(fmiAtt)  &&
                                 !fmi1_import_get_variable_as_boolean(fmiAtt) )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TOutputFileFMU::%s ERROR:\n The output attribute \"%s\" specified as \"Category\" is not an \"ENUM|BOOLEAN\" in the FMU model (FMI 1).",
                                                 __func__,  att->GetName()));
                            break;
                        default:
                            break;
                        }

                        // Get the ValueReference index
                        nfield = fmi1_import_get_variable_vr(fmiAtt);
                        if (_blog)
                            cout << " nfield[" << nfield << "]";

                        // Case of kVector
                        if ( att->getDataType() == URANIE::DataServer::TAttribute::kVector ) {
                            if (_blog)
                                cout << " *** att name[" << att->GetName() << " is kVector :: " <<
                                        " _nCrtOutputVector[" <<_nCrtOutputVector << "/" << _nOutputVector << "]";

                            if ( !_lstOfVector )
                                _lstOfVector = new TArrayI(_nOutputVector);
                            if ( _lstOfVector ) {
                                _lstOfVector->AddAt(nfield, _nCrtOutputVector++);
                            }
                            if (_blog)
                                cout << "   _lstOfVector[" << _nCrtOutputVector << "]    " << endl;
                        }
                    } // End of FMI Attribute ( 1.0 )

                    //
                    if ( nfield > -1 ) {
                        if (_blog)
                            cout << " ***  setFieldOfKey the Attribute[" << att->GetName() << "]" <<
                                    " sfilecrt[" << sfilecrt << "]" <<
                                    " :: skey[" << skey << "]" <<
                                    " nfield[" << nfield << "]" << endl;
                        list<URANIE::DataServer::TAttributeFileKey *>::iterator first = KeyList.begin();
                        att->setFieldOfKey(distance(first, it), nfield);
                    } // End of nfield > -1

                } // End of _sFileName == sfilecrt
            } // End of for  it

            if ( nfield == -1 )
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TOutputFileFMU::%s ERROR\n Output file \"%s\" not found for the attribute \"%s\".",
                                 __func__, _sFileName.Data(), att->GetName()));
            break;
        } // End of Switch _sVersion

    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " *** End of Method [TOutputFileFMU::" << __func__  << "]" <<
                "      _sFileName[" << _sFileName << "] " <<
                "      Attribute [" << att->GetName() << "]" << endl <<
                " ********************************" <<endl;
} // End of Method searchIndexeOutputAttribute

// ----------------------------------------------------------------------
//  searchOutput Method
// ----------------------------------------------------------------------
Bool_t TOutputFileFMU::searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut)
{
    UNUSED(val);
    UNUSED(lstOut);
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method [TOutputFileFMU::" << __func__  << "] indcrt[" << indcrt << "]" << endl;

    // Function UNUSED(searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut))

    return kTRUE;
} // End of Method searchOutput

// ----------------------------------------------------------------------
//  setFMU1Import Method
// ----------------------------------------------------------------------
void TOutputFileFMU::setFMU1Import( fmi1_import_t* crtFMU1Import)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TOutputFileFMU::" << __func__ << endl;


    if( crtFMU1Import != nullptr) {

        _fmi1Import = crtFMU1Import;

        // Make the searchIndexeOutputAttribute action
        Int_t natt = (Int_t)(_attributes.size());
        if (_blog)
            cout << " *** NumberOfAttributes[" << natt << "] in FileName[" << GetName() << "]" << endl;

        if ( natt == 0 ) {
            addDefaultOutput();
            natt = (Int_t)(_attributes.size());
        }
        for (Int_t i = 0; i < natt; i++) {
            searchIndexeOutputAttribute(_attributes[i]);
            if (_blog)
                _attributes[i]->printLog();
        }

        if (_blog)
            cout << " *** method TOutputFileFMU::" << __func__ << "] _nOutputVector[" << _nOutputVector << "]  _lstOfVector.size[" << _lstOfVector->GetSize() << "]" << endl;
    }
}

// ----------------------------------------------------------------------
//  setFMU2Import Method
// ----------------------------------------------------------------------
void TOutputFileFMU::setFMU2Import( fmi2_import_t* crtFMU2Import)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TOutputFileFMU::" << __func__ << endl;

    if( crtFMU2Import != nullptr) {

        _fmi2Import = crtFMU2Import;

        // Make the searchIndexeOutputAttribute action
        Int_t natt = (Int_t)(_attributes.size());
        if (_blog)
            cout << " *** NumberOfAttributes[" << natt << "] in FileName[" << GetName() << "]" << endl;

        if ( natt == 0 ) {
            addDefaultOutput();
            natt = (Int_t)(_attributes.size());
        }
        for (Int_t i = 0; i < natt; i++) {
            searchIndexeOutputAttribute(_attributes[i]);
            if (_blog)
                _attributes[i]->printLog();
        }
        if (_blog)
            cout << " *** method TOutputFileFMU::" << __func__ << "] _nOutputVector[" << _nOutputVector << "]  _lstOfVector.size[" << _lstOfVector->GetSize() << "]" << endl;
    }
}

// ----------------------------------------------------------------------
//  addDefaultOutput Method
// ----------------------------------------------------------------------
void TOutputFileFMU::addDefaultOutput()
{
    FUNCT_TRACE();

    if (_blog)
        cout << endl <<
                " ********************************" <<
                " *** Method [TOutputFileFMU::" << __func__  << "]" <<
                "      _sFileName[" << _sFileName << "] " << endl;


    try {

        Int_t nY = -1;

        switch (_sVersion) {
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s Info:\n The FMI version (%i) of the file \"%s\" is not defined.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_unsupported_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not supported.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_1_enu: {
            fmi1_status_t fmi1status;
            fmi1_import_model_counts_t  mc;
            fmi1_import_collect_model_counts(_fmi1Import, &mc);

            if ( _blog) {
                cout << "  *** num_constants  [" << mc.num_constants << "]" << endl;
                cout << "  *** num_parameters [" << mc.num_parameters << "]" << endl;
                cout << "  *** num_discrete   [" << mc.num_discrete << "]" << endl;
                cout << "  *** num_continuous [" << mc.num_continuous << "]" << endl;
                cout << "  *** num_inputs[" << mc.num_inputs << "]" << endl;
                cout << "  *** num_outputs[" << mc.num_outputs << "]" << endl;
                cout << "  *** num_internal[" << mc.num_internal << "]" << endl;
                cout << "  *** num_causality_none[" << mc.num_causality_none << "]" << endl;
                cout << "  *** num vars :: " <<
                        " real(" << mc.num_real_vars << ")" <<
                        " integer(" << mc.num_integer_vars <<")" <<
                        " boolean(" << mc.num_bool_vars <<")" <<
                        " enum(" << mc.num_enum_vars << ")" <<
                        " string(" << mc.num_string_vars<<")" << endl;
            }

            nY = mc.num_continuous;
            if (nY == 0)
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The number of Outputs is zero (%i).",
                                 __func__, nY));

            if ( _blog)
                cout << " *** nY[" << nY << "]" << endl;

            fmi1_import_variable_list_t *lstOutputVariables = fmi1_import_get_variable_list(_fmi1Import);

            if (!lstOutputVariables)
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The listOfOutputVariables of size [%i] is not allocated.",
                                 __func__, mc.num_outputs));

            size_t nAtt = fmi1_import_get_variable_list_size(lstOutputVariables);
            if ( _blog)
                cout << " *** addDefaultOutput::nAtt(" << nAtt << ")" << endl;


            fmi1_import_variable_t *fmiAtt;
            TString sname, sdescription, sunit;
            Int_t nvr;
            fmi1_causality_enu_t eCausality;
            URANIE::DataServer::TAttribute *att;

            for(size_t iatt=0; iatt<nAtt; iatt++) {
                cout << "      *** iatt(" << iatt << "/" << nAtt << ") ::";
                fmiAtt = fmi1_import_get_variable(lstOutputVariables, iatt);

                eCausality = fmi1_import_get_causality (fmiAtt);
                cout << " eCausality(" << eCausality << ")";

                if ( eCausality == fmi1_causality_enu_output) {

                    sname = fmi1_import_get_variable_name(fmiAtt);
                    cout << " name(" << sname << ")";

                    sdescription = fmi1_import_get_variable_description(fmiAtt);
                    cout << " sdescription(" << sdescription << ")";

                    nvr = fmi1_import_get_variable_vr (fmiAtt);
                    cout << " nvr(" << nvr << ")";


                    att = new URANIE::DataServer::TAttribute( sname, sdescription);

                    // Search the type aND the unit (for real) of the variable
                    sunit ="Without Unit";
                    // Case of REAL
                    fmi1_import_real_variable_t *fmiAsRealAtt = fmi1_import_get_variable_as_real(fmiAtt);
                    if ( fmiAsRealAtt ) {
                        cout << "[REAL]";
                        att->setDataType(URANIE::DataServer::TAttribute::kReal);

                        fmi1_import_display_unit_t * displayUnit = fmi1_import_get_real_variable_display_unit(fmiAsRealAtt);

                        if ( displayUnit) {
                            cout << " displayUnit(" << displayUnit << ")";
                            sunit =  fmi1_import_get_display_unit_name(displayUnit);
                            att->setUnit(sunit);
                        }
                        cout << " sunit(" << sunit << ")";
                        att->printLog();
                    } // End of REAL

                    // Case of INTEGER
                    if ( fmi1_import_get_variable_as_integer(fmiAtt) ) {
                        cout << "[INTEGER]";
                        sunit = "Without Unit [INTEGER] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kReal);
                    } // End Of INTEGER

                    // Case of [ENUM|BOOLEAN]
                    if ( fmi1_import_get_variable_as_enum(fmiAtt) ||
                         fmi1_import_get_variable_as_boolean(fmiAtt) ) {
                        cout << "[ENUM|BOOLEAN]";
                        sunit = "Without Unit [ENUM|BOOLEAN] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kCategory);
                    } // End of [ENUM|BOOLEAN]

                    // Case of STRING
                    if ( fmi1_import_get_variable_as_string(fmiAtt) ) {
                        cout << "[STRING]";
                        sunit = "Without Unit [STRING] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kString);
                    } // End of STRING
                addAttribute(att, att->GetName());
                } // End of Causality == Output
                // end of Line
                cout << endl;
            } // End of for iatt
            //clean
            fmi1_import_free_variable_list(lstOutputVariables);
            break;
        }
        case fmi_version_2_0_enu: {
            fmi2_status_t fmistatus;
            fmi2_import_model_counts_t  mc;
            fmi2_import_collect_model_counts(_fmi2Import, &mc);

            if ( _blog) {
                cout << "  *** num_constants  [" << mc.num_constants << "]" << endl;
                cout << "  *** num_fixed      [" << mc.num_fixed << "]" << endl;
                cout << "  *** num_tunable    [" << mc.num_tunable << "]" << endl;
                cout << "  *** num_discrete   [" << mc.num_discrete << "]" << endl;
                cout << "  *** num_continuous [" << mc.num_continuous << "]" << endl;
                cout << "  *** num_parameters [" << mc.num_parameters << "]" << endl;
                cout << "  *** num_calculated_parameters[" << mc.num_calculated_parameters << "]" << endl;
                cout << "  *** num_inputs[" << mc.num_inputs << "]" << endl;
                cout << "  *** num_outputs[" << mc.num_outputs << "]" << endl;
                cout << "  *** num_local[" << mc.num_local << "]" << endl;
                cout << "  *** num_independent[" << mc.num_independent << "]" << endl;
                cout << "  *** num vars :: " <<
                        " real(" << mc.num_real_vars << ")" <<
                        " integer(" << mc.num_integer_vars <<")" <<
                        " enum(" << mc.num_enum_vars << ")" <<
                        " boolean(" << mc.num_bool_vars <<")" <<
                        " string(" << mc.num_string_vars<<")" << endl;
            }

            nY = mc.num_outputs;
            if (nY == 0)
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The number of Outputs is zero (%i).",
                                 __func__, nY));

            if ( _blog)
                cout << " *** nY[" << nY << "]" << endl;


            Int_t nOrderType = 2;
            fmi2_import_variable_list_t *lstOutputVariables = fmi2_import_get_variable_list(_fmi2Import, nOrderType);

            if (!lstOutputVariables)
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The listOfOutputVariables of size [%i] is not allocated.",
                                 __func__, mc.num_outputs));

            size_t nAtt = fmi2_import_get_variable_list_size(lstOutputVariables);
            if ( _blog)
                cout << " *** addDefaultOutput::nAtt(" << nAtt << ")" << endl;

            fmi2_import_variable_t *fmiAtt;
            TString sname, sdescription, sunit;
            Int_t nvr;
            fmi2_causality_enu_t eCausality;
            URANIE::DataServer::TAttribute *att;

            for(size_t iatt=0; iatt<nAtt; iatt++) {
                cout << "      *** iatt(" << iatt << "/" << nAtt << ") ::";
                fmiAtt = fmi2_import_get_variable(lstOutputVariables, iatt);

                eCausality = fmi2_import_get_causality (fmiAtt);
                cout << " eCausality(" << eCausality << ")";

                if ( eCausality == fmi2_causality_enu_output) {

                    sname = fmi2_import_get_variable_name(fmiAtt);
                    cout << " name(" << sname << ")";

                    sdescription = fmi2_import_get_variable_description(fmiAtt);
                    cout << " sdescription(" << sdescription << ")";

                    nvr = fmi2_import_get_variable_vr (fmiAtt);
                    cout << " nvr(" << nvr << ")";


                    att = new URANIE::DataServer::TAttribute( sname, sdescription);

                    // Search the type aND the unit (for real) of the variable
                    sunit ="Without Unit";
                    // Case of REAL
                    fmi2_import_real_variable_t *fmiAsRealAtt = fmi2_import_get_variable_as_real(fmiAtt);
                    if ( fmiAsRealAtt ) {
                        cout << "[REAL]";
                        att->setDataType(URANIE::DataServer::TAttribute::kReal);

                        fmi2_import_display_unit_t * displayUnit = fmi2_import_get_real_variable_display_unit(fmiAsRealAtt);

                        if ( displayUnit) {
                            cout << " displayUnit(" << displayUnit << ")";
                            sunit =  fmi2_import_get_display_unit_name(displayUnit);
                            att->setUnit(sunit);
                        }
                        cout << " sunit(" << sunit << ")";
                        att->printLog();
                    } // End of REAL

                    // Case of INTEGER
                    if ( fmi2_import_get_variable_as_integer(fmiAtt) ) {
                        cout << "[INTEGER]";
                        sunit = "Without Unit [INTEGER] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kReal);
                    } // End Of INTEGER

                    // Case of [ENUM|BOOLEAN]
                    if ( fmi2_import_get_variable_as_enum(fmiAtt) ||
                         fmi2_import_get_variable_as_boolean(fmiAtt) ) {
                        cout << "[ENUM|BOOLEAN]";
                        sunit = "Without Unit [ENUM|BOOLEAN] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kCategory);
                    } // End of [ENUM|BOOLEAN]

                    // Case of STRING
                    if ( fmi2_import_get_variable_as_string(fmiAtt) ) {
                        cout << "[STRING]";
                        sunit = "Without Unit [STRING] for FMI)";
                        cout << " sunit(" << sunit << ")";
                        att->setDataType(URANIE::DataServer::TAttribute::kString);
                    } // End of STRING
                addAttribute(att, att->GetName());
                } // End of Causality == Output
                // end of Line
                cout << endl;
            } // End of for iatt
            //clean
            fmi2_import_free_variable_list(lstOutputVariables);
            break;
        }   // End of Switch Case fmi_version_2_0_enu:
        } // End of Switch


        nY = (Int_t)(_attributes.size());

        if ( nY == 0 )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TInputFileFMU::%s ERROR:\n The number of Outputs is finally zero (%i).",
                         __func__, nY));



    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }



} // End of Methof addDefaultOutput

// ----------------------------------------------------------------------
//  searchOutput Method
// ----------------------------------------------------------------------
Bool_t TOutputFileFMU::searchOutput(vector<URANIE::DataServer::UEntry*> *out)
  {
    FUNCT_TRACE();
   // Get the working file
    if (_blog) {
        cout << endl <<
                " ************************" << endl <<
                " *** Method TOutputFileFMU::" << __func__  << endl <<
                " ***  PWD[" << gSystem->pwd() << "]" << endl <<
                " ***  _sFileName[" << _sFileName << "]" << endl;
      }

    if (_blog)
        cout << " *** Size of _vecOutputs  [" << _vecOutputs.size()<< "]" << endl <<
                " *** Size of First Vector [" << _vecOutputs[0].size() << "]" << endl <<
                " ***    _nCrtOutputVector [" << _nCrtOutputVector << "]" << endl;
    _nCrtOutputVector = 0;

    // In case the output file does not exist, put the default value
    //Number of attributes
    Int_t natt = (Int_t)(_attributes.size());
    if (_blog)
      cout << " *** natt[" << natt << "]" << endl;

    fmi2_status_t fmi2status;
    fmi1_status_t fmi1status;

    try {
        switch (_sVersion) {
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s INFO:\n The FMI version (%i) of the file \"%s\" is not defined.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_unsupported_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not supported.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_1_enu:
        case fmi_version_2_0_enu:

            TString skey = "", sfilecrt ="";
            for(Int_t iatt=0; iatt<natt; iatt++) {

                URANIE::DataServer::TAttribute* att = _attributes[iatt];
                if (_blog)
                    cout << " *** iatt(" << iatt << "/" << natt << ") name[" << att->GetName() << "]" << endl;

                fmi2_value_reference_t nfield = -1;

                fmi2_real_t rvalue;
                fmi2_integer_t nvalue;
                fmi2_boolean_t bvalue;

                fmi2_string_t svalue;

                TString sBegin, sEnd;
                getBegEndProperties( att, sBegin, sEnd);

                list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();
                for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it) {
                    sfilecrt = (*it)->_sFile;

                    if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0) {
                        skey = (*it)->_sKey;
                        nfield = -2;

                        if (_blog)
                            cout << " ***  Attribute[" << att->GetName() << "]" <<
                                    " sfilecrt[" << sfilecrt.Data() << "]" <<
                                    " :: skey[" << skey << "]" <<
                                    " nfield[" << nfield << "]" << endl;

                        // Field on the current attribute
                        // Get the EventList
                        if ( skey != "__time__") {
                            TEventList *tel = (*it)->_sEventList;
                            if (_blog)
                                tel->Print("all");

                            Int_t ntel = tel->GetN();

                            if ( ntel != 1 )
                                throw URANIE::Exceptions::UErrorExceptions(
                                            __FILE__,
                                            __LINE__,
                                            Form("TInputFileFMU::%s ERROR:\n The attribute \"%s\" in the FMU file \"%s\" has the key \"%s\" several times (%d).",
                                                 __func__, att->GetName(), _sFileName.Data(), skey.Data(), ntel));

                            nfield = tel->GetEntry(0);
                            if (_blog)
                                cout << "  *** nfield[" << nfield << "] by TEventList";
                        }   else {
                            if (_blog)
                                cout << " *** case of the time attribute[" << att->GetName() << "]" << endl;
                        }

                        switch (att->getDataType()) {
                        case URANIE::DataServer::TAttribute::kReal:
                        case URANIE::DataServer::TAttribute::kDefault:

                            rvalue = -1.0 * DLE_MAX;
                            nvalue = - 123456789;

                            if ( _sVersion == fmi_version_2_0_enu ) {
                                fmi2_import_variable_t* fmiAtt = fmi2_import_get_variable_by_name(_fmi2Import, skey.Data());

                                if ( fmi2_import_get_variable_as_real(fmiAtt) )
                                    fmi2status = fmi2_import_get_real(_fmi2Import, &nfield, 1, &rvalue);
                                if ( fmi2_import_get_variable_as_integer(fmiAtt) ) {
                                    fmi2status = fmi2_import_get_integer(_fmi2Import, &nfield, 1, &nvalue);
                                    rvalue = 1.0 * nvalue;
                                }

                            } // end of fmi_version_2_0_enu
                            if ( _sVersion == fmi_version_1_enu ) {
                                fmi1_import_variable_t* fmiAtt = fmi1_import_get_variable_by_name(_fmi1Import, skey.Data());


                                if ( fmi1_import_get_variable_as_real(fmiAtt) )
                                    fmi1status = fmi1_import_get_real(_fmi1Import, &nfield, 1, &rvalue);
                                if ( fmi1_import_get_variable_as_integer(fmiAtt) ) {
                                    fmi1status = fmi1_import_get_integer(_fmi1Import, &nfield, 1, &nvalue);
                                    rvalue = 1.0 * nvalue;
                                }
                            } // end of fmi_version_1_enu
                            if (_blog)
                                cout << " rvalue[" << rvalue << "]";

                            out->at(iatt)->pushEntry(rvalue);

                            break;
                        case URANIE::DataServer::TAttribute::kCategory:

                            if ( _sVersion == fmi_version_2_0_enu ) {
                                fmi2_import_variable_t* fmiAtt = fmi2_import_get_variable_by_name(_fmi2Import, skey.Data());

                                bvalue = fmi2_true;
                                if ( fmi2_import_get_variable_as_boolean(fmiAtt) ) {
                                    fmi2status = fmi2_import_get_boolean(_fmi2Import, &nfield, 1, &bvalue);
                                    out->at(iatt)->pushEntry(bvalue ? Form("kTrue"): Form("kFALSE"));
                                }
                            } // end of fmi_version_2_0_enu

                            break;
                        case URANIE::DataServer::TAttribute::kString:

                            svalue = "__DEFAULT_VALUE__";

                            if ( _sVersion == fmi_version_2_0_enu ) {
                                fmi2status = fmi2_import_get_string(_fmi2Import, &nfield, 1, &svalue);
                            } // end of fmi_version_2_0_enu
                            if ( _sVersion == fmi_version_1_enu ) {
                                fmi1status = fmi1_import_get_string(_fmi1Import, &nfield, 1, &svalue);
                            } // end of fmi_version_1_enu
                            if (_blog)
                                cout << " svalue[" << svalue << "]";

                            out->at(iatt)->pushEntry(Form("%s", svalue));
                            break;

                        case URANIE::DataServer::TAttribute::kVector: {
                            if (_blog)
                                cout << " *** _nCrtOutputVector[" << _nCrtOutputVector << "/" << _nOutputVector << "]" << endl;
                            size_t no = _vecOutputs.size();
                            size_t na = _vecOutputs[0].size();

                            if (_blog) {
                                cout << " *** no[" << no << "] na[" << na << "]" << endl;
                                cout << " *** _svecdel[" << _svecdel << "]" <<
                                        " *** sBegin[" << sBegin << "]" <<
                                        " *** sEnd[" << sEnd << "]" << endl;
                            }

                            vector<double> vecValue;
                            for(size_t i=0; i<no; i++) {
                                vecValue.push_back(_vecOutputs[i][_nCrtOutputVector]);
                            }
                            _nCrtOutputVector++;

                            out->at(iatt)->fillEntry(URANIE::DataServer::UEntry::dumpVectorStr(&vecValue), ",", sBegin, sEnd);

                            break;
                        }
                        default:
                            break;
                        }
                        // Get the Value
                    }
                }
            } // End of for iatt
            break;
        }

    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Free Mermory
    try {
        switch (_sVersion) {
        case fmi_version_1_enu: {
            if (_blog)
                cout << " *** free memory case FMI1" << endl;
            // Terminate the computation
            fmi1status = fmi1_import_terminate_slave(_fmi1Import);

            if(fmi1status != fmi1_status_ok)
                throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI function \"fmi1_import_terminate_slave\" failed.",__func__));
            // Clean
            fmi1_import_free_slave_instance(_fmi1Import);

            break;
        }
        case fmi_version_2_0_enu: {
            if (_blog)
                cout << " *** free memory case FMI2" << endl;
            // Terminate the computation
            fmi2status = fmi2_import_terminate(_fmi2Import);

            if(fmi2status != fmi2_status_ok)
                throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI function \"fmi2_import_terminate\" failed.",__func__));
            // Clean
            fmi2_import_free_instance(_fmi2Import);

            break;
        }
        default:
            break;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    if (_blog)
        cout << endl <<
                " *** End of Method [TOutputFileFMU::" << __func__  << "]" << endl <<
                " ***  PWD[" << gSystem->pwd() << "]" << endl <<
                " ***  _sFileName[" << _sFileName << "]" << endl <<
                " ************************" << endl;

    return kTRUE;
  } // End of Method searchOutput

// ----------------------------------------------------------------------
//  printLog Method
// ----------------------------------------------------------------------
void TOutputFileFMU::printLog(Option_t *option)
{
    FUNCT_TRACE();
    cout << " *** TOutputFileFMU::printLog[" << option << "]" << endl;
    cout << " *** _sFileName[" << _sFileName << "]" << endl;
    cout << " *** _nOutputVector[" << _nOutputVector << "] /" <<
            " _nCrtOutputVector[" << _nCrtOutputVector << "] ---";
    cout << " _lstOfVector ";
    if ( _lstOfVector ) {
        Int_t nY = _lstOfVector->GetSize();
        cout << "size[" << nY << "]" << endl;
        for( Int_t i=0; i<nY; i++)
            cout << "      *** i(" << i << "/" << nY << ") [" << _lstOfVector->At(i) << "]" << endl;
    } else {
        cout << "[0]" << endl;
    }

    Int_t natt = (Int_t)(_attributes.size());
    for (Int_t indx = 0; indx < natt; indx++)
    {
        cout << "index[" << indx + 1 << "/" << natt << "]" << endl;
        _attributes[indx]->printLog();
    }
}

} // End of namespace Launcher
} // End of namespace URANIE
// End of File
