//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileKey.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:52:50 CEST 2006
 
 \brief Implementation of the class URANIE::Launcher::TOutputFileKey
 
 \par Modifications
\li Jean-Baptiste Blanchard 2016/04/08
 - Add the searchOutput with a map to read strings/vectors
 \li Cedric Forgues (2010-12-17 14:34:08)
 - getLog() erased from searchOutput() method to catch exception.
 Corrects the comment F. Gaudier 2010-08-05 12:46:04, third item.
 \li F. Gaudier (2010-08-06 11:26:56)
 - change the deprecated setKey method by setFileKey method
 \li F. Gaudier 2010-08-05 12:46:04
 - In the case of no temporal output (only one value for each attributes), put the value in the good place.
 So allows you can specify output attributes not in the same order that they appear in the output file.
 - Put the default value in the vector of no temporal output
 - When a key is not found in the output file, it can no throw an error exceptions if the log is ON
 \todo (FGA::2010-08-06 09:24:59)  Add a new flag at these classe to pass the exception when not found a key and continue tke launching
 \li F. Gaudier 2010-04-14 10:21:52
 For the addAttribute method, test the cases
 1. if the keyList is not empty,
 1.1 test if it contains one object
 1.1.1 if the filename is the same than the current filename is good
 1.1.2 else throw an error exception
 1.2 if it contains more than one element, throw an error exception
 2. if the keyList is empty, applique the setFileKey method
 \li F. Gaudier 2010-04-12 14:00:55
 Increase the index \b indcrt in the searchOutput method
 \li F. Gaudier 2010/02/23
 throw an exceptions when the key of one attribute is not found in the output file
 set more readable log
 
 */
#include <stdlib.h>
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TMatrixT.h"

// Uranie
#include "TOutputFileKey.h"
#include "UExceptions.h"

// Club
#ifdef OLDCLUB
#include "club/TamponTexte.h"
#include "club/IterateurChamp.h"
#include "club/ClubErreurs.h"
#else
#include "TamponTexte.h"
#endif

ClassImp (URANIE::Launcher::TOutputFileKey)

namespace URANIE
{
namespace Launcher
{

TOutputFileKey::TOutputFileKey(TString str) :
    TOutputFile(str)
{
    setSeparatorCharacter(" =\t\r\n");
    if (_blog)
	cout << "constructing TOutputFileKey file[" << str
	     << "] SeparatorCharacter[" << getSeparatorCharacter() << "]"
	     << endl;
    _sstrbeg = _sstrend = "";
    _svecbeg = _svecend = ""; _svecdel = "";
}
TOutputFileKey::~TOutputFileKey()
{
    if (_blog)
	cout << "destroy TOutputFileKey" << endl;
}

/////////////////////
//// Add attribut to the Output File
/////////////////////
void TOutputFileKey::addAttribute(TString str, TString skey)
{
    URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
    att->delShare();
    addAttribute(att, skey);
}

void TOutputFileKey::addAttribute(URANIE::DataServer::TAttribute* att,
				  TString skey)
{
    if (_blog)
    {
        cout << "********************************" << endl;
        cout << " ***TOutputFileKey::addAttribute attribute[" << att->GetName()
	     << "] TOutputFileKey FileName[" << _sFileName << "]" << endl;
    }
    try
    {
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	    att->getKeyList();

        if (!KeyList.empty())
	{
            Int_t nOutputfile = KeyList.size();
            if (nOutputfile > 1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileKey::addAttribute\n The Output attribute [%s] have %d Output Files: only one is autorised",
							       att->GetName(), nOutputfile));

            if (_blog)
		cout << " ** KeyList :: Size[" << nOutputfile << "]" << endl;
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		KeyList.begin();
            TString s_key = (*it)->_sKey;
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());
            if (_blog)
		cout << " ** KeyList :: s_key[" << s_key << "] sfilename["
		     << sfilename << "]" << endl;
            if (_sFileName != sfilename)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileKey::addAttribute\n The Output attribute [%s] have already an Output File [%s] which is diffrent than the current [%s]",
							       att->GetName(), sfilename.Data(),
							       _sFileName.Data()));

	}
        else
	    att->setFileKey(_sFileName, skey);

        TCodeFile::addAttribute(att);
        if (_blog)
	    att->printLog();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    if (_blog)
    {
        cout << " *** End Of TOutputFileKey::addAttribute attribute["
	     << att->GetName() << "] TOutputFileKey FileName[" << _sFileName
	     << "]" << endl;
        cout << "********************************" << endl;
    }
}

/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileKey::searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileKey::searchOutput indcrt["
	     << indcrt << "]" << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = (Int_t)(_attributes.size());
    if (_blog)
    {
        cout << " *** natt[" << natt << "]" << endl;
    }
    // Get in local the index value to store the value when no temporal output
    Int_t nlocindcrt = indcrt;

    TMatrixT<double> *matOut = new TMatrixT<double>(1, natt);

    // Default Value
    Double_t valdef = 1.23456789;

    Double_t * tempVal = matOut->GetMatrixArray();
    for (Int_t i = 0; i < natt; i++)
    {
        _attributes[i]->getDefaultValue(valdef);
        val[nlocindcrt + i] = valdef;
        tempVal[i] = valdef;
        indcrt++;
    }
    if (_blog)
    {
        cout << " *** Default Value on vector" << endl;
        for (Int_t iout = 0; iout < indcrt; iout++)
	    cout << "   *** output index [" << iout << "] Val[" << val[iout]
		 << "]" << endl;
        cout << " *** End Of Default Value on vector";
        cout << endl << " *** Default Value For the Matrix";
        matOut->Print();
        cout << " *** End Of Default Value For the Matrix" << endl;
    }

    try
    {
        TamponTexte fnomFichier(_sFileName, 0);
        fnomFichier.initSeparateurs(_cSeparators);
        fnomFichier.initCarCommentaires(_cComments);
#ifndef OLDCLUB
	fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
        // Elimination des lignes de commentaires
        fnomFichier.elimineCommentaires();
        // Elimination des lignes vides
        fnomFichier.elimineBlancs();
        // Test si le fichier ainsi obtenu est vide
        if (fnomFichier.estVide())
	{
            throw "URANIE::Launcher::TCode::substitue the file is empty";
	}

        // Set the number of lines in this file
        Int_t nlines = fnomFichier.nombreLignes();

        if (_blog)
	    cout << " ** Number of lines[" << fnomFichier.nombreLignes() << "]"
		 << endl;

        // Cout the number of presence of each output attributes
        //   and set the number maximum
        // loop on the number of attributes
        Int_t nmax = 0;
        if (_blog)
	{
            cout << endl << " ************************" << endl;
            cout << " *** Count the number of presence of each attributes"
		 << endl;
	}
        for (Int_t indx = 0; indx < natt; indx++)
	{
            if (_blog)
		cout << " *** attribute " << indx + 1 << "/" << natt
		     << " : name[" << _attributes[indx]->GetName() << "]";
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
		_attributes[indx]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		KeyList.begin();
            TString skey = (*it)->_sKey;
            Int_t ncpt = 0;
            if (_blog)
		cout << " -- skey [" << skey << "]";
            for (Int_t iline = 1; iline <= nlines; iline++)
	    {
                const char* champ = fnomFichier.champLigne(1, iline);
                TString strchamp(champ);
#ifndef OLDCLUB
                delete[] champ;
#endif
                if (strchamp.CompareTo(skey.Data(), TString::kExact) == 0)
		    ncpt++;
	    }
            if (_blog)
		cout << " ** number of presence [" << ncpt << "]" << endl;
            if (ncpt > nmax)
		nmax = ncpt;
	    //           try {
	    //             if ( ncpt == 0 )
	    //             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
	    //                                              Form("TOutputFileKey::searchOutput method\nThe attribute [%s] with key [%s] was not found in the output file [%s]", _attributes[indx]->GetName(), skey.Data(), _sFileName.Data()));
	    //           } catch (URANIE::Exceptions::UErrorExceptions& ue) {
	    //             ue.printMessage();
	    //             throw ue;
	    //           }

	}
        if (_blog)
	{
            cout
		<< " *** End Of Count the number of presence of each attributes with Maximun["
		<< nmax << "]" << endl;
            cout << " ************************" << endl;
	}

        // Cas de vector
	delete matOut; matOut = NULL;
        matOut = new TMatrixT<double>(nmax, natt);
        Double_t * theVal = matOut->GetMatrixArray();
        Int_t *npresence = new Int_t[natt];
        for (Int_t i = 0; i < natt; i++)
	{
            npresence[i] = 0;
            // \todo Put all the default values in this vector
            for (Int_t j = 0; j < nmax; j++)
		theVal[i * nmax + j] = 1.0 * i;
	}

        if (_blog)
	    cout << " *** Search the output values" << endl;
        // Loop on each of the lines of the file
        for (Int_t iline = 1; iline <= nlines; iline++)
	{
            if (_blog)
		cout << " ************************" << endl << " *** line [ "
		     << iline << "/" << nlines << "]" << endl;
            const char* champfi = fnomFichier.champLigne(1, iline);
            TString strchampfi(champfi);
#ifndef OLDCLUB
            delete[] champfi;
#endif
            if (_blog)
		cout << " *** Current field[" << strchampfi << "]" << endl;

            // Loop on the file attributes
            for (Int_t indx = 0; indx < natt; indx++)
	    {
                list<URANIE::DataServer::TAttributeFileKey *> KeyList =
		    _attributes[indx]->getKeyList();
                list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		    KeyList.begin();
                TString skey = (*it)->_sKey;
                if (_blog)
		    cout << " *** ------------- Field[" << indx << "] value["
			 << strchampfi << "] skey[" << skey << "]" << endl;
                if (strchampfi.CompareTo(skey.Data(), TString::kExact) == 0)
		{
                    // First time that appears the current attribute

                    const char* champ = fnomFichier.champLigne(2, iline);
                    TString strchamp(champ);
#ifndef OLDCLUB
                    delete[] champ;
#endif
                    if (npresence[indx] == 0)
			val[nlocindcrt + indx] = atof(strchamp);
                    Int_t nn = npresence[indx];
                    theVal[nn * natt + indx] = atof(strchamp);
                    npresence[indx]++;
                    if (_blog)
			cout << " *** ++++++++++++++++ Key attribute["
			     << strchamp << "] val["
			     << theVal[nn * natt + indx] << "] presence["
			     << npresence[indx] << "]" << endl;
                    //
                    break;
		}
	    }
	}
        if (_blog)
	    cout << "*** End Of TOutputFileKey::searchOutput indcrt[" << indcrt
		 << "]" << endl << "************************" << endl;

        // Pour le retour
        if (lstOut != NULL)
	{
            if (_blog)
	    {
                cout << " *** Returned Value on vector" << endl;
                for (Int_t iout = 0; iout < indcrt; iout++)
		    cout << "   *** output index [" << iout << "] Val["
			 << val[iout] << "]" << endl;
                cout << " *** End Of Returned Value on vector";
                cout << endl << " *** Returned Value For the Matrix";
                matOut->Print();
                cout << " *** End Of Returned Value For the Matrix" << endl;
	    }
            lstOut->Add(matOut);
	}
	else
	{
	    delete matOut;
	    matOut=NULL;
	}
        // Clean
        delete[] npresence;
        npresence = NULL;
    }
#ifdef OLDCLUB
    catch (ClubErreurs ce)
    {
        ce.correction();
        if( lstOut != NULL )
	{
            if ( _blog )
		matOut->Print();
            lstOut->Add(matOut);
	}
	else
	{	    
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	}
        return kFALSE;
        //throw ("TOutputFileKeyRow::searchOutput Error");
    }
#else
    catch (TamponTexteError&)
    {
        if (lstOut != NULL)
	{
            if (_blog)
		matOut->Print();
            lstOut->Add(matOut);
	}
	else
	{	    
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	}
        return kFALSE;
    } 
#endif  
    catch (URANIE::Exceptions::UErrorExceptions&)
    {
        if (lstOut != NULL)
	{
            if (_blog)
		matOut->Print();
            lstOut->Add(matOut);
	}
	else
	{	    
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	}
        return kFALSE;
    }

    return kTRUE;

}

/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileKey::searchOutput(vector<URANIE::DataServer::UEntry*> *out)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileKey::searchOutput "<<endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = (Int_t)(_attributes.size());
    if (_blog)
	cout << " *** natt[" << natt << "]" << endl;

    // Create a default map containing just the default of Attribute of 1.23456789
    // // Default Value
    for (Int_t i = 0; i < natt; i++)
    {
        TString def=_attributes[i]->getDefault();
        out->at(i)->fillEntry( ((def=="" && out->at(i)->tag==URANIE::DataServer::UEntry::Double ) ? "1.23456789" : def) );
    }

    if (_blog)
	cout << endl << " *** Default Value  "<< "" << " *** End Of Default Value" << endl;

    try
    {
        TamponTexte fnomFichier(_sFileName, 0);
        fnomFichier.initSeparateurs(_cSeparators);
        fnomFichier.initCarCommentaires(_cComments);
#ifndef OLDCLUB
	fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
        // Elimination des lignes de commentaires
        fnomFichier.elimineCommentaires();
        // Elimination des lignes vides
        fnomFichier.elimineBlancs();
        // Test si le fichier ainsi obtenu est vide
        if (fnomFichier.estVide())
	{
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("TOutputFileKey::searchOutput\n The Output file [%s] is empty. Default will be put for all its attributes",
                                                              this->GetName()));
        }


        // Set the number of lines in this file
        Int_t nlines = fnomFichier.nombreLignes();

        if (_blog)
	    cout << " ** Number of lines[" << fnomFichier.nombreLignes() << "]" << endl;

        if (_blog)
	    cout << " *** Search the output values" << endl;

        // Loop on each of the lines of the file
	TString beg, end;
        for (Int_t iline = 1; iline <= nlines; iline++)
	{
            if ( fnomFichier.nombreChampsLigne(iline) < 2 )
                continue;
            
            const char* champ = fnomFichier.champLigne(1, iline);
	    if (_blog)
		cout << " ************************" << endl << " *** line [ " << iline << "/" << nlines << "]"
		     << " champ [" << champ << "]" << endl;

            TString foundkey(champ);

	    const char* valchamp = fnomFichier.champLigne(2, iline);
	    if( strcmp(valchamp,_svecbeg.Data())==0 || strcmp(valchamp,_sstrbeg.Data())==0 )
		if( fnomFichier.nombreChampsLigne(iline) > 2 )
		{
		    delete[] valchamp;
		    valchamp = fnomFichier.champLigne(3, iline);
		}
	    
	    TString val = TString(valchamp).ReplaceAll(";","");

	    if (_blog)
		cout << "*** End Of TOutputFileKey::searchOutput ************************" << endl;
	    
	    for (Int_t indx = 0; indx < natt; indx++)
	    {
		list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[indx]->getKeyList();
		list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin();
		getBegEndProperties( _attributes[indx], beg, end);
		TString skey = (*it)->_sKey;
		
		if( foundkey != skey )
		    continue;

		if( _svecdel == "" )
		    out->at(indx)->pushEntry(val);
		else
		    out->at(indx)->fillEntry(val, _svecdel, beg, end);
	    }
#ifndef OLDCLUB
	    delete[] champ;
	    delete[] valchamp;
#endif
	}
    }
#ifndef OLDCLUB  
    catch (TamponTexteError&)
    {
	if (_blog)
	    cout << " *** End Of catch ClubErreurs" << endl;

        return kFALSE;

    }
#else
    catch (ClubErreurs ce)
    {
        if (_blog)
            cout << " *** End Of catch ClubErreurs" << endl;

        return kFALSE;

    }
#endif
    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
	ue.printMessage();
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
        throw(ue);
    }

    return kTRUE;

}

void TOutputFileKey::printLog(Option_t *option)
{
    cout << "TOutputFileKey::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    Int_t natt = (Int_t)(_attributes.size());
    for (Int_t indx = 0; indx < natt; indx++)
    {
        cout << "index[" << indx + 1 << "/" << natt << "]" << endl;
        _attributes[indx]->printLog();
    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

