//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileXML.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-09
 
 \brief Implementation of the class URANIE::Launcher::TOutputFileXML
 
 \par Modifications
\li Jean-Baptiste Blanchard 2016/04/08
 - Add the searchOutput with a map to read strings/vectors
 \li Simon Pomarede (2012-11-09)
 - Creation of the class
 */
#include <stdlib.h>
#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TMatrixT.h"

// Uranie
#include "TOutputFileXML.h"
#include "UExceptions.h"

// LibXML2
#include <libxml/xpathInternals.h>

ClassImp (URANIE::Launcher::TOutputFileXML)

namespace URANIE
{
namespace Launcher
{

  TOutputFileXML::TOutputFileXML(TString str) :
    TOutputFile(str)
  {
      _sstrbeg = _sstrend = "";
      _svecbeg = _svecend = ""; _svecdel = " ";
  }

  TOutputFileXML::~TOutputFileXML()
  {
    if (_blog)
      cout << "destroy TOutputFileXML" << endl;
  }

  /////////////////////
  //// Add attribute to the Output File
  /////////////////////
  void TOutputFileXML::addAttribute(TString str, TString xmlPath,
				    URANIE::DataServer::TAttributeFileKey::EFileType FileType)
  {
    URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
    att->delShare();
    addAttribute(att, xmlPath, FileType);
  }

  void TOutputFileXML::addAttribute(URANIE::DataServer::TAttribute* att,
				    TString xmlPath,
				    URANIE::DataServer::TAttributeFileKey::EFileType FileType)
  {
    if (_blog)
      {
        cout << "********************************" << endl;
        cout << " ***TOutputFileXML::addAttribute attribute[" << att->GetName()
	     << "] TOutputFileXML FileName[" << _sFileName << "]" << endl;
      }

    try
      {
        // Get the list of files the attribute is linked: an attribute is associated with at most one file
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	  att->getKeyList();
        if (!KeyList.empty())
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       Form(
							    "TOutputFileKey::addAttribute\n The Output attribute [%s] have an Output File yet: only one is autorised",
							    att->GetName()));
	  }

        // The XML output can only deal with kXMLAttribute and kXMLField attributes
        if (!(FileType == URANIE::DataServer::TAttributeFileKey::kXMLAttribute
	      || FileType == URANIE::DataServer::TAttributeFileKey::kXMLField))
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
						       __LINE__,
						       "TOutputFileKey::addAttribute\n Unauthorized XML type; only kXMLAttribute or kXMLField available");
	  }

        // Addition of the file
        att->setFileKey(_sFileName, xmlPath, "%e", FileType);
        TCodeFile::addAttribute(att);

        if (_blog)
	  att->printLog();
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw(ue);
      }

    if (_blog)
      {
        cout << " *** End Of TOutputFileXML::addAttribute attribute["
	     << att->GetName() << "] TOutputFileXML FileName[" << _sFileName
	     << "]" << endl;
        cout << "********************************" << endl;
      }
  }

  /////////////////////
  //// searchOutput
  /////////////////////
  Bool_t TOutputFileXML::searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut)
  {
    // Get the working file
    if (_blog)
      {
        cout << endl
	     << "************************\n*** TOutputFileXML::searchOutput indcrt["
	     << indcrt << "]" << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
      }

    // In case the output file does not exist, put the default value
    //Number of attributes
    Int_t natt = (Int_t)(_attributes.size());
    if (_blog)
      cout << " *** natt[" << natt << "]" << endl;
    TMatrixT<double> *matOut = new TMatrixT<double>(1, natt);

    // Default Value
    Double_t valdef = 1.23456789;
    Double_t *tempVal = new Double_t[natt];
    // Get in local the index value to store the value when no temporal output
    Int_t nlocindcrt = indcrt;

    // Loop on the file attributes
    for (Int_t i = 0; i < natt; i++)
      {
        _attributes[i]->getDefaultValue(valdef);
        val[nlocindcrt + i] = valdef;
        tempVal[i] = valdef;
        indcrt++;
      }
    matOut->Use(1, natt, tempVal);
    if (_blog)
      {
        cout << endl << " *** Default Value";
        matOut->Print();
        cout << " *** End Of Default Value" << endl;
      }

    try
      {
        // Cas de vector
        delete matOut;
        matOut = NULL;
        matOut = new TMatrixT<double>(1, natt);
        Double_t * theVal = new Double_t[natt];
        for (Int_t i = 0; i < natt; i++)
	  {
            theVal[i] = 1.0 * i;
	  }

        if (_blog)
	  cout << " *** Search the output values" << endl;

        // Case of all the matrix
        int ncrt = 0;
        for (Int_t iout = 0; iout < natt; iout++)
	  {
            // Get the XML path of the current attribute
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	      _attributes[iout]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
	      KeyList.begin();
            TString xmlPath = (*it)->_sKey;

            // Get the value from the XML file
            xmlInitParser();
            xmlDocPtr doc = xmlParseFile(_sFileName.Data()); // parsing of the file
            if (doc == NULL)
	      {
                // Clean
                delete[] theVal;
                theVal = NULL;
                throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
								"TOutputFileKey::searchOutput\n Impossible to parse the file [%s]",
								_sFileName.Data()));
	      }
            xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
            xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression(
								(const xmlChar *) xmlPath.Data(), xpathCtx); // evaluation of the XML path
            if (xpathObj == NULL || xpathObj->nodesetval == NULL
		|| xpathObj->nodesetval->nodeTab == NULL)
	      {
                // Clean
                delete[] theVal;
                theVal = NULL;
                throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
								"TOutputFileKey::searchOutput\n The node [%s] cannot be foud in Output File [%s]",
								xmlPath.Data(), _sFileName.Data()));
	      }
            xmlNodePtr cur = xpathObj->nodesetval->nodeTab[0];
            xmlXPathFreeObject(xpathObj);
            xmlChar * charValue = xmlNodeGetContent(cur);
            double value = atof((char *) charValue);

            // Store the value into the data structures
            val[nlocindcrt + iout] = value;
            theVal[iout] = value;
            ncrt += natt;

            // Clean
            xmlFree(charValue);
            xmlXPathFreeContext(xpathCtx);
            xmlFreeDoc(doc);
	  }
        matOut->Use(1, natt, theVal);

        // Return values
        if (lstOut != NULL)
	  {
            if (_blog)
	      {
                cout << " *** Returned Value on vector" << endl;
                for (Int_t iout = 0; iout < indcrt; iout++)
		  cout << "   *** output index [" << iout << "] Val["
		       << val[iout] << "]" << endl;
                cout << " *** End Of Returned Value on vector";
                cout << endl << " *** Returned Value For the Matrix";
                matOut->Print();
                cout << " *** End Of Returned Value For the Matrix" << endl;
	      }
            lstOut->Add(matOut);
	  }
	else
	  {
	    if(matOut!=NULL)
	      delete matOut;
	    matOut=NULL;
	  }
        // Clean
        delete[] theVal;
        theVal = NULL;
        delete[] tempVal;
        tempVal = NULL;
      }

    catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        if (lstOut != NULL)
	  {
            lstOut->Add(matOut);
	  }
	else
	  {
	    if(matOut!=NULL)
	      delete matOut;
	    matOut=NULL;
	  }
        ue.printMessage();
        // Clean
        if (tempVal != NULL)
	  {
            delete[] tempVal;
            tempVal = NULL;
	  }
	//           throw(ue);
        return kFALSE;
      }

    return kTRUE;
  }

  /////////////////////
  //// searchOutput
  /////////////////////
Bool_t TOutputFileXML::searchOutput(vector<URANIE::DataServer::UEntry*> *out)
  {
    // Get the working file
    if (_blog)
      {
        cout << endl
	     << "************************\n*** TOutputFileXML::searchOutput " << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
      }

    // In case the output file does not exist, put the default value
    //Number of attributes
    Int_t natt = (Int_t)(_attributes.size());
    if (_blog)
      cout << " *** natt[" << natt << "]" << endl;

    // Create a default map containing just the default of Attribute of 1.23456789
    // // Default Value
    for (Int_t i = 0; i < natt; i++)
    {
        TString def=_attributes[i]->getDefault();
        out->at(i)->fillEntry( ((def=="" && out->at(i)->tag==URANIE::DataServer::UEntry::Double ) ? "1.23456789" : def) );
    }


    if (_blog)
      cout << endl << " *** Default Value" << "" << " *** End Of Default Value" << endl;


    try
      {
        if (_blog)
	  cout << " *** Search the output values" << endl;

	TString beg, end;
        for (Int_t iout = 0; iout < natt; iout++)
	  {
            // Get the XML path of the current attribute
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	      _attributes[iout]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
	      KeyList.begin();	    
            TString xmlPath = (*it)->_sKey;
	    getBegEndProperties( _attributes[iout], beg, end);
	    
            // Get the value from the XML file
            xmlInitParser();
            xmlDocPtr doc = xmlParseFile(_sFileName.Data()); // parsing of the file
            if (doc == NULL)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							 __LINE__,
							 Form(
							      "TOutputFileKey::searchOutput\n Impossible to parse the file [%s]",
							      _sFileName.Data()));

	    xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
            xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression( (const xmlChar *) xmlPath.Data(), xpathCtx); // evaluation of the XML path
            if (xpathObj == NULL || xpathObj->nodesetval == NULL || xpathObj->nodesetval->nodeTab == NULL)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							 __LINE__,
							 Form(
							      "TOutputFileKey::searchOutput\n The node [%s] cannot be foud in Output File [%s]",
							      xmlPath.Data(), _sFileName.Data()));

            xmlNodePtr cur = xpathObj->nodesetval->nodeTab[0];
            xmlXPathFreeObject(xpathObj);
            xmlChar * charValue = xmlNodeGetContent(cur);
            string value = string((char *) charValue);

	    //cout<<"for entry "<<out->at(iout)->name<<"  pushing  "<<value<<endl;
	    out->at(iout)->fillEntry(value, _svecdel, beg, end);

            // Clean
            xmlFree(charValue);
            xmlXPathFreeContext(xpathCtx);
            xmlFreeDoc(doc);
	    
	  }
	
	if (_blog)
	  cout << " *** End Of Returned Value on vector";
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        return kFALSE;
      }

    return kTRUE;
  }
  

  /////////////////////
  //// printLog
  /////////////////////
  void TOutputFileXML::printLog(Option_t *option)
  {
    cout << "TOutputFileXML::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    Int_t natt = (Int_t)(_attributes.size());
    for (Int_t indx = 0; indx < natt; indx++)
      {
        cout << "index[" << indx + 1 << "/" << natt << "]" << endl;
        _attributes[indx]->printLog();
      }
  }

} // Fin du namespace Launcher
} // Fin du namespace URANIE

