"""
Example of flowrate code with OATMinMax with key
"""
import numpy as np
from rootlogon import ROOT, DataServer, Launcher


def plan_oat_min_max(mytds):
    """Generate the OAT min max doe"""

    # Create the data tuple
    mytds.createTuple()

    natt = mytds.getNAttributes()
    # The number of attributes is incremented by one for the iterator
    number_n = natt+1

    # Allocate the data vector to fill the tuple
    value = np.zeros([number_n])

    # Init the iterator. It is always the first value.
    value[0] = 0
    # Init the data vector with the nominal / mean values of attributes
    for i in range(natt):
        value[1+i] = mytds.getAttribute(i).getMean()

    # Fill the nominal/mean values: nominal pattern
    mytds.getTuple().Fill(value)

    # Loop on each attributes
    for i in range(natt):
        # Case of min values
        value[1+i] = mytds.getAttribute(i).getLowerBound()
        value[0] += 1.0
        # Fill min value for xi attribute and nominal values
        mytds.getTuple().Fill(value)
        # Case of the max value
        value[1+i] = mytds.getAttribute(i).getUpperBound()
        value[0] += 1.0
        # Fill max value for xi attribute and nominal values
        mytds.getTuple().Fill(value)
        # Reset the nominal/mean value
        value[1+i] = mytds.getAttribute(i).getMean()


# Define the DataServer
tds = DataServer.TDataServer("tdsOATMinMaxFlowrate", "OATMinMax for Flowrate")

# Add the study attributes ( min, max and nominal values)
tds.addAttribute(DataServer.TAttribute("rw", 0.05, 0.15))
tds.getAttribute("rw").setMean(0.10)

tds.addAttribute(DataServer.TAttribute("r", 100.0, 50000.0))
tds.getAttribute("r").setMean(25050.)

tds.addAttribute(DataServer.TAttribute("tu", 63070.0, 115600.0))
tds.getAttribute("tu").setMean(89335.)

tds.addAttribute(DataServer.TAttribute("tl", 63.1, 116.0))
tds.getAttribute("tl").setMean(89.55)

tds.addAttribute(DataServer.TAttribute("hu", 990.0, 1110.0))
tds.getAttribute("hu").setMean(1050.)

tds.addAttribute(DataServer.TAttribute("hl", 700.0, 820.0))
tds.getAttribute("hl").setMean(760.)

tds.addAttribute(DataServer.TAttribute("l", 1120.0, 1680.0))
tds.getAttribute("l").setMean(1400.)

tds.addAttribute(DataServer.TAttribute("kw", 9855.0, 12045.0))
tds.getAttribute("kw").setMean(10950.)

# The reference input file
sIn = ROOT.gSystem.pwd() + "/"
sIn += "flowrate_input_with_keys.in"

# Set the reference input file and the key for each input attributes
tds.getAttribute("rw").setFileKey(sIn, "Rw")
tds.getAttribute("r").setFileKey(sIn, "R")
tds.getAttribute("tu").setFileKey(sIn, "Tu")
tds.getAttribute("tl").setFileKey(sIn, "Tl")
tds.getAttribute("hu").setFileKey(sIn, "Hu")
tds.getAttribute("hl").setFileKey(sIn, "Hl")
tds.getAttribute("l").setFileKey(sIn, "L")
tds.getAttribute("kw").setFileKey(sIn, "Kw")

# Construit le plan OAT par avec les bornes (Min, Max) and the nominal values
plan_oat_min_max(tds)

# Scan the data ( 17 = 1 + 2*8 patterns)
tds.Scan("*", "", "colsize = 5 col = 3:4:5:6:5:4:3:4")

# Save the Design of experiments in an ASCII file
tds.exportData("_flowrate_sampler_oat_.dat")

# The output file of the code
fout = Launcher.TOutputFileRow("_output_flowrate_withRow_.dat")
# The attribute in the output file
fout.addAttribute(DataServer.TAttribute("yhat"))

# Instanciation de mon code
mycode = Launcher.TCode(tds, "flowrate -s -k")
# mycode.addInputFile(Launcher.TInputFile(sIn))
# on ajoute le fichier de sortie du code
mycode.addOutputFile(fout)

# Lancement du code
laun = Launcher.TLauncher(tds, mycode)
laun.setSave()
laun.setClean()
# laun.setWorkingDirectory(ROOT.gSystem.pwd() + "/tmpLaunUranie/flowrate")
laun.setVarDraw("yhat:rw", "", "")

laun.run()

tds.exportData("_flowrate_sampler_launcher_oat_.dat")

# Visualisation
Canvas = ROOT.TCanvas("c1", "Graph launchCodeFlowrateOATMinMax",
                      5, 64, 1270, 667)
pad = ROOT.TPad("pad", "pad", 0, 0.03, 1, 1)
pad.Draw()
ROOT.gStyle.SetPalette(1)
pad.Divide(2, 2)
pad.cd(1)
tds.draw("yhat:rw")

pad.cd(3)
tds.draw("yhat")
pad.cd(2)
tds.draw("yhat:rw", "", "colz")

pad.cd(4)
# The parallel plot
tds.draw("rw:r:tu:tl:hu:hl:l:kw:yhat", "", "para")
para = ROOT.gPad.GetListOfPrimitives().FindObject("ParaCoord")

# The output attribute
axis = para.GetVarList().FindObject("yhat")
axis.AddRange(ROOT.TParallelCoordRange(axis, 15.0, 20.0))
para.AddSelection("blue")
para.GetCurrentSelection().SetLineColor(ROOT.kBlue)

axis.AddRange(ROOT.TParallelCoordRange(axis, 155.0, 160.0))
para.AddSelection("blue")
para.GetCurrentSelection().SetLineColor(ROOT.kBlue)
