//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TABC
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TABC.cxx
 \author Rudy Chocat
 \date   2020/07/17
 
 \brief Interface of class URANIE::ABC::TABC 
 
 */

// Root
#include "TH1.h"
#include "TCanvas.h"
#include "TLeaf.h"

// Uranie
#include "TABC.h"
#include "TAttribute.h"

ClassImp (URANIE::Calibration::TABC);
 
namespace URANIE
{
namespace Calibration
{
    
// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------
    

TABC::TABC(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,  int ns, Option_t *option):
    TCalibration(tds, run, ns, option), _tdsPosterior(NULL), _threshold(12345678.9)
{
    if (_blog)
        cout << "constructor TABC by TDataServer[" << tds->GetName() << "]"
             << endl;

    parseOption(option);

    checktdsParContent();

    TABC::initABC();
}
    
TABC::TABC(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option) , _tdsPosterior(NULL), _threshold(12345678.9)
{
    parseOption(option);

    
    checktdsParContent();

    TABC::initABC();
}

TABC::TABC(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _tdsPosterior(NULL), _threshold(12345678.9)
{
    parseOption(option);

    
    checktdsParContent();
    TABC::initABC();
}

TABC::TABC(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TCalibration(tds, fcode, ns, option), _tdsPosterior(NULL), _threshold(12345678.9)
{
    TString opt = option;
    opt.ToLower();
    
    checktdsParContent();
    TABC::initABC();
}

void TABC::checktdsParContent()
{

    // Dans le cas d'un plan SRS
    setMethodName("ABC");

    // Get the number of attributes
    _nPar=_tdsPar->getNAttributes();
    TIter next(_listOfParameters);
    URANIE::DataServer::TAttribute *att=NULL;
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    try
    {
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
            if(stoatt==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TABC::constructor\n\
 The provided dataserver must contains stochastic attributes. Attribute [%s] is not one of them", att->GetName()));

            _vatt.push_back( stoatt );
            if (_blog)
                cout << " ** Storing " << att->GetName() << "] in our attribute vector"<<endl;
        }
        
        if( _vatt.size()==0 )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TABC: construction with tds not containing any TStochasticAttribute");

        // number of attributes different from number of stochastic attributes
	if( int(_vatt.size()) != _nPar )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TABC: construction with _nPar [%d] different from _vatt content [%lu]", _nPar, _vatt.size()));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    
}
// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TABC::~TABC()
{
    if (_blog)
        cout << "************ Destructor URANIE::Calibration::TABC" << endl;

    if (_blog)
        cout << "************ End of Destructor URANIE::Calibration::TABC"<< endl;

    if( _tdsPosterior!=NULL)
        delete _tdsPosterior;
    _tdsPosterior=NULL;

    if(_rand!=NULL)
        delete _rand;
    _rand=NULL;
}
        
// -----------------------------------------------------------------------
//  Initialisation
// -----------------------------------------------------------------------
    
void TABC::initABC()
{
    
    _rand=NULL;
    setMethodName("ABC");
    _Epsilon= 0.01;
    _kEval = 0;
    _kPosterior = 0;
    _tdsPosterior=NULL;

    
}

void TABC::setUpABC()
{
    _rand = new TRandom3(); 
    if(_nSeed != 0)
        _rand->SetSeed(_nSeed);
        
    _tdsPosterior = new URANIE::DataServer::TDataServer( Form("posteriorTds_%s", _tdsPar->GetName()), Form("The posterior tds of %s", _tdsPar->GetTitle()));
    for (Int_t i=0;i<_nPar+1;i++)
    {
        _tdsPosterior->addAttribute(_tdsPar->getAttribute(i));
    }
    _tdsPosterior->createTuple();
    _tdsPar->createTuple();
    
}

// ---------------- Compute discrepancy 
void TABC::computeABC(int nEval)
{
    _tdsPar->createTuple();
    vector<double> dBase(_nPar + 2);
    double res=0;
    // Should be factorised to be created and deleted once
    // Every time you're clling this method if put the same seed so it is dangerous

    for(int k = 0; k <nEval; k++)
    {    
        res=0.0;
        dBase[0]=double(k);
        _values.clear();
        for (Int_t iatt=0;iatt<_nPar; iatt++)
        {
            _values.push_back(_vatt.at(iatt)->getInvCDF(_rand->Rndm()));
            if(_blog)
                cout<<"** in TABC::computeABC, Theta " << iatt << " ---" << _values[iatt] <<endl;
            dBase[1+iatt] = _values[iatt];
        }
        _dFunc->eval(&_values[0],&res);_kEval++;
        if(_blog)
            cout<<"** in TABC::computeABC, res ---" << res <<endl;
        dBase[1+_nPar] = res;
        _tdsPar->getTuple()->Fill(&dBase[0]);
    }
}
void TABC::posteriorToPar()
{
    TList *listOfParametersLeaves = new TList;
    for (int i=0;i<_nPar;i++)
    {
        listOfParametersLeaves->Add(_tdsPosterior->getTuple()->GetLeaf( _listOfParameters->At(i)->GetName() ) );
    }
    listOfParametersLeaves->Add(_tdsPosterior->getTuple()->GetLeaf( "agreement" ) );
    _tdsPar->createTuple();
    
    vector<double> dBase(2+_nPar);
    int iatt=0;
    for(int k = 0; k < _tdsPosterior->getTuple()->GetEntries(); k++)
    {
        _tdsPosterior->getTuple()->GetEntry(k);
        TIter nextLeaf(listOfParametersLeaves);
        dBase[0]=double(k);
        iatt=0;
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
            dBase[1+iatt] = leaf->GetValue();
            iatt++;
        }
        _tdsPar->getTuple()->Fill(&dBase[0]);
    }
    
    delete listOfParametersLeaves; 

    computeAPosterioriForDistribution();

}

// ---------------- Compute discrepancy 
void TABC::computeABCscore(double dEpsilon)   
{
    TList *listOfParametersLeaves = new TList; // Should be factorised to be created and deleted once
    for (Int_t i=0;i<_nPar;i++)
    {
        listOfParametersLeaves->Add(_tdsPar->getTuple()->GetLeaf( _listOfParameters->At(i)->GetName() ) );
    }
    TLeaf *tlf = _tdsPar->getTuple()->GetLeaf("agreement");
    Int_t iatt=0;
    vector<double> dPost(_nPar + 2);

    _tdsPar->exportData( TString("_")+TString(_tdsPar->GetName())+TString("_ABC_not_trimmed_.dat").Data() );
    for(Int_t k = 0; k<_tdsPar->getTuple()->GetEntries(); k++)
    {
        _tdsPar->getTuple()->GetEntry(k);
        if(tlf->GetValue()<=dEpsilon)
        {
            TIter nextLeaf(listOfParametersLeaves);
            dPost[0]=double(k);
            iatt=0;
            while (TLeaf *leaf = (TLeaf *) nextLeaf())
            {
                dPost[1+iatt] = leaf->GetValue();
                iatt++;
            }
            dPost[1+iatt]=tlf->GetValue();
            _tdsPosterior->getTuple()->Fill(&dPost[0]);
            _kPosterior++;
        }
    }
    delete listOfParametersLeaves;
}


void TABC::setGaussianNoise(const char *stdname)
{
    try
    {
        if( _dFunc==NULL )
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, "The distance function is not yet defined, please ensure that is done before setting gaussian noise");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
        
    _dFunc->setGaussianRandomNoise(stdname);
}

} // End of namespace ABC
} // End of namespace URANIE
