//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TCalibration
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TCalibration.cxx
  \author Jean-Baptiste Blanchard
  \date   2020/07/17

  \brief Implementation of the class URANIE::Calibration::TCalibration


*/

// ROOT
#include "TCanvas.h"
#include "TLeaf.h"
#include "TH1.h"

// Uranie
#include "TCalibration.h"
#include "TStandardDistanceFunction.h"

ClassImp (URANIE::Calibration::TCalibration)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------

TCalibration::TCalibration(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    _tdsPar(tds), _tdsObs(NULL), _tdsEval(NULL), _nLauncher(kFunction), _sFunctionName(""), _sEI(varexpinput), _sEO(varexpoutput), _code(NULL),
    _run(NULL), _pFunction(fcn), _dFunc(NULL), _evTuple(NULL), _listOfParameters(NULL), _canvas(NULL), _drawingGarbageCollector(NULL),
    _nSam(ns), _nObs(0), _nPar(0), _nVar(0), _blog(false)
{
    initInputs();
    
    _voutName.clear();
    // Check whether inputs and references are in the reference 
    std::vector<std::string> ParsedLines;
    std::string delimiter = ":";
    std::string token;

    // First inputs
    std::string s=varexpoutput;
    URANIE::DataServer::UEntry::StringParser(s, ":", ParsedLines);
            
    for(unsigned int ii=0; ii<ParsedLines.size(); ii++)
    {
        _voutName.push_back(ParsedLines.at(ii));
        _outputName = _outputName + _voutName.back() + ((ii!=(ParsedLines.size()-1)) ? ":" : "")  ;
    }
    _nVar=_voutName.size();

    parseOption(option);
}

TCalibration::TCalibration(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    _tdsPar(tds), _tdsObs(NULL), _tdsEval(NULL), _nLauncher(kFunction), _sFunctionName(fcn), _sEI(varexpinput), _sEO(varexpoutput), _code(NULL),
    _run(NULL), _pFunction(NULL), _dFunc(NULL), _evTuple(NULL), _listOfParameters(NULL), _canvas(NULL), _drawingGarbageCollector(NULL),
    _nSam(ns), _nObs(0), _nPar(0), _nVar(0), _blog(false)
{
    initInputs();    

    _voutName.clear();
    // Check whether inputs and references are in the reference 
    std::vector<std::string> ParsedLines;
    std::string delimiter = ":";
    std::string token;

    // First inputs
    std::string s=varexpoutput;
    URANIE::DataServer::UEntry::StringParser(s, ":", ParsedLines);
            
    for(unsigned int ii=0; ii<ParsedLines.size(); ii++)
    {
        _voutName.push_back(ParsedLines.at(ii));
        _outputName = _outputName + _voutName.back() + ((ii!=(ParsedLines.size()-1)) ? ":" : "")  ;
    }
    _nVar=_voutName.size();

    parseOption(option);
}

TCalibration::TCalibration(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    _tdsPar(tds), _tdsObs(NULL), _tdsEval(NULL), _nLauncher(kCode), _sFunctionName(""), _sEI(""), _sEO(""), _code(fcode), _run(NULL),
    _pFunction(NULL), _dFunc(NULL), _evTuple(NULL), _listOfParameters(NULL), _canvas(NULL), _drawingGarbageCollector(NULL),
    _nSam(ns), _nObs(0), _nPar(0), _nVar(0), _blog(false)
{
    initInputs();    

    _voutName.clear();
    // Check whether inputs and references are in the reference 
    std::vector<std::string> ParsedLines;
    std::string delimiter = ":";
    std::string token;

    // First inputs
    std::string s=string(_code->getOutputVarNames());
    URANIE::DataServer::UEntry::StringParser(s, ":", ParsedLines);
            
    for(unsigned int ii=0; ii<ParsedLines.size(); ii++)
    {
        _voutName.push_back(ParsedLines.at(ii));
        _outputName = _outputName + _voutName.back() + ((ii!=(ParsedLines.size()-1)) ? ":" : "")  ;
    }
    _nVar=_voutName.size();
    
    parseOption(option);
}

TCalibration::TCalibration(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *runner, Int_t ns, Option_t * option) :
    _tdsPar(tds), _tdsObs(NULL), _tdsEval(NULL), _nLauncher(kRun), _sFunctionName(""), _sEI(""), _sEO(""), _code(NULL), _run(runner),
     _pFunction(NULL), _dFunc(NULL), _evTuple(NULL), _listOfParameters(NULL), _canvas(NULL), _drawingGarbageCollector(NULL),
    _nSam(ns), _nObs(0), _nPar(0), _nVar(0), _blog(false)
{
    initInputs();    
    
    _voutName.clear();
    TObjArray *evalList = runner->getEval()->getOutputList();
    unsigned int ii=0, limit=evalList->GetEntries();
    for(ii=0; ii<limit; ii++)
    {
        _voutName.push_back(evalList->At(ii)->GetName());
        _outputName = _outputName + _voutName.back() + ((ii!=(limit-1)) ? ":" : "")  ;
    }
    _nVar=_voutName.size();
    
    parseOption(option);
}


void TCalibration::initInputs()
{

    setMethodName("Calibration");
        
    _buseMatrix=false;
    _bsaveAll=false;
    _bdontKeepAgreement=false;
    _buseMode=false;
    
    try
    {
        if (_tdsPar == NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TCalibration::constructor \n The TDS is NULL");
        if(_tdsPar->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TCalibration::constructor \n The TDS is empty");
        if( _nLauncher == kRun )
        {
            if (_run == NULL)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TCalibration::constructor \n The runner provided is NULL");
            if(_run->getEval()->getInputList()->GetEntries()==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TCalibration::constructor The provided evaluator (through the runner) has not input attributes.");
            if(_run->getEval()->getOutputList()->GetEntries()==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TCalibration::constructor The provided evaluator (through the runner) has not output attributes.");
        }
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if(_listOfParameters != NULL)
    {
        delete _listOfParameters;
        _listOfParameters = NULL;
    }

    _listOfParameters=_tdsPar->getListOfAttributes();
    _nPar=_listOfParameters->GetEntries();
    _nIterMax=-1;
}


void TCalibration::initResults(vector<string> *ParsedLines)
{
    _tdsEval = new URANIE::DataServer::TDataServer( (string(_tdsObs->GetName())+string("__Eval__")).c_str(),"" );

    string varlist="";
    for (unsigned int is=0; is<ParsedLines->size(); is++)
    {
        _tdsEval->addAttribute( ParsedLines->at(is).c_str() );
        varlist+= ParsedLines->at(is) + ((is<(ParsedLines->size()-1)) ? ":" : "");
    }
    _evTuple = new URANIE::DataServer::TDSNtupleD(_tdsObs->getTuple(),varlist.c_str());
    _tdsEval->setTuple( _evTuple );
    _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), _tdsEval->getIteratorName(), _tdsObs->getIteratorName());

    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
}


// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TCalibration::~TCalibration()
{
    if (_blog)
        cout << " *** destroy TCalibration(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    
    if( _evTuple != NULL )
        delete _evTuple;
    _evTuple=NULL;

    if( _tdsEval != NULL )
        delete _tdsEval;
    _tdsEval=NULL;

    if( _drawingGarbageCollector != NULL )
        delete _drawingGarbageCollector;
    _drawingGarbageCollector=NULL;

    if( _dFunc != NULL )
        delete _dFunc;
    _dFunc=NULL;

    if( _listOfParameters != NULL )
        delete _listOfParameters;
    _listOfParameters=NULL;
}


// -----------------------------------------------------------------------
//  Define properties
// -----------------------------------------------------------------------

void TCalibration::checkReference(URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *reference, const char *weight)
{

    vector<string> ParsedLines;
    unsigned int ii=0;
    try
    {
        // Check whether this is NULL
        if (tdsRef == NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TCalibration::setReference \n The TDS is NULL");
        // Check whether this tds is empty
	else if(tdsRef->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TCalibration::setReference \n The TDS is empty");

        // Check whether inputs and references are in the reference 
        std::string delimiter = ":";
        std::string token;

        // First inputs
        std::string s=input;
        URANIE::DataServer::UEntry::StringParser(s, ":", ParsedLines);
        for(ii=0; ii<ParsedLines.size(); ii++)
            if( ! tdsRef->isAttribute( ParsedLines.at(ii).c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TCalibration::setReference \n The TDS does not contains [%s]",ParsedLines.at(ii).c_str()));

        // Then references
        s=reference;
        _vrefName.clear();
        URANIE::DataServer::UEntry::StringParser(s, ":", _vrefName);
        for(ii=0; ii<_vrefName.size(); ii++)
        {
            if( ! tdsRef->isAttribute( _vrefName.at(ii).c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TCalibration::setReference \n The TDS does not contains [%s]",_vrefName.at(ii).c_str()));
            ParsedLines.push_back(_vrefName.at(ii));
        }
        _referenceName=reference;

        // Then weights
        s=weight;
        vector<string> weiwei;
        URANIE::DataServer::UEntry::StringParser(s, ":", weiwei);
        for(ii=0; ii<weiwei.size(); ii++)
        {
            if( ! tdsRef->isAttribute(weiwei.at(ii).c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TCalibration::setReference \n The TDS does not contains [%s]",weiwei.at(ii).c_str()));
            ParsedLines.push_back(weiwei.at(ii));
        }
        _weightName=weight;

        // Testing number of outputs and references
        if( _vrefName.size() != _voutName.size() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TCalibration::setReference \n The reference provided [%s=>%lu] does not have the same number of variables than the output [%s=>%lu]",
                                                            _referenceName.Data(), _vrefName.size(), _outputName.Data(), _voutName.size()));
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } 

    // Keep the reference as observations
    _tdsObs = tdsRef;
    _nObs = _tdsObs->getNPatterns();

    // Create the final tuple to get the residuals 
    initResults(&ParsedLines);
}

void TCalibration::computeAPosterioriForDistribution()
{

    vector<double> theta(_nPar);
    cout<<((_buseMode)?"Smoother MAP":"A posteriori mean")<<" coordinates are : (";
    double x=0;
    for(int ivar=0; ivar<_nPar; ivar++)
    {
        _tdsPar->getTuple()->Draw( Form("%s >> hplot%s", _vatt.at(ivar)->GetName(),_vatt.at(ivar)->GetName()), "" , "goff" );
        TH1F *h = (TH1F*)gROOT->Get(Form("hplot%s",_vatt.at(ivar)->GetName()));
        x = h->GetMean();
        if(_buseMode)
        {
            h->Smooth(10000);
            int binmax = h->GetMaximumBin();
            x = h->GetXaxis()->GetBinCenter(binmax);
        }
        theta[ivar]=x;
        cout<<x<<((ivar==(_nPar-1))? ")" : ",");
    }
    cout<<endl;
                       
    _dFunc->runConfiguration( &theta[0] );
    
    // Save the aposteriori results
    for(unsigned int ii=0; ii<_vrefName.size(); ii++)
        _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), ( _vrefName.at(ii)+"_aposteriori").c_str(), _voutName.at(ii).c_str() );      
}

void TCalibration::setDistanceAndReference(const char*funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *reference, const char *weight)
{

    checkReference(tdsRef, input, reference, weight);
    
    TString functionAsked=funcName;
    functionAsked.ToLower();
    
    if (functionAsked=="ls")
        _dFunc = new URANIE::Calibration::TLSDistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    else if (functionAsked=="weightedls" && _weightName!="")
        _dFunc = new URANIE::Calibration::TWeightedLSDistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    else if (functionAsked=="relativels")
        _dFunc = new URANIE::Calibration::TRelativeLSDistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    else if (functionAsked=="l1")
        _dFunc = new URANIE::Calibration::TL1DistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    else if (functionAsked=="mahalanobis")
        _dFunc = new URANIE::Calibration::TMahalanobisDistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    // else if (functionAsked=="weighteddiff")
    //     _dFunc = new URANIE::Calibration::TWeightedDiffDistanceFunction(_tdsEval, _tdsObs, _referenceName, _outputName, _weightName);
    else
    {
        try
        {
            if(functionAsked=="weightedls" && _weightName=="")
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TCalibration::setDistanceAndReference \n Weighted LS is chosen while no weight is provide");
            else
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TCalibration::setDistanceAndReference \n the distance function is not recognised [%s]",functionAsked.Data()));
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
    }

    if(_buseMatrix)
        _dFunc->setUseMatrix(_buseMatrix);

    if(_blog)
        _dFunc->setLog();

    switch(_nLauncher)
    {
    case kRun:
        _dFunc->initParameters(_tdsPar, _run);
        break;
    case kFunction:
        if(_pFunction==NULL)
            _dFunc->initParameters(_tdsPar, _sFunctionName.Data(), _sEI.Data(), _sEO.Data());
        else
            _dFunc->initParameters(_tdsPar, _pFunction, _sEI.Data(), _sEO.Data());
        break;
    case kCode:
        _dFunc->initParameters(_tdsPar, _code);
        break;
    case kUnknown:
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TCalibration::setDistanceAndReference \n the launching method is not recognised");
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        break;
    }
    
    if(_bsaveAll)
        _dFunc->setSaveAllEvaluations(_bsaveAll);
}

void TCalibration::setDistanceAndReference(URANIE::Calibration::TDistanceFunction *func, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *reference, const char *weight)
{
    checkReference(tdsRef, input, reference, weight);
    
    _dFunc = func;

    switch(_nLauncher)
    {
    case kRun:
        _dFunc->initParameters(_tdsPar, _run);
        break;
    case kFunction:
        if(_pFunction==NULL)
            _dFunc->initParameters(_tdsPar, _sFunctionName.Data(), _sEI.Data(), _sEO.Data());
        else
            _dFunc->initParameters(_tdsPar, _pFunction, _sEI.Data(), _sEO.Data());
        break;
    case kCode:
        _dFunc->initParameters(_tdsPar, _code);
        break;
    case kUnknown:
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TCalibration::setDistanceAndReference \n the launching method is not recognised");
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        break;
    }
    
    if(_bsaveAll)
        _dFunc->setSaveAllEvaluations(_bsaveAll);

}

void TCalibration::setObservationCovarianceMatrix(TMatrixD &mat)
{
    
    try
    {
        if( _nObs==0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TCalibration::setObservationCovarianceMatrix \n\
 This method should not be called while setDistanceAndReference method has not be called, as no data have been provided.");
        
            
        if( _nObs != mat.GetNrows() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TCalibration::setObservationCovarianceMatrix \n\
 The provided covariance matrix has wrong number of rows [%d]. It should be [%d]", mat.GetNrows(), _nObs));
        
        if( _nObs != mat.GetNcols() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TCalibration::setObservationCovarianceMatrix \n\
 The provided covariance matrix has wrong number of column [%d]. It should be [%d]", mat.GetNcols(), _nObs));

        if( !mat.IsSymmetric() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TCalibration::setObservationCovarianceMatrix \n\
 The provided covariance matrix is not symmetrical.");
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    try
    {
        if(_dFunc == NULL)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         "TCalibration::setObservationCovarianceMatrix \n \
This method shoule be called once the distance function is chosen.\n    \
Call it again once done");
        else
            _dFunc->setObservationCovarianceMatrix(mat);
    }
    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    }
    
    _mObsCovMat.ResizeTo(mat);
    _mObsCovMat.SetMatrixArray( mat.GetMatrixArray() );

    double det2=0;
    _mObsCovMat.Invert(&det2);

}


void TCalibration::estimateParameters(Option_t *option)
{

    computeParameters(option);

    if( _bdontKeepAgreement && _tdsPar->isAttribute("agreement") )
        _tdsPar->delAttribute("agreement");


    // Cleaning the datatree from removed attributes
    TList *listOfRemainingAttributes = _tdsPar->getListOfAttributes("*",false);
    TObjArray *listOfLeaves = _tdsPar->getTuple()->GetListOfLeaves();
    TString leaflist="";
    
    bool replaceIt=false;
    TIter nextLeaf(listOfLeaves);
    while (TLeaf *leaf = (TLeaf *) nextLeaf())
        if(listOfRemainingAttributes->FindObject(leaf->GetName()) != NULL )
            leaflist+=TString(leaf->GetName())+TString(":");
        else
            replaceIt=true;
    
    leaflist.Replace(leaflist.Length()-1, 1, "");    

    if(replaceIt)
    {
        _tdsPar->setTuple(_tdsPar->getTuple(), true, leaflist);
        URANIE::DataServer::TDSNtupleD *testTDSN=NULL;
        if( gDirectory->FindObject( _tdsPar->getTuple()->GetName()+TString("_copied") ) != NULL )
        {
            testTDSN=(URANIE::DataServer::TDSNtupleD*)gDirectory->FindObject( _tdsPar->getTuple()->GetName()+TString("_copied") );
            delete testTDSN;
        }
        testTDSN=NULL;
    }
    delete listOfRemainingAttributes;
}

void TCalibration::estimateCustomResidues(string resName, int n_theta, double *theta)
{

    try{
        if(n_theta != _nPar)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TCalibration::estimateCustomResidues\n \
The dimension of the provided array of value [%d] does not match the dimension of parameter space [%d] is NULL", n_theta, _nPar));

        if (_tdsEval->getTuple()->GetEntries() == 0) {
            cout << "TCalibration::estimateCustomResidues Error : the tdsEval is empty" << endl;
            return;
        }
        
        for(unsigned int ii=0; ii<_vrefName.size(); ii++)
        { 
            string outname = (_vrefName.at(ii)+"_"+resName);
            if(_tdsEval->isAttribute(outname.c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TCalibration::estimateCustomResidues\n \
The flag [%s] has already been used, since the variable [%s] is already stored in the _tdsEval", resName.c_str(), outname.c_str()));
        }

        if(_dFunc==NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TCalibration::estimateCustomResidues\n \
The distance function is not set yet, this function should be called after the estimateParameters one.");
           
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } 

    double pouet=0;
    _dFunc->eval(theta, &pouet);
    
    // Save the aposteriori results    
    for(unsigned int ii=0; ii<_vrefName.size(); ii++)
        _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), ( _vrefName.at(ii)+"_"+resName).c_str(), _voutName.at(ii).c_str() );

}



// -----------------------------------------------------------------------
//  Drawing related methods
// -----------------------------------------------------------------------

void TCalibration::checkCanvasCreation(bool newcan)
{
        
    if (gPad && !newcan)
    {
        _canvas = (TCanvas *) (gPad);
        if( TString(_canvas->GetName()).BeginsWith("__calibrationcan__"))
            _canvas->Clear();
        else
            _canvas = (TCanvas*)gPad;
    }
    else
    {
        stringstream sstr; sstr << "__calibrationcan__" << _drawingGarbageCollector->GetEntries() ;
        _canvas = new TCanvas(sstr.str().c_str(),sstr.str().c_str(),1);
        _drawingGarbageCollector->Add(_canvas);
    }

    //XYZ<<"[91m The pointer to the canvas created or not is "<<_canvas<<"[0m"<<endl;
    
}

void TCalibration::drawResidues(TString sTitre, const char *variable, const char *select, Option_t * option)
{

    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TCalibration::drawResidues for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsEval->getTuple()->GetEntries() == 0) {
        cout << "TCalibration::drawResidues Error : the tdsEval is empty" << endl;
        return;
    }

    if (soption.Contains("nocanv") || soption.Length() == 0)
        return;

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
        newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    int iter=0;

    vector<string> legend;

    vector<string> pouet;
    URANIE::DataServer::UEntry::StringParser(string(soption),",",pouet);
    for(unsigned int iopt = 0 ; iopt < pouet.size(); iopt++)
    {
        TString localopt = pouet.at(iopt);
        if( localopt.Contains("apriori") )
        {
            localopt.ReplaceAll("apriori", "");
            legend.push_back("apriori");
            cleanoption.ReplaceAll("apriori", "");
        }

        if( localopt.Contains("aposteriori") )
        {
            localopt.ReplaceAll("aposteriori", "");
            legend.push_back("aposteriori");
            cleanoption.ReplaceAll("aposteriori", "");
        }

        if( localopt.Contains("custom=") )
        {
            localopt.ReplaceAll("custom=", "");
            legend.push_back(string(localopt));
            cleanoption.ReplaceAll("custom=", "");
            cleanoption.ReplaceAll(localopt, "");
        }
        
    }

    if(legend.size()==0)
    {
        legend.push_back("apriori");
        legend.push_back("aposteriori");
    }
    
    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" || svariable=="" )
        for(unsigned int ivar=0; ivar<_vrefName.size(); ivar++)
            variables.push_back( _vrefName.at(ivar) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if (soption.Contains("vertical"))
    {
        _canvas->Divide( variables.size(), legend.size());
        cleanoption.ReplaceAll("vertical","");
    }
    else
        _canvas->Divide(legend.size(), variables.size());

    cleanoption.ReplaceAll(",", "");
    string cmpname="", leg="", varname="";
    TH1 *h=NULL;
    URANIE::DataServer::TAttribute *att=NULL;
    for (unsigned int itype=0; itype<legend.size(); itype++)
    {

        if(legend.at(itype)=="apriori") { cmpname="_apriori"; leg="prio";}
        else if(legend.at(itype)=="aposteriori") { cmpname="_aposteriori"; leg="post";}
        else { cmpname = "_" + legend.at(itype); leg = legend.at(itype); }
        
        for(unsigned ivar=0; ivar<variables.size(); ivar++)
        {

            _canvas->cd( itype*variables.size() + ivar + 1 );
            
            varname="Delta_"+variables.at(ivar)+cmpname;
            if( !_tdsEval->isAttribute( varname.c_str() ) )
                _tdsEval->addAttribute( varname.c_str(),  (variables.at(ivar)+"-"+variables.at(ivar)+cmpname).c_str() );

            att=(URANIE::DataServer::TAttribute*)_tdsObs->getAttribute( variables.at(ivar).c_str() );

            _tdsEval->getTuple()->Draw( varname.c_str(), select, cleanoption);
            h = (TH1*)gPad->GetPrimitive("htemp");
            h->SetName( ("h"+varname).c_str() );
            h->SetTitle( ((sTitre=="") ? ("h"+varname).c_str() : sTitre.Data() ) );
            h->GetXaxis()->SetTitle( ("#Delta^{"+leg+"} "+string(att->getTitle())).c_str() ) ;            
        }
    }
}

void TCalibration::drawParameters(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TCalibration::drawParameters for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) {
        cout << "TCalibration::drawParameters Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");
    
    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(int ivar=0; ivar<_listOfParameters->GetEntries(); ivar++)
            variables.push_back( string( _listOfParameters->At(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if( !soption.Contains("same") )
    {
        if (soption.Contains("vertical"))
        {
            _canvas->Divide(1, variables.size());
            cleanoption.ReplaceAll("vertical","");
        }
        else
            _canvas->Divide(variables.size(), 1);
    }
    
    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = sselect;
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        _tdsPar->getTuple()->Draw( variables.at(ivar).c_str(), cut.c_str(), cleanoption.Data() );
        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( ("h"+variables.at(ivar)+"_aposteriori").c_str() );
        h->SetTitle( ((sTitre=="") ? ("h"+variables.at(ivar)+"_aposteriori").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle(att->getTitle());
    }
}


// -----------------------------------------------------------------------
//  Log and Options
// -----------------------------------------------------------------------

void TCalibration::parseOption(Option_t *option)
{
    TString soption=option;
    soption.ToLower();

    if(soption.Contains("savealleval "))
        _bsaveAll=true;

    if(soption.Contains("usematrix"))
        _buseMatrix=true;

    if(soption.Contains("noagreement"))
        _bdontKeepAgreement=true;

    if( soption.Contains("mode") )
	_buseMode=true;

}

void TCalibration::printLog(Option_t *option)
{
    cout << " ** TCalibration::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **   ID[" << getID() << "]" << endl;
    cout << " ** _nSam[" << _nSam << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " **  Log of DataServer (unactive)" << endl;
    if ( 0 ) {
        _tdsPar->printLog();
        cout << " **   End of Log of DataServer" << endl;
    }
    cout << " ** End of TCalibration::printLog[" << option << "]" << endl;
}

} // End of namespace Calibration
} // End of namespace URANIE

