
//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDistanceFunction.cxx
 \author Jean-Baptiste Blanchard (jean-baptiste.blanchard@cea.fr)
 \date   

 \brief Implementation of the class  URANIE::Calibration::TDistanceFunction

 */

//// ROOT
#include "TMath.h"

//// URANIE
#include "TDistanceFunction.h"
#include "TLauncher2.h"
#include "TCodeEval.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TLauncherByStep.h"
#ifdef WITH_LIBSSH
#include "TLauncherByStepRemote.h"
#endif
#include "TAttribute.h"

ClassImp (URANIE::Calibration::TDistanceFunction)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors &  initialisation
// -----------------------------------------------------------------------
TDistanceFunction::TDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight) :
    _tdsEval(tdsEval), _tdsObs(tdsObs), _allData(NULL), _run(NULL), _tlcType("TLauncher"), _codeLauncherOpt("nointermediatestep,quiet"), _code(NULL),
    _funcName(""), _pFunction(NULL), _baprioriSet(false), _bsaveAll(false), _buseMatrix(false), _blog(false), _dumpAllTds(false), _keepParValue(false),
    _bdumpDetails(false), _icalc(0), _listOfParameters(NULL), _rand(NULL) 
{
    string outname=output;
    string refname=reference;
    URANIE::DataServer::UEntry::StringParser(refname,":",_vrefName);
    URANIE::DataServer::UEntry::StringParser(outname,":",_voutName);

    try
    {
        if( _tdsEval == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The tdsEval pointer is null. This should not be.");        
        if( _tdsEval->getDataSpecification() == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The tdsEval seems empty. This should not be.");        
        if( _tdsObs == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The tdsObs pointer is null. This should not be.");
        if( _tdsObs->getDataSpecification() == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The tdsObs seems empty. This should not be.");
        if( _vrefName.size() == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The reference string is empty. This should not be.");
        if( _voutName.size() == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::constructor\n The output string is empty. This should not be.");
        if( _vrefName.size() != _voutName.size() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::constructor\n\
The number of variable provided as reference [%lu : %s] and as output of the assessor [%lu : %s] is different !!", _vrefName.size(), refname.c_str(), _voutName.size(), outname.c_str()));
        
        for(unsigned int i=0; i<_vrefName.size(); i++)
            if( !_tdsObs->isAttribute( _vrefName.at(i) ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TDistanceFunction::constructor\n\
The reference name [%s] is not found in the tdsObs provided. This should be fixed.",_vrefName.at(i).c_str()));
        
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _nObs = _tdsObs->getNPatterns();
    _nVar = _vrefName.size();
    
    _yObs.resize(_nVar);
    _yExp.resize(_nVar);

    _myObs.ResizeTo(_nVar, _nObs);
    _myExp.ResizeTo(_nVar, _nObs);

    _varWeights.resize(_nVar,1.);

    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        _yObs.at(_ivar).resize(_nObs);
        _yExp.at(_ivar).resize(_nObs);
    }

    TString sweight=weight;
    if( sweight != "")
    {
        vector<string> weiName;
        URANIE::DataServer::UEntry::StringParser(string(sweight), ":", weiName);
    
        try
        {
            if( weiName.size() != _nVar )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TDistanceFunction::constructor\n \
The weight string seems to have a size [%s=>%lu] different from the nmber of reference variable [%d]. This should not be.",sweight.Data(),weiName.size(), _nVar));

            for(_ivar=0; _ivar<_nVar; _ivar++)
                if( !_tdsObs->isAttribute( weiName.at(_ivar) ) )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TDistanceFunction::constructor\n\
The weight name [%s] is not found in the tdsObs provided. This should be fixed.",weiName.at(_ivar).c_str()));
        
        }
        catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

        // Resize observations weights
        _obsWeights.resize(_nVar);
        for(_ivar=0; _ivar<_nVar; _ivar++)
        {
            _obsWeights.at(_ivar).resize(_nObs);        
            int nWei = _tdsObs->getTuple()->Draw(weiName.at(_ivar).c_str(),"","goff");
            double *dataW = _tdsObs->getTuple()->GetV1();
            _obsWeights.at(_ivar).assign( &dataW[0], &dataW[0]+nWei);
        }
    }
    
    // Noise 
    _nSeed = 0;
    _boolNoise = false;

    getData(true);
}

void TDistanceFunction::fillInOutAtt(URANIE::DataServer::TDataServer *tdsPar)
{

    if(_blog)
        XYZ<<"In fillInOutAtt"<<endl;
    
    _listOfParameters = tdsPar->getListOfAttributes();
    TIter next(_listOfParameters);
    _nPar=0;

    _vparName.clear();
    URANIE::DataServer::TAttribute *att=NULL;
    while( (att = (URANIE::DataServer::TAttribute *)next.Next()) )
    {
        this->addInput(att);
        _vparName.push_back( string(att->GetName()) );
        _nPar++;
    }

    _vParValues.resize(_nPar);

    // Add output attributes
    tdsPar->addAttribute("agreement");    
    this->addOutput(tdsPar->getAttribute("agreement"));
    
    
}

void TDistanceFunction::initParameters(URANIE::DataServer::TDataServer *tdsPar, URANIE::Relauncher::TRun *run)
{
    if(_blog)
        XYZ<<"In initParameters"<<endl;

    fillInOutAtt(tdsPar);
    _run = run;
    URANIE::Relauncher::TCodeEval *foo = dynamic_cast<URANIE::Relauncher::TCodeEval*>( _run->getEval() );
    if(foo != NULL)
        foo->setOldTmpDir();
}

void TDistanceFunction::initParameters(URANIE::DataServer::TDataServer *tdsPar, const char *funcName, const char *varexpinput, const char *varexpoutput)
{
    if(_blog)
        XYZ<<"In initParameters"<<endl;

    fillInOutAtt(tdsPar);
    _funcName = funcName;
    _varexpinput = varexpinput;
    _varexpoutput = varexpoutput;

    for(_ivar=0; _ivar<_nPar; _ivar++)        
        if( !_tdsObs->isAttribute( _vparName.at(_ivar) ) )
            _tdsObs->addAttribute( _vparName.at(_ivar).c_str() );    
}

void TDistanceFunction::initParameters(URANIE::DataServer::TDataServer *tdsPar, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput)
{
    if(_blog)
        XYZ<<"In initParameters"<<endl;

    fillInOutAtt(tdsPar);
    _pFunction=fcn;
    _varexpinput = varexpinput;
    _varexpoutput = varexpoutput;

    for(_ivar=0; _ivar<_nPar; _ivar++)        
        if( !_tdsObs->isAttribute( _vparName.at(_ivar) ) )
            _tdsObs->addAttribute(_vparName.at(_ivar).c_str());  
    
}

void TDistanceFunction::initParameters(URANIE::DataServer::TDataServer *tdsPar, URANIE::Launcher::TCode *code)
{
    if(_blog)
        XYZ<<"In initParameters"<<endl;

    fillInOutAtt(tdsPar);
    _code = code;

    TIter next(_listOfParameters);
    URANIE::DataServer::TAttribute *att=NULL;
    while( (att = (URANIE::DataServer::TAttribute *)next.Next()) )
        _tdsObs->addAttribute( att );        
}

void TDistanceFunction::changeLauncher(TString tlcName)
{
    try
    {
        if( _code == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDistanceFunction::changeLauncher\n The code pointer is NULL....This should not be.");
        
        if( tlcName!="TLauncher" && tlcName!="TLauncherByStep" && tlcName!="TLauncherByStepRemote")
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::changeLauncher\n The type requested [%s] is not in the list TLauncherByStep, TLauncherByStepRemote.",
                                                            tlcName.Data()));
#ifndef WITH_LIBSSH
        if( tlcName=="TLauncherByStepRemote")
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::changeLauncher\n The type requested [%s] is not availlable as libssh is not available.",
                                                            tlcName.Data()));
#endif
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _tlcType=tlcName;
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TDistanceFunction::~TDistanceFunction()
{
    if (_blog)
        cout << "detruit TDistanceFunction" << endl;
    
    if( _allData != NULL)
        delete _allData;
    _allData=NULL;

    if(_listOfParameters != NULL)
        delete _listOfParameters;
    _listOfParameters = NULL;
    
    if(_rand!=NULL)
        delete _rand;
    _rand=NULL;
}


// -----------------------------------------------------------------------
//  Evaluation
// -----------------------------------------------------------------------

void TDistanceFunction::runConfiguration(double *theta)
{
    if(_blog)
        XYZ<<"In runConfiguration"<<endl;
            
    for(_ivar=0; _ivar<_nVar; _ivar++)        
        if(_tdsObs->isAttribute(_voutName.at(_ivar)))
            _tdsObs->delAttribute(_voutName.at(_ivar));

    if(_run != NULL)
    {
        _run->startSlave();
        TIter next(_listOfParameters);
        URANIE::DataServer::TAttribute *att=NULL;
        while( (att = (URANIE::DataServer::TAttribute *)next.Next()) )
            if(_tdsObs->isAttribute(att->GetName()))
                _tdsObs->delAttribute( att->GetName() );

        _tdsObs->deleteTuple();
        _tdsObs->setTuple(_allData, false);
            
        // Estimate the stuff
        if(_run->onMaster())
        {            
            URANIE::Relauncher::TLauncher2 launch(_tdsObs, _run);
            if(_blog)
                XYZ<<"Launcher created"<<endl;
            
            TIter next(&_input);
            int i=0;
            URANIE::DataServer::TAttribute *att=NULL;
            while( (att = (URANIE::DataServer::TAttribute *)next.Next()) )
                launch.addConstantValue(att,theta[i++],_keepParValue);            
            
            if(_blog)
                XYZ<<"Before solverLoop"<<endl;            
            launch.solverLoop();

            if(_blog)
                XYZ<<"After solverLoop"<<endl;            
            _run->stopSlave();        
        }
    }
    else
    {
        TIter next(_listOfParameters);
        URANIE::DataServer::TAttribute *att=NULL;
        while( (att = (URANIE::DataServer::TAttribute *)next.Next()) )
            if(_tdsObs->isAttribute(att->GetName()))
                _tdsObs->delAttribute( att->GetName() );
        
        _tdsObs->deleteTuple();
        _tdsObs->setTuple(_allData, false);
        
        TIter next2(_listOfParameters);
        while( (att = (URANIE::DataServer::TAttribute *)next2.Next()) )
            _tdsObs->addAttribute( att );
        
        for(_ivar=0; _ivar<_nPar; _ivar++)
        {
            auto newBranch = _tdsObs->getTuple()->Branch(_vparName.at(_ivar).c_str(), &_vParValues.at(_ivar), (_vparName.at(_ivar)+"/D").c_str() );
            _vParValues.at(_ivar)=theta[_ivar];
            for (Long64_t i = 0; i < _nObs; i++) 
                newBranch->Fill();            
            _tdsObs->getTuple()->ResetBranchAddress(newBranch);
        }               
        
        if( _code != NULL)
        {
            if( _tlcType=="TLauncher" )
            {
                URANIE::Launcher::TLauncher tlc(_tdsObs, _code, "quiet");
                tlc.setDrawProgressBar(false);
                tlc.run( _codeLauncherOpt.Data() );
            }
            else if( _tlcType=="TLauncherByStep" )
            {
                URANIE::Launcher::TLauncherByStep tlc(_tdsObs, _code, "quiet");
                tlc.setDrawProgressBar(false);
                tlc.run( _codeLauncherOpt.Data() );
            }
            else if( _tlcType=="TLauncherByStepRemote" )
            {
#ifdef WITH_LIBSSH
                URANIE::Launcher::TLauncherByStepRemote tlc(_tdsObs, _code, "quiet");
                tlc.setDrawProgressBar(false);
                tlc.run( _codeLauncherOpt.Data() );
#endif
            }            
        }
        else if (_funcName!="")
        {
            URANIE::Launcher::TLauncherFunction tlf(_tdsObs, _funcName.Data(), _varexpinput.Data(), _varexpoutput.Data());
            tlf.setDrawProgressBar(false);
            tlf.run();            
        }
        else if (_pFunction!=NULL)
        {
            URANIE::Launcher::TLauncherFunction tlf(_tdsObs, _pFunction, _varexpinput.Data(), _varexpoutput.Data());
            tlf.setDrawProgressBar(false);
            tlf.run();            
        }

    }

    if(_dumpAllTds)
        _tdsObs->exportData(Form("Calibration_testnumber_%d.dat",_icalc));

}

void TDistanceFunction::getData(bool reference)
{
    if(reference)
    {
        for(_ivar=0; _ivar<_nVar; _ivar++)        
        {
            _tdsObs->getTuple()->Draw(_vrefName.at(_ivar).c_str(),"","goff");
            vector<double> data(_tdsObs->getTuple()->GetV1(), _tdsObs->getTuple()->GetV1()+_nObs);
            if( _buseMatrix )
                _myObs.InsertRow(_ivar,0, &data[0]);
            else
                _yObs.at(_ivar).assign( &data[0], &data[0]+_nObs);
        }

        if(_allData!=NULL)
        {
            delete _allData;
            _allData=NULL;
        }
        _allData = new URANIE::DataServer::TDSNtupleD( _tdsObs->getTuple() );
    }
    else
    {
        for(_ivar=0; _ivar<_nVar; _ivar++)        
        {

            _tdsObs->getTuple()->Draw( (_voutName.at(_ivar)+":"+string(_tdsObs->getIteratorName())).c_str(),"","goff");
            vector<double> datatmp(_tdsObs->getTuple()->GetV1(), _tdsObs->getTuple()->GetV1()+_nObs);
            vector<double> data; data.resize(_nObs);
            vector<Long64_t> v_iter(_tdsObs->getTuple()->GetV2(), _tdsObs->getTuple()->GetV2()+_nObs);

            Long64_t min=TMath::MinElement(_nObs, &v_iter[0]);
            for(unsigned int ii=0; ii<_nObs; ii++)
            {
                
                data[v_iter[ii]-min]=datatmp[ii];
                if(_boolNoise)
                {
                    data[v_iter[ii]-min]+= _mSigma(_ivar,ii) * _rand->Gaus();
                }
            }
              
            if( _buseMatrix )
                _myExp.InsertRow(_ivar,0,&data[0]);
            else
                _yExp.at(_ivar).assign( &data[0], &data[0]+_nObs);

        }
    
    }
}

int TDistanceFunction::eval(Double_t* theta, Double_t* res, int foo)
{    
    UNUSED(foo);

    if(_blog)
        XYZ<<"In Eval"<<endl;
    
    runConfiguration(theta);
    getData();
    localeval(res);

    if( !_baprioriSet )
    {
        // Save the apriori results
        for(_ivar=0; _ivar<_nVar; _ivar++)        
            _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), (_vrefName.at(_ivar)+"_apriori").c_str(), _voutName.at(_ivar).c_str());

        _baprioriSet=true;
    }
    else if(_bsaveAll){        
        // Save the apriori results
        for(_ivar=0; _ivar<_nVar; _ivar++)        
            _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), Form("%s_iter_%d",_vrefName.at(_ivar).c_str(),_icalc), _voutName.at(_ivar).c_str()); 
    }
    _icalc++;
    
    return 1;
}



// -----------------------------------------------------------------------
// Setters
// -----------------------------------------------------------------------

void TDistanceFunction::setVarWeights(int nwei, double *wei)
{
    try{
	if( nwei != int(_nVar) )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TDistanceFunction::setVarWeights\n\
The array of weights provided has dimension [%d] while it should be [%d]", nwei, _nVar));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    for(_ivar=0; _ivar<_nVar; _ivar++)
        _varWeights.at(_ivar) = wei[_ivar];
    
}

void TDistanceFunction::setVarWeights(vector<double> wei)
{
    try{
	if( wei.size() != _nVar )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TDistanceFunction::setVarWeights\n\
The vector of weights provided has dimension [%lu] while it should be [%d]", wei.size(), _nVar));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    setVarWeights(wei.size(), &wei[0]);
}

void TDistanceFunction::setObservationCovarianceMatrix(TMatrixD &mat)
{
    try
    {
        if( int(_nObs) != mat.GetNrows() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::setObservationCovarianceMatrix \n\
 The provided covariance matrix has wrong number of rows [%d]. It should be [%d]", mat.GetNrows(), _nObs));
        
        if( int(_nObs) != mat.GetNcols() )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::setObservationCovarianceMatrix \n\
 The provided covariance matrix has wrong number of column [%d]. It should be [%d]", mat.GetNcols(), _nObs));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _mObsCovMat.ResizeTo(mat);
    _mObsCovMat.SetMatrixArray( mat.GetMatrixArray() );

    // Invert the covariance matrix
    double det2;
    _mObsCovMat.Invert(&det2);
}

   
void TDistanceFunction::setGaussianRandomNoise(const char *stdname)
{
    TString sstandardDev=stdname;

    vector<string> stdName;
    URANIE::DataServer::UEntry::StringParser(string(sstandardDev), ":", stdName);

    string output="";
    try
    {
        if( stdName.size() != _nVar )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDistanceFunction::setGaussianRandomNoise\n \
The standardDev string seems to have a size [%s=>%lu] different from the nmber of reference variable [%d]. This should not be.",sstandardDev.Data(),stdName.size(), _nVar));

        for(_ivar=0; _ivar<_nVar; _ivar++)
        {
            if( !_tdsObs->isAttribute( stdName.at(_ivar) ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TDistanceFunction::setGaussianRandomNoise\n\
The standardDev name [%s] is not found in the tdsObs provided. This should be fixed.",stdName.at(_ivar).c_str()));
            output+=string(_voutName.at(_ivar))+":";
        }
        
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
        
    output = output.substr(0, output.size()-1);
    try
    {
        throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                  Form( "TDistanceFunction::setGaussianRandomNoise: \
gaussian random noise(s) is added using information in [%s] to modify the output variable(s) [%s]. ", stdname,output.c_str()));
        
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
    }
    
    if(_rand==NULL)
    {
        _rand = new TRandom3();
        if(_nSeed != 0)
            _rand->SetSeed(_nSeed);
    }

    _boolNoise = true;
    _mSigma.ResizeTo(_nVar, _nObs);
    for(_ivar=0; _ivar<_nVar; _ivar++)
    {
        int nStd = _tdsObs->getTuple()->Draw(stdName.at(_ivar).c_str(),"","goff");
        double *dataS = _tdsObs->getTuple()->GetV1();
        _mSigma.InsertRow(_ivar,0, dataS);
    }
}
// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TDistanceFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TDistanceFunction::printLog[" << option << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Calibration
} // Fin du namespace URANIE


