//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TLinearBayesian
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TLinearBayesian.cxx
  \author Jean-Baptiste Blanchard
  \date   2020/07/17

  \brief Implementation of the class URANIE::Calibration::TLinearBayesian


*/

// ROOT
#include "TCanvas.h"
#include "TH1.h"
#include "TF1.h"

// Uranie
#include "TLinearBayesian.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"

ClassImp (URANIE::Calibration::TLinearBayesian)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------

TLinearBayesian::TLinearBayesian(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *runner, Int_t ns, Option_t * option) :
    TCalibration(tds, runner, ns, option), _fTransfoParam(NULL)
{
    _outputName=runner->getEval()->getOutputList()->At(0)->GetName();

    checktdsParContent();
    
    TString soption=option;
    parseOption(soption.Data());    
}

TLinearBayesian::TLinearBayesian(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _fTransfoParam(NULL)
{
    checktdsParContent();

    TString soption=option;
    parseOption(soption.Data());
}

TLinearBayesian::TLinearBayesian(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _fTransfoParam(NULL)
{
    checktdsParContent();

    TString soption=option;
    parseOption(soption.Data());
}

TLinearBayesian::TLinearBayesian(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TCalibration(tds, fcode, ns, option), _fTransfoParam(NULL)
{
    checktdsParContent();

    TString soption=option;
    parseOption(soption.Data());    
}



void TLinearBayesian::checktdsParContent()
{
    setMethodName("LinearBayesian");

    _aPrioriMode.clear();
    TIter next(_listOfParameters);
    URANIE::DataServer::TAttribute *att=NULL;
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    URANIE::DataServer::TUniformDistribution *uniatt=NULL;
    URANIE::DataServer::TNormalDistribution *normatt=NULL;
    try
    {
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
            if(stoatt==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TLinearBayesian::constructor\n\
 The provided dataserver must contains stochastic attributes. Attribute [%s] is not one of them", att->GetName()));


            uniatt = dynamic_cast<URANIE::DataServer::TUniformDistribution *>(att);
            normatt = dynamic_cast<URANIE::DataServer::TNormalDistribution *>(att);         
            if( normatt==0 && uniatt==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TLinearBayesian::constructor\n\
 The provided dataserver must contains uniform (jeffreys) or gaussian prior. Attribute [%s] is not one of them", att->GetName()));            

            if( normatt != 0)
                _aPrioriMode.push_back( normatt->getTheoreticalMode() );
            else if( uniatt != 0)
                _aPrioriMode.push_back( uniatt->getTheoreticalMode() );

            if (_blog)
                cout << " ** Storing " << att->GetName() << "] in our attribute vector"<<endl;
        }
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TLinearBayesian::~TLinearBayesian()
{
    if (_blog)
        cout << " *** destroy TLinearBayesian(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    
}

// -----------------------------------------------------------------------
//  Setters
// -----------------------------------------------------------------------

void TLinearBayesian::setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(distFunc);
    TCalibration::setDistanceAndReference("mahalanobis", tdsRef, input, output, weight);
}

void TLinearBayesian::setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(funcName);
    TCalibration::setDistanceAndReference("mahalanobis", tdsRef, input, output, weight);
}


// -----------------------------------------------------------------------
//  computeParameters
// -----------------------------------------------------------------------

void TLinearBayesian::computeParameters(Option_t *option)
{    

    try
    {
        if( _mObsCovMat.GetNrows() == 0 && _mObsCovMat.GetNcols() == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TLinearBayesian::computeParameters \n\
There is no observation covariance matrix provided which is compulsory to go further\n \
The method setObservationCovarianceMatrix should be called before doing the estimations");

        if( _mH.GetNcols() == 0 && _mH.GetNrows() == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TLinearBayesian::computeParameters \n\
There is no regressor information in the internal matrix. \n \
The method setRegressor should be called before doing the estimations");
        
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    _mParCovariance.ResizeTo(_nPar,_nPar);
    _mParValues.ResizeTo(_nPar,_nVar);

    double pouet=0;
    _dFunc->eval(&_aPrioriMode[0], &pouet);
    
    // Computation of the posterior mean adn variance
    TMatrixD HtOmegaH(_mH*_mObsCovMat, TMatrixD::kMultTranspose, _mH); // (nPar, nObs) * (nObs, nObs) * (nObs, nPar)
    _mParCovariance = HtOmegaH.Invert();
    
    TMatrixD InvHtOmegaH_Ht(_mParCovariance, TMatrixD::kMult, _mH); // (nPar, nObs)
    TMatrixD myObs = _dFunc->getObservationMatrix();
    myObs.T();
    _mParValues = InvHtOmegaH_Ht * _mObsCovMat * myObs; // (nPar, nObs) * (nObs, nObs) * (nObs, nVar)
 
    if(_blog)
    {
        cout<<"***  TLinearBayesian::computeParameters("<<option<<"); Parameters estimated are"<<endl;
        _mParValues.Print();
        cout<<"***  TLinearBayesian::computeParameters("<<option<<"); Parameters covariances estimated are"<<endl;
        _mParCovariance.Print();
    }

    // If transformation is needed
    _mTransfoParValues.ResizeTo(_mParValues);
    if(_fTransfoParam==NULL)
        _mTransfoParValues.SetMatrixArray(_mParValues.GetMatrixArray());
    else 
        (*_fTransfoParam)(_mParValues.GetMatrixArray(), _mTransfoParValues.GetMatrixArray());

    if(_fTransfoParam!=NULL)
    {
        cout<<"***  TLinearBayesian::computeParameters("<<option<<"); Transformed parameters estimated are"<<endl;
        _mTransfoParValues.Print();
    }
    
    vector<double> aPostMode;
    aPostMode.clear();

    // Prepare the data to be filled
    int bias=0;
    if( !_tdsPar->hasTuple() )
        _tdsPar->createTuple();
    else
        bias=_tdsPar->getNPatterns();
    vector<double> tobefilled={1.0+bias};
    
    for(int ipar=0; ipar<_nPar; ipar++)
    {
        URANIE::DataServer::TNormalDistribution buff( (string(_listOfParameters->At(ipar)->GetName())+"_aposteriori").c_str(), _mTransfoParValues(ipar,0), sqrt(_mParCovariance(ipar,ipar))) ;
        aPostMode.push_back( buff.getTheoreticalMode() );
        tobefilled.push_back( _mTransfoParValues(ipar,0) );
    }

    _dFunc->eval(&aPostMode[0], &pouet);
    tobefilled.push_back(pouet);

    // Save the aposteriori results    
    for(unsigned int ii=0; ii<_vrefName.size(); ii++)
        _tdsEval->getTuple()->Merger(_tdsObs->getTuple(), ( _vrefName.at(ii)+"_aposteriori").c_str(), _voutName.at(ii).c_str() );

    //Filling in tdsPar
    _tdsPar->getTuple()->Fill(&tobefilled[0]);
}


void TLinearBayesian::setRegressorName(const char *regName)
{

    _regname=regName;   

    URANIE::DataServer::UEntry::StringParser(_regname, ":", _vRegName);

    try
    {

        if(_tdsObs==NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TLinearBayesian::setRegressorName \n\
 This method should not be called before the setDistanceAndReference one as no data have been provided yet.");
        
        if( _nPar != int(_vRegName.size()) )  
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TLinearBayesian::setRegressorName \n\
 The provided regressor list has wrong number of variables [%lu]. It should be the number of parameter to estimate [%d]", _vRegName.size(), _nPar));

        for(unsigned int it=0; it<_vRegName.size(); it++)
            if( !_tdsObs->isAttribute( _vRegName.at(it).c_str() ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TLinearBayesian::setRegressorName \n\
 The provided regressor list contains a name that is not an attribute of the observation dataserver: [%s]", _vRegName.at(it).c_str()));
        
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
       
    _mH.ResizeTo(_nObs, _nPar);
    _tdsObs->getTuple()->extractData(&_mH, _regname.c_str());
    _mH.T();
}

void TLinearBayesian::drawParameters(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    UNUSED(select);

    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TLinearBayesian::drawParameters for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    bool transfo=false;
    if (soption.Contains("transformed"))
    {
        transfo=true;
        cleanoption.ReplaceAll("transformed","");
    }
    checkCanvasCreation(newcan);

    int iter=0;
    
    bool apriori=false;
    if (soption.Contains("apriori"))
    {
        apriori=true;
        iter++;
        cleanoption.ReplaceAll("apriori","");
    }
    
    bool aposteriori=false;
    if (soption.Contains("aposteriori"))
    {
        aposteriori=true;
        iter++;
        cleanoption.ReplaceAll("aposteriori","");
    }

    if(iter==0)
    {
        iter=2;
        apriori=true;
        aposteriori=true;
    }
    
    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" || svariable=="")
        for(int ivar=0; ivar<_nPar; ivar++)
            variables.push_back( string( _listOfParameters->At(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if( !soption.Contains("same") )
    {
        if (soption.Contains("vertical"))
        {
            _canvas->Divide(1, variables.size());
            cleanoption.ReplaceAll("vertical","");
        }
        else
            _canvas->Divide(variables.size(), 1);
    }
        
    double min=1e15, max=-1e15;    
    double mean=0, std=0;
    URANIE::DataServer::TAttribute *att=NULL;
    URANIE::DataServer::TUniformDistribution *uniatt=NULL;
    URANIE::DataServer::TNormalDistribution *normatt=NULL;
    TF1 *pFunc;
    cleanoption.ReplaceAll(",","");
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        if( !soption.Contains("same") )
            _canvas->cd(ivar+1);

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());
        uniatt=NULL; normatt=NULL;

        for(int iiter=0; iiter<iter; iiter++)
        {
            bool runningapriori=false;
            if( (iter==1 && aposteriori) || (iiter==0 && iter==2) )
            {
                if(transfo)
                    mean=_mTransfoParValues( _listOfParameters->IndexOf(att), 0 );
                else
                    mean=_mParValues( _listOfParameters->IndexOf(att), 0 );

                std = sqrt( _mParCovariance( _listOfParameters->IndexOf(att), _listOfParameters->IndexOf(att) ) );                    
            }
            else if( (iter==1 && apriori) || (iiter==1 && iter==2) )
            {
                runningapriori=true;
                uniatt = dynamic_cast<URANIE::DataServer::TUniformDistribution *>(att);
                normatt = dynamic_cast<URANIE::DataServer::TNormalDistribution *>(att);
                if(uniatt!=NULL)
                {
                    mean=att->getLowerBound();
                    std=att->getUpperBound();
                }
                if(normatt!=NULL)
                {
                    mean=normatt->getParameterMu();
                    std=normatt->getParameterSigma();
                }
            }            

            if(!(uniatt!=NULL && runningapriori))
            {
                min = ((att->isLowerBounded() && mean > att->getLowerBound() ) ? att->getLowerBound() : mean - 3 * std );
                max = ((att->isUpperBounded() && mean < att->getUpperBound() ) ? att->getUpperBound() : mean + 3 * std );
            }
            else
            {
                min = att->getLowerBound();
                max = att->getUpperBound();
            }

            if(uniatt!=NULL && runningapriori)
                _drawingGarbageCollector->Add( new TF1( ("f"+variables.at(ivar)+"_apriori").c_str(),Form("1./(%g-%g)",std,mean), min, max) );
            else if(runningapriori)
                _drawingGarbageCollector->Add( new TF1( ("f"+variables.at(ivar)+"_apriori").c_str(),"gaus", min, max) );
            else
                _drawingGarbageCollector->Add( new TF1( ("f"+variables.at(ivar)+"_aposteriori").c_str(),"gaus", min, max) );
            pFunc=(TF1*)_drawingGarbageCollector->Last();

            if(runningapriori)
                pFunc->SetLineColor(4);
            
            if(uniatt==NULL || !runningapriori)
                pFunc->SetParameters(1./(sqrt(2*TMath::Pi())*std), mean, std);
            
            if(_blog)
                XYZ<<"Param of the gaussian : "<<1./(sqrt(2*TMath::Pi())*std) << "  " << mean << "  "<< std << "  option [" << cleanoption.Data() << "]" << endl;
            pFunc->SetNpx(1000); 
            pFunc->Draw( (cleanoption+((iiter==1)?"same":"")).Data());
            
            if( !soption.Contains("same"))
            {
                pFunc->SetTitle( ((sTitre=="") ? ("f"+variables.at(ivar)+"_aposteriori").c_str() : sTitre.Data() ) );                            
                pFunc->GetXaxis()->SetTitle(att->getTitle());
            }
        }
    }
}


void TLinearBayesian::computePredictionVariance(URANIE::DataServer::TDataServer *tds_new, string outname)
{

    try
    {
        if(tds_new==NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TLinearBayesian::computePredictionVariance \n\
 The provided dataserver is null.");
        
        for(unsigned int it=0; it<_vRegName.size(); it++)
            if( !tds_new->isAttribute( _vRegName.at(it).c_str() ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TLinearBayesian::computePredictionVariance \n\
 The provided regressor list contains a name that is not an attribute of the newly provided dataserver: [%s]", _vRegName.at(it).c_str()));

        if( tds_new->isAttribute( outname.c_str() ) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TLinearBayesian::computePredictionVariance \n\
 The provided output name [%s] is already an attribute of the newly provided dataserver: [%s]", outname.c_str(), tds_new->GetName()));
        
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    TMatrixD m_localH;
    int size =  tds_new->getNPatterns();
    m_localH.ResizeTo(size, _nPar);
    tds_new->getTuple()->extractData(&m_localH, _regname.c_str());

    // Computation of the prediction variance
    TMatrixD HOmegaHt(m_localH*_mParCovariance, TMatrixD::kMultTranspose, m_localH); // (nPar, nObs) * (nObs, nObs) * (nObs, nPar)
    vector<double> variances(0., size);
    for(int i=0; i<size; i++)
        variances.push_back(HOmegaHt(i,i));

    tds_new->addAttributeUsingData(outname, &variances.at(0), size);
}


void TLinearBayesian::printLog(Option_t *option)
{
    cout << " ** TLinearBayesian::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **   ID[" << getID() << "]" << endl;
    cout << " ** _nSam[" << _nSam << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " **  Log of DataServer (unactive)" << endl;
    if ( 0 ) {
        _tdsPar->printLog();
        cout << " **   End of Log of DataServer" << endl;
    }
    cout << " ** End of TLinearBayesian::printLog[" << option << "]" << endl;
}

} // End of namespace Calibration
} // End of namespace URANIE

