//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TMinimisation
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMinimisation.h
 \author Jean-Baptiste Blanchard
 \date   2020/07/17

 \brief Interface of class URANIE::Calibration::TMinimisation

 */
/**

 * @class URANIE::Calibration::TMinimisation
 *
 * @brief Description of the class TMinimisation.
 *

 *
 */
#ifndef TMINIMISATION_H
#define TMINIMISATION_H

#include <iostream>
#include <sstream>
using std::cout;
using std::endl;

// Uranie
#include "UExceptions.h"
#include "TCalibration.h"

#include "TOptimShare.h"

#ifdef WITH_NLOPT
#include "TNlopt.h"
#endif

#ifdef WITH_VIZIR
#include "TVizir2.h"
#include "TVizirSolver.h"
#endif

namespace URANIE {

namespace Calibration {

class TMinimisation: public URANIE::Calibration::TCalibration {
    
    //---------------------------------------------
    /** @name Public attributes
     *
     **/
    //@{
public:

#ifdef WITH_NLOPT    
    URANIE::Reoptimizer::TNloptSolver *_solv;         ///< Solver for local optimisation
#endif
#ifdef WITH_VIZIR
    URANIE::Reoptimizer::TVizirSolver *_solviz;       ///< Solver for vizir optimisation
#endif
    URANIE::Relauncher::TRun *_optimRun;              ///< Optimisation runner
    URANIE::Reoptimizer::TOptimShare *_optimMaster;        ///< Optimisation runner
        
    //@}
    
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    /*!
      Constructor usual where nS is set to be the number of sample produced by the algorithm.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param run: the runner that would be used to produce evaluations 
      \param nS:  useless with this respect
    */    
    TMinimisation(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int nS=1, Option_t *option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TMinimisation(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t *option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TMinimisation(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t *option = "");

    /*! 
     * \brief Default Calibration constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */
    TMinimisation(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, Option_t *option = "");    

    //! Default destructor
    ~TMinimisation();

    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    URANIE::Reoptimizer::TOptimShare *getOptimMaster(){ return _optimMaster; }
    
#ifdef WITH_NLOPT
    void setOptimProperties(URANIE::Reoptimizer::TNloptSolver *solv, Option_t *option="");
#endif
#ifdef WITH_VIZIR
    void setOptimProperties(URANIE::Reoptimizer::TVizirSolverShare *solv, Option_t *option="");
#endif
    //@}

    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read the possible options
    /*!
     */
    virtual void parseOption(Option_t *option = "");
    //@}

    /*!
     * \brief Estimates the index from a Specific TDataServer
     */
    // virtual void estimateIndexes(Option_t * option = "") = 0;

    //void setReference(URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight="");
    void computeParameters(Option_t *option="");
    /*!
     * Check the content of the input dataserver
     */ 
    void checktdsParContent(){}
    void printLog(Option_t *option="");
    //@}

    ClassDef(URANIE::Calibration::TMinimisation, ID_CALIBRATION)
    
};

} // Fin du namespace Calibration
} // Fin du namespace URANIE

#endif
