//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TRejectionABC.h
 \author Rudy Chocat
 \date   2020/07/17

 \brief Function to calibrate with rejection abc 
 */

#ifndef __REJECTION_ABC__
#define __REJECTION_ABC__

// Uranie
#include "TABC.h"

namespace URANIE
{
namespace Calibration
{
    
class TRejectionABC : public URANIE::Calibration::TABC
{

public :

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    
    /*! 
     * \brief Default ABC constructor with the runner argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param run : the runner that contains the assessor to be used on the reference data
     * \param ns : number of sample to be generated (depending on the method)
     */
    TRejectionABC(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns = 1, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the function argument: it contains the assessor to be used   
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TRejectionABC(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),  const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TRejectionABC(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");
    
    /*! 
     * \brief Default ABC constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */    
    TRejectionABC(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, Option_t *option = "");
    
    
    //! Default destructor
    virtual ~TRejectionABC();
    //@}

    //---------------------------------------------
    /** @name Rejection ABC function
     **/
    
    //@{
    /*! 
     * \brief Generate the sample 
     * \param option: possble options are
     * \li "clean": remove internal branche from the tdsPar object at the end
     * \li "gd": another implementation of this algorithm
     */
    void computeParameters(Option_t *option = "");
    //@}
    
    
    ClassDef(URANIE::Calibration::TRejectionABC, ID_CALIBRATION)

};
    
} // Fin du namespace ABC
} // Fin du namespace URANIE
#endif
