//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTLinearBayesian.cxx
 \author Jean-Baptiste Blanchard
 \date   2020/08/27

 \brief  test file for Linear bayesian classes

 */

// ROOT
#include "TROOT.h"
#include "TSystem.h"

// URANIE
#include "UExceptions.h"
#include "TestTLinearBayesian.h"
#include "TTriangularDistribution.h"
#include "TestUserFunctions.h"
#include "TLauncher2.h"
#include "code_paths.h"
#include "uranie_commands.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_LARGE 1.e-2
#define Flow1DFile "Ex2DoE_n100_sd1.75.dat"
#define DEL(x) if(x!=NULL){delete x; x=NULL;} 

// Constructor Destructor
TLinearBayesianTest::TLinearBayesianTest():
    _tds(NULL),_tdsRef(NULL), _tlb(NULL), _seqRun(NULL), _thrRun(NULL),
    _evalJit(NULL), _evalCode(NULL), _code(NULL), _outRow(NULL), _inp(NULL), _outp(NULL), _norm(NULL), _out(NULL)
{
    _workingTestPath = gSystem->pwd();
}

TLinearBayesianTest::~TLinearBayesianTest()
{
    tearDown();
}


void TLinearBayesianTest::cleanWorkingDirectories()
{
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Ex2DoE_n100_sd1.root");
#ifndef WIN32
    UranieTests::removeFolder(_workingTestPath.Data(), "URA_*");
    UranieTests::removeFolder(_workingTestPath.Data(), "URANIE*");
#endif

}

// pre and post treatment
void TLinearBayesianTest::setUp()
{
    _tds = new TDataServer("tds","pouet");
    _norm = new TNormalDistribution("norm", 0., 1.0);
    
    _out=new TAttribute("out");

    _tdsRef = new TDataServer("tdsRef","pouet");
}

void TLinearBayesianTest::setUpException()
{
    _evalJit = new TCJitEval(EmptyFunction);
    _evalJit->addInput(_norm);
    _evalJit->addOutput(_out);

    _seqRun = new TSequentialRun(_evalJit);
}

void TLinearBayesianTest::tearDown()
{  
    DEL(_evalJit);
    DEL(_evalCode);
    DEL(_code);
    DEL(_outRow);
    DEL(_inp);
    DEL(_outp);
    DEL(_seqRun);
    DEL(_thrRun);
    DEL(_tlb);
    DEL(_tdsRef);
    DEL(_tds);
    DEL(_norm);
    DEL(_out);

    cleanWorkingDirectories();
}

void TLinearBayesianTest::setUpFlow1D(string type)
{

    _tdsRef->fileDataRead(Flow1DFile);
    
    // define the parameters
    double mu=750.6, std=0.5;
    _tds->addAttribute( new URANIE::DataServer::TNormalDistribution("hl", mu, std) );
    double binf_search=749., bsup_search=751.;
    _tds->getAttribute("hl")->setBounds( binf_search, bsup_search);
    _tds->getAttribute( "hl" )->delShare();

    // Create interface to assessors
    TCJitEval Reg(flowrateModelnoH);
    Reg.addInput(_tdsRef->getAttribute("rw"));
    Reg.addInput(_tdsRef->getAttribute("l"));    
    Reg.addOutput(new TAttribute("H"));
    ((TAttribute*)Reg.getOutputList()->At(0))->delShare();

    TSequentialRun runnoH(&Reg);    
    runnoH.startSlave();
    if(runnoH.onMaster())
    {
        TLauncher2 l(_tdsRef, &runnoH);
        l.solverLoop();
        runnoH.stopSlave();
    }

    if(type=="TCJitEval")
    {
        // Create interface to assessors
        _evalJit= new TCJitEval(flowrateModel);
        // Add inputs in the correct order
        _evalJit->addInput(_tds->getAttribute("hl"));
        _evalJit->addInput(_tdsRef->getAttribute("rw"));
        _evalJit->addInput(_tdsRef->getAttribute("l"));
    
        //Add the output: the cost value 
        _evalJit->addOutput(_out);

        _seqRun = new TSequentialRun(_evalJit);    
    }
    else if(TString(type).Contains("TCodeEval"))
    {
        // Create interface to assessors
        _evalCode= new TCodeEval((UranieTests::code_path("flowrate") + " -s -k").data());

        // Add inputs in the correct orde
        _inp = new TKeyScript("flowrate_input_with_keys.in");
        _inp->addInput(_tds->getAttribute("hl"),"Hl");    
        _inp->addInput(_tdsRef->getAttribute("rw"),"Rw");
        _inp->addInput(_tdsRef->getAttribute("l"),"L");

        // Add inputs in the correct orde
        _outp = new TKeyResult("_output_flowrate_withKey_.dat");
        _outp->addOutput(_out, "yhat" );    
    
        //Add the output: the cost value 
        _evalCode->addInputFile(_inp);
        _evalCode->addOutputFile(_outp);

        if(type=="TCodeEvalPar")
            _thrRun = new TThreadedRun(_evalCode,4);
        else
            _seqRun = new TSequentialRun(_evalCode);  
    }
    else if(type=="TCode")
    { 
        TString sIn = TString("flowrate_input_with_keys.in");

        // Set the reference input file and the key for each input attributes
        _tdsRef->getAttribute("rw")->setFileKey(sIn, "Rw");
        _tdsRef->getAttribute("l")->setFileKey(sIn, "L");
        _tds->getAttribute("hl")->setFileKey(sIn, "Hl");
 
        // The output file of the code
        _outRow = new TOutputFileRow("_output_flowrate_withRow_.dat");
        // The attribute in the output file
        _outRow->addAttribute(_out);
    
        // Instanciation de mon code
#ifdef WIN32
        _code= new TCode(_tdsRef,"flowrate.exe -s -k" );
#else
        _code= new TCode(_tdsRef,Form("%s/%s", _workingTestPath.Data(),
                                      "../../../../launCHER/teST/coDE/flowREATE/flowrate -s -k"));
#endif 
        // on ajoute le fichier de sortie du code
        _code->addOutputFile( _outRow );
    }

}

void TLinearBayesianTest::transf(double *x, double *res)
{
    res[0] = 1050 - x[0];
}

void TLinearBayesianTest::runFlow1DCJitNoPlots()
{
    setUpFlow1D("TCJitEval");
   
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, _seqRun, 1, "");
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::runFlow1DNoPlotsCodeEvalPar()
{
    setUpFlow1D("TCodeEvalPar");
    
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, _thrRun, 1, "");
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::runFlow1DNoPlotsCodeEval()
{
    setUpFlow1D("TCodeEval");
    
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, _seqRun, 1, "");
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::runFlow1DNoPlots()
{
    setUpFlow1D("Function");
   
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    TString inFile="myflow.C";
    gROOT->LoadMacro( gSystem->PrependPathName(gSystem->pwd(), inFile) );
    
    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, "flowrateModel", "hl:rw:l","out", 1);
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::runFlow1DPointerFunctionNoPlots()
{
    setUpFlow1D("Function");
   
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, flowrateModel, "hl:rw:l","out", 1);
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::runFlow1DNoPlotsCode()
{
    setUpFlow1D("TCode");
   
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(100,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);
    
    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, _code, 1, "");
    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tlb->setObservationCovarianceMatrix(mat);
    _tlb->setRegressorName("H");
    _tlb->setParameterTransformationFunction(transf);
    _tlb->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.72621, data[0], DBL_PRECISION_FINE);
    
}

void TLinearBayesianTest::throwExceptionConstruction()
{
    setUpException();
    
    // no tds
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(NULL, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // tds without attributes
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    // wrong kind of attribute
    _tds->addAttribute("tutu");
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    _tds->delAttribute("tutu");
    // wrong kind of laws
    _tds->addAttribute(new TTriangularDistribution("triang",0., 1., 0.5) );
    _tds->getAttribute("triang")->delShare();
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    _tds->delAttribute("triang");
    _tds->addAttribute(_norm);

    delete _evalJit;
    delete _seqRun;
    _evalJit = new TCJitEval(EmptyFunction);
    _seqRun  = new TSequentialRun(_evalJit);
    // no input in eval
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    delete _evalJit;
    delete _seqRun;
    _evalJit = new TCJitEval(EmptyFunction);
    _evalJit->addInput(_norm);
    _seqRun  = new TSequentialRun(_evalJit);
    // no output in eval
    CPPUNIT_ASSERT_THROW(new TLinearBayesian(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);
            
}


void TLinearBayesianTest::throwExceptionRegressorsAndCovMat()
{
    setUpFlow1D("TCJitEval");
    
    double sd=_tdsRef->getValue("sd_eps",0);
    TMatrixD mat(101,100);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    TMatrixD mat2(100,101);
    for(int ival=0; ival<_tdsRef->getNPatterns(); ival++)
        mat(ival,ival)=(sd*sd);

    _tlb = new URANIE::Calibration::TLinearBayesian(_tds, _seqRun, 1, "");

    // No data so far
    CPPUNIT_ASSERT_THROW(_tlb->setRegressorName("x1:x2");, URANIE::Exceptions::UErrorExceptions);

    // No data so far
    CPPUNIT_ASSERT_THROW(_tlb->setObservationCovarianceMatrix(mat);, URANIE::Exceptions::UErrorExceptions);

    _tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    
    // Wrong dimensions
    CPPUNIT_ASSERT_THROW(_tlb->setRegressorName("sd_eps:Qexp");, URANIE::Exceptions::UErrorExceptions);

    // Wrong dimension
    CPPUNIT_ASSERT_THROW(_tlb->setObservationCovarianceMatrix(mat);, URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tlb->setObservationCovarianceMatrix(mat2);, URANIE::Exceptions::UErrorExceptions);

    // Wrong Name
    CPPUNIT_ASSERT_THROW(_tlb->setRegressorName("Prout");, URANIE::Exceptions::UErrorExceptions);
}

void TLinearBayesianTest::throwExceptionReference()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tlb=new TLinearBayesian(_tds, _seqRun, 10);

    // no tdsref
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("toutou",NULL,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    // no tdsref empty
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("toutou",_tdsRef,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    _tdsRef->fileDataRead(Flow1DFile);
    // wrong input
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("LS",_tdsRef,"rw:lout","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexpmero","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong weight
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference number
    CPPUNIT_ASSERT_THROW(_tlb->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp:sd_eps","sd_eps"), URANIE::Exceptions::UErrorExceptions);
    
}
