// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2019 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TEgoShareSolver.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2019-11-06
 
 \brief  recherche du point minimisant l'EI
*/


#include "TEgoShareSolver.h"

// ROOT

// Uranie

ClassImp(URANIE::MetaModelOptim::TEgoShareSolver);
#ifdef WITH_VIZIR
ClassImp(URANIE::MetaModelOptim::TEgoVzrSolver);
#endif
ClassImp(URANIE::MetaModelOptim::TEgoHjSolver);



/*** TEgoShareSolver ***/
// constructeur
URANIE::MetaModelOptim::TEgoShareSolver::TEgoShareSolver() :
    TEgoSolver(),
    _tds(NULL), _ego(NULL), _run(NULL), _maxim(), _first(1), _many(4),
    _valid_ei(-1.e-8)
{
}

URANIE::MetaModelOptim::TEgoShareSolver::~TEgoShareSolver()
{
    if (_tds) delete _tds;
}

// public
void URANIE::MetaModelOptim::TEgoShareSolver::significantEI(double vmin)
{
    _valid_ei = -vmin;
}

void URANIE::MetaModelOptim::TEgoShareSolver::setManyNewItem(int sz)
{
    if (sz > 0 && sz < 10 ) _many = sz;
}

void URANIE::MetaModelOptim::TEgoShareSolver::build(URANIE::Relauncher::TRun *run,
    vector<URANIE::DataServer::TAttribute*>& in, URANIE::DataServer::TAttribute *ego)
{
    _run =  run;
    _ego = ego;
    _in = in;
    for (_NinDat=0; in[_NinDat]; _NinDat++) ;
}

void URANIE::MetaModelOptim::TEgoShareSolver::next()
{
    if (_first) _first = 0;
    else {
        _tds->createTuple();
        fitnessChanged();
        //extend();
    }
    
    solverLoop();
}

void URANIE::MetaModelOptim::TEgoShareSolver::more()
{
    extend();
    solverLoop();
}

#ifdef WITH_VIZIR
/*** TEgoVzrSolver ***/
// constructeur
URANIE::MetaModelOptim::TEgoVzrSolver::TEgoVzrSolver() :
    TEgoShareSolver(),
    _algo(NULL), _viz(NULL), _signif(1.0e-10), _NG(10000)
{
}

URANIE::MetaModelOptim::TEgoVzrSolver::~TEgoVzrSolver()
{
    if (_viz) delete _viz;
    //if (_algo) delete _algo;
}

// param
void URANIE::MetaModelOptim::TEgoVzrSolver::setSolver(URANIE::Reoptimizer::TVizirSolverShare *algo)
{
    _algo = algo;
}

void URANIE::MetaModelOptim::TEgoVzrSolver::setStepSize(int size)
{
    _NG = size;
}

void URANIE::MetaModelOptim::TEgoVzrSolver::solverLoop()
{
    _viz->solverLoop();
}
#endif

/*** TEgoHjSolver ***/
// constructeur
URANIE::MetaModelOptim::TEgoHjSolver::TEgoHjSolver() :
    TEgoShareSolver(),
    _algo(NULL)
{
    // default
    _Nfirst = 64; 
    _Nstep = 11;
}

URANIE::MetaModelOptim::TEgoHjSolver::~TEgoHjSolver()
{
    if (_algo) delete _algo;
}

// setup
void URANIE::MetaModelOptim::TEgoHjSolver::setSize(int first, int follow)
{
    _Nfirst = first; _Nstep = follow;
}


// helpers
int URANIE::MetaModelOptim::TEgoHjSolver::getNext(long ev, int free, 
    vector<vector<URANIE::DataServer::UEntry*>> &sampl, vector<char> &mark)
{
    int i, j, siz, ret;
    double *tmpval;
    vector<double*> memval;
    
    siz = (free < _many) ? free+1 : _many ;
    memval.resize(siz);
    std::cout << "EGO need " << siz << std::endl;
    // remplissage
    for (i = 0, ret = 0, tmpval = _algo->item(i);
        tmpval && ret < siz;
        i ++, tmpval = _algo->item(i))
    {
        if (tmpval[_NinDat] > _valid_ei) {
            std::cout << tmpval[_NinDat] << "~";
            continue;
        }
        for (j = ret;
            j > 0 && tmpval[_NinDat] < memval[j-1][_NinDat];
            j--) 
        {
            memval[j] = memval[j-1];   
        }
        memval[j] = tmpval;
        ret ++;
    }
    
    // n meilleur
    for( ; tmpval; i++, tmpval = _algo->item(i)) {
        if (tmpval[_NinDat] < memval[ret-1][_NinDat]) {
            // keep
            for (j = ret-2; 
                j >= 0 && tmpval[_NinDat] < memval[j][_NinDat];
                j--) 
            {
                memval[j+1] = memval[j];   
            }
            memval[j+1] = tmpval;
        }
    }
    
    std::cout << "\nselect " << ret << " from " << i << std::endl;
    for (j=0; j<ret; j++, ev++) {
        
        std::cout << "EGO expect " << ev << " " << -memval[j][_NinDat] << std::endl;
        std::cout << "EGO with ";
        //copie
        mark[ev] = FILL;
        *sampl[ev][0]->val.dou = ev ;
        for (i=0; i<_NinDat; i++) {
            std::cout << " " << memval[j][i];
            *sampl[ev][i+1]->val.dou = memval[j][i];
        }
        std::cout << std::endl;
    }
    
    return ret;
}

void URANIE::MetaModelOptim::TEgoHjSolver::solverLoop()
{
    _algo->solverLoop();
}
