

#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

/* Calcul des erreurs par LOO                           */
/* Methode gpLoo1                                       */
/* Input Kin : matrice de variance                      */
/* Input yin : vecteur des observations                 */
/* Input n : dimension Kin[n x n], yin[n]               */
/* Output errorLoo : n erreurs par LOO			*/
/* Output varLoo : n variances des erreurs		*/
/* Output eqmLoo   : erreur quadratique moyenne         */
libpgauss_EXPORT int gpLoo1Error(double *Kin, double *yin, int n, double *errorLoo, double *varLoo, double *eqmLoo) {
  void gpDposl ( double a[], int lda, int n, double b[] );
  int i,j,info;
  double *v, vty, sv, dv, determinant[2], dT, xk, eqm;

  /* Recopie des donnees qui seront modifiees */
  double *K, *pK, *y, *py;
  K = calloc(sizeof(double),n*n);
  y = calloc(sizeof(double),n);
  for(pK = K,i=0;i<n*n;i++) *pK++ = *Kin++;
  for(py = y,i=0;i<n;i++)   *py++ = *yin++;
  /*******************************/

  v    = calloc(sizeof(double),n);
  info = dpofa (K, n, n);    /* decomposition Cholesky */
  if (info == 0) {
    for(i=0;i<n;i++)
	v[i]=1.;

    dposl(K,n,n,v);
    for(vty=0.,sv=0.,i=0;i<n;i++) {vty += v[i]*y[i], sv += v[i];} dv = vty/sv;
    dposl(K,n,n,y);
    for(i=0;i<n;i++) y[i] -= v[i] * dv;
    dtrdi(K,n,n,determinant,11); /* inverse d'une triangulaire superieure inv(L) */
    for(i=0;i<n;i++) {
      dT = - v[i] * v[i] / sv;
      for(j=i;j<n;j++) {
	xk = K[i+j*n];
	dT += xk * xk;
      }
      errorLoo[i] = y[i] / dT;
      varLoo[i]   = 1. / dT;
    }
    for(eqm=0., i=0;i<n;i++) eqm += errorLoo[i] * errorLoo[i];
    eqmLoo[0] = eqm / n;
  }
  else {
    /* printf("gpLoo1 : The leading minor %d is not positive\n", info); */
    for(i=0;i<n;i++) errorLoo[i]=0.;
    for(i=0;i<n;i++) varLoo[i]=0.;
    eqmLoo[0] = -1;
  }
  free(v); v=NULL;
  free(K);  K=NULL;
  free(y);  y=NULL;
  return info;
}

