/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2023 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpML1.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include <math.h>
# include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul du moins log de la vraisemblance              */
/* Methode gpML1                                        */
/* Input Kin : matrice de variance (array double Fortran) */
/* Input yin : vecteur des observations (array double)    */
/* Input n : dimension K[n x n], y[n]                   */
/* cas = 0 general sinon variance en facteur            */
/* Output ml[2] : critere + variance (si cas != 0)    */
/* Si ml[1] = -1 => echec lors decomposition Cholesky    */
libpgauss_EXPORT int gpML1(double *Kin, double *yin, int n, double *ml, int cas)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int i, info;
    double v1, v2, v3, v4, *z;

    /* Recopie des donnees qui seront modifiees */
    double *K, *pK, *y, *py;
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = *yin++;
    /*******************************/

    info = dpofa(K, n, n); /* Decomposition Choleski K = Lt L*/
    if (info == 0)
    {
        gpDposl(K, n, n, y); /* Resolution y = Lt x */
        z = calloc(sizeof(double), n);
        for (i = 0; i < n; i++)
            z[i] = 1.;
        gpDposl(K, n, n, z); /* Resolution z = 1 = Lt x */
        v1 = 0.;
        v2 = 0.;
        v3 = 0.;
        v4 = 0.;
        for (i = 0; i < n; i++)
        {
            v1 += y[i] * y[i];
            v2 += z[i] * z[i];
            v3 += y[i] * z[i];
            v4 += log(K[i + i * n]);
        }
        free(z);
        if (cas == 0)
        { /* general */
            ml[0] = v1 - v3 * v3 / v2 + 2. * v4;
            ml[1] = 0.;
        }
        else
        { /* variance en facteur */
            ml[0] = n * log(v1 - v3 * v3 / v2) + 2. * v4;
            ml[1] = (v1 - v3 * v3 / v2) / n;
        }
    }
    else
    {
        /* printf("gpML1 : The leading minor %d is not positive\n", info); */
        ml[0] = 0.;
        ml[1] = -1.; /* Cholesky fails */
    }
    free(K);        K = NULL;
    free(y);        y = NULL;
    return info;
}

