/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2023 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpModh.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/*
 VERIFIE 25/06/2012
 */
/* Calcul de hbeta,vbeta,gamma,iL,iR,M                */
/* Methode gpModh                                     */
/* Input Hin : matrice des regresseurs pxn            */
/* Input Kin : matrice de variance                    */
/* Input yin : vecteur des observations               */
/* Input n : nombre des observations                  */
/* Input p : nombre de regresseurs                    */
/* Output hbeta : vecteur p estimation de beta        */
/* Output vbeta : matrice de variance pxp de hbeta    */
/* Output gamma : vecteur n                  */
/* Output iL   : inverse de L                         */
/* Output iR   : matrice inverse de R p x p           */
/* Output M    : matrice n x p                        */
libpgauss_EXPORT int gpModh(double *Hin, double *Kin, double *yin, int n, int p, double *hbeta,
        double *vbeta, double *gamma, double *iL, double *iR, double *M)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int i, j, k, ijp, ijn, info, bidon_jpvt;
    double bidon, *qraux, *qty, determinant[2];

    /* Recopie des donnees qui seront modifiees */
    double *H, *pH, *K, *pK, *y, *py;
    H = calloc(sizeof(double), n * p);
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pH = H, i = 0; i < n * p; i++)
        *pH++ = Hin[i];
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = yin[i];
    /*******************************/

    qraux = calloc(sizeof(double), p);
    qty = calloc(sizeof(double), n);

    info = dpofa(K, n, n); /* decomposition Cholesky K = Lt L*/
    if (info == 0)
    {
        gpDposl(K, n, n, y); /* Resolution y = Lt w */
        /* Calcul de M */
        for (j = 0; j < p; j++)
            gpDposl(K, n, n, &(H[j * n])); /* Resolution H = Lt M */
        for (j = 0; j < n * p; j++)
            M[j] = H[j]; /* Sauvegarde de M */
        dqrdc(H, n, n, p, qraux, &bidon_jpvt, 0, 0); /* Decomposition QR */
        /* Recuperation de R et calcule de son inverse */
        for (i = 0; i < p; i++)
        {
            for (j = i; j < p; j++)
                iR[i + j * p] = H[i + j * n];
            for (j = 0; j < i; j++)
                iR[i + j * p] = 0.;
        }
        dtrdi(iR, p, p, determinant, 11); /* inverse d'une triangulaire superieure */
        dqrsl(H, n, n, p, qraux, y, &bidon, qty, &bidon, &bidon, &bidon, 1000); /* Resolution qty = Qt w */
        /* calcul de hbeta VERIFIE */
        for (i = 0; i < p; i++)
        {
            hbeta[i] = 0.;
            for (j = 0; j < p; j++)
                hbeta[i] += iR[i + j * p] * qty[j];
        }
        /* Calcul de vbeta = iR * iRt VERIFIE*/
        for (i = 0; i < p; i++)
        {
            for (j = i; j < p; j++)
            {
                ijp = i + j * p;
                vbeta[ijp] = 0;
                for (k = 0; k < p; k++)
                    vbeta[ijp] += iR[i + k * p] * iR[j + k * p];
                vbeta[j + i * p] = vbeta[ijp];
            }
        }
        /* Calcul de gamma VERIFIE */
        for (i = 0; i < n; i++)
        {
            gamma[i] = yin[i];
            for (j = 0; j < p; j++)
                gamma[i] -= Hin[i + j * n] * hbeta[j];
        }
        dposl(K, n, n, gamma); /* resolution lineaire */
        /* Calcul de iL */
        dtrdi(K, n, n, determinant, 11); /* inverse de L : inverse d'une triangulaire superieure */
        for (i = 0; i < n; i++)
        {
            for (j = 0; j < i; j++)
                iL[i + j * n] = 0.;
            for (j = i; j < n; j++)
            {
                ijn = i + j * n;
                iL[ijn] = K[ijn];
            }
        }
    }
    free(qraux);    qraux = NULL;
    free(qty);      qty = NULL;
    free(H);        H = NULL;
    free(K);        K = NULL;
    free(y);        y = NULL;
    return info;
}
