/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2023 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpModhBayes.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul de beta_post,Qpost,gamma,iL,iW,M            */
/* Methode gpModhBayes                                */
/* Input beta_prior : moyenne a priori              */
/* Input Qprior : matrice de covariances a priori     */
/* Input Hin : matrice des regresseurs pxn            */
/* Input Kin : matrice de variance                    */
/* Input yin : vecteur des observations               */
/* Input n : nombre des observations                  */
/* Input p : nombre de regresseurs                    */
/* Output beta_post : moyenne a posterior de beta     */
/* Output Qpost : variance pxp a posteriori de beta   */
/* Output gamma : vecteur n                  */
/* Output iL   : inverse de L                         */
/* Output iW   : matrice inverse de W p x p           */
/* Output M    : matrice n x p                        */
libpgauss_EXPORT int gpModhBayes(double *beta_prior, double *Qprior, double *Hin, double *Kin,
        double *yin, int n, int p, double *beta_post, double *Qpost,
        double *gamma, double *iL, double *iW, double *M)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int gpInvCholesky(double A[], int n, double iA[]);
    int i, j, k, ijp, ijn, kp, jp, in, info;
    double *qraux, *Mtw, determinant[2], *w, *iQprior, vt1;

    /* Recopie des donnees qui seront modifiees */
    double *H, *pH, *K, *pK, *y, *py;
    H = calloc(sizeof(double), n * p);
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pH = H, i = 0; i < n * p; i++)
        *pH++ = Hin[i];
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = yin[i];
    /*******************************/

    qraux = calloc(sizeof(double), p);
    Mtw = calloc(sizeof(double), p);
    w = calloc(sizeof(double), n);

    /* Calcul de l'inverse de la matrice de covariance Qprior */
    iQprior = calloc(sizeof(double), p * p);
    info = gpInvCholesky(Qprior, p, iQprior);
    if (info != 0)
        return (info); /* Qprior non definie positive */

    info = dpofa(K, n, n); /* decomposition Cholesky K = Lt L*/
    if (info == 0)
    {
        /* calcul de y - H beta_prior */
        for (i = 0; i < n; i++)
        {
            w[i] = y[i];
            for (j = 0; j < p; j++)
                w[i] -= H[i + j * n] * beta_prior[j];
        }
        gpDposl(K, n, n, w); /* Resolution y = Lt w */
        /* Calcul de M */
        for (j = 0; j < p; j++)
            gpDposl(K, n, n, &(H[j * n])); /* Resolution H = Lt M */
        for (j = 0; j < n * p; j++)
            M[j] = H[j]; /* Sauvegarde de M */

        /* Calcul de Qpost */
        for (i = 0; i < p; i++)
        {
            in = i * n;
            for (j = 0; j < p; j++)
            {
                jp = j * p;
                vt1 = iQprior[i + jp];
                for (k = 0; k < n; k++)
                    vt1 += M[k + in] * M[k + j * n];
                Qpost[i + jp] = vt1;
            }
        }
        dpofa(Qpost, p, p); /* Decomposition Cholesky */
        dtrdi(Qpost, p, p, determinant, 11); /* inverse de W : inverse d'une triangulaire superieure */
        for (i = 0; i < p; i++)
        {
            for (j = 0; j < i; j++)
                iW[i + j * p] = 0.;
            for (j = i; j < p; j++)
            {
                ijp = i + j * p;
                iW[ijp] = Qpost[ijp];
            }
        }
        for (i = 0; i < p; i++)
        {
            for (j = i; j < p; j++)
            {
                ijp = i + j * p;
                Qpost[ijp] = 0;
                for (k = 0; k < p; k++)
                {
                    kp = k * p;
                    Qpost[ijp] += iW[i + kp] * iW[j + kp];
                }
                Qpost[j + i * p] = Qpost[ijp];
            }
        }

        /* calcul de beta_post */
        for (i = 0; i < p; i++)
        { /* calcul de Mt w */
            Mtw[i] = 0.;
            in = i * n;
            for (j = 0; j < n; j++)
                Mtw[i] += M[j + in] * w[j];
        }
        for (i = 0; i < p; i++)
        {
            beta_post[i] = beta_prior[i];
            for (j = 0; j < p; j++)
                beta_post[i] += Qpost[i + j * p] * Mtw[j];
        }
        /* Calcul de gamma */
        for (i = 0; i < n; i++)
        {
            gamma[i] = yin[i];
            for (j = 0; j < p; j++)
                gamma[i] -= Hin[i + j * n] * beta_post[j];
        }
        dposl(K, n, n, gamma); /* resolution lineaire */
        /* Calcul de iL */
        dtrdi(K, n, n, determinant, 11); /* inverse de L : inverse d'une triangulaire superieure */
        for (i = 0; i < n; i++)
        {
            for (j = 0; j < i; j++)
                iL[i + j * n] = 0.;
            for (j = i; j < n; j++)
            {
                ijn = i + j * n;
                iL[ijn] = K[ijn];
            }
        }
    }
    free(qraux);    qraux = NULL;
    free(Mtw);      Mtw = NULL;
    free(w);        w = NULL;
    free(iQprior);  iQprior = NULL;
    free(H);        H = NULL;
    free(K);        K = NULL;
    free(y);        y = NULL;
    return info;
}
