/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2023 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   util.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include "linpack_d.h"
#include  <math.h>
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul de l'inverse d'une matrice symetrique definie positive */
/* gpInvCholesky */
/* Input K : matrice n x n    */
/* Input n : taille de K    */
/* Output iK : inverse de K    */
libpgauss_EXPORT int gpInvCholesky(double *K, int n, double *iK)
{
    int i, j, k, ijn, kn, info;
    double determinant[2];
    double * Kt ;
	Kt = calloc(sizeof(double), n * n);
    for (i = 0; i < n * n; i++)
        Kt[i] = K[i];

    info = dpofa(Kt, n, n);
    if (info == 0)
    {
        dtrdi(Kt, n, n, determinant, 11);
        for (i = 0; i < n; i++)
        {
            for (j = 0; j <= i; j++)
            {
                ijn = i + j * n;
                iK[ijn] = 0.;
                for (k = i; k < n; k++)
                {
                    kn = k * n;
                    iK[ijn] += Kt[i + kn] * Kt[j + kn];
                }
                iK[j + i * n] = iK[ijn];
            }
        }
    }
    free(Kt); Kt = NULL;
    return (info);
}

/* Calcul de la decomposition en valeurs singulieres    */
/* gpSvd                        */
/* Input H : matrice de taille n x p avec n > p        */
/* Input n : nombre de lignes de H            */
/* Input p : nombre de colonnes de H            */
/* Output U : matrice U de taille n x p            */
/* Output S : valeurs singuli�res vecteur de taille n    */
/* Output V : matrice V de taille p x p            */
libpgauss_EXPORT int gpSvd(double *H, int n, int p, double *U, double *S, double *V)
{

    double *e;
    double *work;

	
    /*
     job = 100 : calcul de S uniquement
     job = 110 : calcul de S et U n x n
     job = 120 : calcul de S et de U n x p
     job = 111 : calcul de S et U n x n et V
     job = 121 : calcul de S et U n x p et V
     */
	int job, info;
	
	e =(double*) calloc(p,sizeof(int));
	work =(double*) calloc(n,sizeof(int));
	
    job = 121;
    info = dsvdc(H, n, n, p, S, e, U, n, V, p, work, job);

	free(e);     e = NULL;
	free(work);  work = NULL;

    return (info);
}

int prodMatAB(double A[], int n1, int p1, double B[], int n2, int p2,
        double C[])
{
    int i, j, k;
    if (p1 != n2)
        return (1);
    for (i = 0; i < n1; i++)
    {
        for (j = 0; j < p2; j++)
        {
            C[i + j * n1] = 0.;
            for (k = 0; k < p1; k++)
                C[i + j * n1] += A[i + k * n1] * B[k + j * n2];
        }
    }
    return (0);
}

int prodMatAtA(double A[], int n, int p, double C[])
{
    int i, j, k;
    for (i = 0; i < p; i++)
    {
        for (j = 0; j < p; j++)
        {
            C[i + j * p] = 0.;
            for (k = 0; k < n; k++)
                C[i + j * p] += A[k + i * n] * A[k + j * n];
        }
    }
    return (0);
}

/* Decomposition Cholesky
 A = L Lt
 L triangulaire inferieure retourn�e dans la partie inferieure de A
 p : termes diagonaux de L (A partie triangulaire sup non modifiee)
 */
void choldc(double A[], int n, double p[])
{
    int i, j, k, in, kn;
    double sum;
    for (i = 0; i < n; i++)
    {
        in = i * n;
        for (j = i; j < n; j++)
        {
            for (sum = A[i + j * n], k = 0; k < i; k++)
            {
                kn = k * n;
                sum -= A[i + kn] * A[j + kn];
            }
            if (i == j)
                p[i] = sqrt(sum);
            else
                A[j + in] = sum / p[i];
        }
    }
}

/*    Resolution A x = b */
void cholsl(double A[], int n, double p[], double b[], double x[])
{
    int i, k;
    double sum;
    for (i = 0; i < n; i++)
    { /* L y = b stocke dans x */
        for (sum = b[i], k = 0; k < i; k++)
            sum -= A[i + k * n] * x[k];
        x[i] = sum / p[i];
    }
    for (i = n - 1; i >= 0; i--)
    { /* Lt x = y */
        for (sum = x[i], k = i + 1; k < n; k++)
            sum -= A[k + i * n] * x[k];
        x[i] = sum / p[i];
    }
}

/* Calcul de iL triangulaire inferieure */
void cholin(double A[], int n, double p[])
{
    int i, j, k, in;
    double sum;
    for (i = 0; i < n; i++)
    {
        in = i * n;
        A[i + in] = 1.0 / p[i];
        for (j = i + 1; j < n; j++)
        {
            for (sum = 0.0, k = i; k < j; k++)
                sum -= A[j + k * n] * A[k + in];
            A[j + in] = sum / p[j];
        }
    }
}
