//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TANNModeler.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   tue jui 3 10:38:56 CEST 2007

 \brief Implementation of the class URANIE::Optimizer::TOptimizer

 */
/*
 \par Modifications
 2014-07-10 - ABN: added GPU training - integration of Clement Cannone's code

 2010-11-04 16:17:45 - FGA
 - Add the export in a fortran file
 - Add the Logistic function as activation function. The rules
 - Logistic when kCR and kZeroOne as normalization
 - TanH when kMinusOneOne as normalization
 - Add the export in a PMML File with rules as option
 - NEW : create a new PMML file
 - UPDATE : Write the current ANN in the PMML file and if already exist an ANN in the PMML file, erase it.
 - BACKUP : Write the current ANN in the PMML file and if already exist ANNs in the PMML file with modelname "name_I", rename the old with "name_I+1"
 - ADD : Add the current ANN in the PMML file

 2010-02-16 16:21:31 - FGA : Change the getEQM function
 return the saved values of _dLearn and _dtest after the selection ANN in the trainning method

 29/01/2010 09:03 - FGA : Change the getEQM function
 Now to compute the EQM of app/test, call the interactive method InteractiveTESTANN
 with the good number of parameter (first and size).

 11/01/2010 15:51 - FGA : selectModel function
 Add an Exception if the list of solution is empty.

 */

#include <sstream>
using namespace std;

// ROOT
#include "Rtypes.h"
#include "TMath.h"
#include "TLeaf.h"
#include "TROOT.h"
#include "TCanvas.h"
#include "TRandom3.h"
#include "TRegexp.h"

#ifdef WIN32
#pragma warning( disable : 4275 4355 )
#endif

// Uranie
#include "UVersion.h"
#include "TANNModeler.h"
//#include "UMLP.h"
#include "UExceptions.h"
#include "TAttributeFormula.h"
#include "UTimer.h"

// ROOT & XML
#include "TXMLEngine.h"
#ifdef  WITH_LIBXML2
#include "PMMLlib.hxx"
#endif

//OPT++
// #include "opt++.h"

#include "Opt.h"
#include "NLP.h"
#include "NLF.h"
#include "OptQNewton.h"
#include "OptNIPS.h"
#include "OptPDS.h"

#include "OptppExceptions.h"


URANIE::Modeler::TANNModeler *gUranieANNModeler;
//URANIE::Modeler::UMLP *gUranieUMLP;

using namespace OPTPP;

ClassImp (URANIE::Modeler::TANNModeler)

namespace URANIE
{
namespace Modeler
{

const int TANNModeler::OPTIM_MAX_ITER  = 10000;
const int TANNModeler::OPTIM_MAX_FEVAL = 50000;

class TempSolution
{
public:
    TempSolution():
        iter(-1), candidat(-1),
        eqmlearn(std::numeric_limits<double>::max()),
        eqmtest(std::numeric_limits<double>::max()),
        x(),
        lvoVirtuel(-1.0),
        mu(-1.0)
    {
    }
    ~TempSolution()
    {
    }
    int iter;
    int candidat;
    double eqmlearn;
    double eqmtest;
    ColumnVector x;
    double lvoVirtuel;
    double mu;
    //
    inline bool operator<(TempSolution vec) const
    {
        if (( (candidat == 1) && (vec.candidat == 1))
      || ((candidat == 0) && (vec.candidat == 0) ))
        {
            if (eqmtest < vec.eqmtest)
                return true;
            else
                return false;
        }
        else
        {
            if (candidat == 1)
                return true;
            else
                return false;
        }
    }

};

Int_t selectModel(vector<TempSolution> solutions)
{
    Int_t nsol = solutions.size();
    cout << endl << " ** solutions : " << nsol << endl;
    try
    {
        if (nsol == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TANNModeler::selectModel the list of solution is empty");

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    Double_t dMinTest = solutions[0].eqmtest;
    Int_t nMinTest = 0;
    cout << " ** isol[" << nMinTest + 1 << "] iter[" << solutions[nMinTest].iter
            << "] learn[" << solutions[nMinTest].eqmlearn << "] test["
            << solutions[nMinTest].eqmtest << "] *" << endl;
    ;
    for (Int_t isol = 1; isol < nsol; isol++)
    {
        cout << " ** isol[" << isol + 1 << "] iter[" << solutions[isol].iter
                << "] learn[" << solutions[isol].eqmlearn << "] test["
                << solutions[isol].eqmtest << "]";
        if (solutions[isol].eqmtest < dMinTest)
        {
            nMinTest = isol;
            dMinTest = solutions[nMinTest].eqmtest;
            cout << " *" << endl;
        }
        else
            cout << endl;
    }
    return nMinTest;
}

void InteractiveUPDATEMODEL(int, int, ColumnVector)
{
}

void InteractiveINITEQM(int ndim, ColumnVector& x)
{

    Bool_t blog = gUranieANNModeler->getLog();
    if (blog)
    {
        cout << " ** InteractiveINITEQM **" << endl;
        cout << " ** ndim[" << ndim << "]" << endl;
    }

    x = gUranieANNModeler->getWeights();

    if (blog)
    {
        for (int i = 1; i <= ndim; i++)
            cout << " ** i[" << i << "] x[" << x(i) << "]" << endl;
        cout << "** End of InteractiveINITEQM **" << endl;
    }
}

void InteractiveTESTANN(int ipatdeb, int ipatfin, const ColumnVector& weights, double& fx)
{
    Bool_t blog = gUranieANNModeler->getLog();
    if (1)
        cout << "** InteractiveTESTANN Beg[" << ipatdeb << "] End[" << ipatfin
                << "] **" << endl;

    Int_t ninp = gUranieANNModeler->getNInput();
    Int_t nhidden = gUranieANNModeler->getNHidden();
    Int_t noutput = gUranieANNModeler->getNOutput();
    if ( 1 ) {
        cout << " ** ninp[" << ninp << "] nhidden[" << nhidden << "] noutput[" << noutput << "]" << endl;
    }
    //Int_t napp = gUranieANNModeler->getNAPP();

    //Construct the current ANN
    UMLP* crtMLP = new UMLP(ninp, nhidden, noutput);
    crtMLP->setWeights(weights);

    if ( blog )
        crtMLP->setLog();

    // Activation functions on Hidden Layers
    switch (gUranieANNModeler->getNormalization()) {
    case TANNModeler::kMinusOneOne:
    {
        crtMLP->setActivationFunction(UMLP::kTANH);
        break;
    }
    default:
    {
        crtMLP->setActivationFunction(UMLP::kLOGISTIC);
        break;
    }
    }

    // Class of problem
    switch (gUranieANNModeler->getProblem()) {
    case TANNModeler::kRegression:
        crtMLP->setProblem(UMLP::kRegression);
        break;
    case TANNModeler::kClassification:
        crtMLP->setProblem(UMLP::kClassification);
    }

    for (Int_t i = 0; i <= ninp+noutput; i++) {
        crtMLP->setMinMaxValue(i + 1, gUranieANNModeler->getMinValue(i + 1), gUranieANNModeler->getMaxValue(i + 1));
    }

    if( gDirectory != gROOT ) gDirectory = gROOT;
    if( gROOT->FindObject("djdj") ) gROOT->Delete("djdj");
    TNtupleD * ntd = new TNtupleD("djdj", "ddhh", "iter:ecart:ss");
    ntd->SetDirectory( gROOT );
    ntd->SetMarkerColor(2);
    ntd->SetMarkerSize(4);
    ntd->SetMarkerStyle(7);
    Double_t *dVal = new Double_t[ninp + noutput];
    double eqm_crt = 0.0;
    double ecart;
    Int_t ncrt;
    for (int ipat = ipatdeb; ipat < ipatfin; ipat++)
    {
        gUranieANNModeler->setValue(ipat, ncrt, dVal);
        ecart = crtMLP->getEcart(dVal);
        eqm_crt += ecart;
        if (1)
        {
            //cout << "   *** dMinOut[" << dMinOut << "] dMaxOut[" << dMaxOut << "]" << endl;
            cout << "   *** Test (" << ipat << " ) ncrt(" << ncrt << ") ecart[" << ecart << "] eqm_crt[" << eqm_crt << "]" << endl;
        }
        ntd->Fill(1.0 * (ipat - ipatdeb), ecart, eqm_crt);
    }


    fx = eqm_crt / (double) (ipatfin - ipatdeb);

    if ( blog )
        ntd->Scan("*");

    // Clean
    delete[] dVal; dVal = NULL;
    delete crtMLP; crtMLP = NULL;
    delete ntd; ntd = NULL;
}

void InteractiveEQM(int mode, int n, const ColumnVector& weights, double& fx, ColumnVector& grad, int& result)
{
    Bool_t blog = gUranieANNModeler->getLog();

    if (blog)
        cout << endl << endl << " *** InteractiveEQM n[" << n << "] Mode[" << mode << "] **" << endl;

    Int_t ninput = gUranieANNModeler->getNInput();
    Int_t noutput = gUranieANNModeler->getNOutput();

    if ( blog ) {
        cout << "  ** ninput[" << ninput << "]" << endl;
        cout << "  ** noutput[" << noutput << "]" << endl;
    }

    Double_t dweightDecayParameter = gUranieANNModeler->getWeightDecay();
    if (blog)
        cout << "** weightDecay [" << dweightDecayParameter << "]**" << endl;

    Int_t napp = gUranieANNModeler->getNAPP();
    Int_t ntest = gUranieANNModeler->getNTest();
    if (blog)
        cout << "  ** napp[" << napp << "] ntest[" << ntest << "]" << endl;
    Double_t * dVal = gUranieANNModeler->getdval();

    // Load the weights in the current ANN
    gUranieANNModeler->_UMLP->setWeights(weights);

    if (0) {
        gUranieANNModeler->_UMLP->printLog();
        cout << endl;
    }

    double eqm_crt = 0.0;
    double eqm_tst = 0.0;
    ColumnVector g(n);
    Matrix matA(n, napp);
    if (mode & NLPGradient)
    {
        for (int j = 1; j <= n; j++)
        {
            g(j) = 0.0;
        }
    }

    double ecart;
    Double_t fx_test;
    Int_t ncrt;

    gUranieANNModeler->_UMLP->unsetLog();

    // Optimisation performed on the [0; napp) range
//    blog = kTRUE;
    for (int ipat = 0; ipat < napp; ipat++) {
        if (blog)
            cout<< "   *** ipat (" << ipat << "/" << napp << ")" << endl;
        gUranieANNModeler->setValue(ipat, ncrt, dVal);
        ecart =  gUranieANNModeler->_UMLP->getEcart(dVal);
        eqm_crt += ecart; ///// * ecart;
        if ( blog && ( ipat < 5 ) )
            cout << "   *** ncrt (" << ncrt << ") Ecart[" << ecart <<  "] EQM_CRT[" << eqm_crt << "]" <<endl;
        if (mode & NLPGradient) {
            for (int j = 1; j <= n; j++) {
                g(j) += gUranieANNModeler->_UMLP->getWeightsDerivate(j);
                if ( 0 ) // ( blog && ( ipat < 5 ) )
                    cout << "   Gradient(" << j << "/" << n << ") ncrt(" << ncrt << ") = [" << g(j) << "]"  << endl;
            }
        }
        if (mode & NLPHessian) {
            for (int j = 1; j <= n; j++) {
                matA(j, ipat + 1) = gUranieANNModeler->_UMLP->getPartialWeightsDerivate(j);
            }
            //matA(j, ipat + 1) = gUranieUMLP->getPartialWeightsDerivate(j);
        }
    }
    gUranieANNModeler->_UMLP->unsetLog();

    if (blog)
        cout << "  *** eqm_crt[" << eqm_crt << "] napp[" << napp << "] " << endl << endl;

    // But computation of an extra criteria on the test base (i.e. range [napp; napp+ntest)
    for (int ipat = napp; ipat < napp+ntest; ipat++) {
        if (blog)
            cout<< "   *** ipat (" << ipat << "/" << napp+ntest << ")" << endl;
        gUranieANNModeler->setValue(ipat, ncrt, dVal);
    ecart = gUranieANNModeler->_UMLP->getEcart(dVal);
        eqm_tst += ecart; /// * ecart;
        if ( blog && ( ipat < napp+5 ) )
            cout << "   *** ncrt (" << ncrt << ") Ecart[" << ecart <<  "] EQM_TST[" << eqm_tst << "]" <<endl;
    }

    if (blog)
        cout << "  *** eqm_tst[" << eqm_tst << "] ntest[" << ntest << "] " << endl << endl;
    //blog = kFALSE;

    // Criteria computation
    fx = eqm_crt / (double) (napp);
    fx_test = eqm_tst / (double) (ntest);
    if (blog)
        cout << "   *** fx[" << fx << "](" << napp << ") fx_test[" << fx_test << "](" << ntest << ")" << endl;

    gUranieANNModeler->saveIfBetter(weights, fx, fx_test);

    //////////////////////////////////////
    //
    //       -- CASE OF WEIGHT DECAY --
    //
    // Compute the sum of weight square if the weight decay parameter is positive
    //
    //////////////////////////////
    if (dweightDecayParameter > 0)
    {
        Double_t dweightDecayError = 0.0;
        for (int i = 1; i <= n; i++)
            dweightDecayError += weights(i) * weights(i);
        dweightDecayError *= dweightDecayParameter / 2.0;
        fx += dweightDecayError;
        if (blog)
            cout << " -- dweightDecayParameter [" << dweightDecayParameter
                    << "] ** dweightDecayError [" << dweightDecayError << "] **"
                    << endl;
    }
    result = 0;
    if (mode & NLPFunction)
    {
        result = NLPFunction;
    }

    // Gradient computation
    if (mode & NLPGradient)
    {
        for (int j = 1; j <= n; j++)
        {
            //grad(j) = g(j) ;
            grad(j) = g(j) / (double) (napp);
            if (dweightDecayParameter > 0)
                grad(j) += dweightDecayParameter * weights(j);

            //cout << "j[" << j <<"] g["<< g(j) << "]  napp["<< napp << "] grad[" << grad(j) << "]" << endl;
        }
        result = NLPGradient;
    }

//    // Hessian computation
//    if (mode & NLPHessian)
//    {
//        H << matA * matA.t() / (Double_t)(napp);
//        result = NLPHessian;
//    }

    //gUranieUMLP->dumpWeights("totoBis.wei");

    if (blog)
        cout << "** End of InteractiveEQM result[" << result << "] fx[" << fx
                << "]" << endl;
}

TANNModeler::TANNModeler(URANIE::DataServer::TDataServer *tds,
        TString architecture, Double_t dratio, Option_t *option) :
        TModeler(tds, architecture, option), _nSeed(0), _nHiddenLayer(1), _dLearn(numeric_limits<double>::max()), _dTest(numeric_limits<double>::max()),
        _dWeightDecay(-numeric_limits<double>::max()), _elAll(NULL), _bMix(kTRUE), _sArchi(architecture),  _vecWeights(0),
        _tmpSolution(NULL), _listOfAttributes(NULL), _dFcnTol(1e-3), _nNormType(kCR), _nProlemType(kRegression), _dval(0), _blog(kFALSE), _tds(tds)
{
    try
    {
        if (_blog)
        {
            cout << "*** construit TANNModeler tds Name[" << _tds->GetName()
                    << "]" << endl;
            cout << " ** _sArchi[" << _sArchi << "]" << endl;
        }

        SetNameTitle(Form("_modeler_%s", tds->GetName()), "Classe TANNModeler");
        Int_t npat = tds->getNPatterns();

        TString sarchi = _sArchi;

        Int_t nLayer = _sArchi.CountChar(',')+1;
        if (_blog) cout << " ** nLayer[" << nLayer << "]" << endl;
        _nHiddenLayer = nLayer-2;
        if (_blog) cout << " ** _nHiddenLayer[" << _nHiddenLayer << "]" << endl;



        if ( _nHiddenLayer < 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TANNModeler constructor method \n There is no hiden Layer archi[%s], then it is a linear Model: use the TLinearRegression Class.", _sArchi.Data()));

        sarchi.ReplaceAll(",", " ");
        istringstream inStream((char *) sarchi.Data());

        // Case of Input
        inStream >> _sInput;
        _nInput  = _sInput.CountChar(':')  + 1;
        _nNeurons = _nInput;

        // Case of Hidden Layer
        _sHidden ="";
        _nHidden =0;
        Int_t nCrtNeurons;

        string inName;
        for(Int_t i=0; i<_nHiddenLayer; i++) {
            inStream >> inName;
            if (_blog)
                cout << "   ** Hidden Layer (" << i+1 << "/" <<  _nHiddenLayer << ") Name[" << inName.c_str() << "]" << endl;
            if ( i > 0 )
                _sHidden += ",";
            _sHidden += Form("%s", inName.c_str());
            nCrtNeurons = atoi(inName.c_str());
            _nHidden += nCrtNeurons;
            _nNeurons += nCrtNeurons;
            if (_blog)
                cout << "     ** _sHidden[" << _sHidden << "] nCrtNeurons[" << nCrtNeurons << "] _nHidden[" << _nHidden << "] _nNeurons[" << _nNeurons << "]" << endl;
        }

        // Case of the output
        inStream >> _sOutput;
        _nOutput = _sOutput.CountChar(':') + 1;
        _nNeurons += _nOutput;

        if (_sOutput.BeginsWith('@')) {
            _nProlemType = kClassification;
            _sOutput.ReplaceAll("@","");
        }

        if (_blog) {
            cout << " ** _sInput[" << _sInput << "] _nInput [" << _nInput << "]" << endl;
            cout << " ** _sHidden[" << _sHidden << "] _nHidden [" << _nHidden << "]" << endl;
            cout << " ** _sOutput[" << _sOutput << "] _nOutput [" << _nOutput << "]" << endl;
            cout << " ** _nNeurons[" << _nNeurons << "] with the ouput layer" << endl;
        }

        TString soption = TString(option);

        if (dratio < 1)
        {
            _nApp = TMath::Nint(npat * dratio);
        }
        else
        {
            _nApp = TMath::Nint(dratio);

        }
        _nTest = npat - _nApp;
        if (_blog)
            cout << " ** nPatterns[" << npat << "] -- dratio [" << dratio
                    << "] -- App[" << _nApp << "] Test[" << _nTest << "]"
                    << endl;
        if (_nTest < 1)
        {
            cout << "Error : the test size is empty [" << _nTest << "]" << endl;
        }

        // TRandom3 * gRandom = new TRandom3();
        //  gRandom->SetSeed(0);

        if (_elAll != NULL)
        {
            delete _elAll;
            _elAll = NULL;
        }

        _elAll = new URANIE::DataServer::TPatternsEventList("allpat",
                "all patterns", npat);
        for (Int_t i = 0; i < npat; i++)
            _elAll->Enter(i);

        // Mix the data
//        _elAll->Mix();
//        _elAll->Print("all");

        //    init();
        gUranieANNModeler = this;

        // Set up the corresponding UMLP (the actual network)
        _UMLP = new UMLP(_nInput, _sHidden, _nOutput);
        _UMLP->setLog(_blog);

        _nWeight = _UMLP->getNWeights();
        nCrtNeurons = _UMLP->getNNeurons();
        if (_blog)
            cout << " ** nCrtNeurons[" << nCrtNeurons << "] by UMLP class" << endl;

        if ( _nNeurons != nCrtNeurons )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TANNModeler constructor method \n The number of weights is not equal in the twa computation (%d) (%d) by UMLP", _nNeurons, nCrtNeurons));

        switch (gUranieANNModeler->getNormalization())
        {
          case TANNModeler::kMinusOneOne:
      _UMLP->setActivationFunction(UMLP::kTANH);
            break;
          default:
      _UMLP->setActivationFunction(UMLP::kLOGISTIC);
            break;
        }

        switch (_nProlemType) {
        case kRegression:
            _UMLP->setProblem(UMLP::kRegression);
            break;
        case kClassification:
            _UMLP->setProblem(UMLP::kClassification);
            break;
        }

  //	delete gRandom;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

#ifdef  WITH_LIBXML2
TANNModeler::TANNModeler(URANIE::DataServer::TDataServer *tds, const char* pmmlfile, const char* ANNname, Double_t dratio, Option_t *option):
TModeler(tds, TString(",")),
_nSeed(0), _nHiddenLayer(1),
_dWeightDecay(-9999.9),
_elAll(NULL), _bMix(kTRUE),
_vecWeights(0),
_listOfAttributes(NULL),
_dFcnTol(1e-3),
_nNormType(kCR), _nProlemType(kRegression),
_blog(kFALSE),
_tds(tds)
{
    if ( _blog )
    cout << "*** construit TANNModeler tds Name[" << _tds->GetName() << "] PMML[" << pmmlfile << "] Name[" << ANNname << "]" << endl;

    SetNameTitle(Form("_modeler_%s", tds->GetName()), "Classe TANNModeler");
    Int_t npat = tds->getNPatterns();
    string strPmmlfile("");
    string strMod("");
    if ( pmmlfile ) strPmmlfile = string(pmmlfile);
    if ( ANNname ) strMod = string(ANNname);
    PMMLlib::PMMLlib * pmml = new PMMLlib::PMMLlib( strPmmlfile );
    pmml->SetCurrentModel(strMod, PMMLlib::kANN);
    _sArchi = pmml->ReadNetworkStructure();
    _sArchi.ReplaceAll("@", "");
    if ( _blog )
    cout << " ** _sArchi[" << _sArchi << "]" << endl;

    Int_t nnormalizeType = pmml->GetNormalizationType();
    if ( _blog )
    cout << " ** nnormalizeType[" << nnormalizeType << "]" << endl;
    switch( nnormalizeType )
    {
        case 0:
        {
            _nNormType = kMinusOneOne;
        }
        break;
        case 1:
        {
            _nNormType = kCR;
        }
        break;
    }

    _nInput = pmml->GetNbInputs();
    _nHidden = pmml->GetNbNeuronsAtLayer(0);
    _nOutput = pmml->GetNbOutputs();
    if ( _blog)
    cout << " ** _nInput[" << _nInput << "] _nHidden[" << _nHidden << "] _nOutput[" << _nOutput << "]" << endl;

    // Statistiques on the input and output attributes
    _vecMinValues.ReSize(_nInput + _nOutput);
    _vecMaxValues.ReSize(_nInput + _nOutput);

    // Normalize the inputs ( min and max for kMinusOneOne OR mean and std for kCR)
    Double_t *norm = new Double_t[2];
    for(Int_t i=0; i < _nInput; i++)
    {
        //norm =
        pmml->GetNormalisationInput(i, norm);
        if ( _blog)
        cout << "input[" << i<<"] name[" << pmml->GetNameInput(i) << "]";
        _vecMinValues(i+1) = norm[0];
        _vecMaxValues(i+1) = norm[1];
        if ( _blog)
        cout << " -- first[" << _vecMinValues(i+1) << "] second[" << _vecMaxValues(i+1) << "]" << endl;

    }

    // Normalize the outputs ( min and max for kMinusOneOne OR mean and std for kCR)
    for(Int_t i=0; i < _nOutput; i++)
    {
        pmml->GetNormalisationOutput(i, norm);
        if ( _blog)
        cout << "output[" << i<<"] name[" << pmml->GetNameOutput(i) << "]";
        _vecMinValues(_nInput+i+1) = norm[0];
        _vecMaxValues(_nInput+i+1) = norm[1];
        if ( _blog)
        cout << " ** first[" << _vecMinValues(_nInput+i+1) << "] second[" << _vecMaxValues(_nInput+i+1) << "]" << endl;
    }
    delete [] norm; norm = NULL;

    _nNeurons = _nInput + _nHidden + _nOutput;
    _nWeight = _nHidden * ( _nInput + _nOutput + 1 ) + _nOutput;
    if ( _blog )
    cout << " ** _nNeurons[" << _nNeurons<<"] _nWeight[" << _nWeight<<"]"<<endl;
    _vecWeights.ReSize(_nWeight);

    Double_t dval = 0.0;
    Int_t CrtW = 0;
    //  Input Layer to Hidden Layer
    for (Int_t ii = 0; ii < _nHidden; ii++)
    {
        CrtW = ii * ( _nInput + 2) + 3;
        dval = pmml->GetNeuronBias(0, ii);
        if ( _blog )
        cout << " ** Input biais[" <<ii << "] ["<< dval <<"]"<<endl;
        _vecWeights(CrtW++) = pmml->GetNeuronBias(0, ii);
        for (int jj = 0; jj < _nInput; jj++)
        {
            _vecWeights(CrtW++ ) = pmml->GetPrecNeuronSynapse(0, ii, jj);
        }
    }
    //  Hidden Layer to Output ( _nOutput == 1)
    for (Int_t ii = 0; ii < _nOutput; ii++)
    {
        CrtW = 1;
        _vecWeights(CrtW++) = pmml->GetNeuronBias(1, ii);
        if ( _blog )
        cout << " ** nOutput biais[" <<ii << "] ["<< pmml->GetNeuronBias(1, ii) <<"]"<<endl;
        for (Int_t jj = 0; jj < _nHidden; jj++)
        {
            _vecWeights(CrtW) = pmml->GetPrecNeuronSynapse(1, ii, jj);
            CrtW += ( _nInput + 2);
        }
    }

    TString sarchi = _sArchi;
    sarchi.ReplaceAll(",", " ");
    istringstream inStream((char *) sarchi.Data());
    inStream >> _sInput >> _sHidden >> _sOutput;

    //  _nInput = _sInput.CountChar(':')+1;
    //  _nOutput = 1;
    //  _nHidden = atoi(_sHidden.Data());

    if ( _blog )
    {
        cout << " ** _sInput[" << _sInput<<"] _nInput ["<< _nInput << "]"<< endl;
        cout << " ** _sHidden[" << _sHidden<<"] _nHidden ["<< _nHidden << "]"<<endl;
        cout << " ** _sOutput[" << _sOutput<<"] _nOutput ["<< _nOutput << "]"<<endl;
        cout << " ** _nNeurons[" << _nNeurons<<"] _nWeight[" << _nWeight<<"]"<<endl;
    }

    _listOfAttributes = _tds->getListOfAttributes(_sInput.Data());

    TString soption = TString(option);

    if ( dratio < 1 )
    {
        _nApp = TMath::Nint(npat*dratio);
    }
    else
    {
        _nApp = TMath::Nint(dratio);

    }
    _nTest = npat - _nApp;
    if ( _blog )
    cout << " ** nPAt[" << npat << "] -- dratio [" << dratio << "] -- App[" << _nApp <<"] Test[" << _nTest << "]"<<endl;
    if( _nTest < 1 )
    {
        cout << "Error : the test size is empty [" << _nTest << "]"<<endl;
    }

    //    TRandom3 * gRandom = new TRandom3();
    //  gRandom->SetSeed(0);

    if ( _elAll != NULL) {
      delete _elAll; _elAll = NULL;
    }

    _elAll = new URANIE::DataServer::TPatternsEventList("allpat", "all patterns", npat);
    for(Int_t i=0; i<npat; i++)
    _elAll->Enter(i);

    // Mix the data
//    _elAll->Mix();
//    _elAll->Print("all");

    //  init();

    gUranieANNModeler = this;

  // Set up the corresponding UMLP (the actual network)
        //gUranieUMLP = new UMLP(_nInput, _nHidden);
    _UMLP = new UMLP(_nInput, _nHidden);

        switch (gUranieANNModeler->getNormalization())
        {
          case TANNModeler::kMinusOneOne:
            //gUranieUMLP->setActivationFunction(UMLP::kTANH);
      _UMLP->setActivationFunction(UMLP::kTANH);
            break;
          default:
            //gUranieUMLP->setActivationFunction(UMLP::kLOGISTIC);
      _UMLP->setActivationFunction(UMLP::kLOGISTIC);
            break;
        }

  //    	delete gRandom;
    //Clean
    delete pmml;
}
#endif

TANNModeler::~TANNModeler()
{
  //  if ( _blog )
  cout << " ** Destroy TANNModeler ... ";
  //    cout << " *** Delete _elAll ...";
  if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
  if (_elAll != NULL)
      delete _elAll;
  _elAll = NULL;
    }
  //    cout << " Done" << endl << " *** Delete UMLP ...";
  /*if (gUranieUMLP != NULL)
    delete gUranieUMLP;
    gUranieUMLP = NULL;*/
  if (_UMLP != NULL)
    delete _UMLP;
  _UMLP = NULL;

  if(_listOfAttributes != NULL)
    delete _listOfAttributes;
  _listOfAttributes=NULL;

  cout << "Done" << endl;
}

///////////////////////////////
////     Train      ///////////
///////////////////////////////
void TANNModeler::train_cpu(Int_t niter, Int_t nInit, Bool_t doRandomizeWeights)
{
    ColumnVector eqm_test(niter);
    vector<TempSolution> solutions;    ///< vector of solutions

    //Double_t mse_learn, mse_crt;
    Double_t mse_crt;
    Int_t ret_code;
    Bool_t save_candidat;

    Int_t ndim = _nWeight;   //Size of parameter in the ANN
    char trace_file[250];
    char hname[250];
    sprintf(trace_file, "_tstOptCalNips_%s.out", GetName());
    Bool_t haveTrace = kTRUE;
    ColumnVector x_sol(ndim);

    // if (gPad == NULL)
    //     TCanvas * c = new TCanvas("uranie__learn__ANN",
    //             Form("Learn ANN %s", _tds->GetName()), 15, 36, 1098, 787);

    cout << endl;
    // Start timer
    URANIE::DataServer::UTimer timer(1, "Uranie TANNModeler");

    // initialize temporary vector holding one training sample. This is used at every
    // step of the optimization, so just allocate memory once and for all:
    _dval = new Double_t[_nInput+_nOutput];

    for (int iiter = 0; iiter < niter; iiter++)
    {
        TempSolution crtSolution;
        _tmpSolution = &crtSolution;  // so that the saveIfBetter() function can access the current solution
        eqm_test(iiter + 1) = std::numeric_limits<double>::max();
        save_candidat = false;
        // Mix the data
        if (_bMix)
            _elAll->Mix();

        for (int iinit = 0; iinit < nInit; iinit++)
        {
            // Step 1 : unconstrained optim
            //        cout << endl << " ** iter[" << iiter+1 << "/" << niter << "] - init["<< iinit+1 <<"/"<<nInit<<"]  *" << endl;
            if (doRandomizeWeights)
                // initialize with random weights:
                randomizeWeights();

            if (iinit == 0)
                cout << " ** iter[" << iiter + 1 << "/" << niter << "] : "
                        << flush;

            NLF1 *calann = new NLF1(ndim, InteractiveEQM, InteractiveINITEQM);
            //      calann->setDebug();
            //      calann->initFcn();

            // Optim method: Newton, with trust-region
            //OptNewton *OptTrustRegion = new OptNewton(calann, InteractiveUPDATEMODEL);
            OptQNewton *OptTrustRegion = new OptQNewton(calann, InteractiveUPDATEMODEL);

            // options: log file for tracability
            if (iiter == 0 && iinit == 0)
            {
                if (!OptTrustRegion->setOutputFile(trace_file, 0))
                    cerr << "main: output file open failed" << endl;
            }
            else
                OptTrustRegion->setOutputFile(trace_file, haveTrace);

            // Options Optimisation
            OptTrustRegion->setFcnTol(_dFcnTol);
            OptTrustRegion->setMaxIter(OPTIM_MAX_ITER);
            OptTrustRegion->setMaxFeval(OPTIM_MAX_FEVAL);
            OptTrustRegion->setSearchStrategy(TrustRegion);
            OptTrustRegion->optimize();

            ret_code = OptTrustRegion->getReturnCode();

            sprintf(hname, "Solution OptTrustRegion iter[%d/%d] init[%d/%d]",
                    iiter + 1, niter, iinit + 1, nInit);
            OptTrustRegion->printStatus(hname);

            OptTrustRegion->cleanup();

            if (ret_code == 2)  // convergence reached because objective function (EQM) was small enough
            {
                // Do nothing - everything already stored in crtSolution thanks to saveIfBetter().
                mse_crt = crtSolution.eqmtest;
//                mse_learn = calann->getF();
//                InteractiveTESTANN(_nApp, _nApp + _nTest, calann->getXc(), mse_crt);
//                if (0)
//                    cout << "Etape 1 fx[" << mse_crt << "]";
            }
            else // Bad convergence criteria (objective remains too big)
            {
                cout << "   *** ret_code[" << ret_code << "]" << endl;

                // Case of one test only
                if ( ( niter == 1 ) && ( nInit == 1 ) ) {
                    mse_crt = crtSolution.eqmtest;
                } else {
                    mse_crt = std::numeric_limits<double>::max();
                    //mse_learn = mse_crt;
                    cout << "!" << flush;
                }
            }

            // Update of eqm_test array, and flag to decide if we save the candidate solution
            if (mse_crt < eqm_test(iiter + 1))
            {
                eqm_test(iiter + 1) = mse_crt;
                save_candidat = true;

                crtSolution.iter = iiter;
                crtSolution.candidat = 1;
                if (0)
                    cout << "mse_min[" << eqm_test(iiter + 1) << "]" << endl;
                cout << "*";
            }
            else   // Optimisation converged, but is less good than the previous try.
            {
                cout << "=";
            }
            cout << flush;
            if (iinit == nInit - 1)
                cout << " : mse_min[" << eqm_test(iiter + 1) << "]" << endl;

            delete calann;
            calann = NULL;
            delete OptTrustRegion;
            OptTrustRegion = NULL;
        }

        if (save_candidat)
          solutions.push_back(crtSolution);

    }
    delete[] _dval;
    _dval = NULL;

    // Selection of the final retained ANN
    Int_t nfound = selectModel(solutions);

    setWeights(solutions[nfound].x);
    _dLearn = solutions[nfound].eqmlearn;
    _dTest = solutions[nfound].eqmtest;
    if (_blog)
        cout << endl << " ** selected iter [" << nfound << "] Learn[" << _dLearn
                << "] Test[" << _dTest << "]" << endl;
    cout << " ** CPU training finished. Total elapsed time: " << timer.GetElapsedTime() << endl;
}

#ifdef WITH_GPUANN
void TANNModeler::train_gpu(NNLM::NeuronType T, Int_t niter, Int_t nInit, Bool_t randomizeWeights)
{

    try {
        if  ( _nHiddenLayer > 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TANNModeler::train_gpu method \nMulti Hidden Layers (%s) not yet implemented on GPU (use CPU)", _sHidden.Data()));
        if  ( _nOutput > 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TANNModeler::train_gpu method \nSeverall outputs (%s) [%d] not yet implemented on GPU (use CPU)", _sOutput.Data(), _nOutput));
    } catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


  ColumnVector eqm_test(niter);
  Bool_t save_candidat;
  Double_t *dVal = new Double_t[_nInput + 1]; // +1 because there is only one scalar output
  Double_t * weights = new Double_t[_nWeight];
  Int_t unused;

  NNLM::TwoLayersPerceptron mlp_gpu(_nInput,_nHidden,1, T, 0.1, _dFcnTol, OPTIM_MAX_ITER);
  mlp_gpu.setTrainingSampleCount(_nTest+_nApp);

  cout << " ** [GPU] Training on " << _nTest+_nApp << " samples." << endl;

  // Transfer test base to GPU:
  std::vector<NNLM::Vector> xSamples, ySamples;
  xSamples.reserve(_nTest+_nApp);
  ySamples.reserve(_nTest+_nApp);

  for(int i = 0; i < _nTest+_nApp; ++i)
  {
      NNLM::Vector x(_nInput);
      NNLM::Vector y(1);
      gUranieANNModeler->setValue(i, unused, dVal);
      for (int j = 0; j < _nInput; j++)
          x(j) = dVal[j];
      y(0) = dVal[_nInput];
      mlp_gpu.addTrainingSample(x, y);
  }

  URANIE::DataServer::UTimer timer(1, "Uranie TANNModeler");
  // Training
  mlp_gpu.train(niter, nInit);
  cout << " ** GPU training finished. Total elapsed time: " << timer.GetElapsedTime() << endl;

  // Getting back weights
  mlp_gpu.getWeights(weights);

  // Copy the weights to the TANNModeler member (don't use
  // bulk constructor because of type conversion at the element level: GPU side uses its own typedef-d "Real"
  // and the TANNModeler uses another "Real" type through the ColumnVector type).
  for (int i = 1; i <= _nWeight; i++)
      _vecWeights(i) = weights[i-1];

  delete[] dVal;
  delete[] weights;

  _dLearn = mlp_gpu.getLastTrainingLearningError();
  _dTest = mlp_gpu.getLastTrainingTestingError();
}
#endif

void TANNModeler::train(Int_t niter, Int_t nInit, Option_t* option, Bool_t useGPU)
{

    cout << " ** TANNModeler::train niter[" << niter << "] ninit[" << nInit
            << "]" << endl;
    if (useGPU)
      cout << " ** GPU computation has been requested." << endl;

    try
    {
        if (niter < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TANNModeler::train method \n The number of iterations [%d] is negative",
                            niter));
        if (nInit < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TANNModeler::train method \n The number of initialisations [%d] is negative",
                            nInit));
#ifndef WITH_GPUANN
        if (useGPU)
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                          Form("TANNModeler::train method\n GPU computation has been requested, but URANIE was not compiled with GPU-ANN support!"));
#endif

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Treat the options
    TString soption = TString(option);
    soption.ToLower();
    soption.ReplaceAll(" ", "");
    soption.ReplaceAll(",", " ");
    soption += TString(" ");
    if (_blog)
        cout << " ** Option[" << soption << "]" << endl;
    istringstream inStream((char *) soption.Data());
    string inName;
    while (inStream >> inName)
    {
        TString svar(inName.c_str());
        if (_blog)
            cout << " ** svar[" << svar << "]" << endl;
        //////////////////////////////////////
        //
        //       -- CASE OF WEIGHT DECAY --
        //
        if (soption.Contains("weight-decay="))
        {
            TRegexp reg("weight-decay=0.[0-9]*");
            TString out = soption(reg);
            TString sout = out.Remove(0, 13);
            _dWeightDecay = sout.Atof();
        }
        if (soption.Contains("wd="))
        {
            TRegexp reg("wd=0.[0-9]*");
            TString out = soption(reg);
            TString sout = out.Remove(0, 3);
            _dWeightDecay = sout.Atof();
        }
        //
        //////////////////////////////////////
    }
    Bool_t randomizeweights = true;
    if (soption.Contains("+"))
       randomizeweights = false;

    // Init the ANN
    if (randomizeweights)
    {
        cout << " ** init the ANN" << endl;
        init();
    }

    // Use the Seed for the generator
    gRandom->SetSeed(_nSeed);

    if (useGPU)
      {
#ifdef WITH_GPUANN
        switch (_nNormType)
        {
          case kMinusOneOne:
            {
              train_gpu(NNLM::kTanh, niter, nInit, randomizeweights);
              break;
            }
          default:
            {
              train_gpu(NNLM::kSigmoid, niter, nInit, randomizeweights);
              break;
            }
        }

        try
        {
            if ( _nHiddenLayer != 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("Error TANNModeler::train (WITH_GPUANN) Reorder weights is only for one Hidden Layer [%d] and for Uranie v4.1", _nHiddenLayer));
            if ( _nOutput != 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("Error TANNModeler::train (WITH_GPUANN) Reorder weights is only for one Output attribute [%d] and for Uranie v4.1", _nOutput));


            if ( _blog ) {
                printLog();
                cout << " ** Reorder the weights parameters" << endl;
            }

            ColumnVector vecNewWeights(_nWeight);
            Int_t nWCrt = 1;
            Int_t iWcrt = 0;
            Int_t nW = ( _nInput + _nOutput +1 ) * _nHidden + _nOutput;
            if ( _blog )
                cout << " *** _nInput[" << _nInput << "] _nHidden[" << _nHidden << "] _nOutput[" << _nOutput << "] :: nW[" << nW << "] _nWeight[" << _nWeight << "]" << endl;

            if ( nW != _nWeight )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("Error TANNModeler::train (WITH_GPUANN) The number of weights computed [%d] is not equal to the original number of weights [%d]", nW, _nWeight));


            for(Int_t i=0; i<_nHidden; i++) {
                iWcrt = i * (_nInput + 2) + 3;
                // Biais Hidden
                vecNewWeights(nWCrt++) = _vecWeights(iWcrt++);
                // Weights Input-Hidden
                for(Int_t j=0; j<_nInput; j++)
                    vecNewWeights(nWCrt++) = _vecWeights(iWcrt++);
            }
            // Case Of One Output
            // Biais Output
            vecNewWeights(nWCrt++) = _vecWeights(1);
            for(Int_t i=0; i<_nHidden; i++) {
                // Weight Hidden-Output
                iWcrt = i * (_nInput + 2) + 2;
                vecNewWeights(nWCrt++) = _vecWeights(iWcrt);
            }
            if ( _blog )
                cout << " *** nWCrt[" << nWCrt << "]_nWeight[" << _nWeight << "]" << endl;

            if ( nWCrt-1 != _nWeight )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("Error TANNModeler::train (WITH_GPUANN) The number of weights reordered [%d] is not equal to the original number of weights [%d]", nWCrt-1, _nWeight));


            for(Int_t i=0; i<_nWeight; i++) {
                _vecWeights(i+1) = vecNewWeights(i+1);
                if ( _blog )
                    cout << " *** i(" << i << ") ValueTmp[" << vecNewWeights(i+1) << "] ValueTmp[" << _vecWeights(i+1) << "]"  << endl;
            }
            if ( _blog ) {
                printLog();
                cout << " ** End Of Reordonne the weights parameters" << endl;
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

#endif
      }
    else
      train_cpu(niter, nInit, randomizeweights);
}

void TANNModeler::randomizeWeights()
{
    for (Int_t i = 1; i <= _nWeight; i++)
    {
        _vecWeights(i) = -1.00 + 2.00 * gRandom->Rndm();
        if (_blog)
            cout << " ** i[" << i << "] _vecWeights[" << _vecWeights(i) << "]"
                    << endl;
    }
}

void TANNModeler::setNormalization(ENorm n)
{
    _nNormType = n;

    switch (_nNormType)
    {
      case TANNModeler::kMinusOneOne:
        //gUranieUMLP->setActivationFunction(UMLP::kTANH);
    _UMLP->setActivationFunction(UMLP::kTANH);
        break;
      default:
        //gUranieUMLP->setActivationFunction(UMLP::kLOGISTIC);
    _UMLP->setActivationFunction(UMLP::kLOGISTIC);
        break;
    }
}


void TANNModeler::init()
{
    // The weights vector
    _vecWeights.ReSize(_nWeight);

    // The statistics vector for input and output
    _vecMinValues.ReSize(_nInput + _nOutput);
    _vecMaxValues.ReSize(_nInput + _nOutput);

    if(_listOfAttributes != NULL) delete _listOfAttributes;

    TString sallatt = Form("%s:%s", _sInput.Data(), _sOutput.Data());

    if ( _blog ) {
        cout << " ** sallatt input[" << sallatt << "]  _nInput["  << _nInput << "]" << endl;
        cout << " ** sallatt output[" << sallatt << "] _nOutput[" <<  _nOutput << "]" << endl;
    }

    _listOfAttributes = _tds->getListOfAttributes(sallatt.Data());
    Int_t natt = _listOfAttributes->GetEntries();

    if ( _blog )
        cout << " ** _listOfAttributes nATT[" <<  natt << "]" << endl;

    //TLeaf *leafOfOutput = _tds->getTuple()->GetLeaf(_sOutput);

    if (_tds)
        _tds->computeStatistic(sallatt);
    else
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                Form("TANNModeler::init[%s]  ERROR The TTree is NULL",
                        GetName()));

    URANIE::DataServer::TAttribute *att;

    for (Int_t ivar = 0; ivar < natt; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(ivar);
        if ( _blog ) {
            cout << "  *** ivar[" << ivar << "," << natt << "] Name[" << att->GetName() << "] - ";
//            att->printLog();
        }
        switch (_nNormType)
        {
        case kMinusOneOne:
        {
            _vecMinValues(ivar + 1) = att->getMinimum();
            _vecMaxValues(ivar + 1) = att->getMaximum();
            if ( _blog ) cout << "  ** kMinusOneOne*** Min[" << _vecMinValues(ivar + 1) << "] Max[" << _vecMaxValues(ivar + 1) << "]" << endl;

            break;
        }
        default:
        {
            _vecMinValues(ivar + 1) = att->getMean();
            _vecMaxValues(ivar + 1) = att->getStd();
            if ( _blog ) cout << "  ** Default *** Min[" << _vecMinValues(ivar + 1) << "] Max[" << _vecMaxValues(ivar + 1) << "]" << endl;

            break;
        }
        }

        // Store the max/min into the UMLP:
        _UMLP->setMinMaxValue(ivar + 1, _vecMinValues(ivar + 1), _vecMaxValues(ivar + 1));

    }

    if ( _blog)
        _UMLP->printLog();

//    setLog();

    if (1) //////_blog)
    {
        for (Int_t ivar = 0; ivar < _nInput; ivar++)
        {
            att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(ivar);
            cout << " ** Input  (" << ivar + 1 << ") Name[" << att->GetName()
                    << "] Min[" << _vecMinValues(ivar + 1) << "] Max["
                    << _vecMaxValues(ivar + 1) << "]" << endl;
        }
        for (Int_t ivar = 0; ivar < _nOutput; ivar++)
        {
            att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(_nInput+ivar);
            cout << " ** Output  (" << ivar + 1  << ") Name[" << att->GetName()
                    << "] Min[" << _vecMinValues(_nInput+ivar + 1) << "] Max["
                    << _vecMaxValues(_nInput+ivar + 1) << "]" << endl;
        }
        cout << " ** Tolerance  (" << getFcnTol() << ")" << endl;
        cout << " ** sHidden    (" << _sHidden << ") (" << getNHidden() << ")" << endl;
        cout << " ** Nb Weights (" << _nWeight << ")" << endl;
        switch (_nNormType)
        {
        case kMinusOneOne:
        {
            cout << " ** ActivationFunction[TanH]" << endl;
            break;
        }
        default:
        {
            cout << " ** ActivationFunction[LOGISTIC]" << endl;
            break;
        }
        }
        if (_dWeightDecay > 0)
            cout << " ** WeightDecay (" << getWeightDecay() << ")" << endl;
        cout << " **" << endl;
    }

//    unsetLog();
}

void TANNModeler::saveIfBetter(const ColumnVector & weights, Double_t eqm_learn, Double_t eqm_tst)
{
    if (eqm_tst < _tmpSolution->eqmtest)
    {
        _tmpSolution->x = weights;
        _tmpSolution->eqmtest = eqm_tst;
        _tmpSolution->eqmlearn = eqm_learn;
    }
}

///////////////////////////////
////      SEED      ///////////
///////////////////////////////
void TANNModeler::setSeed(Int_t ind)
{
    if (ind == 0)
    {
        time_t secondes;
        time(&secondes);
        _nSeed = (Int_t)(secondes);
    }
    else
    {
        _nSeed = ind;
    }
    if ( _elAll )
      _elAll->setSeed(_nSeed);
}
///////////////////////////////
////     Value      ///////////
///////////////////////////////
void TANNModeler::setValue(Int_t ind, Int_t &ncrt, Double_t *xy, Bool_t normalize)
{
    if ( _blog )
        cout<< "   *** TANNModeler::setValue ind(" << ind << "/" << ncrt << ") _sOutput[" << _sOutput << "] normalize(" << normalize << ")" << endl;

    Int_t natt = _listOfAttributes->GetEntries();

    if ( _blog) {
        cout << " ** _listOfAttributes nATT[" <<  natt << "]" << endl;
        _listOfAttributes->Print("all");
    }


    if (ind > _nApp + _nTest)
    {
        TString serr = Form("Error : index %d is greater than the size %d\n",
                ind, _nApp);
        cout << serr << endl;
        throw serr;
    }
    ncrt = (Int_t) _elAll->GetEntry(ind);
    if ( _blog )
        cout << "   *** ind(" << ind << ") ncrt[" << ncrt << "]" << endl;
    _tds->getTuple()->GetEntry(ncrt);


    // Populate input part of the xy vector
    URANIE::DataServer::TAttribute *att;
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(ivar);
        TLeaf *leaf = _tds->getTuple()->GetLeaf(att->GetName());
        if(normalize)
          {
            switch (_nNormType)
            {
              case kMinusOneOne:
                {
                  xy[ivar] = 2.0 * (leaf->GetValue() - _vecMinValues(ivar + 1))
                        / (_vecMaxValues(ivar + 1) - _vecMinValues(ivar + 1)) - 1.0;
                  break;
                }
              default:
                {
                  xy[ivar] = (leaf->GetValue() - _vecMinValues(ivar + 1))
                        / _vecMaxValues(ivar + 1);
                }
            }
          } else {
          xy[ivar] = leaf->GetValue();
        }

        if ( _blog ) cout << "   i(" << ivar << "/" << natt << ") Val[" << leaf->GetValue() << "] xy[" << xy[ivar] << "], ";
    }

    // Output normalization
    for (Int_t ivar = _nInput; ivar < _nInput+_nOutput; ivar++) {
        att = (URANIE::DataServer::TAttribute *) _listOfAttributes->At(ivar);
        TLeaf *leaf = _tds->getTuple()->GetLeaf(att->GetName());

        switch (_nProlemType) {
        case kRegression:
            if(normalize) {
                switch (_nNormType)
                {
                case kMinusOneOne:
                {
                    xy[ivar] = 2.0 * (leaf->GetValue() - _vecMinValues(ivar + 1))
                            / (_vecMaxValues(ivar + 1) - _vecMinValues(ivar + 1)) - 1.0;
                    break;
                }
                default:
                {
                    xy[ivar] = (leaf->GetValue() - _vecMinValues(ivar + 1))
                            / _vecMaxValues(ivar + 1);
                    break;
                }
                }
            } else {
                xy[_nInput+ivar] = leaf->GetValue();
            }
            break;
        case kClassification:
            if(normalize) {
                xy[ivar] = 1.0 * (leaf->GetValue() - _vecMinValues(ivar + 1))
                        / (_vecMaxValues(ivar + 1) - _vecMinValues(ivar + 1));
            } else {
                xy[ivar] = leaf->GetValue();
            }

        default:
            break;
        }

        if ( _blog ) cout << "   i(" << ivar << "/" << natt << ") Val[" << leaf->GetValue() << "] xy[" << xy[ivar] << "], ";
    }
    if ( _blog ) cout << endl;
}

////////////////////////////////
////  Exports Function    //////
////////////////////////////////
// void TANNModeler::*exportFunction(const char* lang, const char* file,
//         const char* name, Option_t *option)
// {
//     TString lg = lang;
//     lg.ToLower();
//
//     Bool_t ok = kFALSE;
//     cout << endl << "*******************************" << endl;
//     cout << "*** exportFunction lang[" << lg << "] file[" << file << "] name["
//             << name << "]" << endl;
//     try
//     {
//         if (lg.BeginsWith("c"))
//         {
//
//             TModeler::exportFunction("c", file, name);
//             TString sfile = TString(file);
//             if (sfile.Length() > 0)
//                 sfile = Form("%s.C", file);
//             else
//                 sfile = "model.C";
//             gSystem->CopyFile(sfile.Data(), Form("%s.bak", sfile.Data()),
//                     kTRUE);
//
//             sfile = TString(file);
//             if (sfile.Length() > 0)
//                 sfile = Form("%s.C", file);
//             exportModelCplusplus(sfile.Data(), name, option);
//             ok = kTRUE;
//         }
//         if (lg.Contains("pmml"))
//         {
//             exportModelPMML(file, name, option);
//             ok = kTRUE;
//         }
//         if (lg.Contains("fortran"))
//         {
//             TModeler::exportFunction("fortran", file, name);
//             TString sfile = TString(Form("%s.f", file));
//             gSystem->CopyFile(sfile.Data(), Form("%s.bak", sfile.Data()),
//                     kTRUE);
//             exportModelFortran(sfile.Data(), name, option);
//             ok = kTRUE;
//         }
//         if (ok == kFALSE)
//         {
//             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
//                     Form(
//                             "TANNModeler::exportFunction  ERROR The language [%s]chosen is not accepted",
//                             lang));
//         }
//     } catch (URANIE::Exceptions::UErrorExceptions& ue)
//     {
//         ue.printMessage();
//         throw ue;
//         exit(-1);
//     }
//     cout << "*** End Of exportFunction " << endl;
//     cout << "*******************************" << endl;
//
// }

///////////////////////////////
////  Export PMML        //////
///////////////////////////////
void TANNModeler::exportModelPMML(const char* file, const char* name, Option_t *option) const
{

    try
    {
        if ( _nOutput > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TANNModeler::exportModelPMML method\n PMML export for ANN not yet implemented for multi-ouput _nY (%d)", _nY));

        if ( _nHiddenLayer > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TANNModeler::exportModelPMML method\n PMML export for ANN not yet implemented for multi-hidden layers _nHiddenLayer (%d) and _sHidden[%s]", _nHiddenLayer, _sHidden.Data()));

        if ( _nProlemType == kClassification)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TANNModeler::exportModelPMML method\n PMML export for ANN  not yet implemented with \"Classification\" problem");


    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    cout << endl << "*******************************" << endl;

    TString soption = TString(option);  // NEW / UPDATE / BACKUP / ADD
    soption.ToLower();
    enum
    {
        klocBACKUP, klocUPDATE, klocNEW, klocADD
    };
    Int_t nOption = klocBACKUP;
    if (soption.Contains("update"))
        nOption = klocUPDATE;
    if (soption.Contains("new"))
        nOption = klocNEW;


#ifdef  WITH_LIBXML2
    PMMLlib::PMMLlib *pmml = NULL;
    URANIE::DataServer::TAttribute *att;
    Bool_t bGoodOption = kFALSE;
    while(bGoodOption==kFALSE)
    {
        switch(nOption)
        {
            case klocNEW:
                pmml = new PMMLlib::PMMLlib(_blog);
                // Header node
                pmml->SetHeader("myCopyright", "Text Description", "Uranie", "3.3.2", "Compilation date : 2000-01-01");

                // DataDictionnary node
                for(Int_t j=0; j<_nInput; j++)
                {
                    att = (URANIE::DataServer::TAttribute *)_listOfAttributes->At(j);
                    URANIE::DataServer::TAttributeFormula *attformula = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(att);
                    if ( attformula == NULL )
                    {
                        pmml->AddDataField(att->GetName(), att->GetTitle(), "continuous", "double", "ClosedClosed", att->getMinimum(), att->getMaximum(), true);
                    }
                    else
                    {
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, "TANNModeler::exportModelPMML  ERROR PMML cannot deal with TAttributeFormula attributes");
                    }
                }
                att = _tds->getAttribute(_sOutput);
                pmml->AddDataField(att->GetName(), att->GetTitle(), "continuous", "double", "ClosedClosed", att->getMinimum(), att->getMaximum(), true);
                bGoodOption = kTRUE;

            break;
            // end of "case NEW"

            case klocUPDATE:
                try
                {
                    pmml = new PMMLlib::PMMLlib(file,_blog);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New PMML file created"<<endl;
                    nOption = klocNEW;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                try
                {
                    pmml->SetCurrentModel(name, PMMLlib::kANN);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New model created"<<endl;
                    nOption = klocADD;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                pmml->UnlinkNode();
                bGoodOption = kTRUE;

            break;
            // end of "case UPDATE"

            case klocBACKUP:
                try
                {
                    pmml = new PMMLlib::PMMLlib(file,_blog);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New PMML file created"<<endl;
                    nOption = klocNEW;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                try
                {
                    pmml->SetCurrentModel(name, PMMLlib::kANN);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New model created"<<endl;
                    nOption = klocADD;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                pmml->BackupNode();
                bGoodOption = kTRUE;

            break;
            // end of "case BACKUP"

            case klocADD:
                pmml = new PMMLlib::PMMLlib(file,_blog);
                bGoodOption = kTRUE;
            break;
//             end of "case ADD"

        } // end of "switch(nOption)"
    }
    // Neural Network
    double dmin, dmax, orig1, norm1, orig2, norm2;

    switch (_nProlemType) {
    case kRegression:
        pmml->AddNeuralNetwork(name, PMMLlib::kREGRESSION);
        break;
    case kClassification:
        // TODO FGA 2017/12/21
        //pmml->AddNeuralNetwork(name, PMMLlib::kClassification);
        break;
    }


    for(Int_t j=0; j<_nInput; j++)
    {
        att = (URANIE::DataServer::TAttribute *)_listOfAttributes->At(j);
        URANIE::DataServer::TAttributeFormula *attformula = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(att);
        if ( attformula == NULL )
        {
            // Set MiningSchema
            pmml->AddMiningSchema(att->GetName(), "active");

            // Set NeuralInput
            dmin = _vecMinValues(j+1);
            dmax = _vecMaxValues(j+1);
            switch( _nNormType )
            {
                case URANIE::Modeler::TANNModeler::kMinusOneOne:
                orig1 = dmin;
                norm1 = -1;
                orig2 = dmax;
                norm2 = 1;
                break;
                default:
                orig1 = 0;
                norm1 = -dmax/dmin;
                orig2 = dmin;
                norm2 = 0;
                break;
            }
            pmml->AddNeuralInput(j, att->GetName(), "continuous", "float", orig1, norm1, orig2, norm2);
        }
    }
    att = _tds->getAttribute(_sOutput);
    pmml->AddMiningSchema(att->GetName(), "predicted");

    // Set neural layers
    switch( _nNormType )    {
        case kMinusOneOne: {
        pmml->AddNeuralLayer(PMMLlib::kTANH);
        break;
    }
    default: {
        pmml->AddNeuralLayer(PMMLlib::kLOGISTIC);
        break;
    }
    }

    vector<double> weights;
    for(int j=0; j<_nHidden; j++)// hidden layers
    {
        weights.clear();
        weights.resize(_nInput);
        for(int i=0; i<_nInput; i++)
        {
            weights[i] = _vecWeights( j*(_nInput+_nOutput+1)+4+i );
        }
        pmml->AddNeuron(_nInput+j, (double)_vecWeights(j*(_nInput+_nOutput+1)+3), _nInput, 0, weights);
    }

    switch (_nProlemType) {
    case kRegression:
        pmml->AddNeuralLayer(PMMLlib::kIDENTITY);
        break;
    case kClassification:
        pmml->AddNeuralLayer(PMMLlib::kLOGISTIC);
        break;
    }
    for(int j=0; j<_nOutput; j++) // output layers
    {
        weights.clear();
        weights.resize(_nHidden);
        for(int i=0; i<_nHidden; i++)
        {
            weights[i] = _vecWeights( i*(_nInput+_nOutput+1)+2 );
        }
        pmml->AddNeuron(_nInput+_nHidden+j, (double)_vecWeights(1), _nHidden, _nInput, weights);
    }

    // Set NeuralOutput
    dmin = _vecMinValues(_nInput+1);
    dmax = _vecMaxValues(_nInput+1);
    switch( _nNormType )
    {
        case URANIE::Modeler::TANNModeler::kMinusOneOne:
        orig1 = dmin;
        norm1 = -1;
        orig2 = dmax;
        norm2 = 1;
        break;
        default:
        orig1 = 0;
        norm1 = -dmax/dmin;
        orig2 = dmin;
        norm2 = 0;
        break;
    }
    pmml->AddNeuralOutput(_nInput+_nHidden, att->GetName(), "continuous", "float", orig1, norm1, orig2, norm2);

    switch(nOption)
    {
        case klocNEW:
        pmml->Write(file);
        break;
        // end of "case NEW"
        default :
        pmml->Write();
        break;
        // end of "case BACKUP", "UPDATE", "ADD"
    } // end of "switch(nOption)"

    time_t seconds;
    time(&seconds);
    TString header("");
    header.Append("     *********************************************\n");
    header.Append(Form("     ** Uranie v%s -  Date : %s\n",URANIE_RELEASE, URANIE_RELEASE_DATE));
    header.Append(Form("     ** Export Modeler : %s\n",GetName()));
    header.Append(Form("     **      Date : %s",ctime(&seconds)));
    header.Append("     *********************************************\n");
    header.Append("\n");
    header.Append("     *********************************************\n");
    header.Append("     ** TDataServer \n");
    header.Append(Form("     **        Name  : %s **\n",_tds->GetName()));
    header.Append(Form("     **        Title : %s\n",_tds->GetTitle()));
    header.Append(Form("     **     Patterns : %i\n",_tds->getNPatterns()));
    header.Append(Form("     **   Attributes : %i\n",_tds->getNAttributes()));
    header.Append("     *********************************************\n");
    // inputs title
    header.Append("\n");
    header.Append(Form("     INPUT : %i\n",_nInput));
    for(Int_t ivar=0; ivar<_nInput; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *)_listOfAttributes->At(ivar);
        header.Append(Form("             %s : Min[%f] Max[%f] Unit[%s]\n",att->GetName(),att->getMinimum(),att->getMaximum(), att->getUnity().Data()));
    }
    // outputs title
    header.Append("\n");
    att = _tds->getAttribute(_sOutput);
    header.Append(Form("     OUTPUT : %i\n",_nOutput));
    header.Append(Form("             %s : Min[%f] Max[%f] Unit[%s]\n",att->GetName(),att->getMinimum(),att->getMaximum(), att->getUnity().Data()));
    header.Append("\n");
    header.Append(Form("     HIDDEN : %i\n",_nHidden));
    //Clean
    delete pmml;
#endif
    cout << "*** End Of exportModelPMML " << endl;
    cout << "*******************************" << endl;
}

///////////////////////////////
////  Export C++         //////
///////////////////////////////
void TANNModeler::exportModelCplusplus(const char* file, const char* name,
        Option_t *option) const
{
    TString soption = option;
    soption.ToLower();
    cout << " ** exportFunctionCplusplus file[" << file
         << "] name[" << name << "]" << endl;
    // Int_t i, member, source, CrtW, nCrt;
    // nCrt = 0; i = 0; CrtW = 0; source = 0; member = 0;
    // Int_t nVar = _tds->getNAttributes();

    TString sFile = TString(file);
    if (sFile.Length() == 0)
        sFile = Form("%s_nH%03d.C", _sOutput.Data(), _nHidden);

    TString sName = TString(name);
    if (sName.Length() == 0)
        sName = Form("Fct_%s", _sOutput.Data());

    cout << " ** exportFunctionCplusplus sFile[" << sFile
         << "] sName[" << sName << "]" << endl;

    // Open C source file
    std::ofstream sourcefile(sFile.Data());
//    switch (_nNormType)
//    {
//    case kMinusOneOne:
//    {
//        sourcefile << "#define " << sName << "_ActivationFunction(sum)         ( tanh(sum) )"
//                << endl;
//        break;
//    }
//    default:
//    {
//        sourcefile
//                << "#define " << sName << "_ActivationFunction(sum) ( 1.0 / ( 1.0 + exp( -1.0 * sum )) )"
//                << endl;
//    }
//    }

    sourcefile << "void " << sName << "(double *param, double *res)" << endl;
    sourcefile << "{" << endl;

    // Constantes
    sourcefile << "  int nInput   = " << _nInput << ";" << endl;
    sourcefile << "  int nOuput   = " << _nOutput << ";" << endl;
    sourcefile << "  int nHidden  = " << _nHidden << ";" << endl;
    sourcefile << "  double " << sName << "_act[" << _nNeurons << "];" << endl;

    // Normalize data
    sourcefile << endl;
    sourcefile << "  // --- Pretraitment of the inputs and outputs" << endl;

    sourcefile << "  double " << sName << "_minInput["<<_nInput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
    {
        sourcefile << _vecMinValues(ivar + 1) << ((ivar!=(_nInput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            sourcefile << endl << "  ";
    }
    if (_nInput % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    sourcefile << "  double " << sName << "_minOuput["<<_nOutput<<"] = {" << endl << "  ";
    sourcefile << _vecMinValues(_nInput + 1);
    sourcefile << "  };" << endl;

    sourcefile << "  double " << sName << "_maxInput["<<_nInput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
    {
        sourcefile << _vecMaxValues(ivar + 1) << ((ivar!=(_nInput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            sourcefile << endl << "  ";
    }
    if (_nInput % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    sourcefile << "  double " << sName << "_maxOuput["<<_nOutput<<"] = {" << endl << "  ";
    sourcefile << _vecMaxValues(_nInput + 1);
    sourcefile << "  };" << endl;

    // Vector of weights
    sourcefile << endl;
    sourcefile << "  // --- Values of the weights" << endl;
    sourcefile << "  double " << sName << "_valW["<<_nWeight<<"] = {" << endl << "  ";

    for (int i = 0; i < _nWeight; i++)
    {
      sourcefile << _vecWeights(i + 1) << ((i!=(_nWeight-1))?", ":"");
        if ((i + 1) % 5 == 0)
            sourcefile << endl << "  ";
    }
    sourcefile << endl << "  };" << endl;

    // Constants
    sourcefile << "  // --- Constantes";
    sourcefile << endl;
    sourcefile << "  int indNeurone = 0;" << endl;
    sourcefile << "  int CrtW;" << endl;
    sourcefile << "  double sum;" << endl;

    // Input Layer
    sourcefile << endl;
    sourcefile << "  // --- Input Layers" << endl;
    sourcefile << "  for(int i = 0; i < nInput; i++) {" << endl;
    switch (_nNormType)
    {
    case kMinusOneOne:
    {
        sourcefile << "     " << sName
                << "_act[indNeurone++] = 2.0 * ( param[i] - " << sName
                << "_minInput[i] ) / ( " << sName << "_maxInput[i] - " << sName
                << "_minInput[i] ) - 1.0;" << endl;
        break;
    }
    default:
    {
        sourcefile << "     " << sName << "_act[indNeurone++] = ( param[i] - "
                << sName << "_minInput[i] ) / " << sName << "_maxInput[i];"
                << endl;
    }
    }
    sourcefile << "  }" << endl;

    // Hidden Layer
    sourcefile << endl;
    sourcefile << "  // --- Hidden Layers" << endl;
    sourcefile << "  for (int member = 0; member < nHidden; member++) {"
            << endl;
    sourcefile << "     int CrtW = member * ( nInput + 2) + 2;" << endl;
    sourcefile << "     sum = " << sName << "_valW[CrtW++];" << endl;
    sourcefile << "     for (int source = 0; source < nInput; source++) {"
            << endl;
    sourcefile << "         sum += " << sName << "_act[source] * " << sName
            << "_valW[CrtW++];" << endl;
    sourcefile << "       }" << endl;
//    sourcefile << "       " << sName << "_act[indNeurone++] = " << sName << "_ActivationFunction(sum);" << endl;
    switch( _nNormType ) {
    case kMinusOneOne: {
        sourcefile << "     " << sName << "_act[indNeurone++] = tanh(sum);" << endl;
        break;
    }
    default: {
        sourcefile << "     " << sName << "_act[indNeurone++] = 1.0 / ( 1.0 + exp( -1.0 * sum ));" << endl;
    }
    }




    sourcefile << "  }" << endl;

    // Output layer
    sourcefile << endl;
    sourcefile << "  // --- Output" << endl;
    sourcefile << "  for (int member = 0; member < nOuput; member++) {" << endl;
    sourcefile << "    sum = " << sName << "_valW[0];" << endl;
    sourcefile << "    for (int source = 0; source < nHidden; source++) {"
            << endl;
    sourcefile << "      CrtW = source * ( nInput + 2) + 1;" << endl;
    sourcefile << "      sum += " << sName << "_act[nInput+source] * " << sName
            << "_valW[CrtW];" << endl;
    sourcefile << "    }" << endl;
    sourcefile << "    " << sName << "_act[indNeurone++] = sum;" << endl;
    switch (_nNormType)
    {
    case kMinusOneOne:
    {
        sourcefile << "    res[member] = " << sName
                << "_minOuput[member] + 0.5 * ( " << sName
                << "_maxOuput[member] - " << sName
                << "_minOuput[member] ) * ( sum + 1.0);" << endl;
        break;
    }
    default:
    {
        sourcefile << "    res[member] = " << sName << "_minOuput[member] + "
                << sName << "_maxOuput[member] * sum;" << endl;
    }
    }
    sourcefile << "  }" << endl;

    sourcefile << "}" << endl;
    sourcefile.close();
}

void TANNModeler::exportModelCplusplus(std::ofstream * sourcefile) const
{
    Bool_t bTrace = kFALSE;

    TString sName = _sFunctionName; //  "TANN" _sFunctionName

    // product and time
    time_t seconds;
    time(&seconds);


    *sourcefile << "  int nInput   = " << _nInput << ";" << endl;
    *sourcefile << "  int nOuput   = " << _nOutput << ";" << endl;
    //  sourcefile << "  int nWeights = " <<  _nWeight << ";" << endl;
    *sourcefile << "  int nLayers  = " << _nHiddenLayer << ";" << endl;
    *sourcefile << "  int nHidden  = " << _nHidden << ";" << endl;
    TString scrt = _sHidden;
    scrt.ReplaceAll(",", ", ");
    *sourcefile << "  int Hidden[" << _nHiddenLayer<< "]  = {" << scrt << "};" << endl;
    *sourcefile << "  double " << sName << "_act[" << _nNeurons << "];" << endl;

    // Normalisation of data
    *sourcefile << endl;
    *sourcefile << "  // --- Pretraitment of the inputs and outputs" << endl;

    *sourcefile << "  double " << sName << "_minInput["<<_nInput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
    {
      *sourcefile << _vecMinValues(ivar + 1) << ((ivar!=(_nInput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nInput % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    *sourcefile << "  double " << sName << "_minOuput["<<_nOutput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nOutput; ivar++)
    {
      *sourcefile << _vecMinValues(_nInput + 1 +ivar) << ((ivar!=(_nOutput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nOutput % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    *sourcefile << "  double " << sName << "_maxInput["<<_nInput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
    {
        *sourcefile << _vecMaxValues(ivar + 1) << ((ivar!=(_nInput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nInput % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    *sourcefile << "  double " << sName << "_maxOuput["<<_nOutput<<"] = {" << endl << "  ";
    for (Int_t ivar = 0; ivar < _nOutput; ivar++)
    {
      *sourcefile << _vecMaxValues(_nInput + 1 +ivar) << ((ivar!=(_nOutput-1))?", ":"");
        if ((ivar + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    if (_nOutput % 5 != 0)
        *sourcefile << endl;
    *sourcefile << "  };" << endl;

    // Vecteur des poids
    *sourcefile << endl;
    *sourcefile << "  // --- Values of the weights" << endl;
    *sourcefile << "  double " << sName << "_valW["<<_nWeight<<"] = {" << endl << "  ";

    for (int i = 0; i < _nWeight; i++)
    {
        *sourcefile << _vecWeights(i + 1) << ((i!=(_nWeight-1))?", ":"");
        if ((i + 1) % 5 == 0)
            *sourcefile << endl << "  ";
    }
    *sourcefile << endl << "  };" << endl;

    *sourcefile << "  // --- Constantes";
    *sourcefile << endl;
    *sourcefile << "  int indNeurone = 0;" << endl;
    *sourcefile << "  int indFirstSource  = 0;" << endl;
    *sourcefile << "  int indLastSource   = 0;" << endl;
    *sourcefile << "  int CrtW = 0;" << endl;
    *sourcefile << "  double sum;" << endl;

    // couche  entree
    *sourcefile << endl;
    *sourcefile << "  // --- Input Layer" << endl;
    *sourcefile << "  for(int i = 0; i < nInput; i++) {" << endl;
    if( bTrace )
        *sourcefile << "     cout << endl << \"   *** param(\" << param[i]" << "<< \") -- \";" << endl;
    switch (_nNormType)
    {
    case kMinusOneOne:
    {
        *sourcefile << "     " << sName
                    << "_act[indNeurone++] = 2.0 * ( param[i] - " << sName
                    << "_minInput[i] ) / ( " << sName << "_maxInput[i] - " << sName
                    << "_minInput[i] ) - 1.0;" << endl;
        break;
    }
    default:
    {
        *sourcefile << "     " << sName << "_act[indNeurone++] = ( param[i] - "
                    << sName << "_minInput[i] ) / " << sName << "_maxInput[i];"
                    << endl;
    }
    }
    if( bTrace )
        *sourcefile << "     cout << \"Input(\" << " << sName << "_act[indNeurone-1]" << "<< \")\" << endl;" << endl;
    *sourcefile << "  }" << endl;

    // the Hidden layers
    *sourcefile << endl;
    *sourcefile << "  // --- Hidden Layers" << endl;
    *sourcefile << "  indFirstSource  = 0;" << endl;
    *sourcefile << "  indLastSource   = nInput;" << endl;
    if( bTrace )
        *sourcefile << "     cout << \"  input Layer - ndFirstSource (\" << indFirstSource << \") indLastSource (\" << indLastSource << \")\" << endl;" << endl;
    *sourcefile << "  for (int ilayer = 0; ilayer < nLayers; ilayer++) {" << endl;
    *sourcefile << "    // --- current hidden Layer" << endl;
    *sourcefile << "    for (int member = 0; member < Hidden[ilayer]; member++) {" << endl;
    if ( bTrace )
        *sourcefile << "       cout << \"   *** HL CrtW[\" << CrtW+1  << \"]\" << endl; " << endl;
    *sourcefile << "       sum = " << sName << "_valW[CrtW++];" << endl;
    *sourcefile << "       for (int source = indFirstSource; source < indLastSource; source++) {"
                << endl;
    *sourcefile << "           sum += " << sName << "_act[source] * " << sName << "_valW[CrtW++];" << endl;
    *sourcefile << "       }" << endl;
    switch( _nNormType ) {
    case kMinusOneOne: {
        *sourcefile << "       " << sName << "_act[indNeurone++] = tanh(sum);" << endl;
        break;
    }
    default: {
        *sourcefile << "       " << sName << "_act[indNeurone++] = 1.0 / ( 1.0 + exp( -1.0 * sum ));" << endl;
    }
    }


    if( bTrace )
        *sourcefile << "       cout << \"  H1 sum(\" << sum << \") -- Hidden(\" << " << sName << "_act[indNeurone-1]" << "<< \")\" << endl;" << endl;
        *sourcefile << "    }" << endl;



    *sourcefile << "  indFirstSource  = indLastSource;" << endl;
    *sourcefile << "  indLastSource   += Hidden[ilayer];" << endl;

    if( bTrace )
        *sourcefile << "     cout << \"  Layer (\" << ilayer+1 <<  \"/\" <<  nLayers << \") indFirstSource (\" << indFirstSource << \") indLastSource (\" << indLastSource << \")\" << endl;" << endl;
    *sourcefile << "  }" << endl;




    // Output Layer
    *sourcefile << endl;
    *sourcefile << "  // --- Output Layer" << endl;
    if( bTrace )
        *sourcefile << "     cout << \"  Output Layer - ndFirstSource (\" << indFirstSource << \") indLastSource (\" << indLastSource << \")\" << endl;" << endl;
    *sourcefile << "  for (int member = 0; member < nOuput; member++) {" << endl;
    *sourcefile << "    sum = " << sName << "_valW[CrtW++];" << endl;
    *sourcefile << "    for (int source = indFirstSource; source < indLastSource; source++) {" << endl;
    if ( bTrace )
        *sourcefile << "      cout << \"   *** OL CrtW[\" << CrtW+1  << \"]\" << endl; " << endl;
    *sourcefile << "      sum += " << sName << "_act[source] * " << sName << "_valW[CrtW++];" << endl;
    *sourcefile << "    }" << endl;

    switch (_nProlemType) {
    case kRegression:
        *sourcefile << "    " << sName << "_act[indNeurone++] = sum;" << endl;
        switch (_nNormType)
        {
        case kMinusOneOne:
        {
            *sourcefile << "    res[member] = " << sName
                        << "_minOuput[member] + 0.5 * ( " << sName
                        << "_maxOuput[member] - " << sName
                        << "_minOuput[member] ) * ( sum + 1.0);" << endl;
            break;
        }
        default:
        {
            *sourcefile << "    res[member] = " << sName << "_minOuput[member] + "
                        << sName << "_maxOuput[member] * sum;" << endl;
            break;
        }
        }
        break;
    case kClassification:
        *sourcefile << "    " << sName << "_act[indNeurone] = 1.0 / ( 1.0 + exp( -1.0 * sum ));" << endl;
        *sourcefile << "    res[member] = " << sName
                    << "_minOuput[member] + ( " << sName
                    << "_maxOuput[member] - " << sName
                    << "_minOuput[member] ) * " <<  sName << "_act[indNeurone++];" << endl;

        break;
    }
    if ( bTrace )
        *sourcefile << "    cout << \"   *** member[\" << member  << \"] sum[\" << sum  << \"] res[\" << res[member]  << \"]\" << endl; " << endl;
    *sourcefile << "  }" << endl;
}



///////////////////////////////
////  Export Fortran     //////
///////////////////////////////
void TANNModeler::exportModelFortran(const char* file, const char* name,
        Option_t *option)
{
    TString sFile = file;
    sFile.ReplaceAll(".f", "");

    exportFunction((const char *) "for", sFile.Data(), name, option);
}

void TANNModeler::exportModelFortran(std::ofstream * sourcefile) const
{
    TString sName = _sFunctionName; //  "TANN" _sFunctionName
    Int_t  member, source, CrtW;
    // Int_t i, member, source, CrtW, nCrt;
    // 	nCrt = 0; i = 0;
    // int nVar = _tds->getNAttributes();
    // inputs
    URANIE::DataServer::TAttribute *att;


    // product and time
    time_t seconds;
    time(&seconds);

    *sourcefile << setprecision(14);

    *sourcefile << endl;
    *sourcefile << "C --- Neural Networks specifications" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_W(" << _nWeight << ")" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_act(" << _nNeurons << ")" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_minInp(" << _nInput << ")" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_maxInp(" << _nInput << ")" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_minOut(" << _nOutput << ")" << endl;
    *sourcefile << "      DIMENSION val_" << sName << "_maxOut(" << _nOutput << ")" << endl;
    *sourcefile << "      INTEGER nLayers" << endl;
    *sourcefile << "      INTEGER nHidden" << endl;
    *sourcefile << "      INTEGER Hidden(" <<  _nHiddenLayer << ")" << endl;
    *sourcefile << "      DATA Hidden/" <<  _sHidden << "/" << endl;
    *sourcefile << "      INTEGER indNeurone" << endl;
    *sourcefile << "      INTEGER indFirstSource" << endl;
    *sourcefile << "      INTEGER indLastSource" << endl;
    *sourcefile << "      INTEGER indCrtW" << endl;
    *sourcefile << "      DOUBLE PRECISION val_sum" << endl;

    TString sTitle = GetTitle();
    if ( _blog )
        cout << " *** sTitle[" << sTitle << "]" << endl;

    if ( sTitle.Contains("(WITHOUT_PETUM)")) {
        *sourcefile << endl;
        *sourcefile << "C --- Input Values" << endl;
        for (Int_t ivar = 0; ivar < _nInput; ivar++) {
            att = (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(ivar);
            *sourcefile << "      Xvar(" << ivar+1 << ") = " << att->GetName() << endl;
        }
    }

    // Normalisation des donnees
    *sourcefile << endl;
    *sourcefile << "C --- Pretraitment of the inputs" << endl;
    for (Int_t ivar = 0; ivar < _nInput; ivar++)
        *sourcefile << "      val_" << sName << "_minInp(" << ivar+1 << ") = " << _vecMinValues(ivar + 1) << "D0" << endl;
    *sourcefile << endl;

    for (Int_t ivar = 0; ivar < _nInput; ivar++)
        *sourcefile << "      val_" << sName << "_maxInp(" << ivar+1 << ") = " << _vecMaxValues(ivar + 1) << "D0" << endl;
    *sourcefile << endl;

    *sourcefile << "C --- Pretraitment of the outputs" << endl;
    for (Int_t ivar = _nInput; ivar < _nInput+_nOutput; ivar++)
        *sourcefile << "      val_" << sName << "_minOut(" << ivar+1-_nInput << ") = " << _vecMinValues(ivar + 1) << "D0" << endl;
    *sourcefile << endl;

    for (Int_t ivar = _nInput; ivar < _nInput+_nOutput; ivar++)
        *sourcefile << "      val_" << sName << "_maxOut(" << ivar+1-_nInput << ") = " << _vecMaxValues(ivar + 1) << "D0" << endl;
    *sourcefile << endl;

    // Vecteur des poids
    *sourcefile << endl;
    *sourcefile << "C --- Values of the weights" << endl;
    for (int i = 0; i < _nWeight; i++)
        *sourcefile << "      val_" << sName << "_W(" << i+1 << ") = " << _vecWeights(i + 1) << endl;

    // constantes
    *sourcefile << endl;
    *sourcefile << "C --- Other constantes" << endl;
    *sourcefile << "      nInput = " << _nInput << endl;
    *sourcefile << "      nOuput = " << _nOutput << endl;
    *sourcefile << "      nLayers = " << _nHiddenLayer << endl;
    *sourcefile << "      nHidden = " << _nHidden << endl;
    *sourcefile << "      indNeurone = 1" << endl;
    *sourcefile << "      indFirstSource  = 1" << endl;
    *sourcefile << "      indLastSource   = 1" << endl;
    *sourcefile << "      indCrtW = 1" << endl;

    // Input Layer
        *sourcefile << endl;
    *sourcefile << "C --- Input Layer" << endl;
    *sourcefile << "      DO 10 IA = 1, nInput" << endl;
    *sourcefile << "        val_" << sName << "_act(indNeurone) = " << endl;
    switch (_nNormType) {
        case kMinusOneOne: {
            *sourcefile << "     &   2.D0 * (Xvar(IA) - val_" << sName << "_minInp(IA))" << endl;
            *sourcefile << "     &   / (val_" << sName << "_maxInp(IA) - val_" << sName << "_minInp(IA))" << endl;
            *sourcefile << "     &   - 1.D0" << endl;
            break;
        }
        default: {
            *sourcefile << "     &   (Xvar(IA) - (val_" << sName << "_minInp(IA)))" << endl;
            *sourcefile << "     &   / val_" << sName << "_maxInp(IA) " << endl;
        }
        }
    *sourcefile << "        indNeurone = indNeurone + 1" << endl;
    *sourcefile << "   10 CONTINUE" << endl;

    // the Hidden layers
        *sourcefile << endl;
    *sourcefile << "C --- Hidden Layers" << endl;
    *sourcefile << "      indFirstSource  = 1" << endl;
    *sourcefile << "      indLastSource   =  "<< _nInput <<endl;
    *sourcefile << "      DO 50, IL = 1, nLayers" << endl;
    *sourcefile << "C --- current hidden Layer" << endl;
    *sourcefile << "        DO 40, IM = 1, Hidden(IL)" << endl;
    *sourcefile << "          val_sum = val_" << sName << "_W(indCrtW)" << endl;
    *sourcefile << "          indCrtW = indCrtW + 1" << endl;
    *sourcefile << "          DO 30, IS = indFirstSource, indLastSource" << endl;
    *sourcefile << "            val_sum = val_sum + " << endl;
    *sourcefile << "     &        val_" << sName << "_act(IS) * val_" << sName << "_W(indCrtW)" << endl;
    *sourcefile << "            indCrtW = indCrtW + 1" << endl;
    *sourcefile << "   30     CONTINUE" << endl;
    *sourcefile << "        val_" << sName << "_act(indNeurone) = " << endl;
    switch (_nNormType) {
        case kMinusOneOne: {
            *sourcefile << "     & 2.D0 / (1.D0 + DEXP(-2.D0 * val_sum )) - 1.D0" << endl;
            break;
    }
        default: {
            *sourcefile << "     & 1.D0 / (1.D0 + DEXP(-1.D0 * val_sum))" << endl;
            break;
    }
    }
    *sourcefile << "        indNeurone = indNeurone + 1" << endl;
    *sourcefile << "   40   CONTINUE" << endl;
    *sourcefile << "      indFirstSource  = indLastSource  + 1" << endl;
    *sourcefile << "      indLastSource   = indFirstSource + Hidden(IL)" << endl;
    *sourcefile << "   50 CONTINUE" << endl;

    // Output Layer
    *sourcefile << endl;
    *sourcefile << "C --- Output Layer" << endl;
    *sourcefile << "      DO 80, IM = 1, nOuput" << endl;
    *sourcefile << "        val_sum = val_" << sName << "_W(indCrtW)" << endl;
    *sourcefile << "        indCrtW = indCrtW + 1" << endl;
    *sourcefile << "        DO 70, IS = indFirstSource, indLastSource" << endl;
    *sourcefile << "          val_sum = val_sum + " << endl;
    *sourcefile << "     &      val_" << sName << "_act(IS) * val_" << sName << "_W(indCrtW)" << endl;
    *sourcefile << "          indCrtW = indCrtW + 1" << endl;
    *sourcefile << "   70   CONTINUE" << endl;


    switch (_nProlemType) {
    case kRegression:
        *sourcefile << "      val_" << sName << "_act(indNeurone) = val_sum" << endl;
    switch (_nNormType)
    {
    case kMinusOneOne:
    {
            *sourcefile << "      Yvar(IM) = val_" << sName << "_minOut(IM) + 0.5D0" << endl;
            *sourcefile << "     & * ( val_" << sName << "_maxOut(IM) - val_" << sName << "_minOut(IM) )" << endl;
            *sourcefile << "     & * ( val_" << sName << "_act(indNeurone) + 1.D0)" << endl;
        break;
    }
    default:
    {
            *sourcefile << "      Yvar(IM) = val_" << sName << "_minOut(IM)" << endl;
            *sourcefile << "     & + ( val_" << sName << "_maxOut(IM) - val_" << sName << "_minOut(IM) )" << endl;
            *sourcefile << "     & * val_" << sName << "_act(indNeurone)" << endl;
           break;
    }
    }
        break;
    case kClassification:
        *sourcefile << "      val_" << sName << "_act(indNeurone) = " << endl;
        *sourcefile << "     & 1.D0 / (1.D0 + DEXP(-1.D0 * val_sum))" << endl;
        *sourcefile << "      Yvar(IM) = val_" << sName << "_minOut(IM)" << endl;
        *sourcefile << "     & + ( val_" << sName << "_maxOut(IM) - val_" << sName << "_minOut(IM) )" << endl;
        *sourcefile << "     & * val_" << sName << "_act(indNeurone)" << endl;

        break;
    }
    *sourcefile << "      indNeurone = indNeurone + 1" << endl;
    *sourcefile << "   80 CONTINUE" << endl;
}


///////////////////////////////
////      EQM       ///////////
///////////////////////////////
Double_t TANNModeler::getEQM(Option_t *option)
{

    TString opt = option;
    opt.ToLower();

    try
    {
        if (opt.Contains("learn")) {
            return _dLearn;
        } else {
            if (opt.Contains("test")) {
                return _dTest;
            } else {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                             Form("Warning TANNModeler::getEQM the option [%s] is not recognized", opt.Data()));
            }
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    Double_t eqm_crt = -1.0;
    Int_t ndeb = 0, nrun = _nApp;
    if (opt.Contains("test")) {
        ndeb = _nApp;
        nrun = _nTest;
    }

    InteractiveTESTANN(ndeb, ndeb + nrun, _vecWeights, eqm_crt);
    if (_blog)
        cout << " ** Option[" << option << "] :: ndeb[" << ndeb << "] nrun["
             << nrun << "] eqm_crt[" << eqm_crt << "]" << endl;
    return eqm_crt;
}

///////////////////////////////
////      Log       ///////////
///////////////////////////////
void TANNModeler::printLog(Option_t *option)
{
    cout << endl << "************************************" << endl;
    cout << "*** TANNModeler::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **       _sArchi[" << _sArchi << "]" << endl;
    cout << " **        _nSeed[" << _nSeed << "]" << endl;
    cout << " ** _dWeightDecay[" << _dWeightDecay << "]" << endl;
    switch (_nNormType)
    {
    case kCR:
    {
        cout << " ** _nNormType[CenterReduct]" << endl;
    }
        break;
    case kMinusOneOne:
    {
        cout << " ** _nNormType[kMinusOneOne]" << endl;
    }
        break;
    case kZeroOne:
    {
        cout << " ** _nNormType[kZeroOne]" << endl;
    }
        break;
    }
    switch (_nProlemType) {
    case kRegression:
        cout << " ** _nProlemType[Regression]" << endl;
        break;
    case kClassification:
        cout << " ** _nProlemType[Classification]" << endl;
        break;
    }
    if (_vecWeights.size() > 0)
    {
        cout << " ** Weights [";
        for (int i = 0; i < _nWeight; i++)
            cout << i + 1 << "::" << _vecWeights(i + 1) << " ,";
        cout << "]" << endl;
    }
    cout << "************************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

