//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TANisp.cxx
 \author 
 \date   
 
 \brief Implementation of the class URANIE::Modeler::TAnisp
 */
#include <sstream>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
// #include <stdexcept>
#include <exception>

// ROOT
#include "Rtypes.h"
#include "TH1.h"
#include "TGraph.h"
#ifdef ROOT5
#include "Api.h"
#endif
#include "TString.h"
#include "TROOT.h"
#include "TSystem.h"

// URANIE
#include "TAnisp.h"
#include "TAttributeFormula.h"
#include "TPossibilityAttribute.h"

// anisp
#include "anisp.hpp"
#include "nisp_pc.h"

URANIE::Modeler::TAnisp *gUranieTAnisp;

ClassImp (URANIE::Modeler::TAnisp)

namespace URANIE
{
namespace Modeler
{
//Constructor  from a TDataServer object and a void *
#ifdef ROOT5
TAnisp::TAnisp(TDataServer *tds, void *analyticalFunction, TString sinput,
#else
TAnisp::TAnisp(TDataServer *tds, void (*analyticalFunction)(Double_t*, Double_t*), 
        const char *fname, TString sinput,
#endif
        TString soutput, Option_t *option) :
  TModeler(tds, sinput, option), 
  _drawingGarbageCollector(NULL), _tAnalyticalFunction(NULL), _tlauncher(NULL), _sinput(sinput), _soutput(soutput),
#ifdef ROOT5
  _analyticalFunction(NULL), _sFunctionName("")
#else
  _sFunctionName(fname)
#endif
{
    cout << " *** constructor TAnisp with function void[" << analyticalFunction << "]" << endl;
    try
    {

        if (analyticalFunction == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::constructor ERROR : the pointer of \"analyticalFunction\" is empty");

        _analyticalFunction = analyticalFunction;
        _nCible = TAnisp::kAnalyticalFunction;
        initialize();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    cout << " ***** end constructor " << endl;

} // End constructor

//Constructor  from a TDataServer object and a function name
TAnisp::TAnisp(TDataServer *tds, const char* analyticalFunction, TString sinput,
	       TString soutput, Option_t *option) :
  TModeler(tds, sinput, option), 
//         TModeler(tds,  sinput, soutput, option),
  _drawingGarbageCollector(NULL), _tAnalyticalFunction(NULL), _tlauncher(NULL), _sinput(sinput), _soutput(soutput),
  _analyticalFunction(NULL), _sFunctionName(analyticalFunction)
{
    cout << " *** constructor TAnisp  with function name[" << analyticalFunction << "]" << endl;
    
    _nCible = TAnisp::kAnalyticalFunction;
    initialize();

    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    cout << " ***** end constructor " << endl;

} // End constructor

//Constructor  from a TDataServer object and a TCode object
TAnisp::TAnisp(TDataServer *tds, TCode* code, Option_t *option) :
        TModeler(tds,
                NULL != code ?
                        code->getInputVarNames() :
		 TString((tds->getAttribute(0))->GetName()), option),
	_drawingGarbageCollector(NULL), _tAnalyticalFunction(NULL), _tlauncher(NULL), _analyticalFunction(NULL), _sFunctionName("")
{
    cout << " *** constructor TAnisp with Code";
    if (code != NULL)
        cout << "[" << code->GetName() << "]" << endl;
    else
        cout << " == NULL" << endl;

    try
    {

        if (code == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::constructor ERROR : the pointer of \"code\" is empty");
        if (tds != code->getTDataServer())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::constructor ERROR : The tds and the code->_tds are not the same!");

        _sinput = code->getInputVarNames();
        _soutput = code->getOutputVarNames();
        Int_t nOutputAttribute = code->getNOutputAttributes();
        //Int_t nInputAttributes = code->getNInputAttributes();

        if (nOutputAttribute != 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::constructor ERROR : invalid number of output (must be 1)");

        _Code = code;
        _nCible = TAnisp::kCode;
        initialize();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    cout << " ***** end constructor " << endl;

} // End constructor

//Destructor
TAnisp::~TAnisp()
{

  if( _drawingGarbageCollector != NULL)
    delete _drawingGarbageCollector;
  
  if (_tlauncher != NULL)
    delete _tlauncher;
  if (_tAnalyticalFunction != NULL)
    delete _tAnalyticalFunction;

  if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
        if (_tErrorIndicator != NULL)
            delete _tErrorIndicator;  
        if (_tIntegrationIndices != NULL)
            delete _tIntegrationIndices;
        if (_tdsAnisp != NULL)
            delete _tdsAnisp;
    }
  
  if (_tpc != NULL)
    {
      if((_tpc->_pc)->no == 0)
        {
	  free_dmatrix((_tpc->_pc)->beta,(_tpc->_pc)->ny +1);
	  free_dmatrix((_tpc->_pc)->phi,(_tpc->_pc)->nx +1);
	  free_dvector((_tpc->_pc)->psi);
	  free_imatrix((_tpc->_pc)->indmul,(_tpc->_pc)->p +1);
        }
      if((_tpc->_pc)->np == 0)
	free_dmatrix((_tpc->_pc)->target, 1);
      delete _tpc;
    }
  if (_nisp != NULL)
    {
      delete _nisp;   
    }
    
  // if(_blog)
  //   {
  //     if(_hist1 != NULL)
  // 	delete _hist1;
  //     if(_hist2 != NULL)
  // 	delete _hist2;
  //   }

}

// Initialize()
void TAnisp::initialize()
{

    TAttribute* att;
    for (Int_t i_input = 0; i_input < _tds->getNAttributes(); i_input++)
    {
        att = _tds->getAttribute(i_input);
        if (att->isInput())
        {
            if (dynamic_cast<URANIE::DataServer::TPossibilityAttribute *>(att))
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            "TAnisp::constructor ERROR : invalid input attribute");
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    throw ue;
                    exit(-1);
                }
            }

        }
    }

    _tdsAnisp = new TDataServer(Form("__anisp__%s__", _tds->GetName()),
            _tds->GetTitle());
    for (Int_t k = 0; k < _tds->getNAttributes(); k++)
    {
        _tdsAnisp->addAttribute(_tds->getAttribute(k));
    }

    _tdsAnisp->createTuple();

    // TODO mode force
    //_bdrawProgressBar = kFALSE;
    if (_nCible == TAnisp::kCode)
    {
        if (!_tds->isAttribute(_soutput))
        {
            // as nY=1 the only output attribute is the first attribute of the
            // only one output file
            _tds->addAttribute( _Code->getAttribute(0,0));
            (_tds->getAttribute(_soutput))->setOutput();
        }

	_Code->setTDataServer( _tdsAnisp );
        _tlauncher = new TLauncher(_tdsAnisp, _Code);
        _tlauncher->setDrawProgressBar(_bdrawProgressBar);
        _tAnalyticalFunction = NULL;

    }
    else if (_nCible == TAnisp::kAnalyticalFunction)
    {
        if (0 == _soutput.Length())
        {
	    if(_analyticalFunction!=NULL)
	    {
#ifdef ROOT5
		char *funcname = G__p2f2funcname(_analyticalFunction);
		_soutput = TString(funcname);
#else
            _soutput = _sFunctionName;        
#endif
	    }
	    else
	    {
		_soutput = _sFunctionName;
	    }
        }
        if ((0 == _tds->getNOutputAttributes())
                || (!_tds->isAttribute(_soutput)))
        {
            _tds->addAttribute(new TAttribute(_soutput));
            _tds->getAttribute(_soutput)->setOutput();
	    _tds->getAttribute(_soutput)->delShare();
        }
//         if(! _tds->getAttribute(_soutput)->isOutput())
//         {
//         _tds->getAttribute(_soutput)->setOutput();
//         }

        if(_analyticalFunction!=NULL)
            _tAnalyticalFunction = new TLauncherFunction(_tdsAnisp,
							 _analyticalFunction, _sinput, _soutput);
        else
	        _tAnalyticalFunction = new TLauncherFunction(_tdsAnisp,
							 _sFunctionName.Data(), _sinput, _soutput);
        _tAnalyticalFunction->setDrawProgressBar(_bdrawProgressBar);
        _Code = NULL;
        _tlauncher = NULL;
    }
    //build the group the stochastic variables
//     TAttribute * att = 0;
    att = 0;
    TStochasticAttribute *stoatt;
    for (Int_t i = 0; i < _tds->getNInputAttributes(); i++)
    {
        att = _tds->getAttribute(i);
        stoatt = dynamic_cast<TStochasticAttribute *>(att);
        if (NULL != stoatt)
        {
            try
            {
                switch (stoatt->getLawType())
                {
                case TStochasticAttribute::kNormal:
                {
                    _sLaws = _sLaws + "n";
                    break;
                }
                case TStochasticAttribute::kUniform:
                {
                    _sLaws = _sLaws + "u";
                    break;
                }
                default:
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            "TAnisp::constructor ERROR : law unknown");
                    break;
                }
                } //end switch
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
		delete _tAnalyticalFunction;
		delete _tlauncher;
		delete _tdsAnisp;
                throw ue;
                exit(-1);
            }
        }
    } //end for 

    _ninputs = 0;
    for (Int_t i_att = 0; i_att < _tds->getNInputAttributes(); i_att++)
    {
        if (NULL
                != dynamic_cast<TStochasticAttribute *>(_tds->getAttribute(
                        i_att)))
        {
            _ninputs += 1;
        }
    }

    _dTolerance = 0.001;
    _nNumberMaxOfSimulations = 1000;
    _nKMin = 1;
    _nGreaterIndice = 6;
    _nMaxIndices = 1000;
    _nDegreeMax = 15;
    _sRootFilename = "fichierAnisp_";
    _nIterationNumber = 0;
    _blog = kFALSE;
    _tpc = NULL;
    _nisp = NULL;
    _nNumberOfLaunching = 0;
    _nNumberOfPoints = _tds->getNPatterns();
    // _hist1 = NULL;
    // _hist2 = NULL;

    _sLaunchNumber = TString("__uranie__anisp__");

    try
    {
        if (_tds->getNPatterns() > 0)
        {
            if (!_tds->isAttribute(_sLaunchNumber))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        "TAnisp::constructor - Error: TDataServer contained patterns not computed using the Anisp library");
            else
                _nUranieAnisp = (Int_t)(
                        _tds->getValue(_sLaunchNumber,
                                _tds->getNPatterns() - 1));
//         TAttribute * att = new TAttribute(LaunchNumber);
//         att->setOrigin(URANIE::DataServer::TAttribute::kInternal);
//         _tds->addAttribute(att);

        }
        else
        {
            _nUranieAnisp = 0;
            TAttribute *attlaunch = new TAttribute(_sLaunchNumber);
            attlaunch->setOrigin(URANIE::DataServer::TAttribute::kInternal);
	    attlaunch->delShare();
            _tds->addAttribute(attlaunch);
            _tds->createTuple();
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete _tAnalyticalFunction;
	delete _tlauncher;
	delete _tdsAnisp;
        throw ue;
        exit(-1);
    }

    _tErrorIndicator = new TNtupleD("Error Indicators", "Error Indicators",
            "n:error_indicator");

    _tIntegrationIndices = new TNtupleD("Indices", "Integration indices",
            _sinput);

    // _myCanvas = NULL;

    gUranieTAnisp = this;

    att = 0;
    stoatt = 0;
    delete att;
    delete stoatt;
}

// print Log
void TAnisp::printLog(Option_t *option)
{
    TString name;
    cout << " ** TAnisp::printLog[" << option << "]" << endl;
    cout << "      Stochastic variables:  " << _sinput << endl;
    cout << "      output variable:  " << _soutput << endl;
    _tds->printLog();
    cout << "fin of TAnisp::printLog[" << option << "]" << endl;

}

// Set the KMin
void TAnisp::setKMin(Int_t kmin)
{
    try
    {
        if (kmin < 1 || kmin > _ninputs)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setKMin - Error: invalid value for _nKMin (Need to be > 0 and <= _nX)");
        _nKMin = kmin;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the tolerance
void TAnisp::setTolerance(Double_t tol)
{
    try
    {
        if (tol <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setTolerance - Error: invalid value for _dTolerance (Need to be > 0)");
        _dTolerance = tol;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the _nNumberMaxOfSimulations
void TAnisp::setNumberMaxOfSimulations(Int_t nmos)
{
    try
    {
        if (nmos < 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setNumberMaxOfSimulations - Error: invalid value for _nNumberMaxOfSimulations (Need to be > 0)");
        _nNumberMaxOfSimulations = nmos;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the greater value possible for an indice: _nGreaterIndice
void TAnisp::setGreaterIndice(Int_t greatInd)
{
    try
    {
        if (greatInd < 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setGreaterIndice - Error: invalid value for _nGreaterIndice (Need to be > 0)");
        TAttribute * att = 0;
        TStochasticAttribute *stoatt;
        for (Int_t i = 0; i < _tds->getNInputAttributes(); i++)
        {
            att = _tds->getAttribute(i);
            stoatt = dynamic_cast<TStochasticAttribute *>(att);
            if (NULL != stoatt)
            {
                if (stoatt->getLawType() == TStochasticAttribute::kNormal)
                {
                    if (greatInd > 6)
                    {
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                "TAnisp::setGreaterIndice - Error: invalid value for _nGreaterIndice (Need to be < 7 when there is a normal law)");
                        break;
                    }
                }
                else
                {
                    if (greatInd > 12)
                    {
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                "TAnisp::setGreaterIndice - Error: invalid value for _nGreaterIndice (Need to be < 13 )");
                        break;
                    }
                }
            }

        } //end for
        _nGreaterIndice = greatInd;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the _nMaxIndices
void TAnisp::setMaxIndices(Int_t maxInd)
{
    try
    {
        if (maxInd < 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setMaxIndices - Error: invalid value for _nMaxIndices (Need to be > 0)");
        _nMaxIndices = maxInd;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the _nDegreeMax
void TAnisp::setDegreeMax(Int_t degMax)
{
    try
    {
        if (degMax < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setDegreeMax - Error: invalid value for _nDegreeMax (Need to be >= 0)");
        _nDegreeMax = degMax;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

// Set the _RootFilename
void TAnisp::setRootFilename(TString rootName)
{
    Bool_t IsEmpty = kTRUE;
    TString white = " ";
    try
    {
        if (rootName.IsNull())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setRootFilename - Error: invalid value for _sRootFilename (not an empty string)");

        for (Int_t i = 0; i < rootName.Length(); i++)
        {
            if (rootName[i] != white)
            {
                IsEmpty = kFALSE;
                break;
            }
        }
        if (IsEmpty)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setRootFilename - Error: invalid value for _sRootFilename (not an white string)");
//       if(rootName.Length()<8)
//       {
//         Int_t nn = 8 - rootName.Length();
//         for(Int_t i=0;i<nn;i++)
//         rootName = rootName + TString("_");
//         rootName += TString("_");
//       }
        _sRootFilename = rootName;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

}

// Set the _blog parameter
void TAnisp::setlog(Bool_t blog)
{
    if (blog)
    {
        setLog();
    //     if(_myCanvas != NULL)
    //     {
    //         delete _myCanvas;
    //         _myCanvas = NULL;
    //     }
    //     _myCanvas = new TCanvas();
    //     _myCanvas->Divide(2, 1);
    }
    else
    {
        unsetLog();
    //     if (_myCanvas)
    // 	  _myCanvas->~TCanvas();
    }
}

// Set all the Anisp parameters
void TAnisp::setAllAnispParameters(Double_t tol, Int_t kmin, Int_t nmos,
        Int_t greatInd, Int_t maxInd, Bool_t blog, Int_t degMax,
        TString rootName)
{
    setTolerance(tol);
    setKMin(kmin);
    setNumberMaxOfSimulations(nmos);
    setGreaterIndice(greatInd);
    setMaxIndices(maxInd);
    setlog(blog);
//     if (blog)
//         setLog();
//     else 
//         unsetLog();
    setDegreeMax(degMax);
    setRootFilename(rootName);

}

// Set most of the Anisp parameters
void TAnisp::setAnispParameters(Double_t tol, Int_t kmin, Int_t nmos,
        Int_t greatInd, Int_t maxInd, Bool_t blog)
{
    setTolerance(tol);
    setKMin(kmin);
    setNumberMaxOfSimulations(nmos);
    setGreaterIndice(greatInd);
    setMaxIndices(maxInd);
    setlog(blog);

}

// Set the Anisp calcul parameters
void TAnisp::setLightAnispParameters(Double_t tol, Int_t kmin, Int_t nmos)
{
    setTolerance(tol);
    setKMin(kmin);
    setNumberMaxOfSimulations(nmos);

}

void StepByStepOutput(Double_t * errIndicator,
        vector<Int_t *> integrationIndices)
{
    vector < Int_t > inputs;

    for (Int_t i_att = 0; i_att < gUranieTAnisp->_tds->getNInputAttributes(); i_att++)
    {
        if (NULL
                != dynamic_cast<TStochasticAttribute *>(gUranieTAnisp->_tds->getAttribute(
                        i_att)))
        {
            inputs.push_back(i_att);
        }
    }
    Int_t nx = inputs.size();
    
    
    vector<Double_t *> integInd;
//     Int_t nx = gUranieTAnisp->_tds->getNInputAttributes();
    Int_t ni = int(integrationIndices.size());
    Double_t * pti=NULL;
    
    for (Int_t i = 0; i < ni; i++)
    {
        pti = new Double_t[nx];
        for (Int_t j = 0; j < nx; j++)
        {
            pti[j] = integrationIndices[i][j];
        }

        integInd.push_back(pti);
    }
    gUranieTAnisp->EndIterationActions(errIndicator, integInd);

    // Clean
    for (int idel = 0; idel < (int) (integInd.size()); idel++)
    {
        delete[] integInd[idel];
    }
    integInd.clear();
}

void TAnisp::EndIterationActions(Double_t * errIndicator,
        vector<Double_t *> integrationIndices)
{
    _nIterationNumber += 1;
    Int_t numberOfIndices = integrationIndices.size();
    addErrorIndicator(errIndicator);
    for (Int_t i_ind = 0; i_ind < numberOfIndices; i_ind++)
        addIntegrationIndices(integrationIndices[i_ind]);

    if (_blog)
    {
      stringstream sstr; sstr << "TAnisp_Canvas_" << _drawingGarbageCollector->GetEntries() ;
      TCanvas *Can = new TCanvas(sstr.str().c_str(),sstr.str().c_str(),1);
      Can->Divide(2,1);
      Can->cd(1);
      _drawingGarbageCollector->Add(Can);
      //        _myCanvas->cd(1);
        Can->GetPad(1)->SetLogx(kTRUE);
        Can->GetPad(1)->SetLogy(kTRUE);
//         _tErrorIndicator->Draw("error_indicator:n","","entrylist");

        Int_t nErrInd = Int_t(_tErrorIndicator->GetEntries());
        Double_t* X = new Double_t[nErrInd];
        Double_t* Y = new Double_t[nErrInd];
        Double_t * terrind = _tErrorIndicator->GetArgs();

        for (int i = 0; i < nErrInd; i++)
        {
            _tErrorIndicator->GetEntry(i);
            X[i] = terrind[0];
            Y[i] = terrind[1];
        }
        TGraph* graph = new TGraph(nErrInd, X, Y);
	sstr.str(""); sstr<<"Anisp_graph_"<<_drawingGarbageCollector->GetEntries()<<endl;
	graph->SetName( sstr.str().c_str() );
	_drawingGarbageCollector->Add(graph);
        graph->Draw("AL*");

	//        _myCanvas->cd(2);
	Can->cd(2);
        Can->GetPad(2)->SetGrid();
        Can->GetPad(2)->SetFrameFillColor(33);
        Can->GetPad(2)->SetLogx(kFALSE);
        Can->GetPad(2)->SetLogy(kFALSE);

        Int_t nbin = 10; // number of bar charts
        Int_t nx = _ninputs; // number of input variables
        Int_t ninter = int(pow(2.0, Double_t(nx)) - 1); // number of interaction between variables
        if (ninter < nbin)
            nbin = ninter;

        Int_t nentries = Int_t(_tIntegrationIndices->GetEntries()); // number of indices
        if (nentries < nbin)
            nbin = nentries;
        Double_t * pts = _tIntegrationIndices->GetArgs();
        Int_t ** val = new Int_t*[ninter];
        // initialization to 0
        for (Int_t i_inter = 0; i_inter < ninter; i_inter++)
        {
            val[i_inter] = new Int_t[nx + 2];
            for (Int_t j = 0; j < nx + 2; j++)
                val[i_inter][j] = 0;
        }

        // computation of the sum and te max for each indice
        Double_t sum; // sum of indice
        Double_t num; // number of an indice
        Double_t max_ind; // max value of an indice
        for (Int_t i_entries = 0; i_entries < nentries; i_entries++)
        {
            _tIntegrationIndices->GetEntry(i_entries);
            sum = 0;
            num = 0;
            max_ind = 0;
            for (Int_t i = 0; i < nx; i++)
            {
                sum = sum + pts[i];
                if (pts[i] > 1)
                    num = num + pow(2.0, Double_t(i));
                if (pts[i] > max_ind)
                    max_ind = pts[i];
            }
            num = num - 1;

            if (num >= 0 && int(sum) > val[int(num)][0])
            {
                val[int(num)][0] = int(sum);
                val[int(num)][1] = int(max_ind);
                for (Int_t j = 2; j < nx + 2; j++)
                    val[int(num)][j] = int(pts[j - 2]);
            }
        }
        vector<Int_t*> bins;
        vector<Int_t*>::iterator ibins;
        bins.push_back(val[0]);
        Int_t a;
        //ranking of the indices by sum
        for (Int_t i_inter = 1; i_inter < ninter; i_inter++)
        {
            if (Int_t(bins.size()) < nbin)
            {
                a = 1;
                for (ibins = bins.begin(); ibins != bins.end(); ibins++)
                {
                    if ((*ibins)[0] < val[i_inter][0])
                    {
                        bins.insert(ibins, val[i_inter]);
                        a = 0;
                        break;
                    }
                }
                if (a)
                    bins.push_back(val[i_inter]);
            }
            else
            {
                for (ibins = bins.begin(); ibins != bins.end(); ibins++)
                {
                    if ((*ibins)[0] < val[i_inter][0])
                    {
                        bins.insert(ibins, val[i_inter]);
                        //Clean//////////////////
//                         if(bins.back())
//                             delete[] bins.back();
                        /////////////////////////
                        bins.pop_back();
                        break;
                    }
                }
            }
        }
        // if(hist1 != NULL)
        //     delete hist1;
        // if(hist2 != NULL)
        //     delete hist2;
	sstr.str(""); sstr<<"Anisp_somme_"<<_drawingGarbageCollector->GetEntries()<<endl;
        TH1F *hist1 = new TH1F(sstr.str().c_str(), "Integration indices", nbin, 0, nbin);
	sstr.str(""); sstr<<"Anisp_max_"<<_drawingGarbageCollector->GetEntries()<<endl;
        TH1F *hist2 = new TH1F(sstr.str().c_str(), "Integration indices", nbin, 0, nbin);
	_drawingGarbageCollector->Add(hist1);
	_drawingGarbageCollector->Add(hist2);
	
	
        Int_t nvar = 0;
        for (Int_t i_str = 0; i_str < _sinput.Length(); i_str++)
        {
            if (':' == _sinput[i_str])
                nvar += 1;
        }
        nvar += 1;
        TString* InVar = new TString[nvar];
        for (Int_t i = 0; i < nvar; i++)
            InVar[i] = TString("");
        Int_t num_var = 0;
        for (Int_t i_str = 0; i_str < _sinput.Length(); i_str++)
        {
            if (':' != _sinput[i_str])
            {
                InVar[num_var].Append(_sinput[i_str]);
            }
            else
            {
                num_var += 1;

            }

        }

        TString VarList;
        for (Int_t i_bin = 0; i_bin < nbin; i_bin++)
        {
            VarList = "";
            for (Int_t i = 0; i < nx; i++)
            {
                if (bins[i_bin][i + 2] > 1)
                    VarList += InVar[i] + ":";
            }
            if (VarList.Length() > 0)
            {
                VarList.Remove(VarList.Last(':'));
                hist1->Fill(VarList, bins[i_bin][0]);
                hist2->Fill(VarList, bins[i_bin][1]);
            }
        }

        hist1->SetBarWidth(0.45);
        hist1->SetBarOffset(0.1);
        hist1->SetFillColor(49);
        hist1->SetStats(kFALSE);
        hist1->SetMinimum(0);
        hist1->Draw("bar2");

        hist2->SetBarWidth(0.4);
        hist2->SetBarOffset(0.55);
        hist2->SetFillColor(50);
        hist2->SetStats(kFALSE);
        hist2->DrawCopy("bar2,same");

        TLegend *legend = new TLegend(0.55, 0.75, 0.9, 0.9);
	_drawingGarbageCollector->Add(legend);
        legend->AddEntry(hist1, "sum of indices", "f");
        legend->AddEntry(hist2, "max of indices", "f");
        legend->Draw();

        Can->Modified();
        Can->Update();

        // clean
        delete[] X;
        X = NULL;
        delete[] Y;
        Y = NULL;
        delete[] InVar;
        InVar = NULL;
        
//         for (int idel = 0; idel < (int) bins.size(); idel++)
//         {
//             delete[] bins[idel];
//         }
//         bins.clear();
        
        for (Int_t idel = 0; idel < ninter; idel++)
        {
            if(val[idel])
            {
                delete[] val[idel];
                val[idel] = NULL;
            }
        }
        delete[] val;

    }
}

TString TAnisp::getVariablesName(TString inputs)
{
    TString varname;
    if (inputs.Length() == 0)
        varname = _sinput + TString(":") + _soutput + TString(":")
                + _sLaunchNumber;
    else
        varname = inputs + TString(":") + _soutput + TString(":")
                + _sLaunchNumber;

    return varname;
}

void TAnisp::runAnisp()
{
    gUranieTAnisp = this;
    //Int_t nx = _tds->getNInputAttributes();
    TString ResultFile = getResultFile();
    TString StochasticFile = getStochasticFile();
    TString DataAnispFile = getDataAnispFile();
    TString ErrorIndicatorFile = getErrorIndicatorFile();
    TString IndicesFile = getIndicesFile();

    _nUranieAnisp = 1;
    switch ( _nCible ) {
        case TAnisp::kCode:
        {
            _tlauncher->setDrawProgressBar(getDrawProgressBar());
            break;
        }
        case TAnisp::kAnalyticalFunction:
        {
            _tAnalyticalFunction->setDrawProgressBar(getDrawProgressBar());
            break;
        }
    }
    cout << " *** EVAL with [bdrawProgressBar] = ";
    if (getDrawProgressBar() )
        cout << "TRUE";
    cout << endl;

    Anisp::anisp(functionLauncher, StepByStepOutput, _ninputs, _nGreaterIndice,
            _nMaxIndices, _dTolerance, _nKMin, std::string(_sLaws),
            _nNumberMaxOfSimulations, _nDegreeMax, false, _blog,
            std::string(ResultFile), std::string(StochasticFile),
            std::string(DataAnispFile), std::string(ErrorIndicatorFile),
            std::string(IndicesFile));
    setTPolynomialChaos();
    _tds->fillOthersAttributes();

}

void TAnisp::restartAnisp()
{
    gUranieTAnisp = this;

    //Int_t nx = _tds->getNInputAttributes();
    TString ResultFile = getResultFile();
    TString StochasticFile = getStochasticFile();
    TString DataAnispFile = getDataAnispFile();
    TString ErrorIndicatorFile = getErrorIndicatorFile();
    TString IndicesFile = getIndicesFile();

    Int_t numberOfPoints;
    Int_t i;
//     const char* file = string(getDataAnispFile()).c_str();
    TString data_file = getDataAnispFile();
	//DI347
    char *file = new char[data_file.Length() + 1];
    for (i = 0; i < data_file.Length(); i++)
        file[i] = data_file[i];
    file[data_file.Length()] = '\0';

    ifstream result(file, ios::in);
    if (result.fail())
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::restartAnisp ERROR : no data file");

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            delete[] file;
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    }
    result >> numberOfPoints;
    if (numberOfPoints < 1)
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::restartAnisp ERROR : no computed calcul");

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            delete[] file;
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    }

    _nUranieAnisp += 1;
    switch ( _nCible ) {
        case TAnisp::kCode:
        {
            _tlauncher->setDrawProgressBar(getDrawProgressBar());
            break;
        }
        case TAnisp::kAnalyticalFunction:
        {
            _tAnalyticalFunction->setDrawProgressBar(getDrawProgressBar());
            break;
        }
    }
    cout << " *** EVAL with [bdrawProgressBar] = ";
    if (getDrawProgressBar() )
        cout << "TRUE";
    cout << endl;

    Anisp::anisp(functionLauncher, StepByStepOutput, _ninputs, _nGreaterIndice,
            _nMaxIndices, _dTolerance, _nKMin, std::string(_sLaws),
            _nNumberMaxOfSimulations, _nDegreeMax, true, _blog,
            std::string(ResultFile), std::string(StochasticFile),
            std::string(DataAnispFile), std::string(ErrorIndicatorFile),
            std::string(IndicesFile));
    setTPolynomialChaos();
    _tds->fillOthersAttributes();
    
    // Clean
    delete[] file;

}

void TAnisp::launchCalculs(vector<Double_t *> ptExp)
{
    Int_t nx;
    Int_t ns = ptExp.size();
    Double_t * x;
    Double_t * x2 = NULL;
    Double_t ** xx = new Double_t*[ns];
    // Int_t count;
    vector < Int_t > inputs;
    for (Int_t i_att = 0; i_att < _tds->getNInputAttributes(); i_att++)
    {
        if (NULL
                != dynamic_cast<TStochasticAttribute *>(_tds->getAttribute(
                        i_att)))
        {
            inputs.push_back(i_att);
        }
    }
    if (_nCible == TAnisp::kCode)
    {
        nx = _tds->getNInputAttributes();
        x = new Double_t[nx + 2];
        for(Int_t i_ini=0;i_ini<nx+2;i_ini++)
            x[i_ini] = 0;
        //x2 = new Double_t[nx + 3];

	if( _tdsAnisp->isAttribute(_soutput) )
	  _tdsAnisp->delAttribute(_soutput);
	
        _tdsAnisp->createTuple();

        for (Int_t k = 0; k < ns; k++)
        {
            xx[k] = new Double_t[nx + 3];
            x[0] = k + 1; // iterateur
            xx[k][0] = x[0];

            for (Int_t i = 0; i < Int_t(inputs.size()); i++)
            {
                x[inputs[i] + 1] = ptExp[k][i];
            }
            try
            {
                changeOfVariable(x);
            } catch (URANIE::Exceptions::UErrorExceptions& )
            {
                delete[] x;
                x = NULL;
                //delete[] x2;
                x2 = NULL;
                for(Int_t kdel=0;kdel<=k;kdel++)
                {
                    delete[] xx[kdel];
                    xx[kdel] = NULL;
                }
                delete[] xx;
                xx = NULL;
                
//                 std::exception except;
                Anisp::anispException except;
                throw except;
                
                //throw ue;
                exit(-1);
            }
            for (Int_t i = 0; i < nx; i++)
            {
                xx[k][i + 1] = x[i + 1];
            }

            _tdsAnisp->getTuple()->Fill(x);

        }

        //Fill TAttributeFormula
        _tdsAnisp->fillOthersAttributes();
        //Launch the calculs
        _tlauncher->run();
	
        for (Int_t k = 0; k < ns; k++)
        {
            ptExp[k][_ninputs] = _tdsAnisp->getValue(_soutput, k);
            xx[k][nx + 1] = ptExp[k][_ninputs];
            xx[k][nx + 2] = (Double_t)(_nUranieAnisp);
            xx[k][0] = _nNumberOfPoints + k + 1;
            x2 = xx[k];

            _tds->getTuple()->Fill(x2);

        }

        // Clean
        delete[] x;
        x = NULL;
        delete[] x2;
        x2 = NULL;
    }
    else if (_nCible == TAnisp::kAnalyticalFunction)
    {
        if (_nNumberOfLaunching)
            _tdsAnisp->delAttribute(_soutput);

        nx = _tds->getNInputAttributes();
        x = new Double_t[nx + 1];
        //x2 = new Double_t[nx + 3];
        _tdsAnisp->createTuple();
        for (Int_t k = 0; k < ns; k++)
        {
            xx[k] = new Double_t[nx + 3];
            x[0] = k + 1; // iterateur
            xx[k][0] = x[0];

            for (Int_t i = 0; i < Int_t(inputs.size()); i++)
            {
                x[inputs[i] + 1] = ptExp[k][i];
            }
            try
            {
                changeOfVariable(x);
            } catch (URANIE::Exceptions::UErrorExceptions& )
            {
                delete[] x;
                x = NULL;
                //delete[] x2;
                x2 = NULL;
                for(Int_t kdel=0;kdel<=k;kdel++)
                {
                    delete[] xx[kdel];
                    xx[kdel] = NULL;
                }
                delete[] xx;
                xx = NULL;

//                 std::exception except;
                Anisp::anispException except;
                throw except;
                
                //throw ue;
                exit(-1);
            }
           

            for (Int_t i = 0; i < nx; i++)
            {
                xx[k][i + 1] = x[i + 1];
            }

            _tdsAnisp->getTuple()->Fill(x);

        }
        //Fill TAttributeFormula
        _tdsAnisp->fillOthersAttributes();
        //Launch the calculs
        _tAnalyticalFunction->run();
        for (Int_t k = 0; k < ns; k++)
        {
            ptExp[k][_ninputs] = _tdsAnisp->getValue(_soutput, k);
            xx[k][nx + 1] = ptExp[k][_ninputs];
            xx[k][nx + 2] = (Double_t)(_nUranieAnisp);
            xx[k][0] = _nNumberOfPoints + k + 1;
            x2 = xx[k];

            _tds->getTuple()->Fill(x2);
        }
        // Clean
        delete[] x;
        x = NULL;
        delete[] x2;
        x2 = NULL;
    }

    _nNumberOfPoints += ns;
    _nNumberOfLaunching += 1;
    
    //Clean
    for (Int_t k = 0; k < ns-1; k++)
    {
        if(xx[k])
        {
            delete[] xx[k];
            xx[k] = NULL;
        }
    }
    delete[] xx;
}

void functionLauncher(vector<Double_t *> ptEXp)
{
    gUranieTAnisp->launchCalculs(ptEXp);
}

void TAnisp::changeOfVariable(Double_t* x)
{
    Int_t nx = _tds->getNInputAttributes();
    TAttribute * att = 0;
    TStochasticAttribute *stoatt;
    for (Int_t i = 0; i < nx; i++)
    {
        att = _tds->getAttribute(i);
        stoatt = dynamic_cast<TStochasticAttribute *>(att);
        if (NULL != stoatt)
        {
            try
            {
                switch (stoatt->getLawType())
                {
                case TStochasticAttribute::kNormal:
                {
                    x[i + 1] = stoatt->getTheoreticalStdDev() * x[i + 1]
                            + stoatt->getTheoreticalMean();
                    break;
                }
                case TStochasticAttribute::kUniform:
                {
                    Double_t m = (stoatt->getUpperBound()
                            - stoatt->getLowerBound()) / 2.0;
                    x[i + 1] = m * x[i + 1] + stoatt->getLowerBound() + m;
                    break;
                }
                default:
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            "TAnisp::changeOfVariable ERROR : law unknown");

                    break;
                }
                } //end switch
            } catch (URANIE::Exceptions::UErrorExceptions & ue)
            {
                ue.printMessage();
                throw ue;
                exit(-1);
            }
        }
    }
}

void TAnisp::setTPolynomialChaos()
{
    Double_t coef, buff;
    Int_t num, i;
    //Int_t nx = _tds->getNInputAttributes();

    TString res_file = getResultFile();
    // DI347
	char * file = new char[res_file.Length() + 1];
    for (i = 0; i < res_file.Length(); i++)
        file[i] = res_file[i];
    file[res_file.Length()] = '\0';

    ifstream result(file, ios::in);

    if (result.fail()) 
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::setTPolynomialChaos ERROR : no file");

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            delete[] file;
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    }

    vector < Int_t > pc_nums;
    vector < Double_t > pc_coefs;
    vector < Int_t > pc_degrees;
    Int_t deg, pc_glob_deg;
    pc_glob_deg = 0;
    while (result.good())
    {

        result >> num;
        if (result.fail())
            break;
        pc_nums.push_back(num);
        result >> setprecision(24) >> coef;
        pc_coefs.push_back(coef);
        deg = 0;
        for (i = 0; i < _ninputs; i++)
        {
            result >> buff;
            deg = deg + (Int_t)(buff);
        }
        pc_degrees.push_back(deg);
        if (deg > pc_glob_deg)
            pc_glob_deg = deg;

    }

    if(_nisp != NULL)
    {
        delete _nisp;
        _nisp = NULL;
    }
    try{
    _nisp = new TNisp(_tds);
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        delete[] file;
        throw ue;
        exit(-1);
    }
    _nisp->_ns = _tds->getNPatterns();
    if(_tpc != NULL)
    {
        delete _tpc;
        _tpc = NULL;
    }
    _tpc = new TPolynomialChaos(_tds, _nisp);
    _tpc->setDegree(pc_glob_deg);
    for (i = 0; i < (_tpc->_pc)->p + 1; i++)
        _tpc->setCoefficient(1, i, 0);
    for (i = 0; i < Int_t(pc_degrees.size()); i++)
    {
        _tpc->setCoefficient(1, pc_nums[i], pc_coefs[i]);
    }
    (_tpc->_pc)->SetAnova();

    result.close();
    
    // Clean
    delete[] file;

}

TPolynomialChaos TAnisp::getTPolynomialChaos()
{
    if (_tpc == NULL)
    {
        try
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TAnisp::getTPolynomialChaos ERROR : no polynomial");
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    }
    return _tpc[0];
}

void TAnisp::exportModelCplusplus(std::ofstream * sourcefile) const
{
    Double_t coef, buff;
    Int_t num, i, j;
    Int_t nx = _tds->getNInputAttributes();

    TAttribute * att = NULL;
    TStochasticAttribute *stoatt = NULL;
    j = 0;
    for ( i = 0; i < nx; i++)
    {
        att = _tds->getAttribute(i);
        stoatt = dynamic_cast<TStochasticAttribute *>(att);
        if (NULL != stoatt)
        {

            switch (stoatt->getLawType())
            {
            case TStochasticAttribute::kNormal:
            {
                *sourcefile << setprecision(20) << "param[" << j
                        << "] = (param[" << j << "]-("
                        << stoatt->getTheoreticalMean() << "))/"
                        << stoatt->getTheoreticalStdDev() << ";" << endl;
                j += 1;
                break;
            }
            case TStochasticAttribute::kUniform:
            {
                Double_t m = (stoatt->getUpperBound() - stoatt->getLowerBound())
                        / 2.0;
                *sourcefile << setprecision(20) << "param[" << j
                        << "] = (param[" << j << "]-("
                        << stoatt->getLowerBound() << "))/" << m << " -1.0;"
                        << endl;
                j += 1;
                break;
            }
            default:
            {

                break;
            }
            } //end switch
        }
    }

    string tmp_str = string(this->getResultFile());
    const char* file = tmp_str.c_str();
//     const char* file = string(getResultFile()).c_str();
    ifstream result(file, ios::in);
    if (result.fail())
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						 Form("TAnisp::exportModelCplusplus ERROR : no file [%s] found", file));
    vector < Int_t > pc_nums;
    vector < Double_t > pc_coefs;
    vector<Int_t *> pc_degrees;
    Int_t deg, pc_max_deg, npol;
    Int_t* degrees;
    pc_max_deg = 0;
    while (!result.eof())
    {
        result >> num;
        if (result.fail())
            break;
        pc_nums.push_back(num);
        result >> coef;
        pc_coefs.push_back(coef);
        deg = 0;
        degrees = new Int_t[_ninputs];
        for (i = 0; i < _ninputs; i++)
        {
            result >> buff;
            degrees[i] = (Int_t)(buff);
            if (deg < (Int_t)(buff))
                deg = (Int_t)(buff);
        }
        pc_degrees.push_back(degrees);
        if (deg > pc_max_deg)
            pc_max_deg = deg;
    }

    npol = pc_coefs.size();
    *sourcefile << "//**********npol = " << npol << " *********************"<< endl;

    *sourcefile << "double pc_coefs[" << npol << "] = {" ;
    for (i = 0; i < npol; i++)
      *sourcefile << setprecision(20) << pc_coefs[i] << ((i!=(npol-1))?",":"");
    *sourcefile << "};" << endl;
      
    *sourcefile << "int pc_degrees[" << npol << "]["<<_ninputs<<"]={";
    for (i = 0; i < npol; i++)
      {
        for (j = 0; j < _ninputs; j++)
        {
	  *sourcefile << setprecision(20) << pc_degrees[i][j] << ((i==(npol-1)&&(j==_ninputs-1))?"":",");
        }
    }
    *sourcefile << "};" << endl;
    
    Anisp::SPC spc(2, pc_max_deg, 'l');
    vector<Double_t *> legendre; //les coefficients de la famille des polynomes de legendre
    vector<Double_t *> hermite; //les coefficients de la famille des polynomes de hermite
    legendre.reserve(30);
    hermite.reserve(30);
    Double_t* le;
    Double_t* he;

    for (i = 0; i <= pc_max_deg; i++)
    {
        le = new Double_t[i + 1];
        legendre.push_back(le);
        he = new Double_t[i + 1];
        hermite.push_back(he);
    }

    *sourcefile << "const char* law = \"" << _sLaws.Data() << "\";" << endl;

    *sourcefile << "vector<double> legendre[" << legendre.size()<< "];" << endl;
    *sourcefile << "vector<double> hermite[" << hermite.size() << "];"  << endl;
//     

    //---initialization of legendre --------
    // cout<<"********************************************************************"<<endl;
    Anisp::polynome * phi = NULL;
    phi = Anisp::build_legendre(pc_max_deg);
    Double_t * norme;// = new Double_t[pc_max_deg + 1];
    norme = Anisp::norme_legendre(pc_max_deg);

    for (i = 0; i <= pc_max_deg; i++)
    {
        for (j = 0; j <= i; j++)
        {
            legendre[i][j] = phi[i].coef[j] / sqrt(norme[i]);
            *sourcefile << setprecision(20) << "legendre[" << i
                    << "].push_back(" << legendre[i][j] << ");" << endl;
        }
    }

    //---initialization of hermite --------
// cout<<"********************************************************************"<<endl;
    delete[] phi;
    delete[] norme;
    phi = Anisp::build_hermite(pc_max_deg);
    norme = Anisp::norme_hermite(pc_max_deg);

    for (i = 0; i <= pc_max_deg; i++)
    {
        for (j = 0; j <= i; j++)
        {
            hermite[i][j] = phi[i].coef[j] / sqrt(norme[i]);
            *sourcefile << setprecision(20) << "hermite[" << i
                    << "].push_back(" << hermite[i][j] << ");" << endl;
        }
    }
// cout<<"********************************************************************"<<endl;      

    *sourcefile << "double _leg[" << _ninputs << "][" << pc_max_deg + 1 << "];" << endl;
    *sourcefile << "double _her[" << _ninputs << "][" << pc_max_deg + 1 << "];" << endl;
    *sourcefile << "for(int k=0;k<" << _ninputs << ";k++)" << endl;
    *sourcefile << "{" << endl;
    *sourcefile << "    for(int i=0;i<=" << pc_max_deg << ";i++)" << endl;
    *sourcefile << "    {" << endl;
    *sourcefile << "        if(i==0)" << endl;
    *sourcefile << "         {" << endl;
    *sourcefile << "            _leg[k][i] = legendre[0][0];" << endl;
    *sourcefile << "            _her[k][i] = hermite[0][0];" << endl;
    *sourcefile << "         }" << endl;
    *sourcefile << "        else{" << endl;
    *sourcefile
            << "            _leg[k][i] = legendre[i][i-1] + legendre[i][i]*param[k];"
            << endl;
    *sourcefile
            << "            _her[k][i] = hermite[i][i-1] + hermite[i][i]*param[k];"
            << endl;
    *sourcefile << "            for(int j=i-2;j>=0;j--)" << endl;
    *sourcefile << "            {" << endl;
    *sourcefile
            << "                _leg[k][i] = legendre[i][j] + param[k]*_leg[k][i];"
            << endl;
    *sourcefile
            << "                _her[k][i] = hermite[i][j] + param[k]*_her[k][i];"
            << endl;
    *sourcefile << "            }" << endl;
    *sourcefile << "         }" << endl;
    *sourcefile << "    }" << endl;
    *sourcefile << "}" << endl;

    // initialization of y using the mean coefficient
    *sourcefile << "  double y = " << pc_coefs[0] << " ;" << endl;

    *sourcefile << "double val;" << endl;
    *sourcefile << "" << endl;
    *sourcefile << "for(int i=1;i<" << npol << ";i++)" << endl;
    *sourcefile << "{" << endl;
    *sourcefile << "    val = pc_coefs[i];" << endl;
    *sourcefile << "    for(int j=0;j<" << _ninputs << ";j++)" << endl;
    *sourcefile << "        {" << endl;
    *sourcefile << "            if(law[j]=='u')" << endl;
    *sourcefile << "                val = val*_leg[j][pc_degrees[i][j]];"
            << endl;
    *sourcefile << "            else" << endl;
    *sourcefile << "                val = val*_her[j][pc_degrees[i][j]];"
            << endl;
    *sourcefile << "        }" << endl;
    *sourcefile << "    y += val;" << endl;
    *sourcefile << "}" << endl;

    *sourcefile << "  res[0] = y;" << endl;
    result.close();
    
 //Clean

    delete[] phi;
    phi = NULL;
    delete[] norme;
    norme = NULL;
    
    for(unsigned idel=0;idel<pc_degrees.size();idel++)
    {
        delete[] pc_degrees[idel];
        pc_degrees[idel] = NULL;
    }
    pc_degrees.clear();
    
    for(unsigned idel=0;idel<legendre.size();idel++)
    {
        delete[] legendre[idel];
        legendre[idel] = NULL;
    }
    legendre.clear();
    
    for(unsigned idel=0;idel<hermite.size();idel++)
    {
        delete[] hermite[idel];
        hermite[idel] = NULL;
    }
    hermite.clear();
    
    
 /////////////////   
}

//     void TAnisp::tests(Double_t ** ptExp, Int_t n)
//     {
//     ////////test StepByStepOutput///////////////////////
// //     Double_t* err = new Double_t[2];
// // //     double** ind = new double*;
// //         vector<Int_t*> ind;
// //     Int_t* indice;
// // //     ind[0] = new double[8];
// //     Double_t a,b;
// //         err[0] = 1; err[1] = 2;
// // 
// //     for(int i=0;i<20;i++)
// //     {
// //     a = err[0];
// //     b = err[1];
// //     err = new Double_t[2];
// //     err[0] = a+2; err[1] = b+2;
// // //     ind = new double*[i+1];
// //     for(int k=0;k<i+1;k++)
// //     {
// // //     ind[k] = new double[8];
// //         indice = new Int_t[8];
// //     for(int j=0;j<8;j++)
// // //         ind[k][j] = 0;//j+i+k;
// //             indice[j] = 0;
// //     
// //         if(i%3==0)
// // //         ind[k][0] = i;
// //                 indice[0] = i;
// //         else
// //         {
// //         if(i%3==1)
// // //             ind[k][1] = i;
// //             indice[1] = i;
// //         else
// //         {
// // //             ind[k][2] = i;
// // //             ind[k][0] = i;
// //             indice[2] = i;
// //             indice[0] = i;
// //         }
// //         }
// //     ind.push_back(indice);
// //     }
// //     
// //     StepByStepOutput(err,ind);
// //     }
//     //////////fin test////////////////////////
//     
//     
//     /////test launchCalculs/////////////////
//     vector<Double_t*> vptExp;
//     for(int k=0; k<n;k++) 
//         {
//         vptExp.push_back(ptExp[k]);        
//           
//         }
//     
//     launchCalculs(vptExp);
// //     _tds->exportData("exportDataIshigami");
// //         _tds->exportData("exportDataFlowRate");
// //     functionLauncher(vptExp);
// //     _tds->exportData("exportDataIshigami2");
//     
//     //////////fin test////////////////////////
//     
//     }

}//end of Modeler namespace
}  //end of Uranie namespace

