//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TAnisp.h
 \author Thierry Crestaux (CS)
 \date  29 April 2013

 \brief Interface of the class URANIE::Modeler::TANisp

 */
/*!

 @class URANIE::Modeler::TAnisp

 @brief Description of the class TAnisp.

 A wrapper class for the Anisp library which permit to interface the Anisp 
 library with the URANIE platform. The Anisp library is a research code which
 uses the ANISP method developped by the CEA to approximate the Polynomial Chaos
 expansion using adaptive numerical integration and selecting an adapted and
 smaller Polynomial Chaos basis.
 
 \par Polynomial Chaos expansion for uncertainty propagation in numerical simulation 

 Let's consider a model of the form \f$\mathcal{M}(D,y)=0\f$ where \f$D\f$ represent the datas
 and \f$y\f$ the scalar output of the model. In the deterministic case \f$D\f$ is certain and so
 is \f$y\f$. Now, we consider the stochastic case where \f$D\f$ is uncertain and we express that
 by putting the datas that depend of a stochactic event \f$\omega\f$. The model of the
 stochastic case is thus
 \f{equation}{
 \mathcal{M}(D(\omega),y(\omega)) = 0,
 \f}
 and the output is now an uncertain variable.

 Knowing the uncertainties on the datas, we want to propagate them to know
 the uncertainty of the output. The approach chosen is to parametrize these uncertainties by 
 using a finite vector of independent random  variables : 
 \f{equation}{
 \mbox{\boldmath$\xi$} (\omega) =(\xi_1(\omega), \ldots, \xi_d(\omega))
 \f}
 and so our model becomes:
 \f{equation}{
 \mathcal{M}(D(\mbox{\boldmath$\xi$}(\omega)),y(\mbox{\boldmath$\xi$}(\omega))) = 0.
 \f}
 From now on we will stop writing the dependence on \f$\omega\f$.

 All random variable with a finite variance is equal to an infinite Polynomial Chaos expansion:
 \f{equation}{
 y(\mbox{\boldmath$\xi$}) = \sum_{k=0}^{\infty} y_k \Psi_k(\mbox{\boldmath$\xi$}), 
 \f}
 where \f$\Psi_k\f$ are orthonormal polynomials and
 \f{equation}{
 y_{k} = \frac{\left< y, \Psi_{k} \right> } { \| \Psi_{k}\|^2 } \mbox{ with } \left< X,Y \right> = E(XY) 
 \mbox{ and } \| \Psi_{k}\| = 1,
 \f}
 so,
 \f{equation}{
 y_{k} = E( y \Psi_{k} ) = \int_{\Omega_d} y(\mbox{\boldmath$\xi$})\Psi_{k}(\mbox{\boldmath$\xi$}) 
 pdf(\mbox{\boldmath$\xi$}) d\mbox{$\boldmath{\xi}$} . 
 \f}


 The ANISP method consist to approximate this infinite polynomial expansion by a finite adapted polynomial expansion
 where the polynomial coefficients are computed by using  adaptive numerical integration. The integration algorithm
 used is adaptive cubature algorithm. 

 The strong points of this method are that it automates the construction of a adapted cubature formula for
 numerical integration, the selection of finite and adapted set of polynomials and the computation of
 the coefficients.



 \par Description of the links between the methods of the TAnips class.

 The two constructors call the method initialize :
 \dot
 digraph constructorRelations {
 rankdir=LR;
 node [shape=record, fontname=Helvetica, fontsize=10];
 a [ label="Constructor" ];
 b [ label="initialize()"];
 a -> b  [ arrowhead="open", style="dashed" ];
 }
 \enddot
 
 The runAnisp and restartAnisp methods, which are the core of this class as
 they call the anisp function of the Anisp library, activate
 indirectly the call of several methods:

 \dot
 digraph runAnispRelations {
 node [shape=record, fontname=Helvetica, fontsize=10];
 a [ label="runAnisp() (or restartAnisp())"];
 b [ label="anisp function of the Anisp library"];
 c [ label="setTPolynomialChaos()"];
 b1 [ label="functionLauncher()"];
 b2 [ label="StepByStepOutput()"];
 b2_1 [ label="EndIterationActions()"];
 b1_1 [ label="launchCalculs()"];
 b2_1_1 [ label="addErrorIndicator()"];
 b2_1_2 [ label="addIntegrationIndices()"];
 b1_1_1 [ label="changeOfVariable()"];
 
 a -> b  [ arrowhead="open", style="dashed" ];
 a -> c  [ arrowhead="open", style="dashed" ];
 b -> b1  [ arrowhead="open", style="dashed" ];
 b -> b2  [ arrowhead="open", style="dashed" ];
 b1 -> b1_1  [ arrowhead="open", style="dashed" ];
 b2 -> b2_1  [ arrowhead="open", style="dashed" ];
 b2_1 -> b2_1_1  [ arrowhead="open", style="dashed" ];
 b2_1 -> b2_1_2  [ arrowhead="open", style="dashed" ];
 b1_1 -> b1_1_1  [ arrowhead="open", style="dashed" ];
 }
 \enddot

 */
#ifndef TANISP_H
#define TANISP_H

#include <vector>
#include <string>
#include "math.h"

//ROOT
#include "Rtypes.h"
#include "TLeaf.h"
#include "TCanvas.h"
#include "TH1F.h"
#include "TLegend.h"

// uranie
#include "TDataServer.h"
#include "TModeler.h"
#include "TStochAttribut.h"
#include "UExceptions.h"
#include "TLauncherFunction.h"
#include "TCode.h"
#include "TLauncher.h"
#include "TPolynomialChaos.h"
#include "TNisp.h"

// anisp
#include "anisp.hpp"

using namespace URANIE::DataServer;
using namespace URANIE::Exceptions;
using namespace URANIE::Launcher;

namespace URANIE
{
namespace Modeler
{

void StepByStepOutput(Double_t * val, vector<Int_t *> indices);

void functionLauncher(vector<Double_t *> ptEXp);

class TAnisp: public TModeler
{
/////////////members of the class
public:
    //! Enum for TCode or analytical function
    enum ANISP_TYPE
    {
        kAnalyticalFunction, kCode
    };

    /*! 
     \brief Friend function to pass to the anisp function
     \param val : pointer to a Double_t fill with the values of a multi-indice
     \param indices : vector of Double_t * which contains the iterator number, 
     the value of the point where the calcul is launched, 
     the result of this calcul and the number of running of the ANISP method
     */
    friend void StepByStepOutput(Double_t * val, vector<Int_t *> indices);

    /*! 
     \brief Friend function which launchs the calculs
     \param ptExp : vector of Double_t * which contains the iterator number, 
     the value of the point where the calcul is launched, 
     the result of this calcul and the number of running of the ANISP method 
     */
    friend void functionLauncher(vector<Double_t *> ptEXp);

//     public:
//       //! Boolean for edit the log.
//       Bool_t  _blog;

private:
    //! Garbage collector list for drawing
    TObjArray *_drawingGarbageCollector;
    //! Tolerance of the adaptative integration algorithm
    Double_t _dTolerance;
    //! Order minimum of interaction for the adaptative integration
    Int_t _nKMin;
    //! Stop criterion of the adaptative integration algorithm based of the number of simulations
    Int_t _nNumberMaxOfSimulations;
    //! Greater value possible for an indice
    Int_t _nGreaterIndice;
    //! Number maximum of indices of integration (memory size)
    Int_t _nMaxIndices;
    //! Degree maximum possible for the Polynomial Chaos expansion
    Int_t _nDegreeMax;
    //! Root of the output files name.
    TString _sRootFilename;
    //! Laws of probability for each variable
    TString _sLaws;
    //! Number of iteration of adaptative integration algorithm
    Int_t _nIterationNumber;
    //! Object of type TLauncherFunction
    TLauncherFunction * _tAnalyticalFunction;
    //! Object of type TCode
    TCode * _Code;
    //! Object of type TLauncher
    TLauncher * _tlauncher;
    //! Variable containing the ANISP_TYPE
    ANISP_TYPE _nCible;
    //! Object of type TPolynomialChaos
    TPolynomialChaos * _tpc;
    //! Object of type TNisp
    TNisp * _nisp;
    //! List of error indicators from the integration algorithm 
    TNtupleD * _tErrorIndicator;
    //! List of integration indices 
    TNtupleD * _tIntegrationIndices;
    /* //! Canvas for graphical outputs */
    /* TCanvas * _myCanvas; */
    //! Number of use of the launchCalculs method
    Int_t _nNumberOfLaunching;
    //! Number of computed points
    Int_t _nNumberOfPoints;
    //! Names of the input arguments
    TString _sinput;
    //! Names of the output arguments
    TString _soutput;
    //! Object of type TDataServer which contains all the launchs data
    TDataServer* _tdsAnisp;
    //! analytical function
#ifdef ROOT5
    void* _analyticalFunction;
#else
    void (*_analyticalFunction)(Double_t*, Double_t*);
#endif
    //! name of the analytical function
    TString _sFunctionName;
    //! number of stochastic inputs
    Int_t _ninputs;
    //! number of launching of the ANISP method
    Int_t _nUranieAnisp;
    //! name of the TAttribute containing the number of launching
    TString _sLaunchNumber;
    /* //! Object of type TH1F used for graphical output */
    /* TH1F* _hist1; */
    /* //! Object of type TH1F used for graphical output */
    /* TH1F* _hist2; */

public:
    /** @name Constructor and Destructor
     **/
    //@{
    /*!
     \brief Construct an instance of the TAnisp object from a TDataServer object and a void * (analytical function)
     \param tds : pointer to a TDataServer object 
     \param analyticalFunction : void * (analytical function)
     \param sinput : TString value containing the names of the inputs variables
     \param soutput : TString value containing the name of the unique output variable
     \param option : pointer to Option_t value
     */
#ifdef ROOT5
    TAnisp(TDataServer *tds, void *analyticalFunction, TString sinput = "",
            TString soutput = "", Option_t *option = "");
#else
    TAnisp(TDataServer *tds, void (*analyticalFunction)(Double_t*,Double_t*), 
            const char *fname, TString sinput = "",
            TString soutput = "", Option_t *option = "");
#endif
    /*!
     \brief Construct an instance of the TAnisp object from a TDataServer object and a void * (analytical function)
     \param tds : pointer to a TDataServer object 
     \param analyticalFunction : const char* (name of the analytical function)
     \param sinput : TString value containing the names of the inputs variables
     \param soutput : TString value containing the name of the unique output variable
     \param option : pointer to Option_t value
     */
    TAnisp(TDataServer *tds, const char* analyticalFunction, TString sinput = "",
	   TString soutput = "", Option_t *option = "");

    /*!
     \brief Construct an instance of the TAnisp object from a TDataServer object and a TCode object
     \param tds : pointer to a TDataServer object 
     \param code : pointer to a TCode object
     \param option : pointer to Option_t value
     */
    TAnisp(TDataServer *tds, TCode* code, Option_t *option = "");
    //! Destructor
    /*!
     \brief Destruct an instance of TAnisp object. 
     */
    virtual ~TAnisp();
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;

    //! Export the model in Fortran langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TAnisp::exportModelFortran(ofstream * sourcefile) not yet implemented"
                << endl;
    } ;
    
    //! Export the model in a PMML file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPMML(const char* file = "", 
                         const char* name = "",Option_t *option = "") const
    {
        TString soption = option;
        soption.ToLower();
	UNUSED(file,name);
        std::cerr
                << "   --- Method TAnisp::exportModelPMML(ofstream * sourcefile) not yet implemented"
                << endl;
    };
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TAnisp::exportModelPython(ofstream * sourcefile) not yet implemented"
                << endl;
    };
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
//       void          setLog() { _blog = kTRUE;}
//       void          unsetLog() { _blog = kFALSE;}
//       void          changeLog() {  _blog = _blog? kFALSE : kTRUE ; }
//       Bool_t        getLog() { return _blog; }
    /*!
     \brief Print log
     \param option : pointer to an Option_t object
     */
    virtual void printLog(Option_t *option = "");
    //@}

    //---------------------------------------------

    ////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////

    /** @name Setter and getter functions
     **/
    //@{
    /*!
     \brief Set the value of the _nKMin attribute to "kmin"
     \param kmin : int value to set the minimum level of interaction of the integration algorithm
     */
    void setKMin(Int_t kmin);

    /*!
     \brief Get the order of interaction for adaptative integration
     \return the minimum order of interaction 
     */
    Int_t getKMin()
    {
        return _nKMin;
    }

    //! Set the tolerance for the adaptative integration
    /*!
     \brief Set the value of the tolerance for the adaptative integration
     \param tol : double value to set the tolerance criterion of the integration algorithm
     */
    void setTolerance(Double_t tol);

    /*!
     \brief Get the tolerance for the adaptative integration
     \return the tolerance value for the adaptative integration
     */
    Double_t getTolerance()
    {
        return _dTolerance;
    }

    /*!
     \brief Set the criterion of maximum number of simulations before stopping  the adaptative integration
     \param nmos : int value to set the criterion of maximum number of simulations of the integration algorithm
     */
    void setNumberMaxOfSimulations(Int_t nmos);

    /*!
     \brief Get the criterion of maximum number of simulations before stopping  the adaptative integration
     \return the  maximum number of simulations 
     */
    Int_t getNumberMaxOfSimulations()
    {
        return _nNumberMaxOfSimulations;
    }

    /*!
     \brief Set the greater value possible for an indice
     \param greatInd : int value to set the greater value for an indice reachable by the integration algorithm
     */
    void setGreaterIndice(Int_t greatInd);

    /*!
     \brief Get the greater value possible for an indice
     \return the greater value of an indice
     */
    Int_t getGreaterIndice()
    {
        return _nGreaterIndice;
    }

    /*!
     \brief Set the number maximum of indices for the integration
     \param maxInd : int value to set the maximum number of indices for the integration algorithm (memory criterion)
     */
    void setMaxIndices(Int_t maxInd);

    /*!
     \brief Get the number maximum of indices for the integration
     \return the number maximum of indices 
     */
    Int_t getMaxIndices()
    {
        return _nMaxIndices;
    }

    /*!
     \brief Set the degree maximum for the Polynomial Chaos expansion
     \param degMax : int value to set the maximum degree of the Polynomial Chaos expansion
     */
    void setDegreeMax(Int_t degMax);

    /*!
     \brief Get the degree maximum for the Polynomial Chaos expansion
     \return the degree maximum of the polynomial
     */
    Int_t getDegreeMax()
    {
        return _nDegreeMax;
    }

    /*!
     \brief Get the ANISP_TYPE which indicates if ANISP approximate a code (TCode) or a function (void *)
     \return the ANISP_TYPE of the _nCible attribute
     */
    Int_t getCible()
    {
        return _nCible;
    }

    //! Get the number of iteration of the integration algorithm
    /*!
     \brief Get the number of iteration of the integration algorithm
     \return the number of iteration of the integration algorithm
     */
    Int_t getIterationNumber()
    {
        return _nIterationNumber;
    }

    /*!
     \brief Set the _blog parameter
     \param blog : Bool_t value to indicate if we wante the graphical and console outputs
     */
    void setlog(Bool_t blog);

    /*!
     \brief Set the root name of the output files of the Anisp library
     \param rootName : TString value which serves as root for all the names of the output files of the Anisp library
     */
    void setRootFilename(TString rootName);

    /*!
     \brief Get the root name of the output files of the Anisp library
     \return the root name of the output files of the Anisp library
     */
    TString getRootFilename()
    {
        return _sRootFilename;
    }

    /*!
     \brief Get the list of error indicators
     \return the list of error indicators
     */
    TNtupleD* getErrorIndicator()
    {
        return _tErrorIndicator;
    }

    /*!
     \brief Get the list of integration indices
     \return the list of integration indices
     */
    TNtupleD* getIntegrationIndices()
    {
        return _tIntegrationIndices;
    }

    /*!
     \brief Get the TPolynomialChaos computed by the Anisp method
     \return the TPolynomialChaos computed 
     */
    TPolynomialChaos getTPolynomialChaos();

    /*!
     \brief Get a TString containing the name of all attributes separated by ':'
     \param inputs : TString value containing the name of the inputs attributes separated by ':'
     \return names of all attributes
     */
    TString getVariablesName(TString inputs = "");

    //@}

    //////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////

    /** @name Anisp library files
     **/
    //@{
    /*!
     \brief Get the Anisp intern data file name 
     \return the data file name 
     */
    TString getDataAnispFile() const
    {
        return _sRootFilename + "data.dat";
    }

    /*!
     \brief Get the computed polynomial coefficients file name
     \return the computed coefficients file name
     */
    TString getResultFile() const
    {
        return _sRootFilename + TString("res.dat");
    }

    /*!
     \brief Get the stochastic points file name
     \return the stochastic points file name
     */
    TString getStochasticFile() const
    {
        return _sRootFilename + "sto.dat";
    }

    /*!
     \brief Get the error indicator values file name
     \return the error indicator values file name
     */
    TString getErrorIndicatorFile() const
    {
        return _sRootFilename + "error.dat";
    }

    /*!
     \brief Get the integration indices file name
     \return the integration indices file name
     */
    TString getIndicesFile() const
    {
        return _sRootFilename + "indices.dat";
    }

    //@}

    ///////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////

    /** @name Set parameters functions
     **/
    //@{

    /*!
     \brief Set all the Anisp parameters
     \param tol: double value to set the tolerance criterion of the integration algorithm
     \param kmin: int value to set the minimum level of interaction of the integration algorithm
     \param nmos: int value to set the criterion of maximum number of simulations of the integration algorithm
     \param greatInd: int value to set the greater value for an indice reachable by the integration algorithm
     \param maxInd: int value to set the maximum number of indices for the integration algorithm (memory criterion)
     \param blog: Bool_t value to indicate if we wante the graphical and console outputs
     \param degMax: int value to set the maximum degree of the Polynomial Chaos expansion
     \param rootName: TString value which serves as root for all the names of the output files of the Anisp library
     */
    void setAllAnispParameters(Double_t tol, Int_t kmin, Int_t nmos,
            Int_t greatInd, Int_t maxInd, Bool_t blog, Int_t degMax,
            TString rootName);

    /*!
     \brief Set most of the Anisp parameters
     \param tol: double value to set the tolerance criterion of the integration algorithm
     \param kmin: int value to set the minimum level of interaction of the integration algorithm
     \param nmos: int value to set the criterion of maximum number of simulations of the integration algorithm
     \param greatInd: int value to set the greater value for an indice reachable by the integration algorithm
     \param maxInd: int value to set the maximum number of indices for the integration algorithm (memory criterion)
     \param blog: Bool_t value to indicate if we wante the graphical and console outputs
     */
    void setAnispParameters(Double_t tol, Int_t kmin, Int_t nmos,
            Int_t greatInd, Int_t maxInd, Bool_t blog);

    /*!
     \brief Set the Anisp calcul parameters
     \param tol: double value to set the tolerance criterion of the integration algorithm
     \param kmin: int value to set the minimum level of interaction of the integration algorithm
     \param nmos: int value to set the criterion of maximum number of simulations of the integration algorithm
     */
    void setLightAnispParameters(Double_t tol, Int_t kmin, Int_t nmos);

    //@}

    ///////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////

    /** @name Run the ANISP method
     **/
    //@{  
    /*!
     \brief Run the Anisp method to compute a Polynomial Chaos expansion
     */
    void runAnisp();

    /*!
     \brief Restart the Anisp method to compute a Polynomial Chaos expansion from a previous calcul
     */
    void restartAnisp();

    //@}  

private:
    /*!
     \brief Initialize attributes of the TAnisp object 
     */
    void initialize();

    /*!
     \brief Add an error indicator to _tErrorIndicator
     \details
     this vector contains 2 columns of double
     \li number of simulations
     \li value of the error indicator 
     
     \param errInd : pointer to a Double_t containing a number of simulations 
     and a value of the error indicator 
     */
    void addErrorIndicator(Double_t * errInd)
    {
        _tErrorIndicator->Fill(errInd);
    }

    /*!
     \brief Add an error indicator to _tIntegrationIndices
     \details
     this vector contains one column by uncertain variable
     
     \param ind : pointer to a Double_t fill with the values of a multi-indice
     */
    void addIntegrationIndices(Double_t * ind)
    {
        _tIntegrationIndices->Fill(ind);
    }

    /*!
     \brief Launch batchs of calculs during the numerical integration step of the ANISP method 
     \param ptExp : vector of Double_t * which contains the iterator number, 
     the value of the point where the calcul is launched, 
     the result of this calcul and the number of running of the ANISP method 
     */
    void launchCalculs(vector<Double_t *> ptExp);

    /*!
     \brief Run at the end of an iteration of the integration algorithm and 
     exploit the integration indices and error indicators to show graphical 
     output of the progress of the ANISP method  
     \param errIndicator : pointer to a Double_t fill with the values of a multi-indice
     \param integrationIndices : vector of Double_t * which contains the iterator number, 
     the value of the point where the calcul is launched, 
     the result of this calcul and the number of running of the ANISP method
     */
    void EndIterationActions(Double_t * errIndicator,
            vector<Double_t *> integrationIndices);

    /*!
     \brief Do a changement of variable to change standart stochastic values into the problem uncertain values
     \param x : pointer to a Double_t
     */
    void changeOfVariable(Double_t * x);

    /*!
     \brief Open the result file of the Anisp library and use the coefficients stored there
     to create a TPolynomialChaos object and set its coefficients
     */
    void setTPolynomialChaos();

    ClassDef(URANIE::Modeler::TAnisp, ID_MODELER)

};
//end of the class TAnisp

}//end of the namespace Modeler
}     //end of the namespace URANIE

R__EXTERN URANIE::Modeler::TAnisp *gUranieTAnisp;
#endif

