//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TBootstrap.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Fri Feb 24 14:45:26 2006
 
 \brief Class TBootstrap
 
 
 */

#include <stdio.h>
#include <stdlib.h>

#include "Riostream.h"
#include "TEventList.h"
#include "TLeaf.h"
#include "TCanvas.h"
#include "TGraph.h"
#include "TMath.h"
#include "TH2.h"
#include "TProfile.h"
#include "TStyle.h"

#include "TBootstrap.h"
#include "TPMML.h"

const TString global_WeightsFilename = "weights.tmp";

ClassImp (TBootstrap);

TBootstrap::TBootstrap(TMultiLayerPerceptron* mlp_cons, Int_t nbbases_cons,
        Bool_t bLog_cons) :
        TNamed("Bootstrap", "Methode Bootstrap"), _dLowerQuantile(0.10), _dUpperQuantile(
                0.90)
{
    mlp = mlp_cons;
    nb_bases = nbbases_cons;
    // Initialisation du g�n�rateur de nombres al�atoires
    srand (time(NULL));bLog
    = bLog_cons;

    //
    _belApp = new URANIE::DataServer::TPatternsEventList();

    // Get the Learning TEventList on the ANN and the size
    _telApp = mlp_cons->GetfTraining();
//   if (bLog)
    _telApp->Print("all");

    Int_t nb_trainingpatterns = _telApp->GetN();
    if (bLog)
        cout << " ** nb_trainingpatterns[" << nb_trainingpatterns << "]"
                << endl;

}

//______________________________________________________________________________
TBootstrap::~TBootstrap()
{
    if (_belApp)
        delete _belApp;
    _belApp = NULL;
}

/**
 *
 */

void TBootstrap::Train(Int_t nb_epochs, Option_t * option, TString pmml_name)
{
    // Initialisation des poids
    mlp->Randomize();
    // Sauvegarde des poids initiaux    
    mlp->DumpWeights(global_WeightsFilename);
    // On r�cup�re la structure, la base globale et les indices de la base test du r�seau initial
    TEventList * test = mlp->GetTest();
    if (test)
    {
        cout << endl << "***************" << endl;
        cout << "  ** Warning : test data not used actaullly" << endl;
        cout << "***************" << endl;
    }
    Double_t delta = mlp->GetDelta();
    Double_t epsilon = mlp->GetEpsilon();
    Double_t eta = mlp->GetEta();
    Double_t etadecay = mlp->GetEtaDecay();
    Int_t reset = mlp->GetReset();
    Double_t tau = mlp->GetTau();

    TPMML* pmml = new TPMML(pmml_name);
    pmml->Write(mlp, "RECREATE");

    TString smlpname = mlp->GetMLPName();

    cout << "option[" << option << "]" << endl;
    TString opt = option;
    if (!opt.Contains("+"))
    {
        opt += TString(" +");
        cout << "opt[" << opt << "]" << endl;
    }

    TCanvas *cann = new TCanvas();
    cann->Clear();
    cann->Divide(2, 1);

    // Learn the init MLP
    mlp->SetMLPName(Form("%s_init", smlpname.Data()));
    cann->cd(1);
    mlp->Train(nb_epochs, Form("%s current", opt.Data()));
    mlp->Export(Form("mlp_%s", mlp->GetMLPName().Data()), "c++");
    pmml->Write(mlp, "UPDATE");
    mlp->SetMLPName(smlpname);

    // D�claration du perceptron que l'on utilisera pour le bootstrap
    TMultiLayerPerceptron* boot_mlp = new TMultiLayerPerceptron(
            mlp->GetStructure(), mlp->GetData(), _telApp, _telApp,
            mlp->GetType());
    boot_mlp->SetLearningMethod(mlp->GetLearningMethod());

    // Loop on the bootstap databases
    for (Int_t i = 0; i < nb_bases; i++)
    { // nb_bases
        // Generate a bootstrap database
        BuildBootTraining(i);
        if (bLog)
            cout << "* Base n�" << i << " bootstrap created" << endl;
//    TEventList* tel = training;
//     training->Print("all");
        // Construction du mlp bootstrap� � partir des param�tres du r�seau de d�part
        boot_mlp->SetMLPName(Form("%s_%d", smlpname.Data(), i));
        boot_mlp->SetTrainingDataSet((TEventList*) _belApp);
        boot_mlp->SetDelta(delta);
        boot_mlp->SetEpsilon(epsilon);
        boot_mlp->SetEta(eta);
        boot_mlp->SetEtaDecay(etadecay);
        boot_mlp->SetReset(reset);
        boot_mlp->SetTau(tau);
//     boot_mlp->SetLearningMethod(TMultiLayerPerceptron::kStochastic);

        if (bLog)
            cout << "* R�seau bootstrap� cr��e" << endl;

        //cout << boot_mlp->GetEta() << endl;

        // Chargement des poids initiaux, afin de partir d'un m�me point de d�part
        boot_mlp->LoadWeights(global_WeightsFilename);
        // Lancement de l'apprentissage pour le mlp en cours

        cann->cd(1);

        boot_mlp->Train(nb_epochs, Form("%s current", opt.Data()));
        //    boot_mlp->Export(Form("mlp_%d",i),"C++");
        boot_mlp->DumpWeights(Form("mlp_%d", i));
        //cout << "x1: " << boot_mlp->Evaluate(0,x1) << endl;
        //cout << "x4: " << boot_mlp->Evaluate(0,x4) << endl;
        cann->cd(2);
        boot_mlp->DrawResult(0, "comp train nocanv");
        boot_mlp->DrawResult(0, "comp test nocanv same");

        if (bLog)
            cout << "* Apprentissage sur le r�seau \"" << boot_mlp->GetMLPName()
                    << "\" termin�" << endl;
        pmml->Write(boot_mlp, "UPDATE");

        if (bLog)
            cout << "____________________________________________________"
                    << endl;

    }

    cout << endl;
    cout << "____________________________________________________" << endl;
    cout << " ** End of Bootstrap Learning" << endl;
    cout << " ** Detect the epoch star" << endl;
    cout << " ** File : [" << boot_mlp->GetRooFileWeights() << "]" << endl;
    boot_mlp->SetMLPName(mlp->GetMLPName());

    TString srootfilenameweights = boot_mlp->GetRooFileWeights();
    if (srootfilenameweights.Length() == 0)
    {
        cout << " ERROR : no weights save[" << boot_mlp->GetRooFileWeights()
                << "]" << endl;
        cout << "____________________________________________________" << endl;
        cout << " ** End of Detect the epoch star" << endl;
        return;
    }
    TFile *rf = new TFile(srootfilenameweights, "update");
    TNtupleD * nterr = (TNtupleD *) (rf->Get("MLP_Errors"));
    Int_t nEpoch = (Int_t)(nterr->GetMaximum("epoch"));
    Int_t nBoot = (Int_t)(nterr->GetMaximum("boot"));
    cout << " ** nEpoch[" << nEpoch << "]" << endl;
    cout << " ** nBoot[" << nBoot << "]" << endl;
    cout << endl;

    TNtupleD *ntd_quantiles_errors =
            new TNtupleD("MLP_quantiles_Errors", "Errors of training",
                    "epoch:app_low:app_upp:app_mean:app_std:test_low:test_upp:test_mean:test_std");
    ntd_quantiles_errors->SetMarkerColor(1);
    ntd_quantiles_errors->SetMarkerSize(4);
    ntd_quantiles_errors->SetMarkerStyle(7);
    ntd_quantiles_errors->SetLineStyle(1);
    ntd_quantiles_errors->SetLineColor(1);
    ntd_quantiles_errors->SetLineWidth(4);

    // Calcul des Quantiles
    const Int_t nq = 2;
    Double_t xq[nq]; // position where to compute the quantiles in 0.25, and 0.75
    Double_t learq[nq];  // array to contain the quantiles
    Double_t testq[nq];  // array to contain the quantiles
    xq[0] = _dLowerQuantile;
    xq[1] = _dUpperQuantile;

    Double_t dmeanapp, dstdapp;
    Double_t dmeantest, dstdtest;
    Double_t dmintest = -1.0;
    Int_t nepochstar = 0;

    TCanvas *cc = new TCanvas();
    cc->Clear();
    nterr->SetMarkerColor(15);
    nterr->SetLineColor(15);
    nterr->Draw("test:epoch");
    nterr->SetMarkerColor(kBlue);
    nterr->Draw("app:epoch", "boot==0", "lp same");
    nterr->SetMarkerColor(kRed);
    nterr->Draw("test:epoch", "boot==0", "lp same");
    cc->Modified();
    cc->Update();

    TLeaf *lapp = nterr->GetLeaf("app");
    Double_t dvalapp, dvalgenstar, dcoef = 1.0;

    // Epoch loop.
    Int_t nnvisu = (Int_t)(nEpoch / 10);
    for (Int_t iepoch = 0; iepoch <= nEpoch; iepoch++)
    {
//     cout << " ** epoch[" << " iepoch " << "/" << nEpoch << "]";

        nterr->GetEntry(iepoch);
        dmeanapp = dvalapp = lapp->GetValue();

        // Case of the learning
        if (0)
        {
            //// fga::2008-09-20 18:07:53 no need
            nterr->Draw("app", Form("((epoch==%d)&&(boot>0))", iepoch), "goff");
            TMath::Quantiles(nBoot, nq, nterr->GetV1(), learq, xq, kFALSE);
            //     cout << " -- Learn " << xq[0] << "[" << learq[0] << "] - " << xq[1] << "[" << learq[1]<<"] **";
            nterr->Draw("app",
                    Form("((epoch==%d)&&(boot>0)&&(app<%g)&&(app>%g))", iepoch,
                            learq[1], learq[0]), "goff");
            dmeanapp = TMath::Mean(nterr->GetSelectedRows(), nterr->GetV1());
            dstdapp = TMath::RMS(nterr->GetSelectedRows(), nterr->GetV1());

            //     cout << " Nsel[" << nterr->GetSelectedRows()<<"] mu[" << dmeanapp << "] sigma["<<dstdapp << "] ** ";
        }

        // Case of the test
        nterr->Draw("biais", Form("((epoch==%d)&&(boot>0))", iepoch), "goff");
//     cout << "  ** test i[" << iepoch << "] n sel[" << nterr->GetSelectedRows() << "] app["<< dvalapp << "]";
        if (iepoch > 0)
        {
            TMath::Quantiles(nBoot, nq, nterr->GetV1(), testq, xq, kFALSE);
//       cout << " -- Test " << xq[0] << "[" << testq[0] << "] - " << xq[1] << "[" << testq[1]<<"] **";
            nterr->Draw("biais",
                    Form("((epoch==%d)&&(boot>0)&&(biais<%g)&&(biais>%g))",
                            iepoch, testq[1], testq[0]), "goff");
            dstdtest = TMath::Mean(nterr->GetSelectedRows(), nterr->GetV1());
//       dstdtest = TMath::RMS(nterr->GetSelectedRows(), nterr->GetV1());
//       dstdtest = RMSmu(nterr->GetSelectedRows(), nterr->GetV1(), dvalapp);
//            cout << " Nsel[" << nterr->GetSelectedRows()<<"] delta["<<dstdtest << "] ** ";
        }
        else
        {
            testq[0] = testq[1] = dmeantest = TMath::Mean(
                    nterr->GetSelectedRows(), nterr->GetV1());
            dstdtest = 0.0;
        }
        dvalgenstar = dvalapp + dcoef * dstdtest;
        // cout << "  ** RMS sel[" << nterr->GetSelectedRows() << "] sigma["<< dstdtest << "] gen[" << dvalgenstar <<"]";
        if (!iepoch)
            dmintest = dvalgenstar;

        ntd_quantiles_errors->Fill(iepoch, learq[0], learq[1], dvalgenstar,
                dstdapp, dvalapp + testq[0], dvalapp + testq[1], dmeantest,
                dstdtest);

        if (dvalgenstar < dmintest)
        {
            dmintest = dvalgenstar;
            nepochstar = iepoch;
        }

//     cout << "*** nstar[" << nepochstar << "]"<<endl;

        if (iepoch % nnvisu == 0)
        {
//       cout << " **** " << iepoch << " : " << nnvisu << " **" << endl;
//       cc->cd(1);
            ntd_quantiles_errors->SetMarkerColor(kBlack);
            ntd_quantiles_errors->SetLineColor(kBlack);
            ntd_quantiles_errors->Draw("test_low:epoch", "", "lp same");
            ntd_quantiles_errors->Draw("test_upp:epoch", "", "lp same");
            ntd_quantiles_errors->SetMarkerColor(kYellow);
            ntd_quantiles_errors->SetLineColor(kYellow);
            ntd_quantiles_errors->Draw("app_mean:epoch", "", "lp same");

//       cc->cd(2);
//       nterr->SetMarkerColor(2); nterr->Draw("test:epoch");
//       ntd_quantiles_errors->Draw("test_low:epoch","","lp same");
//       ntd_quantiles_errors->Draw("test_upp:epoch","","lp same");
//       ntd_quantiles_errors->Draw("test_mean:epoch","","lp same");
            cc->Modified();
            cc->Update();
        }

        //     cout << endl;
    }

    cc->Print(
            Form("annMLP%sErrorsLearningValidationBootstrapQuantile.png",
                    mlp->GetMLPName().Data()));

    if (1)
    {
        //    gStyle->SetLineWidth(5);

        TCanvas *ccc = new TCanvas();
        ccc->Clear();
//     cc->Divide(2,1);
//     cc->cd(1);

        nterr->SetMarkerColor(2);
        nterr->Draw("test:epoch");
        nterr->SetLineColor(kBlue);
        nterr->SetMarkerColor(kBlue);
        nterr->Draw("app:epoch", "boot==0", "lp");
        nterr->SetLineColor(kRed);
        nterr->SetMarkerColor(kRed);
        nterr->Draw("test:epoch", "boot==0", "lp same");

//     ntd_quantiles_errors->SetLineWidth(2);
//     ntd_quantiles_errors->SetLineColor(15);
//     ntd_quantiles_errors->Draw("app_low:epoch","","l same");
//     ntd_quantiles_errors->Draw("app_upp:epoch","","l same");
        ntd_quantiles_errors->SetLineColor(46);
        ntd_quantiles_errors->SetMarkerColor(kYellow);
        ntd_quantiles_errors->SetLineColor(kYellow);
        ntd_quantiles_errors->SetLineWidth(6);
        ntd_quantiles_errors->Draw("app_mean:epoch", "", "lp same");

//     cc->cd(2);
//     nterr->SetMarkerColor(2); nterr->Draw("test:epoch");
//     nterr->SetMarkerColor(kGreen); nterr->Draw("test:epoch","boot==0","same");
//     ntd_quantiles_errors->Draw("test_low:epoch","","lp same");
//     ntd_quantiles_errors->Draw("test_upp:epoch","","lp same");
//     ntd_quantiles_errors->Draw("test_mean:epoch","","lp same");

        ccc->Print(
                Form("annMLP%sErrorsLearningValidationBootstrap.png",
                        mlp->GetMLPName().Data()));
        ccc->Modified();
        ccc->Update();
    }
//   cout << "  *** min test[" << dmintest << "] epochstar[" << nepochstar<<"]"<<endl;
//   if ( nepochstar > nEpoch/2 )
//     nepochstar = 143;
    cout << "  *** min test[" << dmintest << "] epochstar[" << nepochstar << "]"
            << endl;

    // Chargement des poids initiaux, afin de partir d'un m�me point de d�part
    mlp->LoadWeights(global_WeightsFilename);
    cann->Modified();
    cann->Update();
    cann->cd(1);
    mlp->Train(nepochstar, Form("%s current", opt.Data()));
    pmml->Write(mlp, "UPDATE");
    cann->cd(2);
    mlp->DrawResult(0, "comp train nocanv");
    mlp->DrawResult(0, "comp test nocanv same");
    mlp->Export(Form("mlp_%s", mlp->GetMLPName().Data()), "c++");

    cout << "____________________________________________________" << endl;

}

Bool_t TBootstrap::BuildBootTraining(Int_t iter)
{

    _belApp->Reset();
    Int_t nb_trainingpatterns = _telApp->GetN();

    // Loop on each cell of the array
    for (Int_t i = 0; i < nb_trainingpatterns; i++)
    {
        // Tirage al�atoire d'un nombre entier entre 0 et nb_trainingpatterns
        Long64_t r = (Long64_t)(
                nb_trainingpatterns * ((Double_t) rand())
                        / (((Double_t) RAND_MAX) + 1.0));
        if (bLog)
            cout << i << " [" << r << "] ind(" << _telApp->GetEntry(r) << ")"
                    << endl;
        _belApp->Enter(_telApp->GetEntry(r));
    }

    if (bLog)
        _belApp->Print("all");

    return kTRUE;
}

void TBootstrap::setQuantiles(Double_t dlow, Double_t dupp)
{
    if ((dlow < 0) || (dlow > 1.0))
    {
        cout
                << "  ** Error setQuantiles : the lower quantile is outside of the bound ( 0 < "
                << dlow << " < 1. )" << endl;
        exit(-1);
    }
    if ((dupp < 0) || (dupp > 1.0))
    {
        cout
                << "  ** Error setQuantiles : the upper quantile is outside of the bound ( 0 < "
                << dupp << " < 1. )" << endl;
        exit(-1);
    }
    if (dupp < dlow)
    {
        cout << "  ** Error setQuantiles : the upper quantile (" << dupp
                << ")is lower than the lower (" << dlow << ")" << endl;
        exit(-1);
    }
    _dLowerQuantile = dlow;
    _dUpperQuantile = dupp;

}

Double_t TBootstrap::RMSmu(Long64_t n, const Double_t *a, Double_t mu)
{

    if (n <= 0 || !a)
        return 0;

    Double_t tot2 = 0;
    for (Long64_t i = 0; i < n; i++)
    {
        tot2 += (a[i] - mu) * (a[i] - mu);
    }
    Double_t n1 = 1. / n;
    Double_t rms = TMath::Sqrt(tot2 * n1);
    // cout << " ** N[" << n << "] mu[" << mu << "] RMSmu[" << rms << "] **";
    return rms;

}
