//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGLM.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-14 12:54:45
 
 \brief Interface de la classe URANIE::Modeler::TGLM
 
 */

//ROOT
#include "TDecompSVD.h"
#include "TLeaf.h"
#include "TMath.h"

// Uranie
#include "TGLM.h"

ClassImp (URANIE::Modeler::TGLM)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TGLM::TGLM(URANIE::DataServer::TDataServer *tds, TString architecture,
        Option_t *option) :
        TModeler(tds, architecture, option), _nDistribution(kUnknown), _nLinkFunction(kCanonical), _nIteration(0), 
        _nMaxIteration(5),  _vecBetaInit(0), _dTolerance(1.e-6), _dNullDeviance(-1.0), _dResidualDeviance(-1.0)
{
    if (_blog)
    {
        cout << "construit TGLM tds[" << _tds->GetName() << "] architecture["
                << architecture << "] Option [" << option << "]" << endl;
    }

    try
    {
        if (_nY > 2)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGLM constructor\n The number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
                            _nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    TString opt = option;
    opt.ToLower();
}

TGLM::TGLM(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t * option) :
        TModeler(tds, varexpinput, varexpoutput, option), _nDistribution(kUnknown), _nLinkFunction(kCanonical), _nIteration(0), 
        _nMaxIteration(25), _vecBetaInit(0), _dTolerance(1.e-6), _dNullDeviance(-1.0), _dResidualDeviance(-1.0)
{
    try
    {
        if (_nY > 2)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGLM constructor\n The number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
                            _nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _sFileName = Form("linearmodel_%s", _sOutputAttributes.Data());
    _sFunctionName = Form("linearmodel_Fct_%s", _sOutputAttributes.Data());
    SetName(
            Form("Modeler[TGLM]Tds[%s]Predictor[%s]Target[%s]", _tds->GetName(),
                    varexpinput, _sOutputAttributes.Data()));
    SetTitle(
            Form("TGLM as modeler for TDS[%s] with Target is[%s] versus[%s]",
                    _tds->GetName(), _sOutputAttributes.Data(), varexpinput));
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TGLM::~TGLM()
{
    if (_blog)
        cout << "detruit TGLM" << endl;
}

// -----------------------------------------------------------------------
//  set the Family
// -----------------------------------------------------------------------
void TGLM::setFamily(EDistribution nYDistribution, ELinkFunctin nLinkFunction)
{
    try
    {
        if (nYDistribution == kUnknown)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGLM::setFamily\n You must define the Y Distribution between (kPoisson|kGamma|kBinomial|kNormal),\n and you use the \"kUnknown\" (%d) parameter.",
                            _nDistribution));

        if (nYDistribution == kNormal)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TGLM::setFamily\n The distribution \"kNormal\" (%d) is not yet implemented.\n You can use the TLinearRegression Uranie Classe.",
                            _nDistribution));

        // Set the Y distribution
        _nDistribution = nYDistribution;

        // Set the link function
        if (nLinkFunction == kCanonical)
        {
            // Case of using canonical function
            switch (_nDistribution)
            {
            case kPoisson:
                _nLinkFunction = kLog;
                break;
            case kGamma:
                _nLinkFunction = kInverse;
                break;
            case kBinomial:
                _nLinkFunction = kLogit;
                break;
	    default:
		break;
            }
        }
        else
        {
            if ((_nDistribution == kPoisson)
                    && !((nLinkFunction == kLog) || (nLinkFunction == kIdentity)))
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form(
                                "TGLM::setFamily\n The distribution \"kPoisson\" (%d) is only developped for the (kLog|kIdentity) link functions,\n And you use the (%d) link function.",
                                _nDistribution, nLinkFunction));
            if ((_nDistribution == kGamma) && !(nLinkFunction == kInverse))
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form(
                                "TGLM::setFamily\n The distribution \"kGamma\" (%d) is only developped for the (kInverse) link function,\n And you use the (%d) link function.",
                                _nDistribution, nLinkFunction));
            if ((_nDistribution == kBinomial) && !(nLinkFunction == kLogit))
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form(
                                "TGLM::setFamily\n The distribution \"kBinomial\" (%d) is only developped for the (kLogit) link function,\n And you use the (%d) link function.",
                                _nDistribution, nLinkFunction));

            _nLinkFunction = nLinkFunction;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  set Initial vector of Beta
// -----------------------------------------------------------------------
void TGLM::setInitialBeta(TVectorD vecIni)
{
    Int_t nP = _nX;
    if (_bIntercept)
        nP++;
    // Initial Beta;
    _vecBetaInit.ResizeTo(nP);
    _vecBetaInit = 1.0;

    try
    {
        if (nP != vecIni.GetNrows())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGLM setInitialBeta\n The number of rows of the Initial Beta [%d] is not equal to the number of parameters [%d]",
                            vecIni.GetNrows(), nP));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " ** _vecBetaInit size(" << _vecBetaInit.GetNrows() << ")"
                << endl;
    _vecBetaInit = vecIni;

}
// -----------------------------------------------------------------------
//  fillData
// -----------------------------------------------------------------------
void TGLM::fillData(TMatrixD &Aw, TVectorD &yw, TMatrixD &matY)
{
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TGLM::fillData" << endl;
    }

    Int_t nP = _nX;
    if (_bIntercept)
        nP++;
    Aw.ResizeTo(_nS, nP);
    matY.ResizeTo(_nS, _nY);
    yw.ResizeTo(_nS);
    yw = 1.0;

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputLeaves = new TList();
    TIter next(_listOfInputAttributes);
    if (_blog)
        cout << " ** input" << endl;
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
        if (_blog)
        {
            _tds->getTuple()->GetLeaf(att->GetName())->Print();
            att->printLog();
        }
    }

    TLeaf *leafY = NULL;
    TList *listOfOutputLeaves = new TList();
    if (_nY == 1)
    {
        // Case of the output (only one attribut) : Fill the vector Y
        leafY = _tds->getTuple()->GetLeaf(_sOutputAttributes.Data());
        if (_blog)
        {
            cout << " ** output" << endl;
            _tds->getTuple()->GetLeaf(_sOutputAttributes)->Print();
            _tds->getAttribute(_sOutputAttributes)->printLog();
        }
    }
    else
    {
        TIter nextout(_listOfOutputAttributes);
        if (_blog)
            cout << " ** Output" << endl;
        while (URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) nextout())
        {
            listOfOutputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
            if (_blog)
            {
                _tds->getTuple()->GetLeaf(att->GetName())->Print();
                att->printLog();
            }
        }
    }

    // Fill the matrix and vector of data
    Int_t ivar;
    for (Int_t ipat = 0; ipat < _nS; ipat++)
    {
        _tds->getTuple()->GetEntry(ipat);
        if ((ipat < 5) && (_blog))
            _tds->getTuple()->Show();

        if (_blog)
            cout << " ** ipat[" << ipat << "/" << _nS << "]" << endl;

        // Case of the input attributes : Fill the matrix Aw
        ivar = 0;
        if (_bIntercept)
            Aw(ipat, ivar++) = 1.0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
            Aw(ipat, ivar++) = leaf->GetValue(0);

        // Case of output attribute : Fill the vector yw
        ivar = 0;
        if (_nY == 1)
        {
            yw(ipat) = leafY->GetValue(0);
        }
        else
        {
            TIter nextLeafout(listOfOutputLeaves);
            while (TLeaf *leaf = (TLeaf *) nextLeafout())
            {
                if (ivar == 0)
                    yw(ipat) = leaf->GetValue(0);
                matY(ipat, ivar++) = leaf->GetValue(0);
            }
        }
    }

    delete listOfInputLeaves;
    listOfInputLeaves = NULL;
    delete listOfOutputLeaves;
    listOfOutputLeaves = NULL;

    if (_blog)
    {
        cout << " ** Matrix Aw(" << Aw.GetNrows() << "," << Aw.GetNcols() << ")"
                << endl;
        Aw.Print();
        cout << " ** Vector Yw(" << yw.GetNrows() << ")" << endl;
        yw.Print();
        cout << " ** matrice matY (" << matY.GetNrows() << ","
                << matY.GetNcols() << ")" << endl;
        matY.Print();
    }

    if (_blog)
    {
        cout << " ** End of TGLM::fillData" << endl;
        cout << "**************************************" << endl;
    }
}

// -----------------------------------------------------------------------
//  estimate
// -----------------------------------------------------------------------
void TGLM::estimate(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TGLM::estimate" << endl;
    }
    try
    {
        if (_nDistribution == kUnknown)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TGLM::estimate\n You must define the Y Distribution between (kPoisson|kGamma|kBinomial|kNormal)\n  by the setFamily method before.");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Calculate the mean/Standard deviation on each input and output  attributes
    _tds->computeStatistic(
            Form("%s:%s", _sInputAttributes.Data(), _sOutputAttributes.Data()));

    TMatrixD X(_nS, _nX);
    TMatrixD matrixY(_nS, _nY);
    TVectorD vecY(_nS);

    fillData(X, vecY, matrixY);

    if (_nDistribution == kBinomial)
    {
        for (Int_t i = 0; i < _nS; i++)
        {
            vecY(i) = matrixY(i, 0) / (matrixY(i, 0) + matrixY(i, 1));
        }
    }

    Int_t nP = X.GetNcols();
    _parameterValues.ResizeTo(nP);
    Bool_t ok = kFALSE;

    if (_vecBetaInit.GetNrows() >= 1)
    {
        // Use the init vector of beta
        _parameterValues = _vecBetaInit;
    }
    else
    {
        // Compute initial values of beta0
        TVectorD vecY0(_nS);
        switch (_nDistribution)
        {
        case kPoisson:
            for (Int_t i = 0; i < _nS; i++)
            {
                vecY0(i) = TMath::Log(vecY(i));
            }
            break;
        case kGamma:
            for (Int_t i = 0; i < _nS; i++)
            {
                vecY0(i) = 1.0 / vecY(i);
            }
            break;
        case kBinomial:
            for (Int_t i = 0; i < _nS; i++)
            {
                vecY0(i) = vecY(i);
            }
            break;
        default:
            for (Int_t i = 0; i < _nS; i++)
            {
                vecY0(i) = vecY(i);
            }
        }

        if (_blog)
        {
            cout << " ** Matrix X(" << X.GetNrows() << "," << X.GetNcols()
                    << ")" << endl;
            X.Print();
            cout << " ** Vector vecY(" << vecY.GetNrows() << ")" << endl;
            vecY.Print();
        }

        // Compute the linear regression
        TDecompSVD svd(X);
        _parameterValues = svd.Solve(vecY0, ok);
        if (ok)
        {
            _vecBetaInit.ResizeTo(_parameterValues.GetNrows());
            for (Int_t i = 0; i < _parameterValues.GetNrows(); i++)
                _vecBetaInit(i) = _parameterValues(i);
            if (_blog)
            {
                cout << " ** Solution (" << _parameterValues.GetNrows() << ")"
                        << endl;
                _parameterValues.Print();
            }
        }
    }

    if (_blog)
    {
        cout << " ** INITIAL Vector of parameters ("
                << _parameterValues.GetNrows() << ")" << endl;
        _parameterValues.Print();
    }
    // Compute eta : X * beta0
    TVectorD eta = X * _parameterValues;
    if (_blog)
    {
        cout << " ** Vector eta(" << eta.GetNrows() << ")" << endl;
        eta.Print();
    }

    TVectorD mu(_nS);
    mu.Zero();
    // Compute mu : g^{-1}(eta)
    switch (_nDistribution)
    {
    case kPoisson:
        for (Int_t i = 0; i < _nS; i++)
        {
            if (_nLinkFunction == kLog)
                mu(i) = TMath::Exp(eta(i));
            if (_nLinkFunction == kIdentity)
                mu(i) = eta(i);
        }
        break;
    case kGamma:
        for (Int_t i = 0; i < _nS; i++)
        {
            mu(i) = 1.0 / eta(i);
        }
        break;
    case kBinomial:
        for (Int_t i = 0; i < _nS; i++)
        {
            if (_nLinkFunction == kLogit)
                mu(i) = TMath::Exp(eta(i)) / (1.0 + TMath::Exp(eta(i)));
        }
        break;
    default:
	break;
    }
    if (_blog)
    {
        cout << " ** Vector mu(" << mu.GetNrows() << ")" << endl;
        mu.Print();
    }

    TMatrixD Aw(_nS, nP);
    Aw.Zero();

    TVectorD omega(_nS);
    omega.Zero();
    TVectorD vecZw(_nS);
    vecZw.Zero();

    Bool_t bstop = kFALSE;
    if (_nMaxIteration > 0)
        bstop = kTRUE;
    Int_t niteration = 1;

    while (bstop)
    {
        if (_blog)
        {
            cout << endl << endl << "*****************************************"
                    << endl;
            cout << " **niteration[" << niteration << "/" << _nMaxIteration
                    << "]" << endl;
        }

        switch (_nDistribution)
        {
        case kPoisson:
            if (_blog)
                cout << endl << " *********** case kPoisson[" << _nDistribution
                        << "]" << endl;
            for (Int_t i = 0; i < _nS; i++)
            {
                // Update the weights vector
                if (_nLinkFunction == kLog)
                    omega(i) = sqrt(mu(i));
                if (_nLinkFunction == kIdentity)
                    omega(i) = 1.0 / sqrt(mu(i));
                // Define the weighted "adjusted dependent variable"
                //    Zw = matDiagW * vecZ
                if (_nLinkFunction == kLog)
                    vecZw(i) = omega(i) * (eta(i) + (vecY(i) - mu(i)) / mu(i));
                if (_nLinkFunction == kIdentity)
                    vecZw(i) = omega(i) * vecY(i);
                // Define the weighted matrice
                //    Aw = matDiagW * X
                for (Int_t j = 0; j < nP; j++)
                {
                    Aw(i, j) = omega(i) * X(i, j);
                }
            }
            break;
        case kGamma:
            if (_blog)
                cout << endl << " *********** case kGamma[" << _nDistribution
                        << "]" << endl;
            for (Int_t i = 0; i < _nS; i++)
            {
                // Update the weights vector
                omega(i) = TMath::Abs(mu(i));
                // Define the weighted "adjusted dependent variable"
                //    Zw = matDiagW * vecZ
                vecZw(i) = omega(i)
                        * (eta(i) - (vecY(i) - mu(i)) / (mu(i) * mu(i)));
                // Define the weighted matrice
                //    Aw = matDiagW * X
                for (Int_t j = 0; j < nP; j++)
                {
                    Aw(i, j) = omega(i) * X(i, j);
                }
            }
            break;
        case kBinomial:
            if (_blog)
                cout << endl << " *********** case kBinomial[" << _nDistribution
                        << "]" << endl;
            for (Int_t i = 0; i < _nS; i++)
            {
                // Update the weights vector
                if (_nLinkFunction == kLogit)
                    omega(i) = sqrt(mu(i) * (1.0 - mu(i)));
                // Define the weighted "adjusted dependent variable"
                //    Zw = matDiagW * vecZ
                if (_nLinkFunction == kLogit)
                    vecZw(i) = omega(i)
                            * (eta(i)
                                    + (vecY(i) - mu(i))
                                            / (mu(i) * (1.0 - mu(i))));
                // Define the weighted matrice
                //    Aw = matDiagW * X
                for (Int_t j = 0; j < nP; j++)
                {
                    Aw(i, j) = omega(i) * X(i, j);
                }
            }
            break;
	default:
	    break;
        }

        if (_blog)
        {
            cout << " ** Matrice Diagonal omega(" << omega.GetNrows() << ")"
                    << endl;
            omega.Print();
            cout << " ** Vector vecZw(" << vecZw.GetNrows() << ")" << endl;
            vecZw.Print();
        }

        // Resolve the weighted problem Zw = Aw * beta
        TDecompSVD svd(Aw);
        ok = kFALSE;
        _parameterValues = svd.Solve(vecZw, ok);

        if (ok)
        {
            if (_blog)
            {
                cout << " ** Vector of parameters solution _parameterValues ("
                        << _parameterValues.GetNrows() << ")" << endl;
                _parameterValues.Print();
            }
        }

        // Compute eta : X * beta
        eta = X * _parameterValues;
        if (_blog)
        {
            cout << " ** Vector eta(" << eta.GetNrows() << ")" << endl;
            eta.Print();
        }

        // Compute mu : g^{-1}(eta)
        switch (_nDistribution)
        {
        case kPoisson:
            for (Int_t i = 0; i < _nS; i++)
            {
                if (_nLinkFunction == kLog)
                    mu(i) = TMath::Exp(eta(i));
                if (_nLinkFunction == kIdentity)
                    mu(i) = eta(i);
            }
            break;
        case kGamma:
            for (Int_t i = 0; i < _nS; i++)
            {
                mu(i) = 1.0 / eta(i);
            }
            break;
        case kBinomial:
            for (Int_t i = 0; i < _nS; i++)
            {
                if (_nLinkFunction == kLogit)
                    mu(i) = TMath::Exp(eta(i)) / (1.0 + TMath::Exp(eta(i)));
            }
            break;
	default:
	    break;
        }
        if (_blog)
        {
            cout << " ** Vector mu(" << mu.GetNrows() << ")" << endl;
            mu.Print();
            cout << " ** Vector vecY(" << vecY.GetNrows() << ")" << endl;
            vecY.Print();
        }

        Double_t dDeviance = -1.0;
        Double_t ddev = 0.0;
        // Compute the deviance
        switch (_nDistribution)
        {
        case kPoisson:
            dDeviance = 0.0;
            for (Int_t i = 0; i < _nS; i++)
            {
                dDeviance += vecY(i) * TMath::Log(vecY(i) / mu(i)) - vecY(i)
                        + mu(i);
            }
            break;
        case kGamma:
            dDeviance = 0.0;
            for (Int_t i = 0; i < _nS; i++)
            {
                dDeviance += -1.0 * TMath::Log(vecY(i) / mu(i))
                        + (vecY(i) - mu(i)) / mu(i);
            }
            break;
        case kBinomial:
            dDeviance = 0.0;
            for (Int_t i = 0; i < _nS; i++)
            {
                ddev = matrixY(i, 0) * vecY(i) * TMath::Log(vecY(i) / mu(i));
                if ((1.0 - vecY(i)) > 1e-8)
                {
                    ddev += matrixY(i, 0) * (1.0 - vecY(i))
                            * TMath::Log((1.0 - vecY(i)) / (1.0 - mu(i)));
                }
                dDeviance += ddev;
            }
            break;
        default:
        {
            dDeviance = -1.0;
        }
            break;
        }
        dDeviance *= 2.0;
        Double_t dstep = 1000.0 * _dTolerance;

        if (niteration > 1)
        {
            dstep = TMath::Abs(dDeviance - _dResidualDeviance)
                    / (0.1 + TMath::Abs(dDeviance));
        }
        if (_blog)
            cout << endl << " *********** dDeviance[" << dDeviance << "] dstep["
                    << dstep << "]" << endl;
        if ((niteration == _nMaxIteration) || (dstep < _dTolerance))
        {
            bstop = kFALSE;
            _nIteration = niteration;
        }

        _dResidualDeviance = dDeviance;

        if (_blog)
            cout << " ** End of niteration[" << niteration << "/"
                    << _nMaxIteration << "]" << endl;
        niteration++;
    }

    if ((_nY < 3) && (_bStoreYHat))
    {
        TString sy = _sOutputAttributes;
        if (_nY == 2)
        {
            TString svar = _sOutputAttributes;
            svar.ReplaceAll(" ", "");
            svar.ReplaceAll(":", " ");
            istringstream inStream((char *) svar.Data());
            string inName;
            inStream >> inName;
            sy = TString(inName.c_str());
        }
        if (_blog)
            cout << " sy[" << sy << "]" << endl;

        TString syhat = Form("%shat", sy.Data());
        URANIE::DataServer::TAttribute *attY = _tds->getAttribute(sy);
        _tds->addAttributeByNameTitle(syhat, Form("#hat{%s}", attY->GetTitle()));
        _tds->getAttribute(syhat)->setUnity(attY->getUnity());

//       Double_t tabYValue = 0.0;
        Double_t *tabYValue = new Double_t[5];
        TBranch *tbr = _tds->getTuple()->Branch(syhat, tabYValue,
                Form("%s/D", syhat.Data()));

        for (Int_t is = 0; is < _nS; is++)
        {
            //_tds->getTuple()->GetEntry(is);
            if (_nY == 1)
                tabYValue[0] = mu(is);
            else
                tabYValue[0] = (matrixY(is, 0) + matrixY(is, 1)) * mu(is);
            // Fill the #hat{y} data
            tbr->Fill();
        }
        // clean
        _tds->getTuple()->ResetBranchAddress(tbr);
        delete[] tabYValue;
        tabYValue = NULL;
        //       _tds->saveTuple("UPDATE");
    }

    if (_blog)
    {
        cout << " ** FINAL :: Iteration (" << _nIteration
                << ") Vector of parameters solution ("
                << _parameterValues.GetNrows() << ")" << endl;
        _parameterValues.Print();
    }
}

// -----------------------------------------------------------------------
//  Export Functions
// -----------------------------------------------------------------------
void TGLM::exportModelCplusplus(std::ofstream * sourcefile) const
{
    cout << " ** TGLM::exportModelCplusplus ** " << endl;
    // Model Quality 
    *sourcefile << "  //" << endl;
    *sourcefile << "  //     QUALITY  : " << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  ////////////////////////////// " << endl << endl;

    URANIE::DataServer::TAttribute *att;
    Int_t ivarinc = 0;
    if (_bIntercept)
    {
        *sourcefile << "  // Intercept" << endl;
        *sourcefile << "  double y = " << _parameterValues[0] << ";" << endl;
        ivarinc = 1;
    }
    else
    {
        *sourcefile << "  // No Intercept" << endl;
        *sourcefile << "  double y = 0.0 ;" << endl;
    }
    *sourcefile << endl;
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                ivar);
        *sourcefile << "  // Attribute : " << att->GetName() << endl;
        *sourcefile << "  y += param[" << ivar << "]*"
                << _parameterValues[ivar + ivarinc] << ";" << endl << endl;
    }

    *sourcefile << "  // Return the value" << endl;
    switch (_nDistribution)
    {
    case kBinomial:
        if (_nLinkFunction == kLogit)
            *sourcefile << "  res[0] = 1.0 / ( 1.0 + exp(-1.0 * y));" << endl;
        break;
    default:
        *sourcefile << "  res[0] = y;" << endl;
    }

}
void TGLM::exportModelFortran(std::ofstream * sourcefile) const
{
    UNUSED(sourcefile);
    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
            "TGLM::exportFunction\n The FORTRAN language is not yet implemented.");
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TGLM::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TGLM::printLog[" << option << "]" << endl;
    cout << " ** _nDistribution[" << _nDistribution << "],  _nLinkFunction["
            << _nLinkFunction << "]" << endl;
    cout << " ** Null deviance[" << _dNullDeviance << "] Residual deviance["
            << _dResidualDeviance << "]" << endl;
    if (_parameterValues.GetNrows() > 0)
    {
        cout << " ** Vector of parameters (" << _parameterValues.GetNrows()
                << ")" << endl;
        _parameterValues.Print();
    }
    TModeler::printLog();
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE
