//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGLM.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-14 12:54:45
 
 \brief Interface de la classe URANIE::Modeler::TGLM
 
 */
/*
 @class URANIE::Modeler::TGLM
 
 @brief Description of the class TGLM.
 
 This class implements the \b "Generalized \b Linear \b Models" model

 The algorith implemented is the \b Iterative \b Weighted \b Least \b Squares (\b IWLS ).
 \li Assign starting values to the parameters \f$ beta^{(0)}\f$, generally by resolving the Least Squares Normal equations (i.e. without weights);
 \li Update the \f$ \eta^{(i)}\f$ and \f$ \mu^{(i)}\f$ with parameters \f$ beta^{(i)}\f$
 \li Update the weights using the new estimated mean vector: \f$ \frac{1}{v^(i+1)} = \var(\mu^{i}) * f(\mu^{i}) \f$
 \li Resolve the Weighted Least Squares Normal equations to give \f$ beta^{(i+1)}\f$
 \li Repeat the thre last steps until convergence (\em i.e. \f$ X\hat{\Beta^{i}} - X\hat{\Beta^{i+1}}\f$ is sufficiently close to zero).
 */
#ifndef TGLM_H
#define TGLM_H

//ROOT

// Uranie
#include "TModeler.h"

namespace URANIE
{
namespace Modeler
{
class TGLM: public TModeler
{
    // Attributs
public:
    enum EDistribution
    {
        kUnknown, kPoisson, kGamma, kBinomial, kNormal
    };
    enum ELinkFunctin
    {
        kCanonical, kIdentity, kLogit, kProbit, kLogLog, kLog, kInverse
    };
private:
    EDistribution _nDistribution;                ///< The type of Y distribution
    ELinkFunctin _nLinkFunction;     ///< The link function (default kCanonical)
    Int_t _nIteration;   ///< the number of iteration at convergence (default 0)
    Int_t _nMaxIteration;      ///< the number of maximum iteration (default 25)

    TVectorD _vecBetaInit;                         ///< Inital Vector for beta
    Double_t _dTolerance;   ///< The tolerance to stop algorithm (default 1.e-6)
    Double_t _dNullDeviance;                       ///< The Null Deviance value
    Double_t _dResidualDeviance;                ///< The Residual Deviance value
    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with the TDataServer and the architecture
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param architecture(TString) The architecture (input and output attributes) of the model separated by the caracter ":"
     \param option (Option_t *)[""]  The option to pass
     */
    TGLM(URANIE::DataServer::TDataServer *tds, TString architecture,
            Option_t *option = "");
    //! Default constructor with input and output attributes
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
     \param option (Option_t *)[""]  The option to pass
     */
    TGLM(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");
    //! Default destructor
    virtual ~TGLM();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //! Set the family of error (Distribution of \f$ Y\f$ and Link Function)
    /*!
     \param nYDistribution (EDistribution) The distribution of \f$ Y\f$ (kBinomial|kPoisson|kGamma)
     \param nLinkFunction (ELinkFunctin) The link function. By defulalt, the canonical function, which depend of the Y distribution, is used. [kCanonical](kCanonical|kIdentity|kLogit|kProbit|kLog)
     */
    void setFamily(EDistribution nYDistribution = kUnknown,
            ELinkFunctin nLinkFunction = kCanonical);
    //! Return the type of Y distribution
    EDistribution getYDistribution()
    {
        return _nDistribution;
    }
    //! Return the link function
    ELinkFunctin getLinkFunction()
    {
        return _nLinkFunction;
    }
    //! Set the number of maximum iteration
    void setMaxIteration(Int_t n)
    {
        if (n > -1)
            _nMaxIteration = n;
    }
    //! Get the number iterations at convergence
    Int_t getNIteration()
    {
        return _nIteration;
    }
    //! Get the number of maximum iteration
    Int_t getMaxIteration()
    {
        return _nMaxIteration;
    }
    //! Set the Tolerance
    void setTolerance(Double_t dtol)
    {
        if (dtol > 0)
            _dTolerance = dtol;
    }
    //! Get the Tolerance
    Double_t getTolerance()
    {
        return _dTolerance;
    }
    //! Get the Null Deviance value
    Double_t getNullDeviance()
    {
        return _dNullDeviance;
    }
    //! Get the Residual Deviance value
    Double_t getResidualDeviance()
    {
        return _dResidualDeviance;
    }
    //! Set the initial parameter
    void setInitialBeta(TVectorD vecIni);
    //! Get the initial parameter
    TVectorD getInitialBeta()
    {
        return _vecBetaInit;
    }
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{
    void estimate(Option_t * option = "");
    //! Fill the data in a TMatrix and a TVector
    /*!
     \todo To deplace in the TModeler Class because it is the same for TLinearRegression.
     */
    void fillData(TMatrixD &Aw, TVectorD &yw, TMatrixD &matY);
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const;
    
    
    //! Export the model in a PMML file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPMML(const char* file = "", 
                         const char* name = "",Option_t *option = "") const
    {
        TString soption = option;
        soption.ToLower();
	UNUSED(file,name);
        std::cerr
                << "   --- Method TGLM::exportModelPMML(ofstream * sourcefile) not yet implemented"
                << endl;
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
            "TGLM::exportFunction\n The PMML language is not yet implemented.");        
    };
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TGLM::exportModelPython(ofstream * sourcefile) not yet implemented"
                << endl;
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
            "TGLM::exportFunction\n The Python language is not yet implemented.");        
    };
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TGLM, ID_MODELER)
    //Classe de
};
} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif // TGLM_H
