//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPBuilder.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date

 \brief Interface de la classe URANIE::Modeler::TGPBuilder

 */
/**

 * @class URANIE::Modeler::TGPBuilder
 *
 * @brief Description of the class TGPBuilder.
 *
 * This class allows to build a gaussian processus (kriging) model from a set 
 * of observations and a correlation function. It provides tools to search for 
 * the optimal parameters of the correlation function and compute the parameters 
 * of the model.
 * It wraps the gpModx functions of the "gaussian process Library" (gpLib), 
 * by Jean-Marc Martinez.
 *
 * Remark about normalisation: the observed input variable are automatically 
 * normalised such that they belong to [0,1]. The normalisation coefficients 
 * are exported into the TKriging model and new input coordinates are thus 
 * automatically normalised.
 * As a consequence, the correlation lengths of the correlation 
 * function are normalised as well. 
 * A special case happens when the deterministic trend is not a constant or a 
 * linear combination of the inputs. In that case, the values of the trend 
 * coefficients are calculated in the "original" space.  
 *
 */
#ifndef TGPBUILDER_H
#define TGPBUILDER_H

//// ROOT
#include "TObjArray.h"

//// URANIE
#include "TCorrelationFunction.h"
#include "TShareEval.h"
#include "TDataServer.h"       
#include "TAttribute.h"   
#include "TKriging.h"

//// Jean-Marc's Gaussian Processes Library
extern "C"
{
int gpMod0(double *Kin, double *yin, int n, double *gamma, double *iL);
int gpModh(double *Hin, double *Kin, double *yin, int n, int p, double *hbeta, double *vbeta, double *gamma, double *iL, double *iR, double *M);
int gpModhBayes(double *BetaPrior, double *QPrior, double *Hin, double *Kin, double *yin, int n, int p, double *BetaPost, double *QPost, double *gamma, double *iL, double *iR, double *M);

int gpLoo0Error(double *Kin, double *yin, int n, double *errorLoo, double *varLoo, double *eqmLoo);
int gpLoohError(double *Hin, double *Kin, double *yin, int n, int p, double *errorLoo, double *varLoo, double *eqmLoo);
int gpLoohBayesError(double *BetaPrior, double *Qprior, double *Hin, double *Kin, double *yin, int n, int p, double *errorLoo, double *varLoo, double *eqmLoo);
}
;

namespace URANIE
{
    namespace Modeler
    {
        class TGPBuilder
        {
        
            // Attributes
        private:
            Bool_t _blog; ///< Boolean for edit the log

            void LSort(Double_t *costValues, Int_t *minIndexes, Int_t nbthreads, Int_t n);

        public:
            URANIE::DataServer::TDataServer *_tds; ///< Pointer to the data server containing the observations
        
            TString _sInputAttributes;   ///< input attributes names separated by ":"
            TString _sOutputAttributes;  ///< output attribute name
            Int_t _nX;  ///< Number of input attributes
            Int_t _nY;  ///< Number of output attributes (should be equal to 1)
        
            TCorrelationFunction* correlationFunction; ///< The correlation function
            Bool_t bHas_corrFunction; ///< kFALSE if correlationFunction == NULL, kTRUE otherwise
            Bool_t bVariance_in_factor; ///< if kTRUE, GP variance is calculated analytically. Otherwise, it is a parameter of the cost function.
            Bool_t bHas_measurement_error; ///< if kTRUE, the construction of the GP will take into account the existence of a measurement error variance. 
            
            Int_t _nS;        ///< Number of observations
            Int_t _nP;        ///< Number of coefficients of the deterministic trend
            Double_t reg;     ///< regularisation parameter, used to improve numerical stability of matrix operations
            Double_t tolerance;  ///< Tolerance value used to stop the optimisation process
	    Double_t sigma2;  ///< Variance of the gaussian process
            Double_t alpha;   ///< Quotient of the variance of the measurment error and the variance of the gaussian process.
            Double_t* Vmes;   ///< The covariance matrix of the measurement errors (dimension: _nS * _nS)
            Double_t* H;      ///< Deterministic trend matrix (dimension: _nS * _nP)
            Double_t* C;      ///< The correlation matrix (dimension: _nS * _nS)
            Double_t* K;      ///< The covariance matrix (dimension: _nS * _nS)
            Double_t* xObs;   ///< The matrix of normalised inputs (dimension: _nX * _nS)            
            Double_t* yObs;   ///< Array representation of the outputs (size: _nS)
            Double_t* xNormParams; ///< The matrix of normalisation parameters. Contains min and max value of each input variable: [x1Min, x1Max, x2Min, x2Max,...] (size: _nX * 2)        

            Double_t  _modlengthMax; ///< Maximum value of the correlation length if default behaviour is requested to be changed by the user.
            Double_t  _modlengthMin; ///< Minimum value of the correlation length if default behaviour is requested to be changed by the user.
            Double_t  _lengthMaxScreeningSF; ///< Scale factor to change the Maximum value of the correlation length during screening procedure: new length is SF * lengthMax (0 < SF <=1)
            
            TString trendString;       ///< Character string containing the trend parameters separated by ":" or a trend type ("Const" or "Linear")
            TObjArray* trendCoefList;  ///< List of the formulas of the trend coefficients (_nP elements)
            Bool_t bHas_trend;         ///< kFALSE if no deterministic trend is defined, kTRUE otherwise.
            Bool_t bUse_normalisation; ///< if kFALSE, the matrix H will be computed with unormalised input data. This happens when the user is providing his own trend function. When the trend is constructed automatically from a trend type, this attribute is set to kTRUE. 
            
            Double_t* BetaPrior; ///< a priori mean vector for the deterministic trend parameters (size: _nP)
            Double_t* QPrior;    ///< a priori variance matrix for the deterministic trend parameters (size: _nP x _nP)
            Bool_t bUse_prior;   ///< kFALSE if no prior is used, kTRUE otherwise
            Int_t _nSeed; ///< number of seed for pseudo-random generator

	    Int_t _retOptim; ///< optimisation results

            Bool_t bIsbuildWithDuplicateKandCandCorrFunc;   ///< kTRUE if build with duplicate K, C and CorrFunc, otherwise kFALSE
            
            // Operations
        public:
            //---------------------------------------------
            /** @name Constructor and Destructor
             **/
            //@{
            //! Default Constructor
            TGPBuilder();
            
            //! Construction from a data server, a list of input and output variables and a correlation function object.
            /*!
             \param tdsObs:       a pointer to the data server containing the observations.             
             \param varexpinput:  a character string containing the names of the input variables, separated by colons (":").
             \param varexpoutput: a character string containing the name of the output variable.
             \param corrFunc:     a pointer to the correlation function.
             \param trend:        a string containing the list of deterministic trend coefficients separated by colons ":" (default = ""). (cf. TGPBuilder::createTrend)
             \param option:       a character string containing optional commands. None is avaiblable at this time (default = "").
             \exception UErrorException if varexpoutput contains more than one variable name.
             */
            TGPBuilder(URANIE::DataServer::TDataServer *tdsObs,
                       const char *varexpinput,
                       const char *varexpoutput, 
                       TCorrelationFunction* corrFunc, 
                       const char *trend = "",
                       Option_t   *option = "");

            //! Construction from a data server, a list of input and output variables and a correlation function name.
            /*!
             \param tdsObs:       a pointer to the data server containing the observations.             
             \param varexpinput:  a character string containing the names of the input variables, separated by colons (":").
             \param varexpoutput: a character string containing the name of the output variable.
             \param corrFunc:     name of the correlation function. Available names are: 
                                  "gauss", "isogauss", "exponential", "maternI", "maternII", "maternIII", "matern3/2", "matern5/2", "matern7/2". 
                                  The class name of a correlation function is also valid (e.g. "TExponentialCorrFunction", "TMatern32CorrFunction", etc.)
             \param trend:        a string containing the list of deterministic trend coefficients separated by colons ":" (default = ""). (cf. TGPBuilder::createTrend)
             \param option:       a character string containing optional commands. None is available at this time (default = "").
             \exception UErrorException if varexpoutput contains more than one variable name.
             */
            TGPBuilder(URANIE::DataServer::TDataServer *tdsObs,
                       const char *varexpinput,
                       const char *varexpoutput, 
                       const char *corrFunc, 
                       const char *trend = "",
                       Option_t   *option = "");

            //! Construction from a data server containing model parameters
            /*!
             This constructor works ONLY with a data server generated by the 
             function TGPBuilder::exportGPData. A set of parameters is written 
             in the #TITLE: field of the file header. These parameters are 
             extracted and used to initialise the GPBuilder. The rest of the 
             file contains the observations (inputs and output). After the 
             construction of the TGPBuilder object, the user can call the function 
             TGPBuilder::buildGP to generate the TKriging object. 
             \param tdsObs (URANIE::DataServer::TDataServer*): a pointer to the data server containing the observations.             
             \note a message is displayed if the user needs to define the measurement 
                   error variance or the a priori information on the trend parameters.            
             */
            TGPBuilder(URANIE::DataServer::TDataServer *tdsModelData);

            TGPBuilder* buildWithDuplicateKandCandCorrFunc();
           
            //! Initialisation of the GPBuilder variables
            void initVariables();
            
            //! Initialisation of the matrices
            void initMatrices();
            
            //! Create the matrices for the deterministic trend
            void createTrend();           

            //! Default destructor
            virtual ~TGPBuilder();            
            //@}
        
            //---------------------------------------------
            /** @name Get methods
             **/
            //@{
            //! Return kTRUE if a correlation function is defined, kFALSE otherwise.
            Bool_t hasCorrFunction()
            {
                return bHas_corrFunction;
            }

            //! Return kTRUE if the variance of the GP is determined analytically.
            Bool_t hasVarianceInFactor()
            {
                return bVariance_in_factor;
            }

            //! Return kTRUE if a measurement error should be considered to build the GP
            Bool_t hasMeasurementError()
            {
                return bHas_measurement_error;
            }            
            
            //! Return kTRUE if a deterministic trend has been defined
            Bool_t hasTrend()
            {
                return bHas_trend;
            }
                    
            //! Return kTRUE if prior knowledge on the deterministic trend is used (bayesian approach).
            Bool_t usePrior()
            {
                return bUse_prior;
            }

	    //! Return the return of the optimisation procedure
	    int getReturnOptim()
		{
		    return _retOptim;
		}
	    
            //! Return a pointer to the correlation function of the gaussian process
            TCorrelationFunction* getCorrFunction()
            {
                return correlationFunction;
            }
        
            //! Return the number of input variables
            Int_t getNInputs()
            {
                return _nX;
            }
        
            //! Return the number of observations used to build the GP model
            Int_t getNObs()
            {
                return _nS;
            }
        
            //! Return the number of parameters of the deterministic trend
            Int_t getNTrendParams()
            {
                return _nP;
            }
        
            //! Return the observations inputs as a flat array
            /*!
            For M observations of inputs a, b and c, the returned array is 
            organised as follows: [a1, b1, c1, a2, b2, c2, ... , aM, bM, cM]            
            */
            Double_t* getObsInputs()
            {
                return xObs;
            }
        
            //! Return the observations outputs
            Double_t* getObsOutput()
            {
                return yObs;
            }

            //! Return the normalisation parameters of the inputs as a flat array
            /*!
            Normalisation parameters are the minimum and maximum values of the 
            input variables. For input variables x1, x2, ... xN, they are stored 
            as follows: [x1min, x1max, x2min, x2max, ... , xNmin, xNmax].
            */
            Double_t* getObsNormParams()
            {
                return xNormParams;
            }
            
            //! Return the array of the covariance matrix of the gaussian process
            Double_t* getCovarianceMatrix()
            {
                return K;
            }

            void setKandCandCorrFunc(Double_t* inK, Double_t* inC, TCorrelationFunction* corrFunc)
            {
               K = inK;
               C = inC;
               correlationFunction = corrFunc;
               bIsbuildWithDuplicateKandCandCorrFunc = kTRUE;
            }
        
            //! Return the array of the correlation matrix of the gaussian process
            Double_t* getCorrelationMatrix()
            {
                return C;
            }

            //! Return the array of the covariance matrix of the measurement errors
            Double_t* getMeasurementErrorCovMatrix()
            {
                return Vmes;
            }
        
            //! Return the array of the deterministic trend matrix
            Double_t* getTrendMatrix()
            {
                return H;
            }
        
            //! Return the deterministic trend list of formulas
            TObjArray* getTrendFormula()
            {
                return trendCoefList;
            }
        
            //! Return the deterministic trend character string
            TString getTrendString()
            {
                return trendString;
            }
            
            //! Return the array of the a priori means of the trend parameters.
            Double_t* getTrendPriorMeans()
            {
                return BetaPrior;
            }
        
            //! Return the array of the a priori covariance matrix of the trend parameters
            Double_t* getTrendPriorCovMat()
            {
                return QPrior;
            }
        
            //! Return the variance of the gaussian process
            Double_t getVariance()
            {
                return sigma2;
            }

	    //! Return the tolerance used to stop the optimisation process	    
            Double_t getTolerance()
            {
                return tolerance;
            }

            //! Return the alpha parameter, quotient of the variance of the measurement error and the variance of the gaussian process
            Double_t getAlpha()
            {
                return alpha;
            }

            //! Return the regularisation coefficient
            Double_t getRegularisation()
            {
                return reg;
            }
            
            
            //! Return the seed number
            Int_t getSeed()
            {
                return _nSeed;
            }

            //@}
        
            //---------------------------------------------
            /** @name Set methods
             **/
            //@{
            //! Set the Minimun and maximum boundaries for the correlation length 
            /*!
              This function changes the default value of the Minimun and maximum boundaries used to 
              define the space definition of the correlation length.

             \param themin: a double that is the min value allowed for the correlation length (default being 1e-3).
             \param themax: a double that is the max value allowed for the correlation length (default being 10).
             */
            void setLengthRange(double themin, double themax);

            //! Set the scale factor to reduce the maximum value of the correlation length during screening 
            /*!
              This function changes the value of the scale factor that can be used to reduce the maximum 
              value of the correlation length during the screening procedure .

             \param sf: a double (default is 1) that is the scale factor to define a new lengthMax specifically for the screening, 
             as lengthMaxScr = sf * lengthMax. Exception if sf < 0 or sf > 1
             */
            void setLengthMaxScreeningSF(double sf);
            
            //! Set the correlation function of the gaussian process
            /*!
            The internal correlation function is cloned from the correlation 
            function provided as parameter. This means that the later isn't 
            modified by the GPBuilder. 
             \param corrFunc: a pointer to the correlation function of the gaussian process.
             */
            void setCorrFunction(TCorrelationFunction* corrFunc);

            //! Set the correlation function of the gaussian process from a character string
            /*!
            The name of the correlation function can either be its class name 
            (e.g. "TExponentialCorrFunction", "TMatern32CorrFunction", etc.), 
            or one of these (case insensitive) keywords: 
            "gauss", "isogauss", "exponential", "maternI", "maternII", "maternIII", "matern3/2", "matern5/2", "matern7/2".
             \param corrFunc: name of the correlation function.
             */
            void setCorrFunction(const char* corrFuncType);
            
            //! Set the variance of the gaussian process
            /*!
             \param var: value of the variance.
             */
            void setVariance(Double_t var)
            {
                sigma2 = var;
            }

	    //! Set the tolerance used to stop the optimisation process
            /*!
             \param var: value of the tolerance.
             */
            void setTolerance(Double_t tol)
            {
                tolerance = tol;
            }

            //! Set the regularisation coefficient.
            /*!
            The regularisation coefficient is a small (~ 1e-12 to 1e-6) value 
            that is added to the diagonal of the correlation matrix when it 
            is poorly conditionned. It should only be set when matrix 
            decomposition problems arise.
            \param regularisation: value of the regularisation coefficient. It 
                                   is typically between 1e-12 and 1e-6.
            */
            void setRegularisation(Double_t regularisation)
            {
                reg = regularisation;
            }
            
            //! Indicate if a measurement error should be taken into account or not.
            /*!
            This function sets the bHas_measurement_error flag which indicates 
            wether a measurement error should be considered to 
            build the Gaussian Process. If kTRUE, and the user doesn't provide 
            a covariance matrix (using setMeasurementErrorCovMatrix or setMeasurementErrorVariance), 
            the parameter alpha will be a parameter of the cost function, and 
            the variance of the measurment error will be sigma2*alpha. 
            \param has_measurement_error: flag to tell if measurement 
                   error is taken into account. (default =  kTRUE)
            */
            void setHasMeasurementError(Bool_t has_measurement_error = kTRUE)
            {
                bHas_measurement_error = has_measurement_error;
            }

            //! Set the value of the alpha parameter.
            /*!
            This parameter is used when the variance of the measurement error 
            can be expressed as sigma2*alpha, where sigma2 is the variance of 
            the GP. 
            \note calling this function automatically sets the flags bHas_measurement_error 
                  and bVariance_in_factor to kTRUE.
            \param a: the value of the parameter alpha
            */
            void setAlpha(Double_t a)
            {
                alpha = a;
                bHas_measurement_error = kTRUE;
                bVariance_in_factor = kTRUE;
            }
            
            //! Set the covariance matrix of the measurement errors
            /*!
             The data from covMes are copied into Vmes. covMes MUST be a valid covariance matrix. 
             \note Calling this function automatically sets the flag bHas_measurement_error to kTRUE 
                   and the flag  bVariance_in_factor to kFALSE, meaning that the GP variance is an 
                   input parameter of the cost function. 
             \param covMes:  "flat" array of nS*nS doubles.
             */
            void setMeasurementErrorCovMatrix(Double_t* covMes);

            //! Set the variance of the measurement errors
            /*!
             The diagonal of the matrix Vmes is filled with varMes. 
             \note Calling this function automatically sets the flag bHas_measurement_error to kTRUE 
                   and the flag  bVariance_in_factor to kFALSE, meaning that the GP variance is an 
                   input parameter of the cost function.              
             \param varMes:  variance of the measurement error.
             */
            void setMeasurementErrorVariance(Double_t varMes);
            
            //! Set the deterministic trend character string
            /*!
             This function creates the deterministic trend matrix based on a list
             of trend coefficients. This list is created from a character string.
             It can be either a keyword describing the trend type ("const" for a 
             constant value or "linear" for a linear combination of the inputs) 
             or a set of ROOT style formulas of the data server attributes, 
             separated by ":" (cf. ROOT documentation of TFormula::Analyze). 
             For example, if we have attributes x1 and x2, the following strings
             are valid trends: 
             \li "const" will create a constant trend of the form: a0
             \li "linear" will create a trend of the form: a0 + a1*X1 + a2*X2, where X1 et X2 are the normalised values of x1 and x2
             \li "1:x1:x1*x1:log(x2)" will create a trend of the form: a0 + a1*x1 + a2*x1^2 + a3*log(x2)
             \warning if a coefficient formula doesn't contain any reference to
             an attribute of the data server, then it MUST be a constant expression 
             (like "1", "log(2)", "pi", etc.). Otherwise, an error will occure.
             \param trend: the character string containing the trend type ("const" or "linear"), 
                           or a the list of coefficients separated by colons ":".
             */
             void setTrendString(const char* trend)
             {
                 trendString = TString(trend);
             }
                     
            //! Set the a priori mean and variance of the trend coefficients
            /*!
             This function should not be used if no trend was defined. It allows to set the a priori
             mean and variance of the trend coefficients.              
             \warning the size of the array is no checked !
             \note Calling this function automatically sets the flag bUse_prior to kTRUE.
             \param betaPrior: a priori mean vector for the deterministic trend parameters (size: _nP)
             \param qPrior:    a priori covariance matrix for the deterministic trend parameters (size: _nP x _nP)
             \exception UranieError if no trend was defined.
             */
            void setPriorData(Double_t* betaPrior, Double_t* qPrior);
        
            //! Set wether or not to use a priori knowledge on the deterministic trend parameters.
            /*!
             This function should not be used if no trend and no prior data were defined.
             \param use_prior: if kTRUE, the a priori knowledge is used (default = kTRUE).
             \exception UranieError if no trend was defined.
             */
            void setUsePrior(Bool_t use_prior = kTRUE);
            
            //! Computes the covariance matrix of the observed data.
            void computeCovarianceMatrix(bool print = false);
            
            //! set a seed number
            /*!
             This function set the value of _nSeed which indicates the seed number to
             use by the random generator
             \param nSeed:    the seed number to use by the random generator  
             */
            void setSeed(Int_t nSeed = 0);
            
            //@}
               
            //---------------------------------------------
            /** @name Construction of the Gaussian Process
             **/
            //@{
            
            //! Update the GP builder matrices 
            /*! 
            This function shall be called when the data server containing 
            the observations is modified. This is updating the matrices but 
            doesn't change the GP hyper-parameters (correlation lengths, variance, etc.)
            */
            void updateObservations();
            
            //! Export the minimum information necessary to build a new TKriging object.
            /*!
            This function allows to save into an ASCII file (Salome table format) a set 
            of information and data that can later be read by a TGPBuilder object to 
            recreate the full TKriging object. The saved information is:
            \li the type of the correlation function
            \li the correlation function parameters (not normalised)
            \li the gaussian process variance
            \li if needed, the deterministic trend character string
            \li if needed, the variance of the measurement error (except if it was user defined)
            \li if needed, the regularisation coefficient
            \li the observations inputs and outputs.
            \param fileName (const char*): name of the output file.
            */
            void exportGPData(const char* fileName);

    
            //! Search for the optimal parameters of the Gaussian Process
            /*!
             This function allows to search for the parameters of the correlation 
             function which minimise a given criterion.
             The function works in two steps: 
             1. a screening of the parameter space is performed to find a good 
                starting point for the optimisation;
             2. an optimisation process is performed to refine the results.
             
             To skip the screening step, simply set the screening size to 0. 
             The initial point of the optimisation will then be chosen as the 
             set of default value of the parameters to optimise (see below), 
             or their current values if "reset" if set to kFALSE.
             
             To skip the optimisation step, simply set the optimisation 
             algorithm to "" (empty string). The optimal parameters will then 
             be the one corresponding to the optimal output of the screening.
             
             This function is set to be "user-friendly", and takes care of most 
             of the tedious part of the process.
             This implies some limitations:
             - hyper-parameters ranges can't be set by the user;
             - most of the optimisation parameters can't be set by the user;
             - currently supported correlation functions are: Gaussian, 
             Isotropic Gaussian, Exponential, MaternI, MaternII, MaternIII, 
             Matern3/2, Matern5/2 and Matern7/2.
                     
             The consequence of these limitations is that the "optimal" parameters
             might not be that optimal. Thus, if the resulting model isn't good
             enough, using the "not so user-friendly" method might be
             necessary.
             
             Here is the list of the current "hard coded" ranges and default values for 
             the parameters to optimise:
             \li correlation lengths: [1e-3, 10.0], 0.01;
             \li exponent: [0.5, 2.0], 1.0;
             \li Matern's regularity (\nu): [0.5, 7.5], 1.5;
             \li alpha: [1e-6, 10.0], 0.00316;
             \li GP variance coefficient: [1e-6, 1e2], 0.01; (this value is multiplied by the observation output's variance).
        
             \warning This function throws an exception if the NLOpt optimisation 
                      library isn't available.
             \param criterion (TString): criterion to optimize. Can be "ML" 
                    (maximum likelyhood), "ReML" (restrained maximum likelyhood), 
                    or "LOO" (leave one out cross-validation). "ML" can't be 
                    used in bayesian mode, and ReML can't be used without a 
                    deterministic trend. (default = "ML").
             \param screeningSize (Int_t): number of parameters combinaisons 
                    to evaluate. An LHS sampling of this size is performed on 
                    the parameter space, and evaluated by the cost function. 
                    (default = 100).
             \param optimAlgo (TString): name of the optimisation algorithm. 
                    Can be: "NelderMead", "Subplexe", "Cobyla", "Bobyqa", "Praxis", 
                    "BFGS", "Newton", "MMA", "SLSQP", "VariableMetric". For details 
                    about these algorithms, cf. NLopt website: 
                    http://ab-initio.mit.edu/wiki/index.php/NLopt_Algorithms. 
                    (default = "BFGS").
             \param nbMaxOptimSteps (Int_t): maximum number of optimisation steps. 
                    This can be used when the cost function is long to evaluate
                    and the optimisation algorithm is slow to converge. (default = 1000).
             \param reset (Bool_t): if kTRUE, the parameters to optimise are reset 
                    to their default values. If kFALSE, the current state of the 
                    parameters (correlation lengths, alpha, GP variance, etc.) 
                    is used. (default = kTRUE).
             */
            void findOptimalParameters(TString criterion = "ML",
                                       Int_t screeningSize = 100,
                                       TString optimAlgo = "BFGS",
                                       Int_t nbMaxOptimSteps = 1000,
                                       Bool_t reset = kTRUE,
                                       Option_t * option = "");
        
            //! Build the gaussian process model.
            /*!
             This function computes the parameters of the gaussian process, 
             necessary to perform new predictions. It then creates a TKriging 
             model and return its pointer.
             \note Calls gpMod0 or gpModh function from gpL
             \warning The user must delete the TKriging object by himself.
             \param computeLooErrors (Bool_t): if kTRUE, the Leave One Out 
             errors of the kriging model are computed and stored in the model. 
             (default = kTRUE).
             */
            URANIE::Modeler::TKriging* buildGP(Bool_t computeLooErrors = kTRUE);

	    //! Clean list of attribute
	    void cleanAttributeList(TList *inputlist);
            //@}
        
            //---------------------------------------------
            /** @name Printing Log
             **/
            /** @name Printing Log
            **/
            //@{
                void setLog()
                {
                    _blog = kTRUE;
                }

                void unsetLog()
                {
                    _blog = kFALSE;
                }

                void changeLog()
                {
                    _blog = !_blog;
                }

                Bool_t getLog()
                {
                    return _blog;
                }

                virtual void printLog(Option_t *option = "");
            //@}
        
            ClassDef(URANIE::Modeler::TGPBuilder, ID_MODELER)
            //Classe de
        };    
    } // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
