//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPMLCostFunction.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface de la classe URANIE::Modeler::TGPMLCostFunction

 */
/**

 * @class URANIE::Modeler::TGPMLCostFunction
 *
 * @brief Description of the class TGPMLCostFunction.
 *
 * This class defines a Maximum Likelyhood cost function used to select the 
 * optimal parameters of the correlation function of a Gaussian Process. It is a wrapper 
 * to the gpML0 and gpMLh cost functions provided by the "gaussian process Library" (gpL), 
 * by Jean-Marc Martinez. 
 * NOTE: TGPMLCostFunction is unsuited to bayesian studies. Use TGPReMLCostFunction instead.
 *
 */
#ifndef TGPMLCOSTFUNCTION_H
#define TGPMLCOSTFUNCTION_H    

//// URANIE
#include "Modeler.h"
#include "TGPBuilder.h"
#include "TShareEval.h"
#include "TGPCostFunction.h"

//// Jean-Marc's Gaussian Processes Library
extern "C"
{
int gpML0(double *K, double *y, int n, double *ml, int cas);
int gpMLh(double *H, double *K, double *y, int n, int p, double *ml, int cas);
}
;

namespace URANIE
{
namespace Modeler
{
//class TGPBuilder;
class TGPMLCostFunction: public TGPCostFunction
{

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{            
    //! Standard constructor
    /*!
     A TGPMLCostFunction object need to be associated with a TGPBuilder 
     object which will provide the covariance matrix and the observations.
     \param gpb (TGPBuilder*): pointer to a Gaussian Process Builder object.
     \warning this class can't handle Bayesian studies. Use the TGPReMLCostFunction class instead.
     */
    TGPMLCostFunction(TGPBuilder* gpb);

    //! Default destructor
    virtual ~TGPMLCostFunction();
    //@}

    virtual URANIE::Relauncher::TEval *rebuild(int rank, Bool_t chef = kTRUE);
    virtual Bool_t unbuild();

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    //! Evaluate the cost function.
    /*!
     This function provides compatibility with the Relauncher module. It takes 
     two arrays of double and returns an integer: 1 if all went well, 0 otherwise.
     \param theta (Double_t*): a single set of parameters to optimise.
     \param res (Double_t*): output of the cost function.
     
     \warning If a measurement error is taken into account and its variance is 
     provided by the user, the last element of the array "theta" 
     is the GP variance. If the measurement error variance exists but is unknown, 
     the last element of "theta" is the value of the parameter alpha.
     */
    virtual Int_t eval(Double_t* theta, Double_t* res, int=0);
    virtual Int_t eval(vector<URANIE::DataServer::UEntry*> *, vector<URANIE::DataServer::UEntry*> *, int=0);

    //@}

    //---------------------------------------------

    ClassDef(URANIE::Modeler::TGPMLCostFunction, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
