//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TIsotropicGaussianCorrFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TIsotropicGaussianCorrFunction

 */
#include <iostream>

//// ROOT
#include "TMath.h"

//// URANIE
#include "TIsotropicGaussianCorrFunction.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TIsotropicGaussianCorrFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TIsotropicGaussianCorrFunction::TIsotropicGaussianCorrFunction(Int_t nbDim, Double_t corrLength)
{
    nbDimensions  = nbDim;
    nbParameters  = 1;
    nbCorrLengths = 1;
    _lCorrLengths = new Double_t[nbCorrLengths]; 
    parameters    = _lCorrLengths;
    
    _lCorrLengths[0] = corrLength;
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TIsotropicGaussianCorrFunction::~TIsotropicGaussianCorrFunction()
{
    delete[] _lCorrLengths;
    _lCorrLengths = NULL;
    parameters = NULL;
    if (_blog)
        cout << "detruit TIsotropicGaussianCorrFunction" << endl;
}

// -----------------------------------------------------------------------
//  Set methods
// -----------------------------------------------------------------------

void TIsotropicGaussianCorrFunction::setCorrLengths(Double_t *corrLength)
{
    _lCorrLengths[0] = corrLength[0];
}

void TIsotropicGaussianCorrFunction::setCorrelationLength(Double_t corrLength)
{
    _lCorrLengths[0] = corrLength;
}

// -----------------------------------------------------------------------
// Clone
// -----------------------------------------------------------------------

TCorrelationFunction* TIsotropicGaussianCorrFunction::clone()
{
    TIsotropicGaussianCorrFunction* cfClone = new TIsotropicGaussianCorrFunction(nbDimensions, _lCorrLengths[0]);
    return cfClone;
}

// -----------------------------------------------------------------------
// Evaluation
// -----------------------------------------------------------------------

void TIsotropicGaussianCorrFunction::eval(Int_t n1, Double_t* x1,
        Int_t n2, Double_t* x2, Double_t* C)
{
    Int_t errorCode = gpCorGaussIso(x1, x2, n1, n2, nbDimensions, parameters[0], C);

    try
    {
        if (errorCode == 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TIsotropicGaussianCorrFunction::eval: evaluation failed");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TIsotropicGaussianCorrFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TIsotropicGaussianCorrFunction::printLog[" << option << "]"
            << endl;
    cout << " ** Number of parameters = [" << nbParameters << "]" << endl;
    cout << " ** Correlation length   = [" << parameters[0] << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

