//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TIsotropicGaussianCorrFunction.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface de la classe URANIE::Modeler::TIsotropicGaussianCorrFunction

 */
/**

 * @class URANIE::Modeler::TIsotropicGaussianCorrFunction
 *
 * @brief Description of the class TIsotropicGaussianCorrFunction.
 *
 * This class defines the shape of the correlation between two sets of variables 
 * as an isotropic gaussian:
 *
 * \f[c(x1,x2) = exp\left(-\frac{||x1-x2||^2}{lc^2}\right)\f]
 *
 * where \f$lc\f$ is the correlation length.
 *
 */
#ifndef TISOTROPICGAUSSIANCORRFUNCTION_H
#define TISOTROPICGAUSSIANCORRFUNCTION_H

//// URANIE
#include "TCorrelationFunction.h"

extern "C"
{
int gpCorGaussIso(double *X1, double *X2, int n1, int n2, int d, double lc,
        double *C);
}
;

namespace URANIE
{
namespace Modeler
{
class TIsotropicGaussianCorrFunction: public TCorrelationFunction
{

    // Attributes
private:
public:

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Standard Constructor
    /*!
     \param nbDim (Int_t): the number of input variables
     \param corrLength (Double_t): the correlation length for the function (default = 0.1)
     */
    TIsotropicGaussianCorrFunction(Int_t nbDim, Double_t corrLength = 0.1);

    //! Default destructor
    virtual ~TIsotropicGaussianCorrFunction();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{  
    Double_t getCorrelationLength()
    {
        return _lCorrLengths[0];
    }

    //! Sets the parameters of the isotropic gaussian correlation function
    /*!
     \param newParams (Double_t*): array containing the parameters of the correlation function. 
     The array's size is 1 and contains the correlation length.
     */
    virtual void setParameters(Double_t* newParams)
    {
        setCorrelationLength(newParams[0]);
    }

    //! Sets the correlation lengths
    /*!
    This is a function inherited from TCorrelationFunction. You can use 
    setCorrelationFunction(Double_t) if you don't want to manipulate an 
    array containing a single value.
     \param corrLength (Double_t*):  an array containing the value of the correlation length.
     */
    void setCorrLengths(Double_t *corrLength);
    
    //! Sets the correlation length parameter
    /*!
     \param corrLength (Double_t):  the value of the correlation length.
     */
    void setCorrelationLength(Double_t corrLength);

    //@}

    //---------------------------------------------
    /** @name Tools
     **/
    //@{           
    //! Return a clone of the correlation function
    /*!
    The clone must be destroyed by the object (or user) calling the function.                
    */
    virtual TCorrelationFunction* clone();

    //! Compute the correlation between x1 and x2 
    /*!
     The order of the indices in the flat arrays are the following:
     x[i+j*d] = X[i,j]
     where X is a matrix of size (d x n), i is the line index, and j the 
     column index. The results are stored in the flat array C.  
     \warning Be aware that the matrix corresponding to flat arrays x1 and 
     x2 have input dimensions as LINE, not columns. This means 
     that for a set of point {(a1,b1),(a2,b2),(a3,b3)}, the 
     corresponding matrix X will be:
     \f$ 
     \left( 
     \begin{array}{ccc}
     a1 & a2 & a3 \\
                           b1 & b2 & b3 
     \end{array}
     \right)
     \f$                     
     and the corresponding flat array x = {a1,b1,a2,b2,a3,b3}.     
     \param n1 (Int_t): number of elements of x1
     \param x1 (Double_t*): data set x1 represented as a flat array (size: d * n1)
     \param n2 (Int_t): number of elements of x2
     \param x2 (Double_t*): data set x2 represented as a flat array (size: d * n2)
     \param C  (Double_t*): correlation matrix represented as a flat array (size: n1 * n2)
     
     \exception UErrorExceptions if the gpLib function returns 1 (evaluation failed)
     */
    virtual void eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2,
            Double_t* C);

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TIsotropicGaussianCorrFunction, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
