//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TKriging.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date

 \brief Implementation of the class  URANIE::Modeler::TKriging

 */

//// ROOT
#include "TLeaf.h"
#include "TDecompChol.h"
#include "TDecompSVD.h"
#ifdef ROOT5
#include "TTreeFormula.h"
#else
#include "TFormula.h"
#endif
#include "TObjString.h"

//// URANIE
#include "UExceptions.h"

#include "TLauncher2.h"
#include "TKriging.h"
#include "TGPCostFunction.h"

#include "TMaternICorrFunction.h"
#include "TMaternIICorrFunction.h"
#include "TMaternIIICorrFunction.h"
#include "TExponentialCorrFunction.h"
#include "TGaussianCorrFunction.h"
#include "TIsotropicGaussianCorrFunction.h"

#include "TDSNtupleD.h"

#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"

#ifdef WITH_NLOPT
#include "TNlopt.h"
#include "TNloptDirect.h"
#include "TSequentialRun.h"
#include "TGradientEstimationRun.h"
#endif

//   GPU
#if defined(WITH_GPUANN) && !defined(__CINT__)
#include "GP_gpu.h"
#endif

#ifdef WITH_LIBXML2
#include "PMMLlib.hxx"
#endif

#include <cmath>

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

ClassImp (URANIE::Modeler::TKriging)

namespace URANIE
{
namespace Modeler
{
  // -----------------------------------------------------------------------
  //  Constructors
  // -----------------------------------------------------------------------
  TKriging::TKriging(Int_t nX, Int_t nS, Double_t* xObs, Double_t* xNormParams, Double_t* yObs, Double_t* gamma, Double_t* iL,  
		     Double_t gpVar, Double_t vErrMes, URANIE::Modeler::TCorrelationFunction* corrFunc, TString varNames,
		     Int_t nP, Double_t* hbeta, Double_t* vbeta, Double_t* M, Double_t* iR, TString trend):
    TDoubleEval()
  {
    bool debug=false;
    // ---------------------

    _inputVarNames = "";
    _outputVarName = "";
    
    TString vNames(varNames);
    Int_t i = vNames.Length()-1;
    while( vNames[i] != ':') i--;
    
    _inputVarNames = vNames(0, i);
    _outputVarName = vNames(i+1, vNames.Length()-i);
    
    _inputNamesList = TString(_inputVarNames).Tokenize(":");
    
    _nX = nX;
    _nS = nS;
    _nYDeclared = -1; 
    _nP = nP;
    
    if(_nP > 0)
        bHas_trend = kTRUE;
    else
        bHas_trend = kFALSE;
    
    _sigma2 = gpVar;
    _vErrMes = vErrMes;
    _correlationFunction = corrFunc->clone();
    
    initMatrices();
    
    // The Gaussian Process
    for(i = 0; i < _nX; i++)
    {
      _xNormParams[2*i]     = xNormParams[2*i];
      _xNormParams[2*i + 1] = xNormParams[2*i + 1];
      
      for(Int_t j = 0; j < _nS; j++)
	_xObs[i*_nS+j] = xObs[i*_nS+j];
   }
    
    for(i = 0; i < _nS; i++)   
    {
        _yObs[i] = yObs[i];
        _gamma[i] = gamma[i];
        for(Int_t j = 0; j < _nS; j++)
            _iL[i*_nS+j] = iL[i*nS+j];
    }
    
    // The correlation lengths
    _normCorrLengths = _correlationFunction->getCorrLengths();
    for(i = 0; i < _nX; i++)
        _corrLengths[i] = _normCorrLengths[i] * (_xNormParams[2*i+1] - _xNormParams[2*i]);


    if (debug){
      for(i = 0; i < _nX; i++){
	std::ostringstream ss;
	ss.precision(16);
	
	ss << scientific << i << " Lambda: " << _corrLengths[i] << " lambdaN: " << _normCorrLengths[i];
	ss << "DX: " << (_xNormParams[2*i+1] - _xNormParams[2*i]);
	std::cout << ss.str() << std::endl;
      }
    }


    // The deterministic trend
    _trendString = trend;
    _trendCoefList = NULL;

    if(bHas_trend)
    {
        createTrend();
        
        for(i = 0; i < _nP; i++)
        {
            _hbeta[i] = hbeta[i];
            for(Int_t j = 0; j < _nP; j++)
            {
                _iR[i*_nP+j] = iR[i*_nP+j];
                _vbeta[i*_nP+j] = vbeta[i*_nP+j]; 
            }
        }
        for(i = 0; i < _nS*_nP; i++)
            _M[i] = M[i];        
    }    

    if (debug){
      StatVect("IL", _nS*_nS, _iL);
      StatVect("Gam", _nS, _gamma);
      
      if (_errorLoo != NULL) StatVect("errorLoo", _nS, _errorLoo);
      if (_varLoo != NULL) StatVect("varLoo", _nS, _varLoo);
      std::cout << "eqmLoo: " << _mseLoo << std::endl;
    }


  }

#ifdef WITH_LIBXML2
TKriging::TKriging(TString filename, TString modelName): TDoubleEval()
{

    // -- Declarations
    PMMLlib::PMMLlib * pmml;
    PMMLlib::PMMLType modelType;
    std::string fname;
    // --
    int nbFields;
    std::vector<std::string> fieldNames;
    std::vector<bool> intervalstatus;
    std::vector<double> varMin, varMax;
    // --
    std::set<std::string> inputNames, outputNames;
    PMMLlib::PMMLKernelType kernel;
    double noiseVariance;
    int ndimE;
    std::vector<double> lambda;
    double degree;
    std::map<std::string, std::vector<double> > table;
    std::set<std::string> InputNames, OutputNames;
    int recordCount;
    bool computeLooErrors;
    bool bVariance_in_factor;
    bool bHas_measurement_error;
    double alpha;
    double reg;
    bool debug=false;
    // ------------------------------

    // -- Parametrage implicite
    computeLooErrors=true;

    ///< if kTRUE, GP variance is calculated analytically. Otherwise, it is a parameter of the cost function.
    bVariance_in_factor = true;

    ///< if kTRUE, the construction of the GP will take into account the existence of a measurement error variance. 
    bHas_measurement_error=false;
    alpha=0.;
    reg=0.;
    // ------------------------------
    fname = std::string(filename);

    pmml = new PMMLlib::PMMLlib(fname);
    std::cout << "-->" << std::endl;
    std::cout << pmml->makeLog() << std::endl;
    pmml->importGaussianProcess(modelName.Data());

    pmml->GetDataDictionary(nbFields,fieldNames,intervalstatus, varMin, varMax);
    pmml->GetGaussianProcess(fieldNames,
			     InputNames, OutputNames,
			     kernel, ndimE, noiseVariance, lambda,
			     recordCount, table);
			     
    _inputVarNames = "";
    _outputVarName = "";
    
    std::set<std::string>::iterator ensIt;
    for(ensIt=InputNames.begin(); ensIt!= InputNames.end(); ensIt++){
        _inputVarNames =  _inputVarNames + ":" + *ensIt;
    }

    ensIt = OutputNames.begin();
    _outputVarName = *ensIt;

    _inputNamesList = TString(_inputVarNames).Tokenize(":");
    
    _nX = ndimE;
    _nS = recordCount;
    _nYDeclared = -1;
    _nXDeclared = -1; 
    _nP = 0; // No trend stored in pmml
    
    if(_nP > 0)
        bHas_trend = kTRUE;
    else
        bHas_trend = kFALSE;
    
    _sigma2 = noiseVariance;

    _vErrMes = -1.0;
    _correlationFunction = new TGaussianCorrFunction(_nX);
    
    initMatrices();
    // -------------------

    int indIn=-1;
    int indOut=-1;

    for(size_t k=0; k < fieldNames.size(); k++){
        std::string varname = fieldNames[k];

        std::set<std::string>::iterator itIn=InputNames.find(varname);

        if (itIn != InputNames.end()){
            indIn++;
            //_xNormParams[2*indIn]     = varMin[k];
            //_xNormParams[2*indIn + 1] = varMax[k];

            std::vector<double>::iterator vecItmin, vecItmax;
            vecItmin = std::min_element(table[varname].begin(),table[varname].end());
            vecItmax = std::max_element(table[varname].begin(),table[varname].end());
            _xNormParams[2*indIn]     = *vecItmin;
            _xNormParams[2*indIn + 1] = *vecItmax;
	
            for(Int_t j = 0; j < _nS; j++){
                _xObs[j*_nX + indIn] = table[varname][j];
            }
        }
        else{
            std::set<std::string>::iterator itOut=OutputNames.find(varname);
            if (itOut != OutputNames.end()){
                indOut++;
                if (indOut>0){
                    std::ostringstream ss;
                    ss << "Only one out variable is allowed." << OutputNames.size() << " were found" << std::endl;
                    throw(ss.str());
                }
                for(Int_t j = 0; j < _nS; j++){
                    _yObs[indOut*_nS+j] = table[varname][j];
                }
            }
            else{
                std::cout << "Variable not Found: " << varname << std::endl;
            }
        }

    }
    //std::cout << "@_normCorrLengths: " << _normCorrLengths << std::endl;

    _normCorrLengths = _correlationFunction->getCorrLengths();
    for(Int_t i = 0; i < _nX; i++){
        double vmin,vmax, dv;
        vmin = _xNormParams[2*i];
        vmax = _xNormParams[2*i + 1];
        dv   = (vmax - vmin);

        _corrLengths[i] = _normCorrLengths[i] *dv;

        for(Int_t j = 0; j < _nS; j++){
            double vo, vn;

            vo = _xObs[j*_nX+i];
            vn = (vo - vmin)/dv; // On normalise
            _xObs[j*_nX+i] = vn;
	}
    }

    for(int i = 0; i < _nX; i++){
        _corrLengths[i] = lambda[i];
        _normCorrLengths[i] = lambda[i]/(_xNormParams[2*i+1] - _xNormParams[2*i]);

    }
    Double_t *xC, *xK, *Vmes;
    Double_t *errorLoo, *varLoo;
    Double_t eqmLoo[1];

    xC        = new Double_t[_nS * _nS];
    xK        = new Double_t[_nS * _nS];
    Vmes      = new Double_t[_nS * _nS];
    //
    errorLoo = new Double_t[_nS];
    varLoo   = new Double_t[_nS];



    for (Int_t i = 0; i < _nS * _nS; i++)
    {
        xC[i] = 0.0;
        Vmes[i] = 0.0;
    }
    //StatMat("X", _nX, _nS, _xObs);
    //StatVect(std::string("Y"), _nS, _yObs);

    computeCovarianceMatrix(_nS, _xObs,
			    bVariance_in_factor, 
			    bHas_measurement_error, alpha, reg,
			    _sigma2, Vmes,
			    xC, xK);
    //StatVect(std::string("C"), _nS*_nS, xC);
    //StatVect(std::string("K0"), _nS*_nS, xK);

    if (bVariance_in_factor)
    {
        _vErrMes = alpha * _sigma2;
        for(Int_t i = 0; i < _nS*_nS; i++)
            xK[i] *= _sigma2;
    }
    else
    {
        // The variance of the measurement error was set by the user, so we 
        // suppose it is known.
        _vErrMes = -1.0;
    }
    //StatVect(std::string("K1"), _nS*_nS, xK);

    int returnCode;

    returnCode = gpMod0(xK, _yObs, _nS, _gamma, _iL);
    if(computeLooErrors){
        returnCode += gpLoo0Error(xK, _yObs, _nS, errorLoo, varLoo, eqmLoo);
        setLooErrors(errorLoo, varLoo, eqmLoo[0]);
    }
    if (debug){
        StatVect("IL", _nS*_nS, _iL);
        StatVect("Gam", _nS, _gamma);
        StatVect("errorLoo", _nS, errorLoo);
        StatVect("varLoo", _nS, varLoo);
        std::cout << "eqmLoo: " << eqmLoo[0] << std::endl;
    }

    delete[] xC; xC=NULL;
    delete[] xK; xK=NULL;
    delete[] Vmes; Vmes=NULL;
    delete[] errorLoo; errorLoo=NULL;
    delete[] varLoo; varLoo=NULL;

    // The deterministic trend
    _trendString = "";
    _trendCoefList = NULL;

    delete pmml;

}
#endif

void TKriging::StatMat(std::string title, int n1, int n2, Double_t *v){
    std::vector<double>  tvmin;
    std::vector<double>  tvmax;
    std::vector<double>  tmean;
    std::vector<double>  tvar;
    std::vector<double>  tnorm;
    // --
    StatMat(title, n1, n2, v, tvmin, tvmax, tmean, tvar, tnorm);
  }
  void TKriging::StatMat(std::string title, int n1, int n2, Double_t *v,
			 std::vector<double> &tvmin, std::vector<double>  &tvmax,
			 std::vector<double> &tmean, std::vector<double>  &tvar,
			 std::vector<double> &tnorm){
    // ---------------------
    double vmin, vmax, mean, var, norm;
    Double_t *tab;
    // ---------------------

    tab = new Double_t[n2];
    for(int k=0;k < n1; k++){
      for(int i=0; i<n2;i++){
	tab[i] = v[i*n1 + k];
      }
      std::ostringstream ss;
      ss << title << "[" << k << ":]";
      StatVect(ss.str(), n2, tab, vmin, vmax, mean, var, norm);
      tvmin.push_back(vmin);
      tvmax.push_back(vmax);
      tmean.push_back(mean);
      tvar.push_back(var);
      tnorm.push_back(norm);
    }
    delete[] tab; tab=NULL;
  }

  void TKriging::StatVect(std::string title, int n, Double_t *v){
    double vmin;
    double vmax;
    double mean;
    double var;
    double norm;
    // -- 
    
    StatVect(title, n, v, vmin, vmax, mean, var, norm);
  }
  void TKriging::StatVect(std::string title, int n, Double_t *v, double &vmin, double &vmax, double &mean, double &var, double &norm){

    // --
    double sum, sum2;
    int id1,id2;
    bool debug=false;
    // --
    vmin=v[0];
    vmax=v[0];
    var = 0.;
    mean = 0.;
    norm = 0.;
    sum=0.0;
    sum2=0.0;

    std::ostringstream msg;
    msg.precision(16);
    msg << scientific;
    id1 = std::min(10, n);
    id2 = (id1-1)/2;

    for(int i=0; i < n; i++){
      if (i < id1) msg << v[i] << " ";
      if ((i == id2) && (id1 > 5)) msg << std::endl;

      if (v[i]> vmax) vmax = v[i];
      if (v[i]< vmin) vmin = v[i];
      sum += v[i];
      sum2 += std::pow(v[i],2);
    }
    mean = sum/double(n);
    var = (sum2/double(n)) - std::pow(mean,2);
    norm = (sum2/double(n));

    if (debug){
      std::string sep(" ");
      std::ostringstream ss;
      ss.precision(16);
      ss << scientific;
      ss << "Min:" << sep << vmin << sep << "Max:" << sep << vmax << std::endl;
      ss << "Moy:" << sep << mean << sep << "Var:" << sep << var << sep << "Norm:" << norm;
      std::cout << title << std::endl;
      std::cout << msg.str() << std::endl;
      std::cout << ss.str() << std::endl;
    }
    
  }
  void TKriging::computeCovarianceMatrix(int ns, double *xObs,
					 bool bVariance_in_factor, 
					 bool bHas_measurement_error, double alpha, double reg,
					 double sigma2, double *Vmes,
					 double *C, double *K)
  {

    _correlationFunction->eval(ns, xObs, ns, xObs, C);

    if (bVariance_in_factor)
    {
        for (Int_t i = 0; i < ns * ns; i++)
            K[i] = C[i];
        if (bHas_measurement_error)
        {
            for (Int_t i = 0; i < ns; i++)
                K[i*(ns+1)] += alpha;
        }
        else
        {
            for (Int_t i = 0; i < ns; i++)
                K[i*(ns+1)] += reg;
        }
    }
    else
    {
        for (Int_t i = 0; i < ns * ns; i++)
            K[i] = sigma2 * C[i] + Vmes[i];
    }            
}

void TKriging::createTrend()
{
    // Decide if the trend is user defined or constructed automatically
    if(_trendString.CompareTo("const", TString::kIgnoreCase) == 0)
    {
        TString localTrendString = "1";
        _trendCoefList = localTrendString.Tokenize(":");
        bUse_normalisation = kTRUE;
    }
    else if(_trendString.CompareTo("linear", TString::kIgnoreCase) == 0)
    {
        TString localTrendString = Form("1:%s",_inputVarNames.Data());
        _trendCoefList = localTrendString.Tokenize(":");
        bUse_normalisation = kTRUE;
    }
    else
    {
        _trendCoefList = _trendString.Tokenize(":");
        bUse_normalisation = kFALSE;
    }
    
        
}

void TKriging::initMatrices()
{
    // Build the matrices
    _xObs        = new Double_t[_nX * _nS];
    _xNormParams = new Double_t[_nX * 2];
    _yObs        = new Double_t[_nS];
    _iL          = new Double_t[_nS * _nS];
    _gamma       = new Double_t[_nS];
    
    _normCorrLengths = NULL;
    _corrLengths     = new Double_t[_nX];
         
    _errorLoo    = NULL;
    _varLoo      = NULL;
    _mseLoo      = -1;

    if(bHas_trend)
    {
        _hbeta     = new Double_t[_nP];
        _vbeta     = new Double_t[_nP * _nP];
        _iR        = new Double_t[_nP * _nP];
        _M         = new Double_t[_nS * _nP];
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TKriging::~TKriging()
{
    delete[] _xObs;
    _xObs = NULL;
    delete[] _xNormParams;
    _xNormParams = NULL;
    delete[] _yObs;
    _yObs = NULL;
    delete[] _iL;
    _iL = NULL;
    delete[] _gamma;
    _gamma = NULL;
    delete[] _corrLengths;
    _corrLengths = NULL;
    if (bHas_trend)
    {
        delete[] _hbeta;
        _hbeta = NULL;
        delete[] _vbeta;
        _vbeta = NULL;
        delete[] _iR;
        _iR = NULL;
        delete[] _M;
        _M = NULL;
        delete _trendCoefList;
        _trendCoefList = NULL;
    }
    
    if (_errorLoo)
    {
        delete [] _errorLoo;
        _errorLoo = NULL;
        delete [] _varLoo;
        _varLoo = NULL;
    }

    delete _inputNamesList;
    
    delete _correlationFunction;
}

// -----------------------------------------------------------------------
//  Get and Set method
// -----------------------------------------------------------------------

TDataServer* TKriging::getLooData()
{
    TDataServer *tdsLoo = new TDataServer("tdsLoo","Leave One Out data");   
    for(Int_t i = 0; i < _inputNamesList->GetEntries(); i++)
    {
        TAttribute *attIn = new TAttribute(_inputNamesList->At(i)->GetName());
        tdsLoo->addAttribute(attIn);
	attIn->delShare();
    }
    TAttribute *attOut = new TAttribute(_outputVarName.Data());
    attOut->delShare(); tdsLoo->addAttribute(attOut);    
    TAttribute *attErrLoo = new TAttribute("errorLoo"); 
    attErrLoo->delShare(); tdsLoo->addAttribute(attErrLoo);
    TAttribute *attVarLoo = new TAttribute("varLoo"); 
    attVarLoo->delShare(); tdsLoo->addAttribute(attVarLoo);
    TAttribute *attStdErrLoo = new TAttribute("stdErrLoo"); 
    attStdErrLoo->delShare(); tdsLoo->addAttribute(attStdErrLoo);
    
    // Fill the data server
    tdsLoo->createTuple();
    TDSNtupleD *t = tdsLoo->getTuple();
    Int_t nbBr = t->GetNbranches();

    Double_t *line = new Double_t[nbBr];

    for(Int_t i = 0; i < _nS; i++)
    {
        line[0] = Double_t(i+1);
        for(Int_t j = 0; j < _nX; j++)
            line[j+1] = _xObs[i*_nX+j] * (_xNormParams[2*j+1] - _xNormParams[2*j]) + _xNormParams[2*j];
        line[_nX+1] = _yObs[i];
        line[_nX+2] = _errorLoo[i];
        line[_nX+3] = _varLoo[i];
        line[_nX+4] = _errorLoo[i]/sqrt(_varLoo[i]);
        t->Fill(line);
    }

    tdsLoo->saveTuple("UPDATE");
	
	delete [] line;
    return tdsLoo;
}


void TKriging::setLooErrors(Double_t *err, Double_t *var, Double_t mse)
{
    _errorLoo = new Double_t[_nS];
    _varLoo   = new Double_t[_nS];
    
    for(Int_t i = 0; i < _nS; i++)
    {
        _errorLoo[i] = err[i];
        _varLoo[i]   = var[i];
    }
    _mseLoo = mse;
}

// -----------------------------------------------------------------------
//  Evaluation
// -----------------------------------------------------------------------
void TKriging::estimateWithCov(TDataServer *tdsQuery, TString listOfInputs,
                               TString listOfOutputs, TString realValue, Option_t * option)
{

    try
    {
        if ( !tdsQuery )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TKriging::estimateWithCov method\n The TDataServer is NULL");
    }
     catch (URANIE::Exceptions::UErrorExceptions& ue)
     {
         ue.printMessage();
         throw ue;
     }
     
        if( _blog ) {
        cout << endl;
        cout << "**************************************" << endl;
        cout << " ** TKriging::estimateWithCov" << endl;
        cout << " **    with TDS (" << tdsQuery->GetName () << ")" << endl;
        cout << "**************************************" << endl;
    }

    TString soption=TString(option);
    soption.ToLower();

    TString listOfOutputTypes="D:D:V";
    // Name of the estimation
    TString  EstimName="", VarName="";

    // Default Values
    if ( listOfInputs.Length () == 0 )
        listOfInputs = _inputVarNames;
    cout << " *** With listOfInputs (" << listOfInputs << ")" << endl;

    if ( listOfOutputs.Length () == 0 )
    {
        listOfOutputs = Form("%sEstim:%sVarEstim", _outputVarName.Data (),  _outputVarName.Data ());
        if ( realValue != "")
        {
            EstimName=Form("%sEstim", _outputVarName.Data ());
            VarName=Form("%sVarEstim", _outputVarName.Data ());
        }
    } 
    cout << " *** With listOfOutputs (" << listOfOutputs << ")" << endl;

    // Validate the arguments
    TList * listOfInputsAttributes=NULL;
    try
    {
        if ( listOfInputs.CountChar (':')+1 != _nX )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TKriging::estimateWithCov method\n The number of inputs [%s] is not the good number (%d) as builded [%s]",
                                                            listOfInputs.Data (), _nX, _inputVarNames.Data()));
        if ( listOfOutputs.CountChar (':') != 1 ) // Only 2 attributes : y_Estim and y_VarianceEstim
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TKriging::estimateWithCov method\n The number of ouputs [%s] must be equal to 2",
                                                            listOfOutputs.Data ()));
        if ( tdsQuery ) {
            listOfInputsAttributes = tdsQuery->getListOfAttributes(listOfInputs);
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete listOfInputsAttributes;  listOfInputsAttributes=NULL;
        throw ue;
    }
    delete listOfInputsAttributes;  listOfInputsAttributes=NULL;

    int nSQuery = tdsQuery->getNPatterns();    

    Double_t *yp       = new Double_t[nSQuery];
    Double_t *covmat   = new Double_t[nSQuery*nSQuery];
    Double_t *xNewNorm = new Double_t[_nX*nSQuery];
    Double_t *xNew     = new Double_t[_nX*nSQuery];
    Double_t *x0       = new Double_t[_nX];
    // Double_t *y0       = new Double_t[2];
    Double_t *c0       = new Double_t[nSQuery*_nS];
    Double_t *k0       = new Double_t[nSQuery*_nS];
    Double_t *CQuery   = new Double_t[nSQuery*nSQuery];    

    vector<URANIE::DataServer::UEntry*> y0;
    vector<TLeaf*> varLeaf;
    TObjArray *varName=listOfInputs.Tokenize(":");
    TIter nextin(varName);
    TObject *var;
    while ((var = nextin() )) {
        varLeaf.push_back(tdsQuery->getTuple()->GetLeaf(var->GetName()));
    }
    delete varName; varName=NULL;

    // Filling the database xNew and xNewNorm
    for(int is=0; is<nSQuery; is++)
    {
        tdsQuery->getTuple()->GetEntry(is);        
        // normalisation of x0
        for(Int_t i = 0; i < _nX; i++)
        {
            x0[i] = varLeaf[i]->GetValue();
            xNew[is*_nX+i] = x0[i];
            xNewNorm[is*_nX+i] = ( x0[i] - _xNormParams[i*2] ) / ( _xNormParams[i*2+1] - _xNormParams[i*2] );
        }
    }

    // Evaluate the covariances within the testing basis
    _correlationFunction->eval(nSQuery, xNewNorm, nSQuery, xNewNorm, CQuery);
    
    // Evaluate the covariances with the kriging basis   
    _correlationFunction->eval(nSQuery, xNewNorm, _nS, _xObs, c0);

    // Correct the correlation matrix by the variance of the process
    for(int is=0; is<nSQuery; is++)
    {
        for (Int_t i = 0; i < _nS; i++)
            k0[is*_nS+i] = _sigma2 * c0[is*_nS+i];

        for (Int_t i = 0; i < nSQuery; i++)
            CQuery[is*nSQuery+i] = _sigma2 * CQuery[is*nSQuery+i];
    }

    // Prediction
    if (!bHas_trend)
        gpPred0Cov(k0, CQuery, _sigma2, _gamma, _iL, nSQuery, _nS, yp, covmat);
    else
    {
        Double_t *h0 = new Double_t[_nP*nSQuery];

        // Fill the deterministic trend with the new point
        
        for (Int_t i = 0; i < _nP; i++)
            for (Int_t is = 0; is < nSQuery; is++)
            {
                // retrieve the formula for trend coefficient i
                TObjString* tmpString = (TObjString*) _trendCoefList->At(i);            
                TString coefFormula = tmpString->GetString();
                for (Int_t attIndex = 0; attIndex < _nX; attIndex++)
                {
                    // update the current formula by replacing attribute
                    // attName by its corresponding value.
                    TString attName = _inputNamesList->At(attIndex)->GetName();
                    if(bUse_normalisation)
                        coefFormula = coefFormula.ReplaceAll(attName, Form("%.14e", xNewNorm[is*_nX+attIndex]));
                    else
                        coefFormula = coefFormula.ReplaceAll(attName, Form("%.14e", xNew[is*_nX+attIndex]));

                }
                // The resulting formula contains no more variables. We
                // create a TFormula out of it, and evaluate it (the
                // result is a constant so the value of the parameter
                // of TFormula::Eval() is meaningless.
                TFormula tmpF("tmpF", coefFormula.Data());
                h0[is+i*nSQuery] = tmpF.Eval(0);
            }

        gpPredhCov(h0, k0, CQuery, _sigma2, _hbeta, _gamma, _iL, _iR, _M, nSQuery, _nS, _nP, yp, covmat);
        delete [] h0;
    }

    varName=listOfOutputs.Tokenize(":");
    int iout=0;

    TIter nextout(varName);
    try{
        while ((var = nextout() )) {
            tdsQuery->addAttributeByNameTitle(var->GetName(),var->GetName());
            if(iout==0 && realValue != "") EstimName=var->GetName();
            if(iout==1 && realValue != "") VarName=var->GetName();
            y0.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(var->GetName())));
            iout++;
        }
    }catch (URANIE::Exceptions::UErrorExceptions& ue){
        ue.printMessage();

        //cleaning
        delete [] xNewNorm; delete [] xNew; delete [] x0; delete [] yp;
        delete [] covmat; delete [] c0; delete [] k0; delete [] CQuery;

        for(vector<URANIE::DataServer::UEntry*>::iterator it= y0.begin(); it!=y0.end(); it++)
            delete (*it);

        delete varName;
        throw ue;
    }

    // try to store the covariance matrix
    listOfOutputs+=":_CovarianceMatrix_";
    y0.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, string("_CovarianceMatrix_")));
    tdsQuery->addAttributeByNameTitle("_CovarianceMatrix_","_CovarianceMatrix_");

    TMatrixD *Res=NULL;
    TMatrixD *UncorRes=NULL;
    TMatrixD *W=NULL;
    TMatrixD *iSigma=NULL;
    bool hasPassed=false;
    if( realValue != "" )        
    {
        try
        {
            if ( realValue.CountChar (':') != 0 ) 
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TKriging::estimateWithCov method\n The string provided for realValue implies several attributes [%s]. Only one should be provded.",
                                                                realValue.Data ()));

            if ( !tdsQuery->isAttribute(realValue) ) 
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TKriging::estimateWithCov method\n The string provided for realValue [%s] is not an attribute of the dataserver[%s].",
                                                                realValue.Data (), tdsQuery->GetName()));
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            //cleaning
            delete [] xNewNorm; delete [] xNew; delete [] x0; delete [] yp;
            delete [] covmat; delete [] c0; delete [] k0; delete [] CQuery;

            for(vector<URANIE::DataServer::UEntry*>::iterator it= y0.begin(); it!=y0.end(); it++)
                delete (*it);

            delete varName;
            throw ue;
        }

        bool withChol=true;
        if( soption.Contains("svd") )
            withChol=false;

        tdsQuery->getTuple()->Draw(realValue.Data(),"","nogoff");

        // Fill Matrix with realvalue
        TMatrixD YR(nSQuery,1); YR.Use(0,nSQuery-1,0,0,tdsQuery->getTuple()->GetV1());
        // Fill Matrix with estimated value
        TMatrixD YE(nSQuery,1); YE.Use(0,nSQuery-1,0,0,yp);

        // Fill covariance matrix
        TMatrixD Cov(nSQuery,nSQuery);  TMatrixD Diag(nSQuery,nSQuery);
        Cov.Use(0,nSQuery-1,0,nSQuery-1,covmat);
        for(int is=0; is<nSQuery; is++) Diag(is,is)= ((sqrt(Cov(is,is))!=0) ? 1./sqrt(Cov(is,is)) : 0);

        if( withChol == true)
        {
            // Cholesky decomposition
            TDecompChol toto(Cov);
            bool oui=false;
            oui=toto.Decompose();

            if(!oui)
            {
                try
                {
                    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                 "TKriging::estimateWithCov method\n The Cholesky decomposition of the covariance matrix has failed. No uncorrelated residuals provided.\n Will try the svd instead");
                }catch (URANIE::Exceptions::UWarningExceptions& ue)
                {
                    ue.printMessage();
                }
                withChol = false;
            }
            else
            {
                // Create the whitening matrix from the original Upper matrix of the CHolesky
                TMatrixD U = toto.GetU();
                W=new TMatrixD(TMatrixD::kInverted, U);
                W->T(); // Done !
            }
        }

        if(withChol==false)
        {
            // Do SVD decomposition
            TDecompSVD toto2(Cov);
            TVectorD Sprime=toto2.GetSig();
            
            // Fill a diag matrix with 1/sqrt(singvalue)
            TMatrixD irS(nSQuery,nSQuery);
            for(int ii=0; ii<nSQuery; ii++) irS(ii,ii) = ((Sprime(ii)!=0) ? 1./sqrt(Sprime(ii)) : 0);

            // Get U matrix
            TMatrixD U = toto2.GetU();

            // Compute whitening from assumption that Cov is symmetric definite positive, so U=V;
            W=new TMatrixD(irS, TMatrixD::kMultTranspose, U);
        }

        // Get inverse of Cov
        iSigma=new TMatrixD(*W, TMatrixD::kTransposeMult, *W);
        TString NewVar[2] = {"_Residuals_","_uncorrResiduals_"};
        for(unsigned int ivar=0; ivar<2; ivar++)
        {            
            // Add a branch to the output tree
            listOfOutputs+= TString(":")+NewVar[ivar];
            listOfOutputTypes+=":D";
            y0.push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string(NewVar[ivar])));
            tdsQuery->addAttributeByNameTitle(NewVar[ivar].Data(),"("+realValue+"-"+EstimName+")");
        }

        // Compute the delta and put the results in the matrix pointer
        YR=YR-YE;
        hasPassed=true;
        Res=new TMatrixD(Diag, TMatrixD::kMult, YR);
        UncorRes=new TMatrixD(*W, TMatrixD::kMult, YR);
        // TMatrixD MDComp1(YR, TMatrixD::kTransposeMult, *iSigma);
        // TMatrixD MD(MDComp1, TMatrixD::kMult, YR);
        // MD.Print();
    }

    TDSNtupleD *localRes= new TDSNtupleD("__localTree__","__localTree__Pouet__",listOfOutputs.Data(), listOfOutputTypes.Data());

    for(int is=0; is<nSQuery; is++)
    {
	//value
	y0[0]->clearEntry(); y0[0]->pushEntry(yp[is]);
        y0[1]->clearEntry(); y0[1]->pushEntry(covmat[is*nSQuery+is]);
	y0[2]->clearEntry(); 
	for(int is2=0; is2<nSQuery; is2++)
	    y0[2]->pushEntry(covmat[is*nSQuery+is2]);

        // XYZ<<"val are "<< y0[0] << "  " << y0[1] << "  ("<<is*nSQuery+is <<")" << endl;
        if(hasPassed)
        {
            y0[3]->clearEntry(); y0[3]->pushEntry( (*Res)(is,0) );
	    y0[4]->clearEntry(); y0[4]->pushEntry( (*UncorRes)(is,0) );
        }
        localRes->Fill(&y0);
    }

    if(soption.Contains("dumpcov") )
    {
        std::cout.precision(2);
        cout<<" Covariance matrix" << endl;
        for(int is=0; is<nSQuery; is++)
        {
            for(int js=0; js<nSQuery; js++)
                cout<< std::setw(6) << std::scientific << covmat[is*nSQuery+js] << "  " ;
            cout<< endl;
        }
    }


    tdsQuery->getTuple()->Merger(localRes, string(listOfOutputs), string(listOfOutputs));

    //complete cleaning    
    delete varName;
    if( Res != NULL ) delete Res;
    if( UncorRes != NULL ) delete UncorRes;
    if( W != NULL ) delete W;
    if( iSigma != NULL ) delete iSigma;
     
    delete [] xNewNorm; delete [] xNew; delete [] x0; delete [] yp;
    delete [] covmat; delete [] c0; delete [] k0; delete [] CQuery;

    for(vector<URANIE::DataServer::UEntry*>::iterator it= y0.begin(); it!=y0.end(); it++)
	delete (*it);
    
    delete localRes;
    
}

// -----------------------------------------------------------------------
//  Evaluation
// -----------------------------------------------------------------------
void TKriging::estimate(TDataServer *tdsQuery, TString listOfInputs,
                        TString listOfOutputs, Option_t * option, Bool_t useGPU)
{
    if( _blog ) {
        cout << endl;
        cout << "**************************************" << endl;
        cout << " ** TKriging::estimate" << endl;
        cout << " **    with TDS (" << tdsQuery->GetName () << ")" << endl;
        cout << "**************************************" << endl;
    }

    // Default Values
    if ( listOfInputs.Length () == 0 )
        listOfInputs = _inputVarNames;
    cout << " *** With listOfInputs (" << listOfInputs << ")" << endl;
    if ( listOfOutputs.Length () == 0 )
        listOfOutputs = Form("%sEstim:%sVarEstim", _outputVarName.Data (),  _outputVarName.Data ());
    cout << " *** With listOfOutputs (" << listOfOutputs << ")" << endl;

    // Validate the arguments
    TList * listOfInputsAttributes=NULL;
    try
    {
        if ( !tdsQuery )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TKriging::estimate method\n The TDataServer is NULL");
        if ( listOfInputs.CountChar (':')+1 != _nX )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TKriging::estimate method\n The number of inputs [%s] is not the good number (%d) as builded [%s]",
                                                            listOfInputs.Data (), _nX, _inputVarNames.Data()));
        if ( listOfOutputs.CountChar (':') != 1 ) // Only 2 attributes : y_Estim and y_VarianceEstim
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TKriging::estimate method\n The number of ouputs [%s] must be equal to 2",
                                                            listOfOutputs.Data ()));
        if ( tdsQuery ) {
            listOfInputsAttributes = tdsQuery->getListOfAttributes(listOfInputs);
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete listOfInputsAttributes;  listOfInputsAttributes=NULL;
        throw ue;
    }
    delete listOfInputsAttributes;  listOfInputsAttributes=NULL;


    if (useGPU)
        cout << " ** GPU computation has been requested." << endl;
    try
    {
        cout << " ** test if GPU is OK ..." << endl;
#ifndef WITH_GPUANN
        if (useGPU)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TKriging::estimate method\n GPU computation has been requested, but URANIE was not compiled with GPU support!"));
#endif
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (useGPU) {
#ifdef WITH_GPUANN
        cout << " **  use the GPU to estimate ..." << endl;
        estimate_GPU(tdsQuery, listOfInputs, listOfOutputs, option);
#endif
    } else {
        estimate_CPU(tdsQuery, listOfInputs, listOfOutputs, option);
    }
}

//  CPU method
void TKriging::estimate_CPU(TDataServer *tdsQuery, TString listOfInputs,
                            TString listOfOutputs, Option_t * option)
{
    if( _blog ) {
        cout << endl;
        cout << "**************************************" << endl;
        cout << " ** TKriging::estimate on CPU" << endl;
        cout << " **   option[" << option << "]" << endl;
        cout << "**************************************" << endl;
    }

    URANIE::Relauncher::TLauncher2 *lanceur =
            new URANIE::Relauncher::TLauncher2( tdsQuery,        // data to estimate
                         this,              // model used for the estimation
                         listOfInputs,     // list of the input variables
                         listOfOutputs);  // name given to the model's outputs
    // Launch the estimations
    lanceur->solverLoop();
    // Clean
    delete lanceur;
    this->clean();
}


#if defined(WITH_GPUANN)
//  GPU method
void TKriging::estimate_GPU(TDataServer *tdsQuery, TString listOfInputs,
                            TString listOfOutputs, Option_t * option)
{
    if( _blog ) {
        cout << endl;
        cout << "**************************************" << endl;
        cout << " ** TKriging::estimate on GPU" << endl;
        cout << " **   option[" << option << "]" << endl;
        cout << "**************************************" << endl;
    }

    URANIE::GPU::GP gp_gpu(_nX, (Float_t)(_sigma2), URANIE::GPU::kMatern32);

    // Allocate the vecteur of distance (size of nS of reference TDS)
    Int_t nsq = _nS;
    if ( tdsQuery )
        nsq = tdsQuery->getNPatterns();

    if( _blog )
      cout << " ** _nS[" << _nS << "] nsq[" << nsq << "] nX[" << _nX << "]" << endl;

    // Pass the References data and the inverse Matrix
    Float_t *vecReferenceInputs = new Float_t[_nS*_nX];
    Float_t *vecReferenceOutput = new Float_t[_nS];
    Float_t *matInverse = new Float_t[_nS*_nS];

    for(Int_t ir=0; ir<_nS; ir++) {
        // The inputs
        for(Int_t ii=0; ii<_nX; ii++) {
            vecReferenceInputs[ir*_nX+ii] = (Float_t)(_xObs[ir*_nX+ii]);
        }
        // The output
        vecReferenceOutput[ir] = (Float_t) _yObs[ir];
        // The inverse Matix
        for(Int_t ii=0; ii<_nS; ii++) {
//            if ( ii <= ir )
//                matInverse[ir*_nS+ii] = (Float_t)(2.0);
//            else
//                matInverse[ir*_nS+ii] = (Float_t)(0.0);
            matInverse[ir*_nS+ii] = (Float_t)(_iL[ir*_nS+ii]);
        }
    }
    gp_gpu.setReferencePoints(_nS, vecReferenceInputs, vecReferenceOutput, matInverse);


    // Pass the The correlation lengths in the normalised input space.
    Float_t *vecCorrelationLengths = new Float_t[_nX];
    Float_t *vecMinValues = new Float_t[_nX];
    Float_t *vecMaxValues = new Float_t[_nX];
    for(Int_t ii=0; ii<_nX; ii++) {
        vecCorrelationLengths[ii] = (Float_t)(_normCorrLengths[ii]);
        vecMinValues[ii] = (Float_t)(_xNormParams[2*ii]);
        vecMaxValues[ii] = (Float_t)(_xNormParams[2*ii+1]);
    }
    gp_gpu.setCorrelationLengths(_nX, vecCorrelationLengths);
    gp_gpu.setNormalisationValues(_nX, vecMinValues, vecMaxValues);

    // Pass the parameters of the gaussian process (size: _nS)
    Float_t *vecCoefficients = new Float_t[_nS];
    for(Int_t ii=0; ii<_nS; ii++)
        vecCoefficients[ii] = (Float_t)(_gamma[ii]);
    gp_gpu.setCoefficients(_nS, vecCoefficients);

    // Pass the Query points (inputs)
    Float_t *vecQueryInputs = new Float_t[nsq*_nX];
    TList * listOfInputsAttributes = NULL;
    if ( tdsQuery )
        listOfInputsAttributes = tdsQuery->getListOfAttributes(listOfInputs);
    for(Int_t is=0; is<nsq; is++) {
        for(Int_t ix=0; ix<_nX; ix++) {
            if ( tdsQuery )
                vecQueryInputs[is*_nX+ix] = (Float_t)(tdsQuery->getValue(((URANIE::DataServer::TAttribute *) listOfInputsAttributes->At(ix))->GetName(), is));
            else
                vecQueryInputs[is*_nX+ix] = (Float_t)(_xObs[is*_nX+ix]);
        }
    }
    gp_gpu.setQueryPoints(nsq, vecQueryInputs);

    // Pass the trend
    Float_t dTrend = 0.0;
    if ( ( bHas_trend == kTRUE ) && ( _nP > 0 ) )
        dTrend = _hbeta[0];
    gp_gpu.setTrends(dTrend);

    if ( 0 ) // _blog )
        gp_gpu.printLog();
    gp_gpu.estimate();

    // Fill the Output values (mean + variance)
    Double_t *tabValue = new Double_t[2];
    TString soutput = listOfOutputs;
    soutput.ReplaceAll(":", " ");
    soutput += " ";
    istringstream inStream((char *) soutput.Data());
    string inNameYMean, inNameYVariance;

    inStream >> inNameYMean;
    tdsQuery->addAttribute( new URANIE::DataServer::TAttribute(inNameYMean.c_str(), Form("#hat{%s}",  _outputVarName.Data())));

    inStream >> inNameYVariance;

    if ( _blog )
    cout << " ** inNameYMean[" << inNameYMean << "]  inNameYVariance[" << inNameYVariance << "]" << endl;
    tdsQuery->addAttribute( new URANIE::DataServer::TAttribute(inNameYVariance.c_str(), Form("Var{%s}",  _outputVarName.Data())));
    TBranch *tbr = tdsQuery->getTuple()->Branch("GPgpu", tabValue, Form("%s/D:%s/D", inNameYMean.c_str(), inNameYVariance.c_str()));

    for (Int_t ipat = 0; ipat < nsq; ipat++) {
        tdsQuery->getTuple()->GetEntry(ipat);
        tabValue[0] = (Double_t)(gp_gpu.getQueryMean(ipat));
        tabValue[1] = (Double_t)(gp_gpu.getQueryVariance(ipat));
        tbr->Fill();
    }

    // Clean
    vecReferenceInputs = NULL; delete [] vecReferenceInputs;
    vecReferenceOutput = NULL; delete [] vecReferenceOutput;

    matInverse = NULL; delete [] matInverse;
    vecCorrelationLengths = NULL; delete [] vecCorrelationLengths;
    vecCoefficients = NULL; delete [] vecCoefficients;
    vecQueryInputs = NULL; delete [] vecQueryInputs;
    tdsQuery->getTuple()->ResetBranchAddress(tbr);
    tabValue = NULL; delete [] tabValue;

}
#endif
// -----------------------------------------------------------------------
//  Evaluation
// -----------------------------------------------------------------------

Int_t TKriging::eval(Double_t* x0, Double_t* y0, int ncalc)
{
    UNUSED(ncalc);
    Double_t yp, vp;
    Double_t *x0Norm = new Double_t[_nX];
    Double_t *c0 = new Double_t[_nS];
    Double_t *k0 = new Double_t[_nS];

    // normalisation of x0
    for(Int_t i = 0; i < _nX; i++)
        x0Norm[i] = ( x0[i] - _xNormParams[i*2] ) / ( _xNormParams[i*2+1] - _xNormParams[i*2] );
    
    // Evaluate the covariances    
    _correlationFunction->eval(1, x0Norm, _nS, _xObs, c0);

    for (Int_t i = 0; i < _nS; i++)
        k0[i] = _sigma2 * c0[i];

    // Prediction
    if (!bHas_trend)
        gpPred0(k0, _sigma2, _gamma, _iL, 1, _nS, &yp, &vp);
    else
    {
        Double_t *h0 = new Double_t[_nP];

        // Fill the deterministic trend with the new point
        for (Int_t i = 0; i < _nP; i++)
        {
            // retrieve the formula for trend coefficient i
            TObjString* tmpString = (TObjString*) _trendCoefList->At(i);            
            TString coefFormula = tmpString->GetString();
            for (Int_t attIndex = 0; attIndex < _nX; attIndex++)
            {
                // update the current formula by replacing attribute
                // attName by its corresponding value.
                TString attName = _inputNamesList->At(attIndex)->GetName();
                if(bUse_normalisation)
                    coefFormula = coefFormula.ReplaceAll(attName, Form("%.14e", x0Norm[attIndex]));
                else
                    coefFormula = coefFormula.ReplaceAll(attName, Form("%.14e", x0[attIndex]));
            }
            // The resulting formula contains no more variables. We
            // create a TFormula out of it, and evaluate it (the
            // result is a constant so the value of the parameter
            // of TFormula::Eval() is meaningless.
            TFormula tmpF("tmpF", coefFormula.Data());            
            h0[i] = tmpF.Eval(0);
        }
        gpPredh(h0, k0, _sigma2, _hbeta, _gamma, _iL, _iR, _M, 1, _nS, _nP, &yp, &vp);
        delete [] h0;
    }
    y0[0] = yp;
    y0[1] = vp;
    
    delete [] x0Norm;
    delete [] c0;
    delete [] k0;
    return 1;
}


// -----------------------------------------------------------------------
//  Export Function
// -----------------------------------------------------------------------
void TKriging::exportFunction(const char* lang, const char* file,
        const char* name, Option_t *option)
{

    TString soption = option;
    soption.ToLower();
    cout << endl << "*******************************" << endl;
    cout << "*** TKriging::exportFunction lang[" << lang << "] file[" << file << "] name[" << name << "] soption[" << soption << "]" << endl;
    cout << endl << "*******************************" << endl;

    TString lg = lang;
    lg.ToLower();

    if( lg=="c" || lg=="c++" || lg=="cplusplus")
    {
        std::ofstream ofs ( file, std::ofstream::out);
        exportModelCplusplus(&ofs, name);
    }
    else if( lg=="f" || lg=="fortran")
    {
        std::ofstream ofs ( file, std::ofstream::out);
        exportModelFortran(&ofs);
    }
    else if(lg=="pmml")
        exportModelPMML(file, name, option);
    else
        cout<<"The lang ["<<lg<<"] is not valid... It should be among this list: [c++, f, pmml]"<<endl;

    cout << "*** End Of exportFunction" << endl;
    cout << "*******************************" << endl;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TKriging::exportModelCplusplus(std::ofstream *sourcefile, const char* name)
{

    TString sFunctionName = TString(name);
    if (sFunctionName.Length() == 0)
        sFunctionName = "Uranie_Kriging";
  
    Bool_t bCrtLog = kFALSE;

    TString sCorrelationName = _correlationFunction->ClassName();

    if ( bCrtLog )
        cout << " *** TKriging::exportModelCplusplus sFunctionName[" << sFunctionName << "] sCorrelationName[" << sCorrelationName << "]" << endl;

    //Bool_t bFound = kFALSE;
    *sourcefile << std::scientific;
    // -- the Correlation function
    *sourcefile << "// --- Correlation function" << endl;
    *sourcefile << "void " << sFunctionName << "_gpCorrelation(double *xNews, double *xReferences, int nNews, int nReferences, int nDimension, double *corrLength, double *C) {" << endl;

    // Case of Gaussian
    if ( sCorrelationName.Contains ("TGaussianCorrFunction") ) {
        *sourcefile << "  /* Gaussian correlation */" << endl;
        *sourcefile << "  double dsum;" << endl;
        *sourcefile << "  double ddelta;" << endl;
        *sourcefile << endl;
        //bFound = kTRUE;
        *sourcefile << "  for(int inew = 0; inew < nNews; inew++) {" << endl;
        *sourcefile << "    for(int iref = 0; iref < nReferences; iref++) {" << endl;
        if ( bCrtLog ) *sourcefile << "      if ( iref < nReferences ) printf(\"\\t ** Ref(%05d):\", iref);" << endl;
        *sourcefile << "      dsum = 0.0;" << endl;
        *sourcefile << "      for(int idim = 0; idim < nDimension; idim++) {" << endl;
        *sourcefile << "        ddelta = ( xNews[inew*nDimension+idim] - xReferences[iref*nDimension+idim]) / corrLength[idim];" << endl;
        *sourcefile << "        dsum += ddelta * ddelta;" << endl;
        if ( bCrtLog ) *sourcefile << "        if ( iref < nReferences ) printf(\" (%.5f)\", ddelta);" << endl;
        *sourcefile << "      } // End Of idim" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: Sum(%.5f)\", dsum);" << endl;
        *sourcefile << "    C[inew + iref*nNews] = exp(- dsum);" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: CC(%.5f)\\n\", exp(- dsum);" << endl;
        *sourcefile << "    } // End Of iref" << endl;
        *sourcefile << "  } // End Of inew" << endl;
    }
    // Case of TMatern12CorrFunction
    if ( sCorrelationName.Contains ("TMatern12CorrFunction") ) {
        *sourcefile << "  /* Matern12 correlation */" << endl;
        *sourcefile << "  double dsum;" << endl;
        *sourcefile << "  double ddelta;" << endl;
        *sourcefile << "  double das;" << endl;
        *sourcefile << endl;
        //bFound = kTRUE;
        *sourcefile << "  for(int inew = 0; inew < nNews; inew++) {" << endl;
        *sourcefile << "    for(int iref = 0; iref < nReferences; iref++) {" << endl;
        if ( bCrtLog ) *sourcefile << "      if ( iref < nReferences ) printf(\"\\t ** Ref(%05d):\", iref);" << endl;
        *sourcefile << "      dsum = 0.0;" << endl;
        *sourcefile << "      for(int idim = 0; idim < nDimension; idim++) {" << endl;
        *sourcefile << "        ddelta = ( xNews[inew*nDimension+idim] - xReferences[iref*nDimension+idim]) / corrLength[idim];" << endl;
        *sourcefile << "        dsum += ddelta * ddelta;" << endl;
        if ( bCrtLog ) *sourcefile << "        if ( iref < nReferences ) printf(\" (%.5f)\", ddelta);" << endl;
        *sourcefile << "      } // End Of idim" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: Sum(%.5f)\", dsum);" << endl;
        *sourcefile << "    das = sqrt( 6.0 * dsum);" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: das(%.5f)\", das);" << endl;
        *sourcefile << "    C[inew + iref*nNews] = exp(- das) * (1. + das );" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: CC(%.5f)\\n\", exp(- das) * (1. + das ));" << endl;
        *sourcefile << "    } // End Of iref" << endl;
        *sourcefile << "  } // End Of inew" << endl;

    }
        // Case of TMatern32CorrFunction
    if ( sCorrelationName.Contains ("TMatern32CorrFunction") ) {
        *sourcefile << "  /* Matern32 correlation */" << endl;
        *sourcefile << "  double dsum;" << endl;
        *sourcefile << "  double ddelta;" << endl;
        *sourcefile << "  double das;" << endl;
        *sourcefile << endl;
        //bFound = kTRUE;
        *sourcefile << "  for(int inew = 0; inew < nNews; inew++) {" << endl;
        *sourcefile << "    for(int iref = 0; iref < nReferences; iref++) {" << endl;
        if ( bCrtLog ) *sourcefile << "      if ( iref < nReferences ) printf(\"\\t ** Ref(%05d):\", iref);" << endl;
        *sourcefile << "      dsum = 0.0;" << endl;
        *sourcefile << "      for(int idim = 0; idim < nDimension; idim++) {" << endl;
        *sourcefile << "        ddelta = ( xNews[inew*nDimension+idim] - xReferences[iref*nDimension+idim]) / corrLength[idim];" << endl;
        *sourcefile << "        dsum += ddelta * ddelta;" << endl;
        if ( bCrtLog ) *sourcefile << "        if ( iref < nReferences ) printf(\" (%.5f)\", ddelta);" << endl;
        *sourcefile << "      } // End Of idim" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: Sum(%.5f)\", dsum);" << endl;
        *sourcefile << "    das = sqrt( 6.0 * dsum);" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: das(%.5f)\", das);" << endl;
        *sourcefile << "    C[inew + iref*nNews] = exp(- das) * (1. + das );" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: CC(%.5f)\\n\", exp(- das) * (1. + das ));" << endl;
        *sourcefile << "    } // End Of iref" << endl;
        *sourcefile << "  } // End Of inew" << endl;

    }
    if ( sCorrelationName.Contains ("TMatern52CorrFunction") ) {
        *sourcefile << "  /* Matern52 correlation */" << endl;
        *sourcefile << "  double dsum;" << endl;
        *sourcefile << "  double ddelta;" << endl;
        *sourcefile << "  double das;" << endl;
        //bFound = kTRUE;
        *sourcefile << "  for(int inew = 0; inew < nNews; inew++) {" << endl;
        *sourcefile << "    for(int iref = 0; iref < nReferences; iref++) {" << endl;
        if ( bCrtLog ) *sourcefile << "      if ( iref < nReferences ) printf(\"\\t ** Ref(%05d):\", iref);" << endl;
        *sourcefile << "      dsum = 0.0;" << endl;
        *sourcefile << "      for(int idim = 0; idim < nDimension; idim++) {" << endl;
        *sourcefile << "        ddelta = ( xNews[inew*nDimension+idim] - xReferences[iref*nDimension+idim]) / corrLength[idim];" << endl;
        *sourcefile << "        dsum += ddelta * ddelta;" << endl;
        if ( bCrtLog ) *sourcefile << "        if ( iref < nReferences ) printf(\" (%.5f)\", ddelta);" << endl;
        *sourcefile << "      } // End Of idim" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: Sum(%.5f)\", dsum);" << endl;
        *sourcefile << "    das = sqrt( 10.0 * dsum);" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: das(%.5f)\", das);" << endl;
        *sourcefile << "    C[inew + iref*nNews] = exp(- das) * (1. + das + 10.0 * dsum / 3.0 );" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: CC(%.5f)\\n\", exp(- das) * (1. + das + 10.0 * dsum / 3.0 ) );" << endl;
        *sourcefile << "    } // End Of iref" << endl;
        *sourcefile << "  } // End Of inew" << endl;
    }
    if ( sCorrelationName.Contains ("TMatern72CorrFunction") ) {
        *sourcefile << "  /* Matern72 correlation */" << endl;
        *sourcefile << "  double dsum;" << endl;
        *sourcefile << "  double ddelta;" << endl;
        *sourcefile << "  double das;" << endl;
        //bFound = kTRUE;
        *sourcefile << "  for(int inew = 0; inew < nNews; inew++) {" << endl;
        *sourcefile << "    for(int iref = 0; iref < nReferences; iref++) {" << endl;
        if ( bCrtLog ) *sourcefile << "      if ( iref < nReferences ) printf(\"\\t ** Ref(%05d):\", iref);" << endl;
        *sourcefile << "      dsum = 0.0;" << endl;
        *sourcefile << "      for(int idim = 0; idim < nDimension; idim++) {" << endl;
        *sourcefile << "        ddelta = ( xNews[inew*nDimension+idim] - xReferences[iref*nDimension+idim]) / corrLength[idim];" << endl;
        *sourcefile << "        dsum += ddelta * ddelta;" << endl;
        if ( bCrtLog ) *sourcefile << "        if ( iref < nReferences ) printf(\" (%.5f)\", ddelta);" << endl;
        *sourcefile << "      } // End Of idim" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: Sum(%.5f)\", dsum);" << endl;
        *sourcefile << "    das = sqrt( 14.0 * dsum);" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: das(%.5f)\", das);" << endl;
        *sourcefile << "    C[inew + iref*nNews] = exp(- das) * (1. + das + 2.0 * das * das / 5.0 + das * das * das / 15.0 );" << endl;
        if ( bCrtLog ) *sourcefile << "    if ( iref < nReferences ) printf(\"   :: CC(%.5f)\\n\", exp(- das) * (1. + das + 2.0 * das * das / 5.0 + das * das * das / 15.0 ) );" << endl;
        *sourcefile << "    } // End Of iref" << endl;
        *sourcefile << "  } // End Of inew" << endl;
    }

    *sourcefile << "}" << endl;
    *sourcefile << endl;

    // -- the Trend function
    *sourcefile << "// --- Trend function" << endl;
    *sourcefile << "double " << sFunctionName << "_Trend(double *xNormalized) {" << endl;
    if ( bHas_trend == kFALSE ) {
        *sourcefile << "    /* Null trend [" << _trendString << "] */" << endl;
        *sourcefile << "    return ((double) 0.0 );" << endl;
    } else  {
        *sourcefile << "    /* trend [" << _trendString << "] */" << endl;
        if(_nP == 1)
        {
            *sourcefile << "    return ((double) " << _hbeta[0] << ");" << endl;
        }
        else if(_nP > 1)
        {
            *sourcefile << "    double param["<<_nP<<"]={";
            for(int _ip=0; _ip<_nP; _ip++)
                *sourcefile << _hbeta[_ip] << ((_ip==(_nP-1))?"":", ");
            *sourcefile << "};" << endl;
            *sourcefile << "    double res=param[0];" << endl;
            *sourcefile << "    for(unsigned int _ip=1; _ip <"<<_nP<<"; _ip++) " << endl;
            *sourcefile << "        res += param[_ip] * xNormalized[_ip-1]; " << endl;            
            *sourcefile << "    return res;"<<endl;
        }
    }
    *sourcefile << "}" << endl;
    *sourcefile << endl;

    // -- The function
    *sourcefile << "// --- Function" << endl;
    *sourcefile << "void " << sFunctionName << "(double *x, double *y) {" << endl;
    // Parameters in the fucntion
    *sourcefile << endl;
    *sourcefile << "  // --- Parameters in the fucntion" << endl;
    *sourcefile << "  int nX = " << _nX << ";" << endl;
    *sourcefile << "  int nS = " << _nS << ";" << endl;
    *sourcefile << "  double xn[" << _nX <<"];" << endl;
    *sourcefile << "  double C[" << _nS <<"];" << endl;
    *sourcefile << "  double yp = 0.0;" << endl;

    // References Points
    *sourcefile << endl;
    *sourcefile << "  // --- References Points" << endl;
    *sourcefile << "  double xReferences[" << _nX * _nS << "] = {" << endl;
    for(Int_t ipat=0; ipat < _nS; ipat++) {
        *sourcefile << "  ";
        for(Int_t ivar=0; ivar<_nX; ivar++) {
            *sourcefile << setprecision(14) << _xObs[ipat*_nX+ivar];
            if ( ivar < _nX-1 )
                *sourcefile << ", ";
        }
        if ( ipat < _nS -1 )
            *sourcefile << ", ";
        *sourcefile << endl;
    }
    *sourcefile << "  };"<<endl;

    // Coefficients
    *sourcefile << endl;
    *sourcefile << "  // --- Coefficients " << endl;
    *sourcefile << "  double xCoefficients[" << _nS <<"] = {" << endl;
    *sourcefile << "  ";
    for(Int_t i=0; i < _nS; i++) {
        *sourcefile << _gamma[i];
        if ( i < _nS-1 )
            *sourcefile << ", ";
        if ( (i+1) %5 == 0 ) {
            *sourcefile << endl << "  ";
        }
    }
    *sourcefile << "  };"<<endl;

    // iL
    *sourcefile << endl;
    *sourcefile << "  // --- Inverse of L " << endl;
    *sourcefile << "  double xiL[" << _nS*_nS <<"] = {" << endl;
    *sourcefile << "  ";
    for(Int_t i=0; i < _nS; i++) {
        for(Int_t j=0; j < _nS; j++) {
            *sourcefile << _iL[i*_nS+j];
            if ( (i < _nS-1) ||
              ( (i == _nS-1) && ( j < _nS-1) ))
                *sourcefile << ", ";
        }
        *sourcefile << endl << "  ";
    }
    *sourcefile << "  };"<<endl;

    if ( bUse_normalisation ) {
        *sourcefile << endl;
        *sourcefile << "  // --- Min and Max for normalization" << endl;
        *sourcefile << "  double xMin[" << _nX <<"] = {"<<endl;
        *sourcefile << "    ";
        for(Int_t i = 0; i < _nX; i++) {
            *sourcefile << _xNormParams[i*2];
            if ( i < _nX-1 )
                *sourcefile << ", ";
            if ( (i+1) %5 == 0 ) {
                *sourcefile << endl << "    ";
            }
        }
        *sourcefile << "};"<<endl;

        *sourcefile << "  double xMax[" << _nX <<"] = {"<<endl;
        *sourcefile << "    ";
        for(Int_t i = 0; i < _nX; i++) {
            *sourcefile << _xNormParams[i*2+1];
            if ( i < _nX-1 )
                *sourcefile << ", ";
            if ( (i+1) %5 == 0 ) {
                *sourcefile << endl << "    ";
            }
        }
        *sourcefile << "  };"<<endl;
    }

    *sourcefile << "  double thetaOpt[" << _nX <<"] = {"<<endl;
    *sourcefile << "    ";
    for(Int_t i = 0; i < _nX; i++) {
        *sourcefile << _normCorrLengths[i];
        if ( i < _nX-1 )
            *sourcefile << ", ";
        if ( (i+1) %5 == 0 ) {
            *sourcefile << endl << "    ";
        }
    }
    *sourcefile << "  };"<<endl;


    // Normalize input
    if ( bUse_normalisation ) {
        if ( bCrtLog ) *sourcefile << "     printf(\"\\n\");" << endl;
        *sourcefile << endl;
        *sourcefile << "  // --- Normalize input" << endl;
        *sourcefile << "  for(int i = 0; i < nX; i++) {" << endl;
        *sourcefile << "     xn[i] = ( x[i] - xMin[i] ) / ( xMax[i] - xMin[i] );" << endl;
        if ( bCrtLog ) *sourcefile << "     printf(\" * (%d) V.(%.5f) N.(%.5f)\", i+1, x[i], xn[i]);" << endl;
        *sourcefile << "  }" << endl;
    }
    if ( bCrtLog ) *sourcefile << "     printf(\"\\n\");" << endl;

    // Apply the Correlation function
    *sourcefile << endl;
    *sourcefile << "  // --- Apply the Correlation function" << endl;
    *sourcefile << "  " << sFunctionName << "_gpCorrelation(xn, xReferences, (int) 1, nS, nX, thetaOpt, C);" << endl;

    // Dot scalar betwen C and Coeff
    *sourcefile << endl;
    *sourcefile << "  // --- Dot scalar betwen C and Coeff" << endl;
    *sourcefile << "  for(int i = 0; i < nS; i++) {" << endl;
    *sourcefile << "     yp += xCoefficients[i] * C[i];" << endl;
    if ( bCrtLog ) *sourcefile << "    if ( i < nS ) printf(\" (%05d/%05d) : CC(%.5f) (%.5f) yyp (%.5f) \\n\", i+1, nS, xCoefficients[i], C[i], yp);" << endl;
    *sourcefile << "  }" << endl;
    if ( bCrtLog ) *sourcefile << "     printf(\"\\n\\t ** dProdCoeff (%.5f)\\n\\n\", yp);" << endl;

    // End of the function
    *sourcefile << endl;
    *sourcefile << "  // --- Return the kriging value : Trend + Variance * yp" << endl;
    *sourcefile << "  y[0] = " << sFunctionName << "_Trend(xn) + "  << _sigma2 << " * yp;" << endl;

    *sourcefile << endl;
    *sourcefile << "  // --- Return the variance : yp * iL-1 * yp " << endl;
    *sourcefile << "  y[1] = " << _sigma2 << ";" << endl;
    *sourcefile << "  // --- Dot scalar betwen C and Coeff" << endl;
    *sourcefile << "  for(int i = 0; i < nS; i++) {" << endl;
    *sourcefile << "    yp = 0.0;" << endl;
    *sourcefile << "    for(int j = 0; j <= i; j++) {" << endl;
    *sourcefile << "     yp += xiL[i*nS+j] * C[j];" << endl;
    *sourcefile << "    }" << endl;
    if ( bCrtLog ) *sourcefile << "    if ( i < nS ) printf(\" (%04d/%04d) : yp(%.5f)\\n\", i, nS, yp);" << endl;
    *sourcefile << "    y[1] -= " << _sigma2 * _sigma2 << " * yp * yp;" << endl;
    *sourcefile << "  }" << endl;
    *sourcefile << "  if ( y[1] < 0.0 ) y[1] = 0.0;" << endl;


    *sourcefile << "}" << endl;
    *sourcefile << endl;

}

void TKriging::exportModelPMML(const char* file, const char* name, Option_t *option) const
{

    // ----------------------------
    enum
    {
        klocBACKUP, klocUPDATE, klocNEW, klocADD
    };
    TString soption = TString(option);  // NEW / UPDATE / BACKUP / ADD
    soption.ToLower();
    Int_t nOption = klocBACKUP;
    if (soption.Contains("update"))
        nOption = klocUPDATE;
    if (soption.Contains("new"))
        nOption = klocNEW;
    // ----------------------------
    TString sCorrelationName = _correlationFunction->ClassName();
    try {
        if  (sCorrelationName.CompareTo("URANIE::Modeler::TGaussianCorrFunction", TString::kIgnoreCase) != 0 ){
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form("TKriging::export PMML not yet implemented for (%s)",sCorrelationName.Data()));
	} 
	if (bHas_trend){
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form("TKriging::export PMML not yet implemented with trend"));
	}
    }
    catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }


#ifdef  WITH_LIBXML2
    PMMLlib::PMMLlib *pmml = NULL;
    URANIE::DataServer::TAttribute *att;
    Bool_t bGoodOption = kFALSE;
    while(bGoodOption==kFALSE)
    {
      switch(nOption)
        {
	case klocNEW:
	  pmml = new PMMLlib::PMMLlib(_blog);
	  // Header node
	  pmml->SetHeader("myCopyright", "Text Description", "Uranie", "3.12.0", "Compilation date : 2000-01-01");
	  // DataDictionnary node
	  for(Int_t j=0; j<_nX; j++)
	    {
	      pmml->AddDataField(_inputNamesList->At(j)->GetName(), _inputNamesList->At(j)->GetName(),
				 "continuous", "double", "closedClosed", 
				 _xNormParams[2*j], _xNormParams[2*j+1], true);
	    }
	  
	  pmml->AddDataField(_outputVarName.Data(), _outputVarName.Data(), "continuous", "double", "ClosedClosed", 0., 0., false);
	  bGoodOption = kTRUE;
	  break;
	  // end of "case NEW"

	case klocUPDATE:
	  try
	    {
	      pmml = new PMMLlib::PMMLlib(file,_blog);
	    }
	  catch (std::string msg)
	    {
	      std::cerr <<"\n New PMML file created"<<endl;
	      nOption = klocNEW;
	      if(pmml)
		delete pmml;
	      continue;
	    }
	  try
	    {
	      pmml->SetCurrentModel(name, PMMLlib::kANN);
	    }
	  catch (std::string msg)
	    {
	      std::cerr <<"\n New model created"<<endl;
	      nOption = klocADD;
	      if(pmml)
		delete pmml;
	      continue;
	    }
	  pmml->UnlinkNode();
	  bGoodOption = kTRUE;

	  break;
	  // end of "case UPDATE"

	case klocBACKUP:
	  try
	    {
	      pmml = new PMMLlib::PMMLlib(file,_blog);
	    }
	  catch (std::string msg)
	    {
	      std::cerr <<"\n New PMML file created"<<endl;
	      nOption = klocNEW;
	      if(pmml)
		delete pmml;
	      continue;
	    }
	  try
	    {
	      pmml->SetCurrentModel(name, PMMLlib::kGAUSS);
	    }
	  catch (std::string msg)
	    {
	      std::cerr <<"\n New model created"<<endl;
	      nOption = klocADD;
	      if(pmml)
		delete pmml;
	      continue;
	    }
	  pmml->BackupNode();
	  bGoodOption = kTRUE;
	  
	  break;
	  // end of "case BACKUP"

	case klocADD:
	  pmml = new PMMLlib::PMMLlib(file,_blog);
	  bGoodOption = kTRUE;
	  break;
	  //             end of "case ADD"

        } // end of "switch(nOption)"
    }
    // ---------------------------------------------

    // ------- Gaussian Process Definition
    pmml->AddGaussianProcess(name, PMMLlib::kREGRESSION);

    // ------- Mining Schema
    for(Int_t j=0; j<_nX; j++)
    {
      // Set MiningSchema
      pmml->AddMiningSchema(_inputNamesList->At(j)->GetName(),"active"); // Input
    }
    pmml->AddMiningSchema(_outputVarName.Data(), "predicted"); // Output

    // ------- Kernel Type
    pmml->AddGaussianKernelType(PMMLlib::kARDSquaredExponential, _sigma2, _nX, _corrLengths);

    // ------- TrainingInstances
    std::vector<std::string> inputNames;
    for(Int_t j=0; j<_nX; j++)
      {
	inputNames.push_back(_inputNamesList->At(j)->GetName());
      }

    pmml->AddTrainingInstances(_nS,_nX, inputNames, _outputVarName.Data(), _xObs, _xNormParams, _yObs);
    // -------
    //
    switch(nOption)
    {
        case klocNEW:
        pmml->Write(file);
        break;
        // end of "case NEW"
        default :
        pmml->Write();
        break;
        // end of "case BACKUP", "UPDATE", "ADD"
    } // end of "switch(nOption)"

    //Clean
    delete pmml;
#else
    UNUSED(file,name);
#endif
    cout << "*** End Of exportModelPMML " << endl;
    cout << "*******************************" << endl;
}

#ifdef DVEC
Int_t TKriging::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    UNUSED(ncalc);
    double *inp = new double[in->size()];
    double *outp = new double[out->size()];
    for(unsigned int ival=0; ival<in->size(); ival++) inp[ival] = *in->at(ival)->val.dou;

    int admis=eval(inp, outp);
    
    for(unsigned int ival=0; ival<out->size(); ival++) *out->at(ival)->val.dou = outp[ival];
    delete [] inp; delete [] outp;
    return ((admis!=0) ? getOutDimension(out) : admis);
}
#endif

// -----------------------------------------------------------------------
//  
// -----------------------------------------------------------------------
void TKriging::freeze()
{

    _nYDeclared=_output.GetEntries();
    try {
        if(  _nYDeclared != 2 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form("TKriging::freeze, the list of output attribute is of size [%d].\n\
This should not be: a kriging object returns the estimated value as first attribute, and its prediction uncertainty as second attribute. \n\
Please make sure that you've added both output to your TKriging object.",_nYDeclared));
    }
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        _output.Delete();
        throw ue;
    }


    _nXDeclared=_input.GetEntries();
    try {
        if(  _nXDeclared != _nX )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, Form("TKriging::freeze, the list of input attribute is of size [%d].\n\
This should not be: this kriging object has been consructed with [%d] input attributes.\n\
Please make sure that you've added the correct number of input attributes (in the right order).",_nXDeclared,_nX));
    }
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        _output.Delete();
        throw ue;
    }

    TBaseEval::freeze();
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TKriging::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << "** TKriging::printLog[" << option << "]" << endl;
    cout << "*******************************" << endl;
    cout << " Input Variables:      " << _inputVarNames << endl;
    cout << " Output Variable:      " << _outputVarName << endl;
    cout << " Deterministic trend:  " << _trendString << endl;
    if(_nP > 0)
    {
        // Information about the trend
        cout << " Trend parameters (" << _nP << "): [";
        printf("%.8e", _hbeta[0]);
        for (Int_t i = 1; i < _nP ; i++)
          printf("; %.8e", _hbeta[i]);
        cout << " ]" << endl;
    }
    
    cout << " Correlation function: " << _correlationFunction->ClassName() << endl;
    // Information about the hyper-parameters
    Int_t nbCorrLengths    = getNCorrLengths();     // number of correlation length
    Double_t *CorrLengths  = getCorrLengths();	     // correlation length
    Double_t *nCorrLengths = getCorrLengthsNormalised();  // normalized correlation length
    cout << " Correlation length:   normalised   (not normalised)" << endl;
    for (Int_t i = 0; i < nbCorrLengths ; i++)
        printf("                       %.4e (%.4e )\n", nCorrLengths[i], CorrLengths[i]);

    Int_t nbOtherParameters    = getNCorrFunctionParams() - nbCorrLengths;	// number of hyperparamaters without correlation length
    Double_t *otherParameters  = getCorrFunctionParams();			// values of hyperparamaters without correlation length
    if (nbOtherParameters > 0) {
      cout << " Other parameters (exponential, matern): " << endl;
      for (Int_t i = 0; i < nbOtherParameters ; i++)
          printf("                      %.4e\n", otherParameters[i]);
    }
    cout << endl;
    cout << " Variance of the gaussian process:      " << _sigma2 << endl;
    if(_vErrMes > 0.0)
        cout << " Variance of the Measurement Error:     " << _vErrMes << endl;
    cout << " RMSE (by Leave One Out):               " << getLooRMSE() << endl;
    cout << " Q2:                                    " << getLooQ2() << endl;
    cout << "*******************************" << endl;
    
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

