//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TKriging.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date

 \brief Interface de la classe URANIE::Modeler::TKriging

 */
/**

 * @class URANIE::Modeler::TKriging
 *
 * @brief Description of the class TKriging.
 *
 * This class defines a Kriging model. It can predict the output of new observations
 * using the TKriging::eval function or the URANIE::Relauncher module formalism. It 
 * is created by TGPBuilder::buildGP. It is a wrapper to the `gpPredX` functions from the 
 * "gaussian process Library" (gpL), by Jean-Marc Martinez.
 *
 */
#ifndef TKRIGING_H
#define TKRIGING_H

#include <cmath>

//// ROOT
#include "TObjArray.h"
#include "TMath.h"

//// URANIE
#include "TCorrelationFunction.h"
#include "TStandardEval.h"
#include "TDataServer.h"
#include "TAttribute.h"
#include "TDataServer.h"

//// Jean-Marc's Gaussian Processes Library
extern "C"
{
    void gpPred0(double *Kp, double variance, double *gamma, double *iL, 
                 int nnew, int n, double *yp, double *vp);
    void gpPred0Cov(double *Kp, double *Rp, double variance, double *gamma, double *iL,
                    int nnew, int n, double *yp, double *vp);
    void gpPredh(double *Hp, double *Kp, double variance, double *hbeta, double *gamma, double *iL, double *iR, double *M, 
                 int nnew, int n, int p, double *yp, double *vp);
    void gpPredhCov(double *Hp, double *Kp, double *Rp, double variance, double *hbeta, double *gamma, double *iL, double *iR, double *M, 
                    int nnew, int n, int p, double *yp, double *vp);
}
;

namespace URANIE
{
namespace Modeler
{
class TKriging: public URANIE::Relauncher::TDoubleEval
{

    // Attributes
private:
    Bool_t _blog; ///< Boolean for edit the log
public:
    TString    _outputVarName;  ///< Name of the output variable. This information can be used to remind the user of the original names of the variables.
    TString    _inputVarNames;  ///< Name of the inputs variables separated by ":". This information can be used to remind the user of the original names of the variables.
    TObjArray* _inputNamesList; ///< list of the inputs' names.
    
    Int_t     _nX;          ///< Number of input attributes    
    Int_t     _nS;          ///< Number of observations
    Int_t     _nYDeclared;  ///< Number of attribute added to the tkiging object before running it.
    Int_t     _nXDeclared;  ///< Number of attribute added to the tkiging object before running it.
    Double_t* _xObs;        ///< The matrix of inputs (dimension: _nX * _nS)
    Double_t* _xNormParams; ///< The matrix of normalisation parameters. Contains min and max value of each input variable: [aMin, aMax, bMin, bMax,...] (size: _nX * 2)
    Double_t* _yObs;        ///< Array representation of the outputs (size: _nS)
    Double_t* _iL;          ///< Inverse of the matrix L, Cholesky decomposition of K (size: _nS * _nS)
    Double_t* _gamma;       ///< Array of parameters of the gaussian process (size: _nS)
    
    Double_t              _sigma2;              ///< Variance of the gaussian process
    Double_t              _vErrMes;             ///< variance of the measurement error
    TCorrelationFunction* _correlationFunction; ///< The correlation function
    Double_t*             _corrLengths;         ///< The correlation lengths in the real input space
    Double_t*             _normCorrLengths;     ///< The correlation lengths in the normalised input space. This corresponds to the correlation lengths stored in the correlation function.
    
    Bool_t     bUse_normalisation; ///< if kFALSE, the matrix H will be computed with unormalised input data. This happens when the user is providing his own trend function. When the trend is constructed automatically from a trend type, this attribute is set to kTRUE. 
    Bool_t     bHas_trend;     ///< kFALSE if no deterministic trend is defined, kTRUE otherwise
    TString    _trendString;   ///< Character string containing the trend parameters separated by ":" or a trend type ("const" or "linear")    
    TObjArray* _trendCoefList; ///< List of the formulas of the trend coefficients (_nP elements)
    Int_t      _nP;            ///< Number of coefficients of the deterministic trend
    Double_t*  _hbeta;         ///< Estimates of the deterministic trend parameters (size: _nP)
    Double_t*  _vbeta;         ///< Covariance matrix of the trend parameters (size: _nP*_nP)
    Double_t*  _M;             ///< Matrix M = L^{-T} H (size: _nS * _nP)
    Double_t*  _iR;            ///< Inverse of the matrix R, from the QR decomposition of M (size: _nP * _nP)
    
    Double_t* _errorLoo; ///< If available, vector of the Leave One Out errors for the model.
    Double_t* _varLoo;   ///< If available, vector of the variances of the Leave One Out errors for the model.
    Double_t  _mseLoo;   ///< if available, Leave one Out mean squared error of the model.

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
    Constructs a TKriging from the raw matrices and other stuff produced by 
    TGPBuilder. This constructor is not supposed to be used by a human being.
    \param nX (Int_t): number of variables
    \param nS (Int_t): number of observations
    \param xObs (Double_t*): the normalised observations (size = nX * nS)
    \param xNormParams (Double_t*): the normalisation parameters of xObs (size = 2 * nX)
    \param yObs (Double_t*); the outputs (size = _nS)
    \param gamma (Double_t*): parameters of the Kriging model (size = nS)
    \param iL (Double_t*): inverse of matrix L (size = nS * nS)
    \param gpVar (Double_t): variance of the Gaussian Process
    \param vErrMes (Double_t): variance of the measurement error
    \param corrFunc (TCorrelationFunction*): correlation function of the Gaussian Process
    \param varNames (TString ): name of the variables separated by ":". The last name is the output variable. 
    \param nP (Int_t): number of parameters of the trend (default = 0)
    \param hbeta (Double_t *): parameters of the trend (size = nP) (default = NULL)
    \param M (Double_t*): matrice M (size = nS * nP) (default = NULL)
    \param iR (Double_t*): inverse de la matrice R (size = nP * nP) (default = NULL)
    \param trend (TString): string listing the trend coefficients (default = "")
    \note the TKriging object is copying every information provided. This means 
          that the source of this information can be safely destroyed after 
          creating the TKriging object.
     */
     TKriging(Int_t nX, Int_t nS, Double_t* xObs, Double_t* xNormParams, Double_t* yObs, Double_t* gamma, Double_t* iL,
              Double_t gpVar, Double_t vErrMes, URANIE::Modeler::TCorrelationFunction* corrFunc, TString varNames,
              Int_t nP = 0, Double_t* hbeta = NULL, Double_t* vbeta= NULL, Double_t* M = NULL, Double_t* iR = NULL, TString trend = "");

    //@{
#ifdef WITH_LIBXML2
    //! PMML constructor
    /*!
    Constructs a TKriging from a PMML file
    \param filename (TString): PMML file name from which to extract data
    \param modelName (TString): model name to be read
     */
     TKriging(TString filename, TString modelName);
#endif
    //! Initialisation of the matrices
    void initMatrices();

    //! Default destructor
    virtual ~TKriging();
    //@}

    //---------------------------------------------
    /** @name Get methods
     **/
    //@{
    //! Return the input variable names
    const char* getInputNames()
    {
        return _inputVarNames.Data();
    }

    //! Return the output variable name
    const char* getOutputName()
    {
        return _outputVarName.Data();
    }
    
    //! Return kTRUE if a deterministic trend has been defined
    Bool_t hasTrend()
    {
        return bHas_trend;
    }

    //! Return a pointer to the correlation function of the gaussian process
    TCorrelationFunction* getCorrFunction()
    {
        return _correlationFunction;
    }

    //! Return the number of input parameters
    Int_t getNInputs()
    {
        return _nX;
    }

    //! Return the number of observations used to build the GP model
    Int_t getNObs()
    {
        return _nS;
    }

    //! Return the number of parameters of the correlation function
    Int_t getNCorrFunctionParams()
    {
        return _correlationFunction->getNParameters();
    }

    //! Return the number of correlation lengths
    Int_t getNCorrLengths()
    {
        return _correlationFunction->getNCorrLengths();
    }
    
    //! Return the number of parameters of the deterministic trend
    Int_t getNTrendParams()
    {
        return _nP;
    }

    //! Return the observations inputs as a flat array: [a1, b1, c1, a2, b2, c2,...]
    Double_t* getObsInputs()
    {
        return _xObs;
    }

    //! Return the observations outputs
    Double_t* getObsOutputs()
    {
        return _yObs;
    }

    //! Return the normalisation parameters of the inputs as a flat array: [aMin, aMax, bMin, bMax,...]
    Double_t* getObsNormParams()
    {
        return _xNormParams;
    }
    
    //! Return the array of the Gaussian Process parameters.
    Double_t* getGPParams()
    {
        return _gamma;
    }

    //! Return the array of the trend parameters.
    Double_t* getTrendParams()
    {
        return _hbeta;
    }

    //! Return the array of the trend parameters covariance matrix.
    Double_t* getTrendParamsCovMatrix()
    {
        return _vbeta;
    }
    
    //! Return the deterministic trend list of formulas
    TObjArray* getTrendFormula()
    {
        return _trendCoefList;
    }

    //! Return the deterministic trend string
    TString getTrendString()
    {
        return _trendString;
    }
    
    //! Return the variance of the gaussian process
    Double_t getVariance()
    {
        return _sigma2;
    }

    //! Return the variance of the measurement error
    /*!
    \note if a value of -1.0 is returned, it means that the measurement error 
          used to create the model was user defined. It was not considered 
          useful to saved it with the model.
    */
    Double_t getMeasurementErrorVariance()
    {
        return _vErrMes;
    }

    //! Return the iL matrix as an array (cf. internal variable description)
    Double_t* getiL()
    {
        return _iL;
    }

    //! Return the M matrix as an array (cf. internal variable description)
    Double_t* getM()
    {
        return _M;
    }

    //! Return the iR matrix as an array (cf. internal variable description)
    Double_t* getiR()
    {
        return _iR;
    }
    
    
    //! Return the vector of the leave one out errors
    Double_t* getLooErrors()
    {
        return _errorLoo;
    }
    
    //! Return the vector of the leave one out prediction variances 
    Double_t* getLooVariances()
    {
        return _varLoo;
    }
    
    //! Return the vector of the leave one out errors
    void getLooErrors(double *arr, int size){
        try{
            if(size!=_nS)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TKriging::getLooErrors method\n The size of provided array [%d] is not the size of the output to be dumped [%d]",
                                                                size, _nS));
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

        for(int ival=0; ival<size; ival++)
            arr[ival]=_errorLoo[ival];
        
    }
    
    //! Return the vector of the leave one out prediction variances 
    void getLooVariances(double *arr, int size){
         try{
            if(size!=_nS)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TKriging::getLooErrors method\n The size of provided array [%d] is not the size of the output to be dumped [%d]",
                                                                size, _nS));
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

        for(int ival=0; ival<size; ival++)
            arr[ival]=_varLoo[ival];
    }
    //! Return the leave one out mean squared error
    Double_t getLooRMSE()
    {
        return sqrt(_mseLoo);
    }

    //! Return the Q2 of the model on the observation.
    Double_t getLooQ2()
    {
        Double_t mean = TMath::Mean(_nS, _yObs);
        Double_t sumVar = 0.0;
        for(Int_t i = 0; i < _nS; i++) sumVar += (_yObs[i]-mean)*(_yObs[i]-mean);
        if(_nS > 1)
        {
            Double_t varYObs = sumVar/(_nS-1);
            return 1.0 - _mseLoo/varYObs;
        }
        else
            return 1.0 - _mseLoo/sumVar;                
    }
    
    //! Return a pointer to a TDataServer containing the observations and the Leave One Out results.
    /*!
    \note the user is responsible of the memory allocated for the data server.
    */
    URANIE::DataServer::TDataServer* getLooData();
    
    //! Return the normalised correlation lengths
    /*!
    \note this corresponds to the correlation lengths stored in the correlation function.
    */
    Double_t* getCorrLengthsNormalised()
    {
        return _normCorrLengths;
    }
    
    //! Return the correlation lengths
    /*!
    \note these are NOT the correlation lengths stored in the correlation function. 
    These values are denormalised in order to be interpretable in the "real" input space. 
    */
    Double_t* getCorrLengths()
    {
        return _corrLengths;
    }

    //! Return all the parameters of the correlation function
    /*!
    \warning this is an alias to TCorrelationFunction::getParameters(). 
    This means that the correlation lengths are stored at the *end* of the array, 
    and are *normalised*.
    */
    Double_t* getCorrFunctionParams()
    {
        return _correlationFunction->getParameters();
    }
    
    //---------------------------------------------
    /** @name Set methods
     **/
    //@{
    //! Set the correlation function of the gaussian process
    /*!
     \param corrFunc (URANIE::Modeler::TCorrelationFunction*): a pointer to the correlation function of the gaussian process.
     */
     void setCorrFunction(TCorrelationFunction* corrFunc)
     {
         _correlationFunction = corrFunc;
     }

    //! Set the variance of the gaussian process
    /*!
     \param var (Double_t): value of the variance.
     */
    void setVariance(Double_t var)
    {
        _sigma2 = var;
    }

    //! Set the information about the leave one out error of the model.
    /*! 
    This function is called by the GPBuilder if the user wants the leave one out error of the model. The errors are computed thanks to the gpLooXError functions of the gpLib.
    \param err (Double_t*): vector of size _nS of the errors from Leave One Out.
    \param var (Double_t*): vector of size _nS of the analytical variances of the leave one out errors.
    \param mse (Double_t):  leave one out mean squared error.
    */
    void setLooErrors(Double_t *err, Double_t *var, Double_t mse);
    
    //! Create the list of trend coefficients.
    void createTrend();

    static void StatMat(std::string title, int n1, int n2, Double_t *v);
    static void StatMat(std::string title, int n1, int n2, Double_t *v,
			std::vector<double> &tvmin, std::vector<double>  &tvmax,
			std::vector<double> &tmean, std::vector<double>  &tvar,
			std::vector<double> &tnorm);
    static void StatVect(std::string title, int n, Double_t *v);
    static void StatVect(std::string title, int n, Double_t *v, double &vmin, double &vmax, double &mean, double &var, double &norm);
    void computeCovarianceMatrix(int ns, double *xObs,
				 bool bVariance_in_factor, 
				 bool bHas_measurement_error, double alpha, double reg,
				 double sigma2, double *Vmes,
				 double *C, double *K);

    //@}

    //---------------------------------------------
    /** @name Evaluation
     **/
    //@{
    
    //! Evaluate the gaussian process at the new location x0
    /*!
     This function aims to be compatible with the TEval prototype, which
     will allow to call it from a launcher or an optimizer.
     \param x0 (Double_t*): a single set of input variables values.
     \param y0 (Double_t*): output of the gaussian process for x0.
     \warning the size of x0 is not checked. If it is different from the
     expected number of input variables, this will lead to
     unexpected results or an error.
     \note Calls the gpPred0 function from the gpL
     */
    Int_t eval(Double_t* x0, Double_t* y0, int=0);

#ifdef WITH_GPUANN
      void estimate(URANIE::DataServer::TDataServer *tdsQuery,
                    TString listOfInputs = "",
                    TString listOfOutputs = "",
                    Option_t * option = "",
                    Bool_t useGPU = true);
      void estimate_GPU(URANIE::DataServer::TDataServer *tdsQuery,
                        TString listOfInputs = "",
                        TString listOfOutputs = "",
                        Option_t * option = "");
#else
      void estimate(URANIE::DataServer::TDataServer *tdsQuery,
                    TString listOfInputs = "",
                    TString listOfOutputs = "",
                    Option_t * option = "",
                    Bool_t useGPU = false);
#endif
      void estimate_CPU(URANIE::DataServer::TDataServer *tdsQuery,
                        TString listOfInputs = "",
                        TString listOfOutputs = "",
                        Option_t * option = "");

    void estimateWithCov(URANIE::DataServer::TDataServer *tdsQuery,
                         TString listOfInputs = "",
                         TString listOfOutputs = "",
                         TString realValue = "",
                         Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Tools
    Exporting the kriging model as a stand alone C function is difficult for 
    various reasons. The main problem is that we would have to store potentially 
    huge matrices as text, which is highly unpractical. It was thus decided to 
    postpone this export as long as a workable solution isn't found.
     **/
    //@{

    void exportFunction(const char* lang, const char* file, const char* name, Option_t *option="");
    
    //! Export the model as a C function (not implemented)
    virtual void exportModelCplusplus(std::ofstream *sourcefile, const char* name = "");

    //! Export the model as a Fortran function (not implemented)
    virtual void exportModelFortran(std::ofstream *sourcefile)
    {
	UNUSED(sourcefile);
        std::cout << "TKriging::exportModelFortran not yet implemented" << std::endl;
    }
    virtual void exportModelPMML(const char* file, const char* name, Option_t *option) const;

    void freeze();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TKriging, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
