//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLinearRegression.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu jun 5 10:35:25 CEST 2008

 \brief Implementation of the class  URANIE::Modeler::TLinearRegression

 \par Modifications
 \li F. Gaudier 17/06/2010
 Take into account the new option "nointercept" for exclude the intercept in the linear regression
 */

//// ROOT
#include "TDecompSVD.h"
#include "TDecompChol.h"
#include "TLeaf.h"
#include "TMath.h"
#include "TTreeFormula.h"

//// URANIE
#include "TLinearRegression.h"
#include "TAttributeFormula.h"

#ifdef  WITH_LIBXML2
#include "PMMLlib.hxx"
#endif

#define SVD_LIMIT 40000

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TLinearRegression)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TLinearRegression::TLinearRegression(URANIE::DataServer::TDataServer *tds,
        TString architecture, Option_t *option) :
        TModeler(tds, architecture, option), _dq2(-1.0), _dpress(-1.0)
{

    if (_blog)
    {
        cout << "construit TLinearRegression tds[" << _tds->GetName()
                << "] architecture[" << architecture << "] Option [" << option
                << "]" << endl;
    }

    TString opt = option;
    opt.ToLower();

    // _mParameterCov.ResizeTo(1,1); _mParameterCov.Zero();
}
TLinearRegression::TLinearRegression(URANIE::DataServer::TDataServer *tds,
        const char *varexpinput, const char *varexpoutput, Option_t * option) :
        TModeler(tds, varexpinput, varexpoutput, option), _dq2(-1.0), _dpress(-1.0)
{

    _sFileName = Form("linearmodel_%s", _sOutputAttributes.Data());
    _sFunctionName = Form("linearmodel_Fct_%s", _sOutputAttributes.Data());
    SetName(
            Form("Modeler[LinearRegression]Tds[%s]Predictor[%s]Target[%s]",
                    _tds->GetName(), varexpinput, _sOutputAttributes.Data()));
    SetTitle(
            Form(
                    "LinearRegression as modeler for TDS[%s] with Target is[%s] versus[%s]",
                    _tds->GetName(), _sOutputAttributes.Data(), varexpinput));

    // _mParameterCov.ResizeTo(1,1); _mParameterCov.Zero();
}

#ifdef  WITH_LIBXML2  
TLinearRegression::TLinearRegression(URANIE::DataServer::TDataServer *tds,
        int id,
        const char* pmmlfile,
        const char* LRname,
        Option_t * option) :
TModeler(tds, TString(",")),
_dq2(-1.0),
_dpress(-1.0)
{
    UNUSED(id); // ? UNUSED ?
    if ( _blog )
    cout << "*** construit TLinearRegression tds Name[" << _tds->GetName() << "] PMML[" << pmmlfile << "] Name[" << LRname << "]" << endl;
    SetNameTitle(Form("_modeler_%s", tds->GetName()), "Classe TLinearRegression");

    PMMLlib::PMMLlib * pmml = new PMMLlib::PMMLlib( string(pmmlfile) );
    std::string strMod(LRname);
    pmml->SetCurrentModel(strMod, PMMLlib::kLR);

    // Détermination de _sInputAttributes et _sOutputAttributes
    string architecture = pmml->ReadRegressionStructure();
    unsigned opos = architecture.find ("@");
    string varexpoutput = architecture.substr(opos+1);
    unsigned ipos = architecture.find (",");
    string varexpinput = architecture.substr(0,ipos);
    _sInputAttributes = TString(varexpinput);
    _sOutputAttributes = TString(varexpoutput);

    _nS = _tds->getNPatterns();

    // Gestion des options
    TString soption = TString(option);
    soption.ToLower();
    if (soption.Contains("nointercept"))
    _bIntercept = kFALSE;
    if (soption.Contains("nostoreyhat"))
    _bStoreYHat = kFALSE;

    if ( _listOfInputAttributes != NULL )
      delete _listOfInputAttributes;
    _listOfInputAttributes = NULL;	
    if ( _listOfOutputAttributes  != NULL)
      delete _listOfOutputAttributes;
    _listOfOutputAttributes=NULL;
    try
    {
        //  Get the list of Input attributes
        _listOfInputAttributes = _tds->getListOfAttributes(_sInputAttributes);
        _nX = _listOfInputAttributes->GetEntries();

        if( _blog )
        {
            cout << endl << " **  Input Attributes listing  _sInputAttributes[" << _sInputAttributes << "] _nX[" << _nX << "]" << endl;
            _listOfInputAttributes->Print();
        }

        //  Get the list of Output attributes
        _listOfOutputAttributes = _tds->getListOfAttributes(_sOutputAttributes);
        _nY = _listOfOutputAttributes->GetEntries();
        if( _blog )
        {
            cout << endl << " **  Output Attributes listing _sOutputAttributes[" << _sOutputAttributes << "] _nY[" << _nY << "]" << endl;
            _listOfOutputAttributes->Print();
        }
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	if ( _listOfInputAttributes != NULL )
	  delete _listOfInputAttributes;
	_listOfInputAttributes = NULL;	
	if ( _listOfOutputAttributes  != NULL)
	  delete _listOfOutputAttributes;
	_listOfOutputAttributes=NULL;
        throw ue;
    }
    // Set _parameterValues :
    _bIntercept = pmml->HasIntercept();
    int nNumPred = pmml->GetNumericPredictorNb();
    int nPredTrm = pmml->GetPredictorTermNb();
    Int_t nP = Int_t(nNumPred + nPredTrm);
    Int_t icrt = 0;
    // -> intercept
    if ( _bIntercept )
    {
        nP += 1;
        _parameterValues.ResizeTo(nP);
        _parameterValues[icrt] = pmml->GetRegressionTableIntercept();
        icrt += 1;
    }
    else
    _parameterValues.ResizeTo(nP);
    // -> NumericPredictor            
    for (int i=0; i<nNumPred; i++)
    {
        _parameterValues[icrt] = pmml->GetNumericPredictorCoefficient(i);
        icrt += 1;
    }
    // -> PredictorTerm
    for (int i=0; i<nPredTrm; i++)
    {
        _parameterValues[icrt] = pmml->GetPredictorTermCoefficient(i);
        icrt += 1;
    }
    delete pmml;
    // _mParameterCov.ResizeTo(1,1); _mParameterCov.Zero();
}
#endif   

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TLinearRegression::~TLinearRegression()
{
    if (_blog)
        cout << " ** Destroy  TLinearRegression" << endl;

    if (_blog)
        cout << " ** End Of Destroy  TLinearRegression" << endl;
}

///////////////////////////////////////////////////////////
//  computeIndexes 
///////////////////////////////////////////////////////////
void TLinearRegression::estimate(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
    {
        cout << endl << "**************************************" << endl;
        cout << " ** TLinearRegression::estimate" << endl;
    }

    try
    {
        if (_nY != 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TLinearRegression::estimate   ERROR\nThe number of output must be equal to 1.\nHere, we are %d output attributes with [%s]",
                            _nY, _sOutputAttributes.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);

    }

    // Calculate the mean/Standard deviation on each input and output  attributes
    _tds->computeStatistic(
            Form("%s:%s", _sInputAttributes.Data(), _sOutputAttributes.Data()));
    URANIE::DataServer::TAttribute *attY = _tds->getAttribute(
            _sOutputAttributes);
    Double_t dymean = attY->getMean();

    Int_t nP = _nX;
    if (_bIntercept)
        nP++;
    TMatrixD Aw(_nS, nP);
    TVectorD yw(_nS);
    yw = 1.0;

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputLeaves = new TList();
    TIter next(_listOfInputAttributes);
    if (_blog)
        cout << " ** input" << endl;
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
        if (_blog)
        {
            _tds->getTuple()->GetLeaf(att->GetName())->Print();
            att->printLog();
        }
    }

    // Case of the output (only one attribut) : Fill the vector Y
    TLeaf *leafY = _tds->getTuple()->GetLeaf(_sOutputAttributes.Data());
    if (_blog)
    {
        cout << " ** output" << endl;
        _tds->getTuple()->GetLeaf(_sOutputAttributes)->Print();
        _tds->getAttribute(_sOutputAttributes)->printLog();
    }

    // Fill the matrix and vector of data
    Int_t ivar;
    for (Int_t ipat = 0; ipat < _nS; ipat++)
    {
        _tds->getTuple()->GetEntry(ipat);
        if ((ipat < 5) && (_blog))
            _tds->getTuple()->Show();

        // Case of the input attributes : Fill the matrix Aw
        ivar = 0;
        if (_bIntercept)
            Aw(ipat, ivar++) = 1.0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
            Aw(ipat, ivar++) = leaf->GetValue(0);

        // Case of output attribute : Fill the vector yw
        yw(ipat) = leafY->GetValue(0);
    }
    delete listOfInputLeaves;

    if (_blog)
    {
        cout << " ** Matrix Aw(" << Aw.GetNrows() << "," << Aw.GetNcols() << ")"
                << endl;
        Aw.Print();
        cout << " ** Vector Yw(" << yw.GetNrows() << ")" << endl;
        yw.Print();
    }

    try
    {
       
        if (Aw.GetNcols() > Aw.GetNrows())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::estimate\n We are more parameters to identify than patterns in the data set");
      
        // Compute the linear regression    
        Bool_t ok = kFALSE;    
        _parameterValues.ResizeTo(nP);
        
        if(_nS < SVD_LIMIT )
        {
            TDecompSVD svd(Aw);
            _parameterValues = svd.Solve(yw, ok);
        }
        else
        {
            TDecompChol ch(TMatrixDSym(TMatrixDSym::kAtA,Aw));
            _parameterValues = ch.Solve(TMatrixD(TMatrixD::kTransposed,Aw)*yw,ok);
        }
       
        if (ok)
        {

            if (_blog)
            {
                cout << " ** Vector of parameters solution ("
                        << _parameterValues.GetNrows() << ")" << endl;
                _parameterValues.Print();
            }
            TVectorD yhat = Aw * _parameterValues;
            Double_t *tabValue = new Double_t[1];
            TString syhat = Form("%shat", _sOutputAttributes.Data());

            if (_bStoreYHat)
                _tds->addAttributeByNameTitle( syhat, Form("#hat{%s}", attY->GetTitle()));
            // \todo ( FGA, 2012-03-27 15:27:51): Use copy of the attribute with an another name. 
            //    URANIE::DataServer::TAttribute *yhatatt = _tds->getAttribute(_sOutputAttributes.Data())->Clone(syhat);
            //    yhatatt->SetTitle(Form("#hat{%s}", attY->GetTitle()));
            //      _tds->addAttribute(yhatatt);
            TBranch *tbr = NULL;
            if (_bStoreYHat)
                tbr = _tds->getTuple()->Branch(syhat, tabValue,
                        Form("%s/D", syhat.Data()));


            // Compute the R2, R2-adjusted, PRESS and Q2 criteria 
            Double_t dSumPRESS = 0.0;
            Double_t dSumSCT = 0.0;
            Double_t dSumSCM = 0.0;
            Double_t dSumRes = 0.0;
            Double_t dSumTot = 0.0;

            for (Int_t ipat = 0; ipat < _nS; ipat++)
            {
                tabValue[0] = yhat[ipat];
                if (_bStoreYHat)
                    tbr->Fill();

                dSumRes += (yhat(ipat) - yw(ipat)) * (yhat(ipat) - yw(ipat));
                dSumTot += (yw(ipat) - dymean)* (yw(ipat) - dymean);
                dSumSCT += (yw(ipat) - dymean) * (yw(ipat) - dymean);
                dSumSCM += (yhat(ipat) - dymean) * (yhat(ipat) - dymean);
            }
            
            
//             if( _nS < SVD_LIMIT )
//             {
//                 // compute the H matrix
//                 TMatrixD matH  = computeH(Aw);
//                 for (Int_t ipat = 0; ipat < _nS; ipat++)
//                     dSumPRESS += ((yw(ipat) - yhat(ipat)) / (1. - matH(ipat, ipat)))
//                         * ((yw(ipat) - yhat(ipat)) / (1. - matH(ipat, ipat)));
//             }
//             else
//             {
                // compute the Hii vector
                TVectorD vecH  = computeHii(Aw);
                for (Int_t ipat = 0; ipat < _nS; ipat++)
                    dSumPRESS += ((yw(ipat) - yhat(ipat)) / (1. - vecH(ipat)))
                        * ((yw(ipat) - yhat(ipat)) / (1. - vecH(ipat)));
                    
//             }
            
                _mParameterCov.ResizeTo(Aw.GetNcols(),Aw.GetNcols());
                _mParameterCov = computeCov(Aw);
            
         
            // Compute the R2 : coefficient of determination
            //          _dr2 = 1.0 - dSumRes / dSumTot;
            _dr2 = dSumSCM / dSumSCT;
            // cout << " _dr2 [" << 1.0 - dSumRes / dSumTot << "]" << endl;
            // cout << " _dr2 [" << 1.0 - dSumRes / dSumSCT << "]" << endl;

            // Compute the adjusted R2 : coefficient of determination adjusted
            _dr2a = 1.0 - (1.0 - _dr2) * Double_t(_nS - 1) / Double_t(_nS - nP);

            // Compute the PRESS
            _dpress = dSumPRESS;

            // Compute the Q2
            _dq2 = 1.0 - _dpress / dSumSCT;

            if (_blog)
            {
                cout << " ** Linear Regression OK" << endl;
                cout << " ** dSumRes[" << dSumRes << "] -- dSumTot[" << dSumTot
                        << "] -- dSumSCM[" << dSumSCM << "] -- dSumSCT["
                        << dSumSCT << "]" << endl;
                cout << " ** _dr2[" << _dr2 << "] _dr2a[" << _dr2a << "]"
                        << endl;
                cout << " ** _dpress[" << _dpress << "] _dq2[" << _dq2 << "]"
                        << endl;
            }

            // Save the data in a TFile
            if (_bStoreYHat)
                _tds->saveTuple("UPDATE");

            // Rest the branch addresses
            _tds->getTuple()->ResetBranchAddress(tbr);
            delete[] tabValue; tabValue = NULL;
        }
        else
        {
            if( _nS < SVD_LIMIT )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::estimate\n the SVD decomposition is down");
            else
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPolynomialRegression::estimate\n the Choleski decomposition is down");    
        }

    }

    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    if (_blog)
    {
        cout << " ** End of TLinearRegression::estimate" << endl;
        cout << "**************************************" << endl;
    }

}

// -----------------------------------------------------------------------
//  computeCov
// -----------------------------------------------------------------------
TMatrixD TLinearRegression::computeCov(TMatrixD X)
{
    if (_blog)
    {
        cout << " ** data original" << endl;
        X.Print();
    }
    Int_t npat = _tds->getNPatterns();

    TMatrixD mattXX(X, TMatrixD::kTransposeMult, X);
    if (_blog)
    {
        cout << " ** mat tX X" << endl;
        mattXX.Print(" mat tX X");
    }

    TMatrixD mattXXInv(TMatrixD::kInverted, mattXX);
    if (_blog)
    {
        cout << " ** mat tX X inv" << endl;
        mattXXInv.Print();
    }

    return mattXXInv;
}


// -----------------------------------------------------------------------
//  computeH
// -----------------------------------------------------------------------
TMatrixD TLinearRegression::computeH(TMatrixD X)
{
    if (_blog)
    {
        cout << " ** data original" << endl;
        X.Print();
    }
    Int_t npat = _tds->getNPatterns();

    TMatrixD mattXX(X, TMatrixD::kTransposeMult, X);
    if (_blog)
    {
        cout << " ** mat tX X" << endl;
        mattXX.Print(" mat tX X");
    }

    TMatrixD mattXXInv(TMatrixD::kInverted, mattXX);
    if (_blog)
    {
        cout << " ** mat tX X inv" << endl;
        mattXXInv.Print();
    }

    TMatrixD matXtXXInv(X, TMatrixD::kMult, mattXXInv);
    if (_blog)
    {
        cout << " ** mat X (tX X)inv" << endl;
        matXtXXInv.Print();
    }

    TMatrixD Xt(TMatrixD::kTransposed, X);
    if (_blog)
    {
        cout << " ** mat Xt" << endl;
        Xt.Print();
    }

    TMatrixD matH(matXtXXInv, TMatrixD::kMult, Xt);   
    if (_blog)
    {
        cout << " ** matH" << endl;
        matH.Print();
    }

    if (_blog)
    {
        for (Int_t i = 0; i < npat; i++)
            cout << " ** hii[" << matH(i, i) << "]" << endl;
    }
    return matH;
}



// -----------------------------------------------------------------------
//  computeHii
// -----------------------------------------------------------------------
TVectorD TLinearRegression::computeHii(TMatrixD X)
{
    
    if (_blog)
    {
        cout << " ** data original" << endl;
        X.Print();
    }
    Int_t npat = _tds->getNPatterns();
    TVectorD vecHii(npat);
    Int_t nP = _nX;
    if (_bIntercept)
        nP++;
    
    TMatrixD mattXX(X, TMatrixD::kTransposeMult, X);
    if (_blog)
    {
        cout << " ** mat tX X" << endl;
        mattXX.Print(" mat tX X");
    }

    TMatrixD mattXXInv(TMatrixD::kInverted, mattXX);
    if (_blog)
    {
        cout << " ** mat tX X inv" << endl;
        mattXXInv.Print();
    }
    
    for (Int_t i = 0; i < npat; i++)
    {
        TMatrixD Xi(nP,1);
        for (Int_t j = 0; j < nP; j++)
            Xi(j,0) = X(i,j);
        
        
        TMatrixD Xit(TMatrixD::kTransposed, Xi);
        if (_blog)
        {
            cout << " ** mat Xit" << endl;
            Xit.Print();
        }
        
        TMatrixD matXitXXInv(Xit, TMatrixD::kMult, mattXXInv);
        if (_blog)
        {
            cout << " ** mat Xi (tX X)inv" << endl;
            matXitXXInv.Print();
        }

        
        
        TMatrixD matHii(matXitXXInv, TMatrixD::kMult, Xi);   
        if (_blog)
        {
            cout << " ** matHii" << endl;
            matHii.Print();
        }
        
        vecHii(i) = matHii(0,0);
    
    }

    if (_blog)
    {
        for (Int_t i = 0; i < npat; i++)
            cout << " ** hii[" << vecHii(i) << "]" << endl;
    }
    
    return vecHii;

}

// -----------------------------------------------------------------------
//  Export Functions
// -----------------------------------------------------------------------
void TLinearRegression::exportModelCplusplus(std::ofstream * sourcefile) const
{
    // Model Quality 
    *sourcefile << "  //" << endl;
    *sourcefile << "  //     QUALITY  : " << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  //              R2  = " << getR2() << endl;
    *sourcefile << "  //              R2a = " << getR2Adjusted() << endl;
    *sourcefile << "  //              Q2  = " << getQ2() << endl;
    *sourcefile << "  //" << endl;
    *sourcefile << "  ////////////////////////////// " << endl << endl;

    URANIE::DataServer::TAttribute *att;
    Int_t ivarinc = 0;
    if (_bIntercept)
    {
        *sourcefile << "  // Intercept" << endl;
        *sourcefile << "  double y = " << _parameterValues[0] << ";" << endl;
        ivarinc = 1;
    }
    else
    {
        *sourcefile << "  // No Intercept" << endl;
        *sourcefile << "  double y = 0.0 ;" << endl;
    }
    *sourcefile << endl;
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                ivar);
        *sourcefile << "  // Attribute : " << att->GetName() << endl;
        *sourcefile << "  y += param[" << ivar << "]*"
                << _parameterValues[ivar + ivarinc] << ";" << endl << endl;
    }
    *sourcefile << "  // Return the value" << endl;
    *sourcefile << "  res[0] = y;" << endl;
}

void TLinearRegression::exportModelFortran(std::ofstream * sourcefile) const
{
    // Model Quality 
    *sourcefile << "C --- " << endl;
    *sourcefile << "C ---      QUALITY  : " << endl;
    *sourcefile << "C --- " << endl;
    *sourcefile << "C ---               R2  = " << getR2() << endl;
    *sourcefile << "C ---               R2a = " << getR2Adjusted() << endl;
    *sourcefile << "C ---               Q2  = " << getQ2() << endl;
    *sourcefile << "C --- " << endl;
    *sourcefile << "C --- //////////////////////////// " << endl << endl;

    URANIE::DataServer::TAttribute *att;
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                ivar);
        *sourcefile << "      DOUBLE PRECISION " << att->GetName() << endl;
    }
    *sourcefile << "      DOUBLE PRECISION " << _sOutputAttributes << endl;
    *sourcefile << endl;
    // Case of Intercept
    Int_t ivarinc = 0;
    if (_bIntercept)
    {
        *sourcefile << "C --- Intercept" << endl;
        *sourcefile << "      " << _sOutputAttributes << " = "
                << _parameterValues[0] << endl;
        ivarinc = 1;
    }
    else
    {
        *sourcefile << "C --- No Intercept" << endl;
        *sourcefile << "      " << _sOutputAttributes << " = 0.0 " << endl;
    }
    *sourcefile << endl;
    for (Int_t ivar = 0; ivar < _nX; ivar++)
    {
        att = (URANIE::DataServer::TAttribute *) _listOfInputAttributes->At(
                ivar);
        *sourcefile << "C --- Attribute : " << att->GetName() << endl;
        *sourcefile << "      " << _sOutputAttributes << " = "
                << _sOutputAttributes << " + " << att->GetName() << " * "
                << _parameterValues[ivar + ivarinc] << endl << endl;
    }
}

void TLinearRegression::exportModelPMML(const char* file, const char* name,
        Option_t *option) const
{

    TString soption = TString(option);  // NEW / UPDATE / BACKUP
    soption.ToLower();
    enum
    {
        klocBACKUP, klocUPDATE, klocNEW, klocADD
    };
    Int_t nOption = klocBACKUP;
    if (soption.Contains("update"))
        nOption = klocUPDATE;
    if (soption.Contains("new"))
        nOption = klocNEW;
//     if (soption.Contains("add"))
//         nOption = ADD;

#ifdef WITH_LIBXML2
    PMMLlib::PMMLlib *pmml = NULL;
    URANIE::DataServer::TAttribute *att;
    Bool_t bGoodOption = kFALSE;
    while(bGoodOption==kFALSE)
    {
        switch(nOption)
        {
            case klocNEW:
                pmml = new PMMLlib::PMMLlib(true);
                // Header node
                pmml->SetHeader("myCopyright", "Text Description", "Uranie", "3.3.2", "Compilation date : 2000-01-01");
                // DataDictionnary node
                for(Int_t j=0; j<_nX; j++)
                {
                    att = (URANIE::DataServer::TAttribute *)_listOfInputAttributes->At(j);
                    URANIE::DataServer::TAttributeFormula *attformula = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(att);
                    if ( attformula == NULL )
                    {
                        pmml->AddDataField(att->GetName(), att->GetTitle(), "continuous", "double", "ClosedClosed", att->getMinimum(), att->getMaximum());
                    }
                }
                att = _tds->getAttribute(_sOutputAttributes);
                pmml->AddDataField(att->GetName(), att->GetTitle(), "continuous", "double", "ClosedClosed", att->getMinimum(), att->getMaximum());
                bGoodOption = kTRUE;

            break;
            // end of "case NEW"

            case klocUPDATE:
                try
                {
                    pmml = new PMMLlib::PMMLlib(file,true);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New PMML file created"<<endl;
                    nOption = klocNEW;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                try
                {
                    pmml->SetCurrentModel(name, PMMLlib::kLR);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New model created"<<endl;
                    nOption = klocADD;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                pmml->UnlinkNode();
                bGoodOption = kTRUE;

            break;
            // end of "case UPDATE"

            case klocBACKUP:
                
                try
                {
                    pmml = new PMMLlib::PMMLlib(file,true);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New PMML file created"<<endl;
                    nOption = klocNEW;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                try
                {
                    pmml->SetCurrentModel(name, PMMLlib::kLR);
                }
                catch (std::string msg)
                {
                    std::cerr <<"\n New model created"<<endl;
                    nOption = klocADD;
                    if(pmml)
                        delete pmml;
                    continue;
                }
                pmml->BackupNode();
                bGoodOption = kTRUE;

            break;

            case klocADD:
                pmml = new PMMLlib::PMMLlib(file, true);
                bGoodOption = kTRUE;

            break;

            break;

        } // end of "switch(nOption)"
    }

    // Linear regression
    pmml->AddRegressionModel(name,PMMLlib::kREGRESSION, _tds->getAttribute(_sOutputAttributes)->GetName());
    // Add mining schema
    for(Int_t j=0; j<_nX; j++)
    {
        att = (URANIE::DataServer::TAttribute *)_listOfInputAttributes->At(j);
        URANIE::DataServer::TAttributeFormula *attformula = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(att);
        if ( attformula == NULL )
        {
            // Set MiningSchema
            pmml->AddMiningSchema(att->GetName(), "active");
        }
    }
    att = _tds->getAttribute(_sOutputAttributes);
    pmml->AddMiningSchema(att->GetName(), "predicted");

    // Regression table
    int icrt=0;
    if( _bIntercept )
    {
        cout << " icrt " << icrt << " _parameterValues.size() " << _parameterValues.GetNrows() << endl;

        pmml->AddRegressionTable(_parameterValues[icrt]);
        icrt++;
    }
    else
    {
        pmml->AddRegressionTable();
    }

    std::string neuronName = "";
    for(Int_t j=0; j<_nX; j++)
    {
        neuronName = getInputName(j);
        att = _tds->getAttribute(TString(neuronName));
        URANIE::DataServer::TAttributeFormula *attformula = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(att);
        if ( attformula == NULL )
        {
            pmml->AddNumericPredictor(neuronName, 1, _parameterValues[icrt]);
            icrt++;
        }
        else
        {
            std::vector<std::string> fieldRefVect;
            TTreeFormula *ttf = new TTreeFormula("aat", attformula->getFormula(), _tds->getTuple());
            for(int i=0; ttf->GetLeaf(i); i++)
            {
                fieldRefVect.push_back( ttf->GetLeaf(i)->GetName() );
            }
            pmml->AddPredictorTerm(_parameterValues[icrt], fieldRefVect);
            icrt++;
        }
    }

    switch(nOption)
    {
        case klocNEW:
        pmml->Write(file);
        break;
        // end of "case NEW"
        default :
        pmml->Write();
        break;// end of "case UPDATE", "BACKUP", "ADD"
    } // end of "switch(nOption)"
    
    //Clean
    delete pmml;
#else
    UNUSED(file,name);
#endif
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TLinearRegression::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TLinearRegression::printLog[" << option << "]" << endl;
    TModeler::printLog();
    cout << " ** q2[" << _dq2 << "]" << endl;
    cout << " ** PRESS[" << _dpress << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

