//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLinearRegression.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu jun 5 10:35:25 CEST 2008
 
 \brief Interface de la classe URANIE::Modeler::TLinearRegression
 
 */
/**
 
 * @class URANIE::Modeler::TLinearRegression
 *
 * @brief Description of the class TLinearRegression.
 *
 * To write by the developper.
 *
 */
#ifndef TLINEARREGRESSION_H
#define TLINEARREGRESSION_H
#include "TModeler.h"

#include <iostream>

namespace URANIE
{
namespace Modeler
{
class TLinearRegression: public TModeler
{

private:
    TMatrixD computeH(TMatrixD A);
    TMatrixD computeCov(TMatrixD A);
public:
    TVectorD computeHii(TMatrixD A);
public:
    Double_t _dr2a;                             ///< The Adjusted R2 quality
    Double_t _dq2;                              ///< The Q2 quality
    Double_t _dpress;                              ///< The PRESS quality
    TMatrixD _mParameterCov;
    
    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with the TDataServer
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
     \param 
     */
    TLinearRegression(URANIE::DataServer::TDataServer *tds,
            TString architecture, Option_t *option = "");
    //! Default constructor with input and output attributes
    TLinearRegression(URANIE::DataServer::TDataServer *tds,
            const char *varexpinput, const char *varexpoutput,
            Option_t * option = "");
    //! Constructor with a dataserver and a PMML File
#ifdef  WITH_LIBXML2       
    TLinearRegression (URANIE::DataServer::TDataServer *tds,
            int id,
            const char* pmmlfile,
            const char* LRname,
            Option_t * option="");
#endif                         
    //! Default destructor
    virtual ~TLinearRegression();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //! Return the Adjusted R2 statistic
    Double_t getR2Adjusted() const
    {
        return _dr2a;
    }
    //! Return the Press statistic
    Double_t getPress() const
    {
        return _dpress;
    }
    //! Return the Q2 statistic
    Double_t getQ2() const
    {
        return _dq2;
    }
    //! Return the Q2 statistic
    TMatrixD getParametersCovariance() const
    {
        return _mParameterCov;
    }
    //@}
    //---------------------------------------------
    /** @name Estimate
     **/
    //@{
    void estimate(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const;

    //! Export the model in a PMML file
    void exportModelPMML(const char* file = "", const char* name = "",
            Option_t *option = "") const;
            
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TLinearRegression::exportModelPython(ofstream * sourcefile) not yet implemented"
                << endl;
    };        
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TLinearRegression, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
