//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMaternICorrFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TMaternICorrFunction

 */
#include <iostream>

//// ROOT
#include "TMath.h"

//// URANIE
#include "TMaternICorrFunction.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TMaternICorrFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TMaternICorrFunction::TMaternICorrFunction(Int_t nbDim, Double_t* nu, Double_t* corrLength)
{
    nbDimensions  = nbDim;
    nbParameters  = nbDimensions * 2;
    nbCorrLengths = nbDimensions;
    parameters    = new Double_t[nbParameters];
    _lNu          = new Double_t[nbDimensions];
    _lCorrLengths = new Double_t[nbCorrLengths];

    Int_t offSet = nbParameters - nbCorrLengths;
    
    if(nu)
    {
        for (Int_t i = 0; i < offSet; i++)
        {
            parameters[i] = nu[i];
            _lNu[i] = nu[i];
        }
    }
    else
    {
        for (Int_t i = 0; i < offSet; i++)
        {
            parameters[i] = 1.5;
            _lNu[i] = 1.5;
        }
    }    
    
    if(corrLength)
    {
        for (Int_t i = offSet; i < nbParameters; i++)
        {
            parameters[i] = corrLength[i - offSet];
            _lCorrLengths[i - offSet] = corrLength[i - offSet];
        }
    }
    else
    {
        for (Int_t i = offSet; i < nbParameters; i++)
        {
            parameters[i] = 0.1;
            _lCorrLengths[i - offSet] = 0.1;
        }
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TMaternICorrFunction::~TMaternICorrFunction()
{
    delete[] parameters;
    parameters = NULL;
    delete[] _lCorrLengths;
    _lCorrLengths = NULL;
    delete[] _lNu;
    _lNu = NULL;
    if (_blog)
        cout << "detruit TMaternICorrFunction" << endl;
}

// -----------------------------------------------------------------------
//  Set methods
// -----------------------------------------------------------------------

void TMaternICorrFunction::setParameters(Double_t* newParams)
{
    Int_t offSet = nbParameters - nbCorrLengths;
    for (Int_t i = 0; i < offSet; i++)
    {
        parameters[i] = newParams[i];
        _lNu[i] = newParams[i];
    }
    for (Int_t i = offSet; i < nbParameters; i++)
    {
        parameters[i] = newParams[i];
        _lCorrLengths[i - offSet] = newParams[i];
    }
}

void TMaternICorrFunction::setParameterNu(Double_t* nu)
{
    for (Int_t i = 0; i < nbParameters / 2; i++)
    {
        parameters[i] = nu[i];
        _lNu[i] = nu[i];
    }
}

void TMaternICorrFunction::setCorrLengths(Double_t *corrLength)
{
    Int_t offSet = nbParameters / 2;
    for (Int_t i = offSet; i < nbParameters; i++)
    {
        parameters[i] = corrLength[i - offSet];
        _lCorrLengths[i - offSet] = corrLength[i - offSet];
    }
}

// -----------------------------------------------------------------------
// Clone
// -----------------------------------------------------------------------

TCorrelationFunction* TMaternICorrFunction::clone()
{
    TMaternICorrFunction* cfClone = new TMaternICorrFunction(nbDimensions, _lNu, _lCorrLengths);
    return cfClone;
}

// -----------------------------------------------------------------------
// Evaluation
// -----------------------------------------------------------------------

void TMaternICorrFunction::eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2, Double_t* C)
{
    if (_blog)
        std::cout << "** TMaternICorrFunction::eval **: nu[0] = " << _lNu[0]
                << "; l[0] = " << _lCorrLengths[0] << std::endl;

    Int_t errorCode = gpCorMaternI(x1, x2, n1, n2, nbDimensions, _lCorrLengths, _lNu, C);

    try
    {
        if (errorCode == 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMaternICorrFunction::eval: bessik function failed to converge.");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TMaternICorrFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TMaternICorrFunction::printLog[" << option << "]" << endl;
    cout << " ** Number of parameters = [" << nbParameters << "]" << endl;
    for (Int_t i = 0; i < nbParameters / 2; i++)
        cout << " ** regularity coef " << i << " = [" << parameters[i] << "]"
                << endl;
    for (Int_t i = nbParameters / 2; i < nbParameters; i++)
        cout << " ** correlation length " << i << " = [" << parameters[i] << "]"
                << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

