//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialMultiDimFit.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date  2012-11-28 15:14:38
 
 \brief Interface de la classe URANIE::Modeler::TPolynomialMultiDimFit
 
 */
/**
 
 * @class URANIE::Modeler::TPolynomialMultiDimFit
 *
 * @brief Description of the class TPolynomialMultiDimFit.
 *
 * To write by the developper.
 It is a Uranie Wrapper to the ROOT class \em TMultiDimFit

 \see $ROOTSYS/tutorials/fit/multidimfit.C
 *
 */

#ifndef TPOLYNOMIALMULTIDIMFIT_H
#define TPOLYNOMIALMULTIDIMFIT_H

//ROOT
#include "TMultiDimFit.h"

// Uranie
#include "TModeler.h"
#include "TDataServer.h"
#include "TLinearRegression.h"

namespace URANIE
{
namespace Modeler
{
class TPolynomialMultiDimFit: public TModeler
{
    // Attributs
protected:
private:
    TMultiDimFit *_MultiDimFit;      ///< The object TMultiDimFit
public:

    // Operations
private:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with the TDataServer and architecture
    /*!
     If you don't want a \em intercept for the linear regression, put \b "nointercept" in the option argument.
     
     \code
     TPolynomialMultiDimFit * tpr = new TPolynomialMultiDimFit(tds, TString("i1:i2:i3,out"), "nointercept");
     tpr->setCriteria(0.977);
     tpr->estimate("Q2");
     \endcode
     
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param architecture (TString)[""] The list of input and output attributes to pass to the function. The names of attributes are separated by the character \em ":" and the type (\em Input or \em Output) are separated by the character \em "," 
     \param option (Option_t *)[""] The option to pass
     */
    TPolynomialMultiDimFit(URANIE::DataServer::TDataServer *tds,
            TString architecture, TMultiDimFit::EMDFPolyType type,
            Option_t *option = "");
    //! Default constructor with the TDataServer and input and output attributes
    /*!
     If you don't want a \em intercept for the linear regression, put \b "nointercept" in the option argument.

     \code
     TPolynomialMultiDimFit * tpr = new TPolynomialMultiDimFit(tds, "i1:i2:i3", "out", "nointercept");
     tpr->setCriteria(0.977);
     tpr->estimate("Q2");
     \endcode

     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param varexpinput (const char*) The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)  The list of output attributes separated by the caracter ":"
     \param option (Option_t *)[""] The option to pass
     */
    TPolynomialMultiDimFit(URANIE::DataServer::TDataServer *tds,
            const char *varexpinput, const char *varexpoutput,
            TMultiDimFit::EMDFPolyType type, Option_t * option = "");
    //! Default destructor
    virtual ~TPolynomialMultiDimFit();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    void setMinAngle(Double_t dangle = 1);
    void setMaxAngle(Double_t dangle = 0);
    void setMaxFunctions(Int_t n);
    void setMaxStudy(Int_t n);
    void setMaxTerms(Int_t n);
    void setMaxPowers(const Int_t *powers);
    void setMinRelativeError(Double_t derror);
    void setPowerLimit(Double_t limit = 1e-3);
    void setPowers(const Int_t *powers, Int_t terms);

    //! Get the TMultiDimFit ROOT object
    /*!
     \return The TMultiDimFit Root Object
     */
    TMultiDimFit *getMultiDimFit()
    {
        return _MultiDimFit;
    }
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{
    void estimate(Option_t * option = "");
    //! Fill the data in a TMatrix and a TVector
    /*!
     \todo To deplace in the TModeler Class because it is the same for TLinearRegression.
     */
    void fillData(TMatrixD &Aw, TVectorD &yw);

    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in an external file with a specified langage
    /*!
     It will be the main interface function to the export.
     \param lang (const char *) the langage to export (C|Fortran)
     \param file (const char *) the file to export the modeler. If empty, use the name of the model, plus the extension of the langage.
     \param name (const char *) the name of the function in the export file. If empty, use the name of the model
     \param option (Option_t *)[""] The option to pass
     */
//     void exportFunction(const char* lang, const char* file = "",
//             const char* name = "", Option_t *option = "");
    //! Export the model in C++ langage in a file
    /*!
     \warning Nothing to do; It is just to define this else it is an abstract classe.
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Nothing to do; It is just to define this else it is an abstract classe.
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const;
    
    //! Export the model in a PMML file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPMML(const char* file = "", 
                         const char* name = "",Option_t *option = "") const
    {
        TString soption = option;
        soption.ToLower();
	UNUSED(file,name);
        std::cerr
                << "   --- Method TPolynomialMultiDimFit::exportModelPMML(std::ofstream * sourcefile) not yet implemented"
                << endl;
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,        
          "TPolynomialMultiDimFit::exportFunction\n The PMML format is not yet implemented.");
        
    };
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TPolynomialMultiDimFit::exportModelPython(std::ofstream * sourcefile) not yet implemented"
                << endl;
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,        
          "TPolynomialMultiDimFit::exportFunction\n The Python format is not yet implemented.");
    };
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Modeler::TPolynomialMultiDimFit, ID_MODELER)
    //Classe de
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
