//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   UMLP.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Sep  8 16:04:53 2005
 
 \brief  description of the class UMLP
 
 
 */

#ifndef UMLP_H_INCLUDED 
#define UMLP_H_INCLUDED 

#include "newmatap.h"
using namespace NEWMAT;
#include "time.h"
#include <string>

//URANIE
#include "TDataServer.h"
#include "USynapses.h"

 

// class TXMLDataSpecification;

namespace URANIE
{
namespace Modeler
{
class UMLP
{
public:
    enum EProblem {
        kRegression, kClassification
    };

    enum EActivation
    {
        kTANH, kLOGISTIC
    };
private:
    int _input;
    int _hidden;
    string _sHidden;
    int _output;
    int _nbweight;
    int _nbneurons;
    double _dsigmahat;
    double _dalphaIC;
    double _dStudentQuantile;
    time_t _seed;
    EActivation _nActivationFunction;    ///! The activation function
    EProblem _nProlemType;              ///! The type of problem (Regression, Classification)
    TList *_lstOfSynapses;          ///< the List of Synapses

    /**!
     * Column vector of weights. This is the row-major order representation (line concatenation) of a
     * (K+1)*(M+2) matrix. The first column represent the weights of the links between the (K+1) hidden
     * neurons towards the unique output. The following columns (j) represent the weights of the links between
     * the (j-1)-th input neuron (the first input neuron being a dummy one for biais) and the hidden layer.
     */

    ColumnVector _weights;
    ColumnVector _deriveweights;
    ColumnVector _derivepartialweights;

    ColumnVector _act;
    ColumnVector _layer;
    ColumnVector _potentiel;
    ColumnVector _derive;
    ColumnVector _delta;
    ColumnVector _deltapartial;
    ColumnVector _minValues;
    ColumnVector _maxValues;
    Matrix _tZZinv;

    string _comment; //< Comment of the ANN
    string _title;   //< Title of the ANN
//   TXMLDataSpecification *_tds;
    bool _tds_is_new;
    URANIE::DataServer::TDataServer * _tds;   ///! The TDS
    bool _bLog;

//   Real * ss;

    void exportFunctionFortran(const char* file, const char* name,
            int nIC = 0) const;
    void exportFunctionCplusplus(const char* file, const char* name,
            int nIC = 0) const;

public:
    UMLP(int input, int hidden, int output=1);
    UMLP(int input, const char *shidden, int output=1);
    UMLP(const char* file);
    ~UMLP();
    void init();

    void initWeights(time_t seed = 0);

    void setWeights(ColumnVector weights);
    ColumnVector getWeights()
    {
        return _weights;
    }

    Int_t getNWeights() { return _nbweight; }
    Int_t getNNeurons() { return _nbneurons; }

    double getWeights(int i)
    {
        return _weights(i);
    }
    void evaluate(double *in);
    void evaluateInput(double *in);
    void setMinValue(int ind, double val)
    {
        _minValues(ind) = val;
    }
    void setMaxValue(int ind, double val)
    {
        _maxValues(ind) = val;
    }
    void setMinMaxValue(int ind, double min, double max)
    {
        _minValues(ind) = min;
        _maxValues(ind) = max;
    }

    void setSigmaHat(double val)
    {
        _dsigmahat = val;
    }
    void setAlphaIC(double val)
    {
        _dalphaIC = val;
    }
    double getSigmaHat()
    {
        return _dsigmahat;
    }
    double getAlphaIC()
    {
        return _dalphaIC;
    }

    void setMatrixtZZinv(Matrix & mattZZinv);

    void setStudentQuantile(double val)
    {
        _dStudentQuantile = val;
    }

    void setActivationFunction(EActivation nFunction);

    EActivation getActivationFunction()
    {
      return _nActivationFunction;
    }

    void setProblem( EProblem type);

    EProblem getProblem() {
        return _nProlemType;
    }

    void setComment(string str)
    {
        _comment = str;
    }
    void setTitle(string str)
    {
        _title = str;
    }

    void setTDS(URANIE::DataServer::TDataServer *tds)
    {
      if(_tds!=NULL && _tds_is_new)
	{
	  delete _tds;
	  _tds_is_new=false;
	}
      _tds = tds;
    }
    URANIE::DataServer::TDataServer *getTDS()
    {
        return _tds;
    }

    double getEcart(double *in);
    double getWeightsDerivate(int i);
    double getPartialWeightsDerivate(int i);
    void evaluateWeightsJacobian(double *in);
    void evaluateWeightsJacobian();

    void dumpWeights(const char* file);
    void loadWeights(const char* file);
    void exportWeights(const char* file);
    void exportFunction(const char* lang, const char* file, const char* name,
            int nIC = 0);


    void setLog(bool blog = kTRUE) { _bLog = blog;}
    void unsetLog() { _bLog = kFALSE;}
    void printLog();
};

} // Fin du namespace Modeler
} // Fin du namespace URANIE

#endif // UMLP_H_INCLUDED
