/////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   USynapses.cxx
 \b Created : Fri Dec  1 23:35:21 2017
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \brief Implementation file for the class URANIE::Modeler::USynapses

 \par Current version
 $Date$
 $Revision$
 */

// ROOT
#include "TMath.h"

//URANIE
#include "USynapses.h"

ClassImp (URANIE::Modeler::USynapses)

namespace URANIE
{
  namespace Modeler
  {
    
    USynapses::USynapses() :
      TNamed("_uranie_USynapses_Null_", "NULL USynapses"),
      _nInputNeurons(1), _nOutputNeurons(1),
      _nRank(0),
      _nFirstIndexOfWeights(0),
      _nOuputActivationFunction(kUNKNOWN),
      _bLog(kFALSE)
    {
      init();
    }
    
    USynapses::USynapses(Int_t ni, Int_t no, EActivationFunction typeOfFunction, Int_t nrank, Int_t nFirstIndex) :
      TNamed(Form("_uranie_USynapses_%d_", nrank), Form("USynapses of rank %d",nrank)),
      _nInputNeurons(ni), _nOutputNeurons(no), _nRank(nrank), _nFirstIndexOfWeights(nFirstIndex),
      _nOuputActivationFunction(typeOfFunction),
      _bLog(kFALSE)
    {
      init();
    }


    USynapses::~USynapses()
    {
      cout << " ** Destroy USynapses[" << GetName() << "]" << endl;
    }
    
    //---------------------------------------------
    void USynapses::init()
    {
        if ( _bLog )
            cout << " ** USynapses::init[" << GetName() << "]" << endl;
        try
        {
            if ( _nInputNeurons < 1)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("USynapses::init method\n USynapses (%s) must have positive input (%d)",
                                                                GetName(), _nInputNeurons));
            if ( _nOutputNeurons < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("USynapses::init method\n USynapses (%s) must have positive output (%d)",
                                                                GetName(), _nOutputNeurons));

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        // Compute the last index of weights
            _nLastIndexOfWeights = _nFirstIndexOfWeights + ( _nInputNeurons +1 ) * _nOutputNeurons -1;

        if ( _bLog )
            cout << " ** End Of USynapses::init[" << GetName() << "]" << endl;
    }

    //---------------------------------------------
    Int_t USynapses::getBiasIndex(Int_t no)
    {
        try
          {
            if ( no > _nOutputNeurons)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getBiasIndex method\n USynapses (%s) The output index (%d) must be less than NumberOf Output neurons (%d)",
                                                              GetName(), no, _nOutputNeurons));
            if ( no < 1)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getBiasIndex method\n USynapses (%s) The output index (%d) must have positive",
                                                              GetName(), no));

          } catch (URANIE::Exceptions::UErrorExceptions& ue)
          {
            ue.printMessage();
            throw ue;
          }

        return _nFirstIndexOfWeights + ( no - 1 ) * ( _nInputNeurons +1 );

    }

    Int_t USynapses::getWeightIndex(Int_t ni, Int_t no)
    {
        Int_t nIndexOfBias = 0;
        try
          {
            if ( ni > _nInputNeurons)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getWeightIndex method\n USynapses (%s) The input index (%d) must be less than NumberOf Input neurons (%d)",
                                                              GetName(), ni, _nInputNeurons));
            if ( ni < 1)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getWeightIndex method\n USynapses (%s) The input index (%d) must have positive",
                                                              GetName(), ni));
            if ( no > _nOutputNeurons)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getWeightIndex method\n USynapses (%s) The output index (%d) must be less than NumberOf Output neurons (%d)",
                                                              GetName(), no, _nOutputNeurons));
            if ( no < 1)
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("USynapses::getWeightIndex method\n USynapses (%s) The output index (%d) must have positive",
                                                              GetName(), no));
            nIndexOfBias = getBiasIndex(no);

          } catch (URANIE::Exceptions::UErrorExceptions& ue)
          {
            ue.printMessage();
            throw ue;
          }
        return nIndexOfBias+ni;

        //return _nFirstIndexOfWeights + ( no - 1 ) * ( _nInputNeurons +1 ) + ni;

    }

    //---------------------------------------------
    void USynapses::getActivityDerivate(Double_t pot, Double_t &dact, Double_t &dderivate) {
        switch (_nOuputActivationFunction)
        {
        case kTANH:
        {
            dact = TMath::TanH(pot);
            dderivate =  1.0 - dact * dact;
            break;
        }
        case kLOGISTIC:
        {
            dact = 1.0 / ( 1.0 + exp( -1.0 * pot ));
            dderivate = dact * ( 1.0 - dact);
            break;
        }
        case kIDENTITY:
        {
            dact = pot;
            dderivate = 1.0;
            break;
        }
        case kUNKNOWN:
        {
            exit(0);
            break;
        }
        }
    }


    //---------------------------------------------
    void USynapses::printLog()
    {
      cout << " **** Log UMLP[" << GetName() << "] [" << GetTitle() << "]" << endl; 
      cout << "   ** _nInputNeurons[" << _nInputNeurons << "]"<< endl; 
      cout << "   ** _nOutputNeurons[" << _nOutputNeurons << "]"<< endl; 
      cout << "   ** _nRank[" << _nRank << "]"<< endl; 
      cout << "   ** _nFirstIndexOfWeights[" << _nFirstIndexOfWeights << "]"<< endl; 
      cout << "   ** _nLastIndexOfWeights[" << _nLastIndexOfWeights << "]"<< endl;
      cout << "   ** _nOuputActivationFunction[" << _nOuputActivationFunction << "]"<< endl;
      cout << "   ** _nOuputActivationFunction[";
      switch (_nOuputActivationFunction)
      {
      case kTANH:
      {
          cout << "kTANH";
          break;
      }
      case kLOGISTIC:
      {
          cout << "kLOGISTIC";
          break;
      }
      case kIDENTITY:
      {
          cout << "kIDENTITY";
          break;
      }
      case kUNKNOWN:
      {
          cout << "kUNKNOWN";
          break;
      }
      }
      cout << "]" << endl;


    }
    
    
  } // End Of namespace Modeler
} // End Of namespace URANIE
