//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/*!
 \file   anisp.hpp
 \brief  function of the method ANISP
 */

#ifndef anisp_HPP
#define anisp_HPP

#include <vector>
#include <cstdlib>
#include <stdio.h>
#include <stddef.h>
#include <fstream>
#include <iomanip>
#include <set> 
// #ifdef WIN32
// #include <time.h>
// #else
// #include <sys/time.h>
// #include <sys/resource.h> 
// #endif

#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <exception>

#include "polynome.hpp"
#include "SPC.hpp"
#include "util.hpp"
#include "fct_integ_adaptatif.hpp"

#ifdef WIN32
#include "libanisp_Export.h"
#else
#define libanisp_EXPORT
#endif

namespace Anisp
{
class anispException : public std::exception
{
    
};
    
    
int rho_NO(vector<int*> NO, vector<int*> npt, int i, int j);

//! function which constructs a set of  Polynomial Chaos
/*!
 \brief Initialize a first set of elements of the Polynomial Chaos basis which is the starting 
 point of the selection of an adapted polynomial basis.
 \param Q : pointer to a vector of pointer to int which contains the set of Polynomial Chaos
 \param S : pointer to a vector of int which contains the relevant interactions
 \param H : pointer to a vector of pointer to double
 \param P :pointer to a vector of pointer to double which contains the points, the weight of the realisations
 \param nld : pointer to an int which indicates the number of points in the cubature formula
 \param d : int value which indicates the stochastic dimension
 \param tol : double value which indicates the tolerance value
 \param Delta : double ** which contains the unidimensional quadrature difference formula for an uniform distribution
 \param Delta_n : double ** which contains the unidimensional quadrature difference formula for a normal distribution
 \param l : int value which indicates the maximum value authorized for an indice along every dimensions
 \param loi : string value which indicates the probability distribution of each variables
 */
void construction_Q(vector<int*>* Q, vector<int>* S, vector<double*>*H,
        vector<double*>* P, int* nld, int d, double tol, double** Delta,
        double** Delta_n, int l, string loi);

//void calcul_delta( double** Delta, int l);

//double f(double* x);
//double f2(double* x);

//! function of the method ANISP
/*!
 \brief Run the ANISP method for the approximation of a Polynomial Chaos expansion
 \param f : void * for the function approximated by the Polynomial Chaos expansion
 \param step_by_step : void * for the function which launches the graphical output in the TAnisp object
 \param d : int value  which indicates the stochastic dimension
 \param l : int value which indicates the maximum value authorized for an indice along every dimensions
 \param m : int value which indicate the number maximum of indice possible in the cubature formula (for memory reasons)
 \param tol : double value which indicates the tolerance value
 \param kmin : int value which indicate the minimum level of interaction to be explored by integration algorithm
 \param loi : string value which indicates the probability distribution of each variables
 \param max : 
 \param :
 \param :
 \param :
 \param :
 \param :
 */
libanisp_EXPORT void anisp(void (*f)(vector<double*>),
        void (*step_by_step)(double*, vector<int*>), int d, int l, int m,
        double tol, int kmin, string loi, int max, int nomax, bool rep,
        bool log, string ResultFile, string StochasticFile,
        string DataAnispFile, string ErrorIndicatorFile, string IndicesFile);

}

#endif

