//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#ifndef __NN_MLAYERPERCEPTRON_INCLUDE__
#define __NN_MLAYERPERCEPTRON_INCLUDE__

#include "utils.h"

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>

namespace NeuralNetworkLevenbergMarquardt
{
    typedef boost::numeric::ublas::matrix<Real, boost::numeric::ublas::column_major> Matrix; // Orientation if for cuBLAS compatibility
    typedef boost::numeric::ublas::matrix<Int, boost::numeric::ublas::column_major> MatrixInt; // Orientation if for cuBLAS compatibility
    typedef boost::numeric::ublas::vector<Real> Vector;
    typedef boost::numeric::ublas::vector<Int> VectorInt;
    typedef boost::numeric::ublas::matrix_row<Matrix> MatrixRow;
    typedef boost::numeric::ublas::matrix_row<MatrixInt> MatrixIntRow;

    enum NeuronType
    {
        kOff,           //!< No output: 0
        kLinear,        //!< Linear output (activation function: f: x->x)
        kSigmoid,       //!< Sigmoid activation function: f: x-> 1/(1+exp(-x))
        kTanh,          //!< Hyperbolic tangent activation function: f: x-> [exp(x)-exp(-x)]/[exp(x)+exp(-x)]
        kGauss          //!< Gaussian activation function: f: x-> exp(-x^2)
//        kExternal     //!< External activation function (NOT IMPLEMENTED)
    };

    class NeuronLayer
    {
        public:            
            explicit NeuronLayer(int m, int k, NeuronType t);
            /**! Size of the layer, in neurons **/
            inline size_t getNeuronNumber() const { return K; }

            void setInput( const Vector& );          //!< Feed an input vector to the layer

            /**!  Get the resulting input (including the bias) **/
            inline const Vector& getInput() const { return mInput; }

            /**! Get the resulting output, *once run externally* */
            inline const Vector& getOutput() const { return mOutput; }

            /**  Get the linear output (no function applied) */
            inline const Vector& getLinearOutput() const {return mPreOutput;}

            void run();                                     //!< Runs on the input (previously set with setInput())
            
            void initialize();

            /**! Get a reference to the Kx(M+1) weight matrix */
            inline Matrix& getWeights() { return mWeights; }
        private:
            const int M, K;          //!< Input, hidden
            const NeuronType T;      // Activation function

            Matrix mWeights;        //!< Kx(M+1) weight matrix
            Vector mInput;          //!< Input (size M+1: 1 for the bias, index M; M for the actual input: 0<=m<=M-1)
            Vector mPreOutput;      //!< Before the (non-linear) function is applied
            Vector mOutput;         //!< Output, after the function is applied
            bool mInitialized;      //!< True if the weights have been initialized
    };

    /**! Multilayer perceptron, with only one hidden layer. **/
    class TwoLayersPerceptron
    {
        public:
            static const Real LAMBDA_DEFAULT_FACTOR;        //!< Default Levenberg damping factor
            static const Real LAMBDA_INCREASE_FACTOR;       //!< Factor by which the Levenberg factor increases at each success
            static const Real LAMBDA_DECREASE_FACTOR;       //!< Factor by which the Levenberg factor decreases at each failure
            static const Real LAMBDA_MAX_VALUE;             //!< Maximum value for the Levenberg factor
            static const Real LAMBDA_MIN_VALUE;             //!< Minimum value for the Levenberg factor
            //static const size_t NUMBER_PARAMETERS;          //!< Number R of parameters in the network, including the biases
            static const size_t MAX_TRAINING_ITERATIONS;    //!< Maximum number of iterations (stopping condition)
            static const size_t MAX_NOCHANGE_ITERATIONS;    //!< Maximum number of iterations with no improvement (stopping condition)
            static const Real ERROR_STOPPING_THRESHOLD;     //!< Default stopping condition for the training (mean error < threshold)       
        public:
            /**! \brief Creates a two layer perceptron using the Levenberg-Marquardt algorithm.
            Two-layer perceptron:
                - hidden layer, with the activation function specified as template parameter T
                - output layer, with a linear activation function
            @param learningRate        initial Levenberg damping factor
            @param precisionThreshold    requested precision on mean squared error (stopping condition for training)
            @param maxIterations    maximum number of epochs (stopping condition for training)
            **/
            explicit TwoLayersPerceptron(int m, int k, int p, NeuronType t,
                                         const Real learningRate = LAMBDA_DEFAULT_FACTOR,
                                         const Real precisionThreshold = ERROR_STOPPING_THRESHOLD,
                                         const size_t maxIterations = MAX_TRAINING_ITERATIONS );

            ~TwoLayersPerceptron();
            
            /**! \brief Have we already trained the NN on a non-empty set? **/
            inline bool isTrained() const { return mHasBeenTrained; }
            /**! \brief Returns the time taken by the last training performed, or 0 if none has been done yet. **/
            //inline double getLastTrainingTime() const { return mTrainingTime; }
            /**! \brief Returns the final total (squared) error during the last training, or 0 if none has been done yet. **/
            //inline Real getLastTrainingError() const { return isTrained() ? mTotalError : static_cast<Real>(0.0); }

            /**! \brief Initializes the network (data and libraries).
            Is automatically called by the other functions, if forgotten. **/
            void initialize();
            /**! \brief Reinitializes the network. **/
            void reset();
            /**! \brief Cleans up the resources used by the network. **/
            void cleanup();
            
            /**! \brief Launches the training of the NN: return false if the training pool is empty.
            * Delegates to either the CPU or GPU version, depending on the CPU_PROCESSING_ONLY macro. The samples (teacher values)
            * are automatically normalized according to the type of activation function (e.g., in [0,1] for kSigmoid)
            @param nTrials Number of different training phases tried (best result is kept)
            @return false if something went wrong. **/
            bool train( const size_t nMixTrainSet = 1, const size_t nTrials = 1 );
            
            inline Real getLastTrainingLearningError() const { return mTotalErrorLearn/getLearningSampleCount(); }
            inline Real getLastTrainingTestingError() const { return mTotalErrorTest/getTestingSampleCount(); }
            /**! \brief Return all weights of the MLP. Uses the TANNModeler convention.
             @param w output vector of weights (must be already allocated).
             */
            void getWeights(double *w);
            
            /**! \brief Runs on the input, with the current weights (must have been already trained if used in production).
            @return The output value, normalized. **/
            Real run( const Vector& input );
            /**! \brief Returns the resulting output of the NN (TwoLayersPerceptron::run() must have been called before).
            @return The output value, normalized. **/
            const Real& getOutput() const;
            /**! \brief Returns the resulting output of the NN (TwoLayersPerceptron::run() must have been called before).
            @return The output value, restored to its real range. **/
            Real getDenormalizedOutput() const;
            
            /**! \brief Runs on all the input, with the current weights (must have been already trained if used in production).
            @param input NxM matrix of input.
            @return N-dimensional vector of the (normalized) output values.**/
            Vector runBatch( const Matrix& input );
            /**! \brief Denormalizes a value (e.g., one of those returned by runBatch()), based on the previous training data.
            Will raise an exception if the value nOutput is not in the range of normalized values.
            @param nOutput the normalized value
            @return the value, restored to its real range. **/
            Real denormalizeOutput( const Real nOutput ) const;
            
            /**! \brief Adds a (x,t) pair to the training pool (allows duplicate entries)
             * The training set is separated into two parts ultimately:
             *   - a learning part (getLearningSampleCount())
             *   - a testing part (getTestingSampleCount())
             * **/
            void addTrainingSample( const Vector& input, const Vector& teacherOutput );
            /**! \brief Returns the *actual* size of the training pool (how many samples have been fed to the NN until now) **/
            inline size_t getTrainingSampleCount() const { return mActualTrainingSetSize;}
            inline size_t getLearningSampleCount() const { return mTestSetStartIdx +1; }
            inline size_t getTestingSampleCount() const { return mActualTrainingSetSize - mTestSetStartIdx; }

            /**! \brief Sets the *expected* size of the training pool and the ratio between the training
             * set and the test set.
             * If too many samples are subsequently added, this count will be updated.
             * @return the number of items in the test database **/
            int setTrainingSampleCount( const size_t count,const float ratio = 0.8);
            /**! \brief Returns the minimum value amongst the training samples.
            Must be called *after* the training is done. **/
            Real getMaxTrainingSampleValue() const;
            /**! \brief Returns the maximum value amongst the training samples.
            Must be called *after* the training is done. **/
            Real getMinTrainingSampleValue() const;

        private:        
            const int M, K, P;      //!< Input, hidden, output
            const NeuronType T;     // Activation function
            const int NUMBER_PARAMETERS;  // Total number of weights, including bias weights.

//~            NeuronLayer<0,M,kLinear> mInputLayer;
            NeuronLayer mHiddenLayer;
            NeuronLayer mOutputLayer;
            
            Real mLambda;                   //!< Learning rate used in the Levenberg-Marquardt algorithm
            Real mPrecisionThreshold;       //!< Stopping condition for the training (average error < threshold)
            const size_t mMaxIterations;    //!< Stopping condition for the training (maximum number of epochs)
            /**!\brief Stopping condition for the training (maximum number of "useless" iterations).
            Note that it is in iterations and not epochs - that is, inside an epoch, we can have several iterations,
            during which we try different values of the damping factor lambda, until something improves or the
            maximum value for lambda is reached. Each of these trials is an iteration.
            The parameter mMaxNoChangeIter is currently arbitrarily set to max( mMaxIterations/20, 30 ), that is 5%
            of the maximum number of epochs (bounded below by 30)). **/
            const size_t mMaxNoChangeIter;

            bool mHasBeenTrained;           //!< Has the NN already trained on the pool?
            bool mInitialized;              //!< Has the NN already been initialized?
            bool mHasRunOnInput;            //!< Has the NN method run() already been used?
            bool mSamplesNormalized;        //!< Have the samples already been normalized? (handy when several training trials are made)
            
            /* Variables used in training */
            size_t mTrainingSetSize;        //!< How many samples are to be specified
            size_t mTestSetStartIdx;        //!< Index of the start of the test set in the members mInputValues and mTeacherValues
            size_t mActualTrainingSetSize;  //!< How many samples have already been specified
            Real mMaxSampleValue;           //!< Maximum value amongst the samples
            Real mMinSampleValue;           //!< Minimum value amongst the samples
            Matrix mInputValues;            //!< Training set: the input (row vectors)
            Vector mTeacherValues;          //!< Training set: the reference (teacher) output (N-dim vector)
            Real mTotalErrorLearn;          //!< Total squared error of the current epoch on the learning set
            Real mTotalErrorTest;           //!< Total squared error of the current epoch on the testing set
            Real mBestTotalErrorTest;
            Real mBestTotalErrorLearn;
            double mTrainingTime;           //!< Time took by the last training to complete (0 if none has been done yet)
         
            #ifdef CPU_PROCESSING_ONLY
              Vector mErrors;                   //!< N-dim vector: each row vector is the error for sample n, e_n = (y_n - t_n)
              Matrix mJacobian;                 //!< NxR matrix J=(dei/dwj) (R = NUMBER_PARAMETERS : weights and biases)
              Matrix mHessian;                  //!< RxR matrix H~tJJ (R = NUMBER_PARAMETERS : weights and biases)
              Vector mGradient;                 //!< R-dim vector g = tJ.e, where e = mErrors
              Matrix mHiddenLayerBackupWeights; //!< Backup of the hidden layer weights - in case we try a "bad" move
              Matrix mOutputLayerBackupWeights; //!< Backup of the output layer weights - in case we try a "bad" move
            #else // GPU
              Real* mDeviceErrors;              //!< N-dim vector: each row vector is the error for sample n, e_n = (y_n - t_n) (in device memory)
              Real* mDeviceInputValues;         //!< N_LEARNx(M+1) matrix: device copy of the input values with bias
              Real* mDeviceInputValues_restr;   //!< N_LEARNx(M+1)
              Real* mDeviceTeacherValues;       //!< NxP matrix: device copy of the teacher values
              Real* mDeviceOutputs;             //!< NxP matrix: network output (in device memory)
              Real* mDeviceIdentityRR;          //!< R-dimensional unit vector representing the identity matrix
              Real* mDeviceLMMatrix;            //!< RxR matrix, working copy of H~tJJ for trial and error
              Real* mDeviceJacobian;            //!< NxR matrix J=(dei/dwj) (in device memory)
              Real* mDeviceGradient;            //!< R-dim vector g = tJ.e, where e is the error vector (in device memory)
              Real* mDeviceHessian;             //!< RxR matrix H~tJJ (in device memory)
              Real* mDeviceWeightsHidden;       //!< Kx(M+1) matrix: device copy of the hidden layer weights
              Real* mDeviceWeightsOutput;       //!< Px(K+1) matrix: device copy of the output layer weights
              Real* mDeviceWeightsBackup;       //!< Backup of the hidden and output layer weights (in device memory)
              Real* mDeviceHiddenOutput;        //!< Nx(K+1) matrix: output of the hidden layer (plus bias) (in device memory)
              Real* mDeviceHiddenOutput_restr;  //!< N_LEARNx(K+1)
              Real* mDeviceWeightsBest;
            #endif
            
        private:
          #ifdef CPU_PROCESSING_ONLY
            /**! \brief Main training algorithm (Levenberg-Marquardt), CPU version.
            @param stopCondition will be set to a bit mask indicating which stopping condition was fulfilled.
            @param nbTotalIterations will be set to the number of iterations (epochs) performed
            @return true if the training was successful **/
            bool trainCPU( short& stopCondition, size_t& nbTotalIterations );
            /**! \brief Apply normalization to the training samples, so that the values lay in [a,b] (CPU version)
            The maximum and minimum values were stored in mMaxSampleValue (resp. mMinSampleValue), determined
            as the samples were added. Should only be called once all the samples are known. The actual interval
            might be [-1,1], [0,1] or the whole real line, depending on the chosen activation function. **/
            bool normalizeSamples();
            /**! \brief Computes the N-dimensional vector mErrors, component by component.
                \f$ \mathcal{E} = (\varepsilon_n)_{1\leq n \leq N} = (y_n - t_n)_{1\leq n \leq N} \f$ 
            This is the matrix whose nth row is the P-dimensional error vector for sample n.
            @param n index of the component to fill
            @return Square of the sample error **/
            Real computeErrorForSample( const size_t n );
            /**! \brief Computes the Jacobian matrix J=(dF/dwj), row by row.
            Uses backpropagation: must be called once the neural network has been fed the nth training
            samples, so that the corresponding row of the matrix mErrors is filled.
            @param n index of the row to fill **/
            void computeJacobian( const size_t n );
            /**! \brief Computes the Hessian (approximation) matrix H=tJJ, based on the Jacobian.
            Hence, computeJacobian() must already have been called on all the samples. Stores the
            result in mHessian. **/
            void computeHessian();
            /**! \brief Computes the gradient g=tJ.e, based on the Jacobian and the error vector.
            Hence, computeJacobian() and computeErrorForSample() must already have been called on
            all the samples. Stores the result in mGradient. **/
            void computeGradient();
            /**! \brief Computes the weight modification values Delta(w), for this epoch of the NN training  (CPU version).
                            \f$ \Delta(w) = ( J^T J + \lambda I_{R} )^{-1} J^T \varepsilon \f$ 
            where \f$\varepsilon\f$ is the N-dimensional error vector \f$e=(||\varepsilon_n||)\f$. The Hessian must
            have been already computed. **/
            const Vector computeDeltaWeights();
            /**! \brief Modifies the network weights by substracting the amount computed during the previous phases of the algorithm.
            @param delta The modification (R-dimensional vector) **/
            void modifyWeights( const Vector& delta );
            /**! \brief Restores the previous weights, in case the epoch wasn't successful. **/
            void restoreWeights();
          #else // GPU
            /**! \brief Main training algorithm (Levenberg-Marquardt), GPU version.
            @param stopCondition will be set to a bit mask indicating which stopping condition was fulfilled.
            @param nbTotalIterations will be set to the number of iterations (epochs) performed
            @return true if the training was successful **/
            bool trainGPU( short& stopCondition, size_t& nbTotalIterations );
            
            /**! \brief Sets up and initializes all device memory needed to run the network on some input.
            @param inputValues Pointer to a NxM (column-major) array, the input. If left equal to 0, mInputValues will be used.
            @param number number of samples (input vectors) N. If inputValues is a null pointer, getTrainingSampleCount() will be used. **/
            void setupDeviceMemoryRunOnGPU( const Real* inputValues = 0, const size_t number = 0 );

            /**! Save restriction of the training base to have only the part used for learning (the
             * rest will be used for validation. */
            void saveRestriction();
            /**! Copy weights from host to device */
            void copyWeightsToDevice();

            /**! \brief Sets up and initializes all other device memory needed for the training (calls setupDeviceMemoryRunOnGPU()). **/
            void setupDeviceMemoryGPU();
            /**! \brief Frees and cleans up the device memory allocated by setupDeviceMemoryRunOnGPU(). **/
            void cleanupDeviceMemoryRunOnGPU();
            /**! \brief Frees and cleans up the device memory allocated by setupDeviceMemoryGPU() (calls cleanupDeviceMemoryRunOnGPU()). **/
            void cleanupDeviceMemoryGPU();
            /**! Mix the full training set (i.e. the pure training part and the test part) on GPU
             * @param permIdx is an index (starting at 0) referencing the permutation to use - see member mPermutations */
            void randomizeTrainingSetGPU();
            /**! \brief Runs simultanously on all the training input, with the current weights.
            Calculates the error vector and the squared total error on the way. **/
            void runForTrainingGPU();
            /**! \brief Apply normalization to the training samples, so that the values lay in [a,b] (GPU version).
            The maximum and minimum values were stored in mMaxSampleValue (resp. mMinSampleValue), determined
            as the samples were added. Should only be called once all the samples are known. The actual interval
            might be [-1,1], [0,1] or the whole real line, depending on the chosen activation function.
            @return False if the normalization could not be performed (typically, only one sample) **/
            bool normalizeSamplesGPU();
            /**! \brief Computes the weight modification values Delta(w), for this epoch of the NN training (GPU version).
                \f$ \Delta(w) = ( J^T J + \lambda I_{R} )^{-1} J^T \varepsilon \f$
            where \f$\varepsilon\f$ is the N-dimensional error vector \f$e=(||\varepsilon_n||)\f$, and applies the change
            to the weight matrices. **/
            void computeDeltaWeightsGPU(Real lambda);
          #endif
            
            
            /**! \brief Displays the training settings (output to console). **/
            void information() const;
            /**! \brief Summarizes the training (output to console).
            @param stopCondition    mask: "stopCondition & 0x01" if the training stopped
                                    because the requested precision was reached, for example.
            @param nbIterations     number of epochs actually performed **/
            void summarize( const short stopCondition, size_t nbIterations,  Real totalErrorLearn, Real totalErrorTest ) const;
            
            void resetTeachingSamples();
    };

}
#endif
