//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <stdio.h>
#include <string.h>
#include <cmath>

// Classes et utilitaires Nisp
#include "nisp_pc.h"
#include "nisp_util.h"


using namespace std;

#define HERMITE  1
#define LEGENDRE 2
#define LAGUERRE 3
#define JACOBI   4

// Coefficients du PolynomialChaos
void WriteCoef(PolynomialChaos * pc, ofstream & ficsrc, char *name)
{
    int j, k, ny, p;
    ny = pc->ny;
    p = pc->p;
    char racine[100];
    strcpy(racine, name);
    strcat(racine, "_beta");
    ficsrc << "double " << racine << "[" << ny << "][" << p + 1 << "]={"
            << endl;
    int nl = 0;
    for (j = 1; j <= ny - 1; j++)
    {
        for (k = 0; k <= p; k++)
        {
            if (nl == 8)
            {
                nl = 0;
                ficsrc << endl;
            }
            nl++;
            ficsrc << pc->beta[j][k] << ",";
        }
    }
    for (k = 0; k < p; k++)
    {
        if (nl == 8)
        {
            nl = 0;
            ficsrc << endl;
        }
        nl++;
        ficsrc << pc->beta[ny][k] << ",";
    }
    ficsrc << pc->beta[ny][p] << "};" << endl;
}

// Tableau des multi indices
void WriteIndiceMultiple(PolynomialChaos * pc, ofstream & ficsrc, char *name)
{
    int i, k, nx, p;
    nx = pc->nx;
    p = pc->p;
    char racine[100];
    strcpy(racine, name);
    strcat(racine, "_indmul");
    ficsrc << "int " << racine << "[" << p + 1 << "][" << nx << "]={" << endl;
    for (k = 0; k < p; k++)
    {
        for (i = 1; i <= nx; i++)
            ficsrc << pc->indmul[k][i] << ",";
        if (!(((k + 1) * nx) % 25))
            ficsrc << endl;
    }
    for (i = 1; i < nx; i++)
        ficsrc << pc->indmul[p][i] << ",";
    ficsrc << pc->indmul[p][nx] << "};" << endl;
}

void WriteCodeLegendre(ofstream & ficsrc, char *name)
{
    char racine[100];
    strcpy(racine, name);
    strcat(racine, "_legendre");
    ficsrc << "void " << racine << "(double *phi,double x, int no) {" << endl;
    ficsrc << "   int i;" << endl;
    ficsrc << "   x=2.*x-1.;" << endl;
    ficsrc << "   phi[0]=1.; if(no>0)  phi[1]=x;" << endl;
    ficsrc
            << "   for(i=1;i<no;i++)    phi[i+1]= ((2.*i+1.) * x * phi[i] - i * phi[i-1]) / (i+1.);"
            << endl;
    ficsrc << "   for(i=0;i<=no;i++)   phi[i]  = phi[i] * sqrt(2.* i + 1.);"
            << endl;
    ficsrc << "}" << endl;
}

void WriteCodeHermite(ofstream & ficsrc, char *name)
{
    char racine[100];
    strcpy(racine, name);
    strcat(racine, "_hermite");
    ficsrc << "void " << racine << "(double *phi,double x, int no) {" << endl;
    ficsrc << "   int i,j;" << endl;
    ficsrc << "   double c1,c2;" << endl;
    ficsrc << "   x=x/sqrt(2.);" << endl;
    ficsrc << "   phi[0]=1.; if (no>0) phi[1]=2*x;" << endl;
    ficsrc
            << "   for(i=1;i<no;i++)    phi[i+1]= 2. * x * phi[i] - 2. * i * phi[i-1];"
            << endl;
    ficsrc << "   for(i=0;i<=no;i++) {" << endl;
    ficsrc << "      c1=pow(2., i);" << endl;
    ficsrc << "      for(c2=1.,j=1;j<=i;j++) c2=c2*j;" << endl;
    ficsrc << "      phi[i]  = phi[i] / sqrt(c1*c2);" << endl;
    ficsrc << "   }" << endl;
    ficsrc << "}" << endl;
}
void WriteCodeLaguerre(ofstream & ficsrc, char *name)
{
    char racine[100];
    strcpy(racine, name);
    strcat(racine, "_laguerre");
    ficsrc << "void " << racine << "(double *phi,double x, int no) {" << endl;
    ficsrc << "   int i;" << endl;
    ficsrc << "   phi[0]=1.; if(no>0)  phi[1]=1.-x;" << endl;
    ficsrc
            << "   for(i=1;i<no;i++)    phi[i+1]= ((2.*i+1.-x) * phi[i] - i * phi[i-1]) / (i+1.);"
            << endl;
    ficsrc << "}" << endl;
}

void WriteCodeJacobi(ofstream & ficsrc, char *name, double a, double b)
{
    char racine[100];
    strcpy(racine, name);
    
    strcat(racine, "_jacobi");
    ficsrc << "void " << racine << "(double *phi,double x, int no) {" << endl;
    ficsrc << "   double ai, bi, ci, di, c1, c2, c3;" << endl;
    ficsrc << "   double a = " << a << ";" << endl;
    ficsrc << "   double b = " << b << ";" << endl;
    ficsrc << "   double alpha = b - 1.;" << endl;
    ficsrc << "   double beta  = a - 1.;" << endl;
    ficsrc << "   int i;" << endl;
    ficsrc << "   x=2.*x-1.;" << endl;
    ficsrc << "   phi[0]=1.; if(no>0)  phi[1]=(alpha-beta)/2. + x * (1. +(alpha+beta)/2.);" << endl;
    ficsrc << "   for(i=1;i<no;i++) {" << endl;
    ficsrc << "      ai = 2 * (i + 1.) * (i + alpha + beta + 1.) * (2 * i + alpha + beta);"<< endl;
    ficsrc << "      bi = (2 * i + alpha + beta + 1.) * (alpha - beta) * (alpha + beta);"  << endl;
    ficsrc << "      ci = (2. * i + alpha + beta) * (2. * i + alpha + beta + 1.0) * (2. * i + alpha + beta + 2.0);" << endl;
    ficsrc << "      di = 2 * (i + alpha) * (i + beta) * (2 * i + alpha + beta + 2.);" << endl;
    ficsrc << "      phi[i+1]= ( (bi + x * ci) * phi[i] - di * phi[i-1] ) / ai;" << endl;
    ficsrc << "   }" << endl;
    ficsrc << "   for(i=0;i<=no;i++) {" << endl;
    ficsrc << "      c1 = 1. / (2. * i + a + b - 1.);" << endl;
    ficsrc << "      c2 = lgamma(i + a) + lgamma(i + b) - lgamma(i + 1.) - lgamma(i + a + b - 1.);" << endl;
    ficsrc << "      c3 = lgamma(a + b) - lgamma(a) - lgamma(b);" << endl;
    ficsrc << "      phi[i]  = phi[i] / sqrt(c1 * exp(c2 + c3));" << endl;
    ficsrc << "   }" << endl;
    ficsrc << "}" << endl;
}

// Polynomes orthogonaux
void WriteCodeSrc(PolynomialChaos * pc, ofstream & ficsrc, char *name)
{
    int i, nx, no, p;
    nx = pc->nx;
    no = pc->no;
    p = pc->p;
    // Modification 2015/04/25 : Use xi as vector to normalize input attributes
    // ficsrc << "   double psi[" << p + 1 << "],phi[" << nx << "][" << no + 1 << "],xi[" << nx << "],s;" << endl;
    ficsrc << "   double psi[" << p + 1 << "],phi[" << nx << "][" << no + 1 << "], s;" << endl;
    // ficsrc << "   for(i=0;i<nx;i++) xi[i]=x[i];" << endl;
    // End Of Modification 2015/04/25 : Use xi as vector to normalize input attributes
    char racine[100];
    for (i = 1; i <= nx; i++)
    {
        if (pc->funpol[i - 1] == HERMITE)
        {
            strcpy(racine, name);
            strcat(racine, "_hermite");
            ficsrc << "   " << racine << "(phi[" << i - 1 << "],xi[" << i - 1
                    << "]," << no << ");" << endl;
        }
        else if (pc->funpol[i - 1] == LEGENDRE)
        {
            strcpy(racine, name);
            strcat(racine, "_legendre");
            ficsrc << "   " << racine << "(phi[" << i - 1 << "],xi[" << i - 1
                    << "]," << no << ");" << endl;
        }
        else if (pc->funpol[i - 1] == JACOBI)
        {
            strcpy(racine, name);
            strcat(racine, "_jacobi");
            ficsrc << "   " << racine << "(phi[" << i - 1 << "],xi[" << i - 1
                    << "]," << no << ");" << endl;
        }
        else
        {
            strcpy(racine, name);
            strcat(racine, "_laguerre");
            ficsrc << "   " << racine << "(phi[" << i - 1 << "],xi[" << i - 1
                    << "]," << no << ");" << endl;
        }
    }
    strcpy(racine, name);
    strcat(racine, "_indmul");
    ficsrc << "   for(k=0;k<=p;k++) {" << endl;
    ficsrc << "      for(psi[k]=1.,i=0;i<nx;i++) psi[k]=psi[k]*phi[i]["
            << racine << "[k][i]];" << endl;
    ficsrc << "   }" << endl;

    strcpy(racine, name);
    strcat(racine, "_beta");
    ficsrc << "   for(j=0;j<ny;j++) {" << endl;
    ficsrc << "      for(s=0.,k=0;k<=p;k++) s+=" << racine << "[j][k]*psi[k];"
            << endl;
    ficsrc << "      y[j]=s;" << endl;
    ficsrc << "   }" << endl;

}
// Generation du code source du PolynomialChaos
void PolynomialChaos::GenerateCode(char *pcsrc, char *name)
{

    ofstream ficsrc(pcsrc, ios::out);
    if (!ficsrc)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(PolynomialChaos::GenerateCode) : Problem when open the file "
                << pcsrc << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    ficsrc << "#include <math.h>" << endl << endl;

    WriteCoef(this, ficsrc, name);
    WriteIndiceMultiple(this, ficsrc, name);

    int flaghermite  = 0;
    int flaglegendre = 0;
    int flaglaguerre = 0;
    int flagjacobi   = 0;
    for (int i = 1; i <= nx; i++)
    {
        if (funpol[i - 1] == HERMITE)
        {
            if (!flaghermite)
                WriteCodeHermite(ficsrc, name);
            flaghermite = 1;
        }
        else if (funpol[i - 1] == LEGENDRE)
        {
            if (!flaglegendre)
                WriteCodeLegendre(ficsrc, name);
            flaglegendre = 1;
        }
        else if (funpol[i - 1] == LAGUERRE)
        {
            if (!flaglaguerre)
                WriteCodeLaguerre(ficsrc, name);
            flaglaguerre = 1;
        }
        else if (funpol[i - 1] == JACOBI)
        {
            if (!flagjacobi)
                WriteCodeJacobi(ficsrc, name, gpx->va[i - 1]->a, gpx->va[i - 1]->b);
            flagjacobi = 1;
        }
    }

    ficsrc << "void " << name << "(double *x, double *y) {" << endl;
    ficsrc << "   int i,j,k,nx,ny,no,p;" << endl;
    ficsrc << "   nx=" << nx << ";ny=" << ny << ";" << "no=" << no << ";p=" << p
            << ";" << endl;
    WriteCodeSrc(this, ficsrc, name);
    ficsrc << "}" << endl;
    ficsrc.close();
}

