//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_gva.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>

#include "nisp_gva.h"
#include "nisp_util.h"

using namespace std;

// ===================
// Methodes classe SetRandomVariable
// ===================
SetRandomVariable::SetRandomVariable()
{
    nx = 0;
    np = 0;
}
SetRandomVariable::SetRandomVariable(int n)
{ // par defaut les variables sont uniformes et normalisees 
    nx = n;
    np = 0;
    for (int j = 1; j <= n; j++)
        this->AddRandomVariable(new RandomVariable("Uniforme"));
}
SetRandomVariable::SetRandomVariable(char *fichier)
{
    int k, j;
    ifstream entree(fichier, ios::in);
    if (!entree)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(SetRandomVariable::SetRandomVariable) : problem to open the file "
                << fichier << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    string ttypesampling, type, str;
    int nnp, nnx, dmax;
    double a, b, c, d;
    entree >> str >> str >> str >> ttypesampling;
    entree >> str >> str >> str >> nnp;
    entree >> str >> str >> str >> nnx;
    entree >> str >> str >> str >> dmax;
    for (j = 1; j <= nnx; j++)
    {
        entree >> str >> str >> str >> type >> a;
	if (type == "Exponentielle") {
	  this->AddRandomVariable(new RandomVariable(type, a));
	}
	else if (type == "Beta") {
	  entree >> b >> c >> d;
	  this->AddRandomVariable(new RandomVariable(type, a, b, c, d));
	}
	else {
	  entree >> b;
	  this->AddRandomVariable(new RandomVariable(type, a, b));
	}
    }
    if (nnp)
    {
        x = dmatrix(nnp + 1, nx + 1);
        w = dvector(nnp + 1);
        for (k = 1; k <= nnp; k++)
        {
            for (j = 1; j <= nnx; j++)
                entree >> x[k][j];
            entree >> w[k];
        }
    }
    entree.close();
    nx = nnx;
    np = nnp;
    typesampling = ttypesampling;
    degre_max = dmax;
}
SetRandomVariable::~SetRandomVariable()
{
    FreeMemory();
    FreeMemoryVariable();
}
void SetRandomVariable::FreeMemory()
{
    if (np)
    {
        free_dmatrix(x, np + 1);
        free_dvector(w);
        np = 0;
    }
}

void SetRandomVariable::FreeMemoryVariable()
{
    for (int i = 0; i < (int) va.size(); i++)
    {
        if (va[i] != NULL)
        {
            delete va[i];
            va[i] = NULL;
        }
    }
}

void SetRandomVariable::AddRandomVariable(RandomVariable *x)
{
    va.push_back(x);
    nx = (int) va.size();
}
void SetRandomVariable::AddRandomVariable(string type)
{
    if (type == "Normale")
        AddRandomVariable(new RandomVariable("Normale"));
    else if (type == "Uniforme")
        AddRandomVariable(new RandomVariable("Uniforme"));
    else if (type == "Exponentielle")
        AddRandomVariable(new RandomVariable("Exponentielle"));
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomvariable::AddRandomVariable) : type " << type
                << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
}

void SetRandomVariable::BuildSample(string name, int n)
{
    typesampling = name;
    void montecarlo(SetRandomVariable *, int);
    void lhs(SetRandomVariable *, int);
    void qmcSobol(SetRandomVariable *, int);
    void quadrature(SetRandomVariable *, int);
    void petras(SetRandomVariable *, int);
    void smolyak(SetRandomVariable *, string, int); // JMModif 26 mai 2009
    if (n < 0)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : level or size " << n
                << " < 0" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : stochastic dimension = "
                << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (np)
        FreeMemory();
    if (typesampling == "MonteCarlo")
        montecarlo(this, n);
    else if (typesampling == "Lhs")
        lhs(this, n);
    else if (typesampling == "QmcSobol")
        qmcSobol(this, n);
    else if (typesampling == "Quadrature")
        quadrature(this, n);
    else if (typesampling == "Petras")
        petras(this, n);
    else if (typesampling == "SmolyakGauss")
        smolyak(this, typesampling, n); // JMModif 26 mai 2009
    else if (typesampling == "SmolyakTrapeze")
        smolyak(this, typesampling, n); // JMModif 26 mai 2009
    else if (typesampling == "SmolyakFejer")
        smolyak(this, typesampling, n); // JMModif 26 mai 2009
    else if (typesampling == "SmolyakClenshawCurtis")
        smolyak(this, typesampling, n); // JMModif 28 mai 2009
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : the method " << name
                << " is unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
}
void SetRandomVariable::SetSample(string name, int n)
{
    typesampling = name;
    if (n < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::SetSample) : size " << n << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    degre_max = -1;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : stochastic dimension = "
                << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (np)
        FreeMemory();
    np = n;
    x = dmatrix(n + 1, nx + 1);
    w = dvector(n + 1);
    for (int i = 1; i <= n; i++)
        w[i] = 1. / (double) n;
}
void SetRandomVariable::SetSample(int k, int i, double value)
{
    if (k > np)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::SetSample) : the number of simulation "
                << k << " is > " << np << " = size of sample" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (i < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::SetSample) : the rank of variable " << i
                << " is < 1 " << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (i > nx)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::SetSample) : the rank of variable " << i
                << " is > stochastic dimension = " << nx << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    x[k][i] = value;
}
void SetRandomVariable::BuildSample(string name, int n, int ne)
{
    if (n < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : size " << n << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (ne < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(SetRandomVariable::BuildSample) : number of Lhs for LhsMaxMin = "
                << ne << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    void lhsmaxmin(SetRandomVariable *, int, int);
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(SetRandomVariable::BuildSample) : stochastique dimension = "
                << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (typesampling == "LhsMaxMin")
    {
        if (np)
            FreeMemory();
        typesampling = name;
        lhsmaxmin(this, n, ne);
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::BuildSample) : method " << name
                << " different of LhsMaxMin" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
}
void SetRandomVariable::BuildSample(SetRandomVariable *mother)
{
    if (nx != mother->nx)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(setRandomVariable::BuildSample) : stochastic dimension are not equal between the two set of random variables"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (mother->np == 0)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg
                << "Nisp(SetRandomVariable::BuildSample) : the size of the mother set is empty"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (np)
        FreeMemory();
    typesampling = mother->typesampling;
    np = mother->np;
    degre_max = mother->degre_max;
    x = dmatrix(np + 1, nx + 1);
    w = dvector(np + 1);
    int k, j;
    for (k = 1; k <= np; k++)
        for (j = 1; j <= nx; j++)
            x[k][j] = va[j - 1]->GetValue(mother->va[j - 1], mother->x[k][j]);
    for (k = 1; k <= np; k++)
        w[k] = mother->w[k];
}

void SetRandomVariable::Save(char *fichier)
{
    int i, j;
    ofstream sortie(fichier, ios::out);
    if (!sortie)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(SetRandomVariable::Save) : Problem when open file "
                << fichier << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    sortie << "# Type         : " << typesampling << endl;
    sortie << "# Simulations  : " << np << endl;
    sortie << "# Variables    : " << nx << endl;
    sortie << "# DegreMax     : " << degre_max << endl;
    for (i = 0; i < nx; i++)
    {
        sortie << "# PDF          : " << va[i]->type << " " << va[i]->a;
	if (va[i]->type == "Exponentielle") {
	  sortie << endl;
	}
	else if (va[i]->type == "Beta") {
	  sortie << " " << va[i]->b << " " << va[i]->c << " " << va[i]->d << endl;
	}
	else {
	  sortie << " " << va[i]->b << endl;
	}
    }
    sortie << endl;
    sortie.setf(ios::scientific, ios::floatfield);
    sortie.precision(8);
    for (i = 1; i <= np; i++)
    {
        for (j = 1; j <= nx; j++)
        {
            sortie.width(16);
            sortie << x[i][j];
        }
        sortie.width(16);
        sortie << w[i];
        sortie << endl;
    }
    sortie.close();
}
int SetRandomVariable::GetDimension()
{
    return (nx);
}
int SetRandomVariable::GetSize()
{
    return (np);
}
double SetRandomVariable::GetSample(int k, int i)
{
    return (x[k][i]);
}
void SetRandomVariable::GetLog()
{
    for (int i = 0; i < nx; i++)
        va[i]->GetLog();
}

