//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_gva.h
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief  Class of set of random variables
 */

#ifndef NISP_GVA_H
#define NISP_GVA_H

#include "nisp_va.h" // Class of random variables
#ifdef WIN32
#include "libnisp_Export.h"
#else
#define libnisp_EXPORT
#endif

using namespace std;

//! Class of set of random variables
class libnisp_EXPORT SetRandomVariable
{
public:
    //! number of random variables
    int nx; // Int_t
    //! vector of random variables  va[0],va[1],...,va[nx-1]
    vector<RandomVariable *> va;
    //! size of the sample (number of simulations)
    int np;
    //! maximum value of the degree
    int degre_max;
    //! experimental matrix x[1...np][1...nx]
    double **x;
    //! weights when the sample is built from quadrature w[1...np]
    double *w;
    //! name of the sampling method ("Quadrature","Petras","MonteCarlo","Lhs")
    string typesampling;
public:
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor
    SetRandomVariable();
    //! Constructor from a set of n uniform variables U[0,1]
    SetRandomVariable(int n);
    //! Constructor from a file
    SetRandomVariable(char *file);
    //! Destructor
    ~SetRandomVariable();
    //@}
    //! Free memory : matrix **x and weights *w
    void FreeMemory();
    void FreeMemoryVariable();
    //! Add of a random variable
    void AddRandomVariable(RandomVariable *va);
    //! Add of a random variable via the type
    void AddRandomVariable(string type);
    //! Realisation of a sample by the method "type", size "np"
    void BuildSample(string type, int np);
    //! Realisation of a sample by the method "type", size "np" (and selection one from ne samples)
    void BuildSample(string type, int np, int ne);
    //! Realisation of a sample from another sample of "*gva" (change of variables)
    void BuildSample(SetRandomVariable *gva);
    //! Specification of a sample from another componant (OPENTURNS, URANIE ...)
    void SetSample(string type, int size);
    //! Specification of the element i of the example k of the sample defined by anther componant (OPENTURNS, URANIE, ...)
    void SetSample(int k, int i, double value);
    //! Save in a "file" of the set of random variables and its sample if it exists
    void Save(char *fichier);
    //! Stochastic dimension (nx)
    int GetDimension();
    //! Size of the sample (np)
    int GetSize();
    //! Value of the input i=1,...nx of the example k=1,...np
    double GetSample(int k, int i);
    //! Informations on the set of random variables
    void GetLog();
};

#endif
