//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_qua.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#define _USE_MATH_DEFINES
#include <cmath>
#include <math.h>
#ifdef WIN32
#include "utilWIN32.h"
#endif

using namespace std;

#include "nisp_util.h"

void hermite_compute(int order, double x[], double w[]);
void legendre_compute(int order, double x[], double w[]);
void laguerre_compute(int order, double x[], double w[]);
void j_quadrature_rule( int n, double alpha, double beta, double x[], double w[] );

static int *tx;

void Quadrature(double x[], double w[], int nq, string type, double a, double b)
{
    int i;
    double * xt = dvector(nq);
    double * wt = dvector(nq);

    if (type == "Normale")
    {
        hermite_compute(nq, xt, wt);
        for (i = 1; i <= nq; i++)
        {
            x[i] = sqrt((double) 2.) * xt[i - 1];
            w[i] = wt[i - 1] / sqrt(M_PI);
        }
    }
    else if (type == "Uniforme")
    {
        legendre_compute(nq, xt, wt);
        for (i = 1; i <= nq; i++)
        {
            x[i] = (xt[i - 1] + 1.) / 2.;
            w[i] = wt[i - 1] / 2.;
        }
    }
    else if (type == "Exponentielle")
    {
        laguerre_compute(nq, xt, wt);
        for (i = 1; i <= nq; i++)
        {
            x[i] = xt[i - 1];
            w[i] = wt[i - 1];
        }
    }
    else if (type == "Beta")
    {
        // loi beta(a,b) => coefficient a0=b+1 et b0=a+1;
	double bnp = exp(lgamma(b) + lgamma(a) - lgamma(a+b));
	j_quadrature_rule (nq, b-1, a-1, xt, wt);
	// xt = points [-1,+1] et wt = poids pour une fonction de poids (1-x)^(b-1) * (1+x)^(a-1)
	// Verifie le 09/04/2014
        for (i = 1; i <= nq; i++)
        {
            x[i] = (xt[i - 1] + 1.) / 2.; // points dans [0,1]
            w[i] =  wt[i - 1] / (bnp * pow(2.,a + b - 1.));
        }
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(Quadrature) : law" << type << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    free_dvector(xt);
    free_dvector(wt);
}

void Quadrature0(double x[], double w[], int nq, string type, double a, double b)
{
    int i;
    if (type == "Normale")
    {
        hermite_compute(nq, x, w);
        for (i = 0; i < nq; i++)
        {
            x[i] = sqrt((double) 2.) * x[i];
            w[i] = w[i] / sqrt(M_PI);
        }
    }
    else if (type == "Uniforme")
    {
        legendre_compute(nq, x, w);
        for (i = 0; i < nq; i++)
        {
            x[i] = (x[i] + 1.) / 2.;
            w[i] = w[i] / 2.;
        }
    }
    else if (type == "Exponentielle")
    {
        laguerre_compute(nq, x, w);
    }
    else if (type == "Beta")
    {
        // loi beta(a,b) => coefficient a0=b+1 et b0=a+1;
	double bnp = exp(lgamma(b) + lgamma(a) - lgamma(a+b));
	j_quadrature_rule (nq, b-1, a-1, x, w);
        for (i = 0; i < nq; i++)
        {
            x[i] = (x[i] + 1.) / 2.; // points dans [0,1]
            w[i] =  w[i] / (bnp * pow(2.,a + b - 1.));
        }
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(Quadrature) : law" << type << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
}

void Rec(int i, int n)
{
    if (tx[i] < n)
        tx[i]++;
    else
    {
        tx[i] = 1;
        Rec(i - 1, n);
    }
}

/*********************************************************************************************
 Calcule la quadratrure tensorisee de nx variables aleatoires et recopie dans les tableaux
 yt[1...nt][1...nx] et wt[1...nt]. Les points et poids de quadrature sont respectivement dans
 xq[1...nx][1...nq], wq[1...nx][1...nq]. Remarque : $nt = nq^{nx}$.
 **********************************************************************************************/

/*!
 \fn QuadratureTensorise(double **yt, double *wt, double **xq, double **wq, int nx, int nq, int nt)
 \brief Quadrature tensoris�e
 \param yt Adresse de la matrice des nt points de la quadrature tensorise�e (yt[1,...,nt][1,...,nx])
 \param wt Adresse du vecteur des nt poids correspondants (wt[1,...,nt])
 \param xq Adresse de la matrice des nx quadratures monodimensionnels xt[1,...,nx][1,...,nq]
 \param wq Adresse de la matrice des nx poids associ�s wt[1,...,nx][1,...,nq]
 \param nx Nombre de variables stochastiques
 \param nq Nombre de points par quadrature
 \param nt nombre total de points de la quadrature tensoris�e ($nq^{nx})
 */
void QuadratureTensorise(double **yt, double *wt, double **xq, double **wq,
        int nx, int nq, int nt)
{
    int i, k;

    tx = ivector(nx + 1);
    for (i = 1; i <= nx; i++)
        tx[i] = 1;
    for (k = 1; k <= nt; k++)
    {
        for (i = 1; i <= nx; i++)
            yt[k][i] = xq[i][tx[i]];
        for (wt[k] = 1., i = 1; i <= nx; i++)
            wt[k] = wt[k] * wq[i][tx[i]];
        if (k < nt)
            Rec(nx, nq);
    }
    free_ivector(tx);
}
