//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTCorrelationFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$

 \brief  Fichier de test de la classe URANIE::Modeler::TCorrelationFunction et de ses classes enfants.

 Cr�e le : 2012.11.16



 */

#define DBL_PRECISION_COMPARE 1.e-12

#include "TestTCorrelationFunction.h"
#include "TString.h"

#include "TExponentialCorrFunction.h"
#include "TGaussianCorrFunction.h"
#include "TIsotropicGaussianCorrFunction.h"
#include "TMaternICorrFunction.h"
#include "TMaternIICorrFunction.h"
#include "TMaternIIICorrFunction.h"
#include "TMatern32CorrFunction.h"
#include "TMatern52CorrFunction.h"
#include "TMatern72CorrFunction.h"

#include <iostream>
using namespace std;
using namespace URANIE::Modeler;

TCorrFunctionTest::TCorrFunctionTest()
{

}

TCorrFunctionTest::~TCorrFunctionTest()
{

}

void TCorrFunctionTest::setUp()
{
    x1[0] = 0.0;
    x1[1] = 0.0;
    x1[2] = 0.0;
    x2[0] = 1.0;
    x2[1] = 1.0;
    x2[2] = 1.0;

    lc[0] = 1.0;
    lc[1] = 0.5;
    lc[2] = 0.25;
}

void TCorrFunctionTest::tearDown()
{
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TCorrFunctionTest::testTExponentialCorrFunction()
{    
    Double_t *params;
    
    // Initialise the power parameters
    Double_t p[3] = { 0.5, 1.0, 2.0 };

    // Create the correlation function
    TExponentialCorrFunction expo(3, p, lc);
    
    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(expo.getNParameters(), 6);
    CPPUNIT_ASSERT_EQUAL(expo.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(expo.getNDimensions(), 3);

    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(expo.getPower()[0], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(expo.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = expo.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 2.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.25, DBL_PRECISION_COMPARE);
        
    // Test the value of the output
    Double_t res[1];
    expo.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 5.6027964375372678e-09, DBL_PRECISION_COMPARE);
    
    // Test setting new parameters
    Double_t newParams[6] = {1.1, 1.2, 1.3, 0.1, 0.2, 0.3};
    expo.setParameters(newParams);
    params = expo.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 1.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    expo.setCorrLengths(newL);
    params = expo.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 1.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newPower[3] = {0.5,1.0,2.0};
    TExponentialCorrFunction *cf = dynamic_cast<TExponentialCorrFunction *>(expo.clone());
    cf->setPower(newPower);
    params = cf->getPower();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 2.0, DBL_PRECISION_COMPARE);
    delete cf;    
}

void TCorrFunctionTest::testTGaussianCorrFunction()
{
    Double_t *params;
    
    // Create the correlation function
    TGaussianCorrFunction gauss(3, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(gauss.getNParameters(), 3);
    CPPUNIT_ASSERT_EQUAL(gauss.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(gauss.getNDimensions(), 3);

    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(gauss.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = gauss.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    gauss.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 7.5825604279119066e-10, DBL_PRECISION_COMPARE);
    
    // Test setting new parameters
    Double_t newParams[3] = {0.1, 0.2, 0.3};
    gauss.setParameters(newParams);
    params = gauss.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.3, DBL_PRECISION_COMPARE);

    // Test cloning
    Double_t newL[3] = {0.2,0.4,0.6};
    TGaussianCorrFunction *cf = dynamic_cast<TGaussianCorrFunction *>(gauss.clone());
    cf->setCorrLengths(newL);
    params = cf->getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.6, DBL_PRECISION_COMPARE);
    delete cf;    
}

void TCorrFunctionTest::testTIsotropicGaussianCorrFunction()
{
    Double_t *params;
    
    // Create the correlation function
    TIsotropicGaussianCorrFunction isogauss(3, lc[0]);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(isogauss.getNParameters(), 1);
    CPPUNIT_ASSERT_EQUAL(isogauss.getNCorrLengths(), 1);
    CPPUNIT_ASSERT_EQUAL(isogauss.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(isogauss.getCorrLengths()[0], 1.0, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = isogauss.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    isogauss.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 0.049787068367863944, DBL_PRECISION_COMPARE);

    // Test setting new parameters
    Double_t newParams[1] = {0.1};
    isogauss.setParameters(newParams);
    params = isogauss.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.1, DBL_PRECISION_COMPARE);

    // Test cloning
    Double_t newL = 0.2;
    TIsotropicGaussianCorrFunction *cf = dynamic_cast<TIsotropicGaussianCorrFunction*>(isogauss.clone());
    cf->setCorrelationLength(newL);
    params = cf->getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.2, DBL_PRECISION_COMPARE);
    delete cf;    
}

void TCorrFunctionTest::testTMaternICorrFunction()
{
    Double_t *params;
    
    // Initialise the power parameters
    Double_t nu[3] = { 1.5, 2.5, 3.5 };

    // Create the correlation function
    TMaternICorrFunction maternI(3, nu, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(maternI.getNParameters(), 6);
    CPPUNIT_ASSERT_EQUAL(maternI.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(maternI.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternI.getParameterNu()[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternI.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = maternI.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 2.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 3.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    maternI.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 1.1473154831794043e-06, DBL_PRECISION_COMPARE);
    
    // Test setting new parameters
    Double_t newParams[6] = {1.0, 2.0, 3.0, 0.1, 0.2, 0.3};
    maternI.setParameters(newParams);
    params = maternI.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 2.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 3.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    maternI.setCorrLengths(newL);
    params = maternI.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 2.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 3.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[4], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[5], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newNu[3] = {1.3,2.2,3.1};
    TMaternICorrFunction *cf = dynamic_cast<TMaternICorrFunction*>(maternI.clone());
    cf->setParameterNu(newNu);
    params = cf->getParameterNu();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 2.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 3.1, DBL_PRECISION_COMPARE);
    delete cf;      
}

void TCorrFunctionTest::testTMaternIICorrFunction()
{
    Double_t *params;
    
    // Initialise the parameter nu
    Double_t nu = 3.5;

    // Create the correlation function
    TMaternIICorrFunction maternII(3, nu, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(maternII.getNParameters(), 4);
    CPPUNIT_ASSERT_EQUAL(maternII.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(maternII.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternII.getParameterNu(), 3.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternII.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = maternII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 3.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    maternII.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 1.12952083364676e-06, DBL_PRECISION_COMPARE);

    // Test setting new parameters
    Double_t newParams[4] = {1.5, 0.1, 0.2, 0.3};
    maternII.setParameters(newParams);
    params = maternII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    maternII.setCorrLengths(newL);
    params = maternII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newNu = 2.5;
    TMaternIICorrFunction *cf = dynamic_cast<TMaternIICorrFunction*>(maternII.clone());
    cf->setParameterNu(newNu);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(cf->getParameterNu(), 2.5, DBL_PRECISION_COMPARE);
    delete cf;    

}

void TCorrFunctionTest::testTMaternIIICorrFunction()
{
    Double_t *params;
    
    // Initialise the parameter nu
    Double_t nu = 3.5;

    // Create the correlation function
    TMaternIIICorrFunction maternIII(3, nu, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(maternIII.getNParameters(), 4);
    CPPUNIT_ASSERT_EQUAL(maternIII.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(maternIII.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternIII.getParameterNu(), 3.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(maternIII.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = maternIII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 3.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    maternIII.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 1.68722849111536e-05, DBL_PRECISION_COMPARE);

    // Test setting new parameters
    Double_t newParams[4] = {1.5, 0.1, 0.2, 0.3};
    maternIII.setParameters(newParams);
    params = maternIII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    maternIII.setCorrLengths(newL);
    params = maternIII.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[3], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newNu = 2.5;
    TMaternIIICorrFunction *cf = dynamic_cast<TMaternIIICorrFunction *>(maternIII.clone());
    cf->setParameterNu(newNu);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(cf->getParameterNu(), 2.5, DBL_PRECISION_COMPARE);
    delete cf;        
}

void TCorrFunctionTest::testTMatern32CorrFunction()
{           
    Double_t *params;
    
    // Create the correlation function
    TMatern32CorrFunction matern32(3, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(matern32.getNParameters(), 3);
    CPPUNIT_ASSERT_EQUAL(matern32.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(matern32.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matern32.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = matern32.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    matern32.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 0.00016304384554789177, DBL_PRECISION_COMPARE);
    
    // Test setting new parameters
    Double_t newParams[3] = {0.1, 0.2, 0.3};
    matern32.setParameters(newParams);
    params = matern32.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    matern32.setCorrLengths(newL);
    params = matern32.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newL2[3] = {0.3, 0.5, 0.7};
    TMatern32CorrFunction *cf = dynamic_cast<TMatern32CorrFunction*>(matern32.clone());
    cf->setCorrLengths(newL2);
    params = cf->getCorrLengths();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.7, DBL_PRECISION_COMPARE);
    delete cf;            
}

void TCorrFunctionTest::testTMatern52CorrFunction()
{
    Double_t *params;
    
    // Create the correlation function
    TMatern52CorrFunction matern52(3, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(matern52.getNParameters(), 3);
    CPPUNIT_ASSERT_EQUAL(matern52.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(matern52.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matern52.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = matern52.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    matern52.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 4.3490795841746913e-05, DBL_PRECISION_COMPARE);

    // Test setting new parameters
    Double_t newParams[3] = {0.1, 0.2, 0.3};
    matern52.setParameters(newParams);
    params = matern52.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    matern52.setCorrLengths(newL);
    params = matern52.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newL2[3] = {0.3, 0.5, 0.7};
    TMatern52CorrFunction *cf = dynamic_cast<TMatern52CorrFunction*>(matern52.clone());
    cf->setCorrLengths(newL2);
    params = cf->getCorrLengths();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.7, DBL_PRECISION_COMPARE);
    delete cf;            
}

void TCorrFunctionTest::testTMatern72CorrFunction()
{
    Double_t *params;
    
    // Create the correlation function
    TMatern72CorrFunction matern72(3, lc);

    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(matern72.getNParameters(), 3);
    CPPUNIT_ASSERT_EQUAL(matern72.getNCorrLengths(), 3);
    CPPUNIT_ASSERT_EQUAL(matern72.getNDimensions(), 3);
    
    // Test the parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matern72.getCorrLengths()[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the list of parameters
    params = matern72.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.25, DBL_PRECISION_COMPARE);

    // Test the value of the output
    Double_t res[1];
    matern72.eval(1, x1, 1, x2, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 1.6872284911153616e-05, DBL_PRECISION_COMPARE);

    // Test setting new parameters
    Double_t newParams[3] = {0.1, 0.2, 0.3};
    matern72.setParameters(newParams);
    params = matern72.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.3, DBL_PRECISION_COMPARE);

    // Test setting new correlation lengths
    Double_t newL[3] = {0.2,0.4,0.6};
    matern72.setCorrLengths(newL);
    params = matern72.getParameters();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.4, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.6, DBL_PRECISION_COMPARE);
    
    // Test cloning
    Double_t newL2[3] = {0.3, 0.5, 0.7};
    TMatern72CorrFunction *cf = dynamic_cast<TMatern72CorrFunction*>(matern72.clone());
    cf->setCorrLengths(newL2);
    params = cf->getCorrLengths();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[0], 0.3, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[1], 0.5, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(params[2], 0.7, DBL_PRECISION_COMPARE);
    delete cf;            
}
