//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTGLM.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::DataServer::TGLM
 
 Created : 2012-11-23 22:27:55
 
 */

#define DBL_PRECISION_COMPARE 1.e-10
// Data Header
#include "TestTGLM.h"
#include "TString.h"
#include <iostream>
#include <ctime>

//ROOT
#include "TMath.h"
#include "TLeaf.h"

#include "uranie_commands.h"

using namespace std;

TGLMTest::TGLMTest() :
        _tdsPoisson(NULL), _tdsBinomial(NULL), _tdsGamma(NULL)
{
    _workingTestPath = gSystem->pwd(); 
}

TGLMTest::~TGLMTest()
{

    if (_tdsPoisson != NULL)
        delete _tdsPoisson;
    _tdsPoisson = NULL;
    if (_tdsBinomial != NULL)
        delete _tdsBinomial;
    _tdsBinomial = NULL;
    if (_tdsGamma != NULL)
        delete _tdsGamma;
    _tdsGamma = NULL;

}

////////////////////////////////
///                          ///
///   setUp / tearDown       ///
///                          ///
////////////////////////////////

void TGLMTest::setUp()
{
    _tdsPoisson = new URANIE::DataServer::TDataServer("tdsPoisson",
            "Data from capitalPunishement");
    _tdsPoisson->fileDataRead("capitalPunishement.dat");
    _tdsPoisson->addAttribute("logcri", "log10(cri)");

    _tdsBinomial = new URANIE::DataServer::TDataServer("tdsBinomial",
            "Data from capitalPunishement");
    _tdsBinomial->fileDataRead("cafards.dat");
    _tdsBinomial->addAttribute("proba", "morts/total");
    _tdsBinomial->addAttribute("vivants", "total-morts");

    _tdsGamma = new URANIE::DataServer::TDataServer("tdsGamma",
            "Data from scottish");
    _tdsGamma->fileDataRead("scottish.dat");
    _tdsGamma->addAttribute("inters", "cou*unm");
}

void TGLMTest::tearDown()
{
    delete _tdsPoisson;
    _tdsPoisson = NULL;
    delete _tdsBinomial;
    _tdsBinomial = NULL;
    delete _tdsGamma;
    _tdsGamma = NULL;
    
    UranieTests::removeFile(_workingTestPath.Data(),"uranie_glm_poisson.C");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.C");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.py");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.f");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.f.bak");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TGLMTest::testTGLMPoisson()
{
    cout << endl;
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsPoisson,
            "inc:pov:blk:logcri:sou:deg", "exe");

    // Default values
    CPPUNIT_ASSERT_EQUAL(0, tglm->getNIteration());
    CPPUNIT_ASSERT_EQUAL(25, tglm->getMaxIteration());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kUnknown,
            tglm->getYDistribution());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kCanonical,
            tglm->getLinkFunction());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.e-6, tglm->getTolerance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, tglm->getNullDeviance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, tglm->getResidualDeviance(),
            DBL_PRECISION_COMPARE);

    // Define Y dsitribution
    tglm->setFamily(URANIE::Modeler::TGLM::kPoisson);
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kPoisson,
            tglm->getYDistribution());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kLog, tglm->getLinkFunction());

    tglm->unsetLog();
    tglm->estimate();

    CPPUNIT_ASSERT(_tdsPoisson->isAttribute("exehat"));

    CPPUNIT_ASSERT_EQUAL(17, _tdsPoisson->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(35.2263655437387,
            _tdsPoisson->getValue("exehat", 0), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(8.1965744215200,
            _tdsPoisson->getValue("exehat", 1), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.4171405073942,
            _tdsPoisson->getValue("exehat", 7), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5346209005245,
            _tdsPoisson->getValue("exehat", 15), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.9790935779351,
            _tdsPoisson->getValue("exehat", 16), 1e-04);

    CPPUNIT_ASSERT_EQUAL(5, tglm->getNIteration());
    CPPUNIT_ASSERT_EQUAL(25, tglm->getMaxIteration());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, tglm->getNullDeviance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(18.5916417595409, tglm->getResidualDeviance(),
            DBL_PRECISION_COMPARE);

    TVectorD dparam = tglm->getParameters();

    CPPUNIT_ASSERT_EQUAL(17, _tdsPoisson->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(7, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-6.8014798608846, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0002611017064, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0778180150431, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0949311078141, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6837179514237, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.3011833213366, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-18.7220679980592, dparam(6),
            DBL_PRECISION_COMPARE);

    delete tglm;
    tglm = NULL;

    ////////////////////////////////////////
    // kPoisson,kIdentity
    ////////////////////////////////////////
    URANIE::DataServer::TDataServer *tdsPoisson2 =
            new URANIE::DataServer::TDataServer("tdsPoisson2",
                    "Data from Dobson exemple");
    tdsPoisson2->fileDataRead("poisson.dat");

//     tglm = new URANIE::Modeler::TGLM(tdsPoisson2, "x", "y");
    tglm = new URANIE::Modeler::TGLM(tdsPoisson2, "x,y");
    tglm->setFamily(URANIE::Modeler::TGLM::kPoisson,
            URANIE::Modeler::TGLM::kIdentity);

    TVectorD vecBeta0(2);
    vecBeta0[0] = 7.0;
    vecBeta0[1] = 5.0;
    tglm->setInitialBeta(vecBeta0);

    TVectorD dvecBeta = tglm->getInitialBeta();
    CPPUNIT_ASSERT_EQUAL(2, dvecBeta.GetNrows());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.0, dvecBeta(0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.0, dvecBeta(1), DBL_PRECISION_COMPARE);

    tglm->unsetLog();
    tglm->estimate();

    CPPUNIT_ASSERT(tdsPoisson2->isAttribute("yhat"));

    CPPUNIT_ASSERT_EQUAL(9, tdsPoisson2->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5163328051158,
            tdsPoisson2->getValue("yhat", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5163328051158,
            tdsPoisson2->getValue("yhat", 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.4516332805115,
            tdsPoisson2->getValue("yhat", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(12.3869337559073,
            tdsPoisson2->getValue("yhat", 7), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(12.3869337559073,
            tdsPoisson2->getValue("yhat", 8), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(3, tglm->getNIteration());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, tglm->getNullDeviance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.89465033525568, tglm->getResidualDeviance(),
            DBL_PRECISION_COMPARE);

    TVectorD dparam2 = tglm->getParameters();

    CPPUNIT_ASSERT_EQUAL(2, dparam2.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.45163328051158, dparam2(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.93530047539575, dparam2(1),
            DBL_PRECISION_COMPARE);

    delete tglm;
    delete tdsPoisson2;
}

void TGLMTest::testTGLMBinomial()
{
    cout << endl;
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsBinomial,
            "dose", "morts:vivants");
    // Define Y dsitribution
    tglm->setFamily(URANIE::Modeler::TGLM::kBinomial);
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kBinomial,
            tglm->getYDistribution());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kLogit,
            tglm->getLinkFunction());

    tglm->unsetLog();

    tglm->estimate();

    CPPUNIT_ASSERT(_tdsBinomial->isAttribute("mortshat"));

    CPPUNIT_ASSERT_EQUAL(8, _tdsBinomial->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.4785560292542,
            _tdsBinomial->getValue("mortshat", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(9.8537873246058,
            _tdsBinomial->getValue("mortshat", 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(49.9773704552359,
            _tdsBinomial->getValue("mortshat", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(59.1715511131067,
            _tdsBinomial->getValue("mortshat", 6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(58.7151944091999,
            _tdsBinomial->getValue("mortshat", 7), DBL_PRECISION_COMPARE);

    TVectorD dvecBeta0 = tglm->getInitialBeta();
    CPPUNIT_ASSERT_EQUAL(2, dvecBeta0.GetNrows());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-8.94784295144143, dvecBeta0(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.32493696781173, dvecBeta0(1),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_EQUAL(6, tglm->getNIteration());

    TVectorD dparam = tglm->getParameters();

    CPPUNIT_ASSERT_EQUAL(8, _tdsBinomial->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(2, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-60.459104114019, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(34.121341493816, dparam(1),
            DBL_PRECISION_COMPARE);
    delete tglm;
}

void TGLMTest::testTGLMGamma()
{
    cout << endl;
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsGamma,
            "cou:unm:mor:act:gdp:age:inters", "yes");
    CPPUNIT_ASSERT_EQUAL(7, tglm->getNInput());
   
    // Define Y dsitribution
    tglm->setFamily(URANIE::Modeler::TGLM::kGamma);
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kGamma,
            tglm->getYDistribution());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::TGLM::kInverse,
            tglm->getLinkFunction());

    tglm->setTolerance(1e-8);
    tglm->setMaxIteration(10);
    tglm->unsetLog();
    tglm->estimate();

    CPPUNIT_ASSERT(_tdsGamma->isAttribute("yeshat"));

    CPPUNIT_ASSERT_EQUAL(32, _tdsGamma->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.578043148244767,
            _tdsGamma->getValue("yeshat", 0), 1e-06);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.532733447006391,
            _tdsGamma->getValue("yeshat", 1), 1e-06);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.748350817571181,
            _tdsGamma->getValue("yeshat", 15), 1e-06);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.738260921658244,
            _tdsGamma->getValue("yeshat", 30), 1e-06);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.696151562102958,
            _tdsGamma->getValue("yeshat", 31), 1e-06);

    CPPUNIT_ASSERT_EQUAL(3, tglm->getNIteration());
    CPPUNIT_ASSERT_EQUAL(10, tglm->getMaxIteration());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.e-8, tglm->getTolerance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, tglm->getNullDeviance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.08738851641, tglm->getResidualDeviance(),
            DBL_PRECISION_COMPARE);

    TVectorD dparam = tglm->getParameters();

    CPPUNIT_ASSERT_EQUAL(32, _tdsGamma->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(8, dparam.GetNrows());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.77652702753725, dparam(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00496176829942, dparam(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.20344225895856, dparam(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.00718142873678, dparam(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.01118520129331, dparam(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.46751504201455e-05, dparam(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.05186831119354, dparam(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.00024271749790, dparam(7),
            DBL_PRECISION_COMPARE);
    delete tglm;
}


void TGLMTest::testExportModelCplusplus()
{
    
    time_t seconds;
    string line, date;
    Bool_t bStop = kFALSE;
    Int_t test1 = 0;
    Int_t test2 = 0;
    Int_t test3 = 0;
    Int_t test4 = 0;
     ////////////////////////////////////////
    // kPoisson,kIdentity
    ////////////////////////////////////////
    URANIE::DataServer::TDataServer *tdsPoisson2 =
            new URANIE::DataServer::TDataServer("tdsPoisson2",
                    "Data from Dobson exemple");
    tdsPoisson2->fileDataRead("poisson.dat");

    URANIE::Modeler::TGLM *tglm = new URANIE::Modeler::TGLM(tdsPoisson2, "x,y");
    tglm->setFamily(URANIE::Modeler::TGLM::kPoisson,
            URANIE::Modeler::TGLM::kIdentity);

    TVectorD vecBeta0(2);
    vecBeta0[0] = 7.0;
    vecBeta0[1] = 5.0;
    tglm->setInitialBeta(vecBeta0);
    TVectorD dvecBeta = tglm->getInitialBeta();
    
    tglm->unsetLog();
    tglm->estimate();
    
    tglm->exportFunction("c++", "uranie_glm_poisson.C","glmPoisson");
    
    fstream file("uranie_glm_poisson.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file, line);
        if (line.find("  //     ** Date : ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[24]) && (date[1] == line[25])
                    && (date[2] == line[26]) && (date[3] == line[27])
                    && (date[4] == line[28]) && (date[5] == line[29])
                    && (date[6] == line[30]) && (date[7] == line[31])
                    && (date[8] == line[32]) && (date[9] == line[33])
                    && (date[10] == line[34]) && (date[11] == line[35])
                    && (date[12] == line[36]) && (date[13] == line[37])
                    && (date[14] == line[38]) && (date[15] == line[39])
                    && (date[16] == line[40]) && (date[17] == line[41])
                    && (date[18] == line[42]) && (date[19] == line[43])
                    && (date[20] == line[44]) && (date[21] == line[45])
                    && (date[22] == line[46]) && (date[23] == line[47]))
                test1 = 1;
               
            if(test1==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);
                            
                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);
                
               
                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test1 = 1;
            }
        }
         
        if (line.find("  double y = 7.45163;")
                != string::npos)
            test2 = 1;
        
        if (line.find("  y += param[0]*4.9353;")
                != string::npos)
            test3 = 1;
         
        if (line.find("glmPoisson(double *param, double *res)")
                != string::npos)
            test4 = 1;
         
        if (file.eof())
            bStop = kTRUE;
    }
    
    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
    CPPUNIT_ASSERT_EQUAL(1, test4);

    
    delete tglm;
    delete tdsPoisson2;
    
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

void TGLMTest::throwExceptionTGLMSetterFamily()
{
    cout << endl;
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsBinomial,
            "dose", "morts:vivants");
    tglm->unsetLog();

    CPPUNIT_ASSERT_THROW(tglm->setFamily(),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(tglm->setFamily(URANIE::Modeler::TGLM::kNormal),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kPoisson,
                    URANIE::Modeler::TGLM::kLogit),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kPoisson,
                    URANIE::Modeler::TGLM::kProbit),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kPoisson,
                    URANIE::Modeler::TGLM::kInverse),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kGamma,
                    URANIE::Modeler::TGLM::kIdentity),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kGamma,
                    URANIE::Modeler::TGLM::kLogit),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kGamma,
                    URANIE::Modeler::TGLM::kProbit),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kGamma,
                    URANIE::Modeler::TGLM::kLog),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kBinomial,
                    URANIE::Modeler::TGLM::kIdentity),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kBinomial,
                    URANIE::Modeler::TGLM::kProbit),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kBinomial,
                    URANIE::Modeler::TGLM::kLog),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->setFamily(URANIE::Modeler::TGLM::kBinomial,
                    URANIE::Modeler::TGLM::kInverse),
            URANIE::Exceptions::UWarningExceptions);

    delete tglm;
	
}

void TGLMTest::throwExceptionTGLMInitBeta()
{
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsPoisson,
            "inc:pov:blk:logcri:sou:deg", "exe");

    TVectorD vecBeta0(6);
    vecBeta0.Zero();
    CPPUNIT_ASSERT_THROW(tglm->setInitialBeta(vecBeta0),
            URANIE::Exceptions::UErrorExceptions);
    delete tglm;
}

void TGLMTest::throwExceptionExportFunction()
{
    cout << endl;
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsBinomial,
            "dose", "morts:vivants");
    tglm->unsetLog();

    // Define Y dsitribution
    tglm->setFamily(URANIE::Modeler::TGLM::kBinomial);

    tglm->estimate();

    CPPUNIT_ASSERT_THROW(
            tglm->exportFunction("toto", "filename", "function_name"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->exportFunction("pmml", "filename", "function_name"),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->exportFunction("fortran", "filename", "function_name"),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            tglm->exportFunction("python", "filename", "function_name"),
            URANIE::Exceptions::UWarningExceptions);
    delete tglm;
}



void TGLMTest::throwExceptionConstructorForTooMuchOutputs()
{
    cout << endl;
    _tdsBinomial->addAttribute("zombies", "total-morts");
    CPPUNIT_ASSERT_THROW(
            new URANIE::Modeler::TGLM(_tdsBinomial,
            "dose,morts:vivants:zombies"),
            URANIE::Exceptions::UErrorExceptions);
    
    CPPUNIT_ASSERT_THROW(
            new URANIE::Modeler::TGLM(_tdsBinomial,
            "dose","morts:vivants:zombies"),
            URANIE::Exceptions::UErrorExceptions);
    
    URANIE::Modeler::TGLM * tglm = new URANIE::Modeler::TGLM(_tdsBinomial, "dose","morts:vivants");
    CPPUNIT_ASSERT_THROW(tglm->getOutputName(-1),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(tglm->getOutputName(5),
            URANIE::Exceptions::UErrorExceptions);
    delete tglm;

   
}

