//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTGPCostFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$

 \brief  Test suite for class URANIE::Modeler::TGPCostFunction and its children classes.



 */

#define DBL_PRECISION_COMPARE 1.e-12

#include "TestTGPCostFunction.h"
#include "TString.h"

#include "TGPBuilder.h"
#include "TGPMLCostFunction.h"
#include "TGPReMLCostFunction.h"
#include "TGPLooCostFunction.h"

#include <iostream>
#include "uranie_commands.h"
using namespace std;
using namespace URANIE::Modeler;
using namespace URANIE::DataServer;

TGPCostFunctionTest::TGPCostFunctionTest()
{
   
}

TGPCostFunctionTest::~TGPCostFunctionTest()
{
    delete tdsObs;
}

void TGPCostFunctionTest::setUp()
{
    tdsObs = new TDataServer("tdsObs","");
    tdsObs->fileDataRead("utf-1D-train.dat");
}

void TGPCostFunctionTest::tearDown()
{
    delete tdsObs;
    tdsObs=NULL;
    UranieTests::removeFile(gSystem->pwd(),"utf-1D-train.root");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TGPCostFunctionTest::testTGPMLCostFunction()
{   
    Double_t *data;
    
    // Construction of the GPBuilder without trend
    TGPBuilder gpb(tdsObs, "x1", "y", "isogauss", "");
    
    // Construction of the ML cost function
    TGPMLCostFunction cost(&gpb);
    
    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(cost.getNObs(), tdsObs->getNPatterns());
    
    // Test the pointers
    data = tdsObs->getTuple()->getBranchData("y");
    for(Int_t i = 0; i < tdsObs->getNPatterns(); i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(cost.getObservationVector()[i], data[i], DBL_PRECISION_COMPARE);
    
    data = cost.getCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    data = cost.getCovarianceMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    // Test eval
    Double_t theta[1] = {0.1};
    Double_t res[2]   = {0.0, 0.0};    
    cost.eval(theta, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 15.679503844689815, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[1],  2.44897935558036,  DBL_PRECISION_COMPARE);
}

void TGPCostFunctionTest::testTGPReMLCostFunction()
{    
    Double_t *data;
    
    // Construction of the GPBuilder with a constant trend
    TGPBuilder gpb(tdsObs, "x1", "y", "isogauss", "const");
    
    // Construction of the ML cost function
    TGPReMLCostFunction cost(&gpb);
    
    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(cost.getNObs(), tdsObs->getNPatterns());
    
    // Test the pointers
    data = tdsObs->getTuple()->getBranchData("y");
    for(Int_t i = 0; i < tdsObs->getNPatterns(); i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(cost.getObservationVector()[i], data[i], DBL_PRECISION_COMPARE);
    
    data = cost.getCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    data = cost.getCovarianceMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    // Test eval
    Double_t theta[1] = {0.1};
    Double_t res[2]   = {0.0, 0.0};    
    cost.eval(theta, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 10.292414478302984,  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[1],  1.2355632598998099, DBL_PRECISION_COMPARE);
}

void TGPCostFunctionTest::testTGPLooCostFunction()
{    
    Double_t *data;
    
    // Construction of the GPBuilder without trend
    TGPBuilder gpb(tdsObs, "x1", "y", "isogauss", "");
    
    // Construction of the ML cost function
    TGPLooCostFunction cost(&gpb);
    
    // Test the numbers
    CPPUNIT_ASSERT_EQUAL(cost.getNObs(), tdsObs->getNPatterns());
    
    // Test the pointers
    data = tdsObs->getTuple()->getBranchData("y");
    for(Int_t i = 0; i < tdsObs->getNPatterns(); i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(cost.getObservationVector()[i], data[i], DBL_PRECISION_COMPARE);
    
    data = cost.getCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    data = cost.getCovarianceMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[0], 1.0,                    DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[1], 0.5969223627713421,     DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[2], 0.005456544824590714,   DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[3], 2.5981145523491532e-20, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[4], 2.0636663452106608e-05, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(data[5], 6.803831497717578e-12,  DBL_PRECISION_COMPARE);
    
    // Test eval
    Double_t theta[1] = {0.1};
    Double_t res[2]   = {0.0, 0.0};    
    cost.eval(theta, res);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[0], 2.3017944377386526, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(res[1], 2.3600865999002756,  DBL_PRECISION_COMPARE);
}

