//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTCorrelationFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$

 \brief  Fichier de test de la classe URANIE::Modeler::TCorrelationFunction et de ses classes enfants.

 Cr�e le : 2012.11.16



 */

#define DBL_PRECISION_COMPARE 1.e-10
#define COV_PRECISION_COMPARE 1.e-6

#include "TestTKriging.h"

// std
#include <iostream>
#include <string>

// ROOT
#include "TMath.h"

// URANIE::DataServer
#include "TAttribute.h"

// URANIE::Relauncher
#include "TLauncher2.h"

// URANIE::Modeler
#include "TGPBuilder.h"
#include "TKriging.h"
#include "uranie_commands.h"

#ifdef  WITH_LIBXML2
#include "PMMLlib.hxx"
#endif
//#include "TANNModeler.h"
using namespace std;
using namespace URANIE::Modeler;
using namespace URANIE::DataServer;
using namespace URANIE::Relauncher;


//===============================================================================//
// Class TKrigingTest
//===============================================================================//

TKrigingTest::TKrigingTest()
{

}

TKrigingTest::~TKrigingTest()
{
    UranieTests::removeFile(gSystem->pwd(),"gpBuilder-optim.dat");
    UranieTests::removeFile(gSystem->pwd(),"gpBuilder-screening.dat");
    UranieTests::removeFile(gSystem->pwd(),"utf-1D-test_sub.dat");
}

void TKrigingTest::setUp()
{
    _tdsObs = new TDataServer("tdsObs", "");
    _tdsObs->fileDataRead("utf-1D-train.dat");
}

void TKrigingTest::tearDown()
{
    delete _tdsObs;
    _tdsObs=NULL;
    
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////
#ifdef  WITH_LIBXML2
void TKrigingTest::TestGaussianProcessComputation(void){

  std::cout << std::endl;

  TestGaussianProcessComputationChoix(1);
  TestGaussianProcessComputationChoix(2);
}

void TKrigingTest::TestGaussianProcessComputationChoix(int choix){
  // --
  std::ofstream *sourcefile;
  TDataServer *tdsEstim, *tdsEstimPmml;
  TKriging    *krig, *krigPmml;
  TLauncher2  *launcher, *launcherPmml;

  TAttribute  *yEstim, *vEstim;  
  TAttribute  *yEstimPmml, *vEstimPmml;  

  TGPBuilder  *gpb;
  URANIE::DataServer::TDataServer * tObs;
  std::string trainingFileName;
  std::string testFileName;
  std::string archiName;
  std::string strPmmlfile;
  // --
  // Construction of the kriging model, with a gaussian  correlation function 
  // and no trend
  //TGPBuilder *gpb = new TGPBuilder(_tdsObs, "x1", "y", "gauss", "");

  strPmmlfile = "kriging.pmml";
  if (choix==1){
    trainingFileName = "utf-1D-train.dat";
    testFileName = "utf-1D-train.dat";
    archiName = "x1";
  }
  else if (choix == 2){
    trainingFileName = "utf_4D_train.dat";
    testFileName = "utf_4D_train.dat";
    archiName = "x1:x2:x3:x4";
  }

  tObs = new TDataServer("tObs", "");
  tObs->fileDataRead(trainingFileName.c_str());

  gpb = new TGPBuilder(tObs, archiName.c_str(), "y", "gauss", "");
  gpb->findOptimalParameters("ML", 0, "");
  krig = gpb->buildGP();                 

  sourcefile = new std::ofstream("exemple_kriging.cxx");
  krig->exportModelCplusplus(sourcefile, "");
  sourcefile->close();
  delete sourcefile; sourcefile = NULL;

  krig->exportModelPMML(strPmmlfile.c_str(), "GP", "new");

  krigPmml = new TKriging(strPmmlfile, "GP");

  CPPUNIT_ASSERT_EQUAL( krig->getNInputs(), krigPmml->getNInputs() );
  CPPUNIT_ASSERT_EQUAL( krig->getNObs(), krigPmml->getNObs() );
  CPPUNIT_ASSERT_EQUAL( krig->getNCorrLengths(), krigPmml->getNCorrLengths() );
  Double_t *ptr, *ptrPmml;
  std::vector<double> tvmin, tvminPmml;
  std::vector<double> tvmax, tvmaxPmml;
  std::vector<double> tmean, tmeanPmml;
  std::vector<double> tvar,  tvarPmml;
  std::vector<double> tnorm, tnormPmml;
  double vmin,vminPmml;
  double vmax,vmaxPmml;
  double mean,meanPmml;
  double var,varPmml;
  double norm, normPmml;


  ptr     = krig->getObsInputs();
  ptrPmml = krigPmml->getObsInputs();
  TKriging::StatMat("X", krig->getNInputs(), krig->getNObs(), ptr,tvmin,tvmax,tmean,tvar,tnorm);
  TKriging::StatMat("X", krigPmml->getNInputs(), krigPmml->getNObs(), ptrPmml,tvminPmml,tvmaxPmml,tmeanPmml,tvarPmml,tnormPmml);
  for(int i=0; i<krig->getNInputs(); i++){
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tvmin[i], tvminPmml[i], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tvmax[i], tvmaxPmml[i], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tmean[i], tmeanPmml[i], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tvar[i],  tvarPmml[i],  DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tnorm[i], tnormPmml[i], DBL_PRECISION_COMPARE );
  }

  ptr     = krig->getObsOutputs();
  ptrPmml = krigPmml->getObsOutputs();
  TKriging::StatVect("Y", krig->getNObs(), ptr, vmin,vmax,mean,var,norm);
  TKriging::StatVect("Y", krigPmml->getNObs(), ptrPmml, vminPmml,vmaxPmml,meanPmml,varPmml,normPmml);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmin, vminPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmax, vmaxPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(mean, meanPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(var,  varPmml,  DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(norm, normPmml, DBL_PRECISION_COMPARE );
  // ------------------------------------------------
  ptr     = krig->getGPParams();
  ptrPmml = krigPmml->getGPParams();
  TKriging::StatVect("Gam", krig->getNObs(), ptr, vmin,vmax,mean,var,norm);
  TKriging::StatVect("Gam", krigPmml->getNObs(), ptrPmml, vminPmml,vmaxPmml,meanPmml,varPmml,normPmml);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmin, vminPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmax, vmaxPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(mean, meanPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(var,  varPmml,  DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(norm, normPmml, DBL_PRECISION_COMPARE );
  // ------------------------------------------------
  ptr     = krig->getCorrLengthsNormalised();
  ptrPmml = krigPmml->getCorrLengthsNormalised();
  TKriging::StatVect("LmabdaN", krig->getNInputs(), ptr, vmin,vmax,mean,var,norm);
  TKriging::StatVect("LambdaN", krigPmml->getNInputs(), ptrPmml, vminPmml,vmaxPmml,meanPmml,varPmml,normPmml);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmin, vminPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmax, vmaxPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(mean, meanPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(var,  varPmml,  DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(norm, normPmml, DBL_PRECISION_COMPARE );
  // ------------------------------------------------
  ptr     = krig->getCorrLengths();
  ptrPmml = krigPmml->getCorrLengths();
  TKriging::StatVect("Lambda", krig->getNInputs(), ptr, vmin,vmax,mean,var,norm);
  TKriging::StatVect("Lambda", krigPmml->getNInputs(), ptrPmml, vminPmml,vmaxPmml,meanPmml,varPmml,normPmml);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmin, vminPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmax, vmaxPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(mean, meanPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(var,  varPmml,  DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(norm, normPmml, DBL_PRECISION_COMPARE );
  // ------------------------------------------------
  ptr     = krig->getiL();
  ptrPmml = krigPmml->getiL();
  int krig_NObs = krig->getNObs();
  TKriging::StatVect("iL", krig_NObs*krig_NObs, ptr, vmin,vmax,mean,var,norm);
  int krigPmml_NObs =  krigPmml->getNObs();
  TKriging::StatVect("iL", krigPmml_NObs*krigPmml_NObs, ptrPmml, vminPmml,vmaxPmml,meanPmml,varPmml,normPmml);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmin, vminPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(vmax, vmaxPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(mean, meanPmml, DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(var,  varPmml,  DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(norm, normPmml, DBL_PRECISION_COMPARE );
  // ------------------------------------------------
  yEstim = new TAttribute("yEstim");
  vEstim = new TAttribute("vEstim");

  yEstimPmml = new TAttribute("yEstim");
  vEstimPmml = new TAttribute("vEstim");
  
  tdsEstim = new TDataServer("tdsEstim", "");
  tdsEstim->fileDataRead(testFileName.c_str());
  tdsEstimPmml = new TDataServer("tdsEstim", "");
  tdsEstimPmml->fileDataRead(testFileName.c_str());

  if (choix == 1){
    krig->addInput(tdsEstim->getAttribute("x1"));
    krigPmml->addInput(tdsEstimPmml->getAttribute("x1"));
  }
  else if (choix==2){
    krig->addInput(tdsEstim->getAttribute("x1"));
    krig->addInput(tdsEstim->getAttribute("x2"));
    krig->addInput(tdsEstim->getAttribute("x3"));
    krig->addInput(tdsEstim->getAttribute("x4"));

    krigPmml->addInput(tdsEstimPmml->getAttribute("x1"));
    krigPmml->addInput(tdsEstimPmml->getAttribute("x2"));
    krigPmml->addInput(tdsEstimPmml->getAttribute("x3"));
    krigPmml->addInput(tdsEstimPmml->getAttribute("x4"));
  }
  krig->setOutputs(2, yEstim, vEstim);
  krigPmml->setOutputs(2, yEstimPmml, vEstimPmml);

  launcher = new TLauncher2(tdsEstim, krig);
  launcher->solverLoop();
  launcherPmml = new TLauncher2(tdsEstimPmml, krigPmml);
  launcherPmml->solverLoop();

  tdsEstim->addAttribute("err2", "(y-yEstim)**2");
  tdsEstim->computeStatistic();
  tdsEstimPmml->addAttribute("err2", "(y-yEstim)**2");
  tdsEstimPmml->computeStatistic();

  std::ostringstream oss;
  oss.precision(16);
  oss << scientific;
  
  oss << "yEstim0: " << tdsEstim->getValue("yEstim",0) << std::endl;
  oss << "vEstim0: " << tdsEstim->getValue("vEstim",0) << std::endl;
  oss << "yEstim : " << tdsEstim->getAttribute("yEstim")->getMean() << std::endl;
  oss << "vEstim : " << tdsEstim->getAttribute("vEstim")->getStd() << std::endl;
  oss << "err2   : " << tdsEstim->getAttribute("err2")->getMean() << std::endl;
  std::cout << oss.str() << std::endl;

  /*
  std::string label;
  double val1,val2,dv;
  std::ostringstream srap;
  srap.precision(16);
  srap << scientific;

  label = "yEstim";
  val1 = tdsEstim->getValue(label,0);
  val2 = tdsEstimPmml->getValue(label,0);
  dv   = val2-val1;
  srap << label+"0" << " " << val1 << " " << val2 << " " << dv << std::endl;

  label = "vEstim";
  val1 = tdsEstim->getValue(label,0);
  val2 = tdsEstimPmml->getValue(label,0);
  dv   = val2-val1;
  srap << label+"0" << " " << val1 << " " << val2 << " " << dv << std::endl;

  label = "yEstim";
  val1 = tdsEstim->getAttribute(label)->getMean();
  val2 = tdsEstimPmml->getAttribute(label)->getMean();
  dv   = val2-val1;
  srap << label+ " " << " " << val1 << " " << val2 << " " << dv << std::endl;

  label = "vEstim";
  val1 = tdsEstim->getAttribute(label)->getStd();
  val2 = tdsEstimPmml->getAttribute(label)->getStd();
  dv   = val2-val1;
  srap << label + " "<< " " << val1 << " " << val2 << " " << dv << std::endl;

  label = "err2";
  val1 = tdsEstim->getAttribute(label)->getMean();
  val2 = tdsEstimPmml->getAttribute(label)->getMean();
  dv   = val2-val1;
  srap << label + "   "<< " " << val1 << " " << val2 << " " << dv << std::endl;
  std::cout << srap.str() << std::endl;
  */

  double _yEstim0, _vEstim0;
  double _yEstim, _vEstim;
  double _err2;

  if (choix == 1){
    _yEstim0 = 3.4319179999999999e-01;
    _vEstim0 = 0.0000000000000000e+00;
    _yEstim  = 1.2218459500000001e+00;
    _vEstim  = 4.5324665183683945e-17;
    _err2    = 8.2173010960522059e-33;
  }
  else if (choix ==2 ){
    _yEstim0 = 1.2351319999999999e+00;
    _vEstim0 = 0.0000000000000000e+00;
    _yEstim  = 9.7158255680119932e-01;
    _vEstim  = 2.1793376873652839e-16;
    _err2    = 1.0145914258105386e-30;
  }

  CPPUNIT_ASSERT_DOUBLES_EQUAL(_yEstim0 , tdsEstimPmml->getValue("yEstim",0), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_vEstim0 , tdsEstimPmml->getValue("vEstim",0), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_yEstim  , tdsEstimPmml->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_vEstim  , tdsEstimPmml->getAttribute("vEstim")->getStd(), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(_err2    , tdsEstimPmml->getAttribute("err2")->getMean(), DBL_PRECISION_COMPARE );


  CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsEstim->getValue("yEstim",0),
			       tdsEstimPmml->getValue("yEstim",0)         , DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsEstim->getValue("vEstim",0),
			       tdsEstimPmml->getValue("vEstim",0)         , DBL_PRECISION_COMPARE );
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsEstim->getAttribute("yEstim")->getMean(),
			       tdsEstimPmml->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsEstim->getAttribute("vEstim")->getStd(),
			       tdsEstimPmml->getAttribute("vEstim")->getStd(),DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsEstim->getAttribute("err2")->getMean(),
			       tdsEstimPmml->getAttribute("err2")->getMean(),DBL_PRECISION_COMPARE );

  delete launcher;   launcher = NULL;
  delete tdsEstim;   tdsEstim = NULL;
  delete vEstim;     vEstim = NULL;
  delete yEstim;     yEstim = NULL;
  
  delete launcherPmml;   launcherPmml = NULL;
  delete tdsEstimPmml;   tdsEstimPmml = NULL;
  delete vEstimPmml;     vEstimPmml = NULL;
  delete yEstimPmml;     yEstimPmml = NULL;
  
  delete krigPmml; krigPmml=NULL;
  delete krig;       krig = NULL;
  delete gpb;        gpb = NULL;
  delete tObs;       tObs=NULL;
}

void TKrigingTest::TestGaussianProcessCheck(void){
  // --
  std::ofstream *sourcefile;
  TKriging    *krig;
  TLauncher2  *launcher, *launcherPmml;

  TGPBuilder  *gpb;
  URANIE::DataServer::TDataServer * tObs;
  std::string trainingFileName;
  std::string testFileName;
  std::string archiName;
  std::string strPmmlfile;
  // --
  // Construction of the kriging model, with a matern3/2  correlation function 
  // and a constant trend


  strPmmlfile = "kriging.pmml";

  trainingFileName = "utf-1D-train.dat";
  testFileName = "utf-1D-train.dat";
  archiName = "x1";

  tObs = new TDataServer("tObs", "");
  tObs->fileDataRead(trainingFileName.c_str());

  gpb = new TGPBuilder(tObs, archiName.c_str(), "y", "matern3/2", "const", "const");
  gpb->findOptimalParameters("ML", 0, "");
  krig = gpb->buildGP();                 

  sourcefile = new std::ofstream("exemple_kriging.cxx");
  krig->exportModelCplusplus(sourcefile, "");
  sourcefile->close();
  delete sourcefile; sourcefile = NULL;

  
  CPPUNIT_ASSERT_THROW(krig->exportModelPMML(strPmmlfile.c_str(), "GP", "new"), URANIE::Exceptions::UErrorExceptions);

  delete krig;       krig = NULL;
  delete gpb;        gpb = NULL;
  delete tObs;       tObs=NULL;
}

void TKrigingTest::TestGaussianProcessCheckTrend(void){
  // --
  std::ofstream *sourcefile;
  TKriging    *krig;
  TLauncher2  *launcher, *launcherPmml;

  TGPBuilder  *gpb;
  URANIE::DataServer::TDataServer * tObs;
  std::string trainingFileName;
  std::string testFileName;
  std::string archiName;
  std::string strPmmlfile;
  // --
  // Construction of the kriging model, with a gauss  correlation function 
  // and a constant trend


  strPmmlfile = "kriging.pmml";

  trainingFileName = "utf-1D-train.dat";
  testFileName = "utf-1D-train.dat";
  archiName = "x1";

  tObs = new TDataServer("tObs", "");
  tObs->fileDataRead(trainingFileName.c_str());

  gpb = new TGPBuilder(tObs, archiName.c_str(), "y", "gauss", "const", "const");
  gpb->findOptimalParameters("ML", 0, "");
  krig = gpb->buildGP();                 

  sourcefile = new std::ofstream("exemple_kriging.cxx");
  krig->exportModelCplusplus(sourcefile, "");
  sourcefile->close();
  delete sourcefile; sourcefile = NULL;

  
  CPPUNIT_ASSERT_THROW(krig->exportModelPMML(strPmmlfile.c_str(), "GP", "new"), URANIE::Exceptions::UErrorExceptions);

  delete krig;       krig = NULL;
  delete gpb;        gpb = NULL;
  delete tObs;       tObs=NULL;
}
#endif



void TKrigingTest::testConstructionFromGPBuilder()
{
    // Construction of the kriging model, with a matern3/2 correlation function 
    // and a constant deterministic trend.
    TGPBuilder *gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern3/2", "const");
    gpb->findOptimalParameters("ML", 0, "neldermead");
    //gpb->findOptimalParameters("ML", 0, "");
    TKriging *krig = gpb->buildGP();                 

    // Observations    
    CPPUNIT_ASSERT( strcmp(krig->getInputNames(), "x1") == 0 );
    CPPUNIT_ASSERT( strcmp(krig->getOutputName(), "y" ) == 0 );
    CPPUNIT_ASSERT_EQUAL( 1, krig->getNInputs() );
    CPPUNIT_ASSERT_EQUAL( 6, krig->getNObs() );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.671542372826104073, krig->getObsInputs()[0]    , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.3431918           , krig->getObsOutputs()[0]   , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02215734          , krig->getObsNormParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.9915253           , krig->getObsNormParams()[1], DBL_PRECISION_COMPARE );

    // Correlation function
    TCorrelationFunction *cf = krig->getCorrFunction();
    CPPUNIT_ASSERT( strcmp(cf->ClassName(), "URANIE::Modeler::TMatern32CorrFunction") == 0 );
    CPPUNIT_ASSERT_EQUAL( 1, krig->getNCorrFunctionParams() );
    CPPUNIT_ASSERT_EQUAL( 1, krig->getNCorrLengths() );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.558592607059609, krig->getCorrLengths()[0]          , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.576244140625000068, krig->getCorrLengthsNormalised()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.576244140625000068, krig->getCorrFunctionParams()[0]      , DBL_PRECISION_COMPARE );    
    
    // Trend
    CPPUNIT_ASSERT( krig->hasTrend() );
    CPPUNIT_ASSERT_EQUAL( 1, krig->getNTrendParams() );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.80406678411240495 , krig->getTrendParams()[0]                , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.873684079770569166, krig->getTrendParamsCovMatrix()[0], DBL_PRECISION_COMPARE );
        
    // Model parameters
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.7401714251646172, krig->getVariance()                , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0               , krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.73975757441743095 , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    // Model matrices
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  7.58060702348973714e-01, krig->getiL()[0] , DBL_PRECISION_COMPARE );                                
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  7.58060702348973714e-01, krig->getM()[0], DBL_PRECISION_COMPARE );                                  
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -9.34710693086673317e-01, krig->getiR()[0], DBL_PRECISION_COMPARE );
    
    // Loo
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.412599686143165501, krig->getLooRMSE() , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.864155497921065163, krig->getLooQ2()   , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.12019871362438905e-01, krig->getLooErrors()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 6.43882073052330778e-02, krig->getLooVariances()[0], DBL_PRECISION_COMPARE );
    
        Int_t tdsObs_nattributes_ini = _tdsObs->getNAttributes();
        Int_t tdsObs_npatterns_ini = _tdsObs->getNPatterns();

    TDataServer *tdsLoo = krig->getLooData();
    CPPUNIT_ASSERT_EQUAL(5, tdsLoo->getNAttributes() );
    CPPUNIT_ASSERT( tdsLoo->isAttribute("errorLoo")  );
    CPPUNIT_ASSERT( tdsLoo->isAttribute("varLoo")    );
    CPPUNIT_ASSERT( tdsLoo->isAttribute("stdErrLoo") );
    CPPUNIT_ASSERT_EQUAL(6, tdsLoo->getNPatterns()   );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.12019871362438905e-01, tdsLoo->getValue("errorLoo", 0) , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  6.43882073052330778e-02, tdsLoo->getValue("varLoo", 0)   , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -4.41460552810858942e-01, tdsLoo->getValue("stdErrLoo", 0), DBL_PRECISION_COMPARE );

    _tdsObs->merge(tdsLoo);

    CPPUNIT_ASSERT_EQUAL(tdsObs_nattributes_ini + tdsLoo->getNAttributes(), _tdsObs->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(tdsObs_npatterns_ini, _tdsObs->getNPatterns());
    CPPUNIT_ASSERT( _tdsObs->isAttribute("errorLoo")  );
    CPPUNIT_ASSERT( _tdsObs->isAttribute("varLoo")    );
    CPPUNIT_ASSERT( _tdsObs->isAttribute("stdErrLoo") );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.12019871362438905e-01, _tdsObs->getValue("errorLoo", 0) , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  6.43882073052330778e-02, _tdsObs->getValue("varLoo", 0)   , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -4.41460552810858942e-01, _tdsObs->getValue("stdErrLoo", 0), DBL_PRECISION_COMPARE );
    

    delete tdsLoo;   tdsLoo = NULL;
    delete krig;     krig = NULL;
    delete gpb;      gpb = NULL;

}

void TKrigingTest::testThrowExceptionEstimWithCov()
{
    TDataServer *tdsEstim=NULL;         
    TGPBuilder  *gpb=NULL;
    TKriging    *krig=NULL;
        
    // Prediction using a model based on:
    // no trend, matern3/2 correlation function, ML optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern3/2", "");
    gpb->findOptimalParameters("ML", 0, "neldermead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test.dat");

    //several realvalue
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1", "yEstim:vEstim", "y:P", ""), URANIE::Exceptions::UErrorExceptions);
    // realvalue not attribute
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1", "yEstim:vEstim", "tutu", ""), URANIE::Exceptions::UErrorExceptions);
    //input, not attribute
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "tutu", "yEstim:vEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    // too much attribute
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1:tutu", "yEstim:vEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    //output already existing
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1", "x1:vEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    //tds null
    TDataServer *tdsEstim2=NULL;         
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim2, "x1", "yEstim:vEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    //too much input 
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1:y", "yEstim:vEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    // too much output
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1", "yEstim:vEstim:tutu", "y", ""), URANIE::Exceptions::UErrorExceptions);
    // too few output
    CPPUNIT_ASSERT_THROW(krig->estimateWithCov(tdsEstim, "x1", "yEstim", "y", ""), URANIE::Exceptions::UErrorExceptions);
    
    delete tdsEstim;   tdsEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;
    
}

void TKrigingTest::testThrowExceptionEvalWrongAttributes()
{
    TDataServer *tdsEstim;
    TKriging    *krig;

    TGPBuilder  *gpb;

    URANIE::DataServer::TDataServer * tObs;
    std::string trainingFileName;
    std::string testFileName;
    std::string archiName;

    trainingFileName = "utf-1D-train.dat";
    testFileName = "utf-1D-train.dat";
    archiName = "x1";

    tObs = new TDataServer("tObs", "");
    tObs->fileDataRead(trainingFileName.c_str());

    gpb = new TGPBuilder(tObs, archiName.c_str(), "y", "matern3/2", "const", "const");
    gpb->findOptimalParameters("ML", 100, "Subplexe", 1000, "");
    krig = gpb->buildGP();

    tdsEstim = new TDataServer("tEstim", "");
    tdsEstim->fileDataRead(testFileName.c_str());

    // Too much input
    CPPUNIT_ASSERT_THROW(new TLauncher2(tdsEstim, krig, "x1:x2","yEstim:vEstim"), URANIE::Exceptions::UErrorExceptions);

    delete krig; krig=gpb->buildGP();
    // Too few input
    CPPUNIT_ASSERT_THROW(new TLauncher2(tdsEstim, krig, "","yEstim:vEstim"), URANIE::Exceptions::UErrorExceptions);

    delete krig; krig=gpb->buildGP();
    // Too much output
    CPPUNIT_ASSERT_THROW(new TLauncher2(tdsEstim, krig, "x1","yEstim:vEstim:Trululu"), URANIE::Exceptions::UErrorExceptions);

    delete krig; krig=gpb->buildGP();
    // Too few output
    CPPUNIT_ASSERT_THROW(new TLauncher2(tdsEstim, krig, "x1","yEstim"), URANIE::Exceptions::UErrorExceptions);

    delete krig; krig=gpb->buildGP();
    // unknow input
    CPPUNIT_ASSERT_THROW(new TLauncher2(tdsEstim, krig, "trululu","yEstim:vEstim"), URANIE::Exceptions::UErrorExceptions);

    delete tdsEstim;   tdsEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;
    delete tObs;       tObs = NULL;
}

void TKrigingTest::testestimWithCov()
{
    TDataServer *tdsEstim=NULL;         
    TGPBuilder  *gpb=NULL;
    TKriging    *krig=NULL;
        
    // Prediction using a model based on:
    // no trend, matern3/2 correlation function, ML optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern5/2", "");
    gpb->findOptimalParameters("ML", 100, "Subplexe", 1000,kTRUE,"optimpar=0.500508");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();

    TDataServer *tdsEstimFull = new TDataServer("tdsEstimFull","base de test");
    tdsEstimFull->fileDataRead("utf-1D-test.dat");
    tdsEstimFull->exportData("utf-1D-test_sub.dat","*","tdsEstimFull__n__iter__<=10");
    delete tdsEstimFull;
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test_sub.dat");

    //several realvalue
    krig->estimateWithCov(tdsEstim, "x1", "yEstim:vEstim", "y", "");

    //Contains the result branches
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("yEstim") );
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("vEstim") );
    
    // Contains the new branches
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_Residuals_"));
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_uncorrResiduals_"));
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_CovarianceMatrix_"));

    int nST=tdsEstim->getNPatterns();
    tdsEstim->getTuple()->Draw("_CovarianceMatrix_","","goff");
    double *cov=tdsEstim->getTuple()->GetV1();

    double ref_cov[100] = {0.00295317,-0.00233921,-6.09591e-05,0.000160471,0.000194573,0.000135925,0.000112492,0.000468832,0.00292221,0.000832449,
                           -0.00233921,0.0118647,0.000437124,-0.00011675,-0.00139897,-0.000976848,-0.000808388,-0.00330207,-0.00347769,-0.00576663,
                           -6.09591e-05,0.000437124,0.000192177,-3.03579e-06,-0.000768008,-0.000517773,-0.000426306,-0.000621681,-9.1809e-05,-0.000771487,
                           0.000160471,-0.00011675,-3.03579e-06,9.04871e-06,9.68962e-06,6.76903e-06,5.60206e-06,2.33511e-05,0.000149875,4.14666e-05,
                           0.000194573,-0.00139897,-0.000768008,9.68962e-06,0.0135291,0.00411414,0.00318315,0.00206014,0.000293075,0.00251819,
                           0.000135925,-0.000976848,-0.000517773,6.76903e-06,0.00411414,0.00226914,0.00181232,0.00143008,0.000204733,0.00175248,
                           0.000112492,-0.000808388,-0.000426306,5.60206e-06,0.00318315,0.00181232,0.00145365,0.00118247,0.000169437,0.00144957,
                           0.000468832,-0.00330207,-0.000621681,2.33511e-05,0.00206014,0.00143008,0.00118247,0.00362566,0.000705559,0.0050541,
                           0.00292221,-0.00347769,-9.1809e-05,0.000149875,0.000293075,0.000204733,0.000169437,0.000705559,0.00364509,0.00125191,
                           0.000832449,-0.00576663,-0.000771487,4.14666e-05,0.00251819,0.00175248,0.00144957,0.0050541,0.00125191,0.00767593};

    for(int i=0; i<nST*nST; i++){
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref_cov[i], cov[i], COV_PRECISION_COMPARE );
    }
    
    delete tdsEstim;   tdsEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;
    
}

void TKrigingTest::testestimWithCovLinearTrend()
{
    TDataServer *tdsEstim=NULL;         
    TGPBuilder  *gpb=NULL;
    TKriging    *krig=NULL;
        
    // Prediction using a model based on:
    // no trend, matern3/2 correlation function, ML optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern5/2", "linear");
    gpb->findOptimalParameters("ML", 100, "Subplexe", 1000,kTRUE,"optimpar=0.500508");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();

    TDataServer *tdsEstimFull = new TDataServer("tdsEstimFull","base de test");
    tdsEstimFull->fileDataRead("utf-1D-test.dat");
    tdsEstimFull->exportData("utf-1D-test_sub.dat","*","tdsEstimFull__n__iter__<=10");
    delete tdsEstimFull;
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test_sub.dat");

    //several realvalue
    krig->estimateWithCov(tdsEstim, "x1", "yEstim:vEstim", "y", "");

    //Contains the result branches
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("yEstim") );
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("vEstim") );
    
    // Contains the new branches
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_Residuals_"));
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_uncorrResiduals_"));
    CPPUNIT_ASSERT_EQUAL( kTRUE, tdsEstim->isAttribute("_CovarianceMatrix_"));

    int nST=tdsEstim->getNPatterns();
    tdsEstim->getTuple()->Draw("_CovarianceMatrix_","","goff");
    double *cov=tdsEstim->getTuple()->GetV1();

    double ref_cov[100] = {0.00401064,-0.00297731,-5.63766e-05,0.000216331,-0.000221291,5.90734e-05,5.53366e-05,0.000535197,0.00395596,0.000981267,
                           -0.00297731,0.0167376,0.00058123,-0.000146254,-0.0014398,-0.00120999,-0.00100913,-0.00453983,-0.00459972,-0.0079794,
                           -5.63766e-05,0.00058123,0.000287107,-2.54739e-06,-0.00102502,-0.000747445,-0.000618022,-0.000905351,-0.000100987,-0.00110748,
                           0.000216331,-0.000146254,-2.54739e-06,1.21691e-05,-1.63456e-05,1.64809e-06,1.75544e-06,2.57151e-05,0.000200168,4.75188e-05,
                           -0.000221291,-0.0014398,-0.00102502,-1.63456e-05,0.0181983,0.00557997,0.00431551,0.00262144,-2.17961e-05,0.00313132,
                           5.90734e-05,-0.00120999,-0.000747445,1.64809e-06,0.00557997,0.00327741,0.00262296,0.00198963,0.000166109,0.00239808,
                           5.53366e-05,-0.00100913,-0.000618022,1.75544e-06,0.00431551,0.00262296,0.00210888,0.00165297,0.000143109,0.00199346,
                           0.000535197,-0.00453983,-0.000905351,2.57151e-05,0.00262144,0.00198963,0.00165297,0.00534408,0.000867792,0.00739358,
                           0.00395596,-0.00459972,-0.000100987,0.000200168,-2.17961e-05,0.000166109,0.000143109,0.000867792,0.00505025,0.00156845,
                           0.000981267,-0.0079794,-0.00110748,4.75188e-05,0.00313132,0.00239808,0.00199346,0.00739358,0.00156845,0.0112063};


    for(int i=0; i<nST*nST; i++){
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref_cov[i], cov[i], COV_PRECISION_COMPARE );
    }
    
    delete tdsEstim;   tdsEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;
    
}


void TKrigingTest::testEval()
{
    TDataServer *tdsEstim;         
    TGPBuilder  *gpb;
    TKriging    *krig;
    TAttribute  *yEstim, *vEstim;  
    TLauncher2  *launcher;             
        
    // Prediction using a model based on:
    // no trend, matern3/2 correlation function, ML optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern3/2", "");
    gpb->findOptimalParameters("ML", 0, "neldermead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();

    yEstim = new TAttribute("yEstim");
    vEstim = new TAttribute("vEstim");
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test.dat");
    
    krig->addInput(tdsEstim->getAttribute("x1"));
    krig->setOutputs(2, yEstim, vEstim);
    launcher = new TLauncher2(tdsEstim, krig);
    launcher->solverLoop();
    
    tdsEstim->addAttribute("err2", "(y-yEstim)**2");
    tdsEstim->computeStatistic();
        
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.296729319548e+00, tdsEstim->getValue("yEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.046561305816e-02, tdsEstim->getValue("vEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.004204843655e+00, tdsEstim->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.824802423670e-01, tdsEstim->getAttribute("yEstim")->getStd(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.716201313241e-03, tdsEstim->getAttribute("err2")->getMean(), DBL_PRECISION_COMPARE );    
    
    delete launcher;   launcher = NULL;
    delete tdsEstim;   tdsEstim = NULL;
    delete vEstim;     vEstim = NULL;
    delete yEstim;     yEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;

    // Prediction using a model based on:
    // linear trend, matern3/2 correlation function, ReML optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern3/2", "linear");
    gpb->findOptimalParameters("ReML", 0, "neldermead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();

    yEstim = new TAttribute("yEstim");
    vEstim = new TAttribute("vEstim");
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test.dat");
    
    krig->addInput(tdsEstim->getAttribute("x1"));
    krig->setOutputs(2, yEstim, vEstim);
    launcher = new TLauncher2(tdsEstim, krig);
    launcher->solverLoop();
    
    tdsEstim->addAttribute("err2", "(y-yEstim)**2");
    tdsEstim->computeStatistic();
        
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.271480339456e+00, tdsEstim->getValue("yEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 8.358671541260e-03, tdsEstim->getValue("vEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.946774217166e-01, tdsEstim->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.728760401863e-01, tdsEstim->getAttribute("yEstim")->getStd(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.949137412128e-03, tdsEstim->getAttribute("err2")->getMean(), DBL_PRECISION_COMPARE );    
  
    delete launcher;   launcher = NULL;
    delete tdsEstim;   tdsEstim = NULL;
    delete vEstim;     vEstim = NULL;
    delete yEstim;     yEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;

    // Prediction using a model based on:
    // 1:x1:x1**2 trend, matern3/2 correlation function, Loo optimisation criterion
    gpb = new TGPBuilder(_tdsObs, "x1", "y", "matern3/2", "1:x1:x1**2");
    gpb->findOptimalParameters("Loo", 0, "neldermead");
    //gpb->findOptimalParameters("Loo", 0, "");
    krig = gpb->buildGP();

    yEstim = new TAttribute("yEstim");
    vEstim = new TAttribute("vEstim");
    
    tdsEstim = new TDataServer("tdsEstim", "");
    tdsEstim->fileDataRead("utf-1D-test.dat");
    
    krig->addInput(tdsEstim->getAttribute("x1"));
    krig->setOutputs(2, yEstim, vEstim);
    launcher = new TLauncher2(tdsEstim, krig);
    launcher->solverLoop();
    
    tdsEstim->addAttribute("err2", "(y-yEstim)**2");
    tdsEstim->computeStatistic();
        
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.235218008296e+00, tdsEstim->getValue("yEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.332818095859e-01, tdsEstim->getValue("vEstim",0), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.536745884343e-01, tdsEstim->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 8.046939306827e-01, tdsEstim->getAttribute("yEstim")->getStd(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.124286996122e-01, tdsEstim->getAttribute("err2")->getMean(), DBL_PRECISION_COMPARE );    
  
    delete launcher;   launcher = NULL;
    delete tdsEstim;   tdsEstim = NULL;
    delete vEstim;     vEstim = NULL;
    delete yEstim;     yEstim = NULL;
    delete krig;       krig = NULL;
    delete gpb;        gpb = NULL;
    
}


void TKrigingTest::testCleanInputsOutputs()
{

    TString correl = "gauss";
    Int_t nScreening=20, nMaxIterOpt=100;

    TGPBuilder *gpb = new TGPBuilder(_tdsObs,"x1","y",correl);
    gpb->findOptimalParameters("ML",nScreening, "BFGS", nMaxIterOpt);
    //gpb->findOptimalParameters("ML",nScreening, "", nMaxIterOpt);
    TKriging *kg = gpb->buildGP();

    for(int itest=0; itest<2; itest++)
    {
        stringstream ss; ss << "TEST_" << itest;
        TDataServer *tdsestim = new TDataServer(ss.str().c_str(),"base de test");

        CPPUNIT_ASSERT_EQUAL( 0, tdsestim->getNAttributes() );

        tdsestim->fileDataRead("utf-1D-test.dat");

        CPPUNIT_ASSERT_EQUAL( 2, tdsestim->getNAttributes() );

        CPPUNIT_ASSERT_EQUAL( kFALSE, tdsestim->isAttribute("yEstim") );
        CPPUNIT_ASSERT_EQUAL( kFALSE, tdsestim->isAttribute("vEstim") );

        TLauncher2 *lkri= new TLauncher2(tdsestim, kg, "x1", "yEstim:vEstim");
        lkri->solverLoop();

        CPPUNIT_ASSERT_EQUAL( 4, tdsestim->getNAttributes() );

        CPPUNIT_ASSERT_EQUAL( kTRUE, tdsestim->isAttribute("yEstim") );
        CPPUNIT_ASSERT_EQUAL( kTRUE, tdsestim->isAttribute("vEstim") );

        tdsestim->computeStatistic();

        CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.316848533032e+00, tdsestim->getValue("yEstim",0), DBL_PRECISION_COMPARE );
        CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.737730022300e-04, tdsestim->getValue("vEstim",0), DBL_PRECISION_COMPARE );
        CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.012799891907e+00, tdsestim->getAttribute("yEstim")->getMean(), DBL_PRECISION_COMPARE );
        CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.001182923164e+00, tdsestim->getAttribute("yEstim")->getStd(), DBL_PRECISION_COMPARE );

        TObjArray * inputList = kg->getInputList();
        TObjArray * outputList = kg->getOutputList();

        CPPUNIT_ASSERT_EQUAL( 1, inputList->GetEntries() );
        CPPUNIT_ASSERT_EQUAL( 2, outputList->GetEntries() );

        lkri->clean(); 

        CPPUNIT_ASSERT_EQUAL( 0, inputList->GetEntries() );
        CPPUNIT_ASSERT_EQUAL( 0, outputList->GetEntries() );

        delete lkri;       lkri = NULL;
        delete tdsestim;   tdsestim = NULL;

    }

    delete kg;      kg = NULL;
    delete gpb;     gpb = NULL;

}

