//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTTNisp.cxx
 \author Maud Salvoldi (OCC)
 \date   08/04/2011
 
 \brief 
 
 */

// ROOT
#include "TMath.h"

// cppunit
#include <iostream>
using namespace std;
#define DBL_PRECISION_COMPARE 1.e-12

//uranie
#include "TestTNisp.h"
#include "TDataServer.h"
#include "TSampling.h"
#include "TSampler.h"
#include "TUniformDistribution.h"
#include "TLogNormalDistribution.h"
#include "TBetaDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TLogUniformDistribution.h"
#include "TCauchyDistribution.h"
#include "TTriangularDistribution.h"
#include "TPetras.h"
#include "TLauncherFunction.h"
#include "uranie_commands.h"

TNispTest::TNispTest() :
        _tds(NULL), _tds2(NULL), _nisp(NULL)
{

}
TNispTest::~TNispTest()
{
   
}

void TNispTest::setUp()
{
    _tds = new URANIE::DataServer::TDataServer();
    _tds->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x1",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tds->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x2",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tds->addAttribute(
	new URANIE::DataServer::TUniformDistribution("x3",
						     -1. * TMath::Pi(), TMath::Pi()));
    _tds->getAttribute("x1")->delShare();
    _tds->getAttribute("x2")->delShare();
    _tds->getAttribute("x3")->delShare();

    _tds2 = new URANIE::DataServer::TDataServer();
    _nisp = new URANIE::Modeler::TNisp(_tds);
    _nisp->unsetLog();
}

void TNispTest::tearDown()
{

    if (_tds)
    {
        delete _tds;
        _tds = NULL;
    }
    if (_tds2)
    {
        delete _tds2;
        _tds2 = NULL;
    }
    
    if (_nisp)
    {
        delete _nisp;
        _nisp = NULL;
    }

    UranieTests::removeFile(gSystem->pwd(),"_dataserver_.root");
    UranieTests::removeFile(gSystem->pwd(),"OPT_DEFAULT.out");
    UranieTests::removeFile(gSystem->pwd(),"_tstOptCalNips__modeler_tdsFlowrate.out");

}

void TNispTest::testgetNAttributes()
{
    // verifiy the number of attributes
    CPPUNIT_ASSERT_EQUAL(3, _nisp->getNAttributes());
}
void TNispTest::testgetNPatterns()
{
    // contructor initialization of the parameter _ns
    CPPUNIT_ASSERT_EQUAL(0, _nisp->getNPatterns());

}
void TNispTest::testgetLevel()
{
    // contructor initialization of the paramter _level
    CPPUNIT_ASSERT_EQUAL(0, _nisp->getLevel());
}

void TNispTest::testsetLevel()
{
    // initialization of the parameter _level
    _nisp->setLevel(7);
    CPPUNIT_ASSERT_EQUAL(7, _nisp->getLevel());
}


void TNispTest::testGenerateSampleSmolyakClenshawCurtis()
{
    //generation of the sample: SmolyakClenshawCurtis of level 3
    _nisp->generateSample("SmolyakClenshawCurtis", 3);

    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 69);
    //test the level
    CPPUNIT_ASSERT_EQUAL(3, _nisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.14159265358979,
            (Double_t) _nisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.90245315213943,
            (Double_t) _nisp->_gvu->GetSample(69, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);
}
void TNispTest::testGenerateSamplePetras()
{
    //generation of the sample: Petra level 2
    _nisp->generateSample("Petras", 2);

    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 19);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, _nisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) _nisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) _nisp->_gvu->GetSample(19, 3),
            DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);
    
}
void TNispTest::testGenerateSampleSmolyakGauss()
{
    //generation of the sample: SmolyakGauss level 2
    _nisp->generateSample("SmolyakGauss", 2);

    //test the number of pattern
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 28);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, _nisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) _nisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.43346720558417,
            (Double_t) _nisp->_gvu->GetSample(28, 3), DBL_PRECISION_COMPARE);
    
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);

}
void TNispTest::testGenerateSampleSmolyakTrapeze()
{
    //generation of the sample: SmolyakTrapeze level 2
    _nisp->generateSample("SmolyakTrapeze", 2);

    //test the number of pattern
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 31);
    //test the level  
    CPPUNIT_ASSERT_EQUAL(2, _nisp->getLevel());
    //test the 1st value and the last value of the sample 
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) _nisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.35619449019234,
            (Double_t) _nisp->_gvu->GetSample(31, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);

}
void TNispTest::testGenerateSampleSmolyakFejer()
{
    //generation of the sample: SmolyakFejer level 2  
    _nisp->generateSample("SmolyakFejer", 2);

    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 31);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, _nisp->getLevel());
    //test the 1st value and the last value of the sample   
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) _nisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.90245315213943,
            (Double_t) _nisp->_gvu->GetSample(31, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);

}
void TNispTest::testGenerateSampleMonteCarlo()
{
    //generation of the sample: Monte Carlo of size 351
    _nisp->generateSample("MonteCarlo", 351);

    //test the number of pattern   
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 351);
    //test the level
    CPPUNIT_ASSERT_EQUAL(351, _nisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", 0),
            (Double_t) _nisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", 350),
            (Double_t) _nisp->_gvu->GetSample(351, 3), DBL_PRECISION_COMPARE);

}
void TNispTest::testGenerateSampleLhs()
{
    //generation of the sample: Lhs level 100
    _nisp->generateSample("Lhs", 100);

    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 100);
    //test the level  
    CPPUNIT_ASSERT_EQUAL(100, _nisp->getLevel());
    //test the 1st value and the last value of the sample 
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", 0),
            (Double_t) _nisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", 99),
            (Double_t) _nisp->_gvu->GetSample(100, 3), DBL_PRECISION_COMPARE);
}
void TNispTest::testGenerateSampleQuadrature()
{
    //generation of the sample: Quadrature size 2
    _nisp->generateSample("Quadrature", 2);

    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 27);
    //test the level   
    CPPUNIT_ASSERT_EQUAL(2, _nisp->getLevel());
    //test the 1st value and the last value of the sample  
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.43346720558417,
            (Double_t) _nisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.43346720558417,
            (Double_t) _nisp->_gvu->GetSample(3, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);

}
void TNispTest::testGenerateSampleQmcSobol()
{
    //generation of the sample: QmcSobol size 512
    _nisp->generateSample("QmcSobol", 512);

    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 512);
    //test the level   
    CPPUNIT_ASSERT_EQUAL(512, _nisp->getLevel());
    //test the 1st value and the last value of the sample    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", 0),
            (Double_t) _nisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", 511),
            (Double_t) _nisp->_gvu->GetSample(512, 3), DBL_PRECISION_COMPARE);

}
void TNispTest::throwExceptionGenerateSampleMethod()
{
    //Error in the name of the method used to compute the sample
    CPPUNIT_ASSERT_THROW(_nisp->generateSample("toto", 2),
            URANIE::Exceptions::UErrorExceptions);

}
void TNispTest::throwExceptionGenerateSampleLevel()
{
    //Error in the name of the method used to compute the sample
    CPPUNIT_ASSERT_THROW(_nisp->generateSample("Petras", -2),
            URANIE::Exceptions::UErrorExceptions);

}
void TNispTest::throwExceptionGenerateSampleBadOption()
{
    //Error in the name of option 'savegvx'
    CPPUNIT_ASSERT_THROW(_nisp->generateSample("Petras", 6, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TNispTest::throwExceptionGenerateSampleNotEmpty()
{
    Int_t nombre_simulations = 100;
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", nombre_simulations);
    fsampling->generateSample();
    _nisp->setSample("LHS");

    //The sample already exist
    CPPUNIT_ASSERT_THROW(_nisp->generateSample("Petras", 6),
            URANIE::Exceptions::UWarningExceptions);
    delete fsampling;
}
void TNispTest::testGenerateSampleOption()
{
    //build the sample with savinfof the gvx
    _nisp->generateSample("SmolyakClenshawCurtis", 6, "savegvx");

    // test is the gvx are saved
    CPPUNIT_ASSERT_NO_THROW(_tds->isAttribute("psi_x1"));
    CPPUNIT_ASSERT_NO_THROW(_tds->isAttribute("psi_x2"));
    CPPUNIT_ASSERT_NO_THROW(_tds->isAttribute("psi_x3"));
    
    
    // verify the type of the gvx attributes
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x1")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x2")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x3")->getOrigin(),
            TAttribute::kAttribute);
    // verify two data of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x1", 0),
            (Double_t) _nisp->_gvx->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x3", 5),
            (Double_t) _nisp->_gvx->GetSample(6, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);
    
}
void TNispTest::testsetSample()
{
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", 100);
    fsampling->generateSample();
    _nisp->setSample("LHS");

    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 100);
    //test the level  
    CPPUNIT_ASSERT_EQUAL(-1, _nisp->getLevel());
    delete fsampling;
}


void TNispTest::testsetSampleOptionWithComputedOutputs()
{
    // create the inputs sample
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", 100);
    fsampling->generateSample();
    
    // compute the outputs
    gROOT->LoadMacro("UserFunctions.C");
    URANIE::Launcher::TLauncherFunction* tlf = new URANIE::Launcher::TLauncherFunction(
            _tds, 
            "Ishigami",
            "", "Ishigami:Lineaire");
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    Double_t firstOutput1 = _tds->getValue("Ishigami", 0);
    Double_t lastOutput1 = _tds->getValue("Ishigami", 99);
    
    Double_t firstOutput2 = _tds->getValue("Lineaire", 0);
    Double_t lastOutput2 = _tds->getValue("Lineaire", 99);
    
    _nisp->setSample("LHS", "savegvx");

    // test is the gvx are saved
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x1"));
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x2"));
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x3"));
    
    // test if a psi_Ishigami for the output is not created
    CPPUNIT_ASSERT( ! _tds->isAttribute("psi_Ishigami"));
    // verify the type of the gvx attributes
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x1")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x2")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x3")->getOrigin(),
            TAttribute::kAttribute);
    // verify two data of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x1", 0),
            (Double_t) _nisp->_gvx->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x1", 99),
            (Double_t) _nisp->_gvx->GetSample(100, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x2", 0),
            (Double_t) _nisp->_gvx->GetSample(1, 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x2", 99),
            (Double_t) _nisp->_gvx->GetSample(100, 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x3", 0),
            (Double_t) _nisp->_gvx->GetSample(1, 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x3", 99),
            (Double_t) _nisp->_gvx->GetSample(100, 3), DBL_PRECISION_COMPARE);
    
    // verify two data of the outputs
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Ishigami", 0),
            firstOutput1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Ishigami", 99),
            lastOutput1, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Lineaire", 0),
            firstOutput2, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("Lineaire", 99),
            lastOutput2, DBL_PRECISION_COMPARE);
    
    delete fsampling;
    delete tlf;
}


void TNispTest::testsetSampleOption()
{
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", 100);
    fsampling->generateSample();
    _nisp->setSample("LHS", "savegvx");

    // test is the gvx are saved
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x1"));
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x2"));
    CPPUNIT_ASSERT(_tds->isAttribute("psi_x3"));
    // verify the type of the gvx attributes
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x1")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x2")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(_tds->getAttribute("psi_x3")->getOrigin(),
            TAttribute::kAttribute);
    // verify two data of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x1", 0),
            (Double_t) _nisp->_gvx->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("psi_x3", 99),
            (Double_t) _nisp->_gvx->GetSample(100, 3), DBL_PRECISION_COMPARE);

    delete fsampling;
}

void TNispTest::testsetSampleOptionWithImportedDatas()
{
    /////////////////////////////////////
    // Lecture des simulations realisees
    /////////////////////////////////////
    TDataServer * tds = new TDataServer("PolynomeMultiD", "Polynômes de Chaos");
    tds->fileDataRead("Quadrature.dat");
    
    // Creation d'un nouveau TDataServer
    TDataServer * tdsNisp = new TDataServer("PolynomeMultiD", "Polynômes de Chaos - Nisp");
    tdsNisp->addAttribute( new TNormalDistribution("x1", 1., 0.5));
    tdsNisp->addAttribute( new TUniformDistribution("x2", 1, 2));
    tdsNisp->addAttribute( new TBetaDistribution("x3",2., 2.));    
    tdsNisp->addAttribute("uWeight");
    tdsNisp->addAttribute("y");
    tdsNisp->getAttribute("x1")->delShare();
    tdsNisp->getAttribute("x2")->delShare();
    tdsNisp->getAttribute("x3")->delShare();
        
    //////////////////
    // Objet TNisp  //
    //////////////////
    URANIE::Modeler::TNisp * nisp = new URANIE::Modeler::TNisp(tdsNisp);  
    nisp->setLog();
    // Une astuce pour transférer les données entre les 2 TDataServer
    // Il suffit de passer l'adresse du TNtuple !!!
    tdsNisp->setTuple( tds->getTuple() );
    
    Double_t firstOutput = tdsNisp->getValue("y", 0);
    Double_t lastOutput = tdsNisp->getValue("y", 124);
    
    Double_t firstWeight = tdsNisp->getValue("uWeight", 0);
    Double_t lastWeight = tdsNisp->getValue("uWeight", 124);
    
    nisp->setSample("Quadrature","savegvx");

    // test is the gvx are saved
    CPPUNIT_ASSERT(tdsNisp->isAttribute("psi_x1"));
    CPPUNIT_ASSERT(tdsNisp->isAttribute("psi_x2"));
    CPPUNIT_ASSERT(tdsNisp->isAttribute("psi_x3"));
    
    // test if a psi_Ishigami for the output is not created
    CPPUNIT_ASSERT( ! tdsNisp->isAttribute("psi_y"));
    // verify the type of the gvx attributes
    CPPUNIT_ASSERT_EQUAL(tdsNisp->getAttribute("psi_x1")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(tdsNisp->getAttribute("psi_x2")->getOrigin(),
            TAttribute::kAttribute);
    CPPUNIT_ASSERT_EQUAL(tdsNisp->getAttribute("psi_x3")->getOrigin(),
            TAttribute::kAttribute);
    // verify two data of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x1", 0),
            (Double_t) nisp->_gvx->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x1", 124),
            (Double_t) nisp->_gvx->GetSample(125, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x2", 0),
            (Double_t) nisp->_gvx->GetSample(1, 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x2", 124),
            (Double_t) nisp->_gvx->GetSample(125, 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x3", 0),
            (Double_t) nisp->_gvx->GetSample(1, 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("psi_x3", 124),
            (Double_t) nisp->_gvx->GetSample(125, 3), DBL_PRECISION_COMPARE);
    
    // verify two data of the outputs
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("y", 0),
            firstOutput, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("y", 124),
            lastOutput, DBL_PRECISION_COMPARE);
    
    // verify two data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("uWeight", 0),
            firstWeight, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("uWeight", 124),
            lastWeight, DBL_PRECISION_COMPARE);
   
    
    delete nisp;
    delete tdsNisp;
    delete tds;
}



void TNispTest::testsetSamplePetras()
{
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(_tds,"",8); 
    sampler->generateSample();
    _nisp->setSample("Petras");
    _nisp->setLevel(8);
    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 495);
    //test the level  
    CPPUNIT_ASSERT_EQUAL(8, _nisp->getLevel());   
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("uWeight", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);
    
    delete sampler;
}

void TNispTest::testsetSamplePetrasWithWeight()
{
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(_tds,"",8); 
    sampler->generateSample();
    _tds->getAttribute("uWeight")->SetName("toto");
    _tds->getTuple()->GetLeaf("uWeight")->SetName("toto");
    _nisp->setSample("Petras,toto");
    _nisp->setLevel(8);
    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), 495);
    //test the level  
    CPPUNIT_ASSERT_EQUAL(8, _nisp->getLevel());
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds->isAttribute("toto"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("toto", 0),
             _nisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("toto", _tds->getNPatterns()-1),
             _nisp->_gvx->w[_tds->getNPatterns()], DBL_PRECISION_COMPARE);
    
    delete sampler;
}

void TNispTest::testSetSampleLhsAfterGenerationOfPetrasSample()
{
    Int_t level = 8;
    _nisp->generateSample("Petras",level);
    
    Int_t nSize = 100;
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", nSize);
    fsampling->generateSample();
    _nisp->setSample("LHS");
    //test the number of pattern  
    CPPUNIT_ASSERT_EQUAL(_tds->getNPatterns(), nSize);
    
    // verify to data of x1
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", 0),
             _nisp->_gvu->GetSample(1,1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", nSize/2 -1),
             _nisp->_gvu->GetSample(nSize/2,1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x1", nSize-1),
             _nisp->_gvu->GetSample(nSize,1), DBL_PRECISION_COMPARE);
    
    // verify to data of x2
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x2", 0),
             _nisp->_gvu->GetSample(1,2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x2", nSize/2 -1),
             _nisp->_gvu->GetSample(nSize/2,2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x2", nSize-1),
             _nisp->_gvu->GetSample(nSize,2), DBL_PRECISION_COMPARE);
    
    
    // verify to data of x3
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", 0),
             _nisp->_gvu->GetSample(1,3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", nSize/2 -1),
             _nisp->_gvu->GetSample(nSize/2,3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds->getValue("x3", nSize-1),
             _nisp->_gvu->GetSample(nSize,3), DBL_PRECISION_COMPARE);
    
    delete fsampling;
}



void TNispTest::testTNispLawNormal()
{
    URANIE::DataServer::TNormalDistribution *tnd =
            new URANIE::DataServer::TNormalDistribution("n", 0.0, 1.0);
    _tds2->addAttribute(tnd);
    tnd->delShare();
    URANIE::DataServer::TNormalDistribution *tnd2 =
            new URANIE::DataServer::TNormalDistribution("n2", 0.0, 1.0);
    tnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLegendre);
    _tds2->addAttribute(tnd2);
    tnd2->delShare();
    URANIE::DataServer::TNormalDistribution *tnd3 =
            new URANIE::DataServer::TNormalDistribution("n3", 0.0, 1.0);
    tnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    _tds2->addAttribute(tnd3);
    tnd3->delShare();
    URANIE::Modeler::TNisp *tnisp = NULL;
    CPPUNIT_ASSERT_NO_THROW(tnisp = new URANIE::Modeler::TNisp(_tds2));
    
    tnisp->generateSample("Petras", 2);
    
    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds2->getNPatterns(), 19);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, tnisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) tnisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) tnisp->_gvu->GetSample(19, 3),
            DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds2->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 0),
             tnisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 5),
             tnisp->_gvx->w[6], DBL_PRECISION_COMPARE);
    
    delete tnisp;
}

void TNispTest::testTNispLawUniform()
{
    URANIE::DataServer::TUniformDistribution *tud =
            new URANIE::DataServer::TUniformDistribution("n");
    _tds2->addAttribute(tud);
    tud->delShare();
    URANIE::DataServer::TUniformDistribution *tud2 =
            new URANIE::DataServer::TUniformDistribution("n2",-1,1);
    tud2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kHermite);
    _tds2->addAttribute(tud2);
    tud2->delShare();
    URANIE::DataServer::TUniformDistribution *tud3 =
            new URANIE::DataServer::TUniformDistribution("n3",-1,1);
    tud2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    _tds2->addAttribute(tud3);
    tud3->delShare();
    URANIE::Modeler::TNisp *tnisp = NULL;
    CPPUNIT_ASSERT_NO_THROW(tnisp = new URANIE::Modeler::TNisp(_tds2));
    
    tnisp->generateSample("Petras", 2);
    
    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds2->getNPatterns(), 19);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, tnisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, (Double_t) tnisp->_gvu->GetSample(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, (Double_t) tnisp->_gvu->GetSample(19, 3),
            DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds2->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 0),
             tnisp->_gvu->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 5),
             tnisp->_gvu->w[6], DBL_PRECISION_COMPARE);
    
    delete tnisp;
} 

void TNispTest::testTNispLawLogUniform()
{

    URANIE::DataServer::TLogUniformDistribution *tlnd =
            new URANIE::DataServer::TLogUniformDistribution("n");
    _tds2->addAttribute(tlnd);
    tlnd->delShare();
    URANIE::DataServer::TLogUniformDistribution *tlnd2 =
            new URANIE::DataServer::TLogUniformDistribution("n2");
    tlnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLegendre);
    _tds2->addAttribute(tlnd2);
    tlnd2->delShare();
    URANIE::DataServer::TLogUniformDistribution *tlnd3 =
            new URANIE::DataServer::TLogUniformDistribution("n3");
    tlnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    _tds2->addAttribute(tlnd3);
    tlnd3->delShare();

    URANIE::Modeler::TNisp *tnisp = NULL;
    CPPUNIT_ASSERT_NO_THROW(tnisp = new URANIE::Modeler::TNisp(_tds2));

    tnisp->generateSample("Petras", 2);

    // test the number of pattern
    CPPUNIT_ASSERT_EQUAL(_tds2->getNPatterns(), 19);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, tnisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.316227766016838, (Double_t) tnisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.316227766016838, (Double_t) tnisp->_gvu->GetSample(19, 3), DBL_PRECISION_COMPARE);

    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds2->isAttribute("uWeight"));

    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 0),
             tnisp->_gvu->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 5),
             tnisp->_gvu->w[6], DBL_PRECISION_COMPARE);

    delete tnisp;
}

void TNispTest::testTNispLawLogNormal()
{
    URANIE::DataServer::TLogNormalDistribution *tlnd =
            new URANIE::DataServer::TLogNormalDistribution("n");
    _tds2->addAttribute(tlnd);
    tlnd->delShare();
    URANIE::DataServer::TLogNormalDistribution *tlnd2 =
            new URANIE::DataServer::TLogNormalDistribution("n2");
    tlnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLegendre);
    _tds2->addAttribute(tlnd2);
    tlnd2->delShare();
    URANIE::DataServer::TLogNormalDistribution *tlnd3 =
            new URANIE::DataServer::TLogNormalDistribution("n3");
    tlnd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    _tds2->addAttribute(tlnd3);
    tlnd3->delShare();
    URANIE::Modeler::TNisp *tnisp = NULL;
    CPPUNIT_ASSERT_NO_THROW(tnisp = new URANIE::Modeler::TNisp(_tds2));
    
    tnisp->generateSample("Petras", 2);
    
    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds2->getNPatterns(), 19);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, tnisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0091505172861615, (Double_t) tnisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0091505172861615, (Double_t) tnisp->_gvu->GetSample(19, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds2->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 0),
             tnisp->_gvu->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 5),
             tnisp->_gvu->w[6], DBL_PRECISION_COMPARE);
    
    
    delete tnisp;
}

void TNispTest::testTNispLawBeta()
{
    URANIE::DataServer::TBetaDistribution *tbd =
            new URANIE::DataServer::TBetaDistribution("n");
    _tds2->addAttribute(tbd);
    tbd->delShare();
    URANIE::DataServer::TBetaDistribution *tbd2 =
            new URANIE::DataServer::TBetaDistribution("n2");
    tbd2->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLegendre);
    tbd2->delShare();
    _tds2->addAttribute(tbd2);
    URANIE::DataServer::TBetaDistribution *tbd3 =
            new URANIE::DataServer::TBetaDistribution("n3");
    tbd3->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    _tds2->addAttribute(tbd3);
    tbd3->delShare();
    URANIE::DataServer::TBetaDistribution *tbd4 =
            new URANIE::DataServer::TBetaDistribution("n4");
    tbd4->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kHermite);
    _tds2->addAttribute(tbd4);
    tbd4->delShare();
    URANIE::Modeler::TNisp *tnisp = NULL;
    CPPUNIT_ASSERT_NO_THROW(tnisp = new URANIE::Modeler::TNisp(_tds2));
    
    tnisp->generateSample("Petras", 2);
    
    // test the number of pattern 
    CPPUNIT_ASSERT_EQUAL(_tds2->getNPatterns(), 33);
    //test the level
    CPPUNIT_ASSERT_EQUAL(2, tnisp->getLevel());
    //test the 1st value and the last value of the sample
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.499999999516932, (Double_t) tnisp->_gvu->GetSample(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0310140674507116, (Double_t) tnisp->_gvu->GetSample(19, 3), DBL_PRECISION_COMPARE);
    
    // test if uWeight is an attribute
    CPPUNIT_ASSERT(_tds2->isAttribute("uWeight"));
    
    // verify to data of the weights
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 0),
             tnisp->_gvx->w[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tds2->getValue("uWeight", 5),
             tnisp->_gvx->w[6], DBL_PRECISION_COMPARE);
    
    delete tnisp;
}

void TNispTest::throwExceptionsetSampleEmpty()
{
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds,
            "lhs", 100);

    CPPUNIT_ASSERT_THROW(_nisp->setSample("LHS"),
            URANIE::Exceptions::UErrorExceptions);
    delete fsampling;
}
void TNispTest::throwExceptionsetSampleBadOption()
{
    URANIE::Sampler::TSampling *fsampling = new URANIE::Sampler::TSampling(_tds, "lhs", 100);
    fsampling->generateSample();

    //bad name option
    CPPUNIT_ASSERT_THROW(_nisp->setSample("LHS", "toto"), URANIE::Exceptions::UWarningExceptions);
    delete fsampling;
}
void TNispTest::throwExceptionTNispTdsNull()
{
  delete _tds2;
  _tds2 = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TNisp(_tds2),
            URANIE::Exceptions::UErrorExceptions);
    
    _tds2 = new URANIE::DataServer::TDataServer();    
}

void TNispTest::throwExceptionTNispLawUnknown()
{
    _tds2->addAttribute(
            new URANIE::DataServer::TCauchyDistribution());
    _tds2->getAttribute("")->delShare();
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TNisp(_tds2),
            URANIE::Exceptions::UErrorExceptions);

}

void TNispTest::throwExceptionTNispLawLogTriangular()
{
    _tds2->addAttribute(
            new URANIE::DataServer::TLogTriangularDistribution("c", 0.03, 10.));
    _tds2->getAttribute("c")->delShare();
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TNisp(_tds2),
            URANIE::Exceptions::UErrorExceptions);

}
void TNispTest::throwExceptionTNispTriangular()
{
    _tds2->addAttribute(
            new URANIE::DataServer::TTriangularDistribution("d", 0.2, 9.));
    _tds2->getAttribute("d")->delShare();
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TNisp(_tds2),
            URANIE::Exceptions::UErrorExceptions);

}
void TNispTest::throwExceptionTNispTdsNotAttribute()
{
    URANIE::DataServer::TAttribute *px = new URANIE::DataServer::TAttribute(
            "x");
    px->setOrigin(TAttribute::kInternal);
    px->delShare();
    _tds2->addAttribute(px);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TNisp(_tds2),
            URANIE::Exceptions::UErrorExceptions);

}

void TNispTest::throwExceptionTNispSetLevelNegative()
{
    // negative initialization of the parameter _level
    CPPUNIT_ASSERT_THROW(_nisp->setLevel(-2),URANIE::Exceptions::UErrorExceptions);
}

void TNispTest::throwExceptionTNispReSetLevel()
{
    // initialization of the parameter _level
    _nisp->setLevel(2);
    // reinitialization of the parameter _level
    CPPUNIT_ASSERT_THROW(_nisp->setLevel(5),URANIE::Exceptions::UWarningExceptions);
}

void TNispTest::throwExceptionTNispSetSamplePetrasWithWeight()
{
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(_tds,"",8); 
    sampler->generateSample();
    CPPUNIT_ASSERT_THROW(_nisp->setSample("Petras,utoto"),URANIE::Exceptions::UErrorExceptions);
    
    delete sampler;
}

void TNispTest::throwExceptionTNispSetSamplePetrasWithDifferentLeaf()
{
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(_tds,"",8); 
    sampler->generateSample();
    _tds->getAttribute("uWeight")->SetName("utoto");
    CPPUNIT_ASSERT_THROW(_nisp->setSample("Petras,utoto"),URANIE::Exceptions::UErrorExceptions);
       
    delete sampler;
}

void TNispTest::throwExceptionTNispSetSamplePetrasWithDifferentLeaf2()
{
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(_tds,"",8); 
    sampler->generateSample();
    _tds->getTuple()->GetLeaf("uWeight")->SetName("utoto");
    CPPUNIT_ASSERT_THROW(_nisp->setSample("Petras"),URANIE::Exceptions::UErrorExceptions);
        
    delete sampler;
}

void TNispTest::testPetrasAndNispSampler()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x2",
                    0, 1));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
    
    URANIE::DataServer::TDataServer* tds2 = new URANIE::DataServer::TDataServer();
    tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds2->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x2",
                    0, 1));
    tds2->getAttribute("x1")->delShare();
    tds2->getAttribute("x2")->delShare();
    
    Int_t level = 5;
    // generation of the sampling with TNisp
    URANIE::Modeler::TNisp *nisp = new URANIE::Modeler::TNisp(tds);
    nisp->unsetLog();
    nisp->generateSample("Petras", level);
    
    // generation of the sampling with TPetras
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(tds2,"",level); 
    sampler->generateSample();

    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(33, npat);

    Int_t npatHalf = (Int_t)(npat/2);

    CPPUNIT_ASSERT( npatHalf < npat );

    // verification that the sampling are identical
    CPPUNIT_ASSERT_EQUAL(tds->getNPatterns(), tds2->getNPatterns());
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("uWeight", 0),
             tds2->getValue("uWeight", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("uWeight", npatHalf),
             tds2->getValue("uWeight", npatHalf), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("uWeight", npat-1),
             tds2->getValue("uWeight", npat-1), DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x1", 0),
             tds2->getValue("x1", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x1", npatHalf),
             tds2->getValue("x1", npatHalf), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x1", npat-1),
             tds2->getValue("x1", npat-1), DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x2", 0),
             tds2->getValue("x2", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x2", npatHalf),
             tds2->getValue("x2", npatHalf), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds->getValue("x2", npat-1),
             tds2->getValue("x2", npat-1), DBL_PRECISION_COMPARE);
    
    delete sampler;
    delete nisp;
    delete tds;
    delete tds2;
    
}


void TNispTest::throwExceptionTNispNormalAttributeUnknownStochasticBasis()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    URANIE::DataServer::TNormalDistribution* tud = new URANIE::DataServer::TNormalDistribution("x1",0,1);
    tud->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kUnknownBasis);
    tds->addAttribute( tud );
    tud->delShare();
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TNisp nisp(tds),URANIE::Exceptions::UErrorExceptions);
    
    delete tds;
}

void TNispTest::throwExceptionTNispUniformAttributeUnknownStochasticBasis()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    URANIE::DataServer::TUniformDistribution* tud = new URANIE::DataServer::TUniformDistribution("x1",0,1);
    tud->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kUnknownBasis);
    tds->addAttribute( tud );
    tud->delShare();
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TNisp nisp(tds),URANIE::Exceptions::UErrorExceptions);
    
    delete tds;
}

void TNispTest::throwExceptionTNispBetaAttributeUnknownStochasticBasis()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    URANIE::DataServer::TBetaDistribution* tud = new URANIE::DataServer::TBetaDistribution("x1",1,1);
    tud->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kUnknownBasis);
    tds->addAttribute( tud );
    tud->delShare();
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TNisp nisp(tds),URANIE::Exceptions::UErrorExceptions);
    
    delete tds;
}

void TNispTest::throwExceptionTNispLogNormalAttributeUnknownStochasticBasis()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    URANIE::DataServer::TLogNormalDistribution* tud = new URANIE::DataServer::TLogNormalDistribution("x1");
    tud->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kUnknownBasis);
    tds->addAttribute( tud );
    tud->delShare();
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TNisp nisp(tds),URANIE::Exceptions::UErrorExceptions);
    
    delete tds;
}

