//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTKNN.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Wed Jul  8 22:41:16 2015

 \brief  Fichier de test de la classe URANIE::DataServer::TkNN

 */

#define DBL_PRECISION_COMPARE 1.e-10
// Data Header
#include "TestTkNN.h"
#include "TString.h"
#include <iostream>
#include <ctime>

//ROOT
#include "TMath.h"
#include "TLeaf.h"

#include "uranie_commands.h"
using namespace std;

TkNNTest::TkNNTest() :
        _tdsR(NULL), _tdsQ(NULL),
        _sX("x"), _sY("Y")
{
    _workingTestPath = gSystem->pwd();
}

TkNNTest::~TkNNTest()
{

    if (_tdsR != NULL) {
        delete _tdsR;   _tdsR = NULL;
    }
    if (_tdsQ != NULL) {
        delete _tdsQ; _tdsQ = NULL;
    }
}

////////////////////////////////
///                          ///
///   setUp / tearDown       ///
///                          ///
////////////////////////////////

void TkNNTest::setUp()
{
    _tdsR = new URANIE::DataServer::TDataServer("tdsRef", "Ref Data to kNN");
    _tdsR->fileDataRead("sinusCardinal_Reference_1D.dat");

    _tdsQ = new URANIE::DataServer::TDataServer("tdsQuery", "Query Data to kNN");
    _tdsQ->fileDataRead("sinusCardinal_Query_1D.dat");
}

void TkNNTest::tearDown()
{
    delete _tdsR; _tdsR = NULL;
    delete _tdsQ; _tdsQ = NULL;

    UranieTests::removeFile(_workingTestPath.Data(),"uranie_kNN_ishigami.C");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.toto");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.pmml");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.C");
    UranieTests::removeFile(_workingTestPath.Data(),"filename.py");
}

//////////////////////////////////
////                          ////
////         Tests            ////
////                          ////
//////////////////////////////////

void TkNNTest::testTkNNConstructor()
{
    cout << endl;

    _sY = "SinC";
    Int_t nK = 5;
    TString sYHat1 = Form("%shat%d", _sY.Data(), nK);

    CPPUNIT_ASSERT_EQUAL(500, _tdsR->getNPatterns());
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sX.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sY.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat1.Data()) == kFALSE);


    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX, _sY, nK);
    TkNN->unsetLog();

    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());
    TkNN->setK ( nK+3);
    CPPUNIT_ASSERT_EQUAL(nK+3, TkNN->getK());
    TkNN->setK ( nK);
    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());


    // Default values

    CPPUNIT_ASSERT_EQUAL(1000, _tdsQ->getNPatterns());
    CPPUNIT_ASSERT(_tdsQ->isAttribute(_sX.Data ()));
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat1.Data ()) == kFALSE);

    Int_t nX2 = _tdsQ->getNAttributes ();
    Int_t nS2 = _tdsQ->getNPatterns();

    TkNN->setQueryPoints(_tdsQ);
    TkNN->estimate();
    delete TkNN;
    TkNN = NULL;

    CPPUNIT_ASSERT_EQUAL(nX2+1, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nS2,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat1.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat1.Data()));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  0), _tdsQ->getValue(sYHat1,  0), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  1), _tdsQ->getValue(sYHat1,  1), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  7), _tdsQ->getValue(sYHat1,  7), 2e-02);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 15), _tdsQ->getValue(sYHat1, 15), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 16), _tdsQ->getValue(sYHat1, 16), 2e-03);

    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.04209425708, _tdsQ->getValue(sYHat1,  0), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02084818178, _tdsQ->getValue(sYHat1,  1), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16115863621, _tdsQ->getValue(sYHat1,  7), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.03131272643, _tdsQ->getValue(sYHat1, 15), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.05582854524, _tdsQ->getValue(sYHat1, 16), 1e-04);

    // Construction with architecture

    nK += 5;
    TString sYHat2 = Form("%shat%d", _sY.Data(), nK);

    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat2.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat2.Data()) == kFALSE);

    TkNN = new URANIE::Modeler::TkNN(_tdsR, Form("%s,%d,%s", _sX.Data(), nK, _sY.Data()));
    TkNN->unsetLog();

    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());
    TkNN->setK ( nK+3);
    CPPUNIT_ASSERT_EQUAL(nK+3, TkNN->getK());
    TkNN->setK ( nK);
    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());

    TkNN->setQueryPoints(_tdsQ);
    TkNN->estimate();
    delete TkNN;
    TkNN = NULL;

    CPPUNIT_ASSERT_EQUAL(nX2+2, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nS2,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat2.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat2.Data ()));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  0), _tdsQ->getValue(sYHat1,  0), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  1), _tdsQ->getValue(sYHat1,  1), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  7), _tdsQ->getValue(sYHat1,  7), 6e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 15), _tdsQ->getValue(sYHat1, 15), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 16), _tdsQ->getValue(sYHat1, 16), 2e-03);

    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.04006119444, _tdsQ->getValue(sYHat2,  0), 1e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02084818178, _tdsQ->getValue(sYHat2,  1), 1e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16115863621, _tdsQ->getValue(sYHat2,  7), 5e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.03131272643, _tdsQ->getValue(sYHat2, 15), 1e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.05582854524, _tdsQ->getValue(sYHat2, 16), 1e-03);

}

void TkNNTest::testTkNNEstimateCPU()
{
    Int_t nK = 5;
    _sY = "SinC";

    TString sYHat = Form("%shat%d", _sY.Data(), nK);

    CPPUNIT_ASSERT_EQUAL(500, _tdsR->getNPatterns());
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sX.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sY.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()) == kFALSE);

    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX, _sY, nK);
    TkNN->unsetLog();

    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());

    // Default values

    CPPUNIT_ASSERT_EQUAL(1000, _tdsQ->getNPatterns());
    CPPUNIT_ASSERT(_tdsQ->isAttribute(_sX.Data ()));
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data ()) == kFALSE);

    Int_t nX2 = _tdsQ->getNAttributes ();
    Int_t nS2 = _tdsQ->getNPatterns();

    TkNN->setQueryPoints(_tdsQ);
    TkNN->estimate("",kFALSE);
    delete TkNN;
    TkNN = NULL;

    CPPUNIT_ASSERT_EQUAL(nX2+1, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nS2,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  0), _tdsQ->getValue(sYHat,  0), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  1), _tdsQ->getValue(sYHat,  1), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  7), _tdsQ->getValue(sYHat,  7), 2e-02);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 15), _tdsQ->getValue(sYHat, 15), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 16), _tdsQ->getValue(sYHat, 16), 2e-03);

    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.04209425708, _tdsQ->getValue(sYHat,  0), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02084818178, _tdsQ->getValue(sYHat,  1), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16115863621, _tdsQ->getValue(sYHat,  7), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.03131272643, _tdsQ->getValue(sYHat, 15), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.05582854524, _tdsQ->getValue(sYHat, 16), 1e-04);
}

void TkNNTest::testTkNNEstimateFlowrate()
{
    Int_t nK = 2;

    delete _tdsR; _tdsR = NULL;
    delete _tdsQ; _tdsQ = NULL;

    _tdsR = new URANIE::DataServer::TDataServer("tdsRef", "Ref Data to kNN");
    _tdsR->fileDataRead("flowrate_sampler_launcher_500.dat");


    _tdsQ = new URANIE::DataServer::TDataServer("tdsQuery", "Query Data to kNN");
    _tdsQ->fileDataRead("flowrateUniformDesignReduced.dat");

    _sY = "yhat";
    _sX = "rw:r:tu:tl:hu:hl:l:kw";

    TString sYHat = Form("%shat%d", _sY.Data(), nK);

    CPPUNIT_ASSERT_EQUAL(500, _tdsR->getNPatterns());
    CPPUNIT_ASSERT(_tdsR->isAttribute("rw"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("r"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("tu"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("tl"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("hu"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("hl"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("l"));
    CPPUNIT_ASSERT(_tdsR->isAttribute("kw"));
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sY.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);

    CPPUNIT_ASSERT_EQUAL(24, _tdsQ->getNPatterns());
    CPPUNIT_ASSERT(_tdsQ->isAttribute("rw"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("r"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("tu"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("tl"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("hu"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("hl"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("l"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute("kw"));
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(_sY.Data()) == kFALSE);

    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX, _sY, nK);
    TkNN->unsetLog();

    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());

    // Default values
    Int_t nXQ = _tdsQ->getNAttributes ();
    Int_t nSQ = _tdsQ->getNPatterns();

    TkNN->setQueryPoints(_tdsQ);
    TkNN->estimate("", kFALSE);

    CPPUNIT_ASSERT_EQUAL(nXQ+1, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nSQ,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()));

    Double_t dvalCrt;
    _tdsQ->getTuple()->SetBranchAddress(sYHat, &dvalCrt);
    const Int_t nCmpValues = 24;
    CPPUNIT_ASSERT_EQUAL(nCmpValues, nSQ);
    Double_t vecRefK2[nCmpValues] =
    {  50.236511,  73.281899, 108.769758, 50.227536, 40.209178, 62.560744, 57.412167, 88.974033,
      174.384776,  53.124686,  95.643589, 83.578152, 94.394073, 80.807325, 57.402927, 52.664986,
       41.186605, 172.234948, 114.092378, 57.716301, 73.266212, 36.169852, 88.367790, 47.485553};
    for (Int_t i = 0; i < nCmpValues; i++) {
        _tdsQ->getTuple()->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(vecRefK2[i], dvalCrt, 1.0e-6);
    }

    nK = 10;
    TkNN->setK(nK);
    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());
    sYHat = Form("%shat%d", _sY.Data(), nK);
    TkNN->estimate("", kFALSE);
    delete TkNN;
    TkNN = NULL;

    CPPUNIT_ASSERT_EQUAL(nXQ+2, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nSQ,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()));

    _tdsQ->getTuple()->SetBranchAddress(sYHat, &dvalCrt);

    Double_t vecRefK10[nCmpValues] =
    {  67.019733, 97.659094, 101.007822, 67.013709, 54.892643, 97.709165, 67.697731, 85.180143,
      107.908159, 61.107450,  79.115317, 92.266333, 78.990866, 91.636214, 67.702886, 68.323785,
       60.292479, 97.889227,  93.838176, 67.689036, 97.725845, 60.151722, 84.938442, 65.271641};
    for (Int_t i = 0; i < nCmpValues; i++) {
        _tdsQ->getTuple()->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(vecRefK10[i], dvalCrt, 1.0e-6);
    }
}


#ifdef WITH_GPUANN
void TkNNTest::testTkNNEstimateGPU()
{
    Int_t nK = 5;
    _sY = "SinC";

    TString sYHat = Form("%shat%d", _sY.Data(), nK);

    CPPUNIT_ASSERT_EQUAL(500, _tdsR->getNPatterns());
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sX.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(_sY.Data()));
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()) == kFALSE);

    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX, _sY, nK);
    TkNN->unsetLog();

    CPPUNIT_ASSERT_EQUAL(nK, TkNN->getK());

    // Default values

    CPPUNIT_ASSERT_EQUAL(1000, _tdsQ->getNPatterns());
    CPPUNIT_ASSERT(_tdsQ->isAttribute(_sX.Data ()));
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data ()) == kFALSE);

    Int_t nX2 = _tdsQ->getNAttributes ();
    Int_t nS2 = _tdsQ->getNPatterns();

    TkNN->setQueryPoints(_tdsQ);
    TkNN->estimate("",kTRUE);
    delete TkNN;
    TkNN = NULL;

    CPPUNIT_ASSERT_EQUAL(nX2+1, _tdsQ->getNAttributes ());
    CPPUNIT_ASSERT_EQUAL(nS2,   _tdsQ->getNPatterns ());
    CPPUNIT_ASSERT(_tdsR->isAttribute(sYHat.Data()) == kFALSE);
    CPPUNIT_ASSERT(_tdsQ->isAttribute(sYHat.Data()));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  0), _tdsQ->getValue(sYHat,  0), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  1), _tdsQ->getValue(sYHat,  1), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY,  7), _tdsQ->getValue(sYHat,  7), 2e-02);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 15), _tdsQ->getValue(sYHat, 15), 2e-03);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_tdsQ->getValue(_sY, 16), _tdsQ->getValue(sYHat, 16), 2e-03);

    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.04209425708, _tdsQ->getValue(sYHat,  0), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02084818178, _tdsQ->getValue(sYHat,  1), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16115863621, _tdsQ->getValue(sYHat,  7), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.03131272643, _tdsQ->getValue(sYHat, 15), 1e-04);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.05582854524, _tdsQ->getValue(sYHat, 16), 1e-04);
}
#endif

void TkNNTest::testTkNNExportModelCplusplus()
{

    time_t seconds;
    string line, date;
    Bool_t bStop = kFALSE;
    Int_t test1 = 0;
    Int_t test2 = 0;
    Int_t test3 = 0;
    Int_t test4 = 0;
    _sY = "SinC";

    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX.Data(), _sY.Data(), 5);
    TkNN->unsetLog();
    TkNN->exportFunction("c++", "uranie_kNN_ishigami.C","kNNFunction");

    fstream file("uranie_kNN_ishigami.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file, line);
        if (line.find("void kNNFunction(double *param, double *res)")
                != string::npos)
            test1 = 1;

        if (line.find("  //     ** Date : ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[24]) && (date[1] == line[25])
                    && (date[2] == line[26]) && (date[3] == line[27])
                    && (date[4] == line[28]) && (date[5] == line[29])
                    && (date[6] == line[30]) && (date[7] == line[31])
                    && (date[8] == line[32]) && (date[9] == line[33])
                    && (date[10] == line[34]) && (date[11] == line[35])
                    && (date[12] == line[36]) && (date[13] == line[37])
                    && (date[14] == line[38]) && (date[15] == line[39])
                    && (date[16] == line[40]) && (date[17] == line[41])
                    && (date[18] == line[42]) && (date[19] == line[43])
                    && (date[20] == line[44]) && (date[21] == line[45])
                    && (date[22] == line[46]) && (date[23] == line[47]))
                test2 = 1;

            if(test2==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);

                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);


                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test2 = 1;
            }
        }

        if (line.find("double y = 0.0;")
                != string::npos)
            test3 = 1;

        if (line.find("res[0] = y;")
                != string::npos)
            test4 = 1;

        if (file.eof())
            bStop = kTRUE;
    }

    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
    CPPUNIT_ASSERT_EQUAL(1, test4);


    delete TkNN;
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////
void TkNNTest::throwExceptionTkNNAttributesNotExistQuery()
{
    _sY = "SinC";
    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX.Data(), _sY.Data(), 5);
    TkNN->unsetLog();

    URANIE::DataServer::TDataServer *tdsQ = new URANIE::DataServer::TDataServer();
    tdsQ->fileDataRead("kernel.dat");

    CPPUNIT_ASSERT_EQUAL(100, tdsQ->getNPatterns());
    CPPUNIT_ASSERT(tdsQ->isAttribute(_sX.Data()) == kFALSE);
    CPPUNIT_ASSERT(tdsQ->isAttribute(_sY.Data()) == kFALSE);

    CPPUNIT_ASSERT_THROW(TkNN->setQueryPoints(tdsQ), URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete tdsQ; tdsQ = NULL;
    delete TkNN; TkNN=NULL;
}

void TkNNTest::throwExceptionTkNNExportFunction()
{
    _sY = "SinC";
    URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsR, _sX.Data(), _sY.Data(), 5);
    TkNN->unsetLog();

    CPPUNIT_ASSERT_THROW(
            TkNN->exportFunction("toto", "filename", "function_name"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            TkNN->exportFunction("pmml", "filename", "function_name"),
            URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(
            TkNN->exportFunction("python", "filename", "function_name"),
            URANIE::Exceptions::UWarningExceptions);

    delete TkNN;
}



void TkNNTest::throwExceptionTkNNSeveralOutput()
{
    _sY = "SinC";
    CPPUNIT_ASSERT_THROW(
                URANIE::Modeler::TkNN * TkNN = new URANIE::Modeler::TkNN(_tdsQ, _sX.Data(), Form("%s:y", _sY.Data()), 5),
            URANIE::Exceptions::UErrorExceptions);

}

