//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestUMLP.cxx
 \author Maud Salvoldi (OCC)
 \date   15/11/2011
 
 \brief 
 
 */

// ROOT
#include "TMath.h"
#include "TString.h"

// cppunit
#include <iostream>
using namespace std;
#define DBL_PRECISION_COMPARE 1.e-10

//uranie
#include "TestUMLP.h"
#include "TDataServer.h"
#include "UExceptions.h"
#include "UDisableTest.h"
#include "TestTANNModeler.h"
#include "uranie_commands.h"

/** Test if files are identical. Possibility to ignore the lines that contain a specified pattern.
 *
 * \param refFileName Name of the reference file
 * \param otherFileName Name of the file to compare to refFileName
 * \param ignoreLinePattern Comparison does not take into account lines that contain this string pattern (optionnal)
 * \return kTRUE if files are equal, kFALSE otherwise
 *
 * \note This method was created to compare files on Windows without taking into account 
 * some lines patterns, but has been used on Linux for computing time reasons.
 */
Int_t areFilesEqual(TString refFileName, TString otherFileName,
        TString ignoreLinePattern = "")
{
    // Get content of the files in TStrings
    TString str1, str2;

    ifstream ifs1(refFileName, ios::in);
    str1.ReadFile(ifs1);
    ifs1.close();

    ifstream ifs2(otherFileName, ios::in);
    str2.ReadFile(ifs2);
    ifs2.close();

    // Remove the lines to ignore from the strings
    if (ignoreLinePattern.CompareTo(""))
    {
        Ssiz_t i0, i1;
        i0 = str1.Index(ignoreLinePattern);
        i1 = str1.Index("\n", i0);
        str1.Remove(i0, i1 - i0 + 1);

        i0 = str2.Index(ignoreLinePattern);
        i1 = str2.Index("\n", i0);
        str2.Remove(i0, i1 - i0 + 1);
    }

    // Compare the TStrings
    // first case: no modification of the format
    Int_t cmp1 = str1.CompareTo(str2);
    // second case: windows compatibility format modification
    str2 = str2.ReplaceAll("e+0", "e+");
    str2 = str2.ReplaceAll("e-0", "e-");
    Int_t cmp2 = str1.CompareTo(str2);

    return cmp1 * cmp2;
}

UMLPTest::UMLPTest()
{
    fstream file("umlp_model.wei");
    if (!file.fail())
        remove("umlp_model.wei");
    
    _workingTestPath = gSystem->pwd();

}

UMLPTest::~UMLPTest()
{
    fstream file("umlp_model.wei");
    if (!file.fail())
        remove("umlp_model.wei");
}

void UMLPTest::setUp()
{
    _tds = new URANIE::DataServer::TDataServer();
    _tds->fileDataRead("flowrate_sampler_launcher_500.dat");

    _network = new URANIE::Modeler::UMLP(8, 1);
    _network->setActivationFunction(URANIE::Modeler::UMLP::kTANH);
}

void UMLPTest::tearDown()
{

    if(_tds) {
        delete _tds; _tds = NULL;
    }
    delete _network; _network = NULL;
    
    UranieTests::removeFile(_workingTestPath.Data(),"SNNS_test.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"umlp_model_funct.C");
    UranieTests::removeFile(_workingTestPath.Data(),"umlp_model_funct.f");
    UranieTests::removeFile(gSystem->pwd(),"flowrate_sampler_launcher_500.root");
    UranieTests::removeFile(gSystem->pwd(),"OPT_DEFAULT.out");
    UranieTests::removeFile(gSystem->pwd(),"_tstOptCalNips__modeler_tdsFlowrate.out");

}

void UMLPTest::testUMLPFirstConstructor()
{
    CPPUNIT_ASSERT_EQUAL(11, _network->getNWeights());
    CPPUNIT_ASSERT_EQUAL(10, _network->getNNeurons());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kTANH, _network->getActivationFunction());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kRegression, _network->getProblem());

    ColumnVector tmp_nbweights;
    tmp_nbweights = _network->getWeights();
    CPPUNIT_ASSERT_EQUAL(11, tmp_nbweights.size());

    _network->setProblem(URANIE::Modeler::UMLP::kClassification);
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kClassification, _network->getProblem());
}

void UMLPTest::testUMLPSecondConstructor()
{
    URANIE::Modeler::UMLP *crtModel = new URANIE::Modeler::UMLP("default_8_1_1.wei");

    CPPUNIT_ASSERT_EQUAL(11, crtModel->getNWeights());
    CPPUNIT_ASSERT_EQUAL(10, crtModel->getNNeurons());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kLOGISTIC, crtModel->getActivationFunction());
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kRegression, crtModel->getProblem());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.879897, crtModel->getWeights(1),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.576637,  crtModel->getWeights(2),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.593863, crtModel->getWeights(3),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.302874, crtModel->getWeights(4),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.276781, crtModel->getWeights(5),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.730723, crtModel->getWeights(6),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.248064, crtModel->getWeights(7),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.481357, crtModel->getWeights(8),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.911367, crtModel->getWeights(9),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.759123,  crtModel->getWeights(10), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.260731,  crtModel->getWeights(11), DBL_PRECISION_COMPARE);

    // Clean
    delete crtModel;
}

void UMLPTest::testUMLPsetWeights()
{
    Int_t nw = 11;
    ColumnVector tmpW(11);
    tmpW(1) = -0.00280385;
    tmpW(2) = -1.0598;
    tmpW(3) = 0.415101;
    tmpW(4) = -0.896944;
    tmpW(5) = -0.00212042;
    tmpW(6) = -0.00781024;
    tmpW(7) = 0.000528823;
    tmpW(8) = -0.20873;
    tmpW(9) = 0.222711;
    tmpW(10) = 0.207785;
    tmpW(11) = -0.111458;

    _network->setWeights(tmpW);

    CPPUNIT_ASSERT_EQUAL(nw, _network->getNWeights());

    for(Int_t i=0; i<nw; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tmpW(i+1), _network->getWeights(i+1), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPexportWeights()
{
    ColumnVector tmp(11);
    tmp(1) = -0.00280385;
    tmp(2) = -1.0598;
    tmp(3) = 0.415101;
    tmp(4) = -0.896944;
    tmp(5) = -0.00212042;
    tmp(6) = -0.00781024;
    tmp(7) = 0.000528823;
    tmp(8) = -0.20873;
    tmp(9) = 0.222711;
    tmp(10) = 0.207785;
    tmp(11) = -0.111458;
    _network->setWeights(tmp);

   _network->exportWeights("SNNS_test.dat");
   
   areFilesEqual(TString("SNNS_ref.dat"), TString("SNNS_test.dat"),
        TString("generated at"));
}

void UMLPTest::testUMLPinitWeights()
{
    _network->initWeights(123);
#ifdef WIN32
	CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.97314453125, _network->getWeights(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.16290283203125, _network->getWeights(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.40838623046875, _network->getWeights(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.2001953125, _network->getWeights(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.97528076171875, _network->getWeights(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.80072021484375, _network->getWeights(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.87677001953125, _network->getWeights(7),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9945068359375, _network->getWeights(8),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.404296875, _network->getWeights(9),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.712890625, _network->getWeights(10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.24688720703125, _network->getWeights(11),
            DBL_PRECISION_COMPARE);
#else
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.879897206090391, _network->getWeights(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.576636929996312, _network->getWeights(2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.593863382004201, _network->getWeights(3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.30287373252213, _network->getWeights(4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.276781194843352, _network->getWeights(5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.730722670443356, _network->getWeights(6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.24806394148618, _network->getWeights(7),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.481356682255864, _network->getWeights(8),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.911367496475577, _network->getWeights(9),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.759123058058321, _network->getWeights(10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.2607312053442, _network->getWeights(11),
            DBL_PRECISION_COMPARE);
#endif
}

void UMLPTest::testUMLPloadWeights()
{
    _network->loadWeights("default_8_1_1.wei");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.879897, _network->getWeights(1),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.576637,  _network->getWeights(2),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.593863, _network->getWeights(3),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.302874, _network->getWeights(4),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.276781, _network->getWeights(5),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.730723, _network->getWeights(6),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.248064, _network->getWeights(7),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.481357, _network->getWeights(8),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.911367, _network->getWeights(9),  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.759123,  _network->getWeights(10), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.260731,  _network->getWeights(11), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPsetTDS()
{
    _network->setTDS(_tds);
    TString soutput = "tdsFlowrate";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(soutput.Data(), _network->getTDS()->GetName()));
}

void UMLPTest::testUMLPsetSigmaHatANDgetSigmaHat()
{
    _network->setSigmaHat(0.95);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.95, _network->getSigmaHat(), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPsetAlphaICANDgetAlphaIC()
{
    _network->setAlphaIC(0.87);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.87, _network->getAlphaIC(), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPsetMinMaxValueANDdumpWeights()
{
    _network->initWeights(123);
    _network->setMinMaxValue(1, 0.0500698, 0.149918);
    _network->setMinValue(2, 147.906);
    _network->setMaxValue(2, 49906.3);
    _network->setMinMaxValue(3, 63163.7, 115568);
    _network->setMinMaxValue(4, 63.1692, 115.901);
    _network->setMinMaxValue(5, 990.01, 1109.79);
    _network->setMinMaxValue(6, 700.145, 819.811);
    _network->setMinMaxValue(7, 1120.34, 1679.34);
    _network->setMinMaxValue(8, 9857.37, 12044);
    _network->setMinMaxValue(9, 13.0982, 208.251);

    _network->dumpWeights("umlp_model.wei");
#ifdef WIN32
	Int_t nret = gSystem->Exec("fc umlp_model.wei umlp_model.wei.win32.ref >NUL");
#else
    Int_t nret = gSystem->Exec("diff umlp_model.wei umlp_model.wei.ref");
#endif
    CPPUNIT_ASSERT_EQUAL(0, nret);

}

void UMLPTest::testUMLPgetActivationFunction()
{
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kTANH, _network->getActivationFunction());

    _network->setActivationFunction(URANIE::Modeler::UMLP::kLOGISTIC);
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kLOGISTIC, _network->getActivationFunction());

}

void UMLPTest::testUMLPgetEcart()
{
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kTANH, _network->getActivationFunction());

    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kRegression, _network->getProblem());

    double input[9];
    // Input
    input[0] = 1.456610531e-01;
    input[1] = 4.817264758e+04;
    input[2] = 7.398381011e+04;
    input[3] = 1.035837913e+02;
    input[4] = 9.968027314e+02;
    input[5] = 7.662662390e+02;
    input[6] = 1.606060725e+03;
    input[7] = 1.147627487e+04;

    // Output (YStar)
    input[8] = 0.498392;

    // Load the MLP
    _network->loadWeights("default_8_1_1.wei");

    _network->setMinMaxValue(1, 0.0500698, 0.149918);
    _network->setMinMaxValue(2, 147.906, 49906.3);
    _network->setMinMaxValue(3, 63163.7, 115568);
    _network->setMinMaxValue(4, 63.1692, 115.901);
    _network->setMinMaxValue(5, 990.01, 1109.79);
    _network->setMinMaxValue(6, 700.145, 819.811);
    _network->setMinMaxValue(7, 1120.34, 1679.34);
    _network->setMinMaxValue(8, 9857.37, 12044);
    _network->setMinMaxValue(9, 13.0982, 208.251);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _network->getEcart(input), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPgetWeightsDerivate()
{
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kTANH, _network->getActivationFunction());

    double input[9];
    // Input
    input[0] = 2.0 * ( 1.456610531e-01 - 0.0500698 ) / (0.149918 - 0.0500698) -1.0 ;
    input[1] = 2.0 * ( 4.817264758e+04 - 147.906) / (49906.3 - 147.906) -1.0;
    input[2] = 2.0 * ( 7.398381011e+04 - 63163.7) / (115568 - 63163.7) -1.0;
    input[3] = 2.0 * ( 1.035837913e+02 - 63.1692) / (115.901 - 63.1692) -1.0;
    input[4] = 2.0 * ( 9.968027314e+02 - 990.01) / (1109.79 - 990.01) -1.0;
    input[5] = 2.0 * ( 7.662662390e+02 - 700.145) / (819.811 - 700.145) -1.0;
    input[6] = 2.0 * ( 1.606060725e+03 - 1120.34) / (1679.34 -1120.34 ) -1.0;
    input[7] = 2.0 * ( 1.147627487e+04 - 9857.37) / (12044 - 9857.37) -1.0;

    // Output (YStar)
    input[8] = 0.55567952329128;

    // Load the MLP
    _network->loadWeights("default_8_1_1.wei");

    Int_t nw = _network->getNWeights();
    CPPUNIT_ASSERT_EQUAL( 11, _network->getNWeights());

    _network->evaluate(input);

    for(Int_t i=0; i<nw; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _network->getWeightsDerivate(i+1), DBL_PRECISION_COMPARE);

    // Output (YStar)
    input[8] = 0.498392;
    _network->evaluate(input);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.84263397493170e-03, _network->getWeightsDerivate(1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.34444206786837e-03, _network->getWeightsDerivate(2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.43550280708031e-03, _network->getWeightsDerivate(3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-3.42993337817960e-03, _network->getWeightsDerivate(4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.72875232912809e-02, _network->getWeightsDerivate(nw-1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-4.47003728379437e-02, _network->getWeightsDerivate(nw), DBL_PRECISION_COMPARE);
}

void UMLPTest::testUMLPevaluateWeightsJacobianANDgetPartialWeightsDerivate()
{
    CPPUNIT_ASSERT_EQUAL(URANIE::Modeler::UMLP::kTANH, _network->getActivationFunction());

    double input[9];
    // Input
    input[0] = 2.0 * ( 1.456610531e-01 - 0.0500698 ) / (0.149918 - 0.0500698) -1.0 ;
    input[1] = 2.0 * ( 4.817264758e+04 - 147.906) / (49906.3 - 147.906) -1.0;
    input[2] = 2.0 * ( 7.398381011e+04 - 63163.7) / (115568 - 63163.7) -1.0;
    input[3] = 2.0 * ( 1.035837913e+02 - 63.1692) / (115.901 - 63.1692) -1.0;
    input[4] = 2.0 * ( 9.968027314e+02 - 990.01) / (1109.79 - 990.01) -1.0;
    input[5] = 2.0 * ( 7.662662390e+02 - 700.145) / (819.811 - 700.145) -1.0;
    input[6] = 2.0 * ( 1.606060725e+03 - 1120.34) / (1679.34 -1120.34 ) -1.0;
    input[7] = 2.0 * ( 1.147627487e+04 - 9857.37) / (12044 - 9857.37) -1.0;

    // Output (YStar)
    input[8] = 0.55567952329128;

    // Load the MLP
    _network->loadWeights("default_8_1_1.wei");

    Int_t nw = _network->getNWeights();
    CPPUNIT_ASSERT_EQUAL( 11, _network->getNWeights());

   _network->evaluateWeightsJacobian(input);

   ColumnVector tmp(nw);
   tmp(1) = 1.0;
   tmp(2) = -0.780281;
   tmp(3) = 0.225558;
   tmp(4) = 0.206325;
   tmp(5) = 0.209841;
   tmp(6) = -0.132414;
   tmp(7) = 0.120185;
   tmp(8) = -0.199975;
   tmp(9) = 0.0237054;
   tmp(10) = 0.166421;
   tmp(11) = 0.108433;

    for(Int_t i=0; i<nw; i++) {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tmp(i+1), _network->getPartialWeightsDerivate(i+1), 1e-6);
    }
}

void UMLPTest::testUMLPExportFunction()
{
    time_t seconds;
    string line, date;
    Bool_t bStop = kFALSE;
    Int_t test1 = 0;
    Int_t test2 = 0;
    Int_t test3 = 0;

    // Load the MLP
   _network->loadWeights("default_8_1_1.wei");

    // C++ Export
    _network->exportFunction("c++", "umlp_model_funct.C", "function_umlp");
    fstream file("umlp_model_funct.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file, line);
        if (line.find("  //     ** Uranie  ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[24]) && (date[1] == line[25])
                    && (date[2] == line[26]) && (date[3] == line[27])
                    && (date[4] == line[28]) && (date[5] == line[29])
                    && (date[6] == line[30]) && (date[7] == line[31])
                    && (date[8] == line[32]) && (date[9] == line[33])
                    && (date[10] == line[34]) && (date[11] == line[35])
                    && (date[12] == line[36]) && (date[13] == line[37])
                    && (date[14] == line[38]) && (date[15] == line[39])
                    && (date[16] == line[40]) && (date[17] == line[41])
                    && (date[18] == line[42]) && (date[19] == line[43])
                    && (date[20] == line[44]) && (date[21] == line[45])
                    && (date[22] == line[46]) && (date[23] == line[47]))
                test1 = 1;
                
            if(test1==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);
                            
                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);
                
               
                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test1 = 1;
            }
        }
        if (line.find("  -0.879897, 0.576637, -0.593863, -0.302874, -0.276781,")
                != string::npos)
            test2 = 1;
        if (line.find("void function_umlp(double *param, Double_t *res)")
                != string::npos)
            test3 = 1;
        if (file.eof())
            bStop = kTRUE;
    }
    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
    
    // Reset variables
    bStop = kFALSE;
    test1 = 0;
    test2 = 0;
    test3 = 0;

    // Fortran Export
    _network->exportFunction("fortran", "umlp_model_funct.f", "function_umlp",0);
    fstream file2("umlp_model_funct.f");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file2, line);
        
        if (line.find("C     ** Uranie  ") != string::npos)
        {
            //verify that the date are the same actually and in the file
            if ((date[0] == line[17]) && (date[1] == line[18])
                    && (date[2] == line[19]) && (date[3] == line[20])
                    && (date[4] == line[21]) && (date[5] == line[22])
                    && (date[6] == line[23]) && (date[7] == line[24])
                    && (date[8] == line[25]) && (date[9] == line[26])
                    && (date[10] == line[27]) && (date[11] == line[28])
                    && (date[12] == line[29]) && (date[13] == line[30])
                    && (date[14] == line[31]) && (date[15] == line[32])
                    && (date[16] == line[33]) && (date[17] == line[34])
                    && (date[18] == line[35]) && (date[19] == line[36])
                    && (date[20] == line[37]) && (date[21] == line[38])
                    && (date[22] == line[39]) && (date[23] == line[40]))
                test1 = 1;

            if(test1==0)
            {
                Int_t pos = date.find_last_of(":");
                Int_t pos2 = line.find_last_of(":");
                Int_t date_seconds, line_seconds;
                date_seconds =  (date[pos-5]*10+date[pos-4])*3600
                            +(date[pos-2]*10+date[pos-1])*60
                            +(date[pos+1]*10+date[pos+2]);
                            
                line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
                            +(line[pos2-2]*10+line[pos2-1])*60
                            +(line[pos2+1]*10+line[pos2+2]);
                
               
                if(fabs(date_seconds - line_seconds + 0.0)<3)
                    test1 = 1;
            }
        }
        if (line.find(
                "      SUBROUTINE function_umlp(inp1,inp2,inp3,inp4,inp5,inp6,inp7,inp8,out9)")
                != string::npos)
            test2 = 1;
        if (line.find("      Xnorminp1 = 2.D0 * ( inp1 - 0.05D0 ) / 0.1D0 - 1.0")
                != string::npos)
            test3 = 1;
        if (file2.eof())
            bStop = kTRUE;
    }
    CPPUNIT_ASSERT_EQUAL(1, test1);
    CPPUNIT_ASSERT_EQUAL(1, test2);
    CPPUNIT_ASSERT_EQUAL(1, test3);
}

void UMLPTest::throwExceptionUMLPgetWeightsDerivate()
{
    double input[8];
    input[0] = 1.456610531e-01;
    input[1] = 4.817264758e+04;
    input[2] = 7.398381011e+04;
    input[3] = 1.035837913e+02;
    input[4] = 9.968027314e+02;
    input[5] = 7.662662390e+02;
    input[6] = 1.606060725e+03;
    input[7] = 1.147627487e+04;
    _network->initWeights(123);
    _network->setMinMaxValue(1, 0.0500698, 0.149918);
    _network->setMinMaxValue(2, 147.906, 49906.3);
    _network->setMinMaxValue(3, 63163.7, 115568);
    _network->setMinMaxValue(4, 63.1692, 115.901);
    _network->setMinMaxValue(5, 990.01, 1109.79);
    _network->setMinMaxValue(6, 700.145, 819.811);
    _network->setMinMaxValue(7, 1120.34, 1679.34);
    _network->setMinMaxValue(8, 9857.37, 12044);
    _network->setMinMaxValue(9, 13.0982, 208.251);
    _network->evaluate(input);

    CPPUNIT_ASSERT_THROW(_network->getWeightsDerivate(-5), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_network->getWeightsDerivate(0), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_network->getWeightsDerivate(14), URANIE::Exceptions::UErrorExceptions);
}

void UMLPTest::throwExceptionUMLPgetPartialWeightsDerivate()
{
    double input[8];
    input[0] = 1.456610531e-01;
    input[1] = 4.817264758e+04;
    input[2] = 7.398381011e+04;
    input[3] = 1.035837913e+02;
    input[4] = 9.968027314e+02;
    input[5] = 7.662662390e+02;
    input[6] = 1.606060725e+03;
    input[7] = 1.147627487e+04;
    _network->initWeights(123);
    _network->setMinMaxValue(1, 0.0500698, 0.149918);
    _network->setMinMaxValue(2, 147.906, 49906.3);
    _network->setMinMaxValue(3, 63163.7, 115568);
    _network->setMinMaxValue(4, 63.1692, 115.901);
    _network->setMinMaxValue(5, 990.01, 1109.79);
    _network->setMinMaxValue(6, 700.145, 819.811);
    _network->setMinMaxValue(7, 1120.34, 1679.34);
    _network->setMinMaxValue(8, 9857.37, 12044);
    _network->setMinMaxValue(9, 13.0982, 208.251);
    _network->evaluateWeightsJacobian(input);

    CPPUNIT_ASSERT_THROW(_network->getPartialWeightsDerivate(-5), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_network->getPartialWeightsDerivate(0), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_network->getPartialWeightsDerivate(14), URANIE::Exceptions::UErrorExceptions);
}

void UMLPTest::throwExceptionUMLPFalseNiput()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(-1, 5), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(-1, "5,6,4"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(-1, 5, 4), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(-1, "5,6,4", 1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(5, 2, -2), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(5, "5,6,4", -2), URANIE::Exceptions::UErrorExceptions);
}

void UMLPTest::throwExceptionUMLPFalseNidden()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, -1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, -1, 2), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, "-5,6,4", 2), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, "5,-6,4", 2), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, "5,6,-4", 2), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP(8, "5,6,5,6,5,6,5,6,-4", 2), URANIE::Exceptions::UErrorExceptions);
}

void UMLPTest::throwExceptionUMLPFalseFileName()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("BadWeights_X_1_1.wei"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("BadWeights_8_X_1.wei"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("BadWeights_8_1_X.wei"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("BadWeights_8_1_Y.wei"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("BadWeights_8_1_1_X.wei"), URANIE::Exceptions::UErrorExceptions);
}

void UMLPTest::throwExceptionUMLPFalseFileData()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::UMLP("umlp_network_false_data.wei"), URANIE::Exceptions::UErrorExceptions);
}
