//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////

#include <sstream>
#include <string>

// ROOT
#include "Api.h"

// URANIE
#include "TFCNAnalyticalFunction.h"
#include "UExceptions.h"

URANIE::Optimizer::TFCNAnalyticalFunction *gTFCNAF;

ClassImp (URANIE::Optimizer::TFCNAnalyticalFunction)

namespace URANIE
{
namespace Optimizer
{

//______________________________________________________________________________
void UranieOptimizerFunctionInteractiveMethod(Double_t *param, Double_t *out)
{

    TMethodCall *m = gTFCNAF->getMethodCall();
    if (!m)
        return;

    Long_t args[2];
    args[0] = (Long_t) param;
    args[1] = (Long_t) out;
    m->SetParamPtrs(args, 2);
    Double_t result;
    m->Execute(result);
//       cout << " *** x1[" << param[0] <<"] x2[" << param[1] << "] y[" << out[0] <<"]" << endl;
}
//////////////////////////////////
// Constructor
//////////////////////////////////
TFCNAnalyticalFunction::TFCNAnalyticalFunction(
        URANIE::DataServer::TDataServer *tds, void *fcn, TString sinput,
        TString soutput, TString scost) :
        _tds(tds), _sinput(sinput), _soutput(soutput), _sCost(scost), _blog(
                kFALSE)
{
    if (1)
    {
        cout
                << "************* TFCNAnalyticalFunction Constructor *******************"
                << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
        cout << " ** _sCost[" << _sCost << "]" << endl;
    }

    try
    {

        if (!fcn)
        {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TFCNAnalyticalFunction::TFCNAnalyticalFunction ERROR\n The function is undefined."));
        }

        char *functionanme = G__p2f2funcname(fcn);

        cout << " ** sFunctionName[" << functionanme << "]" << endl;

        if (functionanme != 0)
        {
            cout << "***************** TMethodCall with [" << functionanme
                    << "]**************" << endl;
            _fMethodCall = new TMethodCall();
            _fMethodCall->InitWithPrototype(functionanme,
                    "Double_t*,Double_t*");
        }
        _fMethod = UranieOptimizerFunctionInteractiveMethod;
        gTFCNAF = this;

        init();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        exit(-1);
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}
TFCNAnalyticalFunction::TFCNAnalyticalFunction(
        URANIE::DataServer::TDataServer *tds, TString sfunction, TString sinput,
        TString soutput) :
        _tds(tds), _sinput(sinput), _soutput(soutput), _blog(kFALSE)
{
    if (1)
    {
        cout
                << "************* TFCNAnalyticalFunction Constructor *******************"
                << endl;
        cout << " ** sfunction[" << sfunction << "]" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
    }

    try
    {

        if (sfunction.Length() == 0)
        {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TFCNAnalyticalFunction::TFCNAnalyticalFunction ERROR\n The functionname is empty."));
        }

        const char *functionanme = sfunction.Data();

        cout << " ** sFunctionName[" << functionanme << "]" << endl;

        if (functionanme != 0)
        {
            cout << "***************** TMethodCall with [" << functionanme
                    << "]**************" << endl;
            _fMethodCall = new TMethodCall();
            _fMethodCall->InitWithPrototype(functionanme,
                    "Double_t*,Double_t*");
        }
        _fMethod = UranieOptimizerFunctionInteractiveMethod;
        gTFCNAF = this;

        init();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        exit(-1);
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////
// Destructor
//////////////////////////////////
TFCNAnalyticalFunction::~TFCNAnalyticalFunction()
{
    delete[] _dinputValue;
    _dinputValue = NULL;
    delete[] _doutputValue;
    _doutputValue = NULL;
}

void TFCNAnalyticalFunction::init()
{
    try
    {
        if (_tds != NULL)
        {
            if (_tds->getTuple() == NULL)
            {
                _tds->createTuple();
            }
        }
        _ninput = _tds->getNInputAttributes();
        _noutput = _tds->getNOutputAttributes();

        Int_t ncpt = 0;
        Int_t nvar = _tds->getNAttributes();
        for (Int_t ivar = 0; ivar < nvar; ivar++)
        {
            if (_tds->getAttribute(ivar)->isOutput())
            {
                TString sname = _tds->getAttribute(ivar)->GetName();
                if (sname == _sCost)
                {
                    _nCost = ncpt;
                    break;
                }
                ncpt++;
            }
        }
        cout << " ** TFCNAnalyticalFunction::init _sCost[" << _sCost
                << "] _nCost[ " << _nCost << "]" << endl;

        // le +1 est a cause de l'iterateur
        _dinputValue = new double[_ninput + _noutput + 1];
        _doutputValue = new double[_noutput];

        clean();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        exit(-1);
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

void TFCNAnalyticalFunction::clean()
{
}
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

