//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////

#include <sstream>
#include <string>

// ROOT
//fga 2015/01/12 #include "Api.h"
#include "TPad.h"
#include "TFile.h"

// URANIE
#include "TMultiGenAnalyticalFunction.h"
#include "TAttributeFormula.h"
#include "UExceptions.h"

URANIE::Optimizer::TMultiGenAnalyticalFunction *gTMGAF;

ClassImp (URANIE::Optimizer::TMultiGenAnalyticalFunction)

namespace URANIE
{
namespace Optimizer
{

//______________________________________________________________________________
void UranieOptimizerNewFunctionInteractiveMethod(Double_t *param, Double_t *out)
{

//     cout << " *** UranieOptimizerNewFunctionInteractiveMethod input x1[" << param[0] <<"] x2[" << param[1] << "] y[" << out[0] <<"]" << endl;
    TMethodCall *m = gTMGAF->getMethodCall();
    if (!m)
        return;

    Long_t args[2];
    args[0] = (Long_t) param;
    args[1] = (Long_t) out;
    m->SetParamPtrs(args, 2);
    Double_t result;
    m->Execute(result);
//     cout << " *** UranieOptimizerNewFunctionInteractiveMethod output x1[" << param[0] <<"] x2[" << param[1] << "] y[" << out[0] <<"]" << endl;
}
//////////////////////////////////
// Constructor with void *
//////////////////////////////////

TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction(
        URANIE::DataServer::TDataServer *tds,
#ifdef ROOT5
        void *fcn,
#else
        void  (*fcn)(Double_t*, Double_t*),
#endif
        TString sinput, TString soutput, TString scost) :
    _blog(kFALSE), _sinput(sinput), _dinputValue(0), _xin(0), _codevalue(NULL), _codeindex(NULL), _soutput(soutput), _doutputValue(0), _sCost(scost), _nCostIndex(-1),
        _sWorkingDirectory("URANIE"), _tds(tds), _CostInOutput(0), _fMethodCall(NULL), _informTree(NULL)
{
    if (_blog)
    {
        cout
            << " ************ TMultiGenAnalyticalFunction Constructor *******************"
            << endl;
        cout << " ** _sWorkingDirectory[" << _sWorkingDirectory << "]" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
        cout << " ** _sCost[" << _sCost << "]" << endl;
    }

    try
    {

        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction\n The pointer of the \"TDS\" is empty.");

        if (!fcn)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction ERROR\n The function is undefined."));
        }

        TString functionName ;
#ifdef ROOT5
        functionName = G__p2f2funcname(fcn);

        if (functionName != 0)
        {
	  if (_blog)
	    cout << " ** sFunctionName[" << functionName << "]" << endl;
	  _fMethodCall = new TMethodCall();
	  _fMethodCall->InitWithPrototype(functionName.Data(),
					  "Double_t*,Double_t*");
        }
        _fMethod = UranieOptimizerNewFunctionInteractiveMethod;
#else
        _fMethod = fcn;
        functionName = soutput;
#endif
        gTMGAF = this;

        init();
        if (_blog)
            cout
                    << " ************ End Of TMultiGenAnalyticalFunction Constructor *******************"
                    << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete _fMethodCall;
	_fMethodCall=NULL;
        throw ue;
    }
}

//////////////////////////////////
// Constructor with const char*
//////////////////////////////////
TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction(
    URANIE::DataServer::TDataServer *tds, const char* fcn, TString sinput,
    TString soutput, TString scost) :
    _blog(kFALSE), _sinput(sinput), _dinputValue(0), _xin(0), _codevalue(NULL), _codeindex(NULL),
    _soutput(soutput), _doutputValue(0), _sCost(scost), _nCostIndex(-1),
    _sWorkingDirectory("URANIE"), _tds(tds), _CostInOutput(0), _fMethod(NULL), _fMethodCall(NULL), 
    _informTree(NULL)
{
    if (_blog)
    {
        cout
                << " ************ TMultiGenAnalyticalFunction Constructor *******************"
                << endl;
        cout << " ** _sWorkingDirectory[" << _sWorkingDirectory << "]" << endl;
        cout << " ** sinput[" << sinput << "]" << endl;
        cout << " ** soutput[" << soutput << "]" << endl;
        cout << " ** _sCost[" << _sCost << "]" << endl;
    }

    try
    {

        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction\n The pointer of the \"TDS\" is empty.");

        if (!fcn)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::TMultiGenAnalyticalFunction ERROR\n The function is undefined."));
        }

        if (fcn != NULL)
        {
            if (_blog)
                cout << " ** sFunctionName[" << fcn << "]" << endl;
            _fMethodCall = new TMethodCall();
            _fMethodCall->InitWithPrototype(fcn, "Double_t*,Double_t*");
        }
	

        if (_fMethodCall->IsValid () == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TMultiGenAnalyticalFunction::constructor with function name\n The function \"%s\" is not valid", fcn));
	
        //_fMethod = UranieOptimizerNewFunctionInteractiveMethod;
        //gTMGAF = this;

        init();
        if (_blog)
            cout
                    << " ************ End Of TMultiGenAnalyticalFunction Constructor *******************"
                    << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete _fMethodCall;
	_fMethodCall=NULL;
        throw ue;
    }
}

//////////////////////////////////
// Destructor
//////////////////////////////////
TMultiGenAnalyticalFunction::~TMultiGenAnalyticalFunction()
{
    for(vector<TTreeFormula*>::iterator it= _vinForm.begin(); it!=_vinForm.end(); it++)
	delete (*it);
    _vinForm.clear();    

    if( _informTree != NULL)
	delete _informTree;

    delete[] _dinputValue;
    _dinputValue = NULL;

    delete[] _codevalue;
    _codevalue = NULL;

    delete[] _codeindex;
    _codeindex = NULL;
    
    delete[] _xin;
    _xin = NULL;

    delete[] _doutputValue;
    _doutputValue = NULL;

    delete _fMethodCall;
    _fMethodCall=NULL;
}

//////////////////////////////////
// init
//////////////////////////////////
void TMultiGenAnalyticalFunction::init()
{
    if (_blog)
        cout
                << " *********************************************** TMultiGenAnalyticalFunction::init"
                << endl;
    try
    {

        if (_tds != NULL)
        {
            try
            {
                _tds->getTuple();
            } catch (URANIE::Exceptions::UWarningExceptions&)
            {
                _tds->createTuple();
            }
        }
        _ninput = 0;//_tds->getNInputAttributes();
        _ninformula = 0;
        _noutput = 0; //_tds->getNOutputAttributes();
        _noutformula=0;
        
        // Verify that all inputs attributes are in the TDS
        TString sinpatt(_sinput.Data());
        sinpatt.ReplaceAll(" ", "");
        sinpatt.ReplaceAll(":", " ");
        istringstream inStream((char *) sinpatt.Data());
        string inName;
        while (inStream >> inName)
        {
            if (!_tds->isAttribute(inName.c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TMultiGenAnalyticalFunction::init\n The input attribute \"%s\" is not in the \"TDS\".",
                                inName.c_str()));
        }

        if (_sCost.Length() == 0)
        {
            // First attribute in the soutput string will be the cost attribute
            if (_soutput.CountChar(':') == 0)
            {
                _sCost = _soutput;
            }
            else
            {
                if (_blog)
                    cout << " ** Case of several outputs [" << _soutput << "]"
                            << endl;

                TString soutatt(_soutput.Data());
                soutatt.ReplaceAll(" ", "");
                soutatt.ReplaceAll(":", " ");
                istringstream outStream((char *) soutatt.Data());
                // The First output attribute is the cost
                outStream >> inName;
                _sCost = Form("%s", inName.c_str());
            }
            if (_blog)
                cout << " ** _sCost[" << _sCost << "] _soutput[" << _soutput
                        << "]" << endl;
        }

        Int_t natt = _sCost.CountChar(':') + 1;
        if (natt > 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::init\n Several output attributes: \"%s\" in the sCost attribute.\n Only one is autorised.",
                            _sCost.Data()));

        if (!_tds->isAttribute(_sCost))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::init\n The cost attribute \"%s\" is not in the \"TDS\".",
                            _sCost.Data()));

        if (!_tds->getAttribute(_sCost)->isOutput())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::init\n The cost attribute \"%s\" is not an kOutput attribute.",
                            _sCost.Data()));

        // Looking for the index of the cost attribute
        //  to return the value of the Cost Function
        vector<TString> informulaes;
        TString inputVar="";
        Int_t ncpt = 0;
        Int_t nvar = _tds->getNAttributes();
        
        for (Int_t ivar = 0; ivar < nvar; ivar++)
        {
            
            URANIE::DataServer::TAttributeFormula* isform = dynamic_cast<URANIE::DataServer::TAttributeFormula*>(_tds->getAttribute(ivar));
            if (_tds->getAttribute(ivar)->isOutput())
            {
               TString sname = _tds->getAttribute(ivar)->GetName();
                if (sname == _sCost)
                    _nCostIndex = ncpt;

                ncpt++;
                if(  isform == NULL )
                    _noutput++;
                else
                    _noutformula++;
            }
            else
            {
                TNtupleD *myTree = new TNtupleD("tree", "Example", "x:y:index");
                if(  isform == NULL )
                {
                    _ninput++;
                    inputVar += ((inputVar=="") ? TString("") : TString(":") ) +  _tds->getAttribute(ivar)->GetName() ;
                }
                else
                {
                    _ninformula++;
                    informulaes.push_back( isform->getFormula() );
                }
		delete myTree;
            }

        }
        if (_blog)
            cout << " ** TMultiGenAnalyticalFunction::init _sCost[" << _sCost
                 << "] _nCostIndex[ " << _nCostIndex << "]  nvar [" << nvar << "]" << endl;
            
        // Verify that all outputs attributes are in the TDS whith type==Output
        _CostInOutput = 0;
        TString sout(_soutput.Data());
        sout.ReplaceAll(" ", "");
        sout.ReplaceAll(":", " ");
        istringstream outStream((char *) sout.Data());
        while (outStream >> inName)
        {
            if (!_tds->isAttribute(inName.c_str()))
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TMultiGenAnalyticalFunction::init\n The ouput attribute \"%s\" is not in the \"TDS\".",
                                inName.c_str()));
            if (!_tds->getAttribute(inName.c_str())->isOutput())
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TMultiGenAnalyticalFunction::init\n The output attribute \"%s\" is not an kOutput attribute.",
                                inName.c_str()));
            // The current output is the cost attribute
            if (inName.c_str() == _sCost)
                _CostInOutput = 1;
        }
        if( !_CostInOutput )
            for (Int_t ivar = 0; ivar < nvar; ivar++)
                if( _tds->getAttribute(ivar)->isOutput() &&  dynamic_cast<URANIE::DataServer::TAttributeFormula*>(_tds->getAttribute(ivar)) != NULL  && _tds->getAttribute(ivar)->GetName() == _sCost)
                    _CostInOutput = 2;
                    
        if (_CostInOutput == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenAnalyticalFunction::init\n The cost attribute \"%s\" is not in the list of output attributes \"%s\".",
                            _sCost.Data(), _soutput.Data()));

        if (_noutput == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TMultiGenAnalyticalFunction::init [%s] ERROR\n the Number of output [%d] is null, while it should not",
                                                              _tds->GetName(), _noutput));

        if( _ninformula != 0)
        {
            cout << " ** TMultiAnalyticalFunction::init Creating the formula for new inputs " << endl;

            _informTree = new TNtupleD("___internalTreeOptimizer___", "foo__optim__bar", inputVar);

            for(unsigned int ifor=0; ifor<informulaes.size(); ifor++)
            {
                _vinForm.push_back( new TTreeFormula(Form("Formula_%d",ifor), informulaes.at(ifor).Data(), _informTree));
                if(_vinForm.back()->GetNdim()== 0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TMultiGenAnalyticalFunction::init ERROR\n the provided formula (%s) is incorrect",
                                                                    informulaes.at(ifor).Data()));
            }
        }
        
        // le +1 est a cause de l'iterateur
        _dinputValue = new double[_ninput + _ninformula + _noutput + _noutformula + 1];
        _doutputValue = new double[_noutput];
        _xin = new double[_ninput];
        _codeindex = new int[ _ninput ];
        _codevalue = new double[_ninput ];

        if( _ninformula != 0)
        {
            int counter=0;
            sinpatt = TString(_sinput.Data());
            sinpatt.ReplaceAll(" ", "");
            sinpatt.ReplaceAll(":", " ");
            istringstream Stream((char *) sinpatt.Data());
            while (Stream >> inName)
            {
                for(int ibr = 0; ibr < _tds->getNAttributes(); ibr++)
                    if( TString(inName) == _tds->getAttribute(ibr)->GetName() )
                    {
                        _codeindex[counter] = ibr + _noutput +1 ;
                        //cout<<" for input (" << inName <<","<<counter<<")  codeindex is " << _codeindex[counter] << endl;
                        counter++;
                    }
            }
        }
        
        // Creates the working directory if it doesn't exist
        void * openFileResult = gSystem->OpenDirectory(
                Form("%s/UranieResults", _sWorkingDirectory.Data()));
        if (openFileResult == 0)
            gSystem->mkdir(Form("%s/UranieResults", _sWorkingDirectory.Data()),
                    kTRUE);
        gSystem->FreeDirectory(openFileResult);

        clean();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////
// DoEval
//////////////////////////////////
double TMultiGenAnalyticalFunction::DoEval(const double * x) const
{
    Int_t niter = (Int_t) _tds->getTuple()->GetEntries();
    if (_blog)
        cout << " ** niter[" << niter << "]" << endl;

    // The iteration value
    _dinputValue[0] = 1.0 + niter;

    // The input values
    for (int i = 0; i < _ninput; i++)
    {
        _xin[i] = x[i];
        _dinputValue[1 + i] = x[i];
        if (_blog)
            cout << "**** input i[" << i << "] x[" << x[i] << "]" << endl;
    }

    if( _ninformula != 0)
    {
        _informTree->Fill(_xin);
        for(unsigned int ifor=0; ifor<_vinForm.size(); ifor++)
            _dinputValue[1 + _ninput + _noutput + ifor] = _vinForm.at(ifor)->EvalInstance();
            
        for(int ival=0; ival<_ninput; ival++)
            _codevalue[ival] = _dinputValue[ _codeindex[ival] ];

    }

    // Evaluate the function
    if (_fMethod != NULL)
        (*_fMethod)( ((_ninformula != 0) ? _codevalue : _xin), _doutputValue);
    if (_fMethodCall) {
        if (_blog)
	  cout << " ** fMethodCall Eval [" << niter << "]" << endl;
	uintptr_t args[2];
	Double_t result = -1.0;

    args[0] = reinterpret_cast<uintptr_t>((_ninformula != 0) ? _codevalue : _xin);
    args[1] = reinterpret_cast<uintptr_t>(_doutputValue);
	_fMethodCall->SetParamPtrs(args, 2);
	_fMethodCall->Execute(&result);
    }

    //Fill the output values in the input vector
    // +1 because the first value is the index
    for (int i = 1; i <= (_noutput + _ninformula + _noutformula); i++)
    {
        int test = ((niter==0)? ((i<=_noutput)?_ninformula:((_ninformula==0)?0:((i>(_noutput+_ninformula))?0:-_noutput))):0);
        _dinputValue[_ninput + i  + test ] = ((i<=_noutput) ? _doutputValue[i-1] : 0);
    }
    
    // for (int i = 0; i < _noutput + _nformula; i++)
    //     _dinputValue[_ninput + i + 1] = ((i<_noutput) ? _doutputValue[i] : 0);
    _tds->getTuple()->Fill(_dinputValue);    
    _tds->fillOthersAttributes();
    
    double fval = ((_CostInOutput==1) ? _doutputValue[_nCostIndex] : _tds->getValue(_sCost, niter));
    //cout << " *** output _nCost[" << _nCostIndex << "] _sCost[" << _sCost << "] val[" << fval << "]" << endl;
    
    // Save the TDSNtupleD in a RootFile
    // FGA::Modification 2012-11-26 13:30:21
    // Suppress this save
    // End Of FGA::Modification 2012-11-26 13:30:21
    if (0)
    {
        TString srootfile = Form("%s/UranieResults/uranielauncher.root",
                _sWorkingDirectory.Data());
        if (niter % 5 == 0)
        {
            TString ssavefile = srootfile + TString(".bak");
            gSystem->CopyFile(srootfile.Data(), ssavefile.Data(), kTRUE);
        }
        TFile * rootfile = TFile::Open(srootfile.Data(), "RECREATE");
        _tds->getTuple()->SetDirectory(0);
        _tds->getTuple()->Write();
        rootfile->Close();
        _tds->getTuple()->SetDirectory(gDirectory);
    }

    // Visualisation
    if (1)
    {
        _tds->draw(Form("%s:%s", _sCost.Data(), _tds->getIteratorName()), "",
                "lp");
        gPad->Modified();
        gPad->Update();
    }
 
    return fval; //_doutputValue[_nCostIndex];
}

//////////////////////////////////
// clean
//////////////////////////////////
void TMultiGenAnalyticalFunction::clean()
{
}
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

