//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TMultiGenCode.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date     
 \brief Implementation of the class URANIE::Optimizer::TMultiGenCode

 \par Modifications
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - add a IFDEF WIN32 to replace rm -fr by del /Q on Windows platform
 
 */
#include <sstream>
#include <string>

// ROOT
//fga 2015/01/12 #include "Api.h"
#include "TPad.h"

// URANIE
#include "TMultiGenCode.h"
#include "TAttributeFormula.h"
#include "UExceptions.h"

URANIE::Optimizer::TMultiGenCode *gTMGC;

ClassImp (URANIE::Optimizer::TMultiGenCode)

namespace URANIE
{
namespace Optimizer
{

////////////////
// Constructor 
////////////////
TMultiGenCode::TMultiGenCode(URANIE::DataServer::TDataServer *tds,
        URANIE::Launcher::TCode *code, TString scost) :
    _blog(kFALSE), _dinputValue(0), _xin(0), _codevalue(NULL), _codeindex(NULL), _doutputValue(0), _sCost(scost), _nCost(0), _tds(tds), _code(code), _informTree(NULL), _bintermedstep(false)
{
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenCode::constructor\n The pointer of \"TDataServer\" is empty");
        if (_code == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenCode::constructor\n The pointer of \"TCode\" is empty");
        init();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////
// Destructor
//////////////////////////////////
TMultiGenCode::~TMultiGenCode()
{

    for(vector<TTreeFormula*>::iterator it= _vinForm.begin(); it!=_vinForm.end(); it++)
	delete (*it);
    _vinForm.clear();    

    if( _informTree != NULL)
	delete _informTree;    

    delete[] _dinputValue;
    _dinputValue = NULL;

    delete[] _codevalue;
    _codevalue = NULL;

    delete[] _codeindex;
    _codeindex = NULL;

    delete[] _xin;
    _xin = NULL;

    delete[] _doutputValue;
    _doutputValue = NULL;

}
//////////////////////////////////
// init
//////////////////////////////////
void TMultiGenCode::init()
{
    try
    {
        if (_tds != NULL)
        {
            try
            {
                _tds->getTuple();
            } catch (URANIE::Exceptions::UWarningExceptions&)
            {
                _tds->createTuple();
            }
        }
        _ninput = 0; //_tds->getNInputAttributes();
        _ninformula = 0;
        _noutput = 0; //_tds->getNOutputAttributes();
        _noutformula = 0;

        vector<TString> informulaes;
        TString inputVar="";
        Int_t ncpt = 1;
        Int_t nvar = _tds->getNAttributes();
        for (Int_t ivar = 0; ivar < nvar; ivar++)
        {

            URANIE::DataServer::TAttributeFormula* isform = dynamic_cast<URANIE::DataServer::TAttributeFormula*>(_tds->getAttribute(ivar));
            if (_tds->getAttribute(ivar)->isOutput())
            {
                _soutput = _tds->getAttribute(ivar)->GetName();
                cout << " ** output :: ivar[" << ivar << "/" << nvar << "] name["
                     << _tds->getAttribute(ivar)->GetName() << "]" << endl;
                if (_soutput == _sCost)
                    _nCost = ncpt;
                ncpt++;
                if(  isform == NULL )
                    _noutput++;
                else
                    _noutformula++;
            }
            else
            {
                TNtupleD *myTree = new TNtupleD("tree", "Example", "x:y:index");
                cout << " ** input :: ivar[" << ivar << "/" << nvar << "] name["
                     << _tds->getAttribute(ivar)->GetName() << "]" << endl;
                if(  isform == NULL )
                {
                    _ninput++;
                    inputVar += ((inputVar=="") ? TString("") : TString(":") ) +  _tds->getAttribute(ivar)->GetName() ;
                }
                else
                {
                    _ninformula++;
                    informulaes.push_back( isform->getFormula() );
                }
		delete myTree;
            }
        }
        
        if (_noutput == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TMultiGenCode::init [%s] ERROR\n the Number of output [%d] is null, while it should not",
                            _tds->GetName(), _noutput));
        if (_nCost == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMultiGenCode::init\n the cost attribute is not not found in the TDataServer"));
        
        cout << " ** TMultiGenCode::init _sCost[" << _sCost << "] _nCost[ "
             << _nCost << "]" << endl;

        if( _ninformula != 0)
        {
            cout << " ** TMultiGenCode::init Creating the formula for new inputs " << endl;

            _informTree = new TNtupleD("___internalTreeOptimizer___", "foo__optim__bar", inputVar);

            for(unsigned int ifor=0; ifor<informulaes.size(); ifor++)
            {
                _vinForm.push_back( new TTreeFormula(Form("Formula_%d",ifor), informulaes.at(ifor).Data(), _informTree));
                if(_vinForm.back()->GetNdim()== 0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TMultiGenCode::init ERROR\n the provided formula (%s) is incorrect",
                                                                    informulaes.at(ifor).Data()));
            }
        }

        // The size is increased by one for the iterator
        _dinputValue = new double[_ninput + _ninformula + _noutput + _noutformula + 1];
        _doutputValue = new double[_noutput + 1];
        _xin = new double[_ninput ];     
        _codeindex = new int[ _ninput + 1 ];
        _codeindex[0]=-1;
        _codevalue = new double[_ninput + 1 ];
        
        clean();

        //_code->init();

        // Go to the working directory
        gSystem->cd(_code->getWorkingDirectory());
        // Make the Results directory for distribution
        TString sresultsdirectory = TString("UranieResults");
        void * openFileResult = gSystem->OpenDirectory(sresultsdirectory);
        if (openFileResult == 0)
            if (gSystem->mkdir(sresultsdirectory, kTRUE) == -1)
                throw URANIE::Exceptions::UExceptions(__FILE__,
                __LINE__,
                        Form(
                                "TLauncher::init ERROR : the subdirectory \"%s\" can not be created in the workindirectory \"%s\"",
                                sresultsdirectory.Data(),
                                _code->getWorkingDirectory().Data()));
        gSystem->FreeDirectory(openFileResult);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////
// DoEval
//////////////////////////////////
double TMultiGenCode::DoEval(const double * par) const
{
    Int_t niter = (Int_t) _tds->getTuple()->GetEntries();
    double fval = 0.1;
    if (_code->getLog())
        cout << "iter[" << niter << "]" << endl;
    if (_code != NULL)
    {
        _dinputValue[0] = 1.0 + niter;

        // Remark: If we want to save all computations, we need a constant Bool_t _bSave,
        // and call the the method _code->init() at each new computation.
        _code->init();
        // Pretraitment of the data
        for (int i = 0; i < _ninput; i++)
        {
            _dinputValue[1 + i] = par[i];
	    if (_code->getLog())
                cout << "**** input i[" << i << "] val[" << _dinputValue[1 + i] << "]" << endl;
            
            if( _ninformula !=0 )
                _xin[i] = _dinputValue[1 + i];
        }

        if( _ninformula != 0)
        {
            _informTree->Fill(_xin);
            for(unsigned int ifor=0; ifor<_vinForm.size(); ifor++)
                _dinputValue[1 + _ninput + _noutput + ifor] = _vinForm.at(ifor)->EvalInstance();
            
            if( _codeindex[0] == -1)
            {
                _codeindex[0] = 0; int counter=1;
                for(int ifile=0; ifile<_code->getNInputFiles(); ifile++)
                    for(int iatt=0; iatt<_code->getNInputAttributes(ifile); iatt++)
                    {
                        for(int ibr = 0; ibr < _tds->getNAttributes(); ibr++)
                            if( _code->getInputAttribute(ifile, iatt)->GetName() == _tds->getAttribute(ibr)->GetName() )
                            {
                                _codeindex[counter] = ibr + 1 + _noutput;
                                counter++;
                            }
                    }
          
            }

            for(int ival=0; ival<_ninput+1; ival++)
		_codevalue[ival] = _dinputValue[ _codeindex[ival] ];

        }
        
        int ind = 1;
        _code->preTraitment(ind, ((_ninformula != 0) ? _codevalue : _dinputValue) );

        // Lancement du code
        _code->run();


        // Post-traitment of the data
        ind = 0;
        _code->postTraitment(ind, _doutputValue);
        if (_code->getLog())
        {
            cout << endl << "*************************************" << endl
                    << " *** TMultiGenCode DoEval _nCost[" << _nCost << "]"
                    << endl;
            for (Int_t ii = 0; ii <= _noutput; ii++)
                cout << " ** ii(" << ii << "/" << _nCost << ") val("
                        << _doutputValue[ii] << ")" << endl;
            cout << " *** End Of TMultiGenCode DoEval" << endl
                    << "*************************************" << endl;
	}

        //Fill the output values in the input vector
        // +1 because the first value is the index
        for (int i = 1; i <= (_noutput + _ninformula + _noutformula); i++)
	{
            int test = ((niter==0)? ((i<=_noutput)?_ninformula:((_ninformula==0)?0:((i>(_noutput+_ninformula))?0:-_noutput))):0);
	    // cout << "(" << _ninput + i  + test << ":"<<_ninput<<" + "<<i<<"  + "<<test << ") ";
            _dinputValue[_ninput + i  + test ] = ((i<=_noutput) ? _doutputValue[i] : 0);
	}
        _tds->getTuple()->Fill(_dinputValue);
	_tds->fillOthersAttributes();
		
        fval = _tds->getValue(_sCost, niter);
        if (_code->getLog())
            cout << " *** output _nCost[" << _nCost << "] _sCost[" << _sCost << "] val[" << fval << "]" << endl;
        
        // Export the TDS in the uranieLauncher working directory
        _tds->exportData(Form("_%s_iteration_%d_.dat", _tds->GetName(), niter));

        // Save the TDSNtupleD in a RootFile
        if ( !_bintermedstep )
        {
            TString srootfile = _code->getWorkingDirectory();
            srootfile += "/UranieResults/uranielauncher.root";
            if (niter % 5 == 0)
            {
                TString ssavefile = srootfile + TString(".bak");
                gSystem->CopyFile(srootfile.Data(), ssavefile.Data(), kTRUE);
            }
            TFile * rootfile = TFile::Open(srootfile.Data(), "RECREATE");
            _tds->getTuple()->SetDirectory(0);
            _tds->getTuple()->Write();
            rootfile->Close();
            _tds->getTuple()->SetDirectory(gDirectory);
        }

        // Visualisation
        if (1)
        {
            _tds->draw(Form("%s:%s", _sCost.Data(), _tds->getIteratorName()), "", "lp");
            gPad->Modified();
            gPad->Update();
        }
    }

    return fval;
}

//////////////////////////////////
// clean
//////////////////////////////////
void TMultiGenCode::clean()
{
    TString sworkindirectory = _code->getWorkingDirectory();
    // Place in the good directory
    if (sworkindirectory.Length() != 0)
        gSystem->cd(sworkindirectory);

    // Clean the subdirectory "UranieLauncher_*"
#ifdef WIN32
//    gSystem->Exec("rd /s /q UranieLauncher_*");
	gSystem->Exec( Form("for /d %%i in ( %s\\UranieLauncher_*) do @rmdir /s /q %%i", gSystem->pwd()));
#else
    gSystem->Exec("rm -fr UranieLauncher_*");
#endif    
}
} // Fin du namespace Optimizer
} // Fin du namespace URANIE
