//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TObjective.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   fri jul 6 13:48:06 CEST 2007
 
 \brief Interface of class URANIE::Optimize::TObjective
 
 */
/**
 
 * @class URANIE::Optimizer::TObjective
 *
 * @brief Description of the class TObjective.
 * This class computes the ojective (from L2 point of view).
 *
 * To be written by the developper.
 *
 */
#ifndef TOBJECTIVE_H
#define TOBJECTIVE_H

#include <iostream>
using namespace std;

//ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TMatrixT.h"

// Uranie
#include "Optimizer.h"
#include "TDataServer.h"
#include "TOutputFile.h"

namespace URANIE
{
namespace Optimizer
{
class TObjective: public TNamed
{
public:
    enum EOptimType
    {
        kFunction, kSumOfSquare
    };
public:
    Double_t _dweight;                     ///< Weight of the Objective function
    Double_t _dweightold;  ///< The old Weight of the Objective function to save
    Double_t _dvalue;                      ///< Value  of the Objective function
    Bool_t _blog;                          ///< Boolean to edit the log
    TString _syhat;                        ///< Name of yhat
    TString _systar;                       ///< Name of ystar
    int     _nshare;                        ///< number of object accessing it
    URANIE::DataServer::TDataServer *_tds;     ///< Pointer of a TDS into YHat
    URANIE::Launcher::TOutputFile *_outfile; ///< Pointer of Output file of TCode into YHAT
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor by name only
    /*! 
     
     \param name Name of the objective
     */
    TObjective(TString name);
    //! Default destructor
    virtual ~TObjective();
    //@}
    //---------------------------------------------
    /** @name Value of the Objective
     **/
    //@{
    // Return the value of the objective time the weight
    Double_t getObjective() const
    {
        return _dweight * _dvalue;
    }
    // Return only the value of the objective without time the weight
    Double_t getObjectiveWithoutWeight() const
    {
        return _dvalue;
    }
    // Compute the value of the objective, and return the value time the weight
    virtual Double_t getObjective(Double_t *dvalues, Int_t nrow);
    Bool_t isActive()
    {
        if (TMath::Abs(_dweight) < 0.001)
            return kFALSE;
        else
            return kTRUE;
    }
    void unActive()
    {
        _dweight = 0.0;
    }
    void active()
    {
        _dweight = _dweightold;
    }

    void delShare()
    {
      _nshare--;
    }
    void addShare()
    {
      _nshare++;
    }
    int getShare()
    {
      return _nshare;
    }

    void setWeight(Double_t val)
    {
        _dweight = val;
        _dweightold = val;
    }
    Double_t getWeight()
    {
        return _dweight;
    }

    void setYStar(URANIE::DataServer::TDataServer *tds, TString ystar)
    {
        _tds = tds;
        _systar = ystar;
    }
    void setYHat(URANIE::Launcher::TOutputFile *outfile, TString yhat)
    {
        _outfile = outfile;
        _syhat = yhat;
    }

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Optimizer::TObjective, ID_OPTIMIZER)
    //Classe de
};

} // Fin du namespace Optimizer
} // Fin du namespace URANIE
#endif
