//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOptimizer.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu jul 3 10:38:56 CEST 2007
 
 \brief Implementation of the class URANIE::Optimizer::TOptimizer

 \par Modifications
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - add a IFDEF WIN32 to replace rm -fr by del /Q on Windows platform
 \li C.Forgues (2010-12-21 )
 - optimize()->optimizeOld(), optimizeMinuit2()->optimize().
 \li C.Forgues (2010-10-26 )
 PrintLevel for minuit2 is set to 0. Errors were detected with Valgrind with PrinLevel >= 1.
 \li F. Gaudier (2010-09-21 12:18:59)
 Change the wrapper to the new Minuit2 library (method optimizeMinuit2)
 \todo After the validation on the new wrapper for External Code, and sum of square,
 change the optimize method to optimizeOLD and the optimizeMinuit2 method to optimize.
 \see http://root.cern.ch/drupal/content/numerical-minimization
 \see http://root.cern.ch/viewvc/trunk/math/minuit2/test/testNewMinimizer.cxx?view=markup
 
 */
#include <iostream>
using std::cout;
using std::endl;

// ROOT
//fga 2015/01/12  #include "Api.h"
#include "TTreeFormula.h"
#include "Math/Minimizer.h"
#include "Math/Factory.h"
#include "Math/IFunction.h"

// Minuit2
#include "Minuit2/Minuit2Minimizer.h"

// URANIE
#include "TOptimizer.h"
#include "TFCNAnalyticalFunction.h"
#include "TFCNCode.h"
#include "TAttributeFormula.h"
#include "TSumOfSquares.h"
#include "TMultiGenAnalyticalFunction.h"
#include "TMultiGenCode.h"
#include "TMultiGenSumOfSquares.h"
#include "UExceptions.h"

ClassImp (URANIE::Optimizer::TOptimizer)

namespace URANIE
{
namespace Optimizer
{
TOptimizer::TOptimizer(URANIE::DataServer::TDataServer *tds,
		       URANIE::Launcher::TCode *code) :
    _bsave(kFALSE), _bclean(kFALSE), _blog(kFALSE), _nPrintLevel(0), _tds(tds), _code(code), 
     _nMaxIterations(50), _nMaxFunctionCalls(10000), _dTolerance(0.01),  _objectives(0),
    _fMethod(NULL), _sFunctionName(""), _optimProblem(kMinimizeCode), _optimMethod(kMigrad), _sInput("*"), _sOutput(""), _sCost("")
{
    try
    {
        if( !tds )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the TDS is NULL for a TCode"));
        if( !code )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the Tcode is NULL with TDS (\"%s\")", tds->GetName()));


        if (_blog)
            cout << "constructs TOptimizer by TDataServer[" << tds->GetName()
                 << "] code" << endl;
        SetName(Form("Optimize_Code_%s_TDS_%s", code->GetName(), tds->GetName()));
        SetTitle(
                    Form("Optimize Code[%s] with TDS[%s]", code->GetName(),
                         tds->GetName()));

        // Get the current directory
        _sdirectory = gSystem->pwd();

        // Boucle sur les TAttribut du TDataServer
        Int_t natt = _tds->getNAttributes();
        URANIE::DataServer::TAttribute *att = 0;
        for (Int_t iatt = 0; iatt < natt; iatt++)
        {
            att = tds->getAttribute(iatt);

            if (iatt == 0)
                _sInput = att->GetName();
            else
                _sInput += Form(":%s", att->GetName());

            if (_blog)
                att->printLog();
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            exit(-1);
    }

    // Ouput Attributes
    // \todo : 2009-06-11 11:58:34 put all this intruction in the Tcode::Init method
    TList *listOfDataOutput=NULL;
    try
    {
        if (_blog)
            cout << " ** Output Attributes" << endl;
        listOfDataOutput = new TList();
        TString strOut = Form("n__calculs__/D:%s/D", _tds->getIteratorName());
        Int_t noutfile = _code->getNOutputFiles();
        Int_t noutput = 0;
        if (_blog)
            cout << " ** N output Files [" << noutfile << "]" << endl;
        for (Int_t ioutfile = 0; ioutfile < noutfile; ioutfile++)
        {
            Int_t nattfile = _code->getNOutputAttributes(ioutfile);
            if (_blog)
                cout << " ** N attributes output Files [" << ioutfile << "/"
                        << noutfile << "] - natt[" << nattfile << "]" << endl;
            for (Int_t iattfile = 0; iattfile < nattfile; iattfile++)
            {
                URANIE::DataServer::TAttribute * att = _code->getAttribute(
                        ioutfile, iattfile);
                att->setOutput();
                _tds->addAttribute(att);
                TString sattname = att->GetName();

                if (noutput == 0)
                {
                    _sOutput = sattname;
                    _sCost = sattname;
                }
                else
                    _sOutput += Form(":%s", sattname.Data());

                if (_blog)
                    cout << " ** Id[" << noutput + 1 << "] Name["
                            << att->GetName() << "]" << endl;
                noutput++;
            }
        }

	delete listOfDataOutput;
	listOfDataOutput=NULL;
        if (_blog)
            cout << " ** End Of Output Attributes [" << noutput << "] ***"
                    << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete listOfDataOutput;
	listOfDataOutput=NULL;
        throw ue;
    }
    if (_blog)
        _code->printLog();

    // Init the code
    _code->initFileType();
    
}

TOptimizer::TOptimizer(URANIE::DataServer::TDataServer *tds,
#ifdef ROOT5
                       void* fcn,
#else
                       void (*fcn)(Double_t*, Double_t*),
#endif
		       TString sinput, TString soutput) :
    _sdirectory(""), _bsave(kFALSE), _bclean(kFALSE), _blog(kFALSE), _nPrintLevel(0), _tds(tds), _code(NULL), 
     _nMaxIterations(50), _nMaxFunctionCalls(10000), _dTolerance(0.01),  _objectives(0),
    _fMethod(fcn), _sFunctionName(""), _optimProblem(kMinimizeFunction), _optimMethod(kMigrad), _sInput(sinput), _sOutput(soutput), _sCost("")
{
     if (_blog)
        cout << endl << " ************* TOptimizer with Function "
            << Form("\"%s\"", (char*)fcn) << " Input [" << sinput << "] Ouput ["
            << soutput << "]" << endl;

    // Get the current directory
    _sdirectory = gSystem->pwd();

    try
    {
        if( !tds )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the TDS is NULL for a TLauncherFunction"));

        if( !tds->getDataSpecification() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the list of attributes for the TDS (\"%s\") is EMPTY", tds->GetName()));


        TString functionName;

#ifdef ROOT5
        //     char *functionName = G__p2f2funcname((void *)(double*, Double_t*)(fcn));
        functionName = G__p2f2funcname(fcn);
        try
        {
            if (!functionName)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::TOptimizer TDS[%s] ERROR\n the fonction name \"%s\" is not defined",
                                _tds->GetName(), (char*)fcn));

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            char *fname;
            fname = G__p2f2funcname(&fcn);
            if (fname)
            {
                cout << " ************** fname existe" << endl;
            }
            else
            {
                cout << " -------------- fname existe pas" << endl;
            }
            throw ue;
        }

        
        G__ClassInfo globalscope;
        long dummy;
        G__MethodInfo method = globalscope.GetMethod(functionName.Data(),
                "double*,double*", &dummy);

        if (_blog)
            cout << endl << " ** Function \"" << functionName << "\" found"
                    << endl;
#else
        functionName = soutput;
#endif
        
        SetName(Form("Uranie_Optimizer_Function_%s", functionName.Data()));
        SetTitle(
            Form("Optimize Function[%s] with TDS[%s]", functionName.Data(),
                        tds->GetName()));

        // The input attributes
        if (_sInput.Length() == 0)
        {

            TList * listofallattributes = _tds->getListOfAttributes("*");
            TIter next(listofallattributes);
            // First input attribute
            URANIE::DataServer::TAttribute *att =
                    (URANIE::DataServer::TAttribute *) next();
            _sInput = att->GetName();
            while ( (att = (URANIE::DataServer::TAttribute *) next()) )
            {
                _sInput += Form(":%s", att->GetName());
                if (_blog)
                    cout << " ** crt att[" << att->GetName() << "], _sInput["
                            << _sInput << "]" << endl;
            }
            if (_blog)
                cout << " ** _sInput[" << _sInput << "]" << endl;

	    delete listofallattributes;
        }

        // The output attributes
        if (_sOutput.Length() == 0)
            _sOutput = functionName;
        else
            _sOutput = soutput;

        TString slabel;
        TString satt(_sOutput.Data());
        satt.ReplaceAll(":", " ");
        satt += " ";
        istringstream inStream((char *) satt.Data());

        // The first output attribute
        Int_t ncpt = 1;
        string inName;
        inStream >> inName;
        URANIE::DataServer::TAttribute * att =
                new URANIE::DataServer::TAttribute(inName.c_str());
        att->setOutput();
        att->delShare();
        _tds->addAttribute(att);
        _sCost = att->GetName();
        if (_blog)
            cout << " ** Output Id[1] Name[" << att->GetName() << "]" << endl;

        while (inStream >> inName)
        {
            att = new URANIE::DataServer::TAttribute(inName.c_str());
            att->setOutput();
	    att->delShare();
            _tds->addAttribute(att);
            if (_blog)
                cout << " ** Output Id[" << ncpt++ << "] Name["
                        << att->GetName() << "]" << endl;
        }

        // Create the tuple if necessary
        try
        {
            _tds->getTuple();
        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            _tds->createTuple();
        }
        if (_blog)
            _tds->getTuple()->Print();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

TOptimizer::TOptimizer(URANIE::DataServer::TDataServer *tds, const char* fcn,
		       TString sinput, TString soutput) :
    _sdirectory(""), _bsave(kFALSE), _bclean(kFALSE), _blog(kFALSE), _nPrintLevel(0), _tds(tds), _code(NULL), 
     _nMaxIterations(50), _nMaxFunctionCalls(10000), _dTolerance(0.01),  _objectives(0),
    _fMethod(NULL), _sFunctionName(""), _optimProblem(kMinimizeFunction), _optimMethod(kMigrad), _sInput(sinput), _sOutput(soutput), _sCost("")

{
    cout << endl << " ************* TOptimizer with Function "
	 << Form("\"%s\"", (char*)fcn) << " Input [" << sinput << "] Ouput ["
	 << soutput << "]" << endl;

    // Get the current directory
    _sdirectory = gSystem->pwd();

    _sFunctionName=TString( fcn ) ;

    try
    {
        if( !tds )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the TDS is NULL for a TLauncherFunction"));

        if( !tds->getDataSpecification() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::TOptimizer: the list of attributes for the TDS (\"%s\") is EMPTY", tds->GetName()));


#ifdef ROOT5
	G__ClassInfo globalscope;
	long dummy;
	G__MethodInfo method = globalscope.GetMethod(_sFunctionName.Data(),
						     "double*,double*", &dummy);
// #else
//         _fMethodCall = new TMethodCall();
//         _fMethodCall->InitWithPrototype(fcn, "double*,double*");
//         if (_fMethodCall->IsValid() == kFALSE)    
//             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
//                         Form("TOptimizer::TOptimizer: function non found (\"%s\") ", fcn));
#endif

	if (_blog)
	    cout << endl << " ** Function \"" << fcn << "\" found"
		 << endl;

	SetName(Form("Uranie_Optimizer_Function_%s", fcn));
	SetTitle(
	    Form("Optimize Function[%s] with TDS[%s]", fcn,
		 tds->GetName()));

	// The input attributes
	if (_sInput.Length() == 0)
	{

	    TList * listofallattributes = _tds->getListOfAttributes("*");
	    TIter next(listofallattributes);
	    // First input attribute
	    URANIE::DataServer::TAttribute *att =
		(URANIE::DataServer::TAttribute *) next();
	    _sInput = att->GetName();
	    while ( (att = (URANIE::DataServer::TAttribute *) next()) )
	    {
		_sInput += Form(":%s", att->GetName());
		if (_blog)
		    cout << " ** crt att[" << att->GetName() << "], _sInput["
			 << _sInput << "]" << endl;
	    }
	    if (_blog)
		cout << " ** _sInput[" << _sInput << "]" << endl;

	    delete listofallattributes;
	}

	// The output attributes
	if (_sOutput.Length() == 0)
	    _sOutput = fcn;
	else
	    _sOutput = soutput;

	TString slabel;
	TString satt(_sOutput.Data());
	satt.ReplaceAll(":", " ");
	satt += " ";
	istringstream inStream((char *) satt.Data());

	// The first output attribute
	Int_t ncpt = 1;
	string inName;
	inStream >> inName;
	URANIE::DataServer::TAttribute * att =
	    new URANIE::DataServer::TAttribute(inName.c_str());
	att->setOutput();
	att->delShare();
	_tds->addAttribute(att);
	_sCost = att->GetName();
	if (_blog)
	    cout << " ** Output Id[1] Name[" << att->GetName() << "]" << endl;

	while (inStream >> inName)
	{
	    att = new URANIE::DataServer::TAttribute(inName.c_str());
	    att->setOutput();
	    att->delShare();
	    _tds->addAttribute(att);
	    if (_blog)
		cout << " ** Output Id[" << ncpt++ << "] Name["
		     << att->GetName() << "]" << endl;
	}

	// Create the tuple if necessary
	try
	{
	    _tds->getTuple();
	} catch (URANIE::Exceptions::UWarningExceptions& ue)
	{
	    _tds->createTuple();
	}
	if (_blog)
	    _tds->getTuple()->Print();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    }

}

TOptimizer::~TOptimizer()
{
    if (_blog)
        cout << "destructs TOptimizer" << endl;

    cout<<"In optimizer destructor, vector is size "<<_objectives.size()<<endl;
    for(vector<TObjective*>::iterator it=_objectives.begin(); it!=_objectives.end(); it++)
      {
	(*it)->delShare();
	if( (*it)->getShare()==0)
	  delete (*it);
      }
    _objectives.clear();
}

// -----------------------------------------------------------------------
//  Manipulate the Objectives
// -----------------------------------------------------------------------
void TOptimizer::addObjective(TString name,
        URANIE::DataServer::TDataServer *tds, TString ystar,
        URANIE::Launcher::TOutputFile *outfile, TString yhat, Double_t weight)
{

    try {
        if( !tds )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the reference TDS is NULL with ystar (\"%s\"), yhat (\"%s\") and TOutputfile (\"%s\").", ystar.Data(), yhat.Data(), outfile->GetName()));

        if( !tds->getDataSpecification() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the list of attributes for the reference TDS is EMPTY with ystar (\"%s\"), yhat (\"%s\") and TOutputfile (\"%s\").", ystar.Data(), yhat.Data(), outfile->GetName()));

        if( !tds->isAttribute(ystar) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the reference attribute \"%s\" does not exist in the reference TDataServer (\"%s\", \"%s\").", ystar.Data(), tds->GetName(), tds->GetTitle()));

        if( !outfile )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the TOutputFile is NULL with TDs(\"%s\"), ystar (\"%s\") and yhat (\"%s\")", tds->GetName(), ystar.Data(), yhat.Data()));
        if( !outfile->isAttribute(yhat) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the attribute \"%s\" does not exist in the TOutputfile \"%s\".", yhat.Data(), outfile->GetName()));

        _optimProblem = kSumOfSquare;

        cout << endl << endl << " ** addObjective ystar(" << ystar << ") in TDS (" <<  tds->GetName() << ") yhat(" << yhat << ") in the output file(" << outfile->GetName() << ")" << endl;

        TObjective *tobj = new TObjective(name);
        tobj->setWeight(weight);
        tobj->setYStar(tds, ystar);
        tobj->setYHat(outfile, yhat);

	tobj->addShare();
        _objectives.push_back(tobj);
     }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}
void TOptimizer::addObjective(TString name,
        URANIE::DataServer::TDataServer *tds, TString ystar, TString sigma,
        URANIE::Launcher::TOutputFile *outfile, TString yhat, Double_t weight)
{
    try {
        if( !tds )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the reference TDS is NULL with ystar (\"%s\"), sigma (\"%s\"), yhat (\"%s\") and TOutputfile (\"%s\").", ystar.Data(), sigma.Data(), yhat.Data(), outfile->GetName()));

        if( !tds->getDataSpecification() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the list of attributes for the reference TDS is EMPTY with ystar (\"%s\"), sigma (\"%s\"), yhat (\"%s\") and TOutputfile (\"%s\").", ystar.Data(), sigma.Data(), yhat.Data(), outfile->GetName()));

        if( !tds->isAttribute(ystar) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the reference attribute \"%s\" does not exist in the reference TDataServer (\"%s\", \"%s\").", ystar.Data(), tds->GetName(), tds->GetTitle()));
        if( !tds->isAttribute(sigma) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the sigma attribute \"%s\" does not exist in the reference TDataServer (\"%s\", \"%s\").", sigma.Data(), tds->GetName(), tds->GetTitle()));

        if( !outfile )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the TOutputFile is NULL with TDs(\"%s\"), ystar (\"%s\") and yhat (\"%s\")", tds->GetName(), ystar.Data(), yhat.Data()));
        if( !outfile->isAttribute(yhat) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TOptimizer::addObjective : the attribute \"%s\" does not exist in the TOutputfile \"%s\".", yhat.Data(), outfile->GetName()));
    _optimProblem = kSumOfSquare;

    cout << endl << endl << " ** addObjective ystar(" << ystar << ") and sigma(" << sigma <<") in TDS (" <<  tds->GetName() << ") yhat(" << yhat << ") in the output file(" << outfile->GetName() << ")" << endl;

    TObjectiveUncertainty *tobj = new TObjectiveUncertainty(name);
    tobj->setWeight(weight);
    tobj->setYStar(tds, ystar, sigma);
    tobj->setYHat(outfile, yhat);

    tobj->addShare();
    _objectives.push_back(tobj);
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
   {
       ue.printMessage();
       throw ue;
   }
}

void TOptimizer::unactiveObjective(TString name)
{
    Int_t nobj = (Int_t)(_objectives.size());
    Bool_t bObjectiveFound = kFALSE;
    for (Int_t indx = 0; indx < nobj; indx++)
    {
        TString sname = _objectives[indx]->GetName();
        if (sname == name)
        {
            _objectives[indx]->unActive();
            bObjectiveFound = kTRUE;
            break;
        }
    }
    
    if(!bObjectiveFound)
    {
        try{
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TOptimizer::unactiveObjective  ERROR\n there is not a objective named [%s]",
                            name.Data()));
        }
        catch(URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
        
    }
}
void TOptimizer::activeObjective(TString name)
{
    Int_t nobj = (Int_t)(_objectives.size());
    Bool_t bObjectiveFound = kFALSE;
    for (Int_t indx = 0; indx < nobj; indx++)
    {
        TString sname = _objectives[indx]->GetName();
        if (sname == name)
        {
            _objectives[indx]->active();
            bObjectiveFound = kTRUE;
            break;
        }
    }
    
    if(!bObjectiveFound)
    {
        try{
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TOptimizer::activeObjective  ERROR\n there is not a objective named [%s]",
                            name.Data()));
        }
        catch(URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
        
    }
    
}

void TOptimizer::invertObjectives()
{
    Int_t nobj = (Int_t)(_objectives.size());
    for (Int_t indx = 0; indx < nobj; indx++)
    {
        if (_objectives[indx]->isActive())
            _objectives[indx]->unActive();
        else
            _objectives[indx]->active();
    }
}

// -----------------------------------------------------------------------
//  Add input and output attribute defined respectively from other inputs and outputs
// -----------------------------------------------------------------------
void TOptimizer::addOutputVariable(TString soutput)
{

    if (_blog)
        cout << " ** TOptimizer::addOutputVariable[" << soutput << "]" << endl;

    TTreeFormula *inForm=NULL;
    TString inName;
    
    if( soutput != "")
    {
        try
        {
            soutput.ReplaceAll(":", " ");
            soutput += " ";

            istringstream inStream((char *) soutput.Data());
            while (inStream >> inName)
            {
                if( _blog)
                    cout<<" ** Adding extra output variable ["<<inName<<"]"<<endl;
                inForm=new TTreeFormula(inName.Data(), inName.Data(), _tds->getTuple());
                if (inForm->GetNdim()== 0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form(
                                                                   "TOptimizer::addOutputVariable[%s]  ERROR: the formula [%s] of the output attribute [%s] is not correct",
                                                                   inName.Data(), inName.Data(), inName.Data()));
                else
                {
                    _tds->addAttribute(inName, inName);
                    _tds->getAttribute(inName)->setOutput(); 
                    delete inForm;
                }
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            delete inForm;
            throw ue;
        }
    }

}

// -----------------------------------------------------------------------
//  Select a cost in the output attribute
// -----------------------------------------------------------------------
void TOptimizer::selectCost(TString scost)
{

    if (_blog)
        cout << " ** TOptimizer::selectCost[" << scost << "] soutput[" << _sOutput << "]" << endl;

    try
    {
        Int_t natt = scost.CountChar(':') + 1;

        if (natt != 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TOptimizer::selectCost [%s] ERROR\n the Number of cost is not equal to 1 [%d]",
                            GetName(), natt));

        if(_tds->isAttribute(scost))
        {
            URANIE::DataServer::TAttribute *att = _tds->getAttribute(scost);
            if (!att->isOutput())
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form(
                                                               "TOptimizer::selectCost [%s] ERROR\n the attribute [%s] is not an output attribute",
                                                               GetName(), scost.Data()));
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TOptimizer::selectCost ERROR\n the attribute [%s] is not an existing  attribute",
                                                           scost.Data()));
        _sCost = scost;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Manipulate the Parameters
// -----------------------------------------------------------------------
void TOptimizer::fixParameter(TString name)
{
    try{
        if(!_tds->isAttribute(name))
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::fixParameter\n The attribute \"%s\" is not in the \"TDS\".",
                                name.Data()));
        }
        else
        {
            URANIE::DataServer::TAttribute *att = _tds->getAttribute(name);
            URANIE::DataServer::TAttribute::EOrigin norigin = att->getOrigin();
            switch (norigin)
                {
                case URANIE::DataServer::TAttribute::kConstant:
                    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::fixParameter\n The attribute \"%s\" is already fixed.",
                                name.Data()));
                    break;
                case URANIE::DataServer::TAttribute::kAttribute:
                    att->setOrigin(URANIE::DataServer::TAttribute::kConstant);
                    break;
                default:
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::fixParameter\n The attribute \"%s\" isn't either a kAttribute or a kConstant.",
                                name.Data()));
                    break;
                }
            
        }
    
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } 

}

void TOptimizer::unfixParameter(TString name)
{
    try{
        if(!_tds->isAttribute(name))
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::unfixParameter\n The attribute \"%s\" is not in the \"TDS\".",
                                name.Data()));
        }
        else
        {
            URANIE::DataServer::TAttribute *att = _tds->getAttribute(name);
            URANIE::DataServer::TAttribute::EOrigin norigin = att->getOrigin();
            switch (norigin)
                {
                case URANIE::DataServer::TAttribute::kConstant:
                    att->setOrigin(URANIE::DataServer::TAttribute::kAttribute);
                    break;
                case URANIE::DataServer::TAttribute::kAttribute:
                    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::unfixParameter\n The attribute \"%s\" is already inconstant.",
                                name.Data()));
                    break;
                default:
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TOptimizer::unfixParameter\n The attribute \"%s\" isn't either a kAttribute or a kConstant.",
                                name.Data()));
                    break;
                }
            
        }
    
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } 
    
    
}
void TOptimizer::invertParameters()
{
    Int_t nvar = _tds->getDataSpecification()->getNAttributes();
    if (_blog)
        cout << endl << endl << " **** TOptimizer::invertParameters nvar["
                << nvar << "]" << endl;
    for (Int_t ivar = 0; ivar < nvar; ivar++)
    {
        URANIE::DataServer::TAttribute *att =
                _tds->getDataSpecification()->getAttribute(ivar);
        if (att != NULL) // verifies that the attribute is not null
        {
            if(att->isInput()) // verifies that the attribute is a parameter
            {
                URANIE::DataServer::TAttribute::EOrigin norigin = att->getOrigin();
                cout << " ** att " << ivar << "/" << nvar << " : name["
                        << att->GetName() << "] Origin[" << norigin << "]" << endl;
                switch (norigin)
                {
                case URANIE::DataServer::TAttribute::kConstant:
                    if (_blog)
                        cout << " **   name [ " << att->GetName()
                                << "] Origin[kConstant]" << endl;
                    att->setOrigin(URANIE::DataServer::TAttribute::kAttribute);
                    break;
                case URANIE::DataServer::TAttribute::kAttribute:
                    if (_blog)
                        cout << " **   name [ " << att->GetName()
                                << "] Origin[kAttribute]" << endl;
                    att->setOrigin(URANIE::DataServer::TAttribute::kConstant);
                    break;
                default:
                    break;
                }
            }
        }
        else
        {
            cout << "Pointer [" << ivar << "] is NULL" << endl;
        }
    }
    if (_blog)
        cout << " **** End Of TOptimizer::invertParameters" << endl << endl;
}

// ----------------------------------------------------------------------
//  TOptimizer::init
// ----------------------------------------------------------------------
void TOptimizer::init(Option_t *option)
{
    if (_blog)
        cout << "TOptimizer::init" << endl;

    if (_code != NULL)
        _code->init(option);
}

// ----------------------------------------------------------------------
//  TOptimizer::optimize
// ----------------------------------------------------------------------
void TOptimizer::optimize(Option_t *option)
{
    if (_blog)
        cout << "TOptimizer::optimize" << endl;

    if (_code != NULL)
    {
        // Clean if necessary
        if (_bclean)
            clean(option);
        // Initialisation
        init(option);
    }
    
    TString soption = option;
    soption.ToLower();

    TList* listofattributes = NULL;
    // The optimization process
    TString sLibraryName = "Minuit2";
    TString sMethodName = "Migrad";

    switch (_optimMethod)
    {
    case kSimplex:
        sMethodName = "Simplex";
        break;
    case kCombined:
        sMethodName = "Combined";
        break;
    case kScan:
        sMethodName = "Scan";
        break;
    default:
        sMethodName = "Migrad";
    }
    cout << " **   sLibraryName[" << sLibraryName << "]" << endl;
    cout << " **    sMethodName[" << sMethodName << "]" << endl;

    //    ROOT::Math::Minimizer * minuit2 = ROOT::Math::Factory::CreateMinimizer(sLibraryName.Data(), sMethodName.Data());
    ROOT::Minuit2::Minuit2Minimizer * minuit2 =
            new ROOT::Minuit2::Minuit2Minimizer(sMethodName.Data());

    // The problem
    ROOT::Math::IBaseFunctionMultiDim *tmgPtr=NULL;
    switch (_optimProblem)
    {
    case kMinimizeFunction:
    {
        cout << " ** Problem[kMinimizeFunction]" << endl;
        TMultiGenAnalyticalFunction *tmgAF; 
        if( _fMethod != NULL)
// #ifdef ROOT5
            tmgAF = new TMultiGenAnalyticalFunction(_tds, _fMethod, _sInput, _sOutput, _sCost);
// #else
//             tmgAF = new TMultiGenAnalyticalFunction(
//                 _tds,  _fMethodCall->GetMethodName(), _sInput, _sOutput, _sCost);
// #endif
        else
            tmgAF = new TMultiGenAnalyticalFunction(_tds, _sFunctionName.Data(), _sInput, _sOutput, _sCost);
	
        tmgPtr = tmgAF;
        minuit2->SetFunction(*tmgAF);
        break;
    }
    case kMinimizeCode:
    {
        cout << " ** Problem[kMinimizeCode]" << endl;
        TMultiGenCode * tmgC = new TMultiGenCode(_tds, _code, _sCost);
        tmgPtr = tmgC;
        minuit2->SetFunction(*tmgC);
        if( soption.Contains("nointermed") )
           tmgC->doNotSaveSteps();
           
        break;
    }
    case kSumOfSquare:
    {
        cout << " ** Problem[kSumOfSquare]" << endl;
        TMultiGenSumOfSquares *tmgSOS = new TMultiGenSumOfSquares(_tds, _code);
        tmgPtr = tmgSOS;

        // The objectives to construct the criteria
        Int_t nobj = (Int_t)(_objectives.size());
        for (Int_t indx = 0; indx < nobj; indx++)
        {
            cout << " ** indx[" << indx << "] Name["
                    << _objectives[indx]->GetName() << "]" << endl;
            //        ((TSumOfSquares *)tmgSOS)->addObjective( _objectives[indx] );
            tmgSOS->addObjective(_objectives[indx]);
        }
        // Init the TMultiGenSumOfSquares object
        tmgSOS->init();
        _sCost = tmgSOS->getCriteriaName();
        minuit2->SetFunction(*tmgSOS);
        break;
    }
    }
    cout << " ** _sCost[" << _sCost << "]" << endl;

    try
    {
        // Initial values
        Int_t norigin;
        Int_t nvar = _tds->getNInputAttributes();

        Int_t ncpt = 0;
        Double_t valdef = 1.23456789;
        Double_t valstep = 1.23456789;

        // Pass the Attributes informations to the TMinuit object
	listofattributes = _tds->getListOfAttributes(_sInput);
        if (_blog)
        {
            cout << " ** _sInput[" << _sInput << "]" << endl;
            listofattributes->Print();
        }

        if(_optimProblem==kMinimizeFunction)
        {
            TString localInput="";
            
            TList* listAllAtt = _tds->getListOfAttributes("*");
            TIter nextAll(listAllAtt);
            while (URANIE::DataServer::TAttribute *att =
                   (URANIE::DataServer::TAttribute *) nextAll())
            {
                norigin = att->getOrigin();
                if( att->isInput() && ! listofattributes->Contains(att) )
                    localInput+=att->GetName()+TString(":");
            }

            if( localInput != "")
            {
                localInput += _sInput;
                delete listofattributes;
                listofattributes = _tds->getListOfAttributes(localInput);
            }
            delete listAllAtt;
            
        }
        
        TIter next(listofattributes);
        while (URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) next())
        {
            norigin = att->getOrigin();
            if (norigin == URANIE::DataServer::TAttribute::kAttribute && dynamic_cast<URANIE::DataServer::TAttributeFormula*>(att)==NULL)
            {
                if (_blog)
                    att->printLog();
                att->getDefaultValue(valdef);
                att->getStepValue(valstep);
                minuit2->SetLimitedVariable(ncpt, att->GetName(), valdef,
                        valstep, att->getLowerBound(), att->getUpperBound());
                ncpt++;
            }
            else if (norigin == URANIE::DataServer::TAttribute::kConstant)
            {
                att->getDefaultValue(valdef);
                minuit2->SetFixedVariable(ncpt, att->GetName(),valdef);
                ncpt++;
            }
        }

        // Pass all the parameters of the optimization
        //PrintLevel mis a 0, semble entrainer des problemes sous WINDOWS s'il est superieur ou egal a 1;
        minuit2->SetPrintLevel(_nPrintLevel);
        minuit2->SetMaxIterations(_nMaxIterations);
        minuit2->SetMaxFunctionCalls(_nMaxFunctionCalls);
        minuit2->SetTolerance(_dTolerance);

        printState("Init");
        // launch the Minimization process
        Bool_t bres = minuit2->Minimize(); //_nMaxIterations, _dTolerance);
        if (_blog)
            cout << " ** Optimization finished with status["
                    << minuit2->Status() << "] Ncalls[" << minuit2->NCalls()
                    << "]" << endl;
        if (!bres)
        {
            TDatime theTime;
            TString sTimeExtension = Form("%4d%02d%02d%02d%02d%02d",
                    theTime.GetYear(), theTime.GetMonth(), theTime.GetDay(),
                    theTime.GetHour(), theTime.GetMinute(),
                    theTime.GetSecond());
            TString sErrorInformation;
            switch (minuit2->Status())
            {
            case 1:
                sErrorInformation = "Covariance was made pos defined";
                break;
            case 2:
                sErrorInformation = "Hesse is invalid";
                break;
            case 3:
                sErrorInformation = "Edm is above max";
                break;
            case 4:
                sErrorInformation = Form(
                        "Reached call limit %d with max %d requested",
                        minuit2->NCalls(), _nMaxFunctionCalls);
                break;
            case 5:
                sErrorInformation = "Any other failure";
            }

            gSystem->cd(_sdirectory);
            TString sFileName = Form("_Save_%s_Time_%s_Iteration_%d_.dat",
                    _tds->GetName(), sTimeExtension.Data(),
                    _tds->getNPatterns());
            _tds->exportData(sFileName);
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TOptimizer::optimize ERROR\n the optimization process is terminated with status [%d] \"%s\"\n The current TDS[%s] was exported in the file[%s]",
                            minuit2->Status(), sErrorInformation.Data(),
                            _tds->GetName(), sFileName.Data()));
        }

        // Get the current solution
        const Double_t *xmin = minuit2->X();

        ncpt = 0;
        next.Reset();
        while (URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) next())
        {
            norigin = att->getOrigin();

            if (norigin == URANIE::DataServer::TAttribute::kAttribute && dynamic_cast<URANIE::DataServer::TAttributeFormula*>(att)==NULL)
            {
                att->setDefaultValue(xmin[ncpt++]);
                att->getDefaultValue(valdef);
                if (_blog)
                    cout << "** End of Optimization " << ncpt + 1 << " / "
                            << nvar << " [" << att->GetName() << "]  ["
                            << valdef << "]" << endl;
            }
            else if(norigin == URANIE::DataServer::TAttribute::kConstant)
            {
                ncpt++;
            }
        }

        // Get The Y minimum value
        _tds->getAttribute(_sCost)->setDefaultValue(minuit2->MinValue());


        // Save the Minimum values for all output attributes
        switch (_optimProblem)
        {
        case kMinimizeFunction:
        {
            //Int_t nY = _sOutput.CountChar (':')+1;
            Double_t * dValYToXmin = dynamic_cast<TMultiGenAnalyticalFunction *>(tmgPtr)->getOutputValues();

            TString satt(_sOutput.Data());
            satt.ReplaceAll(":", " ");
            satt += " ";
            istringstream inStream((char *) satt.Data());
            string inName;
            Int_t iCrt = 0;

            inStream >> inName;
            _tds->getAttribute(inName.c_str())->setDefaultValue(dValYToXmin[iCrt++]);
            while (inStream >> inName)
            {
	      // cout<<"Name "<<inName << " is "<<_tds->getAttribute(inName.c_str())<<endl;
	      // cout<<"Name "<<inName << " is "<<_tds->getAttribute(inName.c_str())<<"  whose name is "<<_tds->getAttribute(inName.c_str())->GetName()<<endl;
                _tds->getAttribute(inName.c_str())->setDefaultValue(dValYToXmin[iCrt++]);
            }
            break;
        }
        case kMinimizeCode:
        {
            // TODO: Puts the optimum values for the TCode objects (FGA 2014/05/19)
            break;
        }
        case kSumOfSquare:
        {
            // TODO: Puts the optimum values for the kSumOfSquare objects (FGA 2014/05/19)
            break;
        }
//         default:
 //            break;
    }


        printState("Final");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	if(listofattributes!=NULL)
	  delete listofattributes;
	listofattributes=NULL;
        delete tmgPtr;
        tmgPtr = NULL;
        delete minuit2;
        minuit2 = NULL;
	if (_code != NULL)
	  terminate(option);	
        throw ue;
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
	if(listofattributes!=NULL)
	  delete listofattributes;
	listofattributes=NULL;
	delete tmgPtr;
        tmgPtr = NULL;
        delete minuit2;
        minuit2 = NULL;
	if (_code != NULL)
	  terminate(option);
        throw ue;
    }

    if(listofattributes!=NULL)
	delete listofattributes;
    listofattributes=NULL;

    delete tmgPtr;
    tmgPtr = NULL;

    delete minuit2;
    minuit2 = NULL;

    if (_code != NULL)
      terminate(option);

}

// ----------------------------------------------------------------------
//  TOptimizer::terminate
// ----------------------------------------------------------------------
void TOptimizer::terminate(Option_t *option)
{
    _code->terminate(option);
    if (_bclean && (_bsave == kFALSE))
        clean(option);

    gSystem->cd(_sdirectory);
    if (_blog)
    {
        cout << "PWD[" << gSystem->pwd() << "]" << endl;
        cout << "TLauncher::terminate" << endl;
    }
}

// ----------------------------------------------------------------------
//  TOptimizer::clean
// ----------------------------------------------------------------------
void TOptimizer::clean(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TString sworkindirectory = gSystem->UnixPathName(_code->getWorkingDirectory());
    // Place in the good directory
    if (sworkindirectory.Length() != 0)
        gSystem->cd(sworkindirectory);

    // Clean the subdirectory "UranieLauncher_*"
#ifdef WIN32
    // gSystem->Exec("del /s /q UranieLauncher_*");
	gSystem->Exec("dir /B UranieLauncher_* > tmpFile.txt ");
	ifstream directory("tmpFile.txt", ios::in);
	TString strCmd = "rd /s /q ";
    TString strTmp;
    while (!directory.eof())
    {
        directory >> strTmp;
		strCmd += " ";
		strCmd += strTmp;
    }
	directory.close();
	gSystem->Exec(strCmd.Data());
	gSystem->Exec("del /s /q tmpFile.txt");
#else
    gSystem->Exec("rm -fr UranieLauncher_*");
#endif
}
// -----------------------------------------------------------------------
//  Print the status of the attributes
// -----------------------------------------------------------------------
void TOptimizer::printState(Option_t *option)
{
    TString opt = option;
    cout << " ********** Print state [" << _tds->getTuple()->GetEntries()
            << "] option[" << opt << "]" << endl;
    Int_t nvar = _tds->getNAttributes();
    Double_t valdef = 1.23456789;
    for (Int_t ivar = 0; ivar < nvar; ivar++)
    {
        URANIE::DataServer::TAttribute *att = _tds->getAttribute(ivar); // +1 a cause de l'iterateur
        if (att != NULL)
        {
            // Default value
            valdef = 1.23456789;
            att->getDefaultValue(valdef);
            URANIE::DataServer::TAttribute::EOrigin norigin = att->getOrigin();
            switch (norigin)
            {
            case URANIE::DataServer::TAttribute::kConstant:
                cout << " **   name [ " << att->GetName()
                        << "] Origin[kConstant] Value[" << valdef << "]"
                        << endl;
                break;
            case URANIE::DataServer::TAttribute::kAttribute:
                if( (att->isOutput() && att->GetName() == _sCost ) || (!att->isOutput() && dynamic_cast<URANIE::DataServer::TAttributeFormula*>(att)==NULL) )
                    cout << " **   name [ " << att->GetName() << "] Origin[kAttribute] Value[" << valdef << "]" << endl;
                break;
            default:
                break;
            }
        }
        else
        {
            cout << " ** Pointer [" << ivar << "] is NULL" << endl;
        }

    }
    cout << " ********** End Of Print state [" << _tds->getTuple()->GetEntries()
            << "]" << endl;
}

// ----------------------------------------------------------------------
//  TOptimizer::printLog
// ----------------------------------------------------------------------
void TOptimizer::printLog(Option_t *option)
{
    cout << " ********** TOptimizer::printLog[" << option << "]" << endl;
    cout << " **   Name[" << GetName() << "] Title[" << GetTitle() << "]"
            << endl;
    cout << " **   _sInput[" << _sInput << "] _sOutput[" << _sOutput
            << "] _sCost[" << _sCost << "]" << endl;
    cout << " **   _sdirectory[" << _sdirectory << "] " << endl;
    cout << " **   _nMaxIterations[" << _nMaxIterations
            << "] _nMaxFunctionCalls[" << _nMaxFunctionCalls << "] " << endl;
    cout << " **   _dTolerance[" << _dTolerance << "] " << endl;

    switch (_optimMethod)
    {
    case kMigrad:
        cout << " **    optimMethod[kMigrad]" << endl;
        break;
    case kSimplex:
        cout << " **    optimMethod[kSimplex]" << endl;
        break;
    case kCombined:
        cout << " **    optimMethod[kCombined]" << endl;
        break;
    case kScan:
        cout << " **    optimMethod[kScan]" << endl;
        break;
    }

    switch (_optimProblem)
    {
    case kMinimizeCode:
        cout << " **   optimProblem[kMinimizeCode]" << endl;
        break;
    case kMinimizeFunction:
        cout << " **   optimProblem[kMinimizeFunction]" << endl;
        break;
    case kSumOfSquare:
        cout << " **   optimProblem[kSumOfSquare]" << endl;
        break;
    }

    if (_blog && _tds)
        _tds->printLog();

    if (_blog && _code)
        _code->printLog();

    // Les objectives
    Int_t nobj = (Int_t)(_objectives.size());
    for (Int_t indx = 0; indx < nobj; indx++)
    {
        cout << "Objective index[" << indx + 1 << "/" << nobj << "]" << endl;
        _objectives[indx]->printLog();
    }
    cout << " ********** End Of TOptimizer::printLog[" << option << "]" << endl;
}

} // Fin du namespace Optimizer
} // Fin du namespace URANIE

