//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizir.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2009-06-02.11.27.43
 
 \brief interface uranie pour vizir

 \par Modifications
 \li 2011-02-11 14:49:05 FGA
 Add the feature and method setVarDraw
 \li 2010-10-01 16:22:23 FGA
 Add the attribute _sdirectory to reput the user in the current directory
 \li 2010-09-16 15:09:28 FGA
 Add the parameter \em bAddTheAttributes in the \em setConstraintAndCost method to allow modify the function on the fly
 */

#ifndef ROOT_TVIZIR
#define ROOT_TVIZIR

// Writes log
#include <iostream>

////// ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TMethodCall.h"
#include "TMatrixD.h"

// Uranie
#include "Optimizer.h"
#include "TDataServer.h"
#include "UExceptions.h"
#include "TCode.h"

// Vizir
//! The max number of generation to draw
#define MAXNBGENERATIONS 5
#include "vizir.h"



namespace URANIE
{
namespace Optimizer
{
class TVizir: public TNamed
{
public:
    Double_t survie, zygotie, mutation, precision;
    Int_t intern, param, contrainte, cost, hide;
    Int_t population, duree;
    TMethodCall *fDecode, *fSoft, *fHard;
    TMatrixD minmax;

    Bool_t _blog;                              ///< Boolean to edit the log
    URANIE::DataServer::TDataServer *_tds;      //! Pointer vers un TDS
    URANIE::Launcher::TCode *_code;
    TString _sdirectory;                        ///< Current directory
    vzr_t *vizir;        //! zjdsjj


protected:
    void rearangeInput(TList *, TString &);
    void createTuple(TString &);
private:

public:
    TString _sdraw;                      ///< Variables to visualize
    TString _sopt;                       ///< Option of visualisation
    TString _scut;                       ///< Selection of visualisation
    Int_t _nNbGenerationToDraw; ///< The number of last generation to draw (Default=2)

    //       TVizir();
    TVizir(URANIE::DataServer::TDataServer *tds, int npop);
    virtual ~TVizir();
    // parametrage
    //      void init(URANIE::DataServer::TDataServer *tds, int npop);

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    //! Set the precision
    /*!
     */
    void setPrecision(double prec)
    {
        precision = prec;
    }
    //! Set the survival rate
    void setSurvivalRate(double val)
    {
        survie = val;
    }
    //! Set the proba of homozigotie
    void setHomozygoteRate(double val)
    {
        zygotie = val;
    }
    //! Set the probability of mutation
    void setMutationRate(double val)
    {
        mutation = val;
    }
    //! Set the maximum number of evaluation
    void setMaxEvaluation(int val)
    {
        duree = val;
    }
    //! Set evaluation
    void setDecode(const char *, Int_t phys = 0);
    void setCost(const char *, const char*, int hiden = 0, const char *input = "*");
    void setConstraintAndCost(const char *, const char*, int, int hiden = 0,
            const char *input = "*", Bool_t bAddTheAttributes = kTRUE);
    void setCost(URANIE::Launcher::TCode *, int hiden = 0);
    void setConstraintAndCost(URANIE::Launcher::TCode *, int, int hiden = 0);
    void setHardConstraint(const char *);
    //! Get the last generation index
    Int_t getLastGenerationIndex()
    {
        return vizir1_vzr_generationCount(vizir);
    }
    //! Get the number of generation to draw
    /*!
     */
    Int_t getNbGenerationToDraw()
    {
        return _nNbGenerationToDraw;
    }
    //! Set the number of generation to draw
    /*!
     */
    void setNbGenerationToDraw(Int_t n);
    //! Activates and defines the graphics during the optimization process
    /*!
     By default, there is no graphic during the optimization process
     To activate it, you must defines the informations of the graphics : the X, Y and Z attributes, the select and the options of the graphic.
     \param str (TString) the attributes to visualize ("Z:Y:X")
     \param select (TString) [""] the \em "select" of patterns to visualize
     \param opt (TString) [""] the options of the graphic
     */
    void setVarDraw(const char* varexp, const char* selection = "",
            Option_t* option = "");
    // TString str, TString select ="", TString opt="");
    //@}

    //---------------------------------------------
    /** @name Copy for Thread safe evaluation
     **/
    //@{
    TMethodCall *copyDecode()
    {
        return (fDecode) ? new TMethodCall(*fDecode) : NULL;
    }
    TMethodCall *copyCost()
    {
        return (fSoft) ? new TMethodCall(*fSoft) : NULL;
    }
    TMethodCall *copyHardConstraint()
    {
        return (fHard) ? new TMethodCall(*fHard) : NULL;
    }
    URANIE::Launcher::TCode *copyCode();
    //@}

    //---------------------------------------------
    /** @name Algorithm
     **/
    //@{
    void decode(Double_t*, Double_t*, TMethodCall *);
    int evaluer(Double_t*, Double_t*, TMethodCall*, TMethodCall*, TMethodCall*);
    int evaluer(Double_t*, Double_t*, TMethodCall*, TMethodCall*,
            URANIE::Launcher::TCode*);
    void fitnessChanged();
    void optimize(Option_t *option = "");
    void Results();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Optimizer::TVizir, ID_OPTIMIZER)
    //Classe de
};

} // Fin du namespace Optimizer
} // Fin du namespace URANIE

#endif
