//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   vzr_croiser.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2008-03-17.15.27.21

 \brief operateur stochastique standart de croisement
 */

#include <stdlib.h>

#include "vzr_struct.h"

#include "vzr_mt19937ar.h"



//#define MICHEL

#ifndef COMBINATOIRE

#define XPOL_DEC  0.2
#define XPOL_COEF 1.4 // 1 + 2*XPOL_DEC

struct cross
{
    int nballgene, nbdiscgene;
};

static void creer(void *truc, double **fils)
{
    int i;
    struct cross *self = (struct cross *) truc;

    for (i = 0; i < self->nballgene; i++)
    {
        fils[0][i] = vizir1_genrand_real1();
        fils[1][i] = vizir1_genrand_real1();
    }
}

static void allele(void *truc, double **parent, double *fils)
{
    int i;
    struct cross *self = (struct cross *) truc;

    for (i = 0; i < self->nballgene; i++)
    {
        fils[i] = parent[vizir1_genrand_int(2)][i];
    }
}

static void mutation(void * truc, double **fils)
{
    struct cross *self = (struct cross *) truc;

    fils[0][vizir1_genrand_int(self->nballgene)] = vizir1_genrand_real1();
    fils[1][vizir1_genrand_int(self->nballgene)] = vizir1_genrand_real1();
}

static void heterozygote(void *truc, double **fils)
{
    int i;
    double temp;
    struct cross *self = (struct cross *) truc;

    /* on construit le phenotype */
    for (i = 0; i < self->nballgene; i++)
        if (vizir1_genrand_real1() >= 0.5)
        {
            temp = fils[0][i];
            fils[0][i] = fils[1][i];
            fils[1][i] = temp;
        }

}

static void homozygote(void *truc, double **fils)
{
    int i;
    double temp;
    struct cross *self = (struct cross *) truc;

    for (i = 0; i < self->nballgene - self->nbdiscgene; i++)
    {
#ifdef MICHEL
        temp = vizir1_genrand_real1();
        temp = temp * fils[0][i]+ (1-temp)*fils[1][i];
#else
        temp = XPOL_COEF * vizir1_genrand_real1() - XPOL_DEC;
        temp = temp * fils[0][i] + (1 - temp) * fils[1][i];
        temp = (temp > 1.) ? 1. : temp;
        temp = (temp < 0.) ? 0. : temp;
#endif
        fils[0][i] = fils[1][i] = temp;
    }
    for (i = self->nballgene - self->nbdiscgene; i < self->nballgene; i++)
    {
        temp = (vizir1_genrand_real1() >= 0.5) ? 0 : 1;
        temp = temp * fils[0][i] + (1 - temp) * fils[1][i];
        fils[0][i] = fils[1][i] = temp;
    }
}

static void dealloc(struct vzrcrossover *cross)
{
    free(cross);
}

static struct vzrcrossover *vzrcrossstd_objet(void*info)
{
    struct vzrcrossover *ret;

    ret = (struct vzrcrossover *) malloc(sizeof(struct vzrcrossover));
    ret->data = info;
    ret->sample = &creer;
    ret->allele = &allele;
    ret->mutation = &mutation;
    ret->homozygote = &homozygote;
    ret->heterozygote = &heterozygote;
    ret->dealloc = &dealloc;

    return ret;
}

static void vzrcrossstd_clean(struct vzrcrossoverbuild *self)
{
    free(self->data);
    free(self);
}

struct vzrcrossoverbuild *vizir1_vzrcrossstd_classe(int nball, int nbdisc)
{
    struct cross *dat;
    struct vzrcrossoverbuild *ret;

    dat = (struct cross *) malloc(sizeof(struct cross));

    dat->nballgene = nball;
    dat->nbdiscgene = nbdisc;

    ret = (struct vzrcrossoverbuild *) malloc(sizeof(struct vzrcrossoverbuild));
    ret->data = (void *) dat;
    ret->create = &vzrcrossstd_objet;
    ret->dealloc = &vzrcrossstd_clean;

    return ret;
}

#endif

