//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   vzr_kernel.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2003-12-11.16.10.36

 \brief   ALGORITHME GENETIQUE DIPLOIDE A DOMINANCE VRAIE 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#include <process.h>
#endif

#include "vzr_kernel.h"
#include "vzr_util.h"

/**** Prototype ****/

#include "vzr_naitre.h"
#include "vzr_survivre.h"
#include "vzr_croiser.h"
#include "vzr_mt19937ar.h"

int completer_initial(struct vizir *self, struct vzrwork *wself)
/* retour 0 (faux) si la population est constituee                 */
/*                 sinon retour 1(vrai) si le descendant est rejete*/
/*              ou si la population n'est pas constituee           */

{
    int i;

    for (i = 0; i < self->nbgene; i++)
    {
        self->population[self->ipop][0][i] = wself->descendant[0][i];
        self->population[self->ipop][1][i] = wself->descendant[1][i];
    }

    for (i = 0; i < self->nbcrit; i++)
        self->pop_val[self->ipop][i] = wself->val[i];
    self->ipop++;
    if (self->ipop >= self->N)
    {
        self->ipop = 0;
        // self->rgmax=self->N-self->alpha*self->N;
        return (1);
    }
    return (0);
}

/*-------------------------------------------------------------------------*/
/* retour 0 (faux) si la population n'est pas reconstituee                 */
/*          sinon retour 1(vrai) si la population est  reconstituee     */
int stock(struct vizir *self, struct vzrwork *wself)
{
    int i, j;
    /** ajout du descendant a la population **/
    for (i = 0; i < self->nbgene; i++)
    {
        self->population[self->N - self->ipop - 1][0][i] =
                wself->descendant[0][i];
        self->population[self->N - self->ipop - 1][1][i] =
                wself->descendant[1][i];
    }
    /* descendant_val:  tableau temporaire contenant                     */
    /*               les valeurs des criteres des descendants stockes       **/
    for (i = 0; i < self->nbcrit; i++)
        self->descendant_val[self->ipop][i] = wself->val[i];
    self->ipop++; /*** nombre d'individu cree pendant la generation **********/
    /** test de fin de generation **/
    if (self->ipop < self->N - self->nbsurv)
        return (0);
    /* la population est completement reconstruite */
    //for (i=0;i<self->N-self->nbsurv;i++)
    for (i = 0; i < self->ipop; i++)
        for (j = 0; j < self->nbcrit; j++)
            self->pop_val[self->N - i - 1][j] = self->descendant_val[i][j];
    self->ipop = 0;
    return (1);
}

int completer(struct vizir *self, struct vzrwork *wself)
{
    double fitval;

    fitval = PRO_CALL(self->darwin, rank, self, wself);
    if (fitval <= self->fitseuil)
        if (stock(self, wself))
            return (1);

    return (0);
}

/***************************** central **************/
void vizir1_vzr_run(struct vizir *self)
{
    int i, j, good;
    short tag;
    struct vzrwork wself;
    vzrgene_t *tab;

    /* alloc */
    tab = (vzrgene_t *) calloc(6 * self->nbgene, sizeof(vzrgene_t));
    wself.pere[0] = &tab[0];
    wself.pere[1] = &tab[self->nbgene];
    wself.mere[0] = &tab[2 * self->nbgene];
    wself.mere[1] = &tab[3 * self->nbgene];
    wself.descendant[0] = &tab[4 * self->nbgene];
    wself.descendant[1] = &tab[5 * self->nbgene];
    wself.val = (double *) calloc(self->nbcrit, sizeof(double));
    wself.oper = PRO_CREATE(self->foper);

    pthread_mutex_lock(&(self->mutex));
    wself.fun = PRO_CREATE(self->ffun);

    self->bgstate |= RUN;
    self->bgstate &= ~START;
    self->slave += 1;
    while (1)
    {
        /* test arret non converg� */
        if (self->bgstate & STOP)
            break;
        if (self->nbevalu >= self->maxevalu - self->slave + 1)
            break;
        pthread_mutex_unlock(&(self->mutex));
        /* nouvel individu */
        if (vizir1_reevaluation(self, &wself))
        {
        }
        else if (vizir1_immigration(self, &wself))
        {
        }
        else if (self->nbsurv)
        {
            /* croisement */
            accoupler(self, &wself);
            vizir1_procreer(self, &wself);
        }
        else
        {
            /* generation aleatoire */
            PRO_CALL(wself.oper, sample, wself.descendant);
        }
        good = PRO_CALL(wself.fun, compute, wself.descendant[0], wself.val);
        pthread_mutex_lock(&(self->mutex));
        self->nbevalu++;
        /* test arret converge */
        if (self->bgstate & CONV)
            break;
        /* integration de l'individu */
        if (good)
        {
            self->nadmis++;
            //if (! (self->nadmis%100)) {printf("+"); fflush(stdout);}
            good = (self->nbsurv) ?
                    completer(self, &wself) : completer_initial(self, &wself);
        }
        /* test de fin de generation */
        if (good)
        {
            /* nouvelle generation */
            self->niter++;
            printf("\n%s  Generation : %d, Nombre d'enfants a creer : %d\n",
                    self->id, self->niter, self->N - self->nbsurv);
            printf("%s\tNb d'evaluation : %d, taille de la Z.P. :  %d\n",
                    self->id, self->nbevalu, self->nbelite);
            if (vizir1_parents(self))
                break;
            if (self->generation)
                PRO_CALL(self->generation, react, self);
        }

    }
    self->slave--;
    if (!self->slave)
    {
        tag = 0;
        if (self->bgstate & STOP)
        {
            printf(
                    "%s  ! ! ! ! !   La convergence n'a pas ete atteinte  dans le temps imparti  ! ! ! ! \n",
                    self->id);
            tag = 1;
        }
        if (self->nbevalu >= self->maxevalu)
        {
            printf(
                    "%s  ! ! ! ! !  La convergence n'a pas ete atteinte avec le nb max d'evaluations ! ! ! ! \n",
                    self->id);
            tag = 1;
        }
        if (tag)
        {
            printf(
                    "%s  Sauvegarde de la generation en cours : nb de parents = %d, nb de descendants = %d\n",
                    self->id, self->nbsurv, self->ipop);
            for (i = 0; i < self->ipop; i++)
                for (j = 0; j < self->nbcrit; j++)
                    self->pop_val[self->N - i - 1][j] =
                            self->descendant_val[i][j];
            self->ipop = 0;
            vizir1_parents(self);
        }
        self->bgstate = END;
    }

    pthread_mutex_unlock(&(self->mutex));
    printf("%s  Nombre de generations : %d\n", self->id, self->niter);
    PRO_DEALLOC(wself.oper);
    PRO_DEALLOC(wself.fun);
    free(tab);
    free(wself.val);

}

