//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   vzr_monocontraint.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2008-03-28.12.07.37

 \brief classement specifique au monocritere avec contraintes
 */

#include <stdio.h>
#include <stdlib.h>

#include "vzr_util.h"

#define SEUILspeed 0.7

/** utilitaires communs **/

/* calcul du rang dans la population courante
 * */
int rang(struct vizir *viz, struct vzrwork *wviz, int istart, int ncri)
{
    int ret = 0;
    int j, k, id, tagi, tagj;

    for (j = 0; j < viz->N; j++)
    {
        tagi = tagj = 0;
        for (k = 0; k < ncri; k++)
        {
            id = k + istart;
            if (wviz->val[id] > viz->pop_val[j][id])
            {
                tagi = 1;
                if (tagj)
                    break;
            }
            else if (wviz->val[id] < viz->pop_val[j][id])
            {
                tagj = 1;
                if (tagi)
                    break;
            }
            // else if (valcour[k]==valcomp[k][j]) h+=1;
        }
        if (tagi && !tagj)
            ret++;
    }
    return (ret);
}

/* mise a jour des rangs de la nouvelle population
 * */
void classe(struct vizir *viz, int istart, int size, int irang)
{
    int i, j, k, id, tagi, tagj;

    for (i = 0; i < viz->N; i++)
        viz->pop_val[i][irang] = 0;

    for (i = 1; i < viz->N; i++)
    {
        for (j = 0; j < i; j++)
        {
            tagi = tagj = 0;
            for (k = 0; k < size; k++)
            {
                id = k + istart;
                if (viz->pop_val[i][id] > viz->pop_val[j][id])
                {
                    tagi = 1;
                    if (tagj)
                        break;
                }
                else if (viz->pop_val[i][id] < viz->pop_val[j][id])
                {
                    tagj = 1;
                    if (tagi)
                        break;
                }
                // else egal
            }
            if (tagi && !tagj)
                viz->pop_val[i][irang]++;
            if (tagj && !tagi)
                viz->pop_val[j][irang]++;
        }
    }

}

/** fonction du protocole **/
static double rang_individu(void *truc, struct vizir *viz, struct vzrwork *wviz)
{
    double ret, *seuil_de_f;

    seuil_de_f = (double*) truc;

    wviz->val[viz->rangdomi - 1] = (double) rang(viz, wviz, 0, viz->nbsoft);
    /*viz->val[viz->nbfun+viz->nbsoft+1] = 
     (double)rang(viz, wviz, viz->nbsoft, viz->nbfun);
     ret = viz->val[viz->nbfun+viz->nbsoft+2] = 
     (double)rang(viz, wviz, viz->nbsoft+self->nbfun, 2); */
    ret = wviz->val[viz->rangdomi] = (double) rang(viz, wviz, viz->nbsoft, 2);
    /* Tous les individus verifiant les contraintes
     * et dont la fonction objectif est inf a une valeur
     * calculee a la generation precedente
     * sont conserves
     * */
    if ((int) wviz->val[viz->rangdomi - 1] == 0
            && wviz->val[viz->nbsoft] < *seuil_de_f)
        ret = wviz->val[viz->rangdomi] = -2.0;
    return ret;

}

static void rang_population(void *truc, struct vizir *viz)
{
    int i, nbad;
    double fmax, fmin, delta_f, rgmax, *seuil_de_f;

    seuil_de_f = (double *) truc;

    /* calcul des différents rangs */
    classe(viz, 0, viz->nbsoft, viz->rangdomi - 1);
    // classe(viz, viz->nbsoft, 1, viz->nbsoft+viz->nbfun+1); // usefull ?
    classe(viz, viz->nbsoft, 2, viz->rangdomi);

    /** biaisage **/
    /* calcul des seuils */
    nbad = 0;
    fmax = -1.e+60;
    fmin = 1.e+60;
    for (i = 0; i < viz->N; i++)
        if ((int) viz->pop_val[i][viz->rangdomi - 1] == 0)
        {
            nbad++;
            fmax = max(fmax, viz->pop_val[i][viz->nbsoft]);
            fmin = min(fmin, viz->pop_val[i][viz->nbsoft]);
        }
    if (nbad >= viz->consang)
    {
        delta_f = SEUILspeed * (fmax - fmin);
        *seuil_de_f = (delta_f > viz->prec) ? fmin + delta_f : fmin + viz->prec;
    }
    else
        *seuil_de_f = 1.e+60;

    printf("%s\tfmin : %e, fmax :  %e\n ", viz->id, fmin, fmax);

    /* rang max pour le probleme ne contenant que les contraintes
     * qu'un individu conserve ne pourra pas depasse   */
    rgmax = viz->N - viz->alpha * (viz->N - viz->nbelite);

    /* marquage des bons mauvais */
    for (i = 0; i < viz->N; i++)
    {
        /* les bons */
        if ((int) viz->pop_val[i][viz->rangdomi - 1] == 0
                && viz->pop_val[i][viz->nbsoft] < *seuil_de_f)
        {
            /* On met a -1 le rang des individus conserves
             * (ils seront les premiers par ordre decroissant */
            if (viz->pop_val[i][viz->rangdomi] < 0.5)
                viz->pop_val[i][viz->rangdomi] = -2;
            else
                viz->pop_val[i][viz->rangdomi] = -1;
        }
        /* les mauvais */
        if (viz->pop_val[i][viz->rangdomi - 1] > rgmax)
        {
            /* On met a N+1 lr rang des individus a supprimer
             * (il seront les derniers par ordre decoissant */
            viz->pop_val[i][viz->rangdomi] = viz->N + 1;
        }
    }
    //viz->sous_population[0]=cardad;
    /*    printf(" Nb de points admissibles %d pareto %d rejette %d seuil %f\n",cardad,ipareto,nbrejet,viz->fitseuil);  */
}

static int elite(void *truc, struct vizir *viz)
{
    int i, ret;

    ret = 0;
    for (i = 0; i < viz->N; i++)
    {
        if ((int) viz->pop_val[i][viz->rangdomi] < 0.5)
            ret++;
        else
            break;
    }
    return ret;
}


static void vzrfunstd_clean(struct vzrclasse *self)
{
    free(self->data);
    free(self);
}


/** definition de l'instance de protocole **/

struct vzrclasse *mono_contraint(void)
{
    struct vzrclasse *ret;

    ret = (struct vzrclasse *) malloc(sizeof(struct vzrclasse));
    ret->data = malloc(sizeof(double));
    ret->rank_field_size = 2;
    ret->rank = &rang_individu;
    ret->rerank = &rang_population;
    ret->elite_size = &elite;
    ret->dealloc = &vzrfunstd_clean;
      
    return ret;
}

