//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   rosenbrock.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jui 3 10:38:56 CEST 2007
 
 \brief Utilisation de TMinuit2 sur la fonction de Rosenbrock

 $Revision$
 $State$
 $Date$

 \par Revision
 \li 2017/01/26 (FGA )
 This code an option -o to compute the rosenbrok function into ten patterns included in the "objectiveData.h"
and output into the "_output_rosenbrock_withDataServer_.dat" TDS file

 \li 2015-09-15 (TCX)
 Rosenbrock has now a third parameter fC which default value is 0.0.

 \li 2010-09-26 (CFO)
 Rosenbrock code has now "-d" option for debug mode and "-r" option to use an inputfile without using CLUB. 
 
 The \em Rosenbrock function is defined by the formula
 \f[
 f(x, y) \, = \, 100( y - x^{2})^{2} + ( 1 - x )^{2} + 0(1-y)^{2}
 \f]

 The minimum is reached for \f$ ( \star{x^{0}}, \star{y^{0}} ) \, = \, ( 1.0, 1.0 ) \f$

 A possible starting point is \f$ ( x^{0}, y^{0} ) \, = \, ( -1.2, 1.0 )\f$ and its value is equal to 24.20.

 The parameters of this function are defined in a ASCII file with key = value and Rows formalism
 \li the inputs of the function F : x and y. The associated keys are respectively "x" and "y";
 \li the three parameters of the function F : fA, fB and fC and the associated keys are respectively "a", "b" qnd "c";
 the default values of the Three parameters are $fA = 10.$, $fB = 1.0$ and $fC = 0.0$
 */

#define MAX_LENGTH 256
//! Input file with KEY
#define INPUT_FILE_WITH_KEY "input_rosenbrock_with_keys.dat"
//! Input file with VALUES in ROW
#define INPUT_FILE_WITH_VALUES_ROW "input_rosenbrock_with_values_rows.dat"
//! Output file with KEY
#define OUTPUT_FILE_WITH_KEY "_output_rosenbrock_with_keys_.dat"
//! Output file with VALUES in ROW
#define OUTPUT_FILE_WITH_VALUES_ROW "_output_rosenbrock_with_values_rows_.dat"
//! Output file with VALUES in COL
#define OUTPUT_FILE_WITH_VALUES_COL "_output_rosenbrock_with_values_cols_.dat"
//! Output file with DataServer format
#define OUTPUT_FILE_WITH_DATASERVER "_output_rosenbrock_withDataServer_.dat"

#include "objectiveData.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <fstream>
using namespace std;

// Club
#ifdef OLDCLUB
#include "club/TamponTexte.h"
#include "club/IterateurChamp.h"
#include "club/BaseErreurs.h"
#else
#include "TamponTexte.h"
#endif

//! Set the debig flag
int bdebug = 0;
//! Set the verbose boolean
bool bVerbose = true;



/*!
 Reads file with values in rows.
 */
int readInputFileWithValuesRows(const char *sfile, double *input)
{
    if (bdebug)
        cout << "***** readInputFileWithValuesRows file[" << sfile << "] ... "
                << endl;

    int nfound = -1;

    int nparamToRead = 5;

    std::ifstream fileInput(sfile, ios::in);
    if (fileInput)
    {
        if (bdebug)
            cout << " ** file open ifstream " << endl;
    }
    else
    {
        if (bdebug)
            cout << "********************* " << endl;
        return nfound;
    }
    nfound++;
    for (int i = 0; i < nparamToRead; i++)
    {
        fileInput >> input[i];
        if (!fileInput.good())
            break;
        if (bdebug)
            cout << " ** i[" << i << "/" << nparamToRead << "] value["
                    << input[i] << "]" << endl;
        nfound++;
    }

    fileInput.close();
    return nfound;
}

/*!
 Reads file with keys.
 */
int readInputFileWithKeys(const char *sfile, double *input)
{
    if (bdebug)
        cout << "***** readInputFileWithKeys file[" << sfile << "] ... "
                << endl;

    int nparamToRead = 5;
    int nfound = 0;

    FILE *fin = fopen(sfile, "r");
    if (fin)
    {
        if (bdebug)
            cout << " ** file open" << endl;
    }
    else
    {
        if (bdebug)
            cout << "********************* " << endl;
        return -1;
    }

    TamponTexte fnomFichier(sfile, 0);
    fnomFichier.initSeparateurs(" =\n");
    fnomFichier.initCarCommentaires("#");

    // Elimination des lignes vides
    fnomFichier.elimineBlancs();
    // Elimination des lignes de commentaires
    fnomFichier.elimineCommentaires();
    // Test si le fichier ainsi obtenu est vide
    if (fnomFichier.estVide())
    {
        cerr << "\n le fichier: " << fnomFichier.nom() << " est vide.\n";
        exit(1);
    }
    if (bdebug)
    {
        cout << " ** Ouverture du fichier " << fnomFichier.nom() << " terminee."
                << endl;
        cout << " ** fichier " << fnomFichier.nom() << " : "
                << fnomFichier.nombreLignes() << " lignes initiales" << endl;
    }

    for (int iline = 1; iline <= fnomFichier.nombreLignes(); iline++)
    {
        if (bdebug)
            cout << "i[" << iline << "]" << fnomFichier.nombreChampsLigne(iline)
                    << endl;
        if (fnomFichier.nombreChampsLigne(iline) > 1)
        {
            const char *strclef = fnomFichier.champLigne(1, iline);
            if (bdebug)
                cout << "i[" << iline << "] clef[" << strclef << "]" << endl;

            if (strcmp("x", strclef) == 0)
            {
                input[0] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("y", strclef) == 0)
            {
                input[1] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("a", strclef) == 0)
            {
                input[2] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("b", strclef) == 0)
            {
                input[3] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("c", strclef) == 0)
            {
                input[4] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
        }
    }

    // Close the input file
    fclose(fin);

    if (bdebug)
        cout << "**** End of readInputFileWithKeys file[" << sfile
                << "] - Found [" << nfound << "]" << endl;

    return nfound;
}

//! 
/*! 
 
 \param argc le nombre d'argument
 \param argv la liste des arguments

 \return indice de l'erreur de retour
 */
int main(int argc, char *argv[])
{

    FILE * fout;

    // Type of input file
    bool bFileByKey = false;
    bool bFileByRecreate = false;
    bool bObjectiveMode = false;
    char sInputFile[MAX_LENGTH];
    // Case of input file with keys by default
    sprintf(sInputFile, "%s", INPUT_FILE_WITH_KEY);

    for (int i = 0; i < argc; i++)
    {
        // Mode debug
        if (!strcmp(argv[i], "-d"))
        {
            bdebug = 1;
        }
        // Mode verbose
        if (!strcmp(argv[i], "-v"))
        {
            bVerbose = false;
        }
        // Case of input file with values in row
        if (!strcmp(argv[i], "-r"))
        {
            bFileByRecreate = true;
            if (i != argc - 1)
                sprintf(sInputFile, "%s", argv[i + 1]);
            else
                sprintf(sInputFile, "%s", INPUT_FILE_WITH_VALUES_ROW);
        }
        // Case of input file with keys
        if (!strcmp(argv[i], "-k"))
        {
            bFileByKey = true;
            if (i != argc - 1)
                sprintf(sInputFile, "%s", argv[i + 1]);
        }
        
        // Case of objective mode
        if (!strcmp(argv[i], "-o"))
        {
            bObjectiveMode = true;
            bFileByKey = true;
            if (i != argc - 1)
                sprintf(sInputFile, "%s", argv[i + 1]);
        }
    }

    if (bVerbose)
    {
        cout << "******************" << endl;
        cout << "*** Input  File[" << sInputFile << "]" << endl;
        cout << "******************" << endl;
    }

    if ((bFileByRecreate == false) && (bFileByKey == false))
    {
        cerr << " ** Rosenbrock code " << endl;
        cerr << " ** ERROR : You must select the rows (-r)  OR keys (-k) option" << endl;
        cerr << "  **  Usage : -d : debug mode" << endl;
        cerr << "  **  Usage : -v : verbose mode" << endl;
        cerr << "  **  Usage : -r [input_file = \"" << INPUT_FILE_WITH_VALUES_ROW << "\"] : input file with values in row" << endl;
        cerr << "  **  Usage : -k [input_file = \"" << INPUT_FILE_WITH_KEY << "\"] : input file with values in keys" << endl;
        cerr << "  **  Usage : -o [input_file = \"" << INPUT_FILE_WITH_KEY << "\"] : input file with values in keys and generate the objective mode in the output file \"" << OUTPUT_FILE_WITH_DATASERVER << "\"" << endl;
        cerr << " ** The Output files generated by the Rosenbrock code" << endl;
        cerr << "  **  with keys        :  \"" << OUTPUT_FILE_WITH_KEY << "\"" << endl;
        cerr << "  **  with rows        :  \"" << OUTPUT_FILE_WITH_VALUES_ROW << "\"" << endl;
        cerr << "  **  with objectives  :  \"" << OUTPUT_FILE_WITH_DATASERVER << "\"" << endl;
        exit(-1);
    }
    if (bdebug)
        cout << "with debug " << endl;
    // Allocate parameters vectors
    int nparam = 5;
    int nfound = 0;
    double *dparam = new double[nparam];
    // Default value
    dparam[2] = 10.0;
    dparam[3] = 1.0;
    dparam[4] = 0.0;
    
    ////////////////////////////
    //
    //Reads the input parameters
    //
    ////////////////////////////
    // Cas of values in row
    if (bFileByRecreate)
        nfound = readInputFileWithValuesRows(sInputFile, dparam);

    if (bFileByKey)
        nfound = readInputFileWithKeys(sInputFile, dparam);

    if ((nfound < 2) || (nfound > 5))
    {
        cerr << " ** ERROR : Pb in read file [" << sInputFile << "] found["
                << nfound << "] param[" << nparam << "]" << endl;
        exit(-1);
    }
    if (bVerbose)
    {
        cout << "******************" << endl;
        cout << "*** Input parameters File[" << sInputFile << "] Param["
                << nparam << "]" << endl;
        cout << "******************" << endl;
        for (int i = 0; i < nparam; i++)
            cout << "** i[" << i << "] val[" << dparam[i] << "]" << endl;
        cout << "******************" << endl;
    }

    // Evaluate the function
    double x = dparam[0];
    double y = dparam[1];
    double fA = dparam[2];
    double fB = dparam[3];
    double fC = dparam[4];
    double fval = fA * (y - x * x) * (y - x * x) + fB * (1.0 - x) * (1.0 - x) + fC * (1.0 - y) * (1.0 - y);

    // Save the evaluation in output file
    // Output with keys
    fout = fopen(OUTPUT_FILE_WITH_KEY, "w");
    fprintf(fout, "\nX = %e ;\nY = %e ;\nfval = %e ;\nfA = %e ; \nfB = %e ; \nfC = %e ;", x,
            y, fval, fA, fB, fC);
    fclose(fout);
    // Output with values in row
    // Sortie in line (TOutputFileRow)
    fout = fopen(OUTPUT_FILE_WITH_VALUES_ROW, "w");
    fprintf(fout, "#COLUMN_NAMES: x | y | fval | pA | pB | pC\n\n%e\t%e\t%e\t%e\t%e\t%e",
            x, y, fval, fA, fB, fC);
    fclose(fout);
    // Output with values in col
    // Sortie in column (TOutputFileColumn)
    fout = fopen(OUTPUT_FILE_WITH_VALUES_COL, "w");
    fprintf(fout, "%e\n%e\n%e\n%e\n%e\n%e\n", x, y, fval, fA, fB, fC);
    fclose(fout);
    
    // Output with DataServer
    // Sortie in row
    fout = fopen(OUTPUT_FILE_WITH_DATASERVER, "w");
    fprintf(fout, "#COLUMN_NAMES: x | y | a | b | c | fval | fval2 \n\n");
    if(bObjectiveMode)
    {
        
        for(int iter=0; iter<Nvar; iter++)
        {
            // Evaluate the function
            x = tableauVariableXY[iter][0];
            y = tableauVariableXY[iter][1];
            fval = fA * (y - x * x) * (y - x * x) + fB * (1.0 - x) * (1.0 - x) + fC * (1.0 - y) * (1.0 - y);
            fprintf(fout, "%e\t%e\t%e\t%e\t%e\t%e\t%e\n", x, y, fA, fB, fC, fval, 1.0+fval);
        }
    }
    else
    {
        fprintf(fout, "%e\t%e\t%e\t%e\t%e\t%e\t", x, y, fA, fB, fC, fval);
    }
    fclose(fout);
    

    
    if (bVerbose)
    {
        cout << "******************" << endl;
        cout << "***  Ouput Files [" << OUTPUT_FILE_WITH_KEY << "]" << endl;
        cout << "             and [" << OUTPUT_FILE_WITH_VALUES_ROW << "]"
                << endl;
        cout << "             and [" << OUTPUT_FILE_WITH_VALUES_COL << "]"
                << endl;
        cout << "******************" << endl;
        cout << "**         fval  [" << fval << "]" << endl;
        cout << "******************" << endl;
        cout << "******************" << endl;
    }

    delete[] dparam;
    dparam = NULL;

    return 1;

}
