#include "NLF.h"
#include "NLP.h"
#include "CompoundConstraint.h"
#include "NonLinearInequality.h"
#include "BoundConstraint.h"

using namespace NEWMAT;
using namespace OPTPP;

extern "C" {


void init_hs65(int ndim, ColumnVector& x)
{
  if (ndim != 3)
  {
    exit (1);
  }
  double factor = 0.0;
  x(1) = -5.0  - (factor - 1)*8.6505  ;
  x(2) =  5.0  + (factor - 1)*1.3495  ;
  x(3) =  0.0  - (factor - 1)*4.6204  ;
}

void hs65(int mode, int n, const ColumnVector& x, double& fx, ColumnVector& g, int& result)
{ // Hock and Schittkowski's Problem 65 (the objective fcn)
  double f1, f2, f3, x1, x2, x3;

  if (n != 3) return;

  x1 = x(1);
  x2 = x(2);
  x3 = x(3);
  f1 = x1 - x2;
  f2 = x1 + x2 - 10.0;
  f3 = x3 - 5.0;

  if (mode & NLPFunction) {
    fx  = f1*f1+ (f2*f2)*(1.0/9.0) +f3*f3;
    result = NLPFunction;
  }
  if (mode & NLPGradient) {
    g(1) =  2*f1 + (2.0/9.0)*f2;
    g(2) = -2*f1 + (2.0/9.0)*f2;
    g(3) =  2*f3;
    result = NLPGradient;
  }
}

void hs65_2(int mode, int n, const ColumnVector& x, double& fx, ColumnVector& g, SymmetricMatrix& H, int& result)
{ // Hock and Schittkowski's Problem 65 (the objective fcn)
  double f1, f2, f3, x1, x2, x3;

  if (n != 3) return;

  x1 = x(1);
  x2 = x(2);
  x3 = x(3);
  f1 = x1 - x2;
  f2 = x1 + x2 - 10.0;
  f3 = x3 - 5.0;

  if (mode & NLPFunction) {
    fx  = f1*f1+ (f2*f2)/9.0 +f3*f3;
    result = NLPFunction;
  }
  if (mode & NLPGradient) {
    g(1) =  2*f1 + (2.0/9.0)*f2;
    g(2) = -2*f1 + (2.0/9.0)*f2;
    g(3) =  2*f3;
    result = NLPGradient;
  }
  if (mode & NLPHessian) {
    H(1,1) =  2 + (2.0/9.0);

    H(2,1) = -2 + (2.0/9.0);
    H(2,2) =  2 + (2.0/9.0);

    H(3,1) = 0.0;
    H(3,2) = 0.0;
    H(3,3) = 2.0;
    result = NLPHessian;
  }
}

void ineq_hs65(int mode, int n, const ColumnVector& x, ColumnVector& fx, Matrix& g, int& result)
{ // Hock and Schittkowski's Problem 65 
  double f1, f2, f3, x1, x2, x3;

  if (n != 3) return;

  x1 = x(1);
  x2 = x(2);
  x3 = x(3);
  f1 = x1;
  f2 = x2;
  f3 = x3;

  if (mode & NLPFunction) {
    fx  = 48 - f1*f1 - f2*f2 - f3*f3;
    result = NLPFunction;
  }
  if (mode & NLPGradient) {
    g(1,1) = -2*x1;
    g(2,1) = -2*x2;
    g(3,1) = -2*x3;
    result = NLPGradient;
  }
}

void ineq_hs65_2(int mode, int n, const ColumnVector& x, ColumnVector& fx, Matrix& g, OptppArray<SymmetricMatrix>& H, int& result)
{ // Hock and Schittkowski's Problem 65 
  double f1, f2, f3, x1, x2, x3;
  SymmetricMatrix Htmp(n);

  if (n != 3) return;

  x1 = x(1);
  x2 = x(2);
  x3 = x(3);
  f1 = x1;
  f2 = x2;
  f3 = x3;

  if (mode & NLPFunction) {
    fx(1)  = 48 - f1*f1 - f2*f2 - f3*f3;
    result = NLPFunction;
  }
  if (mode & NLPGradient) {
    g(1,1) = -2*x1;
    g(2,1) = -2*x2;
    g(3,1) = -2*x3;
    result = NLPGradient;
  }
  if (mode & NLPHessian) {
    Htmp(1,1) = -2;
    Htmp(1,2) = 0.0;
    Htmp(1,3) = 0.0;
    Htmp(2,1) = 0.0;
    Htmp(2,2) = -2;
    Htmp(2,3) = 0.0;
    Htmp(3,1) = 0.0;
    Htmp(3,2) = 0.0;
    Htmp(3,3) = -2;

    H[0] = Htmp;
    result = NLPHessian;
  }
}


  CompoundConstraint* create_constraint_hs65_2(int n)
{ // Hock and Schittkowski's Problem 65 
  // (the constraints - nonlinear inequality)

  // Construct the nonlinear equation
  NLP* chs65        = new NLP( new NLF2(n,1,ineq_hs65_2,init_hs65));
  Constraint nleqn = new NonLinearInequality(chs65);
  ColumnVector lower(n); 
  lower << -4.5 << -4.5 << -5.0;
  ColumnVector upper(n); 
  upper <<  4.5 <<  4.5 <<  5.0 ;
  Constraint c1    = new BoundConstraint(n,lower,upper); 
  CompoundConstraint* constraints = new CompoundConstraint(nleqn,c1);
  return constraints;
}

  
} // fin du extern "C"

// fin du fichier de fonction
