//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTFormSorm.cxx
 \author Gilles Arnaud <gilles.arnaud@cea.fr>
 \date   mar mai 21 13:36:02 GMT 2013  

 \brief  test file for FORM SORM classes

 */

#include "TestTFormSorm.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_LARGE 1.e-2

// ROOT
//#include "Api.h"

// Uranie
#include "UExceptions.h"
#include "TDataServer.h"
#include "TAttribute.h"
#include "TNormalDistribution.h"

#include "TCJitEval.h"
//#include "TCIntEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TGradientEstimationRun.h"

#include "TNlopt.h"
#include "TNloptCobyla.h"
#include "TNloptDirect.h"

#include "TSimpleTransform.h"
#include "TFormEval.h"
#include "TSorm.h"

// local
#include "TestUserFunctions.h"

using namespace URANIE::DataServer;
using namespace URANIE::Relauncher;
using namespace URANIE::Reoptimizer;
using namespace URANIE::Reliability;

/* possible direct solver
 * direct
 #define DIRECT TNloptCobyla
 * lagragien direct
 #define DIRECT TNloptBobyqa //doesn't work on this problem
 #define DIRECT TNloptPraxis
 #define DIRECT TNloptNelderMead
 #define DIRECT TNloptSubplexe
 * */
#define DIRECT TNloptCobyla

/* possible gradient solver
 * direct
 #define GRADIENT TNloptMMA
 #define GRADIENT TNloptSLSQP
 * lagragien direct
 #define GRADIENT TNloptLBFGS
 #define GRADIENT TNloptNewton
 #define GRADIENT TNloptVariableMetric
 *
 * */
#define GRADIENT TNloptMMA

// Constructor Destructor
TFormSormTest::TFormSormTest()
{
    // creation of pointers on functions
#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;

    method = globalscope.GetMethod("fselect", "Double_t*,Double_t*", &dummy);
    _p2fFselect = method.PointerToFunc();
#endif
}

TFormSormTest::~TFormSormTest()
{
}

// pre and post treatment
void TFormSormTest::setUp()
{
}

void TFormSormTest::tearDown()
{
}

// test suite
void TFormSormTest::testDirect()
{
    /* inputs */
    TNormalDistribution x("x", 40, 5), fe("x2", 50, 2.5), M("M", 1000, 200);
    /* outputs */
    TAttribute cont("seuil");
    /* starting point */
    double start[] =
    { -1., -1., 1. };

    /* code */
    TSimpleTransform fobj;
    fobj.setParameters(3, &x, &fe, &M);

#ifdef ROOT5
    TCIntEval fcont(_p2fFselect);
#else
    TCJitEval fcont(fselect);
#endif
    fcont.setInputs(3, &x, &fe, &M);
    fcont.setOutputs(1, &cont);

    TFormEval code(&fobj, &fcont);
    TGreaterFit it(0.0);
    code.addConstraint(&cont, &it);

    /* tds */
    TDataServer tdsf("form", "tds for form test");
    code.addAllInputs(&tdsf);
    //TDataServer tdss("sorm", "tds for sorm test");
    //code.addAllInputs(&tdss);

    /*** FORM ***/
    /* optimizer */
    DIRECT solv;
    TNlopt nlo(&tdsf, &code, &solv);
    code.addObjective(&nlo);

    /* resolution */
    nlo.setStartingPoint(3,start);
    nlo.solverLoop();

    /*** SORM ***/
    TSorm sorm(&tdsf, &code);

    sorm.solverLoop();

    /* resultats */
    validate(tdsf);

}

void TFormSormTest::testGradient()
{
    /* inputs */
    TNormalDistribution x("x", 40, 5), fe("x2", 50, 2.5), M("M", 1000, 200);
    /* outputs */
    TAttribute cont("seuil");
    /* starting point */
    double start[] =
    { -1., -1., 1. };

    /* code */
    TSimpleTransform fobj;
    fobj.setParameters(3, &x, &fe, &M);

    TCJitEval fcont(&fselect);
    fcont.setInputs(3, &x, &fe, &M);
    fcont.setOutputs(1, &cont);

    TFormEval code(&fobj, &fcont);
    TGreaterFit it(0.0);
    code.addConstraint(&cont, &it);

    //TSequentialRun trun(&code);
    TThreadedRun run(&code, 4);
    //TGradientEstimationRun run(&trun);
    run.startSlave();
    if (run.onMaster())
    {
        /* tds */
        TDataServer tdsf("form", "tds for form test");
        code.addAllInputs(&tdsf);
        //TDataServer tdss("sorm", "tds for sorm test");
        //code.addAllInputs(&tdss);

        /*** FORM ***/
        /* optimizer */
        GRADIENT solv;
        TNlopt nlo(&tdsf, &run, &solv);
        code.addObjective(&nlo);

        /* resolution */
        nlo.setStartingPoint(3,start);
        nlo.solverLoop();

        /*** SORM ***/
        TSorm sorm(&tdsf, &run);

        sorm.solverLoop();

        /* resultats */
        run.stopSlave();
        validate(tdsf);
    }

}

void TFormSormTest::validate(TDataServer &tds)
{
    // only one result
    Int_t siz = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(1, siz);
    // attributes
    // u_x, u_x2, u_M, betaHL, form, 
    // x, x2, M, seuil, factor, sorm
    CPPUNIT_ASSERT_EQUAL(11, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("u_M"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("seuil"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("form"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("sorm"));
    // values (cobyla direct)
    // -2.289658, -0.677000, 1.8963080, 3.0490734, 0.0011477, 
    // 28.551708, 48.307498, 1379.2616, -1.67e-05, 1.0203699, 0.0011711
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.289658, tds.getValue("u_x", 0),
            DBL_PRECISION_LARGE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.677, tds.getValue("u_x2", 0),
            DBL_PRECISION_LARGE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.8963080, tds.getValue("u_M", 0),
            DBL_PRECISION_LARGE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0011477, tds.getValue("form", 0),
            DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0011711, tds.getValue("sorm", 0),
            DBL_PRECISION_FINE);
    // relation beetween values
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds.getValue("u_x2", 0) * 2.5 + 50,
            tds.getValue("x2", 0), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(
            tds.getValue("form", 0) * tds.getValue("factor", 0),
            tds.getValue("sorm", 0), DBL_PRECISION_FINE);

}
