//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNlopt.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-03-20.16.04.36

 \brief   interface algorithme optimisation de la bibliotheque NLOpt
 */

#include "TNloptShare.h"

#include <stdlib.h>

#include "nlopt.h"
#include "loop.h"

#include "TMasterVar.h"
#include "TFitGetter.h"

#include "uranie_detection.h"
#ifdef Boost_COROUTINE_FOUND
#include <boost/coroutine2/all.hpp>

#define COROUTINE

using namespace boost::coroutines2;
#endif

#define SUIVI
#define IsRUNNING 16384

#define EPS 1.0e-5
#define ZERO 1.e-20

ClassImp (URANIE::Reoptimizer::TNloptSolver);
ClassImp (URANIE::Reoptimizer::TNloptShare);

#include "nlopt_work.h"

namespace URANIE
{
namespace Reoptimizer
{

//TNloptShare
TNloptShare::TNloptShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *Run, TNloptSolver *solv) :
    TOptimShare(tds, Run),  _solv(solv), _retOptim(-99)
{
    create();
}

TNloptShare::TNloptShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TNloptSolver *solv) :
    TOptimShare(tds, fun),  _solv(solv), _retOptim(-99)
{
    create();
}

TNloptShare::~TNloptShare()
{
    TFitGetter *one;
    TIter next(&_equal);

    while ((one = (TFitGetter *) next()))
        delete one;
    delete _crit;
    _crit = NULL;
    for (size_t i = 0; i < _guess.size(); i++)
    {
        delete[] _guess[i];
        //_guess.erase(_guess.begin()+i);
    }
}

/* constructeur helpers */
void TNloptShare::create()
{
    //_coef = 0;
    /* defaut */
    _max = 10000;
    /* blank */
    _crit = NULL;
    /* init */
    _id = 0;
}

void TNloptShare::build()
{
    fillDS();
    _tds->createTuple();
    if (_tdstrc) initTrace();
    //_coef = _NinCod * 2 + 1;
}

/* objective and constraints */
void TNloptShare::setStartingPoint(double *pt)
{
    try{
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                     "TNloptShare::setStartingPoint WARNING : this function with only a (double*) signature is not recommended anymore.\nPlease, use the setStartingPoint(int nb, double *p) that check the size of the provided array and check that it matches the number of inputs.\nFor C++ you can do: vector<double> p{0.2,0.3}; opt.setStartingPoint(p.size(),&p[0]);.\nFor python you can use numpy array to do: p=numpy.array([0.2,0.3]); opt.setStartingPoint(len(p),p);");
    }catch(URANIE::Exceptions::UWarningExceptions &ue)
    {
	ue.printMessage();
    }

    double* copy = new double[_NinDat];

    for (int i = 0; i < _NinDat; i++)
        copy[i] = pt[i];

    _guess.push_back(copy);
}

void TNloptShare::setStartingPoint(int nb, double *pt)
{
    try{
        if(nb != _NinDat)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TNloptShare::setStartingPoint ERROR : the starting point provided as dimension [%d], while the code as [%d] inputs",
                                                            nb,_NinDat));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    double* copy = new double[_NinDat];

    for (int i = 0; i < _NinDat; i++)
        copy[i] = pt[i];

    _guess.push_back(copy);
}

/* nlopt callback */
double nlopt_crit(unsigned n, const double *in, double *grad, void *info)
{
    UNUSED(n);
    double out;
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->critere(work, in, &out, grad);
    return out;
}

void nlopt_inequal(unsigned m, double *out, unsigned n, const double *in,
        double *grad, void *info)
{
    UNUSED(n,m);
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->inequality(work, in, out, grad);
}

void nlopt_equal(unsigned m, double *out, unsigned n, const double *in,
        double *grad, void *info)
{
    UNUSED(n,m);
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->equality(work, in, out, grad);
}

/* callback associated method */
void TNloptShare::calcul(struct nlopt_work *self, Bool_t grad, const double *in)
{
#ifdef SUIVI
    static int a = 0;
#endif
    int i, j;
    long id;
	id = 0;

    for EACH(i, _NinDat) if (in[i] != self->last[i]) break;

    //if (i != _NinDat || (self->lastg == kFALSE && grad == kTRUE))
    if (i != _NinDat || self->lastg != grad)
    {
#ifdef SUIVI
        if (grad)
        {
            if (!(a % 10))
                printf("!");
            else if (!(a % 5))
                printf(";");
            else
                printf(",");
        }
        else
        {
            if (!(a % 10))
                printf("|");
            else if (!(a % 5))
                printf(":");
            else
                printf(".");
        }        
        if (!(++a % 50))
            printf("%04d\n", a);
        fflush(stdout);
#endif
        for EACH(j, _NinDat) self->last[j] = in[j];
        put(self, grad, in);
        self->lastg = grad;
#ifdef COROUTINE
        (*self->yield)();
#else
        self->noerr = get(id);
#endif
        if (_tdstrc) {
            fill(self, _tdstrc, _idtrc);
            _idtrc++;
        }
    }
}

/* solve */
void TNloptShare::solver(struct nlopt_work *ret, int id, double *guess,
        vector<URANIE::DataServer::UEntry*>& codi,
        vector<URANIE::DataServer::UEntry*>& codo,
        vector<URANIE::DataServer::UEntry*>& codg
    )
{
    nlopt_opt solv, subsolv;
    long i;
    unsigned nineq = _const.GetEntries();
    unsigned nequ = _equal.GetEntries();
    double* top = new double[_NinDat];
    double* low = new double[_NinDat];

    //std::cout << _NinDat << " " << nineq << " " << nequ << std::endl;
    // solver
    solv = _solv->build(_NinDat, nineq, nequ, _prec, &subsolv);
    // bounds
    TList *inputs = _tds->getListOfAttributes();
    for EACH(i, _NinDat)
    {
        URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) inputs->At(i);
        //std::cout << att->GetName() << ": lower = " << att->getLowerBound() << ", upper = " << att->getUpperBound() << std::endl;
        low[i] = att->getLowerBound();
        top[i] = att->getUpperBound();
    }
    delete inputs;
    nlopt_set_lower_bounds(solv, low);
    nlopt_set_upper_bounds(solv, top);

    // fun
    nlopt_set_min_objective(solv, nlopt_crit, (void *) ret);
    if (nineq)
    {
        double* tol = new double[nineq];
        for EACH(i, nineq) tol[i] = 1e-7;
        nlopt_add_inequality_mconstraint(solv, nineq, nlopt_inequal,
                (void *) ret, tol);
        delete[] tol;
    }
    if (nequ)
    {
        double* tol = new double[nequ];
        for EACH(i, nequ) tol[i] = 1e-4;
        nlopt_add_equality_mconstraint(solv, nequ, nlopt_equal, (void *) ret,
                tol);
        delete[] tol;
    }
    // stop
    nlopt_set_maxeval(solv, _max);

    ret->self = this;
    ret->state = IsRUNNING;
    ret->solv = solv;
    ret->subsolv = subsolv;
    ret->id = id;
    ret->guess = guess;
    ret->last = (double *) calloc(_NinDat, sizeof(double));
    ret->work = (double *) calloc(_NinCod + _NoutCod, sizeof(double));
    ret->grad = (double *) calloc(_NinCod * _NoutCod, sizeof(double));
    ret->codo = &codo;
    ret->codi = &codi;
    ret->codg = &codg;
    ret->lastg = kFALSE;
            
	
	delete[] top;
	delete[] low;
}

void TNloptShare::run(struct nlopt_work *opt)
{
    int ret;
    
    //std::cout << "\n nlopt start " << opt->solv << std::endl ;
    ret = nlopt_optimize(opt->solv, opt->guess, &(opt->minf));

    switch (ret)
    {
    case NLOPT_FAILURE:
    case NLOPT_INVALID_ARGS:
    case NLOPT_OUT_OF_MEMORY:
    case NLOPT_FORCED_STOP:
        std::cout << "\n nlopt failed " << ret << " !\n";
        break;
    case NLOPT_ROUNDOFF_LIMITED:
        std::cout << "\n roundoff\n";
        fillOptimum(opt);
        break;
    case NLOPT_MAXEVAL_REACHED:
        std::cout << "\n max eval reached\n";
        fillOptimum(opt);
        break;
    default:
        std::cout << std::endl;
        fillOptimum(opt);
    }
    opt->state = ret;
    _retOptim = ret;

}

void TNloptShare::solverLoop()
{
    int i;

    // verif
    if (!_guess.size())
    {
        std::cout << "Need a Starting Point\n";
        return;
    }
    if (!_crit)
    {
        std::cout << "Need an Objective\n";
        return;
    }
    build();

    int siz = _guess.size();
    // solver
    struct nlopt_work * opt = new struct nlopt_work[siz];
    buildSolver(siz, opt);
#ifdef COROUTINE
    long id;
    int compt, ret;
    // coroutine
    auto fibre = new coroutine<void>::pull_type*[siz];
    for (i=0; i<siz; i++)
    {
        fibre[i] = new coroutine<void>::pull_type(
            [i, opt](coroutine<void>::push_type &yield) {
                nlopt_work *coro = &(opt[i]);
                coro->yield = &yield;
                coro->self->run(coro);
                return;
            }
        );
    }
    compt = siz;
    while (compt)
    {
        ret = get(id);
        opt[id].noerr = ret;
        (*fibre[id])();
        if (opt[id].state != IsRUNNING)
        {
            compt--;
        }
    }
    for (i=0; i<siz; i++)
    {
        cleanSolver(&opt[i]);
        delete fibre[i];
    }
    delete[] fibre;
#else // no coroutine
    // data
    for (i = 0; i < siz; i++)
    {
        run(&opt[i]);
        cleanSolver(&opt[i]);
    }

#endif
    delete[] opt;
    
    fillConstant();
}

void TNloptShare::fillOptimum(struct nlopt_work *self)
{
    calcul(self, kFALSE, self->guess);
    fill(self, _tds, _id);
    _id++;
}

void TNloptShare::fill(struct nlopt_work *self, URANIE::DataServer::TDataServer *tds, int id)
{
    auto tmpi = new vector<URANIE::DataServer::UEntry*>(self->codi->begin(), self->codi->end());
    auto tmpo = new vector<URANIE::DataServer::UEntry*>(self->codo->begin(), self->codo->end());
    vector<vector<URANIE::DataServer::UEntry*>> all;
    all.resize(1);
    allocDataEntries(1, all);
    auto tmpa = new vector<URANIE::DataServer::UEntry*>(all[0].begin()+1, all[0].end());
    
    all[0][0]->pushEntry((double) id);
    allFromCode(tmpi, tmpo, tmpa);
    tds->getTuple()->Fill(&all.at(0));
	
    for (auto val : all[0]) delete val;
    all[0].clear();
}

}
} //namespace
