//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.2 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizir2.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2011-03-31.10.57.52

 \brief   methodes communes aux algo de Vizir
 */

#include "TVizir2.h"

#include <stdio.h>

#include "TRun.h"

#include "vizir/vzragent_api.h"
#include "vizir/vzrcore_api.h"
#include "vizir/vzr_macro.h"

ClassImp (URANIE::Reoptimizer::TVizir2);

namespace URANIE
{
namespace Reoptimizer
{

// TVizir2
TVizir2::TVizir2(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *Run, TVizirSolver *solv) :
        TVizirShare(tds, Run, solv)
{
    _agent = NULL;
}

TVizir2::TVizir2(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TVizirSolver *solv) :
        TVizirShare(tds, fun, solv)
{
    _agent = NULL;
}

TVizir2::~TVizir2()
{
    if (_agent)
	{
        vzragent_dealloc(_agent);
	}
	
}

/* solve */
void TVizir2::run(int cpus)
{
    int size, outsize, ret;
    long i, j,k;
    Bool_t decd;

    size = (cpus < 2) ? 1 : cpus - 1;
    /*//size = cpus;*/
    outsize = _obj + _valid + _NoutDat + 1;
    decd = _solv->isCoded();

	struct vzrcore** core = new struct vzrcore*[size];
    double** in = new double*[size];
	for(k=0;k<size;k++) in[k] = new double[_NinDat+_Naux];
    double* phy = new double[_NinDat+_Naux];
    vector<vector<URANIE::DataServer::UEntry*>> codin;
    codin.resize(size);
    allocInCodeEntries(size, codin);
	vector<vector<URANIE::DataServer::UEntry*>> codout;
	codout.resize(size);
	allocOutCodeEntries(size, codout);
    double* out = new double[outsize];
    vector<URANIE::DataServer::UEntry*> *dat;
	vector<vector<URANIE::DataServer::UEntry*>> alltrace;
	alltrace.resize(1);
	allocDataEntries(1, alltrace);

    for EACH(i, size)
    {
        core[i] = vzragent_core(_agent);
        PRO_ASK(core[i], subscribe);
        PRO_CALL(core[i], get, in[i]);
        if (decd) decode(in[i], phy);
        else nodecode(in[i], phy);
        inputForCode(phy, &codin[i]);
        _run->put(i, &codin[i], &codout[i] );
    }

    j = size;
    while (j)
    {
        ret = _run->get(i);
        if (_tdstrc) fillTrace(&codin[i], &codout[i], &alltrace[0]);
        fillChoice(&codout[i], out);
        dat = fillData(&codin[i], &codout[i]);
        for (auto val : codout[i]) val->clearEntry();
        if (PRO_CALL(core[i], put, ret, in[i], out, dat)) {
            j--; break;
        }
        if (PRO_CALL(core[i], get, in[i])) {
            j--; continue;
        }
        if (decd) decode(in[i], phy);
        else nodecode(in[i], phy);
        inputForCode(phy, &codin[i]);
        _run->put(i, &codin[i], &codout[i]);
    }

    for (; j; j--) {
        ret = _run->get(i);
        if (_tdstrc) fillTrace(&codin[i], &codout[i], &alltrace[0]);
    }

    for EACH(i, size)
    {
        PRO_ASK(core[i], unsubscribe);
        PRO_DEALLOC(core[i]);
    }

	delete[] core;
	for(k=0;k<size;k++) {
	    delete[] in[k];
	    for (auto it :  codin[k]) delete it;
	    codin[k].clear();
	    for (auto it : codout[k]) delete it;
	    codout[k].clear();
	}
	for (auto it : alltrace[0]) delete it;
	alltrace[0].clear();

	delete[] in;
	delete[] phy;
	delete[] out;
	
}

void TVizir2::solverLoop()
{
    if (_agent) {
        // resume
        _tds->createTuple();
    }
    else {
        // init
        build();

        _obj = _crit.GetEntries();
        _valid = _const.GetEntries();
        _agent = _solv->build(_NinDat+_Naux, _obj, _valid, _prec);
    }
    run(_run->_ncpus);
    _solv->resume(_agent); 		
    fill(_agent);

    fillConstant();

}

bool TVizir2::isConverged()
{
    return vzragent_converged(_agent);
}

void TVizir2::extend(int nbeval)
{
    return vzragent_extend(_agent, nbeval);
}

void TVizir2::fitnessChanged()
{
    return vzragent_fitness_changed(_agent);
}

int TVizir2::item(int id, double *zone, double *front, int *rank,
        int *age, int *info, double *isol)
{
    double *lzone, *lfront;
    int i, ret;
    Bool_t decd = _solv->isCoded();
    
    ret = _solv->individu(_agent, id, &lzone, &lfront, NULL, rank, age, info, isol);
    if (ret) {
        if (decd) decode(lzone, zone);
        else nodecode(lzone, zone);
        for (i=0; i<_valid+_obj; i++) front[i] = lfront[i];
    }
    return ret;
}


}
} //namespace
