//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.5 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizirShare.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-01-25.04.38.31

 \brief   methodes communes aux algo de Vizir
 */

#include "TVizirShare.h"

#include <stdio.h>

#include "TFitGetter.h"
#include "TMasterVar.h"

#include "vizir/vzragent_api.h"
#include "vizir/vzrcore_api.h"
#include "vizir/vzr_aleas.h"
#include "vizir/vzr_macro.h"

#define EXTRA

ClassImp (URANIE::Reoptimizer::TVizirShare);

namespace URANIE
{
namespace Reoptimizer
{

//TVizirShare
TVizirShare::TVizirShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, TVizirSolver *solv) :
        TOptimShare(tds, run), _solv(solv)
{
    //fillOutDS();
    //_tds->createTuple();
}

TVizirShare::TVizirShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TVizirSolver *solv) :
        TOptimShare(tds, fun), _solv(solv)
{
    //fillOutDS();
    //_tds->createTuple();
}

TVizirShare::~TVizirShare()
{
    TFitGetter *one;
    TIter nexto(&_crit);

    while ((one = (TFitGetter *) nexto()))
        delete one;
}

/* constructeur helpers */
void TVizirShare::build()
{
    URANIE::DataServer::TAttribute *tmp;
    /* TEval output */
    fillDS();
    /* trace */
    if (_tdstrc) initTrace();
    /* TVizirShare output */
#ifdef EXTRA
    tmp = new URANIE::DataServer::TAttribute("rgpareto");
    _tds->addAttribute(tmp);
    _sortie.Add(new URANIE::Relauncher::TPostVar(tmp, _NoutCod, _NinDat+_Naux+_NoutDat));
    tmp->delShare();
    tmp = new URANIE::DataServer::TAttribute("generation");
    _tds->addAttribute(tmp);
    _sortie.Add(new URANIE::Relauncher::TPostVar(tmp, _NoutCod+1, _NinDat+_Naux+_NoutDat+1));
    tmp->delShare();
    _NoutDat += 2;
#endif
    _tds->createTuple();
}

void TVizirShare::setFixedSeed()
{
    mt_init(1);
}

/* objective and constraints */
void TVizirShare::addObjective(URANIE::DataServer::TAttribute *obj,
        TOptimFit *fit)
{
    TFitGetter *one;
    one = new TFitGetter(obj, _fun->getOutputList(), fit);
    _crit.Add(one);
}

/* solve */
void TVizirShare::nodecode(Double_t *coded, Double_t *phys)
{
    int i;
    
    for (i = 0; i < _NinDat+_Naux; i++) phys[i] = coded[i];
}

void TVizirShare::decode(Double_t *coded, Double_t *phys)
{
    Int_t i;
    Double_t low, top;
    URANIE::DataServer::TAttribute *att;

    TList *attributeList = _tds->getListOfAttributes();
    TIter next(attributeList);

    for (i = 0; i < _NinDat+_Naux; i++)
    {
        att = (URANIE::DataServer::TAttribute *) next();
        if (att->isLowerBounded() && att->isUpperBounded()) {
            low = att->getLowerBound();
            top = att->getUpperBound();
            phys[i] = coded[i] * (top - low) + low;
        }
        else phys[i] = coded[i];
    }
    delete attributeList;
}

void TVizirShare::fillChoice(vector<URANIE::DataServer::UEntry*> *all, double *out)
{
    TFitGetter *one;
    TIter nextc(&_const);
    TIter nexto(&_crit);
    int i = 0;

    // obj et critere
    while ((one = (TFitGetter *) nextc()))
        out[i++] = one->from(all);
    while ((one = (TFitGetter *) nexto()))
        out[i++] = one->from(all);
    // code output
    //outputFromCode(all, out + i);
}

vector<URANIE::DataServer::UEntry*> *TVizirShare::fillData(
        vector<URANIE::DataServer::UEntry*> *in, 
        vector<URANIE::DataServer::UEntry*> *out)
{
    auto tmpi = new vector<URANIE::DataServer::UEntry*>(in->begin(), in->end());
    auto tmpo = new vector<URANIE::DataServer::UEntry*>(out->begin(), out->end());
    auto all = new vector<URANIE::DataServer::UEntry*>();
    allocDataEntries(*all);
    auto tmpa = new vector<URANIE::DataServer::UEntry*>(all->begin()+1, all->end());
    
    all->at(0)->pushEntry(-1.0);
    allFromCode(tmpi, tmpo, tmpa);
    return all;
}

void TVizirShare::fillTrace(vector<URANIE::DataServer::UEntry*> *in, 
        vector<URANIE::DataServer::UEntry*> *out,
        vector<URANIE::DataServer::UEntry*> *all)
{
    vector<URANIE::DataServer::UEntry*> *tmpi = 
        new vector<URANIE::DataServer::UEntry*>(in->begin(), in->end());
    vector<URANIE::DataServer::UEntry*> *tmpo = 
        new vector<URANIE::DataServer::UEntry*>(out->begin(), out->end());
    vector<URANIE::DataServer::UEntry*> *tmpa = 
        new vector<URANIE::DataServer::UEntry*>(all->begin()+1, all->end());
    
    all->at(0)->pushEntry((double) _idtrc);
    allFromCode(tmpi, tmpo, tmpa);
    tmpa = new vector<URANIE::DataServer::UEntry*>(all->begin(), all->end()-2);
    _tdstrc->getTuple()->Fill(tmpa);

    _idtrc++;
    for (auto val : *all) val->clearEntry();
    delete tmpa;
}

/* resultats */
void TVizirShare::fill(struct vzragent *ag)
{
    int tmp = 0;
    fill(ag, tmp);
}

void TVizirShare::fill(struct vzragent *ag, int &decal)
{
    int i, j;
    double *zone, *front;
    void *dat;
    int rank, ng;
	int nx = _tds->getNAttributes();
	Bool_t decd = _solv->isCoded();

    double* par = new double[_NinDat+_Naux];
    double* dval = new double[nx + 1];	
    

    int idout = _valid + _obj;

    for (i = 0; _solv->individu(ag, i, &zone, &front, &dat, &rank, &ng); i++, decal++)
    {
        auto data = (vector<URANIE::DataServer::UEntry*>*) dat;
        data->at(0)->pushEntry(1.0 *(decal));
        data->at(_NinDat+_Naux+_NoutDat-1)->pushEntry((double) rank);
        data->at(_NinDat+_Naux+_NoutDat)->pushEntry((double) ng);
        _tds->getTuple()->Fill(data);
    }

	delete[] par;
	delete[] dval;

}


}
} //namespace
