//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizirSolver.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-04-26.11.23.44

 \brief   classe abstraite et racine pour les solvers de Vizir
 */

#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#define R_OK    4
#define W_OK    2
#define access _access
#endif

#include "TVizirSolver.h"

#include <vector>

#include "UEntry.h"

#include "vizir/vzr_macro.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzragent_api.h"
#include "vizir/vzrsort.h"

ClassImp (URANIE::Reoptimizer::TVizirSolverShare);

#define KNVOIS 16
#define KNTAUX 0.7


namespace URANIE
{
namespace Reoptimizer
{

/* utilitaire */
void dealloc_uentry_vector(void *dat)
{
    auto data = (std::vector<URANIE::DataServer::UEntry*> *) dat;
    
    for (auto it : *data) delete it;
    data->clear();
    delete data;
}

/* Class TVizirSolverShare */
void TVizirSolverShare::defaultShare()
{
    _step = _pop = 0;
    _eval = 100000;
    _policy = VZR_DEFLT_POLICY;
    _diversity.algo = undef;
    _cut1 = _cut2 = _continue = 0;
    _resume = NULL;
    _doresume = kFALSE;
    _quiet = kFALSE;
}

void TVizirSolverShare::setSize(int pop, int eval, int step)
{
    if (pop)
        _pop = pop;
    if (eval)
        _eval = eval;
    if (step)
        _step = step;
}

Bool_t TVizirSolverShare::setResume(int pop, char *nfile)
{
    _resume = nfile;
    _doresume = (access(nfile, R_OK | W_OK)) ? kFALSE : kTRUE ;
    if (_doresume) _pop = pop;
    return _doresume;
}

void TVizirSolverShare::resume(vzragent_t *ag)
{
    if (_resume != NULL) vzragent_dump(ag, _resume);
}

void TVizirSolverShare::setStrictConstraintPolicy()
{
    _policy = VZR_STRICT_POLICY;
}

void TVizirSolverShare::setSoftConstraintPolicy()
{
    _policy = VZR_SOFT_POLICY;
}

void TVizirSolverShare::setConstraintPolicy(int val)
{
    switch (val) {
    case 1 :
        _policy = VZR_STRICT_POLICY;
        break;
    case 2 :
        _policy = VZR_SOFT_POLICY;
        break;
    default :
        _policy = VZR_DEFLT_POLICY;
        break;
    }
}

void TVizirSolverShare::setMogaDiversity(int val)
{
    _diversity.algo = moga;
    _diversity.nba = val;
}

void TVizirSolverShare::setCrowdDiversity(int vois)
{
    _diversity.algo = crowd;
    _diversity.nba = vois;
}

void TVizirSolverShare::setPairDiversity(int vois)
{
    
    _diversity.algo = pair ;
    _diversity.nba = vois;
}

void TVizirSolverShare::setIbeaDiversity(double k)
{
    _diversity.algo = ibea;
    _diversity.val = k;
}

void TVizirSolverShare::setKneeDiversity(int vois, double taux)
{
    _diversity.algo = knee;
    if (vois == 0 && taux == 0.0) {
        _diversity.val = 0.0;
        _diversity.nb = 0;
    }
    else {
        _diversity.nb = (vois > 4) ? vois : KNVOIS ;
        _diversity.val = (taux == 0.0) ? KNTAUX : taux ;
    }
}

void TVizirSolverShare::setMoeadDiversity(int cut1, int cut2, int vois)
{
    _diversity.algo = moead;
    _diversity.nba = cut1;
    _diversity.nbb = cut2;
    _diversity.nbc = vois;
    
    _cut1 = cut1; _cut2 = cut2;
}

void TVizirSolverShare::setModalDiversity(int vois)
{
    
    _diversity.algo = modal ;
    _diversity.nba = vois;
}

void TVizirSolverShare::setStoppingCriteria(int stop)
{
    _continue = stop;
}

void TVizirSolverShare::setQuiet()
{
    _quiet = kTRUE;
}

void TVizirSolverShare::buildShare(struct vzrbase *vzr)
{
    vzr->pop.dealloc_data = dealloc_uentry_vector;
    if (_policy) vzrpop_set_policy(&vzr->pop, VZR_POLICY_MASK, _policy);
    switch (_diversity.algo) {
        case undef : break;
        case moga : 
            if (_diversity.nba) vzr->pop.sort = create_moga2();
            else vzr->pop.sort = create_moga();
            break;
        case crowd :
            switch (_diversity.nba) {
            case 0 :
                vzr->pop.sort = create_crowding(); break;
            case 1 :
            case 2 :
            case 3 :
                vzr->pop.sort = create_neighbor(); break;
            default :
                vzr->pop.sort = create_neighbor_full(_diversity.nba); break;
            }
            break;
        case pair :
            vzr->pop.sort = (_diversity.nba == 0) ?
                    create_pair() : create_pair_full(_diversity.nba, 20, 50);
            break;
        case ibea :
            vzr->pop.sort = (_diversity.val == 0.0) ?
                    create_ibea() : create_ibea_full(_diversity.val);
            break;
        case knee :
            if (_diversity.nb == 0 && _diversity.val == 0.0) vzr->pop.sort = create_elbow();
            else vzr->pop.sort = create_elbow_full(_diversity.nb, _diversity.val);
            break;
        case moead :
            vzr->pop.sort = (_diversity.nbc) ?
                    create_moeadd_full(_diversity.nba, _diversity.nbb, _diversity.nbc) : 
                    create_moeadd(_diversity.nba, _diversity.nbb);
            break;
        case modal :
            vzr->pop.sort = (_diversity.nba == 0) ?
                    create_modal() : create_modal_full(_diversity.nba);
            break;
    }
    if (_pop == 0) {
        if (_cut1 ) { //MOEADD
            _pop = moeadd_pop_size(vzr->pop.cost, _cut1, _cut2);
            _step = _pop;
        }
        else {
            _pop=250;
            _step=100;
        }
    }
    if (_step == 0) _step = 2*_pop/5;
    if (_continue) vzrpop_set_policy(&vzr->pop, VZR_PARETO_CONT, VZR_PARETO_CONT);
    if (_quiet) vzr->quiet = 1;
}

/* resultats */
int TVizirSolverShare::individu(struct vzragent *ag, int id, double **zone,
        double **front, void **data, int *rank, int *age, int *info, double *isol)
{
    return vzragent_item(ag, id, zone, front, data, rank, age, info, isol);
}


}
} // namespace

