//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//// $Id$
//// $Author$
//// $Date$
//// $Revision$
//// $State$
////////////////////////////////////////////////////////////////
/*!\file   TestTVizir.cxx
 *   \author Gilles Arnaud <gilles.arnaud@cea.fr>
 *   \date   mer. juin 19 10:48:18 CEST 2013
 * 
 *   \brief  test file for Vizir solver classes and TRun classes
 * 
 * */

#include "TestTVizir.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PREC_FINE 4.e-4
#define DBL_PREC_STD  4.e-3
#define DBL_PREC_LARGE 3.e-2
#define DBL_PREC_HUGE 6.e-2

// ROOT
#include "Rtypes.h"
//#include "Api.h"

//URANIE
#include "UExceptions.h"
#include "TDataServer.h"
#include "TAttribute.h"

//#include "TCIntEval.h"
#include "TCJitEval.h"
#include "TKeyScript.h"
#include "TFlatResult.h"
#include "TCodeEval.h"
#include "TComposeEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"

#include "TVizirGenetic.h"
#include "TVizirSwarm.h"
#include "TVizirSimplex.h"
#include "TVizir2.h"
#include "TVizirIsland.h"

#ifdef WITH_PYTHON
#include "python_commands.h"
#endif

#include "TestUserFunctions.h"
#include "uranie_commands.h"

/* user defined */
#define TOLERANCE 0.001
#define NBmaxEVAL 70000
#define STEPDD 20
#define STEP 77 
#define STOP 231
//#define TAILLEPOP 10 

#define NBGENE 7 // nombre de paramètres

using namespace URANIE::DataServer;
using namespace URANIE::Relauncher;
using namespace URANIE::Reoptimizer;

// Constructor Destructor
TVizirTest::TVizirTest()
{
    _todel.SetOwner(kTRUE);
}

TVizirTest::~TVizirTest()
{
    UranieTests::removeFolder( gSystem->pwd(), "URA_*");
}

// pre and post treatemnt
void TVizirTest::setUp()
{
    // inputs
    _x = new TAttribute("x", -4., 4.);
    _y = new TAttribute("y", -4., 4.);
    _z = new TAttribute("z", -4., 4.);
    // objectives
    _obj1 = new TAttribute("obj1");
    _obj2 = new TAttribute("obj2");
    _obj12 = new TAttribute("obj12");
    // constraints
    _cont1 = new TAttribute("cont1");
    _cont2 = new TAttribute("cont2");
    _cont3 = new TAttribute("cont3");
    _cont = new TAttribute("cont", URANIE::DataServer::TAttribute::kVector);
}

void TVizirTest::tearDown()
{
    _todel.Clear();
    //delete _fun;
    delete _cont;
    delete _cont3;
    delete _cont2;
    delete _cont1;
    delete _obj1;
    delete _obj2;
    delete _obj12;
    delete _x;
    delete _y;
    delete _z;
}

// TEST SUITE
void TVizirTest::mono()
{
    TDataServer tds("tds", "vizir test tds");
    tds.keepFinalTuple(false);
    _fun->addAllInputs(&tds);

    TVizirSwarm solv;
    //TVizirSimplex solv;
    //solv.setSize(STOP, NBmaxEVAL, STEP/2);
    solv.setSize(STOP, NBmaxEVAL);
    solv.setQuiet();
    TVizir2 vzr(&tds, _fun, &solv);
    vzr.setFixedSeed();
    vzr.addObjective(_obj12);

    vzr.solverLoop();
	
    validate(tds);
    validateMono(tds);
	
}

void TVizirTest::multi()
{
    TSequentialRun run(_fun);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tds", "vizir test tds");
        tds.keepFinalTuple(false);
        _fun->addAllInputs(&tds);

        TVizirGenetic solv;
        solv.setCrowdDiversity();
        solv.setSize(STOP, NBmaxEVAL, STEP);
        solv.setQuiet();
        TVizir2 vzr(&tds, &run, &solv);
        vzr.setFixedSeed();
        vzr.addObjective(_obj1);
        vzr.addObjective(_obj2);

        vzr.solverLoop();

        run.stopSlave();

        validate(tds);
        validateMulti(tds);
    }
}

void TVizirTest::many(TVizirSolver &solv)
{
    TSequentialRun run(_fun);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tds", "vizir test tds");
        tds.keepFinalTuple(false);
        _fun->addAllInputs(&tds);

        TVizir2 vzr(&tds, &run, &solv);
        vzr.setFixedSeed();
        vzr.addObjective(_cont1);
        vzr.addObjective(_cont2);
        vzr.addObjective(_cont3);

        vzr.solverLoop();

        run.stopSlave();

        validate(tds);
        validateMany(tds);
    }
}

void TVizirTest::many_pair()
{
    TVizirGenetic solv;
    solv.setPairDiversity();
    solv.setSize(STOP, NBmaxEVAL, STEP);
    solv.setQuiet();
    
    many(solv);
}

void TVizirTest::many_ibea()
{
    TVizirGenetic solv;
    solv.setIbeaDiversity();
    solv.setSize(STOP, NBmaxEVAL, STEP);
    solv.setQuiet();
    
    many(solv);
}

void TVizirTest::many_knee()
{
    TVizirGenetic solv;
    solv.setKneeDiversity();
    solv.setSize(STOP, NBmaxEVAL, STEP);
    solv.setQuiet();
    
    many(solv);
}

void TVizirTest::many_moeadd()
{
    TVizirGenetic solv;
    solv.setMoeadDiversity(STEPDD);
    solv.setQuiet();
    
    many(solv);
}


void TVizirTest::monoConstrained()
{
    TThreadedRun run(_fun, 4);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tds", "vizir test tds");
        tds.keepFinalTuple(false);
        _fun->addAllInputs(&tds);

        //TVizirGenetic solv;
        //solv.setSize(STOP, NBmaxEVAL);
        TVizirSimplex solv;
        solv.setSize(STOP, NBmaxEVAL, STEP/2);
        solv.setQuiet();
        TVizir2 vzr(&tds, &run, &solv);
        vzr.setFixedSeed();
        vzr.addObjective(_obj12);
        TGreaterFit bille(1.0);
        vzr.addConstraint(_cont1, &bille);
        vzr.addConstraint(_cont2, &bille);
        vzr.addConstraint(_cont3, &bille);

        vzr.solverLoop();

        run.stopSlave();

        validate(tds);
        validateMonoC(tds);
    }
}

void TVizirTest::multiConstrained()
{
    TThreadedRun run(_fun, 4);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tds", "vizir test tds");
        tds.keepFinalTuple(false);
        _fun->addAllInputs(&tds);

        TVizirGenetic solv;
        solv.setMoeadDiversity(STOP-1);
        //solv.setSize(STOP, NBmaxEVAL, STEP);
        solv.setQuiet();
        TVizir2 vzr(&tds, &run, &solv);
        vzr.setFixedSeed();
        vzr.addObjective(_obj1);
        vzr.addObjective(_obj2);
        TGreaterFit bille(1.0);
        vzr.addConstraint(_cont1, &bille);
        vzr.addConstraint(_cont2, &bille);
        vzr.addConstraint(_cont3, &bille);

        vzr.solverLoop();

        run.stopSlave();

        validate(tds);
        validateMultiC(tds);
    }
}

void TVizirTest::validate(TDataServer &tds)
{
    Int_t pat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(STOP, pat);
    // attributes
    // x, y, z, obj1, obj2, obj12, cont1, cont2, cont3
    CPPUNIT_ASSERT_EQUAL(11, tds.getNAttributes());
    //CPPUNIT_ASSERT_EQUAL(9, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("z"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("obj1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("obj12"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("cont2"));
    // values
}

void TVizirTest::validateMono(TDataServer &tds)
{
    Int_t i, elmnt[] =
    { 1, STEP, STOP-1, 0 };
    // values
    // theo .75, .75, 0.55, 0.125, 0.125, 0.25, 1.4275, 0.9275, 0.9275
    for (i = 0; elmnt[i]; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.75, tds.getValue("x", elmnt[i]),
                DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.55, tds.getValue("z", elmnt[i]),
                DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.25, tds.getValue("obj12", elmnt[i]),
                DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1.4275, tds.getValue("cont1", elmnt[i]),
                DBL_PREC_LARGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.9275, tds.getValue("cont3", elmnt[i]),
                DBL_PREC_LARGE);
    }
}

void TVizirTest::validateMonoC(TDataServer &tds)
{
    Int_t i, elmnt[] =
    { 1, STEP, STOP-1, 0 };
    // values
    // theo .72963, .72963, 0.59481, ?, ?, 0.25568, 1.41852, 1., 1.
    for (i = 0; elmnt[i]; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.72963, tds.getValue("y", elmnt[i]),
                DBL_PREC_LARGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.59481, tds.getValue("z", elmnt[i]),
                DBL_PREC_LARGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.25568, tds.getValue("obj12", elmnt[i]),
                DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1.41852, tds.getValue("cont1", elmnt[i]),
                DBL_PREC_HUGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("cont2", elmnt[i]),
        		DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("cont3", elmnt[i]),
                DBL_PREC_STD);
    }
}

void TVizirTest::validateMonoCVec(TDataServer &tds)
{
    Int_t i, elmnt[] =
    { 1, STEP, STOP-1, 0 };
    // values
    // theo .72963, .72963, 0.59481, ?, ?, 0.25568, 1.41852, 1., 1.
    for (i = 0; elmnt[i]; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.72963, tds.getValue("y", elmnt[i]),
                DBL_PREC_LARGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.59481, tds.getValue("z", elmnt[i]),
                DBL_PREC_LARGE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.25568, tds.getValue("obj1", elmnt[i]),
                DBL_PREC_STD);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("cont1", elmnt[i]),
                DBL_PREC_STD);
    }
}

void TVizirTest::validateMulti(TDataServer &tds)
{
    Int_t i, id[] =
    { 1, STEP, STOP-1, 0 };
    Double_t x, y, z, pc, d, ref;

    ref = sqrt(2.) / 2;
    // values
    // theo point sur le segment de droite [1.,.5,.55][.5,1.,.55]
    for (i = 0; id[i]; i++)
    {
        // z = .55
        z = tds.getValue("z", id[i]);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.55, z, DBL_PREC_HUGE);
        // produit cartesien
        x = tds.getValue("x", id[i]);
        y = tds.getValue("y", id[i]);
        pc = x * 0.75 + y * 0.75 + z * 0.55;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1.4275, pc, DBL_PREC_HUGE);
        // somme des distances aux deux extremités 
        d = sqrt(tds.getValue("obj1", id[i]))
                + sqrt(tds.getValue("obj2", id[i]));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref, d, DBL_PREC_HUGE);
    }
}

void TVizirTest::validateMany(TDataServer &tds)
{
    Int_t i, id[] =
    { 1, STEP, STOP-1, 0 };
    Double_t x, y, z, pc, d, ref;

    ref = sqrt(2.) / 2;
    // values
    // theo point sur le segment de droite [1.,.5,.55][.5,1.,.55]
    for (i = 0; id[i]; i++)
    {
        // z = .55
        z = tds.getValue("z", id[i]);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, z, 0.22);
        // produit cartesien
        x = tds.getValue("x", id[i]);
        y = tds.getValue("y", id[i]);
        pc = x + y ;
        CPPUNIT_ASSERT(pc > -0.2);
        CPPUNIT_ASSERT(pc < 2.3);
        pc = x - y;
        CPPUNIT_ASSERT(pc > -1.2);
        CPPUNIT_ASSERT(pc < 1.2);    
    }
}

void TVizirTest::validateMultiC(TDataServer &tds)
{
    Int_t i, id[] =
    { 1, STEP, STOP-1, 0 };
    Double_t c3, c2;

    // values
    // theo 
    for (i = 0; id[i]; i++)
    {
        // cont2 ou cont3 saturé
        c2 = tds.getValue("cont2", id[i]);
        c3 = tds.getValue("cont3", id[i]);
        c2 = (c2 < c3) ? c2 : c3;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1., c2, DBL_PREC_LARGE);
    }
}

/**** TVizirJitTest ****/
// Constructor Destructor
TVizirJitTest::TVizirJitTest()
{
}

TVizirJitTest::~TVizirJitTest()
{
}

// pre and post treatemnt
void TVizirJitTest::setUp()
{
    TVizirTest::setUp();
    // direct compiled fun
    TCJitEval *fun_jit = new TCJitEval(select);
    fun_jit->setInputs(3, _x, _y, _z);
    fun_jit->setOutputs(6, _obj1, _obj2, _obj12, _cont1, _cont2, _cont3);

    _fun = fun_jit;
    _todel.Add(fun_jit);
}

/**** TVizirJitTest2 ****/
// Constructor Destructor
TVizirJitTest2::TVizirJitTest2()
{
}

TVizirJitTest2::~TVizirJitTest2()
{
}

// pre and post treatemnt
void TVizirJitTest2::setUp()
{
    TVizirTest::setUp();
    // direct compiled fun
    TCJitEval *fun_jit = new TCJitEval(fselect_vec);
    fun_jit->setInputs(3, _x, _y, _z);
    fun_jit->setOutputs(2, _obj1, _cont);

    TCJitEval *fun_jit2 =  new TCJitEval(fselect_min);
    fun_jit2->addInput(_cont);
    fun_jit2->addOutput(_cont1);

    TComposeEval *fun_comp = new TComposeEval();
    fun_comp->addEval(fun_jit);
    fun_comp->addEval(fun_jit2);
    _fun = fun_comp;
    _todel.Add(fun_comp);
    _todel.Add(fun_jit2);
    _todel.Add(fun_jit);
}

void TVizirJitTest2::monoConstrained()
{
    TThreadedRun run(_fun, 4);
    run.startSlave();
    if (run.onMaster())
    {
        TDataServer tds("tds", "vizir test tds");
        tds.keepFinalTuple(false);
        _fun->addAllInputs(&tds);

        TVizirGenetic solv;
        solv.setSize(STOP, NBmaxEVAL, STEP);
        solv.setQuiet();
        TVizir2 vzr(&tds, &run, &solv);
        vzr.setFixedSeed();
        vzr.addObjective(_obj1);
        TGreaterFit bille(1.0);
        vzr.addConstraint(_cont1, &bille);

        vzr.solverLoop();

        run.stopSlave();

        validateMonoCVec(tds);
    }
}

/**** TVizirIntTest ****/
// Constructor Destructor
TVizirIntTest::TVizirIntTest()
{
#ifdef ROOT5
    // creation of CInt function pointers
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;

    method = globalscope.GetMethod("select", "Double_t*,Double_t*", &dummy);
    _p2fSelect = method.PointerToFunc();
#endif
}

TVizirIntTest::~TVizirIntTest()
{
}

// pre and post treatemnt
void TVizirIntTest::setUp()
{
    TVizirTest::setUp();
    // compiled fun with interpreted handle
#ifdef ROOT5
    TCIntEval *fun_int = new TCIntEval(_p2fSelect);
#else
    TCJitEval *fun_int = new TCJitEval(select);
#endif
    fun_int->setInputs(3, _x, _y, _z);
    fun_int->setOutputs(6, _obj1, _obj2, _obj12, _cont1, _cont2, _cont3);

    _fun = fun_int;
    _todel.Add(fun_int);
}

#ifdef ROOT5
void TVizirIntTest::monoConstrained()
{
    CPPUNIT_ASSERT_THROW(TThreadedRun run(_fun, 4), URANIE::Relauncher::TEval::EvalIsNotThreadable);
}

void TVizirIntTest::multiConstrained()
{
    CPPUNIT_ASSERT_THROW(TThreadedRun run(_fun, 4), URANIE::Relauncher::TEval::EvalIsNotThreadable);
}
#endif

/**** TVizirExtTest ****/
// Constructor Destructor
TVizirExeTest::TVizirExeTest()
{
}

TVizirExeTest::~TVizirExeTest()
{
}

// pre and post treatemnt
void TVizirExeTest::setUp()
{
    TVizirTest::setUp();
    // external code
    TKeyScript *fin = new TKeyScript("testVizirCode.py");
    fin->setInputs(3, _x, "x", _y, "y", _z, "z");
    TFlatResult *fout = new TFlatResult("testVizirCode.res");
    fout->setOutputs(5, _cont1, _obj1, _obj2, _cont2, _cont3);

    string python_command;
#ifdef WITH_PYTHON
    UranieTests::set_python_executable(python_command);
#endif
    python_command += " testVizirCode.py > testVizirCode.res";
    TCodeEval *fun_code = new TCodeEval(python_command.data());
    fun_code->addInputFile(fin);
    fun_code->addOutputFile(fout);

    // jit code
    TCJitEval *fun_jit = new TCJitEval(add);
    fun_jit->addInput(_obj1);
    fun_jit->addInput(_obj2);
    fun_jit->addOutput(_obj12);

    // composition
    TComposeEval *code = new TComposeEval();
    code->addEval(fun_code);
    code->addEval(fun_jit);

    _fun = code;

    _todel.Add(code);
    _todel.Add(fun_jit);
    _todel.Add(fun_code);
    _todel.Add(fin);
    _todel.Add(fout);
}

void TVizirExeTest::tearDown()
{
    TVizirTest::tearDown();
}
