/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2010-2023 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   pso_movestd.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2010-03-15
 
 \brief  deplacement standart des particules
*/

#include "swarm/pso_struct.h"

#ifndef COMBINATOIRE

#include <stdlib.h>
#include <math.h>

#include "pso_movestd.h"

#include "vizir/vzr_aleas.h"
#include "vizir/vzr_macro.h"
#include "vizir/vzr_alloc.h"

#define C1 0.689343
#define C2 1.42694

#define BOUND_BIAS
#define PERTE 5.

/* données 
 * */
struct plan {
    int nbgene;
    double c1, c2;
    void (*fealgood)(int, double*);
    void (*fealbad)(int, double*);
};

struct move {
    struct plan *dim;
    double *posit, *speed;
};

/* utilitaires
 * */
// solution aleatoire
static INLINE double mt_sym(void)
{
    return mt_real1() * 0.6 - 0.3;
}

static INLINE double mt_gauss_positif(void)
{
    return mt_gauss() * 0.25 + 0.5;
}

static INLINE void fill_std(int size, double (*fun)(void), double *descendant)
{
    long i;

    for EACH(i, size) {
        descendant[i] = fun();
    }
}

static void fill_uniform(int size, double *ret) {
    fill_std(size, &mt_real1, ret);
}

static void fill_gauss(int size, double *ret) {
    fill_std(size, &mt_gauss_positif, ret);
}

#ifdef SPHERE
static void fill_sphere(int size, double *ret)
{
    long    i;
    double  cum, r;

    fill_std(size, &mt_gauss, ret);
    cum = 0;
    for EACH(i, size)
        cum += ret[i] * ret[i];
    cum = sqrt(cum);
    r = pow(mt_real1(), size);
    for EACH(i, size)
        ret[i] *= cum * r;
}
#endif

static void fill_sphere_positif(int size, double *ret)
{
    long    i;
    double  cum, r;

    fill_std(size, &mt_gauss, ret);
    cum = 0;
    for EACH(i, size)
        cum += ret[i] * ret[i];
    cum = sqrt(cum);
    r = pow(mt_real1(), size);
    for EACH(i, size)
        ret[i] = fabs(ret[i] *cum * r);
}

void (*filler[])(int, double *) = {
    &fill_uniform, &fill_sphere_positif, &fill_gauss
};

static void confiner(struct move *self, double *vit)
{
    int i;
    double next;
#ifdef BOUND_BIAS
    int j;
    double coef, tmp;

    coef = 1.0;
    j = -1;
    for EACH(i, self->dim->nbgene) {
        if (vit[i] == 0.0) continue; // NaN
        tmp = (vit[i] > 0 ) ? 
                (1.0 - self->posit[i]) / vit[i] :
                -self->posit[i] / vit[i];
        if (tmp < coef) {
            coef = tmp;
            j = i;
        }
    }

    if (coef == 1.) { // in bounds
        for EACH(i, self->dim->nbgene) {
            self->posit[i] += vit[i];
            self->speed[i] = vit[i];
        }
    }
    else if (coef > 0.9) { // scotch on bounds
        for EACH(i, self->dim->nbgene) {
            if (i == j) { // roundoff
                self->posit[i] = (vit[i] > 0) ? 1 : 0;
                self->speed[i] = -coef * vit[i];
            }
            else {
                self->posit[i] += coef * vit[i];
                self->speed[i] = coef * vit[i];
            }
        }
    }
    else { // atenuated rebond
#endif
        for EACH(i, self->dim->nbgene) {
            next = self->posit[i] + vit[i];
            if (next > 1.0) {
                self->posit[i] = 1. - (next-1.)/PERTE;
                self->speed[i] = vit[i]/-PERTE;
            }
            else if (next < 0.0) {
                self->posit[i] = next/-PERTE;
                self->speed[i] = vit[i]/-PERTE;
            }
            else {
                self->posit[i] = next;
                self->speed[i] = vit[i];
            }
        }
#ifdef BOUND_BIAS
    }
#endif
}
    
static void stdmove(struct move *self, void (*fill)(int, double*), double *own, double *other, double *ret)
{
    int i;
    VZRTAB(double, r1, self->dim->nbgene);
    VZRTAB(double, r2, self->dim->nbgene);

    (*fill)(self->dim->nbgene, r1);
    (*fill)(self->dim->nbgene, r2);
    for EACH(i, self->dim->nbgene)
        ret[i] = self->dim->c1 * self->speed[i] + self->dim->c2 * (
                r1[i] * (own[i] - self->posit[i])
                + r2[i] * (other[i] - self->posit[i])
                );
    VZRTAB_OUT(r1);
    VZRTAB_OUT(r2);
}

/* definition des operateurs de mouvement des particules
 * */
// mouvement sans actracteur
static void sample(void *truc, double *fils)
{
    int i;
    struct move *self = (struct move *) truc;
    VZRTAB(double, ret, self->dim->nbgene);

    for EACH(i, self->dim->nbgene) {
        ret[i] = self->speed[i];
    }
    // croisement
    confiner(self, ret);
    VZRTAB_OUT(ret);
    // recopie de la position
    for EACH(i, self->dim->nbgene) {
        fils[i] = self->posit[i];
    }
}

// mouvement avec attracteur
static void goodmove(void *truc, double *glob, double *loc, double *fils)
{
    int i;
    struct move *self = (struct move *) truc;
    VZRTAB(double, ret, self->dim->nbgene);

    // croisement
    stdmove(self, self->dim->fealgood, loc, glob, ret);
    confiner(self, ret);
    VZRTAB_OUT(ret);

    // recopie du parent
    for EACH(i, self->dim->nbgene) {
        fils[i] = self->posit[i];
    }
}

static void badmove(void *truc, double *glob, double *loc, double *fils)
{
    int i;
    struct move *self = (struct move *) truc;
    VZRTAB(double, ret, self->dim->nbgene);

    // croisement
    stdmove(self, self->dim->fealbad, loc, glob, ret);
    confiner(self, ret);
    VZRTAB_OUT(ret);

    // recopie du parent
    for EACH(i, self->dim->nbgene) {
        fils[i] = self->posit[i];
    }
}

static void dealloc(struct psomove *pro)
{
    struct move *self = (struct move *) pro->data;

    DELvecteur(self->posit);
    DELvecteur(self->speed);
    free(self);
    free(pro);
}

static struct psomove *movestd_objet(void *info)
{
    struct psomove *ret;
    struct move *dat;
    struct plan *dim = (struct plan *) info;

    /* data */
    dat = (struct move *) malloc(sizeof(struct move));
    dat->dim = dim;
    NEWvecteur(double, dat->posit, dim->nbgene);
    fill_uniform(dim->nbgene, dat->posit);
    NEWvecteur(double, dat->speed, dim->nbgene);
#ifdef SPHERE
    fill_sphere(dim->nbgene, dat->speed);
#else
    fill_std(dim->nbgene, &mt_sym, dat->speed);
#endif
    /* fermeture */
    ret = (struct psomove *) malloc(sizeof(struct psomove));
    ret->data = (void *) dat;
    ret->sample = &sample;
    ret->goodmove = &goodmove;
    ret->badmove = &badmove;
    ret->dealloc = &dealloc;

    return ret;
}

/* fabrique
 * */
static void movestd_clean(struct psomovebuild *self)
{
    free(self->data);
    free(self);
}

struct psomovebuild *pso_movestd(int nbgene, int good, int bad)
{
    struct psomovebuild *ret;
    struct plan *dat;

    dat = (struct plan *) malloc(sizeof(struct plan));
    dat->nbgene = nbgene;
    dat->c1 = C1; dat->c2 = C2;
    dat->fealgood = (good >=0 && good < 3)? filler[good] : filler[0];
    dat->fealbad = (bad >=0 && bad < 3)? filler[bad] : filler[2];

    /* fermeture */
    ret = (struct psomovebuild *) malloc(sizeof(struct psomovebuild));
    ret->data = (void *) dat;
    ret->create = &movestd_objet;
    ret->dealloc = &movestd_clean;

    return ret;
} 

#endif

