/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2007-2023 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_hongrois.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2007-05-15
 
 \brief  affectation lineaire algorithme hongrois
*/


#include <stdlib.h>
#ifdef WIN32
#define WITHALLOC
#endif

/* booleen */
#define OUI 1
#define NON 0

/* mark ligne colonne */
#define UN -2
#define NoUN -1

/* mark case */
#define IMPOSSIBLE -2
#define INTERDIT -1
#define POSSIBLE 0
#define CHOISI  1

/* status */
#define CONT    0
#define MODIF   1
#define STOP    2
#define FIND    3

void premier_graphe(int siz, double **cout, short **casemark, int *ligmark, int *colmark)
{
    int     i, j, flag;
    double      mem;

    /* un zero par ligne */
    for (i=0; i<siz; i++) {
        flag = OUI;
        for (j=0; j<siz; j++)
            if (casemark[i][j] != INTERDIT) {
                if (flag) {
                    mem = cout[i][j];
                    flag = NON;
                }
                else if (cout[i][j] > mem) {
                    mem = cout[i][j];
                }
            }
        for (j=0; j<siz; j++)
            cout[i][j] -= mem;
    }
    /* un zero par colonne */
    for (j=0; j<siz; j++) {
        flag = OUI;
        for (i=0; i<siz; i++)
            if (casemark[i][j] != INTERDIT) {
                if (flag) {
                    mem = cout[i][j];
                    flag = NON;
                }
                else if (cout[i][j] > mem) {
                    mem = cout[i][j];
                }
            }
        for (i=0; i<siz; i++)
            cout[i][j] -= mem;
    }
    /* mise a jour marque case */
    for (i=0; i<siz; i++) {
        for (j=0; j<siz; j++)
            if (casemark[i][j] == IMPOSSIBLE && cout[i][j] == 0.0)
                casemark[i][j] = POSSIBLE;
        ligmark[i] = colmark[i] = NoUN;
    }
    /* premier couplage heuristique */
    for (i=0; i<siz; i++) {
        for (j=0; j<siz; j++) {
            if (casemark[i][j] == POSSIBLE && colmark[j] == NoUN) {
                colmark[j] = ligmark[i] = UN;
                casemark[i][j] = CHOISI;
                break;
            }
        }
    }
}


int ajout_admissible(int size, double **cout, short **casemark, int *ligmark, int *colmark)
{
    int     i, j, flag;
    double      mem;

    flag = OUI;
    for (i=0; i<size; i++) {
        if (ligmark[i] != UN) {
            /* ligne marquee
             * NoUN ou marque */
            for (j=0; j<size; j++) {
                if (casemark[i][j] != INTERDIT && colmark[j] < 0) {
                    /* colone non marquee
                     * UN ou NoUN */
                    if (flag) {
                        mem = cout[i][j];
                        flag = NON;
                    }
                    else {
                        if (cout[i][j] > mem) {
                            mem = cout[i][j];
                        }
                    }
                }
            }
        }
    }
    if (flag)
        return 0;
    /* mise a jour des couts */
    for (i=0; i<size; i++)
        for (j=0; j<size; j++)
            if (casemark[i][j] != INTERDIT) {
                if (ligmark[i] != UN) {
                    if (colmark[j] < 0)
                        cout[i][j] -= mem;
                }
                else {
                    if (colmark[j] >= 0)
                        cout[i][j] += mem;
                }
            }
    /* mise a jour marque case */
    for (i=0; i<size; i++) {
        if (ligmark[i] >= 0)
            ligmark[i] = UN;
        if (colmark[i] >= 0)
            colmark[i] = UN;
    }
    for (i=0; i<size; i++) {
        for (j=0; j<size; j++)
            if (casemark[i][j] == IMPOSSIBLE && cout[i][j] == 0.0) {
                if (ligmark[i] == NoUN && colmark[j] == NoUN ) {
                    casemark[i][j] = CHOISI;
                    ligmark[i] = colmark[j] = UN;
                }
                else
                    casemark[i][j] = POSSIBLE;
            }
            else if (casemark[i][j] == POSSIBLE && cout[i][j] < 0.0) {
                casemark[i][j] = IMPOSSIBLE;
            }
    }
    return 1;
}


int couplage_maximum(int size, short **cell, int *lig, int *col, int *ligp, int *colp)
{
    int     i, ip, j, jp, flag, status, fil;
    
    status = CONT;
    do {
        /* premier marquage a partir des lignes sans couplage */
        ip=jp=0;
        flag = OUI;
        for (i=0; i<size; i++) {
            if (lig[i] == NoUN) {
                flag = NON;
                for (j=0; j<size; j++) {
                    if (col[j] == UN && cell[i][j] == POSSIBLE) {
                        col[j] = i;
                        colp[jp] = j;
                        jp++;
                    }
                }
            }
        }
        if (flag) {
            /* couplage complet */
            status = FIND;
        }
        else {
            /* couplage partiel */
            do {
                /* recherche choisi, marquage ligne */
                while (jp) {
                    j = colp[--jp];
                    for (i=0; i<size; i++)
                        if (lig[i] == UN && cell[i][j] == CHOISI) {
                            lig[i] = j;
                            ligp[ip++] = i;
                            break; /* qu'un seul un */
                        }
                    
                }
                /* recherche possible , marquage colonne */
                while (ip) {
                    i = ligp[--ip];
                    for (j=0; j<size; j++)
                        if (cell[i][j] == POSSIBLE) {
                            if (col[j] == NoUN) {
                                status = MODIF;
                                col[j] = i;
                                fil = j;
                                break;
                            }
                            else if (col[j] == UN) {
                                col[j] = i;
                                colp[jp++] = j;
                            }
                        }
                    if (status == MODIF) break;
                }
                if (status != MODIF)
                    status = (jp)? CONT : STOP ;
            } while (status == CONT);
            
            if (status == MODIF) {
                /* amelioration du couplage */
                flag = NON;
                j = fil;
                do {
                    if (flag) {
                        j = lig[i];
                        cell[i][j] = POSSIBLE ;
                        flag = NON;
                    }
                    else {
                        i = col[j];
                        cell[i][j] = CHOISI ;
                        flag = OUI;
                    }
                } while (!flag || lig[i] != NoUN);
                /* remise a jour des mark */
                for (i=0; i<size; i++)
                    col[i] = lig[i] = NoUN;
                for (i=0; i<size; i++)
                    for (j=0; j<size; j++)
                        if (cell[i][j] == CHOISI) {
                            lig[i] = col[j] = UN;
                            break;
                        }
                status = CONT;
            }
        }
    } while (status == CONT);

    return (status == FIND)? 0 : 1 ;
}

int hongrois(int siz, double **cout, short **casemark, int *perm)
{
    int     i, j, flag;

#ifdef WITHALLOC
    int     *colmark, *ligmark, *colpile, *ligpile;
    colmark = calloc(siz, sizeof(int));
    ligmark = calloc(siz, sizeof(int));
    colpile = calloc(siz, sizeof(int));
    ligpile = calloc(siz, sizeof(int));
#else
    int colmark[siz], ligmark[siz], colpile[siz], ligpile[siz];
#endif

    premier_graphe(siz, cout, casemark, ligmark, colmark);

    flag = 1; // une seule passe
    while (couplage_maximum(siz, casemark, ligmark, colmark, ligpile, colpile)
            && (flag = ajout_admissible(siz, cout, casemark, ligmark, colmark))
          );
    
    /* permutation */
    if (flag) {
        for (i=0; i<siz; i++)
            for (j=0; j<siz; j++)
                if (casemark[i][j] == CHOISI) {
                    perm[j] = i;
                    break;
                }
    }

#ifdef WITHALLOC
    free(colmark);
    free(ligmark);
    free(colpile);
    free(ligpile);
#endif
    return flag;
    
}

