/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2023 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_ibea.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-01-28
 
 \brief  indicator based evolutionary algorithm
*/

#include <stdlib.h>
#include <math.h>

#include "vzrsort.h"
#include "vzr_ibea.h"

#include "vzr_macro.h"
#include "vzr_alloc.h"

#include "vzritem.h"
#include "vzrpop_api.h"
#include "vzr_heap.h"

//#define TRACE
#ifdef TRACE
#include <stdio.h>
#endif

//#define IBEA
#define K_IBEA 0.02

/**** utilitaires ****/

/* calcul de l'indicateur
 * indicateur epsilon +
 ****/
#define MINI(x) if (mini_m || tmp < mini) {mini_m=0; mini = tmp;}
#define MAXI(x) if (maxi_m || tmp > maxi) {maxi_m=0; maxi = tmp;}
#define MINIMAX(x) if (mini_m || tmp < mini) {mini_m=0; mini = tmp;}\
if (maxi_m || tmp > maxi) {maxi_m=0; maxi = tmp;}

void ibea_indicator(struct ibea *self, struct vzritem *i, struct vzritem *j
        , struct vzrpop *pop, double *ind_ij, double *ind_ji)
{
    int k, kk;
    int maxi_m, mini_m, viab_i, viab_j;
    double mini, maxi, tmp;
    
    /* init */
    maxi_m = mini_m = 1;
    viab_j = viab_i = 1;
    
    /* viable */
    for EACH(k, pop->valid) {
        if (i->front[k] > 0) viab_i = 0;
        if (j->front[k] > 0) viab_j = 0;
    }
    /* boucle contraintes */
    if (!(viab_i && viab_j)) {
        for EACH(k, pop->valid) {
            if (i->front[k] < 0 && j->front[k] < 0) continue;
            tmp = (i->front[k] > 0) ? i->front[k] : 0 ;
            tmp -= (j->front[k] > 0) ? j->front[k] : 0 ;
            tmp /= self->it[k];
            MINIMAX(tmp);
        }
    }
    else k = pop->valid;
    /* boucle objectifs */
    if (viab_i && viab_j) {
        for EACH(kk, pop->cost) {
            tmp = (i->front[k] - j->front[k]) / self->it[k];
            MINIMAX(tmp);
            k++;
        }
    }
    else if ( viab_i && !viab_j) {
        for EACH(kk, pop->cost) {
            tmp = (i->front[k] - j->front[k]) / self->it[k];
            MINI(tmp);
            k++;
        }
    }
    else if ( viab_j && !viab_i) {
        for EACH(kk, pop->cost) {
            tmp = (i->front[k] - j->front[k]) / self->it[k];
            MAXI(tmp);
            k++;
        }
    }
    
    *ind_ij = maxi;
    *ind_ji = -mini;

}

/**** fonction du protocole ****/

/* viable 
 * lazy eval : all individual is viable
 ****/
static int ibea_viable(void *info, struct vzritem *item, struct vzrpop *pop)
{
    double ind_ij, ind_ji;
    int i;
    struct ibea *self = (struct ibea *) pop->info;
    
    item->rank = 0;
    item->isol = 0.;
    for EACH(i, pop->level) {
        ibea_indicator(self, &pop->item[pop->rg[i]], item, pop, &ind_ij, &ind_ji);
        if (ind_ij <= 0) item->rank++;
        item->isol -= exp( -ind_ij/ self->k);
    }
    return (pop->level == pop->stop && pop->item[pop->rg[pop->level-1]].rank != -1) ?
            item->rank <= pop->item[pop->rg[pop->level-1]].rank : 1;
}

/* rank population 
 * evaluate both moga rank and ibea indicator
 ****/
static void ibea_rank(void *info, int size, struct vzrpop *pop)
{
    int i, j, k, kk;
    double mini, maxi, ind_ij, ind_ji;
    struct ibea *self = (struct ibea *) pop->info;
    VZRTAB(double, part, size);
    
    for EACH(i, size) {
        pop->item[i].rank = 0;
        pop->item[i].isol = 0.;
    }
    /* normalisation */
    for EACH(k, pop->valid) {
        maxi = 0.;
        for EACH(i, size) {
            if (pop->item[i].front[k] > maxi) maxi = pop->item[i].front[k];
        }
        self->it[k] = (maxi == 0.) ? 1. : maxi ;
    }
    for EACH(kk, pop->cost) {
        mini = maxi = pop->item[0].front[k];
        for FROMEACH(i, 1, size) {
            if (pop->item[i].front[k] < mini ) mini = pop->item[i].front[k];
            else if (pop->item[i].front[k] > maxi ) maxi = pop->item[i].front[k];
        }
        self->it[k] = (mini == maxi) ? 1. : maxi - mini;
        k++;
    }
    
    /* compare */
    for FROMEACH(i, 1, size) {
        for EACH(j, i) {
            if (pop->item[j].rank == -1) continue;
            ibea_indicator(self, &pop->item[i], &pop->item[j], pop, &ind_ij, &ind_ji);
            /* individu egaux */
            if (ind_ij == 0. && ind_ji == 0.) {
                pop->item[i].rank = -1;
                break;
            }
            /* domination */
            if (ind_ji <= 0) pop->item[i].rank += 1;
            else if (ind_ij <= 0) pop->item[j].rank += 1;
            /* indicateur */
            pop->item[i].isol -= exp( -ind_ji / self->k);
            part[j] = exp( -ind_ij / self->k);
        }
        if (pop->item[i].rank != -1) {
            for EACH(j, i) pop->item[j].isol -= part[j];
        }
    }
    VZRTAB_OUT(part);
}

/* isol 
 * done in rank
 ****/
static void ibea_isol(void *info, int size, struct vzrpop *pop) {}


/* survive
 * classement sur l'indicateur ibea
 ****/
static void ibea_survive(void *info, int size, int soon, struct vzrpop *pop)
{
    struct ibea *self = (struct ibea *) pop->info;
    int i, imax;
    double ind_ij, ind_ji;
    int icur = size - soon - 1;
    
    while (icur >= pop->stop) {
        /* recherche max */
        imax = icur;
        for EACH(i, icur) {
            if (pop->item[pop->rg[i]].isol < pop->item[pop->rg[imax]].isol) imax = i;
        }
        /* switch */
        i = pop->rg[icur];
        pop->rg[icur] = pop->rg[imax];
        pop->rg[imax] = i;
        /* repercution de la supression sur l'indicteur */
        for EACH(i, icur) {
            ibea_indicator(self, &pop->item[pop->rg[icur]], &pop->item[pop->rg[i]], pop
                    , &ind_ij, &ind_ji);
            pop->item[pop->rg[i]].isol += exp(- ind_ij / self->k);
        }
        icur--;
    }         
             
    /* last classement sur l'indicateur ibea (utile?) */
    (void) vzr_rank_item(icur+1, 0, pop->rg, pop->item);
#ifdef TRACE
    for EACH(i, size) printf("ind%d  %d, rank %d, ibea %g\n", i, pop->rg[i]
            ,pop->item[pop->rg[i]].rank, pop->item[pop->rg[i]].isol);
    printf("\n\n");
#endif
}

/* build dealloc */
static void ibea_build(void *info, struct vzrpop *pop)
{
    struct ibea *dat;
    
    dat = (struct ibea *) malloc(sizeof(struct ibea));
    dat->k = *((double *) info);
    NEWvecteur(double, dat->it, pop->valid + pop->cost);
    
    pop->genre = Ibea;
    pop->info = dat;
}

static void ibea_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct ibea *dat = (struct ibea *) pop->info;
    
    if (dat->it) DELvecteur(dat->it);
    free(pop->info);
    free(self->data);
    free(self);
}

/** constructeur **/
struct vzrsort *create_ibea_full(double k)
{
    double *dat;
    struct vzrsort *ret;
    
    dat = (double *) malloc(sizeof(double));
    *dat = k;
    
    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = dat;
    ret->build = &ibea_build;
    ret->valid = &ibea_viable;
    ret->rank = &ibea_rank;
    ret->isol = &ibea_isol;
    ret->survive = &ibea_survive;
    ret->dealloc = &ibea_dealloc;
    return ret;
} 

struct vzrsort *create_ibea2(void)
{
    return create_ibea_full(K_IBEA);
}

struct vzrsort *create_ibea(void)
{
    return create_ibea_full(K_IBEA);
}
