/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_survie.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2013-08-26
 
 \brief  survie de la population
*/


#include <stdio.h>

#include "vizir/vzr_survie.h"

#include "vizir/vzr_macro.h"

#include "vizir/vzr_heap.h"
#include "vizir/vzr_pair.h"
#include "vizir/vzr_aleas.h"

//#define TRACE
//#define DEBUG

void valid_reorder(int size, int *idx)
{
    int i, j;
    
    for EACH(i, size) {
        for EACH(j, i) {
            if (idx[i] == idx[j]) printf("%d et %d = %d\n", i , j, idx[i]);
        }
    }
}

/* survie classique */
void pareto_survive(void *info, int size, int soon, struct vzrpop *self)
{
    /* classement par rang puis isolement*/
    (void) vzr_rank_item(size-soon, 0, self->rg, self->item);
}

/* survie pair version 1 */
void opair_survive(void *info, int size, int soon, struct vzrpop *self)
{
    int i, elit, remov;
    int siz = size-soon;
    //int siz = size/2;
    //int simil[size];
    struct local *loc = (struct local *) info;
	VZRTAB_IN(int, simil);

    int proxi = self->step*loc->ratio/100 - soon;
    if (proxi < 0) proxi = 0;
    
#ifdef DEBUG
    printf("START\n");
    valid_reorder(siz, self->rg);
#endif
    /* apairage des points */
    pair_affectation((struct voisin *) self->info, loc, siz, self) ;    
    /* classement par similarités en V */
    VZRTAB_NEW(int, simil, loc->done);
    pair_vorder(loc, siz, self, simil);
#ifdef DEBUG
    printf(" PAIR\n");
    valid_reorder(siz, self->rg);
#endif
    /* elite */
    elit = 0;
    for EACH(i, siz) if (self->item[self->rg[i]].rank == 0) elit++;
    /* premiere elimination, par similarités */
    if (self->cycle == loc->seuil) loc->cut = loc->fit[simil[proxi]].cost;
    //if (self->cycle > loc->seuil) loc->cut = loc->dist[simil[siz*8/10]] / 16.;
    remov = 0;
    while (loc->fit[simil[remov]].cost == 0. ||
            (loc->fit[simil[remov]].cost <= loc->cut && remov < proxi )
            )
        remov++;
    if (self->cycle > loc->seuil) {
        if (loc->cut < loc->fit[simil[remov]].cost) {
            // indice go down
            loc->cut = loc->fit[simil[remov]].cost;
        }
        else {
            // indice go up
            loc->cut = (loc->cut + loc->fit[simil[remov]].cost) / 2;
        }
    }
    if (siz - remov < elit) remov = siz - elit;
    /* seconde elimination, rank pareto puis simil */
    vzr_rank_item(size-remov-soon, 0, self->rg, self->item);
#ifdef DEBUG
    printf(" RANK\n");
    valid_reorder(siz, self->rg);
#endif
	
#ifdef TRACE2
    printf("RM %d\n", remov);
    printf("SEUIL %g  %g\n", loc->fit[simil[remov]].cost, loc->cut);
#endif
	VZRTAB_OUT(simil);
}

void pair_survive(void *info, int size, int soon, struct vzrpop *self)
{
    int imax, imin, itemp, tag, retag;
    struct local *loc = (struct local *) info;
    VZRTAB_IN(int, simil);

    /*** phase 1 : élimination par fitness ***/
    /* premier classement par rang puis isolement*/
    (void) vzr_rank_item(size-soon, 0, self->rg, self->item);

    /* recherche seuil fitness/isol */
    imin = size - self->step ;
    if (self->item[self->rg[size-self->step]].rank == 0) {
        /* converge, 
         * pas d'élimination systématique
         * seuil minimal
         */ 
        imax = (soon >= self->step) ? imin+1 : size - soon;
        retag = tag = 1;
    }
    else {
        /* non convergé 
         * elimination systematique de la moitié
         * seuil quelconque
         */
        imax = (soon >= self->step/2) ? imin+1 : size - soon - self->step/2 ;
        retag = tag = 0;
    } 
    if (self->item[self->rg[imin]].rank != self->item[self->rg[imax-1]].rank) {
        // search a rank step
        while (imax != imin+1) {
            itemp = (imax+imin)/2;
            if (self->item[self->rg[itemp]].rank == self->item[self->rg[imin]].rank) {
                imin = itemp;
                tag = 1;
            }
            else if (itemp+1 == imax) imax = itemp;
            else if (self->item[self->rg[itemp]].rank == self->item[self->rg[imax-1]].rank) {
                imax = itemp+1;
                tag = retag;
            }
            else if (tag) {
                imax = itemp+1;
                tag = retag;
            }
            else {
                imin = itemp;
                tag = 1;
            }
        }
#ifdef TRACE
        printf("step %d to %d, index %d %d\n", self->item[self->rg[imin]].rank, 
                self->item[self->rg[imax]].rank, imin, imax);
#endif
    }
#ifdef TRACE    
    printf("pair dupli %d simil %d / %d\n", soon, imax-self->level, self->step); 
#endif

    /*** phase 2 : élimination par similarité ***/
    /* apairage des points */
    pair_affectation((struct voisin *) self->info, loc, imax, self) ;    
    /* classement par similarités en V */
    VZRTAB_NEW(int, simil, loc->done);
    pair_vorder(loc, imax, self, simil);
    
    /*** phase 3 : last sort ***/
    (void) vzr_rank_item(size - self->step, 0, self->rg, self->item);
#ifdef TRACE
    if (imax == self->level + self->step) {
        for EACH(itemp, self->level) {
            if (itemp%20==19) printf("%d\n", self->item[self->rg[itemp]].rank);
            else printf("%d  ", self->item[self->rg[itemp]].rank);
        }
    }
#endif
    /*** finalize : maj voisin ***/
    //vzrpop_reorder_voisin(self, size, idx, (struct voisin *) loc);
	VZRTAB_OUT(simil);
	

}
