/*  _______________________________________________________________________

 DAKOTA: Design Analysis Kit for Optimization and Terascale Applications
 Copyright (c) 2001, Sandia National Laboratories.
 This software is distributed under the GNU General Public License.
 For more information, see the README file in the top Dakota directory.
 _______________________________________________________________________ */

//
//      File:         LHSVecMatUtil.C
//
//      Description:  Vector and Matrix Utilities
// blb
// starting indices have been changed to 0!!!
//***********************************************************************
#include <iostream>
#include <cstdio>
#include <cmath>
#include "LHSVecMatUtil.H"
#include <string>
#include <cstring>

void nrerror(const char errortext[])
{
    printf("LHSVecMatUtil runtime error...\n");
    printf("%s \n", errortext);
    printf(".... now exiting.\n");
}

//***********************************************************************

double **matrix(long int nrl, long int nrh, long int ncl, long int nch)
{
#ifdef DEBUG
    cerr << "matrix()\n";
#endif
    void nrerror(const char text[]);
    long int i;
    double **m;

    m = new double *[nrh - nrl + 1];
    if (!m)
        nrerror("Allocation failure in matrix()");
    m -= nrl;

    for (i = nrl; i <= nrh; i++)
    {
        m[i] = new double[nch - ncl + 1];
        if (!m[i])
            nrerror("Allocation failure 2 in matrix()");
        m[i] -= ncl;
    }
    return m;
}

//***********************************************************************
// blb
//  overloaded matrix function that is more standard C++
//     note the starting index is 0,0

double **matrix(long int nrh, long int nch)
{
#ifdef DEBUG
    cerr << "matrix()\n";
#endif
    void nrerror(const char text[]);
    long int i;
    double **m;

    m = new double *[nrh];
    if (!m)
        nrerror("Allocation failure in matrix()");

    for (i = 0; i < nrh; i++)
    {
        m[i] = new double[nch];
        if (!m[i])
            nrerror("Allocation failure 2 in matrix()");
    }
    return m;
}

//***********************************************************************

double *_vector(long int nl, long int nh)
{
#ifdef DEBUG
    cerr << "vector()\n";
#endif
    void nrerror(const char text[]);
    double *v;

    v = new double[nh - nl + 1];
    if (!v)
        nrerror("Allocation failure in vector()");
    return v - nl;
}
//***********************************************************************
// blb
//  overloaded vector function to a more standard form
//    note: the starting index is 0

double *_vector(long int nh)
{
#ifdef DEBUG
    cerr << "vector()\n";
#endif
    void nrerror(const char text[]);
    double *v;

    v = new double[nh];
    if (!v)
        nrerror("Allocation failure in vector()");
    return v;
}

//***********************************************************************

long int *_ivector(long int nl, long int nh)
{
#ifdef DEBUG
    cerr << "ivector()\n";
#endif
    void nrerror(const char text[]);
    long int *v;

    v = new long int[nh - nl + 1];
    if (!v)
        nrerror("Allocation failure in ivector()");
    return v - nl;
}

//***********************************************************************
// blb
//  overloaded ivector function in more standard C++
//    note starting index is 0
long int *_ivector(long int nh)
{
#ifdef DEBUG
    cerr << "ivector()\n";
#endif
    void nrerror(const char text[]);
    long int *v;

    v = new long int[nh];
    if (!v)
        nrerror("Allocation failure in ivector()");
    return v;
}

//***********************************************************************

void Vsum(double *sum, double *vect1, double *vect2, long int N)
{
#ifdef DEBUG
    cerr << "Vsum()\n";
#endif
    long int i;

    for (i = 0; i < N; ++i)
        sum[i] = vect1[i] + vect2[i];
}

//***********************************************************************

void Vdiff(double *diff, double *vect1, double *vect2, long int N)
{
#ifdef DEBUG
    cerr << "Vdiff()\n";
#endif
    long int i;

    for (i = 0; i < N; ++i)
        diff[i] = vect1[i] - vect2[i];
}

//***********************************************************************

void MTxV(double *answer, double **mtrx, long int N, long int M, double *vect,
        long int L)
{
#ifdef DEBUG
    cerr << "MTxV()\n";
#endif
    void nrerror(const char text[]);
    long int i, j;

    if (L != M)
        nrerror("Cannot multiply a NxM matrix with Lx1 vector, M!=L");

    for (i = 0; i < N; ++i)
        for (j = 0; j < M; ++j)
            answer[i] += vect[j] * mtrx[j][i];
}

//***********************************************************************

double VTxV(double *vect1, double *vect2, long int L)
{
#ifdef DEBUG
    cerr << "VTxV()\n";
#endif
    long int i;
    double answer;

    answer = 0.0;
    for (i = 0; i < L; ++i)
        answer += vect1[i] * vect2[i];
    return answer;
}

//***********************************************************************

void invertmatrix(double **thematrix, long int n)
{
#ifdef DEBUG
    cerr << "invertmatrix()\n";
#endif

// blb: shouldn't need to show function declaration
//  void ludcmp(double **M, long int a, long int *ind, double *d);
//  void lubksb(double **M, long int a, long int *ind, double *c);
//  void free_vector(double *v, long int l, long int h);
//  long int *ivector(int m, long int n);
//  double *vector(int m, long int n);
//  double **matrix(int nl, long int nh, long int ml, long int mh);
    long int i, j;
    long int *indx;
    double *col, **theinverse, temp;

    /*  Set up the local matrix and vectors */
    col = _vector(n);
    indx = _ivector(n);
    theinverse = matrix(n, n);

    /*  Compute the inverse of the matrix using ludecomp.o  */
    ludcmp(thematrix, n, indx, &temp);
    for (j = 0; j < n; j++)
    {
        for (i = 0; i < n; i++)
            col[i] = 0.0;
        col[j] = 1.0;
        lubksb(thematrix, n, indx, col);
        for (i = 0; i < n; i++)
            theinverse[i][j] = col[i];
    }

    /*  Put the inverse back into the original matrix  */
    for (i = 0; i < n; ++i)
        for (j = 0; j < n; ++j)
            thematrix[i][j] = theinverse[i][j];
    delete[] col;
//  free_vector(col,1,n);
    /*  free_ivector(indx,1,n);*/
}

/*********************************************
 *  This is the Numerical Recipes routine
 *  for matrix LU decomposition
 *********************************************/

#define TINY 1.0e-20;

//********************************************************************

void ludcmp(double **a, long int n, long int *indx, double *d)
{
#ifdef DEBUG
    cerr << "ludcmp()\n";
#endif
    long int i, j, k;
    long int imax = 0;
    double big, dum, sum, temp;
    double *vv;
    void free_vector(double *v, long int l, long int h);

    vv = _vector(n);
    *d = 1.0;
    for (i = 0; i < n; i++)
    {
        big = 0.0;
        for (j = 0; j < n; j++)
            if ((temp = fabs(a[i][j])) > big)
                big = temp;
        vv[i] = 1.0 / big;
    }
    for (j = 0; j < n; j++)
    {
        for (i = 0; i < j; i++)
        {
            sum = a[i][j];
            for (k = 0; k < i; k++)
                sum -= a[i][k] * a[k][j];
            a[i][j] = sum;
        }
        big = 0.0;
        for (i = j; i < n; i++)
        {
            sum = a[i][j];
            for (k = 0; k < j; k++)
                sum -= a[i][k] * a[k][j];
            a[i][j] = sum;
            if ((dum = vv[i] * fabs(sum)) >= big)
            {
                big = dum;
                imax = i;
            }
        }
        if (j != imax)
        {
            for (k = 0; k < n; k++)
            {
                dum = a[imax][k];
                a[imax][k] = a[j][k];
                a[j][k] = dum;
            }
            *d = -(*d);
            vv[imax] = vv[j];
        }
        indx[j] = imax;
        if (a[j][j] == 0.0)
        {
            a[j][j] = TINY
            ;
            cerr << "Warning! TINY used.\n";
        }
        if (j != n - 1)
        {
            dum = 1.0 / (a[j][j]);
            for (i = j + 1; i < n; i++)
                a[i][j] *= dum;
        }
    }
    delete[] vv;
    //        free_vector(vv,1,n);
}
#undef TINY

//********************************************************************

void lubksb(double **a, long int n, long int *indx, double b[])
{
#ifdef DEBUG
    cerr << "lubksb()\n";
#endif
    long int i, ii = 0, ip, j;
    double sum;

    for (i = 0; i < n; i++)
    {
        ip = indx[i];
        sum = b[ip];
        b[ip] = b[i];
        if (ii)
            for (j = ii; j <= i - 1; j++)
                sum -= a[i][j] * b[j];
        else if (sum)
            ii = i;
        b[i] = sum;
    }
    for (i = n - 1; i >= 0; i--)
    {
        sum = b[i];
        for (j = i + 1; j < n; j++)
            sum -= a[i][j] * b[j];
        b[i] = sum / a[i][i];
    }
}

//********************************************************************

void free_vector(double *v)
{
//        Tcl_Free((char*) (v+nl));
    delete[] v;
    return;
}

void free_matrix(double **mat, long int n)
{
    long int i;
    for (i = 0; i < n; i++)
    {
        delete[] mat[i];
        mat[i] = NULL;
    }

    delete[] mat;
    mat = NULL;
    return;
}

//********************************************************************
//   Functions from Numerical Recipes in C
//********************************************************************

int choldc(double **a, long int n, double p[])
{
    //        void nrerror(const char error_text[]);
    long int i, j, k;
    double sum;
    //cerr << "LatinHypercube::choldc()\n" << "n " << n << endl;;

    for (i = 0; i < n; i++)
    {
        for (j = i; j < n; j++)
        {
            for (sum = a[i][j], k = i - 1; k >= 0; k--)
            {
                sum -= a[i][k] * a[j][k];
            }
            if (i == j)
            {
                if (sum <= 0.0)
                {
                    //nrerror("choldc failed");
                    return 0;
                }
                p[i] = sqrt(sum);
                // a[i][j] = p[i];
            }
            else
            {
                a[j][i] = sum / p[i];
            }
        }
    }

    return 1;
}
/* (C) Copr. 1986-92 Numerical Recipes Software ZI5"=+3. */

//********************************************************************
double gammln(double xx)
{
    double x, y, tmp, ser;
    static double cof[6] =
    { 76.18009172947146, -86.50532032941677, 24.01409824083091,
            -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5 };
    long int j;

    y = x = xx;
    tmp = x + 5.5;
    tmp -= (x + 0.5) * log(tmp);
    ser = 1.000000000190015;
    for (j = 0; j <= 5; j++)
        ser += cof[j] / ++y;
    return -tmp + log(2.5066282746310005 * ser / x);
}
// (C) Copr. 1986-92 Numerical Recipes Software ZI5"=+3.

//********************************************************************
// rnum1 random number generator from lhs.for
//   included for testing purposes only
//   converted to c++ and unnecessary code removed
//   Bruce Bainbridge, Sept. 18, 1997

double rnum1(int *idum)
{
    const double P = 16777216.;
    const double GB = 1136868.;
    const double GS = 6328637.;

    static double rani, ranj;
    static int flag = 0;

    if (!flag)
    {
        flag = 1;
        if (*idum < 0)
            *idum = -(*idum);
        int r = (*idum);
        rani = (int) (r / P);
        ranj = r - rani * P;
    }
    double a = GS * ranj;
    double b = GB * ranj + GS * rani + (int) (a / P);
    ranj = a - (int) (a / P) * P;
    rani = b - (int) (b / P) * P;
    double rmcnp = (rani * P + ranj) / P / P;

    *idum += 1;

    return rmcnp;
}

//********************************************************************
// (C) Copr. 1986-92 Numerical Recipes Software ZI5"=+3.

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(int *idum)
{
    long int j;
    double k;
    static long int idum2 = 123456789;
    static double iy = 0;
    static double iv[NTAB];
    static int flag = 0;
    double temp;

// flag variable introduced to remove need for negative seed value
//  Bruce Bainbridge September 18, 1997

    if (*idum <= 0 || !flag)
    {
        flag = 1;
        if (*idum < 0)
            *idum = -(*idum);
        if (*idum < 1)
            *idum = 1;
        idum2 = (*idum);
        for (j = NTAB + 7; j >= 0; j--)
        {
            k = (*idum) / IQ1;
            *idum = (int) (IA1 * (*idum - k * IQ1) - k * IR1);
            if (*idum < 0)
                *idum += IM1;
            if (j < NTAB)
                iv[j] = *idum;
        }
        iy = iv[0];
    }
    k = (*idum) / IQ1;
    *idum = (int) (IA1 * (*idum - k * IQ1) - k * IR1);
    if (*idum < 0)
        *idum += IM1;
    k = idum2 / IQ2;
    idum2 = (int) (IA2 * (idum2 - k * IQ2) - k * IR2);
    if (idum2 < 0)
        idum2 += IM2;
    j = (int) (iy / NDIV);
    iy = iv[j] - idum2;
    iv[j] = *idum;
    if (iy < 1)
        iy += IMM1;
    if ((temp = AM * iy) > RNMX)
        return RNMX;
    else
        return temp;

    /* random number generator rnum1 from fortran version of lhs code for
     testing only

     const double P = 16777216.;
     const double GB = 1136868.;
     const double GS = 6328637.;

     static double rani, ranj;
     static int flag = 0;

     if (!flag) {
     flag = 1;
     if (*idum < 0) *idum = -(*idum);
     int r = (double)(*idum);
     rani = (int) (r/P);
     ranj = r - rani * P;
     }
     double a = GS * ranj;
     double b = GB * ranj + GS * rani + (int) (a/P);
     ranj = a - (int)(a/P) * P;
     rani = b - (int)(b/P) * P;
     double rmcnp = (rani * P + ranj) /P/P;

     *idum += 1;

     return rmcnp;
     */
}

#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

//********************************************************************
// (C) Copr. 1986-92 Numerical Recipes Software ZI5"=+3.

#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

double ran3(int *idum)
{
    static long int inext, inextp;
    static double ma[56];
    static long int iff = 0;
    long int mj, mk;
    long int i, ii, k;

    if (*idum < 0 || iff == 0)
    {
        iff = 1;
        mj = MSEED - (*idum < 0 ? -*idum : *idum);
        mj = mj % MBIG;
        ma[55] = mj;
        mk = 1;
        for (i = 1; i <= 54; i++)
        {
            ii = (21 * i) % 55;
            ma[ii] = mk;
            mk = mj - mk;
            if (mk < MZ)
                mk += MBIG;
            mj = (long int) ma[ii];
        }
        for (k = 1; k <= 4; k++)
            for (i = 1; i <= 55; i++)
            {
                ma[i] -= ma[1 + (i + 30) % 55];
                if (ma[i] < MZ)
                    ma[i] += MBIG;
            }
        inext = 0;
        inextp = 31;
        *idum = 1;
    }
    if (++inext == 56)
        inext = 1;
    if (++inextp == 56)
        inextp = 1;
    mj = (long int) (ma[inext] - ma[inextp]);
    if (mj < MZ)
        mj += MBIG;
    ma[inext] = mj;
    return mj * FAC;
}
#undef MBIG
#undef MSEED
#undef MZ
#undef FAC

//**************************************************************************

void quicksort(double *a, long int F, long int L)
{
    long int PivotIndex;

    if (F < L)
    {
        PivotIndex = partition(a, F, L);

        // sort region s1 and s2 
        quicksort(a, F, PivotIndex - 1);
        quicksort(a, PivotIndex + 1, L);
    }
}

//**************************************************************************

int partition(double *a, long int F, long int L)
{
    long int firstunknown, lasts1;
    double pivot;

    pivot = a[F];
    lasts1 = F;
    firstunknown = F + 1;

    while (firstunknown <= L)
    {
        if (a[firstunknown] < pivot)
        {
            Swap(&a[firstunknown], &a[lasts1 + 1]);
            lasts1 = lasts1 + 1;
            firstunknown = firstunknown + 1;
        }
        else
            firstunknown = firstunknown + 1;
    }
    Swap(&a[F], &a[lasts1]);
    return lasts1;
}

//**************************************************************************

void Swap(double *a, double *b)
{
    double temp;

    temp = a[0];
    a[0] = b[0];
    b[0] = temp;
}

//**************************************************************************
