//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TBasicSampling.cxx
 *  \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 *  \date   $date$
 *  \brief Implementation of the class URANIE::Sampler::TBasicSampling.
 *
 *  Created the : Fri Oct  1 13:37:23 2004
 *
 */
#include <ctime>
#include <iostream>
#include <fstream>

////// ROOT
#include "TDecompChol.h"
#include "TDecompSVD.h"
#include "TMath.h"
#include "TDatime.h"
#include "TROOT.h"
#include "TRandom3.h"
#include "TObjString.h"

////// URANIE
#include "TBasicSampling.h"
#include "TAttributeFormula.h"
#include "TStochAttribut.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "LatinHypercube.H"

ClassImp (URANIE::Sampler::TBasicSampling)

////////////////////////////////////////////////////////////
//  TOOLS: some useful functions   
////////////////////////////////////////////////////////////

// Function allowing to remove a row and/or a column from a TMatrixD
void removeRowCol(TMatrixD &m, Int_t rowNumber, Int_t colNumber)
{
    Int_t skipRow, skipCol, newNrows, newNcols;

    if (rowNumber >= 0)
    {
        newNrows = m.GetNrows() - 1;
        skipRow = rowNumber;
    }
    else
    {
        newNrows = m.GetNrows();
        skipRow = m.GetNrows();
    }

    if (colNumber >= 0)
    {
        newNcols = m.GetNcols() - 1;
        skipCol = colNumber;
    }
    else
    {
        newNcols = m.GetNcols();
        skipCol = m.GetNcols();
    }

    TMatrixD m2(newNrows, newNcols);
    for (Int_t i = 0; i < m.GetNrows(); i++)
    {
        Int_t i2;
        if (i == skipRow)
            continue;
        else if (i < skipRow)
            i2 = i;
        else
            i2 = i - 1;

        for (Int_t j = 0; j < m.GetNcols(); j++)
        {
            Int_t j2;
            if (j == skipCol)
                continue;
            else if (j < skipCol)
                j2 = j;
            else
                j2 = j - 1;
            m2(i2, j2) = m(i, j);
        }
    }
    m.ResizeTo(m2);
    m = m2;
    //return m2;            
}

// Functions using the Fisher-Yates algorithm to shuffle the contents of an array.
void shuffleInteger(Int_t arraySize, Int_t* array)
{
    for (Int_t i = arraySize - 1; i > 0; i--)
    {
        Int_t indtmp = (Int_t) gRandom->Integer(i + 1);
        Int_t valtmp = array[indtmp];
        array[indtmp] = array[i];
        array[i] = valtmp;
    }
}

void shuffleDouble(Int_t arraySize, Double_t* array)
{
    for (Int_t i = arraySize - 1; i > 0; i--)
    {
        Int_t indtmp = (Int_t) gRandom->Integer(i + 1);
        Double_t valtmp = array[indtmp];
        array[indtmp] = array[i];
        array[i] = valtmp;
    }
}

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TBasicSampling::TBasicSampling(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t samplingSize) :
        TSamplerStochastic(tds, option, samplingSize), _fLhsInput(NULL), _simTolerance(1e-6)
{
    if (_blog)
        cout << "constructor TBasicSampling using TDataServer["
                << tds->GetName() << "]" << endl;

    TString opt = option;
    opt.ToLower();

    // Set the sampling method
    if (opt.Contains("srs"))
    {
        setMethodName("SRS");
        _ntype = kSRS;
        SetNameTitle(Form("Sampling_srs_%d", samplingSize),
                Form("Uranie %d sample with sampling method SRS",
                        samplingSize));

    }
    else
    {
        setMethodName("LHS");
        _ntype = kLHS;
        SetNameTitle(Form("Sampling_lhs_%d", samplingSize),
                Form("Uranie %d sample with sampling method LHS",
                        samplingSize));
    }

    _listOfSimilar = new TList();
    _listOfFormulas = new TList();
}

TBasicSampling::~TBasicSampling()
{
    if (_blog)
        cout << "************ Destroying URANIE::Sampler::TBasicSampling"
                << endl;
    if (_fLhsInput)
    {
        if (_blog)
            cout << " ** Delete _fLhsInput " << endl;
        delete _fLhsInput;
        _fLhsInput = NULL;
    }
    delete _listOfSimilar;
    _listOfFormulas->Delete();
    delete _listOfFormulas;
}

/************************************/
/*           Correlations           */
/************************************/
void TBasicSampling::setCorrelationMatrix(TMatrixD corrMat)
{
    try
    {
        if ((corrMat.GetNcols() != _nX)
                || (corrMat.GetNrows() != corrMat.GetNcols()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TBasicSampling::generateCorrSample - ERROR: invalid correlation matrix: it must be a symetrical square matrix of size %d x ",
                            _nX));

        userCorrMat.ResizeTo(corrMat);
        userCorrMat = corrMat;

        for(int icol=0; icol<corrMat.GetNcols(); icol++) 
	{
	    userCorrMat(icol,icol)=1.0;
	    for(int ilig=0; ilig<corrMat.GetNrows(); ilig++)
	    {
                if( fabs(userCorrMat(ilig,icol) - userCorrMat(icol,ilig)) > 1e-14 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TBasicSampling::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix.\n\
This is not the case as M(%d,%d) != M(%d,%d). Their difference is %g",ilig,icol,icol,ilig, userCorrMat(ilig,icol) - userCorrMat(icol,ilig)));
	    }
	}
		
        Double_t max = TMath::MaxElement(userCorrMat.GetNoElements(),
                userCorrMat.GetMatrixArray());
        Double_t min = TMath::MinElement(userCorrMat.GetNoElements(),
                userCorrMat.GetMatrixArray());
        if (max > 1.0 || min < -1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TBasicSampling::generateCorrSample - ERROR: invalid correlation matrix: correlation values must belong to the interval [-1.0, 1.0]");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TBasicSampling::setUserCorrelation(TString attName1, TString attName2,
        Double_t corrValue)
{
    try
    {
        if (corrValue > 1.0 || corrValue < -1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TBasicSampling::setUserCorrelation - ERROR: correlation values must belong to the interval [-1.0, 1.0]. The given value was: %f",
                            corrValue));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (userCorrMat.GetNcols() != _nX)
    {
        userCorrMat.ResizeTo(_nX, _nX);
        userCorrMat.UnitMatrix();
    }
    Int_t ind1 = _tds->getAttributeIndex(attName1) - 1;
    Int_t ind2 = _tds->getAttributeIndex(attName2) - 1;
    userCorrMat[ind1][ind2] = corrValue;
    userCorrMat[ind2][ind1] = corrValue;
}

void TBasicSampling::findSimilar()
{
    //-- ngi: I'm not at all satisfied of the way I've done this 
    //        function. It is a bit of a hack, actually, and might be 
    //        quite slow when the number of attributes is large.
    //        I'm sure it can be optimised, but at the time of the 
    //        writting (2013.05.15), I couldn't see how. 
    //----
    URANIE::DataServer::TStochasticAttribute *attA, *attB;
    for (Int_t i = 0; i < _nX; i++)
    {
        attA =
                (URANIE::DataServer::TStochasticAttribute *) (_lstOfAttributesToSample->At(
                        i));
        // Check if the first attribute was already found to be 
        // similar to another one.
        if (_listOfSimilar->IndexOf(attA) == -1)
        {
            for (Int_t j = i + 1; j < _nX; j++)
            {
                attB =
                        (URANIE::DataServer::TStochasticAttribute *) (_lstOfAttributesToSample->At(
                                j));
                // Check if the second attribute was already found to be 
                // similar to another one.
                if (_listOfSimilar->IndexOf(attB) == -1)
                {
                    // Check if attB is similar to attA
                    if (1.0 - TMath::Abs(userCorrMat(i, j)) < _simTolerance)
                    {
                        // Add attB to the list of totally dependent attributes 
                        _listOfSimilar->Add(attB);
                        // Compute the parameters of the linear dependency
                        Double_t alpha = TMath::Sign(1.0, userCorrMat(i, j))
                                * attB->getTheoreticalStdDev()
                                / attA->getTheoreticalStdDev();
                        Double_t beta = attB->getTheoreticalMean()
                                - alpha * attA->getTheoreticalMean();
                        // Save the formula
                        TObjString *formula = new TObjString(
                                Form("%.8e * %s + %.8e", alpha, attA->GetName(),
                                        beta));
                        _listOfFormulas->Add(formula);
                    }
                }
            }
        }
    }

    // Clean the list of attribute to sample            
    TMatrixD tmpCorrMat(userCorrMat);
    URANIE::DataServer::TAttribute* att;
    TIter next(_listOfSimilar);
    while ((att = (URANIE::DataServer::TAttribute*) next()) )
    {
        Int_t attIndex = _lstOfAttributesToSample->IndexOf(att);
        removeRowCol(tmpCorrMat, attIndex, attIndex);
        _lstOfAttributesToSample->Remove(att);
    }
    _nX = _lstOfAttributesToSample->GetEntries();

    // Copy the correlation matrix
    _corrMat.ResizeTo(tmpCorrMat);
    _corrMat = tmpCorrMat;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TBasicSampling::generateCorrSample(Option_t *option)
{
    findSimilar();

    bool buseSVD=false;
    bool buseSVDquiet=false;
    TString opt = option;
    opt.ToLower();
    
    if( opt.Contains("svd") )
        buseSVD=true;
    if( opt.Contains("quiet") )
        buseSVDquiet=true;

    parseOption(option);
    
    // Take into account the LHS or SRS cases
    Double_t dpas=-1, dtype=-1, dproba=0;

    switch (_ntype)
    {
    case kLHS:
    {
        dpas = 1.0 / (Double_t)(_nS);
        dtype = 1.0;
    }
        break;
    case kSRS:
    {
        dpas = 1.0;
        dtype = 0.0;
    }
        break;
    default:
	break;
    }

    TMatrixD matUFromC(_corrMat.GetNcols(),_corrMat.GetNcols());
    if(buseSVD)
    {
        if (!buseSVDquiet)
        {
            try{
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                          "TBasicSampling::generateCorrSample: you've asked to use SVD instead of Cholesky to decompose the correlation matrix. ");
            }catch(URANIE::Exceptions::UInfoExceptions &ue)
            {
                ue.printMessage();
            }
        }
        TDecompSVD svdC(_corrMat);
        if (svdC.Decompose() == kFALSE)
        {
            if (_blog)
            {
                _corrMat.Print();
                // print the matrix U
                cout << " ** Print the matrix U" << endl;
                svdC.GetU().Print();
            }
            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TBasicSampling::generateCorrSample : the SVD decomposition of the user correlation matrix is not good");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }
        
        TMatrixD matDFromC(_corrMat.GetNcols(),_corrMat.GetNcols());
        const TVectorD vecDFromC = svdC.GetSig();
        matDFromC.Zero();
        for (int i=0; i<_nX; i++)
            matDFromC(i,i)=sqrt(vecDFromC[i]);

        TMatrixD Ut(TMatrixD::kTransposed, svdC.GetU());
        matUFromC = matDFromC * Ut;
        
        if (_blog)
        { // _blog
            cout << "******* Matrice  de correlation ******** " << endl;
            svdC.Print();
        }   
    }
    else
    {
        TDecompChol dchol(_corrMat);
        dchol.Decompose();
        if (dchol.Decompose() == kFALSE)
        {
            if (_blog)
            {
                _corrMat.Print();
                // print the matrix U
                cout << " ** Print the matrix U" << endl;
                dchol.GetU().Print();
            }
            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TBasicSampling::generateCorrSample: the cholesky decomposition of the user correlation matrix is not good.\n Maybe you can try with SVD decomposition.");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }
        matUFromC = dchol.GetU();
    }

    TMatrixD Y(_nS, _corrMat.GetNcols());
    TMatrixD Z(_nS, _corrMat.GetNcols());
    URANIE::DataServer::TNormalDistribution refNorm("ref", 0.0, 1.0);

    // Prepare the indices
    Int_t * nind = new Int_t[_nS];
    for (Int_t i = 0; i < _nS; i++)
        nind[i] = i;

    // Fill the Y matrix
    for (Int_t j = 0; j < Y.GetNcols(); j++)
    {
        shuffleInteger(_nS, nind);
        for (Int_t i = 0; i < _nS; i++)
        {
            dproba = dpas * gRandom->Rndm() + dtype * dpas * i;
            Y[nind[i]][j] = refNorm.getInvCDF(dproba);
        }
    }

    delete[] nind;

    // Compute the Z matrix
    Z.Mult(Y, matUFromC);
    
    // Fill the tuple
    Double_t * valPattern = new Double_t[Z.GetNcols() + 1];
    createTuple();

    if (_blog)
        cout << "Filling the TTree..." << endl;

    for (Int_t i = 0; i < _nS; i++)
    {
        valPattern[0] = (Double_t)(i + 1);
        for (Int_t j = 0; j < Z.GetNcols(); j++)
        {
            URANIE::DataServer::TStochasticAttribute* att =
                    (URANIE::DataServer::TStochasticAttribute*) _lstOfAttributesToSample->At(j);
            valPattern[j + 1] = att->getInvCDF(refNorm.getCDF(Z(i, j)));
        }
        _ntsample->Fill(valPattern);
    }
    delete[] valPattern;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    fillOtherAttributes();
}

void TBasicSampling::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TSamplerStochastic::init();

    parseOption(option);
    
    TMatrixD matdvalCR(_nS, _nX);

    // Take into account the LHS or SRS cases
    Double_t dpas=-1.0, dtype=-1.0, dproba;

    switch (_ntype)
    {
    case kLHS:
    {
        dpas = 1.0 / (Double_t)(_nS);
        dtype = 1.0;
    }
        break;
    case kSRS:
    {
        dpas = 1.0;
        dtype = 0.0;
    }
        break;
    default:
	break;
    }

    // Prepare the indices
    Int_t * nind = new Int_t[_nS];
    for (Int_t i = 0; i < _nS; i++)
        nind[i] = i;

    // Generate the sample
    Int_t iatt;
    URANIE::DataServer::TStochasticAttribute *curdist;

    // Boucle sur les variables stochastiques
    iatt = 0;
    TIter next(_lstOfAttributesToSample);
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        // Shuffle the indices (Fisher-Yates)
        if (_blog)
            cout << "Shuffling indices..." << endl;

        for (Int_t is = _nS - 1; is > 0; is--)
        {
            Int_t indtmp = (Int_t) gRandom->Integer(is + 1);
            Int_t valtmp = nind[indtmp];
            nind[indtmp] = nind[is];
            nind[is] = valtmp;
        }

        // Generate the values
        if (_blog)
            cout << "Generating the data..." << endl;

        for (Int_t is = 0; is < _nS; is++)
        {
            dproba = dpas * gRandom->Rndm() + dtype * dpas * is;
            matdvalCR(nind[is], iatt) = curdist->getInvCDF(dproba);

            if (_blog)
                cout << "** [" << iatt << "," << is << "] -- random value ["
                        << matdvalCR(nind[is], iatt) << "]" << endl;
        }
        iatt++;
    }
    delete[] nind;
    nind = NULL;

    // Fill the TTree
    Double_t * valPattern = new Double_t[_nX + 1];
    createTuple();

    if (_blog)
        cout << "Filling the TTree..." << endl;

    for (Int_t is = 0; is < _nS; is++)
    {
        valPattern[0] = (Double_t)(is + 1);
        for (Int_t ix = 0; ix < _nX; ix++)
            valPattern[ix + 1] = matdvalCR(is, ix);
        _ntsample->Fill(valPattern);
    }
    delete[] valPattern;
    valPattern = NULL;

    if (_blog)
        cout << "TTree filling done" << endl;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    fillOtherAttributes();
}

void TBasicSampling::fillOtherAttributes()
{
    // Fill the similar attributes (if any)
    Int_t nbSimilar = _listOfSimilar->GetEntries();
    if (nbSimilar > 0)
    {
        for (Int_t i = 0; i < nbSimilar; i++)
        {
            Bool_t update = (i == nbSimilar - 1);
            TObject *att = _listOfSimilar->At(i);
            TObjString *formula = (TObjString*) _listOfFormulas->At(i);
            _tds->fillAttribute(TString(att->GetName()), formula->GetString(), false);
        }
    }
    // Fill the attribute formulas
    TSampler::fillOtherAttributes();
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TBasicSampling::printLog(Option_t *option)
{
    cout << "** TBasicSampling::printLog  [" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nX   [" << _nX << "]" << endl;
    cout << "** _sMethod   [" << _sMethod << "]" << endl;

    if (_fLhsInput)
        _fLhsInput->printlog();

}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
