//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TBasicSampling.h
 *  \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 *  \date   $Date$
 *  \brief Interface of the class URANIE::Sampler::TBasicSampling.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */

/**
 * @class URANIE::Sampler::TBasicSampling
 *
 * @brief Description of the class TBasicSampling.

 This class defines a TSamplerStochastic object for fast random data generation.

 This is a very simple implementation of the random sampling. It will 
 generate a set of n values for each random variable defined in the given 
 TDataServer object.

 The main differences with the TSampling class are:

 1. No correlation can be imposed between the variables.
 2. The strict independence between the generated variables is not garanteed.
 3. There is no lower limit in the number of random samples that can be generated.
 4. The generation is much faster when dealing with a large number of attributes. 

 On any other aspects, the two classes have the same properties.

 In short, if you have a large number of independent random variables to sample 
 in a short time, you should use this class. Otherwise, use TSampling.

 */

#ifndef TBASICSAMPLING_H
#define TBASICSAMPLING_H

#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TNtuple
#include "TNtuple.h"
#endif
#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSamplerStochastic.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataServer.h"
#include "LHSInput.H"

namespace URANIE
{
namespace Sampler
{
class TBasicSampling: public TSamplerStochastic
{
    // Associations

    // Attributes
public:
    enum EType
    {
        kLHS, kSRS, kUnknown
    };
    LHSInput* _fLhsInput;
    EType _ntype;          ///< The type of sampling (SRS, LHS)
    TMatrixD userCorrMat;     ///< correlation matrix defined by the user
    TMatrixD _corrMat;        ///< correlation matrix used internally
    TList* _listOfSimilar; ///< list of attributes totally correlated with other attributes.
    TList* _listOfFormulas; ///< list of formulas linking two totally correlated attributes.
    Double_t _simTolerance; ///< precision to decide if a correlation is equal to 1.0 (or -1.0).
    Bool_t bApplyUserCorrelation; ///< if true, correlated data are produced.

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Creates a TBasicSampling object from a TDataServer, a 
    //! sampling method and the size of the sample to generate
    /*!
     The random variables to sample are the attributes of the TDataServer.
     These attributes must derive from the TStochasticAttribute class.
     The sampling method, given as an option, can be either "lhs" 
     (Latin Hypercube Sampling) or "srs" (Simple Random Sampling). If
     none of these is given, "lhs" is chosen.                
     
     \param tds        (TDataServer *) pointer to the TDataServer.
     \param option     (Option_t *) option string containing the sampling method (Default = "lhs").
     \param sampleSize (Int_t) number of values to generate for each random variable (Default = 1000).
     
     \exception UErrorExceptions if sampleSize <= 0
     */
    TBasicSampling(URANIE::DataServer::TDataServer *tds, Option_t *option =
            "lhs", Int_t sampleSize = 1000);

    //! Default destructor
    virtual ~TBasicSampling();
    //@}

    //---------------------------------------------
    /** @name Correlations
     **/
    //@{
    //! Set the precision of similarity                
    void setTolerance(Double_t tolerance)
    {
        _simTolerance = tolerance;
    }

    //! Set the correlation matrix
    void setCorrelationMatrix(TMatrixD corrMat);

    //! Set user correlation for two attributes
    void setUserCorrelation(TString attName1, TString attName2,
            Double_t corrValue);

    //! Locate totally correlated attributes
    void findSimilar();
    //@}

    //---------------------------------------------            
    /** @name Generate the sample
     **/
    //@{
    //! Generates correlated data following a user defined correlation matrix.
    /*!
     This function uses the "normal copula" method (cf. S. Wang. 
     "Aggregation of correlated risk portfolios: Models and algorithms". 
     Proceedings of the Casualty Actuarial Society, LXXXV:848-939, 1998)
     to create correlated data. This method is equivalent to the 
     Iman & Connover one, but is faster and easier to implement. 
     However, it does not preserve the "LHS" distribution, and 
     might perform worse than I&C on the reconstruction of the 
     theoretical correlation.
     
     In addition, this function automatically detects totally 
     correlated variables, and handle them correctly.
     */
    void generateCorrSample(Option_t *option="");

    //! Generates the sample 
    /*!
     Generates the random data and fills the TDataServer with it.                
     \param option (Option_t *) option string (Default = "")
     */
    void generateSample(Option_t *option = "");

    //! Fills the TAttributeFormula and totally correlated attributes 
    virtual void fillOtherAttributes();
    //@}                       

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{                
    //! Prints the log
    /*!
     \param option (Option_t *) option string (Default = "")
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TBasicSampling, ID_SAMPLER)
    //Definition d'un plan d'exp�rience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
