//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TConstrLHS.h
 *  \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 *  \date   $Date$
 *  \brief Interface of the class URANIE::Sampler::TConstrLHS.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */

/**
 * @class URANIE::Sampler::TConstrLHS
 *
 * @brief Description of the class TConstrLHS.

 This class is used to produce a constr lhs doe. It starts either with an empy tds or 
 an already filled one, with the hypothesis that it has been filled with an LHS doe 

 An simulated annealing algorithm is implemented here to deal with the permutation.
 This implementation has be done by Guillaume Damblin.
 */

#ifndef TCONSTRLHS_H
#define TCONSTRLHS_H

#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TNtuple
#include "TNtuple.h"
#endif
#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSamplerStochastic.h"
#include "TSampling.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Sampler
{
class TConstrLHS: public TSamplerStochastic
{
    
public:

    // Attributes
    /** @name internal variables
     **/
    bool                 _bAlreadyFilled;        ///< If the tree is already filled

    TMatrixD             _currentLHS;            ///< The lhs sample under modification
    TMatrixD             _currentPerm;           ///< a matrix used to do the permutation of the constraint matrices
    
    vector<TMatrixD>     _currentConst;          ///< vector of constraints matrix
    vector<vector<int> > _rowSol;                ///< vector of number of constraint respected by row (by constraint)
    vector<vector<int> > _colSol;                ///< vector of number of constraint respected by column (by constraint) 
    vector<vector<int> > _diag;                  ///< vector of boolean for the diagonal element of the constraint (by constraint)
    vector<map<int, vector<int> > >  _modified;  ///< vector of modification by permutation (by constraint)
    vector<int>          _paramValues;           ///< vector of parameter values used to compute the constraints and fill the constraint matrix
    vector<int>          _nbdiag;                ///< vector of number of diagonal constrainted validated

    TSampling           *_sampling;              ///< Pointer to a TSampling object if tds is empty;

    unsigned int        _ns;                     ///< Local counter for number of samples and variables
    unsigned int        _nx;                     ///< Local counter for number of samples and variables
    unsigned int        _nParam;                 ///< Local counter for number of parameters (inputs) used to compute the constraint
    
    unsigned int        _nConstr;                ///< Number of constraints used in thie class
    unsigned int        _iPerm;                  ///< Number of permutations performed so far
    void (*_fConstrain)(double *, double*);      ///< Parameter transformation function
    
    //@}

    // Operations
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Creates a TConstrLHS object from a TDataServer, and the number of events requested
    /*!
      
      \param tds        (TDataServer *) pointer to the TDataServer.
      \param sampleSize (Int_t) number of values to generate for each random variable.
     
      \exception UErrorExceptions if sampleSize <= 0
    */
    TConstrLHS(URANIE::DataServer::TDataServer *tds, int sampleSize);

    //! Default destructor
    virtual ~TConstrLHS();
    //@}


    //---------------------------------------------            
    /** @name Generate the sample
     **/
    
    //! Generates the sample 
    /*!
      Generates the random data and fills the TDataServer with it.                
      \param option (Option_t *) option string (Default = "")
    */
    void generateSample(Option_t *option = "");

    //! Set the random generator seed
    void setSeed(Int_t ind = 0);

    void addConstraint( void (*fConstrain)(double *, double*), int nConstr, int nbparam, int *param );

    //@}
    
private:
    
    //---------------------------------------------            
    /** @name Specific methods to initialise and fill the doe
     **/

    //! initialise the elements of inner variables used to perform the permutation
    void initElements();
    
    //! Fills the TAttributeFormula and totally correlated attributes 
    virtual void fillOtherAttributes();

    //! create the ntuple
    void createTuple();

    //@}


    //---------------------------------------------
    /** @name Compute the constraint matrix 
     **/
    //@{
      
    //! Compute the distance matrix
    /*!
      Compute the distance matrix

      \param Const (TMatrixD) the distance matrix of the lhs plan under study
      \param lhs (TMatrixD) the lhs plan under study
    */
    void getConstMatrix();

    //!  Print the matrix (for debuging purpose mainly)
    /*!
      Dump a constraint matrix. The specific formatting is done for matrices filled with 0 or 1

      \param dist (TMatrixD) the constraint matrix to dump
     */
    void printMatrix(TMatrixD &dist);
    
    //! find the permutation to be performed both for the lhs plan and all the inner variables used to  
    /*!
      The permutation is factorised in performPermutation below
    */
    void findNextPermutation();
    
    //! perform the permutation defined in findNextPermutation for the lhs plan and all the inner variables used to  
    /*!
      The permutation is performed
    */
    void performPermutation(int Rcon, int Rrow, int Rcol, int Rvar);
    
    //! A DFS based recursive function that returns true if a matching for vertex u is possible
    bool bpm(TMatrixD &lhs, int u, int seen[], int matchR[]);
    
    //! Returns the maximum number of matching from _nS to _nS
    int maxBPM(TMatrixD &lhs);

    
    
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{                
    //! Prints the log
    /*!
      \param option (Option_t *) option string (Default = "")
    */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TConstrLHS, ID_SAMPLER)
    //Definition d'un plan d'exp�rience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
