//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TImportanceSampling.cxx
 \author Maud Salvoldi <ANP/OCC>
 \date   mar jui 12 11:23:07 CEST 2011

 $Revision$
 $Date$

 \brief Implementation of the class URANIE::DataServer::TImportanceSampling

 *********************************************************************  
 */
#include <sstream>
#include <string>
#include <vector>

// ROOT

// URANIE
#include "TSampling.h"
#include "TImportanceSampling.h"

ClassImp (URANIE::Sampler::TImportanceSampling)

namespace URANIE
{
namespace Sampler
{
TImportanceSampling::TImportanceSampling(URANIE::DataServer::TDataServer *tds,
        TString xname, URANIE::DataServer::TStochasticAttribute *att, Int_t ns) :
        TSamplerStochastic(tds, "IS", ns), _tds_IS(NULL), _x(NULL), _is_x(att), _mean(
                0.0), _std(0.0), _proba(0.0), _blog(kFALSE), _sampling(NULL)
{

    try
    {
        if (!tds->isAttribute(xname))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::constructeur - ERROR: this attribut [%s] is unknown",
                            xname.Data()));
        //copy of the tds where the xname attribute is replaced by att
        Int_t nAttributes;
        nAttributes = tds->getNAttributes();
        if (_blog)
            cout << "**number of attribute" << nAttributes << endl;
        _tds_IS =
                new URANIE::DataServer::TDataServer(
                        Form("%s_IS_copy", tds->GetName()),
                        Form(
                                "IS copy of the tds where the attribute %s was subtistuted by %s",
                                xname.Data(), att->GetName()));
        for (Int_t i = 0; i < nAttributes; i++)
        {
            URANIE::DataServer::TAttribute *att_tds = tds->getAttribute(i);

            if (att_tds->isInput())
            {
                if (att_tds->GetName() == xname)
                {
                    _x =
                            dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att_tds);
                    _is_x->setSonAttribute(_x);
                    _tds_IS->addAttribute(_is_x);
                    if (_blog)
                        cout << "Substitution of the attribute " << xname
                                << " by the attribute " << _is_x->GetName()
                                << endl;
                }
                else
                    _tds_IS->addAttribute(
                            dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att_tds));
            }

        }

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TImportanceSampling::~TImportanceSampling()
{
  if(_x->haveSon())
    _x->removeSonAttribute();

  if(_is_x->haveSon())
    _is_x->removeSonAttribute();

  delete _tds_IS;

}

///////////////////////////////////////////////////////////
// Generate the sample
///////////////////////////////////////////////////////////
void TImportanceSampling::generateSample(Option_t *option)
{
    if (_blog)
        cout << " ** TImportanceSampling::generateSample with option[" << option
                << "]" << endl;

    // instanciate the TSampling pointer
    _sampling = new TSampling(_tds_IS, "lhs", _nS);

    _sampling->generateSample();

//      _tds->setTuple(_tds_IS->getTuple());
    if (_blog)
        cout << " ** TImportanceSampling::generateSample TSampling SEED["
                << _sampling->getSeed() << "]" << endl;

    if (_blog)
        cout << " ** End Of TImportanceSampling::generateSample with option["
                << option << "]" << endl;

    delete _sampling;
}

///////////////////////////////////////////////////////////
//  Create the List of Attributes to simulate from the TDataServer
///////////////////////////////////////////////////////////
void TImportanceSampling::createListOfAttributes()
{
    if (_blog)
        cout << " ** TTImportanceSampling::createListOfAttributes" << endl;

    // Loop on the attributes of the TDataServer to select only TStochasticAttributes
    URANIE::DataServer::TAttribute *att = 0;
    URANIE::DataServer::TStochasticAttribute *stoatt;

    TList *attList = _tds->getListOfAttributes();
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {
        att = (URANIE::DataServer::TAttribute*) attList->At(iatt);
        if (_blog)
            cout << " ** att [" << att->GetName() << "]";
        if (att->isInput())
        {

            // Forcast le pointeur pour savoir s'il provient d'un URANIE::DataServer::TStochasticAttributee
            stoatt =
                    dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);

            if (stoatt != NULL)
            {

                if (stoatt->GetName() == _is_x->GetName())
                {
                    _x = stoatt;
//         _is_x->setSonAttribute(_x);
                    _x->setSonAttribute(_is_x);
                    if (_blog)
                        cout << "Substitution of the attribute "
                                << stoatt->GetName() << " by the attribute "
                                << _is_x->GetName() << endl;
                    _lstOfAttributesToSample->Add(_is_x);
                }
                else
                {
                    if (_blog)
                        cout << " ********** Add";

                    _lstOfAttributesToSample->Add(stoatt);
                }

            }
        }
        if (_blog)
            cout << endl;
    }

    _nX = _lstOfAttributesToSample->GetEntries();

    if (_blog)
    {
        cout << " ** Entries : " << _nX << endl;
        _lstOfAttributesToSample->Print();
    }
    if (_blog)
        cout << " ** End of TImportanceSampling::createListOfAttributes"
                << endl;

    delete attList;
}

Double_t TImportanceSampling::computeMean(TString varY)
{
    Double_t val = 0.0;
    _mean = 0.0;
    try
    {
        if (!_tds_IS->isAttribute(varY))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeMean - ERROR: this attribut [%s] is unknown",
                            varY.Data()));

        // attribute varY present
        TLeaf *lu = _tds_IS->getTuple()->GetLeaf(varY);
        TLeaf *lis = _tds_IS->getTuple()->GetLeaf(_is_x->GetName());

        Double_t dvalue = 0.0;
        Double_t dvalueIS = 0.0;
        Double_t dprobaX = -1.0;
        Double_t dprobaIS = -1.0;

        Int_t npat = _tds_IS->getNPatterns();
        for (Int_t i = 0; i < npat; i++)
        {
            dprobaX = -1.0;
            dprobaIS = -1.0;
            _tds_IS->getTuple()->GetEntry(i);
            dvalue = lu->GetValue();
            dvalueIS = lis->GetValue();

            dprobaX = _x->getPDF(dvalueIS);
            dprobaIS = _is_x->getPDF(dvalueIS);

            _mean *= (Double_t)(i) / (Double_t)(i + 1);
            val = dvalue * dprobaX / dprobaIS;
            _mean += val / (Double_t)(i + 1);

            if (_blog)
                cout << " ** leaf[" << lu->GetName() << "] i[" << i << "/"
                        << npat << "] value (" << dvalue << ") IS name["
                        << lis->GetName() << "," << dvalueIS << "] Min("
                        << _x->getLowerBound() << ") - Max("
                        << _x->getUpperBound() << ")   Proba(" << dprobaX << ","
                        << dprobaIS << ") Mean[" << _mean << "]" << endl;
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    return _mean;
}

Double_t TImportanceSampling::computeStd(TString u)
{
    Double_t dmean = 0.0;
    Double_t poids = 0.0;
    Double_t rep = 0.0;
    _std = 0.0;
    try
    {
      
        if (!_tds_IS->isAttribute(u))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeStd - ERROR: this attribut [%s] is unknown",
                            u.Data()));
        // attribute u is present
      
        Int_t npat = _tds_IS->getNPatterns();
        if (npat <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
		 "TImportanceSampling::computeStd - ERROR: Sample is empty. Run generateSample");
      
        TLeaf *lu = _tds_IS->getTuple()->GetLeaf(u);
      
        TLeaf *lis = _tds_IS->getTuple()->GetLeaf(_is_x->GetName());
      
        Double_t dvalue = 0.0;
        Double_t dvalueIS = 0.0;
        Double_t dprobaX, dprobaIS;
      
        for (Int_t i = 0; i < npat; i++)
        {
            dprobaX = -1.0;
            dprobaIS = -1.0;
            _tds_IS->getTuple()->GetEntry(i);
            dvalue = lu->GetValue();
            dvalueIS = lis->GetValue();

            //value balanced
            dprobaX = _x->getPDF(dvalueIS);
            dprobaIS = _is_x->getPDF(dvalueIS);
            poids = dprobaX / dprobaIS;

            //mean
            dmean *= (Double_t)(i) / (Double_t)(i + 1);
            dmean += dvalue * poids / (Double_t)(i + 1);

//       xx *= (Double_t)(i)/(Double_t)(i+1);
//       xx += poids*dvalue*dvalue/(Double_t)(i+1);

            //variance
            if (i > 0)
            {
                _std *= (Double_t)(i - 1) / (Double_t)(i);
                rep = (Double_t)(i + 1) / (Double_t)(i);
                _std += rep * poids * (dvalue - dmean) * (dvalue - dmean)
                        / (Double_t)(i);
            }

            if (_blog)
                cout << " ** leaf[" << lu->GetName() << "] i[" << i << "/"
                        << npat << "] value (" << dvalue << ") IS name["
                        << lis->GetName() << "," << dvalueIS << "] Min("
                        << _x->getLowerBound() << ") - Max("
                        << _x->getUpperBound() << ")   Proba(" << dprobaX << ","
                        << dprobaIS << ") Mean[" << dmean << "] var[" << _std
                        << "]" << endl;

        }
	      
        _std = TMath::Sqrt(_std);

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
            
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
      
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    
    return _std;
}
Double_t TImportanceSampling::computeThreshold(TString varY, Double_t seuil)
{
    _proba = 0.0;
    try
    {
        //verify the entrance data
        if (!_tds_IS->isAttribute(varY))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeTHreshold - ERROR: this attribut [%s] is unknown",
                            varY.Data()));
        TLeaf *lu = _tds_IS->getTuple()->GetLeaf(varY);
        TLeaf *lis = _tds_IS->getTuple()->GetLeaf(_is_x->GetName());

        Double_t dvalue = 0.0;
        Double_t dvalueIS = 0.0;
        Double_t dweight = 0.0;
        Double_t dprobaX, dprobaIS;
        Int_t npat = _tds_IS->getNPatterns();
        Double_t ind;

        for (Int_t i = 0; i < npat; i++)
        {
            // value
            _tds_IS->getTuple()->GetEntry(i);
            dvalue = lu->GetValue();

            // weight
            dprobaX = -1.0;
            dprobaIS = -1.0;
            dvalueIS = lis->GetValue();
            dprobaX = _x->getPDF(dvalueIS);
            dprobaIS = _is_x->getPDF(dvalueIS);
            dweight = dprobaX / dprobaIS;

            //indicatrice value
            if (dvalue > seuil)
                ind = 1.0;
            else
                ind = 0.0;

            _proba *= (Double_t)(i) / (Double_t)(i + 1);
            _proba += ind * dweight / (Double_t)(i + 1);
            if (_blog)
                cout << " ** " << i << "/" << npat << " -- dvalue[" << dvalue
                        << "] dvalueIS[" << dvalueIS << "] weight[" << dweight
                        << "] :: ind[" << ind << "] :: _proba[" << _proba << "]"
                        << endl;

        }
    } catch (URANIE::Exceptions::UErrorExceptions & ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _proba;
}

Double_t TImportanceSampling::computeQuantile(TString sVar, Double_t dAlpha,
        EQuantileMethod nMethod)
{

    _proba = 0.0;
    try
    {

        if (_blog)
        {
            switch (nMethod)
            {
            case kQuantile1:
            {
                cout << " ** TImportanceSampling::computeQuantile with method 1"
                        << endl;
                break;
            }
            case kQuantile2:
            {
                cout << " ** TImportanceSampling::computeQuantile with method 2"
                        << endl;
                break;
            }
            case kQuantile3:
            {
                cout << " ** TImportanceSampling::computeQuantile with method 3"
                        << endl;
                break;
            }
            case kQuantile4:
            {
                cout << " ** TImportanceSampling::computeQuantile with method 4"
                        << endl;
                 break;
            }
	    default:
		break;
            }
        }

        //verify the entrance data
	switch (nMethod)
	{
	case kQuantile3:
	{
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
	 	    "TImportanceSampling::computeQuantile - ERROR\n the algorithm with [TImportanceSampling::kQuantile3] parameter is not yet implemented");
	    break;
	}
	case kQuantile4:
	{
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TImportanceSampling::computeQuantile - ERROR\n the algorithm with  [TImportanceSampling::kQuantile4] parameter is not yet implemented");
	    break;
	}
        case kQuantileUnknown:
	{
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeQuantile - ERROR: the method [%d] is unknown",
                            nMethod));
	    break;
	}
	default:
	    break;
	}

        if (!_tds_IS->isAttribute(sVar))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeQuantile - ERROR: this attribut [%s] is unknown",
                            sVar.Data()));

        if (dAlpha < 0.0 || dAlpha > 1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TImportanceSampling::computeQuantile - ERROR: Alpha value [%e] is not correct",
                            dAlpha));

        Int_t npat = _tds_IS->getNPatterns();

        TLeaf *lu = _tds_IS->getTuple()->GetLeaf(sVar);
        TLeaf *lis = _tds_IS->getTuple()->GetLeaf(_is_x->GetName());

        // Store the vector of values for the sVar attribute and the xIS attribute
        Double_t *valueData = new Double_t[npat];
        Double_t *valueDataXIS = new Double_t[npat];
        for (Int_t i = 0; i < npat; i++)
        {
            _tds_IS->getTuple()->GetEntry(i);
            valueData[i] = lu->GetValue();
            valueDataXIS[i] = lis->GetValue();
            if (_blog)
                cout << " ** " << i << "/" << npat << " -- att[" << sVar
                        << "] dvalue[" << valueData[i] << "] - att["
                        << _is_x->GetName() << "] dvalueIS[" << valueDataXIS[i]
                        << "]" << endl;
        }
        Int_t *nIndex = new Int_t[npat];
        TMath::Sort(npat, valueData, nIndex, 0);
        Double_t dvalueY = 0.0;
        Double_t dvalueXIS = 0.0;
        Double_t dprobaX, dprobaXIS;
        Double_t dweight = 0.0;

        if (_blog)
        {
            cout << " ******* Order up" << endl;
            for (Int_t i = 0; i < npat; i++)
            {
                dvalueY = valueData[nIndex[i]];
                dvalueXIS = valueDataXIS[nIndex[i]];

                dprobaX = _x->getPDF(dvalueXIS);
                dprobaXIS = _is_x->getPDF(dvalueXIS);

                cout << " ** " << i << "/" << npat << " -- att Y[" << sVar
                        << " :: V " << dvalueY << "] att [" << _is_x->GetName()
                        << " :: V " << dvalueXIS << " :: P " << dprobaXIS
                        << "] [" << _x->GetName() << " :: P " << dprobaX
                        << "] -  W1[" << dprobaXIS / dprobaX << "] W2["
                        << dprobaX / dprobaXIS << "] - index[" << nIndex[i]
                        << "]" << endl;
            }
            cout << " ******* End Of Order up" << endl;
        }

        // Loop on the values
        _proba = 0.0;
        Double_t dThreshold = -10.0;
        switch (nMethod)
        {
        case kQuantile1:
        {
            dThreshold = dAlpha * npat;
            for (Int_t i = 0; i < npat; i++)
            {
                dvalueY = valueData[nIndex[i]];
                dvalueXIS = valueDataXIS[nIndex[i]];

                dprobaX = _x->getPDF(dvalueXIS);
                dprobaXIS = _is_x->getPDF(dvalueXIS);
                dweight = dprobaXIS / dprobaX;

                _proba += dweight;

                if (_blog)
                    cout << " ** " << i << "/" << npat << " -- att Y[" << sVar
                            << "] dvalue[" << dvalueY << "] - att IS["
                            << _is_x->GetName() << "] dvalueIS[" << dvalueXIS
                            << "] index[" << nIndex[i] << "] -- weight["
                            << dweight << "] _proba[" << _proba << "]" << endl;

                if (_proba > dThreshold)
                {
                    _proba = dvalueY;
                    break;
                }

            }
            break;
        }
        case kQuantile2:
        {
            dThreshold = (1.0 - dAlpha) * npat;
            for (Int_t i = npat - 1; i > 0; i--)
            {
                dvalueY = valueData[nIndex[i]];
                dvalueXIS = valueDataXIS[nIndex[i]];

                dprobaX = _x->getPDF(dvalueXIS);
                dprobaXIS = _is_x->getPDF(dvalueXIS);
                dweight = dprobaX / dprobaXIS;

                _proba += dweight;

                if (_blog)
                    cout << " ** " << i << "/" << npat << " -- att Y[" << sVar
                            << " :: " << dvalueY << "] - att IS["
                            << _is_x->GetName() << " :: " << dvalueXIS
                            << "] index[" << nIndex[i] << "] -- weight["
                            << dweight << "] _proba[" << _proba << "]" << endl;

                if (_proba > dThreshold)
                {
                    _proba = dvalueY;
                    break;
                }

            }
            break;
	default:
	    break;
        }
	// to be implemented
        // case kQuantile3:
        // {
        //     dThreshold = -3.0;
        //     break;
        // }
        // case kQuantile4:
        // {
        //     dThreshold = -4.0;
        //     break;
        // }
        }
        if (_blog)
            cout << " ** npat[" << npat << "] Alpha[" << dAlpha
                    << "] - dThreshold[" << dThreshold << "]" << endl;

        // Clean
        delete[] valueData;
        valueData = NULL;
        delete[] valueDataXIS;
        valueDataXIS = NULL;
        delete[] nIndex;
        nIndex = NULL;

        return _proba;

    } catch (URANIE::Exceptions::UErrorExceptions & ue)
    {
        ue.printMessage();
        throw ue;
    }
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TImportanceSampling::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TImportanceSampling::printLog  ******" << endl;
    cout << " ** new tds:" << _tds_IS->GetName() << endl;
    cout << " ** attribute to substitute: " << _x->GetName() << endl;
    cout << " ** new attribute: " << _is_x->GetName() << endl;
    _tds_IS->printLog();
    cout << "*******************************" << endl;
}

}      //namespace DataServer
}      //namespace URANIE

