//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*! \file TMCMC.cxx
 * \brief This file presents the object \b TMCMC whom methods  enable to generate from MCMC methods.
 *
 * There are three main MCMC methods implemented in this file: Metropolis-Hastings (variable-at-a-time, random walk), Gibbs et Normal Kernel Coupler.\n These methods have been applied in the case of a multidimensional normal law in order to to test their robustness, then to test a multidimensional gaussian mixing.\n
 This class uses the classes TNormale and TMelange, programmed before. 
 * 
 */

#include "TH1.h"

#include "TMCMC.h"
#include "TObjArray.h"
#include "TLeaf.h"

ClassImp (TMCMC)
/*!  \brief Constructor of the class TMCMC. q is the attribute used as an instrumental law in the broad sense of the word (not only as an instrumental law of algorithms of Metropolis-Hastings).\n _rdm is used in NKC so as to generate numbers located between 0 and 1.
 */
TMCMC::TMCMC() :
        _sdraw("")
{

    //  cout << "construit TMCMC " << endl;
    _size = 0;
//    q = new TNormale();
//    _moy.ResizeTo(1);
//    _moyRB.ResizeTo(1);
//    _moyP.ResizeTo(1);
//    _moyR.ResizeTo(1);
    _rdm = new TRandom3();
    _rdm->SetSeed(0);
    _rdm->Uniform(1.0);
    cvisu = NULL;
    _dataorg = NULL;
}

TMCMC::~TMCMC()
{

}

/*! \fn void TMCMC::varTimeMetropolisSampling (TNormale* f, TVectorD X)
 * \brief Method of simulation of a gaussian vector using the algorithm variable-at-a-time of MH.\n The components of the gaussian vector are updated one by one in each step, the instrumental law being a real gaussian centered at the value of the component to be updated at the previous step, and the variance is the same as the one of the law to be simulated.
 *
 * slow convergence 
 */
TNtupleD* TMCMC::varTimeMetropolisSampling(TNormale* f, TVectorD X)
{
    q = new TNormale();
    TVectorD Y(f->getDim());
    TVectorD M = f->getMean();
    TVectorD A = f->getSigma();
    X.ResizeTo(f->getDim());
    Double_t* tab = new Double_t[f->getDim() + 1];
    tab[0] = 0;
    for (Int_t k = 0; k < f->getDim(); k++)
        tab[k + 1] = X(k);
    (f->getTuple())->Fill(tab);
    TVectorD mean(1);
    TMatrixD sigma(1, 1);
    TVectorD v(2);
    v(0) = 1;
    Double_t u = 0;
    Double_t x = 0;
    Double_t seuil = 0;
    Double_t a = 0;
    Double_t b = 0;
    for (Int_t i = 0; i < _size; i++)
    {
        Y = X;
        tab[0] = i + 1;
        for (Int_t l = 0; l < f->getDim(); l++)
        {
            u = (q->getRandom())->Uniform(1.0);
            sigma(0, 0) = A(l);
            q->setMatEcartTypeCorrelation(sigma);
            mean(0) = X(l);
            q->setMean(mean);
            x = q->getObs();
            Y(l) = x;
            a = q->eval(Y(l));
            mean(0) = Y(l);
            q->setMean(mean);
            b = q->eval(X(l));
            v(1) = (f->eval(Y) * b) / (f->eval(X) * a);
            seuil = v.Min();
            if (u <= seuil)
            {
                X(l) = x;
            }
            tab[l + 1] = X(l);
        }
        (f->getTuple())->Fill(tab);
    }
//  (f->getTuple())->Write();
    return (f->getTuple());
}

/*! \fn void TMCMC::randomWalkMetropolisSampling (TNormale* f, TVectorD X)
 * \brief Method of simulation of a gaussian vector using the algorithm random walk of MH. \n The components are updated at the same time at each step, and the instrumental law is a gaussian vector with mean being the one of the vector simulated at the  previous step and with variance being the one of the law to be simulated multiplied by \f$ \frac{2.38^2}{d} \f$, where \f$ d \f$ is the dimension of the law to be simulated. 
 *
 *relatively fast convergence
 */
TNtupleD* TMCMC::randomWalkMetropolisSampling(TNormale* f, TVectorD X)
{
    Double_t m = 0;
    TVectorD Z(f->getDim() + 1);
    Z.Zero();
    Z.SetSub(1, X);
    (f->getTuple())->Fill(Z.GetMatrixArray());
    Double_t u = 0;
    Double_t seuil = 0;
    TVectorD v(2);
    v(0) = 1;
    TVectorD Y(f->getDim());
    TMatrixD M = f->getMatEcartTypeCorrelation();
    for (Int_t i = 0; i < f->getDim(); i++)
        M(i, i) = (2.38 / sqrt((Float_t)(f->getDim()))) * M(i, i);
    q->setMatEcartTypeCorrelation(M);
    for (Int_t n = 0; n < _size; n++)
    {
        Z(0) = n + 1;
        u = (q->getRandom())->Uniform(1.0);
        q->setMean(X);
        Y = q->getObsMultiDim();
        v(1) = f->eval(Y) / f->eval(X);
        seuil = v.Min();
        if (u <= seuil)
        {
            X = Y;
            m++;
        }
        Z.SetSub(1, X);
        (f->getTuple())->Fill(Z.GetMatrixArray());
    }
    cout << m / _size << endl;
    //(f->getTuple())->Write();
    return (f->getTuple());
}

/*! \fn void TMCMC::gibbsSampling(TNormale* f, TVectorD X )
 * \brief Method implementing the algorithm of Gibbs, used to simulate a gaussian vector.
 *
 * This is a particular case of the method "variable-at-a-time" of MH, where the instrumental laws
 * aimed at updating the components, are their laws of each of them, conditioned under the others.
 *
 * fast convergence 
 */
TNtupleD* TMCMC::gibbsSampling(TNormale* f, TVectorD X)
{
    q = new TNormale();
    TVectorD M = f->getMean();
    TMatrixD A = (f->getMatCovariance()).Invert();
    A.Print();
    X.ResizeTo(f->getDim());
    Double_t* tab = new Double_t[f->getDim() + 1];
    tab[0] = 0;
    for (Int_t k = 0; k < f->getDim(); k++)
        tab[k + 1] = X(k);
    (f->getTuple())->Fill(tab);
    TVectorD mean(1);
    TMatrixD sigma(1, 1);
    for (Int_t i = 0; i < _size; i++)
    {
        tab[0] = i + 1;
        for (Int_t l = 0; l < f->getDim(); l++)
        {
            sigma(0, 0) = 1 / sqrt(A(l, l));
            q->setMatEcartTypeCorrelation(sigma);
            mean(0) = 0;
            for (Int_t j = 0; j < f->getDim(); j++)
                mean(0) = mean(0) + (M(j) - X(j)) * A(j, l) / A(l, l);
            mean(0) = mean(0) + X(l);
            q->setMean(mean);
            X(l) = q->getObs();
            tab[l + 1] = X(l);
        }
        (f->getTuple())->Fill(tab);
    }
    //(f->getTuple())->Write();
    return (f->getTuple());
}

TVectorD TMCMC::getObs(TNormale* f, TVectorD X)
{
    TVectorD M = f->getMean();
    TMatrixD A = (f->getMatCovariance()).Invert();
    X.ResizeTo(f->getDim());
    TVectorD mean(1);
    TMatrixD sigma(1, 1);
    Int_t N = f->getDim();
    for (Int_t l = 0; l < N; l++)
    {
        sigma(0, 0) = 1 / sqrt(A(l, l));
        q->setMatEcartTypeCorrelation(sigma);
        mean(0) = 0;
        for (Int_t j = 0; j < f->getDim(); j++)
            mean(0) = mean(0) + (M(j) - X(j)) * A(j, l) / A(l, l);
        mean(0) = mean(0) + X(l);
        q->setMean(mean);
        X(l) = q->getObs();
    }
    return X;
}

/*! \fn void TMCMC::gibbsSampling(TMelange* mel, TVectorD X )
 * \brief Method implementing the algorithm of Gibbs, in order to simulate a multidimensionnal gaussian mixing.
 *
 * Since the conditional densities of such a law  can not be simulated, a completion step is retained :
 * A variable is introduced which will enable to obtain conditional densities that can be simulated.
 */

TNtupleD* TMCMC::gibbsSampling(TMelange* mel, TVectorD X)
{
//   _moy.ResizeTo(mel->getDimMelange());

    TVectorD noyau(mel->getTailleMelange());
    noyau.Zero();
    Double_t u;
    Int_t z;
    Int_t n;
    Double_t* stock = new Double_t[mel->getDimMelange() + 1];
    stock[0] = 0;
    for (Int_t k = 0; k < mel->getDimMelange(); k++)
        stock[k + 1] = X(k);
    (mel->getTuple())->Fill(stock);
    TVectorD y(mel->getTailleMelange());

    char titre[200];
    if (mel->getDimMelange() == 1)
        sprintf(titre, "Initialisation (%5.2f)", X(0));
    else if (mel->getDimMelange() == 2)
        sprintf(titre, "Initialisation (%5.2f, %5.2f)", X(0), X(1));
    else
        sprintf(titre, "Initialisation (%5.2f, %5.2f, %5.2f)", X(0), X(1),
                X(2));
    cout << titre << endl;

    for (Int_t i = 0; i < _size; i++)
    {
        n = 0;
        u = _rdm->Uniform(1.0);
        stock[0] = i + 1;
        y = mel->LoiDeZ(X);

        while (y(n) < u)
        {
            n++;
        }
        z = n;
        noyau(z)++;q
        ->setMean(mel->getMean(z));
        q->setMatEcartTypeCorrelation(mel->getMatEcartTypeCorrelation(z));
        if (mel->getDimMelange() == 1)
        {
            X[0] = q->getObs();
        }
        else
        {
            X = q->getObsMultiDim();
        }
        TVectorD P = mel->discrimination(X);
        for (Int_t l = 0; l < mel->getDimMelange(); l++)
            stock[l + 1] = X(l);
        (mel->getTuple())->Fill(stock);
    }
    for ( n = 0; n < mel->getDimMelange(); n++)
        X(n) = X(n) / _size;
    if (mel->getDimMelange() == 1)
    {
        Draw1(mel, mel->getTuple(), titre);
    }
    if (mel->getDimMelange() == 2)
    {
        Draw(mel, mel->getTuple(), titre);
    }
    //(mel->getTuple())->Write();
    Int_t somme = 0;
    for (Int_t j = 0; j < mel->getTailleMelange(); j++)
    {
        cout << "noyau" << j + 1 << endl;
        cout << "proportion:" << noyau(j) << endl;
        somme = somme + (Int_t)(noyau(j));
    }
    cout << "------" << endl;
    cout << somme << endl;
    return mel->getTuple();
}

/*! \fn void TMCMC::randomWalkMetropolisSampling(TMelange* mel, TVectorD X )
 * \brief Method implementing the random Walk de MH of a a multidimensionnal gaussian mixing. The instrumental law is the same as observed in the case of a gaussian vector, except that the variance used is the mean value of the variances of each element of the mixing, multiplied by \f$ \frac{2.38^2}{d} \f$, where \f$ d \f$ is the dimension of the law to be simulated.  
 */

TNtupleD* TMCMC::randomWalkMetropolisSampling(TMelange* mel, TVectorD X)
{
    TVectorD noyau(mel->getTailleMelange());
    noyau.Zero();
    TVectorD posteriori(mel->getTailleMelange());
    TMatrixD A = mel->getMatEcartTypeCorrelation();
    for (Int_t i = 0; i < mel->getDimMelange(); i++)
        A(i, i) = (2.38 / sqrt((Float_t)(mel->getDimMelange()))) * A(i, i);
    q->setMatEcartTypeCorrelation(A);
    Double_t u = 0;
    Double_t seuil = 0;
    TVectorD Z(mel->getDimMelange() + 1);
    Z.Zero();
    Z.SetSub(1, X);
    (mel->getTuple())->Fill(Z.GetMatrixArray());
    TVectorD Y(mel->getDimMelange());
    TVectorD v(2);
    v(0) = 1;
    char titre[200];
    sprintf(titre, "Initialisation (%5.2f, %5.2f)", X(0), X(1));
    cout << titre << endl;
    for (Int_t i = 0; i < _size; i++)
    {
        Z(0) = i;
        u = (q->getRandom())->Uniform(1.0);
        q->setMean(X);
        Y = q->getObsMultiDim();
        v(1) = mel->eval(Y) / mel->eval(X);
        seuil = v.Min();
        if (u <= seuil)
        {
            X = Y;
        }
        Z.SetSub(1, X);
        (mel->getTuple())->Fill(Z.GetMatrixArray());
        posteriori = mel->discrimination(X);
        for (Int_t k = 0; k < mel->getTailleMelange(); k++)
        {
            if (posteriori(k) == posteriori.Max())
                noyau(k)++;}
        }

    Int_t somme = 0;
    for (Int_t l = 0; l < mel->getTailleMelange(); l++)
    {
        cout << "noyau" << l + 1 << " = " << noyau(l) << endl;
        somme = somme + (Int_t)(noyau(l));
    }
    cout << "------" << endl;
    cout << somme << endl;

    if (mel->getDimMelange() == 2)
    {
        Draw(mel, mel->getTuple(), titre);
    }

    return (mel->getTuple());
}

/*! \fn void TMCMC::varTimeMetropolisSampling(TMelange* mel, TVectorD X )
 * \brief Method implementing the "variable-at-a-time" of MH for a gaussian mixing. The transformation of the instrumentale law*/

TNtupleD* TMCMC::varTimeMetropolisSampling(TMelange* mel, TVectorD X)
{
    TVectorD noyau(mel->getTailleMelange());
    noyau.Zero();
    TVectorD posteriori(mel->getTailleMelange());
    q = new TNormale();
    TVectorD Y(mel->getDimMelange());
    TVectorD A = mel->getSigma();
    X.ResizeTo(mel->getDimMelange());
    Double_t* tab = new Double_t[mel->getDimMelange() + 1];
    tab[0] = 0;
    for (Int_t k = 0; k < mel->getDimMelange(); k++)
        tab[k + 1] = X(k);
    (mel->getTuple())->Fill(tab);
    TVectorD mean(1);
    TMatrixD sigma(1, 1);
    TVectorD v(2);
    v(0) = 1;
    Double_t u = 0;
    Double_t x = 0;
    Double_t seuil = 0;
    Double_t a = 0;
    Double_t b = 0;
    char titre[200];
    sprintf(titre, "Initialisation (%5.2f, %5.2f)", X(0), X(1));
    cout << titre << endl;
    for (Int_t i = 0; i < _size; i++)
    {
        Y = X;
        tab[0] = i + 1;
        for (Int_t l = 0; l < mel->getDimMelange(); l++)
        {
            u = (q->getRandom())->Uniform(1.0);
            sigma(0, 0) = A(l);
            q->setMatEcartTypeCorrelation(sigma);
            mean(0) = X(l);
            q->setMean(mean);
            x = q->getObs();
            Y(l) = x;
            a = q->eval(Y(l));
            mean(0) = Y(l);
            q->setMean(mean);
            b = q->eval(X(l));
            v(1) = (mel->eval(Y) * b) / (mel->eval(X) * a);
            seuil = v.Min();
            if (u <= seuil)
            {
                X(l) = x;
            }
            tab[l + 1] = X(l);
        }
        (mel->getTuple())->Fill(tab);
        posteriori = mel->discrimination(X);
        for (Int_t k = 0; k < mel->getTailleMelange(); k++)
        {
            if (posteriori(k) == posteriori.Max())
                noyau(k)++;}
        }

    if (mel->getDimMelange() == 2)
    {
        Draw(mel, mel->getTuple(), titre);
    }
    Int_t somme = 0;
    for (Int_t l = 0; l < mel->getTailleMelange(); l++)
    {
        cout << "noyau" << l + 1 << " = " << noyau(l) << endl;
        somme = somme + (Int_t)(noyau(l));
    }
    cout << "------" << endl;
    cout << somme << endl;
    return (mel->getTuple());
}

void TMCMC::Draw1(TMelange* mel, TNtupleD* ntd, Option_t *option)
{
    UNUSED(mel); // ? UNUSED ?
    TString soption = option;
    soption.ToLower();
    if (cvisu == NULL)
        cvisu = new TCanvas("c1", "canvas de visu des MCMC");
    cvisu->Divide(2);

    // courbe au cours du temp
    cvisu->cd(1);
    ntd->Draw("var1:indice", "", "lp");
    // courbe au cours du temp
    cvisu->cd(2);

    ntd->Draw("var1>>hhhttteeemp", "indice>400");
    TH1F *htemp = (TH1F*) gPad->GetPrimitive("hhhttteeemp");
    htemp->Scale(1.0 / htemp->Integral());

    cvisu->Modified();
    cvisu->Update();

}

/*! \fn void  TMCMC::Draw(TMelange* mel, TNtupleD* ntd, Option_t *option)
 \brief Method allowing to visualize the results of the MCMC methods programmed in this class, when the law has to be simulated is 2D dimension. Four visualizations are provided, allowing to follow the evolution of the simulation versus time, to get to know whether the values generated are distributed according to an expected law and after a certain number of iterations. 
 */

void TMCMC::Draw(TMelange* mel, TNtupleD* ntd, Option_t *option)
{
    //cout << "TMCMC::Draw [" << option << "]" << endl;
    // Gestion des options
    TString opt = option;
    opt.ToLower();
    if (ntd == NULL)
        return;
    //  if (opt.Contains("s"))
//     ntd->StartViewer();

    // Search For the two first variables

    TObjArray* leaves = _dataorg->GetListOfLeaves();
    Int_t nLeaf = leaves->GetEntries();
    Int_t ncrt = 0;

    TString svarx, svary;

    for (Int_t i = 0; i < nLeaf; i++)
    {
        TLeaf* leaf = (TLeaf *) leaves->UncheckedAt(i);
        TString sName = leaf->GetName();
        if (sName.Contains("n__iter__", TString::kExact) != kTRUE)
        {
            if (ncrt == 0)
                svarx = sName;
            if (ncrt == 1)
                svary = sName;
            ncrt++;
            if (ncrt == 2)
                i = nLeaf;
        }
    }
    TString svardraw = Form("%s:%s", svary.Data(), svarx.Data());

    if (_sdraw.Length() > 0)
        svardraw = _sdraw;

    Int_t nbloi = mel->getTailleMelange();

    if (cvisu == NULL)
        cvisu = new TCanvas("c1", "canvas de visu des MCMC");

    TPaveText *pt = new TPaveText(0.340981, 0.478261, 0.519778, 0.526087,
            "blNDC");
    pt->SetFillColor(19);
    pt->AddText(option);
    pt->Draw();

    cvisu->Divide(2, 2);

    // nuage de points
    cvisu->cd(1);
    // Les points d'origine
    if (_dataorg != NULL)
        //_dataorg->Draw("x2:x1","","");
        _dataorg->Draw(svardraw.Data(), "", "");
    // Les ellipses fournies par mixmod
    for (Int_t i = 0; i < nbloi; i++)
    {
        TVectorD mean = mel->getMean(i);
        TMatrixD corvar = mel->getMatEcartTypeCorrelation(i);

        TEllipse *ellipse = new TEllipse(mean(0), mean(1), corvar(0, 0),
                corvar(1, 1));
        //TEllipse *ellipse = new TEllipse(mean(1), mean(0), corvar(1,1) , corvar(0,0) );

        ellipse->SetLineColor(4);
        ellipse->SetLineStyle(1);
        ellipse->SetLineWidth(3);
        ellipse->Draw("only");

        cvisu->Modified();
        cvisu->Update();
    }

    // courbe au cours du temp
    cvisu->cd(4);
    ntd->Draw("var2:indice", "", "lp");
    // courbe au cours du temp
    cvisu->cd(3);
    ntd->Draw("var1:indice", "2", "lp");
    // nuage de points
    cvisu->cd(2);
    // convergence des estimateurs
    //c1->Range(-10.0, -10.0, 10., 10.0);
    cvisu->SetBorderSize(2);
    cvisu->SetFrameFillColor(0);

    TVectorD X1(mel->getTailleMelange());
    TVectorD X2(mel->getTailleMelange());
    TVectorD Y1(mel->getTailleMelange());
    TVectorD Y2(mel->getTailleMelange());
    Double_t Xmin;
    Double_t Xmax;
    Double_t Ymin;
    Double_t Ymax;
    for (Int_t j = 0; j < mel->getTailleMelange(); j++)
    {
        TVectorD m = mel->getMean(j);
        TVectorD sig = mel->getSigma(j);

        X1(j) = m(0) - 5 * sig(0);
        Y1(j) = m(1) - 5 * sig(1);
        X2(j) = m(0) + 5 * sig(0);
        Y2(j) = m(1) + 5 * sig(1);
    }
    Xmin = X1.Min();
    Xmax = X2.Max();
    Ymin = Y1.Min();
    Ymax = Y2.Max();
    if (0)
    {
        Y1.Print("vecteur y1");
        cout << "Xmin[" << Xmin << "]" << endl;
        cout << "Xmax[" << Xmax << "]" << endl;
        cout << "Ymin[" << Ymin << "]" << endl;
        cout << "Ymax[" << Ymax << "]" << endl;
    }

    ntd->SetMarkerColor(8);
    if (0)
    {
        gPad->DrawFrame(Xmin, Ymin, Xmax, Ymax);
        ntd->Draw("var2:var1", "", "same");
    }
    else
    {
        gPad->DrawFrame(Xmin, Ymin, Xmax, Ymax);
        ntd->SetMarkerColor(30);
        ntd->Draw("var2:var1", "indice<1000", "same");
        ntd->SetMarkerColor(50);
        ntd->Draw("var2:var1", "indice>1000", "samecontz");
    }

    // Les ellipses fournies par mixmod
    for (Int_t i = 0; i < nbloi; i++)
    {
        TVectorD mean = mel->getMean(i);
        TMatrixD corvar = mel->getMatEcartTypeCorrelation(i);

        TEllipse *ellipse = new TEllipse(mean(0), mean(1), corvar(0, 0),
                corvar(1, 1));
        //TEllipse *ellipse = new TEllipse(mean(1),mean(0), corvar(1,1) , corvar(0, 0) );

        ellipse->SetLineColor(4);
        ellipse->SetLineStyle(1);
        ellipse->SetLineWidth(3);
        ellipse->Draw("only");

        cvisu->Modified();
        cvisu->Update();
    }

    if (_dataorg != NULL)
        _dataorg->Draw(svardraw.Data(), "", "same");
    //_dataorg->Draw("x2:x1","","same");

    cvisu->Modified();
    cvisu->Update();

    Int_t n = mel->getDimMelange();
    TString sLabelVariable(24 * n + 6);
    sLabelVariable += "indice:";
    for (Int_t i = 1; i < mel->getDimMelange(); i++)
    {
        sLabelVariable += "moy";
        sLabelVariable += i;
        sLabelVariable += ":";
        sLabelVariable += "moyRB";
        sLabelVariable += i;
        sLabelVariable += ":";
        sLabelVariable += "moyP";
        sLabelVariable += i;
        sLabelVariable += ":";
        sLabelVariable += "moyR";
        sLabelVariable += i;
        sLabelVariable += ":";
    }
    sLabelVariable += "moy";
    sLabelVariable += n;
    sLabelVariable += ":";
    sLabelVariable += "moyRB";
    sLabelVariable += n;
    sLabelVariable += ":";
    sLabelVariable += "moyP";
    sLabelVariable += n;
    sLabelVariable += ":";
    sLabelVariable += "moyR";
    sLabelVariable += n;

    Double_t moyenne1 = 0;
    Double_t moyenne2 = 0;
    TNtupleD* nt = new TNtupleD("estimations", "convergence",
            sLabelVariable.Data());
    for (Int_t i = 0; i < ntd->GetEntries(); i++)
    {
        ntd->GetEntry(i);
        Double_t* val = ntd->GetArgs();
        moyenne1 = moyenne1 + val[1];
        moyenne2 = moyenne2 + val[2];
        nt->Fill(i, moyenne1 / (i + 1), 0, 0, 0, moyenne2 / (i + 1), 0, 0, 0);
    }

    //cout << "fin de TMCMC::Draw [" << option << "]" << endl;
}

/*! \fn Int_t TMCMC::selectComponentState()
 \brief Method used in the NKC, to determine the vector which will be used as mean of the instrumental law.
 */

Int_t TMCMC::selectComponentState()
{
    TVectorD seuil(11);
    seuil.Zero();
    for (Int_t l = 0; l < 10; l++)
        seuil(l + 1) = seuil(l) + 0.1;
    Double_t u;
    u = _rdm->Uniform(1.0);
    for (Int_t i = 0; i < 10; i++)
    {
        if ((u > seuil(i)) && (u <= seuil(i + 1)))
        {
            return i;
        }
    }
    return 0;
}

/*! \fn TNtupleD* TMCMC::NKCSampling(TMelange* mel, TVectorD* X)
 \brief Method allowing to deal with the "Normal Kernel Coupler", a MMC methode MCMC efficient for the simulation of mixing.
 */

TNtupleD* TMCMC::NKCSampling(TMelange* mel, TVectorD* X)
{
    TVectorD init(X[1].GetNrows() + 1);
    init(0) = 0;
    init.SetSub(1, X[1]);
    (mel->getTuple())->Fill(init.GetMatrixArray());

    TVectorD* Y = new TVectorD[100];
    for (Int_t m = 0; m < 100; m++)
        Y[m].ResizeTo(mel->getDimMelange());
    //  TVectorD tmp(mel->getDimMelange());
    TVectorD resultat(mel->getDimMelange() + 1);

    //construction de la variance de la loi instrumentale
    TMatrixD V = mel->getMatEcartTypeCorrelation();
    for (Int_t h = 0; h < mel->getDimMelange(); h++)
        V(h, h) = sqrt(1.4 * TMath::Power(0.1, 2 / (4 + mel->getDimMelange())))
                * V(h, h);
    q->setMatEcartTypeCorrelation(V);

    //Int_t indiceMelange;
    Int_t ind;
    //Int_t compteur=0;
    //Int_t indice;

    // loi instrumentale
    TMelange* instru = new TMelange(mel->getDimMelange());
    // melange servant par la suite � selectionner les vecteurs � actualiser
    //  TMelange* instru2 = new TMelange(mel->getDimMelange());
    Double_t u;
    TRandom3* rdm = new TRandom3();
    rdm->SetSeed(0);

    Double_t alpha;
    TVectorD v(2);
    v(0) = 1;

    //  Int_t* indiceVecteur= new Int_t[100];

    for (Int_t i = 0; i < _size; i++)
    {
        cout << i << endl;
        Y = X;
        //construction de la loi instrumentale.
        for (Int_t l = 0; l < 100; l++)
            instru->addLoi(X[l], V, 0.01);
        // construction du melange servant � la selection.
        //for(Int_t g=0;g<100;g++)
        //instru2->addLoi(X[g],V,0.01);

        //for(Int_t j=0;j<2;j++)
        //  indiceVecteur[j]=j;
        //mise � jour des vecteurs d'�tats
        // while(instru2->getTailleMelange()>0)
        for (Int_t j = 0; j < 100; j++)
        {
            //selection du vecteur d'etat
            //indiceMelange=selectComponentState(instru2);
            //if (indiceMelange<=instru2->getTailleMelange()-2) 
            // {
            // for(Int_t k=indiceMelange;k<instru2->getTailleMelange();k++)
            //   {
            //    indiceVecteur[k]=k+1;
            //  }
            // }
            //tmp=instru->getMean(j);
            //selection de l'esp�rance.
            ind = selectComponentState();
            q->setMean(X[ind]);
            //if(compteur>0)
            // indice=indiceVecteur[indiceMelange];
            //else
            // indice=indiceMelange;
            Y[j] = q->getObsMultiDim();
            instru->reset();
            Double_t a = instru->eval(X[j]);
            for (Int_t k = 0; k < 100; k++)
                instru->addLoi(Y[k], V, 0.01);
            Double_t b = instru->eval(Y[j]);
            //for(Int_t p=0;p<100;p++)
            // inst->addLoi(Y[p],V,0.01);
            v(1) = (mel->eval(Y[j]) * a) / (mel->eval(X[j]) * b);
            alpha = v.Min();
            u = rdm->Uniform(1.0);
            if (u <= alpha)
                X = Y;
            //suppression du vecteur selectionn� du m�lange pour ne pas le reselectionner avant la prochaine it�ration
            //instru2->removeLoi(indiceMelange);
            //compteur++;
        }

        resultat(0) = i + 1;
        resultat.SetSub(1, X[1]);
        (mel->getTuple())->Fill(resultat.GetMatrixArray());
    }
    return mel->getTuple();
}

/*! \fn TNtupleD* TMCMC::NKCSampling2(TMelange* mel, TVectorD* X)
 \brief Method implementing the NKC using 100 state vectors.
 
 Theorically this method is supposed to work, but we have to face to implementation problems.
 */

TNtupleD* TMCMC::NKCSampling2(TMelange* mel, TVectorD* X)
{
    Double_t stock = 0;
    Int_t compteur = 0;
    Double_t u;
    Int_t indice = 0;
    // Int_t ind;
    TVectorD v(2);
    v(0) = 1;
    TVectorD resultat(mel->getDimMelange() + 1);
    resultat(0) = 0;
    resultat.SetSub(1, X[2]);
    (mel->getTuple())->Fill(resultat.GetMatrixArray());
    TVectorD Y(mel->getDimMelange());
    Double_t a = 0;
    Double_t b = 0;
    Double_t c = 0;
    Double_t d = 0;
    cout << "taille du m�lange " << mel->getTailleMelange() << endl;
    TMatrixD V = mel->getMatEcartTypeCorrelation();
    for (Int_t h = 0; h < mel->getDimMelange(); h++)
        V(h, h) = sqrt(1.4 * TMath::Power(0.1, 2 / (4 + mel->getDimMelange())))
                * V(h, h);
    q->setMatEcartTypeCorrelation(V);
    Int_t nEtat = 10;
    for (Int_t i = 0; i < _size; i++)
    {
        for (Int_t j = 0; j < nEtat; j++)
        {
            compteur++;
            u = _rdm->Uniform(1.0);
            indice = _rdm->Integer(nEtat);
            cout << "indice[" << j << "][" << indice << "] * ";
            q->setMean(X[indice]);
            Y = q->getObsMultiDim();

            for (Int_t m = 0; m < nEtat; m++)
            {
                q->setMean(X[m]);
                stock = q->eval(Y);
                a = a + 0.01 * stock;
            }

            for (Int_t n = 0; n < nEtat; n++)
            {
                if (n == j)
                    q->setMean(Y);
                else
                    q->setMean(X[n]);
                stock = q->eval(X[j]);
                b = b + 0.01 * stock;
            }

            for (Int_t k = 0; k < mel->getTailleMelange(); k++)
            {
                q->setMean(mel->getMean(k));
                q->setMatEcartTypeCorrelation(
                        mel->getMatEcartTypeCorrelation(k));
                stock = q->eval(Y);
                c = c + (mel->getProb(k)) * stock;
                stock = q->eval(X[j]);
                d = d + (mel->getProb(k)) * stock;
            }

            v(1) = (c * b) / (d * a);
            if (u <= v.Min())
            {
                X[j] = Y;
            }
            resultat(0) = compteur;
            resultat.SetSub(1, X[j]);
            (mel->getTuple())->Fill(resultat.GetMatrixArray());
        }
        //resultat(0)=i+1;
        cout << i + 1 << endl;
        //resultat.SetSub(1,X[_rdm->Integer(100)]);
        //(mel->getTuple())->Fill(resultat.GetMatrixArray());
    }
    return mel->getTuple();
}

/** Association of a canvas */
void TMCMC::SetCanvas(TCanvas * canvas)
{
    cvisu = canvas;
}
/** No descriptions */
void TMCMC::SetTuple(TNtupleD *tuple)
{
    _dataorg = tuple;
}
