//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TMaxiMinLHS.cxx
 *  \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 *  \date   $date$
 *  \brief Implementation of the class URANIE::Sampler::TMaxiMinLHS. (based on the algorithm by Guillaume Damblin)
 *
 *  Created the : Thu Apr 13 13:37:23 2004
 *
 */
#include <ctime>
#include <iostream>
#include <iomanip>
#include <fstream>

////// ROOT
#include "TMath.h"
#include "TDatime.h"
#include "TROOT.h"
#include "TRandom3.h"

////// URANIE
#include "TMaxiMinLHS.h"

ClassImp (URANIE::Sampler::TMaxiMinLHS)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TMaxiMinLHS::TMaxiMinLHS(URANIE::DataServer::TDataServer *tds, int sampleSize, double sainitTemp, double saCoeff, int saN, int sainnerN):
        TSamplerStochastic(tds, "lhs", sampleSize), _SAinitTemp(-1), _SACoeff(-1), _SAN(-1), _SAinnerN(-1)
{
    if (_blog)
        cout << "constructor TMaxiMinLHS using TDataServer["
                << tds->GetName() << "]" << endl;

    _vi=NULL; _vj=NULL;
    
    setSAProperties( sainitTemp,  saCoeff,  saN,  sainnerN);

    //Check whether a tree exists or not
    try
    {
	if( tds->getTuple() != NULL )
	{
	    _bAlreadyFilled = (tds->getTuple()->GetEntries() != 0) ;
            int nEntries = tds->getTuple()->GetEntries();
	    if( _nS != nEntries )
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TMaxiMinLHS ERROR : The number of requested sample [%i] is different from the one stored [%i]",
                                                                nEntries, _nS));

	    TList *attList = _tds->getListOfAttributes();
	    _nX = attList->GetEntries();
	    delete attList;
	}
    }catch( URANIE::Exceptions::UWarningExceptions& ue)
     {
	 _bAlreadyFilled = false;
     }catch( URANIE::Exceptions::UErrorExceptions& ue)
     {
	 ue.printMessage();
         if( _vi != NULL)
             delete[] _vi;
         _vi=NULL;
         if( _vj != NULL)
             delete[] _vj;
         _vj=NULL;
	 throw ue;
     }  
	

    //Create the sampler for initial sampling    
    _sampling = ( (_bAlreadyFilled ) ? NULL : new TSampling(tds, "lhs", sampleSize));

    _sMethod=TString("maximin_lhs");

    //Resizing the methods
    _currentLHS.ResizeTo(_nS,_nX);
    _varyingLHS.ResizeTo(_nS,_nX);
    _currentDist.ResizeTo(_nS,_nS);
    _varyingDist.ResizeTo(_nS,_nS);
    _fp.ResizeTo(_nS,_nS);

    _ns=_nS;
    _nx=_nX;

    _vi = new double[_nx];
    _vj = new double[_nx];
    
}

TMaxiMinLHS::~TMaxiMinLHS()
{
    if (_blog)
        cout << "************ Destroying URANIE::Sampler::TMaxiMinLHS"
                << endl;

    if( _sampling != NULL)
	delete _sampling;
    _sampling=NULL;

    if( _vi != NULL)
        delete[] _vi;
    _vi=NULL;
    if( _vj != NULL)
        delete[] _vj;
    _vj=NULL;
}

void TMaxiMinLHS::setSeed(Int_t ind )
{

    if( _sampling != NULL)
	_sampling->setSeed(ind);

    TSamplerStochastic::setSeed(ind);
    
}
////////////////////////////////////////////////////////////
//  maximin stuff
////////////////////////////////////////////////////////////
void TMaxiMinLHS::setSAProperties(double &sainitTemp, double &saCoeff, int &saN, int &sainnerN)
{

    try
    {
	if( sainitTemp < 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMaxiMinLHS::setSAProperties ERROR : The initial temperature cannot be negative [%g]", _SAinitTemp));
	_SAinitTemp = sainitTemp;
	
	if( saCoeff < 0 || saCoeff >= 1)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMaxiMinLHS::setSAProperties ERROR : The temperature coefficient cannot be negative or greater than 1 [%g]", _SACoeff));
	_SACoeff = saCoeff;
	
	if( saN < 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMaxiMinLHS::setSAProperties ERROR : The number of iteration cannot be negative [%i]", _SAN));
	_SAN = saN;
	
	if( sainnerN < 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMaxiMinLHS::setSAProperties ERROR : The inner number of iteration cannot be negative [%i]", _SAinnerN));
	_SAinnerN = sainnerN;
		
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
     {
	 ue.printMessage();
	 throw ue;
     }    
}

Double_t TMaxiMinLHS::getUpdatedLPNorm(Double_t LP_norm_, unsigned int nrowRand1_, unsigned int nrowRand2_)
{
    unsigned int k=0;
    LP_norm_=pow(LP_norm_,50);
    Double_t dst=0;
    // const int local_nx = _nx;
    // Double_t vi[local_nx],vj[local_nx];

    _varyingLHS.ExtractRow(nrowRand1_,k,_vi);
    for (unsigned int i=0; i<_ns ; i++)
    {
	_varyingLHS.ExtractRow(i,k,_vj);
	if (i!=nrowRand1_)
	{
	    for (unsigned int ii=0 ; ii<_nx ; ii++) // 
	    {
		dst=dst+(_vi[ii]-_vj[ii])*(_vi[ii]-_vj[ii]);
	    }         
         
	    _varyingDist(i,nrowRand1_)=_varyingDist(nrowRand1_,i)=pow(dst,0.5);
	    dst=0;
	}
         
    }
                 
    _varyingLHS.ExtractRow(nrowRand2_,k,_vi);
    for (unsigned int i=0; i<_ns ; i++)
    {
	_varyingLHS.ExtractRow(i,k,_vj);
	if (i!=nrowRand2_)
	{
	    for (unsigned int ii=0 ; ii<_nx ; ii++)
	    {
		dst=dst+(_vi[ii]-_vj[ii])*(_vi[ii]-_vj[ii]);
	    }
         
	    _varyingDist(i,nrowRand2_)=_varyingDist(nrowRand2_,i)=pow(dst,0.5);
	    dst=0;	    
	}

    }  

    Double_t Updated_LPnorm;
    Updated_LPnorm=getLPNorm(_varyingDist);
   
    return Updated_LPnorm;
}


Double_t TMaxiMinLHS::getLPNorm(const TMatrixD &Dist_)
{

    for (unsigned int i=0 ; i<_ns ; i++)
    {
	_fp(i,i)=0; 
	for (unsigned int j=i+1 ; j<_ns; j++)
        {
	    _fp(i,j)=1./(pow(Dist_(i,j),50));
	}
    }

    // _fp.Print();
    return pow(_fp.Sum(),1./50.);

}

Double_t TMaxiMinLHS::getMinDist(const TMatrixD &lhs)
{
    const unsigned int  Nrow=lhs.GetNrows();
    TMatrixD Dist(Nrow,Nrow);

    getDistMatrix(Dist, lhs); // creation of the Dist matrix
    for (unsigned int i=0; i<Nrow ; i++)
    {
	Dist(i,i)= DLE_MAX;
    }
    //Dist.Print();
    return Dist.Min();
}

void TMaxiMinLHS::getDistMatrix(TMatrixD &Dist, const TMatrixD &lhs)
{

    unsigned int k=0;
    unsigned int Nrow=lhs.GetNrows();
    const unsigned int Ncol=lhs.GetNcols();
    vector<double> vi, vj; vi.resize(Ncol); vj.resize(Ncol);
    double dst=0;
    
    for (unsigned int i=0 ; i<Nrow ; i++)
    {
	Dist(i,i)=0;  //diag to 0 value
	for (unsigned int j=i+1 ; j<Nrow ; j++)
	{
	    lhs.ExtractRow(i,k,&vi[0]);
	    lhs.ExtractRow(j,k,&vj[0]);
	    for (unsigned int ii=0 ; ii<Ncol ; ii++)
	    {
		dst=dst+(vi[ii]-vj[ii])*(vi[ii]-vj[ii]);
	    }
	    Dist(i,j)=dst;
	    Dist(j,i)=Dist(i,j);
	    dst=0;
        }
    }

    Dist.Sqrt(); //The matrix of L2-distance is constructed

}


void TMaxiMinLHS::Permutation(unsigned int & ncolRand, unsigned int & nrowRand1, unsigned int & nrowRand2)
{  
    _varyingLHS=_currentLHS;

    ncolRand=gRandom->Integer(_nx);
    nrowRand1=gRandom->Integer(_ns);
    nrowRand2=gRandom->Integer(_ns);
    while(nrowRand2==nrowRand1)
	nrowRand2=gRandom->Integer(_ns);

    Double_t auxi=_varyingLHS(nrowRand1,ncolRand);
    _varyingLHS(nrowRand1,ncolRand)=_varyingLHS(nrowRand2,ncolRand);
    _varyingLHS(nrowRand2,ncolRand)=auxi;

}

TMatrixD TMaxiMinLHS::maximinSA_LHS(const TMatrixD &lhs)
{

    unsigned int I=0, row1=0, row2=0, col=0, flag;
    Double_t ref;

    TMatrixD lhs_best=lhs;
    TMatrixD Dist_best(_ns,_ns);
    
    _currentLHS=lhs;
    _varyingLHS=lhs;
    getDistMatrix(_currentDist,_currentLHS);
    
    _varyingDist=_currentDist;
    
    Double_t T=_SAinitTemp;
    Double_t diff, lp_norm, updated_lp_norm;

    lp_norm=getLPNorm(_currentDist);
    ref=lp_norm;

    for (unsigned int i=0; i<_SAN ; i++)
    {
	flag=0;
	I=1;
	if(_blog)
	    cout <<"============================================================================= "<< i << " T "<<T<<"  mindist "<<getMinDist(lhs_best)<<" ref "<<ref<<endl;
	
	while (I<=_SAinnerN)
	{

	    Permutation(col,row1,row2),
	    updated_lp_norm=getUpdatedLPNorm(lp_norm,row1,row2); 

	    diff=min(exp((lp_norm-updated_lp_norm)/T),1.);

	    if (diff==1)
	    {
		lp_norm=updated_lp_norm;
		_currentLHS=_varyingLHS;
		_currentDist=_varyingDist;
		flag=1;
		if (lp_norm<ref)
		{
		    ref=lp_norm;
		    lhs_best=_currentLHS;
		    Dist_best=_currentDist;
                              
		    I=1;
		}
		else
		{
		    I=I+1;
		}
	    }
	    else 
	    {
		bool a=0;
		a=gRandom->Binomial(1,diff);
		if(a==1)
		{
		    lp_norm=updated_lp_norm;
		    _currentLHS=_varyingLHS;
		    _currentDist=_varyingDist;
		    flag=1;
		    if (lp_norm<ref)
		    {
			ref=lp_norm;
			lhs_best=_currentLHS;
			Dist_best=_currentDist;
			I=1;
		    }
		    else
		    {
			I=I+1;
		    }
		}
		else
		{
		    I=I+1;
		    _varyingDist=_currentDist;
		    _varyingLHS=_currentLHS;
		}
	    }
		     

	}
	if (flag==1)
	{
	    T=T*_SACoeff;
	}
	else
	{
//	    cout << "stop" << endl;
	    break;
	}     
    } 

    return lhs_best;       
}

void TMaxiMinLHS::transformMatrix(TMatrixD &RefMat, bool toProbaSpace)
{
    
    URANIE::DataServer::TStochasticAttribute *curdist;
    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0, is=0;
    TMatrixD changed_mat(_nS,_nx);
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        for (is = 0; is < _nS; is++)
            changed_mat(is, iatt) = ( ( toProbaSpace) ? curdist->getCDF(RefMat(is, iatt)) : curdist->getInvCDF(RefMat(is, iatt)) );

        iatt++;
    }

    RefMat = changed_mat;
}

void TMaxiMinLHS::generateSample(Option_t *option)
{

    TString soption = option;
    soption.ToLower();

    parseOption(option);
   
    if( !_bAlreadyFilled )
    {
	_sampling->generateSample();
    }

    _currentLHS = _tds->getMatrix();
    transformMatrix(_currentLHS, true);
    // _currentLHS.Print();

    TMatrixD lhs_optim(_ns,_nx);
    lhs_optim=maximinSA_LHS(_currentLHS);
    transformMatrix(lhs_optim, false);
 
    // Fill the TTree
    Double_t * valPattern = new Double_t[_nx + 1];
    _tds->deleteTuple();
    createTuple();

    if (_blog)
        cout << "Filling the TTree..." << endl;

    for (unsigned int is = 0; is < _ns; is++)
    {
        valPattern[0] = (Double_t)(is + 1);
        for (unsigned int ix = 0; ix < _nx; ix++)
            valPattern[ix + 1] = lhs_optim(is, ix);
        _ntsample->Fill(valPattern);
    }
    delete[] valPattern;
    valPattern = NULL;

    if (_blog)
        cout << "TTree filling done" << endl;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    if ( !_bAlreadyFilled )
	fillOtherAttributes();
}

void TMaxiMinLHS::fillOtherAttributes()
{
    _sampling->fillOtherAttributes();
}

///////////////////////////////////////////////////////////
//  Create the URANIE::DataServer::TDSTNtupleD of data
///////////////////////////////////////////////////////////
void TMaxiMinLHS::createTuple()
{
    try
    {
        TString sLabelVariable = _tds->getIteratorName();
        if (_blog)
            cout << endl
                    << "**************************\n*** TMaxiMinLHS::createTuple sLabelVariable ["
                    << sLabelVariable << "]" << endl;

        if (_nX != _lstOfAttributesToSample->GetEntries() && !_bAlreadyFilled )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMaxiMinLHS::createTuple: The number of stochastic attributes (%d) has been modified (%d).",
                            _nX, _lstOfAttributesToSample->GetEntries()));
        if (_blog)
            cout << " ** TMaxiMinLHS::createTuple Size _nX [" << _nX << "]"
                    << endl;

        // Loop on stochastic attributes to have their names
        URANIE::DataServer::TAttribute *att;
	TList *attList = _tds->getListOfAttributes();
        TIter next(((!_bAlreadyFilled) ? _lstOfAttributesToSample : attList ) );
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            sLabelVariable += Form(":%s", att->GetName());
            if (_blog)
                cout << " ** stoatt[" << att->GetName() << "] sLabelVariable["
                        << sLabelVariable << "]" << endl;
        }

        if (_blog)
            cout << " ** Create the TTree (" << sLabelVariable
                    << ") of data ...";

        //// Create the TTree of data
        //// ngi -- essai d'economie de memoire pour les grands nombres d'attributs
        //Int_t bufsize = 32000;
        //if(_nX > 1000)
        //    bufsize = bufsize / (_nX/1000);
        ////

	// cout<<__LINE__<<" the ntulpe is "<<_ntsample<<endl;
	if(_ntsample!=NULL)
	  delete _ntsample;
	
        _ntsample = new URANIE::DataServer::TDSNtupleD(Form("Sampling_%s", GetName()), GetTitle(), sLabelVariable);

        if (_blog)
            cout
                    << "*** End Of TMaxiMinLHS::createTuple\n**************************"
                    << endl;

        delete attList;
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        if( _vi != NULL)
            delete[] _vi;
        _vi=NULL;
        if( _vj != NULL)
            delete[] _vj;
        _vj=NULL;
        throw ue;
    }
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TMaxiMinLHS::printLog(Option_t *option)
{
    cout << "** TMaxiMinLHS::printLog  [" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nx   [" << _nx << "]" << endl;

}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
