//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/*! \file TMelange.cxx
 * \brief Implementation of the class representing a model of mixing of normal laws.
 *
 *
 *  The implemented methods allow to sample according to a given mixing, to increase this mixing or to simulate a single observation. 
 *
 */
#include "TMelange.h"

#ifndef ROOT_TRandom3
#include "TRandom3.h"
#endif

#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif

#include "TMath.h"

void TMelange::_printlog()
{
    cout << "**** log of the object[TMelange]" << endl;
    cout << " Attr taille[" << _taille << "]" << endl;
    cout << " Attr tailleMelange[" << _tailleMelange << "]" << endl;
    cout << "**** fin log de l'objet[TMelange]" << endl << endl;
}

TMelange::TMelange(Int_t n)
{
    f = new TNormale();
    _taille = 0;
    lLoi = new TList();
    _tailleMelange = 0;
    _dimMelange = n;
    _rand = new TRandom3();
    _rand->SetSeed(0);
    _rand->Uniform(1.0);
    TString sLabelVariable(5 * _dimMelange - 1);
    sLabelVariable += "indice:";
    for (Int_t i = 1; i < _dimMelange; i++)
    {
        sLabelVariable += "var";
        sLabelVariable += i;
        sLabelVariable += ":";
    }
    sLabelVariable += "var";
    sLabelVariable += _dimMelange;
    _nt = new TNtupleD("melange", "mon melange", sLabelVariable.Data());
    //  rootFile = new TFile("TMelange.root", "RECREATE");
}

TMelange::~TMelange()
{
    delete lLoi;
    lLoi = NULL;
}

Double_t TMelange::eval(TVectorD X)
{
    Double_t Eval = 0;
    for (Int_t i = 0; i < _tailleMelange; i++)
    {
        f->setMean(getMean(i));
        f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(i));
        Eval = Eval + getProb(i) * (f->eval(X));
    }
    return Eval;
}

Double_t TMelange::eval(Double_t x)
{
    Double_t Eval = 0;
    for (Int_t i = 0; i < _tailleMelange; i++)
    {
        f->setMean(getMean(i));
        f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(i));
        Eval = Eval + getProb(i) * (f->eval(x));
    }
    return Eval;
}

Double_t TMelange::getProb(Int_t i)
{
    Double_t p = 0;
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing  -1"
                << endl;
        return 0;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        p = list->getProb();
    }
    return p;
}

TVectorD TMelange::getMean(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing -1"
                << endl;
        TVectorD x(1);
        x = 0;
        return x;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        return (list->getMean());
    }
}

TVectorD TMelange::getMean()
{
    TVectorD moyenne(_dimMelange);
    moyenne.Zero();
    TVectorD M(_dimMelange);
    for (Int_t i = 0; i < _tailleMelange; i++)
    {
        M = getMean(i);
        M *= getProb(i);
        moyenne += M;
    }
    return moyenne;
}

TVectorD TMelange::getVariance()
{
    TMatrixD A = getMatCovariance();
    TMatrixDDiag_const B(A);
    TVectorD var(B);
    return var;
}

TVectorD TMelange::getSigma()
{
    TVectorD sigma = getVariance();
    for (Int_t i = 0; i < _dimMelange; i++)
    {
        sigma(i) = sqrt(sigma(i));
    }
    return sigma;
}

TMatrixD TMelange::getMatEcartTypeCorrelation(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing -1"
                << endl;
        TMatrixD x(1, 1);
        x = 0;
        return x;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        return (list->getMatEcartTypeCorrelation());
    }
}

TVectorD TMelange::getSigma(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing -1"
                << endl;
        TVectorD x(1);
        x = 0;
        return x;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        return (list->getSigma());
    }
}

TMatrixD TMelange::getMatCorrelation(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing -1"
                << endl;
        TMatrixD x(1, 1);
        x = 0;
        return x;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        return (list->getMatCorrelation());
    }
}

TMatrixD TMelange::getMatCovariance(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing -1"
                << endl;
        TMatrixD x(1, 1);
        x = 0;
        return x;
    }
    else
    {
        TNormale* list = 0;
        TIter next_list(lLoi);
        Int_t j = 0;
        while ((list = (TNormale*) next_list()) && (j < i))
        {
            j++;
        }
        return (list->getMatCovariance());
    }
}

TMatrixD TMelange::getMatCovariance()
{
    TMatrixD matCovariance(_dimMelange, _dimMelange);
    matCovariance.Zero();
    TMatrixD A(_dimMelange, _dimMelange);
    for (Int_t i = 0; i < _tailleMelange; i++)
    {
        A = getMatCovariance(i);
        A *= getProb(i);
        matCovariance += A;
    }
    return matCovariance;
}

TMatrixD TMelange::getMatCorrelation()
{
    TMatrixD matCorr = getMatCovariance();
    TVectorD sigma = getSigma();
    for (Int_t i = 0; i < _dimMelange; i++)
        for (Int_t j = 0; j < _dimMelange; j++)
            matCorr(i, j) = matCorr(i, j) / (sigma(i) * sigma(j));
    return matCorr;
}

TMatrixD TMelange::getMatEcartTypeCorrelation()
{
    TMatrixD ecartTypeCorr = getMatCorrelation();
    TVectorD sigma = getSigma();
    for (Int_t i = 0; i < _dimMelange; i++)
        ecartTypeCorr(i, i) = sigma(i);
    return ecartTypeCorr;
}

TVectorD TMelange::getSeuil()
{
    TVectorD vectSeuil(_tailleMelange + 1);
    vectSeuil.Zero();

    Int_t i = 0;

    TNormale* list = 0;
    TIter next_list(lLoi);
    while ((list = (TNormale*) next_list()))
    {
        vectSeuil(i + 1) = vectSeuil(i) + list->getProb();
        i++;
    }
    return vectSeuil;
}

/*! \fn void  TMelange::simulationMelange()
 \brief Classic simulation of a mixing
 
 In a firt step, the classic simulation consists in selecting ramdomly an integer, then choosing a sample corresponding to the law which contains the index.
 The law is recovered from thresholds \f$ 0 < s_{1} < \cdots < s_{N} = 1.0\f$.
 For each law, we count the number of times \f$ l\f$ that an integer \f$ U \in [0, 1]\f$ is located betweeen
 \f$ s_{l} < u < s_{l+1}\f$.
 Finally, for each law \f$ l\f$, a \f$ n_{l}\f$ sample is chosen, with the constraint that the sum \f$ \sum_{l}{n_{l}}\f$ is equal to the desired size.

 */
void TMelange::simulationMelange()
{
    // variables pour generer va uniforme sur [0;1]
    Float_t U;

    // variables pour calculer les partitions de [0;1]
    TVectorD seuil = getSeuil();

    // variables servant d'indices dans le parcours de la liste chainee  
    Int_t n = 0;

    // tableau d'effectif: nb de simulation pour chaque loi du melange
    Int_t * nsimu = new Int_t[_tailleMelange];
    for (Int_t k = 0; k < _tailleMelange; k++)
        nsimu[k] = 0;

    for (Int_t j = 0; j < _taille; j++)
    {
        U = (Float_t) _rand->Uniform(1.0);
        for (Int_t i = 0; i < seuil.GetNrows(); i++)
        {
            if ((U > seuil(i)) && (U <= seuil(i + 1)))
                nsimu[i]++;
        }
    }

    TNormale* list = 0;
    TIter next_list(lLoi);
    while ((list = (TNormale*) next_list()))
    {
        list->setTaille(nsimu[n]);
        list->acceptationRejet();
        TNtupleD* nt = list->getTuple();

        addTuple(nt);
        n++;
    }
}

Double_t TMelange::getObs()
{
    Int_t i = 0;
    TVectorD seuil = getSeuil();
    TNormale* element = 0;
    TIter suivant(lLoi);
    Double_t unif = _rand->Uniform(1.0);
    while ((element = (TNormale*) suivant()))
    {
        if ((unif <= seuil(i + 1)) && (unif > seuil(i)))
        {
            return element->getObs();
        }
        i++;
    }

    return 0;
}

void TMelange::addTuple(TNtupleD* t)
{

    Int_t nPat = (Int_t)(t->GetEntries());
    Int_t nCrt = (Int_t)(_nt->GetEntries());

    for (Int_t i = 0; i < nPat; i++)
    {
        t->GetEntry(i);
        Double_t* val = t->GetArgs();
        val[0] += nCrt + 1;
        _nt->Fill(val);
    }

}

void TMelange::addLoi(TVectorD moy, TMatrixD ecartTypeCorr, Double_t proba)
{
    TNormale* g = new TNormale(moy, ecartTypeCorr);
    g->setProb(proba);
    if (g->getDim() == _dimMelange)
    {
        lLoi->Add(g);
        _tailleMelange++;
    }
    else
    {
        cout << "Not the right dimension at this position " << _tailleMelange
                << endl;
    }
}

void TMelange::removeLoi(Int_t i)
{
    if ((i < 0) && (i >= _tailleMelange))
    {
        cout
                << "The position must be set between 0 and the size of the mixing  -1"
                << endl;
    }
    else
    {
        lLoi->RemoveAt(i);
        _tailleMelange--;
    }
}

void TMelange::reset()
{
    lLoi->RemoveAll();
    _tailleMelange = 0;
}

/*! \fn TVectorD TMelange::LoiDeZ(TVectorD X)
 \brief Computes the conditional repartition function for X belonging to a mode.
 */
TVectorD TMelange::LoiDeZ(TVectorD X)
{
    TVectorD proba(_tailleMelange);
    f->setMean(getMean(0));
    f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(0));
    proba(0) = (getProb(0) * (f->eval(X))) / eval(X);
    for (Int_t i = 1; i < _tailleMelange; i++)
    {
        f->setMean(getMean(i));
        f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(i));
        proba(i) = proba(i - 1) + (getProb(i) * (f->eval(X))) / eval(X);
    }
    return proba;
}

/*! \fn TVectorD TMelange::discrimination(TVectorD X)
 \brief  Computes for each mode, the 'a posteriori' probability that X belongs to the mode.
 Enables to undertake a discriminant analysis.
 */
TVectorD TMelange::discrimination(TVectorD X)
{
    TVectorD proba(_tailleMelange);
    f->setMean(getMean(0));
    f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(0));
    for (Int_t i = 0; i < _tailleMelange; i++)
    {
        f->setMean(getMean(i));
        f->setMatEcartTypeCorrelation(getMatEcartTypeCorrelation(i));
        proba(i) = (getProb(i) * (f->eval(X))) / eval(X);
    }
    return proba;
}

/*! \fn void TMelange::remplacer(Int_t i, TVectorD M, TMatrixD A)
 \brief Replaces the component i by a component  N(M,A)
 */
void TMelange::remplacer(Int_t i, TVectorD M, TMatrixD A)
{
    if ((M.GetNrows() != _dimMelange) || (A.GetNrows() != _dimMelange))
        cout << "wrong dimension" << endl;
    else
    {
        TNormale* g = new TNormale(M, A);
        g->setProb(getProb(i));
        lLoi->AddAt(g, i);
    }
}

/*! \fn void TMelange::afficher()
 \brief Prints the characteristics of mixing modes.
 */
void TMelange::afficher()
{
    for (Int_t i = 0; i < getTailleMelange(); i++)
    {
        cout << "   " << endl;
        cout << "Component" << i + 1 << endl;
        getMean(i).Print();
        getMatEcartTypeCorrelation(i).Print();
        cout << "####################################" << endl;
        cout << "####################################" << endl;

    }
}

