//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOATDesign.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr> 
 \date   2012.08.10

 \brief  "One factor At a Time" design of experiment. It has been developped 
 mainly for finite differences and sensitivity computation.


 */
////// ROOT
#include "TMath.h"
#include "TLeaf.h"
#include "TROOT.h"
#include "TRandom3.h"
#include "TObjString.h"

////// URANIE
#include "TOATDesign.h"
#include "TAttribute.h" 
#include "TStochAttribut.h"
#include "UExceptions.h" 

using namespace URANIE::DataServer;
using namespace std;

ClassImp (URANIE::Sampler::TOATDesign)

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TOATDesign::TOATDesign(URANIE::DataServer::TDataServer *tds, TString mode,
        Int_t number) :
        TSamplerStochastic(tds, "", 1), _dStepPrecision(dStepPrecision)
{
    _nominalValueTuple = NULL;

    try
    {
        // Check that the number of modifications is valid
        if (number <= 0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::TOATDesign: the given number of modifications [%d] is negative or null",
                            number));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Check if the tds is empty 
    _bHasData = kTRUE;
    try
    {
        _tds->getTuple(); // raises a warning exception if the tds is empty
    } catch (URANIE::Exceptions::UWarningExceptions&)
    {
        _bHasData = kFALSE;
    }

    // Set the mode and number of modifications.
    setOption(mode, number);

    // We insure that the TObjString objects created on the fly and stored 
    // in the TMap object will be properly deleted with it. 
    _infoRangeMap.SetOwnerKeyValue();
}

TOATDesign::~TOATDesign()
{
    if (_blog)
        cout << "TOATDesign::~TOATDesign: destruction of TOATDesign" << endl;

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	if (_nominalValueTuple)
	    delete _nominalValueTuple;
    }
}

void TOATDesign::setOption(TString mode, Int_t number)
{
    // The mode string is formatted to avoid extra spaces and capital letters.
    mode.ToLower();
    mode.ReplaceAll(" ", "");

    // Setting the sampling mode
    try
    {
        if (mode.Contains("reg"))
        {
            _sMode = "regular";
            _sGenMode = "range";
        }
        else if (mode.Contains("lhs"))
        {
            _sMode = "random";
            _sGenMode = "lhs";
        }
        else if (mode.Contains("srs"))
        {
            _sMode = "random";
            _sGenMode = "srs";
        }
        else if (mode.Contains("rand"))
        {
            _sMode = "random";
            _sGenMode = "srs";
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setOption: unknown mode [%s]",
                            mode.Data()));
    } catch (URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Setting the number of modifications 
    try
    {
        if (number < 2)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setOption: the given number of modification [%d] is too small. Its minimum value is 2",
                            number));
        else if ((_sMode == "regular") && (number % 2 > 0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setOption: the given number of modification [%d] is odd. Only even values are accepted in %s mode.",
                            number, _sMode.Data()));
        else
            _nNbModif = number;
    } catch (URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TOATDesign::setRange(TString changingAttName, Double_t rangeValue,
        Bool_t is_percent)
{
    try
    {
        // Check if the given attribute name exists in the data server
        if (!_tds->isAttribute(changingAttName))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: unknown attribute %s",
                            changingAttName.Data()));

        // The attribute exists in the data server, thus we can retrieve it
        TAttribute *att = _tds->getAttribute(changingAttName);
        Double_t defVal;

        // In case the data server is empty, check that the attribute 
        // has a default value.
        if (!_bHasData && !att->getDefaultValue(defVal))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: attribute %s has no default value",
                            changingAttName.Data()));

        // Check the validity of the range
        if (rangeValue <= 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: range value must be strictly positive [%e]",
                            rangeValue));

        // In case we are in "random" mode, assert that the attribute 
        // is a TStochasticAttribute object.
        if ((_sMode == "random")
                && !att->InheritsFrom(TStochasticAttribute::Class()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: random mode requires that changing attributes be TStochasticAttribute objects. Attribute %s isn't.",
                            changingAttName.Data()));

        // Verify that the attribute is not already in the list.
        if (_changingAttList.FindObject(att) != 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: attribute %s is already in the list of changing attributes.",
                            changingAttName.Data()));

        // Everything is fine, so we can add the attribute to the list 
        // of changing attributes.
        _changingAttList.Add(att);

        // TMap objects only take TObject objects as key and value. 
        // We thus need to use TObjString instead of TString.
        TObjString* key = new TObjString(changingAttName.Data());

        // To keep floating point precision when converting to 
        // string, we need scientific notation and a large number 
        // of significant digits (e.g. %.16e). 
        TObjString* value =
                is_percent ?
                        new TObjString(Form("%.16e%%", rangeValue)) :
                        new TObjString(Form("%.16e", rangeValue));
        _infoRangeMap.Add(key, value);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TOATDesign::setRange(TString changingAttName, TString rangeAttName,
        Bool_t is_percent)
{
    try
    {
        // Check if the given changing attribute name exists in the data server
        if (!_tds->isAttribute(changingAttName))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: unknown attribute %s",
                            changingAttName.Data()));

        // Check if the given range attribute name exists in the data server
        if (!_tds->isAttribute(rangeAttName))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: unknown range attribute %s",
                            rangeAttName.Data()));

        // The attributes exist in the data server, so we can retreive them
        TAttribute *att = _tds->getAttribute(changingAttName);
        TAttribute *rangeAtt = _tds->getAttribute(rangeAttName);
        Double_t defVal;

        // In case the data server is empty, check that the changing attribute 
        // has a default value.
        if (!_bHasData && !att->getDefaultValue(defVal))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: attribute %s has no default value",
                            changingAttName.Data()));

        // In case the data server is empty, check that the range attribute 
        // has a positive default value.                
        if (!_bHasData)
        {
            Bool_t has_default_value = rangeAtt->getDefaultValue(defVal);
            if (!has_default_value)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "*** ERROR in TOATDesign::setRange: range attribute %s has no default value",
                                rangeAttName.Data()));
            else if (defVal <= 0.0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "*** ERROR in TOATDesign::setRange: range attribute %s has a negative or null default value [%f]",
                                rangeAttName.Data(), defVal));
        }

        // In case we are in "random" mode, assert that the changing 
        // attribute is a TStochasticAttribute object.
        if ((_sMode == "random")
                && !att->InheritsFrom(TStochasticAttribute::Class()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: random mode requires that changing attributes be TStochasticAttribute objects. Attribute %s isn't.",
                            changingAttName.Data()));

        // Verify that the changing attribute is not already in the list.
        if (_changingAttList.FindObject(att) != 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATDesign::setRange: attribute %s is already in the list of changing attributes.",
                            changingAttName.Data()));

        // Everything is fine, so we can add the attribute to the list 
        // of changing attributes.                
        _changingAttList.Add(att);

        // TMap objects only take TObject objects as key and value. 
        // We thus need to use TObjString instead of TString.
        TObjString* key = new TObjString(changingAttName.Data());
        TObjString* value =
                is_percent ?
                        new TObjString(Form("%s%%", rangeAttName.Data())) :
                        new TObjString(rangeAttName.Data());
        _infoRangeMap.Add(key, value);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TOATDesign::init()
{
    // Save the nominal values in a TDSNtupleD object and reset the 
    // TDSNtupleD of the data server
    if (_bHasData)
    {
        // The tds contains data => the values stored will be used as 
        // nominal values. 

        // Create the two additionnal attributes and give them values. 
        // __modified_att__ will always be -1 (i.e. no modified att), 
        // while __nominal_set__ will have values from 1 to the number 
        // of events in the tuple. To achieve this easily, 
        // we create these two attributes as TAttributeFormula. Their 
        // "formula" property will be lost when we reset the tuple.                                
        _tds->addAttribute("__nominal_set__",
                Form("%s", _tds->getIteratorName()));
        _tds->addAttribute("__modified_att__", "-1.0");

        // Save the nominal values
        _nominalValueTuple = _tds->getTuple();
        // Discard the tds' tuple
        _tds->setTuple((TDSNtupleD *) NULL);
        // Create a new (empty) one
        _tds->createTuple();
    }
    else
    {
        // The tds is empty => the nominal values are the attributes' 
        // default values

        // Create the internal attributes which identify which set of 
        // nominal values is used (__nominal_set__) and which attribute 
        // is modified (__modified_att__)            
        TAttribute *attNominalSet = new TAttribute("__nominal_set__", "");
	attNominalSet->delShare();
        TAttribute *attModifiedAtt = new TAttribute("__modified_att__", "");
	attModifiedAtt->delShare();
	
        // Default value of the nominal set is 1 because there is only 
        // one set of nominal variables.
        attNominalSet->setDefaultValue(1.0);
        // Default value of the modified attribute is -1 because by 
        // default, no attribute is modified.
        attModifiedAtt->setDefaultValue(-1.0);

        _tds->addAttribute(attNominalSet);
        _tds->addAttribute(attModifiedAtt);

        // Create the new data tree inside the tds, and prepare to store 
        // the nominal values in another one.
        _tds->createTuple();
        _nominalValueTuple = _tds->getTuple();
        // Discard the tds' tuple
        _tds->setTuple((TDSNtupleD *) NULL);
        // Create a new one
        _tds->createTuple();

        // Set the default value of the tds' iterator. This allows us to 
        // manage all the attributes in the same way when setting their 
        // values.
        _tds->getAttribute(_tds->getIteratorName())->setDefaultValue(1.0);

        // Fill the _nominalValueTuple with the default values.
        // We use the list of leaves instead of the list of attributes 
        // because the "getListOfAttributes" function doesn't return the 
        // hidden attributes (like the iterator). By parsing the leaves, 
        // we are sure to access to every attribute.
        // NOTE: this only works if the name of the leaf is the same as 
        //       the name of the corresponding attribute. If this "feature" 
        //       was to change in URANIE, we will have to reimplement this 
        //       procedure (an exception will be thrown, anyway).
        TObjArray* nomLeavesList = _nominalValueTuple->GetListOfLeaves();
        Int_t nbLeaves = nomLeavesList->GetEntries();
        Double_t *values = new Double_t[nbLeaves];
        for (Int_t i = 0; i < nbLeaves; i++)
        {
            TString attName = nomLeavesList->At(i)->GetName();
            TAttribute *att = _tds->getAttribute(attName);
            try
            {
                if( ! att->hasDefaultValue() )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form(
                                                                   "ERROR in TOATDesign the attribute [%s] has no default value. Please set one", attName.Data()));
            
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
            
            att->getDefaultValue(values[i]);            
        }
        _nominalValueTuple->Fill(values);
        delete[] values;
        values = 0;
    }
}

void TOATDesign::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    init();

    // We use the list of leaves instead of the list of attributes 
    // because the "getListOfAttributes" function doesn't return the 
    // hidden attributes (like the iterator). By parsing the leaves, 
    // we are sure to access to every attribute.
    // NOTE: this doesn't work if the name of the leaf is different 
    //       from the name of the corresponding attribute.
    TObjArray *leavesList = _nominalValueTuple->GetListOfLeaves();

    Int_t nbValues = leavesList->GetEntries();
    Double_t *values = new Double_t[nbValues];

    // The order of data inside the data server is changed. Thus, we 
    // need to set the iterator value "manually".
    Double_t iteratorValue = 0.0;

    // Saving some useful indices. These indices are a priori unknown, 
    // so we need to retrieve them.            
    Int_t iteratorIndex = leavesList->IndexOf(
            _nominalValueTuple->GetLeaf(_tds->getIteratorName()));
    Int_t modifAttIndex = leavesList->IndexOf(
            _nominalValueTuple->GetLeaf("__modified_att__"));

    // Get a handle on the data server's tuple. 
    TDSNtupleD *tdsTuple = _tds->getTuple();
    
    for (Int_t nomSet = 1; nomSet <= _nominalValueTuple->GetEntries(); nomSet++)
    {
        // We loop over the nominal sets. The first line we add to the data 
        // server will be the nominal values of all the attributes (except 
        // for the iterator) for the corresponding set. 
        _nominalValueTuple->GetEntry(nomSet - 1);
        iteratorValue += 1.0;
        values[iteratorIndex] = iteratorValue;
        for (Int_t i = 0; i < nbValues; i++)
        {
            if (i == iteratorIndex)
                continue;
            else
                values[i] = ((TLeaf*) leavesList->At(i))->GetValue();
        }
        tdsTuple->Fill(values);

        // Next step is the loop over the changing attributes.
        Double_t dRange = 0.0;
        Double_t dBackupNominal = 0.0;
        Bool_t bIsPercent = kFALSE;

        // Prepare a list of indices for the random sampling
        Int_t *nInd = new Int_t[_nNbModif];
        for (Int_t i = 0; i < _nNbModif; i++)
            nInd[i] = i;

        TIter next(&_changingAttList);
        TAttribute *att = (TAttribute*) next();
        while (att)
        {
            TString attName = att->GetName();
            // Retrieve the (real) attribute index (the one inside the TDSNtupleD)
            Int_t attIndex = leavesList->IndexOf(
                    _nominalValueTuple->GetLeaf(attName.Data()));

            // Backup the attribute's nominal value
            dBackupNominal = values[attIndex];

            // Retrieve the range information.
            try
            {
                if( !_infoRangeMap.Contains(attName.Data()) )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form(
                                                                   "ERROR in TOATDesign the attribute [%s] has no range value. Please set one", attName.Data()));
            
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
               
            TString rangeInfo = _infoRangeMap(attName.Data())->GetName();

            // Check if the range must be interpreted as a percentage of 
            // the nominal value.
            if (rangeInfo.Contains('%'))
            {
                bIsPercent = kTRUE;
                // The "%" was added by us in the setRange function, so 
                // we have to remove it.
                rangeInfo.Remove(rangeInfo.Length() - 1);
            }
            else
                bIsPercent = kFALSE;

            // Retrieve the range value. The string can be interpreted  
            // either as a numerical value or as an attribute name.
            if (rangeInfo.IsFloat())
                dRange = rangeInfo.Atof();
            else
            {
                Int_t attRangeIndex = leavesList->IndexOf(
                        _nominalValueTuple->GetLeaf(rangeInfo.Data()));
                dRange = values[attRangeIndex];
            }

            if (bIsPercent) {
                dRange = dRange * dBackupNominal / 100.0;

                // Case of dRange is lower than StepPrecision (as dBackupNominal == 0.0)
                //   then dRange = getStepValue ELSE dStepPrecision parameter
                if (( TMath::Abs(dRange) < _dStepPrecision ) &&
                        ( !att->getStepValue(dRange) )) {
                        dRange = _dStepPrecision;
                }
            }

            // Now we proceed with data generation.
            if (_sMode == "regular")
            {
                // Compute the step
                Double_t dStep = dRange / _nNbModif;

                // Set the value of the __modified_att__ attribute to the 
                // "attribute index" inside the data server. This might 
                // be different from the index in the list of leaves, but 
                // is easier to access for the user.
                values[modifAttIndex] = _tds->getAttributeIndex(att);

                // Compute the values of the modified attribute and fill 
                // the data server. To simplify the process, the values 
                // are generated by pair, around the nominal value.                        
                for (Int_t i = 0; i < _nNbModif / 2; i++)
                {
                    // Lower bound
                    iteratorValue += 1.0;
                    values[iteratorIndex] = iteratorValue;
                    values[attIndex] = dBackupNominal - dRange / 2.0
                            + i * dStep;
                    _tds->getTuple()->Fill(values);
                    // Upper bound
                    iteratorValue += 1.0;
                    values[iteratorIndex] = iteratorValue;
                    values[attIndex] = dBackupNominal + dRange / 2.0
                            - i * dStep;
                    _tds->getTuple()->Fill(values);
                }
            }
            else if (_sMode == "random")
            {
                // Consider the generic attribute object as a TStochasticAttribute. 
                // This provides access to the getInvCDF function to generate 
                // values from a probability.

                TStochasticAttribute *distrib = (TStochasticAttribute*) att;

                // Backup the attribute's boundaries and adapt them to 
                // the range value.
                //Double_t dBackupLowerBound = distrib->getLowerBound();
                //Double_t dBackupUpperBound = distrib->getUpperBound();
                //distrib->setUpperBound(dBackupNominal + dRange/2.0);
                //distrib->setLowerBound(dBackupNominal - dRange/2.0);
                Double_t pMin = distrib->getCDF(dBackupNominal - dRange / 2.0);
                Double_t pMax = distrib->getCDF(dBackupNominal + dRange / 2.0);
                
                // NOTE: the following lines are heavily inspired from 
                //       TSampling.
                // Take into account the LHS or SRS cases            
                Double_t dpas, dtype, dproba;
                if (_sGenMode == "lhs")
                {
                    //dpas = 1.0 / (Double_t)( _nNbModif );
                    dpas = (pMax - pMin) / (Double_t)(_nNbModif);
                    dtype = 1.0;
                }
                else
                {
                    //dpas = 1.0;
                    //Modification DI333
                    dpas = (pMax - pMin);
                    dtype = 0.0;
                    
                }

                // Shuffle the list of indices (Fisher-Yates)
                for (Int_t is = _nNbModif - 1; is > 0; is--)
                {
                    Int_t indtmp = (Int_t) gRandom->Integer(is + 1);
                    Int_t valtmp = nInd[indtmp];
                    nInd[indtmp] = nInd[is];
                    nInd[is] = valtmp;
                }

                // Set the value of the __modified_att__ attribute to the 
                // "attribute index" inside the data server. This might 
                // be different from the index in the list of leaves, but 
                // is easier to access for the user.                        
                values[modifAttIndex] = _tds->getAttributeIndex(att);

                // Generate the random values and fill the data server.                        
                for (Int_t is = 0; is < _nNbModif; is++)
                {
                    iteratorValue += 1.0;
                    values[iteratorIndex] = iteratorValue;

                    //dproba = dpas * gRandom->Rndm() + dtype * dpas * nInd[is];
                    dproba = dpas * gRandom->Rndm() + dtype * dpas * nInd[is]
                            + pMin;
                    values[attIndex] = distrib->getInvCDF(dproba);
                    _tds->getTuple()->Fill(values);
                }
                // Reset the boundary values.
                //distrib->setUpperBound(dBackupUpperBound);
                //distrib->setLowerBound(dBackupLowerBound);                        
            }

            // Retrieve the nominal value of the attribute
            values[attIndex] = dBackupNominal;

            // Load the next attribute                    
            att = (TAttribute*) next();
        }
        delete[] nInd;
        nInd = 0;
    }
    delete[] values;
    values = 0;
    terminate();
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TOATDesign::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "TOATDesign::printLog" << endl;
    cout << "Mode:                    [" << _sMode << "]" << endl;
    cout << "Generation mode:         [" << _sGenMode << "]" << endl;
    cout << "_dStepPrecision          [" << _dStepPrecision << "]" << endl;
    cout << "number of modifications: [" << _nNbModif << "]" << endl;
    cout << "---" << endl;
    cout << "List of changing attributes: " << endl;
    _changingAttList.ls();
    cout << "---" << endl;
    cout << "Range information: " << endl;
    TIter next(&_changingAttList);
    TAttribute *att = (TAttribute*) next();
    while (att)
    {
        TString attName = att->GetName();
        TString rangeInfo = _infoRangeMap(attName.Data())->GetName();
        cout << attName << ": " << rangeInfo << endl;
        att = (TAttribute*) next();
    }
    cout << "---" << endl;
    cout << "Nominal values tuple: " << endl;
    _nominalValueTuple->Scan("*");
    cout << "End of TOATDesign::printLog" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
