//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPlackettCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 18:08:40 CEST 2012
 
 \brief Implementation de la classe URANIE::Sampler::TPlackettCopula
 
 */
//// ROOT
#include "TRandom3.h"
#include "TMath.h"

////// URANIE
#include "TPlackettCopula.h"
#include "TStochAttribut.h"
#include "TUniformDistribution.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TPlackettCopula)

namespace URANIE
{
namespace Sampler
{
TPlackettCopula::TPlackettCopula(URANIE::DataServer::TDataServer *tds,
        Double_t dparam, Int_t nCalcul) :
        TArchimedianCopula(tds, dparam, nCalcul)
{
    if (_blog)
        cout << " ** constructor TPlackettCopula" << endl;
    try
    {
        if (_dparam < 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPlackettCopula constructor\n The parameter %e must be possitive",
                            _dparam));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TPlackettCopula::~TPlackettCopula()
{
    if (_blog)
        cout << " ** destroy TPlackettCopula" << endl;

    if (_lstOfAttributesToSample != NULL)
      delete _lstOfAttributesToSample;
    _lstOfAttributesToSample = NULL;
}

/*******************************************/
/*           Generate the U Matrix         */
/*******************************************/
void TPlackettCopula::generateUMatrix(Option_t *option)
{
    if (_blog)
        cout << " ** TPlackettCopula::generateUMatrix _dparam[" << _dparam
                << "] [" << option << "]" << endl;

    Double_t dproba;
    Int_t is;

    // Loop on Stochastics Attributes
    URANIE::DataServer::TStochasticAttribute *curdist;
    URANIE::DataServer::TUniformDistribution *guranieCopula =
            new URANIE::DataServer::TUniformDistribution("__guranie__copula__",
                    0.0, 1.0);

    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0;
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << curdist->GetName() << "] Type["
                    << curdist->getLawType() << "]" << endl;
        }

        // Generate the U-sample
        for (is = 0; is < _nS; is++)
        {
            //dproba = dpas * gRandom->Rndm() + dpmin;
            dproba = gRandom->Rndm();
            if (iatt == 0)
                _matUValue(is, iatt) = guranieCopula->getInvCDF(dproba);
            else
            {
                Double_t dvalu = _matUValue(is, iatt - 1);
                Double_t dvalt = gRandom->Rndm();
                Double_t dvala = dvalt * (1. - dvalt);
                Double_t dvalb = _dparam
                        + dvala * (_dparam - 1.) * (_dparam - 1.);
                Double_t dvalc = 2. * dvala
                        * (dvalu * _dparam * _dparam + 1. - dvalu)
                        + _dparam * (1. - 2. * dvala);
                Double_t dvald = TMath::Sqrt(_dparam);
                dvald *= TMath::Sqrt(
                        _dparam
                                + 4. * dvala * dvalu * (1. - dvalu)
                                        * (1. - _dparam) * (1. - _dparam));

                if (_blog)
                    cout << " ** dvalu[" << dvalu << "] dvalt[" << dvalt
                            << "] dvala[" << dvala << "] dvalb[" << dvalb
                            << "] dvalc[" << dvalc << "] dvald[" << dvald << "]"
                            << endl;
                _matUValue(is, iatt) = (dvalc - (1. - 2. * dvalt) * dvald)
                        / (2. * dvalb);
            }
            if (_blog)
                cout << "** is[" << is << "] -- proba [" << dproba
                        << "] -- value[" << _matUValue(is, iatt) << "]" << endl;
        }

        //Increment the index of attribute
        iatt++;
    }

    // Clean
    delete guranieCopula;
    guranieCopula = NULL;

    if (_blog)
        cout << " ** End Of TPlackettCopula::generateUMatrix[" << option << "]"
                << endl;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TPlackettCopula::printLog(Option_t *option)
{
    cout << " ** TPlackettCopula::printLog[" << option << "]" << endl;
    cout << " ** Theta [" << _dparam << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

