//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPlackettCopula.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 18:08:40 CEST 2012
 
 \brief Interface of the class URANIE::Sampler::TPlackettCopula
 
 */
/**
 @class URANIE::Sampler::TPlackettCopula
 
 @brief Description of the class TPlackettCopula.
 
 The \b Plackett \b copula for bivariate is defined by the formula
 \f[
 C(u_1, u_2 ; \theta) \: = \: \frac{\left( [ 1 + (\theta - 1)(u_1+u_2)] - \sqrt{[1+(\theta - 1)(u_1+u_2)]^2 - 4 uv\theta(\theta-1)}\right)}{2(\theta-1)} \: , \: \theta > 0, \theta \neq 1
 \f]
 and
 \f[
 C(u_1, u_2 ; \theta) \: = \: u_1 u_2 \:, \: \theta = 1.0
 \f]
 
 with \f$\theta \in [0, \infty]\f$.

 \par Implementation
 We use the Johnson algorithm (1986) to generate random variates \f$ (u,v)\f$ from a \b Plackett distribution with parameter \f$ \theta > 0\f$.
 \li Generate two independent uniform variate \f$ (u,t) \in (0,1)\f$
 \li Set \f$ a = t(1-t) \f$ <br>
 \f$ b = \theta + a(\theta-1)^2\f$ <br>
 \f$ c = 2a(u\theta^2 + 1 - u) + \theta(1 - 2a)\f$ <br>
 \f$ d = \sqrt{\theta}\sqrt{\theta + 4au(1 - u)(1 - \theta)^2}\f$ <br>
 \f$ v = \frac{c-(1-2t)d}{2b}\f$ <br>
 \li the desired pair is \f$ (u,v)\f$
 
 */
#ifndef TPLACKETTCOPULA_H
#define TPLACKETTCOPULA_H

////// URANIE
#include "TArchimedianCopula.h"
#include "UExceptions.h"

namespace URANIE
{
namespace Sampler
{
class TPlackettCopula: public TArchimedianCopula
{
private:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TPlackettCopula(URANIE::DataServer::TDataServer *tds, Double_t dparam,
            Int_t nCalcul);
    //! Default destructor
    virtual ~TPlackettCopula();
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the matrix of U
    /*!
     \param option (Option_t *) The option to pass.
     */
    void generateUMatrix(Option_t *option);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TPlackettCopula, ID_SAMPLER)
    // TPlackettCopula Class
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
