//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSampler.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Mar  7 14:06:02 2006

 \brief  Creation of the abstract class TSampler


 */

#include <ctime>
#include <iostream>

////// URANIE
#include "TSampler.h"
#include "TAttributeFormula.h"
#include "TStochAttribut.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sampler::TSampler)

namespace URANIE
{
namespace Sampler
{

////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TSampler::TSampler(URANIE::DataServer::TDataServer *tds, Option_t *option,
        Int_t ns) :
    _nS(ns), _ntsample(NULL), _sMethod("_Sampler_Method_Uranie"), _blog(kFALSE), _bupdateFile(kTRUE),  _tds(tds),
    _lstOfAttributesToSample(NULL)
{
    //  setLog();
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TSampler::constructor - ERROR\n The TDS is NULL");
        TString opt = option;
        SetName(Form("Sampling_%s", opt.Data()));
        SetTitle(Form("Sampling with option[%s]", opt.Data()));

        parseOption(option);
        
        _lstOfAttributesToSample = new TList();

        // init the TList to have the number of stochastic attributes
        if (_nS > 0)
            createListOfAttributes();

        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSampler::constructor\n The requested number of points [%i] should be greater than 0",_nS));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        delete _lstOfAttributesToSample;
        throw ue;
    }
}

TSampler::~TSampler()
{
    if (_blog)
        cout << "************ Destructor URANIE::Sampler::TSampler" << endl;

    //cout<<" delete sampler "<< _ntsample << "  test   "<<gROOT->GetList() << "  test2   "<<gROOT->GetList()->IsEmpty()<<endl;
    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
        if ( _ntsample !=NULL)
            if( gROOT->GetList()->Contains(_ntsample) )                
                delete _ntsample;
    }
    _ntsample=NULL;
    
    if (_lstOfAttributesToSample != NULL)
      delete _lstOfAttributesToSample;
    _lstOfAttributesToSample = NULL;

    if (_blog)
        cout << "************ End of Destructor URANIE::Sampler::TSampler"
                << endl;
}

///////////////////////////////////////////////////////////
//  Create the List of Attributes to simulate from the TDataServer
///////////////////////////////////////////////////////////
void TSampler::createListOfAttributes()
{
    if (_blog)
        cout << " ** TSampler::createListOfAttributes" << endl;

    // Loop on the attributes of the TDataServer to select only TStochasticAttributes
    URANIE::DataServer::TAttribute *att = 0;
    URANIE::DataServer::TStochasticAttribute *stoatt;

    //---- Modification to avoid calling TDataServer::getAttribute(Int_t) ----//
    //---- ngi: 2010.06.24
    //
    //            Int_t natt = _tds->getNAttributes();
    //            for(Int_t iatt=0; iatt<natt; iatt++) {
    //                att = _tds->getAttribute(iatt);
    //                if ( _blog )
    //                    cout << " ** att [" << att->GetName() << "]";
    //                // Forcast le pointeur pour savoir s'il provient d'un URANIE::DataServer::TStochasticAttributee
    //                stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
    //                if ( stoatt != NULL ) {
    //                    if ( _blog )
    //                        cout << " ********** Add";
    //                    _lstOfAttributesToSample->Add( stoatt );
    //                }
    //                if ( _blog )
    //                    cout << endl;
    //            }

    TList *attList = _tds->getListOfAttributes();
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {
        att = (URANIE::DataServer::TAttribute*) attList->At(iatt);
        if (_blog)
            cout << " ** att [" << att->GetName() << "]";
        // Forcast le pointeur pour savoir s'il provient d'un URANIE::DataServer::TStochasticAttributee
        stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
        if (stoatt != NULL)
        {
            if (_blog)
                cout << " ********** Add";
            _lstOfAttributesToSample->Add(stoatt);
        }
        if (_blog)
            cout << endl;
    }

    //-----------------------------------------------------------------------//

    _nX = _lstOfAttributesToSample->GetEntries();

    if (_blog)
    {
        cout << " ** Entries : " << _nX << endl;
        _lstOfAttributesToSample->Print();
    }
    if (_blog)
        cout << " ** End of TSampler::createListOfAttributes" << endl;

    delete attList;
}

///////////////////////////////////////////////////////////
//  Create the URANIE::DataServer::TDSTNtupleD of data
///////////////////////////////////////////////////////////
void TSampler::createTuple()
{
    try
    {
        TString sLabelVariable = _tds->getIteratorName();
        TString sTypeVariable = "D";
        if (_blog)
            cout << endl
                 << "**************************\n*** TSampler::createTuple sLabelVariable ["
                 << sLabelVariable << "]" << endl;
        
        if (_nX != _lstOfAttributesToSample->GetEntries())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TSampler::createTuple: The number of stochastic attributes (%d) has been modified (%d).",
                            _nX, _lstOfAttributesToSample->GetEntries()));
        if (_blog)
            cout << " ** TSampler::createTuple Size _nX [" << _nX << "]"
                    << endl;

        // Loop on stochastic attributes to have their names
        URANIE::DataServer::TAttribute *att;
        TIter next(_lstOfAttributesToSample);
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            sLabelVariable += Form(":%s", att->GetName());
            sTypeVariable += Form(":%s", att->getDataTypeStr().Data());
            
            if (_blog)
                cout << " ** stoatt[" << att->GetName() << "] sLabelVariable["
                        << sLabelVariable << "]" << endl;
        }

        if (_blog)
            cout << " ** Create the TTree (" << sLabelVariable
                 << ") of data ...";

        //// Create the TTree of data
        //// ngi -- essai d'economie de memoire pour les grands nombres d'attributs
        //Int_t bufsize = 32000;
        //if(_nX > 1000)
        //    bufsize = bufsize / (_nX/1000);
        ////

	// cout<<__LINE__<<" the ntulpe is "<<_ntsample<<endl;
	if(_ntsample!=NULL)
	  delete _ntsample;
	
        _ntsample = new URANIE::DataServer::TDSNtupleD(Form("Sampling_%s", GetName()), GetTitle(), sLabelVariable, sTypeVariable);
            
        if (_blog)
            cout
                    << "*** End Of TSampler::createTuple\n**************************"
                    << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

///////////////////////////////////////////////////////////
//  Fill the other attributes
///////////////////////////////////////////////////////////
void TSampler::fillOtherAttributes()
{
    // modified by ngi: 2013-04-19
    URANIE::DataServer::TAttributeFormula *att = 0;
    TList *attList = _tds->getListOfAttributes();

    // Search for TAttributeFormula
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {
        att = dynamic_cast<URANIE::DataServer::TAttributeFormula *>(attList->At(
                iatt));
        if (att != NULL)
        {
            //_tds->addAttribute(formulatt->GetName(), formulatt->getFormula());
            // fga::todo mauvais code car il vaudrait mieux avoir la methode tds->fillAttribute(TString sname)
            // fin de fga::todo
            _tds->fillAttribute(att->GetName(),"",_bupdateFile);
        }
    }
    att = 0;
    delete attList;
    //-----------------------------------------------------------------------//            
}

///////////////////////////////////////////////////////////
//  Printing Log
///////////////////////////////////////////////////////////
void TSampler::parseOption(Option_t *option)
{

    TString sopt=option;
    sopt.ToLower();

    if( sopt.Contains("noupdate") )
        _bupdateFile=false;
    
}

///////////////////////////////////////////////////////////
//  Printing Log
///////////////////////////////////////////////////////////
void TSampler::printLog(Option_t *option)
{
    cout << "TSampler::printLog[" << option << "] Name[" << GetName()
            << "] Title[" << GetTitle() << "] ID[" << GetID() << "]" << endl;
    cout << "_nS[" << _nS << "]" << endl;
    cout << "_nX[" << _nX << "]" << endl;
    cout << "_sMethod[" << _sMethod << "]" << endl;
    if (_tds)
    {
        cout << "**   TDataServer printLog" << endl;
        _tds->printLog();
        cout << "**   End of TDataServer printLog" << endl;
    }
    if (_ntsample)
    {
        cout << "**   NtSample" << endl;
        _ntsample->Print();
        //_ntsample->Scan();
        cout << "**   fin de NtSample" << endl;
    }
    cout << "TSampler::fin de printLog[]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
