//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSampler.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Mar  7 14:06:02 2006
 
 \brief  Creation of the abstract class TSampler
 
 
 */
#ifndef TSampler_H
#define TSampler_H

// Writes log
#include <iostream>

////// ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TString.h"

// Uranie
#include "TDataServer.h"
#include "TDSNtupleD.h"
#include "UExceptions.h"
#include "Sampler.h"

/*!
 \class URANIE::Sampler::TSampler
 \brief \b Abstraite class for sampling methods

 This is the \b abstract class for sampling methods (determinist, stochastic).
 Usually, the constructors of these classes follow the following prototype :
 \code
 TSampler( URANIE::DataServer::TDataServer *tds, Option_t *option , Int_t nCalcul )
 \endcode
 where :
 \li \em tds : this is a pointer to a \b TDataServer which contains the attributes to be simulated. It can also contain other attributes which will not be sampled as they do not correspond to a recognized type if this method is used.
 \li \em option : it is a list of characters which precises options of the chosen method. For example for qMC methods, a king of sequence ("sobol" ou "halton") can be chosen.
 \li \em nCalcul : the size of the sample to be generated.
 

 The type of the class must be considered be as a machine with 3 states :
 \dot
 digraph samplerTSamplerMachineEtat {
 node [shape=record, fontname=Helvetica, fontsize=10];
 a [ label="init" ];
 b [ label="generateSample"];
 c [ label="terminate"];
 a -> b -> c [ arrowhead="open", style="dashed" ];
 }
 \enddot
 avec 
 \li \b init : this method enables the initialisation of all the information needed by the method (matrix, vector of the prime numbers for qMC). Among others, it permits to select the list of variables to be simulated (  _lstOfAttributesToSample ) which changes according to the adopted method : the \b TAttributes, borned for the determinist methods \b qMC and the \b TStochAttribut for the stoachastic methods (SRS, LHS).
 \li \b generateSample : the method of samples generation itself. It  consists in filling the \b TDataServer which is passed as argument to the constructor.
 \li \b terminate 
 
 \section Pure_virtual_method the pure virtual methods 
 To make the class useful, the  \em daughter classes of TSampler must define the pure virtual method  \b generateSample. This method returns the \b TDataServer for generating a sample.
 
 \section Methode_virtuelle The virtual methods 
 The sampling  method \b generateSample is seen as a state machine,   


 */

using std::cout;
using std::endl;

namespace URANIE
{
namespace Sampler
{
class TSampler: public TNamed
{

    // Attributes
public:
    Int_t _nS;                             ///< The size of the sample
    Int_t _nX;                             ///< The size of attributes to sample
    URANIE::DataServer::TDSNtupleD* _ntsample;          ///< the tntuple of data
    TString _sMethod;                       ///< The title of the sampler method
    Bool_t _blog;                              ///< Log Printing
    Bool_t _bupdateFile;                       ///< Update the back up file when generating the attributeformula if there is some
    URANIE::DataServer::TDataServer *_tds;      ///< Pointer to a TDS
    TList * _lstOfAttributesToSample; ///< The list of Stochastic Attributes to sample

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a TDataServer, the options and the size of the sample
    TSampler(URANIE::DataServer::TDataServer *tds, Option_t *option,
            Int_t nCalcul);
    //! Default destructor
    virtual ~TSampler();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    //! Returns the ID of the class
    /*!
     \todo Adapt to the right form, and see location of the file which contains these informations.
     */
    Int_t GetID()
    {
        return ID_SAMPLER;
    }
    //! Sets the method name in a global variable
    void setMethodName(TString str)
    {
        _sMethod = str;
    }
    //! Gets the method name
    TString getMethodName()
    {
        return _sMethod;
    }
    //! Return the TDS filling by the sampling algorithm
    /*!
     This method returns the TDS filling by the sampling algorithm
     */
    virtual URANIE::DataServer::TDataServer *getTDS()
    {
        return _tds;
    }

    //! Parse the option
    void parseOption(Option_t * option);
    //@}

    //---------------------------------------------
    /** @name Generates the sample
     **/
    //@{
    //! Generates the Sample
    virtual void generateSample(Option_t *option = "") = 0;
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{
    //! Creates the List of attributes to simulate
    virtual void createListOfAttributes();
    //! Creates the TDSNtupleD of data with only the TStochasticAttributes
    virtual void createTuple();
    //! Returns the TDSNtupleD of data
    URANIE::DataServer::TDSNtupleD * getTuple()
    {
        return _ntsample;
    }
    //! Fills the TDSNtupleD of data with other TFormulaAttributes
    virtual void fillOtherAttributes();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSampler, ID_SAMPLER)
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
