//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TSampling.cxx
 *  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 *  \date   $date$
 *  \brief _descritption_.
 *
 *  Created the : Fri Oct  1 13:37:23 2004
 *

 \par Modifications
 \li Cedric Forgues (C.S.) 2010/01/22
 prototype de TMath:Sort() change pour correspondre a celui de ROOT 5.22.00f

 */
#include <ctime>
#include <iostream>
#include <fstream>

////// ROOT
#include "TDecompChol.h"
#include "TDecompSVD.h"
#include "TMath.h"
#include "TDatime.h"
#include "TROOT.h"
#include "TRandom3.h"

////// URANIE
#include "TSampling.h"
#include "TAttributeFormula.h"
#include "TStochAttribut.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "LatinHypercube.H"

ClassImp (URANIE::Sampler::TSampling)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TSampling::TSampling(URANIE::DataServer::TDataServer *tds, Option_t *option,
        Int_t ns) :
    TSamplerStochastic(tds, option, ns), _fLhsInput(NULL), _buseSVD(false), _buseSVDquiet(false)
{
    if (_blog)
        cout << "constructor TSampling by TDataServer[" << tds->GetName() << "]"
                << endl;

    TString opt = option;
    opt.ToLower();

    // Dans le cas d'un plan SRS
    if (opt.CompareTo("srs", TString::kExact) == 0)
    {
        setMethodName("SRS");
        _ntype = kSRS;
        SetNameTitle(Form("Sampling_%s_%d", opt.Data(), ns),
                Form("Uranie %d sample with method %s", ns, opt.Data()));
        // Dans le cas d'un plan LHS (valeur par defaut ou erreur dans le nom de la methode)
    }
    else
    {
        setMethodName("LHS");
        _ntype = kLHS;
        SetNameTitle(Form("Sampling_lhs_%d", ns),
                Form("Uranie %d sample with method lhs", ns));
    }

    if( opt.Contains("svd") )
        _buseSVD=true;
        
    // Allocate the correlation matrix
    _corrmatrix.ResizeTo(_nX, _nX);
    _corrmatrix.UnitMatrix();
}

TSampling::~TSampling()
{
    if (_blog)
        cout << "************ Destructor URANIE::Sampler::TSampling" << endl;
    if (_fLhsInput)
    {
        if (_blog)
            cout << " ** Delete _fLhsInput " << endl;
        delete _fLhsInput;
        _fLhsInput = NULL;
    }
    if (_blog)
        cout << "************ End of Destructor URANIE::Sampler::TSampling"
                << endl;
}

//Small structure to replace the ranking procedure
struct CompareIndicesByAnotherVectorValues
{
    std::vector<double>* _values;
public:
    CompareIndicesByAnotherVectorValues(std::vector<double>* values) : _values(values) {}
    bool operator() (const int& a, const int& b) const { return _values->at(a) < _values->at(b); }
};

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TSampling::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TSamplerStochastic::init();
    
    if( soption.Contains("svd") )
        _buseSVD=true;

    if( soption.Contains("quiet") )
        _buseSVDquiet=true;

    parseOption(option);
    
    Int_t nS = _nS;
    TMatrixD matdval(nS, _nX);

    Double_t *values = new Double_t[nS];
    Int_t * nind = new Int_t[nS];

    Int_t is;

    // Allocate the index position correlation matrix
    _corrIndexPosition.ResizeTo(_nS, _nX);
    _corrIndexPosition.Zero();

    // Initialize permutation
    for (is = 0; is < nS; is++)
        nind[is] = is + 1;
    Int_t valtmp, indtmp;

    if (0)
    {
        // Create a permutation with the Devroye Algorithm
        for (is = nS - 1; is > 0; is--)
        {
            indtmp = (Int_t) gRandom->Integer(is);
            valtmp = nind[indtmp];
            if (_blog)
                cout << "** is[" << is << "] val [" << nind[is] << "] indtmp["
                        << indtmp << "] val[" << nind[indtmp] << "]";
            nind[indtmp] = nind[is];
            nind[is] = valtmp;
            if (_blog)
                cout << "************* is[" << is << "] val [" << nind[is]
                        << "] indtmp[" << indtmp << "] val[" << nind[indtmp]
                        << "]" << endl;
            ;
        }
    }

    Double_t dpas=-1, dtype=-1, dproba=0;

    // Boucle sur les variables stochastiques
    URANIE::DataServer::TStochasticAttribute *curdist;
    TIter next(_lstOfAttributesToSample);
    Int_t iatt = 0;
    while ((curdist = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        //Double_t dpmin = curdist->getPMin(); 
        //Double_t dpmax = curdist->getPMax(); 

        switch (_ntype)
        {
        case kLHS:
        {
            //dpas = ( dpmax - dpmin) / (Double_t)( nS);
            dpas = 1.0 / (Double_t)(nS);
            dtype = 1.0;
        }
            break;
        case kSRS:
        {
            //dpas = ( dpmax - dpmin);
            dpas = 1.0;
            dtype = 0.0;
        }
            break;
	default:
	    break;
        }

        if (_blog)
        {
            cout << "name[" << curdist->GetName() << "] Type["
                    << curdist->getLawType() << "]" << endl;
        }

	//vector to store the values, original index and valuues' indexes
	vector<double>   v_dval(nS, 0.);
	vector<Long64_t> v_index(nS, 0.);
	vector<double>   v_valindex(nS, 0.);
        // Generate the sample
        for (is = 0; is < nS; is++)
        {
            //dproba = dpas * gRandom->Rndm() + dtype * dpas * is + dpmin;
            dproba = dpas * gRandom->Rndm() + dtype * dpas * is;	    
            matdval(is, iatt) = values[is] = curdist->getInvCDF(dproba);
	    if( dtype==0 )
	    {
		v_dval[is] = values[is] ;
		v_index[is] = is ; //Was 0.0 in the old procedure
	    }
            if (_blog)
                cout << "** is[" << is << "] -- proba [" << dproba
                        << "] -- value[" << values[is] << "]" << endl;
        }

	// // Sorting matrix in the case of SRS drawing
	if(dtype==0)
	{
	    
	    stable_sort(v_index.begin(), v_index.end(), CompareIndicesByAnotherVectorValues(&v_dval) );

	    //Fill the final indexes values
	    for (is = 0; is < nS; is++)
		v_valindex.at(v_index.at(is)) = (Double_t)(is);

	    for (is = 0; is < nS; is++)
		matdval(v_valindex.at(is), iatt) = v_dval.at(is);
	}
	
        if (0)
        {
            // Create a new permutation and Store the vector of the current attribute in the matrix data
            for (is = nS - 1; is > 0; is--)
            {
                indtmp = (Int_t) gRandom->Integer(is);
                valtmp = nind[indtmp];
                if (_blog)
                    cout << "** is[" << is << "] val [" << nind[is]
                            << "] indtmp[" << indtmp << "] val[" << nind[indtmp]
                            << "]";
                nind[indtmp] = nind[is];
                nind[is] = valtmp;
                if (_blog)
                    cout << "--- is[" << is << "] val [" << nind[is]
                            << "] indtmp[" << indtmp << "] val[" << nind[indtmp]
                            << "]";
                ;
                matdval(is, iatt) = values[nind[is] - 1];
                if (_blog)
                    cout << "::: is[" << is << "] [" << matdval(is, iatt) << "]"
                            << endl;
                ;
            }
            // For the last pattern
            if (_blog)
                cout << "last pattern[" << is << "] valeur["
                        << values[nind[is] - 1] << "]" << endl;
            matdval(is, iatt) = values[nind[is] - 1];
        }
        //Increment the index of attribute
        iatt++;
    }

    // Print the data matrix if log
    if (_blog)
        matdval.Print();

    // Create the TTree of data
    createTuple();

    // Get a permutation to fit the correlation matrix (Unity by default)
    if (_blog)
    {
        cout << "***  generateSample :: Print the correlation matrix" << endl;
        _corrmatrix.Print();
    }

    try
    {
        CorrelationFactorization();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	delete[] values;
	values = NULL;
	delete[] nind;
	nind = NULL;
        ue.printMessage();
        throw ue;
    }

    Double_t *valpattern = new Double_t[_nX + 1];

    Int_t nPos;
    if (_blog)
        cout << "Fill the TTree ..." << endl;
    for (is = 0; is < nS; is++)
    {

        if (_blog)
            cout << "Case is[" << is << "] ...";

        valpattern[0] = (Double_t)(is + 1);

        if (_blog)
            cout << " -- val [" << valpattern[0] << "] ...";

        for (Int_t jx = 0; jx < _nX; jx++)
        {
            //      valpattern[jx+1] = matdval(is, jx);
            nPos = (Int_t)(_corrIndexPosition(is, jx)); // is | (Int_t)(_corrIndexPosition(is, jx));
            valpattern[jx + 1] = matdval(nPos, jx);
            if (_blog)
                cout << "var jx[" << jx << "]  nPos[" << nPos << "] val["
                        << valpattern[jx + 1] << "]...";
        }
        _ntsample->Fill(valpattern);
        if (_blog)
            cout << " OK" << endl;
    }
    if (_blog)
        cout << " Done" << endl;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    fillOtherAttributes();

    // Clean
    delete[] values;
    values = NULL;
    delete[] valpattern;
    valpattern = NULL;
    delete[] nind;
    nind = NULL;
}

/**************************************************/
/*   Generate the sample with the Dakota module   */
/**************************************************/
void TSampling::generateSampleByDakota()
{

  TSamplerStochastic::init();
      
    if (_fLhsInput)
    {
        delete _fLhsInput;
        _fLhsInput = NULL;
    }
    _fLhsInput = new LHSInput();

    if (_ntype == kSRS)
        _fLhsInput->SetSampleFlag(1);
    if (_ntype == kLHS)
        _fLhsInput->SetSampleFlag(0);

    _fLhsInput->SetN_of_O(_nS);

    // Boucle sur les variables stochastiques
    URANIE::DataServer::TStochasticAttribute *varsto;
    TIter next(_lstOfAttributesToSample);
    while ((varsto = (URANIE::DataServer::TStochasticAttribute *) next()))
    {
        if (_blog)
        {
            cout << "name[" << varsto->GetName() << "] Type["
                    << varsto->getLawType() << "]" << endl;
            //      curdist->printLog();
        }

        switch (varsto->getLawType())
        {
        case URANIE::DataServer::TStochasticAttribute::kNormal:
        {
            LHSNormal *var = new LHSNormal();
            var->Set_M_V(varsto->getTheoreticalMean(),
                    varsto->getTheoreticalStdDev());
            var->SetName(varsto->GetName());
            if (varsto->isUpperBounded())
                var->SetLowerBound(varsto->getLowerBound());
            var->SetUpperBound(varsto->getUpperBound());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kLogNormal:
        {
            LHSLognormal *var = new LHSLognormal();
            var->Set_M_V(varsto->getTheoreticalMean(),
                    varsto->getTheoreticalStdDev());
            var->SetName(varsto->GetName());
            var->SetLowerBound(varsto->getLowerBound());
            var->SetUpperBound(varsto->getUpperBound());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kUniform:
        {
            LHSUniform *var = new LHSUniform();
            var->SetRange(varsto->getLowerBound(), varsto->getUpperBound());
            var->SetName(varsto->GetName());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kLogUniform:
        {
            LHSLoguniform *var = new LHSLoguniform();
            var->SetRange(varsto->getLowerBound(), varsto->getUpperBound());
            var->SetName(varsto->GetName());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kTriangular:
        {
            LHSTriangular *var = new LHSTriangular();
            var->SetRange(varsto->getLowerBound(), varsto->getUpperBound(),
                    varsto->getTheoreticalMode());
            var->SetName(varsto->GetName());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kLogTriangular:
        {
            LHSLogtriangular *var = new LHSLogtriangular();
            var->SetRange(varsto->getLowerBound(), varsto->getUpperBound(),
                    varsto->getTheoreticalMode());
            var->SetName(varsto->GetName());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kExponential:
        {
            LHSExponential *var = new LHSExponential();
            var->SetRange(varsto->getTheoreticalMean(),
                    varsto->getLowerBound());
            var->SetName(varsto->GetName());
            _fLhsInput->Add(var);
            break;
        }
        case URANIE::DataServer::TStochasticAttribute::kWeibull:
        {
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
	          "TSampling::generateSampleByDakota ERROR : Weibull law not used in Dakota module");
            break;
        }
        default:
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
		  Form("TSampling::generateSampleByDakota ERROR : Loi [%d] non connue\n",
		       varsto->getLawType()));
        }

    }

    LatinHypercube* echan = new LatinHypercube(_fLhsInput);

    // if (_nSeed == 0)
    // {
    //     time_t secondes;
    //     time(&secondes);
    //     echan->SetSeed((int) (secondes));
    // }
    // else
    //     echan->SetSeed(_nSeed);

    echan->SetSeed(_nSeed);

    echan->SetLog(_blog);

    Int_t nvar = _tds->getNAttributes();
    if (_blog)
        cout << "nvar[" << nvar << "]" << endl;

    // Si la matrice _corrmatrix est d�finie
    if (_corrmatrix.GetNrows() > 0)
    {
        // Effectue la d�composition of the correlation matrix
        TDecompChol chol(_corrmatrix);

	try 
	{
	    if (chol.Decompose())
	    {
	        for (Int_t i = 0; i < nvar; i++)
		    for (Int_t j = 0; j < i; j++)
		        echan->addCorr(i + 1, j + 1, _corrmatrix(i, j));
	    }
	    else
	    {
	        if (_blog)
		{
		    cerr << "The correlation matrix is not positive definite" << endl;
		    _corrmatrix.Print();
		}
		// throw "The correlation matrix is not positive definite\n";
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
		 "TSampling::generateSampleByDakota ERROR : The correlation matrix is not positive definite");

	    }
	} catch (URANIE::Exceptions::UErrorExceptions& ue)
	{
	    delete echan;
	    echan = NULL;
	    throw ue;
	}
    }
    // Generation of the sample
    echan->GenSample();

    // Create the TTree of data
    createTuple();

    Double_t *values = new Double_t[nvar + 1];

    for (Int_t i = 0; i < _nS; i++)
    {
        values[0] = (Double_t)(i + 1);
        for (Int_t j = 0; j < nvar; j++)
        {
            values[j + 1] = echan->GetSample(i + 1, j + 1);
        }
        _ntsample->Fill(values);
    }

    // Clean
    delete[] values;
    values = NULL;
    delete echan;
    echan = NULL;

    if (_blog)
        _ntsample->Print();

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    fillOtherAttributes();

}

/************************************/
/*       Correlations Matrix        */
/************************************/
void TSampling::setCorrelationMatrix(TMatrixD corrMat)
{

    try
    {
        if ((corrMat.GetNcols() != _nX)
            || (corrMat.GetNrows() != corrMat.GetNcols()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix of size %d x %d ",
                                                           _nX,_nX));

        _corrmatrix.ResizeTo(corrMat);
        _corrmatrix = corrMat;

        for(int icol=0; icol<corrMat.GetNcols(); icol++)
	{
	    _corrmatrix(icol,icol)=1.0;
	    for(int ilig=0; ilig<corrMat.GetNrows(); ilig++)
	    {
		if( fabs(_corrmatrix(ilig,icol) - _corrmatrix(icol,ilig)) > 1e-14 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TSampling::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix.\n\
This is not the case as M(%d,%d) != M(%d,%d). Their difference is %g",ilig,icol,icol,ilig, _corrmatrix(ilig,icol) - _corrmatrix(icol,ilig)));
	    }
	}
        
        Double_t max = TMath::MaxElement(_corrmatrix.GetNoElements(),
                                         _corrmatrix.GetMatrixArray());
        Double_t min = TMath::MinElement(_corrmatrix.GetNoElements(),
                                         _corrmatrix.GetMatrixArray());
        if (max > 1.0 || min < -1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSampling::setCorrelationMatrix: correlation values must belong to the interval [-1.0, 1.0] (min and max found are [%g,%g])",
                                                            min,max));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

/***************************************************/
/*      Correlation factor                         */
/***************************************************/

void TSampling::setUserCorrelation(URANIE::DataServer::TAttribute *x,
        URANIE::DataServer::TAttribute *y, Double_t value)
{

    try
    {
        if (fabs(value) > 1.0)        
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the correlation factor  between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                                                           x->GetName(), y->GetName(), value));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    Int_t indx = _tds->getAttributeIndex(x);
    Int_t indy = _tds->getAttributeIndex(y);

    setUserCorrelation(indx - 1, indy - 1, value);
}

void TSampling::setUserCorrelation(TString xname, TString yname, double value)
{
    try
    {
    if (fabs(value) > 1.0)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TSampling::setUserCorrelation ERROR : the correlation factor  between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                        xname.Data(), yname.Data(), value));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    Int_t indx = _tds->getAttributeIndex(xname);
    Int_t indy = _tds->getAttributeIndex(yname);

    setUserCorrelation(indx - 1, indy - 1, value);
}

void TSampling::setUserCorrelation(Int_t indx, Int_t indy, Double_t value)
{
    try
    {
        if (fabs(value) > 1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the correlation factor between attributes (%d, %d) is greater 1.0 : [%f]",
                                                           indx, indy, value));
        if (indx == indy)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the two indices are equal (%d, %d)",
                                                           indx, indy));
        if (indx < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the first indice is lesser than 0  (%d)",
                                                           indx));
        if (indy < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the second indice is lesser than 0  (%d)",
                                                           indy));

        if (indx >= _nX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the first indice is greater than max size (%d) (%d)",
                                                           indx, _nX));
        if (indy >= _nX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSampling::setUserCorrelation ERROR : the second indice is greater than max size (%d) (%d)",
                                                           indy, _nX));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    //  if ( _corrmatrix.GetNrows() != _nX ) {
    //    // Si la matrice _matcorr est non nulle (cad existe d�j�),
    //    //   alors on en fait une copie
    //    if ( _corrmatrix.GetNrows() > 0 ) {
    //      TMatrixD corrcrt(_corrmatrix);

    //      // Nouvelle allocation
    //      _corrmatrix.ResizeTo(_nX, _nX);
    //      _corrmatrix.UnitMatrix();

    //      //Recopie l'ancienne sous-matrice dans la nouvelle
    //      for(Int_t i=0; i<corrcrt.GetNrows(); i++) {
    //        for(Int_t j=0; j<i; j++) {
    //      _corrmatrix(i, j) = corrcrt(i, j);
    //      _corrmatrix(j, i) = corrcrt(j, i);
    //        }
    //      }
    //      // La matrice n'existe pas encore
    //    } else {
    //      //  Allocation
    //      _corrmatrix.ResizeTo(_nX, _nX);
    //      _corrmatrix.UnitMatrix();
    //    }
    //  }

    // On sauvegarde les valeurs
    _corrmatrix(indx, indy) = value;
    _corrmatrix(indy, indx) = value;

    if (_blog)
        _corrmatrix.Print();
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TSampling::printLog(Option_t *option)
{
    cout << "TSampling::printLog[" << option << "] Name[" << GetName()
            << "] Title[" << GetTitle() << "] ID[" << GetID() << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nX   [" << _nX << "]" << endl;
    cout << "** _sMethod   [" << _sMethod << "]" << endl;
    _corrmatrix.Print();

    //  if ( _tds ) {
    //    cout << "**   TDataServer" << endl;
    //    _tds->printLog();
    //  }

    if (_fLhsInput)
        _fLhsInput->printlog();

    TSamplerStochastic::printLog(option);
    cout << "End Of TSampling::printLog" << endl;
}

void TSampling::setSaveRang(Int_t rang)
{
    UNUSED(rang);
}

void TSampling::setTypeSampling(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
}

void TSampling::CorrelationFactorization()
{
    if (_blog)
    {
        cout << "TSampling::CorrelationFactorization" << endl;
        cout << "_nS[" << _nS << "]" << endl;
        cout << "_nX[" << _nX << "]" << endl;
    }

    if (_blog)
        cout << " Decomposition de cholesky de la matrice de correlation "
                << endl;

    TMatrixD matUFromC(_nX,_nX);
    TMatrixD matDFromC(_nX,_nX);
    if( _buseSVD )
    {
        if (!_buseSVDquiet)
        {
            try{
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                          "TSampling::CorrelationFactorization : you've asked to use SVD instead of Cholesky to decompose the correlation matrix. ");
            }catch(URANIE::Exceptions::UInfoExceptions &ue)
            {
                ue.printMessage();
            }
        }
        TDecompSVD svdC(_corrmatrix);
        if (svdC.Decompose() == kFALSE)
        {
            if (_blog)
            {
                _corrmatrix.Print();
                // print the matrix U
                cout << " ** Print the matrix U" << endl;
                svdC.GetU().Print();
            }
            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TSampling::CorrelationFactorization : the SVD decomposition of the user correlation matrix is not good");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }
        matUFromC=svdC.GetU();
        const TVectorD vecDFromC = svdC.GetSig();
        matDFromC.Zero();
        for (int i=0; i<_nX; i++)
            matDFromC(i,i)=sqrt(vecDFromC[i]);

        if (_blog)
        { // _blog
            cout << "******* Matrice  de correlation ******** " << endl;
            svdC.Print();
        }   
    }
    else
    {
        TDecompChol cholc(_corrmatrix);
        if (cholc.Decompose() == kFALSE)
        {
            if (_blog)
            {
                _corrmatrix.Print();
                // print the matrix U
                cout << " ** Print the matrix U" << endl;
                cholc.GetU().Print();
            }
            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           "TSampling::CorrelationFactorization : the cholesky decomposition of the user correlation matrix is not good.\n Maybe you can try with SVD decomposition.");
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
        }
        matUFromC = cholc.GetU();

        if (_blog)
        { // _blog
            cout << "******* Matrice  de correlation ******** " << endl;
            cholc.Print();
        }        
    }

    if (_blog)
    { // _blog
        cout << "******* Matrice  U De C (calculated) ******** " << endl;
        matUFromC.Print();
        if( _buseSVD )
        {
            cout << "******* Matrice  D^{1/2} De C (calculated) ******** " << endl;
            matDFromC.Print();
        }
    }

    URANIE::DataServer::TNormalDistribution *_rank_gauss =
            new URANIE::DataServer::TNormalDistribution(
                    "_uranie_rank_gauss_cholesky", 0.0, 1.0);
    if (_blog)
        _rank_gauss->printLog();

    TMatrixD matScoreS(_nS, 3);
    TMatrixD matS(_nS, _nX);
    for (Int_t i = 0; i < _nS; ++i)
    {
        matScoreS(i, 0) = (Double_t)(i + 1);
        matScoreS(i, 1) = (Double_t)(i + 1) / (Double_t)(_nS + 1);
        matScoreS(i, 2) = _rank_gauss->getInvCDF(matScoreS(i, 1));
    }
    if (_blog)
    {
        cout << "*** Initialisation de la Matrice matScoreS" << endl;
        matScoreS.Print("jddjdjj");
    }

    Double_t* dMean = new Double_t[_nX];
    Double_t* dSigma = new Double_t[_nX];

    Int_t deqq_1[10] =
    { 7, 4, 3, 6, 1, 5, 0, 9, 8, 2 };
    Int_t deqq_2[10] =
    { 0, 8, 5, 4, 3, 2, 7, 9, 6, 1 };

    // Bouble sur chaque variable
    for (Int_t i = 0; i < _nX; ++i)
    {
        // Create a permutation with the Devroye Algorithm
        Int_t valtmp, indtmp;
        dMean[i] = 0.;
        dSigma[i] = 0.;
        for (Int_t is = _nS - 1; is > 0; is--)
        {
            indtmp = (Int_t) gRandom->Integer(is);
            valtmp = (Int_t)(matScoreS(indtmp, 0));
            if (_blog)
                cout << "** is[" << is << "] val [" << matScoreS(is, 0)
                        << "] indtmp[" << indtmp << "] val["
                        << matScoreS(indtmp, 0) << "]";
            matScoreS(indtmp, 0) = matScoreS(is, 0);
            matScoreS(is, 0) = valtmp;
            if (_blog)
                cout << "************* is[" << is << "] val ["
                        << matScoreS(is, 0) << "] indtmp[" << indtmp << "] val["
                        << matScoreS(indtmp, 0) << "]" << endl;
            matS(is, i) = matScoreS((Int_t)(matScoreS(is, 0)) - 1, 2);
        }
        // Last index
        matS(0, i) = matScoreS((Int_t)(matScoreS(0, 0)) - 1, 2);

        if (_nS == 10)
        {
            if (i == 0)
                for (Int_t j = 0; j < _nS; j++)
                {
                    matS(j, i) = matScoreS(deqq_1[j], 2);
                }
            if (i == 1)
                for (Int_t j = 0; j < _nS; j++)
                {
                    matS(j, i) = matScoreS(deqq_2[j], 2);
                }
        }
        if (_blog)
        {
            cout << "Matrice  des Scores avant Reduction" << endl;
            matS.Print("");
        }

        // Compute the mean and sigma value using recursive formulae
        for (Int_t j = 1; j <= _nS; j++)
        {
            Double_t v = matS(j - 1, i);

            // Cas de la moyenne
            dMean[i] *= (Double_t)(j - 1) / (Double_t)(j);
            dMean[i] += v / (Double_t)(j);

            // Cas de sigma
            if (j > 1)
            {
                dSigma[i] *= (Double_t)(j - 2) / (Double_t)((j - 1));
                dSigma[i] += (Double_t)(j) /(Double_t)(j-1)  * (v - dMean[i]) * (v - dMean[i]) / (Double_t)(j-1);
            }
            if (_blog)
                cout << "j[" << j << "] val[" << v << "] mean[" << dMean[i]
                        << "] Sigma[" << dSigma[i] << "]" << endl;
        }
        if (_blog)
            cout << "cas de la variable " << i << " matScoreS Mean[" << dMean[i]
                    << "] Sigma[" << TMath::Sqrt(dSigma[i]) << "]" << endl;

        if (_blog)
        {
            cout << "******* Matrice  des Scores apr�s Reduction ******** "
                    << endl;
            matS.Print("");
            cout << endl
                    << "**************************************************** "
                    << endl;
        }
    }

    TMatrixD matTmp(matS);

    // Center and reduce the current column
    for (Int_t i = 0; i < _nX; ++i)
    {
        for (Int_t j = 0; j < _nS; j++)
        {
            matS(j, i) -= dMean[i];
            matTmp(j, i) = matS(j, i);
            matS(j, i) /= TMath::Sqrt((Double_t)(_nS - 1) * dSigma[i]);
        }
    }

    if (_blog)
    {
        cout << endl << "**************************************************** "
                << endl;
        cout
                << "******* Matrice  des Scores apr�s Reduction==centre+reduire ******** "
                << endl;
        matS.Print("");
        cout << endl << "**************************************************** "
                << endl;
        cout << "******* Matrice  des Scores matTmp apr�s centre ******** "
                << endl;
        matTmp.Print("");
        cout << endl << "**************************************************** "
                << endl;
    }

    // We delete the data after use
    delete[] dMean;
    dMean = NULL;
    delete[] dSigma;
    dSigma = NULL;

    if (_blog)
    {
        cout << "******* Matrice  des Scores apr�s Reduction ******** " << endl;
        matS.Print("");
    }
    TMatrixD matCorrS(matS, TMatrixD::kTransposeMult, matS);
    if (_blog)
    {
        cout << "******* Matrice  de correlation des scores ******** " << endl;
        matCorrS.Print("");
    }
    
    if (_blog)
        cout << " Decomposition de cholesky de la matrice de correlation "
                << endl;
    TDecompChol cholcE(matCorrS);
    if (cholcE.Decompose() == kFALSE)
    {
        matCorrS.Print();
        try{
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSampling::CorrelationFactorization : the cholesky decomposition of the score correlation matrix is not good");
        }catch(URANIE::Exceptions::UErrorExceptions &ue)
        {
            ue.printMessage();
            delete _rank_gauss; _rank_gauss=NULL;               
            throw ue;
        }

    }
    const TMatrixD matUFromE = cholcE.GetU();
    if (_blog)
    {
        cout << "******* Matrice  Q (calculated) ******** " << endl;
        matUFromE.Print();
    }

    TMatrixD matUFromEinv(TMatrixD::kInverted, matUFromE);
    if (_blog)
    {
        cout << "******* Matrice  Q-1 (calculated) ******** " << endl;
        matUFromEinv.Print();
    }

    TMatrixD matRacDinvP(_nX,_nX);
    if( _buseSVD )
    {
        matRacDinvP = matUFromEinv * matDFromC;
        matUFromC.T();
    }
    else
        matRacDinvP = matUFromEinv;
    TMatrixD matQinvP(matRacDinvP, TMatrixD::kMult, matUFromC);
    
    if (_blog)
    {
        cout << endl << endl
                << "******* Matrice  (Q-1)t Pt (calculated) ******** " << endl;
        matQinvP.Print();
        cout << endl
                << "******* End of Matrice  (Q-1)t Pt (calculated) ******** "
                << endl;
    }

    TMatrixD SStar(matTmp, TMatrixD::kMult, matQinvP);
    if (_blog)
    {
        cout << "******* Matrice  SStar = S * Q-1t Pt (calculated) ******** "
                << endl;
        SStar.Print();
    }

    if (_blog)
    {
        computeCorrelationMatrix(SStar);
    }

    // Compute the order of the Score transformed Matrix S*

    Double_t* values = new Double_t[_nS];
    Long64_t *index = new Long64_t[_nS];
    //             long long *index = new long long[_nS];

    // Compute rank on each attribute
    for (Int_t i = 0; i < _nX; ++i)
    {
        for (Int_t j = 0; j < _nS; j++)
        {
            values[j] = SStar(j, i);
        }
        TMath::Sort(Long64_t(_nS), values, index, kFALSE);
        for (Int_t j = 0; j < _nS; j++)
        {
            _corrIndexPosition((Int_t) index[j], i) = j;
        }
        if (_blog)
        {
            cout << endl << endl << "struct rank Sampling" << endl << endl;
            for (Int_t j = 0; j < _nS; j++)
            {
                cout << "Sampling [" << j << "] val[" << values[j] << "] rank["
                        << _corrIndexPosition(j, i) << "] index[" << index[j]
                        << "]" << endl;
            }
        }

    }
    // We delete the data after use
    delete[] index;
    index = NULL;
    delete[] values;
    values = NULL;
    delete _rank_gauss; _rank_gauss = NULL;
}


void TSampling::computeCorrelationMatrix(TMatrixD matOrg)
{
    if (_blog)
    {
        cout
                << "**************  TSampling::correlationCalculate ************** "
                << endl;
        cout << "******* Original matrix ******** " << endl;
        matOrg.Print();
    }

    TMatrixD matTmp(matOrg);

    Int_t nx = matTmp.GetNrows();
    Int_t ny = matTmp.GetNcols();
    if (_blog)
        cout << "nx[" << nx << "] ny[" << ny << "]" << endl;

    Double_t* vecMean = new Double_t[ny];
    Double_t* vecSigma = new Double_t[ny];

    // Loop on each column to compute Mean and Variance with recursive formulae
    for (Int_t iy = 0; iy < ny; ++iy)
    {
        // Init means and sigma value to 0.
        vecMean[iy] = 0.;
        vecSigma[iy] = 0.;

        // Fist Step : Compute the mean and sigma value using recursive formulae
        for (Int_t ix = 0; ix < nx; ix++)
        {
            Double_t v = matTmp(ix, iy);

            // Mean case
            vecMean[iy] *= (Double_t)(ix) / (Double_t)(ix + 1);
            vecMean[iy] += v / (Double_t)(ix + 1);

            // Variance=sigma**2 case
            if (ix > 0)
            {
                vecSigma[iy] *= (Double_t)(ix - 1) / (Double_t)(ix);
                vecSigma[iy] += (Double_t)(ix + 1) * (v - vecMean[iy])
                        * (v - vecMean[iy]) / (Double_t)(ix * ix);
            }
            if (_blog)
                cout << "iy[" << iy << "] val[" << v << "] mean[" << vecMean[iy]
                        << "] Sigma[" << vecSigma[iy] << "]" << endl;
        }

        // Second Step : Center and reduce the current column
        for (Int_t ix = 0; ix < nx; ix++)
        {
            matTmp(ix, iy) -= vecMean[iy];
            matTmp(ix, iy) /= TMath::Sqrt((Double_t)(nx - 1) * vecSigma[iy]);
        }
    }

    // We delete the data after use
    delete[] vecMean;
    vecMean = NULL;
    delete[] vecSigma;
    vecSigma = NULL;

    TMatrixD matCorr(matTmp, TMatrixD::kTransposeMult, matTmp);
    if (_blog)
    {
        cout << "******* Matrice  de correlations (calculated) ******** "
                << endl;
        matCorr.Print("");
        cout << "******* Matrice  de correlations user ******** " << endl;
        _corrmatrix.Print("");
        cout << endl
                << "**************  End of TSampling::correlationCalculate ************** "
                << endl << endl;
    }
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
