//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
#ifndef TSequence_H
#define TSequence_H
/**
 
 * @class URANIE::Sampler::TSequence

 * @brief Description of a sequence (qMC)
 *
 * There are two abstract methods to be redefined for each daughter class:
 * \li init : initialization of the sequence;
 * \li setInstance(const ULong_t  n) : setting the current cursor on the position number n
 *
 */
#include "Rtypes.h"
#include "TNamed.h"
#include "TSampler.h"

namespace URANIE
{
namespace Sampler
{
class TSequence: public TNamed
{

    // Attributes
public:
    ULong_t _nDim;             ///< The size of the sequence
    Double_t * _dValue;     //[_nDim]   ///< Pointeur
    static Int_t _seqCounter;  ///< The counter of sequence
    static Bool_t _blog;       ///< Log Printing

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TSequence();
    //! Default constructor
    TSequence(ULong_t ndim);
    //! Default destructor
    virtual ~TSequence();
    //@}

    //---------------------------------------------
    /** @name Abstract method
     These two methods must be redefined for each daughter class
     **/
    //@{
    virtual void setInstance(const ULong_t n) = 0;
    virtual void init() = 0;
    //@}

    inline Double_t operator[](const ULong_t) const;
    inline Double_t getComponent(const ULong_t) const;

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    //! Sets the static variable _blog at kTRUE
    void setLog()
    {
        _blog = kTRUE;
    }
    //! Sets the static variable _blog at kFALSE
    void unsetLog()
    {
        _blog = kFALSE;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSequence, ID_SAMPLER)
    //Classe de g�n�ration par Quasi Monte-Carlo
};

//////////////////////////////////////////////////////////////////////////////

inline Double_t TSequence::operator[](const ULong_t j) const
{
    if (j >= _nDim)
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TSequence::operator[] ERROR : the indice %lu is greater than the dimension[%lu]",
                        j, _nDim));
    }
    return (_dValue[j]);
}

//////////////////////////////////////////////////////////////////////////////

inline Double_t TSequence::getComponent(const ULong_t j) const
{
    if (j >= _nDim)
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TSequence::getComponent ERROR : the indice %lu is greater than the dimension[%lu]",
                        j, _nDim));
    }
    return (_dValue[j]);
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
