//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVQ.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jui 21 14:54:08 2009

 \brief  Implement the abstract class for Vectorial Quantification (VQ) algorithms


 */
#include <ctime>

// ROOT
#include "TRandom3.h"

// URANIE
#include "TVQ.h"

ClassImp (URANIE::Sampler::TVQ);

namespace URANIE
{
namespace Sampler
{

TVQ::TVQ(URANIE::DataServer::TDataServer *tds, Option_t *option, Int_t ns) :
        TSamplerStochastic(tds, option, ns), _fLogger(NULL), _bdrawProgressBar(kTRUE), _vecindexorg(0), 
        _subtds(NULL), _bsaveall(kFALSE), _sattributes("")
{
    TString soption = option;
    soption.ToLower();

    if (soption.Contains("saveall"))
        _bsaveall = kTRUE;
}

// destructor
TVQ::~TVQ()
{
    if (_subtds)
        delete _subtds;
    _subtds = NULL;
}

///////////////////////////////////////////////////////////
//  Create the List of Attributes to simulate from the TDataServer
///////////////////////////////////////////////////////////
void TVQ::createListOfAttributes()
{
    if (_blog)
        cout << " ** TVQ::createListOfAttributes" << endl;

    // Loop on the attributes of the TDataServer to select only TStochasticAttributes
    URANIE::DataServer::TAttribute *att = 0;

    //---- Modification to avoid calling TDataServer::getAttribute(Int_t) ----//
    //---- ngi: 2010.06.24          
    //            
    //            Int_t natt = _tds->getNAttributes();            
    //            for(Int_t iatt=0; iatt<natt; iatt++) {
    //                att = _tds->getAttribute(iatt);
    //                if ( _blog )
    //                    cout << " ** att [" << att->GetName() << "]";
    //                _lstOfAttributesToSample->Add( att );
    //                if ( _blog )
    //                    cout << endl;
    //            }

    TList *attList = _tds->getListOfAttributes();
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {
        att = (URANIE::DataServer::TAttribute*) attList->At(iatt);
        if (_blog)
            cout << " ** att [" << att->GetName() << "]";
        _lstOfAttributesToSample->Add(att);
        if (_blog)
            cout << endl;
    }
    //-----------------------------------------------------------------------//            
    _nX = _lstOfAttributesToSample->GetEntries();

    if (_blog)
    {
        cout << " ** Entries : " << _nX << endl;
        _lstOfAttributesToSample->Print();
    }
    if (_blog)
        cout << " ** End of TSampler::createListOfAttributes" << endl;

    delete attList;
}

/*******************************************/
/*           state = Init                  */
/*******************************************/
void TVQ::init()
{
    if (_blog)
        cout << " ** TVQ::init " << endl;

    createListOfAttributes();

    // Allocate the vector index with the size of the original dataserver
    Int_t nsorg = _tds->getNPatterns();
    _vecindexorg.ResizeTo(nsorg);
    for (Int_t i = 0; i < nsorg; i++)
        _vecindexorg(i) = (Float_t) i;

    // Create the TDataServer which will contain the subsample
    _subtds = new URANIE::DataServer::TDataServer(
            Form("sub_%s", _tds->GetName()),
            Form("sub sample of %s", _tds->GetTitle()));

    // Loop on the attributes
    URANIE::DataServer::TAttribute *tatt;
    TIter next(_lstOfAttributesToSample);
    Int_t ncrt = 0;
    while ((tatt = (URANIE::DataServer::TAttribute *) next()))
    {
        if (_blog)
        {
            cout << " ** att number[" << ncrt + 1 << "] name["
                    << tatt->GetName() << "] " << endl;
            tatt->printLog();
        }
        _subtds->addAttribute(tatt);
        if (ncrt > 0)
            _sattributes += TString(":");
        _sattributes += TString(tatt->GetName());
        ncrt++;
    }
    _subtds->createTuple();

    if (_blog)
        cout << " ** End Of TVQ::init " << endl;
}

/*******************************************/
/*           state = terminate             */
/*******************************************/
void TVQ::terminate()
{
    if (_blog)
        cout << " ** TVQ::terminate " << endl;
    if (_blog)
        cout << " ** End Of TVQ::terminate " << endl;
}

/*******************************************/
/*           Shuffle the _vecindex             */
/*******************************************/
void TVQ::shuffle()
{
    Int_t nS = _vecindexorg.GetNrows();
    if (_blog)
        cout << " ** TVQ::shuffle nS[" << nS << "]" << endl;
    Int_t indtmp, valtmp;
    // Create a permutation with the Devroye Algorithm
    for (Int_t is = nS - 1; is > 0; is--)
    {
        indtmp = (Int_t) gRandom->Integer(is);
        valtmp = (Int_t)(_vecindexorg(indtmp));
        if (_blog)
            cout << "** is[" << is << "] val [" << _vecindexorg(is)
                    << "] indtmp[" << indtmp << "] val[" << _vecindexorg(indtmp)
                    << "]";
        _vecindexorg(indtmp) = _vecindexorg(is);
        _vecindexorg(is) = (Float_t) valtmp;
        if (_blog)
            cout << "************* is[" << is << "] val [" << _vecindexorg(is)
                    << "] indtmp[" << indtmp << "] val[" << _vecindexorg(indtmp)
                    << "]" << endl;
        ;
    }
    if (_blog)
        cout << " ** End Of TVQ::shuffle nS[" << nS << "]" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/

URANIE::DataServer::TDataServer * TVQ::getSubSample(Option_t *option)
{
    generateSample(option);
    return _subtds;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TVQ::printLog(Option_t *option)
{
    cout << "TVQ::PrintLog[" << option << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

// fin du fichier $RCSfile$.
