//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVQ.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jui 21 14:53:46 2009
 
 \brief  Define the abstract class for Vectorial Quantification (VQ) algorithms
 
 
 */
#ifndef TVQ_H
#define TVQ_H

// ROOT
#include "TVector.h"

// Uranie
#include "TSamplerStochastic.h"
#include "UMessageLogger.h"

/*!
 \class URANIE::Sampler::TVQ
 \brief Abstract class for Vectorial Quantification (VQ) algorithms

 The \b generateSample method of these VQ classes can be considered as a machine with 3 states :
 \par Implementation
 The virtual methods to implement for a new VQ class are the \b learn(index)  and \b generateSample methods.

 \par Implemented algorithms :
 \li  \b TNeuralGas : The Neural Gas algorithm (Martinetz and Schulten, 1991)
 
 */

namespace URANIE
{
namespace Sampler
{
class TVQ: public TSamplerStochastic
{
    // Attributes
protected:
    // the mutable declaration is needed to use the logger in const methods
    URANIE::DataServer::UMessageLogger *_fLogger; // message logger
    Bool_t _bdrawProgressBar; ///< Boolean to know if the progress bar has to be drawn
public:
    TVector _vecindexorg;              ///<  The vector of index of the database
    URANIE::DataServer::TDataServer *_subtds; ///< The subsampler returened by the getSubSample method

    Bool_t _bsaveall; ///< The boolean to save all the data in the learning (default = kFALSE)
    TString _sattributes;                        ///< the list of attributes

    // Methods
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    /*!
     \param tds (URANIE::DataServer::TDataServer *) the original data
     \param option (Option_t) contains the options
     \li \em "saveall" Save all the data generate by the algorithm ( _bsaveall = kTRUE).
     When is TRUE, all points during the learning process are store in the TDataServer,
     and when is FALSE, only points at the last iteration are store.
     \param ns (Int_t) the size of the subsample
     */
    TVQ(URANIE::DataServer::TDataServer *tds, Option_t *option, Int_t ns);
    //! Default destructor
    virtual ~TVQ();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    //! Returns the flag SaveAll
    /*!
     \li When is TRUE, all points during the learning process are store in the TDataServer.
     \li When is FALSE, only points at the last iteration are store in the TDataServer.
     */
    Bool_t getSaveAll()
    {
        return _bsaveall;
    }

    //! Set the "draw progress bar" flag
    /*!
     The progress bar will be drawn if and only if the "draw progress bar"
     flag is true.
     */
    void setDrawProgressBar(Bool_t bbool = kTRUE)
    {
        _bdrawProgressBar = bbool;
    }

    //! Get the \em clean flag
    /*!
     \return (Bool_t) the \em clean flag (boolean)
     */
    Bool_t getDrawProgressBar()
    {
        return _bdrawProgressBar;
    }
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{
    //! Creates the list of attributes to select patterns
    void createListOfAttributes();
    //! shuffle the learning database (original TDataServer)
    void shuffle();
    //@}

    //---------------------------------------------
    /** @name Manipulate the sample
     The method to get the subsample TDataServer
     **/
    //@{
    //! Generate the Sample and return the TDataServer 
    /*!
     \param option (Option_t) contains the options
     \li \em "loop=" Defines the number of loop to stop the algorithm. ("loop=10" for example)
     */
    URANIE::DataServer::TDataServer * getSubSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name State machine
     For the learning process, three states can be created
     \li \b init : execute at the begining of the learning
     \li \b learn (index) : the learning step for the pattern number index
     \li \b terminate : the terminate step executing at the end of the learning
     The \b init and \b terminate step  can be used for pre and post treatment purpose.
     **/
    //@{
    //! The pre-processing step
    virtual void init();
    //! The post-processing step
    virtual void terminate();
    //! The learning step for a pattern
    virtual void learn(Int_t index) = 0;
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TVQ, ID_SAMPLER)
    //Classe abstraite de g�n�ration de plan d'exp�rience num�rique
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
