//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTImportanceSampling.cxx
 \author Maud Salvoldi <Areva/OCC>
 \date   $Date$
 
 \brief   Fichier de test de la classe URANIE::DataServer::TImportanceSampling
 
 Created: lun jui 18 09:41:03 CEST 2011
 
 */

// CPPUNIT
#include <iostream>
#define DBL_PRECISION_COMPARE 1.e-10

// Uranie
#include "TestTImportanceSampling.h"
#include "UDisableTest.h"

using namespace std;

TImportanceSamplingTest::TImportanceSamplingTest() :
        _nS(1000)
{
 
}

TImportanceSamplingTest::~TImportanceSamplingTest()
{

}

void TImportanceSamplingTest::setUp()
{
    _tds = new URANIE::DataServer::TDataServer("TU_IS",
					       "test du tirage importance");
    _tds->addAttribute(
	new URANIE::DataServer::TNormalDistribution("rw", 0.0, 1.0));
    _tds->addAttribute(
	new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0));

    _tds->getAttribute("r")->delShare();
    _tds->getAttribute("rw")->delShare();

    _tdstmp = new URANIE::DataServer::TDataServer("tds_flowrate_is",
						  "tds avec importance sampling");    
    _tdstmp->fileDataRead(TString("TDS_Flowrate_isU.dat"));
    _tuple = _tdstmp->getTuple();
    
    _att = new URANIE::DataServer::TNormalDistribution("rw_IS", 2.0, 1.0);
    _tis = new URANIE::Sampler::TImportanceSampling(_tds, "rw", _att, _nS);

    _tdsis = _tis->getTDS();
    _tdsis->setTuple(_tuple);
}

void TImportanceSamplingTest::tearDown()
{

  if (_tis)
    {      
      delete _tis;
      _tis = NULL;
    }

  delete _att;
  _att=NULL;
  
   if(_tds!=NULL)
    delete _tds;
  _tds=NULL;

  if(_tdstmp!=NULL)
    delete _tdstmp;
  _tdstmp=NULL;  
    
}
void TImportanceSamplingTest::throwExceptionTImportanceSamplingTdsNull()
{

    URANIE::DataServer::TDataServer * toto = NULL;    
    URANIE::DataServer::TNormalDistribution *att = new URANIE::DataServer::TNormalDistribution("rw_IS", 0.10, 0.015);
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TImportanceSampling(toto, "rw", att , 100),
			 URANIE::Exceptions::UErrorExceptions);
    delete att;

}

void TImportanceSamplingTest::throwExceptionTImportanceSamplingNameUnknow()
{

  URANIE::DataServer::TNormalDistribution *att = new URANIE::DataServer::TNormalDistribution("rw_IS", 0.10, 0.015);
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TImportanceSampling(_tds, "toto", att, 100),
			 URANIE::Exceptions::UErrorExceptions);

    delete att;
}
void TImportanceSamplingTest::testgetTDS()
{

    // verify the good number of Attributes
    CPPUNIT_ASSERT_EQUAL(_tds->getNAttributes(), _tdsis->getNAttributes());

    // verif that the attribut rw_new is added  
    CPPUNIT_ASSERT(_tdsis->isAttribute("rw_IS"));
    // verif that the attribut rw_new is not present in the original TDS  
    CPPUNIT_ASSERT(_tds->isAttribute("rw_IS") == kFALSE);
    // verif that the original attribut rw is presented in the original TDS  
    CPPUNIT_ASSERT(_tds->isAttribute("rw"));

    // verify that rw is the son of rw_new
    TString _name = _tdsis->getAttribute("rw_IS")->getSonAttribute()->GetName();
    CPPUNIT_ASSERT(!strcmp(_name.Data(), "rw"));

}
void TImportanceSamplingTest::testgenerateSample()
{
    _tis->generateSample();
    // verify the good number of patterns
    CPPUNIT_ASSERT_EQUAL(_nS, _tdsis->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(0, _tds->getNPatterns());

}

void TImportanceSamplingTest::throwExceptioncomputeMeanFalseName()
{
    CPPUNIT_ASSERT_THROW(_tis->computeMean("toto"),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::testcomputeMean()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer("tds", "ksjdj");
    tds->addAttribute( new URANIE::DataServer::TNormalDistribution("nor", 0.0, 1.0));
    tds->getAttribute("nor")->delShare();
      
    int nS = 2000;
    URANIE::DataServer::TNormalDistribution *att = new URANIE::DataServer::TNormalDistribution("nor_IS", 0.1, 0.2);
    URANIE::Sampler::TImportanceSampling *tisCrt = new URANIE::Sampler::TImportanceSampling(tds, "nor", att, _nS);

    tisCrt->generateSample();

    Double_t dmean = tisCrt->computeMean("nor_IS");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, dmean, 0.8);

    tisCrt->getTDS()->addAttribute("y", "nor_IS*nor_IS");
    Double_t dmeanY = tisCrt->computeMean("y");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, dmeanY, 5.0);

    delete tisCrt;
    tisCrt = NULL;

    delete tds;
    delete att;
}

void TImportanceSamplingTest::throwExceptioncomputeStdFalseName()
{
    CPPUNIT_ASSERT_THROW(_tis->computeStd("toto"),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::throwExceptionTestcomputeStdEmptySample()
{
  _tdsis->deleteTuple();

  CPPUNIT_ASSERT_THROW(_tis->computeStd("rw_IS2"),  URANIE::Exceptions::UErrorExceptions);  

}

void TImportanceSamplingTest::testcomputeStd()
{
    // FGA::WARNING  2012-03-13 16:36:36
    // DO NOR USE The computeStd method
    URANIE_INBUILDING_TEST();
    // End Of FGA::WARNING  2012-03-13 16:36:36

    _tis->generateSample("lhs");

//   _tis->setLog();

    Double_t dsigma = _tis->computeStd("rw_IS");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(38.7115, dsigma, 1e-4);

    _tdsis->addAttribute("y", "rw_IS*rw_IS");
    dsigma = _tis->computeStd("y");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, dsigma, 1e-5);

}
void TImportanceSamplingTest::throwExceptioncomputeThresholdFalseName()
{
    CPPUNIT_ASSERT_THROW(_tis->computeThreshold("toto", 0.1),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::testcomputeThreshold()
{
    _tis->generateSample("lhs");

    Double_t dproba = _tis->computeThreshold("rw_IS", 1.96);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.025, dproba, 4e-3);

    dproba = _tis->computeThreshold("rw_IS", 0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.50, dproba, 0.1);

    _tdsis->addAttribute("y", "rw_IS*rw_IS");
    dproba = _tis->computeThreshold("y", 6.5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0055, dproba, 1e-3);
}

void TImportanceSamplingTest::throwExceptionComputeQuantileFalseName()
{
  CPPUNIT_ASSERT_THROW(_tis->computeQuantile("toto", 0.1, URANIE::Sampler::TImportanceSampling::kQuantile1),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::throwExceptionComputeQuantileFalseAlpha()
{
    _tis->generateSample("lhs");

    CPPUNIT_ASSERT_THROW(_tis->computeQuantile("rw_IS", 2.0,  
			       URANIE::Sampler::TImportanceSampling::kQuantile1),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_tis->computeQuantile("rw_IS", -1.0,  
			       URANIE::Sampler::TImportanceSampling::kQuantile1),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::throwExceptionComputeQuantileFalseMethod()
{
    _tis->generateSample("lhs");

    CPPUNIT_ASSERT_THROW(_tis->computeQuantile("rw_IS", 0.1, 
			       URANIE::Sampler::TImportanceSampling::kQuantile3),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_tis->computeQuantile("rw_IS", 0.1,  
			       URANIE::Sampler::TImportanceSampling::kQuantile4),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(_tis->computeQuantile("rw_IS", 0.1,  
			       URANIE::Sampler::TImportanceSampling::kQuantileUnknown),
            URANIE::Exceptions::UErrorExceptions);
}

void TImportanceSamplingTest::testComputeQuantile()
{
    // A cross-check need to be implemented to validate this test
    URANIE_INBUILDING_TEST();

    _tis->generateSample("lhs");

    Double_t q1 = _tis->computeQuantile("rw_IS", 0.1,  
			       URANIE::Sampler::TImportanceSampling::kQuantile1);

    Double_t q2 = _tis->computeQuantile("rw_IS", 0.1,  
			       URANIE::Sampler::TImportanceSampling::kQuantile2);

    cout << q1 << " " << q2<< endl;

}
