//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTOATSamplingMultiNominal.cxx
 \author Nicolas Gilardi
 \date   $Date$

 \brief  Test file of the class URANIE::Sampler::TOATSamplingMultiNominal

 Created: 2012-08-21 

 */

// CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-6

// Uranie
#include "TAttribute.h"
#include "TNormalDistribution.h"
#include "TestTOATSamplingMultiNominal.h"

//ROOT
#include "TMath.h"

using namespace std;

TOATSamplingMultiNominalTest::TOATSamplingMultiNominalTest()
{
}

TOATSamplingMultiNominalTest::~TOATSamplingMultiNominalTest()
{
}

void TOATSamplingMultiNominalTest::setUp()
{
}

void TOATSamplingMultiNominalTest::tearDown()
{
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TOATSamplingMultiNominalTest::testConstructor()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "regular",
                    0), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "regular",
                    -1), URANIE::Exceptions::UErrorExceptions);
}

void TOATSamplingMultiNominalTest::testSetOptions()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "gromf", 2),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "lhs", 1),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "regular",
                    7), URANIE::Exceptions::UErrorExceptions);
}

void TOATSamplingMultiNominalTest::testGetHasData()
{
    cout << endl;

    // TDS with data
    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 4);
    CPPUNIT_ASSERT(mySampling->getHasData());
    delete mySampling;
    mySampling = NULL;

    // TDS without data
    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    x2.setDefaultValue(-3.0);

    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem, "lhs",
            2);
    CPPUNIT_ASSERT(mySampling->getHasData() == kFALSE);
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testGetNumberOfModifications()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 4);
    CPPUNIT_ASSERT(mySampling->getNumberOfModifications() == 4);
    delete mySampling;
    mySampling = NULL;

    // Random mode accepts odd number of modifications
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "random", 7);
    CPPUNIT_ASSERT(mySampling->getNumberOfModifications() == 7);
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testGetSamplingMode()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "regular");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "random", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "srs",
            2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "lhs",
            2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "random");
    delete mySampling;
    mySampling = NULL;

    // Unexpected string containing valid sub string: 
    // "ASLhsKCjk" contains "Lhs" -> "random"; 
    // "sRSTDEre" contains "sRS" -> "random";
    // "TDEre GAjnchdu" contains "re G" -> "regular";
    // The string "reg" is searched first and thus "regular" mode should be chosen.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "ASLhsKCjk kshze,::sRSTDEre GAjnchdu hdzdg", 2);
    CPPUNIT_ASSERT(mySampling->getSamplingMode() == "regular");
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testGetGenerationMode()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Normal situation for "regular" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "range");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "random" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "random", 2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "srs");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "srs" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "srs",
            2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "srs");
    delete mySampling;
    mySampling = NULL;

    // Normal situation for "lhs" mode.
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile, "lhs",
            2);
    CPPUNIT_ASSERT(mySampling->getGenerationMode() == "lhs");
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testSetRangeDouble()
{
    cout << endl;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);
    URANIE::DataServer::TAttribute x3("x3");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    tdsMem.addAttribute(&x3);
    x3.setDefaultValue(1.0);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem,
            "random", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x3 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x3", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // This one should pass without problem:
    mySampling->setRange("x1", 0.5);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem,
            "regular", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", 0.5);
    mySampling->setRange("x3", 10.0, kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // "random" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "random", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x1 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);
    // Exception: unknown attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", 0.5),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: range is negative or null    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 0.0),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", -0.5),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", 0.5);
    mySampling->setRange("x2", 10.0, kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", 1.5),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testSetRangeString()
{
    cout << endl;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);
    URANIE::DataServer::TAttribute x3("x3");
    URANIE::DataServer::TAttribute r1("r1");
    URANIE::DataServer::TAttribute r2("r2");
    URANIE::DataServer::TAttribute r3("r3");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(0.0);
    tdsMem.addAttribute(&x2);
    tdsMem.addAttribute(&x3);
    x3.setDefaultValue(1.0);
    tdsMem.addAttribute(&r1);
    r1.setDefaultValue(0.1);
    tdsMem.addAttribute(&r2);
    r2.setDefaultValue(0.0);
    tdsMem.addAttribute(&r3);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem,
            "random", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r2 has an invalid default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r3 has no default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r3"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x3 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x3", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // This one should pass without problem:
    mySampling->setRange("x1", "r1");
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem,
            "regular", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x2 has no default value
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x2", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r2 has an invalid default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r2"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: r3 has no default value    
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r3"),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", "r1");
    mySampling->setRange("x3", "r1", kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "r1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // "random" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "random", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: x1 is not a TStochasticAttribute object
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;

    // "regular" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);
    // Exception: unknown changing attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("gromf", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    // Exception: unknown range attribute 
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "gromf"),
            URANIE::Exceptions::UErrorExceptions);
    // These ones should pass without problem:
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", "x1", kTRUE);
    // Exception: x1 is already in the list
    CPPUNIT_ASSERT_THROW(mySampling->setRange("x1", "rx1"),
            URANIE::Exceptions::UErrorExceptions);
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testGetChangingAttList()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    TList* changingAttList = mySampling->getChangingAttList();

    CPPUNIT_ASSERT(changingAttList->At(0) == tdsFile.getAttribute("x1"));
    CPPUNIT_ASSERT(changingAttList->At(1) == tdsFile.getAttribute("x2"));
    delete mySampling;
    mySampling = NULL;
}

void TOATSamplingMultiNominalTest::testGetRangeInformation()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    TMap* rangeInfo = mySampling->getRangeInformation();

    CPPUNIT_ASSERT(rangeInfo->GetValue("x1")->GetName() == TString("rx1"));
#ifdef WIN32
    CPPUNIT_ASSERT( rangeInfo->GetValue("x2")->GetName() == TString("1.0000000000000000e+001%") );
#else
    CPPUNIT_ASSERT(
            rangeInfo->GetValue("x2")->GetName()
                    == TString("1.0000000000000000e+01%"));
#endif
    delete mySampling;
    mySampling = NULL;

}

void TOATSamplingMultiNominalTest::testGetNominalValueTuple()
{
    cout << endl;

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler 
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 2);

    // Set the range information
    mySampling->setRange("x1", "rx1");
    mySampling->setRange("x2", 10.0, kTRUE);

    // generate the sample (this creates the nominal value tuple)
    mySampling->generateSample();

    URANIE::DataServer::TDSNtupleD* nomVal = mySampling->getNominalValueTuple();

    Double_t * data = nomVal->getBranchData("x1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.0, data[1], DBL_PRECISION_COMPARE);

    data = nomVal->getBranchData("rx1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.4, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.2, data[1], DBL_PRECISION_COMPARE);

    data = nomVal->getBranchData("x2");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, data[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, data[1], DBL_PRECISION_COMPARE);
    delete mySampling;
    mySampling = NULL;

}

void TOATSamplingMultiNominalTest::testGenerateSample()
{
    cout << endl;

    //-----------------------------//
    // TDS constructed from a file //
    //-----------------------------//

    URANIE::DataServer::TDataServer tdsFile("tdsFile",
            "TDS for OAT sampling loaded from a file");
    tdsFile.fileDataRead("nominalValues.dat");

    // Create the sampler
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsFile,
            "regular", 4);

    // Set the range of attribute x1. It will be the value of attribute rx1.
    mySampling->setRange("x1", "rx1");
    // Set the range of attribute x2. It will be 10.0% of x2's nominal value.
    mySampling->setRange("x2", 10.0, kTRUE);
    // Generate an OAT sampling where only x1 and x2 will vary.
    mySampling->generateSample();

    CPPUNIT_ASSERT(tdsFile.getNAttributes() == 3 + 2); // 3 original attributes plus __nominal_set__ and __modified_att__ . 
    CPPUNIT_ASSERT(tdsFile.getNPatterns() == 2 + (2 * 2 * 4)); // 2 nominal sets, 2 changing attributes, 4 generated values for each.    

    Double_t genData1[90];
    Double_t refData1[90] =
    { 3.00, 0.4, 2.000, 1.0, -1.0, 2.80, 0.4, 2.000, 1.0, 0.0, 3.20, 0.4, 2.000,
            1.0, 0.0, 2.90, 0.4, 2.000, 1.0, 0.0, 3.10, 0.4, 2.000, 1.0, 0.0,
            3.00, 0.4, 1.900, 1.0, 2.0, 3.00, 0.4, 2.100, 1.0, 2.0, 3.00, 0.4,
            1.950, 1.0, 2.0, 3.00, 0.4, 2.050, 1.0, 2.0, 5.00, 0.2, 3.000, 2.0,
            -1.0, 4.90, 0.2, 3.000, 2.0, 0.0, 5.10, 0.2, 3.000, 2.0, 0.0, 4.95,
            0.2, 3.000, 2.0, 0.0, 5.05, 0.2, 3.000, 2.0, 0.0, 5.00, 0.2, 2.850,
            2.0, 2.0, 5.00, 0.2, 3.150, 2.0, 2.0, 5.00, 0.2, 2.925, 2.0, 2.0,
            5.00, 0.2, 3.075, 2.0, 2.0 };

    tdsFile.getTuple()->extractData(genData1, 90,
            "x1:rx1:x2:__nominal_set__:__modified_att__", "", "line");
    for (Int_t i = 0; i < 90; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refData1[i], genData1[i],
                DBL_PRECISION_COMPARE);

    delete mySampling;
    mySampling = NULL;

    //-----------------------------------------//
    // TDS and attributes constructed manually //
    //-----------------------------------------//

    URANIE::DataServer::TNormalDistribution x1("x1", 0.0, 1.0);
    URANIE::DataServer::TNormalDistribution x2("x2", -3.0, 2.0);
    URANIE::DataServer::TAttribute r1("r1");

    URANIE::DataServer::TDataServer tdsMem("tdsMem", "");
    tdsMem.addAttribute(&x1);
    x1.setDefaultValue(5.0);
    tdsMem.addAttribute(&x2);
    x2.setDefaultValue(-3.0);
    tdsMem.addAttribute(&r1);
    r1.setDefaultValue(20.0);

    // "random" mode
    mySampling = new URANIE::Sampler::TOATSamplingMultiNominal(&tdsMem, "lhs",
            9);
    mySampling->setSeed(1234);
    mySampling->setRange("x1", "r1", kTRUE);
    mySampling->generateSample();

    CPPUNIT_ASSERT(tdsMem.getNAttributes() == 3 + 2); // 3 original attributes plus __nominal_set__ and __modified_att__ . 
    CPPUNIT_ASSERT(tdsMem.getNPatterns() == 1 + (1 * 1 * 9)); // 1 nominal set, 1 changing attribute, 9 generated values.    

    Double_t genData2[10];
    Double_t refData2[10] =
    { 5.0000000, 4.5203057, 5.2626381, 4.8355680, 4.6769375, 4.5623774,
            4.7407237, 4.6593369, 4.6161644, 4.5527299 };

    tdsMem.getTuple()->extractData(genData2, 10, "x1", "", "line");
    for (Int_t i = 0; i < 10; i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refData2[i], genData2[i],
                DBL_PRECISION_COMPARE);

    delete mySampling;
    mySampling = NULL;
}

