//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSpaceFilling.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSpaceFilling class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTSpaceFilling.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "TLogUniformDistribution.h"
#include "TLogNormalDistribution.h"
#include "TTriangularDistribution.h"
#include "TLogTriangularDistribution.h"
#include "TExponentialDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

using namespace std;

TSpaceFillingTest::TSpaceFillingTest() :
         _nXDefault(0), _nXCukier(0), _nXSaltelliB(0), _tdsDefault(NULL), _tdsCukier(NULL),  _tdsSaltelliB(NULL)
{


}

TSpaceFillingTest::~TSpaceFillingTest()
{
 
}
void TSpaceFillingTest::setUp()
{

// Create a TDataServer
    _tdsDefault = new URANIE::DataServer::TDataServer();
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TLogUniformDistribution("loguniform", .001,
							10.));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TTriangularDistribution("triangular", 0.0,
							1.0, 0.5));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TLogTriangularDistribution("logtriangular",
							   0.01, 10., 5.0));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TExponentialDistribution("exp", .1, 12.0));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TWeibullDistribution("Weibull", 0.2, 0.4));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
							    1.0, 4.0, 2.0));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TBetaDistribution("Beta", 0.5, 0.5));
    _tdsDefault->addAttribute(
	new URANIE::DataServer::TCauchyDistribution("Cauchy", 1.0, 2.0));
    _tdsDefault->getAttribute("Uniform")->delShare();
    _tdsDefault->getAttribute("loguniform")->delShare();
    _tdsDefault->getAttribute("triangular")->delShare();
    _tdsDefault->getAttribute("logtriangular")->delShare();
    _tdsDefault->getAttribute("exp")->delShare();
    _tdsDefault->getAttribute("Weibull")->delShare();
    _tdsDefault->getAttribute("UnifParts")->delShare();
    _tdsDefault->getAttribute("Beta")->delShare();
    _tdsDefault->getAttribute("Cauchy")->delShare();
    _nXDefault = _tdsDefault->getNInputAttributes();

    // Create a TDataServer
    _tdsCukier = new URANIE::DataServer::TDataServer();
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TLogUniformDistribution("loguniform", .001,
							10.));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TLogNormalDistribution("lognormal", 0.01,
						       3.0));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TTriangularDistribution("triangular", 0.0,
							1.0, 0.5));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TLogTriangularDistribution("logtriangular",
							   0.01, 10., 5.0));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TExponentialDistribution("exp", .1, 12.0));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TWeibullDistribution("Weibull", 0.2, 0.4));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
							    1.0, 4.0, 2.0));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TBetaDistribution("Beta", 0.5, 0.5));
    _tdsCukier->addAttribute(
	new URANIE::DataServer::TCauchyDistribution("Cauchy", 1.0, 2.0));
    _tdsCukier->getAttribute("Uniform")->delShare();
    _tdsCukier->getAttribute("loguniform")->delShare();
    _tdsCukier->getAttribute("Normal")->delShare();
    _tdsCukier->getAttribute("lognormal")->delShare();
    _tdsCukier->getAttribute("triangular")->delShare();
    _tdsCukier->getAttribute("logtriangular")->delShare();
    _tdsCukier->getAttribute("exp")->delShare();
    _tdsCukier->getAttribute("Weibull")->delShare();
    _tdsCukier->getAttribute("UnifParts")->delShare();
    _tdsCukier->getAttribute("Beta")->delShare();
    _tdsCukier->getAttribute("Cauchy")->delShare();
    _nXCukier = _tdsCukier->getNInputAttributes();

    // Create a TDataServer
    _tdsSaltelliB = new URANIE::DataServer::TDataServer();
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TLogUniformDistribution("loguniform", .001,
							10.));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TLogNormalDistribution("lognormal", 0.01,
						       3.0));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TTriangularDistribution("triangular", 0.0,
							1.0, 0.5));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TLogTriangularDistribution("logtriangular",
							   0.01, 10., 5.0));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TExponentialDistribution("exp", .1, 12.0));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TWeibullDistribution("Weibull", 0.2, 0.4));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
							    1.0, 4.0, 2.0));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TBetaDistribution("Beta", 0.5, 0.5));
    _tdsSaltelliB->addAttribute(
	new URANIE::DataServer::TCauchyDistribution("Cauchy", 1.0, 2.0));
    _tdsSaltelliB->getAttribute("Uniform")->delShare();
    _tdsSaltelliB->getAttribute("loguniform")->delShare();
    _tdsSaltelliB->getAttribute("Normal")->delShare();
    _tdsSaltelliB->getAttribute("lognormal")->delShare();
    _tdsSaltelliB->getAttribute("triangular")->delShare();
    _tdsSaltelliB->getAttribute("logtriangular")->delShare();
    _tdsSaltelliB->getAttribute("exp")->delShare();
    _tdsSaltelliB->getAttribute("Weibull")->delShare();
    _tdsSaltelliB->getAttribute("UnifParts")->delShare();
    _tdsSaltelliB->getAttribute("Beta")->delShare();
    _tdsSaltelliB->getAttribute("Cauchy")->delShare();
    _nXSaltelliB = _tdsSaltelliB->getNInputAttributes();

    _samplerDefault = new URANIE::Sampler::TSpaceFilling(_tdsDefault,
							 "SpaceFilling", 100);
    _samplerDefault->unsetLog();
    _samplerDefault->setDefaultFrequencies();

    _samplerCukier = new URANIE::Sampler::TSpaceFilling(_tdsCukier,
							"SpaceFilling", 100, URANIE::Sampler::TSpaceFilling::kCukier);
    _samplerCukier->unsetLog();
    _samplerCukier->setDefaultFrequencies();

    _samplerSaltelliB = new URANIE::Sampler::TSpaceFilling(_tdsSaltelliB,
							   "SpaceFilling", 100, URANIE::Sampler::TSpaceFilling::kSaltelliB);
    _samplerSaltelliB->unsetLog();
    _samplerSaltelliB->setDefaultFrequencies();
}

void TSpaceFillingTest::tearDown()
{
    delete _samplerDefault;
    _samplerDefault = NULL;
    delete _samplerSaltelliB;
    _samplerSaltelliB = NULL;
    delete _samplerCukier;
    _samplerCukier = NULL;
    delete _tdsDefault;
    _tdsDefault = NULL;
    delete _tdsCukier;
    _tdsCukier = NULL;
    delete _tdsSaltelliB;
    _tdsSaltelliB = NULL;
}

void TSpaceFillingTest::testTSpaceFillingConstructor()
{
    //default function option (i.e kSaltelliA)  
    TString sname = "SpaceFilling_1";
    CPPUNIT_ASSERT_EQUAL(0,
            strcmp(sname.Data(), _samplerDefault->getMethodName()));
    CPPUNIT_ASSERT_EQUAL(kFALSE, _samplerDefault->getPermutation());
    CPPUNIT_ASSERT_EQUAL(URANIE::Sampler::TSpaceFilling::kSaltelliA,
            _samplerDefault->getTransformFunction());
    CPPUNIT_ASSERT_EQUAL(19, _samplerDefault->getFrequency(0));
    CPPUNIT_ASSERT_EQUAL(59, _samplerDefault->getFrequency(1));
    CPPUNIT_ASSERT_EQUAL(91, _samplerDefault->getFrequency(2));
    CPPUNIT_ASSERT_EQUAL(113, _samplerDefault->getFrequency(3));
    CPPUNIT_ASSERT_EQUAL(133, _samplerDefault->getFrequency(4));
    CPPUNIT_ASSERT_EQUAL(143, _samplerDefault->getFrequency(5));
    CPPUNIT_ASSERT_EQUAL(149, _samplerDefault->getFrequency(6));
    CPPUNIT_ASSERT_EQUAL(157, _samplerDefault->getFrequency(7));
    CPPUNIT_ASSERT_EQUAL(161, _samplerDefault->getFrequency(8));
    CPPUNIT_ASSERT_EQUAL(_nXDefault, _samplerDefault->_nX);

    //function option kCukier 
    sname = "SpaceFilling_0";
    CPPUNIT_ASSERT_EQUAL(0,
            strcmp(sname.Data(), _samplerCukier->getMethodName()));
    CPPUNIT_ASSERT_EQUAL(kFALSE, _samplerCukier->getPermutation());
    CPPUNIT_ASSERT_EQUAL(URANIE::Sampler::TSpaceFilling::kCukier,
            _samplerCukier->getTransformFunction());
    CPPUNIT_ASSERT_EQUAL(41, _samplerCukier->getFrequency(0));
    CPPUNIT_ASSERT_EQUAL(67, _samplerCukier->getFrequency(1));
    CPPUNIT_ASSERT_EQUAL(105, _samplerCukier->getFrequency(2));
    CPPUNIT_ASSERT_EQUAL(145, _samplerCukier->getFrequency(3));
    CPPUNIT_ASSERT_EQUAL(177, _samplerCukier->getFrequency(4));
    CPPUNIT_ASSERT_EQUAL(199, _samplerCukier->getFrequency(5));
    CPPUNIT_ASSERT_EQUAL(219, _samplerCukier->getFrequency(6));
    CPPUNIT_ASSERT_EQUAL(229, _samplerCukier->getFrequency(7));
    CPPUNIT_ASSERT_EQUAL(235, _samplerCukier->getFrequency(8));
    CPPUNIT_ASSERT_EQUAL(243, _samplerCukier->getFrequency(9));
    CPPUNIT_ASSERT_EQUAL(247, _samplerCukier->getFrequency(10));
    CPPUNIT_ASSERT_EQUAL(_nXCukier, _samplerCukier->_nX);

    //function option kSaltelliB 
    sname = "SpaceFilling_2";
    CPPUNIT_ASSERT_EQUAL(0,
            strcmp(sname.Data(), _samplerSaltelliB->getMethodName()));
    CPPUNIT_ASSERT_EQUAL(kFALSE, _samplerSaltelliB->getPermutation());
    CPPUNIT_ASSERT_EQUAL(URANIE::Sampler::TSpaceFilling::kSaltelliB,
            _samplerSaltelliB->getTransformFunction());
    CPPUNIT_ASSERT_EQUAL(41, _samplerSaltelliB->getFrequency(0));
    CPPUNIT_ASSERT_EQUAL(67, _samplerSaltelliB->getFrequency(1));
    CPPUNIT_ASSERT_EQUAL(105, _samplerSaltelliB->getFrequency(2));
    CPPUNIT_ASSERT_EQUAL(145, _samplerSaltelliB->getFrequency(3));
    CPPUNIT_ASSERT_EQUAL(177, _samplerSaltelliB->getFrequency(4));
    CPPUNIT_ASSERT_EQUAL(199, _samplerSaltelliB->getFrequency(5));
    CPPUNIT_ASSERT_EQUAL(219, _samplerSaltelliB->getFrequency(6));
    CPPUNIT_ASSERT_EQUAL(229, _samplerSaltelliB->getFrequency(7));
    CPPUNIT_ASSERT_EQUAL(235, _samplerSaltelliB->getFrequency(8));
    CPPUNIT_ASSERT_EQUAL(243, _samplerSaltelliB->getFrequency(9));
    CPPUNIT_ASSERT_EQUAL(247, _samplerSaltelliB->getFrequency(10));
    CPPUNIT_ASSERT_EQUAL(_nXSaltelliB, _samplerSaltelliB->_nX);
}

void TSpaceFillingTest::throwExceptionTSpaceFillingConstructorTdsNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TSpaceFilling(tdsnull, "spacefillingnull",
                    100), URANIE::Exceptions::UErrorExceptions);
}

void TSpaceFillingTest::testTSpaceFillingsetTransformFunctionANDgetTransformFunction()
{
    _samplerDefault->setTransformFunction(
            URANIE::Sampler::TSpaceFilling::kSaltelliB);
    CPPUNIT_ASSERT_EQUAL(URANIE::Sampler::TSpaceFilling::kSaltelliB,
            _samplerDefault->getTransformFunction());

    _samplerDefault->setTransformFunction(
            URANIE::Sampler::TSpaceFilling::kCukier);
    CPPUNIT_ASSERT_EQUAL(URANIE::Sampler::TSpaceFilling::kCukier,
            _samplerDefault->getTransformFunction());
}
void TSpaceFillingTest::testTSpaceFillingsetPermutationANDgetPermutation()
{
    _samplerDefault->setPermutation(kTRUE);
    CPPUNIT_ASSERT_EQUAL(kTRUE, _samplerDefault->getPermutation());

}
void TSpaceFillingTest::throwExceptionTSpaceFillingsetPermutationBadValue()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    CPPUNIT_ASSERT_THROW(_samplerDefault->setPermutation(20),
            URANIE::Exceptions::UErrorExceptions);

}
void TSpaceFillingTest::testTSpaceFillingsetFrequenciesANDgetFrequency()
{
    Int_t * nFreq = new Int_t[_nXSaltelliB];
    nFreq[0] = 11;
    for (Int_t i = 1; i < _nXSaltelliB; i++)
        nFreq[i] = nFreq[i - 1] + 10 * i;
    _samplerSaltelliB->setFrequencies(_nXSaltelliB, nFreq);
    CPPUNIT_ASSERT_EQUAL(11, _samplerSaltelliB->getFrequency(0));
    CPPUNIT_ASSERT_EQUAL(21, _samplerSaltelliB->getFrequency(1));
    CPPUNIT_ASSERT_EQUAL(41, _samplerSaltelliB->getFrequency(2));
    CPPUNIT_ASSERT_EQUAL(71, _samplerSaltelliB->getFrequency(3));
    CPPUNIT_ASSERT_EQUAL(111, _samplerSaltelliB->getFrequency(4));
    CPPUNIT_ASSERT_EQUAL(161, _samplerSaltelliB->getFrequency(5));
    CPPUNIT_ASSERT_EQUAL(221, _samplerSaltelliB->getFrequency(6));
    CPPUNIT_ASSERT_EQUAL(291, _samplerSaltelliB->getFrequency(7));
    CPPUNIT_ASSERT_EQUAL(371, _samplerSaltelliB->getFrequency(8));
    CPPUNIT_ASSERT_EQUAL(461, _samplerSaltelliB->getFrequency(9));
    CPPUNIT_ASSERT_EQUAL(561, _samplerSaltelliB->getFrequency(10));

    delete[] nFreq;
    nFreq = NULL;
}
void TSpaceFillingTest::throwExceptionTSpaceFillinggetFrequencyHighIndex()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    Int_t * nFreq = new Int_t[_nXDefault];
    nFreq[0] = 0;
    for (Int_t i = 1; i < _nXDefault; i++)
        nFreq[i] = nFreq[i - 1] + 10 * i;

    CPPUNIT_ASSERT_THROW(_samplerDefault->setFrequencies(_nXDefault, nFreq),
            URANIE::Exceptions::UErrorExceptions);

    delete[] nFreq;

}
void TSpaceFillingTest::throwExceptionTSpaceFillingsetFrequencyHighIndex()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    Int_t * nFreq = new Int_t[_nXDefault + 3];
    nFreq[0] = 11;
    for (Int_t i = 1; i < _nXDefault; i++)
        nFreq[i] = nFreq[i - 1] + 10 * i;

    CPPUNIT_ASSERT_THROW(_samplerDefault->setFrequencies(_nXDefault + 3, nFreq),
            URANIE::Exceptions::UErrorExceptions);

    delete[] nFreq;

}
void TSpaceFillingTest::throwExceptionTSpaceFillingsetFrequencyLessIndex()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    Int_t * nFreq = new Int_t[_nXDefault - 2];
    nFreq[0] = 11;
    for (Int_t i = 1; i < _nXDefault; i++)
        nFreq[i] = nFreq[i - 1] + 10 * i;
    CPPUNIT_ASSERT_THROW(_samplerDefault->setFrequencies(_nXDefault - 2, nFreq),
            URANIE::Exceptions::UErrorExceptions);

    delete[] nFreq;
}
void TSpaceFillingTest::throwExceptionTSpaceFillingsetFrequencyBadValue()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    Int_t * nFreq = new Int_t[_nXDefault];
    nFreq[0] = 11;
    for (Int_t i = 1; i < _nXDefault; i++)
        nFreq[i] = nFreq[i - 1] - 10 * i;
    CPPUNIT_ASSERT_THROW(_samplerDefault->setFrequencies(_nXDefault - 2, nFreq),
            URANIE::Exceptions::UErrorExceptions);

    delete[] nFreq;

}
void TSpaceFillingTest::testTSpaceFillingenerateSample()
{
    _samplerDefault->generateSample();
    _tdsDefault->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.31, _tdsDefault->getValue("Uniform", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.69, _tdsDefault->getValue("Uniform", 99),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5,
            _tdsDefault->getAttribute("Uniform")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.290071394903762,
            _tdsDefault->getAttribute("Uniform")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.002290867653,
            _tdsDefault->getValue("loguniform", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.36515832240159,
            _tdsDefault->getValue("loguniform", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.08409424563184,
            _tdsDefault->getAttribute("loguniform")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.06530761927577,
            _tdsDefault->getAttribute("loguniform")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.452769256906871,
            _tdsDefault->getValue("triangular", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.547230743093128,
            _tdsDefault->getValue("triangular", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5,
            _tdsDefault->getAttribute("triangular")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.204593163407342,
            _tdsDefault->getAttribute("triangular")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.81364027901105,
            _tdsDefault->getValue("logtriangular", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.538089265191451,
            _tdsDefault->getValue("logtriangular", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.85266778689416,
            _tdsDefault->getAttribute("logtriangular")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.02201886730044,
            _tdsDefault->getAttribute("logtriangular")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(29.7195684193184,
            _tdsDefault->getValue("exp", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(13.863295781915,
            _tdsDefault->getValue("exp", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(21.9308519447223,
            _tdsDefault->getAttribute("exp")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(9.68841193035125,
            _tdsDefault->getAttribute("exp")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.30638610732143,
            _tdsDefault->getValue("Weibull", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0002837482351,
            _tdsDefault->getValue("Weibull", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.612174712361146,
            _tdsDefault->getAttribute("Weibull")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.49549807559774,
            _tdsDefault->getAttribute("Weibull")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.96, _tdsDefault->getValue("UnifParts", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.02, _tdsDefault->getValue("UnifParts", 99),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.25,
            _tdsDefault->getAttribute("UnifParts")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.882203393942024,
            _tdsDefault->getAttribute("UnifParts")->getStd(),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.987958380969374,
            _tdsDefault->getValue("Beta", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.01204161903,
            _tdsDefault->getValue("Beta", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5,
            _tdsDefault->getAttribute("Beta")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.35533452725935,//155789,
            _tdsDefault->getAttribute("Beta")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.77760685391507,
            _tdsDefault->getValue("Cauchy", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.777606853915069,
            _tdsDefault->getValue("Cauchy", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0,
            _tdsDefault->getAttribute("Cauchy")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.03526470681894,
            _tdsDefault->getAttribute("Cauchy")->getStd(),
            DBL_PRECISION_COMPARE);

}
void TSpaceFillingTest::testTSpaceFillinggenerateSampleCukier()
{

    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    //case Cukier
    _samplerCukier->generateSample();
    _tdsCukier->computeStatistic("*");
}
void TSpaceFillingTest::testTSpaceFillinggenerateSampleSaltelliB()
{

    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST()
    ;

    //case SaltelliB
    _samplerSaltelliB->generateSample();
    _tdsSaltelliB->computeStatistic("*");

}
void TSpaceFillingTest::testTSpaceFillinggenerateSampleNormalLaw()
{
    URANIE::DataServer::TDataServer *tds =
            new URANIE::DataServer::TDataServer();
    tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
    tds->addAttribute(
            new URANIE::DataServer::TLogNormalDistribution("lognormal", 0.01,
                    3.0));
    tds->getAttribute("Normal")->delShare();
    tds->getAttribute("lognormal")->delShare();
	
    URANIE::Sampler::TSpaceFilling *sampler =
            new URANIE::Sampler::TSpaceFilling(tds, "SpaceFilling", 100);
    sampler->unsetLog();
    sampler->generateSample();
    tds->computeStatistic("*");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.125661346855074, tds->getValue("Normal", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.125661346855074, tds->getValue("Normal", 99),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds->getAttribute("Normal")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.942695113796325,
            tds->getAttribute("Normal")->getStd(), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00687304948729101,
            tds->getValue("lognormal", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00931419379977499,
            tds->getValue("lognormal", 99), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00992276494512779,
            tds->getAttribute("lognormal")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0070378291189924,
            tds->getAttribute("lognormal")->getStd(), DBL_PRECISION_COMPARE);

    delete sampler;
    sampler = NULL;
    delete tds;
    tds = NULL;

}

void TSpaceFillingTest::testTSpaceFillingsetSize()
{
    _samplerDefault->setSize(50);
    _samplerDefault->generateSample();
    CPPUNIT_ASSERT_EQUAL(50, _tdsDefault->getNPatterns());

}
