//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCMN.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   ven oct 10 10:02:29 CEST 2008
 
 \brief Impl�mentation de la classe URANIE::Sensitivity::TCMN
 
 */
//// ROOT
#include "TPad.h"
#include "TProfile.h"
#include "TLeaf.h"
#include "TDirectory.h"
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,18,0)
#include "Math/Math.h"
#include "Math/ProbFuncMathCore.h"
#include "Math/SpecFuncMathCore.h"
#else
#include "TMath.h"
#endif        

//// Uranie
#include "TCMN.h"
#include "TSensitivity.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TCMN)

namespace URANIE
{
namespace Sensitivity
{
TCMN::TCMN()
{
    cout << "construit TCMN" << endl;
}

TCMN::TCMN(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t * option) :
        TSensitivityTest(tds, varexpinput, varexpoutput, option), _nBins(20)
{
    SetName("testCMD");
    SetTitle("my title of the CMN Test");
}

TCMN::~TCMN()
{
    cout << "detruit TCMN" << endl;
}

///////////////////////////////////////////////////////////
//  generateSample 
///////////////////////////////////////////////////////////
void TCMN::generateSample(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    cout << "  ** RAF ***" << endl;
}

///////////////////////////////////////////////////////////
//  computeIndexes 
///////////////////////////////////////////////////////////
void TCMN::computeIndexes(Option_t * option)
{
    cout << endl;
    cout << "*********************************" << endl;
    cout << " ** TCMN::computeIndexes" << endl;
    evaluateIndexes(option);

}

///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TCMN::evaluateIndexes(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    TIter nextouput(_listOfOutputAttributes);
    while (URANIE::DataServer::TAttribute *attouput =
            (URANIE::DataServer::TAttribute *) nextouput())
    {
        cout << " ** case output [" << attouput->GetName() << "]" << endl;

        Double_t dsumy = 0.0;
        Double_t dsumy2 = 0.0;

        // Compute the sum(y*y)
        TLeaf *leafY = _tds->getTuple()->GetLeaf(attouput->GetName());
        // Loop on the TDS for the current leaf : leafY
        for (Int_t ipat = 0; ipat < _nS; ipat++)
        {
            // get the current pattern
            _tds->getTuple()->GetEntry(ipat);

            dsumy += leafY->GetValue(0);
            dsumy2 += leafY->GetValue(0) * leafY->GetValue(0);
        }
        Double_t dmeany = dsumy / (Double_t)(_nS);
        cout << " ** dmeany[" << dmeany << "] dsumy2 [" << dsumy2 << "]"
                << endl;

        TIter nextinput(_listOfInputAttributes);
        while (URANIE::DataServer::TAttribute *attinput =
                (URANIE::DataServer::TAttribute *) nextinput())
        {
            cout << " **** input [" << attinput->GetName() << "]" << endl;
            TString sprofname = TString("gURANIE_drawProfile");
            sprofname += Form("%s__Versus__%s", attouput->GetName(),
                    attinput->GetName());
            cout << "sprofname <<" << sprofname << endl;
            _tds->draw(
                    Form("%s:%s>>%s(%d)", attouput->GetName(),
                            attinput->GetName(), sprofname.Data(), _nBins), "",
                    "goffprofs");
            _tds->draw(Form("%s:%s", attouput->GetName(), attinput->GetName()));
            TProfile *hprofs = (TProfile*) gDirectory->Get(sprofname);
            hprofs->SetMarkerColor(5);
            hprofs->SetMarkerSize(0.7);
            hprofs->SetMarkerStyle(21);
            hprofs->Draw("same");

            Double_t dpartialsum = 0.0;

            Int_t nbinsx = hprofs->GetNbinsX();
            Int_t nentries = 0;
            Int_t nclasses = 0;
            for (Int_t i = 0; i < nbinsx; i++)
            {
                if (hprofs->GetBinEntries(i) > 0)
                {
                    if (_blog)
                        cout << "  ** i " << i << "/" << nbinsx
                                << " *** Center[" << hprofs->GetBinContent(i)
                                << "]";
                    nentries += (Int_t)(hprofs->GetBinEntries(i));
                    if (_blog)
                        cout << " n=[" << hprofs->GetBinEntries(i) << "/"
                                << nentries << "]";
                    dpartialsum += hprofs->GetBinEntries(i)
                            * hprofs->GetBinContent(i)
                            * hprofs->GetBinContent(i);
                    if (_blog)
                        cout << endl;
                    nclasses++;
                }
            }
            Double_t dftest = (dpartialsum - (Double_t)(_nS) * dmeany * dmeany)
                    / (Double_t)(nclasses - 1);
            dftest /= (dsumy2 - dpartialsum) / (Double_t)(_nS - nclasses);
            //         Double_t pvalue = 1.0 - ROOT::Math::fdistribution_prob(dftest, nclasses-1, _nS - nclasses);

            /* fga::modif 2008-10-24 16:31:04
             pourcompil v5.18 _prob n'existe pas
             Double_t pvalue = ROOT::Math::fdistribution_prob(dftest, nclasses-1, _nS - nclasses);
             */

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,18,0)
            Double_t pvalue = ROOT::Math::fdistribution_cdf(dftest,
                    nclasses - 1, _nS - nclasses);
#else
            Double_t pvalue = TMath::FDist(dftest, nclasses-1, _nS - nclasses);
#endif        
            cout << " ***************** dftest[" << dftest << "] nseg["
                    << nclasses - 1 << "] nn[" << _nS - nclasses << "] pvalue["
                    << pvalue << "]" << endl;
            _dpvalue = pvalue;
        }
    }

}

void TCMN::printLog(Option_t *option)
{
    cout << "  ** TCMN::printLog[" << option << "]" << endl;
    cout << "  ** ns[" << _nS << "]" << endl;
    cout << "  ** nBins[" << _nBins << "]" << endl;
    TSensitivityTest::printLog(option);
    cout << "  ** End of TCMN::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

