//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDGSM.cxx
 \author GAUDIER Fabrice <fabrice.gaudier@cea.fr>
 \date   2019/06/14

 \brief Implementation of the class URANIE::Sensitivity::TDGSM

 */

// ROOT
#include "TROOT.h"
#include "TPRegexp.h"
#include "TMatrixD.h"

// Uranie
#include "TDGSM.h"
#include "TOATDesign.h"
#include "TOATSampling.h"
#include "TUniformDistribution.h"
#include "TStochAttribut.h"


//#include "TLauncher.h"
//#include "TLauncherFunction.h"
#include "TSampling.h"
#include "TQMC.h"

using std::cout;
using std::endl;

using namespace URANIE::DataServer;
using namespace URANIE::Sampler;

ClassImp (URANIE::Sensitivity::TDGSM)

namespace URANIE
{
namespace Sensitivity
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TDGSM::TDGSM() :
        TFiniteDifferences()
{
    setMethodName("DGSM");
    _sFunctionName = TString("");
    _sSamplingOption = TString("");
    _sSensitiveAtt = TString("");
    _sSelectedInput = TString("");
    _sSelectedOutput = TString("");

    _nbIn = 0;
    _nbAux = 0;
    _nbOut = 0;
    _nS = 0;

    _sIteratorName = "__modified_att__";
    if (1)
    {
        cout << " ___________________________________________________" << endl;
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
        cout << "---- nS                            = " << _nS << endl;
        cout << " *** _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " *** _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        cout << " ***  _sFunctionName [" << _sFunctionName << "]" << endl;
        cout << " ___________________________________________________" << endl;
    }
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
TDGSM::TDGSM(TDataServer *tdsNominal, void (*fcn)(Double_t*, Double_t*),
        Int_t ns, TString sensitiveAtt, TString outputAtt, TString samplingOption) :
    TFiniteDifferences(tdsNominal, fcn, sensitiveAtt.Data(), outputAtt.Data(), samplingOption)
{
    setMethodName("DGSM");
    _sIteratorName = "__modified_att__";
    _nS = ns;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
        cout << "---- nS                            = " << _nS << endl;
        cout << " ***  _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " *** _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        cout << " ***  _sFunctionName [" << _sFunctionName << "]" << endl;

    }

    parseOption(samplingOption);
}

TDGSM::TDGSM(TDataServer *tdsNominal, const char* fcn,
             Int_t ns, TString sensitiveAtt, TString outputAtt, TString samplingOption ):
    TFiniteDifferences(tdsNominal, fcn, sensitiveAtt.Data(), outputAtt.Data(), samplingOption)
{
    setMethodName("DGSM");
    _sIteratorName = "__modified_att__";
    _nS = ns;
    if (_blog)
    {
        cout << " ___________________________________________________" << endl;
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
        cout << "---- nS                            = " << _nS << endl;
        cout << " *** _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " *** _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        cout << " ***  _sFunctionName [" << _sFunctionName << "] fcn[" << fcn << "]" << endl;
        cout << " ___________________________________________________" << endl;
    }
    parseOption(samplingOption);
}



// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TDGSM::TDGSM(TDataServer *tdsNominal, URANIE::Launcher::TCode *fcode, Int_t ns,
             TString sensitiveAtt, TString samplingOption) :
        TFiniteDifferences(tdsNominal, fcode, sensitiveAtt, samplingOption)
{
    setMethodName("DGSM");
    _sIteratorName = "__modified_att__";
    _nS = ns;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- nS                            = " << _nS << endl;
        cout << " *** _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " *** _sSelectedOutput[" << _sSelectedOutput << "]" << endl;

    }
    parseOption(samplingOption);
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TDGSM::TDGSM(TDataServer *tdsNominal, URANIE::Relauncher::TRun *run, Int_t ns,
             TString sensitiveAtt, TString samplingOption) :
    TFiniteDifferences(tdsNominal, run, sensitiveAtt, samplingOption)
{
    setMethodName("DGSM");
    _sIteratorName = "__modified_att__";
    _nS = ns;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of output / output elements = " << _sSelectedOutput.CountChar(':') + 1 << "/" <<_nbOut << endl;
        cout << "---- nS                            = " << _nS << endl;
    }
    parseOption(samplingOption);
}

// -----------------------------------------------------------------------
//  Constructors with only data
// -----------------------------------------------------------------------
TDGSM::TDGSM(TDataServer *tdsNominal, TString inputAtt, TString outputAtt,
             TString sensitiveAtt) :
        TFiniteDifferences(tdsNominal, inputAtt, outputAtt, sensitiveAtt)
{
    setMethodName("DGSM");
    _sIteratorName = "__modified_att__";

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
        cout << "---- nS                            = " << _nS << endl;
    }
}

TDGSM::~TDGSM()
{
}


///////////////////////////////////////////////////////////
//  init the Sample when TDS have no data
///////////////////////////////////////////////////////////
void TDGSM::initSample(Int_t ns, Option_t * option)
{
    if (_blog)
        cout << endl << "***********************************" << endl
             << " ** TDGSM::initSample ns[" << ns << "] option[" << option << "] _sSamplingOption[" << _sSamplingOption << "]" << endl;
    TString soption = _sSamplingOption + " " + TString(option);
    soption.ToLower();

    if (_blog)
        cout << " ** soption[" << soption << "]" << endl;
    soption.ReplaceAll(",", " ");
    soption += TString(" ");

    enum ESamplerMethod
    {
        kMC, kQMC
    };

    // default values
    ESamplerMethod nSamplerMethod=kMC;
    TString sSamplingOption="srs";

    istringstream inStream((char *) soption.Data());
    string inName;
    while (inStream >> inName)
    {
        TString svar(inName.c_str());
        if (_blog)
            cout << " ** svar[" << svar << "]" << endl;

        // Case of the sampler
        if (svar == "srs") {
            nSamplerMethod = kMC;
            sSamplingOption="srs";
        }
        if (svar == "lhs") {
            nSamplerMethod = kMC;
            sSamplingOption="lhs";
        }
        if (svar == "qmc=sobol") {
            nSamplerMethod = kQMC;
            sSamplingOption="sobol";
        }
        if (svar == "qmc=halton") {
            nSamplerMethod = kQMC;
            sSamplingOption="halton";
        }
    }
    if ( _blog )
        cout << " *** nSamplerMethod[" << nSamplerMethod << "] sSamplingOption[" << sSamplingOption << "]" << endl;

    // Take into account Lower/Upper bounds AND the \delta values
    Double_t dBound;
    Double_t dDelta;
    TStochasticAttribute *stoatt;
    TMatrixD matBounds(_nX, 2);
    matBounds.Zero();
    for (Int_t iatt = 0; iatt < _nX; iatt++) {
        stoatt = dynamic_cast<TStochasticAttribute *>(_listOfInputAttributes->At(iatt));

        if ( stoatt->isLowerBounded () ) {
            dBound = stoatt->getLowerBound();
            matBounds(iatt, 0) = dBound;
            dDelta = TMath::Abs(dBound * _dstepSize );
            stoatt->setLowerBound(dBound + dDelta);
        }
        if ( stoatt->isUpperBounded () ) {
            dBound = stoatt->getUpperBound();
            matBounds(iatt, 1) = dBound;
            dDelta = TMath::Abs(dBound * _dstepSize );
            stoatt->setUpperBound(dBound - dDelta);
        }
    }

    if (_blog) {
        cout << " *** Matrix of Bounds for Bounded distribution" << endl;
        matBounds.Print("Matrix of Bounds");
    }

    // Generate the original Design of Experiments
    TSampler *ts = nullptr;

    switch (nSamplerMethod)
    {
    case kMC:
    {
        ts = new TSampling(_tds, sSamplingOption, ns);
        if ( _nSeed > 0 ) {
            if (_blog)
                cout << " *** pass seed(" << _nSeed << ") to the TSampling object" << endl;
            dynamic_cast<TSampling *>(ts)->setSeed(_nSeed);
        }
        break;
    }
    case kQMC:
    {
        ts = new TQMC(_tds, sSamplingOption, ns);
        break;
    }
    default:
        cout << " ** Default case of TSobol::initSample" << endl;
    }

    if (_blog)
        ts->setLog();

    try{
        ts->generateSample();
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        delete ts; ts=NULL;
        throw ue;
    }
    
    Int_t npat = _tds->getNPatterns();
    if ( _blog )
        cout << " *** npat[" << npat << "] with TSampling object" << endl;


    // reset the lower/upper bounds
    for (Int_t iatt = 0; iatt < _nX; iatt++) {
        stoatt = dynamic_cast<TStochasticAttribute *>(_listOfInputAttributes->At(iatt));

        if ( stoatt->isLowerBounded () )
            stoatt->setLowerBound(matBounds(iatt, 0));
        if ( stoatt->isUpperBounded () )
            stoatt->setUpperBound(matBounds(iatt, 1));
    }

    // Clean
    delete ts; ts = NULL;

    if (_blog)
        cout << " ** End Of TDGSM::initSample ns[" << ns << "]" << endl
             << "***********************************" << endl;
}

///////////////////////////////////////////////////////////
//  Inherited mandatory functions
///////////////////////////////////////////////////////////
void TDGSM::generateSampleByOATDesign(Option_t * option)
{
    if ( _blog )
        cout << "TDGSM::generateSampleByOATDesign options[" << option <<"]" << endl;

    Int_t npat = _tds->getNPatterns();
    if ( _blog ) {
        cout << " *** npat[" << npat << "] initial" << endl;
        _listOfInputAttributes->Print("all");
    }

    if ( npat == 0 ) {
        // Number of eval by points in the DoE
        Int_t nbEval = 2*_nX+1;
        if ( _blog )
            cout << " *** nbEval[" << nbEval << "] / _nS[" << _nS << "]" << endl;

        npat = TMath::FloorNint( _nS/nbEval);
        if ( _blog )
            cout << " *** npat[" << npat << "] to generate base sample" << endl;

        initSample(npat, option);
    }

    Int_t natt = _tds->getNAttributes();
    if ( _blog ) {
        cout << " ***   natt[" << natt   << "]" << endl;
        cout << " *** _nbAux[" << _nbAux << "]" << endl;
        cout << " ***  _nbIn[" << _nbIn  << "]" << endl;
        cout << " ***  _nX[" << _nX  << "]" << endl;
        _listOfInputAttributes->Print("all");
    }

    // Call the TOATDesign class to generate the DoE
    TOATDesign *tmpOATDesign = new TOATDesign(_tds, "regular", 2);
    if ( _blog )
        tmpOATDesign->setLog();


    TString sAtt = "";
    // Set the default values in the tmp TDS for the OAT Sampling
    for(Int_t iatt=0; iatt  < _nX; iatt++) {
        sAtt = _listOfInputAttributes->At(iatt)->GetName();
        if ( _blog )
            cout << " *** iatt(" << iatt << "/" << _nX << ") att[" << sAtt << "]";

        // Set the default value to the attribute
        tmpOATDesign->setRange(sAtt, _dstepSize, kTRUE);
        if ( _blog )
            cout << " _dstepSize[" << _dstepSize << "]" <<endl;
    }

    try {
        tmpOATDesign->generateSample();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if ( _blog ) {
        cout << endl << " After TOATDesign::generateSample method" << endl;
        _tds->getTuple()->SetScanField(-1);
        _tds->scan("*");
    }

    // Clean
    delete tmpOATDesign; tmpOATDesign = NULL;
}

void TDGSM::evaluateIndexes(Option_t * option)
{
    // Shaping the sensitivity matrix
    if (_blog)
    {
        cout << "TDGSM::evaluateIndexes options [" << option <<"]\n *** _iy["<<_iy<<"] _iely["<<_iely << "]" << endl;
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
        cout << " *** _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " *** _sSelectedOutput[" << _sSelectedOutput << "] _iOut[" << _iOut << "]" << endl;
    }

    // Filling the sensitivity matrix
    TObjArray *inputList=NULL;

    inputList = _sSelectedInput.Tokenize(":");

    if ( _blog ) {
        _tds->getTuple()->SetScanField(-1);
        _tds->scan("*");
    }

    Int_t iIn = 0;             // index of the sensitive inputs
    TString selection = "";    // the current selection
    for (Int_t iInAux = 0; iInAux < _nbIn; iInAux++) {
        Double_t *tmpIndex;
        Int_t nCrt=0;
        Int_t nCrt1=0, nCrt2=0;
        Double_t deltaX, deltaY;

        TString inputName = _listOfInputAttributes->At(iInAux)->GetName();
        if ( _blog )
            cout << " *** inputName[" << inputName << "]" << endl;

        // The "__nominal_set__" number change for each orignal patterns.
        // The "__modified_att__" is the index of the input attributes.
        selection = Form("( %s == %d )", _sIteratorName.Data(), iInAux + 1);
        if ( _blog )
            cout << " *** selection[" << selection << "]" << endl;

        Int_t nbValues = (Int_t) _tds->getTuple()->Draw(_tds->getIteratorName(), selection.Data(), "goff");

        if ( _blog )
            cout << " *** Input attribute " << iInAux << ": " << inputName
                 << "; Selection: " << selection << " gives nbValues "<< nbValues << endl;

        // Check if the attribute is a sensitive one
        if (nbValues > 1)
        {
            tmpIndex = _tds->getTuple()->GetV1();

            Int_t *indexOrdered = new Int_t[nbValues];
            TMath::Sort(nbValues, tmpIndex, indexOrdered, kFALSE);

            TVectorD vecIndex(nbValues);

            for(Int_t iCrt=0; iCrt<nbValues; iCrt++) {
                if ( _blog )
                    cout << "  **** tmpVal(" << iCrt << "/" << nbValues << ")[" << tmpIndex[indexOrdered[iCrt]] << "]" << endl;
                vecIndex[iCrt] = 1.0 * indexOrdered[iCrt];
            }
            if ( _blog )
                vecIndex.Print("vec index ordered");

            delete[] indexOrdered;

            // init the cell
            _sMatrix(_iOut, iIn) = 0.0;

            // current output
            TString outputName = _sOutputAttribute;
            if ( _blog ) {
                cout << " *** selection[" << selection << "] for input[" << inputName << "]" << endl;
                _tds->getTuple()->Scan(Form("%s:%s:%s",_tds->getIteratorName(), inputName.Data(), outputName.Data()), selection.Data());
                cout << " *** outputName[" << outputName << "]_iOut[" << _iOut << "]" << endl;
            }

            nbValues = (Int_t)_tds->getTuple()->Draw(inputName+":"+outputName, selection.Data(), "goff");
            Double_t * tmpValX = _tds->getTuple()->GetV1();
            Double_t * tmpValY = _tds->getTuple()->GetV2();
            if ( _blog ) {
                for(Int_t iCrt=0; iCrt<nbValues; iCrt++) {
                    nCrt = (Int_t)(vecIndex[iCrt]);
                    cout << "  **** tmpVal(" << iCrt << "/" << nbValues << ") - nCrt[" <<  nCrt << "] ValX[" << tmpValX[nCrt] << "] ValY[" << tmpValY[nCrt] << "]" << endl;
                }
            }

            for(Int_t ival=0; ival<nbValues; ival+=2) {
                nCrt1 = (Int_t)(vecIndex[ival]);
                nCrt2 = (Int_t)(vecIndex[ival+1]);

                deltaX = tmpValX[nCrt1] - tmpValX[nCrt2];
                deltaY = tmpValY[nCrt1] - tmpValY[nCrt2];
                if (_blog)
                {
                    cout << " *** Output (" << _iy << ")" << outputName << ": ival(" << ival << "/" << nbValues <<")"
                         << "; nCrt1(" << nCrt1 << ") X(" << tmpValX[nCrt1] << " , " << tmpValX[nCrt2] << ")nCrt2(" << nCrt2 << ") Y(" << tmpValY[nCrt1] << " , " << tmpValY[nCrt2] << ")nCrt2(" << nCrt2 << ")  deltaX = " << deltaX
                         << "; deltaY = " << deltaY;
                }
                _sMatrix(_iOut, iIn) += deltaY*deltaY / (deltaX*deltaX);
                if (_blog)
                    cout << " mui(" << deltaY*deltaY / (deltaX*deltaX) << ") smatrix[" << _sMatrix(_iOut, iIn) << "]" << endl;
            }

            _sMatrix(_iOut, iIn) *= 2.0/(1.0*nbValues);

            fillIndex(inputName.Data(), "First", _sMatrix(_iOut, iIn),"--first--");
            fillIndex(inputName.Data(), "Total", _sMatrix(_iOut, iIn),"--total--");

            //            }
            // Increment the index of sensitive inputs
            iIn++;
        }
    }
    _iOut++;

    if ( _blog ) {
        cout << " *** the Matrix of DGSM" << endl;
        _sMatrix.Print();
    }

    inputList->Delete();
    delete inputList;
    inputList=NULL;
}


void TDGSM::parseOption(Option_t *option)
{
    try {
        TString soption(option);
        soption.ToLower();

        // Get the step value of the design
        TDataServer *tdsTmp_OAT = new TDataServer("tmp__DGSM__OAT", " a tempory TDS");
        TString sTmpAttOAT = "__x__dgsm_oat";
        tdsTmp_OAT->addAttribute ( new TUniformDistribution(sTmpAttOAT, 0.0, 1.0));
        tdsTmp_OAT->getAttribute(sTmpAttOAT)->setStepValue(0.01);
        tdsTmp_OAT->getAttribute(sTmpAttOAT)->setDefaultValue(0.5);
        tdsTmp_OAT->getAttribute(sTmpAttOAT)->delShare();
        
        TOATSampling *tmpOATDesign = new TOATSampling(tdsTmp_OAT, _sSamplingOption);

        Double_t dStepOAT = tmpOATDesign->getStepSizeParameter();
        TString sModeOAT = tmpOATDesign->getSamplingMode();
        Bool_t bPercentOAT = tmpOATDesign->isPercentMode();


        if ( !bPercentOAT )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDGSM::parseOption ERROR : Only percentage steps [%s] mandatory soption(%s)", sModeOAT.Data(), soption.Data()));

        if ( ( dStepOAT < 0.) || ( dStepOAT > 100.0)  )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDGSM::parseOption ERROR : The step value [%f] must be between [0.0, 100.0] given by option(%s)", dStepOAT, soption.Data()));

        _dstepSize = dStepOAT / 100.0;

        // Clean
        delete tmpOATDesign; tmpOATDesign = nullptr;
        delete tdsTmp_OAT; tdsTmp_OAT = nullptr;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}


TMatrixD TDGSM::getSensitivityMatrix()
{
    return _sMatrix;
}

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TDGSM::printLog()
{
    cout << "*******************************" << endl;
    cout << "** TDGSM::printLog  ******" << endl;
    TSensitivity::printLog();
    cout << "*******************************" << endl;
}
} // End Of Sensitivity namespace
} // End Of URANIE namespace

