//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDGSM.h
 \author GAUDIER Fabrice <fabrice.gaudier@cea.fr>
 \date   2019/06/14
 
 \brief Header of the class URANIE::Sensitivity::TDGSM
 */
/**

 @class URANIE::Sensitivity::TDGSM

 @brief Description of the class TDGSM

 */
#ifndef TDGSM_H
#define TDGSM_H

// Uranie
#include "TStochAttribut.h"
#include "TFiniteDifferences.h"

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

namespace URANIE
{
namespace Sensitivity
{

class TDGSM: public TFiniteDifferences
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    /*!
     * \brief Default constructor
     */
    TDGSM();

    /*!
     * \brief Constructor with the TDataServer and function
     * \param tdsNominal
     * \param sensitiveAtt
     * \param outputAtt
     * \param samplingOption
     */
    TDGSM(TDataServer *tdsNominal, void (*fcn)(Double_t*,Double_t*),
          Int_t ns, TString sensitiveAtt, TString outputAtt,
          TString samplingOption = "steps=1%");

    /*!
     * \brief Constructor with the TDataServer and function name
     * \param tdsNominal
     * \param fcn
     * \param sensitiveAtt
     * \param outputAtt
     * \param samplingOption
     */
    TDGSM(TDataServer *tdsNominal, const char* fcn,
          Int_t ns,
          TString sensitiveAtt, TString outputAtt,
          TString samplingOption );

    /*!
     * \brief Constructor with the TDataServer and a TCode
     * \param tdsNominal
     * \param fcode
     * \param sensitiveAtt
     * \param samplingOption
     */
    TDGSM(TDataServer *tdsNominal, URANIE::Launcher::TCode *fcode,
          Int_t ns,
          TString sensitiveAtt = "", TString samplingOption = "steps=1%");

    /*!
     * \brief Constructor with the TDataServer and a TRun
     * \param tdsNominal
     * \param run
     * \param sensitiveAtt
     * \param samplingOption
     */
    TDGSM(TDataServer *tdsNominal, URANIE::Relauncher::TRun *run,
          Int_t ns,
          TString sensitiveAtt = "", TString samplingOption = "steps=1%");

    /*!
     * \brief Constructor with a loaded data server
     * \param tdsNominal
     * \param inputAtt
     * \param outputAtt
     * \param sensitiveAtt
     */
    TDGSM(TDataServer *tdsNominal, TString inputAtt,
          TString outputAtt, TString sensitiveAtt = TString(""));

    /*!
     * \brief Default destructor
     */
    virtual ~TDGSM();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Get the step size parameter
    virtual Double_t getStepSizeParameter()
    {
        return _dstepSize;
    }
    /*!
     * \brief Returns the sensitivity matrix
     * \return
     */
    TMatrixD getSensitivityMatrix();

    /*!
     * \brief Returns a pointer to the internal data server
     * \return
     */
    TDataServer* getDataServer()
    {
        return _tds;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    virtual void generateSample(Option_t * option = "")
    {
        generateSampleByOATDesign(option);
    }

    void generateSampleByOATDesign(Option_t * option = "");
    /*!
     * \brief initSample
     * \param ns (Int_t) size of the sample to generate
     * \param option (Option_t) Possible options to generate the initial sample
        \li "srs": generate sample with SRS
        \li "lhs": generate sample with LHS
        \li "qmc=sobol": generate sample with QMC=SOBOL
        \li "qmc=halton": generate sample with QMC=HALTON
     */
    void initSample(Int_t ns, Option_t * option = "");
    //! Read the possible options
    virtual void parseOption(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Compute the indexes
     **/
    //@{

    /*!
     * \brief Compute the sensitivity matrix
     * \param option
     */
    void evaluateIndexes(Option_t * option = "");

    //@}
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    /*!
     * \brief printLog
     */
    virtual void printLog();
    //@}
    ClassDef(URANIE::Sensitivity::TDGSM, ID_SENSITIVITY)

    //---------------------------------------------
    /** @name Attributes
     *
     **/
    //@{
    private:
        Double_t _dstepSize;
    //@}


};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

#endif
