//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFast.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon jun 2 10:02:35 CEST 2008
 
 \brief Interface of class URANIE::Sensitivity::TFast
 
 */
/**
 
 * @class URANIE::Sensitivity::TFast
 *
 * @brief Description of the class TFast.
 *
 * To be written by the developper.
 *
 */
#ifndef TFAST_H
#define TFAST_H

////// URANIE
#include "TSensitivity.h"
#include "TSpaceFilling.h"

#include <iostream>
using std::cout;
using std::endl;

namespace URANIE
{
namespace Sensitivity
{
class TFast: public TSensitivity
{
public:
    enum EFunction
    {
        kCukier, kSaltelliA, kSaltelliB
    };

    // Attributes
public:
    Int_t * _nValFreq;        ///< The array of frequences
    Int_t _nM;                      ///< The max harmonic to consider
    bool b_plotgraph;          ///<Plot the graph
private:
    EFunction _nFunction;       ///< The function G

    // constructor helpers
    void stdInit();
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TFast();
    //! Default constructor with a function
#ifdef ROOT5
    TFast(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns = 100,
            const char *varexpinput = "", const char *varexpoutput = "");
#else
    TFast(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
            const char *varexpinput, const char *varexpoutput, Int_t ns = 100);
#endif
//! Default constructor with a function name
    TFast(URANIE::DataServer::TDataServer *tds, const char* fcn, Int_t ns = 100,
            const char *varexpinput = "", const char *varexpoutput = "");
    //! Default constructor with a tcode
    TFast(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode,
            Int_t ns);
    //! Default constructor with a trun
    TFast(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,
            Int_t ns);
    //! Default destructor
    virtual ~TFast();
    //@}

    //---------------------------------------------
    /** @name Functions to the frequencies and the transformed
     The Fast method is to choose a frequency to each uncertainty attribute and a transform function to "space filling"
     **/
    //@{
    //! Sets the frequencies array
    void setFrequencies(Int_t nfreq, Int_t *valfreq);
    //! Sets the Transform fonction
    void setTransformFunction(EFunction ntype)
    {
        _nFunction = ntype;
    }
    //! Gets the transform function
    EFunction getTransformFunction()
    {
        return _nFunction;
    }
    //! Sets the max number of harmonic
    void setM(Int_t m)
    {
        _nM = m;
    }
    //! Returns the max number of harmonic to consider
    Int_t getM()
    {
        return _nM;
    }

    //@}

    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read option specific to TSobol
    /*!
      Possible options specific to this methods are 
     \li "graph": do the graph of the used frequencies
    */
    virtual void parseOption(Option_t *option = "");
    //@}

    
    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    virtual void generateSample(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Computes the indexes
     **/
    //@{
    //! Evaluates the index from the dataserver
    virtual void evaluateIndexes(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sensitivity::TFast, ID_SENSITIVITY)
    //Classe de
};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE
#endif
