//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFiniteDifferences.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   mer sep 13 12:39:50 CEST 2006

 \brief Implementation of the class URANIE::Sensitivity::TFiniteDifferences

 */

// ROOT
#ifdef ROOT5
#include "Api.h"
#endif
#include "TROOT.h"
#include "TPRegexp.h"
#include "TObjString.h"

// Uranie
#include "TFiniteDifferences.h"
//#include "TLauncher.h"
//#include "TLauncherFunction.h"
#include "TOATSampling.h"

using std::cout;
using std::endl;

using namespace URANIE::DataServer;
using namespace URANIE::Sampler;

ClassImp (URANIE::Sensitivity::TFiniteDifferences)

namespace URANIE
{
namespace Sensitivity
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TFiniteDifferences::TFiniteDifferences() :
        TSensitivity()
{
    _sFunctionName = TString("");
    _sSamplingOption = TString("");
    _sSensitiveAtt = TString("");
    _sSelectedInput = TString("");
    _sSelectedOutput = TString("");

    _nbIn = 0;
    _nbAux = 0;
    _nbOut = 0;

    _sIteratorName = "__modified_att__";
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
#ifdef ROOT5
TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal, void *fcn,
        TString outputAtt, TString sensitiveAtt, TString samplingOption) :
    TSensitivity(tdsNominal, fcn, 1, sensitiveAtt.Data(), outputAtt.Data(),
                samplingOption)
#else
TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        void (*fcn)(Double_t*, Double_t*),
        TString sensitiveAtt, TString outputAtt, TString samplingOption) :
    TSensitivity(tdsNominal, fcn, sensitiveAtt.Data(), outputAtt.Data(), 1,
            samplingOption)
#endif
{
    _sSensitiveAtt = sensitiveAtt;
    _sSamplingOption = samplingOption;
    _sSelectedInput = TString("");
    _sSelectedOutput = outputAtt;
    _sIteratorName = "__modified_att__";

    // Computing the size of the sensitivity matrix
    _sSensitiveAtt.ReplaceAll(" ", "");
    _sSelectedOutput.ReplaceAll(" ", "");
    setMethodName("FiniteDifferences");
    if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    {
        _nbAux = 0;
        _nbIn = tdsNominal->getNAttributes();
    }
    else
    {
        _nbIn = _sSensitiveAtt.CountChar(':') + 1;
        _nbAux = tdsNominal->getNAttributes() - _nbIn;
    }
    _nbOut = _sSelectedOutput.CountChar(':') + 1;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
    }
}

TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal, const char* fcn,
                                       TString sensitiveAtt, TString outputAtt,
                                       TString samplingOption ):
    TSensitivity(tdsNominal, fcn, 1, sensitiveAtt.Data(), outputAtt.Data(), samplingOption)
{
    _sSensitiveAtt = sensitiveAtt;
    _sSamplingOption = samplingOption;
    _sSelectedInput = TString("");
    _sSelectedOutput = outputAtt;
    _sIteratorName = "__modified_att__";

    // Computing the size of the sensitivity matrix
    _sSensitiveAtt.ReplaceAll(" ", "");
    _sSelectedOutput.ReplaceAll(" ", "");
    setMethodName("FiniteDifferences");
    if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    {
        _nbAux = 0;
        _nbIn = tdsNominal->getNAttributes();
    }
    else
    {
        _nbIn = _sSensitiveAtt.CountChar(':') + 1;
        _nbAux = tdsNominal->getNAttributes() - _nbIn;
    }
    _nbOut = _sSelectedOutput.CountChar(':') + 1;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
    }
}



// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        URANIE::Launcher::TCode *fcode, TString sensitiveAtt,
        TString samplingOption) :
        TSensitivity(tdsNominal, fcode, 1)
{
    _sSensitiveAtt = sensitiveAtt;
    _sSamplingOption = samplingOption;
    _sFunctionName = TString("");
    _sSelectedInput = TString("");
    //_sSelectedOutput = TString("");
    setMethodName("FiniteDifferences");
    _sIteratorName = "__modified_att__";

    // Computing the size of the sensitivity matrix
    _sSensitiveAtt.ReplaceAll(" ", "");
    if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    {
        _nbAux = 0;
        _nbIn = tdsNominal->getNAttributes();
    }
    else
    {
        _nbIn = _sSensitiveAtt.CountChar(':') + 1;
        _nbAux = tdsNominal->getNAttributes() - _nbIn;
    }
    _nbOut = _sSelectedOutput.CountChar(':') + 1;
#ifdef HIDE
    //Create the list of output and store it in _sSelectedOutput
    Int_t noutfile = fcode->getNOutputFiles();
    _nbOut=0;
    for (Int_t ioutfile = 0; ioutfile < noutfile; ioutfile++)
      {
  int nattfile = fcode->getNOutputAttributes(ioutfile);
  for (Int_t iattfile = 0; iattfile < nattfile; iattfile++)
    {
            URANIE::DataServer::TAttribute * att = _code->getAttribute(ioutfile, iattfile);
      TString sattname = att->GetName();
      _sSelectedOutput += sattname + TString(":");
      _nbOut += _mAttributeSize[ att->GetName() ];
    }
      }
    _sSelectedOutput.Remove(_sSelectedInput.Length() - 1);
#endif

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
    }
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        URANIE::Relauncher::TRun *run, TString sensitiveAtt,
        TString samplingOption) :
    TSensitivity(tdsNominal, run, 1)
{
    _sSensitiveAtt = sensitiveAtt;
    _sSamplingOption = samplingOption;
    _sFunctionName = TString("");
    _sSelectedInput = TString("");
    //_sSelectedOutput = TString("");
    setMethodName("FiniteDifferences");
    _sIteratorName = "__modified_att__";

    // Computing the size of the sensitivity matrix
    _sSensitiveAtt.ReplaceAll(" ", "");
    if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    {
        _nbAux = 0;
        _nbIn = tdsNominal->getNAttributes();
    }
    else
    {
        _nbIn = _sSensitiveAtt.CountChar(':') + 1;
        _nbAux = tdsNominal->getNAttributes() - _nbIn;
    }
    _nbOut = _sSelectedOutput.CountChar(':') + 1;
#ifdef HIDE
    //_nbOut = 0;
    //Create the list of output and store it in _sSelectedOutput
    Int_t noutfile = fcode->getNOutputFiles();
    _nbOut=0;
    for (Int_t ioutfile = 0; ioutfile < noutfile; ioutfile++)
      {
  int nattfile = fcode->getNOutputAttributes(ioutfile);
  for (Int_t iattfile = 0; iattfile < nattfile; iattfile++)
    {
            URANIE::DataServer::TAttribute * att = _code->getAttribute(ioutfile, iattfile);
      TString sattname = att->GetName();
      _sSelectedOutput += sattname + TString(":");
      _nbOut += _mAttributeSize[ att->GetName() ];
    }
      }
    _sSelectedOutput.Remove(_sSelectedInput.Length() - 1);
#endif

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
  cout << "---- Number of output / output elements = " << _sSelectedOutput.CountChar(':') + 1 << "/" <<_nbOut << endl;
    }
}

// -----------------------------------------------------------------------
//  Constructors with only data
// -----------------------------------------------------------------------
TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        TString inputAtt, TString outputAtt, TString sensitiveAtt) :
        TSensitivity(tdsNominal, inputAtt, outputAtt)
{
    _tds = tdsNominal;
    _sSensitiveAtt = sensitiveAtt;
    _sSamplingOption = TString("");
    _sFunctionName = TString("");
    // _sSelectedInput = inputAtt;
    // _sSelectedOutput = outputAtt;
    setMethodName("FiniteDifferences");
    _sIteratorName = "__modified_att__";

    // Computing the size of the sensitivity matrix
    _sSelectedInput.ReplaceAll(" ", "");
    _sSelectedOutput.ReplaceAll(" ", "");
    _sSensitiveAtt.ReplaceAll(" ", "");
    if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    {
        _nbAux = 0;
        _nbIn = _sSelectedInput.CountChar(':') + 1;
    }
    else
    {
        _nbIn = _sSensitiveAtt.CountChar(':') + 1;
        _nbAux = _sSelectedInput.CountChar(':') + 1 - _nbIn;
    }

    _nbOut = _sSelectedOutput.CountChar(':') + 1;

    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
    }
}

TFiniteDifferences::~TFiniteDifferences()
{
}

///////////////////////////////////////////////////////////
//  Inherited mandatory functions
///////////////////////////////////////////////////////////
void TFiniteDifferences::generateSample(Option_t * option)
{
    if (_blog)
    {
        cout << "TFiniteDifferences::generateSample options: " << option
                << endl;
    }

    // Construction of the design
    TOATSampling *oatsampling = new TOATSampling(_tds, _sSamplingOption);

    if (_blog)
        oatsampling->setLog();

    oatsampling->generateSample(Form("inputs=%s", _sSensitiveAtt.Data()));

    // Fill the list of input parameters
    TList *attList = _tds->getListOfAttributes();
    for (Int_t i = 0; i < attList->GetEntries(); i++)
    {
        TAttribute *att = (TAttribute*) attList->At(i);
        _sSelectedInput += Form("%s:", att->GetName());
    }
    _sSelectedInput.ReplaceAll("__nominal_set__:", "");
    _sSelectedInput.ReplaceAll("__modified_att__:", "");
    _sSelectedInput.Remove(_sSelectedInput.Length() - 1);

    delete attList;
    delete oatsampling;
}

  //Resize the matrix result at the very beginning of the algo
  void TFiniteDifferences::preTreatment()
  {
    if( _iy==0 && _iely==0)
      {
  _sMatrix.ResizeTo(_nbOut, _nbIn);
  _sMatrix.Zero();
      }
  }

void TFiniteDifferences::evaluateIndexes(Option_t * option)
{
    if (_blog)
    {
        cout << "TFiniteDifferences::evaluateIndexes options: " << option
       <<"  _iy  "<<_iy<<"  _iely  "<<_iely
                << endl;
    }

    // Shaping the sensitivity matrix
    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
    }

    // Filling the sensitivity matrix
    TObjArray *inputList=NULL;

    // if (_sSensitiveAtt == "" || _sSensitiveAtt == "*")
    //     inputList = _sSelectedInput.Tokenize(":");
    // else
    //     inputList = _sSensitiveAtt.Tokenize(":");

    inputList = _sSelectedInput.Tokenize(":");

    TObjArray *outputList = _sSelectedOutput.Tokenize(":");

    Int_t iIn = 0; // index of the sensitive inputs
    for (Int_t iInAux = 0; iInAux < _nbIn + _nbAux; iInAux++)
    {
        Double_t *tmpVal;
        Double_t deltaX, deltaY;

        TString inputName = ((TObjString*) inputList->At(iInAux))->GetString();
        //TString inputName = _tds->getAttribute(i)->GetName();

        // The "__nominal_set__" number may change if multiple nominal
        // values are given. This is not possible in the present
        // implementation but the possibility exists, just in case...
        // This selection will extract the two lines corresponding to the
        // variation of the input attribute's values in the OAT.
        TString selection = Form(
                "__nominal_set__ == %d && __modified_att__ == %d", 1,
                iInAux + 1);

        Int_t nbValues = (Int_t) _tds->getTuple()->Draw(inputName.Data(),
                selection.Data(), "goff");

  if (_blog)
    cout << "---- Input attribute " << iInAux << ": " << inputName
         << "; Selection: " << selection << " gives nbValues "<< nbValues << endl;

        // Check if the attribute is a sensitive one
        if (nbValues > 1)
        {
            tmpVal = _tds->getTuple()->GetV1();
            deltaX = tmpVal[1] - tmpVal[0];

      TString outputName = ((TObjString*) outputList->At(_iy))->GetString();
      //TString outputName = _tds->getAttribute(_nbIn + _nbAux + 2 + j)->GetName();
      //double *tmpVal;
      _tds->getTuple()->Draw(outputName.Data(), selection.Data(), "goff");
      tmpVal = _tds->getTuple()->GetV1();
      deltaY = tmpVal[ _nElY + _iely] - tmpVal[ _iely];
      if (_blog)
        {
    cout << "---- Output attribute " << _iy << ": "
         << outputName << "; deltaX = " << deltaX
         << "; deltaY = " << deltaY << endl;
        }
      _sMatrix(_iOut, iIn) = deltaY / deltaX;
      //            }
            // Increment the index of sensitive inputs
            iIn++;
        }
    }
    _iOut++;

    outputList->Delete();
    delete outputList;
    outputList=NULL;

    inputList->Delete();
    delete inputList;
    inputList=NULL;
}

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TFiniteDifferences::printLog()
{
    cout << "*******************************" << endl;
    cout << "** TFiniteDifferences::printLog  ******" << endl;
    TSensitivity::printLog();
    cout << "*******************************" << endl;
}
} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

