//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFiniteDifferences.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   mer sep 13 12:39:50 CEST 2006

 \brief Implementation of the class URANIE::Sensitivity::TFiniteDifferences

 */

// ROOT
#include "TROOT.h"
#include "Api.h"
#include "TPRegexp.h"

// Uranie
#include "TFiniteDifferences.h"
//#include "TLauncher.h"
//#include "TLauncherFunction.h"
#include "TOATSampling2.h"

using std::cout;
using std::endl;

using namespace URANIE::DataServer;
using namespace URANIE::Sampler;

ClassImp (URANIE::Sensitivity::TFiniteDifferences)

namespace URANIE
{
namespace Sensitivity
{
// -----------------------------------------------------------------------
//  Constructors & destructors
// -----------------------------------------------------------------------
TFiniteDifferences::TFiniteDifferences() :
        TSensitivity()
{
}

TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal, void *fcn,
        TString sensitiveAtt, TString ranges, TString inputAtt,
        TString outputAtt) :
        TSensitivity()
{
}

TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        URANIE::Launcher::TCode *fcode, TString sensitiveAtt, TString ranges) :
        TSensitivity(tdsNominal, fcode, 1)
{

    // Fill the list of outputs
    Int_t nbOutFiles = _code->getNOutputFiles();
    for (Int_t i = 0; i < nbOutFiles; i++)
    {
        Int_t nbAtt = _code->getNOutputAttributes(i);
        for (Int_t j = 0; j < nbAtt; j++)
        {
            TAttribute *att = _code->getAttribute(i, j);
            _lOutputs->Add(new TObjString(att->GetName()));
        }
    }
}

TFiniteDifferences::TFiniteDifferences(TDataServer *tdsNominal,
        TString sensitiveAtt, TString inputAtt, TString outputAtt) :
        TSensitivity()
{
}

TFiniteDifferences::~TFiniteDifferences()
{
    if (_tdsOAT)
        delete _tdsOAT;
    _tdsOAT = NULL;
}

///////////////////////////////////////////////////////////
//  Inherited mandatory functions 
///////////////////////////////////////////////////////////
void TFiniteDifferences::generateSample(Option_t * option)
{
    if (_blog)
    {
        cout << "TFiniteDifferences::generateSample options: " << option
                << endl;
    }

    // Construction of the design
    if (!gROOT->GetClass("URANIE::Sampler::TOATSampling2"))
        gSystem->Load("libUranieSampler");

    TOATSampling2 oatSampling(_tdsOAT, "regular", 2);
    for (Int_t i = 0; i < _lSensitiveAtt->GetEntries(); i++)
    {
        Bool_t is_percent = kFALSE;
        TString attName = ((TObjString*) _lSensitiveAtt->At(i))->GetString();
        TString attRange = ((TObjString*) _lRanges->At(i))->GetString();
        if (attRange.Contains("%"))
        {
            is_percent = kTRUE;
            attRange.ReplaceAll("%", "");
        }
        if (attRange.IsFloat())
            oatSampling.setRange(attName, attRange.Atof(), is_percent);
        else
            oatSampling.setRange(attName, attRange, is_percent);
    }

    if (_blog)
        oatSampling.setLog();

    oatSampling.generateSample();
}

void TFiniteDifferences::computeIndexes(Option_t * option)
{
    TString sOption = option;
    sOption.ToLower();

    if (_generateData)
    {
        if (_blog)
        {
            cout << "*** TFiniteDifferences::computeIndexes: generating data..."
                    << endl;
        }

        generateSample();
        try
        {
            if (_nLauncher == kCode)
            {
                // Create the Launcher object and launch the external code
                _code->initFileType();
                _codeLauncher = new URANIE::Launcher::TLauncher(_tdsOAT, _code);
                _codeLauncher->setDrawProgressBar(_bdrawProgressBar);
                if (_blog)
                {
                    cout << "  Launcher constructed" << endl;
                    _codeLauncher->setLog();
                }
                if (sOption.Contains("save"))
                {
                    _codeLauncher->setClean();
                    _codeLauncher->setSave();
                }
                _codeLauncher->run();
            }
            else if (_nLauncher == kFunction)
            {
                // Launch the function
                _functionLauncher = new URANIE::Launcher::TLauncherFunction(
                        _tds, _functioname, _sInputs, _sOutputs);
                _functionLauncher->run();
            }
            else
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        "*** ERROR in TFiniteDifferences::computeIndexes: No launcher available");
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
    }

    // Shaping the sensitivity matrix
    if (_blog)
    {
        cout << "---- Number of parameters          = " << _nbIn << endl;
        cout << "---- Number of auxiliary variables = " << _nbAux << endl;
        cout << "---- Number of responses           = " << _nbOut << endl;
    }

    _sMatrix.ResizeTo(_nbOut, _nbIn);
    _sMatrix.Zero();

    TDSNtupleD *oatDataTree = _tdsOAT->getTuple();

    // Retrieve the number of nominal sets
    TMatrixD matNominal();
    oatDataTree->extractData(&matNominal, "", "__modified_att__ == -1");

    for (Int_t nomSet = 1; nomSet <= nbNominal; nomSet++)
    {
        Double_t* nomValues = oatDataTree->copy
    }

}

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////        
void TFiniteDifferences::printLog()
{
    cout << "*******************************" << endl;
    cout << "** TFiniteDifferences::printLog  ******" << endl;
    TSensitivity::printLog();
    cout << "*******************************" << endl;
}
} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

