//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFiniteDifferences.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   wen sep 13 12:39:50 CEST 2006
 
 \brief This class provides sensitivity measures based on finite difference
 
 */
/**
 
 @class URANIE::Sensitivity::TFiniteDifferences

 @brief Description of the class TFiniteDifferences.

 The finite differences method of sensitivity analysis is among the simplest ones. 
 The resulting sensitivity index of an input variable \f$ x \in X \f$ with respect to an output 
 \f$ y = f(X) \f$ is an estimation of \f$ \frac{\partial f}{\partial x} \f$ around a nominal value x_{nom}. 

 In this implementation of the method, the estimation is obtained by applying an 
 OAT design of experiments (cf. Sampler::TOATSampling2) to the studied model. 
 For each input's nominal value \f$ x_{nom} \f$, we define a range \f$ 2 * \pm\delta_x \f$ and the values: 
 \li \f$ x_{sup} = x_{nom} + \pm\delta_x \f$ 
 \li \f$ x_{inf} = x_{nom} - \pm\delta_x \f$

 Then we can compute:

 \li \f$ d_1 = \frac{f(x_{norm}) - f(x_{inf})}{\pm\delta_x} \f$
 \li \f$ d_2 = \frac{f(x_{sup}) - f(x_{norm})}{\pm\delta_x} \f$

 We can then give an estimate of the partial derivative around \f$ x_{nom} \f$ as:

 \li \f$ \frac{\partial f}{\partial x_{|x = x_{nom}}} \approx d_x = \frac{d1+d2}{2} = \frac{f(x_{sup}) - f(x_{inf})}{2 * \pm\delta_x} \f$

 An indicator of the "linearity" of the function \f$ f \f$ around \f$ x_{nom} \f$ is given by:

 \li \f$ e_x = |d1-d2| = \frac{|f(x_{sup}) + f(x_{inf}) - 2 * f(x_{nom})|}{\pm\delta_x} \f$

 If \f$ y \f$ varies linearly around \f$ x_{nom} \f$, it is expected that \f$ e_x \f$ will be small 
 compared to \f$ d_x \f$. However, this indicator only has a meaning if the function f is "smooth" 
 on the interval \f$ [x_{inf},x_{sup}] \f$, and if \f$ d_x  \neq 0 \f$.

 To use the class TFiniteDifferences, one needs to give it a TDataServer 
 containing a list nominal values, and a model to compute the values of the outputs. The user 
 will then call the \b computeIndexes() \b method which will:

 \li generate the OAT sampling,
 \li call the model on it, 
 \li compute the sensitivity indexes, 
 \li store them in a sensitivity matrix. 

 This matrix can be retrieved using the method \b getSensitivityMatrix() \b .

 (Nicolas Gilardi 2012.10.31)
 */
#ifndef TFiniteDifferences_H
#define TFiniteDifferences_H

// Uranie
#include "TStochAttribut.h"
#include "TSensitivity.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

namespace URANIE
{
namespace Sensitivity
{

//void SensitivityInteractiveMethod(Double_t *param, Double_t *out);      

class TFiniteDifferences: public TSensitivity
{

private:
    URANIE::DataServer::TDataServer* _tdsOAT; ///< a pointer to the data server containing the OAT sampling
    TMatrixD _sMatrix;                               ///< The sensitivity matrix
    TObjArray * _lRanges;   ///< the list of ranges of the sensitives attributes
    TObjArray * _lSensitiveAtt;           ///< The list of sensitive attributes 
    TObjArray * _lInputs;                 ///< the full list of input variables 
    TObjArray * _lOutputs;                         ///< the full list of outputs
    // TString      _sSensitiveAtt;                             ///< The list of sensitive attributes
    // TString      _sRanges;                                   ///< the ranges of the sensitives attributes
    // TString      _sInputs;                                   ///< the full list of input variables
    // TString      _sOutputs;                                  ///< the full list of outputs
    Bool_t _generateData;        ///< if TRUE, one need to generate the sampling
    URANIE::Launcher::TLauncher *_codeLauncher; ///< a pointer to a code launcher
    URANIE::Launcher::TLauncherFunction *_functionLauncher; ///< a pointer to a function launcher
    Int_t _nbIn;                             ///< the number of sensitive inputs
    Int_t _nbAux;                        ///< the number of non sensitive inputs
    Int_t _nbOut;                                     ///< the number of outputs
protected:

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TFiniteDifferences();

    //! Constructor with the TDataServer and function
    /*!      
     This constructor is used when the model is a function.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *): the dataserver containing the nominal values of the input variables.
     \param fcn (void *): the function to analyse.
     \param sensitiveAtt (TString): a list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account.
     \param ranges (TString): a list of range values or range attribute names. It must have either one or as many element as sensitiveAtt. 
     Each numerical value of the list is interpreted as the range of the OAT sampling. Each name is interpreted 
     as the one of the attribute refering to this range in the data server. If a "%" character follows the value 
     or name, the range will be a percentage of the nominal value. 
     If only one value (or attribute name) is given, the corresponding range will be applied to all sensitive 
     attributes.
     Example: sensitiveAtt = "x1:x2:x3:x4", ranges = "0.33:5%:rx3:rx4%": 
     - range for variable x1 is 0.33,
     - range for variable x2 is 5% of the nominal value of x2, 
     - range for variable x3 is the nominal value of attribute rx3, 
     - range for variable x4 is the nominal value of attribute rx4 
     interpeted as a percentage of the nominal value of x4, 
     i.e. if rx4 = 0.1, range for x4 is 0.1% of x4.
     \param inputAtt  (TString): a list of attribute names separated by colons. They represent all the inputs of the 
     problem. They must refer to existing attributes in the "tdsNominal". If the string 
     is equal to "" or "*", all the attributes are taken into account.
     \param outputAtt (TString): a list of names separated by colons. They represent all the responses of 
     the problem. Attributes with these names will be added to the data server.
     */
    TFiniteDifferences(TDataServer *tdsNominal, void *fcn, TString sensitiveAtt,
            TString ranges, TString inputAtt, TString outputAtt);

    //! Constructor with the TDataServer and code
    /*!       
     This constructor is used when the model is an external code.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *):the dataserver containing the nominal values of the input variables.
     \param fcode (URANIE::Launcher::TCode *): the code to analyse.
     \param sensitiveAtt (TString): a list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the attributes are taken into account.
     \param ranges (TString): a list of range values or range attribute names. It must have either one or as many element as sensitiveAtt. 
     Each numerical value of the list is interpreted as the range of the OAT sampling. Each name is interpreted 
     as the one of the attribute refering to this range in the data server. If a "%" character follows the value 
     or name, the range will be a percentage of the nominal value. 
     If only one value (or attribute name) is given, the corresponding range will be applied to all sensitive 
     attributes.
     Example: sensitiveAtt = "x1:x2:x3:x4", ranges = "0.33:5%:rx3:rx4%": 
     - range for variable x1 is 0.33,
     - range for variable x2 is 5% of the nominal value of x2, 
     - range for variable x3 is the nominal value of attribute rx3, 
     - range for variable x4 is the nominal value of attribute rx4 
     interpeted as a percentage of the nominal value of x4, 
     i.e. if rx4 = 0.1, range for x4 is 0.1% of x4.
     */
    TFiniteDifferences(TDataServer *tdsNominal, URANIE::Launcher::TCode *fcode,
            TString sensitiveAtt, TString ranges);

    //! Constructor with a loaded data server
    /*!       
     This constructor is used when the user already have a dataset.
     
     \param tdsNominal (URANIE::DataServer::TDataServer *): this TDS contains the input and output values necessary for the finite 
     differences calculation. It must also contain the attributes 
     "__nominal_set__" and "__modified_att__" which are necessary to know 
     which attribute is modified.
     \param sensitiveAtt (TString): a list of attribute names separated by colons. These attributes are the one that will 
     be taken into account for the sensitivity analysis. If the string is equal to "" or "*", 
     all the input attributes are taken into account.
     \param inputAtt  (TString):  a list of attribute names separated by colons. They represent all the inputs of the 
     problem. They must refer to existing attributes in the "tdsNominal".
     \param outputAtt (TString):  a list of attribute names separated by colons. They represent all the responses of 
     the problem. They must refer to existing attributes in the "tdsNominal".
     \warning When using this constructor with CINT, do not forget to EXPLICITELY pass TString objects for 
     "inputAtt" and "outputAtt" parameters. Otherwise, you may end up calling another constructor !                                            
     */
    TFiniteDifferences(TDataServer *tdsNominal, TString sensitiveAtt,
            TString inputAtt, TString outputAtt);

    //! Default destructor
    virtual ~TFiniteDifferences();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{         
    //! Returns the sensitivity matrix
    TMatrixD getSensitivityMatrix()
    {
        return _sMatrix;
    }

    //! Returns a pointer to the internal data server
    TDataServer* getOATDataServer()
    {
        return _tdsOAT;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the OAT sampling
    /*!
     This function fills a TDS with appropriate variables, creates and uses a
     TOATSampling to generate the design of experiment.

     \param option (Option_t *) just in case we need option in the future...
     (Default = "").
     */
    virtual void generateSample(Option_t * option = "");

    //@}

    //---------------------------------------------
    /** @name Compute the indexes
     **/
    //@{
    //! Compute the sensitivity matrix
    /*! 
     This function construct the sensitivity matrix from the OAT design 
     of experiment. The resulting matrix can be retrieved using the 
     getSensitivityMatrix() function.           

     \param option (Option_t *) just in case we need option in the future...
     (Default = "").
     */
    void computeIndexes(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog();
    //@}
    ClassDef(URANIE::Sensitivity::TFiniteDifferences, ID_SENSITIVITY)
    //Definition of an variable

};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

#endif
