//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   THSIC.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2018/06/22

 \brief Interface of the URANIE::Sensitivity::THSIC class

 */
/**
 *
 * @class URANIE::Sensitivity::THSIC
 *
 * @brief Description of the class THSIC.
 *
 * \section sectionIntro Description of the class THSIC.
 *
 * \subsection subsectionIntro Introduction
 *
 * The goal of this class is to find the list of input attributes which are dependent with the output attributes
 * using the \b "HSIC" values (for "Hilbert-Schmidt Independence Criterion").
 * This approach was introduced by Gretton (2005, 2007).
 *
 * Moreover, in the same time, we compute the \f$ R^{2}_{HSIC} \f$ proposed by Da Veiga (2015) given by the formula
 * \f[
 *  R^{2}_{HSIC}  ( X, Y) \; = \; \frac{ HSIC ( X, Y) }{\sqrt{ HSIC ( X, X) \, \times \, HSIC ( Y, Y)}}
 * \f]
 *
 * \subsection subsectionDraw Draw Indexes
 *
 * We use a \b "Pareto chart" to draw the \b "p-Values" and the \b \f$ R^{2}_{HSIC} \f$.
 *
 * \subsection subsectionBiblio Bibliography
 *
 * \li Da Veiga, S. " Global Sensitivity Analysis with Dependence Measures ", https://arxiv.org/abs/1311.2483, 2013
 * \li Zhang, Q., Filippi, S., Gretton, A., Sejdinovic, D., "Large-Scale Kernel Methods for Independence Testing", 2016
 * \li Gretton, A., Fukumizu, K., Teo, C.H., Song, L., Schoelkopf, B., et Smola, A.,
 *     "A Kernel Statistical Test of Independence,", NIPS 2007
 * \li Gretton, A., Herbrich, R., Smola, A., Bousquet, O., et Schoelkopf, B.,
 *     "Kernel Methods for Measuring Independence", Journal of Machine Learning Research, 6 , pp.2075--2129, 2005
 * \li The Pareto chart, https://en.wikipedia.org/wiki/Pareto_chart
 */
#ifndef THSIC_H
#define THSIC_H

/// ROOT
#include "TVector.h"
#include "TMatrixDSymfwd.h"

/// Uranie
#include "TSensitivityTest.h"

/// Default Values
#define NDefaultBootstrap 500
#define NDefaultThresholdBootstrap 7
#define NDefaultThresholdGamma 7
#define NMinimumPatterns 6
#define dEPSILONSigma 1.e-12

namespace URANIE {
namespace Sensitivity {
class THSIC: public TSensitivityTest {
public:
    enum EEstimatedVariance
    {
        kUnknown,
        kStdEmpirical,
        kMedianeDelta
    };
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*!
     * \brief Default constructor
     */
    THSIC();

    /*!
     * \brief THSIC constructor with data contined in the TDS and the inputs and outputs attributes
     *
     * \param tds
     * \param sInputs
     * \param sOutputs
     * \param option (Option_t *) option to pass [""]
     */
    THSIC(URANIE::DataServer::TDataServer *tds, const char *sInputs, const char *sOutputs, Option_t * option = "");

    /*!
     * \brief Default destructor
     */
    virtual ~THSIC();
    //@}

    //---------------------------------------------
    /** @name Overload Methods
     **/
    //@{
    /*!
     * \brief The generate Sample method
     *
     * Nothing to do.
     * Only the case where the TDS contains the data
     *
     * \param option (Option_t *) option to pass [""]
     */
    virtual void generateSample(Option_t * option = "");

    /*!
     * \brief computeIndexes
     * \param option (Option_t *) option to pass [""]
     */
    void computeIndexes(Option_t * option = "");
    /*!
     * \brief evaluateIndexes
     * \param option (Option_t *) option to pass [""]
     */
    void evaluateIndexes(Option_t * option = "");

    /*!
     * \brief Draw the indexes computed by the method
     *
     * \param sTitle (TString) the title of the graph
     * \param select (const char*) the select argument
     * \param option (Option_t *) option to pass [""]
     */
    void drawIndexes(TString sTitle, const char *select = "", Option_t * option = "");
    //@}


    ///---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    /*!
     * \brief Which method used to estimate the variance of the gaussian kernel ( kUnknown | kStdEmpirical | kMedianeDelta )
     *
     * \li \b kStdEmpirical : the \f$ \sigma \f$ parameters is estimated by the emperical standatd deviation of the attributes
     * \li \b kMedianeDelta : use the median distance between points
     * \return the method used to estimated the variance of the gaussian kernel ( kStdEmpirical | kMedianeDelta )
     */
    EEstimatedVariance getEstimatedVariance() {
        return _nEstimatedVariance;
    }

    /*!
     * \brief Set the method used to estimate the variance of the gaussian kernel ( kUnknown | kStdEmpirical | kMedianeDelta )
     *
     * \param nMethod (EEstimatedVariance) One of the ( kStdEmpirical | kMedianeDelta ) choice
     * \exception UErrorExceptions for the case (kUnknown)
     */
    void setEstimatedVariance(EEstimatedVariance nMethod);

    /*!
     * \brief getNBoostrapSample
     * \return The number of Bootstrap sample (permutation zize) to estimate the pValues
     */
    Int_t getNBoostrapSample() {
        return _nBootstrapSample;
    }

    /*!
     * \brief Set the Boostrap sample permutation Size (default NDefaultBootstrap)
     * \param ns size of the bootstrap permutation size
     */
    void setNBoostrapSample(Int_t ns);

    /*! Set the Threshold (nS/nX) to use the bootstrap permutation test to cumpute the p-value (default NDefaultThresholdBootstrap)
      */
    void setThresholdBootstrap(Int_t n);
    /*!
     * \brief Get the Threshold (nS/nX) to use the bootstrap permutation test to cumpute the p-value (default NDefaultThresholdBootstrap)
     *
     * \return The current Threshold (nS/nX) for using the bootstrap permutation test
      */
    Int_t getThresholdBootstrap() {
        return _nThresholdBootstrap;
    }

    /*!
     * \brief Set the Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default NDefaultThresholdGamma)
     *
     * \param n (Int_t) the Threshold (nS/nX) for using the Gamma approximation
     */
    void setThresholdGamma(Int_t n);

    /*!
     * \brief Get the Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default NDefaultThresholdGamma)
     *
     * \return The current Threshold (nS/nX) for using the Gamma approximation
     */
    Int_t getThresholdGamma() {
        return _nThresholdGamma;
    }
    //@}

    //---------------------------------------------
    /** @name General methods
     *
     **/
    //@{
    /*!
     * \brief generate the \f$ H \f$ matrix
     *
     * The \f$ H \f$ matrix is defined by the formula
     * \f[ H (i, j) \, = \,  1\!\!I_{nS, nS} \: - \: \frac{1}{nS} \: \delta_{i=j} \f]
     * \param option (Option_t *) option to pass [""]
     */
    void generateHMatrix(Option_t *option = "");

    /*!
     * \brief compute the trace of the product \f$ \frac{1}{nS^2} \, K_{X} \times H \, K_{Y} \, H \f$ matrix
     *
     * \param mat (TMatrixD) the \f$ K_{X} \f$ matrix
     * \param crtMatHLH  (TMatrixD) the \f$ H \, K_{Y} \, H \f$ matrix
     * \return (Double_t) the trace od the product matrix
     */
    Double_t computeTrace(TMatrixD mat, TMatrixD crtMatHLH);

        /*!
     * \brief compute the matrix \f$  H \, K \, H \f$ matrix
     *
     * \param mat (TMatrixD) the \f$ K \f$ matrix
     * \return (TMatrixD) the \f$ H \, K \, H \f$ matrix
     */
    TMatrixDSym doubleCentering(const TMatrixDSym& mat);
    
    /*!
     * \brief getMatrixKGaussian
     * \param sAtt
     * \param dvariance
     * \return
     */
    TMatrixDSym getMatrixKGaussian(const char *sAtt, Double_t dvariance);

    /*!
     * \brief getSumOfMatrixBk
     * \param matHLkH
     * \param matHLH
     * \return
     */
    Double_t getSumOfMatrixBk(TMatrixD matHLkH, TMatrixD matHLH);

    /*!
     * \brief estimateSigma
     * \param sAtt
     * \param option (Option_t *) option to pass [""]
     * \return
     */
    Double_t estimateSigma(const char *sAtt, Option_t *option = "");

    /*!
     * \brief Use the unbiased formula to compute \f$ HSIC(X, Y) \f$
     *
     * The unbiased formula can be found in the article of Zhang (2016) (eq. 6 - page 8)
     *
     * \f[ HSIC(X, Y) \, = \, \frac{1}{nS-3} \, \left[ tr(\tilde{K_{X}} \tilde{K_{Y}})  + \frac{1^T \tilde{K_{X}} 1 \times 1^T \tilde{K_{Y}} 1 }{(nS-1)(nS-2)} - \frac{2}{nS-2} 1^T \tilde{K_{X}} \tilde{K_{Y}} 1 \right] \f]
     *
     * where \f$ \tilde{K_{X}} := K_{X} - diag(K_{X}) \f$ ( \em i.e. the kernel matrix with diagonal elements set to zero) and similarly for \f$ \tilde{K_{Y}} \f$.
     * \f$ 1 \f$  is a vector of 1s of relevant dimension.
     *
     * \param matKx
     * \param matKy
     * \return
     */
    Double_t unBiasedHSIC( TMatrixD matKx, TMatrixD matKy);

    /*!
     * \brief Create the list of permuted Bootstrap sample
     *
     * \param theMatK
     * \param tlstIndex
     * \param tlstMatrix
     */
    void permutedBootMatrix(TMatrixDSym& theMatK, TList *tlstIndex, TList *tlstMatrix);
    //@}

    //---------------------------------------------
    /** @name Parse the option
     *
     * The options treated by the THSIC object were :
     * \li "empiri" : Choose the Standard deviation of each attributes to compute the \f$ \sigma\f$ parameter
     * \li "median" : Choose the median distance between points of each attributes to compute the \f$ \sigma\f$ parameter
     *
     * \see TSensitivity::parseOption()
     **/
    //@{
    /*!
     * \brief Parse the option given by the user
     * \param option (Option_t) the string of option
     */
    virtual void parseOption(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    /*!
     * \brief Print the Log
     *
     * \param option (Option_t *) option to pass [""]
     */
    virtual void printLog(Option_t *option = "");

    /*!
     * \brief Print the summary of the HSIC method
     *
     * Print just the results of the method \b NOT \b THE \b LOG
     * Just a test to generalize
     * \param option (Option_t *) option to pass [""]
     */
    void printSummary(Option_t * option = "");
    //@}

    ClassDef(URANIE::Sensitivity::THSIC, ID_SENSITIVITY)

    //---------------------------------------------
    /** @name Private and Public attributes
     *
     **/
    //@{
public:
    TMatrixDSym _matH;                        ///< The H matrix defined by H := ( Kronecker(i,j) - 1/nS)
private:
    EEstimatedVariance _nEstimatedVariance;   //! the method to compute the variance for the gaussian kernel
    Int_t _nBootstrapSample;                  //! The number of bootstrap sample ( permutation )         (default NDefaultBootstrap)
    Int_t _nThresholdBootstrap;                   //! The Threshold (nS/nX) to use the bootstrap permutation test to cumpute the p-value (default 5)
    Int_t _nThresholdGamma;                       //! The Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default 10)
    //@}

};

} // End Of namespace Sensitivity
} // End Of namespace URANIE
#endif
