//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TJohnsonRW.h
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \date   jeu jun 30 10:35:25 CEST 2020
 
 
 \brief Interface of class URANIE::Sensitivity::TJohnsonRW
 
 */
/**
 
 * @class URANIE::Sensitivity::TJohnsonRW
 *
 * @brief Description of the class TJohnsonRW.
 *
 * To be written by the developper.
 *
 */
#ifndef TJOHNSONRW_H
#define TJOHNSONRW_H
#include "TSensitivity.h"

namespace URANIE
{
namespace Sensitivity
{
class TJohnsonRW: public TSensitivity
{


private:

    Double_t _dr2;                 ///< The R2 define by the formula
    Double_t _dr2A;                ///< The R2A define by the formula
    bool _busingCorr;              ///< compute the weight from the correlation matrix.
    bool _bmatCorrSet;             ///< Set the correlation matrix by hand, not using empirical one;
    bool _busingSample;            ///< Using a sample provided by fileDataRead
    TString _sSampMethod;          ///< The sampling method
    
    TMatrixD _mmatCorr;            ///< Matrix of correlation to be used for coefficient calculation
    TMatrixD _mRxxInv;             ///< Inverse input correlation matrix from SVD

    TMatrixD _mLambda;             ///< Lambda matrix from Johnson method
    TMatrixD _mBetanoY;            ///< Beta matrix from Johnson method up to final output multiplication

    int _localYCounter;            ///< Local counter to know how many outputs there are

    // Operations
    
public:

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    
    //! Default constructor
    TJohnsonRW();

    //! Default constructor with the TDataServer
    /*
      \param tds (URANIE::DataServer::TDataServer *) the dataserver
      \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
      \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
      \param 
    */

    TJohnsonRW(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
               const char *varexpoutput, Option_t * option = "");


    //! Default constructor with the name of a function
    TJohnsonRW(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t ns,
            const char *varexpinput = "", const char *varexpoutput = "",
            Option_t * option = "");

     //! Default constructor with a pointer to a function
    TJohnsonRW(URANIE::DataServer::TDataServer *tds,
            void (*fcn)(Double_t*,Double_t*),
            const char *varexpinput, const char *varexpoutput,
            Int_t ns, Option_t * option = "");
  
    //! Default constructor with TCode arg
    /*!
     */
    TJohnsonRW(URANIE::DataServer::TDataServer *tds,
            URANIE::Launcher::TCode *code, Int_t ns, Option_t * option = "");
    //! Default constructor with TRun arg
    /*!
     */

    TJohnsonRW(URANIE::DataServer::TDataServer *tds,
            URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option = "");   
    //! Default destructor
    virtual ~TJohnsonRW();
    //@}


    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read option specific to TJohnsonRW
    /*!
      Possible options specific to this methods are
     \li "regressors": the computation is done on regressors instead of covariance matrix
    */
    void parseOption(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
        In this case, there is noting to do
    **/
    //@{
    void generateSample(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Computes the indexes
     **/
    //@{
    void evaluateIndexes(Option_t * option = "");
    //!PreTreatment for every output
    /*!
      Doing a pre-treatment when running over outputs.
    */
    void preTreatment();
    //@}

    //---------------------------------------------
    /** @name Getters/Setters
     **/
    //@{
    //! get the coefficient of determination R2
    Double_t getR2(){
        cout << endl <<  " TJohnsonRW::getR2() :: this method returns the latest R2 computed, disregarding the algorithm used and if a vector has been analysed." << endl;
        cout << "   ==>  The proper results are now stored in the result tuple with flag __R2__ and __R2A__ " << endl << endl;
        return _dr2;
    }
    //! get the adjuested coefficient of determination R2A
    Double_t getR2A(){
        cout << endl <<  " TJohnsonRW::getR2A() :: this method returns the latest R2A computed, disregarding the algorithm used and if a vector has been analysed." << endl;
        cout << "   ==>  The proper results are now stored in the result tuple with flag __R2__ and __R2A__ " << endl << endl;
        return _dr2A;
    }
    //!Set the correlation matrix by hand
    void setCorrelationMatrix(TMatrixD corrMat);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sensitivity::TJohnsonRW, ID_SENSITIVITY)
    //Classe de
};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE
#endif
