//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TRBD.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer jun 9 15:03:08 CEST 2010
 
 \brief Implementation of class URANIE::Sensitivity::TRBD
 
 */
//using namespace std;

// ROOT
#include "TMath.h"
#include "TFFTRealComplex.h"
#include "TLeaf.h"
#include "TVectorD.h"
#include "TPad.h"

// Uranie
#include "TRBD.h"
#include "TDSNtupleD.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"

ClassImp (URANIE::Sensitivity::TRBD)

namespace URANIE
{
namespace Sensitivity
{
// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
#ifdef ROOT5
TRBD::TRBD(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns,
        const char *varexpinput, const char *varexpoutput) :
    TFast(tds, fcn, ns, varexpinput, varexpoutput)
#else
TRBD::TRBD(URANIE::DataServer::TDataServer *tds, 
        void (*fcn)(Double_t*,Double_t*),
        const char *varexpinput, const char *varexpoutput, Int_t ns) :
    TFast(tds, fcn, varexpinput, varexpoutput, ns)
#endif
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a function name
// -----------------------------------------------------------------------
TRBD::TRBD(URANIE::DataServer::TDataServer *tds, const char* fcn, Int_t ns,
        const char *varexpinput, const char *varexpoutput) :
        TFast(tds, fcn, ns, varexpinput, varexpoutput)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TRBD::TRBD(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *code,
        Int_t ns) :
    TFast(tds, code, ns)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TRBD::TRBD(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,
        Int_t ns) :
    TFast(tds, run, ns)
{
    stdInit();
}

// Constructor helpers
void TRBD::stdInit()
{

    _sIteratorName = Form("rbd__n__iter__%s", _tds->GetName());
    setMethodName("RBD");

}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TRBD::~TRBD()
{
    if (_blog)
        cout << " ** destroy TRBD[" << GetName() << "]" << endl;
}

///////////////////////////////////////////////////////////
//  Frequencies 
///////////////////////////////////////////////////////////
void TRBD::setFrequencies(Int_t nfreq, Int_t valfreq)
{
    try
    {
        if (nfreq == 0)
            nfreq = _nX;
        if (nfreq != _nX)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TRBD::setFrequencies ERROR : the TDataServer \"%s\" do not have the same number of attributes %d than required %d.",
                            GetName(), _nX, nfreq));
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            exit(-1);
    }

    if (_nValFreq == NULL)
        _nValFreq = new Int_t[_nX];

    for (Int_t i = 0; i < _nX; i++)
    {
        _nValFreq[i] = valfreq;
    }
}

///////////////////////////////////////////////////////////
//  generateSample 
///////////////////////////////////////////////////////////
void TRBD::generateSample(Option_t * option)
{

    if (_blog)
    {
        cout << " ***********************************" << endl;
        cout << " ** TRBD::generateSample[" << option << "] Begin _nS[" << _nS
                << "] _nSeed[" << _nSeed << "]" << endl;
    }

    if (_nValFreq == NULL)
        setFrequencies(_nX, 1);

    TFast::generateSample(option);

    // Get the original dataset
    _tds->exportData("_rbd_original_sample.dat");
    URANIE::DataServer::TDSNtupleD *originalNTuple =
            new URANIE::DataServer::TDSNtupleD(_tds->getTuple());

    // Reset the TTree of the TDS _tds to fill premute data
    //    _tds->setTuple((URANIE::DataServer::TDSNtupleD *) NULL);
    _tds->deleteTuple();

    // Reset the tuple of the TDS to fill with the permutation
    _tds->createTuple();

    //////////////////////
    // Permute the data
    //////////////////////
    // Init the matrix
    TMatrixD matPermutation(_nS, _nX);
    for (Int_t is = 0; is < _nS; is++)
        for (Int_t ix = 0; ix < _nX; ix++)
            matPermutation(is, ix) = is;
    if (_blog)
    {
        cout << " ** Initial permutation matrix" << endl;
        matPermutation.Print();
    }

    Int_t indtmp, valtmp;
    gRandom->SetSeed(_nSeed);

    for (Int_t ix = 0; ix < _nX; ix++)
    {
        // Create a permutation with the Devroye Algorithm
        for (Int_t is = _nS - 1; is > 0; is--)
        {
            indtmp = (Int_t)(gRandom->Integer(is));
            valtmp = (Int_t)(matPermutation(indtmp, ix));
            if (_blog)
                cout << "** is[" << is << "] val ["
                        << matPermutation(indtmp, ix) << "] indtmp[" << indtmp
                        << "] val[" << valtmp << "]";
            matPermutation(indtmp, ix) = matPermutation(is, ix);
            matPermutation(is, ix) = valtmp;
            if (_blog)
                cout << "************* is[" << is << "] val ["
                        << matPermutation(is, ix) << "] indtmp[" << indtmp
                        << "] val[" << matPermutation(indtmp, ix) << "]"
                        << endl;
            ;
        }
    }

    if (_blog)
    {
        cout << " ** Final permutation matrix" << endl;
        matPermutation.Print();
    }

    // _nX+2:  first: iterator - last: internal attribute
    Double_t *valpattern = new Double_t[_nX + 2];
    TLeaf *tlinternal = originalNTuple->GetLeaf(getIteratorName());

    for (Int_t is = 0; is < _nS; is++)
    {
        valpattern[0] = (Double_t)(is + 1);
        for (Int_t ix = 0; ix < _nX; ix++)
        {
            originalNTuple->GetEntry((Int_t)(matPermutation(is, ix)));
            //        originalNTuple->Show();
            Double_t* val = originalNTuple->GetArgs();
            valpattern[ix + 1] = val[ix + 1];
            if (_blog)
                cout << " ********* ix[" << ix << "] val[" << val[ix + 1]
                        << "] valpattern[" << valpattern[ix + 1] << "]" << endl;
        }
        // Iternal attribute as rbd__n__iter__
        valpattern[_nX + 1] = tlinternal->GetValue();
        _tds->getTuple()->Fill(valpattern);
    }
    _tds->exportData("_rbd_final_sample.dat");

    // clean
    delete[] valpattern;
    valpattern = NULL;
    delete originalNTuple;
    originalNTuple = NULL;
}

///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TRBD::evaluateIndexes(Option_t * option)
{

  if (_blog)
    {
      cout << "***********************************" << endl;
      cout << " ** TRBD::evaluateIndexes[" << option << "]" << endl;
    }

  sprintf(_sMethod, "%s", _sMethodName.Data());


  TFFTRealComplex *Uranie__RBD__FFTW = new TFFTRealComplex(_nS, kFALSE);
  URANIE::DataServer::TAttribute *atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(_iy));
  TString soutputy = atty->GetName();
  TGraph *gr=NULL;
  
  Double_t *valY = new Double_t[_nS];
  Double_t *valYOrdered = new Double_t[_nS];

  if (_blog)
    cout << " ** Case of Output atty [" << soutputy << "]" << endl;

  // Loop on the TDS to store all the Y value
  for (Int_t ipat = 0; ipat < _nS; ipat++)
    {
      // get the current pattern
      valY[ipat] = _tds->getValue(soutputy,ipat,_iely);
    }

  TVectorD dS1FFTW(_nX);
  dS1FFTW.Zero();
  // Store the Y values in order to the X values
  // Loop on the input attributes
  for (Int_t ix = 0; ix < _nX; ix++)
    {
      URANIE::DataServer::TAttribute *attx =
	(URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(
								       ix));
      if (_blog)
	cout << " ** Input i[" << ix << "/"
	     << _listOfInputAttributes->GetEntries() << "] Name["
	     << attx->GetName() << "]" << endl;

      TLeaf *leafxrk = _tds->getTuple()->GetLeaf(Form("Rk_%s", attx->GetName()));
      if( leafxrk == NULL )
	{
	  _tds->computeRank(attx->GetName());
	  leafxrk = _tds->getTuple()->GetLeaf(Form("Rk_%s", attx->GetName()));
	}

      for (Int_t ipat = 0; ipat < _nS; ipat++)
	{
	  // get the current pattern
	  _tds->getTuple()->GetEntry(ipat);
	  Int_t nind = (Int_t)(leafxrk->GetValue(0)) - 1;
	  valYOrdered[nind] = valY[ipat];
	}
      Int_t nKind = 2;
      Uranie__RBD__FFTW->Init("ES", 1, &nKind);

      Uranie__RBD__FFTW->SetPoints(valYOrdered);
      Uranie__RBD__FFTW->Transform();

      // Compute the total Energy on all the spectrum
      // and Draw the graph of spectrum
      Int_t nhalfn = (Int_t)(_nS / 2);
      gr = new TGraph(nhalfn);

      Double_t dAllEnergy = 0.0;
      Double_t dre, dim;
      for (Int_t i = 1; i <= nhalfn; i++)
	{
	  Uranie__RBD__FFTW->GetPointComplex(i, dre, dim);
	  dAllEnergy += 2. * (dre * dre + dim * dim)
	    / (Double_t)(_nS * _nS);
	  if (_blog)
	    cout << " ** FFTW  vary  i[" << i << "/" << nhalfn
		 << "] dre[" << dre << "] dim[" << dim
		 << "] dAllEnergy[" << dAllEnergy << "]" << endl;
	  gr->SetPoint(i - 1, 1.0 * i,
		       (dre * dre + dim * dim) / (Double_t)(_nS * _nS));
	}

      if (_blog)
	cout << " ** dAllEnergy[" << dAllEnergy << "]" << endl;

      if(b_plotgraph)
      {
	  checkCanvasCreation(true);
	  gr->Draw("alp");
      }

      // Compute the energy along the frequencies of Xi
      dS1FFTW(ix) = 0.0;
      for (Int_t im = 0; im < _nM; im++)
	{
	  Int_t nfre = (im + 1) * _nValFreq[ix];
	  Uranie__RBD__FFTW->GetPointComplex(nfre, dre, dim);
	  dS1FFTW(ix) += 2. * (dre * dre + dim * dim)
	    / (Double_t)(_nS * _nS);
	  if (_blog)
	    cout << " ** FFTW  im[" << im << "] freq[" << nfre
		 << "] dre[" << dre << "] dim[" << dim
		 << "] dS1FFTW[" << dS1FFTW(ix) << "]" << endl;
	}
      dS1FFTW(ix) /= dAllEnergy;

      // Fill the index values
      fillIndex(attx->GetName(), "First", dS1FFTW(ix), "--first--");
      fillIndex(attx->GetName(), "Total", dS1FFTW(ix), "--total--");
      if(!b_plotgraph)
	delete gr;
      else
	  _drawingGarbageCollector->Add(gr);
    }

  if (_blog)
    {
      _ntresult->Scan("*");
      cout << " ** First Order Sensitivity Index by RBD" << endl;
      dS1FFTW.Print();
    }


  // Clean
  delete[] valY;
  valY = NULL;
  delete[] valYOrdered;
  valYOrdered = NULL;
  delete Uranie__RBD__FFTW;
  Uranie__RBD__FFTW = NULL;

  //  drawIndexes("RBD Indice", "", option);

  if (_blog)
    {
      cout << " ** End of TRBD::evaluateIndexes" << endl;
      cout << "***********************************" << endl;
    }
}

///////////////////////////////////////////////////////////
//  Log 
///////////////////////////////////////////////////////////
void TRBD::printLog(Option_t *option)
{
    cout << "**********" << endl << "TRBD::printLog[" << option << "]" << endl;
    TFast::printLog();
    cout << "*** End Of TRBD::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

