//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TRegression.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu jun 5 10:35:25 CEST 2008
 
 \brief Implementation of the class URANIE::Sensitivity::TRegression
 
 */

//// ROOT
#include "TFile.h"
#include "TLeaf.h"
#include "TPad.h"
#include "TCanvas.h"
#include "TDecompSVD.h"

//// URANIE
//#include "TTemporalTree.h"
#include "TRegression.h"
#include "TDSNtupleD.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TRegression)

namespace URANIE
{
namespace Sensitivity
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TRegression::TRegression()
{
    cout << "construit TRegression" << endl;
    _fLogger = new URANIE::DataServer::UMessageLogger("TRegression");
}

TRegression::TRegression(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
    TSensitivity(tds, varexpinput, varexpoutput, option), _nMethod(0), 
    _dr2(0.0), _dr2A(0.0), _dSumOfSquareIndexes(0.0), _sCompleteMethod(""), _fake(NULL)
{

  _fLogger = new URANIE::DataServer::UMessageLogger("TRegression");
    if (_blog)
    {
        cout << "construit TRegression tds[" << _tds->GetName() << "] Option ["
                << option << "]" << endl;
        cout << "varexpinput[" << varexpinput << "]" << endl;
        cout << "varexpoutput[" << varexpoutput << "]" << endl;
    }

    _sIteratorName = Form("regression__n__iter__%s", _tds->GetName());
    setMethodName("Regression");

    //Check all pssible
    TString opt = option;
    opt.ToLower();
    if (opt.Contains("src"))
    {
        if(_fake==NULL)
        {
            _fake = new URANIE::DataServer::TDataServer("__faketds__","pouet");
            _fake->addAttribute("Rn2");
        }
	setMethod(kSRC);
        setMethodName("SRC^2");
	_sCompleteMethod.Append("SRC^2");
    }
    if (opt.Contains("srrc"))
    {
        if(_fake==NULL)
        {
            _fake = new URANIE::DataServer::TDataServer("__faketds__","pouet");
            _fake->addAttribute("Rn2");
        }
        setMethod(kSRRC);
        setMethodName("SRRC^2");
	_sCompleteMethod.Append("SRRC^2");
    }
    if (opt.Contains("pcc"))
    {
        setMethod(kPCC);
        setMethodName("PCC^2");
	_sCompleteMethod.Append("PCC^2");
    }
    if (opt.Contains("prcc"))
    {
        setMethod(kPRCC);
        setMethodName("PRCC^2");
	_sCompleteMethod.Append("PRCC^2");
    }

    try {
        if ( !(_nMethod & 0xF) ) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TRegression::Constructor Error\n method name passed in options [%s] is unknown !\n Possibilities are [src|srrc|pcc|prcc]", opt.Data()));

        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
    {
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TRegression::~TRegression()
{
//    cout << "detruit TRegression" << endl;
    delete _fLogger;
    _fLogger=NULL;

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	delete _fake;
	_fake=NULL;
    }
}

void TRegression::preTreatment()
{
    //Implement new matrix method from Fabrice
    // Initialise the correlation matrix with the correct size
    if(_blog)
    {
        cout << endl << "*********************************  TRegression::preTreatment" << endl;
        cout<< "*** matCorr resize to "<<_nX+_nElY<<" x "<< _nX+_nElY<<endl;
    }

    _mmatCorr.ResizeTo(_nX+_nElY, _nX+_nElY); _mmatCorrRank.ResizeTo(_nX+_nElY, _nX+_nElY);
    if( (_nMethod & (0x1 << kSRC)) || (_nMethod & (0x1 << kPCC)) )
        _mmatCorr = _tds->computeCorrelationMatrix(Form("%s:%s", _sSelectedInput.Data(), _vOutputNames.at(_iy).c_str()));
    if( (_nMethod & (0x1 << kSRRC)) || (_nMethod & (0x1 << kPRCC)) )
        _mmatCorrRank = _tds->computeCorrelationMatrix(Form("%s:%s", _sSelectedInput.Data(), _vOutputNames.at(_iy).c_str()),"","rank");

    if(_blog)
    {
        _mmatCorr.Print();
        cout << endl << "*********************************" << endl;
    }
}
  
///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TRegression::evaluateIndexes(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
    {
        cout << endl << "*********************************" << endl;
        cout << " ** TRegression::evaluateIndexes" << endl;
        cout << " **    _sSelectedInput[" << _sSelectedInput << "]" << endl;
        cout << " **   _sSelectedOutput[" << _sOutputAttribute << "]" << endl;
    }

    try
    {
        try
        {
            _tds->getTuple();
        } catch (URANIE::Exceptions::UWarningExceptions&)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TRegression::computeIndexes Error\n the sample is empty!! ");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    _nS = _tds->getNPatterns();
  

    //Create smaller correaltion matrix in case of vectors: one should always invert a matrix with inputs and the considered output
    TMatrixD RestmatCorr(_nX+1, _nX+1), RestmatCorrRank(_nX+1, _nX+1);
    _mmatCorr.GetSub(0,_nX,0,_nX, RestmatCorr);
    _mmatCorrRank.GetSub(0,_nX,0,_nX, RestmatCorrRank);
    if( _iely!=0)
    {
        for(int ival=0; ival<_nX; ival++)
	{
            if( (_nMethod & (0x1 << kSRC)) || (_nMethod & (0x1 << kPCC)) )
	    {
                RestmatCorr(ival,_nX) = _mmatCorr(ival,_nX+_iely);
                RestmatCorr(_nX,ival) = _mmatCorr(_nX+_iely,ival);
	    }
            if( (_nMethod & (0x1 << kSRRC)) || (_nMethod & (0x1 << kPRCC)) )
	    {
                RestmatCorrRank(ival,_nX) = _mmatCorrRank(ival,_nX+_iely);
                RestmatCorrRank(_nX,ival) = _mmatCorrRank(_nX+_iely,ival);
	    }
	}
    }

    //To get correlation 
    TMatrixD Correl(_nX+1, _nX+1), CorrelRank(_nX+1, _nX+1);
    if( (_nMethod & (0x1 << kSRC)) )
        Correl = RestmatCorr;
    if( (_nMethod & (0x1 << kSRRC)) )
        CorrelRank = RestmatCorrRank;
  
    //Pepite
    double pepite=1e-10;
    for(int ival=0; ival<_nX; ival++)
    {
        if( (_nMethod & (0x1 << kSRC)) || (_nMethod & (0x1 << kPCC)) )
            RestmatCorr(ival,ival) += pepite;
        if( (_nMethod & (0x1 << kSRRC)) || (_nMethod & (0x1 << kPRCC)) )
            RestmatCorrRank(ival,ival) += pepite;
    }

  
    //Invert the matrixes
    double det2=0;
    if( (_nMethod & (0x1 << kSRC)) || (_nMethod & (0x1 << kPCC)) )
        RestmatCorr.Invert(&det2);
    if( (_nMethod & (0x1 << kSRRC)) || (_nMethod & (0x1 << kPRCC)) )
        RestmatCorrRank.Invert(&det2);

    //cout<<"================================================================================="<<endl;
  
    double valSobolCrt=0, dValCrt=0, valSumCrt2[4]={0,0,0,0};
    for(Int_t i=0; i< _nX; i++)
    {
        URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));	
        TString satt = att->GetName();
        //Go throught the method      
        if( _nMethod & (0x1 << kSRC) )
        {
            dValCrt = -1.0 * RestmatCorr(_nX, i) / RestmatCorr(_nX, _nX);
            valSobolCrt = dValCrt*dValCrt;
            valSumCrt2[kSRC] += valSobolCrt;
            setMethodName("SRC^2");
            fillIndex(satt.Data(), "First", valSobolCrt,"--first--");
            fillIndex(satt.Data(), "Total", valSobolCrt,"--total--");
          
            Double_t vSeuil=0.999999, valSobolCrtICLower=0, valSobolCrtICUpper=0;
            getUncertaintyFromRhoTransformation(Correl(_nX, i), valSobolCrtICLower, valSobolCrtICUpper);
            valSobolCrt = Correl(_nX, i)*Correl(_nX, i);
            valSobolCrtICLower = TMath::Min(valSobolCrtICLower,Double_t(vSeuil));
            valSobolCrtICLower = TMath::Max(valSobolCrtICLower,Double_t(0.0));
            valSobolCrtICUpper = TMath::Min(valSobolCrtICUpper,Double_t(vSeuil));
            valSobolCrtICUpper = TMath::Max(valSobolCrtICUpper,Double_t(0.0));          
            fillIndex(satt.Data(), "First", valSobolCrt, "--rho^2--", valSobolCrtICLower, valSobolCrtICUpper);
            fillIndex(satt.Data(), "Total", valSobolCrt, "--rho^2--", valSobolCrtICLower, valSobolCrtICUpper);
        }
        if( _nMethod & (0x1 << kSRRC) )
        {
            dValCrt = -1.0 * RestmatCorrRank(_nX, i) / RestmatCorrRank(_nX, _nX);
            valSobolCrt = dValCrt*dValCrt;
            valSumCrt2[kSRRC] += valSobolCrt;
            setMethodName("SRRC^2");
            fillIndex(satt.Data(), "First", valSobolCrt,"--first--");
            fillIndex(satt.Data(), "Total", valSobolCrt,"--total--");
          
            Double_t vSeuil=0.999999, valSobolCrtICLower=0, valSobolCrtICUpper=0;
            getUncertaintyFromRhoTransformation(CorrelRank(_nX, i), valSobolCrtICLower, valSobolCrtICUpper);
            valSobolCrt = CorrelRank(_nX, i)*CorrelRank(_nX, i);                
            valSobolCrtICLower = TMath::Min(valSobolCrtICLower,Double_t(vSeuil));
            valSobolCrtICLower = TMath::Max(valSobolCrtICLower,Double_t(0.0));
            valSobolCrtICUpper = TMath::Min(valSobolCrtICUpper,Double_t(vSeuil));
            valSobolCrtICUpper = TMath::Max(valSobolCrtICUpper,Double_t(0.0));          
            fillIndex(satt.Data(), "First", valSobolCrt, "--rho^2--", valSobolCrtICLower, valSobolCrtICUpper);
            fillIndex(satt.Data(), "Total", valSobolCrt, "--rho^2--", valSobolCrtICLower, valSobolCrtICUpper);
        }
        if( _nMethod & (0x1 << kPCC) )
        {
            dValCrt = -1.0 * RestmatCorr(_nX, i) / TMath::Sqrt(RestmatCorr(i, i) * RestmatCorr(_nX, _nX));
            valSobolCrt = dValCrt*dValCrt;
            valSumCrt2[kPCC] += valSobolCrt;
            setMethodName("PCC^2");
            fillIndex(satt.Data(), "First", valSobolCrt,"--first--");
            fillIndex(satt.Data(), "Total", valSobolCrt,"--total--");
        }
        if( _nMethod & (0x1 << kPRCC) )
        {
            dValCrt = -1.0 * RestmatCorrRank(_nX, i) / TMath::Sqrt(RestmatCorrRank(i, i) * RestmatCorrRank(_nX, _nX));
            valSobolCrt = dValCrt*dValCrt;
            valSumCrt2[kPRCC] += valSobolCrt;
            setMethodName("PRCC^2");
            fillIndex(satt.Data(), "First", valSobolCrt,"--first--");
            fillIndex(satt.Data(), "Total", valSobolCrt,"--total--");
        }
    }
  
    if( _nMethod & (0x1 << kSRC) )
    {
        setMethodName("SRC^2");
        fillIndex("__sum__", "First", valSumCrt2[kSRC],"--first--");
        fillIndex("__sum__", "Total", valSumCrt2[kSRC],"--total--");
        _dSumOfSquareIndexes=valSumCrt2[kSRC];
        _dr2 = 1.0 - 1.0/ RestmatCorr(_nX, _nX);
        _dr2A = 1.0 - (1.0 - _dr2) * Double_t(_nS) / Double_t(_nS - _nX);
        fillIndex("__R2__", "First", _dr2,"--first--");
        fillIndex("__R2__", "Total", _dr2,"--total--");
        fillIndex("__R2A__", "First", _dr2A,"--first--");
        fillIndex("__R2A__", "Total", _dr2A,"--total--");
    }
    if( _nMethod & (0x1 << kSRRC) )
    {
        setMethodName("SRRC^2");
        fillIndex("__sum__", "First", valSumCrt2[kSRRC],"--first--");
        fillIndex("__sum__", "Total", valSumCrt2[kSRRC],"--total--");
        _dSumOfSquareIndexes=valSumCrt2[kSRRC];
        _dr2 = 1.0 - 1.0/ RestmatCorrRank(_nX, _nX);
        _dr2A = 1.0 - (1.0 - _dr2) * Double_t(_nS) / Double_t(_nS - _nX);
        fillIndex("__R2__", "First", _dr2,"--first--");
        fillIndex("__R2__", "Total", _dr2,"--total--");
        fillIndex("__R2A__", "First", _dr2A,"--first--");
        fillIndex("__R2A__", "Total", _dr2A,"--total--");
    }
    if( _nMethod & (0x1 << kPCC) )
    {
        setMethodName("PCC^2");
        fillIndex("__sum__", "First", valSumCrt2[kPCC],"--first--");
        fillIndex("__sum__", "Total", valSumCrt2[kPCC],"--total--");
        _dSumOfSquareIndexes=valSumCrt2[kPCC];
    }
    if( _nMethod & (0x1 << kPRCC) )
    {
        setMethodName("PRCC^2");
        fillIndex("__sum__", "First", valSumCrt2[kPRCC],"--first--");
        fillIndex("__sum__", "Total", valSumCrt2[kPRCC],"--total--");
        _dSumOfSquareIndexes=valSumCrt2[kPRCC];
    }

    setMethodName(_sCompleteMethod);
}

void TRegression::getUncertaintyFromRhoTransformation(double &rho, double &CImin, double &CImax)
{   

    unsigned int nGen=1000; // nombre de point généré aléatoirement
    double res[2];

    double Zn=0, Rn=0, std = 1/sqrt((double)(_nS - 3) ); //std de Zn
    if( _fake->hasTuple() )
        _fake->deleteTuple();
    _fake->createTuple();

    for(unsigned int it=0; it<nGen; it++)
    {
        res[0] = (double)it;
        Zn = gRandom->Gaus(0.5*TMath::Log((1+rho)/(1-rho)), std);
        Rn = TMath::TanH(Zn);
        res[1] = Rn*Rn;
        _fake->getTuple()->Fill(res);
    }

    double Proba[2]={0.025,0.975}; double Rn2Quant[2]={0,0}; 
    _fake->computeQuantile("Rn2", 2, Proba, Rn2Quant);

    CImin = Rn2Quant[0]; CImax = Rn2Quant[1];

}

void TRegression::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TRegression::printLog[" << option << "]" << endl;
    cout << " **  _nX[" << _nX << "]" << endl;
    cout << " **  _nY[" << _nY << "]" << endl;

    cout << " ** _nMethod[";
    switch (_nMethod)
    {
    case kSRC:
    {
        cout << "SRC";
    }
        break;
    case kSRRC:
    {
        cout << "SRRC";
    }
        break;
    default:
        cout << "***** Unknown ****";
    }
    cout << "]" << endl;
    cout << " ** _dr2[" << _dr2 << "] _dr2A[" << _dr2A << "]" << endl;
    cout << " ** _dSumOfSquareIndexes[" << _dSumOfSquareIndexes << "]" << endl;

    TSensitivity::printLog();
    cout << "*******************************" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

