//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TRegression.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu jun 5 10:35:25 CEST 2008
 
 \brief Interface of class URANIE::Sensitivity::TRegression
 
 */
/**
 
 * @class URANIE::Sensitivity::TRegression
 *
 * @brief Description of the class TRegression.
 *
 * To be written by the developper.
 *
 */
#ifndef TREGRESSION_H
#define TREGRESSION_H
#include "TSensitivity.h"

namespace URANIE
{
namespace Sensitivity
{
class TRegression: public TSensitivity
{

    // Attributes
public:
    enum EMethod
    {
      kSRC, kSRRC, kPCC, kPRCC, kUnknown
    };
private:
    int _nMethod;               ///< The type of method. Changed to int for convenience (see setMethod)

    Double_t _dr2;                  ///< The R2 define by the formula
    Double_t _dr2A;                 ///< The R2A define by the formula
    Double_t _dSumOfSquareIndexes; ///< The sum of square indexes \f$ \sum_{i} S_i^2 \f$

    TMatrixD _mmatCorr;            ///< Matrix of correlation to be used for coefficient calculation
    TMatrixD _mmatCorrRank;        ///< Matrix of rank correlation to be used for coefficient calculation
    URANIE::DataServer::UMessageLogger *_fLogger;   ///<  Message logger
    TString _sCompleteMethod;
    URANIE::DataServer::TDataServer *_fake;
    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TRegression();
    //! Default constructor with the TDataServer
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
     \param 
     */
    TRegression(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "src");
    //! Default destructor
    virtual ~TRegression();
    //@}

    //---------------------------------------------
    /** @name Setter and Getter for the method
     **/
    //@{
    //! Sets the method
    /*!
      The _nMethod is now a int so that one can ask for several calculation at the same time
      Possibilities are now: SRC, SRRC, PCC and PRCC
     */
    void setMethod(EMethod nmethod)
    {
      _nMethod |=  (0x1 << nmethod);
    }
    //! Clear the method index
    void clearMethod()
    {
      _nMethod = 0;
    }
    //! Gets the transform function
    int getMethod()
    {
        return _nMethod;
    }

    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     In this case, there is noting to do
     **/
    //@{
    void generateSample(Option_t * option = "")
    {
    TString soption = option;
    soption.ToLower();
    }
    ;
    //@}

    //---------------------------------------------
    /** @name Computes the indexes
     **/
    //@{
    void evaluateIndexes(Option_t * option = "");
    //!PreTreatment for every output
    /*!
      Doing a pre-treatment when running over outputs.
     */
    void preTreatment();
    //!Compute the CI from the rho estimation (handmade)
    /*!
      Compute the CI from the rho estimation (handmade)
    */
    void getUncertaintyFromRhoTransformation(double &rho, double &CImin, double &CImax);

    //@}

    //---------------------------------------------
    /** @name Getters/Setters
     **/
    //@{
    //! get the coefficient of determination R2
    Double_t getR2()
    {
      (*_fLogger) << Endl <<  " TRegression::getR2() :: this method returns the latest R2 computed, disregarding the algorithm used and if a vector has been analysed." << Endl;
      (*_fLogger) << "   ==>  The proper results are now stored in the result tuple with flag __R2__ and __R2A__ " << Endl << Endl;
      return _dr2;
    }
    //! get the adjuested coefficient of determination R2A
    Double_t getR2A()
    {
      (*_fLogger) << Endl <<  " TRegression::getR2A() :: this method returns the latest R2A computed, disregarding the algorithm used and if a vector has been analysed." << Endl;
      (*_fLogger) << "   ==>  The proper results are now stored in the result tuple with flag __R2__ and __R2A__ " << Endl << Endl;
        return _dr2A;
    }
    //! Get the sum of square indexes
    /*!
     the returned parameter is define by the formula
     \f[
     \sum_{i=1}^{nX}{\alpha_i^2}
     \f]
     */
    Double_t getSumOfSquareIndexes()
    {
        return _dSumOfSquareIndexes;
    }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sensitivity::TRegression, ID_SENSITIVITY)
    //Classe de
};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE
#endif
