//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSobol.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer sep 13 12:39:50 CEST 2006
 
 \brief Implementation of the class URANIE::Sensitivity::TSobol

 \par Modifications
 2010-06-25 09:34:49 - FGA
 Implement the formulas to compute first and total order described in the article of saltelli09

 \par Bibliography
 \em saltelli09 Variance Based sensitivity analysis of model output : Design and estimator for the total sensitivity index,
 \em Computer Physics Communications (2009)
 
 */

#ifdef ROOT5
#include "Api.h"
#endif

#include "TAttribute.h"

#include "mat3D_alloc.h"

#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#include "TMLPAnalyzer.h"

#include "TDirectory.h"
#include "TMath.h"
#include "TLeaf.h"
#include "TSystem.h"
#include "TROOT.h"

// Uranie
#include "TSobol.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TSobol)

namespace URANIE
{
namespace Sensitivity
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TSobol::TSobol() :
    TSensitivity(NULL, (char*) "", 0, "", 0), // ? (char*)
    _nSamplerMethod(kSRS), _nMatrixMiNi(kMiNiSaltelli),
    _nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1)
{

}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
#ifdef ROOT5
TSobol::TSobol(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns,
        const char *varexpinput, const char *varexpoutput, Option_t * option) :
    TSensitivity(tds, fcn, ns, varexpinput, varexpoutput), 
#else
TSobol::TSobol(URANIE::DataServer::TDataServer *tds, 
        void (*fcn)(Double_t*,Double_t*), const char *varexpinput, 
        const char *varexpoutput, Int_t ns, Option_t * option) :
    TSensitivity(tds, fcn, varexpinput, varexpoutput, ns), 
#endif
    _nSamplerMethod(kSRS), _nMatrixMiNi(kMiNiSaltelli), 
    _nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1), _withdatantp(NULL), _preloadedntp(NULL), _useless(NULL)
{
    initForRun(option);
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
TSobol::TSobol(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t ns, const char *varexpinput,
	       const char *varexpoutput, Option_t* option) :
    TSensitivity(tds, fcn, ns, varexpinput, varexpoutput, option), _nSamplerMethod(kSRS),
    _nMatrixMiNi(kMiNiSaltelli), _nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1), _withdatantp(NULL), _preloadedntp(NULL), _useless(NULL)
{
    initForRun(option);
}

// -----------------------------------------------------------------------
//  Constructors with a TDataServer filled
// -----------------------------------------------------------------------
TSobol::TSobol(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t* option) :
        TSensitivity(tds, varexpinput, varexpoutput, option), _nSamplerMethod(kSRS),
	_nMatrixMiNi(kMiNiSaltelli), _nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1), _withdatantp(NULL), _preloadedntp(NULL), _useless(NULL)
{
    initForRun(option);

    try{
        if( !_tds->isAttribute(_sIteratorName ) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSobol::constructor ERROR : the constructor requiring only input and output string is specific:\n\
It should be used only if the dataserver is containing data from a previous TSobol analysis.\n\
It should indeed contain the attribute [%s] which is compulsory to get the permutation idea of every configuration (see peek and freeze method).\n\
This variable is not here so this method is not properly used. Check documentation and contact support in case of problem.",
                                                            _sIteratorName.Data()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}
// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TSobol::TSobol(URANIE::DataServer::TDataServer *tds,
        URANIE::Launcher::TCode *code, Int_t ns, Option_t * option) :
        TSensitivity(tds, code, ns, option), _nSamplerMethod(kSRS), _nMatrixMiNi(kMiNiSaltelli),
	_nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1), _withdatantp(NULL), _preloadedntp(NULL), _useless(NULL)
{
    initForRun(option);
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TSobol::TSobol(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option) :
    TSensitivity(tds, run, ns, option), 
    _nSamplerMethod(kSRS), _nMatrixMiNi(kMiNiSaltelli), 
    _nSiAlgorithm(kSiOnlyNi), _nStiAlgorithm(kStiOnlyNi), _nSimPerIndex(-1), _withdatantp(NULL), _preloadedntp(NULL), _useless(NULL)
{
    initForRun(option);
}

// Constructor's helpers
void TSobol::initForRun(Option_t *option)
{
    _bSM = kFALSE;
    _bWithSample=false;
    _bWithData=false;
    
    if (_blog)
    {
        cout << " ********************************" << endl;
        cout << " ** TSobol constructor with function" << endl;
    }

    _sIteratorName = Form("sobol__n__iter__%s", TString(_tds->getIteratorName()).ReplaceAll("__n__iter__","").Data());
    setMethodName("Sobol");

    parseOption(option);

    if (_blog)
    {
        cout << " ** End of TSobol constructor" << endl;
        cout << " ********************************" << endl;
    }
}

TSobol::~TSobol()
{
    if (_blog)
        cout << " ** Destroy TSobol[" << GetName() << "]" << endl;
    if( _bWithData && _withdatantp != NULL)
    {
        delete _withdatantp;
        _withdatantp=NULL;
    }
    if( _preloadedntp != NULL )
    {
        delete _preloadedntp;
        _preloadedntp=NULL;
    }
}

///////////////////////////////////////////////////////////
//  init the Sample when TDS have no data 
///////////////////////////////////////////////////////////
void TSobol::initSample(Int_t ns, Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << endl << "***********************************" << endl
                << " ** TSobol::initSample ns[" << ns << "]" << endl;

    switch (_nSamplerMethod)
    {
    case kSRS:
    {
        URANIE::Sampler::TSampling *_fsampler = new URANIE::Sampler::TSampling(
                _tds, "srs", ns);
        _fsampler->setSeed(_nSeed);
        _fsampler->generateSample();
        delete _fsampler;
        _fsampler = NULL;
        break;
    }
    case kLHS:
    {
        URANIE::Sampler::TSampling *_fsampler = new URANIE::Sampler::TSampling(
                _tds, "lhs", ns);
        _fsampler->setSeed(_nSeed);
        _fsampler->generateSample();
        delete _fsampler;
        _fsampler = NULL;
        break;
    }
    case kqMCSobol:
    {
        URANIE::Sampler::TQMC *_fsampler = new URANIE::Sampler::TQMC(_tds,
                "sobol", ns);
        _fsampler->generateSample();
        delete _fsampler;
        _fsampler = NULL;
        break;
    }
    case kqMCHalton:
    {
        URANIE::Sampler::TQMC *_fsampler = new URANIE::Sampler::TQMC(_tds,
                "halton", ns);
        _fsampler->generateSample();
        delete _fsampler;
        _fsampler = NULL;
        break;
    }
    default:
        cout << " ** Default case of TSobol::initSample" << endl;
    }
    if (_blog)
        cout << " ** End Of TSobol::initSample ns[" << ns << "]" << endl
                << "***********************************" << endl;

}

///////////////////////////////////////////////////////////
//  parseOption 
///////////////////////////////////////////////////////////
void TSobol::parseOption(Option_t * option)
{
  
  TSensitivity::parseOption(option);
  
    if (_blog)
        cout << endl << "***********************************" << endl
                << " ** TSobol::parseOption option[" << option << "]" << endl;

    TString soption = TString(option);
    soption.ToLower();
        
    soption.ReplaceAll(",", " ");
    soption += TString(" ");
    if (_blog)
        cout << " ** soption[" << soption << "]" << endl;

    istringstream inStream((char *) soption.Data());
    string inName;
    while (inStream >> inName)
    {
        TString svar(inName.c_str());
        if (_blog)
            cout << " ** svar[" << svar << "]" << endl;

        // Case of the sampler
        if (svar == "srs")
            _nSamplerMethod = kSRS;
        if (svar == "lhs")
            _nSamplerMethod = kLHS;
        if (svar == "qmc=sobol")
            _nSamplerMethod = kqMCSobol;
        if (svar == "qmc=halton")
            _nSamplerMethod = kqMCHalton;

        // Case of the MiNi Matrix to generate
        if (svar == "sobol")
	  {
            _nMatrixMiNi = kMiNiSobol;
	    _nSiAlgorithm=kSiAll;
	    _nStiAlgorithm=kStiAll;
	  }
        if (svar == "saltelli")
	  {
            _nMatrixMiNi = kMiNiSaltelli;
	    _nSiAlgorithm=kSiOnlyNi;
	    _nStiAlgorithm=kStiOnlyNi;
	  }

        if( svar == "withdata" && _bWithData == false)
        {
            try{
                if( _tds->isAttribute( _sIteratorName ) )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TSobol::parseOption ERROR : the dataserver already contains a Sobol iterator [%s]\n\
The provided sample appears to have been produced by the TSobol class, this is not correct :\n\
Every sample for which [%s > 100] has been made up by smart permutation, so this sample and option should not be used here !!", _sIteratorName.Data(), _sIteratorName.Data()));
            
                for(unsigned int iout=0; iout<_vOutputNames.size(); iout++)
                {       
                    if( !_tds->isAttribute( _vOutputNames.at(iout).c_str() ) )
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TSobol::parseOption ERROR : the option \"withdata\" has been used but the output [%s] is missing.\n\
This is a wrong usage og this option please check documentation or call support.", _vOutputNames.at(iout).c_str()));
                    else
                        _tds->delAttribute( _vOutputNames.at(iout).c_str() );

                }
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }

            _bWithData=true;

        }
    }
    
    if (_blog)
        cout << " ** End Of TSobol::parseOption option[" << option << "]"
                << endl << "***********************************" << endl;
}

///////////////////////////////////////////////////////////
//  Generate the Sample 
///////////////////////////////////////////////////////////
void TSobol::generateSample(Option_t * option)
{
    if (_blog)
    {
        cout << endl;
        cout << "*********************************" << endl;
        cout << " ** TSobol::generateSample" << endl;
        cout << " ** _nSeed[" << _nSeed << "]" << endl;
    }

    Int_t npat;

    // Create the tuple if necessary
    try
    { 
        _tds->getTuple();
        npat = (Int_t)(_tds->getTuple()->GetEntries());
        _nSimPerIndex = TMath::Nint(TMath::Floor((Double_t)(npat) / 2.0));

        if(_blog)
            XYZ<<" Tuple found, with "<<npat<<"  which propose "<<_nSimPerIndex<<" elements per index"<<endl;
            
        if( _nSimPerIndex < 3 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSobol::generateSample ERROR : the ntuple provided does not contains at least 3 simulation per index:\n\
 (%d patterns for %d entries)",npat, _nX));

        TList *locallist = _tds->getListOfAttributes();
        TIter next(locallist);

        map<string,int> allAttributes;
        URANIE::DataServer::TAttribute *att;
        while( (att = (URANIE::DataServer::TAttribute*)next()) )
            allAttributes[string(att->GetName())]=0;
        
        delete locallist;

        for(unsigned int iout=0; iout<_vOutputNames.size(); iout++)
        {
            // Removing this attributes (if it exists) from the list
            allAttributes[_vOutputNames.at(iout)]++;
            
            if( _tds->isAttribute( _vOutputNames.at(iout).c_str() ) && !_bWithData )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TSobol::generateSample ERROR : the ntuple provided does contains output results (output [%s] has been found).\n\
If you want to provide results and use inputs and outputs as a starter for Matrices M and N (see peek and freeze explanation) you should use the \"withdata\" option.\n\
Please have a look at the documentation and or contact support.", _vOutputNames.at(iout).c_str()));
        }

        TString sinput = _sSelectedInput;
        sinput.ReplaceAll(":", " ");
        istringstream inStream((char *) sinput.Data());
        string inName;
        while (inStream >> inName)
        {
            // Removing this attributes (if it exists) from the list
            allAttributes[inName]++;
            
            if( !_tds->isAttribute( inName.c_str() ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TSobol::generateSample ERROR : the ntuple provided does not contains one needed input [%s].\n\
If you want to provide a sample, please provide all needed inputs.", inName.c_str()));
        }


        for(map<string,int>::iterator itmap=allAttributes.begin(); itmap!=allAttributes.end(); itmap++)
        {
            if( itmap->second==0)
            {
                cout<<"[34m <URANIE::INFO>  The following attribute ["
                    <<itmap->first<<"], brought through the ntuple, has been found ["
                    <<itmap->second<<"] times in input and output lists. => It will be dropped [0m"<<endl;
                _tds->delAttribute(itmap->first);
            }
        }

        _bWithSample=true;
        
        if( 2*_nSimPerIndex != npat )
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                      Form("TSobol::generateSample INFO : the ntuple provided does not contains an even number of event. %d events are dropped.",
                                                           npat-2*_nSimPerIndex));
        
    }catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
    } catch (URANIE::Exceptions::UWarningExceptions&)
    {
        createTuple(option);
        _nSimPerIndex = 0;

        if (_nMatrixMiNi == kMiNiSobol)
            _nSimPerIndex = TMath::Nint(
                    TMath::Ceil(_nS * 1.0 / (2.0 * _nX + 2.0)));
        if (_nMatrixMiNi == kMiNiSaltelli)
            _nSimPerIndex = TMath::Nint(TMath::Ceil(_nS * 1.0 / (_nX + 2.0)));
        if (_blog)
            cout << "_nSimPerIndex[" << _nSimPerIndex << "]" << endl;
        /* \todo FGA::2010-04-23 11:57:58
         ** actuellement un seule grosse matrice de taille 2nS \times nX est generee
         **  et le decoupage est effectue comme suit : M==1->nS et N==nS+1->2nS 
         ** Rendre independant la generation des deux matrices M et N
         ** et attention au cas QMC car dim de la matrice est nS \times (2 * nX)
         */
        initSample(2 * _nSimPerIndex);
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    if (_blog)
        _tds->getTuple()->Print();

    _matM.ResizeTo(_nSimPerIndex, _nX);
    _matN.ResizeTo(_nSimPerIndex, _nX);

    npat = (Int_t)(_tds->getTuple()->GetEntries());
    if (_blog)
    {
        _listOfInputAttributes->Print();
        cout << "_nS[" << _nS << "] _nX[" << _nX << "] _nSimPerIndex["
                << _nSimPerIndex << "] -- npat[" << npat << "]" << endl;
    }

    Int_t ivar;

    // Get The list of Leaf (from TTree) of the list of attributes ( _listOfInputAttributes)
    TList *listOfInputLeaves = new TList();
    TIter next(_listOfInputAttributes);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        listOfInputLeaves->Add(_tds->getTuple()->GetLeaf(att->GetName()));
    }

    URANIE::DataServer::TAttribute* indexSobol =
            new URANIE::DataServer::TAttribute(_sIteratorName);
    indexSobol->setOrigin(URANIE::DataServer::TAttribute::kAttribute); // kInternal / kAttribute
    indexSobol->delShare();
    _tds->getDataSpecification()->addAttribute(indexSobol);

    // Loop on the sampling to fill the Matrix M
    for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
    {
        // get the current pattern
        _tds->getTuple()->GetEntry(((_bWithData||_bWithSample)? 2*ipat : ipat));
        // Fill the input vector
        ivar = 0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
            _matM(ipat, ivar++) = leaf->GetValue(0);
        }
    }

    if (_blog)
    {
        cout << "First Sampling : Matrix M" << endl;
	_matM.Print();
    }

    // Loop on the sampling to fill the Matrix N
    for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
    {
        // get the current pattern
        _tds->getTuple()->GetEntry(((_bWithData||_bWithSample)? 2*ipat +1 : _nSimPerIndex + ipat));
        // Fill the input vector
        ivar = 0;
        TIter nextLeaf(listOfInputLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
        {
            _matN(ipat, ivar++) = leaf->GetValue(0);
        }
    }
    if (_blog)
    {
        cout << "Second Sampling : Matrix N" << endl;
	_matN.Print();
    }

    // Clean
    delete listOfInputLeaves;
    listOfInputLeaves = NULL;
    
    // Nom Refill all the data
    createTuple(option);

    Int_t nSobolIndex = 0;

    TMatrixD _matMi(_nSimPerIndex, _nX);
    TMatrixD _matNi(_nSimPerIndex, _nX);
    // Create the Matrix M_i ( M except the the i^th column which is N_i)
    Double_t *valInpM = new Double_t[_nX + 2];
    Double_t *valInpN = new Double_t[_nX + 2];

    valInpM[1] = -1.0;
    valInpN[1] = 0.0;

    int dataShift=0;
    if( !_bWithData )
    {
        // Case Of M and N 
        for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
        {
            valInpM[0] = 1.0 * nSobolIndex++;

            valInpN[0] = 1.0 * nSobolIndex++;
            for (Int_t ix = 0; ix < _nX; ix++)
            {
                valInpM[2 + ix] = _matM(ipat, ix);
                valInpN[2 + ix] = _matN(ipat, ix);
            }

            _tds->getTuple()->Fill(valInpM);
            _tds->getTuple()->Fill(valInpN);

        }
    }else
    {
        dataShift=2*_nSimPerIndex;
    }
        
    // Case of all the Mi and Ni    
    for (Int_t i = 0; i < _nX; i++)
    {

        valInpM[1] = (Double_t)(i + 1);
        valInpN[1] = (Double_t)(100 * _nX + i + 1);

        // Loop on the sample M
        for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
        {
            if (_nMatrixMiNi == kMiNiSobol)
                valInpM[0] = dataShift + 1.0 * nSobolIndex++;
            valInpN[0] = dataShift + 1.0 * nSobolIndex++;

            for (Int_t ivarcrt = 0; ivarcrt < _nX; ivarcrt++)
            {
	         if (ivarcrt == i)
                {
                    _matMi(ipat, ivarcrt) = _matN(ipat, ivarcrt);
                    _matNi(ipat, ivarcrt) = _matM(ipat, ivarcrt);
                }
                else
                {
                    _matMi(ipat, ivarcrt) = _matM(ipat, ivarcrt);
                    _matNi(ipat, ivarcrt) = _matN(ipat, ivarcrt);
                }
		 valInpM[2 + ivarcrt] = _matMi(ipat, ivarcrt);
		 valInpN[2 + ivarcrt] = _matNi(ipat, ivarcrt);
            }
            // Case Of Mi
            if (_nMatrixMiNi == kMiNiSobol)
                _tds->getTuple()->Fill(valInpM);

            // Case Of Ni
            _tds->getTuple()->Fill(valInpN);
        }

        if (_blog)
        {
            if (_nMatrixMiNi == kMiNiSobol)
            {
                cout << "Matrice Mi ( index = " << i << ")" << endl;
                _matMi.Print();
            }
            cout << "Matrice Ni ( index = " << i << ")" << endl;
	    _matNi.Print();
        }
    }

    // Clean
    delete[] valInpM;
    valInpM = NULL;
    delete[] valInpN;
    valInpN = NULL;

    _tds->exportData("_sobol_sampling_.dat", _sSelectedInput+":"+_sIteratorName);

    if( _bWithData && _blog)
    {
        _withdatantp->SetScanField(-1);
        _withdatantp->Scan("*");
    }

    _tds->getAttribute(_sIteratorName)->setOrigin(
            URANIE::DataServer::TAttribute::kAttribute);

}

void TSobol::loadOtherSobolFile(const char *filename)
{

    try
    {
        if( _preloadedntp!=NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSobol::loadOtherSobolFile ERROR : this function has already been called once. One can only load one file.");
            
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    URANIE::DataServer::TDataServer *pouet = new URANIE::DataServer::TDataServer("__internal__tds__lodaing__sobol__","pouet");
    TString SobIter="";

    try{
        pouet->fileDataRead(filename,false);

        TList *locallist = pouet->getListOfAttributes();
        TIter next(locallist);
        bool foundSobIter=false;
        map<string,int> allAttributes;
        while (URANIE::DataServer::TAttribute *att =
               (URANIE::DataServer::TAttribute *) next()) {

            allAttributes[string(att->GetName())]=0;
            SobIter=TString(att->GetName());
            if(SobIter.Contains("sobol__n__iter__") )
            {
                foundSobIter=true;
                allAttributes[string(att->GetName())]++;
            }
        }
        delete locallist;
        
        if(!foundSobIter)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSobol::loadOtherSobolFile ERROR : the ntuple provided does not contain a sobol iterator (should start with [sobol__n__iter__].");
         
        for(unsigned int iout=0; iout<_vOutputNames.size(); iout++)
        {       
            allAttributes[_vOutputNames.at(iout)]++;
            
            if( !pouet->isAttribute( _vOutputNames.at(iout).c_str() ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TSobol::loadOtherSobolFile ERROR : the ntuple provided does not contain output results (output [%s] has not been found)."
                                                                , _vOutputNames.at(iout).c_str()));
        }
        
        TString sinput = _sSelectedInput;
        sinput.ReplaceAll(":", " ");
        istringstream inStream((char *) sinput.Data());
        string inName;
        while (inStream >> inName)
        {
            allAttributes[inName]++;
            
            if( !pouet->isAttribute( inName.c_str() ) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TSobol::loadOtherSobolFile ERROR : the ntuple provided does not contain input results (input [%s] has not been found)."
                                                                , inName.c_str()));
        }

        for(map<string,int>::iterator itmap=allAttributes.begin(); itmap!=allAttributes.end(); itmap++)
            if( itmap->second==0)
                cout<<"[34m <URANIE::INFO>  The following attribute ["
                    <<itmap->first<<"], brought through the ntuple in TSobol::loadOtherSobolFile, has been found ["
                    <<itmap->second<<"] times in input and output lists. => It will be dropped [0m"<<endl;
        
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        delete pouet;
        throw ue;        
    }
    
    _preloadedntp = new URANIE::DataServer::TDSNtupleD( pouet->getTuple(), _sSelectedInput+":"+_sSelectedOutput);

    if(_blog)
    {
        XYZ<<"Looking at preloaded input and output values"<<endl;
        _preloadedntp->Scan("*");
        
        XYZ<<" Adding both kind of iterator, changing ["<<string(SobIter)<<"] into ["<<string(_sIteratorName)
           <<"] and ["<<pouet->getIteratorName()<<"] into ["<<_tds->getIteratorName()<<"] "<<endl;
    }
    
    _preloadedntp->Merger(pouet->getTuple(), string(_sIteratorName)+":"+string(_tds->getIteratorName()), string(SobIter)+":"+string(pouet->getIteratorName()));
    if(_blog)
    {
        XYZ<<"Final preloaded reshaped sample, to be added to our current analysis"<<endl;        
        //    _preloadedntp->SetScanField(-1);
        _preloadedntp->Scan("*");
    }    

    delete pouet;
}

  // =============================================
  // PRETREATMENT
  // =============================================
  void TSobol::preTreatment()
  {
    if (_blog)
      {
        cout << "***********************************" << endl;
        cout << " ** TSobol::preTreatment" << endl;
      }

    if( _bWithData )
    {
        if(_blog)
            XYZ<<"Putting back the data taken out before submission "<<endl;

        TList list;
        list.Add(_withdatantp);
        list.Add(_tds->getTuple());

        TTree *localmerge = TTree::MergeTrees(&list);
        list.Clear();
        _tds->getTuple()->SetName( "toberemoved" );        
        _tds->deleteTuple();                
        _tds->setTuple((URANIE::DataServer::TDSNtupleD*)localmerge);        
        delete localmerge;
        
        if(_blog)
        {
            XYZZY;
            //_tds->getTuple()->SetScanField(-1);
            _tds->getTuple()->Scan("*");
        }
    }

    if(_preloadedntp != NULL)
    {

        TObjArray *allList = _tds->getTuple()->GetListOfBranches();
        TIter next(allList);
        map<string,int> allAttributes;
        TString neededBranches="";
        bool redotuple=false;
        while (TObject *att = next())
        {
            allAttributes[string(att->GetName())]=0;
            if( _preloadedntp->GetBranch( att->GetName() ) != NULL)
            {
                allAttributes[string(att->GetName())]++;
                neededBranches+=TString(att->GetName())+":";
            }
            else
            {
                cout<<"[34m <URANIE::INFO>  The following attribute ["
                    <<att->GetName()<<"], was brought through the ntuple but not present in the file provided by TSobol::loadOtherSobolFile. => It will be dropped [0m"<<endl;
                redotuple=true;
                _tds->delAttribute(att->GetName());
            }
        }

        if(redotuple)
        {
            URANIE::DataServer::TDSNtupleD *pointer=_tds->getTuple();
            _tds->setTuple( pointer, true, neededBranches.Data());
            delete pointer;
        }

        TList list;
        list.Add(_tds->getTuple());
        list.Add(_preloadedntp);
        
        TTree *localmerge = TTree::MergeTrees(&list);
        list.Clear();
        _tds->getTuple()->SetName( "toberemoved" );        
        _tds->deleteTuple();
        _tds->setTuple((URANIE::DataServer::TDSNtupleD*)localmerge);        
        delete localmerge;

        if(_blog)
        {
            XYZZY;
            //_tds->getTuple()->SetScanField(-1);
            _tds->getTuple()->Scan("*");
        }

        delete _preloadedntp;
        _preloadedntp=NULL;
        _nSimPerIndex = _tds->getTuple()->GetEntries() / (_nX+2);
    }

    /// Look for the Attribute specific for the Sobol Method
    try
      {
        if (_tds->isAttribute(_sIteratorName) == kFALSE)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TSobol::preTreatment ERROR : the TDataServer \"%s\" do not have the iterator attribute \"%s\" for this method.",
							  GetName(), _sIteratorName.Data()));

      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	  throw ue;
      }

    if( _iely==0 && _iy==0)
      {
	////////////////////////////////////
	/// Matrix M and N of the output
	////////////////////////////////////
	/// Get the list of integer for the two matrix M and N with the _sIteratorName
	/// -1 : for the M matrix
	///  0 : for the N matrix
	if( gDirectory != gROOT ) gDirectory = gROOT;
	if( gROOT->FindObject("uranie_list_m") ) gROOT->Delete("uranie_list_m");
	if( gROOT->FindObject("uranie_list_n") ) gROOT->Delete("uranie_list_n");
	_tds->getTuple()->Draw(">>uranie_list_m", Form("%s==-1", _sIteratorName.Data()), "goff");
	_tds->getTuple()->Draw(">>uranie_list_n", Form("%s==0", _sIteratorName.Data()), "goff");

	_telm = (TEventList*) gROOT->Get("uranie_list_m");
	_teln = (TEventList*) gROOT->Get("uranie_list_n");

	Int_t nSizeM = _telm->GetN();
	Int_t nSizeN = _teln->GetN();

	if( _nSimPerIndex == -1) //Case a file is and no sample is generated
	  _nSimPerIndex=nSizeN;
	
	if (_blog)
	  {
	    cout << " ** nSizeM[" << nSizeM << "]" << endl;
	    _telm->Print("all");
	    cout << " ** nSizeN[" << nSizeN << "]" << endl;
	    _teln->Print("all");
	  }
	Int_t nsim = nSizeN;

	_valM.ResizeTo(nsim, _nbOut);
	_valN.ResizeTo(nsim, _nbOut);
	//////////////////////

	//////////////////////
	/// Matrix for the indexes
	///   valSi  for first ordFer
	///   valSti for total order
	/// 5 methods (enum ES[i|ti]Algorithm) are computed in the same time
	///   and 2 indexes by methods
	/// Then nStoredIndexes = 10
	_nMethodComputed = 5;
	_valSi.ResizeTo(_nMethodComputed, _nbOut);
	_valSi.Zero();
	_valSti.ResizeTo(_nMethodComputed, _nbOut);
	_valSti.Zero();
	/// Matrix to store the mean and Sigma for the differents algorithms
	///   5 methods (enum ES[|ti]iAlgorithm) and 2 matrix M and N nStoredIndexes = 10
	/// the 5 firsts are the M matrix
	/// the 5 lasts  are the N matrix
	/// 1. 0 - kSiSobol93          M*M
	/// 2. 1 - kSiSaltelli02       M*N
	/// 3. 2 - Jansen
	/// 4. 3 - kSiSumo10       (M+N)*(M+N)
	/// 5. 4 - kSiMartinez11       N*Ni
	Int_t nStoredIndexes = 10;
	_valMeanMN.ResizeTo(nStoredIndexes, _nbOut);
	_valMeanMN.Zero();
	_valSigmaMN.ResizeTo(nStoredIndexes, _nbOut);
	_valSigmaMN.Zero();
	//////////////////////

      }

    _valMeanYM.clear(); _valMeanYM.resize(_nElY,0);
    _valVarYM.clear(); _valVarYM.resize(_nElY,0);
    _valMeanYN.clear(); _valMeanYN.resize(_nElY,0);
    _valVarYN.clear(); _valVarYN.resize(_nElY,0);
    
    //Compute stat once and for all for M and N
    URANIE::DataServer::TAttribute *atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(_iy));
    TString sSelectComputeStat = Form("%s==-1", _sIteratorName.Data());
    if (_blog)
      cout << " ** M sSelectComputeStat [" << sSelectComputeStat << "]" << endl;
    _tds->computeStatistic(atty->GetName(), sSelectComputeStat);
    for(unsigned int iely=0; iely<_nElY; iely++)
      {
	_valMeanYM.at(iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	_valVarYM.at(iely) = _tds->getAttribute(atty->GetName())->getStd(iely) * _tds->getAttribute(atty->GetName())->getStd(iely);

	/// Case of kSiSobol93|kStiHomma96
	if ((_nSiAlgorithm & 0x1 << kSiSobol93) || (_nStiAlgorithm & 0x1 << kStiHomma96))
	  {
	    _valMeanMN(0, _iOut+iely) = _valMeanMN(5, _iOut+iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	    _valSigmaMN(0, _iOut+iely) = _valSigmaMN(5, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	  }
  
	/// Case of kSiSaltelli02|kStiSobol07
	if ((_nSiAlgorithm & 0x1 << kSiSaltelli02) || (_nStiAlgorithm & 0x1 << kStiSobol07) )
	  {
	    _valMeanMN(1, _iOut+iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	    _valSigmaMN(1, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	  }
  
	/// Case Of kSiJansen99|kStiJansen99
	if ((_nSiAlgorithm & 0x1 << kSiJansen99) || (_nStiAlgorithm & 0x1 << kStiJansen99))
	  {
	    _valSigmaMN(2, _iOut+iely) = _valSigmaMN(7, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	  }
      }

      /// Case of the N matrix
      sSelectComputeStat = Form("%s==0", _sIteratorName.Data());
      if (_blog)
	cout << " ** N sSelectComputeStat [" << sSelectComputeStat << "]" << endl;
      _tds->computeStatistic(atty->GetName(), sSelectComputeStat);
      for(unsigned int iely=0; iely<_nElY; iely++)
	{
	  _valMeanYN.at(iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	  _valVarYN.at(iely) = _tds->getAttribute(atty->GetName())->getStd(iely) * _tds->getAttribute(atty->GetName())->getStd(iely);

	  /// Case of kSiSaltelli02|kStiSobol07
	  if ((_nSiAlgorithm & 0x1 << kSiSaltelli02) || (_nStiAlgorithm & 0x1 << kStiSobol07))
	    {
	      _valMeanMN(6, _iOut+iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	      _valSigmaMN(6, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	    }

	  /// Case of kSiMartinez11|kStiMartinez11
	  if ((_nSiAlgorithm & 0x1 << kSiMartinez11)  || (_nStiAlgorithm & 0x1 << kStiMartinez11))
	    {
	      _valMeanMN(4, _iOut+iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	      _valSigmaMN(4, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	    }
	}


      /// case of the MuN matrix
      sSelectComputeStat = Form("(%s==0)||(%s==-1)", _sIteratorName.Data(), _sIteratorName.Data());
      if (_blog)
	cout << " ** [kSiSumo10|kStiSumo10] sSelectComputeStat [" << sSelectComputeStat << "]" << endl;
      _tds->computeStatistic(atty->GetName(), sSelectComputeStat);      
       for(unsigned int iely=0; iely<_nElY; iely++)
	{
	  /// Case of kSiSumo10|kStiSumo10
	  if ((_nSiAlgorithm & 0x1 << kSiSumo10) || (_nStiAlgorithm & 0x1 << kStiSumo10))
	    {
	      _valMeanMN(3, _iOut+iely) = _valMeanMN(8, _iOut+iely) = _tds->getAttribute(atty->GetName())->getMean(iely);
	      _valSigmaMN(3, _iOut+iely) = _valSigmaMN(8, _iOut+iely) = _tds->getAttribute(atty->GetName())->getStd(iely);
	    }
	}
  }

bool custom_sorter(const std::pair<double,double>& a, const std::pair<double,double>& b)
{
    return a.second < b.second;
}
void our_parallel_sort(vector<double>* values, vector<double> *ref)
{   
    std::vector<std::pair<double,double>> zipped;
    for(size_t i=0; i<values->size(); ++i)
        zipped.push_back(std::make_pair(values->at(i), ref->at(i)));
    
    // Sort the vector of pairs
    std::sort(std::begin(zipped), std::end(zipped), &custom_sorter);

    for(size_t i=0; i<values->size(); i++)
    {
        values->at(i) = zipped[i].first;
        ref->at(i) = zipped[i].second;
    }
}

///////////////////////////////////////////////////////////
///  evaluateIndexes
///////////////////////////////////////////////////////////
void TSobol::evaluateIndexes(Option_t * option)
{
  if (_blog)
    {
      cout << "***********************************" << endl;
      cout << " ** TSobol::evaluateIndexes[" << option << "]" << endl;
    }

  URANIE::DataServer::TAttribute *attx;
  URANIE::DataServer::TAttribute *atty;

  sprintf(_sMethod, "%s", _sMethodName.Data());
  TString soutputy, sSelectComputeStat;
  atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(_iy));
  soutputy = atty->GetName();

  //////////////////////
  /// FGA::remarks 2010-06-25 10:40:57
  /// Use only by the old method to compute ( --first-- and --total--)
  /// Replaced by the two matrix valMeanMN and valSigmaMN
  Double_t *valfirstOrder = new Double_t[_nX];
  Double_t *valTotalOrder = new Double_t[_nX];
  double *Values=NULL, *IterV=NULL;
  vector<vector<double> > M_N_Mi; M_N_Mi.resize(2);
  vector<double> iterVals;
  double mi, ni;
  /// End of FGA::remarks 2010-06-25 10:40:57
  //////////////////////
  stringstream sstr;   
  //Make the loop for components here to avoid recompute statistic so many times
  for( _iely=0; _iely<_nElY; _iely++)
    {

      if (_nElY!=1)
	{
	  sstr.str("");
	  sstr << atty->GetName() <<  "[" << _iely << "]";
	  sprintf(_sOutputAttribute, "%s", sstr.str().c_str() );

	  if (_mAttributeElements.find( string(atty->GetName()) )!=_mAttributeElements.end())
	    {
	      vector<int> sublist =  _mAttributeElements.find( string(atty->GetName()) )->second;
	      if( find( sublist.begin(), sublist.end(), _iely) == sublist.end() )
		{
		  _iOut++;
		  continue;
		}
	    }
	}

      if (_blog)
	{
	  cout << endl << " ** Case of Mean" << endl;
	  _valMeanMN.Print(" ** Case of MeanMN");
	  cout << endl << " ** Case of Sigma" << endl;
	  _valSigmaMN.Print();
	}

      // if (_blog)
      cout << " ** Case of Output atty [" << _sOutputAttribute << "]  nSimPerIndex " << _nSimPerIndex << endl;  

      TString iter_name = _tds->getIteratorName();
      
      _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==-1", _sIteratorName.Data()), "goff" );
      Values = _tds->getTuple()->GetV2(); M_N_Mi[0].assign(Values, Values + _nSimPerIndex); 
      IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
      our_parallel_sort(&M_N_Mi[0], &iterVals);
  
      _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==0", _sIteratorName.Data()), "goff" );
      Values = _tds->getTuple()->GetV2(); M_N_Mi[1].assign(Values, Values + _nSimPerIndex); 
      IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
      our_parallel_sort(&M_N_Mi[1], &iterVals);

      /// Store the two matrix M and N
      for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
	{
	  if (_blog)
	    cout << ipat << "/" << _nSimPerIndex << " --- M" << _telm->GetEntry(ipat)
		 << ", N" << _teln->GetEntry(ipat) << "  ------ atty["
		 << soutputy << "]" << endl;
	  /// Matrix M
	  _valM(ipat, _iOut) = M_N_Mi[0][ipat]; //_tds->getValue(soutputy, _telm->GetEntry(ipat), _iely);
	  /// Matrix N
	  _valN(ipat, _iOut) = M_N_Mi[1][ipat]; //_tds->getValue(soutputy, _teln->GetEntry(ipat), _iely);
	}

      /// Loop on the X attributes
      for (Int_t ix = 0; ix < _nX; ix++)
	{
	  attx = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(ix));
	  if (_blog)
	    {
	      cout << endl;
	      _valSi.Print (Form("*** iy[%d/%d]", _iOut, _nY));
	      cout << endl;
	      cout << " ** Input att [" << attx->GetName()
		   << "] _sIteratorName[" << _sIteratorName << "]_nX["
		   << _nX << "]" << endl;
	    }

	  /// Matrice of the output Of Mi
	  if( gDirectory != gROOT ) gDirectory = gROOT;
	  if( gROOT->FindObject("uranie_list_mi") ) gROOT->Delete("uranie_list_mi");
	  _tds->getTuple()->Draw(">>uranie_list_mi",
				 Form("%s==%d", _sIteratorName.Data(), ix + 1), "goff");

	  TEventList *telmi = (TEventList*) gROOT->Get("uranie_list_mi");
	  try
	    {
	      if (telmi->GetN() == 0 && _nMatrixMiNi!=kMiNiSaltelli)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TSobol::evaluateIndexes ERROR : the TDataServer \"%s\" does not contain the %d Mi matrix.",
								GetName(), ix));
	    }catch(URANIE::Exceptions::UErrorExceptions &ue)
	    {
	      ue.printMessage();
	      throw ue;
	    }

	  /// Matrice of the output Of Ni
	  if( gROOT->FindObject("uranie_list_ni") ) gROOT->Delete("uranie_list_ni");
	  _tds->getTuple()->Draw(">>uranie_list_ni",
				 Form("%s==%d", _sIteratorName.Data(), 100 * _nX + ix + 1), "goff");
	  TEventList *telni = (TEventList*) gROOT->Get("uranie_list_ni");
	  try
	    {
	      if (telni->GetN() == 0)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TSobol::evaluateIndexes ERROR : the TDataServer \"%s\" does not contain the %d Ni matrix.",
								GetName(), ix));
	    }catch(URANIE::Exceptions::UErrorExceptions &ue)
	    {
	      ue.printMessage();
	      throw ue;
	    }

	  /// Case of kSiMartinez11|kStiMartinez11
	  if ((_nSiAlgorithm & 0x1 << kSiMartinez11) || (_nStiAlgorithm & 0x1 << kStiMartinez11))
	    {
	      /// case of the Ni matrix
	      sSelectComputeStat = Form("%s==%d", _sIteratorName.Data(), 100 * _nX + ix + 1);
	      if (_blog)
		cout << " ** [kSiMartinez11|kStiMartinez11] sSelectComputeStat ["
		     << sSelectComputeStat << "]" << endl;
	      _tds->computeStatistic(soutputy, sSelectComputeStat);
	      _valMeanMN(9, _iOut) = _tds->getAttribute(soutputy)->getMean(_iely);
	      _valSigmaMN(9, _iOut) = _tds->getAttribute(soutputy)->getStd(_iely);
	    }

	  /// Init to zero all the indexes
	  for(Int_t imethod=0; imethod<_nMethodComputed; imethod++) {
	    _valSi(imethod, _iOut) = 0.0;
	    _valSti(imethod, _iOut) = 0.0;
	  }

          if( _nMatrixMiNi != kMiNiSaltelli)
          {
              _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==%d", _sIteratorName.Data(), ix + 1), "goff" );
              Values = _tds->getTuple()->GetV2(); M_N_Mi[0].assign(Values, Values + _nSimPerIndex);
              IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
              our_parallel_sort(&M_N_Mi[0], &iterVals);
 
          }
          _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==%d", _sIteratorName.Data(), 100 * _nX + ix + 1), "goff" );
          Values = _tds->getTuple()->GetV2(); M_N_Mi[1].assign(Values, Values + _nSimPerIndex);
          IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
          our_parallel_sort(&M_N_Mi[1], &iterVals);
 
	  /// Loop on the nS simulation
	  for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
	    {
	      //////////////////////////////
	      /// Matrix Mi
	      //////////////////////////////
	      
	      if( _nMatrixMiNi != kMiNiSaltelli)
	      {
		mi=M_N_Mi[0][ipat]; //_tds->getValue(soutputy, telmi->GetEntry(ipat), _iely);
		//////////////////////////////
		/// First Order
		//////////////////////////////
		//////////////////////////////
		/// sobol93  :: f(N)*f(Mi)
		if( _nSiAlgorithm & 0x1 << kSiSobol93)
		  _valSi(0, _iOut) += ( _valN(ipat, _iOut) - _valMeanMN(5, _iOut)) * ( mi - _valMeanMN(0, _iOut));
		/// Jansen99 : [f(N) - f(Mi)]^2
		if( _nSiAlgorithm & 0x1 << kSiJansen99)
		  _valSi(2, _iOut) += 0.5 * (_valN(ipat, _iOut) - mi) * (_valN(ipat, _iOut) - mi);

		//////////////////////////////
		/// Total Order
		//////////////////////////////
		/// Sobol07 f(M)[ f(M) - f(Mi)]
		if( _nStiAlgorithm & 0x1 << kStiSobol07)
		  _valSti(1, _iOut) += ( _valM(ipat, _iOut) - _valMeanMN(1, _iOut)) * (_valM(ipat, _iOut) - mi);
		/// Jansen99 : [f(M) - f(Mi)]^2
		if( _nStiAlgorithm & 0x1 << kStiJansen99)
		  _valSti(2, _iOut) += 0.5 * (_valM(ipat, _iOut) - mi) * (_valM(ipat, _iOut) - mi);
	      }


	      //////////////////////////////
	      /// Matrix Ni
	      //////////////////////////////
	      ni= M_N_Mi[1][ipat];//_tds->getValue(soutputy, telni->GetEntry(ipat), _iely);
	      {
		//////////////////////////////
		/// First Order
		//////////////////////////////
		/// Saltelli2002 ::  f(M)*[f(Ni) - f(N)]
		if( _nSiAlgorithm & 0x1 << kSiSaltelli02)
		  _valSi(1, _iOut) += ( _valM(ipat, _iOut) - _valMeanMN(1, _iOut)) * (ni - _valN(ipat, _iOut));
		/// Sumo2010 :: f(M)f(Ni)
		if( _nSiAlgorithm & 0x1 << kSiSumo10)
		  _valSi(3, _iOut) += ( _valM(ipat, _iOut) - _valMeanMN(3, _iOut)) * ( ni - _valMeanMN(8, _iOut));
		/// Martinez11 :: f(M)f(Ni)
		if( _nSiAlgorithm & 0x1 << kSiMartinez11)
		  _valSi(4, _iOut) += ( _valM(ipat, _iOut) - _valMeanMN(4, _iOut)) * ( ni - _valMeanMN(9, _iOut));

		//////////////////////////////
		/// Total Order
		//////////////////////////////
		//////////////////////////////
		//////////////////////////////
		/// Homma96 f(N)*f(Ni)
		if( _nStiAlgorithm & 0x1 << kStiHomma96)
		  _valSti(0, _iOut) += (_valN(ipat, _iOut) - _valMeanMN(5, _iOut)) * ( ni - _valMeanMN(5, _iOut));
		/// Sumo2010 :: f(M)f(Ni)
		if( _nStiAlgorithm & 0x1 << kStiSumo10)
		  _valSti(3, _iOut) += ( _valN(ipat, _iOut) - _valMeanMN(8, _iOut)) * ( ni - _valMeanMN(3, _iOut));
		/// Martinez11 :: f(M)f(Ni)
		if( _nStiAlgorithm & 0x1 << kStiMartinez11)
		  _valSti(4, _iOut) += ( _valN(ipat, _iOut) - _valMeanMN(4, _iOut)) * ( ni - _valMeanMN(9, _iOut));
	      }

	    }
	  if (_blog)
	    {
	      _valMeanMN.Print("Matrix of Mean for the M and N");
	      _valSigmaMN.Print("Matrix of Standard-Deviation for the M and N");
	    }

	  //////////////////////////////
	  /// Normalize the indexes SI=sum
	  ///   (SI/nS - mu_1*mu_2)/sigma_1*sigma_2
	  //////////////////////////////
	  for (Int_t isi = 0; isi < _nMethodComputed; isi++)
	    {
	      if( _nSiAlgorithm & 0x1 << isi )
		{
		  /// First Order
		  _valSi(isi, _iOut) /= (Double_t)(_nSimPerIndex);
		  //                _valSi(isi, _iOut) -= _valMeanMN(isi, _iOut) * _valMeanMN(isi + _nMethodComputed, _iOut);
		  _valSi(isi, _iOut) /= _valSigmaMN(isi, _iOut) * _valSigmaMN(isi + _nMethodComputed, _iOut);
		}
	      if( _nStiAlgorithm & 0x1 << isi )
		{
		  /// Total Order
		  _valSti(isi, _iOut) /= (Double_t)(_nSimPerIndex);
		  //                _valSti(isi, _iOut) -= _valMeanMN(isi, _iOut) * _valMeanMN(isi + _nMethodComputed, _iOut);
		  _valSti(isi, _iOut) /= _valSigmaMN(isi, _iOut) * _valSigmaMN(isi + _nMethodComputed, _iOut);
		}
	    }


	  //////////////////////////////
	  /// Fill the ntuple of restults with First Orders
	  //////////////////////////////
	  
	  Double_t vSeuil=0.999999;
	  Double_t dvalCrtM11=0, dvalM11ICLower=0, dvalM11ICUpper=0;
	  if( _nSiAlgorithm & 0x1 << kSiMartinez11)
	    {
	      dvalCrtM11 = _valSi(4, _iOut);
	      dvalCrtM11 = TMath::Min(dvalCrtM11,Double_t(vSeuil));
	      dvalCrtM11 = TMath::Max(dvalCrtM11,Double_t(0.0));
	      dvalM11ICLower = TMath::TanH(0.5 * TMath::Log((1.0+dvalCrtM11)/(1.0-dvalCrtM11))-1.96/sqrt(Double_t(_nSimPerIndex-3)));
	      dvalM11ICUpper = TMath::TanH(0.5 * TMath::Log((1.0+dvalCrtM11)/(1.0-dvalCrtM11))+1.96/sqrt(Double_t(_nSimPerIndex-3)));
	      dvalM11ICLower = TMath::Min(dvalM11ICLower,Double_t(vSeuil));
	      dvalM11ICLower = TMath::Max(dvalM11ICLower,Double_t(0.0));
	      dvalM11ICUpper = TMath::Min(dvalM11ICUpper,Double_t(vSeuil));
	      dvalM11ICUpper = TMath::Max(dvalM11ICUpper,Double_t(0.0));
	    }
	  
	  ///  Default First Oder :: Sobol93  :: martinez11 if kSaltelli is chosen for _nMatrixMiNi
	  if( _nMatrixMiNi == kMiNiSobol)
	    fillIndex(attx->GetName(), "First", _valSi(0, _iOut), "--first--");
	  else
	    fillIndex(attx->GetName(), "First", dvalCrtM11, "--first--", dvalM11ICLower, dvalM11ICUpper);
	  /// Other methods
	  if( _nSiAlgorithm & 0x1 << kSiSobol93) fillIndex(attx->GetName(), "First", _valSi(0, _iOut), "sobol93");
	  if( _nSiAlgorithm & 0x1 << kSiSaltelli02) fillIndex(attx->GetName(), "First", _valSi(1, _iOut), "02saltelli");
	  if( _nSiAlgorithm & 0x1 << kSiJansen99) fillIndex(attx->GetName(), "First", 1.0 - _valSi(2, _iOut), "jansen99");
	  if( _nSiAlgorithm & 0x1 << kSiSumo10) fillIndex(attx->GetName(), "First", _valSi(3, _iOut), "sumo10");
	  if( _nSiAlgorithm & 0x1 << kSiMartinez11) fillIndex(attx->GetName(), "First", dvalCrtM11, "martinez11", dvalM11ICLower, dvalM11ICUpper);

	  valfirstOrder[ix] = ((_nMatrixMiNi==kMiNiSobol) ? _valSi(0, _iOut) : dvalCrtM11);
	  
	  //////////////////////////////
	  /// Fill the ntuple of restults with Total Orders
	  //////////////////////////////

	  if( _nStiAlgorithm & 0x1 << kStiMartinez11)
	    {
	      dvalCrtM11 = 1.0 - _valSti(4, _iOut);
	      dvalCrtM11 = TMath::Min(dvalCrtM11,Double_t(vSeuil));
	      dvalCrtM11 = TMath::Max(dvalCrtM11,Double_t(0.0));
	      /// JBB Remplace these two lines to match uncertainty from JMM standalone macro.
	      // dvalStIM11ICLower = TMath::TanH(0.5 * TMath::Log((1.0+dvalCrtStIM11)/(1.0-dvalCrtStIM11))-1.96/sqrt(Double_t(_nSimPerIndex-3))); 
	      // dvalStIM11ICUpper = TMath::TanH(0.5 * TMath::Log((1.0+dvalCrtStIM11)/(1.0-dvalCrtStIM11))+1.96/sqrt(Double_t(_nSimPerIndex-3)));
	      dvalM11ICLower = 1.0 - TMath::TanH(0.5 * TMath::Log((1.0+_valSti(4, _iOut))/(1.0-_valSti(4, _iOut)))+1.96/sqrt(Double_t(_nSimPerIndex-3)));
	      dvalM11ICUpper = 1.0 - TMath::TanH(0.5 * TMath::Log((1.0+_valSti(4, _iOut))/(1.0-_valSti(4, _iOut)))-1.96/sqrt(Double_t(_nSimPerIndex-3)));
	      dvalM11ICLower = TMath::Min(dvalM11ICLower,Double_t(vSeuil));
	      dvalM11ICLower = TMath::Max(dvalM11ICLower,Double_t(0.0));
	      dvalM11ICUpper = TMath::Min(dvalM11ICUpper,Double_t(vSeuil));
	      dvalM11ICUpper = TMath::Max(dvalM11ICUpper,Double_t(0.0));
	    }

	  /// Default Total Order :: Jansen1999 :: martinez11 if kSaltelli is chosen for _nMatrixMiNi
	  if( _nMatrixMiNi == kMiNiSobol)
	    fillIndex(attx->GetName(), "Total", _valSti(2, _iOut), "--total--");
	  else
	    fillIndex(attx->GetName(), "Total", dvalCrtM11, "--total--", dvalM11ICLower, dvalM11ICUpper);
	  /// Other methods
	  if( _nStiAlgorithm & 0x1 << kStiHomma96) fillIndex(attx->GetName(), "Total", 1.0 - _valSti(0, _iOut), "homma96");
	  if( _nStiAlgorithm & 0x1 << kStiSobol07) fillIndex(attx->GetName(), "Total", _valSti(1, _iOut), "07sobol");
	  if( _nStiAlgorithm & 0x1 << kStiJansen99) fillIndex(attx->GetName(), "Total", _valSti(2, _iOut), "jansen99");
	  if( _nStiAlgorithm & 0x1 << kStiSumo10) fillIndex(attx->GetName(), "Total", 1.0 - _valSti(3, _iOut), "sumo10");
	  if( _nStiAlgorithm & 0x1 << kStiMartinez11) fillIndex(attx->GetName(), "Total", dvalCrtM11, "martinez11", dvalM11ICLower, dvalM11ICUpper);
	  
	  valTotalOrder[ix] = ((_nMatrixMiNi==kMiNiSobol) ? _valSti(2, _iOut) : dvalCrtM11);

          //if (_blog )
          cout << " ** Input att [" << attx->GetName()
               << "] First [" << valfirstOrder[ix]
               << "] Total Order[" << valTotalOrder[ix]
               << "]" << endl;
	}

      // Compute the sum
      Double_t dsumS = 0.;
      Double_t dsumSt = 0.0;
      for (Int_t ix = 0; ix < _nX; ix++)
	{
	  if ( valfirstOrder[ix] > 0.0)
	    dsumS += valfirstOrder[ix];
	  if ( valTotalOrder[ix] )
	    dsumSt += valTotalOrder[ix];
	}

      fillIndex("__sum__", "First", dsumS,"--first--");
      fillIndex("__sum__", "Total", dsumSt,"--total--");

      if( _nMatrixMiNi != kMiNiSaltelli)
	if( gROOT->FindObject("uranie_list_mi") ) gROOT->Delete("uranie_list_mi");
      if( gROOT->FindObject("uranie_list_ni") ) gROOT->Delete("uranie_list_ni");

      _iOut++;
    }
  /// Clean
  delete[] valfirstOrder;
  valfirstOrder = NULL;
  delete[] valTotalOrder;
  valTotalOrder = NULL;

  if ( _bWithData  && _withdatantp)
  {
      delete _withdatantp;
      _withdatantp=NULL;
  }
  
  if( _iOut == _nbOut)
    {
      if( gROOT->FindObject("uranie_list_m") ) gROOT->Delete("uranie_list_m");
      if( gROOT->FindObject("uranie_list_n") ) gROOT->Delete("uranie_list_n");
      //drawIndexes(Form("Sensitivity indices  : %s", _sSelectedOutput.Data()), "", option);
    }

  if (_blog)
    {
      cout << " ** End of TSobol::evaluateIndexes" << endl;
      cout << "***********************************" << endl;
    }
}



/*****************************************************/
/*  Create the TNtuple                               */
/*****************************************************/
void TSobol::createTuple(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TString sVariableNames = _tds->getIteratorName();
    sVariableNames += Form(":%s", _sIteratorName.Data());
    TString sVariableTypes = "D:D";
    
    TIter next(_listOfInputAttributes);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        sVariableNames += Form(":%s", att->GetName());
	sVariableTypes += Form(":%s", att->getDataTypeStr().Data());
    }

    if (_blog)
      cout << "sVariableNames[" << sVariableNames << "]" <<"  sVariableTypes ["<< sVariableTypes <<"]" << endl;
    
    URANIE::DataServer::TDSNtupleD *ntd = new URANIE::DataServer::TDSNtupleD(Form("%s__data__tree__", _tds->GetName()),
									     _tds->GetTitle(), sVariableNames, sVariableTypes);
    
    ntd->SetMarkerColor(2);
    ntd->SetMarkerSize(4);
    ntd->SetMarkerStyle(7);
    
    if(_tds->hasTuple())
    {      
        if ( _bWithData )
        {
            _tds->computeStatistic( _tds->getIteratorName() );
            double IterMin=_tds->getAttribute(  _tds->getIteratorName() )->getMinimum();
            if( IterMin != 0 && IterMin != 1)
                XYZ<<" PROBLEM !!"<<endl;

            int shift= ((IterMin==1) ? 0 : -1); // dataserver produced by launcher or relauncher :D (iterator starts at 1 or 0)

            if(_blog)
                XYZ<<"chosen inputs:["<<_sSelectedInput<<"],  outputs:["<<_sSelectedOutput<<"],  soboliterator:["<<_sIteratorName
                   <<"],  tdsiterator:["<<_tds->getIteratorName()<<"],  Cutting:["<<Form("%s>%d", _tds->getIteratorName(), 2*_nSimPerIndex+shift)<<"],  itermin:["<<IterMin<<"]"<<endl;

            _withdatantp = new URANIE::DataServer::TDSNtupleD( _tds->getTuple(), _sSelectedInput+":"+_sSelectedOutput,
                                                               Form("%s>%d", _tds->getIteratorName(), 2*_nSimPerIndex+shift),true);

            //Add the iterator for sobol
            double *iter = new double[1];
            TBranch *bsoboliter = _withdatantp->Branch(_sIteratorName.Data(), iter, Form("%s/D",_sIteratorName.Data()) );
            Long64_t nentries = _withdatantp->GetEntries();
            for (Long64_t i=0;i<nentries;i++)
            {
                _withdatantp->GetEntry(i);
                //                iter[0] = ((i<_nSimPerIndex) ? -1 : 0);//
                iter[0] = ((i%2==0) ? -1 : 0);
                bsoboliter->Fill();
            }
            _withdatantp->ResetBranchAddress(bsoboliter);

            //Add the global iterator            
            TBranch *biter = _withdatantp->Branch(_tds->getIteratorName(), iter, Form("%s/D",_tds->getIteratorName()) );
            for (Long64_t i=0;i<nentries;i++)
            {
                _withdatantp->GetEntry(i);
                iter[0] = i;
                biter->Fill();
            }
            _withdatantp->ResetBranchAddress(biter);
            delete[] iter; iter=NULL;
            
        }
      _tds->deleteTuple();
    }
    _tds->setTuple(ntd);
    delete ntd;
}

void TSobol::printLog()
{
    cout << "*******************************" << endl;
    cout << "** TSobol::printLog  ******" << endl;
    cout << "**  _sMethodName[" << _sMethodName << "]" << endl;
    cout << "**  _sFunctionName[" << _sFunctionName << "]" << endl;
    cout << "**  ENUM      Sampler[" << _nSamplerMethod << "]" << endl;
    cout << "**  ENUM   MatrixMiNi[" << _nMatrixMiNi << "]" << endl;
    cout << "**  ENUM  SiAlgorithm[" << _nSiAlgorithm << "]" << endl;
    cout << "**  ENUM StiAlgorithm[" << _nStiAlgorithm << "]" << endl;
    cout << "**  _nX[" << _nX << "]" << endl;
    cout << "**  _nY[" << _nY << "]" << endl;
    TSensitivity::printLog();
    cout << "*******************************" << endl;
}
} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

