//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSobol.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   wen sep 13 12:39:50 CEST 2006

 \brief Interface of class URANIE::Sensitivity::TSobol

 */
/**

 * @class URANIE::Sensitivity::TSobol
 *
 * @brief Description of the class TSobol.
 *
 * To be written by the developper.
 *
 */
#ifndef TSobol_H
#define TSobol_H
#include "TString.h"
#ifndef ROOT_TMethodCall
#include "TMethodCall.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif

// Uranie
#include "TStochAttribut.h"
#include "TSensitivity.h"
#include "TSampling.h"
#include "TQMC.h"

using std::cout;
using std::endl;

namespace URANIE {
namespace Sensitivity {
class TSobol: public TSensitivity {

    /// list Of Enum
public:
    enum ESamplerMethod
    {
        kSRS, kLHS, kqMCSobol, kqMCHalton
    };    ///< The List sampler method to generate the M and N matrix
    enum EMatrixMiNi
    {
        kMiNiSobol, kMiNiSaltelli
    };              ///< The List of method to generate the Min and/or Ni matrix
    enum ESiAlgorithm
    {
      kSiSobol93=0, kSiSaltelli02, kSiJansen99, kSiSumo10, kSiMartinez11, kSiOnlyNi=26, kSiAll=31
    };            ///< The List of method to compute the first order index. All value change to facilitate the bit-wise comparison
    enum EStiAlgorithm
    {
        kStiHomma96=0,
        kStiSobol07,
        kStiJansen99,
        kStiSumo10,
        kStiMartinez11,
  kStiOnlyNi=25,
        kStiAll=31
    };     ///< The List of method to compute the total order index. All value change to facilitate the bit-wise comparison

    // Operations
private:

protected:
    //---------------------------------------------
    /** @name Visualization and Scan the data
     **/
    //@{
    void initSample(Int_t ns, Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read option specific to TSobol
    /*!
      Possible options specific to this methods are
     \li "srs": generate sample with SRS
     \li "lhs": generate sample with LHS
     \li "qmc=sobol": generate sample with QMC=SOBOL
     \li "qmc=halton": generate sample with QMC=HALTON
     \li "sobol/saltelli": use either the solbol or saltelli implementation (the second one is the default)
    */
    virtual void parseOption(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @namePreTreatment
     **/
    //@{
    //! Resize matrices and fill many vector/matrice to avroid recompute stat too oftently
    void preTreatment();
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TSobol();
    //! Default constructor with the TDataServer
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param fcn (void *) The function to analyze
     \param ns (Int_t) The size (total) of the sampling
     \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
     \param
     */
#ifdef ROOT5
    TSobol(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns,
            const char *varexpinput = "", const char *varexpoutput = "",
            Option_t * option = "");
#else
    TSobol(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
            const char *varexpinput, const char *varexpoutput,
            Int_t ns, Option_t * option = "");
#endif
    //! Default constructor with the TDataServer
    /*
     \param tds (URANIE::DataServer::TDataServer *) the dataserver
     \param fcn (const char*) The function to analyze
     \param ns (Int_t) The size (total) of the sampling
     \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
     \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
     \param
     */
    TSobol(URANIE::DataServer::TDataServer *tds, const char *fcn,
     Int_t ns, const char *varexpinput, const char *varexpoutput,
     Option_t* option="");
    //! Default constructor with a TDataServer filling
    TSobol(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");

    //! Default constructor with a tcode
    TSobol(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode,
            Int_t ns, Option_t * option = "");
    //! Default constructor with a trun
    TSobol(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,
            Int_t ns, Option_t * option = "");
    //! Default destructor
    virtual ~TSobol();
    //@}

    //---------------------------------------------
    /** @name Constructor's helper
     **/
    //@{
    void initForRun(Option_t *option);
    //@}

    
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    void createTuple(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates all the matrix of sample
    /*
     \li from a TDS empty of data . Run the method \ref initSample
     \li from a TDS with data. First Half of the data is M and the other is N

     After mix the data to create the \f$ M_{i}\f$ and \f$ N_{i}\f$ matrix
     */
    virtual void generateSample(Option_t * option = "");

    void loadOtherSobolFile(const char *filename);
    //@}

    //---------------------------------------------
    //@{
    //! Evaluates the index from a Specific TDataServer
    /*
     The TDS must contain an attribute Name + "n__iter__sobol" with values
     \li            -1 : for the M matrix
     \li             0 : for the N matrix
     \li          ix+1 : for the matrix Mi
     \li 100*_nX+ ix+1 : for the matrix Ni
     */
    void evaluateIndexes(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog();
    //@}

    ClassDef(URANIE::Sensitivity::TSobol, ID_SENSITIVITY)

    //---------------------------------------------
    /** @name Private and Public attributes
     *
     **/
    //@{
    Bool_t _bSM;          ///< If true function or Surrogate Model
    Bool_t _bWithSample;    ///< true if data are provided and used as full sample of 2 times ns samples
    Bool_t _bWithData;    ///< true if data are provided and used as full sample of 2 times ns samples

    ESamplerMethod _nSamplerMethod; ///< The sampler method to generate the M and N matrix
    EMatrixMiNi _nMatrixMiNi; ///< The method to generate the Min and/or Ni matrix (Sobol, Saltelli)

    ESiAlgorithm _nSiAlgorithm; ///< The algorithm to use to compute the first order algorithm
    EStiAlgorithm _nStiAlgorithm; ///< The algorithm to use to compute the total order algorithm

    //---------------------------------------------
    /** @name Method
     The method needs two matrix M, N
     **/
    //@{
    Int_t _nSimPerIndex;     ///< The size of the two matrix M et N.
    Int_t _nMethodComputed; ///< number of method used
    TMatrixD _matM;             ///< The matrix M
    TMatrixD _matN;             ///< The matrix M

    TEventList *_telm, *_teln;  ///< m and n lit of event in the ntuple
    TMatrixD _valN, _valM;  ///< values of output for M and N matrix
    TMatrixD _valSi, _valSti;  ///< values of the first and total sobol indexes
    TMatrixD _valMeanMN, _valSigmaMN; ///< complexe matrices to store the mean and variance of used M/N/Mi/Ni matrices
    vector<double> _valMeanYN, _valMeanYM, _valVarYN, _valVarYM;
    URANIE::DataServer::TDSNtupleD *_withdatantp;
    URANIE::DataServer::TDSNtupleD *_preloadedntp;
    int *_useless;
    //@}

};

} // End Of namespace Sensitivity
} // End Of namespace URANIE
#endif
