//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef CPPUNITEST_TDGSM_H
#define CPPUNITEST_TDGSM_H


#include <cppunit/extensions/HelperMacros.h>
//#include <cppunit/portability/CppUnitDeque.h>
#include <cppunit/Test.h>
#include <cppunit/TestFailure.h>
#include <cppunit/TestResultCollector.h>

// ROOT
#include "TMatrixD.h"

// Uranie
#include "TDGSM.h"
#include "UExceptions.h"
#include "TUniformDistribution.h"


class TDGSMTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE (TDGSMTest);
    CPPUNIT_TEST (testDGSMConstructor);
    CPPUNIT_TEST (testDGSMIshigamiFunction);
    CPPUNIT_TEST (testDGSMAffineFunction);
    CPPUNIT_TEST (testDGSMCounterExemple3Function);
    CPPUNIT_TEST (testDGSMCounterExemple3TDS);
    CPPUNIT_TEST (testDGSMInitSample);
    CPPUNIT_TEST (testDGSMGenerateSample);

    CPPUNIT_TEST (throwExceptionDGSMConstructor);
    CPPUNIT_TEST (throwExceptionDGSMInitSample);

    CPPUNIT_TEST_SUITE_END();

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TDGSMTest();

    //! Default destructor
    virtual ~TDGSMTest();
    //@}

    /** @name setUp/tearDown and tools methods
     **/
    //@{
    /*!
     * \brief Nominal setup
     */
    void setUp();

    /*!
     * \brief setUp for the Affine usecase
     */
    void setUpAffine();
    /*!
     * \brief setUp for the Ishigami usecase
     */
    void setUpIshigami();
    /*!
     * \brief setUp for the Flowrate usecase
     */
    void setUpFlowrate();
    /*!
     * \brief Nominal tearDown
     */
    void tearDown();

    //! the clean tempory directories
    void cleanWorkingDirectories(const char *option = "without");
    //@}

    //---------------------------------------------

    /** @name Validation of the indexes for the usecases
     **/
    //@{
    /*!
     * \brief validate the sensivity indexes for the Affine usecase
     */
    void validateValuesAffine();
    /*!
     * \brief validate the sensivity indexes for the Ishigami usecase
     */
    void validateValuesIshigami();
    /*!
     * \brief validate the sensivity indexes for the Flowrate usecase
     */
    void validateValuesFlowrate();
    //@}

    //---------------------------------------------
    /** @name Unit Tests
     The list of unit test
     **/
    //@{
    //! Compute DGSM constructor
    void testDGSMConstructor();
    //! Init the sample points
    void testDGSMInitSample();
    //! Init the sample points
    void testDGSMGenerateSample();
    //! Compute DGSM of a function (ishigami)
    void testDGSMIshigamiFunction();
    //! Compute DGSM of a function (CounterExemple3)
    void testDGSMCounterExemple3Function();
    //! Compute DGSM of a function (CounterExemple3) from a TDS
    void testDGSMCounterExemple3TDS();
    //! Compute DGSM of a function (Affine)
    void testDGSMAffineFunction();
    //@}


    //---------------------------------------------
    /** @name Throw Exceptions
     The list of Throw Exceptions
     **/
    //@{
    //! Throw Exceptions for the constructor
    void throwExceptionDGSMConstructor();

    //! Throw exceptions for the generateSample method
    void throwExceptionDGSMInitSample();
    //@}


private:
    TString _workingTestPath;
    URANIE::Sensitivity::TDGSM * _tsen;
    Int_t _nS;

    // UseCase Flowrate
    Int_t _nsFlowrate ;
    URANIE::DataServer::TUniformDistribution *rw;
    URANIE::DataServer::TUniformDistribution *r;
    URANIE::DataServer::TUniformDistribution *tu;
    URANIE::DataServer::TUniformDistribution *tl;
    URANIE::DataServer::TUniformDistribution *hu;
    URANIE::DataServer::TUniformDistribution *hl;
    URANIE::DataServer::TUniformDistribution *l;
    URANIE::DataServer::TUniformDistribution *kw;
    TMatrixD _refMatFlowrate;

    // UseCase Ishigami
    Int_t _nSIshigami ;
    URANIE::DataServer::TUniformDistribution *x1;
    URANIE::DataServer::TUniformDistribution *x2;
    URANIE::DataServer::TUniformDistribution *x3;
    TMatrixD _refMatIshigami;

    URANIE::DataServer::TDataServer *_tds1, *_tds2;
    URANIE::Launcher::TCode *_mycode;
#ifdef ROOT5
    void * _p2fIshigami, *_p2fFlowrate, *_p2fAffine;
#endif
    URANIE::Launcher::TOutputFile *_fout;
    Bool_t _bLog;
};


#endif  // CPPUNITEST_TDGSM_H
