//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTFiniteDifferences.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$

 \brief  Test File of URANIE::Sensitivity::TFiniteDifferences class


 */

// Uranie
#include "TestTSensitivity.h"
#include "TestTFiniteDifferences.h"

#include "TFiniteDifferences.h"
#include "TDataServer.h" 
#include "TLogUniformDistribution.h"
#include "TCode.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TOutputFileColumn.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TestUserFunctions.h"
#include "uranie_commands.h"

#include "TCJitEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "code_paths.h"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TSystem.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-12
#define LVLE_EPSILON 0.001
#define EPSILON 0.01

using namespace std;

TFiniteDifferencesTest::TFiniteDifferencesTest() :
        _tds(NULL), _tds2(NULL), _tfd(NULL), _mycode(NULL), _p2fIshigami(NULL), _fout(NULL)
{
    _workingTestPath = gSystem->pwd();

#ifdef ROOT5
    //  gROOT->LoadMacro("UserFunctions.C");
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("Ishigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();
#endif

    // Define the reference results
    _refMat.ResizeTo(1, 3);
    _refMat(0, 0) = 2.07047310000000016e+03;
    _refMat(0, 1) = -3.43887775551112372e-07;
    _refMat(0, 2) = -1.66780178571428600e-02;
}

TFiniteDifferencesTest::~TFiniteDifferencesTest()
{
}


void TFiniteDifferencesTest::cleanWorkingDirectories(const char *option)
{
    cout << endl;
    cout << " ** TFiniteDifferencesTest::cleanWorkingDirectories option["
            << option << "] _workingTestPath[" << _workingTestPath << "]"
            << endl;

    UranieTests::removeFile( _workingTestPath.Data(), "_finitedifferences_launching_.dat");
    UranieTests::removeFolder(_workingTestPath.Data(),"URANIE");
    UranieTests::removeFolder(_workingTestPath.Data(),"tmpLanceurUranie");
    UranieTests::removeFolder(_workingTestPath.Data(),"FiniteDifferencesLevelEDatabase.root");    
}

void TFiniteDifferencesTest::setUp()
{
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("setUp");

    rw = new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15);
    r = new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0);
    tu = new URANIE::DataServer::TUniformDistribution("tu", 63070.0, 115600.0);
    tl = new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0);
    hu = new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0);
    hl = new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0);
    l = new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0);
    kw = new URANIE::DataServer::TUniformDistribution("kw", 9855.0, 12045.0);

    rw->setDefaultValue(0.05);
    r->setDefaultValue(33366.67);
    tu->setDefaultValue(63070.0);
    tl->setDefaultValue(116.0); 
    hu->setDefaultValue(1110.0);
    hl->setDefaultValue(768.57);
    l->setDefaultValue(1200.0);
    kw->setDefaultValue(11732.14);

    x1 = new URANIE::DataServer::TUniformDistribution("x_{1}",
            -1.0 * TMath::Pi(), TMath::Pi());
    x2 = new URANIE::DataServer::TUniformDistribution("x_{2}",
            -1.0 * TMath::Pi(), TMath::Pi());
    x3 = new URANIE::DataServer::TUniformDistribution("x_{3}",
            -1.0 * TMath::Pi(), TMath::Pi());

    x1->setDefaultValue(0.);
    x2->setDefaultValue(0.);
    x3->setDefaultValue(0.);

}

void TFiniteDifferencesTest::tearDown()
{
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("tearDown");

    if (_fout)
    {
        delete _fout;
        _fout = NULL;
    }

    if (_mycode != NULL)
    {
        delete _mycode;
        _mycode = NULL;
    }

    if (_tfd)
    {
        delete _tfd;
        _tfd = NULL;
    }

    if (_tds)
      {
        delete _tds;
        _tds = NULL;
      }
    
    if (_tds2)
      {
        delete _tds2;
        _tds2 = NULL;
      }

    delete rw;
    rw = NULL;
    delete r;
    r = NULL;
    delete tu;
    tu = NULL;
    delete tl;
    tl = NULL;
    delete hu;
    hu = NULL;
    delete hl;
    hl = NULL;
    delete l;
    l = NULL;
    delete kw;
    kw = NULL;

    delete x1;
    x1 = NULL;
    delete x2;
    x2 = NULL;
    delete x3;
    x3 = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TFiniteDifferencesTest::setUp1()
{
    _tds2 = new URANIE::DataServer::TDataServer("tds", "");
    _tds2->addAttribute(x1);
    _tds2->addAttribute(x2);
    _tds2->addAttribute(x3);

#ifdef ROOT5
    _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds2, _p2fIshigami,
            "Ishigami", "x_{1}:x_{2}:x_{3}", "minmax");
#else
    _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds2, &Ishigami,
            "x_{1}:x_{2}:x_{3}", "Ishigami", "minmax");
#endif
    _tfd->setDrawProgressBar(kFALSE);
    _tfd->setSeed(123);
}

void TFiniteDifferencesTest::setUp2()
{
    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_values_rows.in";

    TString outputFileName = "_output_flowrate_withRow_.dat";

    rw->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    r->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    tu->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    tl->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    hu->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    hl->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    l->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    kw->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    _tds = new URANIE::DataServer::TDataServer("tds", "");
    _tds->addAttribute(rw);
    _tds->addAttribute(r);
    _tds->addAttribute(tu);
    _tds->addAttribute(tl);
    _tds->addAttribute(hu);
    _tds->addAttribute(hl);
    _tds->addAttribute(l);
    _tds->addAttribute(kw);

    _fout = new URANIE::Launcher::TOutputFileDataServer(outputFileName);
    _fout->addAttribute("yhat");

    _mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);

    _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds, _mycode, "rw:r:l",
            "minmax");
    _tfd->setDrawProgressBar(kFALSE);
    _tfd->setSeed(123);
}


void TFiniteDifferencesTest::testFiniteDifferencesWithCode()
{
    TString inputFileName = gSystem->pwd();
#ifdef WIN32
    inputFileName += "\\";
#else
    inputFileName += "/";
#endif 
    inputFileName += "flowrate_input_with_keys.in";
    TString outputFileName = "_output_flowrate_withKey_.dat";

    rw->setFileKey(inputFileName, "Rw");
    r->setFileKey(inputFileName, "R");
    tu->setFileKey(inputFileName, "Tu");
    tl->setFileKey(inputFileName, "Tl");
    hu->setFileKey(inputFileName, "Hu");
    hl->setFileKey(inputFileName, "Hl");
    l->setFileKey(inputFileName, "L");
    kw->setFileKey(inputFileName, "Kw");

    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer(
            "tds", "");

    tds->addAttribute(rw);
    tds->addAttribute(r);
    tds->addAttribute(tu);
    tds->addAttribute(tl);
    tds->addAttribute(hu);
    tds->addAttribute(hl);
    tds->addAttribute(l);
    tds->addAttribute(kw);

    URANIE::Launcher::TOutputFileKey *fout =
            new URANIE::Launcher::TOutputFileKey(outputFileName);
    fout->addAttribute("yhat");

    TString sWorkingTestPath = gSystem->pwd();
    URANIE::Launcher::TCode *flowrateCode;
    flowrateCode = new URANIE::Launcher::TCode(tds, UranieTests::code_path("flowrate") + " -s -k");
    flowrateCode->addOutputFile(fout);

    URANIE::Sensitivity::TFiniteDifferences *fd =
            new URANIE::Sensitivity::TFiniteDifferences(tds, flowrateCode,
                    "rw:r:l", "minmax");
    fd->setDrawProgressBar(kFALSE);
    fd->computeIndexes("nointermed");
    tds->exportData("_oat_.dat");

    TMatrixD sMat = fd->getSensitivityMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 0), sMat(0, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 1), sMat(0, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 2), sMat(0, 2),
            DBL_PRECISION_COMPARE);

    delete fd;
    fd = NULL;
    delete flowrateCode;
    flowrateCode = NULL;
    delete fout;
    fout = NULL;
    delete tds;
    tds = NULL;
}

// void TFiniteDifferencesTest::testFiniteDifferencesWithFunction()
// {
//    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer("tds","");

//    tds->addAttribute(rw);
//    tds->addAttribute(r);
//    tds->addAttribute(tu);
//    tds->addAttribute(tl);
//    tds->addAttribute(hu);
//    tds->addAttribute(hl);
//    tds->addAttribute(l);
//    tds->addAttribute(kw);
// //
//    gROOT->LoadMacro("UserFunctions.C");
//    URANIE::Sensitivity::TFiniteDifferences *fd = new URANIE::Sensitivity::TFiniteDifferences(tds, 
//                                                                                              flowrateModel, 
//                                                                                              "yhat", 
//                                                                                              "rw:r:l", 
//                                                                                              "minmax");
//    fd->computeIndexes();
//    tds->exportData("_oat_.dat");

//    TMatrixD sMat = fd->getSensitivityMatrix();
//    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0,0), sMat(0,0), DBL_PRECISION_COMPARE);
//    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0,1), sMat(0,1), DBL_PRECISION_COMPARE);
//    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0,2), sMat(0,2), DBL_PRECISION_COMPARE);

// }


#ifdef WITH_LEVELE
// ======================================
//            LevelE
// ======================================
void TFiniteDifferencesTest::checkLevelEResults(TMatrixD mat, vector<int> *elmts)
{

  double wvalref[26]={-7.82570351109e-37,-1.27865573505e-32,-1.3354223697e-29,-2.61616705864e-27,-1.68501367283e-25,-4.99300635274e-24,-8.3807446611e-23,-9.15831109764e-22,
		      -7.16366925584e-21,-5.94895538114e-16,-5.47484116148e-14,-4.23705721116e-13,-9.58410542769e-13,-1.06784741639e-12,-7.62452330005e-13,-4.04786555837e-13,
		      -1.74859218359e-13,-6.51289731781e-14,-1.20917352416e-19,-8.83751111529e-32,-1.83042695287e-30,-7.2648499961e-30,-1.3055403629e-29,-1.69028158234e-29,
		      -1.97347859684e-29,-2.24913710024e-29};


  for(unsigned int ival=0; ival<((elmts==NULL) ? 26 : elmts->size()); ival++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(wvalref[ival], mat(((elmts==NULL) ? ival : elmts->at(ival)),11) , LVLE_EPSILON);
}

void TFiniteDifferencesTest::testLevelEFiniteDifferences()
{

    // Define the DataServer
  _tds = new URANIE::DataServer::TDataServer("tdsLevelE","levele");
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("t", 100, 1000) ); _tds->getAttribute("t")->setDefaultValue(500);
  _tds->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kl", 0.001, 0.01) ); _tds->getAttribute("kl")->setDefaultValue(0.0015);
  _tds->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kc", 0.000001, 0.00001) ); _tds->getAttribute("kc")->setDefaultValue(0.0000015);
  _tds->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v1", 0.001, 0.1) ); _tds->getAttribute("v1")->setDefaultValue(0.0015);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("l1", 100, 500) ); _tds->getAttribute("l1")->setDefaultValue(300);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("r1", 1, 5) ); _tds->getAttribute("r1")->setDefaultValue(3);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("rc1", 3, 30) ); _tds->getAttribute("rc1")->setDefaultValue(16.5);
  _tds->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v2", 0.01, 0.1) ); _tds->getAttribute("v2")->setDefaultValue(0.015);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("l2", 50, 200) ); _tds->getAttribute("l2")->setDefaultValue(125);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("r2", 1, 5) ); _tds->getAttribute("r2")->setDefaultValue(3);
  _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("rc2", 3, 30) ); _tds->getAttribute("rc2")->setDefaultValue(16.5);
  _tds->addAttribute( new URANIE::DataServer::TLogUniformDistribution("w", 100000, 10000000) ); _tds->getAttribute("w")->setDefaultValue(100100);

  _tds->getAttribute( "t" )->delShare();
  _tds->getAttribute( "kl" )->delShare();
  _tds->getAttribute( "kc" )->delShare();
  _tds->getAttribute( "v1" )->delShare();
  _tds->getAttribute( "l1" )->delShare();
  _tds->getAttribute( "r1" )->delShare();
  _tds->getAttribute( "rc1" )->delShare();
  _tds->getAttribute( "v2" )->delShare();
  _tds->getAttribute( "l2" )->delShare();
  _tds->getAttribute( "r2" )->delShare();
  _tds->getAttribute( "rc2" )->delShare();
  _tds->getAttribute( "w" )->delShare();

  TString name="levelE_input_with_values_rows.in";
  _tds->getAttribute("t")->setFileKey(name, "t", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("kl")->setFileKey(name, "kl", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("kc")->setFileKey(name, "kc", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("v1")->setFileKey(name, "v1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("l1")->setFileKey(name, "l1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("r1")->setFileKey(name, "r1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("rc1")->setFileKey(name, "rc1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("v2")->setFileKey(name, "v2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("l2")->setFileKey(name, "l2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("r2")->setFileKey(name, "r2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("rc2")->setFileKey(name, "rc2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds->getAttribute("w")->setFileKey(name, "w", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);

  URANIE::Launcher::TOutputFileRow *out = new URANIE::Launcher::TOutputFileRow("_output_levelE_withRow_.dat");
  out->addAttribute(new URANIE::DataServer::TAttribute("tps", URANIE::DataServer::TAttribute::kVector) );
  out->addAttribute(new URANIE::DataServer::TAttribute("y", URANIE::DataServer::TAttribute::kVector) );
  out->getAttribute(0)->delShare();
  out->getAttribute(1)->delShare();
  
  URANIE::Launcher::TCode *myc;
#ifdef WIN32
  myc = new URANIE::Launcher::TCode(_tds, "levele.exe");
#else
  myc = new URANIE::Launcher::TCode(_tds, Form("%s/%s",gSystem->pwd(),"../../../../launCHER/teST/coDE/levelE/levele 2> /dev/null"));
#endif
  myc->addOutputFile(out);

  _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds, myc);
  //_tfd->setLog();
  _tfd->setSeed(12345);
  _tfd->setDrawProgressBar(false);
  _tfd->computeIndexes("output=y");
  
  TMatrixD mat = _tfd->getSensitivityMatrix();
  checkLevelEResults(mat);
  
  delete out;
  delete myc;
}

void TFiniteDifferencesTest::testLevelEFiniteDifferencesOutputRestrVarexp()
{

  delete _tds; _tds = new URANIE::DataServer::TDataServer();
  _tds->fileDataRead("FiniteDifferencesLevelEDatabase.dat");
  
  _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]");
  _tfd->computeIndexes();

  TMatrixD ntresu = _tfd->getSensitivityMatrix();
  checkLevelEResults(ntresu, _tfd->getAttributeElements(string("y")));
}

void TFiniteDifferencesTest::testLevelEFiniteDifferencesOutputRestrOption()
{
  
  delete _tds; _tds = new URANIE::DataServer::TDataServer();
  _tds->fileDataRead("FiniteDifferencesLevelEDatabase.dat");

  _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y");
  _tfd->computeIndexes("output=y[4]:y[11]:y[22]");

  TMatrixD ntresu = _tfd->getSensitivityMatrix();
  checkLevelEResults(ntresu, _tfd->getAttributeElements(string("y")));
}

void TFiniteDifferencesTest::testLevelEFiniteDifferencesOutputRestrBoth()
{

  delete _tds; _tds = new URANIE::DataServer::TDataServer();
  _tds->fileDataRead("FiniteDifferencesLevelEDatabase.dat");

  _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]");
  _tfd->computeIndexes("output=y[4]:y[11]:y[22]");

  TMatrixD ntresu = _tfd->getSensitivityMatrix();
  checkLevelEResults(ntresu, _tfd->getAttributeElements(string("y")));
}
#endif

void TFiniteDifferencesTest::testFiniteDifferencesWithDataFile()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer(
            "tds", "");
    tds->fileDataRead("_oat_.dat");

    URANIE::Sensitivity::TFiniteDifferences *fd =
            new URANIE::Sensitivity::TFiniteDifferences(tds,
                    TString("rw:r:tu:tl:hu:hl:l:kw"), TString("yhat"),
                    TString("rw:r:l"));
    fd->setDrawProgressBar(kFALSE);
    fd->computeIndexes();
    TMatrixD sMat = fd->getSensitivityMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 0), sMat(0, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 1), sMat(0, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, 2), sMat(0, 2),
            DBL_PRECISION_COMPARE);
    delete fd;
    fd = NULL;
    delete tds;

}

void TFiniteDifferencesTest::validateValuesIshigami()
{
    TMatrixD sMat = _tfd->getSensitivityMatrix();

    TMatrixD refMat2(1, 3);
    refMat2(0, 0) = 3.89804e-17;
    refMat2(0, 1) = 0.;
    refMat2(0, 2) = 0.;

    for (int i = 0; i < 3; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refMat2(0, i), sMat(0, i), 5e-3);
    }
}

void TFiniteDifferencesTest::testIshigamiFiniteDifferencesComputeIndexesWithTLauncherFunction()
{
    setUp1();
    _tfd->computeIndexes();
    validateValuesIshigami();
}

void TFiniteDifferencesTest::testIshigamiFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherFunction()
{
    setUp1();
    _tfd->generateSample();
    _tfd->computeIndexes();
    validateValuesIshigami();
}

void TFiniteDifferencesTest::testIshigamiFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction()
{
    setUp1();
    _tfd->generateSample();
    URANIE::Launcher::TLauncherFunction *tlf =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(_tds2, _p2fIshigami);
#else
            new URANIE::Launcher::TLauncherFunction(_tds2, &Ishigami, "", "Ishigami");
#endif
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    _tfd->computeIndexes();
    validateValuesIshigami();
    
    delete tlf;
}

void TFiniteDifferencesTest::testIshigamiFiniteDifferencesComputeIndexesWithTLauncher2()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("Ishigami");
    out->delShare();
   
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.setInputs(3, x1, x2, x3);
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        _tds2 = new URANIE::DataServer::TDataServer("tds", "");
        _tds2->addAttribute(x1);
        _tds2->addAttribute(x2);
        _tds2->addAttribute(x3);

        _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds2, 
                &run, "x_{1}:x_{2}:x_{3}", "minmax");
        _tfd->setDrawProgressBar(kFALSE);
        _tfd->setSeed(123);
        _tfd->computeIndexes();
        validateValuesIshigami();
        
        run.stopSlave();
    }
}

void TFiniteDifferencesTest::testIshigamiFiniteDifferencesComputeIndexesThreaded()
{

    URANIE::DataServer::TAttribute *outT = new URANIE::DataServer::TAttribute("TestIshigamiT");
    outT->delShare();
   
    URANIE::Relauncher::TCJitEval evalT(&Ishigami);
    evalT.setInputs(3, x1, x2, x3);
    evalT.addOutput(outT);
    
    URANIE::Relauncher::TThreadedRun runT(&evalT,4);
    runT.startSlave();
    if (runT.onMaster()) {
        _tds2 = new URANIE::DataServer::TDataServer("tds", "");
        _tds2->addAttribute(x1);
        _tds2->addAttribute(x2);
        _tds2->addAttribute(x3);

        _tfd = new URANIE::Sensitivity::TFiniteDifferences(_tds2, 
                &runT, "x_{1}:x_{2}:x_{3}", "minmax");
        _tfd->setDrawProgressBar(kFALSE);
        _tfd->setSeed(123);
        _tfd->computeIndexes();
        validateValuesIshigami();
        
        runT.stopSlave();
    }
}


void TFiniteDifferencesTest::validateValuesFlowrate()
{
    TMatrixD sMat = _tfd->getSensitivityMatrix();

    for (int i = 0; i < 3; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(_refMat(0, i), sMat(0, i), 5e-3);
    }
}

void TFiniteDifferencesTest::testFlowrateFiniteDifferencesComputeIndexesWithTLauncherCode()
{
    setUp2();
    _tfd->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TFiniteDifferencesTest::testFlowrateFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherCode()
{
    setUp2();
    _tfd->generateSample();
    _tfd->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TFiniteDifferencesTest::testFlowrateFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode()
{
    setUp2();
    _tfd->generateSample();
    URANIE::Launcher::TLauncher *tlf = new URANIE::Launcher::TLauncher(_tds,
            _mycode); 
    tlf->setDrawProgressBar(kFALSE);
    tlf->run("nointermed");
    _tfd->computeIndexes("nointermed");
    validateValuesFlowrate();
    delete tlf;
}

void TFiniteDifferencesTest::throwExceptionConstructorTooLongOutputFunction() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs, true);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    string inputs_name;
    TList *list=tds.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TFiniteDifferences finite_diff(&tds,
        flowrateModel, inputs_name.data(),
        "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut",
        "steps=1%"),
        URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TFiniteDifferences finite_diff(&tds,
        flowrateModel, inputs_name.data(),
        "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut",
        "steps=1%"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TFiniteDifferencesTest::throwExceptionConstructorTooLongOutputCode() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs, true);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(tds, input_file);

    URANIE::Launcher::TOutputFileColumn output_file("_output_flowrate_withColumn_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");
    
    URANIE::Launcher::TCode code(&tds, "");
    code.addOutputFile(&output_file);

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TFiniteDifferences
        finite_diff(&tds, &code), URANIE::Exceptions::UErrorExceptions);
}

void TFiniteDifferencesTest::throwExceptionConstructorTooLongOutputRun() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs, true);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    for(auto& input: inputs)
        eval.addInput(&input);
    eval.addOutput(&output);

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::DataServer::TDataServer tds("tds", "");
        for (auto& input: inputs)
            tds.addAttribute(&input);
        CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TFiniteDifferences
            finite_diff(&tds, &run), URANIE::Exceptions::UErrorExceptions);
        run.stopSlave();
    }
}
