//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTHSIC.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2019/01/11

 \brief  Fichier de test de la classe URANIE::Sensitivity::THSIC

 */

// Uranie
#include "TestTHSIC.h"
#include "THSIC.h"
#include "UDisableTest.h"
#include "TUniformDistribution.h"
#include "TestUserFunctions.h"

// ROOT
#include "TString.h"
#include "TLeaf.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif
#include "uranie_commands.h"
 
#define EPSILON 0.001
#define SPL_PRECISION_COMPARE 1.e-6
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>
using namespace std;

THSICTest::THSICTest() :
        _tdsA(NULL), _tHSIC(NULL)
{
}


THSICTest::~THSICTest()
{
  UranieTests::removeFolder(gSystem->pwd(),"URANIE");
  UranieTests::removeFile(gSystem->pwd(), "_HSIC_launching_.dat");
  UranieTests::removeFile(gSystem->pwd(), "_HSIC_sampling_.dat");
}

void THSICTest::setUp()
{
    _tdsA = new URANIE::DataServer::TDataServer("tdsA", "my first TDS");
    _tdsA->fileDataRead("flowrateUniformDesign.dat");

    _tdsD = new URANIE::DataServer::TDataServer("tdsD", "TDS for Ishigami function");
}

void THSICTest::tearDown()
{
    if ( _tHSIC != NULL ) {
        delete _tHSIC;
        _tHSIC = NULL;
    }

    delete _tdsA;
    _tdsA = NULL;

    delete _tdsD;
    _tdsD = NULL;

    /// Remove ROOT Files
    UranieTests::removeFile(gSystem->pwd(), "Ishigami_00300.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign.root");

    /// Remove ROOT Objects
    if( gROOT->FindObject("evlHSIC") )
        gROOT->Delete("evlHSIC");
    if( gROOT->FindObject("R2HSic") )
        gROOT->Delete("R2HSic");
    if( gROOT->FindObject("hPValues") )
        gROOT->Delete("hPValues");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void THSICTest::testconstructorHSIC()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");

    CPPUNIT_ASSERT_MESSAGE("The Log flag via getLog()",
                           _tHSIC->getLog() == kFALSE);
    _tHSIC->unsetLog();

    Int_t nS = _tdsA->getNPatterns();

    /// Default Values for the THSIC Object
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Bootstrap permutation (default)",
                                 NDefaultBootstrap, _tHSIC->getNBoostrapSample());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Bootstrap (default)",
                                 NDefaultThresholdBootstrap, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Gamma Approximation (default)",
                                 NDefaultThresholdGamma, _tHSIC->getThresholdGamma());
    CPPUNIT_ASSERT_MESSAGE("Name of the method (HSIC)",
                           strcmp(_tHSIC->GetName(), "HSIC") == 0);
    CPPUNIT_ASSERT_MESSAGE("Name of the method (HSic)",
                           strcmp(_tHSIC->getMethodName(), "HSic") == 0);

    /// Default Values for the TSensitivity Object with the Getter Functions
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input getID()",
                                 ID_SENSITIVITY, _tHSIC->getID());
    CPPUNIT_ASSERT_MESSAGE("The pointer of the result TTree getResultTuple()",
                           _tHSIC->getResultTuple() != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of the result TTree getResultTuple()",
                                 0, (Int_t)(_tHSIC->getResultTuple()->GetEntries()));
    const Int_t nBranches = 8;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of branches of the result TTree getResultTuple()",
                                 nBranches, (Int_t)(_tHSIC->getResultTuple()->GetNbranches()));
    
    TString sBranches[nBranches] = {"Out", "Inp", "Order", "Method", "Algo", "Sobol", "CILower", "CIUpper"};
    for (Int_t i=0; i < nBranches; i++) {
        CPPUNIT_ASSERT_MESSAGE(Form("Status of the branch \"%s\" in the result TTree getResultTuple()", sBranches[i].Data()),
                               _tHSIC->getResultTuple()->GetBranchStatus(sBranches[i].Data()));
    }
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the evaluatuor function getFunctionName() [%s]", _tHSIC->getFunctionName().Data()),
                           strcmp(_tHSIC->getFunctionName(), "") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("The Seed value getSeed()[%d]", _tHSIC->getSeed()),
                                 0, (Int_t)(_tHSIC->getSeed()));
    CPPUNIT_ASSERT_MESSAGE("The noIntermediateSaved flag",
                           _tHSIC->getNoIntermediateSaved() == kFALSE);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the Iterator getIteratorName() [%s]", _tHSIC->getIteratorName()),
                           strcmp(_tHSIC->getIteratorName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the time attribute getTimeName() [%s]", _tHSIC->getTimeName().Data()),
                           strcmp(_tHSIC->getTimeName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("The draw progress bar flag",
                           _tHSIC->getDrawProgressBar());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 8, _tHSIC->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 1, _tHSIC->getNOutput());

    /// Default Values for the TSensitivity Object WITHOUT Getter Functions
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TDS is NULL (default)",
                                 _tHSIC->_tds != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample via _nS",
                                 nS, _tHSIC->_nS);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of elements of Y via _nElY",
                                 0, (Int_t)(_tHSIC->_nElY));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via _nbOut",
                                 0, _tHSIC->_nbOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Counter for output via _iOut",
                                 0, _tHSIC->_iOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iy",
                                 0, (Int_t)(_tHSIC->_iy));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iely",
                                 0, (Int_t)(_tHSIC->_iely));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Type of launcher (default)",
                                 URANIE::Sensitivity::TSensitivity::kUnknown, _tHSIC->_nLauncher);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TCode is NULL (default)",
                                 _tHSIC->_code == NULL);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TRun is NULL (default)",
                                 _tHSIC->_run == NULL);
    CPPUNIT_ASSERT_MESSAGE("The Fact that the input list is provided or not",
                           _tHSIC->_bChosenOutputs == kTRUE);

    ///
    /// Default constructor (empty)
    ///
    cout << endl;

    URANIE::Sensitivity::THSIC *defHSIC = new URANIE::Sensitivity::THSIC();

    CPPUNIT_ASSERT_MESSAGE("The Log flag via getLog()",
                           defHSIC->getLog() == kFALSE);

    defHSIC->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default constructor)",
                                 URANIE::Sensitivity::THSIC::kUnknown, defHSIC->getEstimatedVariance());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Bootstrap permutation (default)",
                                 NDefaultBootstrap, defHSIC->getNBoostrapSample());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Bootstrap (default)",
                                 NDefaultThresholdBootstrap, defHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Gamma Approximation (default)",
                                 NDefaultThresholdGamma, defHSIC->getThresholdGamma());
    CPPUNIT_ASSERT_MESSAGE("Name of the method (HSIC)",
                           strcmp(defHSIC->GetName(), "HSIC") == 0);
    CPPUNIT_ASSERT_MESSAGE("Name of the method (HSic)",
                           strcmp(defHSIC->getMethodName(), "HSic") == 0);

    /// Default Values for the TSensitivity Object with the Getter Functions
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input getID()",
                                 ID_SENSITIVITY, defHSIC->getID());
    CPPUNIT_ASSERT_MESSAGE("The pointer of the result TTree getResultTuple()",
                           defHSIC->getResultTuple() != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of the result TTree getResultTuple()",
                                 0, (Int_t)(defHSIC->getResultTuple()->GetEntries()));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of branches of the result TTree getResultTuple()",
                                 nBranches, (Int_t)(defHSIC->getResultTuple()->GetNbranches()));
    for (Int_t i=0; i < nBranches; i++) {
        CPPUNIT_ASSERT_MESSAGE(Form("Status of the branch \"%s\" in the result TTree getResultTuple()", sBranches[i].Data()),
                               defHSIC->getResultTuple()->GetBranchStatus(sBranches[i].Data()));
    }
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the evaluatuor function getFunctionName() [%s]", defHSIC->getFunctionName().Data()),
                           strcmp(defHSIC->getFunctionName(), "") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("The Seed value getSeed()[%d]", defHSIC->getSeed()),
                                 0, (Int_t)(defHSIC->getSeed()));
    CPPUNIT_ASSERT_MESSAGE("The noIntermediateSaved flag",
                           defHSIC->getNoIntermediateSaved() == kFALSE);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the Iterator getIteratorName() [%s]", defHSIC->getIteratorName()),
                           strcmp(defHSIC->getIteratorName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the time attribute getTimeName() [%s]", defHSIC->getTimeName().Data()),
                           strcmp(defHSIC->getTimeName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("The draw progress bar flag",
                           defHSIC->getDrawProgressBar());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 0, defHSIC->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 0, defHSIC->getNOutput());

    /// Default Values for the TSensitivity Object WITHOUT Getter Functions
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TDS is NULL (default)",
                                 defHSIC->_tds == NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample via _nS",
                                 0, defHSIC->_nS);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of elements of Y via _nElY",
                                 0, (Int_t)(defHSIC->_nElY));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via _nbOut",
                                 0, defHSIC->_nbOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Counter for output via _iOut",
                                 0, defHSIC->_iOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iy",
                                 0, (Int_t)(defHSIC->_iy));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iely",
                                 0, (Int_t)(defHSIC->_iely));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Type of launcher (default)",
                                 URANIE::Sensitivity::TSensitivity::kUnknown, defHSIC->_nLauncher);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TCode is NULL (default)",
                                 defHSIC->_code == NULL);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TRun is NULL (default)",
                                 defHSIC->_run == NULL);
    CPPUNIT_ASSERT_MESSAGE("The Fact that the input list is provided or not",
                           defHSIC->_bChosenOutputs == kFALSE);

    /// Clean
    delete defHSIC; defHSIC = NULL;
}

void THSICTest::testExceptionTHSICEstimatedVariance()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw ", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Value
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    _tHSIC->setEstimatedVariance(URANIE::Sensitivity::THSIC::kMedianeDelta);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kMedianeDelta)",
                                 URANIE::Sensitivity::THSIC::kMedianeDelta, _tHSIC->getEstimatedVariance());

    _tHSIC->setEstimatedVariance(URANIE::Sensitivity::THSIC::kStdEmpirical);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kStdEmpirical)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    _tHSIC->parseOption("median");
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kMedianeDelta) by parseOption",
                                 URANIE::Sensitivity::THSIC::kMedianeDelta, _tHSIC->getEstimatedVariance());

    _tHSIC->parseOption("kUnknown");
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kMedianeDelta) by parseOption with no changes",
                                 URANIE::Sensitivity::THSIC::kMedianeDelta, _tHSIC->getEstimatedVariance());

    _tHSIC->parseOption("empiri");
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kStdEmpirical) by parseOption",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    _tHSIC->parseOption("media");
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kStdEmpirical) by parseOption with no changes",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    CPPUNIT_ASSERT_THROW_MESSAGE("Method to compute the variance of the Gaussian kernel (kUnknown) not authorized",
                                 _tHSIC->setEstimatedVariance(URANIE::Sensitivity::THSIC::kUnknown),
                                 URANIE::Exceptions::UErrorExceptions);
}

void THSICTest::testExceptionTHSICNBoostrapSample()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw ", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Value
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Bootstrap permutation (default)",
                                 NDefaultBootstrap, _tHSIC->getNBoostrapSample());

    Int_t nBoot = 100;
    _tHSIC->setNBoostrapSample(nBoot);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of Bootstrap permutation by setNBoostrapSample(%d)", nBoot),
                                 nBoot, _tHSIC->getNBoostrapSample());
    nBoot = 0;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation by setNBoostrapSample(%d)", nBoot),
                                 _tHSIC->setNBoostrapSample(nBoot),
                                 URANIE::Exceptions::UErrorExceptions);
    nBoot = -1000;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation by setNBoostrapSample(%d)", nBoot),
                                 _tHSIC->setNBoostrapSample(nBoot),
                                 URANIE::Exceptions::UErrorExceptions);
}

void THSICTest::testExceptionTHSICThresholdBootstrap()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Values
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Bootstrap (default)",
                                 NDefaultThresholdBootstrap, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Gamma Approximation (default)",
                                 NDefaultThresholdGamma, _tHSIC->getThresholdGamma());

    Int_t nDefault = NDefaultThresholdGamma;

    Int_t n = nDefault - 1;
    _tHSIC->setThresholdBootstrap(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdBootstrap(%d)", n),
                                 n, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdBootstrap(%d)", n),
                                 nDefault, _tHSIC->getThresholdGamma());
    n = nDefault + 1;
    _tHSIC->setThresholdBootstrap(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdBootstrap(%d)", n),
                                 n, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdBootstrap(%d)", n),
                                 n, _tHSIC->getThresholdGamma());

    n = NDefaultThresholdBootstrap -1;
    nDefault = _tHSIC->getThresholdGamma();
    _tHSIC->setThresholdBootstrap(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdBootstrap(%d)", n),
                                 n, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdBootstrap(%d)", n),
                                 nDefault, _tHSIC->getThresholdGamma());

    n = 0;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation after setThresholdBootstrap(%d)", n),
                                 _tHSIC->setThresholdBootstrap(n),
                                 URANIE::Exceptions::UErrorExceptions);

    n = -100;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation after setThresholdBootstrap(%d)", n),
                                 _tHSIC->setThresholdBootstrap(n),
                                 URANIE::Exceptions::UErrorExceptions);
}

void THSICTest::testExceptionTHSICThresholdGamma()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Values
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Bootstrap (default)",
                                 NDefaultThresholdBootstrap, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Threshold for the Gamma Approximation (default)",
                                 NDefaultThresholdGamma, _tHSIC->getThresholdGamma());

    Int_t nDefault = NDefaultThresholdBootstrap;

    Int_t n = nDefault + 1;
    _tHSIC->setThresholdGamma(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdGamma(%d)", n),
                                 nDefault, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdGamma(%d)", n),
                                 n, _tHSIC->getThresholdGamma());
    n = nDefault - 1;
    _tHSIC->setThresholdGamma(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdGamma(%d)", n),
                                 n, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdGamma(%d)", n),
                                 n, _tHSIC->getThresholdGamma());

    n = NDefaultThresholdGamma + 1;
    nDefault = _tHSIC->getThresholdBootstrap();
    _tHSIC->setThresholdGamma(n);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdGamma(%d)", n),
                                 nDefault, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdGamma(%d)", n),
                                 n, _tHSIC->getThresholdGamma());

    n = 0;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation after setThresholdGamma(%d)", n),
                                 _tHSIC->setThresholdGamma(n),
                                 URANIE::Exceptions::UErrorExceptions);

    n = -100;
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Number of Bootstrap permutation after setThresholdGamma(%d)", n),
                                 _tHSIC->setThresholdGamma(n),
                                 URANIE::Exceptions::UErrorExceptions);
}



void THSICTest::testExceptionTHSICMatH()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Values
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Rows of the _matH matrix (default)",
                                 0, _tHSIC->_matH.GetNrows());

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Rows of the _matH matrix (default)",
                                 0, _tHSIC->_matH.GetNcols());

    _tHSIC->generateHMatrix();

    Int_t nS = _tdsA->getNPatterns();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Rows of the _matH matrix (default)",
                                 nS, _tHSIC->_matH.GetNrows());

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Rows of the _matH matrix (default)",
                                 nS, _tHSIC->_matH.GetNcols());

    Double_t dmatHII = 1.0;
    Double_t dmatHIJ =  -1.0 / (nS);
    dmatHII += dmatHIJ;

    for(Int_t i=0; i<nS; i++) {
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value of the diagonal element H(%d,%d) = %e", i, i, dmatHII),
                                             dmatHII, _tHSIC->_matH(i,i), DBL_PRECISION_COMPARE);
        for(Int_t j=0; j<i; j++) {
            CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value of the triangular element H(%d,%d) = %e", i, j, dmatHIJ),
                                                 dmatHIJ, _tHSIC->_matH(i,j), DBL_PRECISION_COMPARE);
        }
    }
}

void THSICTest::testExceptionTHSICEstimateSigma()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");
    _tHSIC->unsetLog();

    /// Default Value
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    TString satt = "_x1_";
    Double_t dval = -1.0;
    Double_t dSigma = -1.0;

    /// Case of kStdEmpirical
    satt = "rw"; dval = 0.0312212706649727;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsA->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    satt = "tl"; dval = 20.0308780990588;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsA->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    satt = "hu"; dval = 39.9077277664847;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsA->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    satt = "kw"; dval = 728.316030057495;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsA->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    /// Case of kMedianeDelta
    _tHSIC->setEstimatedVariance(URANIE::Sensitivity::THSIC::kMedianeDelta);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kMedianeDelta)",
                                 URANIE::Sensitivity::THSIC::kMedianeDelta, _tHSIC->getEstimatedVariance());

    satt = "rw"; dval = 0.0333;
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, SPL_PRECISION_COMPARE);

    satt = "tl"; dval = 17.64;
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, SPL_PRECISION_COMPARE);

    satt = "hu"; dval = 34.29;
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, SPL_PRECISION_COMPARE);

    satt = "kw"; dval = 625.72;
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, SPL_PRECISION_COMPARE);

    /// Clean
    delete _tHSIC;
    _tHSIC = NULL;

    /// Case of constant values
    _tdsD->addAttribute("X01", 0.0, 1.0);
    _tdsD->addAttribute("X02", 0.0, 1.0);
    _tdsD->addAttribute("X03", 0.0, 1.0);
    _tdsD->addAttribute("Y", 0.0, 1.0);

    _tdsD->createTuple();

    Int_t nS = 100;
    for(Int_t i=0; i<nS; i++) {
        _tdsD->getTuple()->Fill(1.0*nS, 0.010 * i, 8.7500000e-05 * i * i, 0.8562, 9.7500000e-05 * i * (i + 2.5));
    }

//    _tdsD->getTuple()->SetScanField(-1);
//    _tdsD->getTuple()->Scan("*");

    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsD, "X01:X02:X03", "Y");
    _tHSIC->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    satt = "X01"; dval = 0.29011491975882;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsD->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    satt = "X02"; dval = 0.259715295343606;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsD->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);

    satt = "Y"; dval = 0.296245163607674;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsD->isAttribute(satt));
    dSigma = _tHSIC->estimateSigma(satt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                         dval, dSigma, DBL_PRECISION_COMPARE);
    satt = "X03"; dval = 0.296245163607674;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                           _tdsD->isAttribute(satt));

    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Parameter Sigma (kStdEmpirical) for attribute %s < %e", satt.Data(), dEPSILONSigma),
                                 _tHSIC->estimateSigma(satt),
                                 URANIE::Exceptions::UErrorExceptions);

    /// Case of kMedianeDelta
    _tHSIC->setEstimatedVariance(URANIE::Sensitivity::THSIC::kMedianeDelta);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (kMedianeDelta)",
                                 URANIE::Sensitivity::THSIC::kMedianeDelta, _tHSIC->getEstimatedVariance());

    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Parameter Sigma (kMedianeDelta) for attribute %s < %e", satt.Data(), dEPSILONSigma),
                                 _tHSIC->estimateSigma(satt),
                                 URANIE::Exceptions::UErrorExceptions);
}

void THSICTest::testExceptionTHSICComputeIndexes()
{
    Int_t nS = 300;
    _tdsD->fileDataRead(Form("Ishigami_%05d.dat", nS));

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 nS, _tdsD->getNPatterns());

    TString sInputs  ="X01:X02:X03";
    const Int_t nX = 3;

    TString sOutputs ="Y";
    const Int_t nY = 1;


    TString sAtts = Form("%s:%s", sInputs.Data(), sOutputs.Data());
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;
    Double_t dval = -1.0;
    Double_t dSigma;
    cout << endl;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
   }

    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsD, sInputs, sOutputs);
    _tHSIC->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 nX, _tHSIC->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 nY, _tHSIC->getNOutput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Method to compute the variance of the Gaussian kernel (default)",
                                 URANIE::Sensitivity::THSIC::kStdEmpirical, _tHSIC->getEstimatedVariance());

    const Int_t nAtts = nX + nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Attributes  via nAtts", 4, nAtts);
    Double_t dvalSigma[nAtts] = {1.81091840847876, 1.71376550090841, 1.80895247465531, 3.6251921838347};

    inStream.clear();
    inStream.seekg(0);
    Int_t iAtts = 0;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        dSigma = _tHSIC->estimateSigma(satt);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Parameter Sigma (kStdEmpirical) for attribute %s = %e", satt.Data(), dSigma),
                                             dvalSigma[iAtts++], dSigma, DBL_PRECISION_COMPARE);
    }

    _tHSIC->computeIndexes();

    // TTree of results
    TTree *tt = _tHSIC->getResultTuple();

    Int_t nResu =  3 * nX * nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    Double_t dvalR2HSIC[nX] = {0.17373096, 0.01221493, 0.03397067};
    Double_t dvalHSIC[nX] = {0.015046791, 0.001051322, 0.002924192};
    Double_t dvalPValues[nX] = {0.0000000000, 0.1357520523, 0.0006788981};
    inStream.clear();
    inStream.seekg(0);
    iAtts = 0;
    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i+=3, iAtts++) {
        inStream >> inName;
        tt->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of R2HSic for [%s] i(%d) %e", inName.c_str(), iAtts, dvalR2HSIC[iAtts]),
                                              dvalR2HSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of HSIC for [%s] i(%d) %e", inName.c_str(), iAtts, dvalHSIC[iAtts]),
                                              dvalHSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of PValues for [%s] i(%d) %e", inName.c_str(), iAtts, dvalPValues[iAtts]),
                                              dvalPValues[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
    }

    _tHSIC->printSummary();

    TString sFunctionName = Form("Ishigami dataset from file \"Ishigami_%05d.dat\"", nS);
    TString sSelect="My select not empty";
    TString sOptionHSIC="log,legend";

    _tHSIC->drawIndexes(sFunctionName, "", sOptionHSIC);

    _tHSIC->drawIndexes(sFunctionName, sSelect);

    sOptionHSIC="pvalues=0.05";
    _tHSIC->drawIndexes(sFunctionName, "", sOptionHSIC);

    sOptionHSIC="pvalues=-0.5";
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Case of the drawIndexes method with the \"select\" argument not empty (%s)", sSelect.Data()),
                                 _tHSIC->drawIndexes(sFunctionName, "", sOptionHSIC),
                                 URANIE::Exceptions::UErrorExceptions);

    sOptionHSIC="pvalues=1.5";
    _tHSIC->drawIndexes(sFunctionName, "", sOptionHSIC);
}

void THSICTest::testExceptionTHSICComputeIndexesUnbiased()
{
    Int_t nS = 300;
    _tdsD->fileDataRead(Form("Ishigami_%05d.dat", nS));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 nS, _tdsD->getNPatterns());

    TString sInputs  ="X01:X02:X03";
    const Int_t nX = 3;

    TString sOutputs ="Y";
    const Int_t nY = 1;

    TString sAtts = Form("%s:%s", sInputs.Data(), sOutputs.Data());
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;
    Double_t dval = -1.0;
    Double_t dSigma;
    cout << endl;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
    }

    /// Compute with the unbiased formula
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsD, sInputs, sOutputs);
    _tHSIC->unsetLog();

    _tHSIC->computeIndexes("unbiased");

    // TTree of results
    TTree *tt = _tHSIC->getResultTuple();

    Int_t nResu =  3 * nX * nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    _tHSIC->printSummary();

    inStream.clear();
    inStream.seekg(0);
    Int_t iAtts = 0;

    Double_t dvalR2HSIC[nX] = {0.17373096, 0.01221493, 0.03397067};
    Double_t dvalR2HSICUnbiased[nX] = { 0.1682915, 0.005071462503, 0.0260207541338};
    Double_t dvalHSIC[nX] = {0.015046791, 0.001051322, 0.002924192};
    Double_t dvalHSICUnbiased[nX] = {0.01457568358, 0.00043649369, 0.00223986405};
    Double_t dvalPValues[nX] = {0.0000000000, 0.1357520523, 0.0006788981};

    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i+=3, iAtts++) {
        inStream >> inName;
        tt->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of R2HSic unbiased for [%s] i(%d) %e", inName.c_str(), iAtts, dvalR2HSICUnbiased[iAtts]),
                                              dvalR2HSICUnbiased[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of HSIC unbiased for [%s] i(%d) %e", inName.c_str(), iAtts, dvalHSICUnbiased[iAtts]),
                                              dvalHSICUnbiased[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of PValues for [%s] i(%d) %e", inName.c_str(), iAtts, dvalPValues[iAtts]),
                                              dvalPValues[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
    }


}


void THSICTest::testExceptionTHSICComputeIndexesBootStrap()
{
    Int_t nS = 300;
    _tdsD->fileDataRead(Form("Ishigami_%05d.dat", nS));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 nS, _tdsD->getNPatterns());

    TString sInputs  ="X01:X02:X03";
    const Int_t nX = 3;

    TString sOutputs ="Y";
    const Int_t nY = 1;

    TString sAtts = Form("%s:%s", sInputs.Data(), sOutputs.Data());
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;
    Double_t dval = -1.0;
    Double_t dSigma;
    cout << endl;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
    }

    /// Compute with the unbiased formula
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsD, sInputs, sOutputs);
    _tHSIC->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample via _nS",
                                 nS, _tHSIC->_nS);

    _tHSIC->setThresholdBootstrap(nS+1);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdBootstrap(%d)", nS+1),
                                 nS+1, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdBootstrap(%d)", nS+1),
                                 nS+1, _tHSIC->getThresholdGamma());

    TString sFunctionName = Form("Ishigami dataset from file \"Ishigami_%05d.dat\"", nS);
    CPPUNIT_ASSERT_THROW_MESSAGE("Case of the drawIndexes method with the empty TTree result",
                                 _tHSIC->drawIndexes(sFunctionName),
                                 URANIE::Exceptions::UErrorExceptions);

    gRandom->SetSeed(0);
    Int_t nBoot = 200;

    _tHSIC->computeIndexes(Form("nboot=%d", nBoot));

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of Bootstrap permutation by computeIndexes(nboot=%d)", nBoot),
                                 nBoot, _tHSIC->getNBoostrapSample());

    // TTree of results
    TTree *tt = _tHSIC->getResultTuple();

    Int_t nResu =  3 * nX * nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    _tHSIC->printSummary();

    inStream.clear();
    inStream.seekg(0);
    Int_t iAtts = 0;
    Double_t dvalR2HSIC[nX] = {0.17373096, 0.01221493, 0.03397067};
    Double_t dvalHSIC[nX] = {0.015046791, 0.001051322, 0.002924192};
    Double_t dvalPValues[nX] = {0.0000000000, 0.1357520523, 0.0006788981};

    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i+=3, iAtts++) {
        inStream >> inName;
        tt->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of R2HSic for [%s] i(%d) %e", inName.c_str(), iAtts, dvalR2HSIC[iAtts]),
                                              dvalR2HSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of HSIC unbiased for [%s] i(%d) %e", inName.c_str(), iAtts, dvalHSIC[iAtts]),
                                              dvalHSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of PValues for [%s] i(%d) %e", inName.c_str(), iAtts, dvalPValues[iAtts]),
                                              dvalPValues[iAtts], tlSobol->GetValue(), 0.07);
    }
}



void THSICTest::testExceptionTHSICComputeIndexesSeveralOutputs()
{
    Int_t nS = 300;
    _tdsD->fileDataRead(Form("Ishigami_%05d.dat", nS));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 nS, _tdsD->getNPatterns());

    TString sInputs  ="X01:X02:X03";
    const Int_t nX = 3;

    _tdsD->addAttribute("Y1", "2.0*Y");
    _tdsD->addAttribute("Y2", "-5.0*Y");

    TString sOutputs ="Y:Y1:Y2";
    const Int_t nY = 3;

    TString sAtts = Form("%s:%s", sInputs.Data(), sOutputs.Data());
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;
    Double_t dval = -1.0;
    Double_t dSigma;
    cout << endl;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
    }

    /// Compute with the unbiased formula
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsD, sInputs, sOutputs);
    _tHSIC->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample via _nS",
                                 nS, _tHSIC->_nS);


    _tHSIC->setThresholdBootstrap(nS+1);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Bootstrap after setThresholdBootstrap(%d)", nS+1),
                                 nS+1, _tHSIC->getThresholdBootstrap());
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Threshold for the Gamma Approximation after setThresholdBootstrap(%d)", nS+1),
                                 nS+1, _tHSIC->getThresholdGamma());

    TString sFunctionName = Form("Ishigami dataset from file \"Ishigami_%05d.dat\"", nS);
    CPPUNIT_ASSERT_THROW_MESSAGE("Case of the drawIndexes method with the empty TTree result",
                                 _tHSIC->drawIndexes(sFunctionName),
                                 URANIE::Exceptions::UErrorExceptions);

    gRandom->SetSeed(0);
    Int_t nBoot = 200;

    _tHSIC->computeIndexes(Form("nboot=%d", nBoot));

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of Bootstrap permutation by computeIndexes(nboot=%d)", nBoot),
                                 nBoot, _tHSIC->getNBoostrapSample());

    // TTree of results
    TTree *tt = _tHSIC->getResultTuple();

    Int_t nResu =  3 * nX * nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    _tHSIC->printSummary();

    inStream.clear();
    inStream.seekg(0);
    Int_t iAtts = 0;
    Double_t dvalR2HSIC[nX] = {0.17373096, 0.01221493, 0.03397067};
    Double_t dvalHSIC[nX] = {0.015046791, 0.001051322, 0.002924192};
    Double_t dvalPValues[nX] = {0.0000000000, 0.1357520523, 0.0006788981};

    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i+=3, iAtts++) {
        inStream >> inName;
        tt->GetEntry(i);
        if ( iAtts == 3 ) iAtts = 0;
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of R2HSic for [%s] i(%d) %e", inName.c_str(), iAtts, dvalR2HSIC[iAtts]),
                                              dvalR2HSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of HSIC unbiased for [%s] i(%d) %e", inName.c_str(), iAtts, dvalHSIC[iAtts]),
                                              dvalHSIC[iAtts], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
        tt->GetEntry(i+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of PValues for [%s] i(%d) %e", inName.c_str(), iAtts, dvalPValues[iAtts]),
                                              dvalPValues[iAtts], tlSobol->GetValue(), 0.07);
    }
}

////////////////////////////////
///                          ///
///      Exceptions          ///
///                          ///
////////////////////////////////

void THSICTest::throwExceptionTHSICConstructor()
{
    URANIE::DataServer::TDataServer *tdsnull = new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(tdsnull, "x_{1}:x_{2}:x_{3}", "Ishigami", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(NULL, "x_{1}:x_{2}:x_{3}", "Ishigami", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);
    /// Clean
    delete tdsnull;

    /// False input
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(_tdsA, "rw:hu:TheUnKnownAtt", "Ishigami", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);

    /// False output
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(_tdsA, "rw:hu:hl:tu:tl", "TheUnKnownAtt", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);
    /// empty output
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(_tdsA, "rw:hu:hl:tu:tl", "", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::THSIC(_tdsA, "rw:hu:hl:tu:tl", ":", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);

    /// case of NPatterns = 4 < 6 = NMinimumPatterns
    _tdsA->setSelect("rw < 0.060");
    /// Not Enough patterns
    CPPUNIT_ASSERT_THROW_MESSAGE(Form("Case of Lesser points [%d] than the minimum required [%d]", _tdsA->getNPatterns(), NMinimumPatterns),
                                 new URANIE::Sensitivity::THSIC(_tdsA, "rw:hu:hl:tu:tl", "ystar", "myOpt"),
                                 URANIE::Exceptions::UErrorExceptions);
}

void THSICTest::throwExceptionsGenerateSample()
{
    _tHSIC = new URANIE::Sensitivity::THSIC(_tdsA, "rw:r:tu:tl:hu:hl:l:kw", "ystar", "myOpt");
    _tHSIC->unsetLog();

    CPPUNIT_ASSERT_THROW_MESSAGE("Case of the generateSample method",
                                 _tHSIC->generateSample("unKnown"),
                                 URANIE::Exceptions::UWarningExceptions);
}

// End OF File
