//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef CPPUNITEST_THSIC_H
#define CPPUNITEST_THSIC_H

#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>
#include <cppunit/TestFailure.h>
#include <cppunit/TestResultCollector.h>

// Uranie
#include "TSensitivity.h"
#include "THSIC.h"
#include "TDataServer.h"
#include "UExceptions.h"

class THSICTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE (THSICTest);

    CPPUNIT_TEST (testconstructorHSIC);

    CPPUNIT_TEST (testExceptionTHSICEstimatedVariance);
    CPPUNIT_TEST (testExceptionTHSICNBoostrapSample);
    CPPUNIT_TEST (testExceptionTHSICThresholdBootstrap);
    CPPUNIT_TEST (testExceptionTHSICThresholdGamma);
    CPPUNIT_TEST (testExceptionTHSICMatH);
    CPPUNIT_TEST (testExceptionTHSICEstimateSigma);
    CPPUNIT_TEST (testExceptionTHSICComputeIndexes);
    CPPUNIT_TEST (testExceptionTHSICComputeIndexesUnbiased);
    CPPUNIT_TEST (testExceptionTHSICComputeIndexesBootStrap);
    CPPUNIT_TEST (testExceptionTHSICComputeIndexesSeveralOutputs);

    CPPUNIT_TEST (throwExceptionTHSICConstructor);
    CPPUNIT_TEST (throwExceptionsGenerateSample);


    CPPUNIT_TEST_SUITE_END();

public:
    //---------------------------------------------
    /** @name Constructor and Destructor for the CPPUnit tests
     **/
    //@{
    //! Default constructor
    THSICTest();

    //! Default destructor
    virtual ~THSICTest();

    //! the setUp method
    void setUp();
    //! the tearDown method
    void tearDown();
    //@}

    //---------------------------------------------
    /** @name Validation methods for the use cases
     **/
    //@{
    //!   Validation on the Ishigami function
    void validateValuesIshigami();
    //! Validation method on the Flowrate function
    void validateValuesFlowrate();
    //@}


    //---------------------------------------------
    /** @name Unit Tests
     The list of unit test
     **/
    //@{
    //! Test the HSIC constyructor
    void testconstructorHSIC();

    //! Tests and Exceptions for the \b _nEstimatedVariance parameter
    void testExceptionTHSICEstimatedVariance();

    //! Tests and Exceptions for the \b _nBootstrapSample parameter
    void testExceptionTHSICNBoostrapSample();

    //! Tests and Exceptions for the \b _nThresholdBootstrap parameter
    void testExceptionTHSICThresholdBootstrap();

    //! Tests and Exceptions for the \b _nThresholdGamma parameter
    void testExceptionTHSICThresholdGamma();

    //! Tests and Exceptions for the \b _matH parameter
    void testExceptionTHSICMatH();

    //! Tests and Exceptions for the \b estimateSigma method
    void testExceptionTHSICEstimateSigma();

    /*!
     * \brief Tests and Exceptions for the \b computeIndexes method
     *
     * From the R macro "HSIC.Ré with the Ishigami database of 300 patterns stored in
     *  the Ishigami_00300.dat
     *
     * > GSA.HSIC_V1
     * $HSIC
     *         X1          X2          X3
     * 0.015046791 0.001051322 0.002924192
     * $HSIC.norm
     *         X1         X2         X3
     * 0.79100777 0.05526786 0.15372437
     * $HSIC.R2
     *         X1         X2         X3
     * 0.17373096 0.01221493 0.03397067
     * $HSIC.R2.norm
     *         X1         X2         X3
     * 0.78998580 0.05554349 0.15447071
     *
     * $pval
     * [1] 0.0000000000 0.1357520523 0.0006788981
     *
     * And the Uranie implementation
     *
     * 5: ************************************************************************************************************
     * 5: *    Row   *       Out *       Inp *     Order *    Method * Algo.Algo *     Sobol *   CILower *   CIUpper *
     * 5: ************************************************************************************************************
     * 5: *        0 *         Y *       X01 *    R2HSic *      HSic * --first-- * 0.1737309 *        -1 *        -1 *
     * 5: *        1 *         Y *       X01 *      HSic *      HSic *  --hsic-- * 0.0144300 * 0.0131964 * 0.0168970 *
     * 5: *        2 *         Y *       X01 *   pValues *      HSic * --pvalues * 2.489e-22 *        -1 *        -1 *
     * 5: *        3 *         Y *       X02 *    R2HSic *      HSic * --first-- * 0.0122149 *        -1 *        -1 *
     * 5: *        4 *         Y *       X02 *      HSic *      HSic *  --hsic-- * 0.0004352 * -0.000796 * 0.0028994 *
     * 5: *        5 *         Y *       X02 *   pValues *      HSic * --pvalues * 0.1357520 *        -1 *        -1 *
     * 5: *        6 *         Y *       X03 *    R2HSic *      HSic * --first-- * 0.0339706 *        -1 *        -1 *
     * 5: *        7 *         Y *       X03 *      HSic *      HSic *  --hsic-- * 0.0023079 * 0.0010755 * 0.0047728 *
     * 5: *        8 *         Y *       X03 *   pValues *      HSic * --pvalues * 0.0006788 *        -1 *        -1 *
     * 5: ************************************************************************************************************
     *
     */
    void testExceptionTHSICComputeIndexes();
    void testExceptionTHSICComputeIndexesUnbiased();
    void testExceptionTHSICComputeIndexesBootStrap();

    /*!
      \brief Case of several outputs
     */
    void testExceptionTHSICComputeIndexesSeveralOutputs();
    //@}

    //---------------------------------------------
    /** @name Throw Exceptions
     The list of Throw Exceptions
     **/
    //@{
    //! Throw Exceptions for the constructor
    void throwExceptionTHSICConstructor();

    //! Throw exceptions for the generateSample method
    void throwExceptionsGenerateSample();
    //@}

private:
    URANIE::DataServer::TDataServer *_tdsA, *_tdsD;
    URANIE::Sensitivity::THSIC * _tHSIC;
};

#endif  // CPPUNITEST_THSIC_H
