//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTJohnsonRW.cxx
 \author JB Blanchard
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Sensitivity::TJohnsonRW
 
 Cr�e le : jeu d�c 22 12:30:15 CET 2011

 
 */

// Uranie
// #include "TUniformDistribution.h"
#include "TSampling.h"
#include "TLauncher2.h"
#include "TCJitEval.h"
// #include "TOutputFileKey.h"
#include "TOutputFileDataServer.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TestUserFunctions.h"
#include "TestTSensitivity.h"
#include "TestTJohnsonRW.h"
#include "TJohnsonRW.h"
#include "UDisableTest.h"
#include "uranie_commands.h"
#include "TThreadedRun.h"

// ROOT
#include "TString.h"
#include "TCanvas.h"
#include "TTreePlayer.h"
#define EPSILON 0.01 
#define LVLE_EPSILON 0.001
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>
using namespace std;

TJohnsonRWTest::TJohnsonRWTest() :
        tds1(NULL), tds(NULL), tjrw(NULL), tss1(NULL)
{

}

TJohnsonRWTest::~TJohnsonRWTest()
{
    // UranieTests::removeFile(gSystem->pwd(),"Ishigami.root");
    UranieTests::removeFile(gSystem->pwd(),"1000pointsLevelEDB.root");
    UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign.root");
    // UranieTests::removeFile(gSystem->pwd(), "flowrateUniformDesign2D.root");
}


void TJohnsonRWTest::setUp()
{
    tds1 = new URANIE::DataServer::TDataServer();

    tds = new URANIE::DataServer::TDataServer("tdsflowreate", "DataBase flowreate");
    tds->fileDataRead("flowrateUniformDesign.dat");
}

void TJohnsonRWTest::tearDown()
{
    delete tss1;
    tss1 = NULL;
    
    delete tjrw;
    tjrw = NULL;

    delete tds1;
    tds1 = NULL;

    delete tds;
    tds = NULL;
}

TMatrixD TJohnsonRWTest::setUpIshi()
{

    delete tds1; tds1=NULL;
    
    tds1 = new URANIE::DataServer::TDataServer();
    tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("x_{1}", -1.0 * TMath::Pi(), TMath::Pi()));
    tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("x_{2}", -1.0 * TMath::Pi(), TMath::Pi()));
    tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("x_{3}", -1.0 * TMath::Pi(), TMath::Pi()));
    tds1->getAttribute("x_{1}")->delShare();
    tds1->getAttribute("x_{2}")->delShare();
    tds1->getAttribute("x_{3}")->delShare();

    TMatrixD Corr(3,3);
    Corr(0,0)=Corr(1,1)=Corr(2,2)=1.0;
    Corr(1,0)=Corr(0,1)=0.95;
    Corr(2,0)=Corr(0,2)=0.85;
    Corr(1,2)=Corr(2,1)=0.90;

    return Corr;    
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TJohnsonRWTest::testJohnsonRWFromPaper()
{
    // Load the data base in the DataServer
    int _nX=3;
    for(int i=0; i<_nX; i++)
        tds1->addAttribute(Form("x%d",i+1),0.,1.);

    tds1->addAttribute("y",0,1);
    tds1->getAttribute("y")->setOutput();
    
    TMatrixD Corr(_nX+1,_nX+1);
    Corr(0,0)=Corr(1,1)=Corr(2,2)=Corr(3,3)=1.0;
    Corr(1,0)=Corr(0,1)=0.95;
    Corr(2,0)=Corr(0,2)=0.85;
    Corr(1,2)=Corr(2,1)=0.90;
    Corr(3,0)=Corr(0,3)=0.50;    Corr(3,1)=Corr(1,3)=0.40;    Corr(3,2)=Corr(2,3)=0.30;
    
    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "x1:x2:x3", "y");
    tss1->setCorrelationMatrix( Corr );
    tss1->computeIndexes();

    Double_t sobol;
    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);
    
    Double_t dsum = 0.0, jsum=0.0;

    Double_t Comparison[4] = {0.5540898, 0.2870087, 0.1589014, 0.3229729};
    for (Int_t i = 0; i < 4; i++)
    {
        tree->GetEntry(2 * i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i], sobol, 0.001);
        if(i<3)
        {
            dsum += Comparison[i];
            jsum += sobol;
        }
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, jsum, 0.001);
}

void TJohnsonRWTest::testJohnsonRWFromPaperTwoOutputs()
{
    // Load the data base in the DataServer
    int _nX=3;
    for(int i=0; i<_nX; i++)
        tds1->addAttribute(Form("x%d",i+1),0.,1.);

    tds1->addAttribute("y",0,1);
    tds1->addAttribute("z",0,1);

    TMatrixD Corr(_nX+2,_nX+2);
    Corr(0,0)=Corr(1,1)=Corr(2,2)=Corr(3,3)=Corr(4,4)=1.0;
    Corr(1,0)=Corr(0,1)=0.95;
    Corr(2,0)=Corr(0,2)=0.85;
    Corr(1,2)=Corr(2,1)=0.90;
    // Output1
    Corr(3,0)=Corr(0,3)=0.50; Corr(3,1)=Corr(1,3)=0.40; Corr(3,2)=Corr(2,3)=0.30; 
    // Output2
    Corr(4,0)=Corr(0,4)=0.45; Corr(4,1)=Corr(1,4)=0.30; Corr(4,2)=Corr(2,4)=0.30; Corr(4,3)=Corr(3,4)=1.;
    
    
    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "x1:x2:x3", "y:z");
    tss1->setCorrelationMatrix( Corr );
    tss1->computeIndexes();

    Double_t sobol;
    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);
    
    Double_t dsum = 0.0, jsum=0.0;

    Double_t Comparison[8] = {0.5540898, 0.2870087, 0.1589014, 0.3229729, 0.5428918, 0.2746127, 0.1824953, 0.3770270};
    for (Int_t ev = 0; ev < 2; ev++)
    {
	dsum=0; jsum=0;
	for (Int_t i = 0; i < 4; i++)
	{
	    tree->GetEntry(2*i+ev*2*(_nX+1));
	    CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i+4*ev], sobol, 0.001);
	    if(i<3)
	    {
		dsum += Comparison[i];
		jsum += sobol;
	    }
	}
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, jsum, 0.001);
}

void TJohnsonRWTest::testJohnsonRWFromPaperTwoOutputsAsVector()
{
    // Load the data base in the DataServer
    int _nX=3;
    for(int i=0; i<_nX; i++)
        tds1->addAttribute(Form("x%d",i+1),0.,1.);

    tds1->addAttribute(new URANIE::DataServer::TAttribute("y", URANIE::DataServer::TAttribute::kVector ) );

    TMatrixD Corr(_nX+2,_nX+2);
    Corr(0,0)=Corr(1,1)=Corr(2,2)=Corr(3,3)=Corr(4,4)=1.0;
    Corr(1,0)=Corr(0,1)=0.95;
    Corr(2,0)=Corr(0,2)=0.85;
    Corr(1,2)=Corr(2,1)=0.90;
    // Output1
    Corr(3,0)=Corr(0,3)=0.50; Corr(3,1)=Corr(1,3)=0.40; Corr(3,2)=Corr(2,3)=0.30; 
    // Output2
    Corr(4,0)=Corr(0,4)=0.45; Corr(4,1)=Corr(1,4)=0.30; Corr(4,2)=Corr(2,4)=0.30; Corr(4,3)=Corr(3,4)=1.;
    
    
    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "x1:x2:x3", "y");
    tss1->setCorrelationMatrix( Corr );
    tss1->computeIndexes();

    Double_t sobol;
    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);
    
    Double_t dsum = 0.0, jsum=0.0;

    Double_t Comparison[8] = {0.5540898, 0.2870087, 0.1589014, 0.3229729, 0.5428918, 0.2746127, 0.1824953, 0.3770270};
    for (Int_t ev = 0; ev < 2; ev++)
    {
	dsum=0; jsum=0;
	for (Int_t i = 0; i < 4; i++)
	{
	    tree->GetEntry(2*i+ev*2*(_nX+1));
	    CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i+4*ev], sobol, 0.001);
	    if(i<3)
	    {
		dsum += Comparison[i];
		jsum += sobol;
	    }
	}
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, jsum, 0.001);
}


void TJohnsonRWTest::testJohnsonRWFromDatabase()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("flowrateUniformDesign.dat");

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "rw:r:tu:tl:hu:hl:l:kw", "ystar");
    tss1->computeIndexes();
    tss1->printLog();
    
    Double_t sobol;
    TTree * tree = tss1->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);

    Double_t dsum = 0.0, jsum=0.0;

    Double_t Comparison[10] =
        {0.8105524, 0.0087153, 0.0084060, 0.0107096, 0.0421202, 0.0402158, 0.0603393, 0.0189410, 0.96265, 0.9502};
    for (Int_t i = 0; i < 10; i++)
    {
        tree->GetEntry(2 * i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(Comparison[i], sobol, 0.001);
        if(i<8)
        {
            dsum += Comparison[i];
            jsum += sobol;
        }
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dsum, jsum, 0.001);
}

#ifdef WITH_LEVELE
// ===========================================
//               LevelE
// ===========================================
void TJohnsonRWTest::checkLevelEResults(TTree *res, vector<int>* elmts)
{

    double ref[26]={0.138347,0.153089,0.124661,0.0802056,0.0689665,0.0724155,0.0721911,0.0706976,0.0693986,0.0543585,
		    0.0358531,0.0452664,0.0721091,0.0929692,0.10428,0.112806,0.119981,0.125616,0.135217,0.11699,
		    0.136991,0.142465,0.158608,0.194051,0.191359,0.17403};

  
    int npts = res->Draw("Sobol","Inp==\"__R2__\" && Order==\"Total\"","goff");
    double *data=res->GetV1();
//  cout<<"ref[26]={";
    for(int ival=0; ival<npts; ival++)
	CPPUNIT_ASSERT_DOUBLES_EQUAL(ref[ ((elmts==NULL) ? ival : elmts->at(ival)) ], data[ival], LVLE_EPSILON);
//      cout<<data[ival]<<((ival==(npts-1))?"};":",");
    // 
    //cout<<ref[ ((elmts==NULL) ? ival : elmts->at(ival)) ]<<" ?=? "<< data[ival]<<endl;
    //cout<<endl;
			 
}

void TJohnsonRWTest::testLevelEJohnsonRW()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y","SRCSRRC");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res);
}


void TJohnsonRWTest::testLevelEJohnsonRWOutputRestrVarexp()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]","SRCSRRC");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}

void TJohnsonRWTest::testLevelEJohnsonRWOutputRestrOption()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y","SRCSRRC,output=y[8]:y[15]:y[22]");
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}

void TJohnsonRWTest::testLevelEJohnsonRWOutputRestrBoth()
{
    // Load the data base in the DataServer
    tds1->fileDataRead("1000pointsLevelEDB.dat");

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]","SRCSRRC");
    tss1->computeIndexes("output=y[8]:y[15]:y[22]");

    TTree *res=tss1->getResultTuple();
    checkLevelEResults(res, tss1->getAttributeElements(string("y")));
}
#endif

void TJohnsonRWTest::checkIshigamiResults(TTree *res)
{
    double ref[5]={0.3031968, 0.2865275, 0.4102756, 0.2840843, 0.2819301};
  
    int npts = res->Draw("Sobol","Order==\"Total\"","goff");
    double *data=res->GetV1();
    for(int ival=0; ival<npts; ival++)
	CPPUNIT_ASSERT_DOUBLES_EQUAL( ref[ival], data[ival], LVLE_EPSILON);
}

void TJohnsonRWTest::testJohnsonRWIshigamiFunctionLauncher()
{

    TMatrixD Corr=setUpIshi();

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, &Ishigami, "x_{1}:x_{2}:x_{3}","Ishi",1000);
    tss1->setSeed(123456);
    tss1->setInputCorrelationMatrix(Corr);
    tss1->setDrawProgressBar(false);
    tss1->computeIndexes();

    TTree *res=tss1->getResultTuple();
    checkIshigamiResults(res);
    
}

void TJohnsonRWTest::testJohnsonRWIshigamiFunctionRelauncher()
{

    TMatrixD Corr=setUpIshi();

    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("Ishi");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(tds1->getAttribute("x_{1}"));
    eval.addInput(tds1->getAttribute("x_{2}"));
    eval.addInput(tds1->getAttribute("x_{3}"));
    eval.addOutput(out);
 	
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
	tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, &run,1000);
	tss1->setSeed(123456);
	tss1->setInputCorrelationMatrix(Corr);
	tss1->setDrawProgressBar(false);
	tss1->computeIndexes();

	TTree *res=tss1->getResultTuple();
	checkIshigamiResults(res);

	run.stopSlave();		
    }
}

void TJohnsonRWTest::testJohnsonRWIshigamiFunctionThreadedandNot()
{

    TMatrixD Corr=setUpIshi();

    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("TestIshigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(tds1->getAttribute("x_{1}"));
    eval.addInput(tds1->getAttribute("x_{2}"));
    eval.addInput(tds1->getAttribute("x_{3}"));
    eval.addOutput(out);
 	
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
	tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, &run, 100);
	tss1->setSeed(123456);
	tss1->setInputCorrelationMatrix(Corr);
	tss1->setDrawProgressBar(false);
	tss1->computeIndexes();

	run.stopSlave();		
    }

    TTree *T = tss1->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuS.txt");
    T->Scan("*");
    
    tds1->deleteTuple();
    tds1->delAttribute("TestIshigami");

    URANIE::DataServer::TAttribute *outT = new URANIE::DataServer::TAttribute("TestIshigamiT");
    outT->delShare();

    URANIE::Relauncher::TCJitEval evalT(&Ishigami);
    evalT.addInput(tds1->getAttribute("x_{1}"));
    evalT.addInput(tds1->getAttribute("x_{2}"));
    evalT.addInput(tds1->getAttribute("x_{3}"));
    evalT.addOutput(outT);
 	
    URANIE::Relauncher::TThreadedRun runT(&evalT,4);
    runT.startSlave();
    if( runT.onMaster() )
    {

        delete tss1;
        
	tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, &runT, 100);
	tss1->setSeed(123456);
	tss1->setInputCorrelationMatrix(Corr);
	tss1->setDrawProgressBar(false);
	tss1->computeIndexes();

	runT.stopSlave();		
    }

    T = tss1->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuT.txt");
    T->Scan("*");
    
    Int_t nret = UranieTests::areFilesEqual(TString("outpuT.dat"),
                                            TString("outpuS.dat"));
    
    CPPUNIT_ASSERT_EQUAL(nret, 0);        

}

static void constructor_function_long_output_throw_exception(const string& outputs) {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.keepFinalTuple(false);
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver);
    sampling.generateSample();

    URANIE::Launcher::TLauncherFunction launcher(&dataserver, &flowrateModel, "",
        outputs.data());
    launcher.setDrawProgressBar(false);
    launcher.run();

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TJohnsonRW regression(&dataserver,
        inputs_name.data(), outputs.data(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TJohnsonRWTest::throwExceptionConstructorTooLongOutputFunctionFirst() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TJohnsonRWTest::throwExceptionConstructorTooLongOutputFunctionSecond() {
    const string outputs = "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    constructor_function_long_output_throw_exception(outputs);
}

void TJohnsonRWTest::throwExceptionConstructorTooLongOutputCode() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    string inputs_name;
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(dataserver, input_file);

    URANIE::Launcher::TOutputFileDataServer output_file("_output_flowrate_withRow_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");

    URANIE::Launcher::TCode code(&dataserver, "");
    code.addOutputFile(&output_file);

    URANIE::Launcher::TLauncher launcher(&dataserver, &code);
    launcher.setDrawProgressBar(false);
    launcher.run("nointermediatesaved");

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TJohnsonRW regression(&dataserver,
        inputs_name.data(), code.getOutputVarNames(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TJohnsonRWTest::throwExceptionConstructorTooLongOutputRun() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    for (auto& input: inputs)
        dataserver.addAttribute(&input);

    URANIE::Sampler::TSampling sampling(&dataserver, "lhs", 100);
    sampling.generateSample();

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    TList *list=dataserver.getListOfAttributes();
    for (const auto& attribute: *list)
        eval.addInput(dynamic_cast<URANIE::DataServer::TAttribute*>(attribute));
    eval.addOutput(&output);

    string inputs_name;
    for (const auto& input: *eval.getInputList())
        inputs_name += input->GetName() + string(":");
    inputs_name.pop_back();

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::Relauncher::TLauncher2 launcher(&dataserver, &run);
        launcher.solverLoop();
        run.stopSlave();

        CPPUNIT_ASSERT_EQUAL(1, eval.getOutputList()->GetEntries());
        CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TJohnsonRW regression(&dataserver,
            inputs_name.data(), eval.getOutputList()->First()->GetName(), "SRCSRRC"),
            URANIE::Exceptions::UErrorExceptions);
    }
    delete list;
}

void TJohnsonRWTest::throwExceptionConstructorTooLongOutputFromData() {
    URANIE::DataServer::TDataServer dataserver("dataserver", "");
    dataserver.fileDataRead("flowrateUniformDesign.dat");
    TList *list=dataserver.getListOfAttributes();
    const string output_name = list->Last()->GetName();

    string inputs_name;
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    const string output_name_w_sep = ":" + output_name + ":";
    inputs_name.erase(inputs_name.find(output_name_w_sep), output_name_w_sep.size());

    dataserver.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut", output_name);
    delete list; list=dataserver.getListOfAttributes();
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TJohnsonRW(&dataserver, inputs_name.data(),
        list->Last()->GetName(), "SRCSRRC"),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TJohnsonRWTest::throwExceptionTJohnsonRWConstructorTdsNull()
{

    URANIE::DataServer::TDataServer *tdsnull =
            new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TJohnsonRW(tdsnull,
                    "rw:r:tu:tl:hu:hl:l:kw", "ystar", "SRRC"),
            URANIE::Exceptions::UErrorExceptions);

    delete tdsnull;
}
void TJohnsonRWTest::throwExceptionTJohnsonRWConstructorFalseInput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TJohnsonRW(tds, "rw:r:tu:tl:hu:hl:l:toto",
                    "ystar", "SRRC"), URANIE::Exceptions::UErrorExceptions);
}
void TJohnsonRWTest::throwExceptionTJohnsonRWConstructorFalseOutput()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TJohnsonRW(tds, "rw:r:tu:tl:hu:hl:l:kw",
                    "toto", "SRRC"), URANIE::Exceptions::UErrorExceptions);
}

void TJohnsonRWTest::throwExceptionTJohnsonRWCorrelationMatrix()
{

    // Load the data base in the DataServer
    int _nX=3;
    for(int i=0; i<_nX; i++)
        tds1->addAttribute(Form("x%d",i+1),0.,1.);

    tds1->addAttribute("y",0,1);

    tss1 = new URANIE::Sensitivity::TJohnsonRW(tds1, "x1:x2:x3", "y");
    
    TMatrixD Corr(_nX+1,_nX+1);
    Corr.UnitMatrix();    
    Corr(2,3)=1.30; Corr(3,2)=1.30;
    CPPUNIT_ASSERT_THROW(tss1->setCorrelationMatrix( Corr ), URANIE::Exceptions::UErrorExceptions);

    Corr(2,3)=-1.30; Corr(3,2)=-1.30;
    CPPUNIT_ASSERT_THROW(tss1->setCorrelationMatrix( Corr ), URANIE::Exceptions::UErrorExceptions);

    TMatrixD Corr2(_nX+1,_nX+2);
    Corr2.UnitMatrix();
    CPPUNIT_ASSERT_THROW(tss1->setCorrelationMatrix( Corr2 ), URANIE::Exceptions::UErrorExceptions);    

    TMatrixD Corr3(_nX,_nX+1);
    Corr3.UnitMatrix();
    CPPUNIT_ASSERT_THROW(tss1->setCorrelationMatrix( Corr3 ), URANIE::Exceptions::UErrorExceptions);

    TMatrixD Corr4(_nX+1,_nX+1);
    Corr4.UnitMatrix();
    Corr4(1,2)=0.4; Corr4(2,1)=0.40001;
    CPPUNIT_ASSERT_THROW(tss1->setCorrelationMatrix( Corr4 ), URANIE::Exceptions::UErrorExceptions);    
}

