//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTMorris.cxx
 \author maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Unit Test for the URANIE::Sensitivity::TMorris class
 
 Created : 2011-12-22

 */
#define EPSILON 0.01 
#define LVLE_EPSILON 1.e-12
#define SUBLVLE_EPSILON 1.e-10
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>

// Uranie
#include "TestTSensitivity.h"
#include "TestTMorris.h"
#include "TMorris.h"
#include "UDisableTest.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TExponentialDistribution.h"
#include "TCauchyDistribution.h"
#include "TGumbelMaxDistribution.h"
#include "TWeibullDistribution.h"
#include "TGenParetoDistribution.h"
#include "TGammaDistribution.h"
#include "TInvGammaDistribution.h"
#ifdef WITH_MATHMORE
#include "TStudentDistribution.h"
#endif
#include "TOutputFileRow.h"
#include "TOutputFileDataServer.h"
#include "TCode.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TKeyScript.h"
#include "TKeyResult.h"
#include "TCodeEval.h"
#include "TCJitEval.h"
#include "TThreadedRun.h"
#include "TSequentialRun.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#ifdef WITH_PYTHON
#include "python_commands.h"
#endif
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TTreePlayer.h"
#ifdef ROOT5
#include "Api.h"
#endif
#include "TCanvas.h"

using namespace std;

TMorrisTest::TMorrisTest() :
        _tds1(NULL), _tds(NULL), _tds2(NULL), _tmorris(NULL), _mycode(NULL), 
        _p2fIshigami(NULL), _fout(NULL)
{
#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;

    method = globalscope.GetMethod("Ishigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();
    method = globalscope.GetMethod("badPrototypeFunction", "", &dummy);
    _p2fBadPrototypeFunction = method.PointerToFunc();
#endif
}

TMorrisTest::~TMorrisTest()
{
    UranieTests::removeFolder(gSystem->pwd(),"URANIE");
    UranieTests::removeFile(gSystem->pwd(), "testMorrisIshigami.out");
    UranieTests::removeFile(gSystem->pwd(), "_morris_launching_.root");
    UranieTests::removeFile(gSystem->pwd(), "_morris_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(), "TDS_morris_Ishgami.root");
    UranieTests::removeFile(gSystem->pwd(),"morris_gFunction6_Optimized.root");
    UranieTests::removeFile(gSystem->pwd(),"sensitivityMorrisFunctionFlowrateData.txt");
    UranieTests::removeFile(gSystem->pwd(),"TDS_morris_false.root");
    UranieTests::removeFile(gSystem->pwd(),"MorrisLevelEDatabase.root");

}

void TMorrisTest::setUp()
{
    _tds1 = new URANIE::DataServer::TDataServer();

    _tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowreate");
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tu", 63070.0,
                    115600.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("kw", 9855.0,
                    12045.0));

    _tds2 = new URANIE::DataServer::TDataServer("tds2",
            "TDS for Ishigami function");
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{1}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{2}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{3}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
}

void TMorrisTest::tearDown()
{
    if (_fout)
    {
        delete _fout;
        _fout = NULL;
    }

    if (_mycode != NULL)
    {
        delete _mycode;
        _mycode = NULL;
    }

    delete _tmorris;
    _tmorris = NULL;

    delete _tds1;
    _tds1 = NULL;

    _tds2->getAttribute("x_{1}")->delShare();
    _tds2->getAttribute("x_{2}")->delShare();
    _tds2->getAttribute("x_{3}")->delShare();
     delete _tds2;
    _tds2 = NULL;
    
    _tds->getAttribute("rw")->delShare();
    _tds->getAttribute("r")->delShare();
    _tds->getAttribute("tu")->delShare();
    _tds->getAttribute("tl")->delShare();
    _tds->getAttribute("hu")->delShare();
    _tds->getAttribute("hl")->delShare();
    _tds->getAttribute("l")->delShare();
    _tds->getAttribute("kw")->delShare();
     delete _tds;
    _tds = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TMorrisTest::setUp1()
{

    Int_t nreplique = 3;
    Int_t nlevel = 10;
#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique,
            nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, Ishigami, "", "Ishigami", nreplique,
            nlevel);
#endif
    _tmorris->setDrawProgressBar(kFALSE);
    _tmorris->setSeed(123);
}

void TMorrisTest::setUp2()
{
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_values_rows.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    _fout = new URANIE::Launcher::TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

    _mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);

    Int_t nreplique = 3;
    Int_t nlevel = 10;
    _tmorris = new URANIE::Sensitivity::TMorris(_tds, _mycode, nreplique,
            nlevel);
    _tmorris->setDrawProgressBar(kFALSE);
    _tmorris->setSeed(123);
}

void TMorrisTest::throwExceptionTMorrisConstructorTdsNull()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;

    // Case of Analytical function

#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TMorris(NULL, _p2fIshigami, nreplique, nlevel), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TMorris(NULL, &Ishigami, "", "Ishi", nreplique, nlevel), URANIE::Exceptions::UErrorExceptions);
#endif

    // Case of Code
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

#ifdef WIN32
    _mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -s -r 1>NUL");
#else
    TString sRepertoireReference = gSystem->pwd();
    _mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate -r>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    _mycode->addOutputFile(_fout);

    CPPUNIT_ASSERT_THROW( new  URANIE::Sensitivity::TMorris(NULL, _mycode, nreplique, nlevel),
                          URANIE::Exceptions::UErrorExceptions);

}

#ifdef ROOT5
void TMorrisTest::throwExceptionTMorrisConstructorFalseFunction()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;

    CPPUNIT_ASSERT_THROW( new  URANIE::Sensitivity::TMorris(_tds2, _p2fBadPrototypeFunction, nreplique, nlevel),
                          URANIE::Exceptions::UErrorExceptions);
}
#endif

void TMorrisTest::throwExceptionTMorrisConstructorFalseCode()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST();

    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

    URANIE::Launcher::TCode *code = new URANIE::Launcher::TCode();

    Int_t nreplique = 3;
    Int_t nlevel = 10;
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TMorris(_tds, code, nreplique, nlevel),
            URANIE::Exceptions::UErrorExceptions);

    delete code;
}

void TMorrisTest::throwExceptionTMorrisConstructorFalseInput()
{

    _tds1->fileDataRead("TDS_morris_Ishgami.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TMorris(_tds1, "x_{1}:x_{2}:toto",
                    "Ishigami"), URANIE::Exceptions::UErrorExceptions);
}

void TMorrisTest::throwExceptionTMorrisConstructorFalseOutput()
{
    _tds1->fileDataRead("TDS_morris_Ishgami.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TMorris(_tds1, "x_{1}:x_{2}:x_{3}","toto"),
                URANIE::Exceptions::UErrorExceptions);
}

void  TMorrisTest::throwExceptionTMorrisConstructorFalseNReplique()
{
    Int_t nreplique = -5;
    Int_t nlevel = 10;

    // Case of Analytical function
#ifdef ROOT5 
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel),
		URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel),
		URANIE::Exceptions::UErrorExceptions);
#endif

    // Case of Code
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

#ifdef WIN32
    _mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -s -r 1>NUL");
#else
    TString sRepertoireReference = gSystem->pwd();
    _mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate -r>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    _mycode->addOutputFile(_fout);

    CPPUNIT_ASSERT_THROW( new  URANIE::Sensitivity::TMorris(_tds, _mycode, nreplique, nlevel),
                          URANIE::Exceptions::UErrorExceptions);

}

void TMorrisTest::throwExceptionTMorrisConstructorFalseLevel()
{
    Int_t nreplique = 3;
    Int_t nlevel = 1;

    // Case of Analytical function

#ifdef ROOT5 
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel,
		URANIE::Exceptions::UErrorExceptions));
#else
	CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel),
		URANIE::Exceptions::UErrorExceptions);
#endif

    // Case of Code
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    _fout->addAttribute("yhat");

#ifdef WIN32
    _mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -s -r 1>NUL");
#else
    TString sRepertoireReference = gSystem->pwd();
    _mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate -r>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    _mycode->addOutputFile(_fout);

    CPPUNIT_ASSERT_THROW( new  URANIE::Sensitivity::TMorris(_tds, _mycode, nreplique, nlevel),
                          URANIE::Exceptions::UErrorExceptions);
}

void TMorrisTest::throwExceptionTMorrisConstructorFalseLevelandDelta()
{
    Int_t nreplique = 3;
    Int_t nlevel = 5;

    // Common on an Analytical Function and a TCode
#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami,nreplique, nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi",nreplique, nlevel);
#endif
            
    _tmorris->setDrawProgressBar(kFALSE);
    
    CPPUNIT_ASSERT_THROW(_tmorris->computeIndexes(),
            URANIE::Exceptions::UErrorExceptions);

    delete _tmorris;
    Double_t delta = 1.0 * nlevel;
#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami,nreplique, nlevel, delta);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi",nreplique, nlevel, delta);
#endif
            
    _tmorris->setDrawProgressBar(kFALSE);
    
    CPPUNIT_ASSERT_THROW(_tmorris->computeIndexes(),
            URANIE::Exceptions::UErrorExceptions);

    delete _tmorris;
    nlevel = 6;
    delta = 1.0 * nlevel;
#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel, delta);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel, delta);
#endif
            
    _tmorris->setDrawProgressBar(kFALSE);

    CPPUNIT_ASSERT_THROW(_tmorris->computeIndexes(),
            URANIE::Exceptions::UErrorExceptions);

}
void TMorrisTest::throwExceptionTMorrisConstructorFalseDelta()
{
    _tds1->fileDataRead("TDS_morris_false.dat");

    CPPUNIT_ASSERT_THROW(
            new URANIE::Sensitivity::TMorris(_tds1, "x_{1}:x_{2}:x_{3}",
                    "Ishigami"), URANIE::Exceptions::UErrorExceptions);
}

void TMorrisTest::throwExceptionConstructorTooLongOutputFunction(){
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    string inputs_name;
    TList *list = tds.getListOfAttributes();
    for (const auto& attribute: *list)
        inputs_name += attribute->GetName() + string(":");
    inputs_name.pop_back();

    const int n_replica = 3;
    const int n_level = 10;

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris morris(&tds, &flowrateModel,
        inputs_name.data(),
        "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut",
        n_replica, n_level),
        URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris morris(&tds, &flowrateModel,
        inputs_name.data(),
        "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut",
        n_replica, n_level),
        URANIE::Exceptions::UErrorExceptions);
    delete list;
}

void TMorrisTest::throwExceptionConstructorTooLongOutputCode(){
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(tds, input_file);

    URANIE::Launcher::TOutputFileRow output_file("_output_flowrate_withRow_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");

    URANIE::Launcher::TCode code(&tds, "");
    code.addOutputFile(&output_file);

    const int n_replica = 3;
    const int n_level = 10;
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris morris(&tds, &code, n_replica, n_level),
        URANIE::Exceptions::UErrorExceptions);
}

void TMorrisTest::throwExceptionConstructorTooLongOutputRun(){
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    for(auto& input: inputs)
        eval.addInput(&input);
    eval.addOutput(&output);

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::DataServer::TDataServer tds("tds", "");
        for (auto& input: inputs)
            tds.addAttribute(&input);

        const int n_replica = 3;
        const int n_level = 10;
        CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris morris(&tds, &run, n_replica, n_level),
            URANIE::Exceptions::UErrorExceptions);
        run.stopSlave();
    }
}

void TMorrisTest::throwExceptionTMorrisUnboundedInfiniteLaws()
{

    URANIE::DataServer::TDataServer* _localtds = new URANIE::DataServer::TDataServer("tds","tds");
    URANIE::DataServer::TNormalDistribution *n = new URANIE::DataServer::TNormalDistribution("n",0,1);
    n->delShare();
    _localtds->addAttribute(n);

    int nreplica = 3;
    int nlevel = 10;
    CPPUNIT_ASSERT_THROW(URANIE::Sensitivity::TMorris morris(_localtds, &FonctionAffine, "", "Affine", nreplica, nlevel),
                         URANIE::Exceptions::UErrorExceptions);
    delete _localtds;
}

void throwExceptionConstructorUnboundedLaw(URANIE::DataServer::TStochasticAttribute& distribution) {
    URANIE::DataServer::TDataServer data;
    data.addAttribute(&distribution);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(data.getAttribute(0)->IsA()->GetName(),
        true, !data.getAttribute(0)->isBounded());

    const int nreplica = 3;
    const int nlevel = 10;
    URANIE::Sensitivity::TMorris morris(&data, &FonctionAffine, "", "Affine", nreplica, nlevel);
}

void TMorrisTest::throwExceptionConstructorUnboundedNormal() {
    URANIE::DataServer::TNormalDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedLogNormal() {
    URANIE::DataServer::TLogNormalDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedExponential() {
    URANIE::DataServer::TExponentialDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedCauchy() {
    URANIE::DataServer::TCauchyDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedGumbel() {
    URANIE::DataServer::TGumbelMaxDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedWeibull() {
    URANIE::DataServer::TWeibullDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedGenPareto() {
    URANIE::DataServer::TGenParetoDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedGamma() {
    URANIE::DataServer::TGammaDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

void TMorrisTest::throwExceptionConstructorUnboundedInvGamma() {
    URANIE::DataServer::TInvGammaDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}

#ifdef WITH_MATHMORE
void TMorrisTest::throwExceptionConstructorUnboundedStudent() {
    URANIE::DataServer::TStudentDistribution distribution;
    throwExceptionConstructorUnboundedLaw(distribution);
}
#endif

#ifdef WITH_LEVELE
// ===========================================
//               LevelE
// ===========================================
void TMorrisTest::checkLevelEResults(TTree *ntresu, vector<int>* elmts)
{

    double muref[26] ={-1.618518380060e-07,-4.650183213680e-07,-3.354773322130e-07,-8.712905107490e-08,-1.843326058500e-08,-1.550641863190e-08,-2.944494807340e-08,
		       -3.833148503630e-08,-3.587099555910e-08,-9.168778686270e-08,-8.004439580130e-08,-1.464119975770e-07,-2.915795514610e-07,-4.155554034590e-07,
		       -4.870784366910e-07,-5.204479317060e-07,-5.359866121130e-07,-5.456219852630e-07,-6.387857982400e-07,-7.764350752890e-07,-9.021536055700e-07,
		       -8.198758669090e-07,-5.681447935450e-07,-3.227685637670e-07,-1.678553052940e-07,-9.299648145450e-08};

    double mustarref[26] ={1.618518380060e-07,4.650183213680e-07,3.354773322130e-07,8.712905107490e-08,1.843326058500e-08,1.550641863190e-08,2.944494807340e-08,
			   3.833148503630e-08,3.587099555910e-08,9.168778686270e-08,8.004439580130e-08,1.464119975770e-07,2.915795514610e-07,4.155554034590e-07,
			   4.870784366910e-07,5.204479317060e-07,5.359866121130e-07,5.456219852630e-07,6.387857982400e-07,7.764350752890e-07,9.021536055700e-07,
			   8.198758669090e-07,5.681447935450e-07,3.227685637670e-07,1.678553052940e-07,9.299648145450e-08};

    double sigmaref[26] ={3.151931779550e-07,9.973052550900e-07,7.289155219170e-07,2.194532522080e-07,4.639384853440e-08,4.063927656800e-08,9.094404508220e-08,
			  1.170052622040e-07,1.002056699080e-07,1.668079888200e-07,1.235172579060e-07,4.316511140080e-07,8.995466775630e-07,1.275985099900e-06,
			  1.493659601220e-06,1.601234089560e-06,1.657486788020e-06,1.693923662740e-06,1.965731922450e-06,2.390510106560e-06,2.779003622150e-06,
			  2.508948870160e-06,1.701820797670e-06,9.135524755320e-07,4.117102282800e-07,1.708940441510e-07};

    // double muref[26] ={-5.19308057602e-14, -1.77441018581e-13, -1.26437179618e-13, -3.73052470823e-14, -7.84204211377e-15, -2.67980349062e-15, -3.34975461997e-15,
    //                        -4.26803416491e-15, -4.27263636278e-15, -2.9385316067e-14, -2.93477715469e-14, -7.16160923042e-14, -1.45801254428e-13, -2.079177334e-13,
    //                        -2.43690780977e-13, -2.6032174977e-13, -2.67930416936e-13, -2.72453100392e-13, -3.17018061427e-13, -3.85554578937e-13, -4.48065299492e-13,
    //                        -4.0645136056e-13, -2.79985976051e-13, -1.56611009705e-13, -7.83878505546e-14, -4.0166309606e-14};

    //     double mustarref[26] ={5.19308057602e-14, 1.77441018581e-13, 1.26437179618e-13, 3.73052470823e-14, 7.84204211377e-15, 2.67980349062e-15, 3.34975461997e-15,
    //                            4.26803416491e-15, 4.27263636278e-15, 2.9385316067e-14, 2.93477715469e-14, 7.16160923042e-14, 1.45801254428e-13, 2.079177334e-13,
    //                            2.43690780977e-13, 2.6032174977e-13, 2.67930416936e-13, 2.72453100392e-13, 3.17018061427e-13, 3.85554578937e-13, 4.48065299492e-13,
    //                            4.0645136056e-13, 2.79985976051e-13, 1.56611009705e-13, 7.83878505546e-14, 4.0166309606e-14};

    //     double sigmaref[26] ={1.22343473904e-13, 4.94935791287e-13, 3.45228829076e-13, 1.09356626115e-13, 2.33366190591e-14, 5.62986505782e-15, 9.76001852457e-15,
    //                           1.25779723706e-14, 1.0829934409e-14, 6.05834408836e-14, 5.14979085148e-14, 2.16473005265e-13, 4.50123481854e-13, 6.38445201353e-13,
    //                           7.47364230831e-13, 8.01213036461e-13, 8.29418367362e-13, 8.4775430249e-13, 9.84456000545e-13, 1.19711412132e-12, 1.39163445271e-12,
    //                           1.25666087391e-12, 8.52968391384e-13, 4.58696170381e-13, 2.07602152557e-13, 8.6446421377e-14};
    
  
    int ntps =  ntresu->Draw("mu:mustar:sigma"," Input==\"w\"","goff");
    double *mu=ntresu->GetV1(), *mustar=ntresu->GetV2(), *sigma=ntresu->GetV3();

    // stringstream musstr, mustarsstr, sigmasstr;
    for( int ival=0; ival<ntps; ival++)
    {
        // musstr << ", " << std::setprecision(14) << mu[ival];
        // mustarsstr << ", " << std::setprecision(14) << mustar[ival];
        // sigmasstr << ", " << std::setprecision(14) << sigma[ival];

        CPPUNIT_ASSERT_DOUBLES_EQUAL(muref[((elmts==NULL) ? ival : elmts->at(ival))], mu[ival], ((elmts==NULL)? LVLE_EPSILON : SUBLVLE_EPSILON));
	CPPUNIT_ASSERT_DOUBLES_EQUAL(mustarref[((elmts==NULL) ? ival : elmts->at(ival))], mustar[ival], ((elmts==NULL)? LVLE_EPSILON : SUBLVLE_EPSILON));
	CPPUNIT_ASSERT_DOUBLES_EQUAL(sigmaref[((elmts==NULL) ? ival : elmts->at(ival))], sigma[ival], ((elmts==NULL)? LVLE_EPSILON : SUBLVLE_EPSILON));
    }

    // cout<<musstr.str()<<endl;
    // cout<<mustarsstr.str()<<endl;
    // cout<<sigmasstr.str()<<endl;
}

void TMorrisTest::testLevelEMorris()
{

  // Define the DataServer
  _tds1->SetName("tdsLevelE"); _tds1->SetTitle("levele");
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("t", 100, 1000) );
  _tds1->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kl", 0.001, 0.01) );
  _tds1->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kc", 0.000001, 0.00001) );
  _tds1->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v1", 0.001, 0.1) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("l1", 100, 500) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("r1", 1, 5) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("rc1", 3, 30) );
  _tds1->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v2", 0.01, 0.1) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("l2", 50, 200) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("r2", 1, 5) );
  _tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("rc2", 3, 30) );
  _tds1->addAttribute( new URANIE::DataServer::TLogUniformDistribution("w", 100000, 10000000) );

  _tds1->getAttribute( "t" )->delShare();
  _tds1->getAttribute( "kl" )->delShare();
  _tds1->getAttribute( "kc" )->delShare();
  _tds1->getAttribute( "v1" )->delShare();
  _tds1->getAttribute( "l1" )->delShare();
  _tds1->getAttribute( "r1" )->delShare();
  _tds1->getAttribute( "rc1" )->delShare();
  _tds1->getAttribute( "v2" )->delShare();
  _tds1->getAttribute( "l2" )->delShare();
  _tds1->getAttribute( "r2" )->delShare();
  _tds1->getAttribute( "rc2" )->delShare();
  _tds1->getAttribute( "w" )->delShare();

  TString name="levelE_input_with_values_rows.in";
  _tds1->getAttribute("t")->setFileKey(name, "t", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("kl")->setFileKey(name, "kl", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("kc")->setFileKey(name, "kc", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("v1")->setFileKey(name, "v1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("l1")->setFileKey(name, "l1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("r1")->setFileKey(name, "r1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("rc1")->setFileKey(name, "rc1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("v2")->setFileKey(name, "v2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("l2")->setFileKey(name, "l2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("r2")->setFileKey(name, "r2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("rc2")->setFileKey(name, "rc2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds1->getAttribute("w")->setFileKey(name, "w", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);

  URANIE::Launcher::TOutputFileRow *out = new URANIE::Launcher::TOutputFileRow("_output_levelE_withRow_.dat");
  out->addAttribute(new URANIE::DataServer::TAttribute("y", URANIE::DataServer::TAttribute::kVector), 2);
  out->getAttribute(0)->delShare();

  URANIE::Launcher::TCode *myc;
#ifdef WIN32
  myc = new URANIE::Launcher::TCode(_tds1, "levele.exe");
#else
  myc = new URANIE::Launcher::TCode(_tds1, Form("%s/%s",gSystem->pwd(),"../../../../launCHER/teST/coDE/levelE/levele 2> /dev/null "));
#endif
  myc->addOutputFile(out);

  Int_t nlevel = 16, nreplique=10;
  _tmorris = new URANIE::Sensitivity::TMorris(_tds1, myc, nreplique, nlevel);
  _tmorris->setSeed(12345);
  _tmorris->setDrawProgressBar(kFALSE);
  _tmorris->computeIndexes("output=y");

  TTree *ntresu = _tmorris->getMorrisResults();
  checkLevelEResults(ntresu);
  
  delete out;
  delete myc;

}

void TMorrisTest::testLevelEMorrisOutputRestrVarexp()
{

  delete _tds1; _tds1 = new URANIE::DataServer::TDataServer();
  _tds1->fileDataRead("MorrisLevelEDatabase.dat");
  
  _tmorris = new URANIE::Sensitivity::TMorris(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]");
  _tmorris->computeIndexes();

  TTree *ntresu = _tmorris->getMorrisResults();
  checkLevelEResults(ntresu, _tmorris->getAttributeElements(string("y")));
}

void TMorrisTest::testLevelEMorrisOutputRestrOption()
{
  
  delete _tds1; _tds1 = new URANIE::DataServer::TDataServer();
  _tds1->fileDataRead("MorrisLevelEDatabase.dat");

  _tmorris = new URANIE::Sensitivity::TMorris(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y");
  _tmorris->computeIndexes("output=y[0]:y[5]:y[12]");

  TTree *ntresu = _tmorris->getMorrisResults();
  checkLevelEResults(ntresu, _tmorris->getAttributeElements(string("y")));
}

void TMorrisTest::testLevelEMorrisOutputRestrBoth()
{

  delete _tds1; _tds1 = new URANIE::DataServer::TDataServer();
  _tds1->fileDataRead("MorrisLevelEDatabase.dat");

  _tmorris = new URANIE::Sensitivity::TMorris(_tds1, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w","y[0]:y[5]:y[12]");
  _tmorris->computeIndexes("output=y[4]:y[11]:y[15]");

  TTree *ntresu = _tmorris->getMorrisResults();
  checkLevelEResults(ntresu, _tmorris->getAttributeElements(string("y")));
}
#endif

void TMorrisTest::testTMorrissetLevel()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;
    
#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami,nreplique, nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi",nreplique, nlevel);
#endif
            
    _tmorris->setDrawProgressBar(kFALSE);
    
    CPPUNIT_ASSERT_EQUAL(10, _tmorris->getLevel());
    _tmorris->setLevel(8);
    CPPUNIT_ASSERT_EQUAL(8, _tmorris->getLevel());
}
void TMorrisTest::throwExceptionTMorrissetLevel()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;

#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel);
#endif

    _tmorris->setDrawProgressBar(kFALSE);
    _tmorris->setLevel(5);

    CPPUNIT_ASSERT_THROW(_tmorris->computeIndexes(),
            URANIE::Exceptions::UErrorExceptions);
}
void TMorrisTest::testTMorrissetDelta()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;

#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel);
#endif

    _tmorris->setDrawProgressBar(kFALSE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tmorris->getDelta(),
            DBL_PRECISION_COMPARE);
    _tmorris->setDelta(2.5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5, _tmorris->getDelta(),
            DBL_PRECISION_COMPARE);

}
void TMorrisTest::throwExceptionTMorrisetDelta()
{
    Int_t nlevel = 9;
    Int_t nreplique = 3;
    Double_t delta = 9.0;

#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami,  nreplique, nlevel, delta);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel, delta);
#endif

    _tmorris->setDrawProgressBar(kFALSE);
    _tmorris->setDelta(0.0);

    CPPUNIT_ASSERT_THROW(_tmorris->computeIndexes(),
            URANIE::Exceptions::UErrorExceptions);
}
void TMorrisTest::testTMorrisgetReplica()
{
    Int_t nreplique = 3;
    Int_t nlevel = 10;

#ifdef ROOT5
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, _p2fIshigami, nreplique, nlevel);
#else
    _tmorris = new URANIE::Sensitivity::TMorris(_tds2, &Ishigami, "", "Ishi", nreplique, nlevel);
#endif

    _tmorris->setDrawProgressBar(kFALSE);

    CPPUNIT_ASSERT_EQUAL(3, _tmorris->getNReplica());

}


void TMorrisTest::testgFunction6Optimized ()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();
    tds->fileDataRead ("morris_gFunction6_Optimized.dat");

    CPPUNIT_ASSERT_EQUAL(28, tds->getNPatterns ());
    CPPUNIT_ASSERT( tds->isAttribute ("gF6"));
    CPPUNIT_ASSERT( tds->isAttribute ("x1"));
    CPPUNIT_ASSERT( tds->isAttribute ("x2"));
    CPPUNIT_ASSERT( tds->isAttribute ("x3"));
    CPPUNIT_ASSERT( tds->isAttribute ("x4"));
    CPPUNIT_ASSERT( tds->isAttribute ("x4"));
    CPPUNIT_ASSERT( tds->isAttribute ("x6"));
    CPPUNIT_ASSERT( !tds->isAttribute ("x7"));

    URANIE::Sensitivity::TMorris* tmorris = new URANIE::Sensitivity::TMorris(tds, "x1:x2:x3:x4:x5:x6", "gF6");
    tmorris->setDrawProgressBar(kFALSE);
    tmorris->computeIndexes();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.6666667, tmorris->getDelta (), 0.00001);
    CPPUNIT_ASSERT_EQUAL(4, tmorris->getNReplica ());

    Double_t sobol;
    TTree *tree = tmorris->getResultTuple(true);
    tree->SetBranchAddress("Sobol", &sobol);

    gROOT->Delete("eventList;*");
    tree->Draw(">>eventList", "(Order==\"mu\")");
    TEventList *tel = (TEventList*) gDirectory->Get("eventList");
    Int_t nx = 6;
    CPPUNIT_ASSERT_EQUAL(nx, tel->GetN());

    Double_t dvalMu[6] = {-0.006, -0.078, -0.130, -0.004, 0.012, -0.004};
    Double_t dvalMuStar[6] = {0.056, 0.277, 1.760, 1.185, 0.035, 0.099};
    Double_t dvalSigma[6] = {0.064, 0.321, 2.049, 1.370, 0.041, 0.122};

    for (Int_t i = 0; i < 6; i++) {
        tree->GetEntry(tel->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMu[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMuStar[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalSigma[i], sobol, 0.001);
    }

    // Clean
    delete tmorris;
    delete tds;
}

void TMorrisTest::testFonctionAffine()
 {
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();

    tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x1", 0, 1));//-1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", 0, 100000));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
    
    Int_t nlevel = 20;
    URANIE::Sensitivity::TMorris * tmorris = new URANIE::Sensitivity::TMorris(tds, FonctionAffine, "", "FonctionAffine", 10, nlevel, 10);
    //  scmo->setLog();
    tmorris->setDrawProgressBar(kFALSE);
    tmorris->computeIndexes();

  
    Double_t sobol;
    TTree *tree = tmorris->getResultTuple(true);
    tree->SetBranchAddress("Sobol", &sobol);

    gROOT->Delete("eventList;*");
    tree->Draw(">>eventList", "(Order==\"mu\")");
    TEventList *tel = (TEventList*) gDirectory->Get("eventList");
    Int_t nx = 2;
    CPPUNIT_ASSERT_EQUAL(nx, tel->GetN());

    Double_t dvalMu[2] = {1., 200000.};
    Double_t dvalMuStar[2] = {1., 200000.};
    Double_t dvalSigma[2] = {0., 0.};

    for (Int_t i = 0; i < 2; i++) {
        tree->GetEntry(tel->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMu[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMuStar[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalSigma[i], sobol, 0.001);
    }

    // Clean
    delete tmorris;
    delete tds;
 }

    void TMorrisTest::testFonctionAffine2()
 {
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();

    tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x1", 0, 1));//-1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", 0, 100000));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();

    Int_t nlevel = 20;
    URANIE::Sensitivity::TMorris * tmorris = new URANIE::Sensitivity::TMorris(tds, FonctionAffine, "", "FonctionAffine", 10, nlevel, 10);
    //  scmo->setLog();
    tmorris->computeIndexes("relative");

    Double_t sobol;
    TTree *tree = tmorris->getResultTuple(true);
    tree->SetBranchAddress("Sobol", &sobol);

    gROOT->Delete("eventList;*");
    tree->Draw(">>eventList", "(Order==\"mu\")");
    TEventList *tel = (TEventList*) gDirectory->Get("eventList");
    Int_t nx = 2;
    CPPUNIT_ASSERT_EQUAL(nx, tel->GetN());

    Double_t dvalMu[2] = {1., 2.};
    Double_t dvalMuStar[2] = {1., 2.};
    Double_t dvalSigma[2] = {0., 0.};

    for (Int_t i = 0; i < 2; i++) {
        tree->GetEntry(tel->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMu[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMuStar[i], sobol, 0.001);
        tree->GetEntry(tel->GetEntry(i)+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalSigma[i], sobol, 0.001);
    }

    // Clean
    delete tmorris;
    delete tds;
}

void TMorrisTest::validateValuesIshigami()
{
    Double_t sobol;
    TTree *tree = _tmorris->getResultTuple(true);
    tree->SetBranchAddress("Sobol", &sobol);

    gROOT->Delete("eventList;*");
    tree->Draw(">>eventList", "(Order==\"mu\")","goff");
    TEventList *tel = (TEventList*) gDirectory->Get("eventList");
    CPPUNIT_ASSERT_EQUAL(3, tel->GetN());

    // Reference values
    Double_t dvalMu[3]     = { 13.21425, 2.8293223, -3.756228 };
    Double_t dvalMuStar[3] = { 13.21425, 2.8293223,  3.756228 };
    Double_t dvalSigma[3]  = { 13.81881, 2.4502650,  6.505979 };
    // Double_t dvalMu[3]     = { 2.1031139, 0.4503006, -0.597822 };
    // Double_t dvalMuStar[3] = { 2.1031139, 0.4503006, 0.5978223 };
    // Double_t dvalSigma[3]  = { 2.1993319, 0.3899717, 1.0354587 };

    for (Int_t i = 0; i < 3; i++) {
        tree->GetEntry(tel->GetEntry(i));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMu[i], sobol, 0.0001);
        tree->GetEntry(tel->GetEntry(i)+1);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMuStar[i], sobol, 0.0001);
        tree->GetEntry(tel->GetEntry(i)+2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalSigma[i], sobol, 0.0001);
    }
    gROOT->Delete("eventList;*");
}

void TMorrisTest::testIshigamiMorrisWithDrawings()
{
    setUp1();
    _tmorris->computeIndexes();


    TCanvas *Can = new TCanvas("toto","tutu",1);
    _tmorris->drawIndexes("mustar,nonewcanv");
    _tmorris->drawIndexes("mu,nonewcanv");
    
    _tmorris->drawIndexes("#mu - values","","hist,mu,nonewcanv");
    _tmorris->drawIndexes("#mu^{*} - values","","hist,mustar,nonewcanv");
    _tmorris->drawIndexes("#sigma - values","","hist,sigma,nonewcanv");     
    delete Can;

    _tmorris->drawIndexes("mustar");
    _tmorris->drawIndexes("mu");
    
    _tmorris->drawIndexes("#mu - values","","hist,mu");
    _tmorris->drawIndexes("#mu^{*} - values","","hist,mustar");
    _tmorris->drawIndexes("#sigma - values","","hist,sigma");     
    
    _tmorris->drawIndexesAll("");

    _tmorris->drawSample();
    
}

void TMorrisTest::testIshigamiMorrisComputeIndexesWithTLauncherFunction()
{
    setUp1();
    _tmorris->computeIndexes();
    validateValuesIshigami();
}

void TMorrisTest::testIshigamiMorrisComputeIndexesGenerateSampleWithTLauncherFunction()
{
    setUp1();
    _tmorris->generateSample();
    _tmorris->computeIndexes();
    validateValuesIshigami();
}

void TMorrisTest::testIshigamiMorrisComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction()
{
    setUp1();
    _tmorris->generateSample();
    URANIE::Launcher::TLauncherFunction *tlf =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(_tds2, _p2fIshigami,"", "p2fIshigami");
#else
            new URANIE::Launcher::TLauncherFunction(_tds2, &Ishigami,"", "p2fIshigami");
#endif
                    
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    _tmorris->computeIndexes();
    validateValuesIshigami();
    delete tlf;
}

void TMorrisTest::validateValuesFlowrate()
{
    Double_t sobol;
    TTree *tree = _tmorris->getResultTuple(true);
    tree->SetBranchAddress("Sobol", &sobol);
    
    gROOT->Delete("eventList;*");
    tree->Draw(">>eventList", "(Order==\"mu\")","goff");
    TEventList *tel = (TEventList*) gDirectory->Get("eventList");
    Int_t nx = 8;

    CPPUNIT_ASSERT_EQUAL(nx, tel->GetN());


    Double_t dvalMu[8]     = { 1.27478910000000e+02, -6.96420000000046e-02, 4.07999999986686e-04, 4.65953999999991e-01,
                               2.11923420000000e+01, -3.27488220000000e+01, -2.38932540000000e+01, 7.57660800000000e+00};
    Double_t dvalMuStar[8] = { 1.27478910000000e+02, 6.96420000000046e-02, 4.07999999986686e-04, 4.65953999999991e-01,
                               2.11923420000000e+01, 3.27488220000000e+01, 2.38932540000000e+01, 7.57660800000000e+00};
    Double_t dvalSigma[8]  = { 3.45217813212262e+01, 7.94260703044043e-02, 4.39558869768938e-04, 3.30096175312592e-01,
                               8.84988478326154e+00, 3.01460852966718e+01, 8.27818537037846e+00, 2.74575575800106e+00};

    
    // Double_t dvalMu[8]     = { 1.27478910000000e+03, -1.39563126252514e-06, 7.76699029100868e-09, 8.80820415879000e-03,
    //                            1.76602850000000e-01, -2.72906850000000e-01, -4.26665250000000e-02, 3.45963835616439e-03};
    // Double_t dvalMuStar[8] = { 1.27478910000000e+03, 1.39563126252514e-06, 7.76699029100868e-09, 8.80820415879000e-03,
    //                            1.76602850000000e-01, 2.72906850000000e-01, 4.26665250000000e-02, 3.45963835616439e-03};
    // Double_t dvalSigma[8]  = { 3.45217813212262e+02, 1.59170481571953e-06, 8.36776831846445e-09, 6.24000331403766e-03,
    //                            7.37490398605126e-02, 2.51217377472265e-01, 1.47824738756758e-02, 1.25376975251190e-03};
    
    // stringstream musstr, mustarsstr, sigmasstr;
    for (Int_t i = 0; i < 8; i++) {
        tree->GetEntry(tel->GetEntry(i));
        //musstr << ", " << std::scientific << std::setprecision(14) << sobol;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMu[i], sobol, 0.0001);
        tree->GetEntry(tel->GetEntry(i)+1);
        //mustarsstr << ", " << std::scientific << std::setprecision(14) << sobol;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalMuStar[i], sobol, 0.0001);
        tree->GetEntry(tel->GetEntry(i)+2);
        //sigmasstr << ", " << std::scientific << std::setprecision(14) << sobol;
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalSigma[i], sobol, 0.0002);
    }
    
   // cout<<musstr.str()<<endl;
   // cout<<mustarsstr.str()<<endl;
   // cout<<sigmasstr.str()<<endl;

    gROOT->Delete("eventList;*");
}

void TMorrisTest::testFlowrateMorrisComputeIndexesWithTLauncherCode()
{
    setUp2();
    _tmorris->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TMorrisTest::testFlowrateMorrisComputeIndexesWithTCodeEval()
{
    //TString sJDDReference = gSystem->pwd();
    //sJDDReference += TString("/flowrate_input_with_keys.in");

    URANIE::Relauncher::TKeyScript fin("flowrate_input_with_keys.in");
    fin.addInput(_tds->getAttribute("rw"), "Rw");
    fin.addInput(_tds->getAttribute("r"), "R");
    fin.addInput(_tds->getAttribute("tu"), "Tu");
    fin.addInput(_tds->getAttribute("tl"), "Tl");
    fin.addInput(_tds->getAttribute("hu"),  "Hu");
    fin.addInput(_tds->getAttribute("hl"), "Hl");
    fin.addInput(_tds->getAttribute("l"), "L");
    fin.addInput(_tds->getAttribute("kw"), "Kw");

    URANIE::Relauncher::TKeyResult fout("_output_flowrate_withKey_.dat");
    URANIE::DataServer::TAttribute *yhat = new URANIE::DataServer::TAttribute("yhat");
    yhat->delShare();
    fout.addOutput(yhat, "yhat"); // memory leak
    //fout.addOutput(new URANIE::DataServer::TAttribute("d"));

    URANIE::Relauncher::TCodeEval code((UranieTests::code_path("flowrate") + " -s -k").data());
    code.addInputFile(&fin);
    code.addOutputFile(&fout);
    code.setOldTmpDir();

    //URANIE::Relauncher::TThreadedRun run(&code, 4);
    URANIE::Relauncher::TSequentialRun run(&code);
    run.startSlave();
    if (run.onMaster()) {
        Int_t nreplique = 3;
        Int_t nlevel = 10;
        _tmorris = new URANIE::Sensitivity::TMorris(_tds, &run, nreplique,
                nlevel);
        _tmorris->setDrawProgressBar(kFALSE);
        _tmorris->setSeed(123);
        _tmorris->computeIndexes();
        validateValuesFlowrate();
        
        run.stopSlave();
    }

    TTree *T = _tmorris->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuS.txt");
    T->Scan("*");
}


void TMorrisTest::testFlowrateMorrisComputeIndexesWithThreadedandNot()
{
    //TString sJDDReference = gSystem->pwd();
    //sJDDReference += TString("/flowrate_input_with_keys.in");

    URANIE::Relauncher::TKeyScript fin("flowrate_input_with_keys.in");
    fin.addInput(_tds->getAttribute("rw"), "Rw");
    fin.addInput(_tds->getAttribute("r"), "R");
    fin.addInput(_tds->getAttribute("tu"), "Tu");
    fin.addInput(_tds->getAttribute("tl"), "Tl");
    fin.addInput(_tds->getAttribute("hu"),  "Hu");
    fin.addInput(_tds->getAttribute("hl"), "Hl");
    fin.addInput(_tds->getAttribute("l"), "L");
    fin.addInput(_tds->getAttribute("kw"), "Kw");

    URANIE::Relauncher::TKeyResult fout("_output_flowrate_withKey_.dat");
    URANIE::DataServer::TAttribute *yhat = new URANIE::DataServer::TAttribute("yhat");
    yhat->delShare();
    fout.addOutput(yhat, "yhat"); // memory leak

    URANIE::Relauncher::TCodeEval code((UranieTests::code_path("flowrate") + " -s -k").data());
    code.addInputFile(&fin);
    code.addOutputFile(&fout);
    code.setOldTmpDir();

    URANIE::Relauncher::TThreadedRun run(&code,4);
    run.startSlave();
    if (run.onMaster()) {
        Int_t nreplique = 3;
        Int_t nlevel = 10;
        _tmorris = new URANIE::Sensitivity::TMorris(_tds, &run, nreplique,
                                                    nlevel);
        _tmorris->setDrawProgressBar(kFALSE);
        _tmorris->setSeed(123);
        _tmorris->computeIndexes();
        
        run.stopSlave();
    }

    
    TTree *T = _tmorris->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuT.txt");
    T->Scan("*");

    Int_t nret = UranieTests::areFilesEqual(TString("outpuT.dat"),
                                            TString("outpuS.dat"));
    
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    UranieTests::removeFile(gSystem->pwd(),"outpuT.dat");
    UranieTests::removeFile(gSystem->pwd(),"outpuS.dat");
}



void TMorrisTest::testFlowrateMorrisComputeIndexesGenerateSampleWithTLauncherCode()
{
    setUp2();
    _tmorris->generateSample();
    _tmorris->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TMorrisTest::testFlowrateMorrisComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode()
{
    setUp2();
    _tmorris->generateSample();
    URANIE::Launcher::TLauncher *tl = new URANIE::Launcher::TLauncher(_tds,
            _mycode);
    tl->setDrawProgressBar(kFALSE);
    tl->run("nointermed");
    delete tl;
    tl = NULL;
    _tmorris->computeIndexes("nointermed");
    validateValuesFlowrate();
}

#ifdef WITH_PYTHON
void TMorrisTest::testSensitivityMorrisFunctionFlowrateWithPython()
{
    gSystem->Exec(UranieTests::python_command("sensitivityMorrisFunctionFlowrate.py").data());
    ifstream fichier("sensitivityMorrisFunctionFlowrateData.txt"); 
    string level;
    string delta;
    string nreplica;
    /*if (fichier)
    {*/
        getline(fichier, level);
        getline(fichier, delta);
        getline(fichier, nreplica);
        fichier.close();

        CPPUNIT_ASSERT_EQUAL(10, atoi(level.c_str()));
        CPPUNIT_ASSERT_EQUAL(3, atoi(nreplica.c_str()));
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.555555555556, atof(delta.c_str()), DBL_PRECISION_COMPARE); 

    /*} else {
        cout << "erreur d'ouverture du fichier" << endl;
    }*/

}
#endif
