//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTShapley.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2020/09/29

 \brief  Implement the unit tests of the URANIE::Sensitivity::TShapley class
 */

// Uranie
#include "TestTShapley.h"
#include "TShapley.h"
#include "UDisableTest.h"
#include "TUniformDistribution.h"
#include "TestUserFunctions.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TLeaf.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#define EPSILON 0.001
#define SPL_PRECISION_COMPARE 1.e-6
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>
using namespace std;

TShapleyTest::TShapleyTest() :
        _tdsA(NULL), _tdsD(NULL), _TShapley(NULL)
{
    gErrorIgnoreLevel = kWarning;
}


TShapleyTest::~TShapleyTest()
{
  UranieTests::removeFolder(gSystem->pwd(),"URANIE");
  UranieTests::removeFile(gSystem->pwd(), "_Shapley_launching_.dat");
  UranieTests::removeFile(gSystem->pwd(), "_Shapley_sampling_.dat");
}

void TShapleyTest::setUp()
{
    _tdsA = new URANIE::DataServer::TDataServer("tdsFlowrate", "my first TDS");
    _tdsA->fileDataRead("flowrate_shapley_launching.dat");

    _tdsD = new URANIE::DataServer::TDataServer("tdsFlowrate_1", "TDS for Ishigami function");
}

void TShapleyTest::tearDown()
{
    if ( _TShapley != NULL ) {
        delete _TShapley;
        _TShapley = NULL;
    }

    delete _tdsA;
    _tdsA = NULL;

    delete _tdsD;
    _tdsD = NULL;

    /// Remove ROOT Files
    // UranieTests::removeFile(gSystem->pwd(), "Ishigami_00300.root");
    // UranieTests::removeFile(gSystem->pwd(), "flowrate_shapley_launching.root");

    /// Remove ROOT Objects
    // if( gROOT->FindObject("Sampling_Sampling_srs_5_copied") )
    //     gROOT->Delete("Sampling_Sampling_srs_5_copied");
    // if( gROOT->FindObject("Sampling_Sampling_srs_5") )
    //     gROOT->Delete("Sampling_Sampling_srs_5");
    // if( gROOT->FindObject("hPValues") )
    //     gROOT->Delete("hPValues");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TShapleyTest::testconstructorShapley()
{
    _TShapley = new URANIE::Sensitivity::TShapley();

    CPPUNIT_ASSERT_MESSAGE("The Log flag via getLog()", _TShapley->getLog() == kFALSE);
    _TShapley->unsetLog();

    const Int_t nBranches = 8;
    TString sBranches[nBranches] = {"Out", "Inp", "Order", "Method", "Algo", "Sobol", "CILower", "CIUpper"};


    ///
    /// Default constructor (empty)
    ///
    cout << endl;

    URANIE::Sensitivity::TShapley *defShapley = new URANIE::Sensitivity::TShapley();

    CPPUNIT_ASSERT_MESSAGE("The Log flag via getLog()",
                           defShapley->getLog() == kFALSE);

    defShapley->unsetLog();

    CPPUNIT_ASSERT_MESSAGE("Name of the method (Shapley)",
                           strcmp(defShapley->GetName(), "Shapley") == 0);
    CPPUNIT_ASSERT_MESSAGE("Name of the method (Shapley)",
                           strcmp(defShapley->getMethodName(), "Shapley") == 0);

    /// Default Values for the TSensitivity Object with the Getter Functions
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input getID()",
                                 ID_SENSITIVITY, defShapley->getID());
    CPPUNIT_ASSERT_MESSAGE("The pointer of the result TTree getResultTuple()",
                           defShapley->getResultTuple() != NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of the result TTree getResultTuple()",
                                 0, (Int_t)(defShapley->getResultTuple()->GetEntries()));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of branches of the result TTree getResultTuple()",
                                 nBranches, (Int_t)(defShapley->getResultTuple()->GetNbranches()));
    for (Int_t i=0; i < nBranches; i++) {
        CPPUNIT_ASSERT_MESSAGE(Form("Status of the branch \"%s\" in the result TTree getResultTuple()", sBranches[i].Data()),
                               defShapley->getResultTuple()->GetBranchStatus(sBranches[i].Data()));
    }
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the evaluatuor function getFunctionName() [%s]", defShapley->getFunctionName().Data()),
                           strcmp(defShapley->getFunctionName(), "") == 0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("The Seed value getSeed()[%d]", defShapley->getSeed()),
                                 0, (Int_t)(defShapley->getSeed()));
    CPPUNIT_ASSERT_MESSAGE("The noIntermediateSaved flag",
                           defShapley->getNoIntermediateSaved() == kFALSE);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the Iterator getIteratorName() [%s]", defShapley->getIteratorName()),
                           strcmp(defShapley->getIteratorName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE(Form("The Name of the time attribute getTimeName() [%s]", defShapley->getTimeName().Data()),
                           strcmp(defShapley->getTimeName(), "") == 0);
    CPPUNIT_ASSERT_MESSAGE("The draw progress bar flag",
                           defShapley->getDrawProgressBar());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 0, defShapley->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 0, defShapley->getNOutput());

    /// Default Values for the TSensitivity Object WITHOUT Getter Functions
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TDS is NULL (default)",
                                 defShapley->_tds == NULL);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample via _nS",
                                 0, defShapley->_nS);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of elements of Y via _nElY",
                                 0, (Int_t)(defShapley->_nElY));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via _nbOut",
                                 0, defShapley->_nbOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Counter for output via _iOut",
                                 0, defShapley->_iOut);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iy",
                                 0, (Int_t)(defShapley->_iy));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Iterator over number of element via _iely",
                                 0, (Int_t)(defShapley->_iely));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Type of launcher (default)",
                                 URANIE::Sensitivity::TSensitivity::kUnknown, defShapley->_nLauncher);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TCode is NULL (default)",
                                 defShapley->_code == NULL);
    CPPUNIT_ASSERT_MESSAGE("Pointer of the TRun is NULL (default)",
                                 defShapley->_run == NULL);
    CPPUNIT_ASSERT_MESSAGE("The Fact that the input list is provided or not",
                           defShapley->_bChosenOutputs == kFALSE);

    /// Clean
    delete defShapley; defShapley = NULL;
}

void TShapleyTest::testExceptionTShapleyComputeIndexes()
{
    _tdsD->addAttribute( new TUniformDistribution("rw", 0.05, 0.15));
    _tdsD->addAttribute( new TUniformDistribution("r", 100.0, 50000.0));
    _tdsD->addAttribute( new TUniformDistribution("tu", 63070.0, 115600.0));
    _tdsD->addAttribute( new TUniformDistribution("tl", 63.1, 116.0));
    _tdsD->addAttribute( new TUniformDistribution("hu", 990.0, 1110.0));
    _tdsD->addAttribute( new TUniformDistribution("hl", 700.0, 820.0));
    _tdsD->addAttribute( new TUniformDistribution("l", 1120.0, 1680.0));
    _tdsD->addAttribute( new TUniformDistribution("kw", 9855.0, 12045.0));

    _tdsD->getAttribute("rw")->delShare();
    _tdsD->getAttribute("r")->delShare();
    _tdsD->getAttribute("tu")->delShare();
    _tdsD->getAttribute("tl")->delShare();
    _tdsD->getAttribute("hu")->delShare();
    _tdsD->getAttribute("hl")->delShare();
    _tdsD->getAttribute("l")->delShare();
    _tdsD->getAttribute("kw")->delShare();

    TString sInputs  ="rw:r:tu:tl:hu:hl:l:kw";
    const Int_t nX = 8;

    TString sOutputs ="flowrateModel";
    const Int_t nY = 1;

    TString sAtts = sInputs;
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
   }

    TMatrixD matCorr(nX, nX);
    matCorr.UnitMatrix();
    matCorr(0,1) = matCorr(1, 0) = 0.100;
    matCorr(0,2) = matCorr(2, 0) = 0.200;
    matCorr(0,3) = matCorr(3, 0) = 0.300;
    matCorr(0,4) = matCorr(4, 0) = 0.400;

    matCorr(1,2) = matCorr(2, 1) = 0.201;
    matCorr(1,3) = matCorr(3, 1) = 0.301;
    matCorr(1,4) = matCorr(4, 1) = 0.401;

    matCorr(2,3) = matCorr(3, 2) = 0.302;
    matCorr(2,4) = matCorr(4, 2) = 0.402;

    matCorr(3,4) = matCorr(4, 3) = 0.403;


    Int_t nS = 500;
    _TShapley = new URANIE::Sensitivity::TShapley(_tdsD, &flowrateModel, nS, sInputs, sOutputs);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 0, _tdsD->getNPatterns());

    Int_t nM = 1000;
    Int_t nO = 1;
    Int_t nI = 3;

    _TShapley->setNm(nM);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Value of nM(%d))", nM), nM, _TShapley->getNm());

    _TShapley->setNo(nO);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Value of nO(%d))", nO), nO, _TShapley->getNo());

    _TShapley->setNi(nI);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Value of nI(%d))", nI), nI, _TShapley->getNi());

    _TShapley->setSeed(123654789);

    _TShapley->setInputCorrelationMatrix(matCorr);

    _TShapley->setDrawProgressBar(kFALSE);
    _TShapley->generateSample();

    Int_t nsize = nS + (nX-1)*nM*nI*nO;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Size of sample in the file \"Ishigami_%05d.dat\") ", nS),
                                 nsize, _tdsD->getNPatterns());

    _TShapley->computeIndexes();

    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sOutputs.Data()), _tdsD->isAttribute(sOutputs));

    const Int_t nAtts = nX + nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Attributes  via nAtts", 9, nAtts);
    Double_t dvalShapley[nX] = {0.724072, -0.0352185, 0.0315779, 0.0856236, 0.12833, 0.0215925, 0.0272871, 0.0167354};

    inStream.clear();
    inStream.seekg(0);

    // TTree of results
    TTree *tt = _TShapley->getResultTuple();
    tt->Scan("*");

    Int_t nResu =  nX;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    inStream.clear();
    inStream.seekg(0);
    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i++) {
        inStream >> inName;
        tt->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of Shapley for [%s] i(%d) %e", inName.c_str(), i, dvalShapley[i]),
                                              dvalShapley[i], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
    }
    _TShapley->printSummary();
    _TShapley->printLog();
}

void TShapleyTest::testExceptionTShapleyComputeIndexesFromTDS()
{
    Int_t nS = 61000;
    _tdsD->SetName("tdsSample");
    _tdsD->fileDataRead("linear_fromR_shapley.dat");

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Size of sample in the file \"linear_fromR_shapley.dat\") ",
                                 nS, _tdsD->getNPatterns());

    TString sInputs  ="x:u:v";
    const Int_t nX = 3;

    TString sOutputs ="y";
    const Int_t nY = 1;


    TString sAtts = Form("%s:%s", sInputs.Data(), sOutputs.Data());
    sAtts.ReplaceAll(":", " ");
    istringstream inStream((char *) sAtts.Data());
    string inName;
    Double_t dval = -1.0;
    Double_t dSigma;
    cout << endl;

    while (inStream >> inName)
    {
        TString satt(inName.c_str());
        CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", satt.Data()),
                               _tdsD->isAttribute(satt));
   }

    _TShapley = new URANIE::Sensitivity::TShapley(_tdsD, sInputs, sOutputs);
    _TShapley->unsetLog();

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Input  via getNInput()",
                                 nX, _TShapley->getNInput());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Output via getNOutput()",
                                 nY, _TShapley->getNOutput());

    const Int_t nAtts = nX + nY;
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Number of Attributes  via nAtts", 4, nAtts);
    Double_t dvalShapley[nX] = {0.0879683, 0.416015, 0.496017};

    inStream.clear();
    inStream.seekg(0);

    _TShapley->setDrawProgressBar(kFALSE);
    _TShapley->computeIndexes();

    // TTree of results
    TTree *tt = _TShapley->getResultTuple();

    Int_t nResu =  nX;
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Number of resultats %d", nResu),
                                      nResu, (Int_t)(tt->GetEntries()));

    inStream.clear();
    inStream.seekg(0);
    TLeaf *tlSobol = tt->GetLeaf("Sobol");

    for(Int_t i=0; i < nResu; i++) {
        inStream >> inName;
        tt->GetEntry(i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("Value Of Shapley for [%s] i(%d) %e", inName.c_str(), i, dvalShapley[i]),
                                              dvalShapley[i], tlSobol->GetValue(), SPL_PRECISION_COMPARE);
    }

    _TShapley->printSummary();

}


////////////////////////////////
///                          ///
///      Exceptions          ///
///                          ///
////////////////////////////////

void TShapleyTest::throwExceptionTShapleyConstructor()
{
    URANIE::DataServer::TDataServer *tdsnull = new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(tdsnull, "x_{1}:x_{2}:x_{3}", "Ishigami", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(NULL, "x_{1}:x_{2}:x_{3}", "Ishigami", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);
    /// Clean
    delete tdsnull;

    /// False input
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(_tdsA, "rw:hu:TheUnKnownAtt", "Ishigami", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);

    /// False output
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(_tdsA, "rw:hu:hl:tu:tl", "TheUnKnownAtt", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);
    /// empty output
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(_tdsA, "rw:hu:hl:tu:tl", "", "myOpt"),
                         URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TShapley(_tdsA, "rw:hu:hl:tu:tl", ":", "myOpt"),
                          URANIE::Exceptions::UErrorExceptions);

}

void TShapleyTest::throwExceptionsGenerateSample()
{
    _TShapley = new URANIE::Sensitivity::TShapley(_tdsA, "rw:tu:tl:hu:hl", "flowrateModel", "myOpt");
    _TShapley->unsetLog();

    CPPUNIT_ASSERT_THROW_MESSAGE("Case of the generateSample method",
                                 _TShapley->generateSample("unKnown"),
                                 URANIE::Exceptions::UErrorExceptions);
}

// End OF File
