//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestUserFunctions.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-09-14 10:45:11

 \brief Implement the user function 

 The analytical function is givent by the formula
 
 \$[ f(x_1, x_2) = a * ( x_2 - x_1^2 )^2 + b * ( 1.0 - x_1 )^2 \$]

 with \f$ a=10.0, b=1.0 \f$.

 */

// ROOT
#include "TROOT.h"
#include "TMath.h"

/****************** User Function **********************/
//! The true flow rate model
/*!
 It describes the flow rate of water through a borehole from an upper aquifer to a lower aquifer separated by an impermeable rock layer.
 The response variable \f$ y\f$, the \b flow \b water through the borehole in \f$ m^{3}/yr\f$, is given by :
 \f[
 y = f(x) = \frac{2 \pi T_u\left( H_u - H_l\right)}{\ln(\frac{r}{r_{\omega}})\left[ 1 + \frac{2 L T_u}{\ln(\frac{r}{r_{\omega}}) r_{\omega}^2 K_{\omega}} + \frac{T_u}{T_l}\right]}
 \f]

 \param x (double *) the input vector in the order :
 \li \f$ r_{\omega} \in [0.05, 0.15]\f$ : radius of borehole (m)
 \li \f$ r \in [100, 50~000]\f$ : radius of influence {\ix (m)}
 \li \f$ T_u \in [63~070, 115~600]\f$ : Transmissivity of upper aquifer (\f$ m^{2}\f$/year)
 \li \f$ T_l \in [63.1, 116]\f$ : Transmissivity of lower aquifer (\f$ m^{2}\f$/year)
 \li \f$ H_u \in [990, 1~110]\f$ : Potentiometric head of upper aquifer (m)
 \li \f$ H_l \in [700, 820]\f$ : Potentiometric head of lower aquifer (m)
 \li \f$ L \in [1~120, 1~680]\f$ : length of borehole (m)
 \li \f$ K_{\omega} \in [9~855, 12~045]\f$ : hydraulic conductivity of borehole (m)
 \param y (double *) the output vector

 \see waterboreholeHoXu
 */
void flowrateModel(double *x, double *y)
{
    Double_t drw = x[0], dr = x[1];
    Double_t dtu = x[2], dtl = x[3];
    Double_t dhu = x[4], dhl = x[5];
    Double_t dl = x[6], dkw = x[7];

    Double_t dnum = 2.0 * TMath::Pi() * dtu * (dhu - dhl);
    Double_t dlnronrw = TMath::Log(dr / drw);
    Double_t dden =
            dlnronrw
                    * (1.0 + (2.0 * dl * dtu) / (dlnronrw * drw * drw * dkw)
                            + dtu / dtl);

    y[0] = dnum / dden;
}

void Ishigami(double *paramt, Double_t *res)
{
    Double_t a = 7.0;
    Double_t b = 0.1;

    Double_t x1 = paramt[0];
    Double_t x2 = paramt[1];
    Double_t x3 = paramt[2];

    Double_t dsinx1 = TMath::Sin(x1);
    Double_t dsinx2 = TMath::Sin(x2);

    res[0] = dsinx1;
    res[0] += a * dsinx2 * dsinx2;
    res[0] += b * x3 * x3 * x3 * x3 * dsinx1;
}


void FonctionAffine(double *paramt, Double_t *res)
{
    Double_t x1 = paramt[0];
    Double_t x2 = paramt[1];

    res[0] = x1 + 2 * x2;
}

void CounterExemple3Function(double *x, double *y)
{
  Double_t c1=1.0, c2=1.0, c3=1.0, c4=1.0, c12=50.0;

  Double_t dsum=0.0;
  Double_t dpart1 = (x[0] - 0.5);
  Double_t dpart2 = (x[1] - 0.5);

  dsum += c1 * dpart1;
  dsum += c2 * dpart2 ;
  dsum += c3 * (x[2] - 0.5);
  dsum += c4 * (x[3] - 0.5);

  dsum += c12 * dpart1 * dpart2 * dpart2 * dpart2 * dpart2 * dpart2;

  y[0] = dsum;
}

void IshigamiDecaled(double *paramt, Double_t *res)
{
    Double_t dres = 0.0;
    Ishigami(paramt, &dres);

    res[0] = dres + 1.0e4;
}

void badPrototypeFunction()
{
    //Double_t dres = 0.0;
}
/****************** End of User Function **********************/
