//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTestKolmogorovSmirnov.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Tue Jan  6 11:56:31 2009

 \brief Kolmogorov-Smirnov Statistical Test

 \par Modifications
 \li Cedric Forgues (C.S.) 2010/02/17
 Change the prototype TMath:Sort() to adapt to ROOT 5.22.00f
 */

#include <iomanip>

// ROOT
#include "TF1.h"
#include "TH1.h"
#include "TLeaf.h"
#include "TPad.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TRegexp.h"

// Uranie
#include "TTestKolmogorovSmirnov.h"
#include "TStochAttribut.h"
#include "UExceptions.h"

ClassImp (URANIE::UncertModeler::TTestKolmogorovSmirnov)

namespace URANIE
{
namespace UncertModeler
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TTestKolmogorovSmirnov::TTestKolmogorovSmirnov(
        URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        Option_t * option) :
        TTester(tds, TTester::kKolmogorovSmirnov, varexpinput, option)
{
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TTestKolmogorovSmirnov::~TTestKolmogorovSmirnov()
{
    if (_blog)
        cout << " ** Destroy TTestKolmogorovSmirnov" << endl;
}

// -----------------------------------------------------------------------
//  computeTheCurrentScore
// -----------------------------------------------------------------------
Double_t TTestKolmogorovSmirnov::computeTheCurrentScore(Int_t i, Double_t dzi)
{
    Double_t di = 1.0 * i;
    Double_t dn = 1.0 * _nS;

    Double_t dPlus = di / dn - dzi;
    Double_t dMinus = dzi - (di - 1.0) / dn;
    if (_blog) {
        cout << endl << " ** KS :: di[" << di << "] zi[" << std::setw(1) << dzi  << "]";
        cout << " ** dPlus[" << dPlus << "]";
        cout << " ** dMinus[" << dMinus << "]";
    }

    return (TMath::Max(dMinus, dPlus));
}

// -----------------------------------------------------------------------
//  computeModifiedScore
// -----------------------------------------------------------------------
void TTestKolmogorovSmirnov::computeModifiedScore(
        URANIE::DataServer::TStochasticAttribute::ELawType tlaw)
{
    Double_t dsqrtn = TMath::Sqrt(1.0 * _nS);

    switch (_nCase)
    {
    case kCase0:    // Case 0
        _dModifiedScore = _dScore * (dsqrtn + 0.12 + 0.11 / dsqrtn);
        break;
    case kCase3:    // Case 3
    {
        // Set the Modified KolmogorovSmirnov value of the statistic
        switch (tlaw)
        {
        case URANIE::DataServer::TStochasticAttribute::kUniform:
        {
            _dModifiedScore = _dScore;
        }
            break;
        case URANIE::DataServer::TStochasticAttribute::kBeta:
        {
            // Set the Modified KolmogorovSmirnov value of the statistic
            _dModifiedScore = _dScore;
        }
        break;
        case URANIE::DataServer::TStochasticAttribute::kNormal:
        {
            _dModifiedScore = _dScore * (dsqrtn - 0.01 + 0.85 / dsqrtn);
        }
            break;
        case URANIE::DataServer::TStochasticAttribute::kLogNormal:
        {
            _dModifiedScore = _dScore * (dsqrtn - 0.10 + 0.85 / dsqrtn);
        }
            break;
        case URANIE::DataServer::TStochasticAttribute::kExponential:
        {
            _dModifiedScore = _dScore;
        }
            break;
        case URANIE::DataServer::TStochasticAttribute::kWeibull:
        {
            _dModifiedScore = _dScore;
        }
            break;
        case URANIE::DataServer::TStochasticAttribute::kGamma:
        {
            _dModifiedScore = _dScore;
        }
            break;
        default:
        {
             try
             {
                 throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                      Form("TTestKolmogorovSmirnov::computeModifiedScore ** Law not treated [%d] **", tlaw ));
             } catch (URANIE::Exceptions::UErrorExceptions& ue)
             {
               ue.printMessage();
               throw ue;
             }
        }
             break;
        }
    }
    default:
        break;
    }
}

} // Fin du namespace UncertModeler
} // Fin du namespace URANIE

