//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/**
 * @file   TTester.cxx
 * @author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 * @date   Tue Jan  6 10:26:57 2009
 *
 * @brief  Implement the TTester classe
 *
 *
 */
#include <iomanip>

// ROOT
#include "TH1D.h"
#include "TPad.h"
#include "TSystem.h"
#include "TGraph.h"
#include "TLegend.h"
#include "TMath.h"
#include "TLeaf.h"
#include "TRegexp.h"
#include "Math/SpecFuncMathCore.h"

// Uranie
#include "TTester.h"
#include "UExceptions.h"
#include "TStochAttribut.h"
#include "qvalue_A2_Case0.h"
#include "TWeibullDistribution.h"

ClassImp (URANIE::UncertModeler::TTester)

namespace URANIE
{
namespace UncertModeler
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
  TTester::TTester(URANIE::DataServer::TDataServer *tds, ETestType nTestType,
       const char *varexpinput, Option_t * option) :
    _nS(-1), _pValue(-1.0), _dScore(-1.0), _dModifiedScore(-1.0), _sAttributes(varexpinput),
    _listOfLaws(NULL), _bDrawSame(kTRUE), _nTestType(nTestType),
    _nCase(kCaseUnknown), _tds(tds), _blog(kFALSE), _bquiet(false)
{
    if (_blog)
    {
        cout << endl << "Builds TTester by Type tds[" << _tds->GetName()
                << "] Option [" << option << "]" << endl;
        cout << " ** ETestType[" << nTestType << "]" << endl;
        cout << " ** inputs[" << varexpinput << "]" << endl;
    }

    Int_t natt = _sAttributes.CountChar(':') + 1;

    try
    {
        if (natt != 1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTester::constructor   ERROR \n the number of attributes is different to 1 (%d)\n not yet implementd(%s)",
                            natt, _sAttributes.Data()));
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TTester::constructor ERROR : the pointer of \"tds\" is empty");
        if (_tds->getNPatterns() == 0)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TTester::constructor   ERROR \n The TTree is empty [%s] ",
                            GetName()));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    try
      {
  _tds->computeStatistic(varexpinput);
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
  ue.printMessage();
  throw ue;
      }

    TString soption = option ;
    soption.ToLower();

    if(soption.Contains("quiet"))
        _bquiet=true;
    
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TTester::~TTester()
{
    if (_blog)
        cout << " ** Destroy TTester Name[" << GetName() << "] Title["
                << GetTitle() << "]" << endl;
    if (_listOfLaws)
      {
  _listOfLaws->Delete();
        delete _listOfLaws;
      }
    _listOfLaws=NULL;

}

// -----------------------------------------------------------------------
//  computeTheScore
// -----------------------------------------------------------------------
Double_t TTester::computeScore(Option_t * option)
{
    if (_blog)
        cout << " ** TTester::computeScore option[" << option << "]" << endl;
    treatOption(option);

    Int_t nlaw = _listOfLaws->GetEntries();

    try
    {
        if (nlaw == 0)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TTester::computeScore   ERROR \n The list of laws to test is empty [%s] attributes[%s]\n The options are [%s]",
                            _tds->GetName(), _sAttributes.Data(), option));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    TString spar = _sAttributes;
    Bool_t bDraw = kFALSE;

    if (_bDrawSame) {
        if ( gPad != NULL ) {

            TH1F *htemp = (TH1F*) gPad->GetPrimitive(Form("hcdf%s", spar.Data()));

            try {
                if (!htemp)
                    throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                                 "TTester::computeScore method\n The current canvas gPad does not contain an CDF graph; we will draw the CDF.");
            } catch (URANIE::Exceptions::UInfoExceptions& ue)
            {
                if(!_bquiet)
                    ue.printMessage();
                if(_blog)
                    gPad->ls();
                bDraw = kTRUE;
            }
        } else {
            bDraw = kTRUE;
        }

    }

    if (_blog)
        cout << " ******************************" << endl;
    if ( _bDrawSame && bDraw ) {
        _tds->drawCDF(spar,"","legend");
    }
    if (_blog)
        cout << " ******************************" << endl;

    _tds->computeStatistic(spar);
    Double_t dsigma = _tds->getAttribute(spar)->getStd();
    Double_t dmu = _tds->getAttribute(spar)->getMean();
    Double_t dmin = _tds->getAttribute(spar)->getMinimum();
    Double_t dmax = _tds->getAttribute(spar)->getMaximum();

    _tds->draw(spar, "", "goff");
    _nS = _tds->getTuple()->GetSelectedRows();
    if (_blog)
        cout << " ** _nS[" << _nS << "]" << endl;
    Long64_t *index = new Long64_t[(Int_t) _nS];

    TMath::Sort(_nS, _tds->getTuple()->GetV1(), index, kFALSE);
    TLeaf *lx = _tds->getTuple()->GetLeaf(spar);

    Int_t ncrt = 0;
    TLeaf *law=NULL;
    TIter next(_listOfLaws);
    Double_t xcrt = 0., zi=0.0;
    Color_t colcrt = kBlack;
    URANIE::DataServer::TStochasticAttribute::ELawType tlaw=URANIE::DataServer::TStochasticAttribute::kUnknown;
    TVectorD paramValues(5);

    while ((law = (TLeaf*) next()))
    {
        _nCase = TTester::kCaseUnknown;

        TString slaw = law->GetName();
        dsigma = _tds->getAttribute(spar)->getStd();
        dmu = _tds->getAttribute(spar)->getMean();

        if (_blog) {
            cout << " *** law " << ncrt + 1 << "/" << nlaw << " :: slaw[" << slaw << "]";
            cout << "    mu (" << dmu << ") sigma(" << dsigma << ")" << endl;
        }

        //////////////////////////
        // Case of uniform law
        if (slaw.BeginsWith("uniform"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kUniform;
            if (slaw.Length() == 7)
            {
                dmu = _tds->getAttribute(spar)->getMinimum();
                dsigma = _tds->getAttribute(spar)->getMaximum();
                _nCase = TTester::kCase3;
            }
            else
            {
                TRegexp reg("uniform([+/-]?[0-9.]*,[+/-]?[0-9.]*)");
                TString out = slaw(reg);
                try
                {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form(
                                        "TTester::computeScore  ERROR \n The current law is not well defined [%s]",
                                        slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 8), paramValues);
                dmu = paramValues(0);
                dsigma = paramValues(1);
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Uniform Law :: Case(" << _nCase << ")  Min(" << dmu << ") Max(" << dsigma << ") ** ";
        }
        //////////////////////////
        // Case of normal law
        if (slaw.BeginsWith("normal"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kNormal;
            if (slaw.Length() == 6)
            {
                Int_t npat = _tds->getNPatterns();
                dmu = _tds->getAttribute(spar)->getMean();
                dsigma = TMath::Sqrt((1.0*(npat-1.0)/(1.0*npat))) *_tds->getAttribute(spar)->getStd();
                //dsigma = _tds->getAttribute(spar)->getStd();
                _nCase = TTester::kCase3;
            }
            else
            {
                TRegexp reg("normal([+/-]?[0-9.]*,[+]?[0-9.]*)");
                TString out = slaw(reg);
                try
                {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form(
                                        "TTester::computeScore  ERROR \n The current law is not well defined [%s]",
                                        slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 7), paramValues);
                dmu = paramValues(0);
                dsigma = paramValues(1);
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Normal Law :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** ";
        }
        //////////////////////////
        // Case of exponential law
        if (slaw.BeginsWith("exponential"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kExponential;
            if (slaw.Length() == 11)
            {
                Int_t npat = _tds->getNPatterns();
                dmu = _tds->getAttribute(spar)->getMean();
                dmin = _tds->getAttribute(spar)->getMinimum();

                dmu = ( 1.0 * npat * ( dmu - dmin ) ) / ( npat-1.0) ;
                dsigma = dmin - dmu / ( 1.0 * npat);

                _nCase = TTester::kCase3;
            }
            else
            {
                TRegexp reg("exponential([+/-]?[0-9.]*,[+]?[0-9.]*)");
                TString out = slaw(reg);
                try
                {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form(
                                        "TTester::computeScore  ERROR \n The current law is not well defined [%s]",
                                        slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 11), paramValues);
                dmu = 1.0 / paramValues(0);
                dsigma = paramValues(1);
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Exponential Law :: Case(" << _nCase << ")  mu(" << dmu << ") Beta(" << dsigma << ") ** ";
        }
        //////////////////////////
        // Case of lognormal law
        if (slaw.BeginsWith("lognormal"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kLogNormal;
            if (slaw.Length() == 9)
            {
                TString snewatt = Form("__uranie_att_LnNormal_%s", spar.Data());
                if( !_tds->isAttribute( snewatt ) )
                {
                    _tds->addAttribute(snewatt, Form("log(%s)", spar.Data()));
                    _tds->computeStatistic();
                    // _tds->delAttribute(snewatt);
                    _tds->getAttribute(snewatt)->setOrigin(URANIE::DataServer::TAttribute::kInternal);
                }
                _tds->computeStatistic(snewatt);
                Int_t npat = _tds->getNPatterns();
                dmu = _tds->getAttribute(snewatt)->getMean();
                dsigma = TMath::Sqrt((1.0*(npat-1.0)/(1.0*npat))) * _tds->getAttribute(snewatt)->getStd();
//                dsigma = _tds->getAttribute(snewatt)->getStd();
                _nCase = TTester::kCase3;
                paramValues(0) = TMath::Exp(dmu+dsigma*dsigma*0.5);
                paramValues(1) = TMath::Exp(1.6449*dsigma);
           }
            else
            {
                TRegexp reg("lognormal([+/-]?[0-9.]*,[+]?[0-9.]*)");
                TString out = slaw(reg);
                try
                {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form(
                                        "TTester::computeScore  ERROR \n The current law is not well defined [%s]",
                                        slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 10), paramValues);
                // Pass to the Normal law parameters
                dsigma = TMath::Log(paramValues(1)) / 1.6449;
                dmu    = TMath::Log(paramValues(0)) - dsigma*dsigma*0.5;
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Log-Normal law :: Case(" << _nCase << ")  mu_N(" << dmu << ") sigma_N(" << dsigma << ") -- "
                        "mu(" << paramValues(0) << ") EF(" << paramValues(1) << ") ** " << endl;
        }
        //////////////////////////
        // Case of Weibull law
        if (slaw.BeginsWith("weibull"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kWeibull;
            if (slaw.Length() == 7) {
                TString snewatt = Form("__uranie_att_weibull_%s", spar.Data());
                if( !_tds->isAttribute( snewatt ) ) {
                    _tds->addAttribute(snewatt, Form("-1.0*log(%s)", spar.Data()));
                    _tds->getAttribute(snewatt)->setOrigin(URANIE::DataServer::TAttribute::kInternal);
                }
                _tds->computeStatistic(snewatt);
                _nCase = TTester::kCase3;
                try {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                 Form("TTester::computeScore   ERROR \nWeibull law for [%s] estimation by MLE Not Yet implemented", spar.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    throw ue;
                }
                cout << endl << " *** Weibull Law by MLE :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** " << endl;
            } else {
                TRegexp reg("weibull([+/-]?[0-9.]*,[+]?[0-9.]*)");
                TString out = slaw(reg);
                try {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TTester::computeScore  ERROR \n The current law is not well defined [%s]", slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 7), paramValues);
                dmu = paramValues(0);
                dsigma = paramValues(1);
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Weibull law :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** " << endl;

            // Change  the order
            _tds->draw(spar, "", "goff");
            TMath::Sort(_nS, _tds->getTuple()->GetV1(), index, kTRUE);
        }
        //////////////////////////
        // Case of Gamma law
        if (slaw.BeginsWith("gamma"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kGamma;
            if (slaw.Length() == 5) {
                // Not MLE but Moment Method Estimation
                Double_t dparMean =  dmu;
                Double_t dparSigma2 = dsigma * dsigma;

                dmu = ( dparMean * dparMean ) / dparSigma2 ;
                dsigma = dparSigma2 / dparMean;

                _nCase = TTester::kCase3;

                cout << endl << " *** GAMMA Law by MME :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** " << endl;
            } else {
                TRegexp reg("gamma([+/-]?[0-9.]*,[+]?[0-9.]*)");
                TString out = slaw(reg);
                try {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TTester::computeScore  ERROR \n The current law is not well defined [%s]", slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 5), paramValues);
                dmu = paramValues(0);
                dsigma = paramValues(1);
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Gamma law :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** " << endl;

            // Change  the order
            _tds->draw(spar, "", "goff");
            TMath::Sort(_nS, _tds->getTuple()->GetV1(), index, kFALSE);
        }
        //////////////////////////
        // Case of Beta law
        if (slaw.BeginsWith("beta"))
        {
            tlaw = URANIE::DataServer::TStochasticAttribute::kBeta;
            if (slaw.Length() == 4)
            {
                dsigma = _tds->getAttribute(spar)->getStd();
                dmu = _tds->getAttribute(spar)->getMean();
                Double_t bufmu = ((1-dmu)/(dsigma*dsigma) - 1./dmu)*dmu*dmu;
                Double_t bufsi = bufmu*(1./dmu - 1);
                dmu = bufmu;
                dsigma = bufsi;
                _nCase = TTester::kCase3;
            } else {
                TRegexp reg("beta([+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*)");
                TString out = slaw(reg);

                try {
                    if (out.Length() == 0) {
                        TRegexp reg2("beta([+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*)");
                        TString out2 = slaw(reg2);

                        if (out2.Length() == 0)
                            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                                       __LINE__,
                                                                       Form("TTester::computeScore  ERROR \n The current law is not well defined [%s]",
                                                                            slaw.Data()));
                    }
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    // Clean
                    delete[] index;
                    index = NULL;
                    throw ue;
                }

                findParametersLaw(slaw.Remove(0, 5), paramValues);
                dmu = paramValues(0);
                dsigma = paramValues(1);
                if( paramValues.GetNrows()==4 ) {
                    dmin = paramValues(2);
                    dmax = paramValues(3);
                }
                _nCase = TTester::kCase0;
            }
            if (_blog)
                cout << endl << " ** Beta Law :: Case(" << _nCase << ")  Alpha(" << dmu << ") Beta(" << dsigma << ") ** ";
        }


        //////////////////////////

        if (_blog)
            cout << endl << " ** Current Law["  <<  tlaw << "] :: Case(" << _nCase << ")  mu(" << dmu << ") sigma(" << dsigma << ") ** ";

        // Initialize the score value
        _dScore = 0.0;

        // Loop
        for (Long64_t i = 0; i < _nS; i++) {

            _tds->getTuple()->GetEntry(index[i]);
            xcrt = lx->GetValue();

            switch (tlaw)
            {
            case URANIE::DataServer::TStochasticAttribute::kUniform:
            {
                zi = (xcrt - dmu) / (dsigma - dmu);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kNormal:
            {
                zi = ROOT::Math::gaussian_cdf(xcrt, dsigma, dmu);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kLogNormal:
            {
                zi = ROOT::Math::gaussian_cdf(TMath::Log(xcrt), dsigma, dmu);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kExponential:
            {
                zi = 1.0 - TMath::Exp ( -1.0 * ( xcrt - dsigma ) / dmu );
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kWeibull:
            {
                zi = TMath::Exp ( -1.0 * TMath::Exp ( ( TMath::Log(xcrt) - TMath::Log(dmu) ) * dsigma ));
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kGamma:
            {
                zi = ROOT::Math::inc_gamma(dmu, xcrt / dsigma);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kBeta:
            {
                zi = ROOT::Math::beta_cdf( (xcrt-dmin)/(dmax-dmin), dmu, dsigma);
                break;
            }
            default:
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TTester::computeScore ** Law [%d]  not treated **", tlaw ));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                }
                break;
            }
            }

            if (_blog)
                cout << endl << " ** i[" << i << "] xcrt[" << xcrt << "] zi["
                        << std::setw(1) << zi << "]";

            Double_t dval = computeTheCurrentScore((Int_t) i + 1, zi);
            switch (_nTestType)
            {
            case kKolmogorovSmirnov:
            {
                // Set the KolmogorovSmirnov value of the statistic
                _dScore = TMath::Max(_dScore, dval);
                break;
            }
            case kAndersonDarling:
            {
                _dScore -= 1.0 / (1.0 * _nS) * dval;
                break;
            }
            case kCramerVonMises:
            {
                _dScore += dval;
                break;
            }
            default:
                break;
            }

            if (_blog)
                cout << " dval[" << dval << "]  _dScore[" << _dScore << "]";
        } // End Of Loop on nS
        if (_blog)
            cout << endl;

        // Last operation
        switch (_nTestType)
        {
        case kCramerVonMises: //  Add 1.0/(12n)
            _dScore += 1.0 / (12.0 * _nS);
            break;
        case kAndersonDarling:   // Substract -nS
            _dScore -= 1.0 * _nS;
            break;
        default:
            break;
        }

        // Set the Modified value of the statistic
        computeModifiedScore(tlaw);

        // Compute the PValue
        computePValue(tlaw);

        if (_bDrawSame)
        {
            switch (tlaw)
            {
            case URANIE::DataServer::TStochasticAttribute::kUniform:
            {
                addUniformLaw(dmu, dsigma);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kNormal:
            {
                addNormalLaw(dmu, dsigma);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kLogNormal:
            {
                addLogNormalLaw(dmu, dsigma);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kWeibull:
            {
                addWeibullLaw(dmu, dsigma);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kGamma:
            {
                addGammaLaw(dsigma, dmu);
                break;
            }
            case URANIE::DataServer::TStochasticAttribute::kBeta:
            {
                addBetaLaw(dmu, dsigma, dmin, dmax);
                break;
            }
            default:
            {
                 try
                 {
                     throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                         Form("TTester::computeScore ** Law not treated **" ));
                 } catch (URANIE::Exceptions::UErrorExceptions& ue)
                 {
                    ue.printMessage();
                 }
                 break;
            }
            }
        }
        if (_blog)
            cout << endl << " ** " << getTitleStatistic() << " Test :  [" << _dScore
                << "], Modified value " << getTitleStatistic() << "*["
                << _dModifiedScore << "]" << endl;

        ncrt++;
    }

    // Clean
    delete[] index;
    index = NULL;

    return _dScore;
}

// -----------------------------------------------------------------------
//  computePValue_KS_CaseAll
// -----------------------------------------------------------------------
Double_t TTester::computePValue_KS_CaseAll(Double_t dz)
{
    if ( _blog )
        cout << " *** computePValue_KS_CaseAll function : dz[" << dz << "]" << endl;

    Int_t nMAXSUM = 100;
    Double_t dEPS1 = 1.0E-3;
    Double_t dEPS2 = 1.0E-8;
    Double_t dTWO = 2.0;
    if ( _blog ) {
        cout << " *** global parameters" << endl;
        cout << " *** nMAXSUM[" << nMAXSUM << "]" << endl;
        cout << " *** dEPS1[" << dEPS1 << "]" << endl;
        cout << " *** dEPS2[" << dEPS2 << "]" << endl;
    }

    Double_t dCrt = 0.0;
    Double_t dCrtAbs = 0.0;
    Double_t dSum = 0.0;

    // use stat :  D^{\star}
    Double_t dval2 = -2.0 * dz * dz;
    if ( _blog )
        cout << " *** dval2[" << dval2 << "]" << endl;

    for(Int_t i=1; i<nMAXSUM; i++) {
        dCrt = dTWO * TMath::Exp( dval2 * i * i);
        dSum += dCrt;
        if ( (TMath::Abs(dCrt) <= dEPS1 * dCrtAbs) ||
             (TMath::Abs(dCrt) <= dEPS2 * dSum) )
            break;
        dTWO *= -1.0;
        dCrtAbs = TMath::Abs(dCrt);
    }

    dSum = TMath::Min(1.0, dSum);

    if ( _blog )
        cout << " *** dSum[" << dSum << "]" << endl;

    return dSum;
}

// -----------------------------------------------------------------------
//  computePValue_AD_Case0
// -----------------------------------------------------------------------
Double_t TTester::computePValue_AD_Case0(Double_t dz)
      {
    if ( _blog )
        cout << " *** computePValue_AD_Case0 function : dz[" << dz << "] loop over[" << QVALUE_A2_CASE0_nPattern << "]" << endl;

    // Looking for the first index where "dz < z_value" in the array of reference values
    Int_t iCrt = 0;
    for(iCrt=0; iCrt< QVALUE_A2_CASE0_nPattern; iCrt++) {
        if ( dz < z_QVALUE_A2_CASE0[iCrt]) {
            break;
        }
    }

    if ( _blog )
        cout << " *** dz[" << dz << "] iCrt[" << iCrt << "/" << QVALUE_A2_CASE0_nPattern << "] z_[" << z_QVALUE_A2_CASE0[iCrt] << "]"<< endl;

    // Default value in case iCrt==0
    Double_t dCrt = 1.0;

    if ( iCrt > 1 ) {
        if ( iCrt < QVALUE_A2_CASE0_nPattern - 1 ) {
            if ( _blog )
                cout << " *** dz[" << dz << "] iCrt[" << iCrt << "] z_[" << z_QVALUE_A2_CASE0[iCrt-1] << "]("<< pvalue_QVALUE_A2_CASE0[iCrt-1] <<") z_[" << z_QVALUE_A2_CASE0[iCrt] << "]("<< pvalue_QVALUE_A2_CASE0[iCrt] <<")"<< endl;
            Double_t dSlope = ( pvalue_QVALUE_A2_CASE0[iCrt] - pvalue_QVALUE_A2_CASE0[iCrt-1] ) / ( z_QVALUE_A2_CASE0[iCrt] - z_QVALUE_A2_CASE0[iCrt-1] );
            if ( _blog )
                cout << " dslope[" << dSlope << "]" << endl;
            dCrt = pvalue_QVALUE_A2_CASE0[iCrt-1] + dSlope * ( dz - z_QVALUE_A2_CASE0[iCrt-1]);
            if ( _blog )
                cout << " dCrt[" << dCrt << "]" << endl;
        } else {
            // case of dz > 8.0
            dCrt = 1.0e-4;
        }
    }

    return dCrt;
}

// -----------------------------------------------------------------------
//  computePValue_AD_Case3
// -----------------------------------------------------------------------
Double_t TTester::computePValue_AD_Case3(Double_t dz, URANIE::DataServer::TStochasticAttribute::ELawType tlaw)
      {
    if ( _blog )
        cout << " *** computePValue_AD_Case3 function : dz[" << dz << "] law type [" << tlaw << "]" << endl;
    Int_t nCut = 3;
    TVectorD vecCut(nCut);

    // Default value
    Double_t dres = -1.0;
    Double_t dCrt;

    Int_t iCrt=-1;

    switch (tlaw) {
    case URANIE::DataServer::TStochasticAttribute::kNormal:
    case URANIE::DataServer::TStochasticAttribute::kLogNormal: {
        // Table 4.9 page 127 of the "Goodness-of-Fit Techniques" book
        vecCut(0) = 0.200;
        vecCut(1) = 0.340;
        vecCut(2) = 0.600;

        iCrt = nTimesValueLTvalVector(dz, vecCut);

        switch (iCrt) {
        case 4:
            dres = TMath::Exp( 1.2937 - 5.709 * dz + 0.0186 * dz * dz);
            break;
        case 3:
            dres = TMath::Exp( 0.9177 - 4.279 * dz - 1.38 * dz * dz);
            break;
        case 2:
            dCrt = TMath::Exp( -8.318 + 42.796 * dz - 59.938 * dz * dz);
            dres = 1.0 - dCrt;
            break;
        case 1:
            dCrt = TMath::Exp( -13.436 + 101.14 * dz - 223.73 * dz * dz);
            dres = 1.0 - dCrt;
            break;
        default:
            break;
        }
    }
    default:
        break;
      }

    return dres;
}

// -----------------------------------------------------------------------
//  computePValue_CvM_Case3
// -----------------------------------------------------------------------
Double_t TTester::computePValue_CvM_Case3(Double_t dz, URANIE::DataServer::TStochasticAttribute::ELawType tlaw)
{
    if ( _blog )
        cout << " *** computePValue_CvM_Case3 function : dz[" << dz << "] law type [" << tlaw << "]" << endl;
    Int_t nCut = 3;
    TVectorD vecCut(nCut);

    // Default value
    Double_t dres = -2.0;
    Double_t dCrt;

    Int_t iCrt=-1;

    switch (tlaw) {
    case URANIE::DataServer::TStochasticAttribute::kNormal:
    case URANIE::DataServer::TStochasticAttribute::kLogNormal: {
        // Table 4.9 page 127 of the "Goodness-of-Fit Techniques" book
        vecCut(0) = 0.0275;
        vecCut(1) = 0.051;
        vecCut(2) = 0.092;

        iCrt = nTimesValueLTvalVector(dz, vecCut);

        switch (iCrt) {
        case 4:
            dres = TMath::Exp( 1.11 - 34.242 * dz + 12.832 * dz * dz);
            break;
        case 3:
            dres = TMath::Exp( 0.886 - 31.62 * dz + 10.897 * dz * dz);
            break;
        case 2:
            dCrt = TMath::Exp( -5.903 + 179.546 * dz - 1515.29 * dz * dz);
            dres = 1.0 - dCrt;
            break;
        case 1:
            dCrt = TMath::Exp( -13.953 + 775.5 * dz - 12542.61 * dz * dz);
            dres = 1.0 - dCrt;
            break;
        default:
            break;
        }
    }
    default:
        break;
    }

    return dres;

}

// -----------------------------------------------------------------------
//  nTimesValueLTvalVector
// -----------------------------------------------------------------------
Int_t TTester::nTimesValueLTvalVector(Double_t dval, TVectorD vecValues)
{
    if ( _blog ) {
        cout << " *** nTimesValueLTVector function : dval[" << dval << "] size of vector [" << vecValues.GetNrows() << "]" << endl;
        vecValues.Print(Form("Case of test [%s] on att[%s] in the nTimesValueLTvalVector function", getTitleStatistic(), _sAttributes.Data()));
    }

    Int_t nValues = vecValues.GetNrows();
    Int_t iCrt=nValues+1;

    for(Int_t j=0; j<nValues; j++) {
        // Number of times than dval LT vecValues(j)
        if ( dval < vecValues(j) )
            iCrt--;
    }
    if ( _blog )
        cout << " ** End Of Loop in function nTimesValueLTVector :: icrt(" << iCrt << ") dval[" << dval << "] vecValues[" <<  vecValues(iCrt-2) << "]" <<endl;

    return iCrt;
}


// -----------------------------------------------------------------------
//  computePValue
// -----------------------------------------------------------------------
void TTester::computePValue(URANIE::DataServer::TStochasticAttribute::ELawType tlaw)
    {
    if ( _blog ) {
        cout << " *** computePValue with tlaw[" << tlaw << "] case[" << _nCase << "] for attribute[" << _sAttributes << "]" << endl;
        cout << " *** Test type[" << _nTestType << "] Title[" << getTitleStatistic() << "]" << endl;
        cout << " *** Score [" << _dScore << "] Modified[" << _dModifiedScore << "]" << endl;
    }
    _pValue = -2.0;

    switch (_nTestType) {
    case kKolmogorovSmirnov: {
        // Case Of KolmogorovSmirnov Test ("KS")
        _pValue = computePValue_KS_CaseAll(_dModifiedScore);
        break;
    }
    case kAndersonDarling : {
        // Case Of AndersonDarling Test ("AD")
        switch (_nCase) {
        case kCase0: {
            // Case of the parameters of the law are given
            _pValue = computePValue_AD_Case0(_dScore);
            break;
        }
        case kCase3: {
            // Case of the parameters of the law are estimated by ML
            _pValue = computePValue_AD_Case3(_dModifiedScore, tlaw);
            break;
        }
        default:
            break;
        }
        break;
    }
    case kCramerVonMises : {
        // Case Of CramerVonMises Test ("CvM")
        switch (_nCase) {
        case kCase3: {
            // Case of the parameters of the law are estimated by ML
            _pValue = computePValue_CvM_Case3(_dModifiedScore, tlaw);
            break;
        }
        default:
            break;
        }
        break;
    }
    default:
        break;
    }

    if ( _blog )
        cout << " *** PValue -------- [" << _pValue << "]" << endl;
}

// -----------------------------------------------------------------------
//  Adds Uniform law defined by the (min, max) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addUniformLaw(Double_t min, Double_t max, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for (Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt = 0.0;
            if ( i > 0 )
                ycrt = (xcrt - min) / (max - min);
            ycrt = TMath::Max(0.0, ycrt);
            ycrt = TMath::Min(1.0, ycrt);
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ U (%3.3g, %3.3g) ] = %.4f", getTitleStatistic(), min, max, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Adds Normal law defined by the (mu, sigma) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addNormalLaw(Double_t mu, Double_t sigma, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for (Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt = 0.0;
            if ( i > 0 )
                ycrt = ROOT::Math::gaussian_cdf(xcrt, sigma, mu);
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ N (%3.3g, %3.3g) ] = %.4f", getTitleStatistic(), mu, sigma, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Adds LogNormal law defined by the (mu, sigma) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addLogNormalLaw(Double_t mu, Double_t sigma, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for ( Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt = 0.0;
            if ( i > 0 )
                ycrt = ROOT::Math::gaussian_cdf(TMath::Log(xcrt), sigma, mu);
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        // Pass to the LogNormal law parameters
        if (_blog)
            cout << endl << " *** Log-Normal Law by addLogNormalLaw ::  mu_N(" << mu << ") sigma_N(" << sigma << ") -- ";
        mu = TMath::Exp(mu+sigma*sigma*0.5);
        sigma = TMath::Exp(1.6449*sigma);
        if (_blog)
            cout << "mu(" << mu << ") EF(" << sigma << ") ** " << endl;
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ LN (%3.3g, %3.3g) ] = %.4f", getTitleStatistic(), mu, sigma, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Adds Beta law defined by the (mu, sigma, min, max) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addBetaLaw(Double_t mu, Double_t sigma, Double_t min, Double_t max, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for (Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt =  ROOT::Math::beta_cdf( (xcrt-min)/(max-min), mu, sigma);
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ Be (%3.3g, %3.3g, %3.3g, %3.3g) ] = %.4f", getTitleStatistic(), mu, sigma, min, max, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Adds Weibull law defined by the (mu, sigma) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addWeibullLaw(Double_t lambda, Double_t k, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for ( Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt = 0.0;
            if ( i > 0 )
                ycrt = 1.0 - TMath::Exp ( -1.0 * TMath::Power ( xcrt / lambda , k ));
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ W (%3.3g, %3.3g) ] = %.4f", getTitleStatistic(), lambda, k, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Adds Gamma law defined by the (alpha, beta) parameters and the color of the line
// -----------------------------------------------------------------------
void TTester::addGammaLaw(Double_t beta, Double_t alpha, Color_t color, TString stitle)
{
    try {
        TGraph *grCrt = createGraph(color);

        TString sGraphName = Form("hcdf%s",_sAttributes.Data());
        TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);

        Double_t xcrt = 0., ycrt = 0.0;
        for ( Int_t i = 0; i <= htemp->GetNbinsX(); i++)
        {
            xcrt = htemp->GetBinCenter(i);
            ycrt = 0.0;
            if ( i > 0 )
                ycrt = ROOT::Math::inc_gamma(alpha, xcrt / beta);
            grCrt->SetPoint(i, xcrt, ycrt);
        }

        grCrt->Draw("lpsame");

        TLegend *leg = (TLegend*) gPad->GetPrimitive("TPave");
        if (stitle.Length() == 0) {
            TString sEntryTitle = Form("%s [ Ga. (%3.3g, %3.3g) ] = %.4f", getTitleStatistic(), alpha, beta, _dScore);
            if ( _pValue > 0.0 )
                sEntryTitle += Form(", Pval=%.4f", _pValue);
            leg->AddEntry(grCrt, sEntryTitle, "Lp");
        } else {
            leg->AddEntry(grCrt, stitle, "Lp");
        }
    }   catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Create the TGraph Object
// -----------------------------------------------------------------------
TGraph *TTester::createGraph(Color_t color)
{
    TString sGraphName = Form("hcdf%s",_sAttributes.Data());

    TH1F *htemp = (TH1F*) gPad->GetPrimitive(sGraphName);
    try
    {
        if( !htemp )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TTester::computeScore method\n Cannot get the histrogram %s in current pad", sGraphName.Data()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        gPad->GetListOfPrimitives()->ls();
        throw ue;
    }

    TGraph * gr = new TGraph(htemp->GetNbinsX());
    _tds->_drawingGarbageCollector->Add(gr);

    stringstream sstr;
    Int_t i=0;
    sstr << "Graph_" << i;
    while ( gPad->FindObject( sstr.str().c_str() ) != 0)
    {
        i++;
        sstr.str("");
        sstr << "Graph_" << i;
    }

    gr->SetName( sstr.str().c_str() );

    if ( _blog )
        cout << " *** init color(" << color << ") i(" << i << ")" << endl;
    if(color==kBlack) {
        color++;
        if ( i > 0 ) {
            TGraph *grLast =  (TGraph*) gPad->FindObject(Form("Graph_%d", i-1));
            color = grLast->GetLineColor()+1;
        }
        if ( ( color==5) || ( color==10) )
            color++;
    }
    if ( _blog )
        cout << " ** LAST color(" << color << ")" << endl;
    gr->SetLineColor(color);
    gr->SetMarkerColor(color);

    Width_t nwidth = 2;
    gr->SetLineWidth(nwidth);

    return gr;
}

// -----------------------------------------------------------------------
//  Return the parameters of the law contained in the string on the vector
// -----------------------------------------------------------------------
void TTester::findParametersLaw(TString soption, TVectorD &vec)
{
    if (_blog)
        cout << endl << " ***********************" << endl
             << " *** soption[" << soption << "]" << endl
             << " ***********************" << endl;
    Int_t nparams = soption.CountChar(',') + 1;
    vec.ResizeTo(nparams);
    soption.ReplaceAll(")", ",").ReplaceAll("(", "");

    vec(0) = 12.36;
    Ssiz_t nbegin = 0;

    for (Int_t i = 0; i < nparams; i++)
    {
        Ssiz_t nend = soption.Index(',', nbegin);
        TString scrt = soption(nbegin, nend);
        vec(i) = atof(scrt.Data());

        nbegin = nend + 1;
    }
    if (_blog)
        vec.Print();
}

// -----------------------------------------------------------------------
//  Treats the option of the computeScoresMethod
// -----------------------------------------------------------------------
void TTester::treatOption(Option_t *option)
{
  if (_listOfLaws)
    {
      _listOfLaws->Delete();
      delete _listOfLaws;
    }
  _listOfLaws = new TList();

  TString soption(option);
  soption.ToLower();
  soption.ReplaceAll(" ", "");
  soption += TString(" ");
  soption.ReplaceAll(":", " ");

  istringstream inStream((char *) soption.Data());
  string inName;
  Int_t ncrt = 0;
  Bool_t bFound = kFALSE;
  while (inStream >> inName)
  {
      bFound = kFALSE;
      try
      {
          TString svar(inName.c_str());
          if ( _blog )
              cout << " ** svar[" << svar << "]" << endl;
          if ( svar.BeginsWith("normal", TString::kExact) ||
               svar.BeginsWith("lognormal", TString::kExact) ||
               svar.BeginsWith("uniform", TString::kExact) ||
               svar.BeginsWith("exponential", TString::kExact) ||
               svar.BeginsWith("gamma", TString::kExact) ||
               svar.BeginsWith("beta", TString::kExact) ||
               svar.BeginsWith("weibull", TString::kExact)
               ) {

              _listOfLaws->Add( new TNamed(svar.Data(), Form("law number %d", ncrt++)));
              bFound = kTRUE;
          }
          if (svar.BeginsWith("same", TString::kExact))
          {
              _bDrawSame = kTRUE;
              bFound = kTRUE;
          }
          if (svar.BeginsWith("noplot", TString::kExact))
          {
              //_bDraw = kFALSE;
              bFound = kTRUE;
          }

          if (!bFound)
              throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                           Form(
                                                               "TTester::treatOption\n Option [%s] not recognized in the list of options [%s].",
                                                               svar.Data(), soption.Data()));

      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
          ue.printMessage();
          throw ue;
      }
  }
  if (_blog)
      _listOfLaws->Print();

}

// -----------------------------------------------------------------------
//  Log printing
// -----------------------------------------------------------------------
void TTester::printLog(Option_t *option)
{
    cout << endl << "************** TTester::printLog[" << option << "] Name["
            << GetName() << "] Title[" << GetTitle() << "] ID[" << getID()
            << "]" << endl;
    cout << " **   TDataServer Log (unactive)" << endl;
//     _tds->printLog();
    cout << " **   _sAttributes[" << _sAttributes << "]" << endl;
    cout << " **   _nTestType[" << _nTestType << "] statistic["
         << getTitleStatistic() << "] case[" << _nCase << "]" << endl;
    cout << " **   _nS[" << _nS << "]" << endl;
    cout << " **   _dScore[" << _dScore << "] _dModifiedScore["
            << _dModifiedScore << "]" << endl;
    cout << " **   _pValue[" << _pValue << "]" << endl;
    cout << " ** List Of Laws" << endl;
    _listOfLaws->Print();
    cout << "******** TTester::End of printLog[]" << endl;
}

} // Fin du namespace UncertModeler
} // Fin du namespace URANIE

