//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTCirce.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Test File of the class URANIE::DataServer::TCirce
 
 Created: 2011-12-12 09:48:07

 */

#define DBL_PRECISION_COMPARE 1.e-12
#define SPL_PRECISION_COMPARE 1.e-8

#include "TestTCirce.h"
#include "TString.h"
#include "uranie_commands.h"

#include <iostream>
using namespace std;

TCirceTest::TCirceTest()
{
}

TCirceTest::~TCirceTest()
{
    UranieTests::removeFile(gSystem->pwd(), "jdd_circe_cas1.root");
}


////////////////////////////////
///                          ///
///         setUp            ///
///                          ///
////////////////////////////////

void TCirceTest::setUp()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "the TDS 1");
    tds1->fileDataRead("jdd_circe_summerschool2006_dataserver.dat");

    tds2 = new URANIE::DataServer::TDataServer();
}

////////////////////////////////
///                          ///
///         tearDown         ///
///                          ///
////////////////////////////////

void TCirceTest::tearDown()
{
    delete tds1;
    tds1 = NULL;
    delete tds2;
    tds2 = NULL;
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TCirceTest::testGetNTDS()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();
    CPPUNIT_ASSERT_EQUAL(1, tc->getNTDS());
    delete tc;
    tc = NULL;

}

void TCirceTest::testGetNPatterns()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    CPPUNIT_ASSERT_EQUAL(0, tc->getNPatterns());
    tc->init();

    CPPUNIT_ASSERT_EQUAL(150, tc->getNPatterns());
    delete tc;
    tc = NULL;

}

void TCirceTest::testGetNParameters()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    CPPUNIT_ASSERT_EQUAL(2, tc->getNParameters());
    delete tc;
    tc = NULL;

}

void TCirceTest::testGetTolerance()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0e-5, tc->getTolerance(),
            DBL_PRECISION_COMPARE);

    Double_t dtol = 1e-3;
    tc->setTolerance(dtol);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dtol, tc->getTolerance(),
            DBL_PRECISION_COMPARE);

    dtol = -1e-3;
    CPPUNIT_ASSERT_THROW(tc->setTolerance(dtol),
            URANIE::Exceptions::UErrorExceptions);

    delete tc;
    tc = NULL;

}

void TCirceTest::testGetNCMatrix()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    CPPUNIT_ASSERT_EQUAL(1, tc->getNCMatrix());
    Int_t nInitCMatrix = 5;
    tc->setNCMatrix(nInitCMatrix);

    CPPUNIT_ASSERT_EQUAL(nInitCMatrix, tc->getNCMatrix());
    delete tc;
    tc = NULL;

}

void TCirceTest::testSetBVectorInitial()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    TVectorD initBVec(2);
    initBVec(0) = -1.436394;
    initBVec(1) = -1.501561;
    tc->setBVectorInitial(initBVec);

    TVectorD vBiais = tc->getBVectorInitial();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.436394, vBiais(0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.501561, vBiais(1), DBL_PRECISION_COMPARE);

    tc->estimate();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.72955933310684, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;
}

void TCirceTest::testSetCMatrixInitial()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    TMatrixD initCMat(2, 2);
    initCMat.Zero();
    initCMat(0, 0) = 0.142737;
    initCMat(1, 1) = 0.525673;
    tc->setCMatrixInitial(initCMat);

    CPPUNIT_ASSERT_EQUAL(1, tc->getNCMatrix());

    TMatrixD matD = tc->getCMatrixInitial();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.142737, matD(0, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, matD(0, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.525673, matD(1, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, matD(1, 0), DBL_PRECISION_COMPARE);

    tc->estimate();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.7295593331088, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;
}

void TCirceTest::testGetSeed()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    CPPUNIT_ASSERT_EQUAL(1, tc->getNCMatrix());
    Int_t nSeed = 145236789;
    tc->setSeed(nSeed);

    CPPUNIT_ASSERT_EQUAL(nSeed, tc->getSeed());
    delete tc;
    tc = NULL;

}

void TCirceTest::testGetLikelihood()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;

}

void TCirceTest::testGetInputFile()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    TString sJDD = "jdd_circe_summerschool2006_dataserver.dat";

    tc->setInputFile(sJDD);
    TString sfile = gSystem->BaseName(tc->getInputFile());

    CPPUNIT_ASSERT(!strcmp(sfile, sJDD));

    delete tc;
    tc = NULL;
}

void TCirceTest::testEstimate()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    tc->estimate();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.729559333111, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    TVectorD vBiais = tc->getBVector();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.013170460679, vBiais(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.010615495684, vBiais(1),
            DBL_PRECISION_COMPARE);

    TMatrixD matC = tc->getCMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.016122246249, matC(0, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, matC(0, 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.036160164331, matC(1, 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, matC(1, 0), DBL_PRECISION_COMPARE);

    Int_t nInitCMatrix = 5;
    tc->setNCMatrix(nInitCMatrix);
    tc->estimate();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.729559333109, tc->getLikelihood(),
            SPL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;

}

void TCirceTest::testEstimateConfidenceIntervals1()
{
    tds2->fileDataRead("jdd_circe_cas1.dat");

    Int_t nPat = tds2->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(123, nPat);

    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds2, "Yobs", "Ycal", "Sens1");
    tc->unsetLog();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    tc->estimate();

    Int_t nPar = tc->getNParameters();
    CPPUNIT_ASSERT_EQUAL(1, nPar);

    TMatrixD matC = tc->getCMatrix();
    CPPUNIT_ASSERT_EQUAL(nPar, matC.GetNrows());
    CPPUNIT_ASSERT_EQUAL(nPar, matC.GetNcols());
    Double_t dSigma2 = matC(0, 0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.099001530734, dSigma2, DBL_PRECISION_COMPARE);

    TMatrixD matMu  = tc->getMatrixVarianceMu();
    CPPUNIT_ASSERT_EQUAL(nPar, matMu.GetNrows());
    CPPUNIT_ASSERT_EQUAL(nPar, matMu.GetNcols());
    Double_t dVarMu = matMu(0, 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dSigma2 / ( 1.0 * nPat), dVarMu, DBL_PRECISION_COMPARE);

    TMatrixD matSigma2 = tc->getMatrixVarianceSigma();
    CPPUNIT_ASSERT_EQUAL(nPar, matSigma2.GetNrows());
    CPPUNIT_ASSERT_EQUAL(nPar, matSigma2.GetNcols());
    Double_t dVarSigma2 = matSigma2(0, 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0 * dSigma2 * dSigma2 / ( 1.0 * nPat), dVarSigma2, DBL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;

}

void TCirceTest::testEstimateWithUncertainties()
{
    tds1->addAttribute("exp_uncert", "0.05*exp");
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2", "exp_uncert");
    tc->unsetLog();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    tc->estimate();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-2.72203731672689, tc->getLikelihood(),
            DBL_PRECISION_COMPARE);

    delete tc;
    tc = NULL;
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////

void TCirceTest::throwExceptionsConstructor()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TCirce(tds1, "expexp", "code",
                    "sens1,sens2"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TCirce(tds1, "exp", "codecode",
                    "sens1,sens2"), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TCirce(tds1, "exp", "code", ""),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TCirce(tds1, "exp", "code",
                    "sens1,sens2", "uncert"),
            URANIE::Exceptions::UErrorExceptions);
            
}

void TCirceTest::throwExceptionsAddData()
{
    URANIE::UncertModeler::TCirce * tc = new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2");
    tc->unsetLog();

    tds2->fileDataRead("jdd_circe_summerschool2006_dataserver.dat");
    tc->addData(tds2, "exp", "code", "sens1,sens2");

    delete tc;

}

void TCirceTest::throwExceptionsSetSelect()
{
    tds1->setSelect(Form("%s <10",tds1->getIteratorName() ));

        CPPUNIT_ASSERT_THROW(new URANIE::UncertModeler::TCirce(tds1,
            "exp", "code", "sens1,sens2"),URANIE::Exceptions::UErrorExceptions);
}


