//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTTestKolmogorovSmirnov.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$

 \brief  Test File of the class URANIE::DataServer::TTestKolmogorovSmirnov

 Created: wen nov 21 07:59:24 2007

 \par Modifications
 \li M. Salvoldi 18/01/2012
 - add 8 news unit tests
 */

#define DBL_PRECISION_COMPARE 1.e-12
#define DBL_PRECISION_COMPARE2 1.e-3
#include <iostream>
using namespace std;

// Data Header
#include "bliss.h"
#include "youngsmodulus.h"

//ROOT
#include "TString.h"

// URANIE
#include "TTester.h"
#include "TestTTestKolmogorovSmirnov.h"
#include "TUniformDistribution.h"
#include "TSampling.h"

TTestKolmogorovSmirnovTest::TTestKolmogorovSmirnovTest() :
        _tdsBliss(NULL), _tdsYoungModulus(NULL), _tks(NULL)
{
}

TTestKolmogorovSmirnovTest::~TTestKolmogorovSmirnovTest()
{
}

void TTestKolmogorovSmirnovTest::setUp()
{
    TString svar = "w";
    _tdsBliss = new URANIE::DataServer::TDataServer();
    _tdsBliss->addAttribute(svar);
    _tdsBliss->createTuple();

    for (Int_t i = 0; i < bliss_nPattern; i++)
        _tdsBliss->getTuple()->Fill(i, w[i]);
    _tdsBliss->computeStatistic(svar);

    svar = "E";
    _tdsYoungModulus = new URANIE::DataServer::TDataServer();
    _tdsYoungModulus->addAttribute(svar);
    _tdsYoungModulus->createTuple();

    for (Int_t i = 0; i < youngsmodulus_nPattern; i++)
        _tdsYoungModulus->getTuple()->Fill(i, E[i]);
    _tdsYoungModulus->computeStatistic(svar);
}

void TTestKolmogorovSmirnovTest::tearDown()
{
    delete _tks;
    _tks = NULL;
    delete _tdsBliss;
    _tdsBliss = NULL;
    delete _tdsYoungModulus;
    _tdsYoungModulus = NULL;
}

void TTestKolmogorovSmirnovTest::testKSBlissCase3()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();
    Double_t dks   = _tks->computeScore("normal");
    Double_t dksm  = _tks->getModifiedScore();

    //ORIGIN: Goodness-of-fit techniques D'Agostino et Stephens (page 125)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.108, dks, 1.e-3);  // 0.104 without MLE
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.504, dksm, 1.e-3); // 0.483
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase3 for Normal and Bliss",
                                 URANIE::UncertModeler::TTester::kCase3, _tks->getCase());

    dks = _tks->computeScore("uniform");
    dksm = _tks->getModifiedScore();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dks, 0.160526315789, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dksm, 0.160526315789, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase3 for Uniform and Bliss",
                                 URANIE::UncertModeler::TTester::kCase3, _tks->getCase());
}

void TTestKolmogorovSmirnovTest::testKSBlissCase0()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();
    Double_t dks = _tks->computeScore("normal(200,35)");
    Double_t dksm = _tks->getModifiedScore();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.171, dks, 1.e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.790, dksm, 1.e-3);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase0 for Normal and Bliss",
                                 URANIE::UncertModeler::TTester::kCase0, _tks->getCase());
}

void TTestKolmogorovSmirnovTest::testUniformValues()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(new URANIE::DataServer::TUniformDistribution("u", -5.0, 2.0));
    tds->getAttribute("u")->delShare();
    URANIE::Sampler::TSampling *fsamp = new URANIE::Sampler::TSampling(tds, "lhs", 10000);
    fsamp->generateSample();

    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(tds, "u");
    _tks->unsetLog();

    Double_t dks = _tks->computeScore("uniform");
    Double_t dksm = _tks->computeScore("uniform(-5.00000,2.00000)");
    //Double_t dksm2 = _tks->computeScore("uniform(-5.00000,-2.00000)");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dks, 0.000000, DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dksm, 0.000000, DBL_PRECISION_COMPARE2);

    delete fsamp;
    delete tds;
    tds = NULL;

}

void TTestKolmogorovSmirnovTest::testKSYoungsModulusNormalCase3Case0()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsYoungModulus, "E");
    _tks->unsetLog();
    Double_t dks = _tks->computeScore("normal");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dks, 0.1018, .003);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase3 for Normal and YoungsModulus",
                                 URANIE::UncertModeler::TTester::kCase3, _tks->getCase());

    dks = _tks->computeScore("normal(29576, 1507)");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dks, 0.1018, .003);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase0 for Normal and YoungsModulus",
                                 URANIE::UncertModeler::TTester::kCase0, _tks->getCase());
}

void TTestKolmogorovSmirnovTest::testKSYoungsModulusLogNormalCase3Case0()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsYoungModulus, "E");
    _tks->unsetLog();

    Double_t dks = _tks->computeScore("lognormal");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.093186762577, dks, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase3 for LogNormal and YoungsModulus",
                                 URANIE::UncertModeler::TTester::kCase3, _tks->getCase());

    dks = _tks->computeScore("lognormal(29563.63979723,    1.08750915245)");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0899763559277472, dks, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Case of kCase0 for LogNormal and YoungsModulus",
                                 URANIE::UncertModeler::TTester::kCase0, _tks->getCase());
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovConstructorTdsNull()
{
    URANIE::DataServer::TDataServer *tdsNull = NULL;

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestKolmogorovSmirnov(tdsNull, "w"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovConstructorTdsEmpty()
{
    URANIE::DataServer::TDataServer *tdsEmpty =
            new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestKolmogorovSmirnov(tdsEmpty, "w"),
            URANIE::Exceptions::UWarningExceptions);

    delete tdsEmpty;

}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovConstructorMoreAttribut()
{
    _tdsBliss->addAttribute("E");

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w:E"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovConstructorFalseAttribut()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss,
                    "toto"), URANIE::Exceptions::UErrorExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovcomputeScoreNoLaw()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();

    CPPUNIT_ASSERT_THROW(_tks->computeScore(),
            URANIE::Exceptions::UWarningExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovcomputeScoreBadLaw()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();
    CPPUNIT_ASSERT_THROW(_tks->computeScore("invgamma"), URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(_tks->computeScore("qsdfg"), URANIE::Exceptions::UWarningExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovcomputeScoreNormalBadDefined()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();

    CPPUNIT_ASSERT_THROW(_tks->computeScore("normal()"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tks->computeScore("normal(125.0)"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tks->computeScore("normal(125.0, 2.369, 12.02)"),
            URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_tks->computeScore("normal(125.0,-12.0)"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovcomputeScoreUniformBadDefined()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();

    CPPUNIT_ASSERT_THROW(_tks->computeScore("uniform()"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestKolmogorovSmirnovTest::throwExceptionTestKolmogorovSmirnovcomputeScoreLogNormalBadDefined()
{
    _tks = new URANIE::UncertModeler::TTestKolmogorovSmirnov(_tdsBliss, "w");
    _tks->unsetLog();

    CPPUNIT_ASSERT_THROW(_tks->computeScore("lognormal()"),
            URANIE::Exceptions::UErrorExceptions);
}
