//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.6 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TComposeMulti.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2011-11-24.15.37.54

 \brief   composition de fonction avec une fonction multiple
 */

#include "TComposeMulti.h"

ClassImp (URANIE::Relauncher::TComposeMulti);

namespace URANIE
{
namespace Relauncher
{

// Class TComposeMulti
TComposeMulti::TComposeMulti(TEval *fun)
{
    _multi = new TSubMulti(fun);
    _before = new TList();
    _before->SetOwner();
    _after = new TList();
    _after->SetOwner();
}

TComposeMulti::~TComposeMulti()
{
    delete _before;
    _before = NULL;
    delete _multi;
    _multi = NULL;
    delete _after;
    _after = NULL;
}

void TComposeMulti::addBeforeEval(TEval *un)
{
    _before->Add(new TSubEval(un));
}

void TComposeMulti::addAfterEval(TEval *un)
{
    _after->Add(new TSubEval(un));
}

void TComposeMulti::setBeforeEvals(Int_t size, TEval *un, ...)
{
    va_list ap;
    int i;
    TEval *eval;

    va_start(ap, un);
    _before->Add((TObject *) new TSubEval(un));
    for (i = 1; i < size; i++)
    {
        eval = va_arg(ap, TEval*);
        _before->Add((TObject *) new TSubEval(eval));
    }
    va_end(ap);
}

void TComposeMulti::setAfterEvals(Int_t size, TEval *un, ...)
{
    va_list ap;
    int i;
    TEval *eval;

    va_start(ap, un);
    _after->Add((TObject *) new TSubEval(un));
    for (i = 1; i < size; i++)
    {
        eval = va_arg(ap, TEval*);
        _after->Add((TObject *) new TSubEval(eval));
    }
    va_end(ap);
}

void TComposeMulti::addConstantValue(URANIE::DataServer::TAttribute *att,
        double val)
{
    _const.Add(att);
    _val.push_back(val);
}

void TComposeMulti::freeze()
{
    // TEval *fun;
    TSubEval *onefun;
    URANIE::DataServer::TAttribute *att;
    int dec;

    if (_freezed)
        return;
    TBaseEval::freeze();

    /* extraction des listes _input _output */
    dec = 0;
    TIter nextb(_before);
    while ((onefun = (TSubEval *) nextb()))
    {
        dec = onefun->freeze(dec, _const, _input, _output);
    }
    _mfirst = _output.GetEntries();
    dec = _multi->freeze(dec, _const, _input, _output);
    _msize = _output.GetEntries() - _mfirst;
    TIter nexta(_after);
    while ((onefun = (TSubEval *) nexta()))
    {
        dec = onefun->freeze(dec, _const, _input, _output);
    }

    /* identification des inputs par codes */
    /* liste complete */
    TObjArray allput;
    TIter next_cst(&_const);
    while ((att = (URANIE::DataServer::TAttribute *) next_cst()))
        allput.Add(att);
    TIter next_in(&_input);
    while ((att = (URANIE::DataServer::TAttribute *) next_in()))
        allput.Add(att);
    TIter next_out(&_output);
    while ((att = (URANIE::DataServer::TAttribute *) next_out()))
        allput.Add(att);

    /* recherche des inputs */
    _allsize = allput.GetEntries();
    nextb.Reset();
    while ((onefun = (TSubEval *) nextb()))
        onefun->match(allput);
    _multi->match(allput);
    nexta.Reset();
    while ((onefun = (TSubEval *) nexta()))
        onefun->match(allput);

}

void TComposeMulti::_rebuild(int rank, Bool_t chef)
{
    TSubEval *onefun;
    TList *tmpb = _before;
    _before = new TList();
    _before->SetOwner();
    TList *tmpa = _after;
    _after = new TList();
    _after->SetOwner();
 
    TIter nextb(tmpb);
    while ((onefun = (TSubEval*) nextb()))
    {
        _before->Add(onefun->rebuild(rank, chef));
    }
    _multi = _multi->rebuild(rank, chef);
    TIter nexta(tmpa);
    while ((onefun = (TSubEval*) nexta()))
    {
        _after->Add(onefun->rebuild(rank, chef));
    }
}

TEval *TComposeMulti::rebuild(int rank, Bool_t chef)
{
    TComposeMulti *ret = new TComposeMulti(*this);

    ret->_rebuild(rank, chef);

    return (TEval *) ret;
}

Bool_t TComposeMulti::unbuild()
{
    TSubEval *one;

    TIter nextb(_before);
    while ((one = (TSubEval*) nextb()))
    {
        one->unbuild();
    }
    _multi->unbuild();
    TIter nexta(_after);
    while ((one = (TSubEval*) nexta()))
    {
        one->unbuild();
    }
    return kTRUE;
}

int TComposeMulti::meval(Double_t *in, Double_t **out)
{
    Double_t* memo = new Double_t[_allsize];
	Double_t* tmp;
    int i, j, dec, ret, flag;
    int id, loop;
    TSubEval *fun;

    /* copie des constantes */
    TIter next_cst(&_const);
    i = 0;
    while (next_cst())
    {
        memo[i] = _val[i];
        i++;
    }
    /* copie des entrées */
    TIter next_in(&_input);
    j = 0;
    while (next_in())
    {
        memo[i] = in[j];
        i++;
        j++;
    }
    dec = i;

    /* composition des evals */
    /* before */
    TIter nextb(_before);
    while ((fun = (TSubEval *) nextb()))
    {
        flag = fun->eval(memo, memo + dec);
        if (flag != 1)
            return 0;
    }
    /* multi */
    tmp = NULL;
    loop = ret = _multi->meval(memo, &tmp);
    if (!*out)
        *out = new Double_t[ret * _output.GetEntries()];
    /* after */
    TIter nexta(_after);
    id = 0;
    for (i = 0; i < loop; i++, nexta.Reset())
    {
        /* recopie multi ouput */
        for (j = 0; j < _msize; j++)
            memo[j + dec + _mfirst] = tmp[j + i * _msize];
        /* apply after function */
        while ((fun = (TSubEval *) nexta()))
        {
            flag = fun->eval(memo, memo + dec);
            if (flag != 1)
                break;
        }
        /* recopie sorties */
        if (flag != 1)
        {
            ret--;
        }
        else
        {
            for (j = 0; j < _output.GetEntries(); j++, id++)
                (*out)[id] = memo[j + dec];
        }
    }
    delete[] tmp;
    tmp = NULL;
	delete[] memo;
	memo = NULL;
    return ret;

}

}
} //namespace
