//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TFlagScript.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-05-29.11.18.08

 \brief   gestion des fichiers entree de code utilisant les flags
 */

#include "TFlagScript.h"

#include <iostream>
#include <fstream>

//ROOT

ClassImp (URANIE::Relauncher::TFlagScript);

namespace URANIE
{
namespace Relauncher
{

/*** TFlagParser ***/
/* Constructeur et destructeur */
TFlagParser::TFlagParser(const char *name, URANIE::DataServer::TAttribute *att) :
        TNamed(name, name), _att(att)
{
    _line = _istart = _istop = _index = -1;
}

TFlagParser::~TFlagParser(void)
{
}
  
//Bool_t TFlagParser::IsSortable(void) {return kTRUE;}

int TFlagParser::Compare(const TObject *other) const
{
    const TFlagParser *tmp = dynamic_cast<const TFlagParser*>(other);
    int ret;

    if (_line < tmp->_line)
        ret = -1;
    else if (_line == tmp->_line)
    {
        if (_istart > tmp->_istart)
            ret = -1;
        else if (_istart == tmp->_istart)
            ret = 0;
        else
            ret = 1;
    }
    else
        ret = 1;
    return ret;
}

void TFlagParser::build(int id, std::istream &fi, TObjArray *occur)
{
    TRegexp reg(GetName());
    TString titi;
    Ssiz_t deb;

    _index = id;
    /* parse index */
    _line = 0;
    titi.ReadLine(fi, kFALSE);
    deb = 0;
    while (1)
    {
        _istart = reg.Index(titi, &_istop, deb);
        if (_istart != -1)
        {
            // find occurence
            occur->Add(new TFlagParser(*this));
            //std::cout << _istart << ":" << _istop << std::endl;
            // next occurence in line
            deb = _istart + _istop;
        }
        else
        {
            // next line
            if (titi.ReadLine(fi, kFALSE))
            {
                //std::cout << titi.Data() << std::endl;
                _line++;
                deb = 0;
            }
            else
                break;
        }
    }
}

TString &TFlagParser::replace(TString *line, Double_t val)
{
    char strnum[32];
    sprintf(strnum, "%.16g", val);
    return line->Replace(_istart, _istop, strnum);
}

TString &TFlagParser::replace(TString *line, TString val)
{
  return line->Replace(_istart, _istop, val);
}
  
/*** TFlagScript *******/
TFlagScript::TFlagScript(const char *name) :
        TShareScript(name)
{
}

TFlagScript::~TFlagScript(void)
{
    TFlagParser *one;

    TIter next(&_vars);
    while ((one = (TFlagParser*) next()))
    {
        delete one;
        one = NULL;
    }
    TIter nexto(&_occur);
    while ((one = (TFlagParser*) nexto()))
    {
        delete one;
        one = NULL;
    }
}
  
void TFlagScript::addInput(URANIE::DataServer::TAttribute *att,
        const char *hash)
{
    _vars.Add(new TFlagParser(hash, att));
}

void TFlagScript::setInputs(Int_t size, URANIE::DataServer::TAttribute *att,
        const char *hash, ...)
{
    va_list ap;
    int i;
    URANIE::DataServer::TAttribute *attr;
    const char *name;

    _vars.Add(new TFlagParser(hash, att));
    //std::cout << hash << std::endl;
    va_start(ap, hash);
    for (i = 1; i < size; i++)
    {
        attr = va_arg(ap, URANIE::DataServer::TAttribute *);
        name = va_arg(ap, const char *);
        //std::cout << name << std::endl;
        _vars.Add(new TFlagParser(name, attr));
    }
    va_end(ap);
}

void TFlagScript::build(TString &ref, TObjArray *in)
{
    TIter next(&_vars);
#ifdef WIN32
    TString reff = TString::Format("%s\\%s", ref.Data(), _path.Data());
#else	
    TString reff = TString::Format("%s/%s", ref.Data(), _path.Data());
#endif	
    TFlagParser *a;
    int id;

    while ((a = (TFlagParser *) next()))
    {
        // ajout si variable non reference
        if ((id = in->IndexOf(a->_att)) == -1)
        {
            id = in->GetEntries();
            in->Add(a->_att);
        }
        // recherche occurence dans fichier
        std::ifstream f(reff, std::ios::in);
        if (!f)
            std::cout << "No input file\n";
        a->build(id, f, &_occur);
        f.close();
    }

    // trie des substitutions
    _occur.Sort();
}

void TFlagScript::put(TString &ref, TString &work, Double_t *in)
{
    char* reff = new char[ref.Length() + _path.Length() + 3];
    char* copy = new char[work.Length() + _path.Length() + 3];
#ifdef WIN32
    sprintf(reff, "%s\\%s", ref.Data(), _path.Data());
    sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else		
    sprintf(reff, "%s/%s", ref.Data(), _path.Data());
    sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(reff, std::ios::in);
    std::ofstream fout(copy, std::ios::out);
    TIter next(&_occur);
    TString a;
    int iline;
    TFlagParser *match;

    iline = 0;
    match = (TFlagParser *) next();
    while (a.ReadLine(fin, kFALSE))
    {
        if (match && match->_line == iline)
        {
            while (match && match->_line == iline)
            {
                a = match->replace(&a, in[match->_index]);
                //std::cout << b << std::endl;
                match = (TFlagParser *) next();
            }
        }
        fout << a << std::endl;
        iline++;
    }
    fin.close();
    fout.close();
	
	delete[] reff;
	delete[] copy;
}


  
void TFlagScript::put(TString &ref, TString &work, vector<URANIE::DataServer::UEntry*> *in)
{
  char* reff = new char[ref.Length() + _path.Length() + 3];
  char* copy = new char[work.Length() + _path.Length() + 3];
#ifdef WIN32
  sprintf(reff, "%s\\%s", ref.Data(), _path.Data());
  sprintf(copy, "%s\\%s", work.Data(), _path.Data());
#else		
  sprintf(reff, "%s/%s", ref.Data(), _path.Data());
  sprintf(copy, "%s/%s", work.Data(), _path.Data());
#endif	
  ifstream fin(reff, std::ios::in);
  ofstream fout(copy, std::ios::out);
  TIter next(&_occur);
  TString a;
  int iline;
  TFlagParser *match;

  iline = 0;
  match = (TFlagParser *) next();
  while (a.ReadLine(fin, kFALSE))
    {
      if (match && match->_line == iline)
        {
	  while (match && match->_line == iline)
            {
                if(in->at(match->_index)->tag==URANIE::DataServer::UEntry::Vector)
                    a = match->replace(&a, _svecbeg+in->at(match->_index)->dumpStr( _svecdel,"","",_format )+_svecend );
                else if(in->at(match->_index)->tag==URANIE::DataServer::UEntry::String)
                    a = match->replace(&a, _sstrbeg+in->at(match->_index)->dumpStr("","","",_format )+_sstrend );
                else
                    a = match->replace(&a, in->at(match->_index)->dumpStr("","","",_format ) );
	      //std::cout << b << std::endl;
	      match = (TFlagParser *) next();
            }
        }
      fout << a << std::endl;
      iline++;
    }
  fin.close();
  fout.close();
	
  delete[] reff;
  delete[] copy;
}
  
}
} //namespace

