//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TLauncher2.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-09-07.15.13.41

  \brief   execution dun plan dexperience
*/

#include "TLauncher2.h"
#include "TAttribute.h"

#include "TLeaf.h"

#include "TRun.h"

ClassImp (URANIE::Relauncher::TLauncher2);

namespace URANIE
{
namespace Relauncher
{

TLauncher2::TLauncher2(URANIE::DataServer::TDataServer *tds, TRun *run) :
    TNamed("UranieLauncher", "design pattern realisation"),
    TInformMaster(tds, run), _tdserr(NULL)
{
}

TLauncher2::TLauncher2(URANIE::DataServer::TDataServer *tds, TEval *fun) :
    TNamed("UranieLauncher", "design pattern realisation"),
    TInformMaster(tds, fun), _tdserr(NULL)
{
}

TLauncher2::TLauncher2(URANIE::DataServer::TDataServer *tds, TStandardEval *fun
		       , const char *in, const char *out) :
    TNamed("UranieLauncher", "design pattern realisation"),
    TInformMaster(tds, fun, in, out)
{
}

TLauncher2::~TLauncher2(void)
{
}

/* public method */
void TLauncher2::setSaveError(URANIE::DataServer::TDataServer *tdserr)
{
    TList *tmplist;
    URANIE::DataServer::TAttribute *att;
    
    tmplist = _tds->getListOfAttributes("*");
    TIter next(tmplist);
    while ((att = (URANIE::DataServer::TAttribute *) next())) {
        tdserr->addAttribute(att);
    }
    tdserr->createTuple();
    _tdserr = tdserr;
    
    delete tmplist;
}

/* protocole method */

void TLauncher2::solverLoop()
{

    int ipat, ret;
    long id;
    // URANIE::DataServer::TAttribute *att;

    build();

    Int_t dpsize = (Int_t) _sampl->GetEntries();
    Int_t bufsize = (_run->_ncpus < dpsize) ? _run->_ncpus : dpsize;

    vector<URANIE::DataServer::UEntry*> samp;
    vector<vector<URANIE::DataServer::UEntry*> > total; total.resize(bufsize); //total memory of an event
    //prepare buffer for memory for inputs
    vector<vector<URANIE::DataServer::UEntry*> > in, out;
    in.resize(bufsize); out.resize(bufsize);
    TObject *var;

    allocSamplEntries(samp);
    allocInCodeEntries(bufsize, in, true);
    allocOutCodeEntries(bufsize, out);
    allocDataEntries(bufsize, total);

    //Temporary vector used also for simplicity (transmit subrange of information)
    vector<vector<URANIE::DataServer::UEntry*> *> temp_in; temp_in.resize(bufsize);
    int icalc=0;
    for (ipat = 0; ipat < bufsize; ipat++)
    {
	getCurrentPattern(ipat, &samp, &in[ipat]); 
	temp_in[ipat] = new vector<URANIE::DataServer::UEntry*>(in[ipat].begin() + 1, in[ipat].end()-_Naux);
	_run->put(ipat, temp_in[ipat], &out[ipat], icalc ); // Push back list of elements (tupl)
        icalc++;
    }

    //Running on
    for (; ipat < dpsize; ipat++)
    {
	ret = _run->get(id); 
	if (ret)
	    putCurrentPattern(&in[id], &out[id], &total[id]);
	else if (_tdserr)
	    putCurrentAsError((int) *in[id][0]->val.dou, &samp, _tdserr);

        //Clear output (for vectors mainly)
	for(vector<URANIE::DataServer::UEntry*>::iterator it= out[id].begin(); it!=out[id].end(); it++)
	    (*it)->clearEntry();
	delete temp_in[id];

	getCurrentPattern(ipat, &samp, &in[id]);
	temp_in[id] = new vector<URANIE::DataServer::UEntry*>(in[id].begin() + 1, in[id].end()-_Naux);
	_run->put(id, temp_in[id], &(out[id]), icalc);
        icalc++;
    }

    for (ipat = 0; ipat < bufsize; ipat++)
    {
	ret = _run->get(id);
	if (ret)
	    putCurrentPattern(&in[id], &out[id], &total[id]);
	else if (_tdserr)
	    putCurrentAsError((int) *in[id][0]->val.dou, &samp, _tdserr);

        //Clear output (for vectors mainly)
	for(vector<URANIE::DataServer::UEntry*>::iterator it= out[id].begin(); it!=out[id].end(); it++)
	    (*it)->clearEntry();
	delete temp_in[id];
    }

    //End of job, deleteing all pointers
    for (ipat = 0; ipat < bufsize; ipat++)
    {
	for(vector<URANIE::DataServer::UEntry*>::iterator it= in[ipat].begin(); it!=in[ipat].end(); it++)
	    delete (*it);
	in[ipat].clear();
	for(vector<URANIE::DataServer::UEntry*>::iterator it= out[ipat].begin(); it!=out[ipat].end(); it++)
	    delete (*it);
	out[ipat].clear();
	for(vector<URANIE::DataServer::UEntry*>::iterator it= total[ipat].begin(); it!=total[ipat].end(); it++)
	    delete (*it);
	total[ipat].clear();
    }    
    for(auto it : samp ) delete it;
    //for vector<URANIE::DataServer::UEntry*>::iterator it= _bufferin.begin(); it!=_bufferin.end(); it++)
    //	delete (*it);

    fillConstant();
}

}
} //namespace

