//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TMaster.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-04-01.09.16.51

  \brief   classe abstraite de supervision : definition utilitaires
*/

#include "TMaster.h"

#include "UExceptions.h"
#include "TSequentialRun.h"

//#define TRACE

ClassImp (URANIE::Relauncher::TMaster);

namespace URANIE
{
namespace Relauncher
{

// Constructeur Destructeur
TMaster::TMaster(URANIE::DataServer::TDataServer *tds, TRun *run) :
    _tds(tds), _run(run)
{
    _localrun = NULL;
    _fun = run->getEval();
    match_input();
}

TMaster::TMaster(URANIE::DataServer::TDataServer *tds, TEval *fun) :
    _tds(tds), _fun(fun)
{
    _localrun = _run = new TSequentialRun(fun);
    match_input();
    _run->startSlave();
}

TMaster::TMaster(URANIE::DataServer::TDataServer *tds, TStandardEval *fun,
		 const char *in, const char *out) :
    _tds(tds),
    _fun(fun)
{
    TObject *var=NULL;
    URANIE::DataServer::TAttribute *att=NULL;
    
    TString sin(in);
    TString sout(out);

    TObjArray *sin_list=sin.Tokenize(":");
    TIter nextin(sin_list);
    try{
        while ((var = nextin() )) {
        fun->addInput(tds->getAttribute(var->GetName()));
        }
    } catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        delete sin_list;
        throw ue;
    }
    delete sin_list;
 
    TObjArray *sout_list=sout.Tokenize(":");
    TIter nextout(sout_list);
    try{
        while ((var = nextout() )) {
            att = new URANIE::DataServer::TAttribute(var->GetName());
            fun->addOutput(att);
            att->delShare();
        }
    } catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        delete sout_list;
        throw ue;
    } 
    delete sout_list;

    _localrun = _run = new TSequentialRun(fun);
    match_input();
    _run->startSlave();

}

TMaster::~TMaster()
{
    _param.SetOwner();
    _sortie.SetOwner();
    if (_localrun)
    {
        delete _localrun;
        _localrun = NULL;
    }
}

// Constructeur helper
void TMaster::match_input()
{
    Int_t i;
    URANIE::DataServer::TAttribute *att, *satt;

    //_fun->build(); in TRun

    // correspondance input
    TObjArray *icode = _fun->getInputList();
    _NinCod = icode->GetEntries();
    TList *isolv = _tds->getListOfAttributes("*");
    Int_t nis = isolv->GetEntries();

    bool* mark = new bool[nis];
    for (i = 0; i < nis; i++)
	mark[i] = false;
    //std::cout << nin << " " << nall-nin << std::endl;
    // phase un : trie variable constante
    TIter nextc(icode);
    TIter nexts(isolv);
    _NinDat = 0;
    while ((att = (URANIE::DataServer::TAttribute *) nextc()))
    {
        i = 0;
        while ((satt = (URANIE::DataServer::TAttribute*) nexts()) && (att != satt)) i++;
        if (i == nis)
	        _param.Add(new TConstVar(att));
        else {
	        _param.Add(new TSolvVar(att, i));
	        _NinDat++;
	        mark[i] = true;
        }
        nexts.Reset();
    }
    // phase deux : ajout output du code
    _NoutCod = 0;
    TIter nexto(_fun->getOutputList());
    while ((att = (URANIE::DataServer::TAttribute*) nexto()))
    {
#ifdef TRACE
        std::cout << ((TNamed *) att)->GetName() << std::endl;
#endif
        i = 0;
        while ((satt = (URANIE::DataServer::TAttribute*) nexts()) && att != satt ) i++;
        if (i == nis) {
            _sortie.Add(new TCodeVar(att, _NoutCod, nis));
        }
        else {
            _sortie.Add(new TPreVar(att, _NoutCod, i));
            mark[i] = true;
        }
        _NoutCod++;
        nexts.Reset();
    }
    // phase trois : ajout variable auxi
    _Naux = i = 0;
    while ((att = (URANIE::DataServer::TAttribute *) nexts()))
    {
	if (!mark[i])
        {
	    _param.Add(new TAuxiVar(att, i));
	    _Naux++;
        }
	i++;
    }

    // phase quatre : solver dependant
#ifdef TRACE
    std::cout << _NinDat << "/" << _NinCod << " " << _Naux << " " << _NoutCod << std::endl;
#endif

    // finalyse
    delete isolv;
    isolv = NULL;
    delete[] mark;
    mark = NULL;
}

// Variable Modifier
void TMaster::addConstantValue(URANIE::DataServer::TAttribute *att, double val, bool KeepIt)
{
    TObject *var;
    TIter next(&_param);
    try
    {
        while ((var = next()))
        {
            if (((TMasterVar *) var)->match(att))
            {
                TConstVar *tmp = dynamic_cast<TConstVar *>(var);
                if( tmp ==0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TMaster input [%s] has been declared as constant but it seems not to be one.\n If you have added this attribute to the dataserver, please comment this line", att->GetName()));
     
                tmp->setValue(val);
                tmp->saveValue(KeepIt);
                break;
            }
        }
        if (!var)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TMaster input not found, %s", att->GetName()));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
       	ue.printMessage();
	throw ue;        
    }
    
}

void TMaster::addConstantValue(URANIE::DataServer::TAttribute *att, URANIE::DataServer::UEntry *constant, bool KeepIt)
{
    TObject *var;
    TIter next(&_param);
    try
    {
        while ((var = next()))
        {
            if (((TMasterVar *) var)->match(att))
            {
                TConstVar *tmp = dynamic_cast<TConstVar *>(var);
                if( tmp ==0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TMaster input [%s] has been declared as constant but it seems not to be one.\n If you have added this attribute to the dataserver, please comment this line", att->GetName()));
     
                tmp->setValue(constant);
                tmp->saveValue(KeepIt);
                break;
            }
        }
        if (!var)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TMaster input not found, %s", att->GetName()));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
       	ue.printMessage();
	throw ue;        
    }
}

void TMaster::addTemporary(URANIE::DataServer::TAttribute *att)
{
    TObject *var;
    TIter next(&_sortie);
    while ((var = next()))
    {
        if (((TMasterVar *) var)->match(att))
        {
            dynamic_cast<TCodeVar *>(var)->setTemp();
            break;
        }
    }
    try{
        if (!var)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TMaster output not found, %s", att->GetName()));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
       	ue.printMessage();
	throw ue;        
    }
}

// running helper
void TMaster::fillDS()
{
    TInputVar *vari;
    TOutputVar *varo;
    int i;

    // input (verif constant defined)
    TIter nextp(&_param);
    try{
        while ((vari = (TInputVar*) nextp())) vari->isValid();
    }
    catch(URANIE::Exceptions::UErrorExceptions &ue) {
        ue.printMessage();
        throw ue;
    }

    // output (filter tmp soon in tds)
    _NoutDat = i = 0;
    TIter next(&_sortie);
    while ((varo = (TOutputVar *) next())) {
        if (varo->isExported()) {
            _NoutDat++;
            if (! varo->isPre()) {
                varo->reindice(i);
                _tds->addAttribute(varo->getRef());
                i++;
            }
        }
    }
}

// memory
void TMaster::allocInCodeEntries(int siz, vector<vector<URANIE::DataServer::UEntry*>> &in, bool withid)
{
    int ipat, ivar;
    TIter nexti(&_param);
    TMasterVar *var;

    //Reserving memory blocks for Iterators
    if (withid) {
        for (ipat = 0; ipat < siz; ipat++) {
            in[ipat].push_back( new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string("Iterator")) );
        }
    }
    //Reserving memory blocks for inputs
    while ((var = (TMasterVar *) nexti())) {
        for (ipat = 0; ipat < siz; ipat++) {
            var->allocCodeEntry(in[ipat]);
        }
    }

}

void TMaster::allocOutCodeEntries(int siz, vector<vector<URANIE::DataServer::UEntry*>> &out)
{
    int ipat;
    TIter nexto(&_sortie);
    TOutputVar *var;

    while ((var = (TOutputVar *) nexto()))
    {
        if (! var->isPost()) {
            for (ipat = 0; ipat < siz; ipat++) {
                var->allocCodeEntry(out[ipat]);
            }
        }
    }
}

void TMaster::allocGradCodeEntries(int siz, vector<vector<URANIE::DataServer::UEntry*>> &out)
{
    int ipat;
    TIter nexto(&_sortie);
    TOutputVar *var;

    while ((var = (TOutputVar *) nexto()))
    {
        if (! var->isPost()) {
            for (ipat = 0; ipat < siz; ipat++) {
                var->allocGradEntry(out[ipat]);
            }
        }
    }
}

void TMaster::allocDataEntries(int bufsize, vector<vector<URANIE::DataServer::UEntry*>> &all)
{
    int ipat, ivar;
    TObject *obj;
    
    //Reserving cell memory
    for (ipat = 0; ipat < bufsize; ipat++) {
        all[ipat].resize(1+_NinDat+_Naux+_NoutDat);
    }
    
    //Reserving memory blocks for Iterators
    for (ipat = 0; ipat < bufsize; ipat++) {
        all[ipat][0] = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string("Iterator")) ;
    }
    
    //Reserving memory blocks for inputs
    TIter nexti(&_param);
    while ((obj = nexti())) {
        TMasterVar *var = (TMasterVar *) obj;
        if (var->isExported()) {
            for (ipat = 0; ipat < bufsize; ipat++) {
                var->allocDataEntry(1, all[ipat]);
            }
        }
    }

    //Reserving memory blocks for outputs
    TIter nexto(&_sortie);
    while ((obj = nexto())) {
        TMasterVar *var = (TMasterVar *) obj;
        if (var->isExported()) {
            for (ipat = 0; ipat < bufsize; ipat++) {
                var->allocDataEntry(1, all[ipat]);
            }
        }
    }

}

void TMaster::allocDataEntries(vector<URANIE::DataServer::UEntry*> &all)
{
    TMasterVar *var;
    
    //Reserving cell memory
    all.resize(1+_NinDat+_Naux+_NoutDat);
    
    //Reserving memory blocks for Iterators
    all[0] = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, string("Iterator")) ;
    
    //Reserving memory blocks for inputs
    TIter nexti(&_param);
    while ((var = (TMasterVar *) nexti())) {
        if (var->isExported()) var->allocDataEntry(1, all);
    }

    //Reserving memory blocks for outputs
    TIter nexto(&_sortie);
    while ((var = (TMasterVar *) nexto())) {
        if (var->isExported()) var->allocDataEntry(1, all);
    }

}

// data
// double to double
void TMaster::inputForCode(const double *src, double *dest)
{
    TObject *var;
    TIter next(&_param);
    int i = 0;

    while ((var = next()))
    {
        if (!((TMasterVar *) var)->put(i, src, dest))
            break;
        i++;
    }
}

// double to UEntry
void TMaster::inputForCode(const double *src, vector<URANIE::DataServer::UEntry*> *dest)
{
    TObject *var;
    TIter next(&_param);
    int i = 0;

    while ((var = next()))
    {
        if (!((TMasterVar *) var)->put(i, src, dest))
            break;
        i++;
    }
    //delete dest;
}

void TMaster::outputFromCode(vector<URANIE::DataServer::UEntry*> *src, double *dest)
{
    TOutputVar *var;
    TIter next(&_sortie);
    int i;

    i = 0;
    while ((var = (TOutputVar *) next()))
    {
        if ( var->isPost()) i++; 
        else if ( var->get(i, dest, src) ) i++;
    }
    //delete src;
}

// =====================================================
//            Entry part for vectors
// =====================================================
  
void TMaster::inputForCode(const vector<URANIE::DataServer::UEntry*> *src, vector<URANIE::DataServer::UEntry*> *dest)
{
    TObject *var;
    TIter next(&_param);
    int i = 0;

    while ((var = next()))
    {
        if (!((TMasterVar *) var)->put(i, src, dest))
            break;
        i++;
    }

    delete dest;
}

void TMaster::allFromCode(vector<URANIE::DataServer::UEntry*> *srci, vector<URANIE::DataServer::UEntry*> *srco, vector<URANIE::DataServer::UEntry*> *dest)
{
    TInputVar *vari;
    TOutputVar *varo;
    TIter nexti(&_param);
    TIter nexto(&_sortie);
    int i, j; // j use ??

    i = j = 0;
    while ((vari = (TInputVar *) nexti()))
    {
        if ( vari->get(j, dest, srci)) i++;
        j++;
    }

    while ((varo = (TOutputVar *) nexto()))
    {
        if( varo->isExported() ) {
            if ( varo->isPost() ) i++ ;
            else if ( varo->get(i, dest, srco)) i++;
        }
    }

    delete srci; delete srco; delete dest;
}

void TMaster::clean()
{
    _fun->clean();
}

void TMaster::fillConstant()
{
    TString varname="";
    TObject *var;
    TIter next(&_param);    
    while ((var = next()))
    {
        TConstVar *tmp = dynamic_cast<TConstVar *>(var);
        if( tmp !=0)
            if( tmp->getSaveValue())
                _tds->fillConstantAttribute( tmp->getRef()->GetName(), tmp->getValue());
    }
    
    
}


}
} //namespace
