//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TBaseModel.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   fri jul 13 10:11:28 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::XMLProblem::TBaseModel
 
 */

/*!
 @class URANIE::XMLProblem::TBaseModel
 
 @brief Description of the class TBaseModel.
 
 To be written by the developper.
 
 */
#ifndef TBASEMODEL_H
#define TBASEMODEL_H

#define DEFAULTTOLERANCE 0.0001

// ROOT
#include "TNamed.h"

// URANIE
#include "XMLProblem.h"
#include "UExceptions.h"

namespace URANIE
{
  namespace XMLProblem
  {
    class TBaseModel: public TNamed
    {
    public:
      enum ELanguage
      {
        kEMPTY, kC, kCXX, kPMML, kPYTHON, kFORTRAN
      };
      enum ENorm
      {
        kNO, kCR, kMinusOneOne, kZeroOne
      };

    private:
      TString _sinput;
      TString _shidden;
      TString _soutput;

      ELanguage _nlanguage;
      ENorm _nNormType;                   ///! The normalized for input and output

      TString _sfileNameExport;
      TString _sfunctionNameExport;

      Int_t _niteration;
      Int_t _ninitialization;

      Double_t _dlearn;
      Double_t _dtolerance;
      Int_t _nseed;
    public:
      TBaseModel();
      TBaseModel(TString sinput, TString soutput);
      TBaseModel(TString sname, TString sinput, TString shidden, TString soutput);
      //! Default destructor
      virtual ~TBaseModel();

      //---------------------------------------------
      /** @name OtherSetting and Getting
     **/
      //@{
      void setName(TString str)
      {
        SetName(str);
      }
      /*!
     * \brief set the \bf "Input" of the model
     * \param str (TString) the inputs of the model (attributes names separated by chararcter ":")
     */
      void setInput(TString str)
      {
        _sinput = str;
      }
      /*!
     * \brief get the \bf "Input" of the model
     * \return  the \bf "Input" of the model (attributes names separated by chararcter ":")
     */
      TString getInput()
      {
        return _sinput;
      }
      /*!
     * \brief setHidden
     * \param str (TString) the "hidden" of the model (attributes names separated by chararcter ":")
     */
      void setHidden(TString str)
      {
        _shidden = str;
      }
      /*!
     * \brief getHidden
     * \return the \bf "Hidden" of the model
     */
      TString getHidden()
      {
        return _shidden;
      }
      /*!
     * \brief  set the \bf "Output" of the model
     * \param str  (TString) the "outputs" of the model (attributes names separated by chararcter ":")
     */
      void setOutput(TString str)
      {
        _soutput = str;
      }
      /*!
     * \brief getOutput
     * \return  the \bf "Output" of the model (attributes names separated by chararcter ":")
     */
      TString getOutput()
      {
        return _soutput;
      }
      /*!
     * \brief Set the language of the exported code source files
     * \param val (ELanguage) The language of the exported code source files [kCXX | kPMML | kPYTHON | kFORTRAN]
     */
      void setLanguage(ELanguage val)
      {
        _nlanguage = val;
      }
      /*!
     * \brief get the language of the exported code source files
     * \return The language of the exported code source files [kCXX | kPMML | kPYTHON | kFORTRAN]
     */
      ELanguage getLanguage()
      {
        return _nlanguage;
      }
      /*!
     * \brief set the normalisation procedure between [kNO | kCR | kMinusOneOne| kZeroOne]. kCR is the default.
     * \param n (ENorm) the normalisation [kNO | kCR | kMinusOneOne| kZeroOne].
     */
      void setNormalization(ENorm n)
      {
        _nNormType = n;
      }
      //! Get the normalisation parameter
      /*!
     * \brief get the normalisation
     * \return (ENorm) the normalisation [kCR | kMinusOneOne| kZeroOne].
     */
      ENorm getNormalization()
      {
        return _nNormType;
      }
      /*!
     * \brief set the file name of the code source file
     * \param str (TString) the file name of the code source file
     */
      void setFileNameExport(TString str)
      {
        _sfileNameExport = str;
      }
      /*!
     * \brief get the file name of the code source file
     * \return the file name of the code source file
     */
      TString getFileNameExport()
      {
        return _sfileNameExport;
      }
      /*!
     * \brief set the name of the function in the code source file
     * \param str(TString) the name of the function saved in the code source file
     */
      void setFunctionNameExport(TString str)
      {
        _sfunctionNameExport = str;
      }
      /*!
     * \brief get the function name in the code source file
     * \return the function name saved in the code source file
     */
      TString getFunctionNameExport()
      {
        return _sfunctionNameExport;
      }
      /*!
     * \brief set the number of iteration (split of the TDs in learning and test data base). 1 is default.
     * \param n (Int_t) : the number of iteration (split of the TDs in learning and test data base).
     */
      void setNIteration(Int_t n)
      {
        try {
          if (n < 0) {
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TBaseModel::setNIteration Error\n The number of iteration must be positive [%d].", n));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
          ue.printMessage();
          throw ue;
        }
        _niteration = n;
      }
      /*!
     * \brief get the number of iteration (split of the TDs in learning and test data base)
     * \return the number of iteration (split of the TDs in learning and test data base)
     */
      Int_t getNIteration()
      {
        return _niteration;
      }
      /*!
     * \brief set the number of initialisation step (weights for Neural Network learning). 1 is default value.
     * \param n
     */
      void setNInitialization(Int_t n)
      {
        try {
          if (n < 0) {
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TBaseModel::setNInitialization Error\n The number of initialization must be positive [%d].", n));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
          ue.printMessage();
          throw ue;
        }
        _ninitialization = n;
      }
      /*!
     * \brief get the number of initialisation step (weights for Neural Network learning)
     * \return the number of initialisation step (weights for Neural Network learning)
     */
      Int_t getNInitialization()
      {
        return _ninitialization;
      }
      /*!
     * \brief set the proportion of patterns to split the learning dataset in Learning and test databases. It must be in [0.0, 1.0] and 0.8 is default value.
     * \param dval (Double_t) : the proportion of patterns
     */
      void setLearn(Double_t dval)
      {
        try {
          if ( ( dval < 0 ) || ( dval > 1.0 ) )  {
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TBaseModel::setLearn Error\n The proportion must be in [0.0, 1.0], and we set to [%f].", dval));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
          ue.printMessage();
          throw ue;
        }
        _dlearn = dval;
      }
      /*!
     * \brief get the proportion of patterns to split the learning dataset in Learning and test databases.
     * \return the proportion of patterns
     */
      Double_t getLearn()
      {
        return _dlearn;
      }
      /*!
     * \brief Set the tolerance parameter. It defined the stoppig rule for the optimization algorithm. It must be possitive and 1.e-4 is default value.
     * \param dval
     */
      void setTolerance(Double_t dval)
      {
        try {
          if ( dval < 0 )  {
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TBaseModel::setTolerance Error\n The tolerance must be possitive [%f].", dval));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
          ue.printMessage();
          throw ue;
        }
        _dtolerance = dval;
      }
      /*!
     * \brief Get the tolerance parameter.
     * \return
     */
      Double_t getTolerance()
      {
        return _dtolerance;
      }

      /*!
       * \brief set the Seed parameter. It must be positive and 0 is default value.
       * \param n
       */
      void setSeed(Int_t n)
      {
        try {
          if ( n < 0 )  {
              throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                         Form("TBaseModel::setSeed Error\n The seed must be possitive [%d].", n));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
          ue.printMessage();
          throw ue;
        }
        _nseed = n;
      }
      Int_t getSeed()
      {
        return _nseed;
      }
      //@}

      //---------------------------------------------
      /** @name Init the exported files
     **/
      //@{
      void initExport();
      //@}

      //---------------------------------------------
      /** @name The  Log
     **/
      //@{
      void printLog(Option_t *option = "");
      //@}

      ClassDef(URANIE::XMLProblem::TBaseModel, ID_XMLPROBLEM)
      //Classe de
    };

  } // Fin du namespace XMLProblem
} // Fin du namespace URANIE
#endif
