//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2023 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TXMLProblem.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu mai 31 09:31:50 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::XMLProblem::TXMLProblem
 
 */

/*!
 @class URANIE::XMLProblem::TXMLProblem
 
 @brief Description of the class TXMLProblem.
 
 To be written by the developper.
 
 */
#ifndef TXMLPROBLEM_H
#define TXMLPROBLEM_H
// ROOT
#include <TXMLNode.h>

#include "XMLProblem.h"
#include "TDataServer.h"
#include "TBaseModel.h"

//STL
#include <vector>

//typedef struct {
//    int value;
//    std::string msg;
//} UranieError;

namespace URANIE
{
namespace XMLProblem
{

class TXMLProblem: public TNamed
{
private:
    Bool_t _blog;          ///< Log Printing
    Int_t _ndebuglevel;   ///< Level of debug
    Int_t _ntask;         ///< The iterator of the task to do
    URANIE::DataServer::TDataServer* _tds;           ///< The Data Server
    URANIE::DataServer::TDataServer* _tdsdatafield;         ///< The Data Server
    vector<TBaseModel *> _vectorOfModels; ///< The vector of Models
    TList * _listOfToDo; ///> The list of actions to do

protected:
    //---------------------------------------------
    /** @name Parse the data nodes
     **/
    //@{
    //! Parses the node "description"
    virtual void parseNodeHeader(TXMLNode *);
    //! Parses the node "DataDictionary"
    virtual void parseNodeDataDictionary(TXMLNode *);
    //! Parses the node "DataDictionary/DataField"
    virtual void parseNodeDataField(TXMLNode *);
    //! Parses the node "models/model"
    virtual void parseNodePatterns(TXMLNode *, TBaseModel *);
    //! Parses the node "models/patterns/select"
    virtual TString parseNodePatternSelected(TXMLNode *);
    //! Parses the node "modeler/algorithm"
    virtual void parseNodeModelerAlgorithm(TXMLNode *, TBaseModel *);
    //@}

    //---------------------------------------------
    /** @name  The Launcher node
     **/
    //@{
    //! Parses and runs the node "Launcher"
    void parseAndRunNodeLauncher(TXMLNode *);
    //@}

    //---------------------------------------------
    /** @name  The Sampler node
     **/
    //@{
    //! Parses and Runs the node "sampler"
    void parseAndRunNodeSampler(TXMLNode *);
    //@}

    //---------------------------------------------
    /** @name  The Modeler node
     **/
    //@{
    //! Parse and run the node "Modeler"
    void parseAndRunNodeModeler(TXMLNode *);
#ifdef WITH_OPT
    //! Parse the node "Modeler/NeuralNetwork"
    virtual void parseAndRunNodeModelerANN(TXMLNode *, TBaseModel *);
#endif
    //! Parse the node "Modeler/LinearRegression"
    virtual void parseAndRunNodeModelerLinearRegression(TXMLNode *,
            TBaseModel *);
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TXMLProblem();
    //! constructor with the XML File
    TXMLProblem(TString filename, Bool_t blog = kFALSE);
    //! Default destructor
    virtual ~TXMLProblem();
    //@}

    //---------------------------------------------
    /** @name Setter and getter methods
     **/
    //@{
    //! Gets the debug level
    Int_t getLevelDebug()
    {
        return _ndebuglevel;
    }
    //! Sets the debug level
    void setLevelDebug(Int_t val)
    {
        _ndebuglevel = val;
    }
    // Gets the dataserver
    URANIE::DataServer::TDataServer* getDataServer()
    {
        return _tds;
    }
    //@}

    //---------------------------------------------
    /** @name Reads XML data
     **/
    //@{
    //! Reads XML data
    Int_t parseFile(TString filename);
    //@}

    //---------------------------------------------
    /** @name Executes the actions
     **/
    //@{
    //! Initial phase of the code
    virtual void init(Option_t *option = "");
    //! Runs phase of the code
    virtual void run(Option_t *option = "");
    //! Terminates phase of the code
    virtual void terminate(Option_t *option = "");
    //! Cleans the working directory of the code
    virtual void clean(Option_t *option = "");
    /*!
     * \brief add Informations the the Task List to do
     * \param sname (TString) : the name of the task
     * \param stitle (TStrinh) : the task
     */
    void addInfoTask(TString sname, TString stitle);
    //@}

    //---------------------------------------------
    /** @name Log Printing
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::XMLProblem::TXMLProblem, ID_XMLPROBLEM)
    //Classe de
};

} // Fin du namespace XMLProblem
} // Fin du namespace URANIE
#endif
